Package: lava
Type: Package
Title: Latent Variable Models
Version: 1.9.3
Authors@R: c(person("Klaus K.", "Holst", email="klaus@holst.it", role=c("aut", "cre")),
       person("Benedikt", "Sommer", role = "ctb"),
		   person("Brice", "Ozenne", role = "ctb"),
		   person("Thomas", "Gerds", role = "ctb"))
Author: Klaus K. Holst [aut, cre],
  Benedikt Sommer [ctb],
  Brice Ozenne [ctb],
  Thomas Gerds [ctb]
Maintainer: Klaus K. Holst <klaus@holst.it>
Description: A general implementation of Structural Equation Models
	with latent variables (MLE, 2SLS, and composite likelihood
	estimators) with both continuous, censored, and ordinal
	outcomes (Holst and Budtz-Joergensen (2013) <doi:10.1007/s00180-012-0344-y>).
	Mixture latent variable models and non-linear latent variable models
	(Holst and Budtz-Joergensen (2020) <doi:10.1093/biostatistics/kxy082>).
	The package also provides methods for graph exploration (d-separation,
	back-door criterion), simulation of general non-linear latent variable
	models, and estimation of influence functions for a broad range of
	statistical models.
URL: https://kkholst.github.io/lava/, https://github.com/kkholst/lava
BugReports: https://github.com/kkholst/lava/issues
License: Apache License (== 2.0)
LazyLoad: yes
Depends: R (>= 3.5.0)
Imports: cli, future.apply, graphics, grDevices, methods, numDeriv,
        progressr, stats, survival, SQUAREM, utils
Suggests: KernSmooth, Rgraphviz, data.table, ellipse, fields, future,
        geepack, graph, knitr, rmarkdown, igraph (>= 0.6), lavaSearch2,
        lme4 (>= 1.1.35.1), MASS, Matrix (>= 1.6.3), mets (>= 1.1),
        nlme, optimx, polycor, quantreg, quarto, rgl, svglite, targeted
        (>= 0.4), testthat (>= 0.11), vdiffr, visNetwork
VignetteBuilder: quarto
SystemRequirements: Quarto command line tools
        (https://github.com/quarto-dev/quarto-cli).
ByteCompile: yes
Encoding: UTF-8
Config/roxygen2/version: 8.1.0
NeedsCompilation: no
Packaged: 2026-08-19 20:15:19 UTC; kkzh
Repository: CRAN
Date/Publication: 2026-08-20 09:20:18 UTC
Built: R 4.5.2; ; 2026-08-20 13:24:24 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method("%*%",estimate)
S3method("%++%","function")
S3method("%++%",character)
S3method("%++%",default)
S3method("%++%",estimate)
S3method("%++%",lvm)
S3method("%++%",matrix)
S3method("*",estimate)
S3method("+",estimate)
S3method("+",lvm)
S3method("-",estimate)
S3method("/",estimate)
S3method("==",estimate)
S3method("Graph<-",lvm)
S3method("Graph<-",lvmfit)
S3method("Model<-",lvm)
S3method("Model<-",lvmfit)
S3method("Model<-",multigroup)
S3method("Model<-",multigroupfit)
S3method("[",estimate)
S3method("[",sim)
S3method("^",estimate)
S3method("addvar<-",lvm)
S3method("cancel<-",lvm)
S3method("constrain<-",default)
S3method("constrain<-",multigroup)
S3method("constrain<-",multigroupfit)
S3method("covariance<-",lvm)
S3method("covfix<-",lvm)
S3method("distribution<-",lvm)
S3method("edgelabels<-",graphNEL)
S3method("edgelabels<-",lvm)
S3method("edgelabels<-",lvmfit)
S3method("exogenous<-",lvm)
S3method("functional<-",lvm)
S3method("heavytail<-",lvm)
S3method("index<-",lvm)
S3method("index<-",lvmfit)
S3method("intercept<-",lvm)
S3method("intervention<-",lvm)
S3method("intfix<-",lvm)
S3method("kill<-",lvm)
S3method("labels<-",default)
S3method("latent<-",lvm)
S3method("nodecolor<-",default)
S3method("nodecolor<-",lvm)
S3method("nonlinear<-",lvm)
S3method("ordinal<-",lvm)
S3method("parameter<-",lvm)
S3method("parameter<-",lvmfit)
S3method("parfix<-",lvm)
S3method("randomslope<-",lvm)
S3method("regfix<-",lvm)
S3method("regression<-",lvm)
S3method("rmvar<-",lvm)
S3method("transform<-",lvm)
S3method("variance<-",lvm)
S3method(Graph,lvm)
S3method(Graph,lvmfit)
S3method(IC,data.frame)
S3method(IC,default)
S3method(IC,estimate)
S3method(IC,estimate.complik)
S3method(IC,estimate.list)
S3method(IC,glm)
S3method(IC,lvm.mixture)
S3method(IC,matrix)
S3method(IC,mlm)
S3method(IC,multigroupfit)
S3method(IC,multinomial)
S3method(IC,numeric)
S3method(IC,twostageCV)
S3method(Model,default)
S3method(Model,lvm)
S3method(Model,lvmfit)
S3method(Model,multigroup)
S3method(Model,multigroupfit)
S3method(Model,twostageCV)
S3method(Weights,default)
S3method(acos,estimate)
S3method(acosh,estimate)
S3method(addattr,graphNEL)
S3method(addattr,lvm)
S3method(addattr,lvmfit)
S3method(addvar,lvm)
S3method(adjMat,lvm)
S3method(adjMat,lvmfit)
S3method(ancestors,lvm)
S3method(ancestors,lvmfit)
S3method(as.matrix,sim)
S3method(as.numeric,list)
S3method(as.vector,sim)
S3method(asin,estimate)
S3method(asinh,estimate)
S3method(atan,estimate)
S3method(atanh,estimate)
S3method(baptize,lvm)
S3method(bootstrap,lvm)
S3method(bootstrap,lvmfit)
S3method(c,estimate)
S3method(cancel,lvm)
S3method(cbind,sim)
S3method(children,lvm)
S3method(children,lvmfit)
S3method(click,default)
S3method(coef,effects)
S3method(coef,estimate)
S3method(coef,estimate.list)
S3method(coef,lvm)
S3method(coef,lvm.mixture)
S3method(coef,lvmfit)
S3method(coef,multigroup)
S3method(coef,multigroupfit)
S3method(coef,multinomial)
S3method(coef,ordreg)
S3method(coef,pcor)
S3method(coef,summary.estimate)
S3method(coef,summary.lvmfit)
S3method(coef,twostageCV)
S3method(coef,zibreg)
S3method(compare,default)
S3method(confint,effects)
S3method(confint,lvmfit)
S3method(confint,multigroupfit)
S3method(constrain,default)
S3method(correlation,data.frame)
S3method(correlation,lvmfit)
S3method(correlation,matrix)
S3method(cos,estimate)
S3method(cosh,estimate)
S3method(covariance,formula)
S3method(covariance,lvm)
S3method(covfix,lvm)
S3method(density,sim)
S3method(deriv,"function")
S3method(deriv,lvm)
S3method(descendants,lvm)
S3method(descendants,lvmfit)
S3method(distribution,lvm)
S3method(dsep,lvm)
S3method(edgeList,lvm)
S3method(edgeList,lvmfit)
S3method(edgelabels,graphNEL)
S3method(edgelabels,lvm)
S3method(edgelabels,lvmfit)
S3method(effects,lvmfit)
S3method(endogenous,list)
S3method(endogenous,lm)
S3method(endogenous,lvm)
S3method(endogenous,lvmfit)
S3method(endogenous,multigroup)
S3method(estimate,MAR)
S3method(estimate,array)
S3method(estimate,data.frame)
S3method(estimate,default)
S3method(estimate,formula)
S3method(estimate,glm)
S3method(estimate,list)
S3method(estimate,lvm)
S3method(estimate,mlm)
S3method(estimate,multigroup)
S3method(estimate,twostage.lvm)
S3method(exogenous,list)
S3method(exogenous,lm)
S3method(exogenous,lvm)
S3method(exogenous,lvmfit)
S3method(exogenous,multigroup)
S3method(exp,estimate)
S3method(expm1,estimate)
S3method(family,zibreg)
S3method(finalize,lvm)
S3method(formula,lvm)
S3method(formula,lvmfit)
S3method(functional,lvm)
S3method(gof,lvmfit)
S3method(heavytail,lvm)
S3method(iid,default)
S3method(index,lvm)
S3method(index,lvmfit)
S3method(influence,estimate)
S3method(information,data.frame)
S3method(information,glm)
S3method(information,lvm)
S3method(information,lvm.missing)
S3method(information,lvm.mixture)
S3method(information,lvmfit)
S3method(information,multigroup)
S3method(information,multigroupfit)
S3method(information,multinomial)
S3method(information,table)
S3method(information,zibreg)
S3method(intercept,lvm)
S3method(intervention,lvm)
S3method(kappa,data.frame)
S3method(kappa,multinomial)
S3method(kappa,table)
S3method(kill,lvm)
S3method(labels,estimate)
S3method(labels,graphNEL)
S3method(labels,lvm)
S3method(labels,lvmfit)
S3method(latent,list)
S3method(latent,lvm)
S3method(latent,lvmfit)
S3method(latent,multigroup)
S3method(log,estimate)
S3method(log1p,estimate)
S3method(logLik,estimate.complik)
S3method(logLik,lvm)
S3method(logLik,lvm.missing)
S3method(logLik,lvm.mixture)
S3method(logLik,lvmfit)
S3method(logLik,multigroup)
S3method(logLik,multigroupfit)
S3method(logLik,ordreg)
S3method(logLik,pcor)
S3method(logLik,zibreg)
S3method(manifest,list)
S3method(manifest,lvm)
S3method(manifest,lvm.mixture)
S3method(manifest,lvmfit)
S3method(manifest,multigroup)
S3method(matrices,lvm)
S3method(matrices,multigroup)
S3method(merge,estimate)
S3method(merge,estimate.list)
S3method(merge,glm)
S3method(merge,list)
S3method(merge,lm)
S3method(merge,lvm)
S3method(merge,lvmfit)
S3method(merge,multinomial)
S3method(model.frame,estimate)
S3method(model.frame,lvm.mixture)
S3method(model.frame,lvmfit)
S3method(model.frame,multigroupfit)
S3method(model.frame,multinomial)
S3method(modelPar,lvm)
S3method(modelPar,lvmfit)
S3method(modelPar,multigroup)
S3method(modelPar,multigroupfit)
S3method(modelVar,lvm)
S3method(modelVar,lvmfit)
S3method(moments,lvm)
S3method(moments,lvm.missing)
S3method(moments,lvmfit)
S3method(nonlinear,lvm)
S3method(nonlinear,lvmfit)
S3method(nonlinear,twostage.lvm)
S3method(ordinal,lvm)
S3method(parameter,default)
S3method(parameter,estimate)
S3method(parents,lvm)
S3method(parents,lvmfit)
S3method(parfix,lvm)
S3method(parpos,default)
S3method(parpos,lvm)
S3method(parpos,lvmfit)
S3method(parpos,multigroup)
S3method(parpos,multigroupfit)
S3method(pars,default)
S3method(pars,glm)
S3method(pars,lvm)
S3method(pars,lvm.missing)
S3method(pars,mlm)
S3method(pars,survreg)
S3method(path,graphNEL)
S3method(path,lvm)
S3method(path,lvmfit)
S3method(plot,estimate)
S3method(plot,lvm)
S3method(plot,lvm.mixture)
S3method(plot,lvmfit)
S3method(plot,multigroup)
S3method(plot,multigroupfit)
S3method(plot,mvn.mixture)
S3method(plot,sim)
S3method(plot,twostage.lvm)
S3method(predict,lvm)
S3method(predict,lvm.missing)
S3method(predict,lvm.mixture)
S3method(predict,lvmfit)
S3method(predict,multinomial)
S3method(predict,ordreg)
S3method(predict,twostage.lvmfit)
S3method(predict,twostageCV)
S3method(predict,zibreg)
S3method(print,Combine)
S3method(print,bootstrap.lvm)
S3method(print,diagtest)
S3method(print,effects)
S3method(print,equivalence)
S3method(print,estimate)
S3method(print,fix)
S3method(print,gkgamma)
S3method(print,gof.lvmfit)
S3method(print,lvm)
S3method(print,lvm.mixture)
S3method(print,lvm.predict)
S3method(print,lvmfit)
S3method(print,lvmfit.randomslope)
S3method(print,modelsearch)
S3method(print,multigroup)
S3method(print,multigroupfit)
S3method(print,multinomial)
S3method(print,mvn.mixture)
S3method(print,offdiag)
S3method(print,ordinal.lvm)
S3method(print,ordreg)
S3method(print,pcor)
S3method(print,sim)
S3method(print,summary.estimate)
S3method(print,summary.lvm.mixture)
S3method(print,summary.lvmfit)
S3method(print,summary.multigroupfit)
S3method(print,summary.ordreg)
S3method(print,summary.sim)
S3method(print,summary.twostageCV)
S3method(print,summary.zibreg)
S3method(print,test_adj)
S3method(print,transform.lvm)
S3method(print,twostage.lvm)
S3method(print,twostageCV)
S3method(print,zibreg)
S3method(prod,estimate)
S3method(profile,lvmfit)
S3method(randomslope,lvm)
S3method(randomslope,lvmfit)
S3method(rbind,Surv)
S3method(rbind,sim)
S3method(regfix,lvm)
S3method(regression,formula)
S3method(regression,lvm)
S3method(residuals,lvm)
S3method(residuals,lvmfit)
S3method(residuals,multigroupfit)
S3method(residuals,zibreg)
S3method(rmvar,lvm)
S3method(roots,lvm)
S3method(roots,lvmfit)
S3method(score,estimate.complik)
S3method(score,glm)
S3method(score,lm)
S3method(score,lvm)
S3method(score,lvm.missing)
S3method(score,lvm.mixture)
S3method(score,lvmfit)
S3method(score,multigroup)
S3method(score,multigroupfit)
S3method(score,ordreg)
S3method(score,pcor)
S3method(score,survreg)
S3method(score,zibreg)
S3method(sim,default)
S3method(sim,lvm)
S3method(sim,lvmfit)
S3method(sim,mvn.mixture)
S3method(simulate,lvm)
S3method(simulate,lvmfit)
S3method(sin,estimate)
S3method(sinh,estimate)
S3method(sinks,lvm)
S3method(sinks,lvmfit)
S3method(sqrt,estimate)
S3method(stack,estimate)
S3method(stack,glm)
S3method(stack,lvmfit)
S3method(subset,estimate)
S3method(subset,lvm)
S3method(sum,estimate)
S3method(summary,diagtest)
S3method(summary,effects)
S3method(summary,estimate)
S3method(summary,lvm)
S3method(summary,lvm.mixture)
S3method(summary,lvmfit)
S3method(summary,multigroup)
S3method(summary,multigroupfit)
S3method(summary,ordreg)
S3method(summary,sim)
S3method(summary,test_adj)
S3method(summary,twostageCV)
S3method(summary,zibreg)
S3method(tan,estimate)
S3method(tanh,estimate)
S3method(totaleffects,lvmfit)
S3method(tr,matrix)
S3method(transform,estimate)
S3method(transform,lvm)
S3method(twostage,estimate)
S3method(twostage,lvm)
S3method(twostage,lvm.mixture)
S3method(twostage,lvmfit)
S3method(twostage,twostage.lvm)
S3method(variance,formula)
S3method(variance,lvm)
S3method(vars,graph)
S3method(vars,list)
S3method(vars,lm)
S3method(vars,lvm)
S3method(vars,lvmfit)
S3method(vcov,effects)
S3method(vcov,estimate)
S3method(vcov,estimate.list)
S3method(vcov,lvm.mixture)
S3method(vcov,lvmfit)
S3method(vcov,multigroupfit)
S3method(vcov,multinomial)
S3method(vcov,ordreg)
S3method(vcov,pcor)
S3method(vcov,twostageCV)
S3method(vcov,zibreg)
S3method(with,estimate)
export("%++%")
export("%in.closed%")
export("%in.open%")
export("%ni%")
export("Graph<-")
export("Missing<-")
export("Model<-")
export("addvar<-")
export("cancel<-")
export("categorical<-")
export("constrain<-")
export("covariance<-")
export("distribution<-")
export("edgelabels<-")
export("eventTime<-")
export("exogenous<-")
export("functional<-")
export("heavytail<-")
export("index<-")
export("intercept<-")
export("intervention<-")
export("kill<-")
export("labels<-")
export("latent<-")
export("nodecolor<-")
export("nonlinear<-")
export("offdiag<-")
export("ordinal<-")
export("parameter<-")
export("randomslope<-")
export("regression<-")
export("revdiag<-")
export("rmvar<-")
export("timedep<-")
export("transform<-")
export("variance<-")
export(Binary.lvm)
export(By)
export(CoefMat)
export(CoefMat.multigroupfit)
export(Col)
export(Combine)
export(Diff)
export(Expand)
export(GM2.lvm)
export(GM3.lvm)
export(Gamma.lvm)
export(Graph)
export(Grep)
export(IC)
export(IV)
export(Inverse)
export(Missing)
export(Model)
export(NA2x)
export(NR)
export(OR)
export(PD)
export(Print)
export(Range.lvm)
export(Ratio)
export(Sequence.lvm)
export(Specials)
export(Weights)
export(aalenExponential.lvm)
export(addhook)
export(addvar)
export(adjMat)
export(alpha_zmax)
export(ancestors)
export(as.sim)
export(backdoor)
export(baptize)
export(beta.lvm)
export(binary.lvm)
export(binomial.lvm)
export(binomial.rd)
export(binomial.rr)
export(blockdiag)
export(bootstrap)
export(cancel)
export(categorical)
export(children)
export(chisq.lvm)
export(click)
export(closed_testing)
export(colorbar)
export(colsel)
export(commutation)
export(compare)
export(complik)
export(confband)
export(confpred)
export(constant.lvm)
export(constrain)
export(constraints)
export(contr)
export(correlation)
export(covariance)
export(coxExponential.lvm)
export(coxGompertz.lvm)
export(coxWeibull.lvm)
export(csplit)
export(curly)
export(decomp.specials)
export(density.sim)
export(descendants)
export(describecoef)
export(devcoords)
export(diagtest)
export(distribution)
export(dmvn0)
export(dsep)
export(edgeList)
export(edgelabels)
export(endogenous)
export(equivalence)
export(estimate)
export(estimate.default)
export(eventTime)
export(exogenous)
export(expit)
export(finalize)
export(fixsome)
export(foldr)
export(forestplot)
export(fplot)
export(functional)
export(gaussian.lvm)
export(gaussian_logLik.lvm)
export(getSAS)
export(gethook)
export(getoutcome)
export(gkgamma)
export(gof)
export(graph2lvm)
export(heavytail)
export(id.lvm)
export(idplot)
export(igraph.lvm)
export(iid)
export(images)
export(index)
export(information)
export(intercept)
export(intervention)
export(kill)
export(ksmooth2)
export(latent)
export(lava)
export(lava.options)
export(loggamma.lvm)
export(logit)
export(logit.lvm)
export(lognormal.lvm)
export(lvm)
export(makemissing)
export(manifest)
export(mat.lvm)
export(measurement)
export(measurement.error)
export(mixture)
export(modelPar)
export(modelVar)
export(modelsearch)
export(moments)
export(multigroup)
export(multinomial)
export(multinomial.lvm)
export(mvn.lvm)
export(mvnmix)
export(na.pass0)
export(none.lvm)
export(nonlinear)
export(normal.lvm)
export(odds)
export(offdiag)
export(offdiags)
export(ones.lvm)
export(ordinal)
export(ordreg)
export(pairwise.diff)
export(parameter)
export(parents)
export(pareto.lvm)
export(parlabels)
export(parpos)
export(pars)
export(parsedesign)
export(partialcor)
export(path)
export(pcor)
export(pdfconvert)
export(plot.sim)
export(plotConf)
export(plot_region)
export(poisson.lvm)
export(predictlvm)
export(probit.lvm)
export(procformula)
export(randomslope)
export(regfix)
export(regression)
export(revdiag)
export(riskcomp)
export(rmvar)
export(rmvn0)
export(roots)
export(rot2D)
export(rot3D)
export(rotate2)
export(rsq)
export(scheffe)
export(score)
export(sim)
export(sinks)
export(spaghetti)
export(starter.multigroup)
export(startvalues)
export(startvalues0)
export(startvalues1)
export(startvalues2)
export(student.lvm)
export(summary.sim)
export(surface)
export(test_wald)
export(threshold.lvm)
export(tigol)
export(timedep)
export(toformula)
export(totaleffects)
export(tr)
export(trim)
export(twostage)
export(twostageCV)
export(uniform.lvm)
export(var_ic)
export(variance)
export(variances)
export(vars)
export(vec)
export(wait)
export(weibull.lvm)
export(wkm)
export(wrapvec)
export(x2NA)
export(zibreg)
importFrom(future.apply,future_lapply)
importFrom(future.apply,future_mapply)
importFrom(grDevices,col2rgb)
importFrom(grDevices,colorRampPalette)
importFrom(grDevices,colors)
importFrom(grDevices,gray.colors)
importFrom(grDevices,heat.colors)
importFrom(grDevices,palette)
importFrom(grDevices,rainbow)
importFrom(grDevices,rgb)
importFrom(grDevices,topo.colors)
importFrom(grDevices,xy.coords)
importFrom(graphics,abline)
importFrom(graphics,axis)
importFrom(graphics,box)
importFrom(graphics,contour)
importFrom(graphics,contour.default)
importFrom(graphics,curve)
importFrom(graphics,identify)
importFrom(graphics,image)
importFrom(graphics,layout)
importFrom(graphics,lines)
importFrom(graphics,locator)
importFrom(graphics,matplot)
importFrom(graphics,mtext)
importFrom(graphics,par)
importFrom(graphics,plot)
importFrom(graphics,plot.new)
importFrom(graphics,plot.window)
importFrom(graphics,points)
importFrom(graphics,polygon)
importFrom(graphics,rect)
importFrom(graphics,rug)
importFrom(graphics,segments)
importFrom(graphics,text)
importFrom(graphics,title)
importFrom(methods,as)
importFrom(methods,new)
importFrom(stats,AIC)
importFrom(stats,addmargins)
importFrom(stats,approxfun)
importFrom(stats,as.formula)
importFrom(stats,coef)
importFrom(stats,confint)
importFrom(stats,confint.default)
importFrom(stats,cor)
importFrom(stats,cov)
importFrom(stats,cov2cor)
importFrom(stats,density)
importFrom(stats,density.default)
importFrom(stats,deriv)
importFrom(stats,dnorm)
importFrom(stats,effects)
importFrom(stats,family)
importFrom(stats,fft)
importFrom(stats,formula)
importFrom(stats,get_all_vars)
importFrom(stats,glm)
importFrom(stats,influence)
importFrom(stats,lm)
importFrom(stats,logLik)
importFrom(stats,model.frame)
importFrom(stats,model.matrix)
importFrom(stats,model.weights)
importFrom(stats,na.omit)
importFrom(stats,na.pass)
importFrom(stats,nlminb)
importFrom(stats,p.adjust)
importFrom(stats,pchisq)
importFrom(stats,pnorm)
importFrom(stats,predict)
importFrom(stats,printCoefmat)
importFrom(stats,pt)
importFrom(stats,qchisq)
importFrom(stats,qf)
importFrom(stats,qnorm)
importFrom(stats,qt)
importFrom(stats,quantile)
importFrom(stats,rbinom)
importFrom(stats,rchisq)
importFrom(stats,residuals)
importFrom(stats,rgamma)
importFrom(stats,rlnorm)
importFrom(stats,rmultinom)
importFrom(stats,rnorm)
importFrom(stats,rpois)
importFrom(stats,rt)
importFrom(stats,runif)
importFrom(stats,sd)
importFrom(stats,simulate)
importFrom(stats,terms)
importFrom(stats,uniroot)
importFrom(stats,update)
importFrom(stats,update.formula)
importFrom(stats,var)
importFrom(stats,vcov)
importFrom(stats,weights)
importFrom(survival,is.Surv)
importFrom(utils,combn)
importFrom(utils,data)
importFrom(utils,getFromNamespace)
importFrom(utils,getTxtProgressBar)
importFrom(utils,glob2rx)
importFrom(utils,head)
importFrom(utils,methods)
importFrom(utils,modifyList)
importFrom(utils,read.csv)
importFrom(utils,setTxtProgressBar)
importFrom(utils,stack)
importFrom(utils,tail)
importFrom(utils,txtProgressBar)
importFrom(utils,write.table)

LAVA
====

LAVA is a continuous integration system for deploying operating systems onto
physical and virtual hardware for running tests.

This repository contains the dispatcher and server source code.

Installation instructions
=========================

For latest information, see the lava-server documentation:
* [first installation](https://lava.readthedocs.io/en/latest/admin/basic-tutorials/instance/install.html)

Support
=======

Read the [contact](https://lava.readthedocs.io/en/latest/introduction/contact.html) page.

Reporting Bugs
==============

Bugs should first be discussed by subscribing to the
[lava-users](https://lists.lavasoftware.org/mailman3/lists/lava-users.lists.lavasoftware.org/)
mailing list.

Bug reports and new LAVA features are tracked using
[GitLab](https://gitlab.com/lava/lava/issues).

LAVA
====

LAVA is an automated validation architecture primarily aimed at testing
deployments of systems based around the Linux kernel on ARM devices,
specifically ARMv7 and later.

This repository contains the dispatcher and server source code.

Installation instructions
=========================

For latest information, see the lava-server documentation:
* [first installation](https://docs.lavasoftware.org/lava/first-installation.html)
* [installing on Debian](https://docs.lavasoftware.org/lava/installing_on_debian.html)

Support
=======

Read the [support](https://docs.lavasoftware.org/lava/support.html) page.

Reporting Bugs
==============

Bugs should first be discussed by subscribing to the
[lava-users](https://lists.lavasoftware.org/mailman/listinfo/lava-users)
mailing list.

Bug reports and new LAVA features are tracked using
[GitLab](https://git.lavasoftware.org/lava/lava/issues).

Installation instructions
=======================

For latest information, see the lava-server documentation:
https://master.lavasoftware.org/static/docs/v2/first-installation.html
https://master.lavasoftware.org/static/docs/v2/installing_on_debian.html

Support
=======

https://master.lavasoftware.org/static/docs/v2/support.html

Reporting Bugs
==============

Bugs should first be discussed by subscribing to the
lava-users mailing list.
https://lists.linaro.org/mailman/listinfo/lava-users

Bug reports and new LAVA features are tracked using GitLab:
https://git.lavasoftware.org/lava/lava/issues

- [Latent Variable Models (`lava`)](#latent-variable-models-lava)
  - [Installation](#installation)
  - [Citation](#citation)
  - [Examples](#examples)

<!-- README.md is generated from README.Rmd. Please edit that file -->
<!-- badges: start -->

[![R-CMD-check](https://github.com/kkholst/lava/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/kkholst/lava/actions/workflows/R-CMD-check.yaml)
[![coverage](https://codecov.io/github/kkholst/lava/coverage.svg?branch=main)](https://app.codecov.io/github/kkholst/lava?branch=main)
[![cran](https://www.r-pkg.org/badges/version-last-release/lava)](https://CRAN.R-project.org/package=lava)
[![cran-dl](https://cranlogs.r-pkg.org/badges/lava)](https://cranlogs.r-pkg.org/downloads/total/last-month/lava)
<!-- badges: end -->

# Latent Variable Models (`lava`)

<img src="man/figures/logo.png" align="right" height=250 style="float:right; height:250px;" class="pkgdown-hide" alt="lava website"/>

A general implementation of Structural Equation Models with latent
variables (MLE, 2SLS, and composite likelihood estimators) with both
continuous, censored, and ordinal outcomes (Holst and Budtz-Joergensen
(2013) <doi:10.1007/s00180-012-0344-y>). Mixture latent variable models
and non-linear latent variable models (Holst and Budtz-Joergensen (2020)
<doi:10.1093/biostatistics/kxy082>). The package also provides methods
for graph exploration (d-separation, back-door criterion), simulation of
general non-linear latent variable models, and estimation of influence
functions for a broad range of statistical models.

## Installation

``` r
install.packages("lava", dependencies=TRUE)
library("lava")
demo("lava")
```

For graphical capabilities the `Rgraphviz` package is needed (first
install the `BiocManager` package)

``` r
# install.packages("BiocManager")
BiocManager::install("Rgraphviz")
```

or the `igraph` or `visNetwork` packages

``` r
install.packages("igraph")
install.packages("visNetwork")
```

The development version of `lava` may also be installed directly from
`github`:

``` r
# install.packages("remotes")
remotes::install_github("kkholst/lava")
```

## Citation

To cite that `lava` package please use one of the following references

> Klaus K. Holst and Esben Budtz-Joergensen (2013). Linear Latent
> Variable Models: The lava-package. Computational Statistics 28 (4), pp
> 1385-1453. <http://dx.doi.org/10.1007/s00180-012-0344-y>

    @article{lava,
      title = {Linear Latent Variable Models: The lava-package},
      author = {Klaus Kähler Holst and Esben Budtz-Jørgensen},
      year = {2013},
      volume = {28},
      number = {4},
      pages = {1385-1452},
      journal = {Computational Statistics},
      doi = {10.1007/s00180-012-0344-y}
    }

> Klaus K. Holst and Esben Budtz-Jørgensen (2020). A two-stage
> estimation procedure for non-linear structural equation models.
> Biostatistics 21 (4), pp 676-691.
> <http://dx.doi.org/10.1093/biostatistics/kxy082>

    @article{lava_nlin,
      title = {A two-stage estimation procedure for non-linear structural equation models},
      author = {Klaus Kähler Holst and Esben Budtz-Jørgensen},
      journal = {Biostatistics},
      year = {2020},
      volume = {21},
      number = {4},
      pages = {676-691},
      doi = {10.1093/biostatistics/kxy082},
    }

## Examples

### Influence functions

Construct `estimate` objects from parameter coefficients and estimated
influence functions

``` r
a <- estimate(coef=c("a"=0.5), IC=scale(rnorm(10)), id=1:10)
b <- estimate(coef=c("b"=0.8), IC=scale(rnorm(10)), id=6:15)
```

Alternatively, we can construct estimate objects directly from an
existing model object (`glm()`, `mets::phreg()`, `targeted::cate()`, …)

``` r
estimate(modelobj, id, ...)
```

We can now merge the `estimate` objects to obtain their joint
distribution via their estimated influence functions

``` r
e <- c(a, b)
vcov(e) # joint distribution
#>            a          b
#> a 0.09000000 0.01469326
#> b 0.01469326 0.09000000
summary(e, null=c(0, 0))
#> Call: estimate.default(f = FALSE, contrast = contrast, null = ..1, 
#>     vcov = vcov(object), coef = p)
#> ────────────────────────────────────────────────────────────
#>   Estimate Std.Err     2.5% 97.5%  P-value
#> a      0.5     0.3 -0.08799 1.088 0.095581
#> b      0.8     0.3  0.21201 1.388 0.007661
#> ────────────────────────────────────────────────────────────
#> Null Hypothesis: 
#>   [a] = 0
#>   [b] = 0 
#>  
#> chisq = 8.6688, df = 2, p-value = 0.01311
```

Parameter transformations can be calculated directly as in the following
examples.

Products

``` r
a * b
#>   Estimate Std.Err   2.5% 97.5% P-value
#> a      0.4  0.3031 -0.194 0.994  0.1869
```

General transformations

``` r
(3 * cos(a) / sqrt(b) + 1) / a^2
#>   Estimate Std.Err   2.5% 97.5% P-value
#> a    15.77   21.33 -26.03 57.58  0.4596
```

Inner product, sums, and products

``` r
c(iprod=e %*% c(a, b^2), sum=sum(e), prod=prod(e))
#>       Estimate Std.Err    2.5% 97.5%  P-value
#> iprod    0.762  0.6915 -0.5934 2.117 0.270495
#> sum      1.300  0.4576  0.4031 2.197 0.004497
#> prod     0.400  0.3031 -0.1940 0.994 0.186900
```

Exponentiation and renaming of parameter

``` r
c(pow = a^b)
#>     Estimate Std.Err    2.5% 97.5% P-value
#> pow   0.5743   0.282 0.02166 1.127 0.04167
```

Transformation and subsetting

``` r
c(e["a"] * e["b"] / a, e["b"])
#>   Estimate Std.Err  2.5% 97.5%  P-value
#> a      0.8     0.3 0.212 1.388 0.007661
#> b      0.8     0.3 0.212 1.388 0.007661
```

For the `%*%*` operator we can also use a general contrast matrix

``` r
B <- rbind(c(1,-1), c(1,0), c(0,1))
B %*% e
#>           Estimate Std.Err     2.5%  97.5%  P-value
#> [a] - [b]     -0.3  0.3881 -1.06064 0.4606 0.439512
#> a              0.5  0.3000 -0.08799 1.0880 0.095581
#> b              0.8  0.3000  0.21201 1.3880 0.007661
#> ────────────────────────────────────────────────────────────
#> Null Hypothesis: 
#>   [a] - [b] = 0
#>   [a] = 0
#>   [b] = 0 
#>  
#> chisq = 8.6688, df = 2, p-value = 0.01311
plot(B %*% e)
```

![](man/figures/estimate-contrast-1.svg)<!-- -->

### Structural Equation Model

Specify structural equation models with two factors

``` r
m <- lvm()
regression(m) <- y1 + y2 + y3 ~ u1
regression(m) <- z1 + z2 + z3 ~ u2
latent(m) <- ~ u1 + u2
regression(m) <- u2 ~ u1 + x
regression(m) <- u1 ~ x

plot(m)
```

<img src="man/figures/lvm1-1.svg" alt="" style="display: block; margin: auto;" />

Simulation

``` r
d <- sim(m, 100, seed=1)
```

Estimation

``` r
e <- estimate(m, d)
e
#>                     Estimate Std. Error  Z-value   P-value
#> Measurements:                                             
#>    y2~u1             0.95462    0.08083 11.80993    <1e-12
#>    y3~u1             0.98476    0.08922 11.03722    <1e-12
#>     z2~u2            0.97038    0.05368 18.07714    <1e-12
#>     z3~u2            0.95608    0.05643 16.94182    <1e-12
#> Regressions:                                              
#>    u1~x              1.24587    0.11486 10.84694    <1e-12
#>     u2~u1            0.95608    0.18008  5.30910 1.102e-07
#>     u2~x             1.11495    0.25228  4.41951 9.893e-06
#> Intercepts:                                               
#>    y2               -0.13896    0.12458 -1.11537    0.2647
#>    y3               -0.07661    0.13869 -0.55241    0.5807
#>    u1                0.15801    0.12780  1.23644    0.2163
#>    z2               -0.00441    0.14858 -0.02969    0.9763
#>    z3               -0.15900    0.15731 -1.01076    0.3121
#>    u2               -0.14143    0.18380 -0.76949    0.4416
#> Residual Variances:                                       
#>    y1                0.69684    0.14858  4.69004          
#>    y2                0.89804    0.16630  5.40026          
#>    y3                1.22456    0.21182  5.78109          
#>    u1                0.93620    0.19623  4.77084          
#>    z1                1.41422    0.26259  5.38570          
#>    z2                0.87569    0.19463  4.49934          
#>    z3                1.18155    0.22640  5.21883          
#>    u2                1.24430    0.28992  4.29195
```

### Model assessment

Assessing goodness-of-fit, here the linearity between u2 and u1
(requires the `gof` package)

``` r
# install.packages("gof", repos="https://kkholst.github.io/r_repo/")
library("gof")
set.seed(1)
g <- cumres(e, u2 ~ u1)
plot(g)
```

<img src="man/figures/gof1-1.svg" alt="" style="display: block; margin: auto;" />

### Non-linear measurement error model

Simulate non-linear model

``` r
m <- lvm(y1 + y2 + y3 ~ u, u ~ x)
transform(m,u2 ~ u) <- function(x) x^2
regression(m) <- z~u2+u

d <- sim(m,200,p=c("z"=-1, "z~u2"=-0.5), seed=1)
```

Stage 1:

``` r
m1 <- lvm(c(y1[0:s], y2[0:s], y3[0:s]) ~ 1*u, u ~ x)
latent(m1) <- ~ u
(e1 <- estimate(m1, d))
#>                     Estimate Std. Error  Z-value  P-value
#> Regressions:                                             
#>    u~x               1.06998    0.08208 13.03542   <1e-12
#> Intercepts:                                              
#>    u                -0.08871    0.08753 -1.01344   0.3108
#> Residual Variances:                                      
#>    y1                1.00054    0.07075 14.14214         
#>    u                 1.19873    0.15503  7.73233
```

Stage 2

``` r
pp <- function(mu,var,data,...) cbind(u=mu[,"u"], u2=mu[,"u"]^2+var["u","u"])
(e <- measurement.error(e1, z~1+x, data=d, predictfun=pp))
#>             Estimate Std.Err    2.5%   97.5%   P-value
#> (Intercept)  -1.1181 0.13555 -1.3838 -0.8524 1.602e-16
#> x            -0.0537 0.14361 -0.3352  0.2278 7.085e-01
#> u             1.0039 0.12651  0.7560  1.2519 2.093e-15
#> u2           -0.4718 0.05858 -0.5867 -0.3570 7.974e-16
```

``` r
f <- function(p) p[1]+p["u"]*u+p["u2"]*u^2
u <- seq(-1, 1, length.out=100)
plot(e, f, data=data.frame(u))
```

<img src="man/figures/nlin1-1.svg" alt="" style="display: block; margin: auto;" />

### Simulation

Studying the small-sample properties of a mediation analysis

``` r
m <- lvm(y~x, c~1)
regression(m) <- y+x ~ z
eventTime(m) <- t~min(y=1, c=0)
transform(m,S~t+status) <- function(x) survival::Surv(x[,1],x[,2])
```

plot(m)

``` r
plot(m)
```

<img src="man/figures/mediation1-1.svg" alt="" style="display: block; margin: auto;" />

Simulate from model and estimate indirect effects

``` r
future::plan("multicore") # parallelization via future
## progressr::handlers(global=TRUE) # add progress-bar
onerun <- function(...) {
  d <- sim(m, 100)
  m0 <- lvm(S~x+z, x~z)
  e <- estimate(m0, d, estimator="glm")
  vec(summary(effects(e, S~z))$coef[,1:2])
}
val <- sim(onerun, 100)
summary(val, estimate=1:4, se=5:8, short=TRUE)
#> 100 replications                 Time: 7.381s
#> 
#>         Total.Estimate Direct.Estimate Indirect.Estimate S~x~z.Estimate
#> Mean           1.99533         1.00468           0.99066        0.99066
#> SD             0.18662         0.16553           0.17792        0.17792
#> SE             0.18258         0.17507           0.16503        0.16503
#> SE/SD          0.97835         1.05767           0.92755        0.92755
#>                                                                        
#> Min            1.49759         0.56922           0.50331        0.50331
#> 2.5%           1.66826         0.68806           0.67154        0.67154
#> 50%            1.96614         1.00175           0.97848        0.97848
#> 97.5%          2.39620         1.29212           1.35790        1.35790
#> Max            2.63461         1.42005           1.47636        1.47636
#>                                                                        
#> Missing        0.00000         0.00000           0.00000        0.00000
```

Add additional simulations and visualize results

``` r
val <- sim(val,500) ## Add 500 simulations
plot(val, estimate=c("Total.Estimate", "Indirect.Estimate"),
     true=c(2, 1), se=c("Total.Std.Err", "Indirect.Std.Err"),
     scatter.plot=TRUE)
```

<img src="man/figures/simres1-1.svg" alt="" style="display: block; margin: auto;" />

LAVA
====

LAVA is an automated validation architecture primarily aimed at testing
deployments of systems based around the Linux kernel on ARM devices,
specifically ARMv7 and later.

This repository contains the dispatcher and server source code.

Installation instructions
=========================

For latest information, see the lava-server documentation:
* [first installation](https://docs.lavasoftware.org/lava/first-installation.html)
* [installing on Debian](https://docs.lavasoftware.org/lava/installing_on_debian.html)

Support
=======

Read the [support](https://docs.lavasoftware.org/lava/support.html) page.

Reporting Bugs
==============

Bugs should first be discussed by subscribing to the
[lava-users](https://lists.lavasoftware.org/mailman3/lists/lava-users.lists.lavasoftware.org/)
mailing list.

Bug reports and new LAVA features are tracked using
[GitLab](https://gitlab.com/lava/lava/issues).

Jinja2 support in the new dispatcher
====================================

The new dispatcher removes the need for device configuration on the
dispatcher itself, all device configuration occurs through entries in
the database on the master scheduler and templates which assert the
defaults and overrides.

In time, these templates will move into a separate binary package so
that templates can be updated more easily. It is also possible that
override directories can be supported.

The files in the 'devices' directory are examples which will be used
in development and to provide a way to add devices for the new
dispatcher.

Developer builds
================

debian-dev-build.sh is a simple wrapper to build ad-hoc packages for
developers to test their local changes. Packages built from the release
branch would use the tag name alone - note that normally release branch
builds will be done by Debian maintainers from official tarballs released
by Linaro. Packages built from personal branches or the master branch
will use the most recent tag name and have a suffix specifying the short
git hash of the most recent commit on that branch. (This may be the same
as the hash of the commit merged into the release branch to create the
tag.)

Developer builds need various Debian/Ubuntu package tools installed:

$ sudo apt-get install dpkg-dev
$ sudo apt-get --no-install-recommends install devscripts

The lava-server build-dependencies also need to be installed - these
change from time to time but debian-dev-build.sh will check
dpkg-checkbuilddeps and halt and detail any build-dependencies not
already installed.

A log of the Debian package build will be placed alongside the
packages.

The Debian packaging for lava-server and lava-dispatcher are
available at http://github.com/Linaro/

Building KVM images for LAVA
============================

 git clone git://git.linaro.org/lava/lava-vmdebootstrap.git

lava-vmdebootstrap is just a small wrapper around
[vmdebootstrap](https://gitorious.org/vmdebootstrap) which does all the hard
work. What lava-vmdebootstrap does is download the Linaro image overlays
and pass some options that we will always need so that you don't need to.

Example invocation::

 $ sudo ./lava-vmdebootstrap --image=myimage.img

To run the test image, make sure it is writeable:

 $ sudo chmod a+w ./myimage.img

Execute using qemu, e.g. on amd64 using qemu-system-x86_64:

 $ qemu-system-x86_64 ./myimage.img

See vmdebootstrap(1)

Once the overlay packages have been downloaded, you can call vmdebootstrap
directly to create other types of images without needing to modify
lava-vmdebootstrap. e.g. this is a call to vmdebootstrap to create a
KVM image based on Ubuntu:

 sudo vmdebootstrap \
  --custom-package='linaro-overlay_1112.2_all.deb' \
  --custom-package='linaro-overlay-minimal_1112.2_all.deb' \
  --enable-dhcp --no-kernel --package=linux-image \
  --serial-console --serial-console-command='/bin/auto-serial-console' \
  --root-password='root' --hostname='ubuntu' --user=linaro/linaro --sudo \
  --verbose --image=myimage.img

This command extends lava-vmdebootstrap to make a 4G image based on
Debian testing using the UK Debian mirror:

 sudo vmdebootstrap \
  --custom-package='linaro-overlay_1112.2_all.deb' \
  --custom-package='linaro-overlay-minimal_1112.2_all.deb' \
  --enable-dhcp \
  --serial-console --serial-console-command='/bin/auto-serial-console' \
  --root-password='root' \
  --distribution testing --size 4g \
  --mirror http://ftp.uk.debian.org/debian \
  --verbose --image=myimage.img

Docker packaging
================

This repository contains docker image definitions that will be used by lava ci
jobs to build the official lava-server and lava-dispatcher images.

The CI job for this repository is responsible for building, testing and
publishing the base Docker images on hub.lavasoftware.org/lava/pkg/lava/
These images will be used as base images for building official lava docker images.

The docker images will be available at hub.lavasoftware.org/lava/lava/

The [registry](https://git.lavasoftware.org/lava/lava/container_registry)
list the available images.

You can manually build and test the images, by running `./build.sh` which will
detect which architecture to build using `uname -m`.

In GitLab, base images are built and published for aarch64 and amd64
architectures.

Docker packaging
================

This repository contains docker image definitions that will be used by lava ci
jobs to build the official lava-server and lava-dispatcher images.

The CI job for this repository is responsible for building, testing and
publishing the base Docker images on registry.gitlab.com/lava/pkg/lava/
These images will be used as base images for building official lava docker images.

The docker images will be available at registry.gitlab.com/lava/lava/

The [registry](https://gitlab.com/lava/lava/container_registry)
list the available images.

You can manually build and test the images, by running `./build.sh` which will
detect which architecture to build using `uname -m`.

In GitLab, base images are built and published for aarch64 and amd64
architectures.

Apache config only for pipeline workers
=======================================

LAVA V2 workers just have lava-dispatcher installed, so unless a usable
localhost website is available, retrieving files within the dispatcher can
fail.

This is a section of the lava-server apache config, adapted just for
lava-dispatcher installations.

sudo cp share/apache2/lava-dispatcher.conf /etc/apache2/sites-available/
sudo a2ensite lava-dispatcher
sudo a2dissite 000-default

Note: not all V2 workers need a localhost server as long as TFTP is working but
it can be useful for persistent NFS.

If the DocumentRoot and Alias need to be changed in the localhost
configuration, the value in DISPATCHER_DOWNLOAD_DIR also needs to be changed in
/usr/lib/python3/dist-packages/lava_common/constants.py

(Support for reading constants independently of the packaged code is pending.)

# Dockerized LAVA

The [Dockerfile](Dockerfile) in this directory creates an image based on the
LAVA Debian packaging that exists in the repository for lava-server. You can
use it using regular docker commands and an environment file with at least the
following variables:

```
INSTANCE_NAME=<arbitrary name>
DATABASE_URL=<database>://<user>:<password>@<host>:<port>/<database>
SECRET_KEY=<django's secret key>
ALLOWED_HOSTS=<hosts that can connect to the django server>
```

All of these are regular Django settings that would go into settings.py
`or local_settings.py except for DATABASE_URL, which is specific to the
django-environ library. The reason we use that instead of having variables
for each piece of the database configuration is django-environ handles
corner cases like using sqlite's memory backend for us.

Find more information on the format and database-specific quirks in
[django-environ's documentation](https://django-environ.readthedocs.io/en/latest/).

In addition to lava-server there are also Dockerfile setups for a sinmple
packages build [Dockerfile.build](Dockerfile.build), a dispatcher/worker
[Dockerfile.dispatcher](Dockerfile.dispatcher), which depends on the image from
the packages build, and one for providing the `lavacli` command line utility
[Dockerfile.lavacli](Dockerfile.lavacli).

To use the dispatcher/worker image you can specify the following
variables:

```
WORKER_NAME=--name <name>
URL=--url http://<host>:<port>/
WS_URL=--ws-url http://<host>:<ws-port>/ws/
LOGLEVEL=--log-level <level>
TOKEN=--token <token>
```

They are basically the same you will need to set up in the configuration
file if you use the package. Note that lava-server opens two ports, which
by default are 8000 and 8001, the first one is for `URL` and the second
is the web socket port, `WS_URL`.

## Using docker-compose

A [basic docker-compose file](docker-compose.yaml) exists that can serve as a
basis for your own setup. It specifies 2 volumes, `lavaconfig` and `lavadbdata`
to be used for mounting configuration and database directories respectively.
The actual volumes created and used by docker-compose will get a prefix based on
the directory that hosts the compose file.

For instance, if you simply use the files provided in this repository to start
your lava-server instance the volumes will have a `collabora_` prefix:

```
kov@cereja ~/L/lava> docker volume ls
DRIVER    VOLUME NAME
local     collabora_lavaconfig
local     collabora_lavadbdata
kov@cereja ~/L/lava>
```

It is possible to simply use this base compose file along with a minimal local
override, as long as you have the PostgreSQL server set up with the appropriate
user and database created. First create the override file at the root of the
repository to specify the variables:

```
$ cat docker-compose.local.yaml
version: "3"

services:
    lava:
        environment:
            - INSTANCE_NAME=collabora
            - DATABASE_URL=postgresql://lavaserver:lavapass@db:5432/lavaserver
            - SECRET_KEY=a-ib@8x^c!-n=xf!831@ka$0&x2h8k%ml+is_x!+hn&kcjklt)
            - ALLOWED_HOSTS=*
$
```

You can then fire it up from the root of the repository by running
docker-compose like this:

```
$ docker-compose -f docker/collabora/docker-compose.yaml -f docker-compose.local.yaml up
```

### Ports and reverse-proxying

Running the default compose service will open 2 new ports on your localhost:
`8000` is the main LAVA server instance. It is where you should point your main
reverse proxy to. A second port, `8001` is also opened and is for the web
socket used by the dispatcher/worker.

This is what an apache2 mod-proxy configuration would look like:

```
    # Send web socket requests to lava-publisher
    ProxyPass /ws/ ws://127.0.0.1:8001/ws/
    ProxyPassReverse /ws/ ws://127.0.0.1:8001/ws/

    # Send request to Gunicorn
    ProxyPass / http://127.0.0.1:8000/
    ProxyPassReverse / http://127.0.0.1:8000/
    ProxyPreserveHost On
```

### Dispatcher configuration and tokens

When setting up the worker, if you are using a gateway like Apache or Nginx you
don't need to worry about the two different ports as they will all be available
through the same one. If you set up a worker talking directly to the
containers, though, don't forget that there are two separate ports.

The key settings are `URL`, which should point to the server URL, and `WS_URL`
that should point to the server URL with `/ws/` appended at the end. This
suffix is crucial for it to work.

The final key setting is `TOKEN`, which you obtain when you create the worker
on the LAVA server. If you use the Django admin web interface, you will see the
token at the bottom of the page. You can just copy it out of there and put it
in your configuration file.

The token is also shown when you use the `lava-server manage` command line
tool to create the worker on the server, of when querying worker details:

```
# lava-server manage workers details ci.worker
hostname   : my.worker
state      : Offline
health     : Active
description:
token      : kfiq2vPsNqH7Pj22rMCQYN0lLDeV7nMe
devices    : 1
#
```

A worker configuration for the worker above may look something like this:

```
$ cat /etc/lava-dispatcher/lava-worker
# Configuration for lava-worker daemon

# worker name
# Should be set for host that have random hostname (containers, ...)
# The name can be any unique string.
WORKER_NAME="--name my.worker"

# Logging level should be uppercase (DEBUG, INFO, WARN, ERROR)
# LOGLEVEL="DEBUG"

# Server connection
URL="http://lava.internal/"
TOKEN="--token kfiq2vPsNqH7Pj22rMCQYN0lLDeV7nMe"
WS_URL="--ws-url http://lava.internal/ws/"
# HTTP_TIMEOUT="--http-timeout 600"
$
```

Note that `WORKER_NAME` may be very important here as well. If your worker does
not have an FQDN or if you set a name other than the server's FQDN when
creating it on server, then you need to set this to match.

### Dockerized database

If you prefer to not have to deal with the database yourself, you can use the
sample [docker-compose.server-db.yaml](docker-compose.server-db.yaml) override.
You can use it where it is, and it will use the sample database configuration.
Or, if you prefer setting your own username, password, secret key and so on,
you should copy the file over to the root of the repository renaming it to
`docker-compose.local.yaml` so you can use the same command as above, and edit
it to your hearts' content.

If you do prefer to use the provided sample configuration, you can simply run
it like this:

```
$ docker-compose -f docker/collabora/docker-compose.yaml -f docker/collabora/docker-compose.server-db.yaml up
```

### Dockerized dispatcher/worker

We recommend running the worker directly on a host using the Debian package.
The main reason for that is giving access to all of the resources needed to
properly control the boards may be tricky, while at the same time the worker is
small and doesn't have a lot of configuration or state.

If you do prefer to run the worker on docker you can use the
[Dockerfile.dispatcher](Dockerfile.dispatcher) file after building the base
build image from [Dockerfile.build](Dockerfile.build).

### Dockerize all the things!

In the repository you can find
[docker-compose.ci.yaml](docker-compose.ci.yaml), used for our CI test run but
which can be used as a docker-compose override to run everything as a container
on the same host. If you intend to use that you will benefit from reading
through the scripts used by the CI, since you first need to set up a worker on
the server before the dispatcher service will work, for instance.

A good place to start is the [ci-run](ci-run) script, which calls out to the
other scripts that build, set up and start the various services in order.

# labgrid2lava

Convert labgrid environment config to lava device and job definitions.

## Installation

On Debian-based distributions:

```bash
sudo apt install python3-dacite python3-jinja2 python3-yaml
```

Using a Python virtual environment:

```bash
python3 -m venv .venv
source .venv/bin/activate
pip install --upgrade pip
pip install dacite jinja2 PyYAML
```

## Usage example

### Convert

Docker:

```bash
python3 ./converter.py -l examples/docker.yaml -o output/ -d docker

```

RPi3 uboot:

```bash
python3 ./converter.py -l examples/uboot.yaml -o output/ -d bcm2837-rpi-3-b-32
```

LAVA `device.yaml` and `job.yaml` definitions are saved to the `output` directory.

### Run

Original labgrid run command:

```bash
pytest --lg-env examples/docker.yaml test_shell.py
```

Equivalent lava run command:

```bash
lava-run --job-id 1 \
  --output-dir ./1 \
  --device ./output/device.yaml \
  ./output/job.yaml
```

Using the dev version lava-run from this repo:

```bash
PYTHONPATH=../../ python3 ../../lava/dispatcher/lava-run \
  --job-id 1 \
  --output-dir ./1 \
  --device ./output/device.yaml \
  ./output/job.yaml
```

LAVA job pipeline description, logs and result are saved to the specified output
directory.

## Supported Features

Refer to the `LgConfig`, `Resources` and `Drivers` in the `converter.py` for
supported labgrid modules.

When converting a labgrid env config, un-supported modules will be listed. If needed,
the converter can be extended. Here are the steps:

* In the `converter.py`, define a dataclass for each module.
* Update the lava device and job templates to define the equivalent blocks for rendering.
* Rerun the convert command.

## Reference

The converter is for converting labgrid configure into LAVA device and job
definitions for running LAVA jobs natively.

For running labgrid tests in a LAVA Test Shell like other test suites, you can
use the [labgrid test definition](https://github.com/Linaro/test-definitions/tree/master/automated/linux/labgrid)

This directory contains support scripts for lava-test-shell.

The scripts in this directory will be copied into the target device and will be
in $PATH during the lava-test-shell execution.

Distribution-specific scripts can be placed in distro/$distroname, and will
override the ones in the top level directory. For example,
distro/android/lava-test-runner will be used on Android instead of the
lava-test-runner script present at the same directory as this README file.

All scripts have to be named using a "lava-" suffix.

Placeholder for the per-instance device files

This directory can be used for local changes to the device dictionary
but note that files placed here need to be imported into the database
using lava-server device-dictionary ... or lava-tool device-dictionary
... support to become available to test jobs.

This directory contains support scripts for lava-test-shell.

The scripts in this directory will be copied into the target device and will be
in $PATH during the lava-test-shell execution.

Distribution-specific scripts can be placed in distro/$distroname, and will
override the ones in the top level directory. For example,
distro/android/lava-test-runner will be used on Android instead of the
lava-test-runner script present at the same directory as this README file.

All scripts have to be named using a "lava-" prefix.

See also:
https://staging.validation.linaro.org/static/docs/v2/dispatcher-testing.html#lava-dispatcher
or http://localhost/static/docs/v2/dispatcher-testing.html#lava-dispatcher

To rebuild and update a pipeline reference, use:

    import yaml
    with open('/tmp/test.yaml', 'w') as describe:
        yaml.dump(self.job.pipeline.describe(False), describe)

(Avoid opening in binary mode as this would fail with python3.)

Then use:

from lava_dispatcher.tests.test_basic import pipeline_reference

    # Check Pipeline
    description_ref = self.pipeline_reference('kexec.yaml')
    self.assertEqual(description_ref, self.job.pipeline.describe(False))

or just

    self.assertEqual(self.pipeline_reference(filename), self.job.pipeline.describe(False))

The name of the pipeline_ref file should match the name of the equivalent file in sample_jobs.

To change multiple pipeline references at the same time, change
self.update_ref to True in test_basic.StdoutTestCase - always check all changes to
the pipeline reference files *carefully* before sending for review.

Once you have a new unit test function with the pipeline_reference check in place,
the pipeline_reference can be updated by adding a temporary line:

    self.update_ref = True

Re-run the unit test and remove the line. Check that the diff for the pipeline
reference is sane before submitting for review.

For an example of a short unit test using a pipeline_reference, see:
lava_dispatcher/tests/test_uboot.py - def test_transfer_media(self):

See also:
https://staging.validation.linaro.org/static/docs/v2/dispatcher-testing.html#lava-dispatcher
or http://localhost/static/docs/v2/dispatcher-testing.html#lava-dispatcher

To rebuild and update a pipeline reference, use:

    import yaml
    with open('/tmp/test.yaml', 'w') as describe:
        yaml.dump(self.job.pipeline.describe(), describe)

(Avoid opening in binary mode as this would fail with python3.)

Then use:

from tests.lava_dispatcher.test_basic import pipeline_reference

    # Check Pipeline
    description_ref = self.pipeline_reference('kexec.yaml')
    self.assertEqual(description_ref, self.job.pipeline.describe())

or just

    self.assertEqual(self.pipeline_reference(filename), self.job.pipeline.describe())

The name of the pipeline_ref file should match the name of the equivalent file in sample_jobs.

To change multiple pipeline references at the same time, change
self.update_ref to True in test_basic.LavaDispatcherTestCase - always check all changes to
the pipeline reference files *carefully* before sending for review.

Once you have a new unit test function with the pipeline_reference check in place,
the pipeline_reference can be updated by adding a temporary line:

    self.update_ref = True

Re-run the unit test and remove the line. Check that the diff for the pipeline
reference is sane before submitting for review.

For an example of a short unit test using a pipeline_reference, see:
tests/lava_dispatcher/test_uboot.py  - def test_transfer_media(self):

# ----- Deprecated: removed from lava in version 1.9.1 -----
# Monte Carlo simulation-based p-values via estimate().
# This code is preserved for reference only and is NOT loaded by the package.

# ---- estimate_sim ----------------------------------------------------------
# Standalone replacement for the removed estimate(..., R=, null.sim=) pathway.
#
# @param x model object (glm, lvmfit, ...)
# @param f transformation of model parameters
# @param R number of simulations
# @param null.sim mean under the null for simulations
# @param level level of confidence limits
# @param vcov (optional) covariance matrix of parameter estimates
# @param coef (optional) parameter coefficients
# @param labels (optional) names of coefficients
# @param robust if TRUE robust standard errors are used
# @param IC if TRUE the influence function decomposition is used
# @param ... additional arguments
estimate_sim <- function(x = NULL, f, R, null.sim, ...,
                         level = 0.95, vcov, coef, labels,
                         robust = TRUE, IC = robust) {
  if (is.null(f)) stop("Supply function 'f'")
  if (!missing(coef)) {
    pp <- coef
  } else {
    pp <- stats::coef(x)
  }
  if (missing(vcov) || is.null(vcov)) {
    if (!is.null(IC) && ((is.logical(IC) && IC) || length(IC) > 0) && robust) {
      ic_theta <- lava::IC(x)
      V <- var(ic_theta) / nrow(ic_theta)
    } else {
      V <- stats::vcov(x)
    }
  } else {
    V <- cbind(vcov)
  }
  if (missing(null.sim)) null.sim <- rep(0, length(pp))
  est <- f(pp)
  if (is.list(est)) {
    nn <- names(est)
    est <- unlist(est)
    names(est) <- nn
  }
  if (missing(labels)) {
    labels <- colnames(rbind(est))
  }
  res <- simnull(R, f, mu = null.sim, sigma = V, labels = labels)
  structure(res, class = c("estimate.sim", "sim"),
            coef = pp,
            vcov = V,
            f = f,
            estimate = est)
}

simnull <- function(R, f, mu, sigma, labels = NULL) {
  X <- lava::rmvn0(R, mu = mu, sigma = sigma)
  est <- f(mu)
  res <- apply(X, 1, f)
  if (is.list(est)) {
    nn <- names(est)
    est <- unlist(est)
    names(est) <- nn
    res <- matrix(unlist(res), byrow = TRUE, ncol = length(est))
  } else {
    res <- t(rbind(res))
  }
  if (is.null(labels)) {
    labels <- colnames(rbind(est))
    if (is.null(labels))
      labels <- paste0("p", seq_along(est))
  }
  colnames(res) <- labels
  return(res)
}

estimate.estimate.sim <- function(x, f, R = 0, labels, ...) {
  atr <- attributes(x)
  if (R > 0) {
    if (missing(f)) {
      val <- simnull(R, f = atr[["f"]], mu = atr[["coef"]], sigma = atr[["vcov"]])
      res <- rbind(x, val)
      for (a in setdiff(names(atr), c("dim", "dimnames")))
        attr(res, a) <- atr[[a]]
    } else {
      res <- simnull(R, f = f, mu = atr[["coef"]], sigma = atr[["vcov"]])
      for (a in setdiff(names(atr), c("dim", "dimnames", "f")))
        attr(res, a) <- atr[[a]]
      attr(f, "f") <- f
      est <- unlist(f(atr[["coef"]]))
      if (missing(labels)) labels <- colnames(rbind(est))
      attr(res, "estimate") <- est
    }
    if (!missing(labels)) colnames(res) <- labels
    return(res)
  }
  if (missing(f)) {
    if (!missing(labels)) colnames(res) <- labels
    return(x)
  }

  est <- f(atr[["coef"]])
  res <- apply(x, 1, f)
  if (is.list(est)) {
    res <- matrix(unlist(res), byrow = TRUE, ncol = length(est))
  } else {
    res <- t(rbind(res))
  }
  if (missing(labels)) {
    labels <- colnames(rbind(est))
    if (is.null(labels)) labels <- paste0("p", seq_along(est))
  }
  colnames(res) <- labels
  for (a in setdiff(names(atr), c("dim", "dimnames", "f", "estimate")))
    attr(res, a) <- atr[[a]]
  attr(f, "f") <- f
  attr(res, "estimate") <- unlist(est)
  return(res)
}

print.estimate.sim <- function(x, level = .95, ...) {
  quantiles <- c((1 - level) / 2, 1 - (1 - level) / 2)
  est <- attr(x, "estimate")
  mysummary <- function(x, INDEX, ...) {
    x <- as.vector(x)
    res <- c(mean(x, na.rm = TRUE),
             sd(x, na.rm = TRUE),
             quantile(x, quantiles, na.rm = TRUE),
             est[INDEX],
             mean(abs(x) > abs(est[INDEX]), na.rm = TRUE))

    names(res) <- c("Mean", "SD", paste0(quantiles * 100, "%"),
                    "Estimate", "P-value")
    res
  }
  env <- new.env()
  assign("est", attr(x, "estimate"), env)
  environment(mysummary) <- env
  print(summary(x, fun = mysummary, ...))
}

# ---- examples --------------------------------------------------------------
# Monte Carlo approach, simple trend test example
#
# library(lava)
# m <- categorical(lvm(), ~x, K = 5)
# regression(m, additive = TRUE) <- y ~ x
# d <- simulate(m, 100, seed = 1, 'y~x' = 0.1)
# l <- lm(y ~ -1 + factor(x), data = d)
#
# f <- function(x) coef(lm(x ~ seq_along(x)))[2]
# null <- rep(mean(coef(l)), length(coef(l)))
# ## just need to make sure we simulate under H0: slope = 0
# source(system.file("misc/estimate_sim.R", package = "lava"))
# estimate_sim(l, f, R = 1e2, null.sim = null)

##' Read Mplus output files
##'
##' @title Read Mplus output
##' @param infile Mplus output file
##' @param coef Coefficients only
##' @param \dots additional arguments to lower level functions
##' @author Klaus K. Holst
##' @export
##' @seealso getSAS
`getMplus` <-
    function(infile="template.out",coef=TRUE,...) {
        if (coef) {
            start <- "MODEL RESULTS"
            end1 <- "R-SQUARE"
            res0 <- findfileseg(infile,start)[-c(seq(5))]
            res <- sapply(res0,function(x) { val <- strsplit(x," ")[[1]]; val[val!=""] })
            res <- res[unlist(lapply(res, length))!=0]

            coef.idx <- unlist(lapply(res, length))>3
            lab.idx <- which(!coef.idx)
            count <- 0
            mycoef <- c()
            myrownames <- c()
            for (i in seq_along(res)) {
                if (i %in% lab.idx) {
                    count <- count+1
                } else {
                    val <- char2num(res[[i]][-1])
                    if (length(val)<5) val <- c(val,rep(0,5-length(val)))
                    mycoef <- rbind(mycoef, val)
                    myrownames <- c(myrownames,
                                    paste(paste(res[[lab.idx[count]]],collapse=" "),res[[i]][1])
                                    )

                }
            }
            rownames(mycoef) <- myrownames
            colnames(mycoef) <- c("Estimate","Std.Err","Z-value","Std","StdYX")
            return(mycoef)
        }

        start <- "Estimate       S.E.  Est./S.E."
        end1 <- "MODEL RESULTS"
        ## end2 <- "QUALITY OF NUMERICAL RESULTS"
        res <- findfileseg(infile,start,end1);
        cat(paste(res,"\n"))

        res <- findfileseg(infile, "TESTS OF MODEL FIT", "Chi-Square Test of Model Fit for the Baseline Model")
        cat(paste(res,"\n"))
    }



`findfileseg` <-
    function(infile, startstring, endstring,nlines) {
        con <- file(infile, blocking = FALSE)
        inp <- readLines(con)
        close(con)
        linestart <- 1; lineend <- length(inp)

        mycmd1 <- paste0("grep -n \"",startstring,"\" ", infile);  a1 <- system(mycmd1,intern=TRUE);
        if (length(a1)>0)
            linestart <- char2num(strsplit(a1,":")[[1]][1])

        nn <- length(inp)
        if (!missing(nlines)) nn <- linestart+nlines
        if (missing(endstring)) {
            for (i in seq(linestart,nn)) {
                lineend <- i-1
                if (inp[i]==inp[i-1]) break;
            }
        } else {
            mycmd2 <- paste0("grep -n \"",endstring,"\" ", infile);  a2 <- system(mycmd2,intern=TRUE);
            if (length(a2)>0)
                lineend <- char2num(strsplit(a2,":")[[1]][1])
        }

        res <- inp[linestart:lineend-1]
        return(res)
    }


##################################################
### Generate code and run mplus...
##################################################

`mplus` <-
    function(file="template.mplus",wait=TRUE,intern=TRUE,...) {
        if (!file.exists(file)) file <- paste0(file,".mplus")
        if (!file.exists(file)) stop("File does not exist")
        if (!exists("winecmd")) winecmd <- "wine"
        if (!exists("mplus.directory")) mplus.directory <- ""
        mycmd <- paste0(winecmd, " \"", mplus.directory, "mplus.exe\" ", file)
        system(mycmd, wait=wait, intern=TRUE)
        prefix <- strsplit(file, ".", fixed=TRUE)[[1]][1]
        return(getMplus(paste0(prefix,".out"),coef=TRUE))
    }

`toMplus.data.frame` <-
    function(x, datafile="data.tab",
      mplusfile="template.mplus",
      na.string=".", model="!f1 by x1;",
      analysis=NULL,
      categorical=NULL,
      group,
      run=FALSE, techout=FALSE,missing=TRUE,...) {
        write.table(x, file=datafile, sep="\t",
                    quote=FALSE, row.names=FALSE, col.names=FALSE, na=na.string)
        varnames <- c()
        ngroups <- ceiling(ncol(x)/4)

        for (i in seq_len(ngroups)) {
            newline <- c("\t",colnames(x)[((i-1)*4+1):min(ncol(x), (i*4))],"\n")
            varnames <- c(varnames, newline)
        }

        zz <- file(mplusfile, "w")  # open an output file connection
        cat(file=zz, "TITLE: ...\n")
        cat(file=zz, "!-----------------------------------------------------\n")
        cat(file=zz,"DATA:\n\tFILE=\"", datafile, "\";\n")
        cat(file=zz,"VARIABLE:\n\tNAMES ARE\n")
        cat(file=zz, varnames, ";\n")
        if (!missing(group)) {
            groups <- unique(x[,group])
            mygroupdef <- paste("(",paste(groups,groups,sep="=",collapse=","),")")
            cat(file=zz, "GROUPING IS ", group, mygroupdef, ";\n", sep="")
        } else {
            cat(file=zz, "!GROUPING IS g (1=male, 2=female);\n")
        }
        cat(file=zz, "USEVARIABLES=\n", varnames,";\n")
        if (!is.null(categorical))
            cat(file=zz, paste("CATEGORICAL=",paste(categorical,collapse=" "),";\n"))
        cat(file=zz, "MISSING=",na.string,";\n",sep="")
        cat(file=zz, "!IDVARIABLE=?;\n")
        cat(file=zz, "!DEFINE: define new variables here;\n")
        cat(file=zz, "!SAVEDATA: save data and/or results;\n\n")
        if (is.null(analysis)) {
            cat(file=zz, "ANALYSIS: TYPE=MEANSTRUCTURE");
            if (missing) cat(file=zz, " MISSING;\n")
            else cat(file=zz,";\n")
            cat(file=zz, "ESTIMATOR=ML;\n")
            cat(file=zz, "INFORMATION=EXPECTED;\n")
            cat(file=zz, "ITERATIONS=5000;\n")
            cat(file=zz, "CONVERGENCE=0.00005;\n\n")
        } else {
            cat(file=zz,"ANALYSIS:\n")
            cat(file=zz, analysis,"\n")
        }
        cat(file=zz, "!-----------------------------------------------------\n")
        cat(file=zz, "MODEL:\n")
        cat(file=zz, model,"\n")
        cat(file=zz, "!-----------------------------------------------------\n")
        if (!techout)
            cat(file=zz, "OUTPUT: STANDARDIZED;\n")
        else
            cat(file=zz, "OUTPUT: MODINDICES(0); TECH1; TECH2; TECH5; STANDARDIZED;\n")
        cat(file=zz, "!\tSAMPSTAT;RESIDUAL;CINTERVAL;MODINDICES(0);\n")
        cat(file=zz, "!Other output options are:\n")
        cat(file=zz, "!\tSTANDARDIZED;     !Standardized coefficients\n")
        cat(file=zz, "!\tH1SE;             !Standard errors for the H1 model\n")
        cat(file=zz, "!\tH1TECH3;          !Estimated covar,corr matrices for par. estimates\n")
        cat(file=zz, "!\tPATTERNS;         !Summary of missing data patterns\n")
        cat(file=zz, "!\tFSCOEFFICIENT;    !Factor score coefficients and posterior covar matrix\n")
        cat(file=zz, "!\tFSDERTERMINACY;   !Factor score determinacy for each factor\n")
        cat(file=zz, "!\tTECH1;            !Parameter specifications and starting values\n")
        cat(file=zz, "!\tTECH2             !Parameter derivatives;\n")
        cat(file=zz, "!\tTECH3;            !Covar and Corr matrices for estimates\n")
        cat(file=zz, "!\tTECH4;            !Estimated means and covar for the latent variables\n")
        cat(file=zz, "!\tTECH5;            !Optimization matrix\n")
        cat(file=zz, "!\tTECH6;            !Optimization for categorical variables\n")
        cat(file=zz, "!\tTECH7;            !output for type Mixture\n")
        cat(file=zz, "!\tTECH8;            !Output for type mixture\n")
        cat(file=zz, "!\tTECH9;            !Error messages for MC study\n")
        cat(file=zz, "!\tMONTECARLO:              File is\n")
        close(zz)

        if (run & exists("mplus")) {
            res <- mplus(mplusfile)
            outfile <- paste0(strsplit(mplusfile,".",fixed=TRUE)[[1]][1],".out")
            getMplus(outfile)
            return(res)
        }
    }

`toMplus.lvmfit` <-
    function(x, model=NULL, data=model.frame(x), run=TRUE, categorical=NULL,##binary(Model(x)),
      mplusfile="template.mplus", ...) {
        mymodel <- ""
        M <- index(x)$M
        P <- index(x)$P
        nn <- vars(x)
        p <- length(nn)
        lat.var <- latent(x)
        lat.idx <- match(lat.var, vars(x))
        for (i in seq_len(p)) {
            for (j in seq_len(p)) {
                if (M[i,j]!=0) {
                    var1 <- nn[i]; var2 <- nn[j];
                    if (i %in% lat.idx & !(j %in% lat.idx)) {## & !(j %in% lat.idx)) {
                        mymodel <- paste0(mymodel, "\n", var1, " by ", var2, ";")
                    } else {
                        mymodel <- paste0(mymodel, "\n", var2, " on ", var1, ";")
                    }
                }
            }
        }
        for (i in seq_len(p-1)) {
            for (j in ((i+1):p)) {
                if (P[i,j]!=0) {
                    var1 <- nn[i]; var2 <- nn[j];
                    mymodel <- paste0(mymodel, "\n", var1, " with ", var2, ";")
                }
            }
        }
        if (is.null(model))
            model <- mymodel
        mydata <- subset(as.data.frame(data), select=setdiff(nn,lat.var))
        toMplus.data.frame(mydata,model=mymodel,run=run, mplusfile=mplusfile, ...)
    }

##' @export
`information` <-
    function(x,...) UseMethod("information")

###{{{ information.lvm

##' @export
information.lvm <- function(x,p,n,type=ifelse(model=="gaussian",
                                       c("E","hessian","varS","outer","sandwich","robust","num"),"outer"),
                     data,weights=NULL,
                     data2=NULL,
                     model="gaussian",
                     method=lava.options()$Dmethod,
                     inverse=FALSE, pinv=TRUE,
                     score=TRUE,...) {
    if (missing(n))
        n <- NROW(data)
    if (type[1]%in%c("sandwich","robust")) {
        cl <- match.call()
        cl$inverse <- !inverse
        cl$type <- "outer"
        A <- eval.parent(cl)
        cl$inverse <- !(cl$inverse)
        cl$type <- ifelse(type[1]=="sandwich","E","hessian")
        B <- eval.parent(cl)
        return(B%*%A%*%B)
    }
    if (type[1]%in%c("num","hessian","obs","observed")  | (type[1]%in%c("E","hessian") & model!="gaussian")) {
        myf <- function(p0) score(x, p=p0, model=model,data=data, weights=weights,data2=data2,indiv=FALSE,n=n) ##...)
        I <- -numDeriv::jacobian(myf,p,method=method)
        res <- (I+t(I))/2 # Symmetric result
        if (inverse) {
            if (pinv)
                iI <- Inverse(res)
            else
                iI <- solve(res)
            return(iI)
        }
        return(res)
    }
    if (type[1]=="varS" | type[1]=="outer") {
        S <- score(x,p=p,data=na.omit(data),model=model,weights=weights,data2=data2,indiv=TRUE,...)
        res <- t(S)%*%S
        if (inverse) {
            if (pinv)
                iI <- Inverse(res)
            else
                iI <- solve(res)
            return(iI)
        }
        attributes(res)$grad <- colSums(S)
        return(res)
    }

    if (n>1) {
        xfix <- colnames(data)[(colnames(data)%in%parlabels(x,exo=TRUE))]
        xconstrain <- intersect(unlist(lapply(constrain(x),function(z) attributes(z)$args)),manifest(x))

        if (length(xfix)>0 | length(xconstrain)>0) { ##### Random slopes!
            x0 <- x
            if (length(xfix)>0) {
                nrow <- length(vars(x))
                xpos <- lapply(xfix,function(y) which(regfix(x)$labels==y))
                colpos <- lapply(xpos, function(y) ceiling(y/nrow))
                rowpos <- lapply(xpos, function(y) (y-1)%%nrow+1)
                myfix <- list(var=xfix, col=colpos, row=rowpos)
                for (i in seq_along(myfix$var))
                    for (j in seq_along(myfix$col[[i]]))
                        regfix(x0, from=vars(x0)[myfix$row[[i]]][j],to=vars(x0)[myfix$col[[i]]][j]) <-
                            data[1,myfix$var[[i]]]
                index(x0) <- reindex(x0,zeroones=TRUE,deriv=TRUE)
            }
            pp <- modelPar(x0,p)
            p0 <- with(pp, c(meanpar,p,p2))
            myfun <- function(ii) {
                if (length(xfix)>0)
                    for (i in seq_along(myfix$var)) {
                        for (j in seq_along(myfix$col[[i]])) {
                            index(x0)$A[cbind(myfix$row[[i]],myfix$col[[i]])] <- data[ii,myfix$var[[i]]]
                        }
                    }
                ww <- NULL
                if (!is.null(weights))
                    ww <- weights[ii,]
                return(information(x0,p=p,n=1,type=type,weights=ww,data=data[ii,]))
            }
            L <- lapply(seq_len(nrow(data)),function(y) myfun(y))
            val <- apply(array(unlist(L),dim=c(length(p0),length(p0),nrow(data))),c(1,2),sum)
            if (inverse) {
                if (pinv)
                    iI <- Inverse(val)
                else
                    iI <- solve(val)
                return(iI)
            }
            return(val)
        }
    }

    if (!is.null(weights) && is.matrix(weights)) {
        L <- lapply(seq_len(nrow(weights)),function(y) information(x,p=p,n=1,type=type,weights=weights[y,]))
        val <- apply(array(unlist(L),dim=c(length(p),length(p),nrow(weights))),c(1,2),sum)
        if (inverse) {
            if (pinv)
                iI <- Inverse(val)
            else
                iI <- solve(val)
            return(iI)
        }
        return(val)
    }
    mp <- moments(x,p,data=data)
    pp <- modelPar(x,p)
    D <- deriv.lvm(x, meanpar=pp$meanpar, mom=mp, p=p)##, all=length(constrain(x))>0)
    C <- mp$C
    iC <- Inverse(C,det=FALSE, symmetric = TRUE)

    if (is.null(weights)) {
        ##    W <- diag(ncol(iC))
    } else {
        if (length(weights)<ncol(iC)) {
            oldweights <- weights
            weights <- rbind(rep(1,ncol(iC))) ## Ones at exogenous var.
            idx <- index(x)$vars%in%index(x)$exogenous
            print(idx); print(oldweights)
            weights[,idx] <- oldweights
        }
        W <- diag(nrow=as.numeric(weights))
        iW <- W
        diag(iW) <- 1/diag(iW)
    }

    if (is.null(weights)) {
        ## information_Sigma <-  n/2*t(D$dS)%*%((iC)%x%(iC))%*%(D$dS)
        if (lava.options()$devel) {
            information_Sigma <- matrix(0,length(p),length(p))
            imean <- with(index(x)$parBelongsTo,mean)
            information_Sigma[-imean,-imean] <- n/2*t(D$dS[,-imean])%*%kronprod(iC,iC,D$dS[,-imean])
        } else {
            information_Sigma <- n/2*t(D$dS)%*%kronprod(iC,iC,D$dS)
        }
    } else {
        ## information_Sigma <-  n/2*t(D$dS)%*%((iC)%x%(iC%*%W))%*%(D$dS)
        information_Sigma <- n/2*t(D$dS)%*%kronprod(iC,iC%*%W,D$dS)
    }

    dxi <- D$dxi;

    if (!any(dxi>0)) { 
        if (inverse) {
            if (pinv)
                iI <- Inverse(information_Sigma)
            else
                iI <- solve(information_Sigma)
            return(iI)
        }
        return(information_Sigma)
    }
    ii <- index(x)
    if (is.null(weights)) {
        information_mu <- n*t(dxi) %*% (iC) %*% (dxi)
    } else {
        information_mu <- n*t(dxi) %*% (iC%*%W) %*% (dxi)
    }

    if (!(lava.options()$devel)) {
        information <- information_Sigma+information_mu
    } else {
        mparidx <- with(ii$parBelongsTo,c(mean,reg))
        information <- information_Sigma
        information[mparidx,mparidx] <- information[mparidx,mparidx] + information_mu
    }

    if (inverse) {
        if (pinv)
            iI <- Inverse(information, symmetric = TRUE)
        else
            iI <- solve(information)
        return(iI)
    }
    return(information)
}

###}}} information.lvm

###{{{ information.lvmfit

##' @export
information.lvmfit <- function(x,p=pars(x),n=x$data$n,data=model.frame(x),model=x$estimator,weights=Weights(x),
                        data2=x$data$data2,
                        ...) {
    I <- information(x$model0,p=p,n=n,data=data,model=model,
                     weights=weights,data2=data2,...)
    if (ncol(I)<length(p)) {
        I <- blockdiag(I,matrix(0,length(p)-ncol(I),length(p)-ncol(I)))
    }
    return(I)
}

###}}} information.lvmfit


##' @export
information.lvm.missing <- function(x,
                             p=coef(x), estimator=x$estimator,
                             weights=Weights(x$estimate),
                             ...) {
    information(x$estimate$model0, p=p, model=estimator, weights=weights,...)
}

##' @export
information.multigroupfit <- function(x,p=pars(x), weights=Weights(x), estimator=x$estimator, ...) {
    information(x$model0,p=p, weights=weights, model=estimator ,...)
}

##' @export
information.multigroup <- function(x,data=x$data,weights=NULL,p,indiv=FALSE,...) {
    rm <- procrandomslope(x)
    pp <- with(rm, modelPar(model,p)$p)
    parord <- modelPar(rm$model,seq_len(with(rm$model,npar+npar.mean)))$p
    I <- matrix(0,nrow=length(p),ncol=length(p))
    if (!indiv) {
        for (i in seq_len(x$ngroup))
            I[parord[[i]],parord[[i]]] <- I[parord[[i]],parord[[i]]] + information(x$lvm[[i]],p=pp[[i]],data=data[[i]],weights=weights[[i]],...)
    } else {
        I <- list()
        for (i in seq_len(x$ngroup))
            I <- c(I, list(information(x$lvm[[i]],p=pp[[i]],data=data[[i]],weights=weights[[i]],...)))
    }
    return(I)
}

##' Generic method for removing elements of object
##'
##' @title Remove variables from (model) object.
##' @aliases rmvar rmvar<- kill kill<-
##' @param x Model object
##' @param value Vector of variables or formula specifying which nodes to
##' remove
##' @param \dots additional arguments to lower level functions
##' @usage
##' rmvar(x, ...) <- value
##' @seealso \code{cancel}
##' @author Klaus K. Holst
##' @keywords models regression
##' @export
##' @examples
##' m <- lvm()
##' addvar(m) <- ~y1+y2+x
##' covariance(m) <- y1~y2
##' regression(m) <- c(y1,y2) ~ x
##' ### Cancel the covariance between the residuals of y1 and y2
##' cancel(m) <- y1~y2
##' ### Remove y2 from the model
##' rmvar(m) <- ~y2
##'
"rmvar" <- function(x, ...) UseMethod("rmvar")

##' @export
"kill" <- function(x, ...) UseMethod("kill")

##' @export
"kill<-" <- function(x, ..., value) UseMethod("kill<-")

##' @export
"rmvar<-" <- function(x, ..., value) UseMethod("rmvar<-")

##' @export
"kill<-.lvm" <- function(x, ..., value) {
    kill(x,value)
}

##' @export
"rmvar<-.lvm" <- get("kill<-.lvm")

##' @export
"kill.lvm" <- function(x, value, ...) {
    if (inherits(value,"formula")) value <- all.vars(value)
    idx <- which(names(x$exfix)%in%value)
    if (length(idx)>0) {
        x$attributes$parameter[idx] <- x$expar[idx] <- x$exfix[idx] <- NULL
        if (length(x$exfix)==0) {
            x$exfix <- x$expar <- x$attributes$parameter <- NULL
        }
        index(x) <- reindex(x)
    }
    idx <- which(vars(x)%in%value)
    if (length(idx)!=0){
        vv <- vars(x)[idx]
        keep <- setdiff(seq_along(vars(x)),idx)
        x$M <- x$M[keep,keep,drop=FALSE]
        x$par <- x$par[keep,keep,drop=FALSE]
        x$fix <- x$fix[keep,keep,drop=FALSE]
        x$covpar <- x$covpar[keep,keep,drop=FALSE]
        x$covfix <- x$covfix[keep,keep,drop=FALSE]
        x$cov <- x$cov[keep,keep,drop=FALSE]
        x$mean <- (x$mean)[-idx]
        x$exogenous <- setdiff(exogenous(x),vv)
        x$latent[vv] <- NULL
    } else{ ## remove variables that cannot be accessed by vars in the hook
        vv <- value
    }
    myhooks <- gethook("remove.hooks")
    for (f in myhooks) {
        x <- do.call(f, list(x=x,var=vv,...))
    }
    index(x) <- reindex(x)
    return(x)
}

##' @export
"rmvar.lvm" <- get("kill.lvm")

##' Generic method for removing elements of object
##'
##' @title Remove variables from (model) object.
##' @aliases rmvar rmvar<- kill kill<-
##' @param x Model object
##' @param value Vector of variables or formula specifying which nodes to
##' remove
##' @param \dots additional arguments to lower level functions
##' @usage
##' rmvar(x, ...) <- value
##' @seealso \code{cancel}
##' @author Klaus K. Holst
##' @keywords models regression
##' @export
##' @examples
##' m <- lvm()
##' addvar(m) <- ~y1+y2+x
##' covariance(m) <- y1~y2
##' regression(m) <- c(y1,y2) ~ x
##' ## Cancel the covariance between the residuals of y1 and y2
##' cancel(m) <- y1~y2
##' ## Remove y2 from the model
##' rmvar(m) <- ~y2
##'
"rmvar" <- function(x, ...) UseMethod("rmvar")

##' @export
"kill" <- function(x, ...) UseMethod("kill")

##' @export
"kill<-" <- function(x, ..., value) UseMethod("kill<-")

##' @export
"rmvar<-" <- function(x, ..., value) UseMethod("rmvar<-")

##' @export
"kill<-.lvm" <- function(x, ..., value) {
    kill(x,value)
}

##' @export
"rmvar<-.lvm" <- get("kill<-.lvm")

##' @export
"kill.lvm" <- function(x, value, ...) {
    if (inherits(value,"formula")) value <- all.vars(value)
    idx <- which(names(x$exfix)%in%value)
    if (length(idx)>0) {
        x$attributes$parameter[idx] <- x$expar[idx] <- x$exfix[idx] <- NULL
        if (length(x$exfix)==0) {
            x$exfix <- x$expar <- x$attributes$parameter <- NULL
        }
        index(x) <- reindex(x)
    }
    idx <- which(vars(x)%in%value)
    if (length(idx)!=0){
        vv <- vars(x)[idx]
        keep <- setdiff(seq_along(vars(x)),idx)
        x$M <- x$M[keep,keep,drop=FALSE]
        x$par <- x$par[keep,keep,drop=FALSE]
        x$fix <- x$fix[keep,keep,drop=FALSE]
        x$covpar <- x$covpar[keep,keep,drop=FALSE]
        x$covfix <- x$covfix[keep,keep,drop=FALSE]
        x$cov <- x$cov[keep,keep,drop=FALSE]
        x$mean <- (x$mean)[-idx]
        x$exogenous <- setdiff(exogenous(x),vv)
        x$latent[vv] <- NULL
    } else{ ## remove variables that cannot be accessed by vars in the hook
        vv <- value
    }
    myhooks <- gethook("remove.hooks")
    for (f in myhooks) {
        x <- do.call(f, list(x=x,var=vv,...))
    }
    index(x) <- reindex(x)
    return(x)
}

##' @export
"rmvar.lvm" <- get("kill.lvm")

##' @export
logit <- function(p) log(p/(1-p))

##' @export
expit <- function(z) 1/(1+exp(-z))

##' @export
tigol <- expit
##' Calculate prevalence, sensitivity, specificity, and positive and
##' negative predictive values
##'
##' @title Calculate diagnostic tests for 2x2 table
##' @aliases diagtest odds riskcomp OR Ratio Diff
##' @param table Table or (matrix/data.frame with two columns)
##' @param positive Switch reference
##' @param exact If TRUE exact binomial proportions CI/test will be used
##' @param p0 Optional null hypothesis (test prevalenc, sensitivity, ...)
##' @param confint Type of confidence limits
##' @param ... Additional arguments to lower level functions
##' @author Klaus Holst
##' @details Table should be in the format with outcome in columns and
##'     test in rows.  Data.frame should be with test in the first
##'     column and outcome in the second column.
##' @examples
##' M <- as.table(matrix(c(42,12,
##'                        35,28),ncol=2,byrow=TRUE,
##'                      dimnames=list(rater=c("no","yes"),gold=c("no","yes"))))
##' diagtest(M,exact=TRUE)
##' @export
diagtest <- function(table,positive=2,exact=FALSE,p0=NA,confint=c("logit","arcsin","pseudoscore","exact"),...) {
    if (!inherits(table,c("table","data.frame","matrix","multinomial")))
        stop("Expecting a table or data.frame.")
    if (is.table(table)) {
        lev <- dimnames(table)[[2]]
    }
    if (inherits(table,"multinomial")) {
        lev <- dimnames(table$P)[[2]]
    }
    if (!is.table(table) & (is.matrix(table) || is.data.frame(table))) {
        if (is.factor(table[,2])) {
            lev <- levels(table[,2])
        } else
            lev <- unique(table[,2])
    }
    if (is.character(positive)) {
        positive <- match(positive,lev)
    }
    if (!(positive%in%c(1,2))) stop("Expecting and index of 1 or 2.")
    negative <- positive%%2+1L
    if (!is.null(confint) && confint[1]=="exact") exact <- TRUE
    if (exact) {
        if (!is.table(table) && (is.matrix(table) || is.data.frame(table))) {
            table <- base::table(table[,c(1,2),drop=FALSE])
            ##names(dimnames(table)) <- colnames(table)[1:2]
        }
        if (!is.table(table) || nrow(table)!=2 || ncol(table)!=2) stop("2x2 table expected")
        n <- sum(table)
        nc <- colSums(table)
        nr <- rowSums(table)
        test <- TRUE
        if (is.na(p0)) {
            test <- FALSE
            p0 <- 0.5
        }
        ## Prevalence
        p1 <- with(stats::binom.test(nc[positive],n,p=p0),c(estimate,conf.int,p.value))
        ## Test marginal
        p2 <- with(stats::binom.test(nr[positive],n,p=p0),c(estimate,conf.int,p.value))
        ## Sensitivity/Specificity
        sens <- with(stats::binom.test(table[positive,positive],nc[positive],p=p0),c(estimate,conf.int,p.value))
        spec <- with(stats::binom.test(table[negative,negative],nc[negative],p=p0),c(estimate,conf.int,p.value))
        ## PPV,NPV
        ppv <- with(stats::binom.test(table[positive,positive],nr[positive],p=p0),c(estimate,conf.int,p.value))
        npv <- with(stats::binom.test(table[negative,negative],nr[negative],p=p0),c(estimate,conf.int,p.value))
        ## Accuracy
        acc <- with(stats::binom.test(table[positive,positive]+table[negative,negative],n,p=p0),c(estimate,conf.int,p.value))
        ## Symmetry (McNemar):
        ##   number of discordant pairs under null: b~bin(b+c,0.5)
        sym <- with(stats::binom.test(table[positive,negative],table[positive,negative]+table[negative,positive],p=0.5),c(estimate,conf.int,p.value))
        coefmat <- rbind(Prevalence=p1,
                         Test=p2,
                         Sensitivity=sens,
                         Specificity=spec,
                         PositivePredictiveValue=ppv,
                         NegativePredictiveValue=npv,
                         Accuracy=acc,
                         Homogeneity=sym)
        if (!test) coefmat[seq(nrow(coefmat)-1),4] <- NA
        coefmat <- cbind(coefmat[,1,drop=FALSE],NA,coefmat[,-1,drop=FALSE])
        colnames(coefmat) <- c("Estimate","Std.Err","2.5%","97.5%","P-value")
        res <- list(table=table, prop.table=table/sum(table),
                    coefmat=coefmat)
    } else {
        if (inherits(table,"table"))
            M <- multinomial(table)
        else {
            if (inherits(table,"multinomial")) {
                M <- table
                table <- round(M$P*nrow(M$data))
            } else {
                M <- multinomial(table[,1:2],...)
                table <- base::table(table)
            }
        }
        calc_diag <- function(p,...) {
            P <- matrix(p[1:4],2)
            p1 <- sum(P[,positive])
            p2 <- sum(P[positive,])
            res <- c(Prevalence=p1,  ##(p[1]+p[2]),
                     Test=p2,        ##(p[1]+p[3]),
                     Sensitivity=P[positive,positive]/p1,     ## p[1]/(p[1]+p[2]), # Prob test + | given (true) disease (True positive rate)
                     Specificity=P[negative,negative]/(1-p1), ## p[4]/(1-p[1]-p[2]), # Prob test - | given no disease (True negative rate)
                     PositivePredictiveValue=P[positive,positive]/p2,     ## p[1]/(p[1]+p[3]), # Prob disease | test +
                     NegativePredictiveValue=P[negative,negative]/(1-p2), ## p[4]/(1-p[1]-p[3]), # Prob disease free | test -
                     Accuracy=(P[1,1]+P[2,2])/sum(P),
                     Homogeneity=P[negative,positive]-P[positive,negative]
                     )
            if (!is.null(confint)) {
                if (tolower(confint[1])=="logit") {
                    res[seq(length(res)-1)] <- logit(res[seq(length(res)-1)])
                } else if (tolower(confint[1])=="arcsin") {
                    res[seq(length(res)-1)] <- asin(sqrt(res[seq(length(res)-1)]))
                }
            }
            return(res)
        }

        names(dimnames(table)) <- paste0(c("Test:","Outcome:"),names(dimnames(table)))
        prfun <- function(x,...) {
            printCoefmat(x$coefmat[,c(-2)],na.print="",...)
            print(cli::rule())
            cat("\n")
            cat("Prevalence:				Prob( outcome+ )\n")
            cat("Test:					Prob( test+ )\n")
            cat("Sensitivity (True positive rate):	Prob( test+ | outcome+ )\n")
            cat("Specificity (True negative rate):	Prob( test- | outcome- )\n")
            cat("Positive predictive value (Precision):	Prob( outcome+ | test+ )\n")
            cat("Negative predictive value:		Prob( outcome- | test- )\n")
            cat("Accuracy:				Prob( correct classification )\n")
            cat("Homogeneity/Symmetry:			Prob( outcome+ ) - Prob( test+ )\n")
        }

        btransform <- NULL
        if (!is.null(confint)) {
            if (tolower(confint[1])=="logit") {
                btransform <- function(x) {
                    rbind(expit(x[seq(nrow(x)-1),,drop=FALSE]),
                          x[nrow(x),,drop=FALSE])
                }
            } else if (tolower(confint[1])=="pseudoscore") {
                ## TODO, agresti-ryu, biometrika 2010
            } else if (tolower(confint[1])=="arcsin")  {
                btransform <- function(x) {
                    rbind(sin(x[seq(nrow(x)-1),,drop=FALSE])^2,
                          x[nrow(x),,drop=FALSE])
                }
            }
        }
        res <- estimate(M,calc_diag,print=prfun,null=c(rep(p0,7),0),back.transform=btransform,...)
    }
    CI <- confint[1]
    if (exact) CI <- "exact"
    if (is.null(CI)) CI <- "wald"
    class(res) <- NULL
    res <- structure(c(res,
                       list(table=table, prop.table=table/sum(table),
                            confint=CI,
                            positive=positive,
                            negative=negative,
                            levels=dimnames(table)
                            )),
                     class=c("diagtest","estimate"))
    res$call <- match.call()
    rownames(res$coefmat) <- gsub("\\[|\\]","",rownames(res$coefmat))
    names(res$coef) <- rownames(res$coefmat)
    return(res)
}

##' @export
print.diagtest <- function(x,...) {
    cat("Call: ")
    print(x$call)
    cat("Confidence limits: ", x$confint, "\n", sep="")
    print(cli::rule())
    printmany(x$table, x$prop.table,nspace=2, ...)
    cat("\nPositive outcome: '",
        x$levels[[2]][x$positive], "'\n",
      sep = ""
    )
    ## cat("\tNegative outcome: '", x$levels[[2]][x$positive%%2+1],"'\n",sep="")
    print(cli::rule())
    printCoefmat(x$coefmat[, c(-2)], na.print="", ...)
    print(cli::rule())
    cat("\n")
    cat("Prevalence:				Prob( outcome+ )\n")
    cat("Test:					Prob( test+ )\n")
    cat("Sensitivity (True positive rate):	Prob( test+ | outcome+ )\n")
    cat("Specificity (True negative rate):	Prob( test- | outcome- )\n")
    cat("Positive predictive value (Precision):	Prob( outcome+ | test+ )\n")
    cat("Negative predictive value:		Prob( outcome- | test- )\n")
    cat("Accuracy:				Prob( correct classification )\n")
    if (x$confint == "exact") {
        cat("Homogeneity/Symmetry:			Prob( outcome+, test- | discordant ), H0: p=0.5 \n")
    } else {
        cat("Homogeneity/Symmetry:			H0: Prob( outcome+ ) - Prob( test+ ), H0: p=0\n")
    }
    cat("\n")
}

##' @export
summary.diagtest <- function(object, ...) {
    object[c("IC", "print", "id", "compare")] <- NULL
    return(object)
}

context("Inference")

set.seed(1)
a1 <- estimate(coef=1, IC=rnorm(10), id=1:10, labels="a1")
a2 <- estimate(coef=2, IC=rnorm(10), id=1:10, labels="a2")
a <- merge(a1, a2)

b1 <- estimate(coef=0.5, IC=rnorm(10), id=1:10, labels="b1")
b2 <- estimate(coef=0.9, IC=rnorm(10), id=1:10, labels="b2")
b <- merge(b1, b2)

test_that("+.estimate", {
  ## check addition: estimate + numeric with dim(estimate)=1
  e1 <- a1 + c(2,4,6)
  e2 <- estimate(a1, \(p) p + c(2,4,6))
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))

  ## check addition: estimate + numeric with dim(estimate)>1
  e1 <- a + 2
  e2 <- estimate(a, \(p) p + 2)
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))
  e1 <- a + c(2,4)
  e2 <- estimate(a, \(p) p + c(2,4))
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))
  # commutative
  e1 <- c(2,4) + a # 2.dim. estimate
  e2 <- estimate(a, \(p) p + c(2,4))
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))
  e1 <- c(2,4) + a1 # 1.dim
  e2 <- estimate(a1, \(p) p + c(2,4))
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))

  # addition with two estimate objects
  e1 <- a1 + a2
  e2 <- estimate(a, cbind(1, 1))
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))
  e1 <- a + b # 2d. estimates
  e2 <- estimate(merge(a, b), rbind(c(1, 0, 1, 0), c(0, 1, 0, 1)))
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))
})

test_that("-.estimate", {
  ## check minus operator
  e1 <- -a
  e2 <- estimate(a, function(p) -p)
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))

  ## estimate + numeric with dim(estimate)=1
  e1 <- a1 - c(2,4,6)
  e2 <- estimate(a1, \(p) p - c(2,4,6))
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))

  ## estimate + numeric with dim(estimate)>1
  e1 <- a - 2
  e2 <- estimate(a, \(p) p - 2)
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))
  e1 <- a - c(2,4)
  e2 <- estimate(a, \(p) p - c(2,4))
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))

  # reverse order of numeric and estimate
  e1 <- c(2,4) - a # 2.dim. estimate
  e2 <- estimate(a, \(p) c(2,4) - p)
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))
  e1 <- c(2,4) - a1 # 1.dim
  e2 <- estimate(a1, \(p) c(2,4) - p)
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))

  # subtraction of two estimate objects
  e1 <- a1 - a2
  e2 <- estimate(a, cbind(1, -1))
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))
  e1 <- a - b # 2d. estimates
  e2 <- estimate(merge(a, b), rbind(c(1, 0, -1, 0), c(0, 1, 0, -1)))
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))
})

test_that("*.estimate", {
  ## check multiplication: estimate + numeric with dim(estimate)=1
  e1 <- a1 * c(2,4,6)
  e2 <- estimate(a1, \(p) p * c(2,4,6))
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))

  ## estimate + numeric with dim(estimate)>1
  e1 <- a * 2
  e2 <- estimate(a, \(p) p * 2)
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))
  e1 <- a * c(2,4)
  e2 <- estimate(a, \(p) p * c(2,4))
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))
  # change order
  e1 <- c(2,4) * a # 2.dim. estimate
  e2 <- estimate(a, \(p) p * c(2,4))
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))
  e1 <- c(2,4) * a1 # 1.dim
  e2 <- estimate(a1, \(p) p * c(2,4))
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))

  # with two estimate objects
  e1 <- a1 * a2
  e2 <- estimate(a, prod)
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))
  e1 <- a * b # 2d. estimates
  e2 <- estimate(merge(a, b), function(p) c(p[1]*p[3], p[2]*p[4]))
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))
})

test_that("/.estimate", {
  ## check division: estimate + numeric with dim(estimate)=1
  e1 <- a1 / c(2,4,6)
  e2 <- estimate(a1, \(p) p / c(2,4,6))
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))

  ## estimate + numeric with dim(estimate)>1
  e1 <- a / 2
  e2 <- estimate(a, \(p) p / 2)
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))
  e1 <- a / c(2,4)
  e2 <- estimate(a, \(p) p / c(2,4))
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))
  # numeric first
  e1 <- c(2,4) / a # 2.dim. estimate
  e2 <- estimate(a, \(p) c(2,4) / p)
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))
  e1 <- c(2,4) / a1 # 1.dim
  e2 <- estimate(a1, \(p) c(2,4) / p)
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))

  # with two estimate objects
  e1 <- a1 / a2
  e2 <- estimate(a, function(p) p[1] / p[2])
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))
  e1 <- a / b # 2d. estimates
  e2 <- estimate(merge(a, b), function(p) c(p[1]/p[3], p[2]/p[4]))
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))
})

test_that("^.estimate", {
  ## check power-func: estimate + numeric with dim(estimate)=1
  e1 <- a1^0.25
  e2 <- estimate(a1, function(p) p**0.25)
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))
  e1 <- a1 / c(2,4,6)
  e2 <- estimate(a1, \(p) p / c(2,4,6))
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))

  ## estimate + numeric with dim(estimate)>1
  e1 <- a ^ 3
  e2 <- estimate(a, \(p) p ^ 3)
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))
  e1 <- a ^ c(2,4)
  e2 <- estimate(a, \(p) p ^ c(2,4))
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))
  # numeric first
  e1 <- c(2,4) ^ a # 2.dim. estimate
  e2 <- estimate(a, \(p) c(2,4) ^ p)
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))
  e1 <- c(2,4) ^ a1 # 1.dim
  e2 <- estimate(a1, \(p) c(2,4) ^ p)
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))

  # addition with two estimate objects
  e1 <- a1 ^ a2
  e2 <- estimate(a, function(p) p[1] ^ p[2])
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))
  e1 <- a ^ b # 2d. estimates
  e2 <- estimate(merge(a, b), function(p) c(p[1]^p[3], p[2]^p[4]))
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))
})

test_that("math functions", {
  e0 <- c(a, b) # merge
  e <- merge(a, b)
  expect_equivalent(e, e0)

  # subset and merge
  ## e1 <- c(a=a["a1"], b=b["b1"])
  e1 <- merge(a["a1"], b["b1"], labels=c("a", "b"))
  expect_true(length(coef(e1)) == 2L)
  expect_equivalent(names(coef(e1)), c("a", "b"))

  # product
  e1 <- prod(e, labels="prod")
  e2 <- estimate(e, function(p) prod(p))
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))

  # sum
  e1 <- sum(e, labels="sum")
  e2 <- estimate(e, function(p) sum(p))
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))

  # inner-prod
  e1 <- a %*% b
  e2 <- estimate(e, function(p) sum(p[1:2]*p[3:4]))
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))
  e1 <- a %*% c(2, 3)
  e2 <- estimate(a, function(p) sum(p[1]*2 + p[2]*3))
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))

  # log
  e1 <- log(e["a1"])
  e2 <- estimate(e, function(p) log(p[1]))
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))
  e1 <- log(e)
  e2 <- estimate(e, function(p) log(p))
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))

  e1 <- log1p(b)
  e2 <- estimate(b, function(p) log1p(p))
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))

  # exp
  e1 <- exp(e["a1"])
  e2 <- estimate(e, function(p) exp(p[1]))
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))
  e1 <- exp(e)
  e2 <- estimate(e, function(p) exp(p))
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))

  e1 <- expm1(b)
  e2 <- estimate(b, function(p) expm1(p))
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))

  # pow, sqrt
  e1 <- sqrt(e)
  e2 <- estimate(e, function(p) sqrt(p))
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))

  # trigonometric
  for (f in c(cos, sin, tan, acos, asin, atan)) {
      e1 <- f(b)
      e2 <- estimate(b, function(p) f(p))
      testthat::expect_equivalent(IC(e1), IC(e2))
      testthat::expect_equivalent(coef(e1), coef(e2))
  }
  e1 <- sin(asin(b))
  testthat::expect_equivalent(IC(e1), IC(b))
  testthat::expect_equivalent(coef(e1), coef(b))

  # hyberbolic
  for (f in c(cosh, sinh, tanh)) {
      e1 <- f(b)
      e2 <- estimate(b, function(p) f(p))
      testthat::expect_equivalent(IC(e1), IC(e2))
      testthat::expect_equivalent(coef(e1), coef(e2))
  }
  e1 <- acosh(cosh(b))
  testthat::expect_equivalent(IC(e1), IC(b))
  testthat::expect_equivalent(coef(e1), coef(b))
  e1 <- asinh(sinh(b))
  testthat::expect_equivalent(IC(e1), IC(b))
  testthat::expect_equivalent(coef(e1), coef(b))
  e1 <- atanh(tanh(b))
  testthat::expect_equivalent(IC(e1), IC(b))
  testthat::expect_equivalent(coef(e1), coef(b))

})

test_that("custom functions", {
  e1 <- logit(expit(a))
  testthat::expect_equivalent(IC(e1), IC(a))
  testthat::expect_equivalent(coef(e1), coef(a))

  e1 <- 2 * log(a["a2"]) * sqrt(b["b2"])
  e2 <- estimate(merge(a,b),
                 function(p) 2 * log(p["a2"]) * sqrt(p["b2"]))
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))

  e1 <- 1 / cos(a) - sqrt(a^2 + 1) + 1
  e2 <- estimate(a,
                 function(p) 1 / cos(p) - sqrt(p^2 + 1) + 1)
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))

  e1 <- odds(b)
  e2 <- estimate(b, odds)
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))
})

context("estimate calculus")

set.seed(1)
# Generate mean-zero random IC matrix (for internal testing)
center_ic <- function(n, p = 1) {
  x <- matrix(rnorm(n * p), n, p)
  scale(x, center = TRUE, scale = FALSE)
}
a1 <- estimate(coef=1, IC=center_ic(10), id=1:10, labels="a1")
a2 <- estimate(coef=2, IC=center_ic(10), id=1:10, labels="a2")
a <- merge(a1, a2)

b1 <- estimate(coef=0.5, IC=center_ic(10), id=1:10, labels="b1")
b2 <- estimate(coef=0.9, IC=center_ic(10), id=1:10, labels="b2")
b <- merge(b1, b2)

test_that("+.estimate", {
  ## check addition: estimate + numeric with dim(estimate)=1
  e1 <- a1 + c(2,4,6)
  e2 <- estimate(a1, \(p) p + c(2,4,6))
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))

  ## check addition: estimate + numeric with dim(estimate)>1
  e1 <- a + 2
  e2 <- estimate(a, \(p) p + 2)
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))
  e1 <- a + c(2,4)
  e2 <- estimate(a, \(p) p + c(2,4))
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))
  # commutative
  e1 <- c(2,4) + a # 2.dim. estimate
  e2 <- estimate(a, \(p) p + c(2,4))
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))
  e1 <- c(2,4) + a1 # 1.dim
  e2 <- estimate(a1, \(p) p + c(2,4))
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))

  # addition with two estimate objects
  e1 <- a1 + a2
  e2 <- estimate(a, cbind(1, 1))
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))
  e1 <- a + b # 2d. estimates
  e2 <- estimate(merge(a, b), rbind(c(1, 0, 1, 0), c(0, 1, 0, 1)))
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))
})

test_that("-.estimate", {
  ## check minus operator
  e1 <- -a
  e2 <- estimate(a, function(p) -p)
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))

  ## estimate + numeric with dim(estimate)=1
  e1 <- a1 - c(2,4,6)
  e2 <- estimate(a1, \(p) p - c(2,4,6))
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))

  ## estimate + numeric with dim(estimate)>1
  e1 <- a - 2
  e2 <- estimate(a, \(p) p - 2)
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))
  e1 <- a - c(2,4)
  e2 <- estimate(a, \(p) p - c(2,4))
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))

  # reverse order of numeric and estimate
  e1 <- c(2,4) - a # 2.dim. estimate
  e2 <- estimate(a, \(p) c(2,4) - p)
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))
  e1 <- c(2,4) - a1 # 1.dim
  e2 <- estimate(a1, \(p) c(2,4) - p)
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))

  # subtraction of two estimate objects
  e1 <- a1 - a2
  e2 <- estimate(a, cbind(1, -1))
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))
  e1 <- a - b # 2d. estimates
  e2 <- estimate(merge(a, b), rbind(c(1, 0, -1, 0), c(0, 1, 0, -1)))
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))
})

test_that("*.estimate", {
  ## check multiplication: estimate + numeric with dim(estimate)=1
  e1 <- a1 * c(2,4,6)
  e2 <- estimate(a1, \(p) p * c(2,4,6))
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))

  ## estimate + numeric with dim(estimate)>1
  e1 <- a * 2
  e2 <- estimate(a, \(p) p * 2)
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))
  e1 <- a * c(2,4)
  e2 <- estimate(a, \(p) p * c(2,4))
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))
  # change order
  e1 <- c(2,4) * a # 2.dim. estimate
  e2 <- estimate(a, \(p) p * c(2,4))
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))
  e1 <- c(2,4) * a1 # 1.dim
  e2 <- estimate(a1, \(p) p * c(2,4))
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))

  # with two estimate objects
  e1 <- a1 * a2
  e2 <- estimate(a, prod)
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))
  e1 <- a * b # 2d. estimates
  e2 <- estimate(merge(a, b), function(p) c(p[1]*p[3], p[2]*p[4]))
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))
})

test_that("/.estimate", {
  ## check division: estimate + numeric with dim(estimate)=1
  e1 <- a1 / c(2,4,6)
  e2 <- estimate(a1, \(p) p / c(2,4,6))
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))

  ## estimate + numeric with dim(estimate)>1
  e1 <- a / 2
  e2 <- estimate(a, \(p) p / 2)
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))
  e1 <- a / c(2,4)
  e2 <- estimate(a, \(p) p / c(2,4))
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))
  # numeric first
  e1 <- c(2,4) / a # 2.dim. estimate
  e2 <- estimate(a, \(p) c(2,4) / p)
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))
  e1 <- c(2,4) / a1 # 1.dim
  e2 <- estimate(a1, \(p) c(2,4) / p)
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))

  # with two estimate objects
  e1 <- a1 / a2
  e2 <- estimate(a, function(p) p[1] / p[2])
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))
  e1 <- a / b # 2d. estimates
  e2 <- estimate(merge(a, b), function(p) c(p[1]/p[3], p[2]/p[4]))
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))
})

test_that("^.estimate", {
  ## check power-func: estimate + numeric with dim(estimate)=1
  e1 <- a1^0.25
  e2 <- estimate(a1, function(p) p**0.25)
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))
  e1 <- a1 / c(2,4,6)
  e2 <- estimate(a1, \(p) p / c(2,4,6))
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))

  ## estimate + numeric with dim(estimate)>1
  e1 <- a ^ 3
  e2 <- estimate(a, \(p) p ^ 3)
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))
  e1 <- a ^ c(2,4)
  e2 <- estimate(a, \(p) p ^ c(2,4))
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))
  # numeric first
  e1 <- c(2,4) ^ a # 2.dim. estimate
  e2 <- estimate(a, \(p) c(2,4) ^ p)
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))
  e1 <- c(2,4) ^ a1 # 1.dim
  e2 <- estimate(a1, \(p) c(2,4) ^ p)
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))

  # addition with two estimate objects
  e1 <- a1 ^ a2
  e2 <- estimate(a, function(p) p[1] ^ p[2])
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))
  e1 <- a ^ b # 2d. estimates
  e2 <- estimate(merge(a, b), function(p) c(p[1]^p[3], p[2]^p[4]))
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))
})

test_that("math functions", {
  e0 <- c(a, b) # merge
  e <- merge(a, b)
  expect_equivalent(e, e0)

  # subset and merge
  ## e1 <- c(a=a["a1"], b=b["b1"])
  e1 <- merge(a["a1"], b["b1"], labels=c("a", "b"))
  expect_true(length(coef(e1)) == 2L)
  expect_equivalent(names(coef(e1)), c("a", "b"))

  # product
  e1 <- prod(e, labels="prod")
  e2 <- estimate(e, function(p) prod(p))
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))

  # sum
  e1 <- sum(e, labels="sum")
  e2 <- estimate(e, function(p) sum(p))
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))

  # inner-prod
  e1 <- a %*% b
  e2 <- estimate(e, function(p) sum(p[1:2]*p[3:4]))
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))
  e1 <- a %*% c(2, 3)
  e2 <- estimate(a, function(p) sum(p[1]*2 + p[2]*3))
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))

  # log
  e1 <- log(e["a1"])
  e2 <- estimate(e, function(p) log(p[1]))
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))
  e1 <- log(e)
  e2 <- estimate(e, function(p) log(p))
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))

  e1 <- log1p(b)
  e2 <- estimate(b, function(p) log1p(p))
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))

  # exp
  e1 <- exp(e["a1"])
  e2 <- estimate(e, function(p) exp(p[1]))
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))
  e1 <- exp(e)
  e2 <- estimate(e, function(p) exp(p))
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))

  e1 <- expm1(b)
  e2 <- estimate(b, function(p) expm1(p))
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))

  # pow, sqrt
  e1 <- sqrt(e)
  e2 <- estimate(e, function(p) sqrt(p))
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))

  # trigonometric
  for (f in c(cos, sin, tan, acos, asin, atan)) {
      e1 <- f(b)
      e2 <- estimate(b, function(p) f(p))
      testthat::expect_equivalent(IC(e1), IC(e2))
      testthat::expect_equivalent(coef(e1), coef(e2))
  }
  e1 <- sin(asin(b))
  testthat::expect_equivalent(IC(e1), IC(b))
  testthat::expect_equivalent(coef(e1), coef(b))

  # hyberbolic
  for (f in c(cosh, sinh, tanh)) {
      e1 <- f(b)
      e2 <- estimate(b, function(p) f(p))
      testthat::expect_equivalent(IC(e1), IC(e2))
      testthat::expect_equivalent(coef(e1), coef(e2))
  }
  e1 <- acosh(cosh(b))
  testthat::expect_equivalent(IC(e1), IC(b))
  testthat::expect_equivalent(coef(e1), coef(b))
  e1 <- asinh(sinh(b))
  testthat::expect_equivalent(IC(e1), IC(b))
  testthat::expect_equivalent(coef(e1), coef(b))
  e1 <- atanh(tanh(b))
  testthat::expect_equivalent(IC(e1), IC(b))
  testthat::expect_equivalent(coef(e1), coef(b))

})

test_that("custom functions", {
  e1 <- logit(expit(a))
  testthat::expect_equivalent(IC(e1), IC(a))
  testthat::expect_equivalent(coef(e1), coef(a))

  e1 <- 2 * log(a["a2"]) * sqrt(b["b2"])
  e2 <- estimate(merge(a,b),
                 function(p) 2 * log(p["a2"]) * sqrt(p["b2"]))
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))

  e1 <- 1 / cos(a) - sqrt(a^2 + 1) + 1
  e2 <- estimate(a,
                 function(p) 1 / cos(p) - sqrt(p^2 + 1) + 1)
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))

  e1 <- odds(b)
  e2 <- estimate(b, odds)
  testthat::expect_equivalent(IC(e1), IC(e2))
  testthat::expect_equivalent(coef(e1), coef(e2))
})

## ---------------------------------------------------------------------------
## operator_grad shape correctness vs numerical jacobian (Issue C)
## ---------------------------------------------------------------------------
make_est <- function(coefs, seed) {
  set.seed(seed)
  k <- length(coefs)
  ic <- center_ic(20, k)
  nm <- if (is.null(names(coefs))) paste0("p", seq_along(coefs)) else names(coefs)
  names(coefs) <- nm
  estimate(coef = coefs, IC = ic, id = 1:20)
}

ops <- list(
  "+" = `+`,
  "-" = `-`,
  "*" = `*`,
  "/" = `/`,
  "^" = `^`
)

shapes <- list(
  c(1, 1),
  c(1, 3),
  c(3, 1),
  c(3, 3)
)

for (opname in names(ops)) {
  for (sh in shapes) {
    nx <- sh[1]; ny <- sh[2]
    label <- sprintf("operator_grad: %s with shape (%d,%d)", opname, nx, ny)
    test_that(label, {
      ## Use strictly positive coefs to keep ^ and / well-defined.
      a <- make_est(setNames(seq_len(nx) + 1.0, paste0("a", seq_len(nx))),
                    seed = 100 + nx)
      b <- make_est(setNames(seq_len(ny) + 0.5, paste0("b", seq_len(ny))),
                    seed = 200 + ny)
      op_est <- ops[[opname]]
      op_num <- switch(opname,
                       "+" = function(p) p[seq_len(nx)] + p[nx + seq_len(ny)],
                       "-" = function(p) p[seq_len(nx)] - p[nx + seq_len(ny)],
                       "*" = function(p) p[seq_len(nx)] * p[nx + seq_len(ny)],
                       "/" = function(p) p[seq_len(nx)] / p[nx + seq_len(ny)],
                       "^" = function(p) p[seq_len(nx)] ^ p[nx + seq_len(ny)])
      ## Analytic via the operator
      res <- op_est(a, b)
      res_coef <- coef(res)
      ## Numerical jacobian
      pp <- c(coef(a), coef(b))
      Jnum <- numDeriv::jacobian(op_num, pp)
      ## Reconstruct analytic D from IC: IC_res = IC_ab %*% t(D)
      ## We instead validate by comparing variance: an incorrect D shape
      ## would produce wrong vcov dims or wrong values.
      ## Build merged IC manually:
      ic_ab <- cbind(IC(a), IC(b))
      V_expected <- Jnum %*% var_ic(ic_ab) %*% t(Jnum)
      V_actual   <- vcov(res)
      expect_equal(dim(V_actual), dim(V_expected))
      expect_equivalent(V_actual, V_expected, tolerance = 1e-5)
      expect_equal(length(res_coef), max(nx, ny))
    })
  }
}

ordreg_threshold <- function(theta) {
    v <- theta[1]
    if (length(theta)>1) v <- cumsum(c(v,exp(theta[seq(length(theta)-1L)+1L])))
    return(v)
}

ordreg_ithreshold <- function(v) {
    theta <- v[1]
    if (length(v)>1) theta <- c(theta,log(-rev(diff(rev(v)))))
    return(theta)
}

ordreg_dthreshold <- function(theta) {
    K <- length(theta)+1
    Da <- matrix(0,K,K-1)
    Da[seq(K-1),1L] <- 1L
    for (i in seq_len(K-2)+1) Da[seq(i,K-1),i] <- exp(theta[i])
    Da
}

##' Ordinal regression models
##'
##' @title Univariate cumulative link regression models
##' @param formula formula
##' @param data data.frame
##' @param offset offset
##' @param family family (default proportional odds)
##' @param start optional starting values
##' @param fast If TRUE standard errors etc. will not be calculated
##' @param ... Additional arguments to lower level functions
##' @export
##' @author Klaus K. Holst
##' @examples
##' m <- lvm(y~x)
##' ordinal(m,K=3) <- ~y
##' d <- sim(m,100)
##' e <- ordreg(y~x,d)
ordreg <- function(formula,data=parent.frame(),offset,family=stats::binomial("probit"),start,fast=FALSE,...) {
    y <- ordered(model.frame(update(formula,.~0),data)[,1])
    lev <- levels(y)
    X <- model.matrix(update(formula,.~.+1),data=data)[,-1,drop=FALSE]
    up <- new.env()
    assign("h",family$linkinv,envir=up)
    assign("dh",family$mu.eta,envir=up)
    assign("y",as.numeric(y),envir=up)
    assign("X",X,envir=up)
    assign("K",nlevels(y),envir=up)
    assign("n",length(y),envir=up)
    assign("p",NCOL(X),envir=up)
    assign("threshold", function(theta,K) ordreg_threshold(theta[seq(K-1)]), envir=up)
    assign("dthreshold",function(theta,K) ordreg_dthreshold(theta[seq(K-1)]), envir=up)
    ff <- function(theta) -ordreg_logL(theta,up)
    gg <- function(theta) -ordreg_score(theta,up)
    if (missing(start)) start <- with(up,c(rep(-1,up$K-1),rep(0,p)))
    op <- nlminb(start,ff,gg)
    cc <- op$par;
    if (fast) return(structure(cc,threshold=up$threshold(cc,up$K))) ##,up$K)))
    nn <- c(paste(lev[-length(lev)], lev[-1L], sep = "|"),
                   colnames(X))
    I <- -ordreg_hessian(cc,up)
    names(cc) <- nn
    dimnames(I) <- list(nn,nn)
    res <- list(vcov=solve(I),coef=cc,call=match.call(),up=up,opt=op,levels=lev)
    structure(res,class="ordreg")
}

##' @export
print.ordreg <- function(x,...) {
    cat("Call:\n"); print(x$call)
    cat("\nParameter Estimates:\n")
    print(x$coef)
}

##' @export
summary.ordreg <- function(object,alpha=0.95,...) {
    res <- cbind(coef(object),diag(vcov(object))^.5)
    pp <- 1-(1-alpha)/2
    qq <- qnorm(pp)
    res <- cbind(res,res[,1]-res[,2]*qq,res[,1]+res[,2]*qq,2*(1-pnorm(abs(res[,1])/res[,2])))
    colnames(res) <- c("Estimate","Std.Err",paste0(round(c(1-pp,pp)*1000)/10,"%"),"P-value")
    res <- list(coef=res,logLik=logLik(object),AIC=AIC(object))
    class(res) <- "summary.ordreg"
    return(res)
}

##' @export
print.summary.ordreg <- function(x,alpha=0.95,...) {
    cat("AIC: ", x$AIC, "\n\n")
    print(x$coef)
    cat("\n")
}

##' @export
score.ordreg <- function(x,p=coef(x),indiv=FALSE,...) {
    ordreg_score(p,x$up)
    if (!indiv) return(colSums(x$up$score))
    x$up$score
}

##' @export
logLik.ordreg <- function(object,p=coef(object),indiv=FALSE,...) {
    ordreg_logL(p,object$up)
    res <- log(object$up$pr)
    if (!indiv) res <- sum(res)
    structure(res,nall=length(object$up$pr),nobs=object$up$pr,df=length(p),class="logLik")
}

##' @export
coef.ordreg <- function(object,...) object$coef

##' @export
vcov.ordreg <- function(object,...) object$vcov

ordreg_logL <- function(theta,env,indiv=FALSE,...) {
    if (length(theta)!=with(env,p+K-1)) stop("Wrong dimension")
    env$theta <- theta
    if (env$p>0) beta <- with(env,theta[seq(p)+K-1])
    alpha <- with(env, threshold(theta,K))
    env$alpha <- alpha
    env$beta <- beta
    if (env$p>0) eta <- env$X%*%beta else eta <- cbind(rep(0,env$n))
    env$lp <- kronecker(-eta,rbind(alpha),"+")
    F <- with(env,h(lp))
    Pr <- cbind(F,1)-cbind(0,F)
    pr <- Pr[with(env,cbind(seq(n),as.numeric(y)))]
    env$pr <- pr
    sum(log(pr))
}

ordreg_score <- function(theta,env,...) {
    if (!identical(theta,env$theta)) ordreg_logL(theta,env)
    Da <- with(env,dthreshold(theta,K))
    dF <- with(env, cbind(dh(lp),0))
    idx1 <- with(env,which(as.numeric(y)==1))
    S1 <- cbind(Da[as.numeric(env$y),,drop=FALSE],-env$X)
    S1 <- dF[with(env,cbind(seq(n),as.numeric(y)))]*S1
    y2 <- env$y-1; y2[idx1] <- env$K
    S2 <- cbind(Da[y2,,drop=FALSE],-env$X)
    S2 <- dF[cbind(seq(env$n),y2)]*S2
    env$score <- 1/env$pr*(S1-S2)
    colSums(env$score)
}
ordreg_hessian <- function(theta,env,...) {
    numDeriv::jacobian(function(p) ordreg_score(p,env,...),theta,...)
}

##' @export
predict.ordreg <- function(object,p=coef(object),type=c("prob","cumulative"),...) {
    env <- object$up
    env$theta <- p
    if (env$p>0) beta <- with(env,theta[seq(p)+K-1])
    alpha <- with(env, threshold(theta,K))
    env$alpha <- alpha
    env$beta <- beta
    if (env$p>0) eta <- env$X%*%beta else eta <- cbind(rep(0,env$n))
    env$lp <- kronecker(-eta,rbind(alpha),"+")
    F <- with(env,h(lp))
    if (tolower(type)[1]=="cumulative") return(F)            
    Pr <- cbind(F,1)-cbind(0,F)
    colnames(Pr) <- object$levels
    return(Pr)
}

'.onLoad' <- function(libname, pkgname="lava") {
    addhook("heavytail.init.hook","init.hooks")
    addhook("glm.estimate.hook","estimate.hooks")
    addhook("ordinal.estimate.hook","estimate.hooks")
    addhook("cluster.post.hook","post.hooks")
    addhook("ordinal.sim.hook","sim.hooks")
    addhook("color.ordinal","color.hooks")
    addhook("ordinal.remove.hook","remove.hooks")
}

'.onAttach' <- function(libname, pkgname="lava") {
    #desc <- utils::packageDescription(pkgname)
    #packageStartupMessage(desc$Package, " version ",desc$Version)
    lava.options(cluster.index=versioncheck("mets",c(0,2,7)),
                 tobit=versioncheck("lava.tobit",c(0,5)))
}

##' Estimation of parameters in a Latent Variable Model (lvm)
##'
##' Estimate parameters. MLE, IV or user-defined estimator.
##'
##' A list of parameters controlling the estimation and optimization procedures
##' is parsed via the \code{control} argument. By default Maximum Likelihood is
##' used assuming multivariate normal distributed measurement errors. A list
##' with one or more of the following elements is expected:
##'
##' \describe{
##' \item{start:}{Starting value. The order of the parameters can be shown by
##' calling \code{coef} (with \code{mean=TRUE}) on the \code{lvm}-object or with
##' \code{plot(..., labels=TRUE)}. Note that this requires a check that it is
##' actual the model being estimated, as \code{estimate} might add additional
##' restriction to the model, e.g. through the \code{fix} and \code{exo.fix}
##' arguments. The \code{lvm}-object of a fitted model can be extracted with the
##' \code{Model}-function.}
##'
##' \item{starterfun:}{Starter-function with syntax
##' \code{function(lvm, S, mu)}.  Three builtin functions are available:
##' \code{startvalues}, \code{startvalues0}, \code{startvalues1}, ...}
##'
##' \item{estimator:}{ String defining which estimator to use (Defaults to
##' ``\code{gaussian}'')}
##'
##' \item{meanstructure}{Logical variable indicating
##' whether to fit model with meanstructure.}
##'
##' \item{method:}{ String pointing to
##' alternative optimizer (e.g. \code{optim} to use simulated annealing).}
##'
##' \item{control:}{ Parameters passed to the optimizer (default
##' \code{stats::nlminb}).}
##'
##' \item{tol:}{ Tolerance of optimization constraints on lower limit of
##' variance parameters.  } }
##'
##' @param x \code{lvm}-object
##' @param data \code{data.frame}
##' @param estimator String defining the estimator (see details below)
##' @param control control/optimization parameters (see details below)
##' @param missing Logical variable indiciating how to treat missing data.
##' Setting to FALSE leads to complete case analysis. In the other case
##' likelihood based inference is obtained by integrating out the missing data
##' under assumption the assumption that data is missing at random (MAR).
##' @param weights Optional weights to used by the chosen estimator.
##' @param weightsname Weights names (variable names of the model) in case
##' \code{weights} was given as a vector of column names of \code{data}
##' @param data2 Optional additional dataset used by the chosen
##' estimator.
##' @param id Vector (or name of column in \code{data}) that identifies
##' correlated groups of observations in the data leading to variance estimates
##' based on a sandwich estimator
##' @param fix Logical variable indicating whether parameter restriction
##' automatically should be imposed (e.g. intercepts of latent variables set to
##' 0 and at least one regression parameter of each measurement model fixed to
##' ensure identifiability.)
##' @param index For internal use only
##' @param graph For internal use only
##' @param messages Control how much information should be
##' printed during estimation (0: none)
##' @param quick If TRUE the parameter estimates are calculated but all
##' additional information such as standard errors are skipped
##' @param method Optimization method
##' @param param set parametrization (see \code{help(lava.options)})
##' @param cluster Obsolete. Alias for 'id'.
##' @param p Evaluate model in parameter 'p' (no optimization)
##' @param ... Additional arguments to be passed to lower-level functions
##' @return A \code{lvmfit}-object.
##' @author Klaus K. Holst
##' @seealso estimate.default score, information
##' @keywords models regression
##' @export
##' @method estimate lvm
##' @examples
##' dd <- read.table(header=TRUE,
##' text="x1 x2 x3
##'  0.0 -0.5 -2.5
##' -0.5 -2.0  0.0
##'  1.0  1.5  1.0
##'  0.0  0.5  0.0
##' -2.5 -1.5 -1.0")
##' e <- estimate(lvm(c(x1,x2,x3)~u),dd)
##'
##' ## Simulation example
##' m <- lvm(list(y~v1+v2+v3+v4,c(v1,v2,v3,v4)~x))
##' covariance(m) <- v1~v2+v3+v4
##' dd <- sim(m,10000) ## Simulate 10000 observations from model
##' e <- estimate(m, dd) ## Estimate parameters
##' e
##'
##' ## Using just sufficient statistics
##' n <- nrow(dd)
##' e0 <- estimate(m,data=list(S=cov(dd)*(n-1)/n,mu=colMeans(dd),n=n))
##' rm(dd)
##'
##' ## Multiple group analysis
##' m <- lvm()
##' regression(m) <- c(y1,y2,y3)~u
##' regression(m) <- u~x
##' d1 <- sim(m,100,p=c("u,u"=1,"u~x"=1))
##' d2 <- sim(m,100,p=c("u,u"=2,"u~x"=-1))
##'
##' mm <- baptize(m)
##' regression(mm,u~x) <- NA
##' covariance(mm,~u) <- NA
##' intercept(mm,~u) <- NA
##' ee <- estimate(list(mm,mm),list(d1,d2))
##'
##' ## Missing data
##' d0 <- makemissing(d1,cols=1:2)
##' e0 <- estimate(m,d0,missing=TRUE)
##' e0
`estimate.lvm` <-
    function(x, data=parent.frame(),
      estimator=NULL,
      control=list(),
      missing=FALSE,
      weights, weightsname,
      data2,
      id,
      fix,
      index=!quick,
      graph=FALSE,
      messages=lava.options()$messages,
      quick=FALSE,
      method,
      param,
      cluster,
      p,
      ...) {
        cl <- match.call()
        if (!base::missing(param)) {
            oldparam <- lava.options()$param
            lava.options(param=param)
            on.exit(lava.options(param=oldparam))
        }
        if (!base::missing(method)) {
            control["method"] <- list(method)
        }

        Optim <- list(
            iter.max=lava.options()$iter.max,
            trace=ifelse(lava.options()$debug,3,0),
            gamma=lava.options()$gamma,
            gamma2=1,
            ngamma=lava.options()$ngamma,
            backtrack=lava.options()$backtrack,
            lambda=0.05,
            abs.tol=1e-9,
            epsilon=1e-10,
            delta=1e-10,
            rel.tol=1e-10,
            S.tol=1e-5,
            stabil=FALSE,
            start=NULL,
            constrain=lava.options()$constrain,
            method=NULL,
            starterfun="startvalues0",
            information="E",
            meanstructure=TRUE,
            sparse=FALSE,
            tol=lava.options()$tol)

        defopt <- lava.options()[]
        defopt <- defopt[intersect(names(defopt),names(Optim))]
        Optim[names(defopt)] <- defopt
        if (length(control)>0) {
            Optim[names(control)] <- control
        }
        if (is.environment(data)) {
            innames <- intersect(ls(envir=data),vars(x))
            data <- as.data.frame(lapply(innames,function(x) get(x,envir=data)))
            names(data) <- innames
        }
        if (length(exogenous(x)>0)) {
            catx <- categorical2dummy(x,data)
            x <- catx$x; data <- catx$data
        }

        if (!lava.options()$exogenous) exogenous(x) <- NULL

        redvar <- intersect(intersect(parlabels(x),latent(x)),colnames(data))
        if (length(redvar)>0)
            warning(paste("Latent variable exists in dataset",redvar))
        ## Random-slopes:
        xfix <- setdiff(colnames(data)[(colnames(data)%in%parlabels(x,exo=TRUE))],latent(x))
        if (base::missing(fix)) {
            fix <- ifelse(length(xfix)>0,FALSE,TRUE)
        }
        Debug(list("start=",Optim$start))

        if (!base::missing(cluster)) id <- cluster

        ## Weights...
        if (!base::missing(weights)) {
            if (is.character(weights)) {
                weights <- data[,weights,drop=FALSE]
                if (!base::missing(weightsname)) {
                    colnames(weights) <- weightsname
                } else {
                    yvar <- index(x)$endogenous
                    nw <- seq_len(min(length(yvar),ncol(weights)))
                    colnames(weights)[nw] <- yvar[nw]
                }
            }
            weights <- cbind(weights)
        } else {
            weights <- NULL
        }
        if (!base::missing(data2)) {
            if (is.character(data2)) {
                data2 <- data[,data2]
            }
        } else {
            data2 <- NULL
        }
        ## Correlated clusters...
        if (!base::missing(id)) {
            if (is.character(id)) {
                id <- data[,id]
            }
        } else {
            id <- NULL
        }

        Debug("procdata")
        val <- try({
            dd <- procdata.lvm(x,data=data,missing=missing)
            S <- dd$S; mu <- dd$mu; n <- dd$n
            var.missing <- setdiff(vars(x),colnames(S))
        }, silent=TRUE)
        if (inherits(val,"try-error")) {
            var.missing <- setdiff(vars(x),colnames(data))
            S <- NULL; mu <- NULL; n <- nrow(data)
        }

        ##  if (fix) {
        if (length(var.missing)>0) {## Convert to latent:
            new.lat <- setdiff(var.missing,latent(x))
            if (length(new.lat)>0)
                x <- latent(x, new.lat)
        }
        ##}

        ## Run hooks (additional lava plugins)
        myhooks <- gethook()
        for (f in myhooks) {
            res <- do.call(f, list(x=x,data=data,weights=weights,data2=data2,estimator=estimator,optim=Optim))
            if (!is.null(res$x)) x <- res$x
            if (!is.null(res$data)) data <- res$data
            if (!is.null(res$weights)) weights <- res$weights
            if (!is.null(res$data2)) data2 <- res$data2
            if (!is.null(res$optim)) Optim <- res$optim
            if (!is.null(res$estimator)) estimator <- res$estimator
            rm(res)
        }
        if (is.null(estimator)) {
            if (!missing(weights) && !is.null(weights)) {
                estimator <- "normal"
            } else estimator <- "gaussian"
        }
        checkestimator <- function(x,...) {
            ffname <- paste0(x,c("_objective","_gradient"),".lvm")
            exists(ffname[1])||exists(ffname[2])
        }
        if (!checkestimator(estimator)) { ## Try down/up-case version
            estimator <- tolower(estimator)
            if (!checkestimator(estimator)) {
                estimator <- toupper(estimator)
            }
        }
        ObjectiveFun  <- paste0(estimator, "_objective", ".lvm")
        GradFun  <- paste0(estimator, "_gradient", ".lvm")
        if (!exists(ObjectiveFun) & !exists(GradFun))
            stop("Unknown estimator.")


        Method <-  paste0(estimator, "_method", ".lvm")
        if (!exists(Method)) {
            Method <- "nlminb1"
        } else {
            Method <- get(Method)
        }
        NoOptim <- "method"%in%names(control) && is.null(control$method)
        if (is.null(Optim$method) && !(NoOptim)) {
            Optim$method <- if (missing && Method!="nlminb0") "nlminb1" else Method
        }


        if (index) {
            ## Proces data and setup some matrices
            x <- fixsome(x, measurement.fix=fix, S=S, mu=mu, n=n,debug=messages>1)
            if (messages>1)
                message("Reindexing model...\n")
            if (length(xfix)>0) {
                index(x) <- reindex(x,sparse=Optim$sparse,zeroones=TRUE,deriv=TRUE)
            } else {
                x <- updatelvm(x,sparse=Optim$sparse,zeroones=TRUE,deriv=TRUE,mean=TRUE)
            }
        }
        if (is.null(estimator) || estimator==FALSE) {
            return(x)
        }

        if (length(index(x)$endogenous)==0) stop("No observed outcome variables. Check variable names in model and data.")
        if (!Optim$meanstructure) {
            mu <- NULL
        }

        nparall <- index(x)$npar + ifelse(Optim$meanstructure, index(x)$npar.mean+index(x)$npar.ex,0)
        ## Get starting values
        if (!missing(p)) {
            start <- p
            Optim$start <- p
        } else {
          myparnames <- coef(x,mean=TRUE)
          paragree <- FALSE
          paragree_pos <- c()
          if (!is.null(Optim$start)) {
            paragree <- myparnames%in%names(Optim$start)
            paragree_pos <- matrix(0, sum(paragree), 2)
            count <- 0
            for (i in which(paragree)) {
              count <- count + 1
              paragree_pos[count,] <-  c(i, which(myparnames[i]==names(Optim$start))[1])
            }
          }
          ## If starting values are given for all parameters and not named, then
          ## we just keep the provided starting values as they are.
          ## Otherwise, calculate starting values and place user-provided
          ## starting values in the right positions
          if (! (length(Optim$start)==length(myparnames) & sum(paragree)==0)) {
            if (is.null(Optim$starterfun) && lava.options()$param!="relative")
              Optim$starterfun <- startvalues0
            start <- suppressWarnings(do.call(Optim$starterfun, list(x=x,S=S,mu=mu,debug=lava.options()$debug,
                                                                     messages=messages,data=data,...)))
            if (!is.null(x$expar) && length(start)<nparall) {
              ii <- which(index(x)$e1==1)
              start <- c(start, structure(unlist(x$expar[ii]), names=names(x$expar)[ii]))
            }
            for (i in seq_len(NROW(paragree_pos))) {
              start[paragree_pos[i, 1]] <- Optim$start[paragree_pos[i, 2]]
            }
            Optim$start <- start
          }
        }
        ## Missing data
        if (missing) {
            return(estimate.MAR(x=x,data=data,fix=fix,control=Optim,debug=lava.options()$debug,
                                messages=messages,estimator=estimator,weights=weights,data2=data2,cluster=id,...))
        }
        coefname <- coef(x,mean=Optim$meanstructure,fix=FALSE);
        names(Optim$start) <- coefname

        ## Non-linear parameter constraints involving observed variables? (e.g. nonlinear regression)
        constr <- lapply(constrain(x), function(z)(attributes(z)$args))
        xconstrain <- intersect(unlist(constr), manifest(x))
        xconstrainM <- TRUE
        XconstrStdOpt <- TRUE
        if (length(xconstrain)>0) {
            constrainM <- names(constr)%in%unlist(x$mean)
            for (i in seq_len(length(constr))) {
                if (!constrainM[i]) {
                    if (any(constr[[i]]%in%xconstrain)) {
                        xconstrainM <- FALSE
                        break;
                    }
                }
            }
            if (xconstrainM & ( (is.null(control$method) || Optim$method=="nlminb0") & (lava.options()$test & estimator=="gaussian") ) ) {
                XconstrStdOpt <- FALSE
                Optim$method <- "nlminb0"
                if (is.null(control$constrain)) control$constrain <- TRUE
            }
        }

        ## Setup optimization constraints
        lowmin <- -Inf
        lower <- rep(lowmin,length(Optim$start))
        if (length(Optim$constrain)==1 & Optim$constrain)
            lower[variances(x)+index(x)$npar.mean] <- Optim$tol
        if (any(Optim$constrain)) {
            if (length(Optim$constrain)!=length(lower))
                constrained <- is.finite(lower)
            else
                constrained <- Optim$constrain
            lower[] <- -Inf
            Optim$constrain <- TRUE
            constrained <- which(constrained)
            nn <- names(Optim$start)
            CS <- Optim$start[constrained]
            CS[CS<0] <- 0.01
            Optim$start[constrained] <- log(CS)
            names(Optim$start) <- nn
        }
        ## Fix problems with starting values?
        Optim$start[is.nan(unlist(Optim$start))] <- 0

        ObjectiveFun  <- paste0(estimator, "_objective", ".lvm")
        GradFun  <- paste0(estimator, "_gradient", ".lvm")
        if (!exists(ObjectiveFun) & !exists(GradFun)) stop("Unknown estimator.")

        InformationFun <- paste0(estimator, "_hessian", ".lvm")

        mymodel <- x
        myclass <- "lvmfit"

        ## Random slopes?
        if (length(xfix)>0 | (length(xconstrain)>0 & XconstrStdOpt | !lava.options()$test)) { ## Yes
            x0 <- x

            if (length(xfix)>0) {
                myclass <- c("lvmfit.randomslope",myclass)
                nrow <- length(vars(x))
                xpos <- lapply(xfix,function(y) which(regfix(x)$labels==y))
                colpos <- lapply(xpos, function(y) ceiling(y/nrow))
                rowpos <- lapply(xpos, function(y) (y-1)%%nrow+1)
                myfix <- list(var=xfix, col=colpos, row=rowpos)
                x0 <- x
                for (i in seq_along(myfix$var))
                    for (j in seq_len(length(myfix$col[[i]])))
                        regfix(x0, from=vars(x0)[myfix$row[[i]][j]],
                               to=vars(x0)[myfix$col[[i]][j]]) <-
                            colMeans(data[,myfix$var[[i]],drop=FALSE])
                x0 <- updatelvm(x0,zeroones=TRUE,deriv=TRUE)
                x <- x0
                ## Alter start-values/constraints:
                new.par.idx <- which(coef(mymodel,mean=TRUE,fix=FALSE)%in%coef(x0,mean=TRUE,fix=FALSE))
                if (length(Optim$start)>length(new.par.idx))
                    Optim$start <- Optim$start[new.par.idx]
                lower <- lower[new.par.idx]
                if (Optim$constrain) {
                    constrained <- match(constrained,new.par.idx)
                }
            }
            mydata <- as.matrix(data[,manifest(x0)])

            myObj <- function(pp) {
                if (Optim$constrain) {
                    pp[constrained] <- exp(pp[constrained])
                }
                myfun <- function(ii) {
                    if (length(xfix)>0)
                        for (i in seq_along(myfix$var)) {
                            x0$fix[cbind(rowpos[[i]],colpos[[i]])] <- index(x0)$A[cbind(rowpos[[i]],colpos[[i]])] <- data[ii,xfix[i]]
                        }
                    if (is.list(data2)) {
                        res <- do.call(ObjectiveFun, list(x=x0, p=pp, data=mydata[ii,], n=1, weights=weights[ii,], data2=data2[ii,]))
                    } else {
                        res <- do.call(ObjectiveFun, list(x=x0, p=pp, data=mydata[ii,], n=1, weights=weights[ii,], data2=data2))
                    }
                    return(res)
                }
                sum(sapply(seq_len(nrow(data)),myfun))
            }

            myGrad <- function(pp) {
                if (Optim$constrain) {
                    pp[constrained] <- exp(pp[constrained])
                }
                myfun <- function(ii) {
                    if (length(xfix)>0)
                        for (i in seq_along(myfix$var)) {
                            x0$fix[cbind(rowpos[[i]],colpos[[i]])] <- index(x0)$A[cbind(rowpos[[i]],colpos[[i]])] <- data[ii,xfix[i]]
                        }
                    if (is.list(data2)) {
                        rr <- do.call(GradFun, list(x=x0, p=pp, data=mydata[ii,,drop=FALSE], n=1, weights=weights[ii,], data2=data2))
                    } else
                    {
                        rr <- do.call(GradFun, list(x=x0, p=pp, data=mydata[ii,,drop=FALSE], n=1, weights=weights[ii,], data2=data2[ii,]))
                    }
                    return(rr)
                }
                ss <- rowSums(rbind(sapply(seq_len(nrow(data)),myfun)))
                if (Optim$constrain) {
                    ss[constrained] <- ss[constrained]*pp[constrained]
                }
                return(ss)
            }


            myInfo <- function(pp) {
                myfun <- function(ii) {
                    if (length(xfix)>0)
                        for (i in seq_along(myfix$var)) {
                            x0$fix[cbind(rowpos[[i]],colpos[[i]])] <- index(x0)$A[cbind(rowpos[[i]],colpos[[i]])] <- data[ii,xfix[i]]
                        }
                    if (is.list(data2)) {
                        res <- do.call(InformationFun, list(p=pp, obj=myObj, x=x0, data=data[ii,],
                                                            n=1, weights=weights[ii,], data2=data2))
                    } else {
                        res <- do.call(InformationFun, list(p=pp, obj=myObj, x=x0, data=data[ii,],
                                                            n=1, weights=weights[ii,], data2=data2[ii,]))
                    }
                    return(res)
                }
                L <- lapply(seq_len(nrow(data)),function(x) myfun(x))
                val <- apply(array(unlist(L),dim=c(length(pp),length(pp),nrow(data))),c(1,2),sum)
                if (!is.null(attributes(L[[1]])$grad)) {
                    attributes(val)$grad <- colSums (
                                       matrix( unlist(lapply(L,function(i) attributes(i)$grad)) , ncol=length(pp), byrow=TRUE)
                                   )
                }
                return(val)
            }

        } else { ## No, standard model

            ## Non-linear parameter constraints involving observed variables? (e.g. nonlinear regression)
            xconstrain <- c()
            for (i in seq_len(length(constrain(x)))) {
                z <- constrain(x)[[i]]
                xx <- intersect(attributes(z)$args,manifest(x))
                if (length(xx)>0) {
                    warg <- setdiff(attributes(z)$args,xx)
                    wargidx <- which(attributes(z)$args%in%warg)
                    exoidx <- which(attributes(z)$args%in%xx)
                    parname <- names(constrain(x))[i]
                    y <- names(which(unlist(lapply(intercept(x),function(x) x==parname))))
                    el <- list(i,y,parname,xx,exoidx,warg,wargidx,z)
                    names(el) <- c("idx","endo","parname","exo","exoidx","warg","wargidx","func")
                    xconstrain <- c(xconstrain,list(el))
                }
            }
            yconstrain <- unlist(lapply(xconstrain,function(x) x$endo))
            iconstrain <- unlist(lapply(xconstrain,function(x) x$idx))

            MkOffset <- function(pp,grad=FALSE) {
                if (length(xconstrain)>0) {
                    Mu <- matrix(0,nrow(data),length(vars(x))); colnames(Mu) <- vars(x)
                    M <- modelVar(x,p=pp,data=data)
                    M$parval <- c(M$parval,  x$mean[unlist(lapply(x$mean,is.numeric))])
                    for (i in seq_len(length(xconstrain))) {
                        pp <- unlist(M$parval[xconstrain[[i]]$warg]);
                        myidx <- with(xconstrain[[i]],order(c(wargidx,exoidx)))
                        D <- cbind(rbind(pp)%x%cbind(rep(1,nrow(Mu))),
                                   data[,xconstrain[[i]]$exo,drop=FALSE])[,myidx,drop=FALSE]
                        mu <- try(xconstrain[[i]]$func(D),silent=TRUE)
                        if (is.data.frame(mu)) mu <- mu[,1]
                        if (inherits(mu,"try-error") || NROW(mu)!=NROW(Mu)) {
                            ## mu1 <- with(xconstrain[[i]],
                            ##            apply(data[,exo,drop=FALSE],1,
                            ##                  function(x) func(unlist(c(pp,x))[myidx])))
                            mu <- apply(D,1,xconstrain[[i]]$func)
                        }
                        Mu[,xconstrain[[i]]$endo] <- mu
                    }
                    offsets <- Mu%*%t(M$IAi)[,endogenous(x)]
                    return(offsets)
                }
                return(NULL)
            }

            myObj <- function(pp) {
                if (Optim$constrain) {
                    pp[constrained] <- exp(pp[constrained])
                }
                offset <- MkOffset(pp)
                mu0 <- mu; S0 <- S; x0 <- x
                if (!is.null(offset)) {
                    x0$constrain[iconstrain] <- NULL
                    data0 <- data[,manifest(x0)]
                    data0[,endogenous(x)] <- data0[,endogenous(x)]-offset
                    pd <- procdata.lvm(x0,data=data0)
                    S0 <- pd$S; mu0 <- pd$mu
                    x0$mean[yconstrain] <- 0
                }
                do.call(ObjectiveFun, list(x=x0, p=pp, data=data, S=S0, mu=mu0, n=n, weights=weights
                                          ,data2=data2, offset=offset
                                           ))
            }
            if (!exists(ObjectiveFun) || is.null(get(ObjectiveFun)))
              myObj <- NULL

            myGrad <- function(pp) {
                if (Optim$constrain)
                    pp[constrained] <- exp(pp[constrained])
                S <- do.call(GradFun, list(x=x, p=pp, data=data, S=S, mu=mu, n=n, weights=weights
                                         , data2=data2##, offset=offset
                                           ))
                if (Optim$constrain) {
                    S[constrained] <- S[constrained]*pp[constrained]
                }
                if (is.null(mu) & index(x)$npar.mean>0) {
                    return(S[-c(seq_len(index(x)$npar.mean))])
                }
                if (length(S)<length(pp))  S <- c(S,rep(0,length(pp)-length(S)))

                return(S)
            }
            myInfo <- function(pp) {
                I <- do.call(InformationFun, list(p=pp,
                                                  obj=myObj,
                                                  x=x, data=data,
                                                  S=S, mu=mu,
                                                  n=n,
                                                  weights=weights, data2=data2,
                                                  type=Optim$information
                                                  ))
                if (is.null(mu) && index(x)$npar.mean>0) {
                    return(I[-seq_len(index(x)$npar.mean),-seq_len(index(x)$npar.mean)])
                }
                return(I)
            }

        }

        myHess <- function(pp) {
            p0 <- pp
            if (Optim$constrain)
                pp[constrained] <- exp(pp[constrained])
            I0 <- myInfo(pp)
            attributes(I0)$grad <- NULL
            D <- attributes(I0)$grad
            if (is.null(D)) {
                D <- myGrad(p0)
                attributes(I0)$grad <- D
            }
            if (Optim$constrain) {
                I0[constrained,-constrained] <- apply(I0[constrained,-constrained,drop=FALSE],2,function(x) x*pp[constrained]);
                I0[-constrained,constrained] <- t(I0[constrained,-constrained])
                if (sum(constrained)==1) {
                    I0[constrained,constrained] <- I0[constrained,constrained]*outer(pp[constrained],pp[constrained]) - D[constrained]
                } else {
                    I0[constrained,constrained] <- I0[constrained,constrained]*outer(pp[constrained],pp[constrained]) - diag(D[constrained],ncol=length(constrained))
                }
            }
            return(I0)
        }
        if (is.null(tryCatch(get(InformationFun),error = function (x) NULL)))
            myInfo <- myHess <- NULL
        if (is.null(tryCatch(get(GradFun),error = function (x) NULL)))
            myGrad <- NULL

        if (messages>1) message("Optimizing objective function...")
        if (Optim$trace>0 & (messages>1)) message("\n")
        ## Optimize with lower constraints on the variance-parameters
        if ((is.data.frame(data) | is.matrix(data)) && nrow(data)==0) stop("No observations")
        if (!missing(p)) {
            opt <- list(estimate=p)

        } else {
            if (!is.null(Optim$method)) {
                optarg <- list(start=Optim$start, objective=myObj, gradient=myGrad, hessian=myHess, lower=lower, control=Optim, debug=debug)
                if (length(Optim$method)>1) {
                    Optim$optimx.method <- Optim$method
                }
                if (!is.null(Optim$optimx.method)) {
                    Optim$method <- "optimx"
                }
                if (Optim$method%in%c("optimx","optim")) {
                    optimcontrolnames <-
                        c("trace",
                          "follow.on",
                          "save.failures",
                          "maximize",
                          "all.methods",
                          "kkt",
                          "kkttol",
                          "kkt2tol",
                          "starttests",
                          "dowarn",
                          "badval",
                          "usenumDeriv",
                          "fnscale",
                          "parscale",
                          "ndeps",
                          "maxit",
                          "abstol",
                          "reltol",
                          #"alpha","beta","gamma",
                          "REPORT",
                          "type",
                          "lmm",
                          "factr",
                          "pgtol")
                    if (!is.null(optarg$control)) {
                        optarg$control[names(optarg$control)%ni%optimcontrolnames] <- NULL
                    }
                    args <- names(formals(get(Optim$method)))
                    names(optarg)[1] <- "par"
                    if (is.null(optarg$upper)) optarg$upper <- Inf
                    if (!is.null(optarg[["objective"]])) names(optarg)[2] <- "fn"
                    if (!is.null(optarg[["gradient"]])) names(optarg)[3] <- "gr"
                    ##if (!is.null(optarg[["hessian"]])) names(optarg)[4] <- "hess"
                    optarg$hessian <- NULL
                    optarg[names(optarg)%ni%args] <- NULL
                }
                if (!is.null(Optim$optimx.method)) optarg$method <- Optim$optimx.method
                opt <- do.call(Optim$method,
                               optarg)
                if (inherits(opt,"optimx")) {
                    opt <- list(par=coef(opt)[1,])
                }
                if (is.null(opt$estimate))
                    opt$estimate <- opt$par
                if (Optim$constrain) {
                    opt$estimate[constrained] <- exp(opt$estimate[constrained])
                }

                if (XconstrStdOpt & !is.null(myGrad))
                    opt$gradient <- as.vector(myGrad(opt$par))
                else {
                    opt$gradient <- numDeriv::grad(myObj,opt$par)
                }
            } else {
                if (!NoOptim) {
                    opt <- do.call(ObjectiveFun, list(x=x,data=data,control=control,...))
                    opt$gradient <- rep(0,length(opt$estimate))
                } else {
                    opt <- list(estimate=Optim$start,
                                gradient=rep(0,length(Optim$start)))
                }
            }
        }
        if (!is.null(opt$convergence)) {
            if (opt$convergence != 0)
                warning("Lack of convergence. Increase number of iteration or change starting values.")
        } else if (!is.null(opt$gradient) && mean(opt$gradient)^2>1e-3)
            warning("Lack of convergence. Increase number of iteration or change starting values.")
        if (quick) {
            return(opt$estimate)
        }
        ## Calculate std.err:
        pp <- rep(NA,length(coefname)); names(pp) <- coefname
        pp.idx <- NULL
        if (!is.null(names(opt$estimate))) {
            pp[names(opt$estimate)] <- opt$estimate
            pp.idx <- na.omit(match(coefname,names(opt$estimate)))
        } else {
            pp[] <- opt$estimate
            pp.idx <- seq(length(pp))
        }
        ## TODO:
        ## if (length(pp.idx)!=length(pp)) {
        ##     pp <- rep(NA,length(coefname)); names(pp) <- coefname
        ##     pp[] <- opt$estimate
        ##     pp.idx <- seq(length(pp))
        ## }

        suppressWarnings(mom <- tryCatch(modelVar(x, pp, data=data),error=function(x)NULL))
        if (NoOptim) {
            asVar <- matrix(NA,ncol=length(pp),nrow=length(pp))
        } else {

            if (messages>1) message("\nCalculating asymptotic variance...\n")
            asVarFun  <- paste0(estimator, "_variance", ".lvm")

            if (!exists(asVarFun)) {
                if (is.null(myInfo)) {
                    if (!is.null(myGrad))
                        myInfo <- function(pp)
                            numDeriv::jacobian(myGrad,pp,method=lava.options()$Dmethod)
                    else
                        myInfo <- function(pp)
                            numDeriv::hessian(myObj,pp)
                }
                I <- myInfo(opt$estimate)
                asVar <- tryCatch(Inverse(I),
                                  error=function(e) matrix(NA, length(opt$estimate), length(opt$estimate)))
            } else {
                asVar <- tryCatch(do.call(asVarFun,
                                          list(x=x,p=opt$estimate,data=data,opt=opt)),
                                  error=function(e) matrix(NA, length(opt$estimate), length(opt$estimate)))
            }

            if (any(is.na(asVar))) {
                warning("Problems with asymptotic variance matrix. Possibly non-singular information matrix!")
            }
            if (!is.null(attributes(asVar)$pseudo) && attributes(asVar)$pseudo) {
                warning("Near-singular covariance matrix, using pseudo-inverse!")
            }
            diag(asVar)[diag(asVar)==0] <- NA
        }

        mycoef <- matrix(NA,nrow=nparall,ncol=4)
        mycoef[pp.idx,1] <- opt$estimate ## Will be finished during post.hooks

        res <- list(model=x, call=cl, coef=mycoef,
                    vcov=asVar, mu=mu, S=S, ##A=A, P=P,
                    model0=mymodel, ## Random slope hack
                    estimator=estimator, opt=opt,expar=x$expar,
                    data=list(model.frame=data, S=S, mu=mu,
                              C=mom$C, v=mom$v, n=n,
                              m=length(latent(x)), k=length(index(x)$manifest), data2=data2),
                    weights=weights, data2=data2,
                    cluster=id,
                    pp.idx=pp.idx,
                    graph=NULL, control=Optim)

        class(res) <- myclass

        myhooks <- gethook("post.hooks")
        for (f in myhooks) {
            res0 <- do.call(f,list(x=res))
            if (!is.null(res0))
                res <- res0
        }

        if(graph) {
            res <- edgelabels(res,type="est")
        }

        return(res)
    }

###{{{ estimate.lvm

##' Estimation of parameters in a Latent Variable Model (lvm)
##'
##' Estimate parameters. MLE, IV or user-defined estimator.
##'
##' A list of parameters controlling the estimation and optimization procedures
##' is parsed via the \code{control} argument. By default Maximum Likelihood is
##' used assuming multivariate normal distributed measurement errors. A list
##' with one or more of the following elements is expected:
##'
##' \describe{
##' \item{start:}{Starting value. The order of the parameters can be shown by
##' calling \code{coef} (with \code{mean=TRUE}) on the \code{lvm}-object or with
##' \code{plot(..., labels=TRUE)}. Note that this requires a check that it is
##' actual the model being estimated, as \code{estimate} might add additional
##' restriction to the model, e.g. through the \code{fix} and \code{exo.fix}
##' arguments. The \code{lvm}-object of a fitted model can be extracted with the
##' \code{Model}-function.}
##'
##' \item{starterfun:}{Starter-function with syntax
##' \code{function(lvm, S, mu)}.  Three builtin functions are available:
##' \code{startvalues}, \code{startvalues0}, \code{startvalues1}, ...}
##'
##' \item{estimator:}{ String defining which estimator to use (Defaults to
##' ``\code{gaussian}'')}
##'
##' \item{meanstructure}{Logical variable indicating
##' whether to fit model with meanstructure.}
##'
##' \item{method:}{ String pointing to
##' alternative optimizer (e.g. \code{optim} to use simulated annealing).}
##'
##' \item{control:}{ Parameters passed to the optimizer (default
##' \code{stats::nlminb}).}
##'
##' \item{tol:}{ Tolerance of optimization constraints on lower limit of
##' variance parameters.  } }
##'
##' @param x \code{lvm}-object
##' @param data \code{data.frame}
##' @param estimator String defining the estimator (see details below)
##' @param control control/optimization parameters (see details below)
##' @param missing Logical variable indiciating how to treat missing data.
##' Setting to FALSE leads to complete case analysis. In the other case
##' likelihood based inference is obtained by integrating out the missing data
##' under assumption the assumption that data is missing at random (MAR).
##' @param weights Optional weights to used by the chosen estimator.
##' @param weightsname Weights names (variable names of the model) in case
##' \code{weights} was given as a vector of column names of \code{data}
##' @param data2 Optional additional dataset used by the chosen
##' estimator.
##' @param id Vector (or name of column in \code{data}) that identifies
##' correlated groups of observations in the data leading to variance estimates
##' based on a sandwich estimator
##' @param fix Logical variable indicating whether parameter restriction
##' automatically should be imposed (e.g. intercepts of latent variables set to
##' 0 and at least one regression parameter of each measurement model fixed to
##' ensure identifiability.)
##' @param index For internal use only
##' @param graph For internal use only
##' @param messages Control how much information should be
##' printed during estimation (0: none)
##' @param quick If TRUE the parameter estimates are calculated but all
##' additional information such as standard errors are skipped
##' @param method Optimization method
##' @param param set parametrization (see \code{help(lava.options)})
##' @param cluster Obsolete. Alias for 'id'.
##' @param p Evaluate model in parameter 'p' (no optimization)
##' @param ... Additional arguments to be passed to lower-level functions
##' @return A \code{lvmfit}-object.
##' @author Klaus K. Holst
##' @seealso estimate.default score, information
##' @keywords models regression
##' @export
##' @method estimate lvm
##' @examples
##' dd <- read.table(header=TRUE,
##' text="x1 x2 x3
##'  0.0 -0.5 -2.5
##' -0.5 -2.0  0.0
##'  1.0  1.5  1.0
##'  0.0  0.5  0.0
##' -2.5 -1.5 -1.0")
##' e <- estimate(lvm(c(x1,x2,x3)~u),dd)
##'
##' ## Simulation example
##' m <- lvm(list(y~v1+v2+v3+v4,c(v1,v2,v3,v4)~x))
##' covariance(m) <- v1~v2+v3+v4
##' dd <- sim(m,10000) ## Simulate 10000 observations from model
##' e <- estimate(m, dd) ## Estimate parameters
##' e
##'
##' ## Using just sufficient statistics
##' n <- nrow(dd)
##' e0 <- estimate(m,data=list(S=cov(dd)*(n-1)/n,mu=colMeans(dd),n=n))
##' rm(dd)
##'
##' ## Multiple group analysis
##' m <- lvm()
##' regression(m) <- c(y1,y2,y3)~u
##' regression(m) <- u~x
##' d1 <- sim(m,100,p=c("u,u"=1,"u~x"=1))
##' d2 <- sim(m,100,p=c("u,u"=2,"u~x"=-1))
##'
##' mm <- baptize(m)
##' regression(mm,u~x) <- NA
##' covariance(mm,~u) <- NA
##' intercept(mm,~u) <- NA
##' ee <- estimate(list(mm,mm),list(d1,d2))
##'
##' ## Missing data
##' d0 <- makemissing(d1,cols=1:2)
##' e0 <- estimate(m,d0,missing=TRUE)
##' e0
`estimate.lvm` <-
    function(x, data=parent.frame(),
      estimator=NULL,
      control=list(),
      missing=FALSE,
      weights, weightsname,
      data2,
      id,
      fix,
      index=!quick,
      graph=FALSE,
      messages=lava.options()$messages,
      quick=FALSE,
      method,
      param,
      cluster,
      p,
      ...) {
        cl <- match.call()
        if (!base::missing(param)) {
            oldparam <- lava.options()$param
            lava.options(param=param)
            on.exit(lava.options(param=oldparam))
        }
        if (!base::missing(method)) {
            control["method"] <- list(method)
        }

        Optim <- list(
            iter.max=lava.options()$iter.max,
            trace=ifelse(lava.options()$debug,3,0),
            gamma=lava.options()$gamma,
            gamma2=1,
            ngamma=lava.options()$ngamma,
            backtrack=lava.options()$backtrack,
            lambda=0.05,
            abs.tol=1e-9,
            epsilon=1e-10,
            delta=1e-10,
            rel.tol=1e-10,
            S.tol=1e-5,
            stabil=FALSE,
            start=NULL,
            constrain=lava.options()$constrain,
            method=NULL,
            starterfun="startvalues0",
            information="E",
            meanstructure=TRUE,
            sparse=FALSE,
            tol=lava.options()$tol)

        defopt <- lava.options()[]
        defopt <- defopt[intersect(names(defopt),names(Optim))]
        Optim[names(defopt)] <- defopt
        if (length(control)>0) {
            Optim[names(control)] <- control
        }
        if (is.environment(data)) {
            innames <- intersect(ls(envir=data),vars(x))
            data <- as.data.frame(lapply(innames,function(x) get(x,envir=data)))
            names(data) <- innames
        }
        if (length(exogenous(x)>0)) {
            catx <- categorical2dummy(x,data)
            x <- catx$x; data <- catx$data
        }

        if (!lava.options()$exogenous) exogenous(x) <- NULL

        redvar <- intersect(intersect(parlabels(x),latent(x)),colnames(data))
        if (length(redvar)>0)
            warning(paste("Latent variable exists in dataset",redvar))
        ## Random-slopes:
        xfix <- setdiff(colnames(data)[(colnames(data)%in%parlabels(x,exo=TRUE))],latent(x))
        if (base::missing(fix)) {
            fix <- ifelse(length(xfix)>0,FALSE,TRUE)
        }
        Debug(list("start=",Optim$start))

        if (!base::missing(cluster)) id <- cluster

        ## Weights...
        if (!base::missing(weights)) {
            if (is.character(weights)) {
                weights <- data[,weights,drop=FALSE]
                if (!base::missing(weightsname)) {
                    colnames(weights) <- weightsname
                } else {
                    yvar <- index(x)$endogenous
                    nw <- seq_len(min(length(yvar),ncol(weights)))
                    colnames(weights)[nw] <- yvar[nw]
                }
            }
            weights <- cbind(weights)
        } else {
            weights <- NULL
        }
        if (!base::missing(data2)) {
            if (is.character(data2)) {
                data2 <- data[,data2]
            }
        } else {
            data2 <- NULL
        }
        ## Correlated clusters...
        if (!base::missing(id)) {
            if (is.character(id)) {
                id <- data[,id]
            }
        } else {
            id <- NULL
        }

        Debug("procdata")
        val <- try({
            dd <- procdata.lvm(x,data=data,missing=missing)
            S <- dd$S; mu <- dd$mu; n <- dd$n
            var.missing <- setdiff(vars(x),colnames(S))
        }, silent=TRUE)
        if (inherits(val,"try-error")) {
            var.missing <- setdiff(vars(x),colnames(data))
            S <- NULL; mu <- NULL; n <- nrow(data)
        }

        ##  if (fix) {
        if (length(var.missing)>0) {## Convert to latent:
            new.lat <- setdiff(var.missing,latent(x))
            if (length(new.lat)>0)
                x <- latent(x, new.lat)
        }
        ##}

        ## Run hooks (additional lava plugins)
        myhooks <- gethook()
        for (f in myhooks) {
            res <- do.call(f, list(x=x,data=data,weights=weights,data2=data2,estimator=estimator,optim=Optim))
            if (!is.null(res$x)) x <- res$x
            if (!is.null(res$data)) data <- res$data
            if (!is.null(res$weights)) weights <- res$weights
            if (!is.null(res$data2)) data2 <- res$data2
            if (!is.null(res$optim)) Optim <- res$optim
            if (!is.null(res$estimator)) estimator <- res$estimator
            rm(res)
        }
        if (is.null(estimator)) {
            if (!missing(weights) && !is.null(weights)) {
                estimator <- "normal"
            } else estimator <- "gaussian"
        }
        checkestimator <- function(x,...) {
            ffname <- paste0(x,c("_objective","_gradient"),".lvm")
            exists(ffname[1])||exists(ffname[2])
        }
        if (!checkestimator(estimator)) { ## Try down/up-case version
            estimator <- tolower(estimator)
            if (!checkestimator(estimator)) {
                estimator <- toupper(estimator)
            }
        }
        ObjectiveFun  <- paste0(estimator, "_objective", ".lvm")
        GradFun  <- paste0(estimator, "_gradient", ".lvm")
        if (!exists(ObjectiveFun) & !exists(GradFun))
            stop("Unknown estimator.")


        Method <-  paste0(estimator, "_method", ".lvm")
        if (!exists(Method)) {
            Method <- "nlminb1"
        } else {
            Method <- get(Method)
        }
        NoOptim <- "method"%in%names(control) && is.null(control$method)
        if (is.null(Optim$method) && !(NoOptim)) {
            Optim$method <- if (missing && Method!="nlminb0") "nlminb1" else Method
        }


        if (index) {
            ## Proces data and setup some matrices
            x <- fixsome(x, measurement.fix=fix, S=S, mu=mu, n=n,debug=messages>1)
            if (messages>1)
                message("Reindexing model...\n")
            if (length(xfix)>0) {
                index(x) <- reindex(x,sparse=Optim$sparse,zeroones=TRUE,deriv=TRUE)
            } else {
                x <- updatelvm(x,sparse=Optim$sparse,zeroones=TRUE,deriv=TRUE,mean=TRUE)
            }
        }
        if (is.null(estimator) || estimator==FALSE) {
            return(x)
        }

        if (length(index(x)$endogenous)==0) stop("No observed outcome variables. Check variable names in model and data.")
        if (!Optim$meanstructure) {
            mu <- NULL
        }

        nparall <- index(x)$npar + ifelse(Optim$meanstructure, index(x)$npar.mean+index(x)$npar.ex,0)
        ## Get starting values
        if (!missing(p)) {
            start <- p
            Optim$start <- p
        } else {
          myparnames <- coef(x,mean=TRUE)
          paragree <- FALSE
          paragree_pos <- c()
          if (!is.null(Optim$start)) {
            paragree <- myparnames%in%names(Optim$start)
            paragree_pos <- matrix(0, sum(paragree), 2)
            count <- 0
            for (i in which(paragree)) {
              count <- count + 1
              paragree_pos[count,] <-  c(i, which(myparnames[i]==names(Optim$start))[1])
            }
          }
          ## If starting values are given for all parameters and not named, then
          ## we just keep the provided starting values as they are.
          ## Otherwise, calculate starting values and place user-provided
          ## starting values in the right positions
          if (! (length(Optim$start)==length(myparnames) & sum(paragree)==0)) {
            if (is.null(Optim$starterfun) && lava.options()$param!="relative")
              Optim$starterfun <- startvalues0
            start <- suppressWarnings(do.call(Optim$starterfun, list(x=x,S=S,mu=mu,debug=lava.options()$debug,
                                                                     messages=messages,data=data,...)))
            if (!is.null(x$expar) && length(start)<nparall) {
              ii <- which(index(x)$e1==1)
              start <- c(start, structure(unlist(x$expar[ii]), names=names(x$expar)[ii]))
            }
            for (i in seq_len(NROW(paragree_pos))) {
              start[paragree_pos[i, 1]] <- Optim$start[paragree_pos[i, 2]]
            }
            Optim$start <- start
          }
        }
        ## Missing data
        if (missing) {
            return(estimate.MAR(x=x,data=data,fix=fix,control=Optim,debug=lava.options()$debug,
                                messages=messages,estimator=estimator,weights=weights,data2=data2,cluster=id,...))
        }
        coefname <- coef(x,mean=Optim$meanstructure,fix=FALSE);
        names(Optim$start) <- coefname

        ## Non-linear parameter constraints involving observed variables? (e.g. nonlinear regression)
        constr <- lapply(constrain(x), function(z)(attributes(z)$args))
        xconstrain <- intersect(unlist(constr), manifest(x))
        xconstrainM <- TRUE
        XconstrStdOpt <- TRUE
        if (length(xconstrain)>0) {
            constrainM <- names(constr)%in%unlist(x$mean)
            for (i in seq_len(length(constr))) {
                if (!constrainM[i]) {
                    if (any(constr[[i]]%in%xconstrain)) {
                        xconstrainM <- FALSE
                        break;
                    }
                }
            }
            if (xconstrainM & ( (is.null(control$method) || Optim$method=="nlminb0") & (lava.options()$test & estimator=="gaussian") ) ) {
                XconstrStdOpt <- FALSE
                Optim$method <- "nlminb0"
                if (is.null(control$constrain)) control$constrain <- TRUE
            }
        }

        ## Setup optimization constraints
        lowmin <- -Inf
        lower <- rep(lowmin,length(Optim$start))
        if (length(Optim$constrain)==1 & Optim$constrain)
            lower[variances(x)+index(x)$npar.mean] <- Optim$tol
        if (any(Optim$constrain)) {
            if (length(Optim$constrain)!=length(lower))
                constrained <- is.finite(lower)
            else
                constrained <- Optim$constrain
            lower[] <- -Inf
            Optim$constrain <- TRUE
            constrained <- which(constrained)
            nn <- names(Optim$start)
            CS <- Optim$start[constrained]
            CS[CS<0] <- 0.01
            Optim$start[constrained] <- log(CS)
            names(Optim$start) <- nn
        }
        ## Fix problems with starting values?
        Optim$start[is.nan(unlist(Optim$start))] <- 0

        ObjectiveFun  <- paste0(estimator, "_objective", ".lvm")
        GradFun  <- paste0(estimator, "_gradient", ".lvm")
        if (!exists(ObjectiveFun) & !exists(GradFun)) stop("Unknown estimator.")

        InformationFun <- paste0(estimator, "_hessian", ".lvm")

        mymodel <- x
        myclass <- "lvmfit"

        ## Random slopes?
        if (length(xfix)>0 | (length(xconstrain)>0 & XconstrStdOpt | !lava.options()$test)) { ## Yes
            x0 <- x

            if (length(xfix)>0) {
                myclass <- c("lvmfit.randomslope",myclass)
                nrow <- length(vars(x))
                xpos <- lapply(xfix,function(y) which(regfix(x)$labels==y))
                colpos <- lapply(xpos, function(y) ceiling(y/nrow))
                rowpos <- lapply(xpos, function(y) (y-1)%%nrow+1)
                myfix <- list(var=xfix, col=colpos, row=rowpos)
                x0 <- x
                for (i in seq_along(myfix$var))
                    for (j in seq_len(length(myfix$col[[i]])))
                        regfix(x0, from=vars(x0)[myfix$row[[i]][j]],
                               to=vars(x0)[myfix$col[[i]][j]]) <-
                            colMeans(data[,myfix$var[[i]],drop=FALSE])
                x0 <- updatelvm(x0,zeroones=TRUE,deriv=TRUE)
                x <- x0
                ## Alter start-values/constraints:
                new.par.idx <- which(coef(mymodel,mean=TRUE,fix=FALSE)%in%coef(x0,mean=TRUE,fix=FALSE))
                if (length(Optim$start)>length(new.par.idx))
                    Optim$start <- Optim$start[new.par.idx]
                lower <- lower[new.par.idx]
                if (Optim$constrain) {
                    constrained <- match(constrained,new.par.idx)
                }
            }
            mydata <- as.matrix(data[,manifest(x0)])

            myObj <- function(pp) {
                if (Optim$constrain) {
                    pp[constrained] <- exp(pp[constrained])
                }
                myfun <- function(ii) {
                    if (length(xfix)>0)
                        for (i in seq_along(myfix$var)) {
                            x0$fix[cbind(rowpos[[i]],colpos[[i]])] <- index(x0)$A[cbind(rowpos[[i]],colpos[[i]])] <- data[ii,xfix[i]]
                        }
                    if (is.list(data2)) {
                        res <- do.call(ObjectiveFun, list(x=x0, p=pp, data=mydata[ii,], n=1, weights=weights[ii,], data2=data2[ii,]))
                    } else {
                        res <- do.call(ObjectiveFun, list(x=x0, p=pp, data=mydata[ii,], n=1, weights=weights[ii,], data2=data2))
                    }
                    return(res)
                }
                sum(sapply(seq_len(nrow(data)),myfun))
            }

            myGrad <- function(pp) {
                if (Optim$constrain) {
                    pp[constrained] <- exp(pp[constrained])
                }
                myfun <- function(ii) {
                    if (length(xfix)>0)
                        for (i in seq_along(myfix$var)) {
                            x0$fix[cbind(rowpos[[i]],colpos[[i]])] <- index(x0)$A[cbind(rowpos[[i]],colpos[[i]])] <- data[ii,xfix[i]]
                        }
                    if (is.list(data2)) {
                        rr <- do.call(GradFun, list(x=x0, p=pp, data=mydata[ii,,drop=FALSE], n=1, weights=weights[ii,], data2=data2))
                    } else
                    {
                        rr <- do.call(GradFun, list(x=x0, p=pp, data=mydata[ii,,drop=FALSE], n=1, weights=weights[ii,], data2=data2[ii,]))
                    }
                    return(rr)
                }
                ss <- rowSums(rbind(sapply(seq_len(nrow(data)),myfun)))
                if (Optim$constrain) {
                    ss[constrained] <- ss[constrained]*pp[constrained]
                }
                return(ss)
            }


            myInfo <- function(pp) {
                myfun <- function(ii) {
                    if (length(xfix)>0)
                        for (i in seq_along(myfix$var)) {
                            x0$fix[cbind(rowpos[[i]],colpos[[i]])] <- index(x0)$A[cbind(rowpos[[i]],colpos[[i]])] <- data[ii,xfix[i]]
                        }
                    if (is.list(data2)) {
                        res <- do.call(InformationFun, list(p=pp, obj=myObj, x=x0, data=data[ii,],
                                                            n=1, weights=weights[ii,], data2=data2))
                    } else {
                        res <- do.call(InformationFun, list(p=pp, obj=myObj, x=x0, data=data[ii,],
                                                            n=1, weights=weights[ii,], data2=data2[ii,]))
                    }
                    return(res)
                }
                L <- lapply(seq_len(nrow(data)),function(x) myfun(x))
                val <- apply(array(unlist(L),dim=c(length(pp),length(pp),nrow(data))),c(1,2),sum)
                if (!is.null(attributes(L[[1]])$grad)) {
                    attributes(val)$grad <- colSums (
                                       matrix( unlist(lapply(L,function(i) attributes(i)$grad)) , ncol=length(pp), byrow=TRUE)
                                   )
                }
                return(val)
            }

            ##################################################
        } else { ## No, standard model

            ## Non-linear parameter constraints involving observed variables? (e.g. nonlinear regression)
            xconstrain <- c()
            for (i in seq_len(length(constrain(x)))) {
                z <- constrain(x)[[i]]
                xx <- intersect(attributes(z)$args,manifest(x))
                if (length(xx)>0) {
                    warg <- setdiff(attributes(z)$args,xx)
                    wargidx <- which(attributes(z)$args%in%warg)
                    exoidx <- which(attributes(z)$args%in%xx)
                    parname <- names(constrain(x))[i]
                    y <- names(which(unlist(lapply(intercept(x),function(x) x==parname))))
                    el <- list(i,y,parname,xx,exoidx,warg,wargidx,z)
                    names(el) <- c("idx","endo","parname","exo","exoidx","warg","wargidx","func")
                    xconstrain <- c(xconstrain,list(el))
                }
            }
            yconstrain <- unlist(lapply(xconstrain,function(x) x$endo))
            iconstrain <- unlist(lapply(xconstrain,function(x) x$idx))

            MkOffset <- function(pp,grad=FALSE) {
                if (length(xconstrain)>0) {
                    Mu <- matrix(0,nrow(data),length(vars(x))); colnames(Mu) <- vars(x)
                    M <- modelVar(x,p=pp,data=data)
                    M$parval <- c(M$parval,  x$mean[unlist(lapply(x$mean,is.numeric))])
                    for (i in seq_len(length(xconstrain))) {
                        pp <- unlist(M$parval[xconstrain[[i]]$warg]);
                        myidx <- with(xconstrain[[i]],order(c(wargidx,exoidx)))
                        D <- cbind(rbind(pp)%x%cbind(rep(1,nrow(Mu))),
                                   data[,xconstrain[[i]]$exo,drop=FALSE])[,myidx,drop=FALSE]
                        mu <- try(xconstrain[[i]]$func(D),silent=TRUE)
                        if (is.data.frame(mu)) mu <- mu[,1]
                        if (inherits(mu,"try-error") || NROW(mu)!=NROW(Mu)) {
                            ## mu1 <- with(xconstrain[[i]],
                            ##            apply(data[,exo,drop=FALSE],1,
                            ##                  function(x) func(unlist(c(pp,x))[myidx])))
                            mu <- apply(D,1,xconstrain[[i]]$func)
                        }
                        Mu[,xconstrain[[i]]$endo] <- mu
                    }
                    offsets <- Mu%*%t(M$IAi)[,endogenous(x)]
                    return(offsets)
                }
                return(NULL)
            }

            myObj <- function(pp) {
                if (Optim$constrain) {
                    pp[constrained] <- exp(pp[constrained])
                }
                offset <- MkOffset(pp)
                mu0 <- mu; S0 <- S; x0 <- x
                if (!is.null(offset)) {
                    x0$constrain[iconstrain] <- NULL
                    data0 <- data[,manifest(x0)]
                    data0[,endogenous(x)] <- data0[,endogenous(x)]-offset
                    pd <- procdata.lvm(x0,data=data0)
                    S0 <- pd$S; mu0 <- pd$mu
                    x0$mean[yconstrain] <- 0
                }
                do.call(ObjectiveFun, list(x=x0, p=pp, data=data, S=S0, mu=mu0, n=n, weights=weights
                                          ,data2=data2, offset=offset
                                           ))
            }
            if (!exists(ObjectiveFun) || is.null(get(ObjectiveFun)))
              myObj <- NULL

            myGrad <- function(pp) {
                if (Optim$constrain)
                    pp[constrained] <- exp(pp[constrained])
                S <- do.call(GradFun, list(x=x, p=pp, data=data, S=S, mu=mu, n=n, weights=weights
                                         , data2=data2##, offset=offset
                                           ))
                if (Optim$constrain) {
                    S[constrained] <- S[constrained]*pp[constrained]
                }
                if (is.null(mu) & index(x)$npar.mean>0) {
                    return(S[-c(seq_len(index(x)$npar.mean))])
                }
                if (length(S)<length(pp))  S <- c(S,rep(0,length(pp)-length(S)))

                return(S)
            }
            myInfo <- function(pp) {
                I <- do.call(InformationFun, list(p=pp,
                                                  obj=myObj,
                                                  x=x, data=data,
                                                  S=S, mu=mu,
                                                  n=n,
                                                  weights=weights, data2=data2,
                                                  type=Optim$information
                                                  ))
                if (is.null(mu) && index(x)$npar.mean>0) {
                    return(I[-seq_len(index(x)$npar.mean),-seq_len(index(x)$npar.mean)])
                }
                return(I)
            }

        }

        myHess <- function(pp) {
            p0 <- pp
            if (Optim$constrain)
                pp[constrained] <- exp(pp[constrained])
            I0 <- myInfo(pp)
            attributes(I0)$grad <- NULL
            D <- attributes(I0)$grad
            if (is.null(D)) {
                D <- myGrad(p0)
                attributes(I0)$grad <- D
            }
            if (Optim$constrain) {
                I0[constrained,-constrained] <- apply(I0[constrained,-constrained,drop=FALSE],2,function(x) x*pp[constrained]);
                I0[-constrained,constrained] <- t(I0[constrained,-constrained])
                if (sum(constrained)==1) {
                    I0[constrained,constrained] <- I0[constrained,constrained]*outer(pp[constrained],pp[constrained]) - D[constrained]
                } else {
                    I0[constrained,constrained] <- I0[constrained,constrained]*outer(pp[constrained],pp[constrained]) - diag(D[constrained],ncol=length(constrained))
                }
            }
            return(I0)
        }
        if (is.null(tryCatch(get(InformationFun),error = function (x) NULL)))
            myInfo <- myHess <- NULL
        if (is.null(tryCatch(get(GradFun),error = function (x) NULL)))
            myGrad <- NULL

        if (messages>1) message("Optimizing objective function...")
        if (Optim$trace>0 & (messages>1)) message("\n")
        ## Optimize with lower constraints on the variance-parameters
        if ((is.data.frame(data) | is.matrix(data)) && nrow(data)==0) stop("No observations")
        if (!missing(p)) {
            opt <- list(estimate=p)

        } else {
            if (!is.null(Optim$method)) {
                optarg <- list(start=Optim$start, objective=myObj, gradient=myGrad, hessian=myHess, lower=lower, control=Optim, debug=debug)
                if (length(Optim$method)>1) {
                    Optim$optimx.method <- Optim$method
                }
                if (!is.null(Optim$optimx.method)) {
                    Optim$method <- "optimx"
                }
                if (Optim$method%in%c("optimx","optim")) {
                    optimcontrolnames <-
                        c("trace",
                          "follow.on",
                          "save.failures",
                          "maximize",
                          "all.methods",
                          "kkt",
                          "kkttol",
                          "kkt2tol",
                          "starttests",
                          "dowarn",
                          "badval",
                          "usenumDeriv",
                          "fnscale",
                          "parscale",
                          "ndeps",
                          "maxit",
                          "abstol",
                          "reltol",
                          #"alpha","beta","gamma",
                          "REPORT",
                          "type",
                          "lmm",
                          "factr",
                          "pgtol")
                    if (!is.null(optarg$control)) {
                        optarg$control[names(optarg$control)%ni%optimcontrolnames] <- NULL
                    }
                    args <- names(formals(get(Optim$method)))
                    names(optarg)[1] <- "par"
                    if (is.null(optarg$upper)) optarg$upper <- Inf
                    if (!is.null(optarg[["objective"]])) names(optarg)[2] <- "fn"
                    if (!is.null(optarg[["gradient"]])) names(optarg)[3] <- "gr"
                    ##if (!is.null(optarg[["hessian"]])) names(optarg)[4] <- "hess"
                    optarg$hessian <- NULL
                    optarg[names(optarg)%ni%args] <- NULL
                }
                if (!is.null(Optim$optimx.method)) optarg$method <- Optim$optimx.method
                opt <- do.call(Optim$method,
                               optarg)
                if (inherits(opt,"optimx")) {
                    opt <- list(par=coef(opt)[1,])
                }
                if (is.null(opt$estimate))
                    opt$estimate <- opt$par
                if (Optim$constrain) {
                    opt$estimate[constrained] <- exp(opt$estimate[constrained])
                }

                if (XconstrStdOpt & !is.null(myGrad))
                    opt$gradient <- as.vector(myGrad(opt$par))
                else {
                    opt$gradient <- numDeriv::grad(myObj,opt$par)
                }
            } else {
                if (!NoOptim) {
                    opt <- do.call(ObjectiveFun, list(x=x,data=data,control=control,...))
                    opt$gradient <- rep(0,length(opt$estimate))
                } else {
                    opt <- list(estimate=Optim$start,
                                gradient=rep(0,length(Optim$start)))
                }
            }
        }
        if (!is.null(opt$convergence)) {
            if (opt$convergence != 0)
                warning("Lack of convergence. Increase number of iteration or change starting values.")
        } else if (!is.null(opt$gradient) && mean(opt$gradient)^2>1e-3)
            warning("Lack of convergence. Increase number of iteration or change starting values.")
        if (quick) {
            return(opt$estimate)
        }
        ## Calculate std.err:
        pp <- rep(NA,length(coefname)); names(pp) <- coefname
        pp.idx <- NULL
        if (!is.null(names(opt$estimate))) {
            pp[names(opt$estimate)] <- opt$estimate
            pp.idx <- na.omit(match(coefname,names(opt$estimate)))
        } else {
            pp[] <- opt$estimate
            pp.idx <- seq(length(pp))
        }
        ## TODO:
        ## if (length(pp.idx)!=length(pp)) {
        ##     pp <- rep(NA,length(coefname)); names(pp) <- coefname
        ##     pp[] <- opt$estimate
        ##     pp.idx <- seq(length(pp))
        ## }

        suppressWarnings(mom <- tryCatch(modelVar(x, pp, data=data),error=function(x)NULL))
        if (NoOptim) {
            asVar <- matrix(NA,ncol=length(pp),nrow=length(pp))
        } else {

            if (messages>1) message("\nCalculating asymptotic variance...\n")
            asVarFun  <- paste0(estimator, "_variance", ".lvm")

            if (!exists(asVarFun)) {
                if (is.null(myInfo)) {
                    if (!is.null(myGrad))
                        myInfo <- function(pp)
                            numDeriv::jacobian(myGrad,pp,method=lava.options()$Dmethod)
                    else
                        myInfo <- function(pp)
                            numDeriv::hessian(myObj,pp)
                }
                I <- myInfo(opt$estimate)
                asVar <- tryCatch(Inverse(I),
                                  error=function(e) matrix(NA, length(opt$estimate), length(opt$estimate)))
            } else {
                asVar <- tryCatch(do.call(asVarFun,
                                          list(x=x,p=opt$estimate,data=data,opt=opt)),
                                  error=function(e) matrix(NA, length(opt$estimate), length(opt$estimate)))
            }

            if (any(is.na(asVar))) {
                warning("Problems with asymptotic variance matrix. Possibly non-singular information matrix!")
            }
            if (!is.null(attributes(asVar)$pseudo) && attributes(asVar)$pseudo) {
                warning("Near-singular covariance matrix, using pseudo-inverse!")
            }
            diag(asVar)[diag(asVar)==0] <- NA
        }

        mycoef <- matrix(NA,nrow=nparall,ncol=4)
        mycoef[pp.idx,1] <- opt$estimate ## Will be finished during post.hooks

        res <- list(model=x, call=cl, coef=mycoef,
                    vcov=asVar, mu=mu, S=S, ##A=A, P=P,
                    model0=mymodel, ## Random slope hack
                    estimator=estimator, opt=opt,expar=x$expar,
                    data=list(model.frame=data, S=S, mu=mu,
                              C=mom$C, v=mom$v, n=n,
                              m=length(latent(x)), k=length(index(x)$manifest), data2=data2),
                    weights=weights, data2=data2,
                    cluster=id,
                    pp.idx=pp.idx,
                    graph=NULL, control=Optim)

        class(res) <- myclass

        myhooks <- gethook("post.hooks")
        for (f in myhooks) {
            res0 <- do.call(f,list(x=res))
            if (!is.null(res0))
                res <- res0
        }

        if(graph) {
            res <- edgelabels(res,type="est")
        }

        return(res)
    }

###}}} estimate.lvm

##' Function to compute the Scheffe corrected confidence
##' interval for the regression line
##'
##' @title Calculate simultaneous confidence limits by Scheffe's method
##' @param model Linear model
##' @param newdata new data frame
##' @param level confidence level (0.95)
##' @export
##' @examples
##' x <- rnorm(100)
##' d <- data.frame(y=rnorm(length(x),x),x=x)
##' l <- lm(y~x,d)
##' plot(y~x,d)
##' abline(l)
##' d0 <- data.frame(x=seq(-5,5,length.out=100))
##' d1 <- cbind(d0,predict(l,newdata=d0,interval="confidence"))
##' d2 <- cbind(d0,scheffe(l,d0))
##' lines(lwr~x,d1,lty=2,col="red")
##' lines(upr~x,d1,lty=2,col="red")
##' lines(lwr~x,d2,lty=2,col="blue")
##' lines(upr~x,d2,lty=2,col="blue")
scheffe <- function(model,newdata=model.frame(model),level=0.95) {
    df <- model$df.residual
    p <- model$rank
    alpha <- 1-level
    ## Scheffe value uses 1-tailed F critical value
    scheffe.crit <- sqrt(p*qf(1-alpha,p,df))
    ci <- predict(model,newdata,interval="confidence",level=level)
    delta <- scheffe.crit/qt(1-alpha/2,df)
    ci[,2] <- ci[,1] -(ci[,1]-ci[,2])*delta
    ci[,3] <- ci[,1] +(ci[,3]-ci[,1])*delta
    return(ci)
}

Moments <- function(x,p,data,conditional=TRUE,...) {

}

##' @export
`moments` <-
  function(x,...) UseMethod("moments")

##' @export
moments.lvmfit <- function(x, p=pars(x),...) moments(Model(x),p=p,...)

##' @export
moments.lvm.missing <- function(x, p=pars(x), ...) {
    idx <- match(coef(Model(x)),names(coef(x)))
    moments.lvmfit(x,p=p[idx],...)
}


##' @export
moments.lvm <- function(x, p, debug=FALSE, conditional=FALSE, data=NULL, latent=FALSE, ...) {
### p: model-parameters as obtained from e.g. 'startvalues'.
###       (vector of regression parameters and variance parameters)
### meanpar: mean-parameters (optional)

  ii <- index(x)
  pp <- modelPar(x,p)
  AP <- with(pp, matrices(x,p,meanpar=meanpar,epars=p2,data=data,...))
  P <- AP$P
  v <- AP$v
  if (!is.null(v)) {
    names(v) <- ii$vars
  }

  J <- ii$J
  if (conditional) {
    J <- ii$Jy
    if (latent) {
       J <- diag(nrow=length(ii$vars))[sort(c(ii$endo.idx,ii$eta.idx)),,drop=FALSE]
    }
    px <- ii$px
    exo <- exogenous(x)
    ## if (missing(row)) { 
    v <- rbind(v) %x% cbind(rep(1,nrow(data)))
    if (length(ii$exo.idx)>0) {
        v[,ii$exo.idx] <- as.matrix(data[,exo])
    }
    ## } else {
    ##     if (!is.null(v)) 
    ##         v[exo] <- as.numeric(data[row,exo])
    ## }
    P <-  px%*% tcrossprod(P, px)
  }

  Im <- diag(nrow=nrow(AP$A))
  if (ii$sparse) {
    IAi <- with(AP, as(Inverse(Im-t(A)),"sparseMatrix"))
    ##IAi <- as(solve(Matrix::Diagonal(nrow(A))-t(A)),"sparseMatrix")
    G <- as(J%*%IAi,"sparseMatrix")
  } else {
    IAi <- Inverse(Im-t(AP$A))
    G <- J%*%IAi
  }

  xi <- NULL
  if (!is.null(v)) {
      xi <- v%*%t(G) ## Model-specific mean vector
  }
  Cfull <- as.matrix(IAi %*% tcrossprod(P,IAi))
  C <- as.matrix(J %*% tcrossprod(Cfull,J))

  return(list(Cfull=Cfull, C=C, v=v, e=AP$e, xi=xi, A=AP$A, P=P, IAi=IAi, J=J, G=G, npar=ii$npar, npar.reg=ii$npar.reg, npar.mean=ii$npar.mean, npar.ex=ii$npar.ex, parval=AP$parval, constrain.idx=AP$constrain.idx, constrainpar=AP$constrainpar))
}

Moments <- function(x,p,data,conditional=TRUE,...) {

}

##' @export
`moments` <-
  function(x,...) UseMethod("moments")

##' @export
moments.lvmfit <- function(x, p=pars(x),...) moments(Model(x),p=p,...)

##' @export
moments.lvm.missing <- function(x, p=pars(x), ...) {
    idx <- match(coef(Model(x)),names(coef(x)))
    moments.lvmfit(x,p=p[idx],...)
}


##' @export
moments.lvm <- function(x, p, debug=FALSE, conditional=FALSE, data=NULL, latent=FALSE, ...) {
  ## p: model-parameters as obtained from e.g. 'startvalues'.
  ##       (vector of regression parameters and variance parameters)
  ## meanpar: mean-parameters (optional)

  ii <- index(x)
  pp <- modelPar(x,p)
  AP <- with(pp, matrices(x,p,meanpar=meanpar,epars=p2,data=data,...))
  P <- AP$P
  v <- AP$v
  if (!is.null(v)) {
    names(v) <- ii$vars
  }

  J <- ii$J
  if (conditional) {
    J <- ii$Jy
    if (latent) {
       J <- diag(nrow=length(ii$vars))[sort(c(ii$endo.idx,ii$eta.idx)),,drop=FALSE]
    }
    px <- ii$px
    exo <- exogenous(x)
    ## if (missing(row)) { 
    v <- rbind(v) %x% cbind(rep(1,nrow(data)))
    if (length(ii$exo.idx)>0) {
        v[,ii$exo.idx] <- as.matrix(data[,exo])
    }
    ## } else {
    ##     if (!is.null(v)) 
    ##         v[exo] <- as.numeric(data[row,exo])
    ## }
    P <-  px%*% tcrossprod(P, px)
  }

  Im <- diag(nrow=nrow(AP$A))
  if (ii$sparse) {
    IAi <- with(AP, as(Inverse(Im-t(A)),"sparseMatrix"))
    ##IAi <- as(solve(Matrix::Diagonal(nrow(A))-t(A)),"sparseMatrix")
    G <- as(J%*%IAi,"sparseMatrix")
  } else {
    IAi <- Inverse(Im-t(AP$A))
    G <- J%*%IAi
  }

  xi <- NULL
  if (!is.null(v)) {
      xi <- v%*%t(G) ## Model-specific mean vector
  }
  Cfull <- as.matrix(IAi %*% tcrossprod(P,IAi))
  C <- as.matrix(J %*% tcrossprod(Cfull,J))

  return(list(Cfull=Cfull, C=C, v=v, e=AP$e, xi=xi, A=AP$A, P=P, IAi=IAi, J=J, G=G, npar=ii$npar, npar.reg=ii$npar.reg, npar.mean=ii$npar.mean, npar.ex=ii$npar.ex, parval=AP$parval, constrain.idx=AP$constrain.idx, constrainpar=AP$constrainpar))
}

##' Extract information matrix
##'
##' Extract the (observed or expected) information matrix from a model object.
##' @param x Model object
##' @param ... Additional arguments to lower level functions
##' @return Information matrix or its inverse (depending on arguments).
##' @export
`information` <-
    function(x,...) UseMethod("information")

##' @export
information.lvm <- function(x,p,n,type=ifelse(model=="gaussian",
                                       c("E","hessian","varS","outer","sandwich","robust","num"),"outer"),
                     data,weights=NULL,
                     data2=NULL,
                     model="gaussian",
                     method=lava.options()$Dmethod,
                     inverse=FALSE, pinv=TRUE,
                     score=TRUE,...) {
    if (missing(n))
        n <- NROW(data)
    if (type[1]%in%c("sandwich","robust")) {
        cl <- match.call()
        cl$inverse <- !inverse
        cl$type <- "outer"
        A <- eval.parent(cl)
        cl$inverse <- !(cl$inverse)
        cl$type <- ifelse(type[1]=="sandwich","E","hessian")
        B <- eval.parent(cl)
        return(B%*%A%*%B)
    }
    if (type[1]%in%c("num","hessian","obs","observed")  | (type[1]%in%c("E","hessian") & model!="gaussian")) {
        myf <- function(p0) score(x, p=p0, model=model,data=data, weights=weights,data2=data2,indiv=FALSE,n=n) ##...)
        I <- -numDeriv::jacobian(myf,p,method=method)
        res <- (I+t(I))/2 # Symmetric result
        if (inverse) {
            if (pinv)
                iI <- Inverse(res)
            else
                iI <- solve(res)
            return(iI)
        }
        return(res)
    }
    if (type[1]=="varS" | type[1]=="outer") {
        S <- score(x,p=p,data=na.omit(data),model=model,weights=weights,data2=data2,indiv=TRUE,...)
        res <- t(S)%*%S
        if (inverse) {
            if (pinv)
                iI <- Inverse(res)
            else
                iI <- solve(res)
            return(iI)
        }
        attributes(res)$grad <- colSums(S)
        return(res)
    }

    if (n>1) {
        xfix <- colnames(data)[(colnames(data)%in%parlabels(x,exo=TRUE))]
        xconstrain <- intersect(unlist(lapply(constrain(x),function(z) attributes(z)$args)),manifest(x))

        if (length(xfix)>0 | length(xconstrain)>0) { ## Random slopes!
            x0 <- x
            if (length(xfix)>0) {
                nrow <- length(vars(x))
                xpos <- lapply(xfix,function(y) which(regfix(x)$labels==y))
                colpos <- lapply(xpos, function(y) ceiling(y/nrow))
                rowpos <- lapply(xpos, function(y) (y-1)%%nrow+1)
                myfix <- list(var=xfix, col=colpos, row=rowpos)
                for (i in seq_along(myfix$var))
                    for (j in seq_along(myfix$col[[i]]))
                        regfix(x0, from=vars(x0)[myfix$row[[i]]][j],to=vars(x0)[myfix$col[[i]]][j]) <-
                            data[1,myfix$var[[i]]]
                index(x0) <- reindex(x0,zeroones=TRUE,deriv=TRUE)
            }
            pp <- modelPar(x0,p)
            p0 <- with(pp, c(meanpar,p,p2))
            myfun <- function(ii) {
                if (length(xfix)>0)
                    for (i in seq_along(myfix$var)) {
                        for (j in seq_along(myfix$col[[i]])) {
                            index(x0)$A[cbind(myfix$row[[i]],myfix$col[[i]])] <- data[ii,myfix$var[[i]]]
                        }
                    }
                ww <- NULL
                if (!is.null(weights))
                    ww <- weights[ii,]
                return(information(x0,p=p,n=1,type=type,weights=ww,data=data[ii,]))
            }
            L <- lapply(seq_len(nrow(data)),function(y) myfun(y))
            val <- apply(array(unlist(L),dim=c(length(p0),length(p0),nrow(data))),c(1,2),sum)
            if (inverse) {
                if (pinv)
                    iI <- Inverse(val)
                else
                    iI <- solve(val)
                return(iI)
            }
            return(val)
        }
    }

    if (!is.null(weights) && is.matrix(weights)) {
        L <- lapply(seq_len(nrow(weights)),function(y) information(x,p=p,n=1,type=type,weights=weights[y,]))
        val <- apply(array(unlist(L),dim=c(length(p),length(p),nrow(weights))),c(1,2),sum)
        if (inverse) {
            if (pinv)
                iI <- Inverse(val)
            else
                iI <- solve(val)
            return(iI)
        }
        return(val)
    }
    mp <- moments(x,p,data=data)
    pp <- modelPar(x,p)
    D <- deriv.lvm(x, meanpar=pp$meanpar, mom=mp, p=p)##, all=length(constrain(x))>0)
    C <- mp$C
    iC <- Inverse(C,det=FALSE, symmetric = TRUE)

    if (is.null(weights)) {
        ##    W <- diag(ncol(iC))
    } else {
        if (length(weights)<ncol(iC)) {
            oldweights <- weights
            weights <- rbind(rep(1,ncol(iC))) ## Ones at exogenous var.
            idx <- index(x)$vars%in%index(x)$exogenous
            print(idx); print(oldweights)
            weights[,idx] <- oldweights
        }
        W <- diag(nrow=as.numeric(weights))
        iW <- W
        diag(iW) <- 1/diag(iW)
    }

    if (is.null(weights)) {
        ## information_Sigma <-  n/2*t(D$dS)%*%((iC)%x%(iC))%*%(D$dS)
        if (lava.options()$devel) {
            information_Sigma <- matrix(0,length(p),length(p))
            imean <- with(index(x)$parBelongsTo,mean)
            information_Sigma[-imean,-imean] <- n/2*t(D$dS[,-imean])%*%kronprod(iC,iC,D$dS[,-imean])
        } else {
            information_Sigma <- n/2*t(D$dS)%*%kronprod(iC,iC,D$dS)
        }
    } else {
        ## information_Sigma <-  n/2*t(D$dS)%*%((iC)%x%(iC%*%W))%*%(D$dS)
        information_Sigma <- n/2*t(D$dS)%*%kronprod(iC,iC%*%W,D$dS)
    }

    dxi <- D$dxi;

    if (!any(dxi>0)) { 
        if (inverse) {
            if (pinv)
                iI <- Inverse(information_Sigma)
            else
                iI <- solve(information_Sigma)
            return(iI)
        }
        return(information_Sigma)
    }
    ii <- index(x)
    if (is.null(weights)) {
        information_mu <- n*t(dxi) %*% (iC) %*% (dxi)
    } else {
        information_mu <- n*t(dxi) %*% (iC%*%W) %*% (dxi)
    }

    if (!(lava.options()$devel)) {
        information <- information_Sigma+information_mu
    } else {
        mparidx <- with(ii$parBelongsTo,c(mean,reg))
        information <- information_Sigma
        information[mparidx,mparidx] <- information[mparidx,mparidx] + information_mu
    }

    if (inverse) {
        if (pinv)
            iI <- Inverse(information, symmetric = TRUE)
        else
            iI <- solve(information)
        return(iI)
    }
    return(information)
}

##' @export
information.lvmfit <- function(x,p=pars(x),n=x$data$n,data=model.frame(x),model=x$estimator,weights=Weights(x),
                        data2=x$data$data2,
                        ...) {
    I <- information(x$model0,p=p,n=n,data=data,model=model,
                     weights=weights,data2=data2,...)
    if (ncol(I)<length(p)) {
        I <- blockdiag(I,matrix(0,length(p)-ncol(I),length(p)-ncol(I)))
    }
    return(I)
}

##' @export
information.lvm.missing <- function(x,
                             p=coef(x), estimator=x$estimator,
                             weights=Weights(x$estimate),
                             ...) {
    information(x$estimate$model0, p=p, model=estimator, weights=weights,...)
}

##' @export
information.multigroupfit <- function(x,p=pars(x), weights=Weights(x), estimator=x$estimator, ...) {
    information(x$model0,p=p, weights=weights, model=estimator ,...)
}

##' @export
information.multigroup <- function(x,data=x$data,weights=NULL,p,indiv=FALSE,...) {
    rm <- procrandomslope(x)
    pp <- with(rm, modelPar(model,p)$p)
    parord <- modelPar(rm$model,seq_len(with(rm$model,npar+npar.mean)))$p
    I <- matrix(0,nrow=length(p),ncol=length(p))
    if (!indiv) {
        for (i in seq_len(x$ngroup))
            I[parord[[i]],parord[[i]]] <- I[parord[[i]],parord[[i]]] + information(x$lvm[[i]],p=pp[[i]],data=data[[i]],weights=weights[[i]],...)
    } else {
        I <- list()
        for (i in seq_len(x$ngroup))
            I <- c(I, list(information(x$lvm[[i]],p=pp[[i]],data=data[[i]],weights=weights[[i]],...)))
    }
    return(I)
}

##' @export
deriv.function <- function(expr, parameter_, ..., parameter_.increment =.Machine$double.xmin) {
    p <- length(parameter_)
    f0 <- expr(parameter_)
    z0 <- numeric(p)
    res <- matrix(NA,nrow=length(f0),ncol=p)
    for (i in seq(p)) {
        z <- z0; z[i] <- parameter_.increment*1i
        res[,i] <- Im(expr(parameter_+z,...))/parameter_.increment
    }
    res       
}

##' @export
deriv.lvm <- function(expr, p, mom, conditional=FALSE, meanpar=TRUE, mu=NULL, S=NULL, second=FALSE, zeroones=FALSE, all=!missing(mom),...) {

    if (missing(mom) & !missing(p)) {
        mom <- modelVar(expr,p,conditional=conditional,...)
        all <- TRUE
        if (mom$npar==length(p))
            meanpar <- NULL
    }

    ii <- index(expr)
    npar.total <- npar <- ii$npar; npar.reg <- ii$npar.reg
    npar.mean <- ifelse(is.null(meanpar),0,ii$npar.mean)
    npar.ex <- ii$npar.ex
    meanpar <- seq_len(npar.mean)
    nn <- expr$parpos

    if (is.null(nn))
    {
        nn <- matrices2(expr, seq_len(npar+npar.mean+npar.ex));
        nn$A[ii$M0!=1] <- 0
        nn$P[ii$P0!=1] <- 0
        nn$v[ii$v0!=1] <- 0
        nn$e[ii$e0!=1] <- 0
    }

    regr.idx <- seq_len(npar.reg) + npar.mean
    var.idx <- seq_len(npar-npar.reg) + (npar.mean + npar.reg)
    mean.idx <- seq_len(npar.mean)
    npar.total <- npar+length(mean.idx)
    epar.idx <- seq_len(npar.ex)+npar.total
    npar.total <- npar.total+length(epar.idx)

    if (zeroones | is.null(ii$dA)) {
        dimA <- length(ii$A)
        if (ii$sparse) { ## Not used yet...
            if (!requireNamespace("Matrix",quietly=TRUE)) stop("package Matrix not available")
            dP <- dA <- Matrix::Matrix(0, nrow=dimA, ncol=npar.total)
        } else {
            dP <- dA <- matrix(0, nrow=dimA, ncol=npar.total)
        }
        if (npar.reg>0) {
            dA[,regr.idx] <- sapply(regr.idx, function(i) izero(which(t(nn$A)==i),nrow(dA)) )
        }
        if (npar>npar.reg) {
            dP[,var.idx] <- sapply(var.idx, function(i) izero(which(nn$P==i),nrow(dA)) )
        }
        res <- list(dA=dA, dP=dP)

        {
            if (ii$sparse) {
                dv <- Matrix::Matrix(0, nrow=length(expr$mean), ncol=npar.total)
            } else {
                dv <- matrix(0, nrow=length(expr$mean), ncol=npar.total)
            }
            if (!is.null(meanpar) & npar.mean>0)
                dv[,mean.idx] <- sapply(mean.idx, function(i) izero(which(nn$v==i),length(expr$mean)) )
            res <- c(res, list(dv=dv))
        }
    } else {
        res <- with(ii, list(dA=dA, dP=dP, dv=dv))
        for (pp in nn$parval) {
            res$dP[attributes(pp)$cov.idx,pp] <- 1
            res$dv[attributes(pp)$m.idx,pp] <- 1
        }
    }

    if (!all) return(res)
    ## Non-linear constraints:
    cname <- constrainpar <- c()
    if (!missing(p) && length(index(expr)$constrain.par)>0) {
        for (pp in index(expr)$constrain.par) {
            myc <- constrain(expr)[[pp]]
            if (!is.null(myc)) {
                parval <- mom$parval
                vals <- c(parval,constrainpar,mom$v,mom$e)[attributes(myc)$args]
                fval <- try(myc(unlist(vals)),silent=TRUE)
                fmat <- inherits(fval,"try-error")
                if (fmat) fval <- myc(rbind(unlist(vals)))
                if (!is.null(attributes(fval)$grad)) {
                    if (fmat) {
                        Gr <- attributes(fval)$grad(rbind(unlist(vals)))
                    } else {
                        Gr <- attributes(fval)$grad(unlist(vals))
                    }
                } else {
                    if (fmat) {
                        Gr <- as.numeric(numDeriv::jacobian(myc, rbind(unlist(vals))))
                    } else {
                        Gr <- as.numeric(numDeriv::jacobian(myc, unlist(vals)))
                    }
                }
                mat.idx <- mom$constrain.idx[[pp]]
                cname <- c(cname,pp)
                attributes(fval)$grad <- Gr
                attributes(fval)$vals <- vals
                constrainpar <- c(constrainpar,list(fval)); names(constrainpar) <- cname

                for (jj in seq_len(length(vals))) {
                    allpars <- c(nn$A[attributes(vals[[jj]])$reg.idx[1]],
                                 nn$P[attributes(vals[[jj]])$cov.idx[1]],
                                 nn$v[attributes(vals[[jj]])$m.idx[1]],
                                 nn$e[attributes(vals[[jj]])$e.idx[1]]
                                 )
                    if (!is.null(mat.idx$cov.idx))
                        res$dP[mat.idx$cov.idx,allpars] <- Gr[jj]
                    if (!is.null(mat.idx$reg.idx))
                        res$dA[mat.idx$reg.tidx,allpars] <- Gr[jj]
                    if (!is.null(res$dv) & !is.null(mat.idx$m.idx))
                        res$dv[mat.idx$m.idx,allpars] <- Gr[jj]
                }
            }
        }
    }

    if (is.null(ii$Kkk)) {
        nobs <- nrow(mom$J)
        ii$Ik <- diag(nrow=nobs)
        ii$Im <- diag(nrow=ncol(ii$A))
        ##    ii$Kkk <- commutation(nobs,sparse=FALSE)
    }

    K <- nobs
    ## if (N>10) {
    if (!lava.options()$devel) {
        dG <- with(mom, kronprod(t(IAi),G,res$dA))
        G3 <- with(mom, kronprod(G,G,res$dP))
        GP <- with(mom,G%*%P)
        G1 <- with(mom, kronprod(GP,ii$Ik,dG))
        G2 <- G1[as.vector(matrix(seq_len(K^2),K,byrow=TRUE)),]
        dS <- G1+G2+G3
    } else {
        dG <- with(mom, kronprod(t(IAi),G,res$dA[,ii$parBelongsTo$reg,drop=FALSE]))
        G3 <- with(mom, kronprod(G,G,res$dP[,ii$parBelongsTo$cov,drop=FALSE]))
        GP <- with(mom,G%*%P)
        G1 <- with(mom, kronprod(GP,ii$Ik,dG))
        G2 <- G1[as.vector(matrix(seq_len(K^2),K,byrow=TRUE)),]
        dS <- matrix(0,nrow=nrow(G1),ncol=ncol(res$dA))
        dS[,ii$parBelongsTo$reg] <- G1+G2;  dS[,ii$parBelongsTo$cov] <- G3
    }

    res <- c(res, list(dG=dG, dS=dS))

    if (!is.null(mom$v)) {
        if (lava.options()$devel) {
            dG <- with(mom, kronprod(t(IAi),G,res$dA[,with(ii$parBelongsTo,c(mean,reg)),drop=FALSE]))
        }
        dxi <-
            with(mom, kronprod(rbind(v),dG))
        ##  with(mom, kronprod(rbind(v),ii$Ik,dG))
        if (is.matrix(mom$v) && nrow(mom$v)>1) {
            ## reorder
            k <- nrow(dxi)/nrow(mom$v)
            idx0 <- seq(nrow(mom$v))*k-k+1
            idx <- unlist(lapply(1:k,function(x) idx0+x-1))
            dxi <- dxi[idx,,drop=FALSE]
        }

        if (!is.null(res$dv)) {
            if (!(lava.options()$devel)) {
                if (is.matrix(mom$v) && nrow(mom$v)>1) {
                    dxi <- dxi + (mom$G%*%res$dv)%x%cbind(rep(1,nrow(mom$v)))
                } else {
                    dxi <- dxi+ mom$G%*%res$dv
                }
            } else {
                dxi <- dxi+ mom$G%*%res$dv[,with(ii$parBelongsTo,c(mean,reg))]
            }
        }
        res <- c(res, list(dxi=dxi))
        if (!is.null(mu)) {
            muv <- rbind(mu-mom$xi)
            dT <- suppressMessages(-t(ii$Ik%x%muv + muv%x%ii$Ik) %*% dxi)
            res <- c(res, list(dT=dT))
        }
    }


    if (second) {
        k <- nrow(ii$A)
        K <- ii$Kkk ## commutation(k,k)
        I <- ii$Ik ## diag(k)
        I2 <- diag(nrow=k*k)
        d2S1 <-  t(
        (I %x% K %x% I) %*% (
            ( I2 %x% as.vector(mom$G) )%*% dG +
            ( as.vector(mom$P) %x% I2 )%*% (dP)
        ) %*% t(dG)
        )
        d2S2 <- K%*%d2S1
        d2S3 <- t(
        (I %x% K %x% I) %*% (
            ( I2 %x% as.vector(mom$G) )%*% dG +
            ( as.vector(mom$G) %x% I2 )%*% dG
        ) %*% t(dP)
        )
        vec.d2S <- d2S1+d2S2+d2S3
        res <- c(res, list(d2vecS=vec.d2S))
    }

    return(res)
}

##' Add time-varying covariate effects to model
##'
##' @title Time-dependent parameters
##' @param object Model
##' @param formula Formula with rhs specifying time-varying covariates
##' @param rate Optional rate parameters. If given as a vector this
##' parameter is interpreted as the raw (baseline-)rates within each
##' time interval defined by \code{timecut}.  If given as a matrix the
##' parameters are interpreted as log-rates (and log-rate-ratios for
##' the time-varying covariates defined in the formula).
##' @param timecut Time intervals
##' @param type Type of model (default piecewise constant intensity)
##' @param ... Additional arguments to lower level functions
##' @author Klaus K. Holst
##' @aliases timedep timedep<-
##' @export
##' @examples
##'
##' ## Piecewise constant hazard
##' m <- lvm(y~1)
##' m <- timedep(m,y~1,timecut=c(0,5),rate=c(0.5,0.3))
##'
##' \dontrun{
##' d <- sim(m,1e4); d$status <- TRUE
##' dd <- mets::lifetable(Surv(y,status)~1,data=d,breaks=c(0,5,10));
##' exp(coef(glm(events ~ offset(log(atrisk)) + -1 + interval, dd, family=poisson)))
##' }
##'
##'
##' ## Piecewise constant hazard and time-varying effect of z1
##' m <- lvm(y~1)
##' distribution(m,~z1) <- Binary.lvm(0.5)
##' R <- log(cbind(c(0.2,0.7,0.9),c(0.5,0.3,0.3)))
##' m <- timedep(m,y~z1,timecut=c(0,3,5),rate=R)
##'
##' \dontrun{
##' d <- sim(m,1e4); d$status <- TRUE
##' dd <- mets::lifetable(Surv(y,status)~z1,data=d,breaks=c(0,3,5,Inf));
##' exp(coef(glm(events ~ offset(log(atrisk)) + -1 + interval+z1:interval, dd, family=poisson)))
##' }
##'
##'
##'
##' ## Explicit simulation of time-varying effects
##' m <- lvm(y~1)
##' distribution(m,~z1) <- Binary.lvm(0.5)
##' distribution(m,~z2) <- binomial.lvm(p=0.5)
##' #variance(m,~m1+m2) <- 0
##' #regression(m,m1[m1:0] ~ z1) <- log(0.5)
##' #regression(m,m2[m2:0] ~ z1) <- log(0.3)
##' regression(m,m1 ~ z1,variance=0) <- log(0.5)
##' regression(m,m2 ~ z1,variance=0) <- log(0.3)
##' intercept(m,~m1+m2) <- c(-0.5,0)
##' m <- timedep(m,y~m1+m2,timecut=c(0,5))
##'
##' \dontrun{
##' d <- sim(m,1e5); d$status <- TRUE
##' dd <- mets::lifetable(Surv(y,status)~z1,data=d,breaks=c(0,5,Inf))
##' exp(coef(glm(events ~ offset(log(atrisk)) + -1 + interval + interval:z1, dd, family=poisson)))
##' }
timedep <- function(object,formula,rate,timecut,type="coxExponential.lvm",...) {
    if (missing(timecut)) stop("'timecut' needed")
    ##if (inherits(formula,"formula"))
    ff <- getoutcome(formula)
    simvars <- attributes(ff)$x
    if (is.null(object$attributes$simvar)) {
        object$attributes$simvar <- list(simvars)
        names(object$attributes$simvar) <- ff
        object$attributes$timedep <- object$attributes$simvar
    } else {
        object$attributes$simvar[[ff]] <- simvars
        object$attributes$timedep[[ff]] <- simvars
    }
    if (missing(rate)) rate <- rep(1,length(timecut))
    
    args <- list(timecut=timecut,rate=rate,...)
    covariance(object,ff) <- 1
    distribution(object,ff) <- do.call(type,args)
    return(object)
}

##' @export
"timedep<-" <- function(object,...,value) {
    timedep(object,value,...)
}

##' @export
print.fix <- function(x,exo=FALSE,...) {
    switch(attributes(x)$type,
           reg = cat("Regression parameters:\n"),
           cov = cat("Covariance parameters:\n"),
           mean = cat("Intercept parameters:\n"))
    M <- linconstrain(x,print=TRUE)
    invisible(x)
}

linconstrain <- function(x,print=TRUE,indent="  ",exo=FALSE,...) {
    idx <- seq_len(attributes(x)$nvar)
    idx0 <- setdiff(idx,attributes(x)$exo.idx)
    if (!exo & attributes(x)$type!="reg")
        idx <- idx0
    if (attributes(x)$type=="mean") {
        if (length(idx)>0){
            M <- rbind(unlist(x[idx]))
            rownames(M) <- ""
            M[is.na(M)] <- "*"
        } else {
            M <- NULL
        }
    } else {
        if (length(x$rel)==0) {
            M <- NULL
        } else {
            M <- x$rel[idx,idx,drop=FALSE]
            M[M==0] <- NA
            M[M==1] <- "*"
            M[which(!is.na(x$labels[idx,idx]))] <- x$labels[idx,idx][which(!is.na(x$labels[idx,idx]))]
            M[which(!is.na(x$values[idx,idx]))] <- x$values[idx,idx][which(!is.na(x$values[idx,idx]))]
            if (attributes(x)$type=="reg")
                M <- t(M[,idx0,drop=FALSE])
        }
    }
    if (print) {
        M0 <- M
        if (NROW(M)>0)
            rownames(M0) <- paste(indent,rownames(M))
        print(M0,quote=FALSE,na.print="",...)
    }
    invisible(M)
}

"intfix" <- function(object,...) UseMethod("intfix")

`intfix<-` <- function(object,...,value) UseMethod("intfix<-")

##' Fix mean parameters in 'lvm'-object
##'
##' Define linear constraints on intercept parameters in a \code{lvm}-object.
##'
##'
##' The \code{intercept} function is used to specify linear constraints on the
##' intercept parameters of a latent variable model. As an example we look at
##' the multivariate regression model
##'
##' \deqn{ E(Y_1|X) = \alpha_1 + \beta_1 X} \deqn{ E(Y_2|X) = \alpha_2 + \beta_2
##' X}
##'
##' defined by the call
##'
##' \code{m <- lvm(c(y1,y2) ~ x)}
##'
##' To fix \eqn{\alpha_1=\alpha_2} we call
##'
##' \code{intercept(m) <- c(y1,y2) ~ f(mu)}
##'
##' Fixed parameters can be reset by fixing them to \code{NA}.  For instance to
##' free the parameter restriction of \eqn{Y_1} and at the same time fixing
##' \eqn{\alpha_2=2}, we call
##'
##' \code{intercept(m, ~y1+y2) <- list(NA,2)}
##'
##' Calling \code{intercept} with no additional arguments will return the
##' current intercept restrictions of the \code{lvm}-object.
##'
##' @aliases intercept intercept<- intercept.lvm intercept<-.lvm
##' @param object \code{lvm}-object
##' @param vars character vector of variable names
##' @param value Vector (or list) of parameter values or labels (numeric or
##' character) or a formula defining the linear constraints (see also the
##' \code{regression} or \code{covariance} methods).
##' @param \dots Additional arguments
##' @usage
##' \method{intercept}{lvm}(object, vars, ...) <- value
##' @return
##'
##' A \code{lvm}-object
##' @note
##'
##' Variables will be added to the model if not already present.
##' @author Klaus K. Holst
##' @seealso \code{\link{covariance<-}}, \code{\link{regression<-}},
##' \code{\link{constrain<-}}, \code{\link{parameter<-}},
##' \code{\link{latent<-}}, \code{\link{cancel<-}}, \code{\link{kill<-}}
##' @keywords models regression
##' @export
##' @examples
##'
##'
##' ## A multivariate model
##' m <- lvm(c(y1,y2) ~ f(x1,beta)+x2)
##' regression(m) <- y3 ~ f(x1,beta)
##' intercept(m) <- y1 ~ f(mu)
##' intercept(m, ~y2+y3) <- list(2,"mu")
##' intercept(m) ## Examine intercepts of model (NA translates to free/unique paramete##r)
##'
##'
"intercept" <- function(object,...) UseMethod("intercept")

##' @export
intercept.lvm <- function(object,value,...) {
    if (!missing(value)) {
        intercept(object,...) <- value
        return(object)
    }
    res <- object$mean; attr(res,"type") <- "mean"
    attr(res,"exo.idx") <- index(object)$exo.idx
    attr(res,"nvar") <- length(res)
    class(res) <- "fix"
    return(res)
}

intfix.lvm <- intercept.lvm

##' @export
"intercept<-" <- function(object,...,value) UseMethod("intercept<-")

##' @export
"intercept<-.lvm" <- function(object, vars,...,value) {
    if (!missing(vars) && inherits(value,"formula")) value <- all.vars(value)
    if (inherits(value,"formula")) {
        lhs <- getoutcome(value)
        yy <- decomp.specials(lhs)
        if ((inherits(value[[3]],"logical") && is.na(value[[3]]))) {
            intfix(object,yy) <- NA
            return(object)
        }
        tt <- terms(value)
        xf <- attributes(terms(tt))$term.labels
        res <- lapply(xf,decomp.specials)[[1]]

        myvalue <- char2num(as.list(res))
        myvalue <- lapply(myvalue, function(x) ifelse(x=="NA",NA,x))
        intfix(object,yy) <- myvalue
        object$parpos <- NULL
        return(object)
    }
    if (inherits(vars,"formula")) {
        vars <- all.vars(vars)
    }
    object$mean[vars] <- value
    newindex <- reindex(object)
    object$parpos <- NULL
    index(object)[names(newindex)] <- newindex
    return(object)
}

#' @export
`intfix<-.lvm` <- function(object, ..., value) {
  intercept(object, ...) <- value
  object
}

"covfix" <- function(object,...) UseMethod("covfix")

#' @exportS3Method
covfix.lvm <- function(object,...) {
    res <- list(rel=object$cov, labels=object$covpar, values=object$covfix); attr(res,"type") <- "cov"
    attr(res,"exo.idx") <- index(object)$exo.idx
    attr(res,"nvar") <- NROW(res$rel)
    class(res) <- "fix"
    return(res)
}

`covfix<-` <- function(object,...,value) UseMethod("covfix<-")

#' @export
`covfix<-.lvm` <- function(object, var1, var2=var1, pairwise=FALSE, exo=FALSE, ..., value) {

    if (inherits(var1,"formula")) {
        var1 <- all.vars(var1)
    }
    if (inherits(var2,"formula")) {
        var2 <- all.vars(var2)
    }
    object <- addvar(object,c(var1,var2),reindex=FALSE,...)

    allvars <- c(var1,var2)
    xorg <- exogenous(object)
    exoset <- setdiff(xorg,allvars)

    if (!exo & length(exoset)<length(xorg)) {
        exogenous(object) <- exoset
    }

    if (inherits(value,"formula")) value <- all.vars(value)
    if (pairwise) {
        p <- 0
        K <- length(var1)*(length(var1)-1)/2
        if (length(value)==1)
            value <- rep(value,K)
        if (length(value)!=K) stop("Wrong number of parameters")
        for (i in seq_len(length(var1)-1)) {
            for (j in seq(i+1,length(var1))) {
                p <- p+1
                valp <- char2num(value[[p]])
                if (is.na(value[[p]]) | value[[p]]=="NA") {
                    object$covfix[var1[i],var1[j]] <- object$covpar[var1[i],var1[j]] <- NA
                    object$covfix[var1[j],var1[i]] <- object$covpar[var1[j],var1[i]] <- NA
                }
                else {
                    object$cov[var1[i],var1[j]] <-  object$cov[var1[j],var1[i]] <- 1
                    if (is.numeric(value[[p]]) | !is.na(valp)) {
                        object$covfix[var1[i],var1[j]] <- object$covfix[var1[j],var1[i]] <- valp
                        object$covpar[var1[i],var1[j]] <- object$covpar[var1[j],var1[i]] <- NA
                    } else {
                        object$covpar[var1[i],var1[j]] <- object$covpar[var1[j],var1[i]] <- value[[p]]
                        object$covfix[var1[i],var1[j]] <- object$covfix[var1[j],var1[i]] <- NA
                    }
                }
            }
        }
        newindex <- reindex(object)
        object$parpos <- NULL
        index(object)[names(newindex)] <- newindex
        return(object)
    }


    if (is.null(var2)) {
        if (length(value)==1)
            value <- rep(value,length(var1))
        if (length(value)!=length(var1)) stop("Wrong number of parameters")
        for (i in seq_along(var1)) {
            vali <- char2num(value[[i]])
            if (is.na(value[[i]]) | value[[i]]=="NA") {
                object$covfix[var1[i],var1[i]] <- object$covpar[var1[i],var1[i]] <- NA
            }
            else {
                if (is.numeric(value[[i]]) | !is.na(vali)) {
                    object$covfix[var1[i],var1[i]] <- vali
                    object$covpar[var1[i],var1[i]] <- NA
                } else {
                    object$covfix[var1[i],var1[i]] <- NA
                    object$covpar[var1[i],var1[i]] <- value[[i]]
                }
            }
        }
        newindex <- reindex(object)
        object$parpos <- NULL
        index(object)[names(newindex)] <- newindex
        return(object)
    }

    if (length(var1)==length(var2) & length(var1)==length(value)) {
        p <- 0
        for (i in seq_along(var1)) {
            p <- p+1
            valp <- char2num(value[[p]])
            if (is.na(value[[p]]) | value[[p]]=="NA") {
                object$covfix[var1[i],var2[i]] <- object$covpar[var1[i],var2[i]] <- NA
                object$covfix[var2[i],var1[i]] <- object$covpar[var2[i],var1[i]] <- NA
            }
            else {
                object$cov[var1[i],var2[i]] <-  object$cov[var2[i],var1[i]] <- 1
                if (is.numeric(value[[p]]) | !is.na(valp)) {
                    object$covfix[var1[i],var2[i]] <- object$covfix[var2[i],var1[i]] <- valp
                    object$covpar[var1[i],var2[i]] <- object$covpar[var2[i],var1[i]] <- NA
                } else {
                    object$covpar[var1[i],var2[i]] <- object$covpar[var2[i],var1[i]] <- value[[p]]
                    object$covfix[var1[i],var2[i]] <- object$covfix[var2[i],var1[i]] <- NA
                }
            }
        }
        newindex <- reindex(object)
        object$parpos <- NULL
        index(object)[names(newindex)] <- newindex
        return(object)
    }


    K <- length(var1)*length(var2)
    if (length(value)==1)
        value <- rep(value,K)
    if (length(value)!=K) stop("Wrong number of parameters")

    p <- 0
    for (i in seq_along(var1)) {
        for (j in seq_along(var2)) {
            if (!pairwise | var1[i]!=var2[j]) {
                p <- p+1
                valp <- char2num(value[[p]])
                if (is.na(value[[p]]) | value[[p]]=="NA") {
                    object$covfix[var1[i],var2[j]] <- object$covpar[var1[i],var2[j]] <- NA
                    object$covfix[var2[j],var1[i]] <- object$covpar[var2[j],var1[i]] <- NA
                }
                else {
                    object$cov[var1[i],var2[j]] <-  object$cov[var2[j],var1[i]] <- 1
                    if (is.numeric(value[[p]]) | !is.na(valp)) {
                        object$covfix[var1[i],var2[j]] <- object$covfix[var2[j],var1[i]] <- valp
                        object$covpar[var1[i],var2[j]] <- object$covpar[var2[j],var1[i]] <- NA
                    } else {
                        object$covpar[var1[i],var2[j]] <- object$covpar[var2[j],var1[i]] <- value[[p]]
                        object$covfix[var1[i],var2[j]] <- object$covfix[var2[j],var1[i]] <- NA
                    }
                }
            }
        }
    }
    newindex <- reindex(object)
    object$parpos <- NULL
    index(object)[names(newindex)] <- newindex
    return(object)
}

##' @export
"regfix" <- function(object,...) UseMethod("regfix")

##' @export
regfix.lvm <- function(object,...) {
    res <- list(rel=index(object)$M, labels=object$par, values=object$fix); attr(res,"type") <- "reg"
    attr(res,"exo.idx") <- index(object)$exo.idx
    attr(res,"nvar") <- NROW(res$rel)
    class(res) <- "fix"
    return(res)
}

"regfix<-" <- function(object,...,value) UseMethod("regfix<-")

#' @export
`regfix<-.lvm` <- function(object, to, from=NULL, exo=lava.options()$exogenous, variance, y,x, ..., value) {
    if (!missing(y)) {
      if (inherits(y,"formula")) {
        y <- all.vars(y)
      }
      to <- y
    }
    if (!missing(x)) {
        if (inherits(x,"formula")) x <- all.vars(x)
        from <- x
    }
    if (is.null(to)) stop("variable list needed")

    if (inherits(to,"formula")) {
        val <- procformula(object,to,exo=exo)
        constrain_formula <- to
        object <- val$object
        ys <- val$ys
        xs <- val$xs
        if (!missing(variance))
            covariance(object, ys) <- variance
        to <- ys; from <- xs
    } else {
        object <- addvar(object, c(to, from), reindex=FALSE, ...)
        newexo <- from
        notexo <- to
        curvar <- index(object)$var
        if (exo) {
            oldexo <- exogenous(object)
            newexo <- setdiff(newexo, c(notexo, curvar))
            exogenous(object) <- union(newexo, setdiff(oldexo, notexo))
        }
    }

    if (inherits(value,"formula")) value <- all.vars(value)

    nonlinear_function <- NULL
    if (is.function(value)) {
      vfun <- value
      aname <- setdiff(names(formals(value)), "...")
      if (is.null(to)) { # formula was apparently given as  ~outcome
        to <- from[1]
        from <- NULL
      }
      pname <- parameter(object)
      if (length(from)==0) {
        constrain_formula <- toformula(to, aname)
        from <- setdiff(aname, pname)
        nonlinear_function <- function(...) {
          as.vector(do.call(vfun,
                    structure(as.list(as.data.frame(...)), names=aname)))
        }
      } else {
        nonlinear_function <- function(...) as.vector(vfun(...))
        ##nonlinear_function <- vfun
        from <- setdiff(from, pname)
      }
      object <- addvar(object, c(to, from))
      value <- 0
    }

    if (length(from)==length(to) && length(from)==length(value)
        && is.null(nonlinear_function)) {
        for (i in seq_along(from)) {
            if (object$M[from[i], to[i]]==0) { ## Not adjacent! ##!isAdjacent(Graph(object), from[i], to[i])) {
                object <- regression(object, to=to[i], from=from[i])
            }
            vali <- char2num(value[[i]])
            if (is.na(value[[i]]) | value[[i]]=="NA") {
                object$fix[from[i], to[i]] <- object$par[from[i], to[i]] <- NA
            }
            else {
                if (is.numeric(value[[i]]) | !is.na(vali)) {
                    object$fix[from[i], to[i]] <- vali
                    object$par[from[i], to[i]] <- NA
                } else {
                    object$par[from[i], to[i]] <- value[[i]]
                    object$fix[from[i], to[i]] <- NA
                }
            }
        }
        newindex <- reindex(object)
        object$parpos <- NULL
        index(object)[names(newindex)] <- newindex
        return(object)
    }
    for (i in from) {
        for (j in to) {
            if (object$M[i,j]==0) { ##!isAdjacent(Graph(object), i, j)) {
                object <- regression(object, to=j, from=i)
            }
        }
    }


    K <- length(from)*length(to)
    if (length(value)==1)
        value <- rep(value, K)
    if (length(value)!=K) stop("Wrong number of parameters")

    for (j in seq_along(to)) {
        for (i in seq_along(from)) {
            p <- (j-1)*length(from) + i
            valp <- char2num(value[[p]])
            if (is.na(value[[p]]) | value[[p]]=="NA")
                object$fix[from[i], to[j]] <- object$par[from[i], to[j]] <- NA
            else {
                if (is.numeric(value[[p]]) | !is.na(valp)) {
                    object$fix[from[i], to[j]] <- valp
                    object$par[from[i], to[j]] <- NA
                } else {
                    object$par[from[i], to[j]] <- value[[p]]
                    object$fix[from[i], to[j]] <- NA
                }
            }
        }
    }
    newindex <- reindex(object)
    object$parpos <- NULL
    index(object)[names(newindex)] <- newindex
    index(object) <- reindex(object)

    if (!is.null(nonlinear_function)) {
      constrain(object, constrain_formula, endogenous=FALSE, ...) <- nonlinear_function
    }

    return(object)
}

"parfix<-" <- function(x,...,value) UseMethod("parfix<-")

#' @export
`parfix<-.lvm` <- function(x,idx,...,value) {
    parfix(x,idx,value,...)
}

"parfix" <- function(x,...) UseMethod("parfix")


## m <- lvm(c(y[m:v]~b*x))
## constrain(m,b~a) <- base::identity

#' @exportS3Method
parfix.lvm <- function(x,idx,value,fix=FALSE,...) {
    object <- Model(x)
    if (fix)
        object <- fixsome(object)
    if (length(idx)!=length(value))
        value <- rep(value,length.out=length(idx))
    value <- as.list(value)
    I <- index(object)
    V <- with(I, matrices2(Model(object), seq_len(npar.mean+npar+npar.ex)))
    V$A[I$M0!=1] <- 0; V$P[I$P0!=1] <- 0
    v.fix <- which(V$v%in%idx) ## Intercepts
    vval <- V$v[v.fix]
    v.ord <- match(vval,idx)
    Pval <- V$P[V$P%in%idx] ## Variance/covariance
    P.fix <- which(matrix(V$P%in%idx,nrow=nrow(V$P)),arr.ind=TRUE)
    P.ord <- match(Pval,idx)
    Aval <- V$A[which(V$A%in%idx)] ## Regression parameters
    A.fix <- which(matrix(V$A%in%idx,nrow=nrow(V$A)),arr.ind=TRUE)
    A.ord <- match(Aval,idx)
    e.fix <- which(V$e%in%idx)
    eval <- V$e[e.fix]
    e.ord <- match(eval,idx)
    for (i in seq_len(length(e.fix))) {
        object$exfix[[e.fix[i]]] <- value[[e.ord[i]]]
    }
    for (i in seq_len(length(v.fix))) {
        object$mean[[v.fix[i]]] <- value[[v.ord[i]]]
    }
    for (i in seq_len(nrow(A.fix))) {
        if (is.numeric(value[[ A.ord[i] ]])){
            object$fix[A.fix[i,1],A.fix[i,2]] <- value[[A.ord[i]]]
            object$par[A.fix[i,1],A.fix[i,2]] <- NA
        } else {
            object$par[A.fix[i,1],A.fix[i,2]] <- value[[A.ord[i]]]
            object$fix[A.fix[i,1],A.fix[i,2]] <- NA
        }
    }
    for (i in seq_len(nrow(P.fix))) {
        if (is.numeric(value[[ P.ord[i] ]])) {
            object$covfix[P.fix[i,1],P.fix[i,2]] <- value[[P.ord[i]]]
            object$covpar[P.fix[i,1],P.fix[i,2]] <- NA
        } else {
            object$covpar[P.fix[i,1],P.fix[i,2]] <- value[[P.ord[i]]]
            object$covfix[P.fix[i,1],P.fix[i,2]] <- NA
        }
    }
    newindex <- reindex(object)
    object$parpos <- NULL
    index(object)[names(newindex)] <- newindex
    attributes(object)$fixed <- list(v=v.fix,A=A.fix,P=P.fix,e=e.fix)
    return(object)
}

###{{{ print.fix

##' @export
print.fix <- function(x,exo=FALSE,...) {
    switch(attributes(x)$type,
           reg = cat("Regression parameters:\n"),
           cov = cat("Covariance parameters:\n"),
           mean = cat("Intercept parameters:\n"))
    M <- linconstrain(x,print=TRUE)
    invisible(x)
}

linconstrain <- function(x,print=TRUE,indent="  ",exo=FALSE,...) {
    idx <- seq_len(attributes(x)$nvar)
    idx0 <- setdiff(idx,attributes(x)$exo.idx)
    if (!exo & attributes(x)$type!="reg")
        idx <- idx0
    if (attributes(x)$type=="mean") {
        if (length(idx)>0){
            M <- rbind(unlist(x[idx]))
            rownames(M) <- ""
            M[is.na(M)] <- "*"
        } else {
            M <- NULL
        }
    } else {
        if (length(x$rel)==0) {
            M <- NULL
        } else {
            M <- x$rel[idx,idx,drop=FALSE]
            M[M==0] <- NA
            M[M==1] <- "*"
            M[which(!is.na(x$labels[idx,idx]))] <- x$labels[idx,idx][which(!is.na(x$labels[idx,idx]))]
            M[which(!is.na(x$values[idx,idx]))] <- x$values[idx,idx][which(!is.na(x$values[idx,idx]))]
            if (attributes(x)$type=="reg")
                M <- t(M[,idx0,drop=FALSE])
        }
    }
    if (print) {
        M0 <- M
        if (NROW(M)>0)
            rownames(M0) <- paste(indent,rownames(M))
        print(M0,quote=FALSE,na.print="",...)
    }
    invisible(M)
}

###}}} print.fix

###{{{ intfix

"intfix" <- function(object,...) UseMethod("intfix")

"intfix<-" <- function(object,...,value) UseMethod("intfix<-")

##' Fix mean parameters in 'lvm'-object
##'
##' Define linear constraints on intercept parameters in a \code{lvm}-object.
##'
##'
##' The \code{intercept} function is used to specify linear constraints on the
##' intercept parameters of a latent variable model. As an example we look at
##' the multivariate regression model
##'
##' \deqn{ E(Y_1|X) = \alpha_1 + \beta_1 X} \deqn{ E(Y_2|X) = \alpha_2 + \beta_2
##' X}
##'
##' defined by the call
##'
##' \code{m <- lvm(c(y1,y2) ~ x)}
##'
##' To fix \eqn{\alpha_1=\alpha_2} we call
##'
##' \code{intercept(m) <- c(y1,y2) ~ f(mu)}
##'
##' Fixed parameters can be reset by fixing them to \code{NA}.  For instance to
##' free the parameter restriction of \eqn{Y_1} and at the same time fixing
##' \eqn{\alpha_2=2}, we call
##'
##' \code{intercept(m, ~y1+y2) <- list(NA,2)}
##'
##' Calling \code{intercept} with no additional arguments will return the
##' current intercept restrictions of the \code{lvm}-object.
##'
##' @aliases intercept intercept<- intercept.lvm intercept<-.lvm
##' @param object \code{lvm}-object
##' @param vars character vector of variable names
##' @param value Vector (or list) of parameter values or labels (numeric or
##' character) or a formula defining the linear constraints (see also the
##' \code{regression} or \code{covariance} methods).
##' @param \dots Additional arguments
##' @usage
##' \method{intercept}{lvm}(object, vars, ...) <- value
##' @return
##'
##' A \code{lvm}-object
##' @note
##'
##' Variables will be added to the model if not already present.
##' @author Klaus K. Holst
##' @seealso \code{\link{covariance<-}}, \code{\link{regression<-}},
##' \code{\link{constrain<-}}, \code{\link{parameter<-}},
##' \code{\link{latent<-}}, \code{\link{cancel<-}}, \code{\link{kill<-}}
##' @keywords models regression
##' @export
##' @examples
##'
##'
##' ## A multivariate model
##' m <- lvm(c(y1,y2) ~ f(x1,beta)+x2)
##' regression(m) <- y3 ~ f(x1,beta)
##' intercept(m) <- y1 ~ f(mu)
##' intercept(m, ~y2+y3) <- list(2,"mu")
##' intercept(m) ## Examine intercepts of model (NA translates to free/unique paramete##r)
##'
##'
"intercept" <- function(object,...) UseMethod("intercept")

##' @export
intercept.lvm <- function(object,value,...) {
    if (!missing(value)) {
        intercept(object,...) <- value
        return(object)
    }
    res <- object$mean; attr(res,"type") <- "mean"
    attr(res,"exo.idx") <- index(object)$exo.idx
    attr(res,"nvar") <- length(res)
    class(res) <- "fix"
    return(res)
}

intfix.lvm <- intercept.lvm

##' @export
"intercept<-" <- function(object,...,value) UseMethod("intercept<-")

##' @export
"intercept<-.lvm" <- function(object, vars,...,value) {
    if (!missing(vars) && inherits(value,"formula")) value <- all.vars(value)
    if (inherits(value,"formula")) {
        lhs <- getoutcome(value)
        yy <- decomp.specials(lhs)
        if ((inherits(value[[3]],"logical") && is.na(value[[3]]))) {
            intfix(object,yy) <- NA
            return(object)
        }
        tt <- terms(value)
        xf <- attributes(terms(tt))$term.labels
        res <- lapply(xf,decomp.specials)[[1]]

        myvalue <- char2num(as.list(res))
        myvalue <- lapply(myvalue, function(x) ifelse(x=="NA",NA,x))
        intfix(object,yy) <- myvalue
        object$parpos <- NULL
        return(object)
    }
    if (inherits(vars,"formula")) {
        vars <- all.vars(vars)
    }
    object$mean[vars] <- value
    newindex <- reindex(object)
    object$parpos <- NULL
    index(object)[names(newindex)] <- newindex
    return(object)
}

"intfix<-.lvm" <- function(object, ..., value) {
  intercept(object, ...) <- value
  object
}

###}}} intfix

###{{{ covfix

"covfix" <- function(object,...) UseMethod("covfix")

covfix.lvm <- function(object,...) {
    res <- list(rel=object$cov, labels=object$covpar, values=object$covfix); attr(res,"type") <- "cov"
    attr(res,"exo.idx") <- index(object)$exo.idx
    attr(res,"nvar") <- NROW(res$rel)
    class(res) <- "fix"
    return(res)
}


"covfix<-" <- function(object,...,value) UseMethod("covfix<-")

"covfix<-.lvm" <- function(object, var1, var2=var1, pairwise=FALSE, exo=FALSE, ..., value) {

    if (inherits(var1,"formula")) {
        var1 <- all.vars(var1)
    }
    if (inherits(var2,"formula")) {
        var2 <- all.vars(var2)
    }
    object <- addvar(object,c(var1,var2),reindex=FALSE,...)

    allvars <- c(var1,var2)
    xorg <- exogenous(object)
    exoset <- setdiff(xorg,allvars)

    if (!exo & length(exoset)<length(xorg)) {
        exogenous(object) <- exoset
    }

    if (inherits(value,"formula")) value <- all.vars(value)
    if (pairwise) {
        p <- 0
        K <- length(var1)*(length(var1)-1)/2
        if (length(value)==1)
            value <- rep(value,K)
        if (length(value)!=K) stop("Wrong number of parameters")
        for (i in seq_len(length(var1)-1)) {
            for (j in seq(i+1,length(var1))) {
                p <- p+1
                valp <- char2num(value[[p]])
                if (is.na(value[[p]]) | value[[p]]=="NA") {
                    object$covfix[var1[i],var1[j]] <- object$covpar[var1[i],var1[j]] <- NA
                    object$covfix[var1[j],var1[i]] <- object$covpar[var1[j],var1[i]] <- NA
                }
                else {
                    object$cov[var1[i],var1[j]] <-  object$cov[var1[j],var1[i]] <- 1
                    if (is.numeric(value[[p]]) | !is.na(valp)) {
                        object$covfix[var1[i],var1[j]] <- object$covfix[var1[j],var1[i]] <- valp
                        object$covpar[var1[i],var1[j]] <- object$covpar[var1[j],var1[i]] <- NA
                    } else {
                        object$covpar[var1[i],var1[j]] <- object$covpar[var1[j],var1[i]] <- value[[p]]
                        object$covfix[var1[i],var1[j]] <- object$covfix[var1[j],var1[i]] <- NA
                    }
                }
            }
        }
        newindex <- reindex(object)
        object$parpos <- NULL
        index(object)[names(newindex)] <- newindex
        return(object)
    }


    if (is.null(var2)) {
        if (length(value)==1)
            value <- rep(value,length(var1))
        if (length(value)!=length(var1)) stop("Wrong number of parameters")
        for (i in seq_along(var1)) {
            vali <- char2num(value[[i]])
            if (is.na(value[[i]]) | value[[i]]=="NA") {
                object$covfix[var1[i],var1[i]] <- object$covpar[var1[i],var1[i]] <- NA
            }
            else {
                if (is.numeric(value[[i]]) | !is.na(vali)) {
                    object$covfix[var1[i],var1[i]] <- vali
                    object$covpar[var1[i],var1[i]] <- NA
                } else {
                    object$covfix[var1[i],var1[i]] <- NA
                    object$covpar[var1[i],var1[i]] <- value[[i]]
                }
            }
        }
        newindex <- reindex(object)
        object$parpos <- NULL
        index(object)[names(newindex)] <- newindex
        return(object)
    }

    if (length(var1)==length(var2) & length(var1)==length(value)) {
        p <- 0
        for (i in seq_along(var1)) {
            p <- p+1
            valp <- char2num(value[[p]])
            if (is.na(value[[p]]) | value[[p]]=="NA") {
                object$covfix[var1[i],var2[i]] <- object$covpar[var1[i],var2[i]] <- NA
                object$covfix[var2[i],var1[i]] <- object$covpar[var2[i],var1[i]] <- NA
            }
            else {
                object$cov[var1[i],var2[i]] <-  object$cov[var2[i],var1[i]] <- 1
                if (is.numeric(value[[p]]) | !is.na(valp)) {
                    object$covfix[var1[i],var2[i]] <- object$covfix[var2[i],var1[i]] <- valp
                    object$covpar[var1[i],var2[i]] <- object$covpar[var2[i],var1[i]] <- NA
                } else {
                    object$covpar[var1[i],var2[i]] <- object$covpar[var2[i],var1[i]] <- value[[p]]
                    object$covfix[var1[i],var2[i]] <- object$covfix[var2[i],var1[i]] <- NA
                }
            }
        }
        newindex <- reindex(object)
        object$parpos <- NULL
        index(object)[names(newindex)] <- newindex
        return(object)
    }


    K <- length(var1)*length(var2)
    if (length(value)==1)
        value <- rep(value,K)
    if (length(value)!=K) stop("Wrong number of parameters")

    p <- 0
    for (i in seq_along(var1)) {
        for (j in seq_along(var2)) {
            if (!pairwise | var1[i]!=var2[j]) {
                p <- p+1
                valp <- char2num(value[[p]])
                if (is.na(value[[p]]) | value[[p]]=="NA") {
                    object$covfix[var1[i],var2[j]] <- object$covpar[var1[i],var2[j]] <- NA
                    object$covfix[var2[j],var1[i]] <- object$covpar[var2[j],var1[i]] <- NA
                }
                else {
                    object$cov[var1[i],var2[j]] <-  object$cov[var2[j],var1[i]] <- 1
                    if (is.numeric(value[[p]]) | !is.na(valp)) {
                        object$covfix[var1[i],var2[j]] <- object$covfix[var2[j],var1[i]] <- valp
                        object$covpar[var1[i],var2[j]] <- object$covpar[var2[j],var1[i]] <- NA
                    } else {
                        object$covpar[var1[i],var2[j]] <- object$covpar[var2[j],var1[i]] <- value[[p]]
                        object$covfix[var1[i],var2[j]] <- object$covfix[var2[j],var1[i]] <- NA
                    }
                }
            }
        }
    }
    newindex <- reindex(object)
    object$parpos <- NULL
    index(object)[names(newindex)] <- newindex
    return(object)
}

###}}} covfix

###{{{ regfix

##' @export
"regfix" <- function(object,...) UseMethod("regfix")

##' @export
regfix.lvm <- function(object,...) {
    res <- list(rel=index(object)$M, labels=object$par, values=object$fix); attr(res,"type") <- "reg"
    attr(res,"exo.idx") <- index(object)$exo.idx
    attr(res,"nvar") <- NROW(res$rel)
    class(res) <- "fix"
    return(res)
}

"regfix<-" <- function(object,...,value) UseMethod("regfix<-")

"regfix<-.lvm" <- function(object, to, from=NULL, exo=lava.options()$exogenous, variance, y,x, ..., value) {
    if (!missing(y)) {
      if (inherits(y,"formula")) {
        y <- all.vars(y)
      }
      to <- y
    }
    if (!missing(x)) {
        if (inherits(x,"formula")) x <- all.vars(x)
        from <- x
    }
    if (is.null(to)) stop("variable list needed")

    if (inherits(to,"formula")) {
        val <- procformula(object,to,exo=exo)
        constrain_formula <- to
        object <- val$object
        ys <- val$ys
        xs <- val$xs
        if (!missing(variance))
            covariance(object, ys) <- variance
        to <- ys; from <- xs
    } else {
        object <- addvar(object, c(to, from), reindex=FALSE, ...)
        newexo <- from
        notexo <- to
        curvar <- index(object)$var
        if (exo) {
            oldexo <- exogenous(object)
            newexo <- setdiff(newexo, c(notexo, curvar))
            exogenous(object) <- union(newexo, setdiff(oldexo, notexo))
        }
    }

    if (inherits(value,"formula")) value <- all.vars(value)

    nonlinear_function <- NULL
    if (is.function(value)) {
      vfun <- value
      aname <- setdiff(names(formals(value)), "...")
      if (is.null(to)) { # formula was apparently given as  ~outcome
        to <- from[1]
        from <- NULL
      }
      pname <- parameter(object)
      if (length(from)==0) {
        constrain_formula <- toformula(to, aname)
        from <- setdiff(aname, pname)
        nonlinear_function <- function(...) {
          as.vector(do.call(vfun,
                    structure(as.list(as.data.frame(...)), names=aname)))
        }
      } else {
        nonlinear_function <- function(...) as.vector(vfun(...))
        ##nonlinear_function <- vfun
        from <- setdiff(from, pname)
      }
      object <- addvar(object, c(to, from))
      value <- 0
    }

    if (length(from)==length(to) && length(from)==length(value)
        && is.null(nonlinear_function)) {
        for (i in seq_along(from)) {
            if (object$M[from[i], to[i]]==0) { ## Not adjacent! ##!isAdjacent(Graph(object), from[i], to[i])) {
                object <- regression(object, to=to[i], from=from[i])
            }
            vali <- char2num(value[[i]])
            if (is.na(value[[i]]) | value[[i]]=="NA") {
                object$fix[from[i], to[i]] <- object$par[from[i], to[i]] <- NA
            }
            else {
                if (is.numeric(value[[i]]) | !is.na(vali)) {
                    object$fix[from[i], to[i]] <- vali
                    object$par[from[i], to[i]] <- NA
                } else {
                    object$par[from[i], to[i]] <- value[[i]]
                    object$fix[from[i], to[i]] <- NA
                }
            }
        }
        newindex <- reindex(object)
        object$parpos <- NULL
        index(object)[names(newindex)] <- newindex
        return(object)
    }
    for (i in from) {
        for (j in to) {
            if (object$M[i,j]==0) { ##!isAdjacent(Graph(object), i, j)) {
                object <- regression(object, to=j, from=i)
            }
        }
    }


    K <- length(from)*length(to)
    if (length(value)==1)
        value <- rep(value, K)
    if (length(value)!=K) stop("Wrong number of parameters")

    for (j in seq_along(to)) {
        for (i in seq_along(from)) {
            p <- (j-1)*length(from) + i
            valp <- char2num(value[[p]])
            if (is.na(value[[p]]) | value[[p]]=="NA")
                object$fix[from[i], to[j]] <- object$par[from[i], to[j]] <- NA
            else {
                if (is.numeric(value[[p]]) | !is.na(valp)) {
                    object$fix[from[i], to[j]] <- valp
                    object$par[from[i], to[j]] <- NA
                } else {
                    object$par[from[i], to[j]] <- value[[p]]
                    object$fix[from[i], to[j]] <- NA
                }
            }
        }
    }
    newindex <- reindex(object)
    object$parpos <- NULL
    index(object)[names(newindex)] <- newindex
    index(object) <- reindex(object)

    if (!is.null(nonlinear_function)) {
      constrain(object, constrain_formula, endogenous=FALSE, ...) <- nonlinear_function
    }

    return(object)
}

###}}} regfix

###{{{ parfix

"parfix<-" <- function(x,...,value) UseMethod("parfix<-")

"parfix<-.lvm" <- function(x,idx,...,value) {
    parfix(x,idx,value,...)
}

"parfix" <- function(x,...) UseMethod("parfix")


## m <- lvm(c(y[m:v]~b*x))
## constrain(m,b~a) <- base::identity

parfix.lvm <- function(x,idx,value,fix=FALSE,...) {
    object <- Model(x)
    if (fix)
        object <- fixsome(object)
    if (length(idx)!=length(value))
        value <- rep(value,length.out=length(idx))
    value <- as.list(value)
    I <- index(object)
    V <- with(I, matrices2(Model(object), seq_len(npar.mean+npar+npar.ex)))
    V$A[I$M0!=1] <- 0; V$P[I$P0!=1] <- 0
    v.fix <- which(V$v%in%idx) ## Intercepts
    vval <- V$v[v.fix]
    v.ord <- match(vval,idx)
    Pval <- V$P[V$P%in%idx] ## Variance/covariance
    P.fix <- which(matrix(V$P%in%idx,nrow=nrow(V$P)),arr.ind=TRUE)
    P.ord <- match(Pval,idx)
    Aval <- V$A[which(V$A%in%idx)] ## Regression parameters
    A.fix <- which(matrix(V$A%in%idx,nrow=nrow(V$A)),arr.ind=TRUE)
    A.ord <- match(Aval,idx)
    e.fix <- which(V$e%in%idx)
    eval <- V$e[e.fix]
    e.ord <- match(eval,idx)
    for (i in seq_len(length(e.fix))) {
        object$exfix[[e.fix[i]]] <- value[[e.ord[i]]]
    }
    for (i in seq_len(length(v.fix))) {
        object$mean[[v.fix[i]]] <- value[[v.ord[i]]]
    }
    for (i in seq_len(nrow(A.fix))) {
        if (is.numeric(value[[ A.ord[i] ]])){
            object$fix[A.fix[i,1],A.fix[i,2]] <- value[[A.ord[i]]]
            object$par[A.fix[i,1],A.fix[i,2]] <- NA
        } else {
            object$par[A.fix[i,1],A.fix[i,2]] <- value[[A.ord[i]]]
            object$fix[A.fix[i,1],A.fix[i,2]] <- NA
        }
    }
    for (i in seq_len(nrow(P.fix))) {
        if (is.numeric(value[[ P.ord[i] ]])) {
            object$covfix[P.fix[i,1],P.fix[i,2]] <- value[[P.ord[i]]]
            object$covpar[P.fix[i,1],P.fix[i,2]] <- NA
        } else {
            object$covpar[P.fix[i,1],P.fix[i,2]] <- value[[P.ord[i]]]
            object$covfix[P.fix[i,1],P.fix[i,2]] <- NA
        }
    }
    newindex <- reindex(object)
    object$parpos <- NULL
    index(object)[names(newindex)] <- newindex
    attributes(object)$fixed <- list(v=v.fix,A=A.fix,P=P.fix,e=e.fix)
    return(object)
}

###}}} parfix

##' Two-stage (non-linear) measurement error
##'
##' Two-stage measurement error
##' @param model1 Stage 1 model
##' @param formula Formula specifying observed covariates in stage 2 model
##' @param data data.frame
##' @param predictfun Predictions to be used in stage 2
##' @param id1 Optional id-vector of stage 1
##' @param id2 Optional id-vector of stage 2
##' @param ... Additional arguments to lower level functions
##' @seealso stack.estimate
##' @export
##' @examples
##' m <- lvm(c(y1,y2,y3)~u,c(y3,y4,y5)~v,u~~v,c(u,v)~x)
##' transform(m,u2~u) <- function(x) x^2
##' transform(m,uv~u+v) <- prod
##' regression(m) <- z~u2+u+v+uv+x
##' set.seed(1)
##' d <- sim(m,1000,p=c("u,u"=1))
##'
##' ## Stage 1
##' m1 <- lvm(c(y1[0:s],y2[0:s],y3[0:s])~1*u,c(y3[0:s],y4[0:s],y5[0:s])~1*v,u~b*x,u~~v)
##' latent(m1) <- ~u+v
##' e1 <- estimate(m1,d)
##'
##' pp <- function(mu,var,data,...) {
##'     cbind(u=mu[,"u"],u2=mu[,"u"]^2+var["u","u"],
##'           v=mu[,"v"],uv=mu[,"u"]*mu[,"v"]+var["u","v"])
##' }
##' (e <- measurement.error(e1, z~1+x, data=d, predictfun=pp))
##'
##' ## uu <- seq(-1,1,length.out=100)
##' ## pp <- estimate(e,function(p,...) p["(Intercept)"]+p["u"]*uu+p["u2"]*uu^2)$coefmat
##' if (interactive()) {
##'     plot(e,intercept=TRUE,line=0)
##'     dev.off()
##'     f <- function(p) p[1]+p["u"]*u+p["u2"]*u^2
##'     u <- seq(-1,1,length.out=100)
##'     plot(e, f, data=data.frame(u), ylim=c(-.5,2.5))
##' }
measurement.error <- function(model1, formula, data=parent.frame(),
                       predictfun=function(mu,var,data,...) mu[,1]^2+var[1],
                       id1, id2, ...) {
    if (!inherits(model1,c("lvmfit","lvm.mixture"))) stop("Expected lava object ('lvmfit','lvm.mixture',...)")
    if (missing(formula)) stop("formula needed for stage two (right-hand side additional covariates)")
    p1 <- coef(model1,full=TRUE)
    uhat <- function(p=p1) {
        P <- predict(model1,p=p,x=manifest(model1))
        cbind(predictfun(P,attributes(P)$cond.var,data))
    }
    if (missing(id1)) id1 <- seq(nrow(model.frame(model1)))
    if (missing(id2)) id2 <- seq(nrow(model.frame(model1)))
    if (!inherits(model1,"estimate"))
      e1 <- estimate(NULL, coef=p1, id=id1,
                     IC=IC(model1))
    u <- uhat()
    X0 <- model.matrix(formula, data)
    Y <- model.frame(formula,data)[,1]
    X <- cbind(X0,u)
    stage.two <- lm(Y~-1+X)
    names(stage.two$coefficients) <- colnames(X)
    if (!inherits(stage.two,"estimate"))
        e2 <- estimate(stage.two, id=id2)
    U <- function(alpha=p1,beta=coef(stage.two)) {
        X <- cbind(X0,uhat(alpha))
        r <- (Y-X%*%beta)/summary(stage.two)$sigma^2
        apply(X,2,function(x) sum(x*r))
    }
    ## Ia <- -numDeriv::jacobian(function(p) U(p),p1)
    stacked <- stack(e1,e2,U=U)
    class(stacked) <- NULL
    res <- c(stacked,
             list(naive=e2,lm=coef(summary(stage.two)),
                  fun=predictfun))
    structure(res,class=c("measurement.error","estimate"))
}

goodmankruskal_gamma <- function(P,...) {
    nr <- nrow(P); nc <- ncol(P)
    Pconc <- 0
    for (i in seq_len(nr-1)) {
        h <- seq(i+1,nr)
        for (j in seq_len(nc-1)) {
                k <- seq(j+1,nc)
                Pconc <- Pconc+2*P[i,j]*sum(P[h,k])
            }
    }
    Pdisc <- 0
    for (i in seq_len(nr-1)) {
        h <- seq(i+1,nr)
        for (j in (seq_len(nc-1)+1)) {
            k <- seq(1,j-1)
            Pdisc <- Pdisc+2*P[i,j]*sum(P[h,k])
        }
    }
    list(C=Pconc,D=Pdisc,gamma=(Pconc-Pdisc)/(Pconc+Pdisc))
}


##' @export
gkgamma <- function(x,data=parent.frame(),strata=NULL,all=FALSE,IC=TRUE,...) {
    if (inherits(x,"formula")) {
        xf <- getoutcome(x,sep="|")
        xx <- attr(xf,"x")
        if (length(xx)==0) stop("Not a valid formula")
        yx <- update(as.formula(paste0(xf,"~.")),xx[[1]])
        if (length(xx)>1) {
            strata <- interaction(model.frame(xx[[2]],data=data))
            x <- yx
        } else {
            x <- model.frame(yx,data=data)
        }
    }
    if (!is.null(strata)) {
        dd <- split(data,strata)
        gam <- lapply(dd,function(d,...) gkgamma(x,data=d,...),
                      ...,
                      IC=TRUE,
                      keep=1:2)
        mgam <- Reduce(function(x,y,...) merge(x,y,...),gam)
        ps <- estimate(multinomial(strata),data=data,...)
        mgam <- merge(mgam,ps)
        psi <- 2*length(gam)+seq(length(coef(ps)))
        res <- estimate(mgam,function(p,...) {
            k <- length(p)/3
            cd <- lapply(seq(k),function(x) p[(1:2)+2*(x-1)])
            dif <- unlist(lapply(cd,function(x) x[1]-x[2]))
            tot <- unlist(lapply(cd,function(x) x[1]+x[2]))
            gam <- dif/tot ## Conditional gammas given Z=z
            px2 <- p[psi]^2
            pgamma <- sum(dif*px2)/sum(tot*px2)
            #weights <- px2*tot/sum(px2*tot)
            #pgamma <- sum(weights*gam)
            c(gam,pgamma=pgamma)
        },labels=c(paste0("\u03b3:",names(dd)),"pgamma"),
        IC=IC)
        if (!IC) {
            for (i in seq_along(gam))
                gam[[i]][c("IC","id")] <- NULL
        }
        homtest <- estimate(res,lava::contr(seq_along(gam),length(gam)+1),IC=FALSE)
        attributes(res) <- c(attributes(res),
                             list(class=c("gkgamma","estimate"),
                                  cl=match.call(),
                                  strata=gam,
                                  homtest=homtest))
        return(res)
    }
    if (is.table(x) || is.data.frame(x) || is.matrix(x)) {
        x <- multinomial(x)
    }
    if (!inherits(x,"multinomial")) stop("Expected table, data.frame or multinomial object")
    structure(estimate(x,function(p) {
        P <- x$position; P[] <- p[x$position]
        goodmankruskal_gamma(P)
    },IC=IC,data=data,...),
    cl=match.call(),
    class=c("gkgamma","estimate"))
}

##' @export
print.gkgamma <- function(x,call=TRUE,...) {
    if (call) {
        cat("Call: ")
        print(attr(x,"cl"))
        print(cli::rule())
    }
    n <- x$n

    if (!is.null(attr(x,"strata"))) {
        cat("Strata:\n\n")
        for (i in seq_along(attr(x,"strata"))) {
            with(attributes(x), cat(paste0(names(strata)[i],
                                           " (n=",strata[[i]]$n,
                                           if (strata[[i]]$ncluster<strata[[i]]$n) paste0(",clusters=",strata[[i]]$ncluster),
                                           "):\n",sep="")))
            e <- attr(x,"strata")[[i]]
            print.estimate(e,type=0)
            cat("\n")
        }
        print(cli::rule())
        cat("\n")
        n <- sum(unlist(lapply(attr(x,"strata"),"[[","n")))
    }
    k <- x$ncluster
    if (!is.null(n) && !is.null(k) && k<n) {
        cat("n = ",n,", clusters = ",k,"\n\n",sep="")
    } else {
        if (!is.null(n)) {
            cat("n = ",n,"\n\n",sep="")
        } else if (!is.null(k)) {
            cat("n = ",k,"\n\n",sep="")
        }
    }
    if (!is.null(attr(x,"strata"))) {
        cat("Gamma coefficient:\n\n")
    }
    class(x) <- "estimate"
    print(x)
    ## if (!is.null(attr(x,"homtest"))) {
    ##     print(cli::rule())
    ##     cat("Homogeneity test:\n\n")
    ##     with(attr(x,"homtest")$compare,
    ##          cat("\u03c7\u00b2 = ",statistic,
    ##              ", df = ",parameter,
    ##              ", p-value = ",p.value,"\n",sep=""))
    ## }
    invisible(x)
}

##' @export
"sim" <- function(x,...) UseMethod("sim")

##' Monte Carlo simulation
##'
##' Applies a function repeatedly for a specified number of replications or over
##' a list/data.frame with plot and summary methods for summarizing the Monte
##' Carlo experiment. Can be parallelized via the future package (use the
##' [future::plan()] function).
##' @export
##' @param x function or 'sim' object
##' @param R Number of replications or data.frame with parameters
##' @param f Optional function (i.e., if x is a matrix)
##' @param colnames Optional column names
##' @param seed (optional) Seed (needed with cl=TRUE)
##' @param args (optional) list of named arguments passed to (mc)mapply
##' @param iter If TRUE the iteration number is passed as first argument to
##'   (mc)mapply
##' @param mc.cores Optional number of cores. Will use parallel::mcmapply
##'   instead of future
##' @param progressr.message Optional message for the progressr progress-bar
##' @param estimate.index If return object inherits from `estimate` then only
##'   these column indices are extracted (estimate, se, lower, upper, p-val)
##' @param ... Additional arguments to [future.apply::future_mapply()]
##' @aliases sim sim.default as.sim
##' @seealso [summary.sim()] [plot.sim()] [sim.lvm()]
##' @details To parallelize the calculation use the [future::plan()] function
##'   (e.g., `future::plan(multisession())` to distribute the calculations over
##'   the `R` replications on all available cores). The output is controlled via
##'   the progressr package (e.g., `progressr::handlers(global=TRUE)` to enable
##'   progress information).
##' @examples
##' m <- lvm(y~x+e)
##' distribution(m,~y) <- 0
##' distribution(m,~x) <- uniform.lvm(a=-1.1,b=1.1)
##' transform(m,e~x) <- function(x) (1*x^4)*rnorm(length(x),sd=1)
##'
##' onerun <- function(iter=NULL,...,n=2e3,b0=1,idx=2) {
##'     d <- sim(m,n,p=c("y~x"=b0))
##'     l <- lm(y~x,d)
##'     res <- c(coef(summary(l))[idx,1:2],
##'              confint(l)[idx,],
##'              estimate(l,only.coef=TRUE)[idx,2:4])
##'     names(res) <- c("Estimate","Model.se","Model.lo","Model.hi",
##'                     "Sandwich.se","Sandwich.lo","Sandwich.hi")
##'     res
##' }
##' val <- sim(onerun,R=10,b0=1)
##' val
##'
##' val <- sim(val,R=40,b0=1) ## append results
##' summary(val,estimate=c(1,1),confint=c(3,4,6,7),true=c(1,1))
##'
##' summary(val,estimate=c(1,1),se=c(2,5),names=c("Model","Sandwich"))
##' summary(val,estimate=c(1,1),se=c(2,5),true=c(1,1),
##'         names=c("Model","Sandwich"),confint=TRUE)
##'
##' if (interactive()) {
##'     plot(val,estimate=1,c(2,5),true=1,
##'          names=c("Model","Sandwich"),polygon=FALSE)
##'     plot(val,estimate=c(1,1),se=c(2,5),main=NULL,
##'          true=c(1,1),names=c("Model","Sandwich"),
##'          line.lwd=1,col=c("gray20","gray60"),
##'          rug=FALSE)
##'     plot(val,estimate=c(1,1),se=c(2,5),true=c(1,1),
##'          names=c("Model","Sandwich"))
##' }
##'
##' f <- function(a=1, b=1) {
##'   rep(a*b, 5)
##' }
##' R <- Expand(a=1:3, b=1:3)
##' sim(f, R)
##' sim(function(a,b) f(a,b), 3, args=c(a=5,b=5))
##' sim(function(iter=1,a=5,b=5) iter*f(a,b), iter=TRUE, R=5)
sim.default <- function(x = NULL, R = 100, f = NULL,
                        colnames = NULL,
                        seed = NULL, args = list(),
                        iter = FALSE, mc.cores,
                        progressr.message = NULL,
                        estimate.index = 1:2,
                        ...) {
  stm <- proc.time()
  oldtm <- rep(0, 5)
  if (!exists(".Random.seed", envir = .GlobalEnv, inherits = FALSE)) {
    runif(1)
  }
  if (is.null(seed)) {
    RNGstate <- get(".Random.seed", envir = .GlobalEnv)
  } else {
    R.seed <- get(".Random.seed", envir = .GlobalEnv)
    set.seed(seed)
    RNGstate <- structure(seed, kind = as.list(RNGkind()))
    on.exit(assign(".Random.seed", R.seed, envir = .GlobalEnv))
  }
  olddata <- NULL
  dots <- list(...)
  mycall <- match.call(expand.dots = FALSE)
  if (inherits(x, c("data.frame", "matrix"))) olddata <- x
  if (inherits(x, "sim")) {
    oldtm <- attr(x, "time")
    oldcall <- attr(x, "call")
    x <- attr(x, "f")
    if (!is.null(f)) x <- f
    ex <- oldcall[["..."]]
    for (nn in setdiff(names(ex), names(dots))) {
      dots[[nn]] <- ex[[nn]]
      val <- list(ex[[nn]])
      names(val) <- nn
      mycall[["..."]] <- c(mycall[["..."]], list(val))
    }
  } else {
    if (!is.null(f)) x <- f
    if (!is.function(x)) stop("Expected a function or 'sim' object.")
  }
  if (is.null(x)) stop("Must give new function argument 'f'.")
  # `par.index`: list for storing position of estimate, se, confint when
  # the returned simulation object is an `estimate` object
  par.index <- list()
  res <- val <- NULL
  on.exit({
    if (is.null(colnames) && !is.null(val)) {
      if (is.matrix(val[[1]])) {
        colnames <- base::colnames(val[[1]])
      } else {
        colnames <- names(val[[1]])
      }
    }
    base::colnames(res) <- colnames
    if (!is.null(olddata)) res <- rbind(olddata, res)
    attr(res, "call") <- mycall
    attr(res, "f") <- x
    cls <- ifelse(is.data.frame(res), "data.frame", "matrix")
    class(res) <- c("sim", cls)
    attr(res, "time") <- proc.time() - stm + oldtm
    attr(res, "par.index") <- par.index
    return(res)
  })
  if (inherits(R, c("matrix", "data.frame")) || length(R) > 1) {
    if (is.list(R)) {
      parval <- R
      ## list of parameters for each iteration
    } else if (inherits(R, c("matrix", "data.frame"))) {
      parval <- as.data.frame(R)
      names(parval) <- colnames(R)
      R <- NROW(parval)
    }
  } else {
    parval <- as.data.frame(1:R)
    names(parval) <- NULL
  }
  estimate.index <- sort(estimate.index)
  if (!all(estimate.index %in% 1:5)) {
    stop("wrong estimate.index")
  }
  repl <- NROW(parval)
  pb <- progressr::progressor(steps = repl)
  robx <- function(iter__, ...) {
    if (!is.null(progressr.message)) {
      pb(message = progressr.message(...))
    } else {
      pb()
    }
    res <- tryCatch(x(...), error = function(e) NA)
    is_estimate <- inherits(res, c("estimate", "targeted"))
    if (is_estimate) {
      idx <- intersect(seq_len(5L), estimate.index)
      cmat <- lava::parameter(res)[, idx, drop=FALSE]
      res <- as.vector(cmat)
      cn <- colnames(cmat)
      nam <- c()
      for (i in idx) {
        newn <- rownames(cmat)
        if (i > 1L) newn <- paste(cn[i], newn, sep=".")
        nam <- c(nam, newn)
      }
      names(res) <- nam
    }
    return(structure(res, "estimate" = is_estimate))
  }
  if (iter || !is.data.frame(parval)) {
    formals(robx)[[1]] <- NULL
  }

  if (is.data.frame(parval)) {
    pp <- c(
      as.list(parval), dots,
      list(FUN = robx, SIMPLIFY = FALSE, MoreArgs = as.list(args))
    )
  } else { ## parameters as a list
    for (i in seq_along(parval)) {
      parval[i] <- c(parval[i], dots, args)
    }
    pp <- c(
      list(parval),
      list(FUN = robx, SIMPLIFY = FALSE)
    )
  }
  if (is.null(pp$future.seed)) {
    pp$future.seed <- TRUE
  }
  if (!missing(mc.cores)) {
    pp$future.seed <- NULL
    pp$mc.cores <- mc.cores
  }

  if (!missing(mc.cores)) {
    if (is.null(mc.cores)) {
      val <- do.call(mapply, pp)
    } else {
      val <- do.call(parallel::mcmapply, pp)
    }
  } else {
    val <- do.call(future.apply::future_mapply, pp)
  }
  res <- do.call(rbind, val)
  if (is.null(res)) {
    res <- matrix(NA, ncol=length(val[[1]]), nrow=repl)
  }
  if (attr(val[[1]], "estimate")) {
    if (ncol(res) %% length(estimate.index) != 0L) {
      warning("unexpected number of columns")
    }
    idx <- seq_len(ncol(res))
    np <- ncol(res) / length(estimate.index)
    if (1 %in% estimate.index) {
      par.index[["estimate"]] <- seq_len(np)
      idx <- setdiff(idx, par.index[["estimate"]])
    }
    if (2 %in% estimate.index) {
      par.index[["se"]] <- idx[seq_len(np)]
      idx <- setdiff(idx, par.index[["se"]])
    }
    if (3 %in% estimate.index) {
      par.index[["lower"]] <- idx[seq_len(np)]
      idx <- setdiff(idx, par.index[["lower"]])
    }
    if (4 %in% estimate.index) {
      par.index[["upper"]] <- idx[seq_len(np)]
      idx <- setdiff(idx, par.index[["upper"]])
    }
    if (5 %in% estimate.index) {
      par.index[["pval"]] <- idx[seq_len(np)]
      idx <- setdiff(idx, par.index[["pval"]])
    }
  }
  res
}

##' @export
cbind.sim <- function(x, ...) {
  res <- cbind(as.data.frame(x), ...)
  as.sim(res)
}

##' @export
rbind.sim <- function(x, ...) {
  res <- rbind(as.data.frame(x), ...)
  as.sim(res)
}

##' @export
as.vector.sim <- function(x, mode="any") {
  as.vector(x[,,drop=TRUE], mode=mode)
}

##' @export
as.matrix.sim <- function(x, ...) {
  if (inherits(x, "data.frame")) {
    return(as.matrix(as.data.frame(x)))
  }
  class(x) <- "matrix"
  attr(x, "call") <- NULL
  attr(x, "f") <- NULL
  attr(x, "time") <- NULL
  x
}

##' @export
"[.sim" <- function(x, i, j, drop = FALSE) {
  atr <- attributes(x)
  if (!is.null(dim(x))) {
    class(x) <- class(x)[2]
  } else {
    class(x) <- class(x)[-1]
  }
  x <- NextMethod("[", drop=drop)
  atr.keep <- c("call", "time")
  if (missing(j)) atr.keep <- c(atr.keep, "f")
  attributes(x)[atr.keep] <- atr[atr.keep]
  if (!drop) class(x) <- c("sim", class(x))
  return(x)
}

##' @export
"as.sim" <- function (object, name, ...) {
  if (is.vector(object)) {
    cl <- ifelse(inherits(class(object), "data.frame"), "data.frame", "matrix")
    object <- structure(cbind(object), class=c("sim", cl))
    if (!missing(name)) colnames(object) <- name
    return(object)
  }
  structure(object, class=c("sim", class(object)))
}

Time <- function(sec,print=FALSE,...) {
    h <- sec%/%3600
    m0 <- (sec%%3600)
    m <- m0%/%60
    s <- m0%%60
    res <- c(h=h, m=m, s=s)
    if (print) {
        if (h>0) cat(h, "h ", sep="")
        if (m>0) cat(m, "m ", sep="")
        cat(s, "s", sep="")
        return(invisible(res))
    }
    return(res)
}


##' Generic print method
##'
##' Nicer print method for tabular data. Falls back to standard print method for
##' all other data types.
##' @export
##' @param x object to print
##' @param n number of rows to show from top and bottom of tabular data
##' @param digits precision
##' @param ... additional arguments to print method
Print <- function(x, n=5,
                  digits=max(3, getOption("digits")-3), ...) {
    mat <- !is.null(dim(x))
    if (!mat) {
      if (is.vector(x)) {
          x <- cbind(x)
          colnames(x) <- ""
      } else {
        print(x, ...)
        return(invisible(x))
      }
    } 
    if (is.null(rownames(x))) {
        rownames(x) <- seq(nrow(x))
    }
    sep <- rbind("---"=rep('', ncol(x)))
    if (n<1) {
        print(x, quote=FALSE, digits=digits, ...)
    } else {
      if (NROW(x)<=(2*n)) {
        hd <- base::format(x, digits=digits, ...)
        print(hd, quote=FALSE, ...)
      } else {
        hd <- base::format(utils::head(x, n), digits=digits, ...)
        tl <- base::format(utils::tail(x, n), digits=digits, ...)
        print(rbind(base::as.matrix(hd), sep, base::as.matrix(tl)),
              quote=FALSE, ...)
      }
    }
    invisible(x)
}

##' @export
print.sim <- function(x, ...) {
    s <- summary(x, minimal=TRUE, ...)
    attr(x, "f") <- attr(x, "call") <- NULL
    Print(x, ...)
    cat("\n")
    if (nrow(x)>1) print(s, extra=FALSE, ...)
    return(invisible(x))
}

##' @export
print.summary.sim <- function(x,
                              group=list(c("^mean$","^sd$","^se$","^se/sd$","^coverage"),
                                   c("^min$","^[0-9.]+%$","^max$"),
                                   c("^na$","^missing$"),
                                   c("^true$","^bias$","^rmse$")),
                      lower.case=TRUE,
                      na.print="",
                      digits = max(3, getOption("digits") - 2),
                      quote=FALSE,
                      time=TRUE,
                      extra=TRUE,
                      ...) {c
    if (extra) {
        cat(attr(x,"n")," replications",sep="")
        if (time && !is.null(attr(x,"time"))) {
            cat("\t\t\t\t\tTime: ")
            Time(attr(x,"time")["elapsed"],print=TRUE)
        }
        cat("\n\n")
    }

    nn <- rownames(x)
    if (lower.case)  nn <- tolower(nn)
    gg <- lapply(group,
                 function(x) unlist(lapply(x,function(v) grep(v,nn))))
    gg <- c(gg,list(setdiff(seq_along(nn),unlist(gg))))

    x0 <- c()
    ng <- length(gg)
    for (i in seq(ng)) {
        x0 <- rbind(x0, x[gg[[i]],,drop=FALSE],
        { if(i<ng && length(gg[[i+1]])>0) NA})
    }

    print(structure(x0,class="matrix")[,,drop=FALSE],digits=digits,quote=quote,na.print=na.print,...)
    if (extra) cat("\n")
    invisible(x)
}


##' Summary method for 'sim' objects
##'
##' Summary method for 'sim' objects
##' @export
##' @export summary.sim
##' @param object sim object
##' @param estimate (optional) columns with estimates
##' @param se (optional) columns with standard error estimates
##' @param confint (optional) list of pairs of columns with confidence limits
##' @param true (optional) vector of true parameter values
##' @param fun (optional) summary function
##' @param names (optional) names of estimates
##' @param unique.names if TRUE, unique.names will be applied to column names
##' @param minimal if TRUE, minimal summary will be returned
##' @param level confidence level (0.95)
##' @param quantiles quantiles (0,0.025,0.5,0.975,1)
##' @param df degrees of freedom in t-distribution used for constructing CIs
##'   (default Gaussian approximation)
##' @param ... additional levels to lower-level functions
summary.sim <- function(object,estimate=NULL,se=NULL,
                confint,
                true=NULL,
                fun,names=NULL,unique.names=TRUE,minimal=FALSE,
                level=0.95,
                quantiles=c(0,.025,0.5,.975,1),
                df = Inf,
                ...) {
    if (is.list(estimate)) {
        est <- estimate
        if (is.null(names)) names <- base::names(est)
        estimate <- c()
        nse  <- is.null(se)
        ntrue <- is.null(true)
        elen <- unlist(lapply(est,length))
        est <- lapply(est, function(e) c(e, rep(NA,max(elen)-length(e))))
        for (e in est) {            
            estimate <- c(estimate,e[1])
            if (length(e)>1 && nse) se <- c(se,e[2])
            if (length(e)>2 && ntrue) true <- c(true,e[3])
        }
        cl <- match.call()
        cl[c("estimate","se","true","names")] <- list(estimate,se,true,names)
    }
    par.index <- attr(object, "par.index")
    if (is.null(estimate) && length(par.index) > 0) {
      estimate <- par.index[["estimate"]]
    }
    np <- length(estimate)
    if (is.null(se) && length(par.index) > 0) {
      se <- par.index[["se"]]
    }
    if (missing(confint)) {
      lo <- par.index[["lower"]]
      up <- par.index[["upper"]]
      if (length(par.index) > 0 &&
          (!is.null(lo) || !is.null(up))) {
        if (is.null(lo)) lo <- rep(-Inf, np)
        if (is.null(up)) up <- rep(Inf, np)
        confint <- list()
        for (i in seq_len(np)) {
          confint <- c(confint, list(c(lo[i], up[i])))
        }
      } else confint <- !is.null(se)&&!is.null(true)
    }
    if (minimal) {
        fun <- function(x,se,confint,...) {
            res <- c(Mean=mean(x,na.rm=TRUE),
                    SD=sd(x,na.rm=TRUE))
            if (!missing(se) && !is.null(se)) {
                res <- c(res, c(SE=mean(se,na.rm=TRUE)))
                res <- c(res, c("SE/SD"=res[["SE"]]/res[["SD"]]))
            }            
            return(res)
        }
    }
    mfun <- function(x,...) {
        res <- c(mean(x,na.rm=TRUE),
                 sd(x,na.rm=TRUE),
                 if (length(quantiles)>0) quantile(x,quantiles,na.rm=TRUE),
                 mean(is.na(x)))
        if (length(quantiles)>0) {
            nq <- paste0(quantiles*100,"%")
            idx <- which(quantiles==1)
            if (length(idx)>0) nq[idx] <- "Max"
            idx <- which(quantiles==0)
            if (length(idx)>0) nq[idx] <- "Min"
        }
        names(res) <- c("Mean","SD",
                        if (length(quantiles)>0) nq,
                        "Missing")
        res
    }
    tm <- attr(object,"time")
    N <- max(length(estimate),length(se),length(true))    
    if (!is.null(estimate)) estimate <- rep(estimate,length.out=N)
    if (!is.null(se)) se <- rep(se,length.out=N)
    if (!is.null(true)) {
        if (is.null(estimate)) N <- ncol(object)
        true <- rep(true,length.out=N)
    }
    
    if (!is.null(estimate) && is.character(estimate)) {
        estimate <- match(estimate,colnames(object))
    }
    if (is.list(confint)) confint <- unlist(confint)
    if (!missing(fun)) {
        if (!is.null(estimate)) m.est <- object[,estimate,drop=FALSE]
        else m.est <- object
        m.se <- NULL
        if (!is.null(se)) m.se <- object[,se,drop=FALSE]
        m.ci <- NULL
        if (!is.null(confint)) m.ci <- object[,confint,drop=FALSE]
        res <- lapply(seq(ncol(m.est)),
                      function(i,...) fun(m.est[,i,drop=TRUE],
                                          se=m.se[,i,drop=TRUE],
                                          confint=m.ci[,1:2+(i-1)*2],...,
                                          INDEX=i),...)
        res <- matrix(unlist(res),nrow=length(res[[1]]),byrow=FALSE)
        if (is.null(dim(res))) {
            res <- rbind(res)
        }
        if (is.null(rownames(res))) {
            rownames(res) <- names(fun(0,m.se,m.ci,INDEX=1,...))
            if (is.null(rownames(res))) rownames(res) <- rep("",nrow(res))
        }
        if (is.null(colnames(res))) {
            colnames(res) <- colnames(m.est)
        }
        return(structure(res,
                    n=NROW(object),
                    time=tm,
                    class=c("summary.sim","matrix")))
    }

    if (!is.null(estimate)) {
        est <- apply(object[,estimate,drop=FALSE],2,mfun)
    } else {
        est <- apply(object,2,mfun)
    }

    if (!is.null(true)) {
        if (length(true)!=ncol(est)) {
            ##stop("'true' should be of same length as 'estimate'.")
            true <- rep(true,length.out=ncol(estimate))
        }
        est <- rbind(est,
                     rbind(True=true),rbind(Bias=est["Mean",]-true),
                     rbind(RMSE=((est["Mean",]-true)^2+(est["SD",])^2)^.5)
                     )
    }
    if (!is.null(se)) {
        if (is.character(se)) {
            se <- match(se,colnames(object))
        }
        if (length(se)!=ncol(est)) stop("'se' should be of same length as 'estimate'.")
        est <- rbind(est, SE=apply(object[,se,drop=FALSE],2,
                                   function(x) val <- c(mean(x,na.rm=TRUE))))
        est <- rbind(est,"SE/SD"=est["SE",]/est["SD",])
    }
    if (!is.null(confint) && (length(confint)>1 || confint)) {
        if (is.character(confint)) {
            confint <- match(confint,colnames(object))
        }
        if (length(confint)==1 && confint) {
            if (is.null(se)) stop("Supply confidence limits or SE")
            confint <- c()
            pos <- ncol(object)
            for (i in seq_along(estimate)) {
              z <- 1-(1-level)/2
              if (is.na(se[i])) {
                CI <- matrix(NA, nrow=NROW(object), ncol=2)
              } else {
                q <- qt(z, df = df)
                CI <- cbind(object[,estimate[i]] - q*object[,se[i]],
                            object[,estimate[i]] + q*object[,se[i]])
              }
              colnames(CI) <- NULL
              object <- cbind(object,CI)
              confint <- c(confint,pos+1:2)
              pos <- pos+2
            }
        }
        if (length(confint)!=2*length(estimate)) stop("'confint' should be of length 2*length(estimate).")
        Coverage <- c()
        for (i in seq_along(estimate)) {
            Coverage <- c(Coverage,
                          mean((object[,confint[2*(i-1)+1]]<true[i]) & (object[,confint[2*i]]>true[i]),na.rm=TRUE))
        }
        est <- rbind(est,Coverage=Coverage)
    }

    if (!is.null(names)) {
         if (length(names)<ncol(est)) {
            uest <- unique(estimate)
            names <- names[match(estimate,uest)]
        }
        colnames(est) <- names

    }
    if (unique.names && !is.null(colnames(est))) {
        colnames(est) <- make.unique(colnames(est))
    }
    est[is.nan(est)] <- NA

    return(structure(est,
                     n=NROW(object),
                     time=tm,
                     class=c("summary.sim","matrix")))
}

context("composite likelihood")


test_that("complik", {
  m <- lvm(c(y1,y2,y3)~b*x+1*u[0],latent=~u)
  ordinal(m,K=2) <- ~y1+y2+y3
  d <- sim(m,50,seed=1)
  if (requireNamespace("mets", quietly=TRUE)) {
    e1 <- complik(m,d,control=list(trace=0),type="all")
    dl <- mets::fast.reshape(d)
    g <- lme4::glmer(y~x+(1|id), family=binomial("probit"), data=dl)
    # MLE and complik should agree reasonably
    # even though e1 is fitted with different intercepts so the
    # two models are not identical (lmer nested within complik)
    expect_true((coef(e1)["y1~x"]-lme4::fixef(g)["x"])<0.1)
    expect_true((vcov(e1)["y1~x","y1~x"]-vcov(g)["x","x"])<0.025)

    expect_equivalent(vcov(e1), var_ic(IC(e1)))
    expect_true(inherits(logLik(e1), "logLik"))
    expect_true(mean(score(e1)^2)<1e-4)
  }
})

context("Multiple testing")

test_that("closed_testing", {
  m <- lvm()
  regression(m, c(y1, y2, y3, y4) ~ x) <- c(0, 0.25, 0, 0.25)
  regression(m, to=endogenous(m), from="u") <- 1
  variance(m, endogenous(m)) <- 1
  set.seed(1)
  d <- sim(m, 200)
  l1 <- lm(y1~x, d)
  l2 <- lm(y2~x, d)
  l3 <- lm(y3~x, d)
  l4 <- lm(y4~x, d)
  e <- merge(l1, l2, l3, l4, subset = 2)

  adj <- closed_testing(e)

  H1 <- list(
    h1234 = lava::compare(e, contrast = diag(4)),
    h123 = lava::compare(e, contrast = diag(4)[c(1, 2, 3), , drop = FALSE]),
    h124 = lava::compare(e, contrast = diag(4)[c(1, 2, 4), , drop = FALSE]),
    h134 = lava::compare(e, contrast = diag(4)[c(1, 3, 4), , drop = FALSE]),
    h12 = lava::compare(e, contrast = diag(4)[c(1, 2), , drop = FALSE]),
    h13 = lava::compare(e, contrast = diag(4)[c(1, 3), , drop = FALSE]),
    h14 = lava::compare(e, contrast = diag(4)[c(1, 4), , drop = FALSE]),
    h1 = lava::compare(e, contrast = diag(4)[c(1), , drop = FALSE])
  )

  expect_true(adj$raw.pval[[4]] == H1$h1234$p.value)

  expect_true(adj$p.value[1] ==
              max(lapply(H1, function(x) x$p.value) |> unlist()))

  H2 <- list(
    h1234 = lava::compare(e, contrast = diag(4)),
    h234 = lava::compare(e, contrast = diag(4)[c(2, 3, 4), , drop = FALSE]),
    h123 = lava::compare(e, contrast = diag(4)[c(1, 2, 3), , drop = FALSE]),
    h124 = lava::compare(e, contrast = diag(4)[c(1, 2, 4), , drop = FALSE]),
    h12 = lava::compare(e, contrast = diag(4)[c(1, 2), , drop = FALSE]),
    h23 = lava::compare(e, contrast = diag(4)[c(2, 3), , drop = FALSE]),
    h24 = lava::compare(e, contrast = diag(4)[c(2, 4), , drop = FALSE]),
    h2 = lava::compare(e, contrast = diag(4)[c(2), , drop = FALSE])
  )

  expect_true(adj$raw.pval[[4]] == H1$h1234$p.value)

  expect_true(adj$p.value[2] ==
              max(lapply(H2, function(x) x$p.value) |> unlist()))

})

library(future)
context("Generic simulation framework")

f <- function(iter=list(), ...) {
  if (!is.list(iter) || is.null(iter$i)) return(0)
    return(iter$i)
}
R <- list(
  list(i = 2),
  list(i = 4)
)

test_that("sim.default, list in put", {
  res <- sim(f, R)
  testthat::expect_true(length(R) == nrow(res))
  testthat::expect_identical(
    as.vector(res),
    unname(unlist(R))
  )
})

test_that("sim.default with estimate objects", {
  onerun <- function(...) estimate(coef=runif(2),
                                   vcov=diag(runif(2)),
                                   labels=c("a","b"))
  res <- sim(onerun, 100)
  s <- summary(res)
  expect_true(ncol(s) == 2L)
  expect_equivalent(colnames(s), c("a", "b"))
  expect_equivalent(s["SE",], colMeans(res[, c("a.Std.Err", "b.Std.Err")]))
  expect_equivalent(s["SD",], c(sd(res[,"a"]), sd(res[,"b"])))
})

test_that("sim.default exports seed sequences as attribute", {
  foo <- function() runif(1)
  if (requireNamespace("future",quietly=TRUE)) future::plan("sequential")
  result <- sim(foo, R = 5, future.seed = 42L)
  seeds <- attr(result, "seeds")

  expect_true(is.list(seeds))
  expect_equal(length(seeds), 5L)
  expect_true(is.integer(seeds[[1]]))
})

test_that("sim.default exported seeds reproduce results (sequential)", {
  foo <- function() runif(1)
  if (requireNamespace("future",quietly=TRUE)) future::plan("sequential")
  result <- sim(foo, R = 5, future.seed = 42L)
  seeds <- attr(result, "seeds")

  old_seed <- get(".Random.seed", envir = .GlobalEnv)
  on.exit(assign(".Random.seed", old_seed, envir = .GlobalEnv))

  for (i in seq_len(5)) {
    assign(".Random.seed", seeds[[i]], envir = .GlobalEnv)
    expect_equal(foo(), as.numeric(result[i, 1]))
  }
})

test_that("sim.default exported seeds reproduce results (mc.cores = 1)", {
  skip_on_os("windows")
  foo <- function() runif(1)
  result <- sim(foo, R = 5, mc.cores = 1L)
  seeds <- attr(result, "seeds")

  old_seed <- get(".Random.seed", envir = .GlobalEnv)
  on.exit(assign(".Random.seed", old_seed, envir = .GlobalEnv))

  for (i in seq_len(5)) {
    assign(".Random.seed", seeds[[i]], envir = .GlobalEnv)
    expect_equal(foo(), as.numeric(result[i, 1]))
  }
})

test_that("sim.default subsets", {
  onerun <- function(...) estimate(coef=runif(2),
                                   vcov=diag(runif(2)),
                                   labels=c("a","b"))
  res <- sim(onerun, 100, estimate.index = 1)

  expect_true(ncol(res) == 2L)
  expect_true(nrow(res) == 100L)

  expect_true(nrow(
    rbind(res, res)
  ) == 200L)

  expect_true(nrow(
    rbind(res[1:20, 1], res[1:50, 1])
  ) == 70L)

  expect_true(ncol(
    cbind(res[1:50, 1], res[1:50, 2])
  ) == 2L)

  expect_true(ncol(
    cbind(res[1:50, 1:2], res[1:50, 1:2])
  ) == 4L)

})

context("Composite likelihood")


test_that("complik", {
  m <- lvm(c(y1,y2,y3)~b*x+1*u[0],latent=~u)
  ordinal(m,K=2) <- ~y1+y2+y3
  d <- sim(m,50,seed=1)
  if (requireNamespace("mets", quietly=TRUE)) {
    e1 <- complik(m,d,control=list(trace=0),type="all")
    dl <- mets::fast.reshape(d)
    g <- lme4::glmer(y~x+(1|id), family=binomial("probit"), data=dl)
    # MLE and complik should agree reasonably
    # even though e1 is fitted with different intercepts so the
    # two models are not identical (lmer nested within complik)
    expect_true((coef(e1)["y1~x"]-lme4::fixef(g)["x"])<0.1)
    expect_true((vcov(e1)["y1~x","y1~x"]-vcov(g)["x","x"])<0.025)

    expect_equivalent(vcov(e1), var_ic(IC(e1)))
    expect_true(inherits(logLik(e1), "logLik"))
    expect_true(mean(score(e1)^2)<1e-4)
  }
})

library(future)
context("Generic simulation framework")

f <- function(iter=list(), ...) {
  if (!is.list(iter) || is.null(iter$i)) return(0)
    return(iter$i)
}
R <- list(
  list(i = 2),
  list(i = 4)
)

test_that("sim.default, list in put", {
  res <- sim(f, R)
  testthat::expect_true(length(R) == nrow(res))
  testthat::expect_identical(
    as.vector(res),
    unname(unlist(R))
  )
})

test_that("sim.default with estimate objects", {
  onerun <- function(...) estimate(coef=runif(2),
                                   vcov=diag(runif(2)),
                                   labels=c("a","b"))
  res <- sim(onerun, 10)
  s <- summary(res)
  expect_true(ncol(s) == 2L)
  expect_equivalent(colnames(s), c("a", "b"))
  expect_equivalent(s["SE",], colMeans(res[, c("a.Std.Err", "b.Std.Err")]))
  expect_equivalent(s["SD",], c(sd(res[,"a"]), sd(res[,"b"])))
})

test_that("sim.default with summary.estimate objects", {
  onerun <- function(...) {
    estimate(coef = runif(2), vcov = diag(runif(2)), labels = c("a","b")) |>
      summary()
  }
  res <- sim(onerun, 5)
  s <- summary(res)
  expect_true(ncol(s) == 2L)
  expect_equivalent(colnames(s), c("a", "b"))
  expect_equivalent(s["SE",], colMeans(res[, c("a.Std.Err", "b.Std.Err")]))
  expect_equivalent(s["SD",], c(sd(res[,"a"]), sd(res[,"b"])))

  # with concatenated summary.estimate objects
  onerun <- function(...) {
    s1 <- estimate(
      coef = runif(2), vcov = diag(runif(2)), labels = c("a","b")
    ) |> summary()
    s2 <- estimate(
      coef = runif(2), vcov = diag(runif(2)), labels = c("c","d")
    ) |> summary(transform = exp) # blanks Std.Err column in coefmat
    c(s1, s2)
  }
  res <- sim(onerun, 5)
  expect_true(all(is.na(res[, c("c.Std.Err", "d.Std.Err")])))

  s <- summary(res)
  expect_true(ncol(s) == 4L)
  expect_equivalent(colnames(s), c("a", "b", "c", "d"))

  # with concatenated summary.estimate objects + extra
  onerun <- function(...) {
    s1 <- estimate(
      coef = runif(2), vcov = diag(runif(2)), labels = c("a","b")
    ) |> summary()
    c(s1, niter = rpois(1, 10), converged = 1)
  }
  res <- sim(onerun, 5)
  expect_true(all(c("niter", "converged") %in% colnames(res)))
  expect_true(all(res[, "converged"] == 1))
})

test_that("sim.default exports seed sequences as attribute", {
  foo <- function() runif(1)
  if (requireNamespace("future",quietly=TRUE)) future::plan("sequential")
  result <- sim(foo, R = 5, future.seed = 42L)
  seeds <- attr(result, "seeds")

  expect_true(is.list(seeds))
  expect_equal(length(seeds), 5L)
  expect_true(is.integer(seeds[[1]]))
})

test_that("sim.default exported seeds reproduce results (sequential)", {
  foo <- function() runif(1)
  if (requireNamespace("future",quietly=TRUE)) future::plan("sequential")
  result <- sim(foo, R = 5, future.seed = 42L)
  seeds <- attr(result, "seeds")

  old_seed <- get(".Random.seed", envir = .GlobalEnv)
  on.exit(assign(".Random.seed", old_seed, envir = .GlobalEnv))

  for (i in seq_len(5)) {
    assign(".Random.seed", seeds[[i]], envir = .GlobalEnv)
    expect_equal(foo(), as.numeric(result[i, 1]))
  }
})

test_that("sim.default exported seeds reproduce results (mc.cores = 1)", {
  skip_on_os("windows")
  foo <- function() runif(1)
  result <- sim(foo, R = 5, mc.cores = 1L)
  seeds <- attr(result, "seeds")

  old_seed <- get(".Random.seed", envir = .GlobalEnv)
  on.exit(assign(".Random.seed", old_seed, envir = .GlobalEnv))

  for (i in seq_len(5)) {
    assign(".Random.seed", seeds[[i]], envir = .GlobalEnv)
    expect_equal(foo(), as.numeric(result[i, 1]))
  }
})

test_that("sim.default subsets", {
  onerun <- function(...) estimate(coef=runif(2),
                                   vcov=diag(runif(2)),
                                   labels=c("a","b"))
  res <- sim(onerun, 100, estimate.index = 1)

  expect_true(ncol(res) == 2L)
  expect_true(nrow(res) == 100L)

  expect_true(nrow(
    rbind(res, res)
  ) == 200L)

  expect_true(nrow(
    rbind(res[1:20, 1], res[1:50, 1])
  ) == 70L)

  expect_true(ncol(
    cbind(res[1:50, 1], res[1:50, 2])
  ) == 2L)

  expect_true(ncol(
    cbind(res[1:50, 1:2], res[1:50, 1:2])
  ) == 4L)

})

test_that("sim.default with estimate.extra appends extras", {
  onerun <- function(...) {
    e <- estimate(coef = runif(2), vcov = diag(runif(2)),
                  labels = c("a", "b"))
    c(e, converged = 1, niter = sample(5:20, 1))
  }
  res <- sim(onerun, 20)
  pi <- attr(res, "par.index")

  # Correct dimensions: 2 estimates + 2 SE + 2 extras = 6 columns
  expect_equal(ncol(res), 6L)
  expect_equal(nrow(res), 20L)

  # par.index tracks all components

  expect_equal(pi$estimate, 1:2)
  expect_equal(pi$se, 3:4)
  expect_equal(pi$extra, 5:6)

  # Extra columns have correct names
  expect_true("converged" %in% colnames(res))
  expect_true("niter" %in% colnames(res))

  # All converged values are 1
  expect_true(all(res[, "converged"] == 1))
  # niter values are in expected range
  expect_true(all(res[, "niter"] >= 5 & res[, "niter"] <= 20))
})

test_that("summary.sim ignores extra columns", {
  onerun <- function(...) {
    e <- estimate(coef = runif(2), vcov = diag(runif(2)),
                  labels = c("a", "b"))
    c(e, flag = 1)
  }
  res <- sim(onerun, 30)
  s <- summary(res)

  # Summary only covers the estimate columns, not extras
  expect_equal(ncol(s), 2L)
  expect_equal(colnames(s), c("a", "b"))
})

test_that("summary.sim gives NA SE for extras when estimate is explicit", {
  onerun <- function(...) {
    e <- estimate(coef = runif(2), vcov = diag(runif(2)),
                  labels = c("a", "b"))
    c(e, conv = 1)
  }
  res <- sim(onerun, 20)
  # Include extra column in explicit estimate argument
  s <- summary(res, estimate = c(1, 2, 5))

  expect_equal(ncol(s), 3L)
  # SE is NA for the extra column
  expect_true(is.na(s["SE", "conv"]))
  expect_true(is.na(s["SE/SD", "conv"]))
  # SE is computed for the estimate columns
  expect_false(is.na(s["SE", "a"]))
  expect_false(is.na(s["SE", "b"]))
})

context("estimate lvm")

test_that("estimate.lvm cluster", {
  set.seed(1)
  m <- lvm(y ~ x)
  d <- sim(m,100)
  l <- lm(y ~ x, data=d)
  e <- estimate(m, data=d, id=1:nrow(d))
  e2 <- estimate(l, id=1:nrow(d))
  expect_true(mean((vcov(e2)-vcov(e)[c(1,2),c(1,2)])^2)<1e-6)
})

test_that("estimate.list", {
  d <- data.frame(y=rnorm(10), x=rnorm(10))
  l <- lm(y~x, data=d)
  e <- estimate(list(l, l, l))
  cc <- coef(e)
  expect_true(length(cc)==3L)
  expect_equivalent(cc[[1]], coef(l))
  expect_true(inherits(e[[1]], "estimate"))
})

context("estimate.lvm")

test_that("estimate.lvm cluster", {
  set.seed(1)
  m <- lvm(y ~ x)
  d <- sim(m,100)
  l <- lm(y ~ x, data=d)
  e <- estimate(m, data=d, id=1:nrow(d))
  e2 <- estimate(l, id=1:nrow(d))
  expect_true(mean((vcov(e2)-vcov(e)[c(1,2),c(1,2)])^2)<1e-6)
})

test_that("estimate.list", {
  d <- data.frame(y=rnorm(10), x=rnorm(10))
  l <- lm(y~x, data=d)
  e <- estimate(list(l, l, l))
  cc <- coef(e)
  expect_true(length(cc)==3L)
  expect_equivalent(cc[[1]], coef(l))
  expect_true(inherits(e[[1]], "estimate"))
})

context("Inference")

test_that("estimate.default", {
    m <- lvm(c(y1,y2)~x+z,y1~~y2)
##    set.seed(1)
    d <- sim(m,20)

    l1 <- lm(y1~x+z,d)
    l2 <- lm(y2~x+z,d)
    ll <- merge(l1,l2)
    testthat::expect_equivalent(ll$coefmat[,1],c(coef(l1),coef(l2)))

    e1 <- estimate(l1)
    f1 <- estimate(l1,function(x) x^2, use=2)
    testthat::expect_true(coef(l1)["x"]^2==f1$coefmat[1])

    e1b <- estimate(NULL,coef=coef(l1),vcov=vcov(estimate(l1)))
    e1c <- estimate(NULL,coef=coef(l1), IC=IC(l1))
    testthat::expect_equivalent(vcov(e1b),vcov(e1c))
    testthat::expect_equivalent(var_ic(IC(e1)), vcov(e1b))

    f1b <- estimate(e1b,function(x) x^2)
    testthat::expect_equivalent(f1b$coefmat[2,,drop=FALSE],f1$coefmat)

    h1 <- estimate(l1,cbind(0,1,0))
    testthat::expect_true(h1$coefmat[,5]==e1$coefmat["x",5])

    ## GEE
    if (requireNamespace("geepack",quietly=TRUE)) {
        dd <- reshape(d, direction='long', varying=list(c('y1','y2')), v.names='y')
        dd <- dd[order(dd$id),]
        ## dd <- mets::fast.reshape(d)
        l <- lm(y~x+z,dd)
        g1 <- estimate(l,id=dd$id)
        g2 <- geepack::geeglm(y~x+z,id=dd$id,data=dd)
        testthat::expect_equivalent(g1$coefmat[,c(1,2,5)],
                          as.matrix(summary(g2)$coef[,c(1,2,4)]))
    }

    ## Several parameters
    e1d <- estimate(l1, function(x) list("X"=x[2],"Z"=x[3]))
    testthat::expect_equivalent(e1d$coefmat,e1$coefmat[-1,])

    testthat::expect_true(rownames(estimate(l1, function(x) list("X"=x[2],"Z"=x[3]),keep="X")$coefmat)=="X")
    testthat::expect_true(rownames(estimate(l1, labels=c("a"), function(x) list("X"=x[2],"Z"=x[3]),keep="X")$coefmat)=="a")


    a0 <- estimate(l1,function(p,data) p[1]+p[2]*data[,"x"], average=TRUE)
    a1 <- estimate(l1,function(p,data) p[1]+p[2]*data[,"x"]+p[3], average=TRUE)
    a <- merge(a0,a1,labels=c("a0","a1"))
    estimate(a,diff)
    testthat::expect_equivalent(estimate(a,diff)$coefmat,e1$coefmat[3,,drop=FALSE])

})

# Helper function to manually compute Wald statistic
compute_wald <- function(B, p, S, null) {
  z <- (B %*% p - null)
  V <- B %*% S %*% t(B)
  q <- t(z) %*% Inverse(V) %*% z
  return(structure(q[1], df=qr(V)$rank))
}

set.seed(1)
a1 <- estimate(coef = 1,   IC = rnorm(10), id = 1:10, labels = "a1")
a2 <- estimate(coef = 2,   IC = rnorm(10), id = 1:10, labels = "a2")
a3 <- estimate(coef = 3,   IC = rnorm(10), id = 1:10, labels = "a3")
a4 <- estimate(coef = 4,   IC = rnorm(10), id = 1:10, labels = "a4")
a  <- merge(a1, a2)           # 2-dimensional
a3d <- merge(a1, a2, a3)      # 3-dimensional
a4d <- merge(a1, a2, a3, a4)  # 4-dimensional


test_that("summary.estimate compared with estimate", {
  B <- rbind(c(1,-1, 0), c(0, 1,-1), c(1,0,-1))
  null <- c(1,2,3)
  q <- compute_wald(B, coef(a3d), vcov(a3d), null)
  df <- attr(q, "df")
  e1 <- estimate(a3d, contrast=B)
  e1b <- estimate(a3d, contrast=B, null=null)
  e1a <- estimate(e1, contrast=diag(3), null=null)
  expect_equal(unname(df), unname(e1a$compare$parameter))
  expect_equal(unname(df), unname(e1b$compare$parameter))
  expect_true(abs(pchisq(q, df=df, lower.tail=FALSE) - e1a$compare$p.value)<1e-16)
  expect_true(abs(pchisq(q, df=df, lower.tail=FALSE) - e1b$compare$p.value)<1e-16)
  expect_true(abs(q - e1a$compare$statistic) < 1e-9)
  expect_true(abs(q - e1b$compare$statistic) < 1e-9)
  # function spec.
  e2 <- estimate(a3d, function(p) c(p[1]-p[2], p[2]-p[3], p[1]-p[3]))
  e2b <- estimate(a3d, function(p) c(p[1]-p[2], p[2]-p[3], p[1]-p[3]), null=null)
  e2a <- estimate(e2, contrast=diag(3), null=null)
  expect_equal(unname(df), unname(e2a$compare$parameter))
  expect_equal(unname(df), unname(e2b$compare$parameter))
  expect_true(abs(pchisq(q, df=df, lower.tail=FALSE) - e2a$compare$p.value)<1e-16)
  expect_true(abs(pchisq(q, df=df, lower.tail=FALSE) - e2b$compare$p.value)<1e-16)
  expect_true(abs(q - e2a$compare$statistic) < 1e-9)
  expect_true(abs(q - e2b$compare$statistic) < 1e-9)
  # summary
  e3 <- summary(e1, null=null)
  expect_equal(unname(df), unname(e3$compare$parameter))
  expect_true(abs(pchisq(q, df=df, lower.tail=FALSE) - e3$compare$p.value)<1e-16)
  expect_true(abs(q - e3$compare$statistic) < 1e-9)
})

test_that("summary.estimate compared with estimate", {
  B <- rbind(c(2,-1, 0), c(0, 3,-1), c(1,0,-3), c(1,0,0))
  null <- c(1,0,1,0)
  q <- compute_wald(B, coef(a3d), vcov(a3d), null)
  df <- attr(q, "df")
  e1 <- estimate(a3d, contrast=B)
  e1b <- estimate(a3d, contrast=B, null=null)
  e1a <- estimate(e1, contrast=diag(4), null=null)
  expect_equal(unname(df), unname(e1a$compare$parameter))
  expect_equal(unname(df), unname(e1b$compare$parameter))
  expect_true(abs(pchisq(q, df=df, lower.tail=FALSE) - e1a$compare$p.value)<1e-16)
  expect_true(abs(pchisq(q, df=df, lower.tail=FALSE) - e1b$compare$p.value)<1e-16)
  expect_true(abs(q - e1a$compare$statistic) < 1e-9)
  expect_true(abs(q - e1b$compare$statistic) < 1e-9)
  # summary
  e3 <- summary(e1, null=null)
  expect_equal(unname(df), unname(e3$compare$parameter))
  expect_true(abs(pchisq(q, df=df, lower.tail=FALSE) - e3$compare$p.value)<1e-16)
  expect_true(abs(q - e3$compare$statistic) < 1e-9)
})

test_that("1D: Identity contrast, null = 0", {
  B    <- matrix(1, nrow = 1, ncol = 1)
  null <- 0
  e    <- estimate(a1, contrast = B, null = null)
  p <- coef(a1)
  S <- vcov(a1)
  expected <- compute_wald(B, p, S, null)
  expect_equivalent(e$compare$statistic, expected)
})

test_that("1D: Identity contrast, null = 1", {
  B    <- matrix(1, nrow = 1, ncol = 1)
  null <- 1
  e <- estimate(a1, contrast = B, null = null)
  p <- coef(a1)
  S <- vcov(a1)
  expected <- compute_wald(B, p, S, null)
  expect_equivalent(e$compare$statistic, expected)
})

test_that("1D: Scalar contrast (scaling), null = 0", {
  B    <- matrix(2, nrow = 1, ncol = 1)
  null <- 0
  e <- estimate(a1, contrast = B, null = null)
  p <- coef(a1)
  S <- vcov(a1)
  expected <- compute_wald(B, p, S, null)
  expect_equivalent(e$compare$statistic, expected)
})

test_that("2D: Identity contrast, null = c(0, 0)", {
  B    <- diag(2)
  null <- c(0, 0)
  e <- estimate(a, contrast = B, null = null)
  p <- coef(a)
  S <- vcov(a)
  expected <- compute_wald(B, p, S, null)
  expect_equivalent(e$compare$statistic, expected)
})

test_that("2D: Identity contrast, null equals true coefs", {
  B    <- diag(2)
  null <- c(1, 2)
  e <- estimate(a, contrast = B, null = null)
  p <- coef(a)
  S <- vcov(a)
  expected <- compute_wald(B, p, S, null)
  expect_equivalent(e$compare$statistic, expected)
  # Wald statistic should be very small (close to 0)
  # since null = true coefs
  expect_lt(e$compare$statistic, 1e-10)
})

test_that("2D: Identity contrast, null = c(3, 4)", {
  B    <- diag(2)
  null <- c(3, 4)
  e <- estimate(a, contrast = B, null = null)
  p <- coef(a)
  S <- vcov(a)
  expected <- compute_wald(B, p, S, null)
  expect_equivalent(e$compare$statistic, expected)
})

test_that("2D: Difference contrast c(1,-1), null = 0", {
  B    <- matrix(c(1, -1), nrow = 1)
  null <- 0
  e <- estimate(a, contrast = B, null = null)
  p <- coef(a)
  S <- vcov(a)
  expected <- compute_wald(B, p, S, null)
  expect_equivalent(e$compare$statistic, expected)
})

test_that("2D: Difference contrast c(1,-1), null = -1", {
  B    <- matrix(c(1, -1), nrow = 1)
  null <- -1  # Testing H0: a1 - a2 = -1 (true, since 1 - 2 = -1)
  e <- estimate(a, contrast = B, null = null)
  p <- coef(a)
  S <- vcov(a)
  expected <- compute_wald(B, p, S, null)
  expect_equivalent(e$compare$statistic, expected)
  # Wald stat ~ 0 since null equals true difference
  expect_lt(e$compare$statistic, 1e-10)
})

test_that("2D: Sum contrast c(1,1), null = 3", {
  B    <- matrix(c(1, 1), nrow = 1)
  null <- 3  # True sum = 1 + 2 = 3
  e <- estimate(a, contrast = B, null = null)
  p <- coef(a)
  S <- vcov(a)
  expected <- compute_wald(B, p, S, null)
  expect_equivalent(e$compare$statistic, expected)
  # Wald stat ~ 0
  expect_lt(e$compare$statistic, 1e-10)
})

test_that("2D: Scaling contrast, null = c(2, 6)", {
  B    <- 2 * diag(2)
  null <- c(2, 6)  # True: 2*c(1,2) = c(2,4), so null != true
  e <- estimate(a, contrast = B, null = null)
  p <- coef(a)
  S <- vcov(a)
  expected <- compute_wald(B, p, S, null)
  expect_equivalent(e$compare$statistic, expected)
})

test_that("3D: Identity contrast, null = true coefs", {
  B    <- diag(3)
  null <- c(1, 2, 3)
  e <- estimate(a3d, contrast = B, null = null)
  p <- coef(a3d)
  S <- vcov(a3d)
  expected <- compute_wald(B, p, S, null)
  expect_equivalent(e$compare$statistic, expected)
  expect_lt(e$compare$statistic, 1e-10)
})

test_that("3D: Identity contrast, null = c(0, 0, 0)", {
  B    <- diag(3)
  null <- c(0, 0, 0)
  e <- estimate(a3d, contrast = B, null = null)
  p <- coef(a3d)
  S <- vcov(a3d)
  expected <- compute_wald(B, p, S, null)
  expect_equivalent(e$compare$statistic, expected)
})

test_that("3D: Pairwise differences, null = c(-1, -2)", {
  # H0: a1-a2 = -1, a2-a3 = -1 (true: 1-2=-1, 2-3=-1)
  B <- matrix(c(
    1, -1,  0,
    0,  1, -1
  ), nrow = 2, byrow = TRUE)
  null <- c(-1, -1)
  e <- estimate(a3d, contrast = B, null = null)
  p <- coef(a3d)
  S <- vcov(a3d)
  expected <- compute_wald(B, p, S, null)
  expect_equivalent(e$compare$statistic, expected)
  expect_lt(e$compare$statistic, 1e-10)
})

test_that("3D: Single-row contrast, null = 6", {
  # H0: a1 + a2 + a3 = 6 (true: 1+2+3=6)
  B    <- matrix(c(1, 1, 1), nrow = 1)
  null <- 6
  e <- estimate(a3d, contrast = B, null = null)
  p <- coef(a3d)
  S <- vcov(a3d)
  expected <- compute_wald(B, p, S, null)
  expect_equivalent(e$compare$statistic, expected)
  expect_lt(e$compare$statistic, 1e-10)
})

test_that("4D: Identity contrast, null equals true coefs", {
  B    <- diag(4)
  null <- c(1, 2, 3, 4)
  e <- estimate(a4d, contrast = B, null = null)
  p <- coef(a4d)
  S <- vcov(a4d)
  expected <- compute_wald(B, p, S, null)
  expect_equivalent(e$compare$statistic, expected)
  expect_lt(e$compare$statistic, 1e-10)
})

test_that("4D: Non-square contrast (2x4), null = c(0, 0)", {
  # H0: a1 - a2 = 0, a3 - a4 = 0
  B <- matrix(c(
    1, -1,  0,  0,
    0,  0,  1, -1
  ), nrow = 2, byrow = TRUE)
  null <- c(0, 0)
  e <- estimate(a4d, contrast = B, null = null)
  p <- coef(a4d)
  S <- vcov(a4d)
  expected <- compute_wald(B, p, S, null)
  expect_equivalent(e$compare$statistic, expected)
})

test_that("4D: Non-square contrast, null equals true values", {
  # H0: a1-a2 = -1, a3-a4 = -1 (true: 1-2=-1, 3-4=-1)
  B <- matrix(c(
    1, -1,  0,  0,
    0,  0,  1, -1
  ), nrow = 2, byrow = TRUE)
  null <- c(-1, -1)
  e <- estimate(a4d, contrast = B, null = null)
  p <- coef(a4d)
  S <- vcov(a4d)
  expected <- compute_wald(B, p, S, null)
  expect_equivalent(e$compare$statistic, expected)
  expect_lt(e$compare$statistic, 1e-10)
})

test_that("Degrees of freedom equals nrow(B)", {
  B    <- diag(2)
  null <- c(0, 0)
  e    <- estimate(a, contrast = B, null = null)
  expect_identical(unname(e$compare$parameter), nrow(B))
  B    <- matrix(c(1, -1, 0, 0, 1, -1), nrow = 2, byrow = TRUE)
  null <- c(0, 0)
  e    <- estimate(a3d, contrast = B, null = null)
  expect_equal(unname(e$compare$parameter), nrow(B))
})

context("estimate.default id/cluster")

test_that("index.estimate", {
    m <- lvm(c(y1,y2)~x+z, y1~~y2)
    set.seed(1)
    d <- sim(m,20)

    l1 <- lm(y1~x+z, data=d)
    e1 <- estimate(l1)
    testthat::expect_equivalent(index(e1), rownames(d))
    testthat::expect_true(inherits(index(e1), "character"))
    V <- vcov(e1)
    index(e1) <- as.numeric(index(e1))
    expect_equivalent(vcov(e1), V)
    testthat::expect_true(inherits(index(e1), "numeric"))
})

context("estimate.default id/cluster")

test_that("index.estimate", {
    m <- lvm(c(y1,y2)~x+z, y1~~y2)
    set.seed(1)
    d <- sim(m,20)

    l1 <- lm(y1~x+z, data=d)
    e1 <- estimate(l1)
    testthat::expect_equivalent(index(e1), rownames(d))
    testthat::expect_true(inherits(index(e1), "character"))
    V <- vcov(e1)
    index(e1) <- as.numeric(index(e1))
    expect_equivalent(vcov(e1), V)
    testthat::expect_true(inherits(index(e1), "numeric"))
})


test_that("estimate index order", {
  n <- 20
  d <- data.frame(
    y = rnorm(n),
    a = rbinom(n,1,0.5),
    id = 1:n,
    id2 = n:1
  )

  # check that rownames of data.frame are used as default
  g <- glm(y ~ a, data=d)
  testthat::expect_identical(
              rownames(d),
              index(estimate(g))
            )

  # check that order is preserved (not sorted by default)
  d2 <- d[n:1,]
  g2 <- glm(y ~ a, data=d2)
  testthat::expect_identical(
              rev(rownames(d)),
              index(estimate(g2))
            )

  # user supplied id
  testthat::expect_identical(
              d$id,
              index(estimate(g, id=d$id))
            )

  # user supplied id, order preserved
  testthat::expect_identical(
              d$id2,
              index(estimate(g, id=d$id2))
            )

  # check that sort argument works
  testthat::expect_identical(
              sort(rownames(d)),
              index(merge(estimate(g2), sort=TRUE))
            )

  # check that id also works with transformations
  e <- estimate(g,
           function(x) x,
           id = d$id2)
  testthat::expect_identical(
              d$id2,
              index(e)
            )
  e1 <- estimate(g, id=d$id)
  testthat::expect_equivalent(
              IC(e),
              IC(e1),
              )
})

context("Graphics functions")
library("vdiffr")

test_that("color, devcoords", {
  cur <- palette()
  old <- lava:::mypal()
  testthat::expect_equivalent(col2rgb(cur),col2rgb(old))
  testthat::expect_equivalent(col2rgb(palette()),col2rgb(lava:::mypal(set=FALSE)))

  testthat::expect_equivalent(Col("red",0.5),rgb(1,0,0,0.5))
  testthat::expect_equivalent(Col(c("red","blue"),0.5),rgb(c(1,0),c(0,0),c(0,1),0.5))
  testthat::expect_equivalent(Col(c("red","blue"),c(0.2,0.5)),rgb(c(1,0),c(0,0),c(0,1),c(0.2,0.5)))
  testthat::expect_equivalent(Col(rgb(1,0,0),0.5),rgb(1,0,0,0.5))

  plot(0,xlim=c(0,1),ylim=c(0,1),type="n",ann=FALSE,axes=FALSE)
  devc1 <- devcoords()
  par(mar=c(0,0,0,0))
  plot(0,xlim=c(0,1),ylim=c(0,1),type="n",ann=FALSE,axes=FALSE)
  devc2 <- devcoords()
  figx <- c("fig.x1","fig.x2","fig.y1","fig.y2")
  devx <- c("dev.x1","dev.x2","dev.y1","dev.y2")
  testthat::expect_equivalent(devc1[figx],devc2[devx])
})

test_that("plotConf", {
  skip_on_cran()

  set.seed(1)
  x <- rnorm(50)
  y <- rnorm(50,x)
  z <- rbinom(50,1,0.5)
  d <- data.frame(y,z,x)
  l <- lm(y~x*z)
  myplot <- function(...) {
    plotConf(l,var1="x",var2="z",col=c("black","blue"),alpha=0.5,legend=FALSE)
  }
  expect_doppelganger("plotConf1", myplot)
  ##par(mar=c(0,0,0,0))
  ## newd <- data.frame(x=seq(min(x),max(x),length.out=100))
  ## l0 <- lm(y~x,subset(d,z==0))
  ## ci0 <- predict(l0,newdata=newd,interval="confidence")
  ## l1 <- lm(y~x,subset(d,z==1))
  ## ci1 <- predict(l1,newdata=newd,interval="confidence")
  ## plot(y~x,col=c("black","blue")[z+1],pch=16,ylim=c(min(ci0,ci1,y),max(ci0,ci1,y)))
  ## lines(newd$x,ci0[,1],col="black",lwd=2)
  ## lines(newd$x,ci1[,1],col="blue",lwd=2)
  ## confband(newd$x,lower=ci0[,2],upper=ci0[,3],polygon=TRUE,col=Col("black",0.5),border=FALSE)
  ## confband(newd$x,lower=ci1[,2],upper=ci1[,3],polygon=TRUE,col=Col("blue",0.5),border=FALSE)
  ## points(y~x,col=c("black","blue")[z+1],pch=16)
  #testthat::expect_true(grcompare(d1,d2,threshold=5))
  l <- lm(y~z)
  myplot <- function(...) {
    plotConf(l,var2="z",var1=NULL,jitter=0,col="black",alpha=0.5,xlim=c(.5,2.5),ylim=range(y))
  }
  expect_doppelganger("plotConf2", myplot)
  ## par(mar=c(0,0,0,0))
  ## plot(y~I(z+1),ylim=range(y),xlim=c(0.5,2.5),pch=16,col=Col("black",0.5))
  ## l0 <- lm(y~-1+factor(z))
  ## confband(1:2,lower=confint(l0)[,1],upper=confint(l0)[,2],lwd=3,
  ##          center=coef(l0))
  })

test_that("forestplot", {
  skip_on_cran()

  set.seed(1)
  K <- 20
  est <- rnorm(K); est[c(3:4,10:12)] <- NA
  se <- runif(K,0.2,0.4)
  x <- cbind(est,est-2*se,est+2*se,runif(K,0.5,2))
  rownames(x) <- unlist(lapply(letters[seq(K)],function(x) paste(rep(x,4),collapse="")))
  rownames(x)[which(is.na(est))] <- ""
  signif <- sign(x[,2])==sign(x[,3])
  myplot <- function(...) forestplot(x)
  expect_doppelganger("forestplot1", myplot)
})

test_that("plot.sim", {
  skip_on_cran()
  skip_on_ci() # Skips the test on GitHub Actions

  set.seed(1)
  val <- as.sim(data.frame(a=rnorm(10), a.se=1, b=rnorm(10), b.se=1))
  myplot <- function(...) {
    function() plot(val, ...)
  }
  expect_doppelganger(
    "plot.sim-1",
    myplot(estimate=c(1,3), se=c(2,4), alpha=0.2,
           legend=NULL, angle=c(45,-45))
  )
  ## expect_doppelganger(
  ##   "plot.sim-2",
  ##   myplot(estimate=1, se=2, true=0, legend=NULL,
  ##          col=c("darkred", "gray"), alpha=1,
  ##          density=c(2,2), angle=c(45,-45))
  ## )
  ## expect_doppelganger(
  ##   "plot.sim-3",
  ##   myplot(estimate=c(1,3), se=c(2,4), true=0,
  ##          plot.type = "single", alpha=1, legend=NULL,
  ##          col=1:2,
  ##          density=c(2,2), angle=c(45,-45))
  ## )
  mydensity <- function(...) {
    plot(0, xlim=c(-3,3), ylim=c(0,1), type="n")
    density(val, estimate=1, legend=NULL, add=TRUE)
  }
  expect_doppelganger("density.sim-1", mydensity)
})

test_that("plot.estimate", {
  skip_on_cran()

  set.seed(1)
  e1 <- estimate(coef=1, IC=1:10-mean(1:10), id=1:10)
  e2 <- estimate(coef=1.1, IC=0:9-mean(0:9), id=1:10)
  myplot <- function(...) suppressWarnings(plot(c(e1, e2), null=1))
  val <- expect_doppelganger("plot.estimate-1", myplot)
})

test_that("spaghetti", {
  skip_on_cran()

  K <- 5
  y <- "y"%++%seq(K)
  m <- lvm()
  regression(m,y=y,x=~u) <- 1
  regression(m,y=y,x=~s) <- seq(K)-1
  regression(m,y=y,x=~x) <- "b"
  d <- sim(m,5,seed=1)
  dd <- mets::fast.reshape(d);
  dd$num <- dd$num+rnorm(nrow(dd),sd=0.5) ## Unbalanced
  myplot <- function(...) {
    spaghetti(y~num,dd,id="id",lty=1,col=Col(1,.4),trend=TRUE,trend.col="darkblue")
  }
  expect_doppelganger("spaghetti-1", myplot)
})

test_that("plot.lvm", {
  skip_on_cran()

  ## TODO
  m <- lvm(y~1*u[0:1],u~1*x)
  latent(m) <- ~u
  ## expect_doppelganger("plot.lvm-1", {
  ##   plot(m)
  ## })
  ## d <- sim(m,20,seed=1)
  ## e <- estimate(m,d)
  ## expect_doppelganger("plot.lvmfit-1", {
  ## plot(e)
  ## })
  ## expect_doppelganger("plot.lvmfit-2", {
  ##   plot(lava:::beautify(m))
  ## })
  g <- plot(lava:::beautify(m), noplot=TRUE)
  testthat::expect_true(inherits(g, "graphNEL"))
  g <- igraph.lvm(m)
  testthat::expect_true(inherits(g,"igraph"))
})

test_that("curly", {
  skip_on_cran()

  myplot <- function() {
  plot(0,0,type="n",axes=FALSE,xlab="",ylab="")
  curly(x=c(1,0),y=c(0,1),lwd=2,text="a")
  curly(x=c(1,0),y=c(0,1),lwd=2,text="b",theta=pi)
  curly(x=-0.5,y=0,shape=1,theta=pi,text="c")
  curly(x=0,y=0,shape=1,theta=0,text="d")
  curly(x=0.5,y=0,len=0.2,theta=pi/2,col="blue",lty=2)
  curly(x=0.5,y=-0.5,len=0.2,theta=-pi/2,col="red",shape=1e3,text="e")
  }
  expect_doppelganger("curly-1", myplot)
})

test_that("ksmooth", {
  skip_on_cran()

  ## TODO
  expect_true(TRUE)
})

test_that("images", {
  skip_on_cran()

  ## TODO
  expect_true(TRUE)
})

test_that("labels,edgelabels", {
  skip_on_cran()

  ## TODO
  expect_true(TRUE)
})

test_that("colorbar", {
  skip_on_cran()

  ## TODO
  expect_true(TRUE)
})

test_that("fplot", {
  skip_on_cran()

  ## TODO
  expect_true(TRUE)
})

test_that("interactive", {
  skip_on_cran()

  ## TODO
  expect_true(TRUE)
})

test_that("pdfconvert", {
  skip_on_cran()

  ## TODO
  expect_true(TRUE)
})

context("estimate.default")

test_that("estimate.default misc", {
    m <- lvm(c(y1,y2)~x+z,y1~~y2)
##    set.seed(1)
    d <- sim(m,20)

    l1 <- lm(y1~x+z,d)
    l2 <- lm(y2~x+z,d)
    ll <- merge(l1,l2)
    testthat::expect_equivalent(ll$coefmat[,1],c(coef(l1),coef(l2)))

    e1 <- estimate(l1)
    f1 <- estimate(l1,function(x) x^2, use=2)
    testthat::expect_true(coef(l1)["x"]^2==f1$coefmat[1])

    e1b <- estimate(NULL,coef=coef(l1),vcov=vcov(estimate(l1)))
    e1c <- estimate(NULL,coef=coef(l1), IC=IC(l1))
    testthat::expect_equivalent(vcov(e1b),vcov(e1c))
    testthat::expect_equivalent(var_ic(IC(e1)), vcov(e1b))

    f1b <- estimate(e1b,function(x) x^2)
    testthat::expect_equivalent(f1b$coefmat[2,,drop=FALSE],f1$coefmat)

    h1 <- estimate(l1,cbind(0,1,0))
    testthat::expect_true(h1$coefmat[,5]==e1$coefmat["x",5])

    ## GEE
    if (requireNamespace("geepack",quietly=TRUE)) {
        dd <- reshape(d, direction='long', varying=list(c('y1','y2')), v.names='y')
        dd <- dd[order(dd$id),]
        ## dd <- mets::fast.reshape(d)
        l <- lm(y~x+z,dd)
        g1 <- estimate(l,id=dd$id)
        g2 <- geepack::geeglm(y~x+z,id=dd$id,data=dd)
        testthat::expect_equivalent(g1$coefmat[,c(1,2,5)],
                          as.matrix(summary(g2)$coef[,c(1,2,4)]))
    }

    ## Several parameters
    e1d <- estimate(l1, function(x) list("X"=x[2],"Z"=x[3]))
    testthat::expect_equivalent(e1d$coefmat,e1$coefmat[-1,])

    testthat::expect_true(rownames(estimate(l1, function(x) list("X"=x[2],"Z"=x[3]),keep="X")$coefmat)=="X")
    testthat::expect_true(rownames(estimate(l1, labels=c("a"), function(x) list("X"=x[2],"Z"=x[3]),keep="X")$coefmat)=="a")


    a0 <- estimate(l1,function(p,data) p[1]+p[2]*data[,"x"], average=TRUE)
    a1 <- estimate(l1,function(p,data) p[1]+p[2]*data[,"x"]+p[3], average=TRUE)
    a <- merge(a0,a1,labels=c("a0","a1"))
    estimate(a,diff)
    testthat::expect_equivalent(estimate(a,diff)$coefmat,e1$coefmat[3,,drop=FALSE])

})

# Helper function to manually compute Wald statistic
compute_wald <- function(B, p, S, null) {
  z <- (B %*% p - null)
  V <- B %*% S %*% t(B)
  q <- t(z) %*% Inverse(V) %*% z
  return(structure(q[1], df=qr(V)$rank))
}

set.seed(1)
# Generate mean-zero random IC matrix (for internal testing)
center_ic <- function(n, p = 1) {
  x <- matrix(rnorm(n * p), n, p)
  scale(x, center = TRUE, scale = FALSE)
}
a1 <- estimate(coef = 1,   IC = center_ic(10), id = 1:10, labels = "a1")
a2 <- estimate(coef = 2,   IC = center_ic(10), id = 1:10, labels = "a2")
a3 <- estimate(coef = 3,   IC = center_ic(10), id = 1:10, labels = "a3")
a4 <- estimate(coef = 4,   IC = center_ic(10), id = 1:10, labels = "a4")
a  <- merge(a1, a2)           # 2-dimensional
a3d <- merge(a1, a2, a3)      # 3-dimensional
a4d <- merge(a1, a2, a3, a4)  # 4-dimensional


test_that("summary.estimate compared with estimate", {
  B <- rbind(c(1,-1, 0), c(0, 1,-1), c(1,0,-1))
  null <- c(1,2,3)
  q <- compute_wald(B, coef(a3d), vcov(a3d), null)
  df <- attr(q, "df")
  e1 <- estimate(a3d, f=B)
  e1b <- summary(estimate(a3d, f=B), null=null)
  e1a <- summary(estimate(e1, f=diag(3)), null=null)
  expect_equal(unname(df), unname(e1a$compare$parameter))
  expect_equal(unname(df), unname(e1b$compare$parameter))
  expect_true(abs(pchisq(q, df=df, lower.tail=FALSE) - e1a$compare$p.value)<1e-16)
  expect_true(abs(pchisq(q, df=df, lower.tail=FALSE) - e1b$compare$p.value)<1e-16)
  expect_true(abs(q - e1a$compare$statistic) < 1e-9)
  expect_true(abs(q - e1b$compare$statistic) < 1e-9)
  # function spec.
  e2 <- estimate(a3d, function(p) c(p[1]-p[2], p[2]-p[3], p[1]-p[3]))
  e2b <- summary(estimate(a3d, function(p) c(p[1]-p[2], p[2]-p[3], p[1]-p[3])), null=null)
  e2a <- summary(estimate(e2, f=diag(3)), null=null)
  expect_equal(unname(df), unname(e2a$compare$parameter))
  expect_equal(unname(df), unname(e2b$compare$parameter))
  expect_true(abs(pchisq(q, df=df, lower.tail=FALSE) - e2a$compare$p.value)<1e-16)
  expect_true(abs(pchisq(q, df=df, lower.tail=FALSE) - e2b$compare$p.value)<1e-16)
  expect_true(abs(q - e2a$compare$statistic) < 1e-9)
  expect_true(abs(q - e2b$compare$statistic) < 1e-9)
  # summary
  e3 <- summary(e1, null=null)
  expect_equal(unname(df), unname(e3$compare$parameter))
  expect_true(abs(pchisq(q, df=df, lower.tail=FALSE) - e3$compare$p.value)<1e-16)
  expect_true(abs(q - e3$compare$statistic) < 1e-9)
})

test_that("summary.estimate compared with estimate", {
  B <- rbind(c(2,-1, 0), c(0, 3,-1), c(1,0,-3), c(1,0,0))
  null <- c(1,0,1,0)
  q <- compute_wald(B, coef(a3d), vcov(a3d), null)
  df <- attr(q, "df")
  e1 <- estimate(a3d, f=B)
  e1b <- summary(estimate(a3d, f=B), null=null)
  e1a <- summary(estimate(e1, f=diag(4)), null=null)
  expect_equal(unname(df), unname(e1a$compare$parameter))
  expect_equal(unname(df), unname(e1b$compare$parameter))
  expect_true(abs(pchisq(q, df=df, lower.tail=FALSE) - e1a$compare$p.value)<1e-16)
  expect_true(abs(pchisq(q, df=df, lower.tail=FALSE) - e1b$compare$p.value)<1e-16)
  expect_true(abs(q - e1a$compare$statistic) < 1e-9)
  expect_true(abs(q - e1b$compare$statistic) < 1e-9)
  # summary
  e3 <- summary(e1, null=null)
  expect_equal(unname(df), unname(e3$compare$parameter))
  expect_true(abs(pchisq(q, df=df, lower.tail=FALSE) - e3$compare$p.value)<1e-16)
  expect_true(abs(q - e3$compare$statistic) < 1e-9)
})

test_that("1D: Identity contrast, null = 0", {
  B    <- matrix(1, nrow = 1, ncol = 1)
  null <- 0
  e    <- summary(estimate(a1, f = B), null = null)
  p <- coef(a1)
  S <- vcov(a1)
  expected <- compute_wald(B, p, S, null)
  expect_equivalent(e$compare$statistic, expected)
})

test_that("1D: Identity contrast, null = 1", {
  B    <- matrix(1, nrow = 1, ncol = 1)
  null <- 1
  e <- summary(estimate(a1, f = B), null = null)
  p <- coef(a1)
  S <- vcov(a1)
  expected <- compute_wald(B, p, S, null)
  expect_equivalent(e$compare$statistic, expected)
})

test_that("1D: Scalar contrast (scaling), null = 0", {
  B    <- matrix(2, nrow = 1, ncol = 1)
  null <- 0
  e <- summary(estimate(a1, f = B), null = null)
  p <- coef(a1)
  S <- vcov(a1)
  expected <- compute_wald(B, p, S, null)
  expect_equivalent(e$compare$statistic, expected)
})

test_that("2D: Identity contrast, null = c(0, 0)", {
  B    <- diag(2)
  null <- c(0, 0)
  e <- summary(estimate(a, f = B), null=null)
  p <- coef(a)
  S <- vcov(a)
  expected <- compute_wald(B, p, S, null)
  expect_equivalent(e$compare$statistic, expected)
})

test_that("2D: Identity contrast, null equals true coefs", {
  B    <- diag(2)
  null <- c(1, 2)
  e <- summary(estimate(a, f = B), null = null)
  p <- coef(a)
  S <- vcov(a)
  expected <- compute_wald(B, p, S, null)
  expect_equivalent(e$compare$statistic, expected)
  # Wald statistic should be very small (close to 0)
  # since null = true coefs
  expect_lt(e$compare$statistic, 1e-10)
})

test_that("2D: Identity contrast, null = c(3, 4)", {
  B    <- diag(2)
  null <- c(3, 4)
  e <- summary(estimate(a, f = B), null = null)
  p <- coef(a)
  S <- vcov(a)
  expected <- compute_wald(B, p, S, null)
  expect_equivalent(e$compare$statistic, expected)
})

test_that("2D: Difference contrast c(1,-1), null = 0", {
  B    <- matrix(c(1, -1), nrow = 1)
  null <- 0
  e <- summary(estimate(a, f = B), null = null)
  p <- coef(a)
  S <- vcov(a)
  expected <- compute_wald(B, p, S, null)
  expect_equivalent(e$compare$statistic, expected)
})

test_that("2D: Difference contrast c(1,-1), null = -1", {
  B    <- matrix(c(1, -1), nrow = 1)
  null <- -1  # Testing H0: a1 - a2 = -1 (true, since 1 - 2 = -1)
  e <- summary(estimate(a, f = B), null = null)
  p <- coef(a)
  S <- vcov(a)
  expected <- compute_wald(B, p, S, null)
  expect_equivalent(e$compare$statistic, expected)
  # Wald stat ~ 0 since null equals true difference
  expect_lt(e$compare$statistic, 1e-10)
})

test_that("2D: Sum contrast c(1,1), null = 3", {
  B    <- matrix(c(1, 1), nrow = 1)
  null <- 3  # True sum = 1 + 2 = 3
  e <- summary(estimate(a, f = B), null = null)
  p <- coef(a)
  S <- vcov(a)
  expected <- compute_wald(B, p, S, null)
  expect_equivalent(e$compare$statistic, expected)
  # Wald stat ~ 0
  expect_lt(e$compare$statistic, 1e-10)
})

test_that("2D: Scaling contrast, null = c(2, 6)", {
  B    <- 2 * diag(2)
  null <- c(2, 6)  # True: 2*c(1,2) = c(2,4), so null != true
  e <- summary(estimate(a, f = B), null = null)
  p <- coef(a)
  S <- vcov(a)
  expected <- compute_wald(B, p, S, null)
  expect_equivalent(e$compare$statistic, expected)
})

test_that("3D: Identity contrast, null = true coefs", {
  B    <- diag(3)
  null <- c(1, 2, 3)
  e <- summary(estimate(a3d, f = B), null = null)
  p <- coef(a3d)
  S <- vcov(a3d)
  expected <- compute_wald(B, p, S, null)
  expect_equivalent(e$compare$statistic, expected)
  expect_lt(e$compare$statistic, 1e-10)
})

test_that("3D: Identity contrast, null = c(0, 0, 0)", {
  B    <- diag(3)
  null <- c(0, 0, 0)
  e <- summary(estimate(a3d, f = B), null = null)
  p <- coef(a3d)
  S <- vcov(a3d)
  expected <- compute_wald(B, p, S, null)
  expect_equivalent(e$compare$statistic, expected)
})

test_that("3D: Pairwise differences, null = c(-1, -2)", {
  # H0: a1-a2 = -1, a2-a3 = -1 (true: 1-2=-1, 2-3=-1)
  B <- matrix(c(
    1, -1,  0,
    0,  1, -1
  ), nrow = 2, byrow = TRUE)
  null <- c(-1, -1)
  e <- summary(estimate(a3d, f = B), null = null)
  p <- coef(a3d)
  S <- vcov(a3d)
  expected <- compute_wald(B, p, S, null)
  expect_equivalent(e$compare$statistic, expected)
  expect_lt(e$compare$statistic, 1e-10)
})

test_that("3D: Single-row contrast, null = 6", {
  # H0: a1 + a2 + a3 = 6 (true: 1+2+3=6)
  B    <- matrix(c(1, 1, 1), nrow = 1)
  null <- 6
  e <- summary(estimate(a3d, f = B), null = null)
  p <- coef(a3d)
  S <- vcov(a3d)
  expected <- compute_wald(B, p, S, null)
  expect_equivalent(e$compare$statistic, expected)
  expect_lt(e$compare$statistic, 1e-10)
})

test_that("4D: Identity contrast, null equals true coefs", {
  B    <- diag(4)
  null <- c(1, 2, 3, 4)
  e <- summary(estimate(a4d, f = B), null = null)
  p <- coef(a4d)
  S <- vcov(a4d)
  expected <- compute_wald(B, p, S, null)
  expect_equivalent(e$compare$statistic, expected)
  expect_lt(e$compare$statistic, 1e-10)
})

test_that("4D: Non-square contrast (2x4), null = c(0, 0)", {
  # H0: a1 - a2 = 0, a3 - a4 = 0
  B <- matrix(c(
    1, -1,  0,  0,
    0,  0,  1, -1
  ), nrow = 2, byrow = TRUE)
  null <- c(0, 0)
  e <- summary(estimate(a4d, f = B), null = null)
  p <- coef(a4d)
  S <- vcov(a4d)
  expected <- compute_wald(B, p, S, null)
  expect_equivalent(e$compare$statistic, expected)
})

test_that("4D: Non-square contrast, null equals true values", {
  # H0: a1-a2 = -1, a3-a4 = -1 (true: 1-2=-1, 3-4=-1)
  B <- matrix(c(
    1, -1,  0,  0,
    0,  0,  1, -1
  ), nrow = 2, byrow = TRUE)
  null <- c(-1, -1)
  e <- summary(estimate(a4d, f = B), null = null)
  p <- coef(a4d)
  S <- vcov(a4d)
  expected <- compute_wald(B, p, S, null)
  expect_equivalent(e$compare$statistic, expected)
  expect_lt(e$compare$statistic, 1e-10)
})

test_that("Degrees of freedom equals nrow(B)", {
  B    <- diag(2)
  null <- c(0, 0)
  e    <- summary(estimate(a, f = B), null = null)
  expect_identical(unname(e$compare$parameter), nrow(B))
  B    <- matrix(c(1, -1, 0, 0, 1, -1), nrow = 2, byrow = TRUE)
  null <- c(0, 0)
  e    <- summary(estimate(a3d, f = B), null = null)
  expect_equal(unname(e$compare$parameter), nrow(B))
})

test_that("estimate warns when user-supplied IC has non-zero mean", {
  set.seed(42)
  ic_bad <- rnorm(50, mean = 10)
  expect_warning(
    estimate(coef = c(a = 1), IC = ic_bad, id = 1:50),
    "mean zero"
  )
  ## IC with empirical mean zero — no warning
  ic_good <- center_ic(50)
  expect_no_warning(
    estimate(coef = c(a = 1), IC = ic_good, id = 1:50)
  )
})

test_that("merge.estimate warns when input IC has non-zero mean", {
  set.seed(43)
  ic_bad <- rnorm(50, mean = 5)
  a <- suppressWarnings(estimate(coef = c(a = 1), IC = ic_bad, id = 1:50))
  b <- estimate(coef = c(b = 2), IC = center_ic(50), id = 1:50)
  expect_warning(merge(a, b), "mean zero")
})

test_that("IC mean-zero warning can be suppressed via lava.options", {
  set.seed(44)
  ic_bad <- rnorm(50, mean = 10)
  old <- lava.options(check.ic = FALSE)
  on.exit(lava.options(old))
  expect_no_warning(
    estimate(coef = c(a = 1), IC = ic_bad, id = 1:50)
  )
})

test_that("estimate.default parsedesign dispatch", {
  # Single character name selects that coefficient.
  e <- estimate(a3d, "a2")
  expect_equal(e$coefmat[, 1], 2)
  ef <- estimate(a3d, \(x) x["a2"])
  expect_equal(e$coefmat, ef$coefmat)

  # Multiple symbolic ... args: arithmetic on quoted names captured unevaluated.
  e <- estimate(a3d, "a1", "a2" - "a1", 2 * "a3" - 3 * "a1")
  expect_equal(e$coefmat[1, 1], 1)
  expect_equal(e$coefmat[2, 1], 2 - 1)
  expect_equal(e$coefmat[3, 1], 2 * 3 - 3 * 1)

  # Numeric (non-matrix) f treated as parameter index.
  e1 <- estimate(a3d, 2, 3)
  e2 <- estimate(a3d, "a2", "a3")
  expect_equal(e1$coefmat[, 1], e2$coefmat[, 1])

  # Multi-match wildcard "a*" produces pairwise contrasts (first vs rest).
  e_multi <- estimate(a3d, "a*")
  expect_equal(unname(e_multi$coefmat[, 1]), c(1 - 2, 1 - 3))

  # regex argument: pattern ".*2" diverges between glob and regex semantics.
  #   regex=FALSE: literal ".*2" matches nothing -> falls back to all coefs.
  #   regex=TRUE:  regex ".*2" matches "a2" only.
  e_glob  <- suppressWarnings(estimate(a3d, ".*2", regex = FALSE))
  e_regex <- suppressWarnings(estimate(a3d, ".*2", regex = TRUE))
  expect_equal(rownames(e_glob$coefmat), c("a1", "a2", "a3"))
  expect_equal(rownames(e_regex$coefmat), "a2")

  # Character f with user-supplied coef vector (no model object).
  pp <- c("(Intercept)" = 1.0, x = 2.0, z = -0.5)
  V  <- diag(3)
  dimnames(V) <- list(names(pp), names(pp))
  e <- estimate(f = "x", coef = pp, vcov = V)
  expect_equal(e$coefmat[, 1], 2.0)

  # Character contrasts combine with null hypothesis vector.
  e <- summary(estimate(a3d, "a1", "a1"), null = c(0, 1))
  expect_true(e$coefmat[1, "P-value"] != e$coefmat[2, "P-value"])
})

test_that("estimate.default keep with regex=TRUE", {
  e0 <- estimate(a3d, keep = ".*2", regex = TRUE)
  expect_equal(rownames(e0$coefmat), "a2")
  # the regex behavior differs from the above tests when supplying strings
  # to obtain contrasts (no matches return all coefficients)
  e1 <- estimate(a3d, keep = ".*2") # no literal matches return object with NAs
  expect_true(all(is.na(e1$coefmat)))
  expect_true(nrow(e1$coefmat) == 1)
})

test_that("robust argument backwards compatibility", {
  d <- data.frame(y = rnorm(50), x = rnorm(50))
  g <- lm(y ~ x, data = d)

  # Both robust=TRUE and robust=FALSE emit a deprecation warning
  e0 <- expect_warning(estimate(g, robust = FALSE), "deprecated and ignored")
  e1 <- expect_warning(estimate(g, robust = TRUE), "deprecated and ignored")
  expect_equal(e0$coefmat, e1$coefmat)

  e <- estimate(g)
  # The robust argument is ignored: results are identical to the default
  # (sandwich SEs)
  expect_equal(e$coefmat, e0$coefmat)

  expect_equal(e$coefmat, e0$coefmat)

  # model-based SE can be obtained either via logical variable or supplying
  # covariance matrix
  e0m <- estimate(g, vcov = TRUE)
  expect_false(all(e$coefmat == e0m$coefmat))
  e1m <- estimate(g, vcov = vcov(g))
  expect_equal(e0m$coefmat, e1m$coefmat)
})

test_that("initialization of object without names", {
  p <- c(a=1, b=2, c=3, d=4)
  e0 <- estimate(coef=p, vcov=diag(4))
  e1 <- estimate(coef=1:4, vcov=diag(4))
  expect_equal(coef(e0), coef(e1), check.attributes=FALSE)
  expect_equal(vcov(e0), vcov(e1), check.attributes=FALSE)

  e <- estimate(coef=1:4, vcov=diag(4), f=list(1,2,3))
  expect_equal(coef(e), coef(e0)[c(1:3)], check.attributes=FALSE)
})

library(testthat)

context("nonlinear lvm")

## simulated data
f <- function(x) - 0.25*x^2
m <- lvm(x1+x2+x3~eta1, y1+y2+y3~eta2, latent=~eta1+eta2)
functional(m, eta2~eta1) <- f
d <- sim(m,1000,seed=1,latent=TRUE)

## Setup two-stage model
m1 <- lvm(x1+x2+x3~eta1,latent=~eta1)
m2 <- lvm(y1+y2+y3~eta2,latent=~eta2)
mm <- twostage(m1,m2,formula=eta2~eta1,type="spline")

test_that("twostage", {
  nonlinear(m2,type="quadratic") <- eta2~eta1
  a <- twostage(m1,m2,data=d)
  expect_true(abs(coef(a)["eta2~eta1_1"])<0.1)
  expect_true(abs(coef(a)["eta2~eta1_2"]+0.25)<0.1)
})

test_that("cv", {
  m <- list(lm(Sepal.Length~1, data=iris),
           lm(Sepal.Length~Species, data=iris),
           lm(Sepal.Length~Species * Petal.Length, data=iris))
  x <- cv(m, rep=2, nfolds=5, data=iris)
  expect_equivalent(dim(x$cv), c(2,5,3,1))
  rmse.1 <- mean(x$cv[,,1,1])
  rmse.2 <- mean(x$cv[,,2,1])
  rmse.3 <- mean(x$cv[,,3,1])
  expect_true(rmse.1>rmse.2)
  expect_true(rmse.2>rmse.3)
})

## testthat("twostage cv", {
##   ma <- lvm(c(x1,x2,x3)~u,latent=~u)
##   ms <- functional(ma, y~u, value=function(x) -.4*x^2)
##   d <- sim(ms,50)#,seed=1)
##   ea <- estimate(ma,d)
##   mb <- lvm()
##   mb1 <- nonlinear(mb,type="linear",y~u)
##   mb2 <- nonlinear(mb,type="quadratic",y~u)
##   mb3 <- nonlinear(mb,type="spline",knots=c(-3,-1,0,1,3),y~u)
##   mb4 <- nonlinear(mb,type="spline",knots=c(-3,-2,-1,0,1,2,3),y~u)
##   ff <- lapply(list(mb1,mb2,mb3,mb4),
##                function(m) function(data,...) twostage(ma,m,data=data,st.derr=FALSE))
##   a <- lava:::cv(ff,data=d,rep=1,K=2)
## }

##' Converts strings to formula
##'
##' Converts a vector of predictors and a vector of responses (characters) i#nto
##' a formula expression.
##'
##'
##' @param y vector of predictors
##' @param x vector of responses
##' @return An object of class \code{formula}
##' @author Klaus K. Holst
##' @seealso \code{\link{as.formula}},
##' @keywords models utilities
##' @examples
##'
##' toformula(c("age","gender"), "weight")
##'
##' @export
toformula <- function (y = ".", x = ".")
{
    xst <- x[1]
    xn <- length(x)
    if (xn > 1)
        for (i in 2:length(x)) {
            xst <- paste(xst, "+", x[i])
        }
    yst <- y[1]
    yn <- length(y)
    if (yn > 1) {
        yst <- paste0("c(", yst)
        for (i in 2:length(y)) {
            yst <- paste0(yst, ", ", y[i])
        }
        yst <- paste0(yst, ")")
    }
    ff <- paste(yst, "~", xst)
    return(as.formula(ff))
}

##' Plot method for 'estimate' objects
##'
##' Plot method for 'estimate' objects
##' @export
##' @param x estimate object
##' @param f function of parameter coefficients and data parsed on to 'estimate'.
##'          If omitted a forest-plot will be produced.
##' @param idx Index of parameters (default all)
##' @param intercept include intercept in forest-plot
##' @param data data.frame
##' @param confint Add confidence limits
##' @param type plot type ('l')
##' @param xlab x-axis label
##' @param ylab y-axis label
##' @param col color
##' @param add add plot to current device
##' @param ... additional arguments to lower-level functions
plot.estimate <- function(x,f,idx,intercept=FALSE,data,confint=TRUE,type="l",xlab="x",ylab="f(x)",col=1,add=FALSE,...) {
    if (!missing(f) && !is.null(f)) {
        data <- as.list(data)
        env <- new.env()
        for (y in names(data)) {
            assign(y,data[[y]],env)
        }
        environment(f) <- env
        pp <- estimate(x,f,..., vcov=vcov(x),iid=FALSE)$coefmat
        if (!add) suppressWarnings(plot(data[[1]],pp[,1],xlab=xlab,ylab=ylab,type=type,...))
        else lines(data[[1]],pp[,1],xlab=xlab,ylab=ylab,type=type,col=col,...)
        if (confint) confband(data[[1]],pp[,3],pp[,4],polygon=TRUE,col=Col(col),lty=0)
        return(invisible(pp))
    }
    if (!is.null(x$coefmat)) {
        pp <- x$coefmat[,c(1,3,4),drop=FALSE]
    } else {
        pp <- cbind(coef(x),confint(x))
    }
    if (!missing(idx)) pp <- pp[idx,,drop=FALSE]
    if (!intercept) {
        idx <- match("(Intercept)",rownames(pp))
        if (length(idx)>0 && !any(is.na(idx))) pp <- pp[-idx,,drop=FALSE]
    }
    forestplot(pp[rev(seq(nrow(pp))),,drop=FALSE],...)
}

'.onLoad' <- function(libname, pkgname="lava") {
    addhook("heavytail_init_hook","init.hooks")
    addhook("glm_estimate_hook","estimate.hooks")
    addhook("ordinal_estimate_hook","estimate.hooks")
    addhook("cluster_post_hook","post.hooks")
    addhook("ordinal_sim_hook","sim.hooks")
    addhook("color_ordinal","color.hooks")
    addhook("ordinal_remove_hook","remove.hooks")
    lava.options(cluster.index = packagecheck("mets"))
}

'.onAttach' <- function(libname, pkgname="lava") {
    # desc <- utils::packageDescription(pkgname)
    # packageStartupMessage(desc$Package, " version ",desc$Version)
}

##' Generic method for finding indeces of model parameters
##'
##' @title Generic method for finding indeces of model parameters
##' @param x Model object
##' @param \dots Additional arguments
##' @author Klaus K. Holst
##' @export
`parpos` <-
  function(x,...) UseMethod("parpos")

##' @export
parpos.default <- function(x,p,...) {
  if (is.numeric(p)) return(p)
  na.omit(match(coef(x),p))
}

##' @export
parpos.multigroup <- function(x,p,mean=TRUE,...) {
  if (missing(p)) {
    p <- unique(unlist(lapply(x$lvm, function(z) setdiff(parlabels(z),names(constrain(z))) )))
  }
  if (!is.character(p)) p <- names(p)
  p0 <- rep(NA,with(x,npar+npar.mean));
  names(p0) <- c(x$mean,x$par)
  for (i in seq_along(x$lvm)) {
    cur <- parpos(x$lvm[[i]],p=p)
    if (length(cur)>0) {
      p0[c(x$meanpos[[i]],x$parpos[[i]])[cur]] <- names(cur)
      M <- na.omit(match(names(cur),p))
      if (length(M)>0)
        p <- p[-M]
    }
    if (length(p)==0) break;
  }
  p1 <- which(!is.na(match(x$name,p)))
  p0[p1] <- x$name[p1]
  return(structure(which(!is.na(p0)),name=p0))
##  return(p0)
}

##' @export
parpos.multigroupfit <- function(x,...) parpos.multigroup(x$model0,...)

##' @export
parpos.lvm <- function(x,p,mean=TRUE,...) {
  if (!missing(p)) {
    if (!is.character(p)) p <- names(p)
    cc1 <- coef(Model(x),mean=mean,fix=FALSE)
    cc2 <- coef(Model(x),mean=mean,fix=FALSE,labels=TRUE)
    idx1 <- na.omit(match(p,cc1))
    idx11 <- na.omit(match(p,cc2))
    res <- (union(idx1,idx11));
    if (length(res)!=length(p)) {
      names(res) <- cc1[res]
    } else {
      names(res) <- p
    }
    ##    res <- idx1; res[!is.na(idx11)] <- idx11[!is.na(idx11)]
    ##    names(res) <- p
    ord <- order(res)
    res <- sort(res)
    attributes(res)$ord <- ord
    return(res)
  }
  if (mean)
    nn <- with(index(x),matrices2(x,seq_len(npar+npar.mean+npar.ex))) ## Position of parameters
  else nn <- with(index(x),matrices(x,seq_len(npar),NULL,seq_len(npar.ex)+npar))
  nn$A[index(x)$M0!=1] <- 0
  nn$P[index(x)$P0!=1] <- 0
  nn$v[index(x)$v0!=1] <- 0
  nn$e[index(x)$e0!=1] <- 0
  nn
}

##' @export
parpos.lvmfit <- parpos.lvm

###{{{ coef.lvm

##' @export
`coef.lvm` <-
    function(object, mean=TRUE, fix=TRUE,
      symbol=lava.options()$symbol,
      messages=lava.options()$messages,
      p, data, vcov, type=9,
      labels=lava.options()$coef.names, ...) {
        if (fix)
            object <- fixsome(object,measurement.fix=FALSE)
        if (!missing(p)) {
            coefs <- matrix(NA,nrow=length(p),ncol=4); coefs[,1] <- p
            rownames(coefs) <- c(coef(object,mean=TRUE,fix=FALSE)[c(seq_len(index(object)$npar.mean))],
            {if (index(object)$npar>0) paste0("p",seq_len(index(object)$npar)) },
            {if (index(object)$npar.ex>0) paste0("e",seq_len(index(object)$npar.ex))} )
            if (missing(vcov)) {
                if (!missing(data) && !is.null(data)) {
                    I <- information(object,p=p,data=data,type="E")
                    myvcov <- solve(I)
                } else {
                    myvcov <- matrix(NA,length(p),length(p))
                }
                object$vcov <- myvcov
            } else object$vcov <- vcov
            coefs[,2] <- sqrt(diag(object$vcov))
            coefs[,3] <- coefs[,1]/coefs[,2]
            coefs[,4] <-  2*(pnorm(abs(coefs[,3]),lower.tail=FALSE))
            colnames(coefs) <- c("Estimate","Std. Error", "Z value", "Pr(>|z|)")
            object$coefficients <- coefs;
            return(coef.lvmfit(object,type=type,labels=labels,symbol=symbol,...))
        }


        ## Free regression/covariance parameters
        AP <- matrices(object, paste0("p",seq_len(index(object)$npar)))
        A <- AP$A; A[index(object)$M1==0] <- "0" ## Only free parameters
        P <- AP$P; P[index(object)$P1==0] <- "0"; P[upper.tri(P)] <- "0"
        nn <- vars(object)

        res <- c()
        resname <- c()

        ii <- which(t(A)!="0",arr.ind=TRUE)
        rname <- paste(nn[ii[,1]],nn[ii[,2]],sep=symbol[1])
        if (labels) {
            rname2 <- t(regfix(Model(object))$labels)[ii]
            rname[which(!is.na(rname2))] <- rname2[which(!is.na(rname2))]
        }
        res <- rname
        resname <- c(resname,t(A)[ii])

        ii <- which(P!="0",arr.ind=TRUE)
        if (length(symbol)<2)
            rname <- paste(nn[ii[,2]],nn[ii[,1]],sep=lava.options()$symbol[2])
        else
            rname <- paste(nn[ii[,2]],nn[ii[,1]],sep=symbol[2])
        if (labels) {
            rname2 <- (covfix(Model(object))$labels)[ii]
            rname[which(!is.na(rname2))] <- rname2[which(!is.na(rname2))]
        }
        res <- c(res,rname)
        resname <- c(resname,P[ii])

        names(res) <- resname
        resnum <- sapply(resname, function(s) char2num(substr(s,2,nchar(s))))
        res <- res[order(resnum)]
        if (mean) {
            nmean <- sum(index(object)$v1==1)
            if (nmean>0) {

                if (!labels)
                    res <- c(vars(object)[index(object)$v1==1], res)
                else {
                    mres <- c()
                    for (i in seq_len(length(index(object)$v1))) {
                        val <- index(object)$v1[i]
                        if (val==1) {
                            if (!is.na(intfix(Model(object))[[i]])) {
                                mres <- c(mres, intfix(Model(object))[[i]])
                            }
                            else
                                mres <- c(mres, vars(object)[i])
                        }
                    }
                    res <- c(mres,res)
                }
                names(res)[seq_len(nmean)] <- paste0("m",seq_len(nmean))
            }
        }
        if (!is.null(object$expar) && sum(index(object)$e1==1)>0) {
            n2 <- names(object$expar)[index(object)$e1==1]
            if (labels) {
                count <- 0
                for (i in seq_len(length(index(object)$e1))) {
                    if (index(object)$e1[i]==1) {
                        val <- object$exfix[[i]]
                        count <- count+1
                        if(!is.na(val)) n2[count] <- val
                    }
                }
            }
            names(n2) <- paste0("e",seq_len(length(n2)))
            res <- c(res,n2)
        }

        if (messages>1) {
            cat(paste(res, collapse="\n")); cat("\n")
        }
        if (!is.null(object$order)) res <- res[object$order]
        res
    }

###}}}

###{{{ coef.lvmfit

##' @export
`coef.lvmfit` <-
    function(object, type=ifelse(missing(type),-1,2),
      symbol=lava.options()$symbol,
      data, std=NULL,
      labels=lava.options()$coef.names,
      ##labels=TRUE,
      vcov,
      vcov.type, reliability=FALSE, second=FALSE, ...) {

        res <- (pars.default(object,...))
        if (type<0 && !is.null(names(res))) return(res)

        if (is.null(object$control$meanstructure)) meanstructure <- TRUE
        else meanstructure <- object$control$meanstructure
        npar <- index(object)$npar; npar.mean <- index(object)$npar.mean*meanstructure
        npar.ex <- index(object)$npar.ex

        para <- parameter(Model(object))
        para.idx <- which(vars(object)%in%para)


        if (inherits(object,"lvm.missing")) {
            if (length(object$cc)==0) {## No complete cases
                coefs <- coef(object$estimate)
                c1 <- coef(Model(object),mean=TRUE,fix=FALSE)
                c1. <- coef(Model(object),mean=FALSE,fix=FALSE)
                nn <- gsub("^[0-9]*@","",names(coefs))
                myorder <- match(c1,nn)
                myorder.reg <- order(na.omit(match(nn,c1.)))
                myorder.extra <- c()
            } else {
                myorder <- na.omit(modelPar(object$multigroup,seq_len(npar+npar.mean))$p[[object$cc]])
                myorder.reg <- na.omit(modelPar(object$multigroup,seq_len(npar))$p[[object$cc]])
                myorder.extra <- seq_len(index(object)$npar.ex)+length(myorder)
                myorder <- c(myorder,myorder.extra)

            }
        } else {
            myorder <- seq_len(npar+npar.mean)
            myorder.reg <- seq_len(npar)
            myorder.extra <- seq_len(index(object)$npar.ex)+length(myorder)
            myorder <- c(myorder,myorder.extra)
        }

        if (type<0) {
            names(res)[seq_len(length(myorder))] <- coef(Model(object),fix=FALSE, mean=meanstructure, symbol=symbol)[order(myorder)]
            return(res)
        }


        latent.var <- latent(object)
        Type <- Var <- From <- c()

        Astd <- Pstd <- vstd <- NULL
        if (!is.null(std)) {
            stdCoef <- stdcoef(object)
            {
                switch(tolower(std),
                       latent = {Astd=stdCoef$Astar; Pstd=stdCoef$Pstar; vstd=stdCoef$vstar},
                       y = {Astd=stdCoef$AstarY; Pstd=stdCoef$PstarY; vstd=stdCoef$vstarY},
                       xy = {Astd=stdCoef$AstarXY; Pstd=stdCoef$PstarXY; vstd=stdCoef$vstarXY},
                       yx = {Astd=stdCoef$AstarXY; Pstd=stdCoef$PstarXY; vstd=stdCoef$vstarXY}
                       )
            }
        }
        myparnames <- paste0("p",seq_len(npar+npar.ex))[myorder.reg]

        p <- matrices(Model(object), myparnames)
        A <- p$A
        P <- p$P
        mycoef <- object$coef
        if (!missing(vcov.type) | !missing(vcov)) {
            if (!missing(vcov)) {
                mycoef[,2] <- sqrt(diag(vcov))[myorder]
            } else {
                if (!missing(data))
                    myvcov <- information(object,type=vcov.type,data=data,inverse=TRUE)
                else
                    myvcov <- information(object,type=vcov.type,inverse=TRUE)
                mycoef[,2] <- sqrt(diag(myvcov))[myorder]
            }
            mycoef[,3] <- mycoef[,1]/mycoef[,2]
            mycoef[,4] <-  2*(pnorm(abs(mycoef[,3]),lower.tail=FALSE))
        }

        coefs <- mycoef[myorder,,drop=FALSE]
        nn <- colnames(A)

        free <- A!="0"
        free[index(object)$M1!=1] <- FALSE
        nlincon <- matrix(Model(object)$par%in%names(constrain(Model(object))),nrow(A))
        if (missing(data)) {
            data <- matrix(0,ncol=length(index(Model(object))$manifest)); colnames(data) <- index(Model(object))$manifest
        }
        nlincon.estimates.full<- constraints(object,second=second,data=data)
        nlincon.estimates <- nlincon.estimates.full[,-(5:6),drop=FALSE]
        matched <- c()
        res <- c()
        for (i in seq_len(ncol(A)))
            for (j in seq_len(nrow(A))) {
                val <- A[j,i]
                if (val!="0") {
                    matching <- match(val,rownames(coefs))
                    matched <- c(matched,matching)
                    if (!is.na(matching)) {
                        if (free[j,i])
                            newrow <- matrix(coefs[matching,],nrow=1)
                        else {
                            newrow <- matrix(c(coefs[matching,1],NA,NA,NA), nrow=1)
                        }
                    } else {
                        Debug(list("(i,j)", i, ",", j))
                        if (nlincon[j,i]) {
                            newrow <- matrix(nlincon.estimates[Model(object)$par[j,i],],nrow=1)
                        } else {
                            newrow <- matrix(c(Model(object)$fix[j,i], NA, NA, NA), nrow=1)
                        }
                    }
                    if (!is.null(std)) {
                        newrow <- cbind(newrow,Astd[j,i])
                    }
                    if (labels & !is.na(regfix(Model(object))$labels[j,i])) {
                        rownames(newrow) <- regfix(Model(object))$labels[j,i]
                        if (labels>1) {
                            newst <- paste0(nn[i],symbol[1],nn[j])
                            if (rownames(newrow)!=newst)
                                rownames(newrow) <- paste(rownames(newrow),newst,sep=":")
                        }
                    } else {
                        rownames(newrow) <- paste0(nn[i],symbol[1],nn[j])
                    }
                    if (free[j,i] | type>2) {
                        res <- rbind(res, newrow)
                        Type <- c(Type,"regression")
                        Var <- c(Var, nn[i])
                        From <- c(From, nn[j])
                    }
                }
            }
        free.var <- P!="0"
        free.var[index(object)$P1!=1] <- FALSE
        nlincon.var <- matrix(Model(object)$covpar%in%names(constrain(Model(object))),nrow(P))

        if (type>0)
            ## Variance estimates:
            for (i in seq_len(ncol(p$P)))
                for (j in seq(i,nrow(p$P))) {
                    val <- p$P[j,i]

                    if (!(i%in%para.idx))
                        if (val!="0" & !any(vars(object)[c(i,j)]%in%index(Model(object))$exogenous))
                            if (type>1 | !all(vars(object)[c(i,j)]%in%index(Model(object))$manifest))
                            {
                                matching <- match(val,rownames(coefs))
                                matched <- c(matched,matching)

                                if (!is.na(matching)) {
                                    if (free.var[j,i])
                                        newrow <- matrix(coefs[matching,],nrow=1)
                                    else
                                        newrow <- matrix(c(coefs[matching,1],NA,NA,NA), nrow=1)
                                    ## We don't want to report p-values of tests on the boundary of the parameter space
                                    if (i==j)
                                        newrow[,4] <- NA
                                } else {
                                    Debug(list("(i,j)", i, ",", j))
                                    if (nlincon.var[j,i]) {
                                        newrow <- matrix(nlincon.estimates[Model(object)$covpar[j,i],],nrow=1)
                                    } else {
                                        newrow <- matrix(c(Model(object)$covfix[j,i], NA, NA, NA), nrow=1)
                                    }
                                }
                                if (!missing(std)) {
                                    newrow <- cbind(newrow,Pstd[i,j])
                                }
                                if (length(symbol)<2) {
                                    if (nn[i]!=nn[j]) {
                                        part2 <- paste(nn[i],nn[j],sep=lava.options()$symbol[2])
                                    } else part2 <- nn[i]
                                } else {
                                    part2 <- paste0(nn[i],symbol[2],nn[j])
                                }
                                if (labels & !is.na(covfix(Model(object))$labels[j,i])) {
                                    rownames(newrow) <- covfix(Model(object))$labels[j,i]
                                    if (labels>1) {
                                        if (rownames(newrow)!=part2)
                                            rownames(newrow) <- paste(rownames(newrow),part2,sep=":")
                                    }
                                } else {
                                    rownames(newrow) <- part2
                                }
                                if ((free.var[j,i]) | type>2) {
                                    res <- rbind(res, newrow)
                                    Type <- c(Type,"variance")
                                    Var <- c(Var, nn[i])
                                    From <- c(From, nn[j])
                                }
                            }
                }

        ## Mean parameter:
        nlincon.mean <- lapply(Model(object)$mean, function(x) x%in%names(constrain(Model(object))) )

        if (type>0 & npar.mean>0) {
            midx <- seq_len(npar.mean)
            rownames(coefs)[midx] <- paste0("m",myorder[midx])
            munames <- rownames(coefs)[seq_len(npar.mean)]
            meanpar <- matrices(Model(object), myparnames, munames)$v
            for (i in seq_len(length(meanpar))) {
                if (!index(Model(object))$vars[i]%in%index(Model(object))$exogenous) {
                    val <- meanpar[i]
                    matching <- match(val,rownames(coefs))
                    if (!is.na(matching)) {
                        if (index(object)$v1[i]==1)  ## if free-parameter
                            newrow <- matrix(coefs[matching,],nrow=1)
                        else
                            newrow <- matrix(c(coefs[matching,1],NA,NA,NA), nrow=1)
                    } else {
                        if (nlincon.mean[[i]]) {
                            newrow <- matrix(nlincon.estimates[Model(object)$mean[[i]],],nrow=1)
                        } else {
                            newrow <- matrix(c(as.numeric(meanpar[i]), NA, NA, NA), nrow=1)          }
                    }
                    if (!missing(std)) {
                        newrow <- cbind(newrow,vstd[i])
                    }
                    if (labels & !(is.na(intfix(Model(object))[[i]]) | is.numeric(intfix(Model(object))[[i]]))) {
                        rownames(newrow) <- intfix(Model(object))[[i]]
                        if (labels>1) {
                            if (rownames(newrow)!=index(Model(object))$vars[i])
                                rownames(newrow) <- paste(rownames(newrow),index(Model(object))$vars[i],sep=":")
                        }
                    } else {
                        rownames(newrow) <- index(Model(object))$vars[i]
                    }
                    if ((index(object)$v1[i]==1) | type>2) {
                        res <- rbind(res, newrow)
                        Type <- c(Type,ifelse(!(i%in%para.idx),"intercept","parameter"))
                        Var <- c(Var, index(Model(object))$vars[i])
                        From <- c(From, NA)
                    }
                }
            }
        }

        if (type>0 && length(myorder.extra>0)) {
            cc <- coefs[myorder.extra,,drop=FALSE]
            rownames(cc) <- rownames(index(object)$epar)[which(index(object)$e1==1)]
            cc <- cbind(cc,rep(NA,ncol(res)-ncol(cc)))
            res <- rbind(res,cc)
            Type <- c(Type,rep("extra",length(myorder.extra)))
            Var <- c(Var,rep(NA,length(myorder.extra)))
            From <- c(From,rep(NA,length(myorder.extra)))
        }

        mycolnames <- colnames(coefs)
        if (!is.null(std)) mycolnames <- c(mycolnames, paste("std",std,sep="."))
        colnames(res) <- mycolnames
        attributes(res)$type <- Type
        attributes(res)$var <- Var
        attributes(res)$from <- From
        attributes(res)$latent <- latent.var
        attributes(res)$nlincon <- nlincon.estimates.full
        return(res)
    }

###}}} coef.lvmfit

###{{{ coef.multigroup

##' @export
coef.multigroup <- function(object,...) {
    return(object$parpos)
}

###}}} coef.multigroup

###{{{ coef.multigroupfit

##' @export
coef.multigroupfit <-
    function(object, type=0,vcov, ext=FALSE,
      labels=lava.options()$coef.names,
      symbol=lava.options()$symbol,
      covsymb=NULL,groups=NULL,...) {

        if (type==0) {
            res <- pars(object);
            if (is.null(names(res))) names(res) <- object$model$name
            return(res)
        }
        if (type==1) {
            theta <- pars(object)
            if (missing(vcov))
                theta.sd <- sqrt(diag(object$vcov))
            else
                theta.sd <- sqrt(diag(vcov))
            res <- cbind(theta,theta.sd,(Z <- theta/theta.sd),2*(pnorm(abs(Z),lower.tail=FALSE)))
            if (is.null(rownames(res)))
                rownames(res) <- object$model$name
            colnames(res) <- c("Estimate","Std. Error", "Z value", "Pr(>|z|)")
            return(res)
        }

        cc <- coef(object, type=1, symbol=symbol, ...)
        model <- Model(object)
        parpos <- modelPar(model, seq_len(nrow(cc)))$p
        npar.mean <- object$model$npar.mean
        npar <- object$model$npar
        mynames <- c()
        if (npar.mean>0) {
            mynames <- unlist(object$model$meanlist)
            mynames <- names(mynames)[!duplicated(mynames)]
        }
        if (npar>0) {
            mynames <- c(mynames,object$model$par)
        }
        res <- list()
        misrow <- list()
        parpos2 <- list()
        if (is.null(groups)) groups <- seq(model$ngroup)
        if (length(groups)==0) groups <- seq(model$ngroup)
        for (i in groups) {
            orignames <- coef(object$model0$lvm[[i]],fix=FALSE,mean=object$meanstructure, messages=0, symbol=lava.options()$symbol)
            if (ext) {
                newnames. <- coef(Model(model)[[i]],fix=FALSE, mean=object$meanstructure, messages=0, labels=labels, symbol=symbol)
                newnames <- coef(Model(model)[[i]],fix=FALSE, mean=object$meanstructure, messages=0, labels=labels,symbol=lava.options()$symbol)
                newcoef <- matrix(NA,ncol=4,nrow=length(newnames))
                rownames(newcoef) <- newnames.
                idx <- match(orignames,newnames)
                newcoef[idx,] <- cc[parpos[[i]],,drop=FALSE]
                newparpos <- rep(NA,length(newnames))
                newparpos[idx] <- parpos[[i]]
                parpos2 <- c(parpos2, list(newparpos))
                misrow <- c(misrow, list(setdiff(seq_len(length(newnames)),idx)))
            } else {
                newcoef <- cc[parpos[[i]],,drop=FALSE]
                rownames(newcoef) <- orignames
            }
            colnames(newcoef) <- colnames(cc)
            ## Position of variance parameters:
            varpos <- variances(Model(model)[[i]],mean=FALSE)
            ## Number of parameters resp mean-parameters
            p <- nrow(newcoef); p0 <- length(coef(Model(model)[[i]],fix=FALSE, mean=FALSE, messages=0))
            newcoef[(p-p0) + varpos,4] <- NA
            res <- c(res, list(newcoef))
        }

        if (ext) {
            for (i in seq(length(groups))) {
                if (length(misrow[[i]])>0) {
                    nn <- rownames(res[[i]])[misrow[[i]]]
                    for (j in setdiff(seq_len(length(groups)),i)) {
                        nn2 <- rownames(res[[j]])
                        matching <- na.omit(match(nn,nn2))
                        matching <- setdiff(matching,misrow[[j]])
                        if (length(matching)>0) {
                            idxj <- match(nn2[matching],nn2)
                            idxi <- match(nn2[matching],rownames(res[[i]]))
                            res[[i]][nn2[matching],] <- res[[j]][nn2[matching],]
                            parpos2[[i]][idxi] <- parpos2[[j]][idxj]
                            nn <- setdiff(nn,nn2[matching])
                        }
                        if (length(nn)<1) break;
                    }
                }
            }
            attributes(res)$parpos <- parpos2
        }
        return(res)
    }

###}}}

###{{{ CoefMat

##' @export
CoefMat.multigroupfit <- function(x,type=9,
                           labels=lava.options()$coef.names,
                           symbol=lava.options()$symbol[1],
                           data=NULL,groups=seq(Model(x)$ngroup),...) {

    cc <- coef(x,type=type,ext=TRUE,symbol=symbol,data=data,groups=groups)
    parpos <- attributes(cc)$parpos
    res <- c()
    nlincon.estimates <- c()
    nlincon.names <- c()
    count <- k <- 0

    for (i in groups) {
        k <- k+1
        m0 <- Model(Model(x))[[i]]
        mycoef <- cc[[k]]
        npar <- index(m0)$npar
        npar.mean <- index(m0)$npar.mean
        if (npar>0)
            rownames(mycoef)[(seq(npar))+npar.mean] <- paste0("p",seq(npar))
        m0$coefficients <- mycoef
        m0$opt$estimate <- mycoef[,1]
        Vcov <- vcov(x)[parpos[[k]],parpos[[k]],drop=FALSE]; colnames(Vcov) <- rownames(Vcov) <- rownames(mycoef)
        m0$vcov <- Vcov
        cc0 <- coef.lvmfit(m0,type=type,labels=labels,symbol=symbol)
        attributes(cc0)$dispname <- x$opt$dispname
        res <- c(res, list(CoefMat(cc0)))
        newnlin <- attributes(cc0)$nlincon
        if (length(newnlin)>0)
            if (count==0) {
                count <- count+1
                nlincon.estimates <- newnlin
                nlincon.names <- rownames(newnlin)
            } else {
                for (j in seq_len(NROW(newnlin))) {
                    if (!(rownames(newnlin)[j]%in%nlincon.names)) {
                        nlincon.estimates <- rbind(nlincon.estimates,newnlin[j,,drop=FALSE])
                        nlincon.names <- c(nlincon.names,rownames(newnlin)[j])
                    }
                }
            }
    }
    rownames(nlincon.estimates) <- nlincon.names
    attributes(res)$nlincon <- nlincon.estimates
    return(res)
}


##' @export
CoefMat <- function(x,
             digits = max(3, getOption("digits") - 2),
             type=9,
             symbol=lava.options()$symbol[1],...) {
    cc <- x
    if (!is.matrix(x)) {
        cc <- coef(x,type=type,symbol=symbol,...)
    }
    res <- c()
    mycoef <- format(round(cc,max(1,digits)),digits=digits)
    mycoef[,4] <- formatC(cc[,4],digits=digits-1,format="g",
                          preserve.width="common",flag="")
    mycoef[is.na(cc)] <- ""
    mycoef[cc[,4]<1e-12,4] <- "  <1e-12"

    M <- ncol(cc)
    Nreg <- sum(attributes(cc)$type=="regression")
    Nvar <- sum(attributes(cc)$type=="variance")
    Nint <- sum(attributes(cc)$type=="intercept")

    latent.var <- attributes(cc)$latent


    if (Nreg>0) {
        reg.idx <- which(attributes(cc)$type=="regression")
        latent.from <- which(attributes(cc)$from[reg.idx]%in%latent.var)
        latent.from <- latent.from[which(is.na(match(attributes(cc)$var[latent.from],latent.var)))]

        reg.idx <- setdiff(reg.idx,latent.from)
        Nmeas <- length(latent.from)
        if (Nmeas>0) {
            first.entry <- c()
            for (i in latent.var) {
                pos <- match(i,attributes(cc)$from[latent.from])
                if (!is.na(pos))
                    first.entry <- c(first.entry, pos)
            }
            res <- rbind(res, c("Measurements:",rep("",M)))
            count <- 0
            Delta <- FALSE
            for (i in latent.var) {
                count <- count+1
                Delta <- !Delta
                Myidx <- which(attributes(cc)$from==i & attributes(cc)$type=="regression" & !(attributes(cc)$var%in%latent.var))

                prefix <- ifelse(Delta,"  ","   ")
                for (j in Myidx) {
                    newrow <- mycoef[j,]
                    newname <- rownames(cc)[j]
                    res <- rbind(res,c(paste(prefix,newname),newrow))
                }
            }
        }
        if ((Nreg-Nmeas)>0) {
            responses <- unique(attributes(cc)$var[reg.idx])
            first.entry <- c()
            for (i in responses) {
                pos <- match(i,attributes(cc)$var[reg.idx])
                first.entry <- c(first.entry, pos)
            }
            res <- rbind(res, c("Regressions:",rep("",M)))
            count <- 0
            Delta <- FALSE
            for (i in reg.idx) {
                count <- count+1
                newrow <- mycoef[i,]
                newname <- rownames(cc)[i]
                if (count%in%first.entry) Delta <- !Delta
                prefix <- ifelse(Delta,"  ","   ")
                res <- rbind(res,c(paste(prefix,newname),newrow))
            }
        }
    }


    if (Nint>0) {
        int.idx <- which(attributes(cc)$type=="intercept")
        res <- rbind(res, c("Intercepts:",rep("",M)))
        for (i in int.idx) {
            newrow <- mycoef[i,]
            newname <- rownames(cc)[i]
            res <- rbind(res,c(paste("  ",newname),newrow))
        }
    }
    par.idx <- which(attributes(cc)$type=="parameter")
    parres <- rbind(c("Additional Parameters:",rep("",M)))
    for (i in par.idx) {
        newrow <- mycoef[i,]
        newname <- rownames(cc)[i]
        parres <- rbind(parres,c(paste("  ",newname),newrow))
    }
    extra.idx <- which(attributes(cc)$type=="extra")
    for (i in extra.idx) {
        newrow <- mycoef[i,]
        newname <- rownames(cc)[i]
        parres <- rbind(parres,c(paste("  ",newname),newrow))
    }
    if (nrow(parres)>1) res <- rbind(res,parres)



    if (Nvar>0) {
        var.idx <- which(attributes(cc)$type=="variance")
        vname <- "Residual Variances:"
        if (!is.list(x)) {
            if (!is.null(attributes(x)$dispname)) vname <- attributes(x)$dispname
        } else if (!is.null(x$opt$dispname)) vname <- x$opt$dispname
        res <- rbind(res, c(vname,rep("",M)))
        for (i in var.idx) {
            newrow <- mycoef[i,]
            newname <- rownames(cc)[i]
            res <- rbind(res,c(paste("  ",newname),newrow))
        }
    }
    res0 <- res[,-1]
    rownames(res0) <- format(res[,1],justify="left")
    res0
}

###}}} CoefMat

###{{{ standardized coefficients

stdcoef <- function(x,p=coef(x),...) {
    M0 <- moments(x,p=p,...)
    A <- t(M0$A)
    P <- M0$P
    v <- M0$v
    C <- M0$Cfull
    N <- diag(sqrt(diag(C)),ncol=nrow(C)); colnames(N) <- rownames(N) <- vars(x)
    iN <- N; diag(iN)[diag(N)>0] <- 1/diag(iN)[diag(N)>0]
    diag(iN)[diag(N)==0] <- NA
    Nn <- N; Nn[] <- 0; diag(Nn) <- 1
    Nn[latent(x),latent(x)] <- N[latent(x),latent(x)]
    iNn <- Nn; diag(iNn) <- 1/diag(Nn)
    Ny <- Nn;
    Ny[endogenous(x),endogenous(x)] <- N[endogenous(x),endogenous(x)]
    iNy <- Ny; diag(iNy) <- 1/diag(Ny)
    ## Standardized w.r.t. latent,y and x:
    AstarXY <- t(iN%*%A%*%N)
    PstarXY <- iN%*%P%*%iN
    if (!is.null(v))
        vstarXY <- iN%*%v
    else
        vstarXY <- NULL
    pstdXY <- pars(Model(x),A=AstarXY,P=PstarXY,v=vstarXY)
    ## Standardized w.r.t. latent, y:
    AstarY <- t(iNy%*%A%*%Ny)
    PstarY <- iNy%*%P%*%iNy
    if (!is.null(v))
        vstarY <- iNy%*%v
    else
        vstarY <- NULL
    ## Standardized w.r.t. latent only:
    Astar <- t(iNn%*%A%*%Nn)
    Pstar <- iNn%*%P%*%iNn
    if (!is.null(v))
        vstar <- iNn%*%v
    else
        vstar <- NULL
    pstd <- pars(Model(x),A=Astar,Pstar,v=vstar)
    k <- length(p)-length(pstd)

    res <- list(par=cbind(p,c(pstd,rep(NA,k)),c(pstdXY,rep(NA,k))),
                AstarXY=AstarXY, PstarXY=PstarXY, vstarXY=vstarXY,
                AstarY=AstarY, PstarY=PstarY, vstarY=vstarY,
                Astar=Astar, Pstar=Pstar, vstar=vstar)
    return(res)
}

###}}} standardized coefficients

##' @export
`coef.lvm` <-
    function(object, mean=TRUE, fix=TRUE,
      symbol=lava.options()$symbol,
      messages=lava.options()$messages,
      p, data, vcov, type=9,
      labels=lava.options()$coef.names, ...) {
        if (fix)
            object <- fixsome(object,measurement.fix=FALSE)
        if (!missing(p)) {
            coefs <- matrix(NA,nrow=length(p),ncol=4); coefs[,1] <- p
            rownames(coefs) <- c(coef(object,mean=TRUE,fix=FALSE)[c(seq_len(index(object)$npar.mean))],
            {if (index(object)$npar>0) paste0("p",seq_len(index(object)$npar)) },
            {if (index(object)$npar.ex>0) paste0("e",seq_len(index(object)$npar.ex))} )
            if (missing(vcov)) {
                if (!missing(data) && !is.null(data)) {
                    I <- information(object,p=p,data=data,type="E")
                    myvcov <- solve(I)
                } else {
                    myvcov <- matrix(NA,length(p),length(p))
                }
                object$vcov <- myvcov
            } else object$vcov <- vcov
            coefs[,2] <- sqrt(diag(object$vcov))
            coefs[,3] <- coefs[,1]/coefs[,2]
            coefs[,4] <-  2*(pnorm(abs(coefs[,3]),lower.tail=FALSE))
            colnames(coefs) <- c("Estimate","Std. Error", "Z value", "Pr(>|z|)")
            object$coefficients <- coefs;
            return(coef.lvmfit(object,type=type,labels=labels,symbol=symbol,...))
        }


        ## Free regression/covariance parameters
        AP <- matrices(object, paste0("p",seq_len(index(object)$npar)))
        A <- AP$A; A[index(object)$M1==0] <- "0" ## Only free parameters
        P <- AP$P; P[index(object)$P1==0] <- "0"; P[upper.tri(P)] <- "0"
        nn <- vars(object)

        res <- c()
        resname <- c()

        ii <- which(t(A)!="0",arr.ind=TRUE)
        rname <- paste(nn[ii[,1]],nn[ii[,2]],sep=symbol[1])
        if (labels) {
            rname2 <- t(regfix(Model(object))$labels)[ii]
            rname[which(!is.na(rname2))] <- rname2[which(!is.na(rname2))]
        }
        res <- rname
        resname <- c(resname,t(A)[ii])

        ii <- which(P!="0",arr.ind=TRUE)
        if (length(symbol)<2)
            rname <- paste(nn[ii[,2]],nn[ii[,1]],sep=lava.options()$symbol[2])
        else
            rname <- paste(nn[ii[,2]],nn[ii[,1]],sep=symbol[2])
        if (labels) {
            rname2 <- (covfix(Model(object))$labels)[ii]
            rname[which(!is.na(rname2))] <- rname2[which(!is.na(rname2))]
        }
        res <- c(res,rname)
        resname <- c(resname,P[ii])

        names(res) <- resname
        resnum <- sapply(resname, function(s) char2num(substr(s,2,nchar(s))))
        res <- res[order(resnum)]
        if (mean) {
            nmean <- sum(index(object)$v1==1)
            if (nmean>0) {

                if (!labels)
                    res <- c(vars(object)[index(object)$v1==1], res)
                else {
                    mres <- c()
                    for (i in seq_len(length(index(object)$v1))) {
                        val <- index(object)$v1[i]
                        if (val==1) {
                            if (!is.na(intfix(Model(object))[[i]])) {
                                mres <- c(mres, intfix(Model(object))[[i]])
                            }
                            else
                                mres <- c(mres, vars(object)[i])
                        }
                    }
                    res <- c(mres,res)
                }
                names(res)[seq_len(nmean)] <- paste0("m",seq_len(nmean))
            }
        }
        if (!is.null(object$expar) && sum(index(object)$e1==1)>0) {
            n2 <- names(object$expar)[index(object)$e1==1]
            if (labels) {
                count <- 0
                for (i in seq_len(length(index(object)$e1))) {
                    if (index(object)$e1[i]==1) {
                        val <- object$exfix[[i]]
                        count <- count+1
                        if(!is.na(val)) n2[count] <- val
                    }
                }
            }
            names(n2) <- paste0("e",seq_len(length(n2)))
            res <- c(res,n2)
        }

        if (messages>1) {
            cat(paste(res, collapse="\n")); cat("\n")
        }
        if (!is.null(object$order)) res <- res[object$order]
        res
    }

##' @export
`coef.lvmfit` <-
    function(object, type=ifelse(missing(type),-1,2),
      symbol=lava.options()$symbol,
      data, std=NULL,
      labels=lava.options()$coef.names,
      ##labels=TRUE,
      vcov,
      vcov.type, reliability=FALSE, second=FALSE, ...) {

        res <- (pars.default(object,...))
        if (type<0 && !is.null(names(res))) return(res)

        if (is.null(object$control$meanstructure)) meanstructure <- TRUE
        else meanstructure <- object$control$meanstructure
        npar <- index(object)$npar; npar.mean <- index(object)$npar.mean*meanstructure
        npar.ex <- index(object)$npar.ex

        para <- parameter(Model(object))
        para.idx <- which(vars(object)%in%para)


        if (inherits(object,"lvm.missing")) {
            if (length(object$cc)==0) {## No complete cases
                coefs <- coef(object$estimate)
                c1 <- coef(Model(object),mean=TRUE,fix=FALSE)
                c1. <- coef(Model(object),mean=FALSE,fix=FALSE)
                nn <- gsub("^[0-9]*@","",names(coefs))
                myorder <- match(c1,nn)
                myorder.reg <- order(na.omit(match(nn,c1.)))
                myorder.extra <- c()
            } else {
                myorder <- na.omit(modelPar(object$multigroup,seq_len(npar+npar.mean))$p[[object$cc]])
                myorder.reg <- na.omit(modelPar(object$multigroup,seq_len(npar))$p[[object$cc]])
                myorder.extra <- seq_len(index(object)$npar.ex)+length(myorder)
                myorder <- c(myorder,myorder.extra)

            }
        } else {
            myorder <- seq_len(npar+npar.mean)
            myorder.reg <- seq_len(npar)
            myorder.extra <- seq_len(index(object)$npar.ex)+length(myorder)
            myorder <- c(myorder,myorder.extra)
        }

        if (type<0) {
            names(res)[seq_len(length(myorder))] <- coef(Model(object),fix=FALSE, mean=meanstructure, symbol=symbol)[order(myorder)]
            return(res)
        }


        latent.var <- latent(object)
        Type <- Var <- From <- c()

        Astd <- Pstd <- vstd <- NULL
        if (!is.null(std)) {
            stdCoef <- stdcoef(object)
            {
                switch(tolower(std),
                       latent = {Astd=stdCoef$Astar; Pstd=stdCoef$Pstar; vstd=stdCoef$vstar},
                       y = {Astd=stdCoef$AstarY; Pstd=stdCoef$PstarY; vstd=stdCoef$vstarY},
                       xy = {Astd=stdCoef$AstarXY; Pstd=stdCoef$PstarXY; vstd=stdCoef$vstarXY},
                       yx = {Astd=stdCoef$AstarXY; Pstd=stdCoef$PstarXY; vstd=stdCoef$vstarXY}
                       )
            }
        }
        myparnames <- paste0("p",seq_len(npar+npar.ex))[myorder.reg]

        p <- matrices(Model(object), myparnames)
        A <- p$A
        P <- p$P
        mycoef <- object$coef
        if (!missing(vcov.type) | !missing(vcov)) {
            if (!missing(vcov)) {
                mycoef[,2] <- sqrt(diag(vcov))[myorder]
            } else {
                if (!missing(data))
                    myvcov <- information(object,type=vcov.type,data=data,inverse=TRUE)
                else
                    myvcov <- information(object,type=vcov.type,inverse=TRUE)
                mycoef[,2] <- sqrt(diag(myvcov))[myorder]
            }
            mycoef[,3] <- mycoef[,1]/mycoef[,2]
            mycoef[,4] <-  2*(pnorm(abs(mycoef[,3]),lower.tail=FALSE))
        }

        coefs <- mycoef[myorder,,drop=FALSE]
        nn <- colnames(A)

        free <- A!="0"
        free[index(object)$M1!=1] <- FALSE
        nlincon <- matrix(Model(object)$par%in%names(constrain(Model(object))),nrow(A))
        if (missing(data)) {
            data <- matrix(0,ncol=length(index(Model(object))$manifest)); colnames(data) <- index(Model(object))$manifest
        }
        nlincon.estimates.full<- constraints(object,second=second,data=data)
        nlincon.estimates <- nlincon.estimates.full[,-(5:6),drop=FALSE]
        matched <- c()
        res <- c()
        for (i in seq_len(ncol(A)))
            for (j in seq_len(nrow(A))) {
                val <- A[j,i]
                if (val!="0") {
                    matching <- match(val,rownames(coefs))
                    matched <- c(matched,matching)
                    if (!is.na(matching)) {
                        if (free[j,i])
                            newrow <- matrix(coefs[matching,],nrow=1)
                        else {
                            newrow <- matrix(c(coefs[matching,1],NA,NA,NA), nrow=1)
                        }
                    } else {
                        Debug(list("(i,j)", i, ",", j))
                        if (nlincon[j,i]) {
                            newrow <- matrix(nlincon.estimates[Model(object)$par[j,i],],nrow=1)
                        } else {
                            newrow <- matrix(c(Model(object)$fix[j,i], NA, NA, NA), nrow=1)
                        }
                    }
                    if (!is.null(std)) {
                        newrow <- cbind(newrow,Astd[j,i])
                    }
                    if (labels & !is.na(regfix(Model(object))$labels[j,i])) {
                        rownames(newrow) <- regfix(Model(object))$labels[j,i]
                        if (labels>1) {
                            newst <- paste0(nn[i],symbol[1],nn[j])
                            if (rownames(newrow)!=newst)
                                rownames(newrow) <- paste(rownames(newrow),newst,sep=":")
                        }
                    } else {
                        rownames(newrow) <- paste0(nn[i],symbol[1],nn[j])
                    }
                    if (free[j,i] | type>2) {
                        res <- rbind(res, newrow)
                        Type <- c(Type,"regression")
                        Var <- c(Var, nn[i])
                        From <- c(From, nn[j])
                    }
                }
            }
        free.var <- P!="0"
        free.var[index(object)$P1!=1] <- FALSE
        nlincon.var <- matrix(Model(object)$covpar%in%names(constrain(Model(object))),nrow(P))

        if (type>0)
            ## Variance estimates:
            for (i in seq_len(ncol(p$P)))
                for (j in seq(i,nrow(p$P))) {
                    val <- p$P[j,i]

                    if (!(i%in%para.idx))
                        if (val!="0" & !any(vars(object)[c(i,j)]%in%index(Model(object))$exogenous))
                            if (type>1 | !all(vars(object)[c(i,j)]%in%index(Model(object))$manifest))
                            {
                                matching <- match(val,rownames(coefs))
                                matched <- c(matched,matching)

                                if (!is.na(matching)) {
                                    if (free.var[j,i])
                                        newrow <- matrix(coefs[matching,],nrow=1)
                                    else
                                        newrow <- matrix(c(coefs[matching,1],NA,NA,NA), nrow=1)
                                    ## We don't want to report p-values of tests on the boundary of the parameter space
                                    if (i==j)
                                        newrow[,4] <- NA
                                } else {
                                    Debug(list("(i,j)", i, ",", j))
                                    if (nlincon.var[j,i]) {
                                        newrow <- matrix(nlincon.estimates[Model(object)$covpar[j,i],],nrow=1)
                                    } else {
                                        newrow <- matrix(c(Model(object)$covfix[j,i], NA, NA, NA), nrow=1)
                                    }
                                }
                                if (!missing(std)) {
                                    newrow <- cbind(newrow,Pstd[i,j])
                                }
                                if (length(symbol)<2) {
                                    if (nn[i]!=nn[j]) {
                                        part2 <- paste(nn[i],nn[j],sep=lava.options()$symbol[2])
                                    } else part2 <- nn[i]
                                } else {
                                    part2 <- paste0(nn[i],symbol[2],nn[j])
                                }
                                if (labels & !is.na(covfix(Model(object))$labels[j,i])) {
                                    rownames(newrow) <- covfix(Model(object))$labels[j,i]
                                    if (labels>1) {
                                        if (rownames(newrow)!=part2)
                                            rownames(newrow) <- paste(rownames(newrow),part2,sep=":")
                                    }
                                } else {
                                    rownames(newrow) <- part2
                                }
                                if ((free.var[j,i]) | type>2) {
                                    res <- rbind(res, newrow)
                                    Type <- c(Type,"variance")
                                    Var <- c(Var, nn[i])
                                    From <- c(From, nn[j])
                                }
                            }
                }

        ## Mean parameter:
        nlincon.mean <- lapply(Model(object)$mean, function(x) x%in%names(constrain(Model(object))) )

        if (type>0 & npar.mean>0) {
            midx <- seq_len(npar.mean)
            rownames(coefs)[midx] <- paste0("m",myorder[midx])
            munames <- rownames(coefs)[seq_len(npar.mean)]
            meanpar <- matrices(Model(object), myparnames, munames)$v
            for (i in seq_len(length(meanpar))) {
                if (!index(Model(object))$vars[i]%in%index(Model(object))$exogenous) {
                    val <- meanpar[i]
                    matching <- match(val,rownames(coefs))
                    if (!is.na(matching)) {
                        if (index(object)$v1[i]==1)  ## if free-parameter
                            newrow <- matrix(coefs[matching,],nrow=1)
                        else
                            newrow <- matrix(c(coefs[matching,1],NA,NA,NA), nrow=1)
                    } else {
                        if (nlincon.mean[[i]]) {
                            newrow <- matrix(nlincon.estimates[Model(object)$mean[[i]],],nrow=1)
                        } else {
                            newrow <- matrix(c(as.numeric(meanpar[i]), NA, NA, NA), nrow=1)          }
                    }
                    if (!missing(std)) {
                        newrow <- cbind(newrow,vstd[i])
                    }
                    if (labels & !(is.na(intfix(Model(object))[[i]]) | is.numeric(intfix(Model(object))[[i]]))) {
                        rownames(newrow) <- intfix(Model(object))[[i]]
                        if (labels>1) {
                            if (rownames(newrow)!=index(Model(object))$vars[i])
                                rownames(newrow) <- paste(rownames(newrow),index(Model(object))$vars[i],sep=":")
                        }
                    } else {
                        rownames(newrow) <- index(Model(object))$vars[i]
                    }
                    if ((index(object)$v1[i]==1) | type>2) {
                        res <- rbind(res, newrow)
                        Type <- c(Type,ifelse(!(i%in%para.idx),"intercept","parameter"))
                        Var <- c(Var, index(Model(object))$vars[i])
                        From <- c(From, NA)
                    }
                }
            }
        }

        if (type>0 && length(myorder.extra>0)) {
            cc <- coefs[myorder.extra,,drop=FALSE]
            rownames(cc) <- rownames(index(object)$epar)[which(index(object)$e1==1)]
            cc <- cbind(cc,rep(NA,ncol(res)-ncol(cc)))
            res <- rbind(res,cc)
            Type <- c(Type,rep("extra",length(myorder.extra)))
            Var <- c(Var,rep(NA,length(myorder.extra)))
            From <- c(From,rep(NA,length(myorder.extra)))
        }

        mycolnames <- colnames(coefs)
        if (!is.null(std)) mycolnames <- c(mycolnames, paste("std",std,sep="."))
        colnames(res) <- mycolnames
        attributes(res)$type <- Type
        attributes(res)$var <- Var
        attributes(res)$from <- From
        attributes(res)$latent <- latent.var
        attributes(res)$nlincon <- nlincon.estimates.full
        return(res)
    }

##' @export
coef.multigroup <- function(object,...) {
    return(object$parpos)
}

##' @export
coef.multigroupfit <-
    function(object, type=0,vcov, ext=FALSE,
      labels=lava.options()$coef.names,
      symbol=lava.options()$symbol,
      covsymb=NULL,groups=NULL,...) {

        if (type==0) {
            res <- pars(object);
            if (is.null(names(res))) names(res) <- object$model$name
            return(res)
        }
        if (type==1) {
            theta <- pars(object)
            if (missing(vcov))
                theta.sd <- sqrt(diag(object$vcov))
            else
                theta.sd <- sqrt(diag(vcov))
            res <- cbind(theta,theta.sd,(Z <- theta/theta.sd),2*(pnorm(abs(Z),lower.tail=FALSE)))
            if (is.null(rownames(res)))
                rownames(res) <- object$model$name
            colnames(res) <- c("Estimate","Std. Error", "Z value", "Pr(>|z|)")
            return(res)
        }

        cc <- coef(object, type=1, symbol=symbol, ...)
        model <- Model(object)
        parpos <- modelPar(model, seq_len(nrow(cc)))$p
        npar.mean <- object$model$npar.mean
        npar <- object$model$npar
        mynames <- c()
        if (npar.mean>0) {
            mynames <- unlist(object$model$meanlist)
            mynames <- names(mynames)[!duplicated(mynames)]
        }
        if (npar>0) {
            mynames <- c(mynames,object$model$par)
        }
        res <- list()
        misrow <- list()
        parpos2 <- list()
        if (is.null(groups)) groups <- seq(model$ngroup)
        if (length(groups)==0) groups <- seq(model$ngroup)
        for (i in groups) {
            orignames <- coef(object$model0$lvm[[i]],fix=FALSE,mean=object$meanstructure, messages=0, symbol=lava.options()$symbol)
            if (ext) {
                newnames. <- coef(Model(model)[[i]],fix=FALSE, mean=object$meanstructure, messages=0, labels=labels, symbol=symbol)
                newnames <- coef(Model(model)[[i]],fix=FALSE, mean=object$meanstructure, messages=0, labels=labels,symbol=lava.options()$symbol)
                newcoef <- matrix(NA,ncol=4,nrow=length(newnames))
                rownames(newcoef) <- newnames.
                idx <- match(orignames,newnames)
                newcoef[idx,] <- cc[parpos[[i]],,drop=FALSE]
                newparpos <- rep(NA,length(newnames))
                newparpos[idx] <- parpos[[i]]
                parpos2 <- c(parpos2, list(newparpos))
                misrow <- c(misrow, list(setdiff(seq_len(length(newnames)),idx)))
            } else {
                newcoef <- cc[parpos[[i]],,drop=FALSE]
                rownames(newcoef) <- orignames
            }
            colnames(newcoef) <- colnames(cc)
            ## Position of variance parameters:
            varpos <- variances(Model(model)[[i]],mean=FALSE)
            ## Number of parameters resp mean-parameters
            p <- nrow(newcoef); p0 <- length(coef(Model(model)[[i]],fix=FALSE, mean=FALSE, messages=0))
            newcoef[(p-p0) + varpos,4] <- NA
            res <- c(res, list(newcoef))
        }

        if (ext) {
            for (i in seq(length(groups))) {
                if (length(misrow[[i]])>0) {
                    nn <- rownames(res[[i]])[misrow[[i]]]
                    for (j in setdiff(seq_len(length(groups)),i)) {
                        nn2 <- rownames(res[[j]])
                        matching <- na.omit(match(nn,nn2))
                        matching <- setdiff(matching,misrow[[j]])
                        if (length(matching)>0) {
                            idxj <- match(nn2[matching],nn2)
                            idxi <- match(nn2[matching],rownames(res[[i]]))
                            res[[i]][nn2[matching],] <- res[[j]][nn2[matching],]
                            parpos2[[i]][idxi] <- parpos2[[j]][idxj]
                            nn <- setdiff(nn,nn2[matching])
                        }
                        if (length(nn)<1) break;
                    }
                }
            }
            attributes(res)$parpos <- parpos2
        }
        return(res)
    }

##' @export
CoefMat.multigroupfit <- function(x,type=9,
                           labels=lava.options()$coef.names,
                           symbol=lava.options()$symbol[1],
                           data=NULL,groups=seq(Model(x)$ngroup),...) {

    cc <- coef(x,type=type,ext=TRUE,symbol=symbol,data=data,groups=groups)
    parpos <- attributes(cc)$parpos
    res <- c()
    nlincon.estimates <- c()
    nlincon.names <- c()
    count <- k <- 0

    for (i in groups) {
        k <- k+1
        m0 <- Model(Model(x))[[i]]
        mycoef <- cc[[k]]
        npar <- index(m0)$npar
        npar.mean <- index(m0)$npar.mean
        if (npar>0)
            rownames(mycoef)[(seq(npar))+npar.mean] <- paste0("p",seq(npar))
        m0$coefficients <- mycoef
        m0$opt$estimate <- mycoef[,1]
        Vcov <- vcov(x)[parpos[[k]],parpos[[k]],drop=FALSE]; colnames(Vcov) <- rownames(Vcov) <- rownames(mycoef)
        m0$vcov <- Vcov
        cc0 <- coef.lvmfit(m0,type=type,labels=labels,symbol=symbol)
        attributes(cc0)$dispname <- x$opt$dispname
        res <- c(res, list(CoefMat(cc0)))
        newnlin <- attributes(cc0)$nlincon
        if (length(newnlin)>0)
            if (count==0) {
                count <- count+1
                nlincon.estimates <- newnlin
                nlincon.names <- rownames(newnlin)
            } else {
                for (j in seq_len(NROW(newnlin))) {
                    if (!(rownames(newnlin)[j]%in%nlincon.names)) {
                        nlincon.estimates <- rbind(nlincon.estimates,newnlin[j,,drop=FALSE])
                        nlincon.names <- c(nlincon.names,rownames(newnlin)[j])
                    }
                }
            }
    }
    rownames(nlincon.estimates) <- nlincon.names
    attributes(res)$nlincon <- nlincon.estimates
    return(res)
}


##' @export
CoefMat <- function(x,
             digits = max(3, getOption("digits") - 2),
             type=9,
             symbol=lava.options()$symbol[1],...) {
    cc <- x
    if (!is.matrix(x)) {
        cc <- coef(x,type=type,symbol=symbol,...)
    }
    res <- c()
    mycoef <- format(round(cc,max(1,digits)),digits=digits)
    mycoef[,4] <- formatC(cc[,4],digits=digits-1,format="g",
                          preserve.width="common",flag="")
    mycoef[is.na(cc)] <- ""
    mycoef[cc[,4]<1e-12,4] <- "  <1e-12"

    M <- ncol(cc)
    Nreg <- sum(attributes(cc)$type=="regression")
    Nvar <- sum(attributes(cc)$type=="variance")
    Nint <- sum(attributes(cc)$type=="intercept")

    latent.var <- attributes(cc)$latent


    if (Nreg>0) {
        reg.idx <- which(attributes(cc)$type=="regression")
        latent.from <- which(attributes(cc)$from[reg.idx]%in%latent.var)
        latent.from <- latent.from[which(is.na(match(attributes(cc)$var[latent.from],latent.var)))]

        reg.idx <- setdiff(reg.idx,latent.from)
        Nmeas <- length(latent.from)
        if (Nmeas>0) {
            first.entry <- c()
            for (i in latent.var) {
                pos <- match(i,attributes(cc)$from[latent.from])
                if (!is.na(pos))
                    first.entry <- c(first.entry, pos)
            }
            res <- rbind(res, c("Measurements:",rep("",M)))
            count <- 0
            Delta <- FALSE
            for (i in latent.var) {
                count <- count+1
                Delta <- !Delta
                Myidx <- which(attributes(cc)$from==i & attributes(cc)$type=="regression" & !(attributes(cc)$var%in%latent.var))

                prefix <- ifelse(Delta,"  ","   ")
                for (j in Myidx) {
                    newrow <- mycoef[j,]
                    newname <- rownames(cc)[j]
                    res <- rbind(res,c(paste(prefix,newname),newrow))
                }
            }
        }
        if ((Nreg-Nmeas)>0) {
            responses <- unique(attributes(cc)$var[reg.idx])
            first.entry <- c()
            for (i in responses) {
                pos <- match(i,attributes(cc)$var[reg.idx])
                first.entry <- c(first.entry, pos)
            }
            res <- rbind(res, c("Regressions:",rep("",M)))
            count <- 0
            Delta <- FALSE
            for (i in reg.idx) {
                count <- count+1
                newrow <- mycoef[i,]
                newname <- rownames(cc)[i]
                if (count%in%first.entry) Delta <- !Delta
                prefix <- ifelse(Delta,"  ","   ")
                res <- rbind(res,c(paste(prefix,newname),newrow))
            }
        }
    }


    if (Nint>0) {
        int.idx <- which(attributes(cc)$type=="intercept")
        res <- rbind(res, c("Intercepts:",rep("",M)))
        for (i in int.idx) {
            newrow <- mycoef[i,]
            newname <- rownames(cc)[i]
            res <- rbind(res,c(paste("  ",newname),newrow))
        }
    }
    par.idx <- which(attributes(cc)$type=="parameter")
    parres <- rbind(c("Additional Parameters:",rep("",M)))
    for (i in par.idx) {
        newrow <- mycoef[i,]
        newname <- rownames(cc)[i]
        parres <- rbind(parres,c(paste("  ",newname),newrow))
    }
    extra.idx <- which(attributes(cc)$type=="extra")
    for (i in extra.idx) {
        newrow <- mycoef[i,]
        newname <- rownames(cc)[i]
        parres <- rbind(parres,c(paste("  ",newname),newrow))
    }
    if (nrow(parres)>1) res <- rbind(res,parres)



    if (Nvar>0) {
        var.idx <- which(attributes(cc)$type=="variance")
        vname <- "Residual Variances:"
        if (!is.list(x)) {
            if (!is.null(attributes(x)$dispname)) vname <- attributes(x)$dispname
        } else if (!is.null(x$opt$dispname)) vname <- x$opt$dispname
        res <- rbind(res, c(vname,rep("",M)))
        for (i in var.idx) {
            newrow <- mycoef[i,]
            newname <- rownames(cc)[i]
            res <- rbind(res,c(paste("  ",newname),newrow))
        }
    }
    res0 <- res[,-1]
    rownames(res0) <- format(res[,1],justify="left")
    res0
}

stdcoef <- function(x,p=coef(x),...) {
    M0 <- moments(x,p=p,...)
    A <- t(M0$A)
    P <- M0$P
    v <- M0$v
    C <- M0$Cfull
    N <- diag(sqrt(diag(C)),ncol=nrow(C)); colnames(N) <- rownames(N) <- vars(x)
    iN <- N; diag(iN)[diag(N)>0] <- 1/diag(iN)[diag(N)>0]
    diag(iN)[diag(N)==0] <- NA
    Nn <- N; Nn[] <- 0; diag(Nn) <- 1
    Nn[latent(x),latent(x)] <- N[latent(x),latent(x)]
    iNn <- Nn; diag(iNn) <- 1/diag(Nn)
    Ny <- Nn;
    Ny[endogenous(x),endogenous(x)] <- N[endogenous(x),endogenous(x)]
    iNy <- Ny; diag(iNy) <- 1/diag(Ny)
    ## Standardized w.r.t. latent,y and x:
    AstarXY <- t(iN%*%A%*%N)
    PstarXY <- iN%*%P%*%iN
    if (!is.null(v))
        vstarXY <- iN%*%v
    else
        vstarXY <- NULL
    pstdXY <- pars(Model(x),A=AstarXY,P=PstarXY,v=vstarXY)
    ## Standardized w.r.t. latent, y:
    AstarY <- t(iNy%*%A%*%Ny)
    PstarY <- iNy%*%P%*%iNy
    if (!is.null(v))
        vstarY <- iNy%*%v
    else
        vstarY <- NULL
    ## Standardized w.r.t. latent only:
    Astar <- t(iNn%*%A%*%Nn)
    Pstar <- iNn%*%P%*%iNn
    if (!is.null(v))
        vstar <- iNn%*%v
    else
        vstar <- NULL
    pstd <- pars(Model(x),A=Astar,Pstar,v=vstar)
    k <- length(p)-length(pstd)

    res <- list(par=cbind(p,c(pstd,rep(NA,k)),c(pstdXY,rep(NA,k))),
                AstarXY=AstarXY, PstarXY=PstarXY, vstarXY=vstarXY,
                AstarY=AstarY, PstarY=PstarY, vstarY=vstarY,
                Astar=Astar, Pstar=Pstar, vstar=vstar)
    return(res)
}

##' @export
formula.lvm <- function(x,char=FALSE,all=FALSE,...) {
    A <- index(x)$M
    res <- c()
    for (i in seq_len(ncol(A))) {
        if (all || !(colnames(A)[i]%in%c(index(x)$exogenous,parameter(x)) )) {
            f <- paste(colnames(A)[i],"~ 1")
            if (any(A[,i]!=0)) {
                f <- (paste(colnames(A)[i],"~",paste(colnames(A)[A[,i]!=0],collapse="+")))
            }
            if (!char)
                f <- formula(f)
            res <- c(res, list(f))
        }
    }
    return(res)
}


##' @export
formula.lvmfit <- formula.lvm

##' @export
IC <- function(x, ...) UseMethod("IC")

##' @export
influence.estimate <- function(model, ...)
    IC(model, ...)


##' Extract influence function from model object
##'
##' Extract i.i.d. decomposition (influence function) from model object
##' @aliases IC var_ic
##' @param x model object
##' @param id (optional) id/cluster variable
##' @param bread (optional) Inverse of derivative of mean score function
##' @param ... additional arguments
##' @examples
##' m <- lvm(y~x+z)
##' distribution(m, ~y+z) <- binomial.lvm("logit")
##' d <- sim(m,1e3)
##' g <- glm(y~x+z,data=d,family=binomial)
##' var_ic(IC(g))
##' @export
IC.default <- function(x, bread, id=NULL,
                       ...) {
  if (any(paste("iid", class(x), sep=".") %in% methods("iid"))) {
    ## 'iid' method exists for the specific class.
    ## This is a scaled version of the influence function, hence
    ## we need to rescale.
    cl <- match.call()
    cl[[1]] <- substitute(iid)
    ii <- eval.parent(cl)
    if (!is.null(attr(ii, "bread"))) {
      attr(res, "bread") <- attr(res, "bread")*NROW(res)
    }
    ii <- ii*NROW(ii)
    return(ii)
  }
  if (!any(paste("score", class(x), sep=".") %in% methods("score"))) {
    warning("Not available for this class")
    return(NULL)
  }

  ## if (folds>0 || maxsize>0 ||
  ##     (!missing(id) && lava.options()$cluster.index)) {
  ##   if (!requireNamespace("mets", quietly=TRUE)) stop("Requires 'mets'")
  ## }
  ## if (folds>0) {
  ##   U <- Reduce(
  ##     "rbind",
  ##     mets::divide.conquer(function(data) score(x, data = data, ...),
  ##                          id = id,
  ##                          data = data, size = round(nrow(data) / folds)
  ##       )
  ##   )
  ## } else {
  U <- score(x, indiv=TRUE, ...)
  pp <- pars(x)
  if (!missing(bread) && is.null(bread)) {
    bread <- suppressWarnings(vcov(x)*NROW(U))
  }
  if (missing(bread)) bread <- attributes(U)$bread
  if (is.null(bread)) {
    bread <- attributes(x)$bread
    if (is.null(bread)) bread <- x$bread
    if (is.null(bread)) {
      ## if (maxsize>0) {
      ##   ff <- function(p) {
      ##     colSums(Reduce(
      ##       "rbind",
      ##       mets::divide.conquer(function(data)
      ##         score(x, data = data, p = p, ...),
      ##         data = data, size = maxsize
      ##       )
      ##     ))
      ##   }
      ##   I <- -numDeriv::jacobian(ff, pp, method = lava.options()$Dmethod)
      ## } else {
      I <- -numDeriv::jacobian(function(p) {
        score(x, p = p, indiv = FALSE, ...)
      }, pp, method = lava.options()$Dmethod)
      bread <- Inverse(I)*NROW(U)
    }
  }

  ic0 <- U%*%bread
  if (!missing(id)) {
  N <- nrow(ic0)
  if (!lava.options()$cluster.index) {
    ic0 <- matrix(unlist(by(ic0, id, colSums)),
                  byrow=TRUE, ncol=ncol(bread))
  } else {
    ic0 <- mets::cluster.index(id, mat=ic0, return.all=FALSE)
  }
  ic0 <- ic0*NROW(ic0)/length(id)
  attributes(ic0)$N <- N
}
colnames(ic0) <- colnames(U)
return(structure(ic0, bread=bread))
}

##' @export
IC.multigroupfit <- function(x, ...) IC.default(x, combine=TRUE, ...)

##' @export
IC.matrix <- function(x, ...) {
    p <- NCOL(x)
    n <- NROW(x)
    mu <- colMeans(x, na.rm = TRUE)
    S <- var(x, use = "pairwise.complete.obs") * (n - 1) / n
    ic1 <- t(t(x)-mu)
    ic2 <- matrix(ncol = (p + 1) * p / 2, nrow = n)
    pos <- 0
    nn <- c()
    cc <- mu
    for (i in seq(p))
        for (j in seq(i, p)) {
            pos <- pos+1
            cc <- c(cc, S[i, j])
            ic2[, pos] <- (ic1[, i] * ic1[, j]) - cc[length(cc)]
            nn <- c(nn, paste(colnames(x)[c(i, j)],
                              collapse = lava.options()$symbols[2]
            ))
        }
    colnames(ic1) <- colnames(x)
    colnames(ic2) <- nn
    names(cc) <- c(colnames(ic1), colnames(ic2))
    res <- cbind(ic1, ic2)
    rownames(res) <- rownames(x)
    structure(res,
              coef=cc,
              mean=mu, var=S)
}

##' @export
IC.numeric <- function(x, ...) {
    n <- length(x)
    mu <- mean(x)
    S <- var(x) * (n - 1) / n
    ic1 <- t(t(x) - mu)
    structure(cbind(mean=ic1, var=(ic1^2 - S)),
              coef=c(mean=mu, var=S), mean=mu, var=S)
}

##' @export
IC.data.frame <- function(x, ...) {
  if (!all(apply(x[1, , drop=FALSE], 2,
                 function(x) inherits(x, c("numeric", "integer")))))
        stop("Don't know how to handle data.frames of this type")
    IC(as.matrix(x))
}

##v <- lmerplot(l1,varcomp=TRUE,colorkey=TRUE,lwd=0,col=rainbow(20))
lmerplot <- function(model,x,id,y,transform,re.form=NULL,varcomp=FALSE,colorbar=TRUE,mar=c(4,4,4,6),col,index=seq(50),...) {
    if (varcomp) {
        Z <- lme4::getME(model,"Z")
        nn <- unlist(lapply(lme4::getME(model,"Ztlist"),nrow))
        ve <- lme4::getME(model,"sigma")^2
        vu <- varcomp(model,profile=FALSE)$varcomp
        L <- Matrix::Diagonal(sum(nn),rep(vu,nn))
        V <- Z%*%L%*%(Matrix::t(Z))
        Matrix::diag(V) <- Matrix::diag(V)+ve
        cV <- Matrix::cov2cor(V)
        if (!is.null(index)) {
            index <- intersect(seq(nrow(cV)),index)
            cV <- cV[index,index,drop=FALSE]
        }
        if (colorbar) { opt <- par(mar=mar) }
        if (missing(col)) col <- rev(gray.colors(16,0,1))
        image(seq(nrow(cV)),seq(ncol(cV)),as.matrix(cV),xlab="",ylab="",col=col,zlim=c(0,1),...)
        if (colorbar) {
            uu <- devcoords()
            xrange <- c(uu$fig.x2,uu$dev.x2)
            xrange <- diff(xrange)/3*c(1,-1)+xrange
            yrange <- c(uu$fig.y1,uu$fig.y2)
            colorbar(direction="vertical",x.range=xrange,y.range=yrange,clut=col,values=seq(0,1,length.out=length(col)),srt=0,position=2)
            par(opt)
        }
        return(invisible(V))
    }
    if (missing(y)) y <- model.frame(model)[,1]
    yhat <- predict(model)
    if (!is.null(re.form)) ymean <- predict(model,re.form=re.form)
    if (!missing(transform)) {
        yhat <- transform(yhat)
        if (!is.null(re.form)) ymean <- transform(ymean)
        y <- transform(y)
    }
    plot(y ~ x, col=Col(id,0.3), pch=16,...)
    if (!is.null(re.form)) points(ymean ~ x, pch="-",cex=4);
    for (i in unique(id)) {
        idx <- which(id==i)
        lines(yhat[idx]~x[idx],col=i)
    }
}

varcomp <- function(x,profile=TRUE,...) {
    cc <- cbind(lme4::fixef(x),diag(as.matrix(vcov(x)))^.5)
    cc <- cbind(cc,cc[,1]-qnorm(0.975)*cc[,2],cc[,1]+qnorm(0.975)*cc[,2],
                2*(1-pnorm(abs(cc[,1])/cc[,2])))
    pr <- NULL
    if (profile) pr <- confint(x)
    colnames(cc) <- c("Estimate","Std.Err","2.5%","97.5%","p-value")
    vc <- lme4::VarCorr(x)
    res <- structure(list(coef=lme4::fixef(x), vcov=as.matrix(vcov(x)),
                          coefmat=cc,
                          confint=pr,
                          varcomp=vc[[1]][,],
                          residual=attributes(vc)$sc^2
                          ),
                     class="estimate.lmer")
    res
}

###{{{ summary.lvm

##' @export
`summary.lvm` <-
function(object,...) {
  k <- length(vars(object))
  ## cat("Latent Variable Model \n\twith: ", k, " variables.\n", sep="");
  print(object,print.transform=FALSE,...)
  if (length(transform(object))>0) {
      cat("\nTransformations:\n")
      print(transform(object),quote=FALSE,...)
  }
  cat("\n")
  if (length(index(object))>0)
      cat("Number of free parameters: ", with(index(object),npar+npar.mean+npar.ex),"\n", sep="")

  if (k==0)
    return()
  ##cat("Npar=", index(object)$npar, "+", index(object)$npar.mean, "\n", sep="")
  cat("\n")
  print(regression(object),...)
  print(covariance(object),...)
  print(intercept(object),...)
  if (length(object$exfix)>0) {
    cat("Additional parameters:\n")
    val <- unlist(object$exfix)
    M <- rbind(val); colnames(M) <- names(val)
    rownames(M) <- "   "
    print(M,quote=FALSE,...)
  }
  if (length(constrain(object))>0) {
    cat("Non-linear constraints:\n")
    print(constrain(object),quote=FALSE,...)
  }
  ## printmany(object$cov, printmany(object$covpar, object$covfix, name1="Labels:", name2="Fixed:", print=FALSE), name1="covariance:")
  cat("\n")
}

###}}} summary.lvm

###{{{ summary.lvmfit

##' @export
`summary.lvmfit` <-
function(object,std="xy", type=9, labels=2, ...) {
  cc <- CoefMat(object,labels=labels,std=std,type=type,...)
  mycoef <- coef(object,type=9)
  nlincon <- attributes(mycoef)$nlincon
  nonexo <- setdiff(vars(object),index(Model(object))$exogenous)
  attributes(mycoef) <- attributes(mycoef)[1:2]
  mygof <- object$opt$summary.message
  if (is.null(mygof)) {
    mygof <- gof
  }
  if (class(object)[1]=="lvm.missing") {
    nn <- unlist(lapply(object$multigroup$data, nrow))
    nc <- nn[object$cc]
    if (length(nc)==0) nc <- 0
    ngroup <- object$multigroup$ngroup
    res <- list(object=object, coef=mycoef, coefmat=cc, nlincon=nlincon, gof=mygof(object), n=sum(nn), nc=nc, ngroup=ngroup,
                varmat=modelVar(object)$P[nonexo,nonexo], latent=latent(object), opt=object$opt, vcov=vcov(object), estimator=object$estimator, rsq=rsq(object))
  } else {
    n <- nrow(model.frame(object))
    if (is.null(n)) n <- model.frame(object)$n
    res <- list(coef=mycoef, coefmat=cc, nlincon=nlincon, gof=mygof(object), n=n, nc=n, latent=latent(object),
                opt=object$opt, vcov=vcov(object), estimator=object$estimator, rsq=rsq(object))##, varmat=modelVar(object)$P[nonexo,nonexo])
  }
  class(res) <- "summary.lvmfit"
  res
}

##' @export
print.summary.lvmfit <- function(x,varmat=TRUE,...) {
  if (!is.null(x$control$method)) {
    l2D <- sum(x$opt$grad^2)
    rnkV <- qr(x$vcov)$rank
    if (l2D>1e-2) warning("Possible problems with convergence!")
    cat("||score||^2=",l2D,"\n",sep="")
    np <- nrow(x$vcov)
    if (rnkV<np) warning("Possible problems with identification (rank(informaion)=",rnkV,"<",np,"!")
  }
  cat("Latent variables:", x$latent, "\n")
  cat("Number of rows in data=",x$n,sep="")
  if (x$nc!=x$n) {
    cat(" (",x$nc," complete cases, ", x$ngroup, " groups)",sep="")
  }; cat("\n")
  print(cli::rule())
  print(x$coefmat,quote=FALSE,right=TRUE)
##  if (varmat) {
##    cat("\nResidual covariance matrix:\n")
##    print(x$varmat)
##  }
  if (!is.null(x$nlincon)) {
    cat("\nNon-linear constraints:\n")
    printCoefmat(x$nlincon,signif.stars=FALSE)
  }
  print(cli::rule())
  cat("Estimator:",x$estimator,"\n")
  print(cli::rule())
  if (!is.null(x$gof)) {
    if (class(x$gof)[1]=="list") {
      for (i in x$gof) {
        print(i)
      }
    } else {
      print(x$gof,optim=FALSE)
    }
    print(cli::rule())
  }
  if (!is.null(x$rsq)) {
      if (!is.list(x$rsq)) {
          cat("R-square\n")
          print(round(x$rsq,3),quote=FALSE)
      } else {
          for (i in seq_len(length(x$rsq))) {
              cat(names(x$rsq)[i],"\n")
              print(round(x$rsq[[i]],3),quote=FALSE)
          }
      }
  }
  invisible(x)
}

##' @export
coef.summary.lvmfit <- function(object,...) object$coef

###}}} summary.lvmfit

###{{{ summary.multigroupfit

##' @export
summary.multigroupfit <- function(object,groups=NULL,...) {
  if (is.null(groups) | length(groups)==0) {
    if (object$model$missing) {
      groups <- object$model$complete
      if (length(groups)==0)
        groups <- seq_len(object$model0$ngroup)
    } else {
      groups <- seq_len(object$model$ngroup)
    }
  }
  cc <- CoefMat.multigroupfit(object,groups=groups,...)
  res <- list(coef=coef(object,type=2,groups=groups,...), object=object, coefmat=cc, gof=gof(object), object=object, opt=object$opt, latent=object$latent, estimator=object$estimator)
  class(res) <- "summary.multigroupfit"
  res
}

##' @export
print.summary.multigroupfit <- function(x,...) {
  l2D <- sum(x$opt$grad^2)
  if (l2D>1e-2) warning("Possible problems with convergence!")
  cat("||score||^2=",l2D,"\n")
  cat("Latent variables:", x$latent, "\n")
  print(x$object,...)
  ##print(x$coefmat,quote=FALSE,right=TRUE)
  print(cli::rule())
  if (!is.null(attributes(x$coefmat)$nlincon)) {
    cat("Non-linear constraints:\n")
    print(attributes(x$coefmat)$nlincon)
    print(cli::rule())
  }
  cat("Estimator:",x$estimator,"\n")
  print(cli::rule())
  if (!is.null(x$gof)) {
    print(x$gof)
    print(cli::rule())
  }
  invisible(x)
}

###}}} summary.multigroupfit

###{{{ summary.multigroup

##' @export
summary.multigroup <- function(object,...) {
  for (m in object$lvm)
    print(m, ...)
  print(object)
  invisible(object)
}

###}}}

##' Add time-varying covariate effects to model
##'
##' @title Time-dependent parameters
##' @param object Model
##' @param formula Formula with rhs specifying time-varying covariates
##' @param rate Optional rate parameters. If given as a vector this
##' parameter is interpreted as the raw (baseline-)rates within each
##' time interval defined by \code{timecut}.  If given as a matrix the
##' parameters are interpreted as log-rates (and log-rate-ratios for
##' the time-varying covariates defined in the formula).
##' @param timecut Time intervals
##' @param type Type of model (default piecewise constant intensity)
##' @param ... Additional arguments to lower level functions
##' @author Klaus K. Holst
##' @aliases timedep timedep<-
##' @export
##' @examples
##'
##' ## Piecewise constant hazard
##' m <- lvm(y~1)
##' m <- timedep(m,y~1,timecut=c(0,5),rate=c(0.5,0.3))
##'
##' \dontrun{
##' d <- sim(m,1e4); d$status <- TRUE
##' dd <- mets::lifetable(Surv(y,status)~1,data=d,breaks=c(0,5,10));
##' exp(coef(glm(events ~ offset(log(atrisk)) + -1 + interval, dd, family=poisson)))
##' }
##'
##'
##' ## Piecewise constant hazard and time-varying effect of z1
##' m <- lvm(y~1)
##' distribution(m,~z1) <- dist_ones(0.5)
##' R <- log(cbind(c(0.2,0.7,0.9),c(0.5,0.3,0.3)))
##' m <- timedep(m,y~z1,timecut=c(0,3,5),rate=R)
##'
##' \dontrun{
##' d <- sim(m,1e4); d$status <- TRUE
##' dd <- mets::lifetable(Surv(y,status)~z1,data=d,breaks=c(0,3,5,Inf));
##' exp(coef(glm(events ~ offset(log(atrisk)) + -1 + interval+z1:interval, dd, family=poisson)))
##' }
##'
##'
##'
##' ## Explicit simulation of time-varying effects
##' m <- lvm(y~1)
##' distribution(m,~z1) <- dist_ones(0.5)
##' distribution(m,~z2) <- dist_bernoulli(p=0.5)
##' #variance(m,~m1+m2) <- 0
##' #regression(m,m1[m1:0] ~ z1) <- log(0.5)
##' #regression(m,m2[m2:0] ~ z1) <- log(0.3)
##' regression(m,m1 ~ z1,variance=0) <- log(0.5)
##' regression(m,m2 ~ z1,variance=0) <- log(0.3)
##' intercept(m,~m1+m2) <- c(-0.5,0)
##' m <- timedep(m,y~m1+m2,timecut=c(0,5))
##'
##' \dontrun{
##' d <- sim(m,1e5); d$status <- TRUE
##' dd <- mets::lifetable(Surv(y,status)~z1,data=d,breaks=c(0,5,Inf))
##' exp(coef(glm(events ~ offset(log(atrisk)) + -1 + interval + interval:z1, dd, family=poisson)))
##' }
timedep <- function(object,formula,rate,timecut,type="dist_cox_exponential",...) {
    if (missing(timecut)) stop("'timecut' needed")
    ##if (inherits(formula,"formula"))
    ff <- getoutcome(formula)
    simvars <- attributes(ff)$x
    if (is.null(object$attributes$simvar)) {
        object$attributes$simvar <- list(simvars)
        names(object$attributes$simvar) <- ff
        object$attributes$timedep <- object$attributes$simvar
    } else {
        object$attributes$simvar[[ff]] <- simvars
        object$attributes$timedep[[ff]] <- simvars
    }
    if (missing(rate)) rate <- rep(1,length(timecut))
    
    args <- list(timecut=timecut,rate=rate,...)
    covariance(object,ff) <- 1
    distribution(object,ff) <- do.call(type,args)
    return(object)
}

##' @export
"timedep<-" <- function(object,...,value) {
    timedep(object,value,...)
}

##' @export
"latent<-" <- function(x,...,value) UseMethod("latent<-")

##' @export
"latent<-.lvm" <- function(x, clear=FALSE,..., value) {
    if (inherits(value,"formula")) {
        return(latent(x,all.vars(value),clear=clear,...))
    }
    latent(x, var=value, clear=clear,...)
}

##' @export
`latent` <-
    function(x,...) UseMethod("latent")

##' @export
`latent.lvm` <- function(x,var,clear=FALSE,messages=lava.options()$messages,...) {
    if (missing(var)) {
        latentidx <- unlist(x$latent)
        if (length(latentidx)>0)
            return(names(latentidx))
        else
            return(NULL)
    }
    if (inherits(var,"formula")) var <- all.vars(var)
    if (clear) {
        x$noderender$shape[var] <- "rectangle"
        x$latent[var] <- NULL
        ## intfix(x,var) <- NA
    } else {
        if (!all(var%in%vars(x))) {
            addvar(x,messages=messages,reindex=FALSE,) <- setdiff(var,vars(x))
        }
        x$noderender$shape[var] <- "ellipse"
        x$latent[var] <- TRUE
        ord <- intersect(var,ordinal(x))
        if (length(ord)>0) ordinal(x,K=NULL) <- ord
    }

    xorg <- exogenous(x)
    exoset <- setdiff(xorg,var)
    if (length(exoset)<length(xorg)) {
        exogenous(x) <- exoset
    }

    index(x) <- reindex(x)
    return(x)
}

##' @export
`latent.lvmfit` <-
    function(x,clear=FALSE,...) {
        latent(Model(x),...)
    }

##' @export
latent.list <- function(x,...) {
    latlist <- c()
    for (i in seq_along(x)) {
        latlist <- c(latlist, latent(x[[i]]))
    }
    latlist <- unique(latlist)
    return(latlist)
}

##' @export
`latent.multigroup` <-
    function(x,...) {
        latent(Model(x))
    }

###{{{ distribution

##' @export
"distribution<-" <- function(x,...,value) UseMethod("distribution<-")

##' @export
"distribution" <- function(x,...,value) UseMethod("distribution")

##' @export
"distribution<-.lvm" <- function(x,variable,parname=NULL,init,mdist=FALSE,...,value) {
    if (inherits(variable,"formula")) variable <- all.vars(variable)
    dots <- list(...)

    if (!missing(value)) {
        for (obj in c("variable","parname","init","mdist"))
            if (!is.null(attr(value,obj)) && eval(substitute(missing(a),list(a=obj))))
                assign(obj,attr(value,obj))

        if (is.function(value) & mdist & is.null(parname)) parname <- TRUE
    }
    if (!is.null(parname) || length(dots)>0) {
        if (length(parname)>1 || (is.character(parname))) {
            if (missing(init)) {
                parameter(x,start=rep(1,length(parname))) <- parname
            } else {
                parameter(x,start=init) <- parname
            }
            gen <- function(n,p,...) {
                args <- c(n,as.list(p[parname]),dots)
                names(args) <- names(formals(value))[seq(length(parname)+1)]
                do.call(value,args)
            }
        } else {
            gen <- value
            if ("..."%ni%names(formals(gen))) formals(gen) <- c(formals(gen),alist(...=))
            formals(gen) <- modifyList(formals(gen),dots)
        }
        distribution(x,variable,mdist=TRUE) <- list(gen)
        return(x)
    }

    if (length(variable)==1 && !mdist) {
        addvar(x) <- as.formula(paste("~",variable))
        if (is.numeric(value)) value <- list(value)
        if (!is.null(attributes(value)$mean)) intercept(x,variable) <- attributes(value)$mean
        if (!is.null(attributes(value)$variance)) variance(x,variable,exo=TRUE) <- attributes(value)$variance
        x$attributes$distribution[[variable]] <- value
        ## Remove from 'mdistribution'
        vars <- which(names(x$attributes$mdistribution$var)%in%variable)
        for (i in vars) {
            pos <- x$attributes$mdistribution$var[[i]]
            x$attributes$mdistribution$fun[pos] <- NULL
            x$attributes$mdistribution$var[which(x$attributes$mdistribution$var==pos)] <- NULL
            above <- which(x$attributes$mdistribution$var>pos)
            if (length(above)>0)
                x$attributes$mdistribution$var[above] <- lapply(x$attributes$mdistribution$var[above],function(x) x-1)
        }
        return(x)
    }

    if (is.list(value) && length(value)==1 && (is.function(value[[1]]) || is.null(value[[1]]))) {
        addvar(x) <- variable
        ## Multivariate distribution
        if (is.null(x$attributes$mdistribution)) x$attributes$mdistribution <- list(var=list(), fun=list())
        vars <- x$attributes$mdistribution$var

        if (any(ii <- which(names(vars)%in%variable))) {
            num <- unique(unlist(vars[ii]))
            vars[which(unlist(vars)%in%num)] <- NULL
            newfunlist <- list()
            numleft <- unique(unlist(vars))
            for (i in seq_along(numleft)) {
                newfunlist <- c(newfunlist, x$attributes$mdistribution$fun[[numleft[i]]])
                ii <- which(unlist(vars)==numleft[i])
                vars[ii] <- i
            }
            K <- length(numleft)
            x$attributes$mdistribution$var <- vars
            x$attributes$mdistribution$fun <- newfunlist
        } else {
            K <- length(x$attributes$mdistribution$fun)
        }
        if (length(distribution(x))>0)
            distribution(x,variable) <- rep(list(NULL),length(variable))

        x$attributes$mdistribution$var[variable] <- K+1
        x$attributes$mdistribution$fun <- c(x$attributes$mdistribution$fun,value)

        return(x)
    }

    if ((length(value)!=length(variable) & length(value)!=1))
        stop("Wrong number of values")
    for (i in seq_along(variable))
        if (length(value)==1) {
            distribution(x,variable[i],...) <- value
        } else {
            distribution(x,variable[i],...) <- value[[i]]
        }
    return(x)

}

##' @export
"distribution.lvm" <- function(x,var,value,...,multivariate=FALSE) {
    if (!missing(value)) {
        distribution(x,var,...) <- value
        return(x)
    }
    if (multivariate) return(x$attributes$mdistribution)
    x$attributes$distribution[var]
}

###}}} distribution

###{{{ normal/gaussian

##' @export
normal.lvm <- function(link="identity",mean,sd,log=FALSE,...) {
    rnormal <- if(log) rlnorm else rnorm
    fam <- stats::gaussian(link); fam$link <- link
    f <- function(n,mu,var,...) rnormal(n,fam$linkinv(mu),sqrt(var))
    if (!missing(mean)) attr(f,"mean") <- mean
    if (!missing(sd)) attr(f,"variance") <- sd^2
    attr(f,"family") <- fam
    return(f)
}

##' @export
gaussian.lvm <- normal.lvm

##' @export
lognormal.lvm <- function(...) structure(normal.lvm(...,log=TRUE),family=list(family="log-normal",...))

###}}} normal/gaussian

###{{{ poisson

##' @export
poisson.lvm <- function(link="log",lambda,...) {
    fam <- stats::poisson(link); fam$link <- link
    f <- function(n,mu,...) {
        if (missing(n)) {
            return(fam)
        }
        rpois(n,fam$linkinv(mu))
    }
    if (!missing(lambda)) attr(f,"mean") <- fam$linkfun(lambda)
    attr(f,"family") <- fam
    attr(f,"var") <- FALSE
    return(f)
}

###}}} poisson

###{{{ pareto

## @examples
## m <- lvm()
## categorical(m,K=3) <- ~x
## distribution(m,~y) <- pareto.lvm(lambda=1)
## regression(m,additive=FALSE) <- y~x
## regression(m) <- y~z
## d <- sim(m,1e4,p=c("y~x:0"=1,"y~x:1"=1,"y~x:2"=exp(1)))
##
## X <- model.matrix(y~-1+factor(x)+z,data=d)
## mlogL <- function(theta) {
##     lambda <- exp(theta[1])
##     mu <- exp(X%*%theta[-1])
##     -sum(log(lambda*mu*(1+mu*d$y)^{-lambda-1}))
## }
## nlminb(rep(0,ncol(X)+1),mlogL)
##' @export
pareto.lvm <- function(lambda=1,...) {   ## shape: lambda, scale: mu
    ## Density f(y): lambda*mu*(1+mu*y)^{-lambda-1}
    ## Survival S(y): (1+mu*y)^{-lambda}
    ## Inverse CDF: u -> ((1-u)^{-1/lambda}-1)/mu
    f <- function(n,mu,var,...) {
        ((1-runif(n))^(-1/lambda)-1)/exp(mu)
    }
    attr(f,"family") <- list(family="pareto",
                             par=c(lambda=lambda))
    return(f)
}

###}}} pareto
###{{{ threshold

##' @export
threshold.lvm <- function(p,labels=NULL,...) {
    if (sum(p)>1 || any(p<0 | p>1)) stop("wrong probability vector") ;
    if (!is.null(labels))
        return(function(n,...) {
            return(cut(rnorm(n),breaks=c(-Inf,qnorm(cumsum(p)),Inf),labels=labels))
        })
    function(n,...)
        cut(rnorm(n),breaks=c(-Inf,qnorm(cumsum(p)),Inf))
}

###}}} threshold

##' @export
multinomial.lvm <- function(prob, labels=NULL) {
  if (sum(prob)<1)
    prob <- c(prob, 1-sum(prob))
  if (is.null(labels)) {
    labels = seq(0, length(prob)-1)
  }
  structure(function(n, ...) sample(labels, n, prob=prob, replace=TRUE),
            family=list(family='multinomial', par=prob))
}

###{{{ binomial

##' @export
binomial.lvm <- function(link="logit",p,size=1) {
    if (substitute(link)==quote(identity)) {
        link <- "identity"
    }
    fam <- stats::binomial(link); fam$link <- link
    f <- function(n,mu,var,...) {
        if (missing(n)) {
            return(fam)
        }
        rbinom(n,size,fam$linkinv(mu))
    }
    attr(f,"family") <- fam
    attr(f,"var") <- FALSE
    if (!missing(p)) attr(f,"mean") <- fam$linkfun(p)
    ## f <- switch(link,
    ##             logit =
    ##             function(n,mu,var,...) rbinom(n,1,tigol(mu)),
    ##             cloglog =
    ##             function(n,mu,var,...) rbinom(n,1,1-exp(-exp(1-mu))),
    ##             function(n,mu,var=1,...) rbinom(n,1,pnorm(mu,sd=sqrt(var)))
    ##             ### function(n,mu=0,var=1,...) (rnorm(n,mu,sqrt(var))>0)*1
    ##             )
    ##}
    return(f)
}

##' @export
logit.lvm <- binomial.lvm("logit")

##' @export
probit.lvm <- binomial.lvm("probit")

###}}} binomial

###{{{ Gamma

##' @export
Gamma.lvm <- function(link="inverse",shape,rate,unit=FALSE,var=FALSE,log=FALSE,...) {
    fam <- stats::Gamma(link); fam$link <- link
    rgam <- if (!log) rgamma else function(...) log(rgamma(...))
    if (!missing(shape) & !missing(rate))
        f <- function(n,mu,var,...) rgam(n,shape=shape,rate=rate)
    if (!missing(shape) & missing(rate)) {
        if (unit)
            f <- function(n,mu,var,...) rgam(n,shape=shape,rate=shape)
        else if (var)
            f <- function(n,mu,var,...) rgam(n,shape=shape,rate=sqrt(shape/var))
        else
            f <- function(n,mu,var,...) rgam(n,shape=shape,rate=shape/fam$linkinv(mu))
    }
    if (missing(shape) & !missing(rate)) {
        if (unit)
            f <- function(n,mu,var,...) rgam(n,shape=shape,rate=rate)
        else if (var)
            f <- function(n,mu,var,...) rgam(n,shape=rate^2*var,rate=rate)
        else
            f <- function(n,mu,var,...) rgam(n,shape=rate*fam$linkinv(mu),rate=rate)
    }
    if (missing(shape) & missing(rate)) {
        if (var)
            f <- function(n,mu,var,...) rgam(n,shape=var,rate=1)
        else
            f <- function(n,mu,var,...) rgam(n,shape=fam$linkinv(mu),rate=1)
    }
    attr(f,"family") <- fam
    attr(f,"var") <- FALSE
    par <- NULL
    if (!missing(shape) & !missing(rate))
      par <- c(shape=shape, rate=rate)
    attr(f, "family") <- list(family="gamma", par=par)
    return(f)
}

##' @export
loggamma.lvm <- function(...) Gamma.lvm(...,log=TRUE)

###}}} Gamma

###{{{ chisq

##' @export
chisq.lvm <- function(df=1,...) {
  f <- function(n,mu,var,...) mu + rchisq(n,df=df)
  attr(f, "family") <- list(family="chisq", par=c(df=df))
  return(f)
}

###}}} chisq

###{{{ student (t-distribution)

##' @export
student.lvm <- function(df=2,mu,sigma,...) {
    f <- function(n,mu,var,...) mu + sqrt(var)*rt(n,df=df)
    if (!missing(mu)) attr(f,"mean") <- mu
    if (!missing(sigma)) attr(f,"variace") <- sigma^2
    attr(f, "family") <- list(family="student-t")
    return(f)
}

###}}} student (t-distribution)

###{{{ uniform

##' @export
uniform.lvm <- function(a,b, value=NULL) {
  if (!is.null(value)) {
    f <- function(n, mu, var, ...)
      sample(value, size=n, replace=TRUE)
    attr(f, "family") <- list(family="discrete uniform")
    return(f)
  }
  if (!missing(a) & !missing(b))
    f <- function(n, mu, var,...) mu+runif(n,a,b)
  else
    f <- function(n, mu, var,...)
      mu+(runif(n,-1,1)*sqrt(12)/2*sqrt(var))
  attr(f, "family") <- list(family="uniform")
  return(f)
}

###}}} uniform

###{{{ weibull
## see also eventTime.R for coxWeibull

##' @export
weibull.lvm <- function(intercept=0, sigma=.5, scale, shape) {
    ## accelerated failure time (AFT) regression
    ## parametrization.
    ##
    ## We parametrize the Weibull distribution (without covariates) as follows:
    ## hazard(t) = shape / scale * (t/scale)^(shape-1)
    ## The hazard is:
    ## - rising if shape > 1
    ## - declining if shape <1
    ## - constant if shape=1
    ##
    ## PH regresion:
    ## hazard(t|x) =  shape / scale * (t/scale)^(shape-1)*exp(x'b)
    ## AFT regression
    ## log(T) = a0 + x'B + s*W
    ## Where W is extreme value distributed and,
    ## a0 = log(scale), B=-sigma*b, sigma = 1/shape
    if (missing(shape)) shape <- 1/sigma
    if (missing(scale)) scale <- exp(intercept)
    f <- function(n,mu=0,var,...) {
        (- log(runif(n)) * exp(log(scale)*shape) * exp(mu*shape))^{1/shape}
        ## scale * (-log(1-runif(n)))^{1/shape}
        ## (- (log(runif(n)) / (1/scale)^(shape) * exp(-mu)))^(1/shape)
    }
    attr(f,"family") <- list(family="weibull",
                             regression="AFT",
                             par=c(shape=shape,scale=scale))
    return(f)
}

###}}} weibull

###{{{ sequence

##' @export
id.lvm <- function(...) Sequence.lvm(integer=TRUE)

##' @export
Sequence.lvm <- function(a=0,b=1,integer=FALSE) {
  if (integer) {
    f <- function(n,...) seq(n)
    attr(f, "family") <- list(family="sequence")
    return(f)
  }
  if (is.null(a) || is.null(b)) {
    if (!is.null(a)) {
      f <- function(n,...) seq(a,length.out=n)
    } else {
      f <- function(n,...) seq(n)-(n-b)
    }
  } else {
    f <- function(n,...) seq(a,b,length.out=n)
  }
  attr(f, "family") <- list(family="sequence")
  return(f)
}

###}}} sequence

##' @export
none.lvm <- function(...) {
  f <- function(n, mu, ...) {
    return(mu)
  }
  attr(f, "family") <- list(family="intervention")
  return(f)
}

##' @export
constant.lvm <- function(value=NA) {
  f <- function(n, mu, ...) {
    if (!is.na(value)) return(rep(value, n))
    return(mu)
  }
  attr(f, "family") <- list(family="constant", par=list(value=value))
  return(f)
}

###{{{ ones

##' @export
ones.lvm <- function(p=1,interval=NULL) {
    f <- function(n,...) {
        if (!is.null(interval)) {
            val <- rep(0L,n)
            if (!is.list(interval)) interval <- list(interval)
            for (i in seq_along(interval)) {
                ii <- interval[[i]]
                lo <- round(ii[1]*n)
                hi <- round(ii[2]*n)
                val[seq(lo,hi)] <- 1L
            }
            return(val)
        }
        if (p==0) return(rep(0L,n))
        val <- rep(1L,n)
        if (p>0 && p<1) val[seq(n*(1-p))] <- 0L
        val
    }
    attr(f, "family") <- list(family="binary (deterministic)")
    return(f)

}

##' @export
Binary.lvm <- ones.lvm

##' @export
binary.lvm <- ones.lvm

###}}} ones

###{{{ beta

##' @export
beta.lvm <- function(alpha=1,beta=1,scale=TRUE) {
    ## CDF: F(x) = B(x,alpha,beta)/B(alpha,beta)
    ## Mean: alpha/(alpha+beta)
    ## Var: alpha*beta/((alpha+beta)^2*(alpha+beta+1))
    if (scale)
        f <- function(n,mu,var,...) {
            m <- alpha/(alpha+beta)
            v <- alpha*beta/((alpha+beta)^2*(alpha+beta+1))
            y <- stats::rbeta(n,shape1=alpha,shape2=beta)
            mu+(y-m)*sqrt(var/v)
        }
    else
      f <- function(n,mu,var,...) stats::rbeta(n,shape1=alpha,shape2=beta)
    attr(f, "family") <- list(family="beta", par=c(alpha=alpha, beta=beta))
    return(f)
}

###}}} beta


##' @export
mvn.lvm <- function(N=2,rho=0.5,sigma=NULL,parname="rho") {
    f <- function(n,rho) {
        if (is.null(sigma)) {
            sigma <- diag(nrow=N)*(1-rho) + rho
        }
        rmvn0(n,sigma=sigma)
    }
    if (!is.null(sigma)) parname <- TRUE
    structure(f,parname=parname,init=rho,mdist=TRUE)
}


###{{{ Gaussian mixture

##' @export
GM2.lvm <- function(...,parname=c("Pr","M1","M2","V1","V2"),init=c(0.5,-4,4,1,1)) {
    f <- function(n,pr,m1,m2,v1,v2) {
        y1 <- rnorm(n,m1,v1^0.5)
        if (pr>=1) return(y1)
        z <- rbinom(n,1,pr)
        y2 <- rnorm(n,m2,v2^0.5)
        return(z*y1+(1-z)*y2)
    }
    structure(f,parname=parname,init=init)
}

##' @export
GM3.lvm <- function(...,parname=c("Pr1","Pr2","M1","M2","M3","V1","V2","V3"),init=c(0.25,0.5,-4,0,4,1,1,1)) {
    f <- function(n,pr1,pr2,m1,m2,m3,v1,v2,v3) {
        p <- c(pr1,pr2,1-pr1-pr2)
        y1 <- rnorm(n,m1,v1^0.5)
        y2 <- rnorm(n,m2,v2^0.5)
        y3 <- rnorm(n,m3,v3^0.5)
        z <- stats::rmultinom(n,1,p)
        rowSums(cbind(y1,y2,y3)*t(z))
    }
    structure(f,parname=parname,init=init)
}

###}}} Gaussian mixture

context("Influence functions")

test_that("GEE", {
  if (requireNamespace("geepack",quietly=TRUE)) {
  d <- lvm(y ~ x, ~ id) |>
    distribution(~id, uniform.lvm(value=seq(1:20))) |>
    sim(100, seed=1)
  d0 <- d[order(d$id), ]
  g <- geepack::geeglm(y ~ x, data=d0, id=d0$id)
  V <- summary(g)$cov.scaled
  g0 <- glm(y ~ x, data=d)
  V0 <- vcov(estimate(g0, id = d$id))
  testthat::expect_true(sum((V - V0)^2) < 1e-12)
  }
})

test_that("merge, IC, estimate with 'id' argument", {
  d <- data.frame(id=c("a","a","b","b","b","b","c","c","d"),
                id1=c("a","a","b1","b1","b2","b2","c","c","d"),
                y=rnorm(9), x=rnorm(9))
  d$id0 <- as.numeric(as.factor(d$id))

  l <- glm(y ~ x, data=d)
  e1 <- estimate(l, id=d$id1)
  V1 <- vcov(e1)
  e0 <- estimate(e1, id=c(1,2,2,3,4))
  V0 <- vcov(e0)

  if (requireNamespace("geepack",quietly=TRUE)) {
    V <- summary(geepack::geeglm(y ~ x, id=d$id0, data=d))$cov.scaled
    testthat::expect_true(sum((V - V0)^2) < 1e-12)
  }

  e <- merge(estimate(l), estimate(l), id=list(d$id, d$id1))
  testthat::expect_true(sum((V1 - vcov(e)[3:4,3:4])^2) < 1e-12)
  testthat::expect_true(sum((V0 - vcov(e)[1:2,1:2])^2) < 1e-12)

  ee <- estimate(e, id=c(1,2,3,4,2,2))
  VV <- vcov(ee)
  testthat::expect_true(sum((VV[1:2,1:2] - V0)^2) < 1e-12)
  testthat::expect_true(sum((VV[3:4,3:4] - V0)^2) < 1e-12)
})

test_that("negative binomial regression (glm.nb)", {
  set.seed(1)
  n <- 500
  z <- rgamma(n, .5, .5)
  x <- rnorm(n)
  lam <- z * exp(x)
  y <- rpois(n, lam)
  if (requireNamespace("MASS",quietly=TRUE)) {
  m <- MASS::glm.nb(y ~ x)
  testthat::expect_true(abs(lava:::logL.glm(m) - logLik(m)) < 1e-6)
  p <- coef(m)+1
  u1 <- as.vector(numDeriv::jacobian(function(p) lava:::logL.glm(m, p = p), p))
  u2 <- score(m, p = p)
  testthat::expect_true(sum((u1 - u2)^2) < 1e-6)
  p <- coef(m)
  u1 <- as.vector(numDeriv::jacobian(function(p) lava:::logL.glm(m, p = p), p))
  u2 <- score(m, p = p)
  testthat::expect_true(sum((u1 - u2)^2) < 1e-6)
  }
})

test_that("quasipossion", {
  set.seed(1)
  n <- 500
  z <- rgamma(n, .5, .5)
  x <- rnorm(n)
  lam <- z * exp(x)
  y <- rpois(n, lam)
  m1 <- glm(y ~ x, family=poisson)
  m2 <- glm(y ~ x, family = quasipoisson)
  i1 <- IC(m1)
  i2 <- IC(m2)
  testthat::expect_true(sum((i1 - i2)^2) < 1e-6)
})

test_that("merge back.transform", {
  tmp <- estimate(coef = 1, IC = c(1,2,3), id = 1:3, f = exp, back.transform = log)
  tmp2 <- estimate(coef = 2, IC = c(3,2,1), id = 1:3, f = exp, back.transform = log)

  expect_warning(
    m <- merge(tmp, tmp2) # should throw a warning
  )
  expect_equivalent(coef(m), exp(c(1, 2)))
})

test_that("IC.numeric", {
  x <- rnorm(100)
  res <- IC(x)
  expect_true(ncol(res) == 2L)
  ic1 <- IC(lm(x ~ 1))
  expect_true(mean((ic1-res[,1])^2) < 1e-9)
  expect_equal(mean(x), unname(attr(res, "coef")["mean"]))
  expect_true(var(x)*(length(x)-1)/length(x) == attr(res, "coef")["var"])

  res <- IC(cbind(x1=x, x2=x))
  expect_true(NCOL(res) == 5) # 2 mean, 2 var, 1 cov
  expect_equal(mean(x), unname(attr(res, "coef")["x1"]))
  expect_equal(mean(x), unname(attr(res, "coef")["x2"]))
  expect_equal(var(x)*(length(x)-1)/length(x), unname(attr(res, "coef")[3]))
  expect_equal(var(x)*(length(x)-1)/length(x), unname(attr(res, "coef")[5]))
})

y <- rnorm(100)
x <- rnorm(100)
d <- data.frame(y=y,x=x)
e <- estimate(l <- lm(y ~ x, data=d))

test_that("influence", {
  expect_equivalent(influence(e), IC(e))
})

test_that("IC misc", {
  ic0 <- IC(e)
  id <- rep(1:10, length.out=nrow(d))
  opt <- lava.options(cluster.index=FALSE)
  ic1 <- IC(l, id=id)
  lava.options(cluster.index=TRUE)
  ic2 <- IC(l, id=id)
  expect_equivalent(ic1, ic2)

  ic <- IC(l, folds=10)
  ## lava.options(opt)
})

context("Influence functions")

test_that("GEE", {
  if (requireNamespace("geepack",quietly=TRUE)) {
  d <- lvm(y ~ x, ~ id) |>
    distribution(~id, uniform.lvm(value=seq(1:20))) |>
    sim(100, seed=1)
  d0 <- d[order(d$id), ]
  g <- geepack::geeglm(y ~ x, data=d0, id=d0$id)
  V <- summary(g)$cov.scaled
  g0 <- glm(y ~ x, data=d)
  V0 <- vcov(estimate(g0, id = d$id))
  testthat::expect_true(sum((V - V0)^2) < 1e-12)
  }
})

test_that("merge, IC, estimate with 'id' argument", {
  d <- data.frame(id=c("a","a","b","b","b","b","c","c","d"),
                id1=c("a","a","b1","b1","b2","b2","c","c","d"),
                y=rnorm(9), x=rnorm(9))
  d$id0 <- as.numeric(as.factor(d$id))

  l <- glm(y ~ x, data=d)
  e1 <- estimate(l, id=d$id1)
  V1 <- vcov(e1)
  e0 <- estimate(e1, id=c(1,2,2,3,4))
  V0 <- vcov(e0)

  if (requireNamespace("geepack",quietly=TRUE)) {
    V <- summary(geepack::geeglm(y ~ x, id=d$id0, data=d))$cov.scaled
    testthat::expect_true(sum((V - V0)^2) < 1e-12)
  }

  e <- merge(estimate(l), estimate(l), id=list(d$id, d$id1))
  testthat::expect_true(sum((V1 - vcov(e)[3:4,3:4])^2) < 1e-12)
  testthat::expect_true(sum((V0 - vcov(e)[1:2,1:2])^2) < 1e-12)

  ee <- estimate(e, id=c(1,2,3,4,2,2))
  VV <- vcov(ee)
  testthat::expect_true(sum((VV[1:2,1:2] - V0)^2) < 1e-12)
  testthat::expect_true(sum((VV[3:4,3:4] - V0)^2) < 1e-12)
})

test_that("negative binomial regression (glm.nb)", {
  set.seed(1)
  n <- 500
  z <- rgamma(n, .5, .5)
  x <- rnorm(n)
  lam <- z * exp(x)
  y <- rpois(n, lam)
  if (requireNamespace("MASS",quietly=TRUE)) {
  m <- MASS::glm.nb(y ~ x)
  testthat::expect_true(abs(lava:::logL.glm(m) - logLik(m)) < 1e-6)
  p <- coef(m)+1
  u1 <- as.vector(numDeriv::jacobian(function(p) lava:::logL.glm(m, p = p), p))
  u2 <- score(m, p = p)
  testthat::expect_true(sum((u1 - u2)^2) < 1e-6)
  p <- coef(m)
  u1 <- as.vector(numDeriv::jacobian(function(p) lava:::logL.glm(m, p = p), p))
  u2 <- score(m, p = p)
  testthat::expect_true(sum((u1 - u2)^2) < 1e-6)
  }
})

set.seed(1)
n <- 500
z <- rgamma(n, .5, .5)
x <- rnorm(n)
lam <- z * exp(x)
y <- rpois(n, lam)
d <- data.frame(y,x,z)

test_that("quasipossion", {
  m1 <- glm(y ~ x, family=poisson, data=d)
  m2 <- glm(y ~ x, family = quasipoisson, data=d)
  i1 <- IC(m1)
  i2 <- IC(m2)
  testthat::expect_true(sum((i1 - i2)^2) < 1e-6)

  # check that iid decomposition is correctly scaled
  ii <- iid(m1)
  expect_equivalent(crossprod(ii), var_ic(i1))
})

test_that("merge back.transform", {
  suppressWarnings(
    tmp <- estimate(coef = 1, IC = c(1,2,3), id = 1:3, f = exp, back.transform = log)
  )
  suppressWarnings(
    tmp2 <- estimate(coef = 2, IC = c(3,2,1), id = 1:3, f = exp, back.transform = log)
  )

  expect_warning(
    m <- merge(tmp, tmp2) # should throw a warning
  )
  expect_equivalent(coef(m), exp(c(1, 2)))
})

test_that("IC.numeric", {
  x <- rnorm(100)
  res <- IC(x)
  expect_true(ncol(res) == 2L)
  ic1 <- IC(lm(x ~ 1))
  expect_true(mean((ic1-res[,1])^2) < 1e-9)
  expect_equal(mean(x), unname(attr(res, "coef")["mean"]))
  expect_true(var(x)*(length(x)-1)/length(x) == attr(res, "coef")["var"])

  res <- IC(cbind(x1=x, x2=x))
  expect_true(NCOL(res) == 5) # 2 mean, 2 var, 1 cov
  expect_equal(mean(x), unname(attr(res, "coef")["x1"]))
  expect_equal(mean(x), unname(attr(res, "coef")["x2"]))
  expect_equal(var(x)*(length(x)-1)/length(x), unname(attr(res, "coef")[3]))
  expect_equal(var(x)*(length(x)-1)/length(x), unname(attr(res, "coef")[5]))
})

y <- rnorm(100)
x <- rnorm(100)
d <- data.frame(y=y,x=x)
e <- estimate(l <- lm(y ~ x, data=d))

test_that("influence", {
  expect_equivalent(influence(e), IC(e))
})

test_that("IC misc", {
  ic0 <- IC(e)
  id <- rep(1:10, length.out=nrow(d))
  opt <- lava.options(cluster.index=FALSE)
  ic1 <- IC(l, id=id)
  lava.options(cluster.index=TRUE)
  ic2 <- IC(l, id=id)
  expect_equivalent(ic1, ic2)

  ic <- IC(l, folds=10)
  ## lava.options(opt)
})

context("Simulation")

test_that("Constrain, transform I", {
    m <- lvm(,~y+x)
    distribution(m,~x) <- Sequence.lvm()
    transform(m,y~x) <- function(x) x
    with(sim(m,10),testthat::expect_equivalent(y,x))

    m <- lvm(y~1,~x)
    distribution(m,~x) <- Sequence.lvm()
    intercept(m,~y) <- "ym"
    covariance(m,~y) <- 0.001
    constrain(m,ym~x) <- function(x) x
    d <- simulate(m,200)
    testthat::expect_true(mean((d$y-d$x)^2)<0.1)


})

test_that("Missing", {
    m <- lvm(y~1)
    m <- Missing(m,y~1,r~x)
    set.seed(1)
    d <- simulate(m,1e3,seed=1)
    testthat::expect_equal(sum(d$r),sum(!is.na(d$y0)))

    g <- glm(r~x,data=d,family=binomial)
    testthat::expect_true(all.equal(coef(g), c(0, 1),
                                    tolerance = 0.2,
                                    check.attributes = FALSE
    ))
})

test_that("sim.default I", {
    m <- lvm(y~x+e)
    distribution(m,~y) <- 0
    distribution(m,~x) <- uniform.lvm(a=-1.1,b=1.1)
    transform(m,e~x) <- function(x) (1*x^4)*rnorm(length(x),sd=1)

    onerun <- function(iter=NULL,...,n=2e3,b0=1,idx=2) {
        d <- sim(m,n,p=c("y~x"=b0))
        l <- lm(y~x,d)
        res <- c(coef(summary(l))[idx,1:2],
                 confint(l)[idx,],
                 estimate(l,only.coef=TRUE)[idx,2:4])
        names(res) <- c("Estimate","Model.se","Model.lo","Model.hi",
                        "Sandwich.se","Sandwich.lo","Sandwich.hi")
        res
    }

    val <- sim(onerun,R=2,b0=1,n=10,messages=0)
    testthat::expect_true(nrow(val)==2)
    val <- sim(val,R=2,b0=1,n=10,type=0) ## append results
    testthat::expect_true(nrow(val)==4)

    s1 <- summary(val,
                  estimate = c(1, 1), confint = c(3, 4, 6, 7),
                  true = c(1, 1), names = c("Model", "Sandwich")
    )
    testthat::expect_true(length(grep("Coverage",rownames(s1)))>0)
    testthat::expect_equivalent(colnames(s1),c("Model","Sandwich"))
    
    val <- sim(onerun,R=2,cl=TRUE,seed=1,messages=0,mc.cores=1)
    testthat::expect_true(val[1,1]!=val[1,2])
        
    onerun2 <- function(a,b,...) {
        return(cbind(a=a,b=b,c=a-1,d=a+1))
    }
    R <- data.frame(a=1:2,b=3:4)
    dm <- capture.output(val2 <- sim(onerun2,R=R,messages=1,mc.cores=1))
    testthat::expect_true(all(R-val2[,1:2]==0))
    res <- summary(val2)
    testthat::expect_equivalent(res["Mean",],c(1.5,3.5,0.5,2.5))

    testthat::expect_output(print(val2[1,]),"a b c d")
    testthat::expect_output(print(val2[1,]),"1 3 0 2")
       
    res <- summary(val2,estimate="a",se="b",true=1.5,confint=c("c","d"))
    testthat::expect_true(res["Coverage",]==1)
    testthat::expect_true(res["SE/SD",]==mean(val2[,"b"])/sd(val2[,"a"]))
      
})

test_that("distributions", {
    m <- lvm(y1~x)
    distribution(m,~y1) <- binomial.lvm("probit")
    distribution(m,~y2) <- poisson.lvm()
    distribution(m,~y3) <- normal.lvm(mean=1,sd=2)
    distribution(m,~y3) <- lognormal.lvm()
    distribution(m,~y3) <- pareto.lvm()
    distribution(m,~y3) <- loggamma.lvm()
    distribution(m,~y3) <- weibull.lvm()
    distribution(m,~y3) <- chisq.lvm()
    distribution(m,~y3) <- student.lvm(mu=1,sigma=1)    

    testthat::expect_output(print(distribution(m)$y2),"Family: poisson")
    testthat::expect_output(print(distribution(m)$y1),"Family: binomial")
    latent(m) <- ~u    
    testthat::expect_output(print(m),"binomial\\(probit\\)")
    testthat::expect_output(print(m),"poisson\\(log\\)")

    ## Generator:
    m <- lvm()
    distribution(m,~y,TRUE) <- function(n,...) {
        res <- exp(rnorm(n)); res[seq(min(n,5))] <- 0
        return(res)
    }
    d <- sim(m,10)
    testthat::expect_true(all(d[1:5,1]==0))
    testthat::expect_true(all(d[6:10,1]!=0))

    m <- lvm()
    distribution(m,~y,parname="a",init=2) <- function(n,a,...) {
        rep(1,n)*a
    }
    testthat::expect_true(all(sim(m,2)==2))
    testthat::expect_true(all(sim(m,2,p=c(a=10))==10))
    testthat::expect_equivalent(sim(m,2,p=c(a=10)),sim(m,2,a=10))

    ## Multivariate distribution
    m <- lvm()
    rmr <- function(n,rho,...) rmvn0(n,sigma=diag(2)*(1-rho)+rho)
    distribution(m,~y1+y2,rho=0.9) <- rmr
    testthat::expect_equivalent(c("y1","y2"),colnames(d <- sim(m,5)))

    ## Special 'distributions'
    m <- lvm()
    distribution(m,~x1) <- Sequence.lvm(int=TRUE)
    distribution(m,~x2) <- Sequence.lvm(a=1,b=2)
    distribution(m,~x3) <- Sequence.lvm(a=NULL,b=2)
    distribution(m,~x4) <- Sequence.lvm(a=2,b=NULL)
    ex <- sim(m,5)
    testthat::expect_equivalent(ex$x1,1:5)
    testthat::expect_equivalent(ex$x2,seq(1,2,length.out=5))
    testthat::expect_equivalent(ex$x3,seq(-2,2))
    testthat::expect_equivalent(ex$x4,seq(2,6))

    m <- lvm()
    distribution(m,~x1) <- Binary.lvm()
    distribution(m,~x2) <- Binary.lvm(p=0.5)
    distribution(m,~x3) <- Binary.lvm(interval=c(0.4,0.6))
    ex <- sim(m,10)
    testthat::expect_equivalent(ex$x1,rep(1,10))
    testthat::expect_equivalent(ex$x2,c(rep(0,5),rep(1,5)))
    testthat::expect_equivalent(ex$x3,c(0,0,0,1,1,1,0,0,0,0))

    m <- lvm()
    testthat::expect_error(distribution(m,~y) <- threshold.lvm(p=c(0.5,.75)))
    distribution(m,~y) <- threshold.lvm(p=c(0.25,.25))
    set.seed(1)
    testthat::expect_equivalent(1:3,sort(unique(sim(m,200))[,1]))

    ## distribution(m,~y) <- threshold.lvm(p=c(0.25,.25),labels=letters[1:3])
    ## testthat::expect_equivalent(c("a","b","c"),sort(unique(sim(m,200))[,1]))
        
})

test_that("eventTime", {
    m <- lvm(eventtime~x)
    distribution(m,~eventtime) <- coxExponential.lvm(1/100)
    distribution(m,~censtime) <- coxWeibull.lvm(1/500)
    eventTime(m) <- time~min(eventtime=1,censtime=0)

    set.seed(1)
    d <- sim(m,100)
    testthat::expect_equivalent((d$time<d$cens)*1L,d$status)

    ## TODO: plot(m)
    testthat::expect_output(print(m),"Event History Model")

    ## Time varying effect
    m <- lvm(y~1)
    distribution(m,~z1) <- Binary.lvm(0.5)
    R <- log(cbind(c(0.2,0.7,0.9),c(0.5,0.3,0.3)))
    m <- timedep(m,y~z1,timecut=c(0,3,5),rate=R)
   
    ## sim(m,100)
    ## d <- sim(m,1e4); d$status <- TRUE
    ## dd <- mets::lifetable(survival::Surv(y,status)~z1,data=d,breaks=c(0,3,5,Inf));
    ## exp(coef(glm(events ~ offset(log(atrisk)) + -1 + interval+z1:interval, dd, family=poisson)))

})

context("Simulation")

test_that("Constrain, transform I", {
    m <- lvm(,~y+x)
    distribution(m,~x) <- Sequence.lvm()
    transform(m,y~x) <- function(x) x
    with(sim(m,10),testthat::expect_equivalent(y,x))

    m <- lvm(y~1,~x)
    distribution(m,~x) <- Sequence.lvm()
    intercept(m,~y) <- "ym"
    covariance(m,~y) <- 0.001
    constrain(m,ym~x) <- function(x) x
    d <- simulate(m,200)
    testthat::expect_true(mean((d$y-d$x)^2)<0.1)


})

test_that("Missing", {
    m <- lvm(y~1)
    m <- Missing(m,y~1,r~x)
    set.seed(1)
    d <- simulate(m,1e3,seed=1)
    testthat::expect_equal(sum(d$r),sum(!is.na(d$y0)))

    g <- glm(r~x,data=d,family=binomial)
    testthat::expect_true(all.equal(coef(g), c(0, 1),
                                    tolerance = 0.2,
                                    check.attributes = FALSE
    ))
})

test_that("sim.default I", {
    m <- lvm(y~x+e)
    distribution(m,~y) <- 0
    distribution(m,~x) <- uniform.lvm(a=-1.1,b=1.1)
    transform(m,e~x) <- function(x) (1*x^4)*rnorm(length(x),sd=1)

    onerun <- function(iter=NULL,...,n=2e3,b0=1,idx=2) {
        d <- sim(m,n,p=c("y~x"=b0))
        l <- lm(y~x,d)
        res <- c(coef(summary(l))[idx,1:2],
                 confint(l)[idx,],
                 coef(estimate(l), mat=TRUE)[idx,2:4])
        names(res) <- c("Estimate","Model.se","Model.lo","Model.hi",
                        "Sandwich.se","Sandwich.lo","Sandwich.hi")
        res
    }

    val <- sim(onerun,R=2,b0=1,n=10,messages=0)
    testthat::expect_true(nrow(val)==2)
    val <- sim(val,R=2,b0=1,n=10,type=0) ## append results
    testthat::expect_true(nrow(val)==4)

    s1 <- summary(val,
                  estimate = c(1, 1), confint = c(3, 4, 6, 7),
                  true = c(1, 1), names = c("Model", "Sandwich")
    )
    testthat::expect_true(length(grep("Coverage",rownames(s1)))>0)
    testthat::expect_equivalent(colnames(s1),c("Model","Sandwich"))
    
    val <- sim(onerun,R=2,cl=TRUE,seed=1,messages=0,mc.cores=1)
    testthat::expect_true(val[1,1]!=val[1,2])
        
    onerun2 <- function(a,b,...) {
        return(cbind(a=a,b=b,c=a-1,d=a+1))
    }
    R <- data.frame(a=1:2,b=3:4)
    dm <- capture.output(val2 <- sim(onerun2,R=R,messages=1,mc.cores=1))
    testthat::expect_true(all(R-val2[,1:2]==0))
    res <- summary(val2)
    testthat::expect_equivalent(res["Mean",],c(1.5,3.5,0.5,2.5))

    testthat::expect_output(print(val2[1,]),"a b c d")
    testthat::expect_output(print(val2[1,]),"1 3 0 2")
       
    res <- summary(val2,estimate="a",se="b",true=1.5,confint=c("c","d"))
    testthat::expect_true(res["Coverage",]==1)
    testthat::expect_true(res["SE/SD",]==mean(val2[,"b"])/sd(val2[,"a"]))
      
})

test_that("distributions", {
    m <- lvm(y1~x)
    distribution(m,~y1) <- binomial.lvm("probit")
    distribution(m,~y2) <- poisson.lvm()
    distribution(m,~y3) <- normal.lvm(mean=1,sd=2)
    distribution(m,~y3) <- lognormal.lvm()
    distribution(m,~y3) <- pareto.lvm()
    distribution(m,~y3) <- loggamma.lvm()
    distribution(m,~y3) <- weibull.lvm()
    distribution(m,~y3) <- chisq.lvm()
    distribution(m,~y3) <- student.lvm(mu=1,sigma=1)    

    testthat::expect_output(print(distribution(m)$y2),"Family: poisson")
    testthat::expect_output(print(distribution(m)$y1),"Family: binomial")
    latent(m) <- ~u    
    testthat::expect_output(print(m),"binomial\\(probit\\)")
    testthat::expect_output(print(m),"poisson\\(log\\)")

    ## Generator:
    m <- lvm()
    distribution(m,~y,TRUE) <- function(n,...) {
        res <- exp(rnorm(n)); res[seq(min(n,5))] <- 0
        return(res)
    }
    d <- sim(m,10)
    testthat::expect_true(all(d[1:5,1]==0))
    testthat::expect_true(all(d[6:10,1]!=0))

    m <- lvm()
    distribution(m,~y,parname="a",init=2) <- function(n,a,...) {
        rep(1,n)*a
    }
    testthat::expect_true(all(sim(m,2)==2))
    testthat::expect_true(all(sim(m,2,p=c(a=10))==10))
    testthat::expect_equivalent(sim(m,2,p=c(a=10)),sim(m,2,a=10))

    ## Multivariate distribution
    m <- lvm()
    rmr <- function(n,rho,...) rmvn0(n,sigma=diag(2)*(1-rho)+rho)
    distribution(m,~y1+y2,rho=0.9) <- rmr
    testthat::expect_equivalent(c("y1","y2"),colnames(d <- sim(m,5)))

    ## Special 'distributions'
    m <- lvm()
    distribution(m,~x1) <- Sequence.lvm(int=TRUE)
    distribution(m,~x2) <- Sequence.lvm(a=1,b=2)
    distribution(m,~x3) <- Sequence.lvm(a=NULL,b=2)
    distribution(m,~x4) <- Sequence.lvm(a=2,b=NULL)
    ex <- sim(m,5)
    testthat::expect_equivalent(ex$x1,1:5)
    testthat::expect_equivalent(ex$x2,seq(1,2,length.out=5))
    testthat::expect_equivalent(ex$x3,seq(-2,2))
    testthat::expect_equivalent(ex$x4,seq(2,6))

    m <- lvm()
    distribution(m,~x1) <- Binary.lvm()
    distribution(m,~x2) <- Binary.lvm(p=0.5)
    distribution(m,~x3) <- Binary.lvm(interval=c(0.4,0.6))
    ex <- sim(m,10)
    testthat::expect_equivalent(ex$x1,rep(1,10))
    testthat::expect_equivalent(ex$x2,c(rep(0,5),rep(1,5)))
    testthat::expect_equivalent(ex$x3,c(0,0,0,1,1,1,0,0,0,0))

    m <- lvm()
    testthat::expect_error(distribution(m,~y) <- threshold.lvm(p=c(0.5,.75)))
    distribution(m,~y) <- threshold.lvm(p=c(0.25,.25))
    set.seed(1)
    testthat::expect_equivalent(1:3,sort(unique(sim(m,200))[,1]))

    ## distribution(m,~y) <- threshold.lvm(p=c(0.25,.25),labels=letters[1:3])
    ## testthat::expect_equivalent(c("a","b","c"),sort(unique(sim(m,200))[,1]))
        
})

test_that("eventTime", {
    m <- lvm(eventtime~x)
    distribution(m,~eventtime) <- coxExponential.lvm(1/100)
    distribution(m,~censtime) <- coxWeibull.lvm(1/500)
    eventTime(m) <- time~min(eventtime=1,censtime=0)

    set.seed(1)
    d <- sim(m,100)
    testthat::expect_equivalent((d$time<d$cens)*1L,d$status)

    ## TODO: plot(m)
    testthat::expect_output(print(m),"Event History Model")

    ## Time varying effect
    m <- lvm(y~1)
    distribution(m,~z1) <- Binary.lvm(0.5)
    R <- log(cbind(c(0.2,0.7,0.9),c(0.5,0.3,0.3)))
    m <- timedep(m,y~z1,timecut=c(0,3,5),rate=R)
   
    ## sim(m,100)
    ## d <- sim(m,1e4); d$status <- TRUE
    ## dd <- mets::lifetable(survival::Surv(y,status)~z1,data=d,breaks=c(0,3,5,Inf));
    ## exp(coef(glm(events ~ offset(log(atrisk)) + -1 + interval+z1:interval, dd, family=poisson)))

})

library(testthat)

context("mixture.lvm")

setup_mixture_data <- function(seed = 42, n = 500) {
  set.seed(seed)
  m0 <- lvm(list(y ~ x + z, x ~ z))
  distribution(m0, ~z) <- binomial.lvm()
  d <- sim(m0, n, p = c("y~z" = 2, "y~x" = 1))
  list(data = d, model = lvm(y ~ x, x ~ 1))
}

s <- setup_mixture_data()
m <- baptize(s$model)
intercept(m, ~x + y) <- NA
result <- mixture(m, k = 2, data = s$data,
                    control = list(tol = 1e-4, trace = 0))

test_that("mixture: result contains expected components", {
  expect_s3_class(result, "lvm.mixture")
  expect_true(!is.null(result$theta))
  expect_true(!is.null(result$prob))
  expect_true(!is.null(result$gamma))
  expect_true(!is.null(result$member))
  expect_true(!is.null(result$vcov))
  expect_true(!is.null(result$k))
  ## mixture: correct number of mixture components (k=2)
  expect_equal(result$k, 2)
})

test_that("mixture: mixing probabilities sum to 1", {
  prob <- tail(result$prob, 1)  # last row = final estimates
  expect_equal(sum(prob), 1, tolerance = 1e-6)
  expect_true(all(prob >= 0))
  expect_true(all(prob <= 1))
})

test_that("mixture: posterior probabilities (gamma)", {
  #  sum to 1 per observation
  row_sums <- rowSums(result$gamma)
  expect_equal(row_sums, rep(1, nrow(s$data)), tolerance = 1e-6)
  ## gamma has correct dimensions
  expect_equal(dim(result$gamma), c(nrow(s$data), 2))
})

test_that("coef.mixture", {
  cc <- coef(result)
  expect_true(is.numeric(cc))
  expect_true(!is.null(names(cc)))
  ## coef(list=TRUE) returns a list of length k
  cc_list <- coef(result, list = TRUE)
  expect_true(is.list(cc_list))
  expect_equal(length(cc_list), 2)
})

test_that("logLik.mixture", {
  ll <- logLik(result)
  expect_true(is.finite(ll))
  expect_equal(class(ll), "logLik")
  expect_equal(attr(ll, "nobs"), nrow(s$data))
})

test_that("vcov.mixture", {
  V <- vcov(result)
  expect_true(is.matrix(V))
  expect_equal(nrow(V), ncol(V))
  np <- length(coef(result))
  expect_equal(dim(vcov(result)), c(np, np))
})

test_that("mixture: score() at MLE is approximately zero", {
  sc <- score(result)
  expect_true(sum(sc^2) < 1)  # score norm should be small at MLE
  sc_indiv <- score(result, indiv = TRUE)
  np <- length(coef(result))
  expect_equal(dim(sc_indiv), c(nrow(s$data), np))
})

test_that("mixture: summary() returns object of class 'summary.lvm.mixture'", {
  s_result <- summary(result)
  expect_s3_class(s_result, "summary.lvm.mixture")
})

test_that("mixture: summary() AIC is finite", {
  s_result <- summary(result)
  expect_true(is.finite(s_result$AIC))
})

##' Generic method for labeling elements of an object
##'
##' @title Label elements of object
##' @param x Object
##' @param \dots Additional arguments
##' @author Klaus K. Holst
##' @export
`baptize` <- function(x,...) UseMethod("baptize")


##' @export
baptize.lvm <- function(x,labels,overwrite=FALSE,unique=FALSE,...) {
  p <- describecoef(x, mean=TRUE)
  sym <- lava.options()$symbols
  MeanFix <- intfix(x)
  RegFix <- regfix(x)
  CovFix <- covfix(x)
  count <- 0
  curlab <- parlabels(x)
  coef(x)
  for (i in seq_along(p)) {
    p0 <- p[[i]]
    if (attributes(p0)$type=="reg") {
      curfix <- RegFix$values[p0[2],p0[1]]
      curlab <- RegFix$labels[p0[2],p0[1]]
      if (all(is.na(c(curfix,curlab))) | overwrite) {
        count <- count+1
        st <- ifelse(missing(labels),paste(p0[1],p0[2],sep=sym[1]),labels[count])
        regfix(x,from=p0[2],to=p0[1]) <- st
      }
    } else if (attributes(p0)$type=="cov") {
      curfix <- CovFix$values[p0[2],p0[1]]
      curlab <- CovFix$labels[p0[2],p0[1]]
      if (all(is.na(c(curfix,curlab))) | overwrite) {
        count <- count+1
        st <- ifelse(missing(labels),paste(p0[1],p0[2],sep=sym[2]),labels[count])
        covfix(x,p0[2],p0[1],exo=FALSE) <- st
      }
    } else { ## Mean parameter
      curfix <- MeanFix[[p0]]
      if (length(curfix)>0)
      if (is.na(curfix) | overwrite) {
        count <- count+1
        st <- ifelse(missing(labels),p0,labels[count])
        intfix(x,p0) <- st
      }
    }
  }
  if (index(x)$npar.ex>0) {
    x$exfix[is.na(x$exfix)] <- names(x$exfix)[is.na(x$exfix)]
    index(x) <- reindex(x)
  }
  return(x)
}

#' @export
"sim" <- function(x,...) UseMethod("sim")

#' Monte Carlo simulation
#'
#' Applies a function repeatedly for a specified number of replications or over
#' a list/data.frame with plot and summary methods for summarizing the Monte
#' Carlo experiment. Can be parallelized via the future package (use the
#' [future::plan()] function).
#' @export
#' @param x function or 'sim' object
#' @param R Number of replications or data.frame with parameters
#' @param f Optional function (i.e., if x is a matrix)
#' @param colnames Optional column names
#' @param seed (optional) Seed (needed with cl=TRUE)
#' @param args (optional) list of named arguments passed to (mc)mapply
#' @param iter If TRUE the iteration number is passed as first argument to
#'   (mc)mapply
#' @param mc.cores Optional number of cores. Will use parallel::mcmapply
#'   instead of future
#' @param progressr.message Optional message for the progressr progress-bar
#' @param estimate.index If return object inherits from `estimate` then only
#'   these column indices are extracted (estimate, se, lower, upper, p-val)
#' @param ... Additional arguments to [future.apply::future_mapply()]
#' @aliases sim sim.default as.sim
#' @seealso [summary.sim()] [plot.sim()] [sim.lvm()]
#' @details To parallelize the calculation use the [future::plan()] function
#'   (e.g., `future::plan(multisession())` to distribute the calculations over
#'   the `R` replications on all available cores). The output is controlled via
#'   the progressr package (e.g., `progressr::handlers(global=TRUE)` to enable
#'   progress information).
#' @examples
#' m <- lvm(y~x+e)
#' distribution(m,~y) <- 0
#' distribution(m,~x) <- dist_uniform(a=-1.1,b=1.1)
#' transform(m,e~x) <- function(x) (1*x^4)*rnorm(length(x),sd=1)
#'
#' onerun <- function(iter=NULL,...,n=2e3,b0=1,idx=2) {
#'     d <- sim(m,n,p=c("y~x"=b0))
#'     l <- lm(y~x,d)
#'     res <- c(coef(summary(l))[idx,1:2],
#'              confint(l)[idx,],
#'              coef(estimate(l), mat=TRUE)[idx,2:4])
#'     names(res) <- c("Estimate","Model.se","Model.lo","Model.hi",
#'                     "Sandwich.se","Sandwich.lo","Sandwich.hi")
#'     res
#' }
#' val <- sim(onerun,R=10,b0=1)
#' val
#'
#' val <- sim(val,R=40,b0=1) ## append results
#' summary(val,estimate=c(1,1),confint=c(3,4,6,7),true=c(1,1))
#'
#' summary(val,estimate=c(1,1),se=c(2,5),names=c("Model","Sandwich"))
#' summary(val,estimate=c(1,1),se=c(2,5),true=c(1,1),
#'         names=c("Model","Sandwich"),confint=TRUE)
#'
#' if (interactive()) {
#'     plot(val,estimate=1,c(2,5),true=1,
#'          names=c("Model","Sandwich"),polygon=FALSE)
#'     plot(val,estimate=c(1,1),se=c(2,5),main=NULL,
#'          true=c(1,1),names=c("Model","Sandwich"),
#'          line.lwd=1,col=c("gray20","gray60"),
#'          rug=FALSE)
#'     plot(val,estimate=c(1,1),se=c(2,5),true=c(1,1),
#'          names=c("Model","Sandwich"))
#' }
#'
#' f <- function(a=1, b=1) {
#'   rep(a*b, 5)
#' }
#' R <- Expand(a=1:3, b=1:3)
#' sim(f, R)
#' sim(function(a,b) f(a,b), 3, args=c(a=5,b=5))
#' sim(function(iter=1,a=5,b=5) iter*f(a,b), iter=TRUE, R=5)
#'
#' ## Returning estimate objects with extra per-iteration information
#' onerun2 <- function(...) {
#'   x <- rnorm(50)
#'   y <- 0.5*x + rnorm(50)
#'   e <- estimate(lm(y ~ x))
#'   c(e, converged = 1, niter = sample(5:20, 1), drop.ic = TRUE)
#'   # works identically with summary.estimate objects
#'   # c(summary(e), converged = 1, niter = sample(5:20, 1), drop.ic = TRUE)
#' }
#' val2 <- sim(onerun2, R = 10)
#' val2
#' # extra columns are stored but excluded from default summary statistics
#' idx <- c("(Intercept)", "x", "niter")
#' summary(val2, estimate=idx, true=c(0,0.5,NA))
sim.default <- function(x = NULL, R = 100, f = NULL,
                        colnames = NULL,
                        seed = NULL, args = list(),
                        iter = FALSE, mc.cores,
                        progressr.message = NULL,
                        estimate.index = 1:2,
                        ...) {
  stm <- proc.time()
  oldtm <- rep(0, 5)
  if (!exists(".Random.seed", envir = .GlobalEnv, inherits = FALSE)) {
    runif(1)
  }
  if (is.null(seed)) {
    RNGstate <- get(".Random.seed", envir = .GlobalEnv)
  } else {
    R.seed <- get(".Random.seed", envir = .GlobalEnv)
    set.seed(seed)
    RNGstate <- structure(seed, kind = as.list(RNGkind()))
    on.exit(assign(".Random.seed", R.seed, envir = .GlobalEnv))
  }
  olddata <- NULL
  dots <- list(...)
  mycall <- match.call(expand.dots = FALSE)
  if (inherits(x, c("data.frame", "matrix"))) olddata <- x
  if (inherits(x, "sim")) {
    oldtm <- attr(x, "time")
    oldcall <- attr(x, "call")
    x <- attr(x, "f")
    if (!is.null(f)) x <- f
    ex <- oldcall[["..."]]
    for (nn in setdiff(names(ex), names(dots))) {
      dots[[nn]] <- ex[[nn]]
      val <- list(ex[[nn]])
      names(val) <- nn
      mycall[["..."]] <- c(mycall[["..."]], list(val))
    }
  } else {
    if (!is.null(f)) x <- f
    if (!is.function(x)) stop("Expected a function or 'sim' object.")
  }
  if (is.null(x)) stop("Must give new function argument 'f'.")
  # `par.index`: list for storing position of estimate, se, confint when
  # the returned simulation object is an `estimate` object
  par.index <- list()
  seed_sequence <- NULL
  res <- val <- NULL
  on.exit({
    if (is.null(colnames) && !is.null(val)) {
      if (is.matrix(val[[1]])) {
        colnames <- base::colnames(val[[1]])
      } else {
        colnames <- names(val[[1]])
      }
    }
    base::colnames(res) <- colnames
    if (!is.null(olddata)) res <- rbind(olddata, res)
    attr(res, "call") <- mycall
    attr(res, "f") <- x
    cls <- ifelse(is.data.frame(res), "data.frame", "matrix")
    class(res) <- c("sim", cls)
    attr(res, "time") <- proc.time() - stm + oldtm
    attr(res, "par.index") <- par.index
    attr(res, "seeds") <- seed_sequence
    return(res)
  })
  if (inherits(R, c("matrix", "data.frame")) || length(R) > 1) {
    if (is.list(R)) {
      parval <- R
      ## list of parameters for each iteration
    } else if (inherits(R, c("matrix", "data.frame"))) {
      parval <- as.data.frame(R)
      names(parval) <- colnames(R)
      R <- NROW(parval)
    }
  } else {
    parval <- as.data.frame(1:R)
    names(parval) <- NULL
  }
  estimate.index <- sort(estimate.index)
  if (!all(estimate.index %in% 1:5)) {
    stop("wrong estimate.index")
  }
  repl <- NROW(parval)
  pb <- progressr::progressor(steps = repl)
  robx <- function(iter__, ...) {
    rng_seed <- if (exists(".Random.seed", envir = .GlobalEnv, inherits = FALSE))
                  get(".Random.seed", envir = .GlobalEnv)
                else NULL
    if (!is.null(progressr.message)) {
      pb(message = progressr.message(...))
    } else {
      pb()
    }
    res <- tryCatch(x(...), error = function(e) NA)
    is_estimate <- inherits(res, c("estimate", "summary.estimate", "targeted"))
    extra <- NULL

    if (is_estimate) {
      extra <- attributes(res)$extra # defaults to NULL if attribute doesn't
      # exist // only extract extra attribute for estimate-type objects
      idx <- intersect(seq_len(5L), estimate.index) # parameters to keep
      cmat <- lava::parameter(res)[, idx, drop = FALSE]
      res <- as.vector(cmat)
      cn <- colnames(cmat)
      nam <- c()
      for (i in seq_along(idx)) {
        newn <- rownames(cmat)
        if (idx[i] > 1L) newn <- paste(newn, cn[i], sep=".")
        nam <- c(nam, newn)
      }
      names(res) <- nam
    }
    if (!is.null(extra)) { # append extra par. not part of the estimate object
      res <- c(res, extra)
    }
    return(structure(res,
                     "estimate" = is_estimate,
                     "n.extra" = length(extra),
                     ".rng_seed" = rng_seed))
  }
  if (iter || !is.data.frame(parval)) {
    formals(robx)[[1]] <- NULL
  }

  if (is.data.frame(parval)) {
    pp <- c(
      as.list(parval), dots,
      list(FUN = robx, SIMPLIFY = FALSE, MoreArgs = as.list(args))
    )
  } else { ## parameters as a list
    for (i in seq_along(parval)) {
      parval[i] <- c(parval[i], dots, args)
    }
    pp <- c(
      list(parval),
      list(FUN = robx, SIMPLIFY = FALSE)
    )
  }
  if (is.null(pp$future.seed)) {
    pp$future.seed <- TRUE
  }
  if (!missing(mc.cores)) {
    pp$future.seed <- NULL
    pp$mc.cores <- mc.cores
  }

  if (!missing(mc.cores)) {
    if (is.null(mc.cores)) {
      val <- do.call(mapply, pp)
    } else {
      val <- do.call(parallel::mcmapply, pp)
    }
  } else {
    val <- do.call(future.apply::future_mapply, pp)
  }
  seed_sequence <- lapply(val, attr, ".rng_seed")
  n.extra <- attr(val[[1]], "n.extra")
  val <- lapply(val, function(v) {
    attr(v, ".rng_seed") <- NULL
    attr(v, "n.extra") <- NULL
    v
  })
  res <- do.call(rbind, val)
  if (is.null(res)) {
    res <- matrix(NA, ncol=length(val[[1]]), nrow=repl)
  }
  if (attr(val[[1]], "estimate")) {
    ncol_est <- ncol(res) - n.extra
    if (ncol_est %% length(estimate.index) != 0L) {
      warning("unexpected number of columns")
    }
    idx <- seq_len(ncol_est)
    np <- ncol_est / length(estimate.index)
    if (1 %in% estimate.index) {
      par.index[["estimate"]] <- seq_len(np)
      idx <- setdiff(idx, par.index[["estimate"]])
    }
    if (2 %in% estimate.index) {
      par.index[["se"]] <- idx[seq_len(np)]
      idx <- setdiff(idx, par.index[["se"]])
    }
    if (3 %in% estimate.index) {
      par.index[["lower"]] <- idx[seq_len(np)]
      idx <- setdiff(idx, par.index[["lower"]])
    }
    if (4 %in% estimate.index) {
      par.index[["upper"]] <- idx[seq_len(np)]
      idx <- setdiff(idx, par.index[["upper"]])
    }
    if (5 %in% estimate.index) {
      par.index[["pval"]] <- idx[seq_len(np)]
      idx <- setdiff(idx, par.index[["pval"]])
    }
    if (n.extra > 0L) {
      par.index[["extra"]] <- seq.int(ncol_est + 1L, ncol(res))
    }
  }
  res
}

#' @export
cbind.sim <- function(x, ...) {
  res <- cbind(as.data.frame(x), ...)
  as.sim(res)
}

#' @export
rbind.sim <- function(x, ...) {
  res <- rbind(as.data.frame(x), ...)
  as.sim(res)
}

#' @export
as.vector.sim <- function(x, mode="any") {
  as.vector(x[,,drop=TRUE], mode=mode)
}

#' @export
as.matrix.sim <- function(x, ...) {
  if (inherits(x, "data.frame")) {
    return(as.matrix(as.data.frame(x)))
  }
  class(x) <- "matrix"
  attr(x, "call") <- NULL
  attr(x, "f") <- NULL
  attr(x, "time") <- NULL
  x
}

#' @export
"[.sim" <- function(x, i, j, drop = FALSE, keep.index = FALSE) {
  atr <- attributes(x)
  if (!is.null(dim(x))) {
    class(x) <- class(x)[2]
  } else {
    class(x) <- class(x)[-1]
  }
  cname <- colnames(x)
  x <- NextMethod("[", drop=drop)
  atr.keep <- c("call", "time")
  if (keep.index) atr.keep <- c(atr.keep, "par.index")
  if (missing(j)) {
    atr.keep <- c(atr.keep, "f")
  } else {
    if (keep.index) {
      pidx <- atr[["par.index"]]
      if (!is.null(pidx)) {
        if (is.character(j)) {
        j <- which(cname %in% j)
        }
        pidx <- lapply(pidx, function(x) which(j %in% x))
        atr[["par.index"]] <- pidx
      }
    }
  }
  attributes(x)[atr.keep] <- atr[atr.keep]
  if (!drop) class(x) <- c("sim", class(x))
  return(x)
}

#' @export
"as.sim" <- function (object, name, ...) {
  if (is.vector(object)) {
    cl <- ifelse(inherits(class(object), "data.frame"), "data.frame", "matrix")
    object <- structure(cbind(object), class=c("sim", cl))
    if (!missing(name)) colnames(object) <- name
    return(object)
  }
  structure(object, class=c("sim", class(object)))
}

Time <- function(sec,print=FALSE,...) {
    h <- sec%/%3600
    m0 <- (sec%%3600)
    m <- m0%/%60
    s <- m0%%60
    res <- c(h=h, m=m, s=s)
    if (print) {
        if (h>0) cat(h, "h ", sep="")
        if (m>0) cat(m, "m ", sep="")
        cat(s, "s", sep="")
        return(invisible(res))
    }
    return(res)
}


#' Generic print method
#'
#' Nicer print method for tabular data. Falls back to standard print method for
#' all other data types.
#' @export
#' @param x object to print
#' @param n number of rows to show from top and bottom of tabular data
#' @param digits precision
#' @param ... additional arguments to print method
Print <- function(x, n=5,
                  digits=max(3, getOption("digits")-3), ...) {
    mat <- !is.null(dim(x))
    if (!mat) {
      if (is.vector(x)) {
          x <- cbind(x)
          colnames(x) <- ""
      } else {
        print(x, ...)
        return(invisible(x))
      }
    }
    if (is.null(rownames(x))) {
        rownames(x) <- seq(nrow(x))
    }
    sep <- rbind("---"=rep('', ncol(x)))
    if (n<1) {
        print(x, quote=FALSE, digits=digits, ...)
    } else {
      if (NROW(x)<=(2*n)) {
        hd <- base::format(x, digits=digits, ...)
        print(hd, quote=FALSE, ...)
      } else {
        hd <- base::format(utils::head(x, n), digits=digits, ...)
        tl <- base::format(utils::tail(x, n), digits=digits, ...)
        print(rbind(base::as.matrix(hd), sep, base::as.matrix(tl)),
              quote=FALSE, ...)
      }
    }
    invisible(x)
}

#' @export
print.sim <- function(x, ...) {
    s <- summary(x, minimal=TRUE, ...)
    attr(x, "f") <- attr(x, "call") <- NULL
    Print(x, ...)
    cat("\n")
    if (nrow(x)>1) print(s, extra=FALSE, ...)
    return(invisible(x))
}

#' @export
print.summary.sim <- function(x,
                              group=list(c("^mean$","^sd$","^se$","^se/sd$",
                                           "^coverage","^length"),
                                   c("^min$","^[0-9.]+%$","^max$"),
                                   c("^na$","^missing$"),
                                   c("^true$","^bias$","^rmse$")),
                      lower.case=TRUE,
                      na.print="",
                      digits = max(3, getOption("digits") - 2),
                      quote=FALSE,
                      time=TRUE,
                      extra=TRUE,
                      ...) {
    if (extra) {
        cat(attr(x,"n")," replications",sep="")
        if (time && !is.null(attr(x,"time"))) {
            cat("\t\t\t\t\tTime: ")
            Time(attr(x,"time")["elapsed"],print=TRUE)
        }
        cat("\n\n")
    }

    nn <- rownames(x)
    if (lower.case)  nn <- tolower(nn)
    gg <- lapply(group,
                 function(x) unlist(lapply(x,function(v) grep(v,nn))))
    gg <- c(gg,list(setdiff(seq_along(nn),unlist(gg))))

    x0 <- c()
    ng <- length(gg)
    for (i in seq(ng)) {
        x0 <- rbind(x0, x[gg[[i]],,drop=FALSE],
        { if(i<ng && length(gg[[i+1]])>0) NA})
    }

    print(structure(x0,class="matrix")[,,drop=FALSE],digits=digits,quote=quote,na.print=na.print,...)
    if (extra) cat("\n")
    invisible(x)
}


#' Summary method for 'sim' objects
#'
#' Summary method for 'sim' objects
#' @export
#' @export summary.sim
#' @param object sim object
#' @param estimate (optional) columns with estimates
#' @param se (optional) columns with standard error estimates
#' @param confint (optional) list of pairs of columns with confidence limits
#' @param true (optional) vector of true parameter values
#' @param fun (optional) summary function
#' @param names (optional) names of estimates
#' @param unique.names if TRUE, unique.names will be applied to column names
#' @param minimal if TRUE, minimal summary will be returned
#' @param level confidence level (0.95)
#' @param quantiles quantiles (0,0.025,0.5,0.975,1)
#' @param df degrees of freedom in t-distribution used for constructing CIs
#'   (default Gaussian approximation)
#' @param ... additional levels to lower-level functions
summary.sim <- function(object,estimate=NULL,se=NULL,
                confint,
                true=NULL,
                fun,names=NULL,unique.names=TRUE,minimal=FALSE,
                level=0.95,
                quantiles=c(0,.025,0.5,.975,1),
                df = Inf,
                ...) {
    if (is.list(estimate)) {
        est <- estimate
        if (is.null(names)) names <- base::names(est)
        estimate <- c()
        nse  <- is.null(se)
        ntrue <- is.null(true)
        elen <- unlist(lapply(est,length))
        est <- lapply(est, function(e) c(e, rep(NA,max(elen)-length(e))))
        for (e in est) {
            estimate <- c(estimate,e[1])
            if (length(e)>1 && nse) se <- c(se,e[2])
            if (length(e)>2 && ntrue) true <- c(true,e[3])
        }
        cl <- match.call()
        cl[c("estimate","se","true","names")] <- list(estimate,se,true,names)
    }
    par.index <- attr(object, "par.index")
    if (is.null(estimate) && length(par.index) > 0) {
      estimate <- par.index[["estimate"]]
    } else {
      if (is.character(estimate)) {
        estimate <- match(estimate, base::colnames(object))
      }
    }
    np <- length(estimate)
    if (is.null(se) && length(par.index) > 0 && !is.null(par.index[["se"]])) {
      # Map each requested estimate column to its SE column (NA if unknown)
      se <- rep(NA_integer_, np)
      pi_est <- par.index[["estimate"]]
      pi_se <- par.index[["se"]]
      for (i in seq_len(np)) {
        j <- match(estimate[i], pi_est)
        if (!is.na(j)) se[i] <- pi_se[j]
      }
    }
    if (missing(confint)) {
      lo <- par.index[["lower"]]
      up <- par.index[["upper"]]
      if (length(par.index) > 0 &&
          (!is.null(lo) || !is.null(up))) {
        pi_est <- par.index[["estimate"]]
        if (is.null(lo)) lo <- rep(-Inf, length(pi_est))
        if (is.null(up)) up <- rep(Inf, length(pi_est))
        confint <- list()
        for (i in seq_len(np)) {
          j <- match(estimate[i], pi_est)
          if (!is.na(j)) {
            confint <- c(confint, list(c(lo[j], up[j])))
          } else {
            confint <- c(confint, list(c(NA_integer_, NA_integer_)))
          }
        }
      } else confint <- !is.null(se)&&!is.null(true)
    }
    if (minimal) {
        fun <- function(x,se,confint,...) {
            res <- c(Mean=mean(x,na.rm=TRUE),
                    SD=sd(x,na.rm=TRUE))
            if (!missing(se) && !is.null(se)) {
                res <- c(res, c(SE=mean(se,na.rm=TRUE)))
                res <- c(res, c("SE/SD"=res[["SE"]]/res[["SD"]]))
            }
            return(res)
        }
    }
    mfun <- function(x,...) {
        res <- c(mean(x,na.rm=TRUE),
                 sd(x,na.rm=TRUE),
                 if (length(quantiles)>0) quantile(x,quantiles,na.rm=TRUE),
                 mean(is.na(x)))
        if (length(quantiles)>0) {
            nq <- paste0(quantiles*100,"%")
            idx <- which(quantiles==1)
            if (length(idx)>0) nq[idx] <- "Max"
            idx <- which(quantiles==0)
            if (length(idx)>0) nq[idx] <- "Min"
        }
        names(res) <- c("Mean","SD",
                        if (length(quantiles)>0) nq,
                        "Missing")
        res
    }
    tm <- attr(object,"time")
    N <- max(length(estimate),length(se),length(true))
    if (!is.null(estimate)) estimate <- rep(estimate,length.out=N)
    if (!is.null(se)) se <- rep(se,length.out=N)
    if (!is.null(true)) {
        if (is.null(estimate)) N <- ncol(object)
        true <- rep(true,length.out=N)
    }

    if (!is.null(estimate) && is.character(estimate)) {
        estimate <- match(estimate,colnames(object))
    }
    if (is.list(confint)) confint <- unlist(confint)
    if (!missing(fun)) {
        if (!is.null(estimate)) m.est <- object[,estimate,drop=FALSE]
        else m.est <- object
        m.se <- NULL
        if (!is.null(se)) m.se <- object[,se,drop=FALSE]
        m.ci <- NULL
        if (!is.null(confint)) m.ci <- object[,confint,drop=FALSE]
        res <- lapply(seq(ncol(m.est)),
                      function(i,...) fun(m.est[,i,drop=TRUE],
                                          se=m.se[,i,drop=TRUE],
                                          confint=m.ci[,1:2+(i-1)*2],...,
                                          INDEX=i),...)
        res <- matrix(unlist(res),nrow=length(res[[1]]),byrow=FALSE)
        if (is.null(dim(res))) {
            res <- rbind(res)
        }
        if (is.null(rownames(res))) {
            rownames(res) <- names(fun(0,m.se,m.ci,INDEX=1,...))
            if (is.null(rownames(res))) rownames(res) <- rep("",nrow(res))
        }
        if (is.null(colnames(res))) {
            colnames(res) <- colnames(m.est)
        }
        return(structure(res,
                    n=NROW(object),
                    time=tm,
                    class=c("summary.sim","matrix")))
    }

    if (!is.null(estimate)) {
        est <- apply(object[,estimate,drop=FALSE],2,mfun)
    } else {
        est <- apply(object,2,mfun)
    }

    if (!is.null(true)) {
        if (length(true)!=ncol(est)) {
            ##stop("'true' should be of same length as 'estimate'.")
            true <- rep(true,length.out=ncol(estimate))
        }
        est <- rbind(est,
                     rbind(True=true),rbind(Bias=est["Mean",]-true),
                     rbind(RMSE=((est["Mean",]-true)^2+(est["SD",])^2)^.5)
                     )
    }
    if (!is.null(se)) {
        if (is.character(se)) {
            se <- match(se,colnames(object))
        }
        if (length(se)!=ncol(est)) stop("'se' should be of same length as 'estimate'.")
        est <- rbind(est, SE=apply(object[,se,drop=FALSE],2,
                                   function(x) val <- c(mean(x,na.rm=TRUE))))
        est <- rbind(est,"SE/SD"=est["SE",]/est["SD",])
    }
    if (!is.null(confint) && (length(confint)>1 || confint)) {
        if (is.character(confint)) {
            confint <- match(confint,colnames(object))
        }
        if (length(confint)==1 && confint) {
            if (is.null(se)) stop("Supply confidence limits or SE")
            confint <- c()
            pos <- ncol(object)
            for (i in seq_along(estimate)) {
              z <- 1-(1-level)/2
              if (is.na(se[i])) {
                CI <- matrix(NA, nrow=NROW(object), ncol=2)
              } else {
                q <- qt(z, df = df)
                CI <- cbind(object[,estimate[i]] - q*object[,se[i]],
                            object[,estimate[i]] + q*object[,se[i]])
              }
              colnames(CI) <- NULL
              object <- cbind(object,CI)
              confint <- c(confint,pos+1:2)
              pos <- pos+2
            }
        }
        if (length(confint)!=2*length(estimate)) stop("'confint' should be of length 2*length(estimate).")
        Coverage <- c()
        Length <- c()
        for (i in seq_along(estimate)) {
            upper <- object[,confint[2*i]][[1]]
            lower <- object[,confint[2*(i-1)+1]][[1]]
            Coverage <- c(
              Coverage,
              mean((lower < true[i]) & (upper > true[i]), na.rm=TRUE)
            )
            Length <- c(
              Length,
              mean(abs(upper - lower), na.rm = TRUE)
            )
        }
        est <- rbind(est,Coverage=Coverage, Length=Length)
    }

    if (!is.null(names)) {
         if (length(names)<ncol(est)) {
            uest <- unique(estimate)
            names <- names[match(estimate,uest)]
        }
        colnames(est) <- names

    }
    if (unique.names && !is.null(colnames(est))) {
        colnames(est) <- make.unique(colnames(est))
    }
    est[is.nan(est)] <- NA

    return(structure(est,
                     n=NROW(object),
                     time=tm,
                     class=c("summary.sim","matrix")))
}

##' @export
"parameter<-" <- function(x,...,value) UseMethod("parameter<-")

##' @export
"parameter<-.lvmfit" <- function(x,...,value) {
  parameter(Model(x),...) <- value
  return(x)
}


##' @export
"parameter<-.lvm" <- function(x,constrain,start,remove=FALSE,...,value) {
  if (inherits(value,"formula")) value <- all.vars(value)
  x <- rmvar(x, value)
  if (remove) {
      x$expar[value] <- NULL
      x$exfix[value] <- NULL
      x$attributes$parameter[value] <- NULL
      index(x) <- reindex(x)
      return(x)
      
  }
  if (!missing(start)) {
      if (length(start) != length(value)) stop("'start' and 'value' should be of the same lengths")
      start <- as.list(start)
      names(start) <- value
  } else {
      start <- as.list(rep(0,length(value))); names(start) <- value
  }
  if (!missing(constrain)) {
      newfix <- constrain
      if (!is.list(newfix)) newfix <- as.list(newfix)
  } else {
      newfix <- as.list(value);
  }
  names(newfix) <- value
  x$expar[value] <- start
  x$exfix[value] <- newfix
  index(x) <- reindex(x)
  x$attributes$parameter[value] <- TRUE
  return(x)
}

##' @export
parameter <- function(x,...,value) UseMethod("parameter")

##' @export
parameter.default <- function(x, var, ...) {
    if (missing(var)) return (names(unlist(x$attributes$parameter)))
    parameter(x, ...) <- var
    return(x)
}

set.seed(1)
center_ic <- function(n, p = 1) {
  x <- matrix(rnorm(n * p), n, p)
  scale(x, center = TRUE, scale = FALSE)
}
a1 <- estimate(coef = 1, IC = center_ic(10), id = 1:10, labels = "a1")
a2 <- estimate(coef = 2, IC = center_ic(10), id = 1:10, labels = "a2")
a3 <- estimate(coef = 3, IC = center_ic(10), id = 1:10, labels = "a3")

test_that("c.summary.estimate concatenates coefficient matrices", {
  # test basic functionality without any extra variables
  s1 <- summary(a1)
  s2 <- summary(a2)
  s3 <- summary(a3)
  cc <- c(s1, s2, s3)

  # Result is a summary.estimate object
  expect_s3_class(cc, "summary.estimate")

  # coefmat is the row-bind of the inputs (preserving rows and columns)
  expect_equal(rownames(cc$coefmat), c("a1", "a2", "a3"))
  expect_equal(colnames(cc$coefmat), colnames(s1$coefmat))
  expect_equivalent(cc$coefmat, rbind(s1$coefmat, s2$coefmat, s3$coefmat))

  # coef.summary.estimate returns the combined coef vector
  expect_equal(coef(cc), cc$coefmat[, 1])
  expect_equal(parameter(cc), cc$coefmat)

  # original summaries are retained in the objects element
  expect_length(cc$objects, 3L)
  expect_equal(cc$objects[[2]]$coefmat, s2$coefmat)

  # check variance covariance matrix
  expect_equal(dim(vcov(cc)), c(3L, 3L))
})

test_that("c.summary.estimate with single argument returns input unchanged", {
  s1 <- summary(a1)
  expect_identical(c(s1), s1)
})

test_that("c.summary.estimate carries a custom print method", {
  cc <- c(summary(a1), summary(a2))
  expect_type(cc$print, "closure")
  out <- capture.output(print(cc))
  expect_true(any(grepl("Concatenated summary.estimate objects", out)))
  expect_true(any(grepl("a1", out)))
  expect_true(any(grepl("a2", out)))
})

test_that("c.summary.estimate with extra variables", {
  s1 <- summary(a1)
  ss <- c(s1, niter = 1) # single numerical argument
  expect_equal(attributes(ss)$extra, c(niter = 1))

  ss <- c(s1, 1) # single unnamed variable
  expect_equal(attributes(ss)$extra, 1)

  expect_equal(attributes(c(ss, a = 1))$extra, c(1, a = 1))

  # with single named vector
  ss <- c(s1, c(niter = 1, cc = 2)) # single numerical argument
  expect_equal(attributes(ss)$extra, c(niter = 1, cc = 2))

  ss <- c(s1, niter = 1, cc = 2) # multiple numerical arguments
  expect_equal(attributes(ss)$extra, c(niter = 1, cc = 2))

  # two summary.estimate objects, of which one has extra attributes
  ss <- c(s1, ss)
  expect_equal(attributes(ss)$extra, c(niter = 1, cc = 2))

  # two summary.estimate objects, with both having extra attributes
  ss <- c(c(s1, aa = 1), c(s1, bb = 2))
  expect_equal(attributes(ss)$extra, c(aa = 1, bb = 2))

  # adding additional extra to existing extra attribute
  ss <- c(ss, c(cc = 3))
  expect_equal(attributes(ss)$extra, c(aa = 1, bb = 2, cc = 3))
}
)

test_that("summary.estimate class tests", {
  # tests for future regression when removing estimate class "inheritance"
  # from summary.estimate object class definition
  ss <- summary(a1)
  expect_true(
    all(c(
      inherits(ss, "summary.estimate"),
      inherits(ss, "estimate")
    ))
  )

  # same for concatenation
  ss <- c(summary(a1), aa = 1)
  expect_true(
    all(c(
      inherits(ss, "summary.estimate"),
      inherits(ss, "estimate")
    ))
  )
})

pzmax <- function(alpha, S) {
  ## P(Zmax > z) Family wise error rate, Zmax = max |Z_i|
  if (!requireNamespace("mets",quietly=TRUE))
    stop("'mets' package required")
  k <- nrow(S)
  zz <- qnorm(1-alpha/2)
  unlist(lapply(zz, function(z)
    1 - mets::pmvn(lower=rep(-z, k),
                   upper=rep(z, k),
                   sigma=cov2cor(S))
    ))
}

##' @export
alpha_zmax <- function(object, method, alpha = 0.05, ...) {
  if (!inherits(object, "estimate")) {
    stop("Expected an 'estimate' object")
  }
  dots <- list(...)
  null <- object$compare$null
  if (!("null" %in% names(dots)) && !is.null(null)) {
    dots$null <- null
  }
  object <- do.call(estimate, c(list(object), dots))
  est <- parameter(object)[, c(1, 5), drop = FALSE]
  padj <- pzmax(est[, 2], vcov(object))
  res <- cbind(est, padj)
  colnames(res)[3] <- paste0("Adj.", colnames(res)[2])
  f <- function(a) pzmax(a, vcov(object)) - alpha
  adj <- uniroot(f, lower = 0, upper = alpha)$root
  attributes(res)["adjusted.significance.level"] <- adj
  res
}

##' Closed testing procedure
##'
##' Given p hypotheses \eqn{H_1, \ldots, H_p}{H1, ..., Hp} all
##' \eqn{2^p-1}{2^p-1} intersection hypotheses are calculated and adjusted
##' p-values are obtained for \eqn{H_j}{Hj} is calculated as the max p-value of
##' all intersection hypotheses containing Hj. Example, for p=3, the adjusted
##' p-value for \eqn{H_1}{H1} will be obtained from \eqn{\{(H1, H2, H3),
##' (H1,H2), (H1,H3), (H1)\}}.
##'
##' \if{html}{
##'   \figure{closedtesting.svg}{options: width=500}
##' }
##' \if{latex}{
##'   \figure{closedtesting.pdf}{options: width=7cm}
##' }
##' @aliases closed_testing alpha_zmax
##' @param object `estimate` object
##' @param test function that conducts hypothesis test. See details below.
##' @param ... Additional arguments passed to `test`
##' @export
##' @details The function `test` should be a function `function(object, index,
##'   ...)` which as its first argument takes an `estimate` object and with
##'   an argument `index` which is a integer vector specifying which
##'   subcomponents of `object` to test. The ellipsis argument can be any other
##'   arguments used in the test function. The function [test_wald()] is an
##'   example of valid test function (which has an additional argument `null` in
##'   reference to the above mentioned ellipsis arguments).
##' @references Marcus, R; Peritz, E; Gabriel, KR (1976).
##'   "On closed testing procedures with special reference to ordered analysis of variance".
##'   Biometrika. 63 (3): 655–660.
##' @examples
##' m <- lvm()
##' regression(m, c(y1,y2,y3,y4)~x) <- c(0, 0.25, 0, 0.25)
##' regression(m, to=endogenous(m), from="u") <- 1
##' variance(m,endogenous(m)) <- 1
##' set.seed(1)
##' d <- sim(m, 200)
##' l1 <- lm(y1~x,d)
##' l2 <- lm(y2~x,d)
##' l3 <- lm(y3~x,d)
##' l4 <- lm(y4~x,d)
##'
##' (a <- merge(l1, l2, l3, l4, subset=2))
##' if (requireNamespace("mets",quietly=TRUE)) {
##'    alpha_zmax(a)
##' }
##' adj <- closed_testing(a)
##' adj
##' adj$p.value
##' summary(adj)
closed_testing <- function(object, test = test_wald, ...) {
  if (!inherits(object, "estimate")) stop("`estimate` object needed")
  idx <- seq_along(coef(object))
  if (length(idx) > 15)
    stop("Too many tests. Consider some other adjustment method")
  combs <- pvals <- c()
  for (i in seq_along(idx)) {
    co <- combn(idx, i)
    pp <- numeric(ncol(co))
    for (j in seq_along(pp)) {
      ii <- co[, j]
      tt <- test(object, index = ii, ...)
      pp[j] <- tt$p.value
    }
    combs <- c(combs, list(co))
    pvals <- c(pvals, list(pp))
  }
  pmax <- c()
  for (k in seq_along(idx)) {
    pk <- c()
    for (i in seq_along(idx)) {
      cols <- apply(combs[[i]], 2, function(x) k %in% x)
      pk <- c(pk, pvals[[i]][which(cols)])
    }
    pmax <- c(pmax, max(pk))
  }
  res <- cbind(coef(object), pmax)
  colnames(res) <- c("Estimate", "adj.p")
  rownames(res) <- names(coef(object))
  names(pmax) <- names(coef(object))
  structure(
    list(
      adjp = res,
      call = match.call(),
      args = list(...),
      hypotheses = combs,
      raw.pval = pvals,
      estimate = coef(object),
      p.value = pmax),
    class = "test_adj"
  )
}


#' @export
print.test_adj <- function(x, ...) {
  cat("Call: ")
  print(x$call, quote = FALSE)
  cat("\n")
  print(x$adjp)
}

#' @export
summary.test_adj <- function(object, ...) {
  cat("Call: ")
  print(object$call, quote = FALSE)
  # Print adjusted p-values
  cli::cli_h2("Adjusted p-values")
  print(object$adjp)
  cat("\n")
  # Print raw p-values for intersection hypotheses
  cli::cli_h2("Raw p-values for intersection hypotheses")
  param_names <- names(object$estimate)
  for (i in seq_along(object$hypotheses)) {
    cat(sprintf("%d-way intersections:\n", i))
    comb_matrix <- object$hypotheses[[i]]
    pvals <- object$raw.pval[[i]]
    for (j in seq_len(ncol(comb_matrix))) {
      indices <- comb_matrix[, j]
      hyp_names <- param_names[indices]
      # Create hypothesis notation
      hyp_str <- paste0("{", paste(hyp_names, collapse = ", "), "}")
      # Print with p-value
      cat(sprintf("  %-40s p = %.4f\n", hyp_str, pvals[j]))
    }
    cat("\n")
  }
  invisible(object)
}

#' @export
test_wald <- function(par,
                      vcov,
                      null = NULL,
                      index = NULL) {
  if (!inherits(par, "estimate")) {
    par <- lava::estimate(par=par, vcov=vcov)
  }
  if (is.null(null)) {
    null <- rep(0, length(coef(par)))
  }
  B <- diag(length(coef(par)))
  if (!is.null(index)) {
    if (length(coef(par)) < length(index)) stop("wrong `index`")
    null <- null[index]
    B <- B[index, , drop=FALSE]
  }
  lava::compare(par, contrast = B, null=null)
}

##' @export
IC <- function(x, ...) UseMethod("IC")

##' @export
influence.estimate <- function(model, ...)
    IC(model, ...)


##' Extract influence function from model object
##'
##' Extract i.i.d. decomposition (influence function) from model object
##' @aliases IC
##' @seealso [var_ic], [iid]
##' @param x model object
##' @param id (optional) id/cluster variable
##' @param bread (optional) Inverse of derivative of mean score function
##' @param ... additional arguments
##' @examples
##' m <- lvm(y~x+z)
##' distribution(m, ~y+z) <- dist_bernoulli("logit")
##' d <- sim(m,1e3)
##' g <- glm(y~x+z,data=d,family=binomial)
##' var_ic(IC(g))
##' @return Matrix with rows corresponding to observations and columns to
##'   parameters, representing the estimated influence function. Attributes
##'   include `bread` (derivative of the estimating equation).
##' @export
IC.default <- function(x, bread, id=NULL,
                       ...) {
  if (any(paste("iid", class(x), sep=".") %in% methods("iid"))) {
    ## 'iid' method exists for the specific class.
    ## This is a scaled version of the influence function, hence
    ## we need to rescale.
    cl <- match.call()
    cl[[1]] <- substitute(iid)
    ii <- eval.parent(cl)
    if (!is.null(attr(ii, "bread"))) {
      attr(res, "bread") <- attr(res, "bread")*NROW(res)
    }
    ii <- ii*NROW(ii)
    return(ii)
  }
  if (!any(paste("score", class(x), sep=".") %in% methods("score"))) {
    warning("Not available for this class")
    return(NULL)
  }

  ## if (folds>0 || maxsize>0 ||
  ##     (!missing(id) && lava.options()$cluster.index)) {
  ##   if (!requireNamespace("mets", quietly=TRUE)) stop("Requires 'mets'")
  ## }
  ## if (folds>0) {
  ##   U <- Reduce(
  ##     "rbind",
  ##     mets::divide.conquer(function(data) score(x, data = data, ...),
  ##                          id = id,
  ##                          data = data, size = round(nrow(data) / folds)
  ##       )
  ##   )
  ## } else {
  U <- score(x, indiv=TRUE, ...)
  pp <- pars(x)
  if (!missing(bread) && is.null(bread)) {
    bread <- suppressWarnings(vcov(x)*NROW(U))
  }
  if (missing(bread)) bread <- attributes(U)$bread
  if (is.null(bread)) {
    bread <- attributes(x)$bread
    if (is.null(bread)) bread <- x$bread
    if (is.null(bread)) {
      I <- -numDeriv::jacobian(function(p) {
        score(x, p = p, indiv = FALSE, ...)
      }, pp, method = lava.options()$Dmethod)
      bread <- Inverse(I)*NROW(U)
    }
  }

  ic0 <- U%*%bread
  if (!missing(id)) {
  N <- nrow(ic0)
  if (!lava.options()$cluster.index) {
    ic0 <- matrix(unlist(by(ic0, id, colSums)),
                  byrow=TRUE, ncol=ncol(bread))
  } else {
    ic0 <- mets::cluster.index(id, mat=ic0, return.all=FALSE)
  }
  ic0 <- ic0*NROW(ic0)/length(id)
  attributes(ic0)$N <- N
}
colnames(ic0) <- colnames(U)
return(structure(ic0, bread=bread))
}

##' @export
IC.multigroupfit <- function(x, ...) IC.default(x, combine=TRUE, ...)

##' @export
IC.matrix <- function(x, ...) {
    p <- NCOL(x)
    n <- NROW(x)
    mu <- colMeans(x, na.rm = TRUE)
    S <- var(x, use = "pairwise.complete.obs") * (n - 1) / n
    ic1 <- t(t(x)-mu)
    ic2 <- matrix(ncol = (p + 1) * p / 2, nrow = n)
    pos <- 0
    nn <- c()
    cc <- mu
    for (i in seq(p))
        for (j in seq(i, p)) {
            pos <- pos+1
            cc <- c(cc, S[i, j])
            ic2[, pos] <- (ic1[, i] * ic1[, j]) - cc[length(cc)]
            nn <- c(nn, paste(colnames(x)[c(i, j)],
                              collapse = lava.options()$symbols[2]
            ))
        }
    colnames(ic1) <- colnames(x)
    colnames(ic2) <- nn
    names(cc) <- c(colnames(ic1), colnames(ic2))
    res <- cbind(ic1, ic2)
    rownames(res) <- rownames(x)
    structure(res,
              coef=cc,
              mean=mu, var=S)
}

##' @export
IC.numeric <- function(x, ...) {
    n <- length(x)
    mu <- mean(x)
    S <- var(x) * (n - 1) / n
    ic1 <- t(t(x) - mu)
    structure(cbind(mean=ic1, var=(ic1^2 - S)),
              coef=c(mean=mu, var=S), mean=mu, var=S)
}

##' @export
IC.data.frame <- function(x, ...) {
  if (!all(apply(x[1, , drop=FALSE], 2,
                 function(x) inherits(x, c("numeric", "integer")))))
        stop("Don't know how to handle data.frames of this type")
    IC(as.matrix(x))
}

updatelvm <- function(x,mean=TRUE,...) {
    index(x) <- reindex(x,mean=mean,...)
    x$parpos <- parpos(x,mean=mean,...)
    return(x)
}

##' Generic method for extract index of an object
##'
##' @export
##' @param x object from which to extract index
##' @param ... further arguments to be passed to or from other methods.
##' @seealso [index<-]
##' @keywords utilities
##' @rdname index-method
"index" <- function(x,...) UseMethod("index")

##' Generic method for setting the index of an object
##'
##' @export
##' @param x object on which to set the index
##' @param value the new index
##' @param ... further arguments to be passed to or from other methods.
##' @seealso [index]
##' @keywords utilities
##' @rdname index-method
##' @export
"index<-" <- function(x,...,value) UseMethod("index<-")

##' @title Extract the parameter indicies of a lvm object
##'
##' @description Extracts the matrices with indices of model parameters from a
##'   latent variable model (`lvm`). Returns a list with
##' * A:   Matrix with fixed parameters and ones where parameters are free
##' * J:   Manifest variable selection matrix
##' * M0:  Index of free regression parameters
##' * M1:  Index of free and _unique_ regression parameters
##' * P:   Matrix with fixed variance parameters and ones where parameters are free
##' * P0:  Index of free variance parameters
##' * P1:  Index of free and _unique_ regression parameters
##' * npar.var:   Number of covariance parameters
##' @rdname index-lvm
##' @inherit index
##' @aliases index.lvmfit index.lvm
##' @seealso [modelPar()]
##' @export
"index.lvm" <- function(x,...) { x$index }

##' @export
"index.lvmfit" <- function(x,...) { index(Model(x)) }

##' @rdname index-lvm
##' @param value new index
##' @aliases index<-.lvmfit index<-.lvm
##' @export
"index<-.lvm" <- function(x,...,value)  { x$index <- value; return(x) }

##' @export
"index<-.lvmfit" <- function(x,...,value) { Model(x)$index <- value; return(x) }


`reindex` <-
    function(x, sparse=FALSE,standard=TRUE,zeroones=FALSE,deriv=FALSE,mean=TRUE) { ## Extract indices of parameters from model
        x$parpos <- NULL
        M <- x$M

        eta <- latent(x) ## Latent variables/Factors
        m <- length(eta)
        obs <- manifest(x)  ## Manifest/Observed variables
        endo <- endogenous(x)
        exo <- exogenous(x) ##,index=FALSE)

        allvars <- vars(x)
        eta.idx <- na.omit(match(eta,allvars))
        obs.idx <- na.omit(match(obs,allvars))
        exo.idx <- na.omit(match(exo,allvars))
        exo.obsidx <- na.omit(match(exo,obs))
        endo.obsidx <- na.omit(match(endo,obs))

        fix.idx <- !is.na(x$fix) ## Index of fixed parameters
        covfix.idx <- !is.na(x$covfix) ## Index of fixed covariance parameters

        constrain.par <- NULL
        if (length(constrain(x))>0) constrain.par <- names(constrain(x))

        M0 <- M;  M0[fix.idx] <- 0 ## Matrix of indicators of free regression-parameters (removing fixed parameters)
        M1 <- M0; ## Matrix of indiciator of free _unique_ regression parameters (removing fixed _and_ duplicate parameters)
        parname <- unique(x$par[!is.na(x$par)])
        for (p in parname) {
            ii <- which(x$par==p)
            if (length(ii)>1)
                M1[ii[-1]] <- 0
            if (p %in% constrain.par)
                M0[ii] <- M1[ii] <- 0
        }
        npar.reg <- sum(M1) ## Number of free regression parameters

        P <- x$cov;

        P0 <- P;  P0[covfix.idx] <- 0 ## Matrix of indicators of free covariance-parameters (removing fixed parameters)
        if (length(exo.idx)>0)
            P0[exo.idx,exo.idx] <- 0 ## 6/1-2011
        P1 <- P0 ## Matrix of indiciator of free _unique_ variance parameters (removing fixed _and_ duplicate parameters)
        covparname <- unique(x$covpar[!is.na(x$covpar)])
        for (p in covparname) {
            ii <- which(x$covpar==p)
            if (length(ii)>1)
                P1[ii[-1]] <- 0
            if (p%in%c(parname,constrain.par))
                P0[ii] <- P1[ii] <- 0
        }

        npar.var <- sum(c(diag(P1),P1[lower.tri(P1)]))

        A <- M
        A[fix.idx] <- x$fix[fix.idx] ## ... with fixed parameters in plac
        P[covfix.idx] <- x$covfix[covfix.idx] ## ... with fixed parameters in plac

        px <- Jy <- J <- diag(nrow=length(vars(x)))
        if (m>0) {
            J[eta.idx,eta.idx] <- 0; J <- J[-eta.idx,,drop=FALSE]
        } ## Selection matrix (selecting observed variables)
        {
            ## Selection matrix (selection endogenous variables)
            if (length(c(eta.idx,exo.idx))>0) {
                Jy[c(eta.idx,exo.idx),c(eta.idx,exo.idx)] <- 0; Jy <- Jy[-c(eta.idx,exo.idx),,drop=FALSE]
            }
            ## Cancelation matrix (cancels rows with exogenous variables)
            px[exo.idx,exo.idx] <- 0
        }

        ## Creating indicitor of free mean-parameters
        fixed <- sapply(x$mean, function(y) is.numeric(y) & !is.na(y))
        named <- sapply(x$mean, function(y) is.character(y) & !is.na(y))
        mparname <- NULL
        if (length(named)>0)
            mparname <- unlist(unique(x$mean[named]))
        v0 <- rep(1,length(x$mean)) ## Vector of indicators of free mean-parameters

        v0[exo.idx] <- 0
        if (length(fixed)>0) v0[fixed] <- 0;
        v1 <- v0
        for (p in mparname) {
            idx <- which(x$mean==p)
            if (length(idx)>1) {
                v1[idx[-1]] <- 0
            }
            if (p%in%c(parname,covparname,constrain.par))
                v0[idx] <- v1[idx] <- 0
        } ## duplicate parameters

        ###
        ### Extra parameters
        ###
        efixed <- sapply(x$exfix, function(y) is.numeric(y) & !is.na(y))
        enamed <- sapply(x$exfix, function(y) is.character(y) & !is.na(y))
        if(length(enamed)>0){
            eparname <- unlist(unique(x$exfix[enamed]))
        } else{
            eparname <- NULL
        }
        ## Extra parameters
        e0 <- rep(1,length(x$expar)) ## Indicators of free extra par.
        if (length(efixed)>0)
            e0[efixed] <- 0
        e1 <- e0
        for (p in eparname) {
            idx <- which(x$exfix==p)
            if (length(idx)>1) {
                e1[idx[-1]] <- 0
            }
            if (p%in%c(parname,covparname,constrain.par,mparname))
                e0[idx] <- e1[idx] <- 0
        } ## duplicate parameters


        ## Return:
        ## Adjacency-matrix (M)
        ## Matrix of regression-parameters (0,1) _with_ fixed parameters (A)
        ## Matrix of variance-parameters (indicators 0,1) (P)
        ## Manifest selection matrix (J),
        ## Position of variables matrix (Apos),
        ## Position of covariance variables matrix (Ppos),
        ## Position/Indicator matrix of free regression parameters (M0)
        res <- list(vars=allvars, manifest=obs, exogenous=exo, latent=eta,
                    endogenous=endo,
                    exo.idx=exo.idx, eta.idx=eta.idx,
                    exo.obsidx=exo.obsidx, endo.obsidx=endo.obsidx,
                    obs.idx=obs.idx,
                    endo.idx=setdiff(obs.idx,exo.idx))

        if (standard) {
            res <- c(res, list(M=M, A=A, P=P,
                               P0=P0, P1=P1,
                               M0=M0, M1=M1,
                               v0=v0, v1=v1,
                               e0=e0, e1=e1,
                               npar=(npar.reg+npar.var),
                               npar.reg=npar.reg,
                               npar.var=npar.var,
                               npar.mean=sum(v1),
                               npar.ex=sum(e1),
                               constrain.par=constrain.par))
            which.diag <- NULL
            if (length(P1)>0)
                which.diag <- which(diag(P1==1))

            res <- c(res, list(parname.all=parname, parname=setdiff(parname,constrain.par),
                               which.diag=which.diag,
                               covparname.all=covparname,
                               covparname=setdiff(covparname,constrain.par),
                               meanfixed=fixed, meannamed=named,
                               mparname.all=mparname,
                               mparname=setdiff(mparname,constrain.par),
                               eparname.all=eparname,
                               eparname=setdiff(eparname,constrain.par),
                               J=J, Jy=Jy, px=px, sparse=sparse))

            parname.all.reg.idx <- parname.all.reg.tidx <-
                parname.reg.tidx <- parname.reg.idx <- c()
            for (p in res$parname.all) {
                ipos <- which((x$par==p))
                tipos <- which(t(x$par==p))
                if (p%in%res$parname) {
                    parname.reg.idx <- c(parname.reg.idx, list(ipos))
                    parname.reg.tidx <- c(parname.reg.tidx, list(tipos))
                }
                parname.all.reg.idx <- c(parname.all.reg.idx, list(ipos))
                parname.all.reg.tidx <- c(parname.all.reg.tidx, list(tipos))
            };
            if (length(parname.reg.idx)>0) {
                names(parname.reg.idx) <- names(parname.reg.tidx) <- res$parname
            }
            if (length(parname.all.reg.idx)>0) {
                names(parname.all.reg.idx) <- names(parname.all.reg.tidx) <- res$parname.all
            }
            covparname.all.idx <- covparname.idx <- c()
            for (p in res$covparname.all) {
                ipos <- which(x$covpar==p)
                if (p%in%res$covparname)
                    covparname.idx <- c(covparname.idx, list(ipos))
                covparname.all.idx <- c(covparname.all.idx, list(ipos))
            };
            if (length(covparname.idx)>0)
                names(covparname.idx) <- res$covparname
            if (length(covparname.all.idx)>0)
                names(covparname.all.idx) <- res$covparname.all

            mparname.all.idx <- mparname.idx <- c()
            for (p in res$mparname.all) {
                ipos <- which(x$mean==p)
                if (p%in%mparname)
                    mparname.idx <- c(mparname.idx, list(ipos))
                mparname.all.idx <- c(mparname.all.idx, list(ipos))
            };
            if (length(mparname.idx)>0)
                names(mparname.idx) <- res$mparname
            if (length(mparname.all.idx)>0)
                names(mparname.all.idx) <- res$mparname.all

            eparname.all.idx <- eparname.idx <- c()
            for (p in res$eparname.all) {
                ipos <- which(x$exfix==p)
                if (p%in%eparname)
                    eparname.idx <- c(eparname.idx, list(ipos))
                eparname.all.idx <- c(eparname.all.idx, list(ipos))
            };
            if (length(eparname.idx)>0)
                names(eparname.idx) <- res$eparname
            if (length(eparname.all.idx)>0)
                names(eparname.all.idx) <- res$eparname.all


            res <- c(res, list(mparname.idx=mparname.idx,
                               covparname.idx=covparname.idx,
                               parname.reg.idx=parname.reg.idx,
                               parname.reg.tidx=parname.reg.tidx,
                               mparname.all.idx=mparname.all.idx,
                               eparname.all.idx=eparname.all.idx,
                               covparname.all.idx=covparname.all.idx,
                               parname.all.reg.idx=parname.all.reg.idx,
                               parname.all.reg.tidx=parname.all.reg.tidx
                               ))

        } else {
            res <- index(x)
        }

        if (zeroones) {
            if (sparse) {
                if (!requireNamespace("Matrix",quietly=TRUE)) stop("package Matrix not available")
                Ik <- Matrix::Diagonal(length(obs))
                Im <- Matrix::Diagonal(ncol(A))
                Kkk <- NULL
                J <- as(J, "sparseMatrix")
                Jy <- as(Jy, "sparseMatrix")
                px <- as(px, "sparseMatrix")

            } else {
                Ik <- diag(nrow=length(obs))
                Im <- diag(nrow=ncol(A))
            }
            Kkk <- NULL


            res[c("Ik","Im","Kkk")] <- NULL
            res <- c(res, list(Ik=Ik, Im=Im, Kkk=Kkk))
        }
        if (deriv && length(P)>0) {
            if (res$npar.mean>0 & mean)
                D <- deriv.lvm(x,meanpar=rep(1,res$npar.mean),zeroones=TRUE)
            else
                D <- deriv.lvm(x,meanpar=NULL,zeroones=TRUE)
            res[c("dA","dP","dv")] <- NULL
            res <- c(res, list(dA=D$dA, dP=D$dP, dv=D$dv))
        }

        if (length(P)>0)
            res <- c(res,mat.lvm(x,res))

        return(res)
    }

context("estimate.default and estimate_calculus issues")

library(testthat)
library(lava)


test_that("estimate.array mean formula is correct", {
  ## The variance computation in estimate.array currently evaluates
  ## mean(y - mean(y)^2) instead of mean((y - mean(y))^2).
  set.seed(1)
  n <- 100
  x <- matrix(rnorm(2*n, mean = 3, sd = 2), ncol = 2)
  e <- estimate(x)
  expected <- apply(x, 2, function(y) mean(y))
  expect_equivalent(coef(e), expected, tolerance = 1e-8)
  x0 <- apply(x, 2, function(y) y-mean(y))
  var_expected <- crossprod(x0) / n^2
  expect_equivalent(vcov(e), var_expected, tolerance = 1e-8)
})

test_that("estimate.array variance formula is correct", {
  ## The variance computation in estimate.array currently evaluates
  ## mean(y - mean(y)^2) instead of mean((y - mean(y))^2).
  set.seed(1)
  x <- matrix(rnorm(200, mean = 3, sd = 2), ncol = 2)
  e <- estimate(x, type = "variance")
  expected <- apply(x, 2, function(y) mean((y - mean(y))^2))
  expect_equivalent(coef(e), expected, tolerance = 1e-8)
})

test_that("estimate.array variance handles NA without poisoning", {
  ## Inner mean(y) inside the variance formula does not pass na.rm=TRUE.
  set.seed(2)
  x <- matrix(rnorm(100), ncol = 2)
  x[1, 1] <- NA
  expect_no_error(e <- estimate(x, type = "variance"))
  expect_true(all(is.finite(coef(e))))
})

test_that("estimate.array quantile", {
  set.seed(1)
  y <- cbind(rnorm(100))
  q <- 0.75
  est <- quantile(y, type = 1L, probs=q)
  dens <- density(y)
  dens.est <- approxfun(dens)(est)
  infl <- (q - (y <= est)) / dens.est
  e1 <- estimate(coef = est, IC = infl)
  e2 <- estimate(y, type="quantile1", probs=q)
  # check manual vs quantile1 calc.
  expect_equivalent(IC(e1), IC(e2))
  expect_equivalent(coef(e1), coef(e2))
  # default for stats::quantile is type=7
  e3 <- estimate(y, type="quantile7", probs=q)
  expect_true(coef(e1) != coef(e3))
  expect_equivalent(coef(e3), quantile(y, probs=q))
  # check that multiple quantiles works
  qq <- c(0.25, 0.5, 0.75)
  e4 <- estimate(y, type="quantile7", probs=qq)
  expect_equivalent(coef(e4), quantile(y, qq))
})

test_that("estimate.array id", {
  set.seed(1)
  y <- cbind(rnorm(10))
  id <- 1:10
  e <- estimate(y, id=id, type="mean")
  expect_equivalent(index(e), id)
  # check correct order
  expect_equivalent(IC(e), y-mean(y))
  # string id
  ids <- rev(paste0("i", id))
  e <- estimate(y, id=ids)
  expect_equivalent(index(e), ids)
  expect_equivalent(IC(e), y-mean(y))
  # rownames id
  ys <- y
  rownames(ys) <- ids
  e <- estimate(ys)
  expect_equivalent(index(e), ids)
  expect_equivalent(IC(e), ys-mean(ys))
})

##' @export
`score` <-
function(x,...) UseMethod("score")

###{{{ score.lvm

##' @export
score.lvm <- function(x, data, p, model="gaussian", S, n, mu=NULL, weights=NULL, data2=NULL, debug=FALSE, reindex=FALSE, mean=TRUE, constrain=TRUE, indiv=TRUE,...) {

  cl <- match.call()
  lname <- paste0(model,"_score.lvm")
  if (!exists(lname)) {
    lname <- paste0(model,"_gradient.lvm")
    mygrad <- get(lname)
    scoreFun <- function(...) -mygrad(...)
    if (is.null(mygrad)) {
      stop("Missing gradient")
    }
  } else {
    scoreFun <- get(lname)
  }

  if (missing(data) || is.null(data)) {
    cl[[1]] <- scoreFun
    score <- eval.parent(cl)
    return(rbind(score))
  }

  if (is.null(index(x)$dA) | reindex)
    x <- updatelvm(x,zeroones=TRUE,deriv=TRUE)

  xfix <- colnames(data)[(colnames(data)%in%parlabels(x,exo=TRUE))]
  xconstrain <- intersect(unlist(lapply(constrain(x),function(z) attributes(z)$args)),index(x)$manifest)

  Debug(xfix,debug)
  if (missing(n)) {
    n <- nrow(data)
  }

  if (length(xfix)>0 | length(xconstrain)>0) { ##### Random slopes!
    x0 <- x
    if (length(xfix)>0) {
      Debug("random slopes...",debug)
      nrow <- length(vars(x))
      xpos <- lapply(xfix,function(y) which(regfix(x)$labels==y))
      colpos <- lapply(xpos, function(y) ceiling(y/nrow))
      rowpos <- lapply(xpos, function(y) (y-1)%%nrow+1)
      myfix <- list(var=xfix, col=colpos, row=rowpos)
      for (i in seq_along(myfix$var))
        for (j in seq_along(myfix$col[[i]])) {
          regfix(x0, from=vars(x0)[myfix$row[[i]][j]],to=vars(x0)[myfix$col[[i]][j]]) <-
            data[1,myfix$var[[i]]]
        }
      index(x0) <- reindex(x0,zeroones=TRUE,deriv=TRUE)
    }
    pp <- modelPar(x0,p)
    ##p0 <- with(pp, c(meanpar,p,p2))

    myfun <- function(ii) {
      if (length(xfix)>0)
        for (i in seq_along(myfix$var)) {
          index(x0)$A[cbind(myfix$row[[i]],myfix$col[[i]])] <- data[ii,myfix$var[[i]]]
        }
      return(scoreFun(x0,data=data[ii,], p=with(pp,c(meanpar,p,p2)),weights=weights[ii,,drop=FALSE],data2=data2[ii,,drop=FALSE],model=model,debug=debug,indiv=indiv,...))
    }
    score <- t(sapply(seq_len(nrow(data)),myfun))
    if (!indiv) {
      score <- colSums(rbind(score))
    }
    if (length(score)<length(p))  score <- c(score,rep(0,length(p)-length(score)))
    return(score)
  }
  cl$constrain <- FALSE
  cl[[1]] <- scoreFun
  score <- eval.parent(cl)
  if (is.null(dim(score))) score <- rbind(score)
  if (NCOL(score)<length(p))  score <- cbind(rbind(score),rep(0,length(p)-NCOL(score)))

#  score <- eval(cl,parent.frame())
  return(score)
}

###}}} score.lvm

###{{{ score.lvm.missing

##' @export
score.lvm.missing <- function(x,
                       p=pars(x), estimator=x$estimator,
                       weights=Weights(x$estimate),
                       data2=x$estimate$data2,
                       indiv=FALSE,
                       list=FALSE,
                       ...) {
    dots <- list(...)
    dots$combine <- FALSE
    S <- do.call("score",c(list(x=x$estimate$model0,p=p, model=estimator, weights=weights, data2=data2, indiv=indiv),dots))
    if (indiv & !list) {
        S0 <- matrix(ncol=length(p),nrow=length(x$order))
        rownames(S0) <- seq_len(nrow(S0))
        myorder <- x$orderlist
        if (length(x$allmis)>0)
            myorder[[x$allmis]] <- NULL
        for (i in seq_along(S))
            S0[myorder[[i]],] <- S[[i]]
        if (length(x$allmis)>0) {
            S0 <- S0[-x$orderlist[[x$allmis]],]
        }
        S0[is.na(S0)] <- 0
        colnames(S0) <- names(coef(x))
        return(S0)
    }
    return(S)
}

###}}} score.lvm.missing

###{{{ score.multigroupfit

##' @export
score.multigroupfit <- function(x,p=pars(x), weights=Weights(x), estimator=x$estimator, data2=x$data2, ...) {
  score(x$model0, p=p, weights=weights, data2=data2, model=estimator,...)
}

###}}} score.multigroupfit

###{{{ score.multigroup

##' @export
score.multigroup <- function(x,data=x$data,weights=NULL,data2=NULL,p,indiv=combine,combine=FALSE,...) {
  rm <- procrandomslope(x)
  pp <- with(rm, modelPar(model,p)$p)
  parord <- modelPar(rm$model,seq_len(with(rm$model,npar+npar.mean)))$p
  S <- list()
  for (i in seq_len(x$ngroup)) {
    S0 <- rbind(score(x$lvm[[i]],p=pp[[i]],data=data[[i]],weights=weights[[i]],data2=data2[[i]],indiv=indiv,...))
    S1 <- matrix(ncol=length(p),nrow=nrow(S0))
    S1[,parord[[i]]] <- S0
    S <- c(S, list(S1))
  }
  if (combine) {
      S <- Reduce("rbind",S); S[is.na(S)] <- 0
      if (!indiv) S <- colSums(S)
      return(S)
  }
  if (indiv) return(S)
  res <- matrix(0,nrow=1,ncol=length(p))
  for (i in seq_len(x$ngroup))
    res[,parord[[i]]] <- res[,parord[[i]]]  + S[[i]][,parord[[i]]]
  return(as.vector(res))
}

###}}} score.multigroup

###{{{ score.lvmfit

##' @export
score.lvmfit <- function(x, data=model.frame(x), p=pars(x), model=x$estimator, weights=Weights(x), data2=x$data$data2, ...) {
  res <- score(x$model0, data = data, p = p, model = model, weights = weights, data2 = data2, ...)
  tryCatch(rownames(res) <- rownames(data), error = function(...) NULL)
  res
}

###}}} score.lvmfit

##' Plot method for 'estimate' objects
##'
##' Plot method for 'estimate' objects
##' @export
##' @param x estimate object
##' @param f function of parameter coefficients and data parsed on to
##'   'estimate'. If omitted a forest-plot will be produced.
##' @param idx Index of parameters (default all)
##' @param intercept include intercept in forest-plot
##' @param data data.frame
##' @param confint Add confidence limits
##' @param type plot type ('l')
##' @param xlab x-axis label
##' @param ylab y-axis label
##' @param col color
##' @param add add plot to current device
##' @param null null value for forest-plot
##' @param ... additional arguments to lower-level functions
plot.estimate <- function(x, f, idx,
                          intercept=FALSE, data,
                          confint=TRUE,
                          type="l",
                          xlab="x", ylab="f(x)",
                          col=1, add=FALSE, null=0, ...) {
  if (!missing(f) && !is.null(f)) {
    data <- as.list(data)
    env <- new.env()
    for (y in names(data)) {
      assign(y,data[[y]],env)
    }
    environment(f) <- env
    pp <- estimate(x, f, ..., vcov=vcov(x), IC=FALSE)$coefmat
    if (!add) suppressWarnings(plot(data[[1]],pp[,1],xlab=xlab,ylab=ylab,type=type,...))
    else lines(data[[1]],pp[,1],xlab=xlab,ylab=ylab,type=type,col=col,...)
    if (confint) confband(data[[1]],pp[,3],pp[,4],polygon=TRUE,col=Col(col),lty=0)
    return(invisible(pp))
  }
  if (!is.null(x$coefmat)) {
    pp <- x$coefmat[,c(1,3,4),drop=FALSE]
  } else {
    pp <- cbind(coef(x),confint(x))
  }
  if (!missing(idx)) pp <- pp[idx,,drop=FALSE]
  if (!intercept) {
    idx <- match("(Intercept)",rownames(pp))
    if (length(idx)>0 && !any(is.na(idx))) pp <- pp[-idx,,drop=FALSE]
  }
  forestplot(pp[rev(seq(nrow(pp))),,drop=FALSE], line=null, ...)
}

##' @export
`score` <-
function(x,...) UseMethod("score")

##' @export
score.lvm <- function(x, data, p, model="gaussian", S, n, mu=NULL, weights=NULL, data2=NULL, debug=FALSE, reindex=FALSE, mean=TRUE, constrain=TRUE, indiv=TRUE,...) {

  cl <- match.call()
  lname <- paste0(model,"_score.lvm")
  if (!exists(lname)) {
    lname <- paste0(model,"_gradient.lvm")
    mygrad <- get(lname)
    scoreFun <- function(...) -mygrad(...)
    if (is.null(mygrad)) {
      stop("Missing gradient")
    }
  } else {
    scoreFun <- get(lname)
  }

  if (missing(data) || is.null(data)) {
    cl[[1]] <- scoreFun
    score <- eval.parent(cl)
    return(rbind(score))
  }

  if (is.null(index(x)$dA) | reindex)
    x <- updatelvm(x,zeroones=TRUE,deriv=TRUE)

  xfix <- colnames(data)[(colnames(data)%in%parlabels(x,exo=TRUE))]
  xconstrain <- intersect(unlist(lapply(constrain(x),function(z) attributes(z)$args)),index(x)$manifest)

  Debug(xfix,debug)
  if (missing(n)) {
    n <- nrow(data)
  }

  if (length(xfix)>0 | length(xconstrain)>0) { ## Random slopes!
    x0 <- x
    if (length(xfix)>0) {
      Debug("random slopes...",debug)
      nrow <- length(vars(x))
      xpos <- lapply(xfix,function(y) which(regfix(x)$labels==y))
      colpos <- lapply(xpos, function(y) ceiling(y/nrow))
      rowpos <- lapply(xpos, function(y) (y-1)%%nrow+1)
      myfix <- list(var=xfix, col=colpos, row=rowpos)
      for (i in seq_along(myfix$var))
        for (j in seq_along(myfix$col[[i]])) {
          regfix(x0, from=vars(x0)[myfix$row[[i]][j]],to=vars(x0)[myfix$col[[i]][j]]) <-
            data[1,myfix$var[[i]]]
        }
      index(x0) <- reindex(x0,zeroones=TRUE,deriv=TRUE)
    }
    pp <- modelPar(x0,p)
    ##p0 <- with(pp, c(meanpar,p,p2))

    myfun <- function(ii) {
      if (length(xfix)>0)
        for (i in seq_along(myfix$var)) {
          index(x0)$A[cbind(myfix$row[[i]],myfix$col[[i]])] <- data[ii,myfix$var[[i]]]
        }
      return(scoreFun(x0,data=data[ii,], p=with(pp,c(meanpar,p,p2)),weights=weights[ii,,drop=FALSE],data2=data2[ii,,drop=FALSE],model=model,debug=debug,indiv=indiv,...))
    }
    score <- t(sapply(seq_len(nrow(data)),myfun))
    if (!indiv) {
      score <- colSums(rbind(score))
    }
    if (length(score)<length(p))  score <- c(score,rep(0,length(p)-length(score)))
    return(score)
  }
  cl$constrain <- FALSE
  cl[[1]] <- scoreFun
  score <- eval.parent(cl)
  if (is.null(dim(score))) score <- rbind(score)
  if (NCOL(score)<length(p))  score <- cbind(rbind(score),rep(0,length(p)-NCOL(score)))

#  score <- eval(cl,parent.frame())
  return(score)
}

##' @export
score.lvm.missing <- function(x,
                       p=pars(x), estimator=x$estimator,
                       weights=Weights(x$estimate),
                       data2=x$estimate$data2,
                       indiv=FALSE,
                       list=FALSE,
                       ...) {
    dots <- list(...)
    dots$combine <- FALSE
    S <- do.call("score",c(list(x=x$estimate$model0,p=p, model=estimator, weights=weights, data2=data2, indiv=indiv),dots))
    if (indiv & !list) {
        S0 <- matrix(ncol=length(p),nrow=length(x$order))
        rownames(S0) <- seq_len(nrow(S0))
        myorder <- x$orderlist
        if (length(x$allmis)>0)
            myorder[[x$allmis]] <- NULL
        for (i in seq_along(S))
            S0[myorder[[i]],] <- S[[i]]
        if (length(x$allmis)>0) {
            S0 <- S0[-x$orderlist[[x$allmis]],]
        }
        S0[is.na(S0)] <- 0
        colnames(S0) <- names(coef(x))
        return(S0)
    }
    return(S)
}

##' @export
score.multigroupfit <- function(x,p=pars(x), weights=Weights(x), estimator=x$estimator, data2=x$data2, ...) {
  score(x$model0, p=p, weights=weights, data2=data2, model=estimator,...)
}

##' @export
score.multigroup <- function(x,data=x$data,weights=NULL,data2=NULL,p,indiv=combine,combine=FALSE,...) {
  rm <- procrandomslope(x)
  pp <- with(rm, modelPar(model,p)$p)
  parord <- modelPar(rm$model,seq_len(with(rm$model,npar+npar.mean)))$p
  S <- list()
  for (i in seq_len(x$ngroup)) {
    S0 <- rbind(score(x$lvm[[i]],p=pp[[i]],data=data[[i]],weights=weights[[i]],data2=data2[[i]],indiv=indiv,...))
    S1 <- matrix(ncol=length(p),nrow=nrow(S0))
    S1[,parord[[i]]] <- S0
    S <- c(S, list(S1))
  }
  if (combine) {
      S <- Reduce("rbind",S); S[is.na(S)] <- 0
      if (!indiv) S <- colSums(S)
      return(S)
  }
  if (indiv) return(S)
  res <- matrix(0,nrow=1,ncol=length(p))
  for (i in seq_len(x$ngroup))
    res[,parord[[i]]] <- res[,parord[[i]]]  + S[[i]][,parord[[i]]]
  return(as.vector(res))
}

##' @export
score.lvmfit <- function(x, data=model.frame(x), p=pars(x), model=x$estimator, weights=Weights(x), data2=x$data$data2, ...) {
  res <- score(x$model0, data = data, p = p, model = model, weights = weights, data2 = data2, ...)
  tryCatch(rownames(res) <- rownames(data), error = function(...) NULL)
  res
}

##' @export
`summary.lvm` <-
function(object,...) {
  k <- length(vars(object))
  ## cat("Latent Variable Model \n\twith: ", k, " variables.\n", sep="");
  print(object,print.transform=FALSE,...)
  if (length(transform(object))>0) {
      cat("\nTransformations:\n")
      print(transform(object),quote=FALSE,...)
  }
  cat("\n")
  if (length(index(object))>0)
      cat("Number of free parameters: ", with(index(object),npar+npar.mean+npar.ex),"\n", sep="")

  if (k==0)
    return()
  ##cat("Npar=", index(object)$npar, "+", index(object)$npar.mean, "\n", sep="")
  cat("\n")
  print(regression(object),...)
  print(covariance(object),...)
  print(intercept(object),...)
  if (length(object$exfix)>0) {
    cat("Additional parameters:\n")
    val <- unlist(object$exfix)
    M <- rbind(val); colnames(M) <- names(val)
    rownames(M) <- "   "
    print(M,quote=FALSE,...)
  }
  if (length(constrain(object))>0) {
    cat("Non-linear constraints:\n")
    print(constrain(object),quote=FALSE,...)
  }
  ## printmany(object$cov, printmany(object$covpar, object$covfix, name1="Labels:", name2="Fixed:", print=FALSE), name1="covariance:")
  cat("\n")
}

##' @export
`summary.lvmfit` <-
function(object,std="xy", type=9, labels=2, ...) {
  cc <- CoefMat(object,labels=labels,std=std,type=type,...)
  mycoef <- coef(object,type=9)
  nlincon <- attributes(mycoef)$nlincon
  nonexo <- setdiff(vars(object),index(Model(object))$exogenous)
  attributes(mycoef) <- attributes(mycoef)[1:2]
  mygof <- object$opt$summary.message
  if (is.null(mygof)) {
    mygof <- gof
  }
  if (class(object)[1]=="lvm.missing") {
    nn <- unlist(lapply(object$multigroup$data, nrow))
    nc <- nn[object$cc]
    if (length(nc)==0) nc <- 0
    ngroup <- object$multigroup$ngroup
    res <- list(object=object, coef=mycoef, coefmat=cc, nlincon=nlincon, gof=mygof(object), n=sum(nn), nc=nc, ngroup=ngroup,
                varmat=modelVar(object)$P[nonexo,nonexo], latent=latent(object), opt=object$opt, vcov=vcov(object), estimator=object$estimator, rsq=rsq(object))
  } else {
    n <- nrow(model.frame(object))
    if (is.null(n)) n <- model.frame(object)$n
    res <- list(coef=mycoef, coefmat=cc, nlincon=nlincon, gof=mygof(object), n=n, nc=n, latent=latent(object),
                opt=object$opt, vcov=vcov(object), estimator=object$estimator, rsq=rsq(object))##, varmat=modelVar(object)$P[nonexo,nonexo])
  }
  class(res) <- "summary.lvmfit"
  res
}

##' @export
print.summary.lvmfit <- function(x,varmat=TRUE,...) {
  if (!is.null(x$control$method)) {
    l2D <- sum(x$opt$grad^2)
    rnkV <- qr(x$vcov)$rank
    if (l2D>1e-2) warning("Possible problems with convergence!")
    cat("||score||^2=",l2D,"\n",sep="")
    np <- nrow(x$vcov)
    if (rnkV<np) warning("Possible problems with identification (rank(informaion)=",rnkV,"<",np,"!")
  }
  cat("Latent variables:", x$latent, "\n")
  cat("Number of rows in data=",x$n,sep="")
  if (x$nc!=x$n) {
    cat(" (",x$nc," complete cases, ", x$ngroup, " groups)",sep="")
  }; cat("\n")
  print(cli::rule())
  print(x$coefmat,quote=FALSE,right=TRUE)
##  if (varmat) {
##    cat("\nResidual covariance matrix:\n")
##    print(x$varmat)
##  }
  if (!is.null(x$nlincon)) {
    cat("\nNon-linear constraints:\n")
    printCoefmat(x$nlincon,signif.stars=FALSE)
  }
  print(cli::rule())
  cat("Estimator:",x$estimator,"\n")
  print(cli::rule())
  if (!is.null(x$gof)) {
    if (class(x$gof)[1]=="list") {
      for (i in x$gof) {
        print(i)
      }
    } else {
      print(x$gof,optim=FALSE)
    }
    print(cli::rule())
  }
  if (!is.null(x$rsq)) {
      if (!is.list(x$rsq)) {
          cat("R-square\n")
          print(round(x$rsq,3),quote=FALSE)
      } else {
          for (i in seq_len(length(x$rsq))) {
              cat(names(x$rsq)[i],"\n")
              print(round(x$rsq[[i]],3),quote=FALSE)
          }
      }
  }
  invisible(x)
}

##' @export
coef.summary.lvmfit <- function(object,...) object$coef

##' @export
summary.multigroupfit <- function(object,groups=NULL,...) {
  if (is.null(groups) | length(groups)==0) {
    if (object$model$missing) {
      groups <- object$model$complete
      if (length(groups)==0)
        groups <- seq_len(object$model0$ngroup)
    } else {
      groups <- seq_len(object$model$ngroup)
    }
  }
  cc <- CoefMat.multigroupfit(object,groups=groups,...)
  res <- list(coef=coef(object,type=2,groups=groups,...), object=object, coefmat=cc, gof=gof(object), object=object, opt=object$opt, latent=object$latent, estimator=object$estimator)
  class(res) <- "summary.multigroupfit"
  res
}

##' @export
print.summary.multigroupfit <- function(x,...) {
  l2D <- sum(x$opt$grad^2)
  if (l2D>1e-2) warning("Possible problems with convergence!")
  cat("||score||^2=",l2D,"\n")
  cat("Latent variables:", x$latent, "\n")
  print(x$object,...)
  ##print(x$coefmat,quote=FALSE,right=TRUE)
  print(cli::rule())
  if (!is.null(attributes(x$coefmat)$nlincon)) {
    cat("Non-linear constraints:\n")
    print(attributes(x$coefmat)$nlincon)
    print(cli::rule())
  }
  cat("Estimator:",x$estimator,"\n")
  print(cli::rule())
  if (!is.null(x$gof)) {
    print(x$gof)
    print(cli::rule())
  }
  invisible(x)
}

##' @export
summary.multigroup <- function(object,...) {
  for (m in object$lvm)
    print(m, ...)
  print(object)
  invisible(object)
}

###{{{ labels

##' Define labels of graph
##'
##' Alters labels of nodes and edges in the graph of a latent variable model
##'
##'
##' @aliases labels<- labels labels<-.default labels.lvm labels.lvmfit
##' labels.graphNEL edgelabels edgelabels<- edgelabels<-.lvm nodecolor
##' nodecolor<- nodecolor<-.default
##' @author Klaus K. Holst
##' @export
##' @keywords graphs aplot
##' @examples
##' m <- lvm(c(y,v)~x+z)
##' regression(m) <- c(v,x)~z
##' labels(m) <- c(y=expression(psi), z=expression(zeta))
##' nodecolor(m,~y+z+x,border=c("white","white","black"),
##'           labcol="white", lwd=c(1,1,5),
##'           lty=c(1,2)) <-  c("orange","indianred","lightgreen")
##' edgelabels(m,y~z+x, cex=c(2,1.5), col=c("orange","black"),labcol="darkblue",
##'            arrowhead=c("tee","dot"),
##'            lwd=c(3,1)) <- expression(phi,rho)
##' edgelabels(m,c(v,x)~z, labcol="red", cex=0.8,arrowhead="none") <- 2
##' if (interactive()) {
##'     plot(m,addstyle=FALSE)
##' }
##'
##' m <- lvm(y~x)
##' labels(m) <- list(x="multiple\nlines")
##' if (interactive()) {
##' op <- par(mfrow=c(1,2))
##' plot(m,plain=TRUE)
##' plot(m)
##' par(op)
##'
##' d <- sim(m,100)
##' e <- estimate(m,d)
##' plot(e,type="sd")
##' }
##' @param object \code{lvm}-object.
##' @param value node label/edge label/color
##' @param to Formula specifying outcomes and predictors defining relevant
##' edges.
##' @param \dots Additional arguments (\code{lwd}, \code{cex}, \code{col},
##' \code{labcol}), \code{border}.
##' @param var Formula or character vector specifying the nodes/variables to
##' alter.
##' @param border Colors of borders
##' @param labcol Text label colors
##' @param shape Shape of node
##' @param lwd Line width of border
##' @usage
##' \method{labels}{default}(object, ...) <- value
##' \method{edgelabels}{lvm}(object, to, ...) <- value
##' \method{nodecolor}{default}(object, var=vars(object),
##' border, labcol, shape, lwd, ...) <- value
`labels<-` <- function(object,...,value) UseMethod("labels<-")

##' @export
`labels<-.default` <- function(object,...,value) {
  labels(object,value)
}

##' @export
labels.graphNEL <- function(object,lab=NULL,...) {
  if (is.null(lab))
    return(graph::nodeRenderInfo(object)$label)
  graph::nodeRenderInfo(object) <- list(label=lab)
  names(graph::nodeRenderInfo(object)$label) <- graph::nodes(object);
  return(object)
}

##' @export
labels.lvmfit <- function(object,lab=NULL,...) {
  if (is.null(lab)) return(object$noderender$label)
  object$noderender$label <- lab
  return(object)
}

##' @export
`labels.lvm` <- function(object,lab=NULL,...) {
  if (is.null(lab))
    return(object$noderender$label)
  if (is.null(object$noderender$label))
    object$noderender$label <- lab
  else
    object$noderender$label[names(lab)] <- lab
  return(object)
}
###}}} labels

###{{{ edgelabels

##' @export
"edgelabels<-.lvmfit" <- function(object,to,from,est=TRUE,edges=NULL,cex=1,...,value) {
  if (is.null(edges))  {
    if (inherits(to,"formula")) {
      yy <- decomp.specials(getoutcome(to))
      from <- setdiff(all.vars(to),yy)
      to <- yy
    }
    edges <- paste(from,to,sep="~")
  }

  edges. <- paste0("\"", edges, "\"")
  fromto <- edge2pair(edges)
  val <- c()
  for (i in seq_along(edges)) {
    val <- c(val,
             formatC(effects(object,from=fromto[[i]][1],to=fromto[[i]][2],messages=0)$directef[[1]])
             )
  }
  if (est)
    mytext <- paste("c(", paste(paste0(edges.,"=expression(",as.character(value),"==\"",val,"\")"),collapse=","),")")
  else
    mytext <- paste("c(", paste(paste0(edges.,"=expression(",as.character(value),")"),collapse=","),")")
  graph::edgeRenderInfo(Graph(object))$label <- eval(parse(text=mytext))
  graph::edgeRenderInfo(Graph(object))$cex[edges] <- cex
  return(object)
}

##' @export
edgelabels.lvmfit <- function(object,value,type,pthres,intercept=FALSE,format.fun=formatC,...) {
    if (!missing(value)) {
        edgelabels(object,...) <- value
        return(object)
    }
    if (missing(type))
        return(graph::edgeRenderInfo(Graph(object))$label)


    Afix <- index(object)$A ## Matrix with fixed parameters and ones where parameters are free
    ##Pfix <- index(object)$P ## Matrix with fixed covariance parameters and ones where param
    ##mfix <- index(object)$v0

    npar.mean <- index(object)$npar.mean
    Par <- object$coef
    mpar <- c()
    if (npar.mean>0) {
        mpar <- do.call(format.fun,list(Par[seq_len(npar.mean)]))
        Par <- Par[-seq_len(npar.mean),,drop=FALSE]
    }
    Par <-
        switch(type,
               sd = paste0(do.call(format.fun,list(Par[,1,drop=FALSE])), " (", do.call(format.fun,list(Par[,2,drop=FALSE])), ")"),
               est = do.call(format.fun,list(Par[,1,drop=FALSE])),
               pval = do.call(format.fun,list(Par[,4,drop=FALSE])),
               name = rownames(Par),
               none = ""
               )
    AP <- matrices(Model(object), Par,mpar) ## Ignore expar
    A <- AP$A; P <- AP$P
    P[exogenous(object),exogenous(object)] <- NA

    gr <- finalize(Model(object), ...)
    Anz <- A; Anz[Afix==0] <- NA
    gr <- edgelabels(gr, lab=Anz)
    Pnz <- P; Pnz[Model(object)$cov==0] <- NA
    if (intercept) {
        idx <- which(!is.na(diag(Pnz)))
        diag(Pnz)[idx] <- paste(paste0("[",AP$v[idx],"]"),diag(Pnz)[idx],sep="\n")
    }
    gr <- edgelabels(gr, lab=Pnz, expr=!intercept)
    Graph(object) <- gr
    return(object)
}

##' @export
`edgelabels` <- function(object, ...) UseMethod("edgelabels")

##' @export
`edgelabels<-` <- function(object,...,value) UseMethod("edgelabels<-")

##' @export
`edgelabels<-.lvm` <- function(object,to,...,value) {
  edgelabels(object,to=to, lab=value,...)
}

##' @export
`edgelabels<-.graphNEL` <- function(object,...,value) {
  edgelabels(object,lab=value,...)
}

##' @export
`edgelabels.graphNEL` <- function(object, lab=NULL, to=NULL, from=NULL, cex=1.5, lwd=1, lty=1, col="black", labcol="black", arrowhead="closed",
                                  expr=TRUE,
                                  debug=FALSE,...) {
  if (is.null(lab)) {
    return(graph::edgeRenderInfo(object)$label)
  }
  if (inherits(to,"formula")) {
    yy <- decomp.specials(getoutcome(to))
    from <- all.vars(to[[3]])##setdiff(all.vars(to),yy)
    if (length(from)==0) from <- yy
    to <- yy
  }

  M <- as(object, Class="matrix")
  nodes <- graph::nodes(object)

  if (is.null(graph::edgeRenderInfo(object)$label))
      graph::edgeRenderInfo(object)$label <- expression()


  if (!is.null(lab)) {
    if (!is.null(from) & !is.null(to)) {
      estr <- paste0("\"",from,"~",to,"\"")
      estr2 <- paste0(from,"~",to)
      if (length(lab)!=length(estr2)) lab <- rep(lab,length(estr2))
      if (length(col)!=length(estr2)) col <- rep(col,length(estr2))
      if (length(cex)!=length(estr2)) cex <- rep(cex,length(estr2))
      if (length(lwd)!=length(estr2)) lwd <- rep(lwd,length(estr2))
      if (length(lty)!=length(estr2)) lty <- rep(lty,length(estr2))
      if (length(arrowhead)!=length(estr2))
          arrowhead <- rep(arrowhead,length(estr2))
      if (length(labcol)!=length(estr2))
          labcol <- rep(labcol,length(estr2))

      curedges <- names(graph::edgeRenderInfo(object)$label)
       Debug(estr,debug)

      estr2.idx <- which(estr2%in%curedges)
      newstr.idx <- setdiff(seq_along(estr2),estr2.idx)
      newstr <- estr2[newstr.idx]
      estr2 <- estr2[estr2.idx]
      if (length(estr2)>0) {
        if (!is.null(lab))
          graph::edgeRenderInfo(object)$label[estr2] <- lab[estr2.idx]
        if (!is.null(cex))
            graph::edgeRenderInfo(object)$cex[estr2] <- cex[estr2.idx]
        if (!is.null(col))
            graph::edgeRenderInfo(object)$col[estr2] <- col[estr2.idx]
        if (!is.null(lwd))
            graph::edgeRenderInfo(object)$lwd[estr2] <- lwd[estr2.idx]
        if (!is.null(lty))
            graph::edgeRenderInfo(object)$lty[estr2] <- lty[estr2.idx]
        if (!is.null(labcol))
            graph::edgeRenderInfo(object)$textCol[estr2] <- labcol[estr2.idx]
        if (!is.null(arrowhead))
            graph::edgeRenderInfo(object)$arrowhead[estr2] <- arrowhead[estr2.idx]
      }
      if (length(newstr)>0) {

          if (!is.null(lab))
              graph::edgeDataDefaults(object)$futureinfo$label[newstr] <-
                  lab[newstr.idx]
        if (!is.null(cex))
            graph::edgeDataDefaults(object)$futureinfo$cex[newstr] <-
                cex[newstr.idx]
        if (!is.null(col))
            graph::edgeDataDefaults(object)$futureinfo$col[newstr] <-
                col[newstr.idx]
        if (!is.null(lwd))
            graph::edgeDataDefaults(object)$futureinfo$lwd[newstr] <-
                lwd[newstr.idx]
        if (!is.null(lty))
            graph::edgeDataDefaults(object)$futureinfo$lty[newstr] <-
                lty[newstr.idx]
        if (!is.null(labcol))
            graph::edgeDataDefaults(object)$futureinfo$textCol[newstr] <-
                labcol[newstr.idx]
        if (!is.null(arrowhead))
            graph::edgeDataDefaults(object)$futureinfo$arrowhead[newstr] <-
                arrowhead[newstr.idx]
      }
      return(object)
    }

    ## Used by "edgelabels.lvmfit"
    for (r in seq_len(nrow(M)))
      for (s in seq_len(ncol(M))) {
        if (M[r,s]!=0 & !is.na(lab[r,s])) {
          estr <- paste0("\"",nodes[r],"~",nodes[s],"\"")
          estr2 <- paste0(nodes[r],"~",nodes[s])
          Debug(estr, debug)
          if (expr)
            st <- eval(parse(text=paste0("expression(",lab[r,s],")")))
          else
            st <- lab[r,s]
          graph::edgeRenderInfo(object)$label[estr2] <- st
        }
      }
  }

  return(object)
}



##' @export
`edgelabels.lvm` <- function(object, lab=NULL, to=NULL, from=NULL,
                             cex=1.5, lwd=1, lty=1, col="black",
                             labcol="black", arrowhead="closed",
                             expr=TRUE, debug=FALSE,...) {
    if (inherits(to,"formula")) {
        yy <- decomp.specials(getoutcome(to))
        from <- all.vars(to[[3]])
        if (length(from)==0) from <- yy
        to <- yy
    }
    if (is.null(lab)) {
        res <- c(object$edgerender$label,object$edgerender$futureinfo$label)
        if (!is.null(to) && !is.null(from)) {
            estr <- apply(Expand(from,to),1,function(x) paste0(x,collapse="~"))
            res <- res[estr]
        }
        return(res)
    }

  M <- object$M
  nodes <- colnames(M)

  if (is.null(object$edgerender$label))
    object$edgerender$label <- expression()


  if (!is.null(lab)) {
    if (!is.null(from) & !is.null(to)) {
      estr <- paste0("\"",from,"~",to,"\"")
      estr2 <- paste0(from,"~",to)
      if (length(lab)!=length(estr2)) lab <- rep(lab,length(estr2))
      if (length(col)!=length(estr2)) col <- rep(col,length(estr2))
      if (length(cex)!=length(estr2)) cex <- rep(cex,length(estr2))
      if (length(lwd)!=length(estr2)) lwd <- rep(lwd,length(estr2))
      if (length(lty)!=length(estr2)) lty <- rep(lty,length(estr2))
      if (length(labcol)!=length(estr2)) labcol <- rep(labcol,length(estr2))
      if (length(arrowhead)!=length(estr2)) arrowhead <- rep(arrowhead,length(estr2))

      curedges <- names(object$edgerender$label)
       Debug(estr,debug)

      estr2.idx <- which(estr2%in%curedges)
      newstr.idx <- setdiff(seq_along(estr2),estr2.idx)
      newstr <- estr2[newstr.idx]
      estr2 <- estr2[estr2.idx]
      if (length(estr2)>0) {
        if (!is.null(lab))
          object$edgerenderlabel[estr2] <- lab[estr2.idx]
        if (!is.null(cex))
          object$edgerender$cex[estr2] <- cex[estr2.idx]
        if (!is.null(col))
          object$edgerender$col[estr2] <- col[estr2.idx]
        if (!is.null(lwd))
          object$edgerender$lwd[estr2] <- lwd[estr2.idx]
        if (!is.null(lty))
          object$edgerender$lty[estr2] <- lty[estr2.idx]
        if (!is.null(labcol))
          object$edgerender$textCol[estr2] <- labcol[estr2.idx]
        if (!is.null(arrowhead))
          object$edgerender$arrowhead[estr2] <- arrowhead[estr2.idx]
      }
      if (length(newstr)>0) {

        if (!is.null(lab))
          object$edgerender$futureinfo$label[newstr] <-
            lab[newstr.idx]
        if (!is.null(cex))
          object$edgerender$futureinfo$cex[newstr] <-
            cex[newstr.idx]
        if (!is.null(col))
          object$edgerender$futureinfo$col[newstr] <-
            col[newstr.idx]
        if (!is.null(lwd))
          object$edgerender$futureinfo$lwd[newstr] <-
            lwd[newstr.idx]
        if (!is.null(lty))
          object$edgerender$futureinfo$lty[newstr] <-
            lty[newstr.idx]
        if (!is.null(labcol))
          object$edgerender$futureinfo$textCol[newstr] <-
            labcol[newstr.idx]
        if (!is.null(arrowhead))
          object$edgerender$futureinfo$arrowhead[newstr] <-
            arrowhead[newstr.idx]
      }
      return(object)
    }

    ## Used by "edgelabels.lvmfit"
    for (r in seq_len(nrow(M)))
      for (s in seq_len(ncol(M))) {
        if (M[r,s]!=0 & !is.na(lab[r,s])) {
          estr <- paste0("\"",nodes[r],"~",nodes[s],"\"")
          estr2 <- paste0(nodes[r],"~",nodes[s])
          Debug(estr, debug)
          if (expr)
            st <- eval(parse(text=paste0("expression(",lab[r,s],")")))
          else
            st <- lab[r,s]
          object$edgerender$label[estr2] <- st
        }
      }
  }
  return(object)
}

###}}} edgelabels

##' Define labels of graph
##'
##' Alters labels of nodes and edges in the graph of a latent variable model
##'
##'
##' @aliases labels<- labels labels<-.default labels.lvm labels.lvmfit
##' @aliases labels.graphNEL edgelabels edgelabels<- edgelabels<-.lvm nodecolor
##' @aliases nodecolor<- nodecolor<-.default
##' @author Klaus K. Holst
##' @export
##' @keywords graphs aplot
##' @examples
##' m <- lvm(c(y,v)~x+z)
##' regression(m) <- c(v,x)~z
##' labels(m) <- c(y=expression(psi), z=expression(zeta))
##' nodecolor(m,~y+z+x,border=c("white","white","black"),
##'           labcol="white", lwd=c(1,1,5),
##'           lty=c(1,2)) <-  c("orange","indianred","lightgreen")
##' edgelabels(m,y~z+x, cex=c(2,1.5), col=c("orange","black"),labcol="darkblue",
##'            arrowhead=c("tee","dot"),
##'            lwd=c(3,1)) <- expression(phi,rho)
##' edgelabels(m,c(v,x)~z, labcol="red", cex=0.8,arrowhead="none") <- 2
##' if (interactive()) {
##'     plot(m,addstyle=FALSE)
##' }
##'
##' m <- lvm(y~x)
##' labels(m) <- list(x="multiple\nlines")
##' if (interactive()) {
##' op <- par(mfrow=c(1,2))
##' plot(m,plain=TRUE)
##' plot(m)
##' par(op)
##'
##' d <- sim(m,100)
##' e <- estimate(m,d)
##' plot(e,type="sd")
##' }
##' @param object \code{lvm}-object.
##' @param value node label/edge label/color
##' @param to Formula specifying outcomes and predictors defining relevant
##' edges.
##' @param \dots Additional arguments (\code{lwd}, \code{cex}, \code{col},
##' \code{labcol}), \code{border}.
##' @param var Formula or character vector specifying the nodes/variables to
##' alter.
##' @param border Colors of borders
##' @param labcol Text label colors
##' @param shape Shape of node
##' @param lwd Line width of border
##' @usage
##' \method{labels}{default}(object, ...) <- value
##' \method{edgelabels}{lvm}(object, to, ...) <- value
##' \method{nodecolor}{default}(object, var=vars(object),
##' border, labcol, shape, lwd, ...) <- value
`labels<-` <- function(object,...,value) UseMethod("labels<-")

##' @export
`labels<-.default` <- function(object,...,value) {
  labels(object,value)
}

##' @export
labels.graphNEL <- function(object,lab=NULL,...) {
  if (is.null(lab))
    return(graph::nodeRenderInfo(object)$label)
  graph::nodeRenderInfo(object) <- list(label=lab)
  names(graph::nodeRenderInfo(object)$label) <- graph::nodes(object);
  return(object)
}

##' @export
labels.lvmfit <- function(object,lab=NULL,...) {
  if (is.null(lab)) return(object$noderender$label)
  object$noderender$label <- lab
  return(object)
}

##' @export
`labels.lvm` <- function(object,lab=NULL,...) {
  if (is.null(lab))
    return(object$noderender$label)
  if (is.null(object$noderender$label))
    object$noderender$label <- lab
  else
    object$noderender$label[names(lab)] <- lab
  return(object)
}

##' @export
"edgelabels<-.lvmfit" <- function(object,to,from,est=TRUE,edges=NULL,cex=1,...,value) {
  if (is.null(edges))  {
    if (inherits(to,"formula")) {
      yy <- decomp.specials(getoutcome(to))
      from <- setdiff(all.vars(to),yy)
      to <- yy
    }
    edges <- paste(from,to,sep="~")
  }

  edges. <- paste0("\"", edges, "\"")
  fromto <- edge2pair(edges)
  val <- c()
  for (i in seq_along(edges)) {
    val <- c(val,
             formatC(effects(object,from=fromto[[i]][1],to=fromto[[i]][2],messages=0)$directef[[1]])
             )
  }
  if (est)
    mytext <- paste("c(", paste(paste0(edges.,"=expression(",as.character(value),"==\"",val,"\")"),collapse=","),")")
  else
    mytext <- paste("c(", paste(paste0(edges.,"=expression(",as.character(value),")"),collapse=","),")")
  graph::edgeRenderInfo(Graph(object))$label <- eval(parse(text=mytext))
  graph::edgeRenderInfo(Graph(object))$cex[edges] <- cex
  return(object)
}

##' @export
edgelabels.lvmfit <- function(object,value,type,pthres,intercept=FALSE,format.fun=formatC,...) {
    if (!missing(value)) {
        edgelabels(object,...) <- value
        return(object)
    }
    if (missing(type))
        return(graph::edgeRenderInfo(Graph(object))$label)


    Afix <- index(object)$A ## Matrix with fixed parameters and ones where parameters are free
    ##Pfix <- index(object)$P ## Matrix with fixed covariance parameters and ones where param
    ##mfix <- index(object)$v0

    npar.mean <- index(object)$npar.mean
    Par <- object$coef
    mpar <- c()
    if (npar.mean>0) {
        mpar <- do.call(format.fun,list(Par[seq_len(npar.mean)]))
        Par <- Par[-seq_len(npar.mean),,drop=FALSE]
    }
    Par <-
        switch(type,
               sd = paste0(do.call(format.fun,list(Par[,1,drop=FALSE])), " (", do.call(format.fun,list(Par[,2,drop=FALSE])), ")"),
               est = do.call(format.fun,list(Par[,1,drop=FALSE])),
               pval = do.call(format.fun,list(Par[,4,drop=FALSE])),
               name = rownames(Par),
               none = ""
               )
    AP <- matrices(Model(object), Par,mpar) ## Ignore expar
    A <- AP$A; P <- AP$P
    P[exogenous(object),exogenous(object)] <- NA

    gr <- finalize(Model(object), ...)
    Anz <- A; Anz[Afix==0] <- NA
    gr <- edgelabels(gr, lab=Anz)
    Pnz <- P; Pnz[Model(object)$cov==0] <- NA
    if (intercept) {
        idx <- which(!is.na(diag(Pnz)))
        diag(Pnz)[idx] <- paste(paste0("[",AP$v[idx],"]"),diag(Pnz)[idx],sep="\n")
    }
    gr <- edgelabels(gr, lab=Pnz, expr=!intercept)
    Graph(object) <- gr
    return(object)
}

##' @export
`edgelabels` <- function(object, ...) UseMethod("edgelabels")

##' @export
`edgelabels<-` <- function(object,...,value) UseMethod("edgelabels<-")

##' @export
`edgelabels<-.lvm` <- function(object,to,...,value) {
  edgelabels(object,to=to, lab=value,...)
}

##' @export
`edgelabels<-.graphNEL` <- function(object,...,value) {
  edgelabels(object,lab=value,...)
}

##' @export
`edgelabels.graphNEL` <- function(object, lab=NULL, to=NULL, from=NULL, cex=1.5, lwd=1, lty=1, col="black", labcol="black", arrowhead="closed",
                                  expr=TRUE,
                                  debug=FALSE,...) {
  if (is.null(lab)) {
    return(graph::edgeRenderInfo(object)$label)
  }
  if (inherits(to,"formula")) {
    yy <- decomp.specials(getoutcome(to))
    from <- all.vars(to[[3]])##setdiff(all.vars(to),yy)
    if (length(from)==0) from <- yy
    to <- yy
  }

  M <- as(object, Class="matrix")
  nodes <- graph::nodes(object)

  if (is.null(graph::edgeRenderInfo(object)$label))
      graph::edgeRenderInfo(object)$label <- expression()


  if (!is.null(lab)) {
    if (!is.null(from) & !is.null(to)) {
      estr <- paste0("\"",from,"~",to,"\"")
      estr2 <- paste0(from,"~",to)
      if (length(lab)!=length(estr2)) lab <- rep(lab,length(estr2))
      if (length(col)!=length(estr2)) col <- rep(col,length(estr2))
      if (length(cex)!=length(estr2)) cex <- rep(cex,length(estr2))
      if (length(lwd)!=length(estr2)) lwd <- rep(lwd,length(estr2))
      if (length(lty)!=length(estr2)) lty <- rep(lty,length(estr2))
      if (length(arrowhead)!=length(estr2))
          arrowhead <- rep(arrowhead,length(estr2))
      if (length(labcol)!=length(estr2))
          labcol <- rep(labcol,length(estr2))

      curedges <- names(graph::edgeRenderInfo(object)$label)
       Debug(estr,debug)

      estr2.idx <- which(estr2%in%curedges)
      newstr.idx <- setdiff(seq_along(estr2),estr2.idx)
      newstr <- estr2[newstr.idx]
      estr2 <- estr2[estr2.idx]
      if (length(estr2)>0) {
        if (!is.null(lab))
          graph::edgeRenderInfo(object)$label[estr2] <- lab[estr2.idx]
        if (!is.null(cex))
            graph::edgeRenderInfo(object)$cex[estr2] <- cex[estr2.idx]
        if (!is.null(col))
            graph::edgeRenderInfo(object)$col[estr2] <- col[estr2.idx]
        if (!is.null(lwd))
            graph::edgeRenderInfo(object)$lwd[estr2] <- lwd[estr2.idx]
        if (!is.null(lty))
            graph::edgeRenderInfo(object)$lty[estr2] <- lty[estr2.idx]
        if (!is.null(labcol))
            graph::edgeRenderInfo(object)$textCol[estr2] <- labcol[estr2.idx]
        if (!is.null(arrowhead))
            graph::edgeRenderInfo(object)$arrowhead[estr2] <- arrowhead[estr2.idx]
      }
      if (length(newstr)>0) {

          if (!is.null(lab))
              graph::edgeDataDefaults(object)$futureinfo$label[newstr] <-
                  lab[newstr.idx]
        if (!is.null(cex))
            graph::edgeDataDefaults(object)$futureinfo$cex[newstr] <-
                cex[newstr.idx]
        if (!is.null(col))
            graph::edgeDataDefaults(object)$futureinfo$col[newstr] <-
                col[newstr.idx]
        if (!is.null(lwd))
            graph::edgeDataDefaults(object)$futureinfo$lwd[newstr] <-
                lwd[newstr.idx]
        if (!is.null(lty))
            graph::edgeDataDefaults(object)$futureinfo$lty[newstr] <-
                lty[newstr.idx]
        if (!is.null(labcol))
            graph::edgeDataDefaults(object)$futureinfo$textCol[newstr] <-
                labcol[newstr.idx]
        if (!is.null(arrowhead))
            graph::edgeDataDefaults(object)$futureinfo$arrowhead[newstr] <-
                arrowhead[newstr.idx]
      }
      return(object)
    }

    ## Used by "edgelabels.lvmfit"
    for (r in seq_len(nrow(M)))
      for (s in seq_len(ncol(M))) {
        if (M[r,s]!=0 & !is.na(lab[r,s])) {
          estr <- paste0("\"",nodes[r],"~",nodes[s],"\"")
          estr2 <- paste0(nodes[r],"~",nodes[s])
          Debug(estr, debug)
          if (expr)
            st <- eval(parse(text=paste0("expression(",lab[r,s],")")))
          else
            st <- lab[r,s]
          graph::edgeRenderInfo(object)$label[estr2] <- st
        }
      }
  }

  return(object)
}

##' @export
`edgelabels.lvm` <- function(object, lab=NULL, to=NULL, from=NULL,
                             cex=1.5, lwd=1, lty=1, col="black",
                             labcol="black", arrowhead="closed",
                             expr=TRUE, debug=FALSE,...) {
    if (inherits(to,"formula")) {
        yy <- decomp.specials(getoutcome(to))
        from <- all.vars(to[[3]])
        if (length(from)==0) from <- yy
        to <- yy
    }
    if (is.null(lab)) {
        res <- c(object$edgerender$label,object$edgerender$futureinfo$label)
        if (!is.null(to) && !is.null(from)) {
            estr <- apply(Expand(from,to),1,function(x) paste0(x,collapse="~"))
            res <- res[estr]
        }
        return(res)
    }

  M <- object$M
  nodes <- colnames(M)

  if (is.null(object$edgerender$label))
    object$edgerender$label <- expression()


  if (!is.null(lab)) {
    if (!is.null(from) & !is.null(to)) {
      estr <- paste0("\"",from,"~",to,"\"")
      estr2 <- paste0(from,"~",to)
      if (length(lab)!=length(estr2)) lab <- rep(lab,length(estr2))
      if (length(col)!=length(estr2)) col <- rep(col,length(estr2))
      if (length(cex)!=length(estr2)) cex <- rep(cex,length(estr2))
      if (length(lwd)!=length(estr2)) lwd <- rep(lwd,length(estr2))
      if (length(lty)!=length(estr2)) lty <- rep(lty,length(estr2))
      if (length(labcol)!=length(estr2)) labcol <- rep(labcol,length(estr2))
      if (length(arrowhead)!=length(estr2)) arrowhead <- rep(arrowhead,length(estr2))

      curedges <- names(object$edgerender$label)
       Debug(estr,debug)

      estr2.idx <- which(estr2%in%curedges)
      newstr.idx <- setdiff(seq_along(estr2),estr2.idx)
      newstr <- estr2[newstr.idx]
      estr2 <- estr2[estr2.idx]
      if (length(estr2)>0) {
        if (!is.null(lab))
          object$edgerenderlabel[estr2] <- lab[estr2.idx]
        if (!is.null(cex))
          object$edgerender$cex[estr2] <- cex[estr2.idx]
        if (!is.null(col))
          object$edgerender$col[estr2] <- col[estr2.idx]
        if (!is.null(lwd))
          object$edgerender$lwd[estr2] <- lwd[estr2.idx]
        if (!is.null(lty))
          object$edgerender$lty[estr2] <- lty[estr2.idx]
        if (!is.null(labcol))
          object$edgerender$textCol[estr2] <- labcol[estr2.idx]
        if (!is.null(arrowhead))
          object$edgerender$arrowhead[estr2] <- arrowhead[estr2.idx]
      }
      if (length(newstr)>0) {

        if (!is.null(lab))
          object$edgerender$futureinfo$label[newstr] <-
            lab[newstr.idx]
        if (!is.null(cex))
          object$edgerender$futureinfo$cex[newstr] <-
            cex[newstr.idx]
        if (!is.null(col))
          object$edgerender$futureinfo$col[newstr] <-
            col[newstr.idx]
        if (!is.null(lwd))
          object$edgerender$futureinfo$lwd[newstr] <-
            lwd[newstr.idx]
        if (!is.null(lty))
          object$edgerender$futureinfo$lty[newstr] <-
            lty[newstr.idx]
        if (!is.null(labcol))
          object$edgerender$futureinfo$textCol[newstr] <-
            labcol[newstr.idx]
        if (!is.null(arrowhead))
          object$edgerender$futureinfo$arrowhead[newstr] <-
            arrowhead[newstr.idx]
      }
      return(object)
    }

    ## Used by "edgelabels.lvmfit"
    for (r in seq_len(nrow(M)))
      for (s in seq_len(ncol(M))) {
        if (M[r,s]!=0 & !is.na(lab[r,s])) {
          estr <- paste0("\"",nodes[r],"~",nodes[s],"\"")
          estr2 <- paste0(nodes[r],"~",nodes[s])
          Debug(estr, debug)
          if (expr)
            st <- eval(parse(text=paste0("expression(",lab[r,s],")")))
          else
            st <- lab[r,s]
          object$edgerender$label[estr2] <- st
        }
      }
  }
  return(object)
}

##' @export
`exogenous` <-
    function(x,...) UseMethod("exogenous")

##' @export
"exogenous<-" <- function(x,...,value) UseMethod("exogenous<-")

##' @export
`exogenous<-.lvm` <- function(x,
                       xfree=TRUE,
                       ...,value) {
    if (inherits(value,"formula")) {
        value <- all.vars(value)
    }
    not.in <- !(value%in%vars(x))
    if (any(not.in)) {
        addvar(x,reindex=FALSE) <- value[not.in]
    }
    xorg <- exogenous(x)
    x$exogenous <- value
    if (!is.null(value) & xfree) {
        notexo.idx <- xorg[which(!(xorg%in%value))]
        if (length(notexo.idx)>0) { ##  & mom) {
            if (length(notexo.idx)>1) {
                covariance(x,notexo.idx,pairwise=TRUE,exo=TRUE) <- NA
            }
            covariance(x,notexo.idx,vars(x),exo=TRUE) <- NA
            intercept(x,notexo.idx) <- x$mean[notexo.idx]
        }
    }
    index(x) <- reindex(x)
    return(x)
}

##' @export
`exogenous.lvm` <-
    function(x,variable,latent=FALSE,index=TRUE,...) {
        if (!missing(variable)) {
            exogenous(x) <- variable
            return(x)
        }

        if (!index) {
            if (latent) {
                allvars <- vars(x)
            } else {
                allvars <- manifest(x)
            }
            M <- x$M
            res <- c()
            for (i in allvars)
                if (!any(M[,i]==1) & !any(is.na(x$cov[i,]))) # & any(M[i,]==1))
                    res <- c(res, i)
            return(res)
        }
        if (is.null(x$exogenous)) return(x$exogenous)
        if (all(!is.na(x$exogenous)) & !latent) {
            return(x$exogenous[x$exogenous%in%index(x)$manifest])
        }
        if (!latent)
            return(index(x)$exogenous)
        return(exogenous(x,latent=latent,index=FALSE,...))
    }

##' @export
`exogenous.lvmfit` <-
    function(x,...) {
        exogenous(Model(x),...)
    }

##' @export
exogenous.list <- function(x,...) {
    exolist <- c()
    endolist <- c()
    for (i in seq_along(x)) {
        exolist <- c(exolist, exogenous(x[[i]]))
        endolist <- c(endolist, endogenous(x[[i]]))
    }
    endolist <- unique(endolist)
    exolist <- unique(exolist)
    return(exolist[!(exolist%in%endolist)])
}

##' @export
`exogenous.multigroup` <-
    function(x,...) {
        exogenous(Model(x))
    }

##' @export
`exogenous.lm` <-
    function(x,...) {
        attr(getoutcome(formula(x)),"x")
    }

##' @export
"distribution<-" <- function(x,...,value) UseMethod("distribution<-")

##' @export
"distribution" <- function(x,...,value) UseMethod("distribution")

##' @export
"distribution<-.lvm" <- function(x,
                                 variable,
                                 parname=NULL,
                                 init,
                                 mdist=FALSE,
                                 ...,
                                 value) {
    if (inherits(variable,"formula")) variable <- all.vars(variable)
    dots <- list(...)

    if (!missing(value)) {
        for (obj in c("variable","parname","init","mdist"))
            if (!is.null(attr(value,obj)) && eval(substitute(missing(a),list(a=obj))))
                assign(obj,attr(value,obj))

        if (is.function(value) & mdist & is.null(parname)) parname <- TRUE
    }
    if (!is.null(parname) || length(dots)>0) {
        if (length(parname)>1 || (is.character(parname))) {
            if (missing(init)) {
                parameter(x,start=rep(1,length(parname))) <- parname
            } else {
                parameter(x,start=init) <- parname
            }
            gen <- function(n,p,...) {
                args <- c(n,as.list(p[parname]),dots)
                names(args) <- names(formals(value))[seq(length(parname)+1)]
                do.call(value,args)
            }
        } else {
            gen <- value
            if ("..."%ni%names(formals(gen))) formals(gen) <- c(formals(gen),alist(...=))
            formals(gen) <- modifyList(formals(gen),dots)
        }
        distribution(x,variable,mdist=TRUE) <- list(gen)
        return(x)
    }

    if (length(variable)==1 && !mdist) {
        addvar(x) <- as.formula(paste("~",variable))
        if (is.numeric(value)) value <- list(value)
        if (!is.null(attributes(value)$mean)) intercept(x,variable) <- attributes(value)$mean
        if (!is.null(attributes(value)$variance)) variance(x,variable,exo=TRUE) <- attributes(value)$variance
        x$attributes$distribution[[variable]] <- value
        ## Remove from 'mdistribution'
        vars <- which(names(x$attributes$mdistribution$var)%in%variable)
        for (i in vars) {
            pos <- x$attributes$mdistribution$var[[i]]
            x$attributes$mdistribution$fun[pos] <- NULL
            x$attributes$mdistribution$var[which(x$attributes$mdistribution$var==pos)] <- NULL
            above <- which(x$attributes$mdistribution$var>pos)
            if (length(above)>0)
                x$attributes$mdistribution$var[above] <- lapply(x$attributes$mdistribution$var[above],function(x) x-1)
        }
        return(x)
    }

    if (is.list(value) && length(value)==1 && (is.function(value[[1]]) || is.null(value[[1]]))) {
        addvar(x) <- variable
        ## Multivariate distribution
        if (is.null(x$attributes$mdistribution)) x$attributes$mdistribution <- list(var=list(), fun=list())
        vars <- x$attributes$mdistribution$var

        if (any(ii <- which(names(vars)%in%variable))) {
            num <- unique(unlist(vars[ii]))
            vars[which(unlist(vars)%in%num)] <- NULL
            newfunlist <- list()
            numleft <- unique(unlist(vars))
            for (i in seq_along(numleft)) {
                newfunlist <- c(newfunlist, x$attributes$mdistribution$fun[[numleft[i]]])
                ii <- which(unlist(vars)==numleft[i])
                vars[ii] <- i
            }
            K <- length(numleft)
            x$attributes$mdistribution$var <- vars
            x$attributes$mdistribution$fun <- newfunlist
        } else {
            K <- length(x$attributes$mdistribution$fun)
        }
        if (length(distribution(x))>0)
            distribution(x,variable) <- rep(list(NULL),length(variable))

        x$attributes$mdistribution$var[variable] <- K+1
        x$attributes$mdistribution$fun <- c(x$attributes$mdistribution$fun,value)

        return(x)
    }

    if ((length(value)!=length(variable) & length(value)!=1))
        stop("Wrong number of values")
    for (i in seq_along(variable))
        if (length(value)==1) {
            distribution(x,variable[i],...) <- value
        } else {
            distribution(x,variable[i],...) <- value[[i]]
        }
    return(x)

}

##' @export
"distribution.lvm" <- function(x,var,value,...,multivariate=FALSE) {
    if (!missing(value)) {
        distribution(x,var,...) <- value
        return(x)
    }
    if (multivariate) return(x$attributes$mdistribution)
    x$attributes$distribution[var]
}

################################################################################

##' @export
dist_gaussian <- function(link="identity",mean,sd,log=FALSE,...) {
    rnormal <- if(log) rlnorm else rnorm
    fam <- stats::gaussian(link); fam$link <- link
    f <- function(n,mu,var,...) rnormal(n,fam$linkinv(mu),sqrt(var))
    if (!missing(mean)) attr(f,"mean") <- mean
    if (!missing(sd)) attr(f,"variance") <- sd^2
    attr(f,"family") <- fam
    return(f)
}

##' @export
dist_lognormal <- function(...) {
  structure(normal.lvm(...,log=TRUE),family=list(family="log-normal",...))
}

##' @export
dist_poisson <- function(link="log",lambda,...) {
    fam <- stats::poisson(link); fam$link <- link
    f <- function(n,mu,...) {
        if (missing(n)) {
            return(fam)
        }
        rpois(n,fam$linkinv(mu))
    }
    if (!missing(lambda)) attr(f,"mean") <- fam$linkfun(lambda)
    attr(f,"family") <- fam
    attr(f,"var") <- FALSE
    return(f)
}


## @examples
## m <- lvm()
## categorical(m,K=3) <- ~x
## distribution(m,~y) <- pareto.lvm(lambda=1)
## regression(m,additive=FALSE) <- y~x
## regression(m) <- y~z
## d <- sim(m,1e4,p=c("y~x:0"=1,"y~x:1"=1,"y~x:2"=exp(1)))
##
## X <- model.matrix(y~-1+factor(x)+z,data=d)
## mlogL <- function(theta) {
##     lambda <- exp(theta[1])
##     mu <- exp(X%*%theta[-1])
##     -sum(log(lambda*mu*(1+mu*d$y)^{-lambda-1}))
## }
## nlminb(rep(0,ncol(X)+1),mlogL)
##' @export
dist_pareto <- function(lambda=1,...) {   ## shape: lambda, scale: mu
    ## Density f(y): lambda*mu*(1+mu*y)^{-lambda-1}
    ## Survival S(y): (1+mu*y)^{-lambda}
    ## Inverse CDF: u -> ((1-u)^{-1/lambda}-1)/mu
    f <- function(n,mu,var,...) {
        ((1-runif(n))^(-1/lambda)-1)/exp(mu)
    }
    attr(f,"family") <- list(family="pareto",
                             par=c(lambda=lambda))
    return(f)
}

##' @export
dist_threshold <- function(p,labels=NULL,...) {
    if (sum(p)>1 || any(p<0 | p>1)) stop("wrong probability vector") ;
    if (!is.null(labels))
        return(function(n,...) {
            return(cut(rnorm(n),breaks=c(-Inf,qnorm(cumsum(p)),Inf),labels=labels))
        })
    function(n,...)
        cut(rnorm(n),breaks=c(-Inf,qnorm(cumsum(p)),Inf))
}

##' @export
dist_multinomial <- function(prob, labels=NULL) {
  if (sum(prob)<1)
    prob <- c(prob, 1-sum(prob))
  if (is.null(labels)) {
    labels = seq(0, length(prob)-1)
  }
  structure(function(n, ...) sample(labels, n, prob=prob, replace=TRUE),
            family=list(family='multinomial', par=prob))
}

##' @export
dist_bernoulli <- function(link="logit",p,size=1) {
    if (substitute(link)==quote(identity)) {
        link <- "identity"
    }
    fam <- stats::binomial(link); fam$link <- link
    f <- function(n,mu,var,...) {
        if (missing(n)) {
            return(fam)
        }
        rbinom(n,size,fam$linkinv(mu))
    }
    attr(f,"family") <- fam
    attr(f,"var") <- FALSE
    if (!missing(p)) attr(f,"mean") <- fam$linkfun(p)
    ## f <- switch(link,
    ##             logit =
    ##             function(n,mu,var,...) rbinom(n,1,tigol(mu)),
    ##             cloglog =
    ##             function(n,mu,var,...) rbinom(n,1,1-exp(-exp(1-mu))),
    ##             function(n,mu,var=1,...) rbinom(n,1,pnorm(mu,sd=sqrt(var)))
    ##             ### function(n,mu=0,var=1,...) (rnorm(n,mu,sqrt(var))>0)*1
    ##             )
    ##}
    return(f)
}

##' @export
dist_gamma <- function(link="inverse",shape,rate,unit=FALSE,var=FALSE,log=FALSE,...) {
    fam <- stats::Gamma(link); fam$link <- link
    rgam <- if (!log) rgamma else function(...) log(rgamma(...))
    if (!missing(shape) & !missing(rate))
        f <- function(n,mu,var,...) rgam(n,shape=shape,rate=rate)
    if (!missing(shape) & missing(rate)) {
        if (unit)
            f <- function(n,mu,var,...) rgam(n,shape=shape,rate=shape)
        else if (var)
            f <- function(n,mu,var,...) rgam(n,shape=shape,rate=sqrt(shape/var))
        else
            f <- function(n,mu,var,...) rgam(n,shape=shape,rate=shape/fam$linkinv(mu))
    }
    if (missing(shape) & !missing(rate)) {
        if (unit)
            f <- function(n,mu,var,...) rgam(n,shape=shape,rate=rate)
        else if (var)
            f <- function(n,mu,var,...) rgam(n,shape=rate^2*var,rate=rate)
        else
            f <- function(n,mu,var,...) rgam(n,shape=rate*fam$linkinv(mu),rate=rate)
    }
    if (missing(shape) & missing(rate)) {
        if (var)
            f <- function(n,mu,var,...) rgam(n,shape=var,rate=1)
        else
            f <- function(n,mu,var,...) rgam(n,shape=fam$linkinv(mu),rate=1)
    }
    attr(f,"family") <- fam
    attr(f,"var") <- FALSE
    par <- NULL
    if (!missing(shape) & !missing(rate))
      par <- c(shape=shape, rate=rate)
    attr(f, "family") <- list(family="gamma", par=par)
    return(f)
}

##' @export
dist_loggamma <- function(...) dist_gamma(..., log=TRUE)

##' @export
dist_chisq <- function(df=1,...) {
  f <- function(n,mu,var,...) mu + rchisq(n,df=df)
  attr(f, "family") <- list(family="chisq", par=c(df=df))
  return(f)
}

##' @export
dist_t <- function(df=2, mu, sigma,...) {
    f <- function(n,mu,var,...) mu + sqrt(var)*rt(n,df=df)
    if (!missing(mu)) attr(f,"mean") <- mu
    if (!missing(sigma)) attr(f,"variace") <- sigma^2
    attr(f, "family") <- list(family="student-t")
    return(f)
}

##' @export
dist_uniform <- function(a,b, value=NULL) {
  if (!is.null(value)) {
    f <- function(n, mu, var, ...)
      sample(value, size=n, replace=TRUE)
    attr(f, "family") <- list(family="discrete uniform")
    return(f)
  }
  if (!missing(a) & !missing(b))
    f <- function(n, mu, var,...) mu+runif(n,a,b)
  else
    f <- function(n, mu, var,...)
      mu+(runif(n,-1,1)*sqrt(12)/2*sqrt(var))
  attr(f, "family") <- list(family="uniform")
  return(f)
}

## see also eventTime.R for dist_cox_weibull
##' @export
dist_weibull <- function(intercept=0, sigma=.5, scale, shape) {
    ## accelerated failure time (AFT) regression
    ## parametrization.
    ##
    ## We parametrize the Weibull distribution (without covariates) as follows:
    ## hazard(t) = shape / scale * (t/scale)^(shape-1)
    ## The hazard is:
    ## - rising if shape > 1
    ## - declining if shape <1
    ## - constant if shape=1
    ##
    ## PH regresion:
    ## hazard(t|x) =  shape / scale * (t/scale)^(shape-1)*exp(x'b)
    ## AFT regression
    ## log(T) = a0 + x'B + s*W
    ## Where W is extreme value distributed and,
    ## a0 = log(scale), B=-sigma*b, sigma = 1/shape
    if (missing(shape)) shape <- 1/sigma
    if (missing(scale)) scale <- exp(intercept)
    f <- function(n,mu=0,var,...) {
        (- log(runif(n)) * exp(log(scale)*shape) * exp(mu*shape))^{1/shape}
        ## scale * (-log(1-runif(n)))^{1/shape}
        ## (- (log(runif(n)) / (1/scale)^(shape) * exp(-mu)))^(1/shape)
    }
    attr(f,"family") <- list(family="weibull",
                             regression="AFT",
                             par=c(shape=shape,scale=scale))
    return(f)
}

##' @export
dist_seqint <- function(...) Sequence.lvm(integer=TRUE)

##' @export
dist_seq <- function(a=0,b=1,integer=FALSE) {
  if (integer) {
    f <- function(n,...) seq(n)
    attr(f, "family") <- list(family="sequence")
    return(f)
  }
  if (is.null(a) || is.null(b)) {
    if (!is.null(a)) {
      f <- function(n,...) seq(a,length.out=n)
    } else {
      f <- function(n,...) seq(n)-(n-b)
    }
  } else {
    f <- function(n,...) seq(a,b,length.out=n)
  }
  attr(f, "family") <- list(family="sequence")
  return(f)
}

##' @export
dist_none <- function(...) {
  f <- function(n, mu, ...) {
    return(mu)
  }
  attr(f, "family") <- list(family="intervention")
  return(f)
}

##' @export
dist_const <- function(value=NA) {
  f <- function(n, mu, ...) {
    if (!is.na(value)) return(rep(value, n))
    return(mu)
  }
  attr(f, "family") <- list(family="constant", par=list(value=value))
  return(f)
}

##' @export
dist_ones <- function(p=1,interval=NULL) {
    f <- function(n,...) {
        if (!is.null(interval)) {
            val <- rep(0L,n)
            if (!is.list(interval)) interval <- list(interval)
            for (i in seq_along(interval)) {
                ii <- interval[[i]]
                lo <- round(ii[1]*n)
                hi <- round(ii[2]*n)
                val[seq(lo,hi)] <- 1L
            }
            return(val)
        }
        if (p==0) return(rep(0L,n))
        val <- rep(1L,n)
        if (p>0 && p<1) val[seq(n*(1-p))] <- 0L
        val
    }
    attr(f, "family") <- list(family="binary (deterministic)")
    return(f)

}

##' @export
dist_beta <- function(alpha=1,beta=1,scale=TRUE) {
    ## CDF: F(x) = B(x,alpha,beta)/B(alpha,beta)
    ## Mean: alpha/(alpha+beta)
    ## Var: alpha*beta/((alpha+beta)^2*(alpha+beta+1))
    if (scale)
        f <- function(n,mu,var,...) {
            m <- alpha/(alpha+beta)
            v <- alpha*beta/((alpha+beta)^2*(alpha+beta+1))
            y <- stats::rbeta(n,shape1=alpha,shape2=beta)
            mu+(y-m)*sqrt(var/v)
        }
    else
      f <- function(n,mu,var,...) stats::rbeta(n,shape1=alpha,shape2=beta)
    attr(f, "family") <- list(family="beta", par=c(alpha=alpha, beta=beta))
    return(f)
}

##' @export
dist_mvn <- function(N=2,rho=0.5,sigma=NULL,parname="rho") {
    f <- function(n,rho) {
        if (is.null(sigma)) {
            sigma <- diag(nrow=N)*(1-rho) + rho
        }
        rmvn0(n,sigma=sigma)
    }
    if (!is.null(sigma)) parname <- TRUE
    structure(f,parname=parname,init=rho,mdist=TRUE)
}

##' @export
dist_gaussian_mixture2 <- function(...,
                                   parname=c("Pr",
                                             "M1","M2",
                                             "V1","V2"),
                                   init=c(0.5,
                                          -4,4,
                                          1,1)) {
    f <- function(n,pr,m1,m2,v1,v2) {
        y1 <- rnorm(n,m1,v1^0.5)
        if (pr>=1) return(y1)
        z <- rbinom(n,1,pr)
        y2 <- rnorm(n,m2,v2^0.5)
        return(z*y1+(1-z)*y2)
    }
    structure(f,parname=parname,init=init)
}

##' @export
dist_gaussian_mixture3 <- function(...,
                                   parname=c("Pr1","Pr2",
                                             "M1","M2","M3",
                                             "V1","V2","V3"),
                                   init=c(0.25,0.5,
                                          -4,0,4,
                                          1,1,1)) {
    f <- function(n,pr1,pr2,m1,m2,m3,v1,v2,v3) {
        p <- c(pr1,pr2,1-pr1-pr2)
        y1 <- rnorm(n,m1,v1^0.5)
        y2 <- rnorm(n,m2,v2^0.5)
        y3 <- rnorm(n,m3,v3^0.5)
        z <- stats::rmultinom(n,1,p)
        rowSums(cbind(y1,y2,y3)*t(z))
    }
    structure(f,parname=parname,init=init)
}

##' Faster plot via RGL
##' @title fplot
##' @export
##' @examples
##' if (interactive()) {
##' data(iris)
##' fplot(Sepal.Length ~ Petal.Length+Species, data=iris, size=2, type="s")
##' }
##' @param x X variable
##' @param y Y variable
##' @param z Z variable (optional)
##' @param xlab x-axis label
##' @param ylab y-axis label
##' @param ... additional arggument to lower-level plot functions
##' @param z.col color (use argument alpha to set transparency)
##' @param data data.frame
##' @param add if TRUE use current active device
##' @param aspect aspect ratio
##' @param zoom zoom level
fplot <- function(x,y,z=NULL,xlab,ylab,...,z.col=topo.colors(64), 
                  data=parent.frame(),add=FALSE,aspect=c(1,1),zoom=0.8) {
    if (!requireNamespace("rgl",quietly=TRUE)) stop("Requires 'rgl'")    
    if (inherits(x,"formula")) {
        y <- getoutcome(x)
        x <- attributes(y)$x
        if (length(x)>1) {
            z <- as.numeric(with(data, get(x[2])))
        }
        if (length(x)==0) {
            x <- seq(nrow(data))
            if (missing(xlab)) xlab <- "Index"
        } else {
            if (missing(xlab)) xlab <- x[1]
            x <- with(data, get(x[1]))
        }
        if (missing(ylab)) ylab <- y
        y <- with(data, get(y))
    } else {
        if (missing(y)) {
            y <- x
            if (missing(ylab)) ylab <- deparse(substitute(x))
            x <- seq(nrow(data))            
            if (missing(xlab)) xlab <- "Index"       
        } else {
            if (missing(xlab)) xlab <- deparse(substitute(x))
            if (missing(ylab)) ylab <- deparse(substitute(y))
        }
    }
    rgl::.check3d()
    if (!is.null(z)) {
        ncol <- length(z.col);
        glut <- approxfun(seq(min(z),max(z),length.out=ncol),seq(ncol))
        rgl::plot3d(x,y,0,col=z.col[round(glut(z))],xlab=xlab,ylab=ylab,add=add,...)
    } else {
        rgl::plot3d(x,y,0,xlab=xlab,ylab=ylab,add=add,...)
        
    }
    rgl::view3d(0,0,fov=0,zoom=zoom)
    rgl::aspect3d(c(aspect,1))
}

##v <- lmerplot(l1,varcomp=TRUE,colorkey=TRUE,lwd=0,col=rainbow(20))
lmerplot <- function(model,x,id,y,transform,re.form=NULL,varcomp=FALSE,colorbar=TRUE,mar=c(4,4,4,6),col,index=seq(50),...) {
    if (varcomp) {
        Z <- lme4::getME(model, "Z")
        nn <- unlist(lapply(lme4::getME(model, "Ztlist"), nrow))
        ve <- lme4::getME(model, "sigma")^2
        vu <- varcomp(model,profile=FALSE)$varcomp
        L <- Matrix::Diagonal(sum(nn), rep(vu, nn))
        V <- Z%*%L%*%(Matrix::t(Z))
        Matrix::diag(V) <- Matrix::diag(V)+ve
        cV <- Matrix::cov2cor(V)
        if (!is.null(index)) {
            index <- intersect(seq_len(nrow(cV)), index)
            cV <- cV[index,index,drop=FALSE]
        }
        if (colorbar) { opt <- par(mar=mar) }
        if (missing(col)) col <- rev(gray.colors(16,0,1))
        image(seq_len(nrow(cV)), seq_len(ncol(cV)),
              as.matrix(cV),
            xlab = "", ylab = "", col = col, zlim = c(0, 1), ...
        )
        if (colorbar) {
            uu <- devcoords()
            xrange <- c(uu$fig.x2,uu$dev.x2)
            xrange <- diff(xrange)/3*c(1,-1)+xrange
            yrange <- c(uu$fig.y1,uu$fig.y2)
            colorbar(
              direction = "vertical", x.range = xrange, y.range = yrange,
              clut = col, values = seq(0, 1, length.out = length(col)), srt = 0, position = 2
            )
            par(opt)
        }
        return(invisible(V))
    }
    if (missing(y)) y <- model.frame(model)[,1]
    yhat <- predict(model)
    if (!is.null(re.form)) ymean <- predict(model,re.form=re.form)
    if (!missing(transform)) {
        yhat <- transform(yhat)
        if (!is.null(re.form)) ymean <- transform(ymean)
        y <- transform(y)
    }
    plot(y ~ x, col=Col(id,0.3), pch=16, ...)
    if (!is.null(re.form)) points(ymean ~ x, pch="-", cex=4);
    for (i in unique(id)) {
        idx <- which(id==i)
        lines(yhat[idx]~x[idx],col=i)
    }
}

varcomp <- function(x, profile=TRUE, ...) {
  if (inherits(x, "lme")) {
    coefs <- nlme::fixef(x)
    varcomp <- nlme::getVarCov(x)
    pr <- nlme::intervals(x)
    resvar <- summary(x)$sigma
  } else if (inherits(x, "lmerMod")) {
    coefs <- lme4::fixef(x)
    varcomp <- lme4::VarCorr(x)[[1]][, ]
    resvar <- attributes(lme4::VarCorr(x))$sc^2
    pr <- confint(x) ## Profile likelihood intervals
  } else {
    stop("unrecognized model object")
  }
  cc <- cbind(coefs, diag(as.matrix(vcov(x)))^.5)
  cc <- cbind(cc,cc[,1]-qnorm(0.975)*cc[,2],
              cc[,1]+qnorm(0.975)*cc[,2],
              2*(1-pnorm(abs(cc[,1])/cc[,2])))
  colnames(cc) <- c("Estimate","Std.Err","2.5%","97.5%","p-value")
  res <- structure(list(coef=coefs,
                        vcov=as.matrix(vcov(x)),
                        coefmat=cc,
                        confint=pr,
                        varcomp=varcomp,
                        varresidual=resvar
                        ),
                   class="estimate.lmer")
  res
}

pzmax <- function(alpha, S) {
  ## P(Zmax > z) Family wise error rate, Zmax = max |Z_i|
  if (!requireNamespace("mets",quietly=TRUE))
    stop("'mets' package required")
  k <- nrow(S)
  zz <- qnorm(1-alpha/2)
  unlist(lapply(zz, function(z)
    1 - mets::pmvn(lower=rep(-z, k),
                   upper=rep(z, k),
                   sigma=cov2cor(S))
    ))
}

##' @export
alpha_zmax <- function(object, alpha = 0.05, ...) {
  if (!inherits(object, c("estimate", "summary.estimate"))) {
    stop("Expected an 'estimate' or 'summary.estimate' object")
  }
  if (!inherits(object, "summary.estimate")) {
    object <- summary(object, ...)
  }
  est <- parameter(object)[, c(1, 5), drop = FALSE]
  padj <- pzmax(est[, 2], vcov(object))
  res <- cbind(est, padj)
  colnames(res)[3] <- paste0("Adj.", colnames(res)[2])
  f <- function(a) pzmax(a, vcov(object)) - alpha
  adj <- uniroot(f, lower = 0, upper = alpha)$root
  attributes(res)["adjusted.significance.level"] <- adj
  res
}

##' Closed testing procedure
##'
##' Given \eqn{k} hypotheses \eqn{H_1, \ldots, H_k}{H1, ..., Hp} all
##' \eqn{2^k-1}{2^k-1} intersection hypotheses are calculated and adjusted
##' p-values are obtained for \eqn{H_j}{Hj} is calculated as the max
##' \eqn{p}-value of all intersection hypotheses containing \eqn{H_j}{Hj}.
##' Example, for \eqn{k=3}, the adjusted \eqn{p}-value for \eqn{H_1}{H1} will be
##' obtained from \eqn{\{(H_1, H_2, H_3), (H_1,H_2), (H_1,H_3), (H_1)\}}.
##'
##' \if{html}{
##'   \figure{closedtesting.svg}{options: width=500}
##' }
##' \if{latex}{
##'   \figure{closedtesting.pdf}{options: width=7cm}
##' }
##' @aliases closed_testing alpha_zmax
##' @param object `estimate` object
##' @param test function that conducts hypothesis test. See details below.
##' @param ... Additional arguments passed to `test`
##' @export
##' @details The function `test` should be a function `function(object, index,
##'   ...)` which as its first argument takes an `estimate` object and with
##'   an argument `index` which is a integer vector specifying which
##'   subcomponents of `object` to test. The ellipsis argument can be any other
##'   arguments used in the test function. The function [test_wald()] is an
##'   example of valid test function (which has an additional argument `null` in
##'   reference to the above mentioned ellipsis arguments).
##' @references Marcus, R; Peritz, E; Gabriel, KR (1976).
##'   "On closed testing procedures with special reference to ordered analysis of variance".
##'   Biometrika. 63 (3): 655–660.
##' @examples
##' m <- lvm()
##' regression(m, c(y1,y2,y3,y4)~x) <- c(0, 0.25, 0, 0.25)
##' regression(m, to=endogenous(m), from="u") <- 1
##' variance(m,endogenous(m)) <- 1
##' set.seed(1)
##' d <- sim(m, 200)
##' l1 <- lm(y1~x,d)
##' l2 <- lm(y2~x,d)
##' l3 <- lm(y3~x,d)
##' l4 <- lm(y4~x,d)
##'
##' (a <- merge(l1, l2, l3, l4, subset=2))
##' if (requireNamespace("mets",quietly=TRUE)) {
##'    alpha_zmax(a)
##' }
##' adj <- closed_testing(a, test = test_wald, null = 0)
##' adj
##' adj$p.value
##' summary(adj)
closed_testing <- function(object, test = test_wald, ...) {
  if (!inherits(object, c("estimate", "summary.estimate"))) {
    stop("`estimate` or `summary.estimate` object needed")
  }
  idx <- seq_along(coef(object))
  if (length(idx) > 15)
    stop("Too many tests. Consider some other adjustment method")
  combs <- pvals <- c()
  for (i in seq_along(idx)) {
    co <- combn(idx, i)
    pp <- numeric(ncol(co))
    for (j in seq_along(pp)) {
      ii <- co[, j]
      tt <- test(object, index = ii, ...)
      pp[j] <- tt$p.value
    }
    combs <- c(combs, list(co))
    pvals <- c(pvals, list(pp))
  }
  pmax <- c()
  for (k in seq_along(idx)) {
    pk <- c()
    for (i in seq_along(idx)) {
      cols <- apply(combs[[i]], 2, function(x) k %in% x)
      pk <- c(pk, pvals[[i]][which(cols)])
    }
    pmax <- c(pmax, max(pk))
  }
  res <- cbind(coef(object), pmax)
  colnames(res) <- c("Estimate", "adj.p")
  rownames(res) <- names(coef(object))
  names(pmax) <- names(coef(object))
  structure(
    list(
      adjp = res,
      call = match.call(),
      args = list(...),
      hypotheses = combs,
      raw.pval = pvals,
      estimate = coef(object),
      p.value = pmax),
    class = "test_adj"
  )
}

#' @export
print.test_adj <- function(x, ...) {
  cat("Call: ")
  print(x$call, quote = FALSE)
  cat("\n")
  print(x$adjp)
}

#' @export
summary.test_adj <- function(object, ...) {
  cat("Call: ")
  print(object$call, quote = FALSE)
  # Print adjusted p-values
  cli::cli_h2("Adjusted p-values")
  print(object$adjp)
  cat("\n")
  # Print raw p-values for intersection hypotheses
  cli::cli_h2("Raw p-values for intersection hypotheses")
  param_names <- names(object$estimate)
  for (i in seq_along(object$hypotheses)) {
    cat(sprintf("%d-way intersections:\n", i))
    comb_matrix <- object$hypotheses[[i]]
    pvals <- object$raw.pval[[i]]
    for (j in seq_len(ncol(comb_matrix))) {
      indices <- comb_matrix[, j]
      hyp_names <- param_names[indices]
      # Create hypothesis notation
      hyp_str <- paste0("{", paste(hyp_names, collapse = ", "), "}")
      # Print with p-value
      cat(sprintf("  %-40s p = %.4f\n", hyp_str, pvals[j]))
    }
    cat("\n")
  }
  invisible(object)
}

#' @export
test_wald <- function(par,
                      vcov,
                      null = NULL,
                      index = NULL) {
  if (!inherits(par, c("estimate", "summary.estimate"))) {
    par <- lava::estimate(par=par, vcov=vcov)
  }
  np <- length(coef(par))
  if (is.null(null)) {
    null <- 0
  }
  null <- rep(null, length.out = np)
  B <- diag(1, nrow = np)
  if (!is.null(index)) {
    if (length(coef(par)) < length(index)) stop("wrong `index`")
    null <- null[index]
    B <- B[index, , drop=FALSE]
  }
  lava::compare(par, contrast = B, null=null)
}

## kmeans++
## Assign centre of first cluster randomly from observations.
## 1) Calculate min distances of observations to current clusters, D
## 2) Sample new cluster centre from distribution p(x) = D^2(x)/sum(D^2)
## Repeat 1-2 until all clusters are assigned
kmpp <- function(y,k=2) {
    Dist <- function(y1,y2) sum(y1-y2)^2
    n <- NROW(y)
    ii <- numeric(k)
    u <- runif(k)
    ii[1] <- sample(n,1)
    D <- matrix(0,n,k-1)
    hasmets <- requireNamespace("mets", quietly=TRUE)
    for (i in seq_len(k-1)+1) {
        D[,i-1] <- apply(y,1,function(x) Dist(x,y[ii[i-1],]))
        D2 <- apply(D[,seq(i-1),drop=FALSE],1,min)
        pdist <- cumsum(D2/sum(D2))
        if (hasmets)
            ii[i] <- mets::fast.approx(pdist,u[i])
        else
            ii[i] <- which.min(abs(pdist-u[i]))
    }
    return(ii)
}


##' Weighted K-means
##'
##' Weighted K-means via Lloyd's algorithm
##' @param x Data (or formula)
##' @param mu Initial centers (or number centers chosen randomly among x)
##' @param data optional data frmae
##' @param weights Optional weights
##' @param iter.max Max number of iterations
##' @param n.start Number of restarts
##' @param init method to create initial centres (default kmeans++)
##' @param ... Additional arguments to lower level functions
##' @export
##' @author Klaus K. Holst
##'
wkm <- function(x, mu, data, weights=rep(1,NROW(x)), iter.max=20, n.start=5, init="kmpp", ...) { ## Lloyd's algorithm
    if (inherits(x, "formula")) x <- stats::model.matrix(x,data=data)
    x <- cbind(x)
    random.start <- TRUE
    if (is.list(mu)) {
        random.start <- FALSE
        n.start=1
        K <- length(mu)
    } else {
        K <- mu
    }
    sswmin <- Inf
    mus <- ssws <- NULL
    cl0 <- rep(1,NROW(x))
    for (k in seq(n.start)) {
        if (random.start) {
            if (!exists(init)) { ## Random select centres
                idx <- sample(NROW(x),K)
            } else {
                idx <- do.call(init, list(x, K))
            }
            mu <- lapply(idx, function(i) cbind(x)[i,,drop=TRUE])
        }
        mus <- c(mus, list(mu))
        for (i in seq(iter.max)) {
            d <- Reduce(cbind,lapply(mu, function(m) weights*colSums((t(x)-m)*(t(x)-m))))
            cl <- apply(d,1,which.min)
            for (j in seq_along(mu)) {
                idx <- which(cl==j)
                if (length(idx)) {
                    mu[[j]] <- colSums(cbind(apply(x[idx,,drop=FALSE],2,
                                            function(x) x*weights[idx])))/sum(weights[idx])
                }

            }
            if (sum(cl0-cl)==0L) break; # No change in assigment
        }
        ssw <- sum(d[cbind(seq(NROW(d)),cl)])
        ssws <- c(ssws,ssw)
        if (ssw < sswmin) {
            sswmin <- ssw
            clmin <- cl
            mumin <- mu
        }
    }
    mu <- structure(mu,class="by",dim=K,dimnames=list(class=seq(K)))
    withinclusterss <- as.vector(by(d[cbind(seq(NROW(d)),cl)],cl,sum))
    return(list(cluster=cl,
           center=mu, ssw=withinclusterss))
}

context("Weighted K-means (wkm)")

## Helper: best-permutation accuracy of cluster labels vs truth
.cluster_accuracy <- function(pred, truth) {
    tab <- table(pred, truth)
    K <- nrow(tab)
    perms_list <- function(k) {
        if (k == 1L) return(list(1L))
        sub <- perms_list(k - 1L)
        out <- list()
        for (s in sub) {
            for (i in seq_len(k)) {
                out[[length(out) + 1L]] <- append(s, k, after = i - 1L)
            }
        }
        out
    }
    perms <- perms_list(K)
    n_total <- sum(tab)
    best <- 0
    for (p in perms) {
        ## p is a permutation of columns (truth) to align with rows (pred)
        acc <- sum(diag(tab[, p, drop = FALSE])) / n_total
        if (acc > best) best <- acc
    }
    best
}

## ---------------------------------------------------------------------------
## Return structure
## ---------------------------------------------------------------------------
test_that("wkm returns list with cluster, center, ssw of expected types", {
    set.seed(1)
    x <- matrix(rnorm(60), ncol = 2)
    res <- wkm(x, mu = 2, n.start = 1)
    testthat::expect_type(res, "list")
    testthat::expect_named(res, c("cluster", "center", "ssw"),
                           ignore.order = TRUE)
    testthat::expect_length(res$cluster, nrow(x))
    testthat::expect_true(inherits(res$center, "by"))
    testthat::expect_equal(length(res$center), 2L)
    testthat::expect_true(all(res$ssw >= 0))
})

## ---------------------------------------------------------------------------
## Two well-separated clusters: correctness of clustering and centers
## ---------------------------------------------------------------------------
test_that("wkm separates two well-separated Gaussian blobs", {
    set.seed(2)
    n <- 100
    x <- rbind(matrix(rnorm(n * 2, mean = -3), ncol = 2),
               matrix(rnorm(n * 2, mean =  3), ncol = 2))
    truth <- c(rep(1L, n), rep(2L, n))
    res <- wkm(x, mu = 2, n.start = 5)
    testthat::expect_equal(length(unique(res$cluster)), 2L)
    ## centers (in some order) close to (-3,-3) and (3,3)
    cents <- do.call(rbind, lapply(res$center, as.numeric))
    ## sort by first coordinate
    cents <- cents[order(cents[, 1]), ]
    testthat::expect_lt(max(abs(cents[1, ] - c(-3, -3))), 0.5)
    testthat::expect_lt(max(abs(cents[2, ] - c(3, 3))), 0.5)
    ## cluster purity >= 95%
    acc <- .cluster_accuracy(res$cluster, truth)
    testthat::expect_gte(acc, 0.95)
})

## ---------------------------------------------------------------------------
## Custom initial centers (random.start = FALSE branch)
## ---------------------------------------------------------------------------
test_that("wkm accepts list of initial centers and forces n.start = 1", {
    set.seed(3)
    n <- 80
    x <- rbind(matrix(rnorm(n * 2, mean = -3), ncol = 2),
               matrix(rnorm(n * 2, mean =  3), ncol = 2))
    truth <- c(rep(1L, n), rep(2L, n))
    res <- wkm(x, mu = list(c(-3, -3), c(3, 3)), n.start = 7)
    testthat::expect_equal(length(res$center), 2L)
    acc <- .cluster_accuracy(res$cluster, truth)
    testthat::expect_gte(acc, 0.95)
})

## ---------------------------------------------------------------------------
## Univariate input
## ---------------------------------------------------------------------------
test_that("wkm handles univariate input via cbind", {
    set.seed(4)
    x <- c(rnorm(40, -2), rnorm(40, 2))
    res <- wkm(x, mu = 2, n.start = 3)
    testthat::expect_length(res$cluster, length(x))
    testthat::expect_equal(length(res$center), 2L)
})

## ---------------------------------------------------------------------------
## Formula interface
## ---------------------------------------------------------------------------
test_that("wkm formula interface works with data argument", {
    set.seed(5)
    res <- wkm(~ Sepal.Length + Sepal.Width, data = iris, mu = 3, n.start = 3)
    testthat::expect_length(res$cluster, nrow(iris))
    testthat::expect_lte(length(unique(res$cluster)), 3L)
})

## ---------------------------------------------------------------------------
## Weights effect: heavily weighted observations dominate within-cluster mean
## ---------------------------------------------------------------------------
test_that("wkm weights influence cluster centers", {
    testthat::skip_on_cran()
    set.seed(6)
    n <- 100
    ## Two well-separated blobs; within blob 2 add a heavily weighted shift
    blob1 <- matrix(rnorm(n * 2, mean = -3), ncol = 2)
    blob2 <- matrix(rnorm(n * 2, mean =  3), ncol = 2)
    x <- rbind(blob1, blob2)

    ## Uniform weights -> blob 2 center near (3, 3)
    res_u <- wkm(x, mu = list(c(-3, -3), c(3, 3)),
                 weights = rep(1, nrow(x)), n.start = 1)
    cu2 <- as.numeric(res_u$center[[2]])

    ## Heavily up-weight the *first half* of blob 2 (which is centered above
    ## the blob mean by construction of rnorm + sort by index).  The new
    ## center should be measurably different from the uniform-weight center.
    w <- rep(1, nrow(x))
    w[(n + 1):(n + n %/% 2)] <- 50
    res_w <- wkm(x, mu = list(c(-3, -3), c(3, 3)),
                 weights = w, n.start = 1)
    cw2 <- as.numeric(res_w$center[[2]])

    ## The weighted center should differ noticeably from the uniform center
    testthat::expect_gt(sqrt(sum((cw2 - cu2)^2)), 0.05)
    ## And it should equal the weighted mean of blob 2 observations assigned
    ## to cluster 2 (sanity check on lines 71-72).
    cl2_idx <- which(res_w$cluster == 2)
    expected <- colSums(x[cl2_idx, , drop = FALSE] * w[cl2_idx]) / sum(w[cl2_idx])
    testthat::expect_equal(cw2, as.numeric(expected), tolerance = 1e-8)
})

## ---------------------------------------------------------------------------
## iter.max = 1 still returns valid output
## ---------------------------------------------------------------------------
test_that("wkm runs with iter.max = 1", {
    set.seed(7)
    x <- matrix(rnorm(40), ncol = 2)
    res <- wkm(x, mu = 2, iter.max = 1, n.start = 1)
    testthat::expect_length(res$cluster, nrow(x))
    testthat::expect_equal(length(res$center), 2L)
})

## ---------------------------------------------------------------------------
## init argument: kmpp and fallback
## ---------------------------------------------------------------------------
test_that("wkm accepts init='kmpp' and falls back when init does not exist", {
    set.seed(8)
    x <- matrix(rnorm(60), ncol = 2)
    res1 <- wkm(x, mu = 2, init = "kmpp", n.start = 1)
    testthat::expect_length(res1$cluster, nrow(x))
    set.seed(8)
    res2 <- wkm(x, mu = 2, init = "this_init_does_not_exist", n.start = 1)
    testthat::expect_length(res2$cluster, nrow(x))
    testthat::expect_equal(length(res2$center), 2L)
})

## ---------------------------------------------------------------------------
## Reproducibility under set.seed
## ---------------------------------------------------------------------------
test_that("wkm is reproducible under set.seed", {
    x <- matrix(rnorm(80), ncol = 2)
    set.seed(42)
    a <- wkm(x, mu = 2, n.start = 3)
    set.seed(42)
    b <- wkm(x, mu = 2, n.start = 3)
    testthat::expect_equal(a$cluster, b$cluster)
    testthat::expect_equal(a$ssw, b$ssw)
})

## ---------------------------------------------------------------------------
## Regression: K = 1 should not crash (was apply(d,1,which.min) bug)
## ---------------------------------------------------------------------------
test_that("wkm runs with K = 1", {
    set.seed(11)
    x <- matrix(rnorm(40), ncol = 2)
    res <- wkm(x, mu = list(c(0, 0)), n.start = 1)
    testthat::expect_length(res$cluster, nrow(x))
    testthat::expect_true(all(res$cluster == 1L))
    ## with one cluster the center should equal the (weighted) mean
    expected <- colMeans(x)
    testthat::expect_equal(as.numeric(res$center[[1]]),
                           as.numeric(expected), tolerance = 1e-8)
})

## ---------------------------------------------------------------------------
## Regression: best-start (lowest SSW) is actually returned across n.start
## ---------------------------------------------------------------------------
test_that("wkm returns the best-SSW solution across multiple starts", {
    set.seed(13)
    n <- 80
    x <- rbind(matrix(rnorm(n * 2, mean = -3), ncol = 2),
               matrix(rnorm(n * 2, mean =  3), ncol = 2))
    res <- wkm(x, mu = 2, n.start = 10)
    ## Recompute SSW from returned cluster + centers; must match res$ssw
    centers <- do.call(rbind, lapply(res$center, as.numeric))
    sq <- vapply(seq_len(nrow(x)),
                 function(i) sum((x[i, ] - centers[res$cluster[i], ])^2),
                 numeric(1))
    recomputed <- as.vector(by(sq, res$cluster, sum))
    testthat::expect_equal(sort(recomputed), sort(res$ssw),
                           tolerance = 1e-8)
})

## ---------------------------------------------------------------------------
## Regression: kmpp distance metric (sum of squared diffs, not (sum diff)^2)
## ---------------------------------------------------------------------------
test_that("kmpp distance metric distinguishes points with equal coord-sum", {
    ## (5,-5) and (-5,5) have identical coordinate sums.  Under the buggy
    ## metric sum(y1-y2)^2 they have distance 0, which makes the kmeans++
    ## sampling distribution degenerate (D2/sum(D2) = NaN) and kmpp errors
    ## out.  With the correct metric sum((y1-y2)^2) the points are far
    ## apart and kmpp must complete and return two valid indices.
    y <- rbind(c( 5, -5),
               c(-5,  5))
    for (s in 1:5) {
        set.seed(s)
        idx <- lava:::kmpp(y, k = 2)
        testthat::expect_length(idx, 2L)
        testthat::expect_true(all(idx %in% c(1L, 2L)))
    }
})

##' Finds the unique commutation matrix K:
##' \eqn{K vec(A) = vec(A^t)}
##'
##' @title Finds the unique commutation matrix
##' @param m rows
##' @param n columns
##' @author Klaus K. Holst
##' @export
commutation <- function(m, n=m) {
    if (inherits(m,"matrix")) {
        n <- ncol(m)
        m <- nrow(m)
    }
    H <- function(i,j) { ## mxn-matrix with 1 at (i,j)
        Hij <- matrix(0, nrow=m, ncol=n)
        Hij[i,j] <- 1
        Hij
    }
    K <- matrix(0,m*n,m*n)
    for (i in seq_len(m))
        for (j in seq_len(n))
            K <- K + H(i,j)%x%t(H(i,j))
    K
}

impute0 <- function(object,rows,idx,na.action=na.omit,value,...) {
    if (missing(rows) && missing(idx)) {
        df <- na.action(object,...) 
        rows <- attr(df,"na.action")  
    }
    if (!missing(idx)) {
        obs1 <- setdiff(seq(length(object)),idx)[1]
    } else {
        obs1 <- setdiff(seq(NROW(object)),rows)[1]
    }
    if (missing(value)) {
        fobs <- object[obs1]
        if (is.logical(fobs)) value <- FALSE
        else if (is.character(fobs)) value <- fobs
        else if (is.factor(fobs)) value <- levels(fobs)[1]
        else value <- 0
    }
    if (!missing(idx)) {
        object[idx] <- value
        return(object)
    }
    if (is.matrix(object)) {
        object[rows,] <- value
    } else {
        object[rows] <- value
    }
    return(object)
}

##' @export 
na.pass0 <- function(object,all=TRUE,na.action=na.omit, ...) {
    ## Fill in "zeros" in the design matrix where we have missing data    
    df <- na.action(object,...)
    idx <- attr(df,"na.action")
    if (is.matrix(object) || is.vector(object)) {
        object <- impute0(object,rows=idx,...)
    } else {
        for (i in seq_len(NCOL(object))) {
            object[[i]] <- impute0(object[[i]],rows=idx,...)
        }
    }
    if (!is.null(idx))
        return(structure(object,na.action=structure(idx,class="pass0")))
    return(object)
}

updatelvm <- function(x,mean=TRUE,...) {
    index(x) <- reindex(x,mean=mean,...)
    x$parpos <- parpos(x,mean=mean,...)
    return(x)
}

##' Generic method for extract index of an object
##'
##' @export
##' @param x object from which to extract index
##' @param ... further arguments to be passed to or from other methods.
##' @seealso [index<-]
##' @keywords utilities
##' @rdname index-method
"index" <- function(x,...) UseMethod("index")

##' Generic method for setting the index of an object
##'
##' @export
##' @param x object on which to set the index
##' @param value the new index
##' @param ... further arguments to be passed to or from other methods.
##' @seealso [index]
##' @keywords utilities
##' @rdname index-method
##' @export
"index<-" <- function(x,...,value) UseMethod("index<-")

##' @title Extract the parameter indicies of a lvm object
##'
##' @description Extracts the matrices with indices of model parameters from a
##'   latent variable model (`lvm`). Returns a list with
##' * A:   Matrix with fixed parameters and ones where parameters are free
##' * J:   Manifest variable selection matrix
##' * M0:  Index of free regression parameters
##' * M1:  Index of free and _unique_ regression parameters
##' * P:   Matrix with fixed variance parameters and ones where parameters are free
##' * P0:  Index of free variance parameters
##' * P1:  Index of free and _unique_ regression parameters
##' * npar.var:   Number of covariance parameters
##' @rdname index-lvm
##' @inherit index
##' @aliases index.lvmfit index.lvm
##' @seealso [modelPar()]
##' @export
"index.lvm" <- function(x,...) { x$index }

##' @export
"index.lvmfit" <- function(x,...) { index(Model(x)) }

##' @rdname index-lvm
##' @param value new index
##' @aliases index<-.lvmfit index<-.lvm
##' @export
"index<-.lvm" <- function(x,...,value)  { x$index <- value; return(x) }

##' @export
"index<-.lvmfit" <- function(x,...,value) { Model(x)$index <- value; return(x) }


`reindex` <-
    function(x, sparse=FALSE,standard=TRUE,zeroones=FALSE,deriv=FALSE,mean=TRUE) { ## Extract indices of parameters from model
        x$parpos <- NULL
        M <- x$M

        eta <- latent(x) ## Latent variables/Factors
        m <- length(eta)
        obs <- manifest(x)  ## Manifest/Observed variables
        endo <- endogenous(x)
        exo <- exogenous(x) ##,index=FALSE)

        allvars <- vars(x)
        eta.idx <- na.omit(match(eta,allvars))
        obs.idx <- na.omit(match(obs,allvars))
        exo.idx <- na.omit(match(exo,allvars))
        exo.obsidx <- na.omit(match(exo,obs))
        endo.obsidx <- na.omit(match(endo,obs))

        fix.idx <- !is.na(x$fix) ## Index of fixed parameters
        covfix.idx <- !is.na(x$covfix) ## Index of fixed covariance parameters

        constrain.par <- NULL
        if (length(constrain(x))>0) constrain.par <- names(constrain(x))

        M0 <- M;  M0[fix.idx] <- 0 ## Matrix of indicators of free regression-parameters (removing fixed parameters)
        M1 <- M0; ## Matrix of indiciator of free _unique_ regression parameters (removing fixed _and_ duplicate parameters)
        parname <- unique(x$par[!is.na(x$par)])
        for (p in parname) {
            ii <- which(x$par==p)
            if (length(ii)>1)
                M1[ii[-1]] <- 0
            if (p %in% constrain.par)
                M0[ii] <- M1[ii] <- 0
        }
        npar.reg <- sum(M1) ## Number of free regression parameters

        P <- x$cov;

        P0 <- P;  P0[covfix.idx] <- 0 ## Matrix of indicators of free covariance-parameters (removing fixed parameters)
        if (length(exo.idx)>0)
            P0[exo.idx,exo.idx] <- 0 ## 6/1-2011
        P1 <- P0 ## Matrix of indiciator of free _unique_ variance parameters (removing fixed _and_ duplicate parameters)
        covparname <- unique(x$covpar[!is.na(x$covpar)])
        for (p in covparname) {
            ii <- which(x$covpar==p)
            if (length(ii)>1)
                P1[ii[-1]] <- 0
            if (p%in%c(parname,constrain.par))
                P0[ii] <- P1[ii] <- 0
        }

        npar.var <- sum(c(diag(P1),P1[lower.tri(P1)]))

        A <- M
        A[fix.idx] <- x$fix[fix.idx] ## ... with fixed parameters in plac
        P[covfix.idx] <- x$covfix[covfix.idx] ## ... with fixed parameters in plac

        px <- Jy <- J <- diag(nrow=length(vars(x)))
        if (m>0) {
            J[eta.idx,eta.idx] <- 0; J <- J[-eta.idx,,drop=FALSE]
        } ## Selection matrix (selecting observed variables)
        {
            ## Selection matrix (selection endogenous variables)
            if (length(c(eta.idx,exo.idx))>0) {
                Jy[c(eta.idx,exo.idx),c(eta.idx,exo.idx)] <- 0; Jy <- Jy[-c(eta.idx,exo.idx),,drop=FALSE]
            }
            ## Cancelation matrix (cancels rows with exogenous variables)
            px[exo.idx,exo.idx] <- 0
        }

        ## Creating indicitor of free mean-parameters
        fixed <- sapply(x$mean, function(y) is.numeric(y) & !is.na(y))
        named <- sapply(x$mean, function(y) is.character(y) & !is.na(y))
        mparname <- NULL
        if (length(named)>0)
            mparname <- unlist(unique(x$mean[named]))
        v0 <- rep(1,length(x$mean)) ## Vector of indicators of free mean-parameters

        v0[exo.idx] <- 0
        if (length(fixed)>0) v0[fixed] <- 0;
        v1 <- v0
        for (p in mparname) {
            idx <- which(x$mean==p)
            if (length(idx)>1) {
                v1[idx[-1]] <- 0
            }
            if (p%in%c(parname,covparname,constrain.par))
                v0[idx] <- v1[idx] <- 0
        } ## duplicate parameters

        ##
        ## Extra parameters
        ##
        efixed <- sapply(x$exfix, function(y) is.numeric(y) & !is.na(y))
        enamed <- sapply(x$exfix, function(y) is.character(y) & !is.na(y))
        if(length(enamed)>0){
            eparname <- unlist(unique(x$exfix[enamed]))
        } else{
            eparname <- NULL
        }
        ## Extra parameters
        e0 <- rep(1,length(x$expar)) ## Indicators of free extra par.
        if (length(efixed)>0)
            e0[efixed] <- 0
        e1 <- e0
        for (p in eparname) {
            idx <- which(x$exfix==p)
            if (length(idx)>1) {
                e1[idx[-1]] <- 0
            }
            if (p%in%c(parname,covparname,constrain.par,mparname))
                e0[idx] <- e1[idx] <- 0
        } ## duplicate parameters


        ## Return:
        ## Adjacency-matrix (M)
        ## Matrix of regression-parameters (0,1) _with_ fixed parameters (A)
        ## Matrix of variance-parameters (indicators 0,1) (P)
        ## Manifest selection matrix (J),
        ## Position of variables matrix (Apos),
        ## Position of covariance variables matrix (Ppos),
        ## Position/Indicator matrix of free regression parameters (M0)
        res <- list(vars=allvars, manifest=obs, exogenous=exo, latent=eta,
                    endogenous=endo,
                    exo.idx=exo.idx, eta.idx=eta.idx,
                    exo.obsidx=exo.obsidx, endo.obsidx=endo.obsidx,
                    obs.idx=obs.idx,
                    endo.idx=setdiff(obs.idx,exo.idx))

        if (standard) {
            res <- c(res, list(M=M, A=A, P=P,
                               P0=P0, P1=P1,
                               M0=M0, M1=M1,
                               v0=v0, v1=v1,
                               e0=e0, e1=e1,
                               npar=(npar.reg+npar.var),
                               npar.reg=npar.reg,
                               npar.var=npar.var,
                               npar.mean=sum(v1),
                               npar.ex=sum(e1),
                               constrain.par=constrain.par))
            which.diag <- NULL
            if (length(P1)>0)
                which.diag <- which(diag(P1==1))

            res <- c(res, list(parname.all=parname, parname=setdiff(parname,constrain.par),
                               which.diag=which.diag,
                               covparname.all=covparname,
                               covparname=setdiff(covparname,constrain.par),
                               meanfixed=fixed, meannamed=named,
                               mparname.all=mparname,
                               mparname=setdiff(mparname,constrain.par),
                               eparname.all=eparname,
                               eparname=setdiff(eparname,constrain.par),
                               J=J, Jy=Jy, px=px, sparse=sparse))

            parname.all.reg.idx <- parname.all.reg.tidx <-
                parname.reg.tidx <- parname.reg.idx <- c()
            for (p in res$parname.all) {
                ipos <- which((x$par==p))
                tipos <- which(t(x$par==p))
                if (p%in%res$parname) {
                    parname.reg.idx <- c(parname.reg.idx, list(ipos))
                    parname.reg.tidx <- c(parname.reg.tidx, list(tipos))
                }
                parname.all.reg.idx <- c(parname.all.reg.idx, list(ipos))
                parname.all.reg.tidx <- c(parname.all.reg.tidx, list(tipos))
            };
            if (length(parname.reg.idx)>0) {
                names(parname.reg.idx) <- names(parname.reg.tidx) <- res$parname
            }
            if (length(parname.all.reg.idx)>0) {
                names(parname.all.reg.idx) <- names(parname.all.reg.tidx) <- res$parname.all
            }
            covparname.all.idx <- covparname.idx <- c()
            for (p in res$covparname.all) {
                ipos <- which(x$covpar==p)
                if (p%in%res$covparname)
                    covparname.idx <- c(covparname.idx, list(ipos))
                covparname.all.idx <- c(covparname.all.idx, list(ipos))
            };
            if (length(covparname.idx)>0)
                names(covparname.idx) <- res$covparname
            if (length(covparname.all.idx)>0)
                names(covparname.all.idx) <- res$covparname.all

            mparname.all.idx <- mparname.idx <- c()
            for (p in res$mparname.all) {
                ipos <- which(x$mean==p)
                if (p%in%mparname)
                    mparname.idx <- c(mparname.idx, list(ipos))
                mparname.all.idx <- c(mparname.all.idx, list(ipos))
            };
            if (length(mparname.idx)>0)
                names(mparname.idx) <- res$mparname
            if (length(mparname.all.idx)>0)
                names(mparname.all.idx) <- res$mparname.all

            eparname.all.idx <- eparname.idx <- c()
            for (p in res$eparname.all) {
                ipos <- which(x$exfix==p)
                if (p%in%eparname)
                    eparname.idx <- c(eparname.idx, list(ipos))
                eparname.all.idx <- c(eparname.all.idx, list(ipos))
            };
            if (length(eparname.idx)>0)
                names(eparname.idx) <- res$eparname
            if (length(eparname.all.idx)>0)
                names(eparname.all.idx) <- res$eparname.all


            res <- c(res, list(mparname.idx=mparname.idx,
                               covparname.idx=covparname.idx,
                               parname.reg.idx=parname.reg.idx,
                               parname.reg.tidx=parname.reg.tidx,
                               mparname.all.idx=mparname.all.idx,
                               eparname.all.idx=eparname.all.idx,
                               covparname.all.idx=covparname.all.idx,
                               parname.all.reg.idx=parname.all.reg.idx,
                               parname.all.reg.tidx=parname.all.reg.tidx
                               ))

        } else {
            res <- index(x)
        }

        if (zeroones) {
            if (sparse) {
                if (!requireNamespace("Matrix",quietly=TRUE)) stop("package Matrix not available")
                Ik <- Matrix::Diagonal(length(obs))
                Im <- Matrix::Diagonal(ncol(A))
                Kkk <- NULL
                J <- as(J, "sparseMatrix")
                Jy <- as(Jy, "sparseMatrix")
                px <- as(px, "sparseMatrix")

            } else {
                Ik <- diag(nrow=length(obs))
                Im <- diag(nrow=ncol(A))
            }
            Kkk <- NULL


            res[c("Ik","Im","Kkk")] <- NULL
            res <- c(res, list(Ik=Ik, Im=Im, Kkk=Kkk))
        }
        if (deriv && length(P)>0) {
            if (res$npar.mean>0 & mean)
                D <- deriv.lvm(x,meanpar=rep(1,res$npar.mean),zeroones=TRUE)
            else
                D <- deriv.lvm(x,meanpar=NULL,zeroones=TRUE)
            res[c("dA","dP","dv")] <- NULL
            res <- c(res, list(dA=D$dA, dP=D$dP, dv=D$dv))
        }

        if (length(P)>0)
            res <- c(res,mat.lvm(x,res))

        return(res)
    }

##' @export
model.frame.lvmfit <- function(formula, all=FALSE,...) {
  mydata <- formula$data$model.frame
  if (!is.data.frame(mydata) & !is.matrix(mydata))
    return(mydata)
  if (all) return(mydata)
##  xfix <- colnames(mydata)[(colnames(mydata)%in%parlabels(formula$model0,exo=TRUE))]
  xfix <- colnames(mydata)[(colnames(mydata)%in%parlabels(formula$model0))]
  return( mydata[,c(manifest(formula),xfix),drop=FALSE] )
}

##' @export
model.frame.multigroupfit <- function(formula,...) {
  mydata <- formula$model$data
  return(mydata)
}

##' Conformal predicions using locally weighted conformal inference with a split-conformal algorithm
##'
##' @title Conformal prediction
##' @param object Model object (lm, glm or similar with predict method) or formula (lm)
##' @param data data.frame
##' @param newdata New data.frame to make predictions for
##' @param alpha Level of prediction interval
##' @param mad Conditional model (formula) for the MAD (locally-weighted CP)
##' @param ... Additional arguments to lower level functions
##' @return data.frame with fitted (fit), lower (lwr) and upper (upr) predictions bands.
##' @examples
##' set.seed(123)
##' n <- 200
##' x <- seq(0,6,length.out=n)
##' delta <- 3
##' ss <- exp(-1+1.5*cos((x-delta)))
##' ee <- rnorm(n,sd=ss)
##' y <- (x-delta)+3*cos(x+4.5-delta)+ee
##' d <- data.frame(y=y,x=x)
##'
##' newd <- data.frame(x=seq(0,6,length.out=50))
##' cc <- confpred(lm(y~splines::ns(x,knots=c(1,3,5)),data=d), data=d, newdata=newd)
##' if (interactive()) {
##' plot(y~x,pch=16,col=lava::Col("black"),ylim=c(-10,10),xlab="X",ylab="Y")
##' with(cc,
##'      lava::confband(newd$x,lwr,upr,fit,
##'         lwd=3,polygon=TRUE,col=Col("blue"),border=FALSE))
##' }
##' @export
confpred <- function(object,data,newdata=data,alpha=0.05,mad,...) { ## Split-conformal algorithm
    dd <- csplit(data,0.5)
    muhat.new <- predict(object,newdata=newdata, ...) ## New predictions
    if (inherits(object,"formula")) { ## Training
        suppressWarnings(object <- do.call("lm",list(object,data=dd[[1]],...)))
    } else {
        suppressWarnings(object <- update(object, data=dd[[1]], ...))
    }
    muhat.2 <- predict(object,newdata=dd[[2]], ...)   ## Ranking
    R2 <- abs(dd[[2]][,1]-muhat.2)
    if (missing(mad)) mad <- formula(object)
    if (is.null(mad)) {
        mad.new <- rep(1,nrow(newdata))
    } else { ## Locally-weighted conformal inference
        if (names(dd[[2]])[1] %ni% names(newdata)) {
            newdata <- cbind(0,newdata); names(newdata)[1] <- names(dd[[2]])[1]
        }
        X0 <- model.matrix(mad,data=newdata)
        if (inherits(mad,"formula")) {
            X2 <- model.matrix(mad,dd[[2]])
            mad.obj <- stats::lm.fit(x=X2,y=R2)
            mad2 <- X2%*%mad.obj$coefficients
            mad.new <- X0%*%mad.obj$coefficients
        } else {
            mad.obj <- do.call(mad,list(y=R2,x=dd[[2]]))
            mad2 <- predict(mad.obj,newdata=dd[[2]])
            mad.new <- predict(mad.obj,newdata=newdata)
        }
        R2 <- R2/mad2
    }
    k <- ceiling((nrow(data)/2+1)*(1-alpha))
    if (k==0) k <- 1
    if (k>length(R2)) k <- length(R2)
    q <- sort(R2)[k] ## 1-alpha quantile
    lo <- muhat.new - q*mad.new
    up <- muhat.new + q*mad.new
    data.frame(fit=muhat.new,lwr=lo,upr=up)
}

Isqrt <- function(X) {
    eX <- eigen(X);
    with(eX, vectors %*% diag(1/sqrt(values),nrow=length(values)) %*% t(vectors))
}


##' @export
residuals.multigroupfit <- function(object,data=model.frame(object),p=coef(object), k, ...) {
  pp <- modelPar(object,p,...)
  if (!missing(k)) return(residuals(object$model$lvm[[k]],data=data[[k]],p=pp$p[[k]],...))
  res <- c()
  for (i in seq(length(pp$p))) {
    res <- c(res, list(residuals(object$model$lvm[[i]],data=data[[i]],p=pp$p[[i]],...)))
  }
  return(res)
}


##' @export
residuals.lvmfit <- function(object,data=model.frame(object),p=coef(object),...) {
  residuals(Model(object), data=data, p=p, ...)
}

##' @export
residuals.lvm <- function(object,data=model.frame(object),std=FALSE,p=coef(object),...) {
  Y <- setdiff(manifest(object), exogenous(object))
  Pr <- predict(object,p=p,data=data)
  PrY <- Pr[,Y,drop=FALSE]
  ##  y <- endogenous(object)[match(endogenous(object),manifest(object))]
  r <- as.matrix(data[,Y,drop=FALSE]-(PrY))
  res <- r

  if (std) {
    S <- attributes(Pr)$cond.var;
    if (length(Y)>1) {
      res <- r%*%Isqrt(S)
    } else res <- 1/sqrt(S[1,1])*r
  }
  colnames(res) <- colnames(r)
  res
}

##' Polychoric correlation
##'
##' Maximum likelhood estimates of polychoric correlations
##' @param x Variable 1
##' @param y Variable 2
##' @param X Optional covariates
##' @param start Optional starting values
##' @param ... Additional arguments to lower level functions
##' @export
pcor <- function(x,y,X,start,...) {
    if (!requireNamespace("mets",quietly=TRUE)) stop("'mets' package required")

    if (is.numeric(x) && is.numeric(y)) {
        e <- estimate(covariance(lvm(),x~y))
        return(estimate(e,function(p) list(rho=p[5]/(p[3]*p[4])^.5), IC=TRUE))
    }

    n1 <- 1+seq(nlevels(x)-1)
    n2 <- n1[length(n1)]+seq(nlevels(y)-1)
    if (missing(start)) {
        f <- as.formula(ifelse(missing(X),"~1","~X"))
        start <- c(0.5,
                   attr(lava::ordreg(update(f,x~.),fast=TRUE,family=stats::binomial("probit")),"threshold"),
                   attr(lava::ordreg(update(f,y~.),fast=TRUE,family=stats::binomial("probit")),"threshold"))
    }

    ii <- mets::fast.pattern(cbind(as.numeric(x),as.numeric(y)),categories=max(length(unique(x)),length(unique(y))))

    nn <- table(x,y)
    ff <- function(theta) {
        -sum(as.vector(nn)*log(polycor0(theta[1],theta[n1],theta[n2])))
    }
    gg <- function(theta) {
        pp <- polycor0(theta[1],theta[n1],theta[n2],onlyP=FALSE)
        np <- as.vector(nn)/as.vector(pp$p)
        -colSums(apply(pp$dp,2,function(x) np*x))
    }
    nn[nn==0] <- .5
    p0 <- as.vector(nn)/sum(nn)
    logL0 <- sum(as.vector(nn)*log(p0))
    suppressWarnings(op <- nlminb(start,ff,gg))
    cc <- op$par
    names(cc) <- c("rho",paste(rownames(nn),"x",sep=".")[-1], paste(colnames(nn),"y",sep=".")[-1])
    V <- solve(numDeriv::jacobian(function(p) gg(p), cc))

    res <- list(coef=cc, vcov=V, tab=nn, logLik0=logL0, logLik=-ff(cc), n1=n1, n2=n2, opt=op, idx=ii)
    structure(res,class="pcor")
}

##' @export
coef.pcor <- function(object,...) object$coef

##' @export
vcov.pcor <- function(object,...) object$vcov

##' @export
logLik.pcor <- function(object,p=coef(object),...) {
    u <- polycor0(p[1],p[object$n1],p[object$n2],onlyP=TRUE)
    np <- sum(as.vector(object$tab)*log(as.vector(u)))
    nobs <- sum(object$tab)/2
    structure(np,nall=nobs,nobs=nobs,df=length(p),class="logLik")
}

##' @export
print.pcor <- function(x,...) {
    res <- cbind(coef(x),diag(vcov(x))^0.5)
    colnames(res) <- c("Estimate","Std.Err")
    print(res)
    df <- length(x$tab)-nrow(res)
    q <- with(x, 2*(logLik0-logLik) )
    cat("\nDeviance = ", q, ", df = ",df,"\n")
}

##' @export
score.pcor <- function(x,p=coef(x),indiv=FALSE,...) {
    u <- polycor0(p[1],p[x$n1],p[x$n2],onlyP=FALSE)
    if (!indiv) {
        np <- as.vector(x$tab)/as.vector(u$p)
        return(colSums(apply(u$dp,2,function(x) np*x)))
    }
    U <- u$dp;
    U <- apply(u$dp,2,function(x) x/as.vector(u$p))
    Pos <- matrix(0,nrow=prod(dim(x$tab)),ncol=2)
    count <- 0
    for (j in seq(ncol(x$tab)))
        for (i in seq(nrow(x$tab))) {
            count <- count+1
            Pos[count,] <- c(i,j)
        }
    pos <- match(data.frame(t(x$idx$pattern)),data.frame(t(Pos)))
    return(U[pos[x$idx$group+1],])
}



polycor0 <- function(rho,a0,b0,onlyP=TRUE,...) {
    k1 <- length(a0); k2 <- length(b0)
    S <- diag(c(1-rho,1-rho))+rho
    P <- matrix(0,nrow=k1,ncol=k2)
    P1 <- pnorm(a0,sd=1)
    P2 <- pnorm(b0,sd=1)
    set.seed(1)
    for (i in seq(k1))
        for (j in seq(k2)) P[i,j] <- mets::pmvn(lower=c(-Inf,-Inf),upper=c(a0[i],b0[j]),sigma=S)

    PP <- Drho <- matrix(0,nrow=k1+1,ncol=k2+1)
    pmvn0 <- function(i,j,sigma=S) {
        if (i==0 | j==0) return(0)
        if (i==(k1+1) & j==(k2+1)) return(1)
        if (i==(k1+1)) return(P2[j])
        if (j==(k2+1)) return(P1[i])
        P[i,j]
    }

    dpmvn0 <- function(i,j,type=1,k) {
        if (i==0 | j==0) return(0)
        if (i==(k1+1) & j==(k2+1)) return(0)
        if (i==(k1+1)) {
            if (type==3 && k==j) return(dnorm(b0[j]))
            return(0)
        }
        if (j==(k2+1)) {
            if (type==2 && k==i) return(dnorm(a0[i]))
            return(0)
        }
        if (type==1) ## rho
            return(dmvn0(c(a0[i],b0[j]),sigma=S))
        if (type==2) { ## threshold a
            if (k!=i) return(0)
            return(dnorm(a0[i])*pnorm((b0[j]-rho*a0[i])/sqrt(1-rho^2)))
        } ## threshold b
        if (k!=j) return(0)
        dnorm(b0[j])*pnorm((a0[i]-rho*b0[j])/sqrt(1-rho^2))
    }

    for (i in seq(k1+1))
        for (j in seq(k2+1)) {
            PP[i,j] <- pmvn0(i,j) + pmvn0(i-1,j-1) -
                pmvn0(i-1,j) - pmvn0(i,j-1)
            Drho[i,j] <- dpmvn0(i,j) + dpmvn0(i-1,j-1) -
                dpmvn0(i-1,j) - dpmvn0(i,j-1)
        }
    if (onlyP) return(PP)

    Da <- matrix(0,length(PP),k1)
    for (k in seq(k1))
        for (i in seq(k1+1))
            for (j in seq(k2+1)) {
                pos <- i + (k1+1)*(j-1)
                Da[pos,k] <- dpmvn0(i,j,type=2,k=k) + dpmvn0(i-1,j-1,type=2,k=k) -
                    dpmvn0(i-1,j,type=2,k=k) - dpmvn0(i,j-1,type=2,k=k)
            }

    Db <- matrix(0,length(PP),k2)
    for (k in seq(k2))
        for (i in seq(k1+1))
            for (j in seq(k2+1)) {
                pos <- i + (k1+1)*(j-1)
                Db[pos,k] <- dpmvn0(i,j,type=3,k=k) + dpmvn0(i-1,j-1,type=3,k=k) -
                    dpmvn0(i-1,j,type=3,k=k) - dpmvn0(i,j-1,type=3,k=k)
            }

    list(p=PP,dp=cbind(as.vector(Drho),Da,Db))

}

##' Set global options for \code{lava}
##'
##' Extract and set global parameters of \code{lava}. In particular optimization
##' parameters for the \code{estimate} function.
##'
##' \itemize{
##'   \item \code{param}: 'relative' (factor loading and variance of one
##' endogenous variables in each measurement model are fixed to one), 'absolute'
##' (mean and variance of latent variables are set to 0 and 1, respectively),
##' 'hybrid' (intercept of latent variables is fixed to 0, and factor loading of
##' at least one endogenous variable in each measurement model is fixed to 1),
##' 'none' (no constraints are added)
##'   \item \code{layout}: One of 'dot','fdp','circo','twopi','neato','osage'
##'   \item \code{messages}: Set to 0 to disable various output messages
##'   \item ...  }
##'
##' see \code{control} parameter of the \code{estimate} function.
##'
##' @param \dots Arguments
##' @return \code{list} of parameters
##' @author Klaus K. Holst
##' @keywords models
##' @examples
##'
##' \dontrun{
##' lava.options(iter.max=100,messages=0)
##' }
##'
##' @export
lava.options <- function(...) {
    dots <- list(...)
    newopt <- curopt <- get("options",envir=lava.env)
    if (length(dots)==0)
        return(curopt[order(names(curopt))])
    if (length(dots)==1 && is.list(dots[[1]]) && is.null(names(dots))) {
        dots <- dots[[1]]
    }
    idx <- which(names(dots)!="")
    newopt[names(dots)[idx]] <- dots[idx]
    assign("options",newopt,envir=lava.env)
    invisible(curopt)
}

##' @export
gethook <- function(hook="estimate.hooks",...) {
    get(hook,envir=lava.env)
}

char2num <- function(x,...) {
    idx <- grep("^[-]*[0-9\\.]+",x,perl=TRUE,invert=TRUE)
    if (length(idx)>0) x[idx] <- NA
    as.numeric(x)
}

##' @export
addhook <- function(x,hook="estimate.hooks",...) {
    newhooks <- unique(c(gethook(hook),x))
    assign(hook,newhooks,envir=lava.env)
    invisible(newhooks)
}

packagecheck <- function(pkg) {
  nzchar(system.file(package=pkg))
}

versioncheck <- function(pkg="lava",geq,sep=".",...) {
  if (!is.character(pkg)) return(FALSE)
  if (packagecheck(pkg)) {
    xyz <- char2num(strsplit(as.character(utils::packageVersion(pkg)),
                             split=sep,fixed=TRUE)[[1]])
  } else {
    return(FALSE)
  }
  if (missing(geq)) return(xyz)
  for (i in seq(min(length(xyz),length(geq)))) {
    if (xyz[i]>geq[i]) return(TRUE)
    if (xyz[i]<geq[i]) return(FALSE)
  }
  if (length(xyz)>=length(geq)) return(TRUE)
  return(FALSE)
}

lava.env <- new.env()
assign("init.hooks",c(),envir=lava.env)
assign("remove.hooks",c(),envir=lava.env)
assign("estimate.hooks",c(),envir=lava.env)
assign("color.hooks",c(),envir=lava.env)
assign("sim.hooks",c(),envir=lava.env)
assign("post.hooks",c(),envir=lava.env)
assign("print.hooks",c(),envir=lava.env)
assign("plot.post.hooks",c(),envir=lava.env)
assign("plot.hooks",c(),envir=lava.env)
assign("options", list(
                      trace=0,
                      tol=1e-6,
                      gamma=1,
                      backtrack="wolfe",
                      ngamma=0,
                      iter.max=300,
                      eval.max=250,
                      allow.negative.variance=FALSE,
                      progressbarstyle=3,
                      itol=1e-16,
                      cluster.index=packagecheck("mets"),
                      Dmethod="simple",
                      messages=ifelse(interactive(), 1, 0),
                      parallel=TRUE,
                      param="relative",
                      sparse=FALSE,
                      test=TRUE,
                      coef.names=FALSE,
                      constrain=FALSE,
                      graph.proc="beautify",
                      regex=FALSE,
                      min.weight=1e-3,
                      exogenous=TRUE,
                      plot.engine="Rgraphviz",
                      node.color=c(exogenous="lightblue",endogenous="orange",
                                   latent="yellowgreen",transform="lightgray"),
                      edgecolor=FALSE,
                      layout="dot",
                      ## symbols=c("<-","<->"),
                      symbols=c("~","~~"),
                      devel=FALSE,
                      debug=FALSE), envir=lava.env)

context("merge.estimate")

library(testthat)
library(lava)

set.seed(42)
n    <- 100
ic1  <- matrix(rnorm(n * 2), nrow=n, ncol=2) |> scale(center=TRUE, scale=FALSE)
ic2  <- matrix(rnorm(n * 2), nrow=n, ncol=2) |> scale(center=TRUE, scale=FALSE)
e_ic1  <- estimate(coef=c(a=1, b=2), IC=ic1, id=1:n)
e_ic2  <- estimate(coef=c(c=3, d=4), IC=ic2, id=1:n+50)
e_noic  <- estimate(coef=c(e=5, f=6), vcov=diag(2))
e_noic2 <- estimate(coef=c(g=7, h=8), vcov=diag(2) * 2)

test_that("coefficients are concatenated correctly", {
  m <- merge(e_ic1, e_noic)
  expect_equal(coef(m), c(a=1, b=2, e=5, f=6))
})

test_that("vcov has correct dimensions", {
  m <- merge(e_ic1, e_noic)
  expect_equal(dim(vcov(m)), c(4L, 4L))
})

test_that("IC block is non-NA, cross-terms are NA", {
  m <- merge(e_ic1, e_noic)
  V <- vcov(m)
  expect_true(all(!is.na(V[1:2, 1:2])))
  expect_true(all(!is.na(V[3:4, 3:4])))
  expect_true(all(is.na(V[1:2, 3:4])))
  expect_true(all(is.na(V[3:4, 1:2])))
})

test_that("IC block matches direct merge of IC objects", {
  m_ic   <- merge(e_ic1, e_ic2)
  m_both <- merge(e_ic1, e_ic2, e_noic)
  expect_equal(vcov(m_both)[1:4, 1:4], vcov(m_ic))
})

test_that("non-IC block matches individual vcov", {
  m_both <- merge(e_ic1, e_ic2, e_noic)
  expect_equal(vcov(m_both)[5:6, 5:6], vcov(e_noic))
})

test_that("two non-IC objects: correct block placement and NA cross-terms", {
  m <- merge(e_ic1, e_noic, e_noic2)
  V <- vcov(m)
  expect_equal(dim(V), c(6L, 6L))
  expect_true(all(!is.na(V[1:2, 1:2])))
  expect_true(all(!is.na(V[3:4, 3:4])))
  expect_true(all(!is.na(V[5:6, 5:6])))
  expect_true(all(is.na(V[1:2, 3:6])))
  expect_true(all(is.na(V[3:4, c(1:2, 5:6)])))
  expect_true(all(is.na(V[5:6, 1:4])))
})

test_that("all non-IC objects: diagonal blocks filled, cross-terms NA", {
  m <- merge(e_noic, e_noic2)
  V <- vcov(m)
  expect_true(all(!is.na(V[1:2, 1:2])))
  expect_true(all(!is.na(V[3:4, 3:4])))
  expect_true(all(is.na(V[1:2, 3:4])))
  expect_true(all(is.na(V[3:4, 1:2])))
})

test_that("c.estimate with 'extra' attr", {
  e1 <- c(estimate(coef = c(a = 1), vcov = matrix(0.1)), e1.iter = 2)
  e2 <- c(estimate(coef = c(b = 2), vcov = matrix(0.2)), e2.iter = 3)
  e3 <- c(estimate(coef = c(c = 2), vcov = matrix(0.3)))

  e <- c(e1, e2) # two estimate.extra objects
  expect_equal(c(coef(e1), coef(e2)), coef(e))
  expect_equal(c(vcov(e1), vcov(e2)),
               diag(vcov(e)), check.attributes=FALSE)
  expect_equal(c(attr(e1, "extra"), attr(e2, "extra")),
                 attr(e, "extra"))

  e <- c(e1, e3) # estimatee.xtra + estimate
  expect_equal(c(coef(e1), coef(e3)), coef(e))
  expect_equal(c(vcov(e1), vcov(e3)),
               diag(vcov(e)), check.attributes=FALSE)
  expect_equal(attr(e1, "extra"), attr(e, "extra"))

  e <- c(e3, e1) # estimate + estimate.extra
  expect_equal(c(coef(e3), coef(e1)), coef(e))
  expect_equal(c(vcov(e3), vcov(e1)),
               diag(vcov(e)), check.attributes=FALSE)
  expect_equal(attr(e1, "extra"), attr(e, "extra"))

  e <- c(e1, e2, d=2) # estimate + scalar/extra
  expect_equal(c(coef(e1), coef(e2)), coef(e))
  expect_equal(c(vcov(e1), vcov(e2)),
               diag(vcov(e)), check.attributes=FALSE)
  expect_equal(c(attr(e1, "extra"), attr(e2, "extra"), d=2),
               attr(e, "extra"))

  e <- c(e3, c(a = 1, b = 2)) # named scalar vector via unnamed argument
  expect_equal(c(a = 1, b = 2), attr(e, "extra"))

  e <- c(e3, e1.iter = 1) # duplicated name of extra variable
  e <- c(e, e1)
  expect_equal(c(e1.iter = 1, e1.iter = 2), attr(e, "extra"))

  e <- c(e1, converged = FALSE) # boolean extra variable
  # internal conversion to numeric
  expect_equal(c(e1.iter = 2, converged = 0), attr(e, "extra"))
})

test_that("estimate preserves original order with stack=TRUE", {
  set.seed(1)
  n <- 10
  x <- seq_len(n)
  y <- rnorm(n, sd=0.1) + x
  id1 <- as.character(x + 5)
  d <- data.frame(y, x)
  rownames(d) <- id1

  g <- glm(y~x, data=d)
  e <- estimate(g, id=id1)
  X <- model.matrix(g)
  r <- residuals(g)
  ic1 <- cbind(r, r*d$x) %*% solve(crossprod(X))*nrow(X)
  ic2 <- IC(e)
  expect_equivalent(ic1,ic2)
})

test_that("merge preserves order (of first IC)", {
  set.seed(1)
  id1 <- as.character(1:10 + 5)
  id2 <- as.character(1:10)
  a <- estimate(coef=1, IC=scale(rnorm(length(id1))), id=id1)
  b <- estimate(coef=2, IC=scale(rnorm(length(id2))), id=id2)

  expect_equivalent(id1, rownames(IC(a)))
  expect_equivalent(id2, rownames(IC(b)))

  e <- merge(a, b)
  expect_equivalent(rownames(IC(e))[seq_along(id1)],
                    rownames(IC(a)))

  ic1 <- IC(e)[seq_along(id1),1,drop=FALSE]
  expect_equivalent(IC(a)* 3/2, ic1) # same expect for IPW due to incomplete data
})

test_that("cluster.index vs lava native impl.", {
  op <- lava.options(cluster.index = FALSE)
  e1 <- merge(e_ic1, e_ic2)
  lava.options(cluster.index = TRUE)
  e2 <- merge(e_ic1, e_ic2)
  expect_equal(e1, e2)
  lava.options(op)
})

context("Utility functions")

test_that("By", {
    b1 <- By(datasets::CO2,~Treatment+Type,colMeans,~conc)
    b2 <- By(datasets::CO2,c('Treatment','Type'),colMeans,'conc')
    testthat::expect_equivalent(b1,b2)
    ## require('data.table')
    ## t1 <- as.data.frame(data.table(datasets::CO2)[,mean(uptake),by=.(Treatment,Type,conc>500)])
    d0 <- transform(datasets::CO2,conc500=conc>500)
    t1 <- by(d0[,"uptake"],d0[,c("Treatment","Type","conc500")],mean)
    t2 <- By(datasets::CO2,~Treatment+Type+I(conc>500),colMeans,~uptake)
    testthat::expect_true(inherits(t2,"array"))
    testthat::expect_equivalent(sort(t2),sort(t1))
})

test_that("Expand", {
    dd <- Expand(iris, Sepal.Length=2:8, Species=c("virginica","setosa"))
    testthat::expect_identical(levels(iris$Species),levels(dd$Species))
    testthat::expect_true(nrow(dd)==14)
    
    d0 <- datasets::warpbreaks[,c("wool","tension")]
    T <- table(d0)
    d1 <- Expand(T)
    testthat::expect_identical(dim(d0),dim(d1))
    testthat::expect_identical(table(d1),T)

    testthat::expect_identical(expand.grid(1:2,1:2),Expand(1:2,1:2))
    testthat::expect_identical(expand.grid(a=1:2,b=1:2),Expand(a=1:2,b=1:2))
})

test_that("%in%, %nin%, %in.open%, %in.closed%", {
  expect_equivalent(1:10 %ni% c(1,5,10),
                    !c(TRUE,FALSE,FALSE,FALSE,TRUE,
                      FALSE,FALSE,FALSE,FALSE,TRUE))

  expect_false(1 %in.open% c(1,4))
  expect_true(1 %in.closed% c(1,4))
  expect_error(1 %in.open% c(1,4,5))
  expect_error(1 %in.closed% c(1,4,5))


})

test_that("formulas", {
    f <- toformula(c('y1','y2'),'x'%++%1:5)
    ff <- getoutcome(f)
    testthat::expect_equivalent(trim(ff,all=TRUE),"c(y1,y2)")
    testthat::expect_true(length(attr(ff,'x'))==5)
})

test_that("trim", {
    testthat::expect_true(length(grep(" ",trim(" test ")))==0)    
    testthat::expect_true(length(gregexpr(" ",trim(" t e s t "))[[1]])==3)
    testthat::expect_true(length(grep(" ",trim(" t e s t ",all=TRUE)))==0)
})

test_that("Matrix operations:", {
    ## vec operator
    testthat::expect_equivalent(vec(diag(3)),c(1,0,0,0,1,0,0,0,1))
    testthat::expect_true(nrow(vec(diag(3),matrix=TRUE))==9)

    ## commutaion matrix
    A <- matrix(1:16 ,ncol=4)
    K <- commutation(A)
    testthat::expect_equivalent(K%*%as.vector(A),vec(t(A),matrix=TRUE))
})

test_that("blockdiag:", {
    ## Block diagonal
    A <- diag(3)+1
    B <- blockdiag(A,A,A,pad=NA)
    testthat::expect_equivalent(dim(B),c(9,9))
    testthat::expect_true(sum(is.na(B))==81-27)

    B <- matrix(1, ncol=2, nrow=2)
    res <- blockdiag(A, B, B)
    expect_equivalent(res[1:3, 1:3], A)
    expect_equivalent(res[4:5, 4:5], B)
    expect_equivalent(res[6:7, 6:7], B)
    expect_true(sum(res==0) == (7*7 - 3*3 - 2*2 - 2*2))
})

test_that("wrapvev", {
    testthat::expect_equivalent(wrapvec(5,2),c(3,4,5,1,2))
    testthat::expect_equivalent(wrapvec(seq(1:5),-1),c(5,1,2,3,4))
})

test_that("revdiag", {
    A <- revdiag(1:3)
    testthat::expect_equivalent(A,matrix(c(0,0,1,0,2,0,3,0,0),3))
    testthat::expect_equivalent(1:3,revdiag(A))
    revdiag(A) <- 4
    testthat::expect_equivalent(rep(4,3),revdiag(A))
    diag(A) <- 0
    offdiag(A) <- 5
    testthat::expect_true(sum(offdiag(A))==6*5)
})

test_that("Inv, matrix inverse", {
    A <- matrix(0,3,3)
    offdiag(A,type=3) <- 1:6
    B <- crossprod(A)
    
    testthat::expect_equivalent(solve(A),Inverse(A))
    testthat::expect_equivalent(det(B),attr(Inverse(B,chol=TRUE),"det"))
})

test_that("Grep", {
  d <- Grep(iris, "Sepal")
  expect_true(nrow(d) == nrow(iris))
  expect_equal(colnames(d), c("Sepal.Length", "Sepal.Width"))
  m <- as.matrix(iris[,1:4])
  d <- Grep(m, "Sepal")
  expect_true(nrow(d) == nrow(iris))
  expect_equal(colnames(d), c("Sepal.Length", "Sepal.Width"))

  d <- Grep(iris, "Sepal", subset=FALSE)
  expect_equivalent(colnames(d), c("index", "name"))
  expect_true(nrow(d)==2L)
})

test_that("All the rest", {
    testthat::expect_false(lava:::versioncheck(NULL))
    testthat::expect_true(lava:::versioncheck("lava",c(1,4,1)))

    op <- lava.options(debug=TRUE)
    testthat::expect_true(lava.options()$debug)
    ## lava.options(op)
     lava.options(debug=FALSE)

    A <- diag(2); colnames(A) <- c("a","b")    
    testthat::expect_output(lava:::printmany(A,A,2,rownames=c("A","B"),bothrows=FALSE),"a b")
    testthat::expect_output(lava:::printmany(A,A[1,,drop=FALSE],2,rownames=c("A","B"),bothrows=FALSE),"a b")
    testthat::expect_output(lava:::printmany(A,A,2,rownames=c("A","B"),name1="no.1",name2="no.2",
                            bothrows=TRUE),"no.1")
})

test_that("napass.0", {
  n <- 10
  d <- data.frame(y=rnorm(n), a=rbinom(n,1,0.5)*2-1)
  idx1 <- c(1,3,5)
  idx2 <- c(1,3,8,9)
  d$y[idx1] <- NA
  d$a[idx2] <- NA
  d0 <- na.pass0(d)
  expect_true(nrow(d0)==n)
  expect_true(all(d0$a[idx2]==0))
  expect_true(all(d0$a[idx1]==0))
})

test_that("strip_bracket", {
  expect_equal(strip_bracket("[a]+[b]"), "[a]+[b]")
  expect_equal(strip_bracket("[-[a]-[b]]"), "-[a]-[b]")
  expect_equal(strip_bracket("[[a]]"), "[a]")
  expect_equal(strip_bracket("[2[a]+[b]]"), "2[a]+[b]")
  # vector
  expect_equal(strip_bracket(c("[a]", "[[a]]")), c("a", "[a]"))
})

test_that("frobnorm", {
  x <- rmvn0(100, sigma=diag(3))
  y <- rmvn0(100, sigma=diag(3))
  expect_equal(frobnorm(x, y), sum((x-y)^2)^.5)
})

context("Utility functions")

test_that("By", {
    b1 <- By(datasets::CO2,~Treatment+Type,colMeans,~conc)
    b2 <- By(datasets::CO2,c('Treatment','Type'),colMeans,'conc')
    testthat::expect_equivalent(b1,b2)
    ## require('data.table')
    ## t1 <- as.data.frame(data.table(datasets::CO2)[,mean(uptake),by=.(Treatment,Type,conc>500)])
    d0 <- transform(datasets::CO2,conc500=conc>500)
    t1 <- by(d0[,"uptake"],d0[,c("Treatment","Type","conc500")],mean)
    t2 <- By(datasets::CO2,~Treatment+Type+I(conc>500),colMeans,~uptake)
    testthat::expect_true(inherits(t2,"array"))
    testthat::expect_equivalent(sort(t2),sort(t1))
})

test_that("Expand", {
    dd <- Expand(iris, Sepal.Length=2:8, Species=c("virginica","setosa"))
    testthat::expect_identical(levels(iris$Species),levels(dd$Species))
    testthat::expect_true(nrow(dd)==14)
    
    d0 <- datasets::warpbreaks[,c("wool","tension")]
    T <- table(d0)
    d1 <- Expand(T)
    testthat::expect_identical(dim(d0),dim(d1))
    testthat::expect_identical(table(d1),T)

    testthat::expect_identical(expand.grid(1:2,1:2),Expand(1:2,1:2))
    testthat::expect_identical(expand.grid(a=1:2,b=1:2),Expand(a=1:2,b=1:2))
})

test_that("%in%, %nin%, %in.open%, %in.closed%", {
  expect_equivalent(1:10 %ni% c(1,5,10),
                    !c(TRUE,FALSE,FALSE,FALSE,TRUE,
                      FALSE,FALSE,FALSE,FALSE,TRUE))

  expect_false(1 %in.open% c(1,4))
  expect_true(1 %in.closed% c(1,4))
  expect_error(1 %in.open% c(1,4,5))
  expect_error(1 %in.closed% c(1,4,5))


})

test_that("formulas", {
    f <- toformula(c('y1','y2'),'x'%++%1:5)
    ff <- getoutcome(f)
    testthat::expect_equivalent(trim(ff,all=TRUE),"c(y1,y2)")
    testthat::expect_true(length(attr(ff,'x'))==5)
})

test_that("trim", {
    testthat::expect_true(length(grep(" ",trim(" test ")))==0)    
    testthat::expect_true(length(gregexpr(" ",trim(" t e s t "))[[1]])==3)
    testthat::expect_true(length(grep(" ",trim(" t e s t ",all=TRUE)))==0)
})

test_that("Grep", {
  d <- Grep(iris, "Sepal")
  expect_true(nrow(d) == nrow(iris))
  expect_equal(colnames(d), c("Sepal.Length", "Sepal.Width"))
  m <- as.matrix(iris[,1:4])
  d <- Grep(m, "Sepal")
  expect_true(nrow(d) == nrow(iris))
  expect_equal(colnames(d), c("Sepal.Length", "Sepal.Width"))

  d <- Grep(iris, "Sepal", subset=FALSE)
  expect_equivalent(colnames(d), c("index", "name"))
  expect_true(nrow(d)==2L)
})

test_that("All the rest", {
    testthat::expect_false(lava:::versioncheck(NULL))
    testthat::expect_true(lava:::versioncheck("lava",c(1,4,1)))

    op <- lava.options(debug=TRUE)
    testthat::expect_true(lava.options()$debug)
    ## lava.options(op)
     lava.options(debug=FALSE)

    A <- diag(2); colnames(A) <- c("a","b")    
    testthat::expect_output(lava:::printmany(A,A,2,rownames=c("A","B"),bothrows=FALSE),"a b")
    testthat::expect_output(lava:::printmany(A,A[1,,drop=FALSE],2,rownames=c("A","B"),bothrows=FALSE),"a b")
    testthat::expect_output(lava:::printmany(A,A,2,rownames=c("A","B"),name1="no.1",name2="no.2",
                            bothrows=TRUE),"no.1")
})

test_that("na.pass0", {
  n <- 10
  d0 <- data.frame(y=abs(rnorm(n))+1, a=rbinom(n,1,0.5)*2-1)
  d <- d0
  idx1 <- c(1,3,5)
  idx2 <- c(1,3,8,9)
  d$y[idx1] <- NA
  d$a[idx2] <- NA
  d1 <- na.pass0(d)
  expect_true(nrow(d1)==n)
  expect_true(all(d1$y[idx1]==0))
  expect_true(all(d1$a[idx2]==0))
  expect_true(all(d1$a[setdiff(1:10, idx2)]!=0))
  expect_true(all(d1$y[setdiff(1:10, idx1)]!=0))
  y1 <- na.pass0(d$y)
  expect_equivalent(y1, d1$y)
  a1 <- na.pass0(d$a)
  expect_equivalent(a1, d1$a)

  ## Setting entire rows to zero where ever there is a missing entry
  d2 <- na.pass0(d, row.wise = TRUE)
  expect_true(nrow(d2)==n)
  expect_true(all(d2$y[c(idx1,idx2)]==0))
  expect_true(all(d2$a[c(idx1,idx2)]==0))
  expect_true(all(d2$a[setdiff(1:10, c(idx1, idx2))]!=0))
  expect_true(all(d2$y[setdiff(1:10, c(idx1, idx2))]!=0))
})

test_that("strip_bracket", {
  expect_equal(strip_bracket("[a]+[b]"), "[a]+[b]")
  expect_equal(strip_bracket("[-[a]-[b]]"), "-[a]-[b]")
  expect_equal(strip_bracket("[[a]]"), "[a]")
  expect_equal(strip_bracket("[2[a]+[b]]"), "2[a]+[b]")
  # vector
  expect_equal(strip_bracket(c("[a]", "[[a]]")), c("a", "[a]"))
})

test_that("frobnorm", {
  x <- rmvn0(100, sigma=diag(3))
  y <- rmvn0(100, sigma=diag(3))
  expect_equal(frobnorm(x, y), sum((x-y)^2)^.5)
})

context("Constraints")

test_that("Simple linear constraint",{
    m1 <- lvm(y[m:v] ~ f(x,beta)+f(z,beta2))
    constrain(m1,beta2~psi) <- function(x) 2*x
    testthat::expect_output(summary(m1),"Non-linear constraints:")

    lava:::matrices.lvm(m1,1:2,0,3)
    d1 <- sim(m1,100)
    e1 <- estimate(m1,d1)
    testthat::expect_output(print(e1),"y~x")
    testthat::expect_warning(e1,NA)

    testthat::expect_true((constraints(e1)[1]-coef(lm(y~x+z,d1))["z"])^2<1e-9)
    s <- summary(e1)
    testthat::expect_output(print(s),"Non-linear constraints:")

    testthat::expect_equivalent(dim(coef(s)), c(length(coef(e1))+1,4))
    testthat::expect_equivalent(dim(coef(e1,2)), c(length(coef(e1)),4))
})

test_that("constrain (Fishers z-transform)", {
    set.seed(1)
    m <- lvm(c(y1[m1:v1],y2[m2:v2])~x)
    covariance(m,y1~y2) <- "C"
    d <- sim(m,100)
    e <- estimate(m,d)
    constrain(e,rho~C+v1+v2) <-
        function(x) x[1]/(x[2]*x[3])^0.5
    cc1 <- parameter(correlation(e))
    cc2 <- constraints(e)
    testthat::expect_equivalent(cc2["rho",1],cc1["y1~y2",1])
    constrain(e,z~C+v1+v2) <- function(x) {
        f <- function(p) p[1]/sqrt(p[2]*p[3])
        res <- atanh(f(x))
        df <- function(p) c(1/sqrt(p[2]*p[3]), -f(p)/(2*p[2]), -f(p)/(2*p[3]))
        datanh <- function(r) 1/(1-r^2)
        attributes(res)$grad <- function(p) datanh(f(p))*df(p)
        attributes(res)$inv <- tanh
        return(res)
    }
    cc2 <- constraints(e)
    testthat::expect_equal(cc2["z",2],0.1)
    testthat::expect_equivalent(cc2["inv(z)",1],cc1["y1~y2",1])
})

test_that("Non-linear in exogenous variables", {
    d <- data.frame(x=1:5,y=c(0.5,1.5,2,3,3.5))
    m <- lvm(y[m] ~ 1)
    addvar(m) <- ~x
    parameter(m) <- ~a+b
    constrain(m,m~a+b+x) <- function(z) z[1]+z[2]*z[3]
    e <- estimate(m,d,control=list(method="NR"))
    testthat::expect_true(mean(coef(lm(y~x,d))-coef(e)[c("a","b")])^2<1e-3)
})

if (lava:::versioncheck('mets', c(1,0)))
test_that("Probit constraints", {
    x <- transform(data.frame(lava:::rmvn0(1000,sigma=0.5*diag(2)+0.5)),
                   X1=as.numeric(cut(X1,breaks=3))-1,X2=as.numeric(cut(X2,breaks=3))-1)
    m <- covariance(lvm(),X1~X2)
    ordinal(m,K=3,constrain=list("t1","t2")) <- ~X1
    ordinal(m,K=3,constrain=list("t1","t2")) <- ~X2
    ##        e <- estimate(m,x)
    e <- estimate(list(m,m),list(x[1:500,],x[501:1000,]),estimator="normal")
    res <- estimate(e)
    testthat::expect_true(length(coef(res))==4)
})

test_that("Multiple group constraints I", {
    m1 <- lvm(y[m:v] ~ f(x,beta)+f(z,beta2))
    d1 <- sim(m1,500,seed=1); d2 <- sim(m1,500,seed=2)
    ##coef(estimate(m1,d1))
    constrain(m1,beta2~psi) <- function(x) 2*x
    m2 <- lvm(y[m:v] ~ f(x,beta2) + z)
    constrain(m2,beta2~psi) <- function(x) 2*x
    mg <- multigroup(list(m1,m2),list(d1,d2))
    ee <- estimate(mg)
    testthat::expect_true(length(coef(ee))==5)
    testthat::expect_equivalent(constraints(ee)[1],2*coef(ee)["psi@1"]) # Est
    testthat::expect_equivalent(constraints(ee)[2],2*coef(ee,2)[[1]]["psi",2]) # Std.Err
})

test_that("Multiple group constraints II", {
  data("twindata",package="lava")
  twinwide <- reshape(twindata,direction="wide",
                      idvar="id",timevar="twinnum")
  l <- lvm(~bw.1+bw.2)
  covariance(l) <- bw.1 ~ bw.2
  e <- estimate(l,subset(twinwide,zyg.1=="MZ"),control=list(method="NR"))
  B <- cbind(1,-1); colnames(B) <- c("bw.1,bw.1","bw.2,bw.2")
  colnames(B) <- gsub(",",lava.options()$symbols[2],colnames(B))
  lava::compare(e,contrast=B)
  B2 <- rbind(c(1,-1,0,0),c(0,0,1,-1))
  colnames(B2) <- c("bw.1","bw.2","bw.1,bw.1","bw.2,bw.2")
  colnames(B2) <- gsub(",",lava.options()$symbols[2],colnames(B2))

  lava::compare(e,contrast=B2)

  l <- lvm(~bw.1+bw.2)
  covariance(l) <- bw.1 ~ bw.2
  intercept(l,~bw.1+bw.2) <- "m"
  covariance(l,~bw.1+bw.2) <- "s"
  covariance(l,bw.1~bw.2) <- "r1"
  l2 <- l
  covariance(l2,bw.1~bw.2) <- "r2"

  DZ <- subset(twinwide,zyg.1=="MZ")
  MZ <- subset(twinwide,zyg.1=="DZ")
  ## e <- estimate(l,MZ)
  ## e2 <- estimate(l2,DZ)

  parameter(l) <- ~r2
  parameter(l2) <- ~r1
  ee <- estimate(list(MZ=l,DZ=l2),list(MZ,DZ),control=list(method="NR",tol=1e-9,constrain=FALSE))
  testthat::expect_true(mean(score(ee)^2)<1e-9)

  constrain(ee,h~r2+r1) <- function(x) 2*(x[1]-x[2])
  ce <- constraints(ee)
  testthat::expect_equivalent(constraints(ee)[1],2*diff(coef(ee)[3:4]))
  testthat::expect_true(length(coef(ee))==4)
  testthat::expect_true(nrow(ce)==1)
  testthat::expect_true(all(!is.na(ce)))
})

##' Set global options for \code{lava}
##'
##' Extract and set global parameters of \code{lava}. In particular optimization
##' parameters for the \code{estimate} function.
##'
##' \itemize{
##'   \item \code{param}: 'relative' (factor loading and variance of one
##' endogenous variables in each measurement model are fixed to one), 'absolute'
##' (mean and variance of latent variables are set to 0 and 1, respectively),
##' 'hybrid' (intercept of latent variables is fixed to 0, and factor loading of
##' at least one endogenous variable in each measurement model is fixed to 1),
##' 'none' (no constraints are added)
##'   \item \code{layout}: One of 'dot','fdp','circo','twopi','neato','osage'
##'   \item \code{messages}: Set to 0 to disable various output messages
##'   \item ...  }
##'
##' see \code{control} parameter of the \code{estimate} function.
##'
##' @param \dots Arguments
##' @return \code{list} of parameters
##' @author Klaus K. Holst
##' @keywords models
##' @examples
##'
##' \dontrun{
##' lava.options(iter.max=100,messages=0)
##' }
##'
##' @export
lava.options <- function(...) {
    dots <- list(...)
    newopt <- curopt <- get("options",envir=lava.env)
    if (length(dots)==0)
        return(curopt[order(names(curopt))])
    if (length(dots)==1 && is.list(dots[[1]]) && is.null(names(dots))) {
        dots <- dots[[1]]
    }
    idx <- which(names(dots)!="")
    newopt[names(dots)[idx]] <- dots[idx]
    assign("options",newopt,envir=lava.env)
    invisible(curopt)
}

##' @export
gethook <- function(hook="estimate.hooks",...) {
    get(hook,envir=lava.env)
}

char2num <- function(x,...) {
    idx <- grep("^[-]*[0-9\\.]+",x,perl=TRUE,invert=TRUE)
    if (length(idx)>0) x[idx] <- NA
    as.numeric(x)
}

##' @export
addhook <- function(x,hook="estimate.hooks",...) {
    newhooks <- unique(c(gethook(hook),x))
    assign(hook,newhooks,envir=lava.env)
    invisible(newhooks)
}

packagecheck <- function(pkg) {
  nzchar(system.file(package=pkg))
}

versioncheck <- function(pkg="lava",geq,sep=".",...) {
  if (!is.character(pkg)) return(FALSE)
  if (packagecheck(pkg)) {
    xyz <- char2num(strsplit(as.character(utils::packageVersion(pkg)),
                             split=sep,fixed=TRUE)[[1]])
  } else {
    return(FALSE)
  }
  if (missing(geq)) return(xyz)
  for (i in seq(min(length(xyz),length(geq)))) {
    if (xyz[i]>geq[i]) return(TRUE)
    if (xyz[i]<geq[i]) return(FALSE)
  }
  if (length(xyz)>=length(geq)) return(TRUE)
  return(FALSE)
}

lava.env <- new.env()
assign("init.hooks",c(),envir=lava.env)
assign("remove.hooks",c(),envir=lava.env)
assign("estimate.hooks",c(),envir=lava.env)
assign("color.hooks",c(),envir=lava.env)
assign("sim.hooks",c(),envir=lava.env)
assign("post.hooks",c(),envir=lava.env)
assign("print.hooks",c(),envir=lava.env)
assign("plot.post.hooks",c(),envir=lava.env)
assign("plot.hooks",c(),envir=lava.env)
assign("options", list(
                      trace=0,
                      tol=1e-6,
                      gamma=1,
                      backtrack="wolfe",
                      ngamma=0,
                      iter.max=300,
                      eval.max=250,
                      allow.negative.variance=FALSE,
                      progressbarstyle=3,
                      itol=(.Machine$double.eps)**.5, # Tolerance for pseudo-inverses
                      cluster.index=packagecheck("mets"), # Whether to use mets::cluster.index
                      Dmethod="simple", # Type of numerical derivative for optimization problems
                      messages=ifelse(interactive(), 1, 0), # extra messages
                      parallel=TRUE, # parallelization enabled
                      param="relative", # parametrization type for latent variable models
                      sparse=FALSE, #
                      test=TRUE,
                      coef.names=FALSE,
                      constrain=FALSE,
                      graph.proc="beautify",
                      regex=FALSE,
                      min.weight=1e-3,
                      exogenous=TRUE,
                      plot.engine="Rgraphviz", # rgraphviz, igraph, visnetwork, ...
                      node.color=c(exogenous="lightblue",endogenous="orange",
                                   latent="yellowgreen",transform="lightgray"),
                      edgecolor=FALSE,
                      layout="dot", # graphviz graph layout function
                      ## symbols=c("<-","<->"),
                      symbols=c("~","~~"),
                      devel=FALSE, # development flag, extra debug information
                      check.ic=TRUE, # check if influence functions has zero-mean
                      check.ic.tol=1e-6, # tolerance for checking zero-mean mean(ic)/rms(ic) > check.ic.tol
                      debug=FALSE),
       envir=lava.env)

#' @export
estimate <- function(x, ...) UseMethod("estimate")

#' Influence function based inference
#'
#' Primary tool for obtaining parameter estimates with robust (sandwich)
#' standard errors, applying the delta method, and testing linear hypotheses.
#' The function returns an object of class `estimate` which serves as a
#' general container for parameter estimates and their influence functions
#' (IFs). Three calling conventions are supported:
#'
#' - `estimate(x, ...)` -- extract estimates from a model object
#' - `estimate(coef=, IC=, ...)` -- construct from coefficients and IF matrix
#' - `estimate(coef=, vcov=, ...)` -- construct from coefficients and
#'   covariance matrix
#'
#' @param x model object (`glm`, `lvmfit`, ...) or an existing `estimate`
#'   object. When two model objects are supplied (e.g., `estimate(g, g0)`) a
#'   likelihood-ratio test is performed.
#' @param f transformation of model parameters. Accepts several input types:
#'   - A **function** `f(p)` or `f(p, data)`: applies the delta method.
#'     When `f` returns a named list the names are used as parameter labels.
#'   - A **matrix**: used as a contrast (linear combination) matrix.
#'   - A **numeric vector** of parameter indices: converted to a contrast
#'     that selects and differences those parameters.
#'   - A **list** of indices: each element selects one parameter.
#'   - **Character** expressions: supports wildcards (`"?"`, `"*"`) and
#'     arithmetic on parameter names (e.g., `"z" - "x"`, `2 * "z" - 3 * "x"`).
#' @param ... additional arguments to lower level functions
#' @param data `data.frame` used by `f` when the transformation depends on
#'   covariates (see `average`). Defaults to `model.frame(x)`.
#' @param id (optional) cluster identifier. Can be a vector of cluster IDs, a
#'   one-sided formula (evaluated in `data`), a single character column name,
#'   or a logical scalar (`TRUE` for one-to-one matching, `FALSE` for
#'   independence). When supplied, the IF is aggregated within clusters to
#'   produce cluster-robust standard errors.
#' @param stack if `TRUE` (default) the influence function contributions are
#'   summed within each cluster defined by `id`. Set to `FALSE` to keep the
#'   un-stacked (per-observation) decomposition.
#' @param average if `TRUE` the function computes the standardized
#'   (marginalized) estimate \eqn{\hat\Psi = P_n f(X; \hat\theta)}, i.e.,
#'   the empirical mean of `f(p, data)` over all rows of `data`. The
#'   influence function accounts for both the empirical averaging and the
#'   parameter estimation uncertainty (see Details).
#' @param subset (optional) logical vector, expression evaluated in `data`, or
#'   column name. When used together with `average = TRUE`, the average is
#'   conditioned on the subpopulation where `subset` is `TRUE`, yielding a
#'   conditional marginalized estimate.
#' @param level level of confidence limits (default 0.95)
#' @param IC if `TRUE` (default) the influence function matrix is estimated and
#'   stored in the returned object (extract with the [IC] method). Can also be
#'   a user-supplied IF matrix (one row per observation, one column per
#'   parameter), which is used directly instead of estimating it from `x`.
#' @param type type of small-sample correction for cluster-robust variance.
#'   One of:
#'   - `"robust"` (default): no correction.
#'   - `"df"`: applies \eqn{n/(n-p)} correction (Mancl & DeRouen, 2001).
#'   - `"mbn"`: Morel-Bokossa-Neerchal (2003) correction.
#'   - `"hc3"`: leverage-adjusted HC3-type correction (blended with
#'     `var.adj`).
#'   - `"hc4"`: Cribari-Neto (2004) leverage-adjusted correction.
#' @param var.adj blending parameter for the HC3 leverage adjustment
#'   (default 0.25). Controls the weight between observation-level empirical
#'   leverage and the average leverage \eqn{p/n}.
#' @param keep (optional) index of parameters to keep from final result.
#'   Accepts integer indices, character names, or (with `regex = TRUE`)
#'   perl-compatible regular expressions.
#' @param use (optional) index of parameters to use in calculations. The
#'   selected parameters are first extracted (via `keep`) and then the
#'   remaining arguments (`f`, `contrast`, etc.) are applied to this subset.
#' @param regex if `TRUE` use perl-compatible regular expressions for `keep`
#'   and `use` arguments
#' @param ignore.case ignore case in regular expressions
#' @param contrast (optional) contrast matrix for a final Wald test. When
#'   supplied together with `null`, tests \eqn{H_0: B\theta = b_0}.
#' @param null (optional) null hypothesis vector \eqn{b_0} to test against
#'   (default 0)
#' @param vcov (optional) covariance matrix of parameter estimates, or a
#'   logical. If `TRUE`, [stats::vcov] is used to obtain the (model-based)
#'   covariance matrix from `x`, yielding non-robust standard errors. If a
#'   matrix is supplied it is used directly. When omitted or `FALSE`, robust
#'   standard errors are computed from the influence function.
#' @param coef (optional) named parameter vector. Used instead of
#'   `coef(x)` when constructing an `estimate` object without a model.
#' @param df degrees of freedom for t-based inference (default: `NULL` for
#'   Gaussian approximation; when set, confidence intervals and p-values use
#'   the t-distribution with `df` degrees of freedom)
#' @param print (optional) custom print function for the resulting `estimate`
#'   object
#' @param labels (optional) character vector of coefficient names
#' @param label.width (optional) max display width of labels
#' @param only.coef (Deprecated) if `TRUE` only the coefficient matrix is
#'   returned. Use `parameter(estimate(...))` instead.
#' @param back.transform (optional) function applied to the point estimates
#'   and confidence interval bounds *after* inference is performed on the
#'   original scale. Useful for variance-stabilizing transformations, e.g.,
#'   compute CIs on the `atanh` (Fisher z) scale and back-transform with
#'   `tanh`.
#'
#' @details
#'
#' # Influence functions and robust standard errors
#'
#' An estimator \eqn{\widehat{\theta}} is *regular and asymptotically
#' linear* (RAL) when it admits the iid decomposition
#' \deqn{\sqrt{n}(\widehat{\theta}-\theta) =
#' \frac{1}{\sqrt{n}}\sum_{i=1}^n \mathrm{IC}(Z_i; P) + o_p(1)}
#' where \eqn{\mathrm{IC}} is the unique *influence function* satisfying
#' \eqn{E\{\mathrm{IC}(Z; P)\} = 0}. By the central limit theorem
#' \deqn{\sqrt{n}(\widehat{\theta}-\theta)
#' \overset{d}{\longrightarrow}
#' N(0,\; \mathrm{Var}\{\mathrm{IC}(Z; P)\})}
#' and the asymptotic variance is consistently estimated by the empirical
#' variance of the plugin IF estimate, yielding robust (sandwich) standard
#' errors. The estimated IF can be extracted with the [IC] method.
#'
#' # Parameter transformations (delta method)
#'
#' When `f` is a function \eqn{\phi: R^p \to R^m}, the delta method is
#' applied:
#' \deqn{\sqrt{n}\{\phi(\widehat{\theta}) - \phi(\theta)\} =
#' \frac{1}{\sqrt{n}}\sum_{i=1}^n
#' \nabla\phi(\theta)\,\mathrm{IC}(Z_i; P) + o_p(1)}
#' Derivatives are computed numerically via [numDeriv::jacobian] unless the
#' function returns an attribute `"grad"` with the analytic Jacobian.
#'
#' Alternatively, `estimate` objects support direct arithmetic operations
#' (e.g., `a * b`, `exp(a)`, `a^b`) which apply the delta method with
#' *exact* (analytical) derivatives computed automatically. This influence
#' function calculus allows building complex transformations from simple
#' building blocks without numerical differentiation. See the last example
#' section ("influence function calculus") and
#' `vignette("influencefunction", package = "lava")` for details.
#'
#' # Averaging and marginalization
#'
#' When `average = TRUE` and `f(p, data)` depends on covariates, the
#' target parameter is the standardized (marginalized) estimate
#' \eqn{\Psi = E\{f(X;\theta)\}}. The IF for the averaged estimate
#' accounts for both the empirical averaging and parameter estimation
#' uncertainty:
#' \deqn{\mathrm{IC}_\Psi(Z; P) = f(X;\theta) - \Psi +
#' [E\nabla_\theta f(X;\theta)]\,\phi(Z; P)}
#' When `subset` is also specified, the average is conditioned on the
#' subpopulation, yielding a conditional marginalized estimate.
#'
#' # Cluster-robust standard errors
#'
#' When `id` is supplied, the per-observation IF contributions are summed
#' within clusters (when `stack = TRUE`), producing the cluster-level IF
#' \eqn{\widetilde{\mathrm{IC}}(Z_i; P) = \sum_{k=1}^{N_i}
#' \frac{n}{N}\mathrm{IC}(Z_{ik}; P)}.
#' The resulting variance estimate is equivalent to the GEE working
#' independence sandwich estimator.
#'
#' For full theoretical background and worked examples see
#' `vignette("influencefunction", package = "lava")`.
#'
#' @export
#' @export estimate.default
#' @examples
#'
#' ## Simulation from logistic regression model
#' m <- lvm(y~x+z);
#' distribution(m,y~x) <- binomial.lvm("logit")
#' d <- sim(m,1000)
#' g <- glm(y~z+x,data=d,family=binomial())
#' g0 <- glm(y~1,data=d,family=binomial())
#'
#' ## LRT
#' estimate(g, g0)
#'
#' ## Plain estimates (robust standard errors)
#' estimate(g)
#'
#' ## Testing contrasts
#' estimate(g, null=0)
#' estimate(g, rbind(c(1,1,0), c(1,0,2)))
#' estimate(g, rbind(c(1,1,0), c(1,0,2)), null=c(1,2))
#' estimate(g, 2:3) ## same as cbind(0,1,-1)
#' estimate(g, as.list(2:3)) ## same as rbind(c(0,1,0),c(0,0,1))
#' ## Alternative syntax
#' estimate(g, "z", "z"-"x", 2*"z"-3*"x")
#' estimate(g, "?")  ## Wildcards
#' estimate(g, "*Int*", "z")
#' estimate(g, "1", "2"-"3", null = c(0,1))
#' estimate(g, 2, 3)
#'
#' ## Usual (non-robust) confidence intervals
#' estimate(g, vcov=TRUE)
#' estimate(g, vcov=vcov(g))
#'
#' ## Transformations
#' estimate(g, function(p) p[1]+p[2])
#'
#' ## Multiple parameters
#' e <- estimate(g, function(p) c(p[1]+p[2], p[1]*p[2]))
#' e
#' vcov(e)
#'
#' ## Label new parameters
#' estimate(g, function(p) list("a1"=p[1]+p[2], "b1"=p[1]*p[2]))
#' #'
#' ## Multiple group
#' m <- lvm(y~x)
#' m <- baptize(m)
#' d2 <- d1 <- sim(m,50,seed=1)
#' e <- estimate(list(m,m),list(d1,d2))
#' estimate(e) ## Wrong
#' ee <- estimate(e, id=rep(seq(nrow(d1)), 2)) ## Clustered
#' ee
#' estimate(lm(y~x,d1))
#'
#' ## Marginalize
#' f <- function(p,data)
#'   list(p0=expit(p["(Intercept)"] + p["z"]*data[,"z"]),
#'        p1=expit(p["(Intercept)"] + p["x"] + p["z"]*data[,"z"]))
#' e <- estimate(g, f, average=TRUE)
#' e
#' estimate(e,diff)
#' estimate(e,cbind(1,1))
#'
#' ## Clusters and subset (conditional marginal effects)
#' d$id <- rep(seq(nrow(d)/4),each=4)
#' estimate(g,function(p,data)
#'          list(p0=expit(p[1] + p["z"]*data[,"z"])),
#'          subset=d$z>0, id=d$id, average=TRUE)
#'
#' ## More examples with clusters:
#' m <- lvm(c(y1,y2,y3)~u+x)
#' d <- sim(m,10)
#' l1 <- glm(y1~x,data=d)
#' l2 <- glm(y2~x,data=d)
#' l3 <- glm(y3~x,data=d)
#'
#' ## Some random id-numbers
#' id1 <- c(1,1,4,1,3,1,2,3,4,5)
#' id2 <- c(1,2,3,4,5,6,7,8,1,1)
#' id3 <- seq(10)
#'
#' ## Un-stacked and stacked i.i.d. decomposition
#' IC(estimate(l1,id=id1,stack=FALSE))
#' IC(estimate(l1,id=id1))
#'
#' ## Combined i.i.d. decomposition
#' e1 <- estimate(l1,id=id1)
#' e2 <- estimate(l2,id=id2)
#' e3 <- estimate(l3,id=id3)
#' (a2 <- merge(e1,e2,e3))
#'
#' ## If all models were estimated on the same data we could use the
#' ## syntax:
#' ## Reduce(merge,estimate(list(l1,l2,l3)))
#'
#' ## Same:
#' IC(a1 <- merge(l1,l2,l3,id=list(id1,id2,id3)))
#'
#' IC(merge(l1,l2,l3,id=TRUE)) # one-to-one (same clusters)
#' IC(merge(l1,l2,l3,id=FALSE)) # independence
#'
#'
#' # ------ influence function calculus -------
#' a <- estimate(coef = c("a" = 0.5), IC = scale(rnorm(10), scale=FALSE), id = 1:10)
#' b <- estimate(coef = c("b" = 0.8), IC = scale(rnorm(10), scale=FALSE), id = 1:10)
#'
#' e <- c(a, b) # merge
#' merge(a, b)
#' c(e1=a, b) # naming of par
#' labels(e, c("p1", "p2")) # renaming parameters
#' e["a"] # subset
#' subset(e, "a")
#'
#' # pipes
#' # c(a, b) |>
#' #  transform(function(x) x^2) |>
#' #  subset("a") |>
#' #  labels("sq")
#'
#' # Parameter transformation with automatic calculation of derivatives
#' a * b
#' (3 * cos(a) / sqrt(b) + 1) / a
#' expit(c(a,b))
#' c(sum=sum(e), sum2=a+b,
#'   prod=prod(e), prod2=a*b)
#' e %*% e # inner prod.
#' c(1, 2) %*% e
#' c(pow = a^b)
#' a^c(0.5, 2)
#' c(b=e["a"] * e["b"] / a, also.b=e["b"])
#'
#' B <- rbind(c(1,-1), c(1,0), c(0,1))
#' B %*% e
#' e == 1 # wald-test, null-hypothesis H0: b=1
#' e == c(1,2)
#' B %*% e == 1
#' @aliases estimate estimate.default merge.estimate
#' @aliases estimate.mlm
#' @seealso [estimate.array], [merge.estimate], [contr], [parsedesign],
#'   [pairwise.diff], `summary.estimate`, `coef.estimate`, `vcov.estimate`,
#'   `transform.estimate`, `labels.estimate`, `IC.estimate`
#' @method estimate default
#' @export
estimate.default <- function(x=NULL, f=NULL, ..., data, id,
                             stack=TRUE,
                             average=FALSE, subset,
                             level=0.95,
                             IC=TRUE,
                             type=c("robust", "df", "mbn"),
                             var.adj,
                             keep, use,
                             regex=FALSE, ignore.case=FALSE,
                             contrast, null, vcov, coef,
                             df=NULL,
                             print=NULL, labels, label.width,
                             only.coef=FALSE, back.transform=NULL) {
  cl <- match.call(expand.dots = TRUE)
  cal <- match.call()
  if ("iid" %in% names(cl)) {
    stop("The 'iid' argument is obsolete. Please use the 'IC' argument")
  }
  if ("R" %in% names(cl) || "null.sim" %in% names(cl)) {
    stop("The 'R' and 'null.sim' arguments have been removed. ")
  }
  if ("robust" %in% names(cl)) {
    warning(
      "The 'robust' argument is deprecated and ignored. ",
      "Robust standard errors are now always computed. ",
      "Use the 'vcov' argument to compute model-based SEs."
    )
  }
  if (!missing(use)) {
    p0 <- c(
      "f", "contrast", "only.coef",
      "subset", "average", "keep", "labels", "null"
    )
    cl0 <- cl
    cl0[c("use", p0)] <- NULL
    cl0$keep <- use
    cl$x <- eval(cl0, parent.frame())
    cl[c("vcov", "use")] <- NULL
    return(eval(cl, parent.frame()))
  }
  expr <- suppressWarnings(inherits(try(f, silent=TRUE), "try-error"))
  if (!missing(coef)) {
    pp <- coef
  } else {
    pp <- suppressWarnings(try(stats::coef(x), silent = TRUE))
    if (inherits(x, "survreg") && length(pp) < NROW(x$var)) {
      pp <- c(pp, scale=x$scale)
    }
  }
  if (expr || is.character(f) || (is.numeric(f)
    && !is.matrix(f))) { ## || is.call(f)) {
    dots <- lapply(substitute(placeholder(...))[-1], function(x) x)
    args <- c(list(
      coef = names(pp),
      x = substitute(f),
      regex = regex
    ), dots)
    f <- do.call(parsedesign, args)
  }
  contrast.transform <- TRUE  # parameter estimates should be transformed
                              # according to contrast matrix
  if (!is.null(f) && !is.function(f)) {
    if (!(is.matrix(f) || is.vector(f)))
      return(compare(x, f, ...)) ## LRT
    if (missing(contrast)) contrast <- f
    if (length(f)==1L && (is.logical(f)||is.numeric(f)) && !f) { # f = 0 or FALSE
      contrast.transform <- FALSE # Wald-test and do not alter the parameter estimates
    }
    f <- NULL
  }

  if (lava.options()$cluster.index) {
    if (!requireNamespace("mets", quietly=TRUE)) stop("'mets' package required")
  }
  if (missing(data))
    data <- tryCatch(model.frame(x), error=function(...) NULL)
  alpha <- 1 - level
  alpha.str <- paste(c(alpha/2, 1 -alpha/2)*100, "", sep="%")
  nn <- NULL
  if (((is.logical(IC) && IC) || length(IC)>0) &&
      (missing(vcov) || is.null(vcov) ||
       (is.logical(vcov) && vcov[1]==FALSE && !is.na(vcov[1])))) {
    ## If user supplied vcov, then don't estimate IC
    if (!is.logical(IC)) {
      ic_theta <- cbind(IC)
      if (NCOL(ic_theta) != length(pp)) {
        warning("Wrong dimension of influence function IC")
      }
      if (lava.options()$check.ic) {
        check_ic_mean_zero(ic_theta)
      }
      IC <- TRUE
    } else {
      suppressWarnings(ic_theta <- IC(x))
    }
  } else {
    if (!is.null(x) && (missing(vcov) ||
                        (is.logical(vcov) && !is.na(vcov)[1])))
      suppressWarnings(vcov <- stats::vcov(x))
    ic_theta <- NULL
  }

  if (any(is.na(ic_theta))) {
    ## Rescale each column according to I(obs)/pr(obs)
    for (i in seq_len(NCOL(ic_theta))) {
      pr <- mean(!is.na(ic_theta[, i]))
      ic_theta[, i] <- ic_theta[, i]/pr
    }
    ic_theta[is.na(ic_theta)] <- 0
  }

  if (!missing(subset)) {
    e <- substitute(subset)
    expr <- suppressWarnings(inherits(try(subset, silent=TRUE), "try-error"))
    if (expr) subset <- eval(e, envir=data)
    if (is.character(subset)) subset <- data[, subset]
    if (is.numeric(subset)) subset <- subset > 0
  }
  idstack <- NULL
  ## Preserve id from 'estimate' object
  if (missing(id) && inherits(x, "estimate") && !is.null(x$id))
    id <- x$id
  if (!missing(id) && IC) {
    if (is.null(ic_theta)) stop("'IC' method needed")
    nprev <- nrow(ic_theta)
    if (inherits(id, "formula")) {
      id <- interaction(get_all_vars(id, data))
    }
    ## e <- substitute(id)
    ## expr <- suppressWarnings(inherits(try(id,silent=TRUE),"try-error"))
    ## if (expr) id <- eval(e,envir=data)
    ##if (!is.null(data)) id <- eval(e, data)
    if (is.logical(id) && length(id)==1) {
      id <- if(is.null(ic_theta)) seq_len(nrow(data)) else seq_len(nprev)
      stack <- FALSE
    }
    if (is.character(id) && length(id)==1)
      id <- data[, id, drop=TRUE]
    if (!is.null(ic_theta)) {
      if (length(id)!=nprev) {
        if (!is.null(x$na.action) &&
            (length(id)==length(x$na.action) + nprev)) {
          warning("Applying na.action")
          id <- id[-x$na.action]
        } else stop("Dimensions of i.i.d decomposition and 'id' does not agree")
      }
    } else {
      if (length(id)!=nrow(data)) {
        if (!is.null(x$na.action) &&
            (length(id)==length(x$na.action)+nrow(data))) {
          warning("Applying na.action")
          id <- id[-x$na.action]
        } else stop("Dimensions of IC and 'id' does not agree")
      }
    }
    if (stack) {
      N <- nrow(ic_theta)
      clidx <- NULL
      atr <- attributes(ic_theta)
      atr$dimnames <- NULL
      atr$dim <- NULL
      if (!lava.options()$cluster.index) {
        ic_theta <- matrix(unlist(by(ic_theta, id, colSums)),
                           byrow=TRUE, ncol=ncol(ic_theta))
        attributes(ic_theta)[names(atr)] <- atr
        idstack <- sort(unique(id))
      } else {
        clidx <- mets::cluster.index(id, mat=ic_theta, return.all=TRUE)
        ic_theta <- with(clidx, X)
        attributes(ic_theta)[names(atr)] <- atr
        idstack <- id[as.vector(clidx$firstclustid)+1]
      }
      ic_theta <- ic_theta*NROW(ic_theta)/length(id)
      if (is.null(attributes(ic_theta)$N)) {
        attributes(ic_theta)$N <- N
      }
    } else idstack <- id
  } else {
    if (!is.null(data)) idstack <- rownames(data)
  }
  if (!is.null(ic_theta) && (length(idstack)==nrow(ic_theta))) {
    rownames(ic_theta) <- idstack
  }
  if (inherits(x, "lm") && family(x)$family == "gaussian"
      && is.null(df) && !missing(vcov)) {
    df <- x$df.residual
  }
  if (!is.null(ic_theta) && (missing(vcov) || is.null(vcov))) {
    V <- var_ic(ic_theta)
    ## Small-sample corrections for clustered data
    K <- NROW(ic_theta)
    N <- attributes(ic_theta)$N
    if (is.null(N)) N <- K
    p <- NCOL(ic_theta)
    adj0 <- K/(K-p) ## Mancl & DeRouen, 2001
    adj1 <- K/(K-1) ## Mancl & DeRouen, 2001
    adj2 <- (N-1)/(N-p)*(K/(K-1)) ## Morel,Bokossa & Neerchal, 2003
    if (tolower(type[1])=="mbn" && !is.null(attributes(ic_theta)$bread)) {
      V0 <- V
      iI0 <- attributes(ic_theta)$bread
      I0 <- Inverse(iI0)
      delta <- min(0.5, p / (K - p))
      phi <- max(1, tr(I0%*%V0)*adj2/p)
      V <- adj2*V0 + delta*phi*iI0
    }
    if (tolower(type[1])=="df") {
      V <- adj0*V
    }
    if (tolower(type[1])=="df1") {
      V <- adj1*V
    }
    if (tolower(type[1])=="df2") {
      V <- adj2*V
    }
    if (tolower(type[1])%in%c("hc3", "hc4")) {
      ic <- cbind(ic_theta)
      S <- Inverse(crossprod(ic), tol=sqrt(.Machine$double.eps))
      h_emp <- rowSums((ic %*% S) * ic) # empirical h, lev.
      n <- nrow(ic)
      ## h_emp <- pmin(h_emp, 0.99) * (n-1)/n  # Truncate leverage to prevent division by zero
      if (tolower(type[1])=="hc3") {
        ## phi_norm <- sqrt(rowSums(ic^2))
        ## ex_kurt <- (mean((phi_norm - mean(phi_norm))^4) / var(phi_norm)^2) - 3
        ## alpha <- exp(-max(0, ex_kurt) / 25)
        v.alpha <- ifelse(missing(var.adj), 0.25, var.adj)
        h <- v.alpha * h_emp + (1 - v.alpha) * (ncol(ic) / n)
        adj <- 1 / (1 - h)
      } else {
        ## Cribari-Neto (2004)
        delta <- pmin(1.5, n/ncol(ic) * h_emp)
        adj <- 1 / (1 - h_emp)**delta
      }
      for (i in seq_len(NCOL(ic))) ic[, i] <- ic[, i] * adj
      V <- var_ic(ic)
    }
  } else {
    if (!missing(vcov)) {
      if (length(vcov) == 1 && is.na(vcov)) {
        vcov <- matrix(NA, length(pp), length(pp))
      }
      V <- cbind(vcov)
    } else {
      suppressWarnings(V <- stats::vcov(x))
    }
  }

  derivative <- NULL
  if (!is.null(f)) {
    form <- names(formals(f))
    dots <- ("..."%in%names(form))
    form0 <- setdiff(form, "...")
    parname <- "p"

    if (!is.null(form)) parname <- form[1] # unless .Primitive
    if (length(form0)==1 && !(form0%in%c("object", "data"))) {
      ##names(formals(f))[1] <- "p"
      parname <- form0
    }
    if (!is.null(ic_theta)) {
      arglist <- c(list(object=x, data=data, p=vec(pp)), list(...))
      names(arglist)[3] <- parname
    } else {
      arglist <- c(list(object=x, p=vec(pp)), list(...))
      names(arglist)[2] <- parname
    }
    if (!dots) {
      arglist <- arglist[intersect(form0, names(arglist))]
    }
    newf <- NULL
    if (length(form)==0) {
      arglist <- list(vec(pp))
      newf <- function(...) do.call("f", list(...))
      val <- do.call("f", arglist)
    } else {
      val <- do.call("f", arglist)
      if (is.list(val)) {
        nn <- names(val)
        val <- do.call("cbind", val)
        newf <- function(...) do.call("cbind", f(...))
      }
    }
    k <- NCOL(val)
    N <- NROW(val)
    D <- attributes(val)$grad
    if (!is.null(D)) derivative <- D
    if (is.null(D)) {
      D <- numDeriv::jacobian(function(p, ...) {
        if (length(form)==0) arglist[[1]] <- p
        else arglist[[parname]] <- p
        if (is.null(newf))
          return(do.call("f", arglist))
        return(do.call("newf", arglist)) }, pp)
    }
    if (is.null(ic_theta)) {
      pp <- structure(as.vector(val), names=names(val))
      V <- D%*%V%*%t(D)
    } else {
      if (!average || (N<NROW(data))) {  ## transformation not depending on data
        pp <- structure(as.vector(val), names=names(val))
        ic_theta <- ic_theta%*%t(D)
        V <- var_ic(ic_theta)
      } else {
        if (k>1) { ## More than one parameter (and depends on data)
          if (!missing(subset)) { ## Conditional estimate
            val <- apply(val, 2, function(x) x*subset)
          }
          D0 <- matrix(nrow=k, ncol=length(pp))
          for (i in seq_len(k)) {
            D1 <- D[seq(N)+(i-1)*N, , drop=FALSE]
            if (!missing(subset)) ## Conditional estimate
              D1 <- apply(D1, 2, function(x) x*subset)
            D0[i, ] <- colMeans(D1)
          }
          D <- D0
          ic2 <- ic_theta%*%t(D)
        } else { ## Single parameter
          if (!missing(subset)) { ## Conditional estimate
            val <- val*subset
            D <- apply(rbind(D), 2, function(x) x*subset)
          }
          D <- colMeans(rbind(D))
          ic2 <- ic_theta%*%D
        }
        pp <- vec(colMeans(cbind(val)))
        ic1 <- (cbind(val)-rbind(pp)%x%cbind(rep(1, N)))
        if (NROW(ic_theta)==NROW(ic1)) {
          rownames(ic1) <- rownames(ic_theta)
        }
        if (!missing(id)) {
          if (!lava.options()$cluster.index)
            ic1 <- matrix(unlist(by(ic1, id, colSums)),
                          byrow=TRUE, ncol=ncol(ic1))
          else {
            ic1 <- mets::cluster.index(id, mat=ic1, return.all=FALSE)
          }
          ic1 <- ic1 * NROW(ic1) / length(id)
        }
        if (!missing(subset)) { ## Conditional estimate
          phat <- mean(subset)
          ic3 <- cbind(-1/phat^2 * (subset-phat)) ## check
          if (!missing(id)) {
            if (!lava.options()$cluster.index) {
              ic3 <- matrix(unlist(by(ic3, id, colSums)),
                            byrow=TRUE, ncol=ncol(ic3))
            } else {
              ic3 <- mets::cluster.index(id, mat=ic3, return.all=FALSE)
            }
          }
          ic3 <- ic3*NROW(ic3)/length(id)
          ic_theta <- (ic1+ic2)/phat + rbind(pp)%x%ic3
          pp <- pp/phat
          V <- var_ic(ic_theta)
        } else {
          if (nrow(ic1)!=nrow(ic2)) {
            message("Assuming independence between model iid decomposition and new data frame") #nolint
            V <- var_ic(ic1) + var_ic(ic2)
          } else {
            ic_theta <- ic1+ic2
            V <- var_ic(ic_theta)
          }
        }
      }
    }
  }

  if (is.null(V)) {
    res <- cbind(pp, NA, NA, NA, NA)
  } else {
    if (length(pp)==1)
      res <- rbind(c(pp, diag(V)^0.5))
    else
      res <- cbind(pp, diag(V)^0.5)
    beta0 <- res[, 1]

    if (!missing(null) && missing(contrast))
      beta0 <- beta0-null
    if (!is.null(df)) {
      za <- qt(1-alpha/2, df=df)
      pval <- 2*pt(abs(res[, 1]/res[, 2]), df=df, lower.tail=FALSE)
    } else {
      za <- qnorm(1-alpha/2)
      pval <- 2*pnorm(abs(res[, 1]/res[, 2]), lower.tail=FALSE)
    }
    res <- cbind(res, res[, 1]-za*res[, 2], res[, 1] + za*res[, 2], pval)
  }
  colnames(res) <- c("Estimate", "Std.Err", alpha.str, "P-value")

  if (nrow(res)>0)
    if (!is.null(nn)) {
      rownames(res) <- nn
    } else {
      nn <- attributes(res)$varnames
      if (!is.null(nn))
        rownames(res) <- nn
      if (is.null(rownames(res)))
        rownames(res) <- paste0("p", seq_len(nrow(res)))
    }

  if (NROW(res)==0L) {
    coefs <- NULL
  } else {
    coefs <- res[, 1, drop=TRUE]
    names(coefs) <- rownames(res)
  }
  res <- structure(list(coef=coefs, coefmat=res, vcov=V,
                        IC=NULL, print=print, id=idstack),
                   class="estimate")
  if (IC) ## && is.null(back.transform))
    res$IC <- ic_theta
  if (length(coefs)==0L) return(res)

  if (!missing(contrast) || !missing(null)) {
    p <- length(res$coef)
    if (missing(contrast)) contrast <- diag(nrow=p)
    if (missing(null)) null <- 0
    if (is.vector(contrast) || is.list(contrast)) {
      contrast <- contr(contrast, names(res$coef), ...)
    }
    cc <- compare(res, contrast=contrast, null=null,
                  vcov=V, level=level, df=df)
    class(res) <- "list"
    res <- structure(c(res, list(compare=cc)), class="estimate")
    if (contrast.transform) {
      if (!is.null(df)) {
        pval <- with(cc, pt(abs(estimate[, 1]-null)/estimate[, 2],
                            df=df, lower.tail=FALSE)*2)
      } else {
        pval <- with(cc, pnorm(abs(estimate[, 1]-null)/estimate[, 2],
                               lower.tail=FALSE)*2)
      }
      res$coefmat <- with(cc, cbind(estimate, pval))
      colnames(res$coefmat)[5] <- "P-value"
      rownames(res$coefmat) <- cc$cnames
      if (!is.null(res$IC)) {
        res$IC <- res$IC%*%t(contrast)
        colnames(res$IC) <- cc$cnames
      }
      res$compare$estimate <- NULL
      res$coef <- res$compare$coef
      res$vcov <- res$compare$vcov

      names(res$coef) <- strip_bracket(rownames(res$coefmat))
      rownames(res$coefmat) <- names(res$coef)
    }
  }

  if (!is.null(back.transform)) {
    res$coefmat[, c(1, 3, 4)] <- do.call(back.transform,
                                         list(res$coefmat[, c(1, 3, 4)]))
    res$coefmat[, 2] <- NA
  }

  if (!missing(keep) && !is.null(keep)) {
    if (is.character(keep)) {
      if (regex) {
        nn <- rownames(res$coefmat)
        keep <- unlist(lapply(keep, function(x) {
          grep(x, nn,
               perl = TRUE,
               ignore.case = ignore.case
               )
        }))
      } else {
        keep <- match(keep, rownames(res$coefmat))
      }
    }
    res$coef <- res$coef[keep]
    res$coefmat <- res$coefmat[keep, , drop=FALSE]
    if (!is.null(res$IC)) res$IC <- res$IC[, keep, drop=FALSE]
    res$vcov <- res$vcov[keep, keep, drop=FALSE]
  }

  res <- labels.estimate(
    object = res, str = labels, label.width = label.width
  )

  if (only.coef) {
    .Deprecated(
      msg = paste0(
        "The 'only.coef' argument is deprecated. ",
        "Use `parameter(estimate(...))` instead."
      )
    )
    return(res$coefmat)
  }
  res$call <- cal
  res$back.transform <- back.transform
  res$n <- nrow(data)
  res$ncluster <- if (!is.null(ic_theta)) nrow(ic_theta) else nrow(data)
  res$derivative <- derivative
  return(structure(res, class="estimate"))
}

##' @export
print.estimate <- function(x, type=0L, digits=4L, width=25L,
                           std.error=TRUE, p.value=TRUE,
                           sep=cli::symbol[["line"]],
                           sep.which,
                           sep.labels=NULL,
                           indent=" ", unique.names=TRUE,
                           na.print="", ...) {

  if (!is.null(x$print)) {
    x$print(x, digits=digits, width=width, ...)
    return(invisible(x))
  }
  if (type>0 && !is.null(x$call)) {
    cat("Call: ")
    print(x$call)
    print(cli::rule(width=min(cli::console_width(),60)))
  }
  if (type>0) {
    if (!is.null(x[["n"]]) && !is.null(x[["k"]])) {
      cat("n = ", x[["n"]], ", clusters = ", x[["k"]], "\n\n", sep="")
    } else {
      if (!is.null(x[["n"]])) {
        cat("n = ", x[["n"]], "\n\n", sep="")
      }
      if (!is.null(x[["k"]])) {
        cat("n = ", x[["k"]], "\n\n", sep="")
      }
    }
  }

  cc <- x$coefmat
  if (!is.null(rownames(cc)) && unique.names)
    rownames(cc) <- make.unique(
      unlist(lapply(rownames(cc),
                    function(x) toString(x, width=width)))
    )
  if (!std.error) cc <- cc[, -2, drop=FALSE]
  if (!p.value) cc <- cc[, -ncol(cc), drop=FALSE]

  sep.pos <- c()
  if (missing(sep.which) && !is.null(x$model.index)) {
    sep.which <- unlist(lapply(x$model.index,
                               function(x)
                                 tail(x, 1)))[-length(x$model.index)]
  }
  if (missing(sep.which)) sep.which <- NULL

  if (!is.null(sep.which)) {
    sep0 <- 0%in%sep.which
    if (sep0)
      sep.which <- setdiff(sep.which, 0)
    cc0 <- c()
    sep.which <- c(0, sep.which, nrow(cc))
    N <- length(sep.which)-1
    for (i in seq(N)) {
      if ((sep.which[i]+1)<=nrow(cc))
        cc0 <- rbind(cc0, cc[seq(sep.which[i]+1, sep.which[i+1]), , drop=FALSE])
      if (i<N) {
        cc0 <- rbind(cc0, NA)
        sep.pos <- c(sep.pos, nrow(cc0))
      }
    }
    if (sep0) {
      sep.pos <- c(1, sep.pos+1)
      cc0 <- rbind(NA, cc0)
    }
    cc <- cc0
  }
  if (!is.null(sep.labels)) {
    sep.labels <- rep(sep.labels, length.out=length(sep.pos))
    rownames(cc)[sep.pos] <- sep.labels
    rownames(cc)[-sep.pos] <- paste0(indent, rownames(cc)[-sep.pos])
  } else {
    if (length(sep.pos)>0)
      rownames(cc)[sep.pos] <- rep(paste0(rep(sep, max(nchar(rownames(cc)))),
                                          collapse=""), length(sep.pos))
  }
  print(cc, digits=digits, na.print=na.print, ...)

  if (!is.null(x$compare)) {
    print(cli::rule(width=min(cli::console_width(),60)))
    cat(x$compare$method[3], "\n")
    cat(paste(" ", x$compare$method[-(1:3)], collapse="\n"), "\n")
    if (length(x$compare$method)>=4) {
      out <- character()
      out <- with(x$compare, c(out, paste(names(statistic),
                                          "=", format(round(statistic, 4)))))
      out <- with(x$compare, c(out, paste(names(parameter),
                                          "=", format(round(parameter, 3)))))
      fp  <- with(x$compare, format.pval(p.value, digits = digits))
      out <- c(out, paste("p-value", if (substr(fp, 1L, 1L) == "<")
                                       fp else paste("=", fp)))
      cat(" ", strwrap(paste(out, collapse = ", ")), sep = "\n")
    }
  }
}

#' @export
vcov.estimate <- function(object, list=FALSE, ...) {
  res <- object$vcov
  nn <- names(coef(object, ...))
  if (list && !is.null(object$model.index)) {
    return(lapply(object$model.index, function(x) object$vcov[x, x]))
  }
  dimnames(res) <- list(nn, nn)
  res
}

#' @export
coef.estimate <- function(object,
                          mat=FALSE,
                          list=FALSE,
                          ...) {
  if (mat) return(object$coefmat)
  if (!is.null(object$back.transform)) {
    warning("Return estimates on original scale (before 'back.transform')")
  }
  if (list && !is.null(object$model.index)) {
    return(lapply(object$model.index, function(x) object$coef[x]))
  }
  object$coef
}

with_unique_warnings <- function(expr) {
  # utility function that prevents the same warning being cast more than once
  seen <- character(0)
  withCallingHandlers(expr, warning = function(w) {
    msg <- conditionMessage(w)
    if (msg %in% seen) {
      invokeRestart("muffleWarning")
    } else {
      seen <<- c(seen, msg)
    }
  })
}

#' @export
summary.estimate <- function(object,
                             contrast,
                             ...) {
  with_unique_warnings({
    p <- coef(object)
    if (missing(contrast)) contrast <- diag(1,nrow=length(p))#as.list(seq_along(p))
    test <- estimate(coef=p,
                     vcov=vcov(object),
                     f = FALSE,
                     contrast = contrast,
                     ...)
    class(test) <- "NULL"
    test$compare <- test$compare
    object <- test[c("coef", "coefmat", "vcov", "call",
                     "ncluster", "model.index", "compare")]
    class(object) <- "summary.estimate"
    object
  })
}

#' @export
coef.summary.estimate <- function(object, ...) {
  object$coefmat
}

#' @export
transform.estimate <- function(`_data`, ...) {
  estimate(`_data`, ...)
}

#' @export
labels.estimate <- function(object, str, label.width, ...) {
  if (!missing(str)) {
    names(object$coef) <- str
    if (!is.null(object$IC))
      colnames(object$IC) <- str
    if (!is.null(object$vcov))
      colnames(object$vcov) <- rownames(object$vcov) <- str
    rownames(object$coefmat) <- str
  }
  if (!missing(label.width)) {
    rownames(object$coefmat) <- make.unique(
      unlist(lapply(rownames(object$coefmat),
                    function(x) toString(x, width = label.width)))
    )
  }
  return(object)
}

#' @export
parameter.estimate <- function(x, ...) {
  return(x$coefmat)
}

#' @export
print.summary.estimate <- function(x, ...) {
  print.estimate(x, type=2L, ...)
}

#' @export
IC.estimate <- function(x, ...) {
  if (is.null(x$IC)) return(NULL)
  dimn <- dimnames(x$IC)
  if (!is.null(dimn)) {
    dimn[[2]] <- names(coef(x))
  } else {
    dimn <- list(NULL, names(coef(x)))
  }
  structure(x$IC, dimnames=dimn)
}

#' @export
model.frame.estimate <- function(formula, ...) {
  NULL
}

context("Censored observations")

library("survival")

test_that("right censoring", {
  set.seed(1)
  n <- 500
  x <- rnorm(n)
  y0 <- rnorm(n) + x
  rcens <- rexp(n, 1)
  y <- pmin(y0, rcens)
  rstatus <- y0 < rcens
  d <- data.frame(
    y = y,
    x = x,
    status = rstatus)
  d$ys <- with(d, Surv(y, status, type="right"))

  ms <- lvm(ys ~ x)
  e <- estimate(ms, data = d)
  es <- survreg(ys ~ x, data=d, dist="gaussian")

  expect_true(abs(logLik(e)-logLik(es))<1e-3)
  expect_true(mean((coef(e)[1:2]-coef(es))^2)<1e-3)
})

test_that("left censoring", {
  set.seed(1)
  n <- 500
  x <- rnorm(n)
  y0 <- rnorm(n) + x
  lcens <- -rexp(n, 1)
  y <- pmax(y0, lcens)
  lstatus <- y0 > lcens
  d <- data.frame(
    y = y,
    x = x,
    status = lstatus)
  d$ys <- with(d, Surv(y, status, type="left"))

  ms <- lvm(ys ~ x)
  e <- estimate(ms, data = d)
  es <- survreg(ys ~ x, data=d, dist="gaussian")

  expect_true(abs(logLik(e)-logLik(es))<1e-3)
  expect_true(mean((coef(e)[1:2]-coef(es))^2)<1e-3)
})

test_that("left and right censoring", {
  set.seed(1)
  n <- 500
  x <- rnorm(n)
  y0 <- rnorm(n) + x
  lcens <- -rexp(n, 2)
  rcens <- rexp(n, 1)
  y <- pmax(y0, lcens)
  y <- pmin(y, rcens)
  lstatus <- y0>lcens
  rstatus <- y0<rcens
  d <- data.frame(
    y = y,
    x = x,
    lstatus = lstatus,
    rstatus = rstatus)
  yleft <- rep(-Inf, n)
  yleft[d$lstatus] <- d$y[d$lstatus]
  yright <- rep(Inf, n)
  yright[d$rstatus] <- d$y[d$rstatus]
  d$ys <- Surv(yleft, yright, type="interval2")

  ms <- lvm(ys ~ x)
  e <- estimate(ms, data = d)
  es <- survreg(ys ~ x, data=d, dist="gaussian")

  expect_true(abs(logLik(e)-logLik(es))<1e-3)
  expect_true(mean((coef(e)[1:2]-coef(es))^2)<1e-3)
})

test_that("interval censoring", {
  set.seed(1)
  n <- 1e4
  x <- rnorm(n) # covariate (no effects.lvmfit)
  true_times <- rweibull(n, shape = 1.5, scale = 50) # true event timestamp
  # Define fixed inspection points every 10 time units
  visits <- seq(0, 100, by = 10)
  # left and right interval (obs: right censored after last visit)
  yleft <- sapply(true_times, function(t) max(visits[visits <= t]))
  yright <- sapply(true_times, function(t) min(visits[visits > t], Inf))

  d <- data.frame(
    x = x,
    yleft = yleft,
    yright = yright
  )
  d$ys <- Surv(yleft, yright, type="interval2")

  ms <- lvm(ys ~ x)
  e <- estimate(ms, data = d)
  es <- survreg(ys ~ x, data=d, dist="gaussian")

  expect_true(abs(logLik(e)-logLik(es))<1e-3)
  expect_true(mean((coef(e)[1:2]-coef(es))^2)<1e-3)
})

context("estimate.default and estimate_calculus issues")

library(testthat)
library(lava)


test_that("estimate.array mean formula is correct", {
  ## The variance computation in estimate.array currently evaluates
  ## mean(y - mean(y)^2) instead of mean((y - mean(y))^2).
  set.seed(1)
  n <- 100
  x <- matrix(rnorm(2*n, mean = 3, sd = 2), ncol = 2)
  e <- estimate(x)
  expected <- apply(x, 2, function(y) mean(y))
  expect_equivalent(coef(e), expected, tolerance = 1e-8)
  x0 <- apply(x, 2, function(y) y-mean(y))
  var_expected <- crossprod(x0) / n^2
  expect_equivalent(vcov(e), var_expected, tolerance = 1e-8)
})

test_that("estimate.array variance formula is correct", {
  ## The variance computation in estimate.array currently evaluates
  ## mean(y - mean(y)^2) instead of mean((y - mean(y))^2).
  set.seed(1)
  x <- matrix(rnorm(200, mean = 3, sd = 2), ncol = 2)
  e <- estimate(x, type = "variance")
  expected <- apply(x, 2, function(y) mean((y - mean(y))^2))
  expect_equivalent(coef(e), expected, tolerance = 1e-8)
})

test_that("estimate.array variance handles NA without poisoning", {
  ## Inner mean(y) inside the variance formula does not pass na.rm=TRUE.
  set.seed(2)
  x <- matrix(rnorm(100), ncol = 2)
  x[1, 1] <- NA
  expect_no_error(e <- estimate(x, type = "variance"))
  expect_true(all(is.finite(coef(e))))
})

context("merge.estimate")

library(testthat)
library(lava)

set.seed(42)
n    <- 100
ic1  <- matrix(rnorm(n * 2), nrow=n, ncol=2) |> scale(center=TRUE, scale=FALSE)
ic2  <- matrix(rnorm(n * 2), nrow=n, ncol=2) |> scale(center=TRUE, scale=FALSE)
e_ic1  <- estimate(coef=c(a=1, b=2), IC=ic1, id=1:n)
e_ic2  <- estimate(coef=c(c=3, d=4), IC=ic2, id=1:n+50)
e_noic  <- estimate(coef=c(e=5, f=6), vcov=diag(2))
e_noic2 <- estimate(coef=c(g=7, h=8), vcov=diag(2) * 2)

test_that("coefficients are concatenated correctly", {
  m <- merge(e_ic1, e_noic)
  expect_equal(coef(m), c(a=1, b=2, e=5, f=6))
})

test_that("vcov has correct dimensions", {
  m <- merge(e_ic1, e_noic)
  expect_equal(dim(vcov(m)), c(4L, 4L))
})

test_that("IC block is non-NA, cross-terms are NA", {
  m <- merge(e_ic1, e_noic)
  V <- vcov(m)
  expect_true(all(!is.na(V[1:2, 1:2])))
  expect_true(all(!is.na(V[3:4, 3:4])))
  expect_true(all(is.na(V[1:2, 3:4])))
  expect_true(all(is.na(V[3:4, 1:2])))
})

test_that("IC block matches direct merge of IC objects", {
  m_ic   <- merge(e_ic1, e_ic2)
  m_both <- merge(e_ic1, e_ic2, e_noic)
  expect_equal(vcov(m_both)[1:4, 1:4], vcov(m_ic))
})

test_that("non-IC block matches individual vcov", {
  m_both <- merge(e_ic1, e_ic2, e_noic)
  expect_equal(vcov(m_both)[5:6, 5:6], vcov(e_noic))
})

test_that("two non-IC objects: correct block placement and NA cross-terms", {
  m <- merge(e_ic1, e_noic, e_noic2)
  V <- vcov(m)
  expect_equal(dim(V), c(6L, 6L))
  expect_true(all(!is.na(V[1:2, 1:2])))
  expect_true(all(!is.na(V[3:4, 3:4])))
  expect_true(all(!is.na(V[5:6, 5:6])))
  expect_true(all(is.na(V[1:2, 3:6])))
  expect_true(all(is.na(V[3:4, c(1:2, 5:6)])))
  expect_true(all(is.na(V[5:6, 1:4])))
})

test_that("all non-IC objects: diagonal blocks filled, cross-terms NA", {
  m <- merge(e_noic, e_noic2)
  V <- vcov(m)
  expect_true(all(!is.na(V[1:2, 1:2])))
  expect_true(all(!is.na(V[3:4, 3:4])))
  expect_true(all(is.na(V[1:2, 3:4])))
  expect_true(all(is.na(V[3:4, 1:2])))
})

context("merge.estimate")

library(testthat)
library(lava)

set.seed(42)
n    <- 100
ic1  <- matrix(rnorm(n * 2), nrow=n, ncol=2) |> scale(center=TRUE, scale=FALSE)
ic2  <- matrix(rnorm(n * 2), nrow=n, ncol=2) |> scale(center=TRUE, scale=FALSE)
e_ic1  <- estimate(coef=c(a=1, b=2), IC=ic1, id=1:n)
e_ic2  <- estimate(coef=c(c=3, d=4), IC=ic2, id=1:n+50)
e_noic  <- estimate(coef=c(e=5, f=6), vcov=diag(2))
e_noic2 <- estimate(coef=c(g=7, h=8), vcov=diag(2) * 2)

test_that("coefficients are concatenated correctly", {
  m <- merge(e_ic1, e_noic)
  expect_equal(coef(m), c(a=1, b=2, e=5, f=6))
})

test_that("vcov has correct dimensions", {
  m <- merge(e_ic1, e_noic)
  expect_equal(dim(vcov(m)), c(4L, 4L))
})

test_that("IC block is non-NA, cross-terms are NA", {
  m <- merge(e_ic1, e_noic)
  V <- vcov(m)
  expect_true(all(!is.na(V[1:2, 1:2])))
  expect_true(all(!is.na(V[3:4, 3:4])))
  expect_true(all(is.na(V[1:2, 3:4])))
  expect_true(all(is.na(V[3:4, 1:2])))
})

test_that("IC block matches direct merge of IC objects", {
  m_ic   <- merge(e_ic1, e_ic2)
  m_both <- merge(e_ic1, e_ic2, e_noic)
  expect_equal(vcov(m_both)[1:4, 1:4], vcov(m_ic))
})

test_that("non-IC block matches individual vcov", {
  m_both <- merge(e_ic1, e_ic2, e_noic)
  expect_equal(vcov(m_both)[5:6, 5:6], vcov(e_noic))
})

test_that("two non-IC objects: correct block placement and NA cross-terms", {
  m <- merge(e_ic1, e_noic, e_noic2)
  V <- vcov(m)
  expect_equal(dim(V), c(6L, 6L))
  expect_true(all(!is.na(V[1:2, 1:2])))
  expect_true(all(!is.na(V[3:4, 3:4])))
  expect_true(all(!is.na(V[5:6, 5:6])))
  expect_true(all(is.na(V[1:2, 3:6])))
  expect_true(all(is.na(V[3:4, c(1:2, 5:6)])))
  expect_true(all(is.na(V[5:6, 1:4])))
})

test_that("all non-IC objects: diagonal blocks filled, cross-terms NA", {
  m <- merge(e_noic, e_noic2)
  V <- vcov(m)
  expect_true(all(!is.na(V[1:2, 1:2])))
  expect_true(all(!is.na(V[3:4, 3:4])))
  expect_true(all(is.na(V[1:2, 3:4])))
  expect_true(all(is.na(V[3:4, 1:2])))
})

test_that("c.estimate with 'extra' attr", {
  e1 <- c(estimate(coef = c(a = 1), vcov = matrix(0.1)), e1.iter = 2)
  e2 <- c(estimate(coef = c(b = 2), vcov = matrix(0.2)), e2.iter = 3)
  e3 <- c(estimate(coef = c(c = 2), vcov = matrix(0.3)))

  e <- c(e1, e2) # two estimate.extra objects
  expect_equal(c(coef(e1), coef(e2)), coef(e))
  expect_equal(c(vcov(e1), vcov(e2)),
               diag(vcov(e)), check.attributes=FALSE)
  expect_equal(c(attr(e1, "extra"), attr(e2, "extra")),
                 attr(e, "extra"))

  e <- c(e1, e3) # estimatee.xtra + estimate
  expect_equal(c(coef(e1), coef(e3)), coef(e))
  expect_equal(c(vcov(e1), vcov(e3)),
               diag(vcov(e)), check.attributes=FALSE)
  expect_equal(attr(e1, "extra"), attr(e, "extra"))

  e <- c(e3, e1) # estimate + estimate.extra
  expect_equal(c(coef(e3), coef(e1)), coef(e))
  expect_equal(c(vcov(e3), vcov(e1)),
               diag(vcov(e)), check.attributes=FALSE)
  expect_equal(attr(e1, "extra"), attr(e, "extra"))

  e <- c(e1, e2, d=2) # estimate + scalar/extra
  expect_equal(c(coef(e1), coef(e2)), coef(e))
  expect_equal(c(vcov(e1), vcov(e2)),
               diag(vcov(e)), check.attributes=FALSE)
  expect_equal(c(attr(e1, "extra"), attr(e2, "extra"), d=2),
               attr(e, "extra"))

  e <- c(e3, c(a = 1, b = 2)) # named scalar vector via unnamed argument
  expect_equal(c(a = 1, b = 2), attr(e, "extra"))

  e <- c(e3, e1.iter = 1) # duplicated name of extra variable
  e <- c(e, e1)
  expect_equal(c(e1.iter = 1, e1.iter = 2), attr(e, "extra"))

  e <- c(e1, converged = FALSE) # boolean extra variable
  # internal conversion to numeric
  expect_equal(c(e1.iter = 2, converged = 0), attr(e, "extra"))
})

## summary.estimate() requires IC=TRUE on the input estimate when applying
## small-sample corrections (type / var.adj). Per Q5.

context("summary.estimate IC requirement")

make_glm <- function() {
  set.seed(1)
  d <- data.frame(
    y = rbinom(100, 1, 0.5),
    x = rnorm(100),
    z = rnorm(100)
  )
  glm(y ~ x + z, data = d, family = binomial())
}

test_that("summary(..., type=) errors when input estimate has IC=FALSE", {
  # variance correction via the type argument requires e$IC to be not null
  g <- make_glm()
  e <- estimate(g, IC = FALSE)
  expect_error(
    summary(e, type = "df"),
    "IC=TRUE",
    fixed = TRUE
  )
})

test_that("summary(..., null=) works when input estimate has IC=FALSE", {
  ## null/contrast do NOT require IC, only type/var.adj do.
  g <- make_glm()
  e <- estimate(g, IC = FALSE)
  expect_silent(summary(e, null = 1))
})

## Tests for soft deprecation of null/contrast/type/var.adj in estimate.default

context("estimate.default deprecation warnings")

make_glm <- function() {
  set.seed(1)
  d <- data.frame(
    y = rbinom(100, 1, 0.5),
    x = rnorm(100),
    z = rnorm(100)
  )
  glm(y ~ x + z, data = d, family = binomial())
}

test_that("estimate(..., null=) emits deprecation warning", {
  g <- make_glm()
  expect_warning(estimate(g, null = 0), class = "deprecatedWarning")
})

test_that("estimate(..., back.transform=) emits deprecation warning", {
  g <- make_glm()
  expect_warning(
    ss <- estimate(g, back.transform = exp),
    class = "deprecatedWarning"
  )

  expect_equal(exp(coef(g)), coef(ss))
})

test_that("estimate(..., level=) emits deprecation warning", {
  g <- make_glm()
  s0 <- estimate(g)
  expect_warning(
    ss <- estimate(g, level = 0.5),
    class = "deprecatedWarning"
  )
  expect_true(all(confint(s0) != confint(ss)))

})


test_that("estimate(..., null=0) warns even when null=0 (Q7)", {
  g <- make_glm()
  expect_warning(estimate(g, null = 0), class = "deprecatedWarning")
})

test_that("estimate(..., contrast=) emits deprecation warning", {
  g <- make_glm()
  B <- rbind(c(1, 0, 0), c(0, 1, 0))
  expect_warning(estimate(g, contrast = B), class = "deprecatedWarning")
})

test_that("estimate(..., type=) emits deprecation warning", {
  g <- make_glm()
  expect_warning(estimate(g, type = "df"), class = "deprecatedWarning")
})

test_that("estimate(..., var.adj=) emits deprecation warning", {
  g <- make_glm()
  expect_warning(
    estimate(g, type = "hc3", var.adj = 0.5),
    class = "deprecatedWarning"
  )
})

test_that("multiple deprecated args produce exactly one deprecation warning", {
  g <- make_glm()
  ws <- capture_warnings(estimate(g, null = 1, type = "df"))
  dep_msgs <- grep("deprecated", ws, value = TRUE)
  expect_equal(length(dep_msgs), 1L)
})

test_that("internally-derived contrast from f does NOT warn (Q8)", {
  g <- make_glm()
  ws <- capture_warnings(estimate(g, "x"))
  expect_equal(length(grep("deprecated", ws)), 0L)
  ws <- capture_warnings(estimate(g, 2:3))
  expect_equal(length(grep("deprecated", ws)), 0L)
  ws <- capture_warnings(estimate(g, rbind(c(1, 0, 0), c(0, 1, 0))))
  expect_equal(length(grep("deprecated", ws)), 0L)
})

test_that("plain estimate() emits no deprecation warning (Q4)", {
  g <- make_glm()
  ws <- capture_warnings(estimate(g))
  expect_equal(length(grep("deprecated", ws)), 0L)
})

##' @export
profile.lvmfit <- function(fitted,idx,tau,...) {
  mm <- parfix(Model(fitted),idx,tau)
  index(mm) <- reindex(mm,zeroones=TRUE,deriv=TRUE)
  fixed <- attributes(mm)$fixed
  plogl <- function(tau0) {
    for (i in fixed$v) {
      mm$mean[[i]] <- tau0
    }
    for (i in seq_len(nrow(fixed$A))) {
        index(mm)$A[fixed$A[i,1],fixed$A[i,2]] <-
            mm$fix[fixed$A[i,1],fixed$A[i,2]] <- tau0
    }
    for (i in seq_len(nrow(fixed$P))) {
        index(mm)$P[fixed$P[i,1],fixed$P[i,2]] <-
            mm$covfix[fixed$P[i,1],fixed$P[i,2]] <- tau0
    }
    for (i in length(fixed$e)) {
        index(mm)$exfix[i] <- tau0
    }
    dots <- list(...)
    if (!is.null(dots$control))
      control <- dots$control
    else
      control <- list()
    control$start <- coef(fitted)
    dots$control <- control
    dots$index <- FALSE
    dots$fix <- FALSE
    dots$messages <- 0
    dots$quick <- TRUE
    dots$data <- model.frame(fitted)
    dots$data2 <- fitted$data2
    dots$estimator <- fitted$estimator
    dots$weights <- fitted$weights
    dots$x <- mm
    ee <- do.call("estimate",dots)
    return(logLik(mm,p=ee,data=dots$data,data2=dots$data2,weights=fitted$weights,model=dots$estimator))
  }
  val <- sapply(tau,plogl)
  attributes(val) <- NULL
  val
}

profci.lvmfit <- function(x,parm,level=0.95,interval=NULL,curve=FALSE,n=20,lower=TRUE,upper=TRUE,...) {
  ll <- logLik(x)-qchisq(level,1)/2
  pp <- function(tau) (profile.lvmfit(x,parm,tau) - ll)
  tau0 <- coef(x)[parm]
  tau0.sd <- x$vcov[parm,parm]^0.5
  if (is.null(interval)) {
      interval <- tau0 + 3*c(-1,1)*tau0.sd
      if (parm%in%(variances(x)+index(x)$npar.mean))
          interval[1] <- max(1e-5,interval[1])
  }
  if (curve) {
    xx <- seq(interval[1],interval[2],length.out=n)
    val <- sapply(xx,pp)
    res <- cbind(par=xx,val=val)
    return(res)
  }
  low <- up <- NA
  if (lower)
    low <- uniroot(pp,interval=c(interval[1],tau0))$root
  if (upper)
    up <- uniroot(pp,interval=c(tau0,interval[2]))$root
  ##  res <- rbind(lower$root,upper$root); rownames(res) <- coef()
  return(c(low,up))
}

## kmeans++
## Assign centre of first cluster randomly from observations.
## 1) Calculate min distances of observations to current clusters, D
## 2) Sample new cluster centre from distribution p(x) = D^2(x)/sum(D^2)
## Repeat 1-2 until all clusters are assigned
kmpp <- function(y,k=2) {
    Dist <- function(y1,y2) sum((y1-y2)^2)
    n <- NROW(y)
    ii <- numeric(k)
    u <- runif(k)
    ii[1] <- sample(n,1)
    D <- matrix(0,n,k-1)
    hasmets <- requireNamespace("mets", quietly=TRUE)
    for (i in seq_len(k-1)+1) {
        D[,i-1] <- apply(y,1,function(x) Dist(x,y[ii[i-1],]))
        D2 <- apply(D[,seq(i-1),drop=FALSE],1,min)
        pdist <- cumsum(D2/sum(D2))
        if (hasmets)
            ii[i] <- mets::fast.approx(pdist,u[i])
        else
            ii[i] <- which.min(abs(pdist-u[i]))
    }
    return(ii)
}


##' Weighted K-means
##'
##' Weighted K-means via Lloyd's algorithm
##' @param x Data (or formula)
##' @param mu Initial centers (or number centers chosen randomly among x)
##' @param data optional data frmae
##' @param weights Optional weights
##' @param iter.max Max number of iterations
##' @param n.start Number of restarts
##' @param init method to create initial centres (default kmeans++)
##' @param ... Additional arguments to lower level functions
##' @export
##' @author Klaus K. Holst
##' @examples
##' ## Two well-separated Gaussian blobs in 2-D
##' set.seed(1)
##' x <- rbind(matrix(rnorm(100, mean = -3), ncol = 2),
##'            matrix(rnorm(100, mean =  3), ncol = 2))
##' res <- wkm(x, mu = 2)
##' table(res$cluster)
##' res$center
##'
##' ## Supply explicit initial centers (as a list)
##' res2 <- wkm(x, mu = list(c(-3, -3), c(3, 3)))
##'
##' ## Weighted clustering: up-weight the second blob
##' w <- c(rep(1, 50), rep(10, 50))
##' res3 <- wkm(x, mu = 2, weights = w)
##'
##' ## Formula interface on a data.frame
##' wkm(~ Sepal.Length + Sepal.Width, data = iris, mu = 3)
wkm <- function(x, mu, data, weights=rep(1,NROW(x)), iter.max=20, n.start=5, init="kmpp", ...) { ## Lloyd's algorithm
    if (inherits(x, "formula")) x <- stats::model.matrix(x,data=data)
    x <- cbind(x)
    random.start <- TRUE
    if (is.list(mu)) {
        random.start <- FALSE
        n.start=1
        K <- length(mu)
    } else {
        K <- mu
    }
    sswmin <- Inf
    ssws <- NULL
    dmin <- NULL
    cl0 <- rep(1,NROW(x))
    for (k in seq(n.start)) {
        if (random.start) {
            if (!exists(init)) { ## Random select centres
                idx <- sample(NROW(x),K)
            } else {
                idx <- do.call(init, list(x, K))
            }
            mu <- lapply(idx, function(i) cbind(x)[i,,drop=TRUE])
        }
        for (i in seq(iter.max)) {
            d <- cbind(Reduce(cbind,lapply(mu, function(m) weights*colSums((t(x)-m)*(t(x)-m)))))
            cl <- apply(d,1,which.min)
            for (j in seq_along(mu)) {
                idx <- which(cl==j)
                if (length(idx)) {
                    mu[[j]] <- colSums(cbind(apply(x[idx,,drop=FALSE],2,
                                            function(x) x*weights[idx])))/sum(weights[idx])
                }

            }
            if (sum(cl0-cl)==0L) break; # No change in assigment
        }
        ssw <- sum(d[cbind(seq(NROW(d)),cl)])
        ssws <- c(ssws,ssw)
        if (ssw < sswmin) {
            sswmin <- ssw
            clmin <- cl
            mumin <- mu
            dmin <- d
        }
    }
    mu <- structure(mumin,class="by",dim=K,dimnames=list(class=seq(K)))
    withinclusterss <- as.vector(by(dmin[cbind(seq(NROW(dmin)),clmin)],clmin,sum))
    return(list(cluster=clmin,
           center=mu, ssw=withinclusterss))
}

impute0 <- function(object,rows,idx,na.action=na.omit,value,...) {
    if (missing(rows) && missing(idx)) {
        df <- na.action(object,...) 
        rows <- attr(df,"na.action")  
    }
    if (!missing(idx)) {
        obs1 <- setdiff(seq(length(object)),idx)[1]
    } else {
        obs1 <- setdiff(seq(NROW(object)),rows)[1]
    }
    if (missing(value)) {
        fobs <- object[obs1]
        if (is.logical(fobs)) value <- FALSE
        else if (is.character(fobs)) value <- fobs
        else if (is.factor(fobs)) value <- levels(fobs)[1]
        else value <- 0
    }
    if (!missing(idx)) {
        object[idx] <- value
        return(object)
    }
    if (is.matrix(object)) {
        object[rows,] <- value
    } else {
        object[rows] <- value
    }
    return(object)
}

##' @title Handle Missing Values in Objects
##' @description Returns the object with missing data replaced by zeros. This is
##'   sometimes useful for example when working with inverse probability
##'   weighting of the complete-case data.
##' @seealso [na.pass()], [na.omit()], [na.fail()]
##' @export
##' @param object data.frame, vector
##' @param na.action function used to identify missing values
##' @param ... additional arguments to lower level functions
##' @examples
##' d <- data.frame(y=c(1,1,NA,2,NA,2), r=c(1,1,0,1,1,1))
##' na.pass0(d)
##' glm(y ~ 1, weights=d$r, data=d, na.action=na.pass0)
na.pass0 <- function(object, na.action=na.omit, ...) {
    ## Fill in "zeros" in the design matrix where we have missing data    
    df <- na.action(object,...)
    idx <- attr(df,"na.action")
    if (is.matrix(object) || is.vector(object)) {
        object <- impute0(object,rows=idx,...)
    } else {
        for (i in seq_len(NCOL(object))) {
            object[[i]] <- impute0(object[[i]],rows=idx,...)
        }
    }
    if (!is.null(idx))
        return(structure(object,na.action=structure(idx,class="pass0")))
    return(object)
}

##' Add an observed event time outcome to a latent variable model.
##'
##' For example, if the model 'm' includes latent event time variables
##' are called 'T1' and 'T2' and 'C' is the end of follow-up (right censored),
##' then one can specify
##'
##' \code{eventTime(object=m,formula=ObsTime~min(T1=a,T2=b,C=0,"ObsEvent"))}
##'
##' when data are simulated from the model
##' one gets 2 new columns:
##'
##' - "ObsTime": the smallest of T1, T2 and C
##' - "ObsEvent": 'a' if T1 is smallest, 'b' if T2 is smallest and '0' if C is smallest
##'
##' Note that "ObsEvent" and "ObsTime" are names specified by the user.
##'
##' @author Thomas A. Gerds, Klaus K. Holst
##' @keywords survival models regression
##' @examples
##'
##' # Right censored survival data without covariates
##' m0 <- lvm()
##' distribution(m0,"eventtime") <- coxWeibull.lvm(scale=1/100,shape=2)
##' distribution(m0,"censtime") <- coxExponential.lvm(rate=1/10)
##' m0 <- eventTime(m0,time~min(eventtime=1,censtime=0),"status")
##' sim(m0,10)
##'
##' # Alternative specification of the right censored survival outcome
##' ## eventTime(m,"Status") <- ~min(eventtime=1,censtime=0)
##'
##' # Cox regression:
##' # lava implements two different parametrizations of the same
##' # Weibull regression model. The first specifies
##' # the effects of covariates as proportional hazard ratios
##' # and works as follows:
##' m <- lvm()
##' distribution(m,"eventtime") <- coxWeibull.lvm(scale=1/100,shape=2)
##' distribution(m,"censtime") <- coxWeibull.lvm(scale=1/100,shape=2)
##' m <- eventTime(m,time~min(eventtime=1,censtime=0),"status")
##' distribution(m,"sex") <- binomial.lvm(p=0.4)
##' distribution(m,"sbp") <- normal.lvm(mean=120,sd=20)
##' regression(m,from="sex",to="eventtime") <- 0.4
##' regression(m,from="sbp",to="eventtime") <- -0.01
##' sim(m,6)
##' # The parameters can be recovered using a Cox regression
##' # routine or a Weibull regression model. E.g.,
##' \dontrun{
##'     set.seed(18)
##'     d <- sim(m,1000)
##'     library(survival)
##'     coxph(Surv(time,status)~sex+sbp,data=d)
##'
##'     sr <- survreg(Surv(time,status)~sex+sbp,data=d)
##'     library(SurvRegCensCov)
##'     ConvertWeibull(sr)
##'
##' }
##'
##' # The second parametrization is an accelerated failure time
##' # regression model and uses the function weibull.lvm instead
##' # of coxWeibull.lvm to specify the event time distributions.
##' # Here is an example:
##'
##' ma <- lvm()
##' distribution(ma,"eventtime") <- weibull.lvm(scale=3,shape=1/0.7)
##' distribution(ma,"censtime") <- weibull.lvm(scale=2,shape=1/0.7)
##' ma <- eventTime(ma,time~min(eventtime=1,censtime=0),"status")
##' distribution(ma,"sex") <- binomial.lvm(p=0.4)
##' distribution(ma,"sbp") <- normal.lvm(mean=120,sd=20)
##' regression(ma,from="sex",to="eventtime") <- 0.7
##' regression(ma,from="sbp",to="eventtime") <- -0.008
##' set.seed(17)
##' sim(ma,6)
##' # The regression coefficients of the AFT model
##' # can be tranformed into log(hazard ratios):
##' #  coef.coxWeibull = - coef.weibull / shape.weibull
##' \dontrun{
##'     set.seed(17)
##'     da <- sim(ma,1000)
##'     library(survival)
##'     fa <- coxph(Surv(time,status)~sex+sbp,data=da)
##'     coef(fa)
##'     c(0.7,-0.008)/0.7
##' }
##'
##'
##' # The following are equivalent parametrizations
##' # which produce exactly the same random numbers:
##'
##' model.aft <- lvm()
##' distribution(model.aft,"eventtime") <- weibull.lvm(intercept=-log(1/100)/2,sigma=1/2)
##' distribution(model.aft,"censtime") <- weibull.lvm(intercept=-log(1/100)/2,sigma=1/2)
##' sim(model.aft,6,seed=17)
##'
##' model.aft <- lvm()
##' distribution(model.aft,"eventtime") <- weibull.lvm(scale=100^(1/2), shape=2)
##' distribution(model.aft,"censtime") <- weibull.lvm(scale=100^(1/2), shape=2)
##' sim(model.aft,6,seed=17)
##'
##' model.cox <- lvm()
##' distribution(model.cox,"eventtime") <- coxWeibull.lvm(scale=1/100,shape=2)
##' distribution(model.cox,"censtime") <- coxWeibull.lvm(scale=1/100,shape=2)
##' sim(model.cox,6,seed=17)
##'
##' # The minimum of multiple latent times one of them still
##' # being a censoring time, yield
##' # right censored competing risks data
##'
##' mc <- lvm()
##' distribution(mc,~X2) <- binomial.lvm()
##' regression(mc) <- T1~f(X1,-.5)+f(X2,0.3)
##' regression(mc) <- T2~f(X2,0.6)
##' distribution(mc,~T1) <- coxWeibull.lvm(scale=1/100)
##' distribution(mc,~T2) <- coxWeibull.lvm(scale=1/100)
##' distribution(mc,~C) <- coxWeibull.lvm(scale=1/100)
##' mc <- eventTime(mc,time~min(T1=1,T2=2,C=0),"event")
##' sim(mc,6)
##'
##'
##' @export
##' @aliases eventTime<-
##' @param object Model object
##' @param formula Formula (see details)
##' @param eventName Event names
##' @param \dots Additional arguments to lower levels functions
eventTime <- function(object,formula,eventName="status",...) {
    if (missing(formula)) return(object$attributes$eventHistory)
    if (inherits(eventName,"formula")) eventName <- all.vars(eventName)
    ff <- as.character(formula)
    timeName <- all.vars(update.formula(formula,"~1"))
    if (length(timeName)==0){
        timeName <- "observedTime"
        rhs <- ff[[2]]
    }else{
        rhs <- ff[[3]]
    }
    ## rhs <- tolower(rhs)
    latentTimes <- strsplit(rhs,"[(,)]")[[1]]
    if (latentTimes[1]!="min")
        stop(paste("Formula ",formula," does not have the required form, ",
                   "e.g. ~min(T1=1,T2=2,C=0), see (examples in) help(eventTime)."))
    latentTimes <- latentTimes[-1]
    NT <- length(latentTimes)
    events <- vector(NT,mode="character")
    for (lt in seq_len(NT)){
        tmp <- strsplit(latentTimes[lt],"=")[[1]]
        stopifnot(length(tmp) %in% c(1,2))
        if (length(tmp)==1){
            events[lt] <- as.character(lt)
            latentTimes[lt] <- tmp
        }
        else{
            events[lt] <- tmp[2]
            latentTimes[lt] <- tmp[1]
        }
    }
    events <- gsub(" ","",events)
    eventnum <- char2num(events)
    if (all(!is.na(eventnum))) {
        events <- eventnum
    } else {
        events <- gsub("\"","",events)
    }

    addvar(object) <- timeName
    eventTime <- list(names=c(timeName,eventName),
                      latentTimes=gsub(" ","",latentTimes),
                      events=events
                      )

    transform(object,
              y=eventTime$names,
              x=eventTime$latentTimes) <-
        function(z) {
            idx <- apply(z,1,which.min)
            cbind(z[cbind(seq(NROW(z)),idx)],
                  eventTime$events[idx])
        }

    if (is.null(object$attributes$eventHistory)) {
        object$attributes$eventHistory <- list(eventTime)
        names(object$attributes$eventHistory) <- timeName
    } else {
        object$attributes$eventHistory[[timeName]] <- eventTime
    }
    return(object)
}

##' @export
"eventTime<-" <- function(object,...,value) {
    eventTime(object,value,...)
}

## addhook("color_eventHistory","color.hooks")
## color.eventHistory <- function(x,subset=vars(x),...) {
##   return(list(vars=intersect(subset,binary(x)),col="indianred1"))
## }

addhook("plothook_eventHistory","plot.post.hooks")
plothook_eventHistory <- function(x,...) {
    eh <- x$attributes$eventHistory
    for (f in eh) {
        x <- regression(x,to=f$names[1],from=f$latentTimes)
        latent(x) <- f$latentTimes
        kill(x) <- f$names[2]
    }
    timedep <- x$attributes$timedep
    for (i in seq_len(length(timedep))) {
        x <- regression(x,to=names(timedep)[i],from=timedep[[i]])
    }
    return(x)
}

addhook("colorhook_eventHistory","color.hooks")
colorhook_eventHistory <- function(x,subset=vars(x),...) {
    return(list(vars=intersect(subset,unlist(x$attributes$timedep)),col="lightblue4"))
}

addhook("print_eventHistory","print.hooks")
print_eventHistory <- function(x,...) {
    eh <- x$attributes$eventHistory
    timedep <- x$attributes$timedep
    if (is.null(eh) & is.null(timedep)) return(NULL)
    ehnames <- unlist(lapply(eh,function(x) x$names))
    cat("Event History Model\n")
    ff <- formula(x,char=TRUE,all=TRUE)
    R <- c()
    for (f in ff) {
        oneline <- as.character(f);
        y <- gsub(" ","",strsplit(f,"~")[[1]][1])
        if (!(y %in% ehnames)) {
            col1 <- as.character(oneline)
            D <- attributes(distribution(x)[[y]])$family
            col2 <- "Normal"
            if (!is.null(D$family)) col2 <- paste0(D$family)
            if (!is.null(D$link)) col2 <- paste0(col2,"(",D$link,")")
            if (!is.null(D$par)) col2 <- paste0(col2,"(",paste(D$par,collapse=","),")")
            R <- rbind(R,c(col1,"  ",col2))
        }
    }
    for (y in names(eh)) {
        col1 <- paste0(y, " = min(",paste(eh[[y]]$latentTimes,collapse=","),")")
        eh[[y]]$names[2]
        col2 <- paste0(eh[[y]]$names[2], " := {",paste(eh[[y]]$events,collapse=","),"}")
        R <- rbind(R,c(col1,"",col2))
    }
    rownames(R) <- rep("",nrow(R)); colnames(R) <- rep("",ncol(R))
    print(R,quote=FALSE,...)
    cat("\n")
    for (i in seq_len(length(timedep))) {
        cat("Time-dependent covariates:\n\n")
        cat(paste("",names(timedep)[i],"~", paste(timedep[[i]],collapse="+")),"\n")
    }
    TRUE
}

## addhook("simulate_eventHistory","sim.hooks")

## simulate.eventHistory <- function(x,data,...){
##   if (is.null(eventTime(x))) {
##     return(data)
##   }
##   else{
##     for (eh in eventTime(x)) {
##       if (any((found <- match(eh$latentTimes,names(data),nomatch=0))==0)){
##         warning("Cannot find latent time variable: ",
##                 eh$latentTimes[found==0],".")
##       }
##       else{
##         for (v in seq_along(eh$latentTimes)) {
##           if (v==1){ ## initialize with the first latent time and event
##             eh.time <- data[,eh$latentTimes[v]]
##             eh.event <- rep(eh$events[v],NROW(data))
##           } else{ ## now replace if next time is smaller
##             ## in case of tie keep the first event
##             eh.event[data[,eh$latentTimes[v]]<eh.time] <- eh$events[v]
##             eh.time <- pmin(eh.time,data[,eh$latentTimes[v]])
##           }
##         }
##       }
##       data[,eh$names[1]] <- eh.time
##       data[,eh$names[2]] <- eh.event
##     }
##     return(data)
##   }
## }



##' @export
coxWeibull.lvm <- function(scale=1/100,shape=2, param=1) {
    ## proportional hazard (Cox) parametrization.
    ##
    ## Here we parametrize the Weibull distribution
    ## (without covariates) as
    ##
    ## hazard(t) = scale * shape * t^(shape-1)
    ##
    ## The linear predictor (LP) enters like this
    ##
    ## hazard(t) = = scale * exp(LP) * shape * t^(shape-1)
    ##
    ## Thus, we simulate
    ##
    ## T = (scale^{-1} * exp(-LP) * -log(U))^{shape-1})
    ##
    ## The hazard is:
    ## - rising if shape > 1
    ## - declining if shape <1
    ## - constant if shape=1
    ##
    ## scale = exp(b0 + b1*X)
    ##
    ## An alternative parametrization, which can be found in ABGK and
    ## Kalbfleisch&Prentice, can be obtained with \code{param=2}. This leads to
    ## hazard(t) = scale*shape*(scale*t)^shape, and cum.hazard(t) = (scale*t)^shape.
    ## \code{param=3} gives the same parametrization as \code{rweibull}, i.e.,
    ## scale[rweibull] := 1/scale^shape
    if (tolower(param)%in%c("2","kp","abgk"))
       scale <- scale^(shape)
    if (tolower(param)%in%c("3","r","rweibull")) {
       scale <- scale^(-shape)
    }
    f <- function(n,mu=0,Scale=scale,Shape=shape,...) {
        (- log(runif(n)) / (Scale * exp(mu)))^(1/Shape)
    }
    ff <- formals(f)
    expr <- "(- log(runif(n)) / (Scale * exp(mu)))^{1/Shape}"
    if (inherits(scale,"formula")) scale <- all.vars(scale)[1]
    if (is.character(scale)) {
        names(ff)[3] <- scale
        expr <- gsub("Scale",scale,expr)
    }
    if (inherits(shape,"formula")) shape <- all.vars(shape)[1]
    if (is.character(shape)) {
        names(ff)[4] <- shape
        expr <- gsub("Shape",shape,expr)
    }
    formals(f) <- ff
    e <- parse(text=expr)
    body(f) <- as.call(c(as.name("{"), e))
    attr(f,"family") <- list(family="weibull",
                             regression="PH",
                             par=c(shape=shape,scale=scale))
    return(f)
}


##' @export
coxExponential.lvm <- function(scale=1,rate=1,timecut){
    if (missing(timecut)) {
        return(coxWeibull.lvm(shape=1, rate))
    }
    if (NROW(rate)>length(timecut))
        stop("Number of time-intervals (cuts) does not agree with number of rate parameters (beta0)")
    par <- paste(timecut,rate,sep=":")
    if (is.matrix(rate)) par <- "..."
    timecut <- c(timecut,Inf)
    f <- function(n,mu=0,...) {
        Ai <- function() {
            vals <- matrix(0,ncol=length(timecut)-1,nrow=n)
            ival <- numeric(n)
            if (is.matrix(rate)) {
                mu <- cbind(mu[,1],cbind(1,as.matrix(mu[,-1]))%*%t(rate))
                rate <- rep(1,length(timecut)-1)
            }
            for (i in seq(length(timecut)-1)) {
                u <- -log(runif(n)) ##rexp(n,1)
                if (NCOL(mu)>1) {
                    vals[,i] <-  timecut[i] + u*exp(-mu[,1]-mu[,i+1])/(rate[i])
                } else {
                    vals[,i] <-  timecut[i] + u*exp(-mu)/(rate[i])
                }
                idx <- which(vals[,i]<=timecut[i+1] & ival==0)
                ival[idx] <- vals[idx,i]
            }
            ival
        }
        Ai()
    }
    attributes(f)$family <- list(family="CoxExponential",par=par)
    return(f)
}

##' @export
aalenExponential.lvm <- function(rate=1,timecut=0){
    if (missing(timecut)==1) {
        return(coxWeibull.lvm(shape=1,rate))
    }

    if (length(rate)>length(timecut))
        stop("Number of time-intervals (cuts) does not agree with number of rate parameters (beta0)")
    par <- paste(timecut,rate,sep=":")
    if (is.matrix(rate)) par <- "..."
    timecut <- c(timecut,Inf)
    f <- function(n,mu,...) {
        Ai <- function() {
            vals <- matrix(0,ncol=length(timecut)-1,nrow=n)
            ival <- numeric(n)
            if (is.matrix(rate)) {
                mu <- cbind(mu[,1],cbind(1,as.matrix(mu[,-1]))%*%t(rate))
                rate <- rep(1,length(timecut)-1)
            }
            for (i in seq(length(timecut)-1)) {
                u <- -log(runif(n)) ##rexp(n,1)
                if (NCOL(mu)>1) {
                    vals[,i] <-  timecut[i] + u/(rate[i]+mu[,1]+mu[,i+1])
                } else {
                    vals[,i] <-  timecut[i] + u/(rate[i]+mu)
                }
                idx <- which(vals[,i]<=timecut[i+1] & ival==0)
                ival[idx] <- vals[idx,i]
            }
            ival
        }
        Ai()
    }
    attributes(f)$family <- list(family="aalenExponential",par=par)
    return(f)
}


##' @export
coxGompertz.lvm <- function(shape=1,scale) {
    f <- function(n,mu,var,...) {
        (1/shape) * log(1 - (shape/scale) * (log(runif(n)) * exp(-mu)))
    }
    attr(f,"family") <- list(family="gompertz",par=c(shape,scale))
    return(f)
}

context("procformula robustness against as.character.formula masking")

# formula.tools overwrites as.character.formula, changing it from returning
# a length-3 vector c("~", "LHS", "RHS") to a single collapsed string.
# These tests verify that procformula, eventTime, and predict.zibreg work
# correctly regardless of whether as.character.formula has been masked.

mock_as_character_formula <- function(x, ...) {
  paste(deparse(x), collapse = " ")
}

test_that("procformula handles standard two-sided formulas", {
  res <- procformula(value = y ~ x + z)
  expect_equal(res$ys, "y")
  expect_true("x" %in% res$xs)
  expect_true("z" %in% res$xs)
})

test_that("procformula handles covariance formula (y1 ~ ~y2)", {
  m <- lvm()
  res <- procformula(m, value = y1 ~ ~y2)
  expect_equal(res$ys, "y1")
  expect_true(res$iscovar)
})

test_that("procformula handles multivariate LHS", {
  res <- procformula(value = c(y1, y2) ~ x + z)
  expect_true("y1" %in% res$ys)
  expect_true("y2" %in% res$ys)
  expect_true("x" %in% res$xs)
  expect_true("z" %in% res$xs)
})

test_that("procformula handles one-sided formula", {
  res <- procformula(value = ~ x + z)
  expect_null(res$ys)
  expect_true("x" %in% res$xs)
  expect_true("z" %in% res$xs)
})

test_that("procformula handles link function syntax", {
  res <- procformula(value = y ~ exp(x + z))
  expect_equal(res$invlink, "exp")
  expect_true("x" %in% res$xs)
  expect_true("z" %in% res$xs)
})

test_that("procformula results match when as.character.formula is masked", {
  formulas <- list(
    y ~ x + z,
    y1 ~ ~y2,
    c(y1, y2) ~ x + z,
    ~ x + z,
    y ~ x + z + ~w
  )
  for (f in formulas) {
    # Run with normal dispatch
    res_normal <- procformula(value = f)
    # Temporarily mask as.character.formula
    registerS3method("as.character", "formula", mock_as_character_formula,
                     envir = asNamespace("base"))
    on.exit(
      {
        rm_env <- asNamespace("base")
        if (exists("as.character.formula", envir = rm_env))
          rm("as.character.formula", envir = rm_env)
      },
      add = TRUE
    )
    res_masked <- procformula(value = f)
    # Restore
    rm_env <- asNamespace("base")
    if (exists("as.character.formula", envir = rm_env))
      rm("as.character.formula", envir = rm_env)
    expect_identical(res_normal$ys, res_masked$ys,
                     info = paste("ys mismatch for", deparse(f)))
    expect_identical(res_normal$xs, res_masked$xs,
                     info = paste("xs mismatch for", deparse(f)))
    expect_identical(res_normal$iscovar, res_masked$iscovar,
                     info = paste("iscovar mismatch for", deparse(f)))
    expect_identical(res_normal$invlink, res_masked$invlink,
                     info = paste("invlink mismatch for", deparse(f)))
  }
})

test_that("lvm model building works when as.character.formula is masked", {
  registerS3method("as.character", "formula", mock_as_character_formula,
                   envir = asNamespace("base"))
  on.exit(
    {
      rm_env <- asNamespace("base")
      if (exists("as.character.formula", envir = rm_env))
        rm("as.character.formula", envir = rm_env)
    },
    add = TRUE
  )
  # These should not error
  m <- lvm(y ~ x + z)
  expect_true("y" %in% vars(m))
  expect_true("x" %in% vars(m))

  m2 <- lvm(c(y1, y2) ~ x, y1 ~ ~y2)
  expect_true("y1" %in% vars(m2))
  expect_true("y2" %in% vars(m2))
})

test_that("eventTime works when as.character.formula is masked", {
  registerS3method("as.character", "formula", mock_as_character_formula,
                   envir = asNamespace("base"))
  on.exit(
    {
      rm_env <- asNamespace("base")
      if (exists("as.character.formula", envir = rm_env))
        rm("as.character.formula", envir = rm_env)
    },
    add = TRUE
  )
  m <- lvm(y ~ x)
  # Two-sided eventTime formula
  expect_silent(eventTime(m, time ~ min(T1 = 1, T2 = 2, C = 0)))
  # One-sided eventTime formula
  expect_silent(eventTime(m, ~ min(T1 = 1, T2 = 0)))
})

##' @export
`endogenous` <-
    function(x,...) UseMethod("endogenous")

##' @export
`endogenous.lvmfit` <-
    function(x,...) {
        endogenous(Model(x),...)
    }

##' @export
`endogenous.lvm` <-
    function(x,top=FALSE,latent=FALSE,...) {
        observed <- manifest(x)
        if (latent) observed <- vars(x)
        if (top) {
            M <- x$M
            res <- c()
            for (i in observed)
                if (!any(M[i,]==1))
                    res <- c(res, i)
            return(res)
        }
        exo <- exogenous(x)
        return(setdiff(observed,exo))
    }

##' @export
endogenous.list <- function(x,...) {
    endolist <- c()
    for (i in seq_along(x)) {
        endolist <- c(endolist, endogenous(x[[i]]))
    }
    endolist <- unique(endolist)
    return(endolist)
}

##' @export
`endogenous.multigroup` <-
    function(x,...) {
        endogenous(Model(x))
    }

##' @export
`endogenous.lm` <-
    function(x,...) {
        getoutcome(formula(x))[1]
    }

##' Add color-bar to plot
##'
##' @title Add color-bar to plot
##' @param clut Color look-up table
##' @param x.range x range
##' @param y.range y range
##' @param values label values
##' @param digits number of digits
##' @param label.offset label offset
##' @param srt rotation of labels
##' @param cex text size
##' @param border border of color bar rectangles
##' @param alpha Alpha (transparency) level 0-1
##' @param position Label position left/bottom (1) or top/right (2) or no text (0)
##' @param direction horizontal or vertical color bars
##' @param \dots additional low level arguments (i.e. parsed to \code{text})
##' @export
##' @examples
##' \dontrun{
##' plotNeuro(x,roi=R,mm=-18,range=5)
##' colorbar(clut=Col(rev(rainbow(11,start=0,end=0.69)),0.5),
##'          x=c(-40,40),y.range=c(84,90),values=c(-5:5))
##'
##' colorbar(clut=Col(rev(rainbow(11,start=0,end=0.69)),0.5),
##'          x=c(-10,10),y.range=c(-100,50),values=c(-5:5),
##'          direction="vertical",border=1)
##' }
colorbar <- function(clut=Col(rev(rainbow(11,start=0,end=0.69)),alpha),
              x.range=c(-.5,.5),y.range=c(-.1,.1),
              values=seq(clut),digits=2,label.offset,srt=45,
              cex=0.5,border=NA,
              alpha=0.5,
              position=1,
              direction=c("horizontal","vertical"),...) {
    nlut <- length(clut)
    X <- length(agrep(tolower(direction[1]),"horizontal"))>0
    scale <- ifelse(X,diff(x.range),diff(y.range))/nlut
    barsize <- ifelse(X,diff(y.range),diff(x.range))
    if (missing(label.offset)) label.offset <- barsize/3
    delta <- ifelse(X,x.range[1],y.range[1])
    if (!is.null(values)) dM <- diff(range(values))/(nlut-1)
    for (i in seq_len(nlut+1)-1) {
        pos <- delta + (i-1)*scale
        if (X) {
            x1 <- pos; x2 <- pos+scale; y1 <- y.range[1]; y2 <- y.range[2]
        } else {
            y1 <- pos; y2 <- pos+scale; x1 <- x.range[1]; x2 <- x.range[2]
        }
        if (i>0)
            rect(x1,y1,x2,y2, col=clut[i], border=border, xpd=TRUE)
    }
    if (!is.null(values)) {
        for (i in seq_len(nlut+1)-1) {
            pos <- delta + (i-1)*scale
            rund <- format(round(min(values)+dM*i,max(1,digits)),digits=digits)
            x0 <- pos+(1+0.5)*scale; y0 <- y.range[2]+label.offset
            if (!X) {
                y0 <- x0;
                if (position==1) x0 <- x.range[1]-label.offset
                if (position==2) x0 <- x.range[1]+label.offset*5
                if (position==3) x0 <- x.range[1]+label.offset*1
            }
            if (i<nlut)
                text(x0,y0,rund,cex=cex,srt=srt,xpd=TRUE,...)
        }
    }
}

##' Extract graph
##'
##' Extract or replace graph object
##'
##'
##' @aliases Graph Graph<-
##' @usage
##'
##' Graph(x, ...)
##'
##' Graph(x, ...) <- value
##'
##' @param x Model object
##' @param value New \code{graphNEL} object
##' @param \dots Additional arguments to be passed to the low level functions
##' @author Klaus K. Holst
##' @seealso \code{\link{Model}}
##' @keywords graphs models
##' @export
##' @examples
##'
##' m <- lvm(y~x)
##' Graph(m)
##'
##' @export
`Graph` <-
    function(x,...) UseMethod("Graph")

##' @export
`Graph.lvm` <-
    function(x,add=FALSE,...) {
        if ((is.null(x$graph) || length(x$graph)==0) & add) {
            m <- Model(x)
            return(plot(m,noplot=TRUE))
        }
        else return(x$graph)
    }

##' @export
`Graph.lvmfit` <- function(x,...) Graph.lvm(x,...)

##' @export
"Graph<-" <- function(x,...,value) UseMethod("Graph<-")

##' @export
"Graph<-.lvmfit" <- function(x,...,value) { x$graph <- value; return(x) }

##' @export
"Graph<-.lvm" <- function(x,...,value) { x$graph <- value; return(x) }

###{{{

`matrices` <-
    function(x,...) UseMethod("matrices")

###{{{ matrices.lvm

##' @export
mat.lvm <- function(x,ii=index(x),...) {
    A <- ii$A ## Matrix with fixed parameters and ones where parameters are free
    M1 <- ii$M1 ## Index of free and _unique_ regression parameters
    P <- ii$P  ## Matrix with fixed variance parameters and ones where parameters are free
    P1 <- ii$P1 ## Index of free and _unique_ regression parameters

    constrain.par <- names(constrain(x))
    parval <- list()


    parBelongsTo <- list(mean=seq_len(ii$npar.mean),
                         reg=seq_len(ii$npar.reg)+ii$npar.mean,
                         cov=seq_len(ii$npar.var)+ii$npar.mean+ii$npar.reg,
                         epar=seq_len(ii$npar.ex)+with(ii,npar.reg+npar.var+npar.mean),
                         cpar=numeric())

    idxA <- which(M1==1)
    pidxA <- parBelongsTo$reg
    if (ii$npar.reg>0) {
        A[idxA] <- pidxA
        for (p in ii$parname) {
            idx <- which((x$par==p))
            newval <- A[idx[1]]
            attributes(newval)$reg.idx <- idx
            attributes(newval)$reg.tidx <- which(t(x$par==p))
            parval[[p]] <- newval
            if (length(idx)>1) {
                idxA <- c(idxA,idx[-1])
                pidxA <- c(pidxA,rep(A[idx[1]],length(idx)-1))
                A[idx] <- A[idx[1]]
            }
        } ## duplicate parameters
    }

    pars.var <- parBelongsTo$cov
    idxdiag <- (seq(ncol(P1))-1)*ncol(P1) + seq(ncol(P1))
    idxP <- idxdiag[which(P1[idxdiag]==1)]
    pidxP <- pars.var[seq_len(length(idxP))]
    P[idxP] <- pidxP

    pars.off.diag <- pars.var
    if (length(pidxP)>0) {
        pars.off.diag <- pars.off.diag[-seq_len(length(pidxP))]
    }

    counter <- 0
    if (length(pars.off.diag)>0 & ncol(P)>1)
        for (i in seq_len(ncol(P1)-1))
            for (j in seq(i+1,nrow(P1))) {
                if (ii$P1[j,i]!=0) {
                    counter <- counter+1
                    pos <- c(j+(i-1)*ncol(P1),
                             i+(j-1)*ncol(P1))
                    P[j,i] <- P[i,j] <- pars.off.diag[counter]
                    idxP <- c(idxP,pos); pidxP <- c(pidxP,P[j,i],P[i,j])
                }
            }

    if (length(ii$covparname)>0)
        for (p in ii$covparname) {
            idx <- which(x$covpar==p)
            isOffDiag <- !(idx[1]%in%idxdiag)
            if (!(p%in%ii$parname)) {
                parval[[p]] <- P[idx[1]]
            }
            attributes(parval[[p]])$cov.idx <- idx
            if (length(idx)>1+isOffDiag) {
                P[idx[-seq(1+isOffDiag)]] <- parval[[p]]
            }
            if (ii$npar.reg>0 && p%in%ii$parname) {
                parBelongsTo$reg <- c(parBelongsTo$reg,p)
                idx.reg <- which(x$par==p)
                P[idx] <- A[idx.reg[1]]
                atr <- attributes(parval[[p]])
                parval[[p]] <- A[idx.reg[1]]
                attributes(parval[[p]]) <- atr
                idxP <- c(idxP,idx)
                pidxP <- c(pidxP,rep(P[idx[1]],length(idx)))
            } else {
                idxP <- c(idxP,idx[-seq(1+isOffDiag)])
                pidxP <- c(pidxP,rep(P[idx[1]],length(idx)-1-isOffDiag))
            }
        } ## duplicate parameters

    idxM <- c()
    pidxM <- seq_len(ii$npar.mean)
    v <- NULL

    ## named <- sapply(x$mean, function(y) is.character(y) & !is.na(y))
    fixed <- sapply(x$mean, function(y) is.numeric(y) & !is.na(y))
    v <- rep(0,length(x$mean))
    names(v) <- colnames(P)
    if (ii$npar.mean>0) {
        idxM <- which(ii$v1==1)
        v[idxM] <- pidxM
    }
    if (any(fixed))
        v[fixed] <- unlist(x$mean[fixed])

    for (p in ii$mparname) {
        idx <- which(x$mean==p)
        if (!(p%in%c(ii$parname,ii$covparname))) {
            if (length(idx)>1) {
                pidxM <- c(pidxM,rep(v[idx[1]],length(idx)-1))
                idxM <- c(idxM,idx[-1])
            }
            parval[[p]] <- v[idx[1]]
            v[idx] <- parval[[p]]
        }
        attributes(parval[[p]])$m.idx <- idx
        if (p %in% ii$covparname & !(p %in% ii$parname)) {
            parBelongsTo$cov <- c(parBelongsTo$cov,p)
            idx.2 <- which(x$covpar==p)
            v[idx] <- P[idx.2[1]]
            pidxM <- c(pidxM,rep(P[idx.2[1]],length(idx)))
            idxM <- c(idxM,idx)
        }
        if (p %in% ii$parname) {
            parBelongsTo$reg <- c(parBelongsTo$reg,p)
            idx.2 <- which(x$par==p)
            v[idx] <- A[idx.2[1]]
            pidxM <- c(pidxM,rep(A[idx.2[1]],length(idx)))
            idxM <- c(idxM,idx)
        }
    }

    ## Ex-parameters
    idxE <- NULL
    pidxE <- parBelongsTo$epar
    ## named <- sapply(x$exfix, function(y) is.character(y) & !is.na(y))
    fixed <- sapply(x$exfix, function(y) is.numeric(y) & !is.na(y))
    epar <- rep(0,length(x$exfix))
    names(epar) <- names(x$expar)
    if (!(ii$npar.ex==0)) {
        idxE <- which(ii$e1==1)
        epar[idxE] <- pidxE
    }
    if (any(fixed))
        epar[fixed] <- unlist(x$exfix[fixed])
    for (p in ii$eparname) {
        idx <- which(x$exfix==p)
        if (!(p%in%c(ii$parname,ii$covparname,ii$mparname))) {
            if (length(idx)>1) {
                idxE <- c(idxE,idx[-1])
                pidxE <- c(pidxE,rep(epar[idx[1]],length(idx)-1))
            }
            parval[[p]] <- epar[idx[1]]
        }
        attributes(parval[[p]])$e.idx <- idx

        if (length(idx)>1)
            epar[idx[-1]] <- parval[[p]]
        if (p %in% setdiff(ii$covparname,c(ii$parname,ii$mparname))) {
            parBelongsTo$cov <- c(parBelongsTo$cov,p)
            idx.2 <- which(x$covpar==p)
            epar[idx] <- P[idx.2[1]]
            pidxE <- c(pidxE,rep(P[idx.2[1]],length(idx)))
            idxE <- c(idxE,idx)
        }
        if (p %in% setdiff(ii$parname,ii$mparname)) {
            parBelongsTo$reg <- c(parBelongsTo$reg,p)
            idx.2 <- which(x$par==p)
            epar[idx] <- A[idx.2[1]]
            pidxE <- c(pidxE,rep(A[idx.2[1]],length(idx)))
            idxE <- c(idxE,idx)
        }
        if (p %in% ii$mparname) {
            parBelongsTo$mean <- c(parBelongsTo$mean,p)
            idx.2 <- which(x$mean==p)
            epar[idx] <- v[idx.2[1]]
            pidxE <- c(pidxE,rep(v[idx.2[1]],length(idx)))
            idxE <- c(idxE,idx)
        }
    }
    ee <- cbind(idxE,pidxE); rownames(ee) <- names(x$expar)[ee[,1]]

    ## Constrained...
    constrain.par <- names(constrain(x))
    constrain.idx <- NULL

    if (length(constrain.par)>0) {
        constrain.idx <- list()
        for (p in constrain.par) {
            reg.tidx <- reg.idx <- cov.idx <- m.idx <- e.idx <- NULL
            myc <- constrain(x)[[p]]
            xargs <- manifest(x)[na.omit(match(attributes(myc)$args,manifest(x)))]
            if (length(xargs)>0) {
                parval[xargs] <- 0
            }
            if (p%in%ii$parname.all) {
                reg.idx <- which(x$par==p)
                reg.tidx <- which(t(x$par==p))
            }
            if (p%in%ii$covparname.all) {
                cov.idx <- which(x$covpar==p)
            }
            if (p%in%ii$mparname.all) {
                m.idx <- which(x$mean==p)
            }
            if (p%in%ii$eparname.all) {
                e.idx <- which(x$exfix==p)
            }
            constrain.idx[[p]] <- list(reg.idx=reg.idx,reg.tidx=reg.tidx,cov.idx=cov.idx,m.idx=m.idx,e.idx=e.idx)
        }
    }

    parBelongsTo <- lapply(parBelongsTo,function(x) sort(unique(x)))

    meanl <- idxM
    if (!is.null(meanl)) meanl <- cbind(meanl, pidxM)
    regl <- idxA
    if (!is.null(regl)) regl <- cbind(regl, pidxA)
    covl <- idxP
    if (!is.null(covl)) covl <- cbind(covl, pidxP)

    return(list(mean=meanl,
                reg=regl,
                cov=covl,
                epar=ee,
                parval=parval,
                constrain.idx=constrain.idx,
                parBelongsTo=parBelongsTo))

}

##' @export
matrices.lvm <- function(x,pars,meanpar=NULL,epars=NULL,data=NULL,...) {
    ii <- index(x)
    pp <- c(rep(NA,ii$npar.mean),pars,epars)
    ##v <- NULL
    v <- ii$v0
    if (!is.null(meanpar) && length(meanpar)>0) {
        pp[seq(ii$npar.mean)] <- meanpar
        v[ii$mean[,1]] <- meanpar[ii$mean[,2]]
    }

    A <- ii$A
    A[ii$reg[,1]] <- pp[ii$reg[,2]]
    P <- ii$P
    P[ii$cov[,1]] <- pp[ii$cov[,2]]
    e <- NULL

    if (length(x$expar)>0) {
        e <- rep(0,length(x$expar))
        fixed <- sapply(x$exfix, function(y) is.numeric(y) & !is.na(y))
        if (any(fixed))
            e[fixed] <- unlist(x$exfix[fixed])
        if (nrow(ii$epar)>0)
            e[ii$epar[,1]] <- pp[ii$epar[,2]]
        names(e) <- names(x$expar)
    }

    parval <- lapply(ii$parval,function(x) {
        res <- pp[x];
        attributes(res) <- attributes(x);
        res })
    ## Constrained...
    constrain.par <- names(constrain(x))
    constrain.idx <- NULL
    cname <- constrainpar <- c()
    if (length(constrain.par)>0 && is.numeric(c(pars,meanpar))) {
        constrain.idx <- list()
        for (p in constrain.par) {
            cname <- c(cname,p)
            myc <- constrain(x)[[p]]
            xargs <- manifest(x)[na.omit(match(attributes(myc)$args,manifest(x)))]
            if (length(xargs)>0) {
                if (!is.null(data)) {
                    parval[xargs] <- (data)[xargs]
                } else parval[xargs] <- 0
            }
            val <- unlist(c(parval,constrainpar,x$mean,e)[attributes(myc)$args])
            cpar <- myc(val);
            constrainpar <- c(constrainpar,list(cpar)); names(constrainpar) <- cname
            if (p%in%ii$parname.all) {
                if (!is.null(val))
                    A[ii$constrain.idx[[p]]$reg.idx] <- cpar
            }
            if (p%in%ii$covparname.all) {
                if (!is.null(val))
                    P[ii$constrain.idx[[p]]$cov.idx] <- cpar
            }
            if (p%in%ii$mparname.all) {
                if (!is.null(val))
                    v[ii$constrain.idx[[p]]$m.idx] <- cpar
            }
            if (p%in%ii$eparname.all) {
                if (!is.null(val))
                    e[ii$constrain.idx[[p]]$e.idx] <- cpar
            }
        }
    }

    return(list(A=A, P=P, v=v, e=e, parval=parval,
                constrain.idx=ii$constrain.idx, constrainpar=constrainpar))
}

###}}} matrices.lvm

###{{{ matrices.multigroup

##' @export
matrices.multigroup <- function(x, p, ...) {
    pp <- modelPar(x,p)
    res <- list()
    for (i in seq_len(x$ngroup))
        res <- c(res, list(matrices2(x$lvm[[i]],pp$p[[i]])))
    return(res)
}

###}}}

matrices2 <- function(x,p,...) {
    m0 <- p[seq_len(index(x)$npar.mean)]
    p0 <- p[with(index(x),seq_len(npar)+npar.mean)]
    e0 <- p[with(index(x),seq_len(npar.ex)+npar.mean+npar)]
    matrices(x,p0,m0,e0,...)
}

###{{{ matrices, to be superseeded by above definition

matrices.lvm <- function(x,pars,meanpar=NULL,epars=NULL,data=NULL,...) {
    ii <- index(x)
    A <- ii$A ## Matrix with fixed parameters and ones where parameters are free
    M1 <- ii$M1 ## Index of free and _unique_ regression parameters
    P <- ii$P  ## Matrix with fixed variance parameters and ones where parameters are free
    P1 <- ii$P1 ## Index of free and _unique_ regression parameters

    constrain.par <- names(constrain(x))
    parval <- list()

    if (ii$npar.reg>0) {
        A[which(M1==1)] <- pars[seq_len(ii$npar.reg)]
        for (p in ii$parname) {
            idx <- which((x$par==p))
            newval <- A[idx[1]]
            attributes(newval)$reg.idx <- idx
            attributes(newval)$reg.tidx <- which(t(x$par==p))
            parval[[p]] <- newval
            if (length(idx)>1) {
                A[idx[-1]] <- parval[[p]]
            }
        } ## duplicate parameters
    }

    if (ii$npar.reg==0) {
        pars.var <- pars
    } else {
        pars.var <- pars[-seq_len(ii$npar.reg)]
    }

    diag(P)[ii$which.diag] <- pars.var[seq_along(ii$which.diag)]
    pars.off.diag <- pars.var
    if (length(ii$which.diag)>0)
        pars.off.diag <- pars.off.diag[-seq_along(ii$which.diag)]
    counter <- 0
    if (length(pars.off.diag)>0 & ncol(P)>1)
        for (i in seq_len(ncol(P1)-1))
            for (j in seq(i+1,nrow(P1))) {
                if (ii$P1[j,i]!=0) {
                    counter <- counter+1
                    P[j,i] <- pars.off.diag[counter]
                }
            }

    if (length(ii$covparname)>0)
        for (p in ii$covparname) {
            idx <- which(x$covpar==p)
            if (!(p%in%ii$parname)) {
                parval[[p]] <- P[idx[1]]
            }
            attributes(parval[[p]])$cov.idx <- idx
            if (length(idx)>1) {
                P[idx[-1]] <- parval[[p]]
            }
            if (ii$npar.reg>0 && p%in%ii$parname) {
                idx.reg <- which(x$par==p)
                P[idx] <- A[idx.reg[1]]
                atr <- attributes(parval[[p]])
                parval[[p]] <- A[idx.reg[1]] ###?????
                attributes(parval[[p]]) <- atr
            }
        } ## duplicate parameters
    P[upper.tri(P)] <- t(P)[upper.tri(P)]  ## Symmetrize...


    v <- NULL
    {
        ## named <- sapply(x$mean, function(y) is.character(y) & !is.na(y))
        fixed <- sapply(x$mean, function(y) is.numeric(y) & !is.na(y))
        v <- rep(0,length(x$mean))
        names(v) <- colnames(P)
        if (!(is.null(meanpar) | ii$npar.mean==0))
            v[ii$v1==1] <- meanpar
        if (any(fixed))
            v[fixed] <- unlist(x$mean[fixed])

        for (p in ii$mparname) {
            idx <- which(x$mean==p)

            if (!(p%in%c(ii$parname,ii$covparname))) {
                parval[[p]] <- v[idx[1]]
            }
            attributes(parval[[p]])$m.idx <- idx

            if (length(idx)>1)
                v[idx[-1]] <- parval[[p]]
            if (p %in% ii$covparname & !(p %in% ii$parname)) {
                idx.2 <- which(x$covpar==p)
                v[idx] <- P[idx.2[1]]
            }
            if (p %in% ii$parname) {
                idx.2 <- which(x$par==p)
                v[idx] <- A[idx.2[1]]
            }
        }
    }


    ## Ex-parameters
    e <- NULL
    {
        ## named <- sapply(x$exfix, function(y) is.character(y) & !is.na(y))
        fixed <- sapply(x$exfix, function(y) is.numeric(y) & !is.na(y))

        e <- rep(0,length(x$exfix))

        names(e) <- names(x$expar)
        if (!(is.null(epars) | ii$npar.ex==0))
            e[which(ii$e1==1)] <- epars
        if (any(fixed))
            e[fixed] <- unlist(x$exfix[fixed])
        for (p in ii$eparname) {
            idx <- which(x$exfix==p)
            if (!(p%in%c(ii$parname,ii$covparname,ii$mparname))) {
                parval[[p]] <- e[idx[1]]
            }
            attributes(parval[[p]])$e.idx <- idx

            if (length(idx)>1)
                e[idx[-1]] <- parval[[p]]
            if (p %in% setdiff(ii$covparname,c(ii$parname,ii$mparname))) {
                idx.2 <- which(x$covpar==p)
                e[idx] <- P[idx.2[1]]
            }
            if (p %in% setdiff(ii$parname,ii$mparname)) {
                idx.2 <- which(x$par==p)
                e[idx] <- A[idx.2[1]]
            }
            if (p %in% ii$mparname) {
                idx.2 <- which(x$mean==p)
                e[idx] <- v[idx.2[1]]
            }
        }
    }

    ## Constrained...
    constrain.idx <- NULL
    cname <- constrainpar <- c()
    if (length(constrain.par)>0 && is.numeric(c(pars,meanpar,e))) {
        constrain.idx <- list()
        for (p in constrain.par) {
            cname <- c(cname,p)
            reg.tidx <- reg.idx <- cov.idx <- m.idx <- e.idx <- NULL
            myc <- constrain(x)[[p]]
            xargs <- manifest(x)[na.omit(match(attributes(myc)$args,manifest(x)))]
            if (length(xargs)>0) {
                if (!is.null(data)) {
                    parval[xargs] <- (data)[xargs]
                } else parval[xargs] <- 0
            }
            val <- rbind(unlist(c(parval,constrainpar,x$mean,e)[attributes(myc)$args]))
            cpar <- myc(val);
            constrainpar <- c(constrainpar,list(cpar)); names(constrainpar) <- cname
            if (p%in%ii$parname.all) {
                reg.idx <- which(x$par==p)
                reg.tidx <- which(t(x$par==p))
                if (!is.null(val))
                    A[reg.idx] <- cpar##myc(val)
            }
            if (p%in%ii$covparname.all) {
                cov.idx <- which(x$covpar==p)
                if (!is.null(val))
                    P[cov.idx] <- cpar##myc(val)
            }
            if (p%in%ii$mparname.all) {
                m.idx <- which(x$mean==p)
                if (!is.null(val))
                    v[m.idx] <- cpar##myc(val)
            }
            if (p%in%ii$eparname.all) {
                e.idx <- which(x$exfix==p)
                if (!is.null(val))
                    e[e.idx] <- cpar##myc(val)
            }
            constrain.idx[[p]] <- list(reg.idx=reg.idx,reg.tidx=reg.tidx,cov.idx=cov.idx,m.idx=m.idx,e.idx=e.idx)
        }
    }

    if (x$index$sparse & !is.character(class(pars)[1])) {
        A <- as(A,"sparseMatrix")
        P <- as(P,"sparseMatrix")
        v <- as(v,"sparseMatrix")
    }
    return(list(A=A, P=P, v=v, e=e, parval=parval, constrain.idx=constrain.idx, constrainpar=constrainpar))
}

###}}} matrices Obsolete

ordinal_remove_hook <- function(x,var,...) {
    ordinal(x,K=0) <- var
    return(x)
}

color_ordinal <- function(x,subset=vars(x),...) {
    return(list(vars=intersect(subset,ordinal(x)),col="indianred1"))
}

ordinal_sim_hook <- function(x,data,p,modelpar,...) {
    ovar <- ordinal(x)

    for (i in seq_len(length(ovar))) {
        if (attributes(ovar)$liability[i]) {
            idx <- attributes(ovar)$idx[[ovar[i]]]
            if (length(idx)==0) {
                breaks <- c(-Inf,0,Inf)
            } else {
                breaks <- c(-Inf,ordreg_threshold(modelpar$e[idx]),Inf)
            }
            z <- cut(data[,ovar[i]],breaks=breaks)
            data[,ovar[i]] <- as.numeric(z)-1
        }
        K <- attributes(ovar)$K[i]
        lab <- attributes(ovar)$labels[ovar[i]][[1]]
        if (!is.null(lab))
            data[,ovar[i]] <- factor(data[,ovar[i]],
                                     levels=seq(K)-1,
                                     labels=lab)
    }
    return(data)
}

ordinal_estimate_hook <- function(x,data,weights,data2,estimator,...) {
    dots <- list(...)

    nestimator <- c("normal")
    nestimator2 <- c("tobit","tobitw","gaussian")

    ord <- ordinal(x)
    bin <- NULL
    if (is.null(estimator) && length(ord)>0) estimator <- nestimator[1]

    ## Binary outcomes -> censored regression
    if (is.null(dim(data))) return(NULL)
    if (is.null(estimator) || estimator%in%c(nestimator2,nestimator)) {
        for (i in setdiff(lava::endogenous(x),bin)) {
            if (is.character(data[,i]) | is.factor(data[,i])) { # Transform binary 'factor'
                y <- as.factor(data[,i])
                data[,i] <- as.numeric(y)-1
                estimator <- nestimator[1]
                ordinal(x,K=nlevels(y)) <- i
            }
        }
        ord <- ordinal(x)
        if (length(ord)>0 && !is.null(estimator) && estimator%in%nestimator2) {
          estimator <- nestimator[1]
        }
    }

    ## Transform 'Surv' objects
    data2 <- mynames <- NULL
    if (is.null(estimator) || estimator%in%nestimator[1] ) {
        for (i in setdiff(lava::endogenous(x),c(bin,ord))) {
            if (survival::is.Surv(data[,i])) {
                S <- data[,i]
                y1 <- S[,1]
                if (attributes(S)$type=="left")  {
                    y2 <- y1
                    y1[S[,2]==0] <- -Inf
                }
                if (attributes(S)$type=="right") {
                    y2 <- y1
                    y2[S[,2]==0] <- Inf
                }
                if (attributes(S)$type=="interval2") {
                    y2 <- S[,2]
                }
                if (attributes(S)$type=="interval") {
                    y2 <- S[,2]
                    y2[S[,3]==1L] <- y1[S[,3]==1L]
                }
                if (!(attributes(S)$type%in%c("left","right","interval2","interval"))) stop("Surv type not supported.")
                mynames <- c(mynames,i)
                y2 <- cbind(y2)
                colnames(y2) <- i
                data2 <- cbind(data2,y2)
                data[,i] <- y1
                estimator <- "normal"
            }
        }
    }

    return(c(list(x=x,data=data,weights=weights,data2=data2,estimator=estimator),dots))
}


##' Define variables as ordinal
##'
##' Define variables as ordinal in latent variable model object
##' @export
##' @aliases ordinal ordinal<-
##' @param x Object
##' @param ... additional arguments to lower level functions
##' @param value variable (formula or character vector)
##' @examples
##' if (requireNamespace("mets")) {
##' m <- lvm(y + z ~ x + 1*u[0], latent=~u)
##' ordinal(m, K=3) <- ~y+z
##' d <- sim(m, 100, seed=1)
##' e <- estimate(m, d)
##' }
##' 
"ordinal<-" <- function(x,...,value) UseMethod("ordinal<-")

##' @export
"ordinal<-.lvm" <- function(x,...,value) {
  ordinal(x, value, ...)
}

##' @export
"ordinal" <- function(x,...) UseMethod("ordinal")

##' @export
print.ordinal.lvm <- function(x,...) {
  cat(rep("_",28),"\n",sep="")
  for (i in x) {
    val <- attr(x,"fix")[[i]]
    if (length(val)==0)
      cat(paste(i,"binary",sep=":"),"\n")
    else print(unlist(attr(x,"fix")[[i]]),quote=FALSE)
    cat(rep("_",28),"\n",sep="")
  }
}

##' @export
`ordinal.lvm` <- function(x,var=NULL,K=2, constrain, breaks=NULL, p, liability=TRUE, labels, exo=FALSE, ...) {
    if (inherits(var,"formula")) {
        var <- all.vars(var)
    }
    if (is.null(var)) {
        ordidx <- unlist(x$attributes$ordinal)
        KK <- unlist(x$attributes$nordinal)
        idx <- x$attributes$ordinalparname
        fix <- lapply(idx,function(z) x$exfix[z])
        liability <- x$attributes$liability
        labels <- x$attributes$labels
        if (length(ordidx)>0) {
            val <- names(ordidx)
            return(structure(val,K=KK,idx=idx,fix=fix,liability=liability,labels=labels,class="ordinal.lvm"))
        }
        else
            return(NULL)
    }
    if (K[1]==0L || is.null(K[1]) || (is.logical(K) & !K[1])) {
        idx <- na.omit(match(var,names(x$attributes$ordinal)))
        if (length(idx)>0) {
            pp <- unlist(x$attributes$ordinalparname[idx])
            if (!is.null(pp)) parameter(x,remove=TRUE) <- pp
            if (!is.null(x$attributes$ordinalparname))
                x$attributes$ordinalparname <- x$attributes$ordinalparname[-idx]
            x$attributes$ordinal <- x$attributes$ordinal[-idx]
            ##x$attributes$labels[var] <- NULL
            x$attributes$type <- x$attributes$type[-idx]
            x$attributes$constrainY <- x$attributes$constrainY[setdiff(names(x$attributes$constrainY),var)]
            x$attributes$liability <- x$attributes$liability[-idx]
            x$attributes$nordinal <- x$attributes$nordinal[-idx]
            x$attributes$normal <- x$attributes$normal[-idx]
            exo <- intersect(var,exogenous(x,latent=TRUE))
            if (length(exo)>0) {
                intercept(x,var) <- NA
                covariance(x,var) <- NA
                exogenous(x) <- union(exogenous(x),exo)
            }
        }
        return(x)
    }
    
    if (!missing(p)) breaks <- qnorm(cumsum(p))
    if (!is.null(breaks)) {
        breaks <- ordreg_ithreshold(breaks)
        K <- length(breaks)+1
    }
    if (!missing(labels)) K <- length(labels)
    if (length(var)>length(K)) K <- rep(K[1],length(var))
    if (length(var)==1 && !missing(constrain)) constrain <- list(constrain)
    if (length(var)>1) {
        if (!missing(labels) && !is.list(labels)) labels <- rep(list(labels),length(var))
        if (!missing(breaks) && !is.list(breaks)) breaks <- rep(list(breaks),length(var))
        if (!missing(constrain) && !is.list(constrain)) constrain <- rep(list(constrain),length(var))
    }

    addvar(x) <- var
    for (i in seq_len(length(var))) {
        if (K[i]>2 || (K[i]==2 && !liability)) {
            parname <- paste0(var[i],":",paste(seq(K[i]-1)-1,seq(K[i]-1),sep="|"))
            newpar <- if (is.null(breaks)) {
                rep(-1,K[i]-1)
            } else if (is.list(breaks)) breaks[[i]] else breaks
            if (length(newpar)<K[i]-1) stop("Wrong number of starting values")
            newfix <- if (missing(constrain))
                rep(list(NA),length(newpar)) else constrain[[i]]
            if (length(newfix)<K[i]-1) stop("Wrong number of constraints")
            names(newpar) <- names(newfix) <- parname

            parameter(x,newfix,start=newpar) <- names(newfix)
            ## pex <- parname%in%names(x$expar)
            ## if (any(pex <- parname%in%names(x$expar))) {
            ##     if (!all(pex)) stop("Cannot change number of categories! Re-specify model.")
            ##     x$attributes$iordinal[var] <- list(idx)
            ## }
            x$attributes$ordinalparname[var[i]] <- list(names(newfix))
        }
        x$attributes$type[var[i]] <- ifelse(K[i]>2,"categorical","binary")
        if (K[i]>2) intfix(x,var[i],NULL) <- 0
        if (!liability) {
            mytr <- function(y,p,idx,...) {
                breaks <- c(-Inf,ordreg_threshold(p[idx]),Inf)
                as.numeric(cut(y,breaks=breaks))-1
            }
            myalist <- substitute(alist(y=,p=,idx=pp),
                                  list(pp=x$attributes$ordinalparname[[var[i]]]))
            formals(mytr) <- eval(myalist)
            transform(x,var[i],post=FALSE) <- mytr

        }
    }
    x$attributes$liability[var] <- liability
    x$attributes$ordinal[var] <- TRUE
    if (!missing(labels)) {
        if (length(var)==1) labels <- list(labels)
        x$attributes$labels[var] <- labels
    }
    x$attributes$nordinal[var] <- K
    x$attributes$normal[var] <- FALSE
    covfix(x,var,NULL,exo=exo) <- 1
    if (is.null(index(x))) index(x) <- reindex(x)
    return(x)
}

`matrices` <-
    function(x,...) UseMethod("matrices")

##' @export
mat.lvm <- function(x,ii=index(x),...) {
    A <- ii$A ## Matrix with fixed parameters and ones where parameters are free
    M1 <- ii$M1 ## Index of free and _unique_ regression parameters
    P <- ii$P  ## Matrix with fixed variance parameters and ones where parameters are free
    P1 <- ii$P1 ## Index of free and _unique_ regression parameters

    constrain.par <- names(constrain(x))
    parval <- list()


    parBelongsTo <- list(mean=seq_len(ii$npar.mean),
                         reg=seq_len(ii$npar.reg)+ii$npar.mean,
                         cov=seq_len(ii$npar.var)+ii$npar.mean+ii$npar.reg,
                         epar=seq_len(ii$npar.ex)+with(ii,npar.reg+npar.var+npar.mean),
                         cpar=numeric())

    idxA <- which(M1==1)
    pidxA <- parBelongsTo$reg
    if (ii$npar.reg>0) {
        A[idxA] <- pidxA
        for (p in ii$parname) {
            idx <- which((x$par==p))
            newval <- A[idx[1]]
            attributes(newval)$reg.idx <- idx
            attributes(newval)$reg.tidx <- which(t(x$par==p))
            parval[[p]] <- newval
            if (length(idx)>1) {
                idxA <- c(idxA,idx[-1])
                pidxA <- c(pidxA,rep(A[idx[1]],length(idx)-1))
                A[idx] <- A[idx[1]]
            }
        } ## duplicate parameters
    }

    pars.var <- parBelongsTo$cov
    idxdiag <- (seq(ncol(P1))-1)*ncol(P1) + seq(ncol(P1))
    idxP <- idxdiag[which(P1[idxdiag]==1)]
    pidxP <- pars.var[seq_len(length(idxP))]
    P[idxP] <- pidxP

    pars.off.diag <- pars.var
    if (length(pidxP)>0) {
        pars.off.diag <- pars.off.diag[-seq_len(length(pidxP))]
    }

    counter <- 0
    if (length(pars.off.diag)>0 & ncol(P)>1)
        for (i in seq_len(ncol(P1)-1))
            for (j in seq(i+1,nrow(P1))) {
                if (ii$P1[j,i]!=0) {
                    counter <- counter+1
                    pos <- c(j+(i-1)*ncol(P1),
                             i+(j-1)*ncol(P1))
                    P[j,i] <- P[i,j] <- pars.off.diag[counter]
                    idxP <- c(idxP,pos); pidxP <- c(pidxP,P[j,i],P[i,j])
                }
            }

    if (length(ii$covparname)>0)
        for (p in ii$covparname) {
            idx <- which(x$covpar==p)
            isOffDiag <- !(idx[1]%in%idxdiag)
            if (!(p%in%ii$parname)) {
                parval[[p]] <- P[idx[1]]
            }
            attributes(parval[[p]])$cov.idx <- idx
            if (length(idx)>1+isOffDiag) {
                P[idx[-seq(1+isOffDiag)]] <- parval[[p]]
            }
            if (ii$npar.reg>0 && p%in%ii$parname) {
                parBelongsTo$reg <- c(parBelongsTo$reg,p)
                idx.reg <- which(x$par==p)
                P[idx] <- A[idx.reg[1]]
                atr <- attributes(parval[[p]])
                parval[[p]] <- A[idx.reg[1]]
                attributes(parval[[p]]) <- atr
                idxP <- c(idxP,idx)
                pidxP <- c(pidxP,rep(P[idx[1]],length(idx)))
            } else {
                idxP <- c(idxP,idx[-seq(1+isOffDiag)])
                pidxP <- c(pidxP,rep(P[idx[1]],length(idx)-1-isOffDiag))
            }
        } ## duplicate parameters

    idxM <- c()
    pidxM <- seq_len(ii$npar.mean)
    v <- NULL

    ## named <- sapply(x$mean, function(y) is.character(y) & !is.na(y))
    fixed <- sapply(x$mean, function(y) is.numeric(y) & !is.na(y))
    v <- rep(0,length(x$mean))
    names(v) <- colnames(P)
    if (ii$npar.mean>0) {
        idxM <- which(ii$v1==1)
        v[idxM] <- pidxM
    }
    if (any(fixed))
        v[fixed] <- unlist(x$mean[fixed])

    for (p in ii$mparname) {
        idx <- which(x$mean==p)
        if (!(p%in%c(ii$parname,ii$covparname))) {
            if (length(idx)>1) {
                pidxM <- c(pidxM,rep(v[idx[1]],length(idx)-1))
                idxM <- c(idxM,idx[-1])
            }
            parval[[p]] <- v[idx[1]]
            v[idx] <- parval[[p]]
        }
        attributes(parval[[p]])$m.idx <- idx
        if (p %in% ii$covparname & !(p %in% ii$parname)) {
            parBelongsTo$cov <- c(parBelongsTo$cov,p)
            idx.2 <- which(x$covpar==p)
            v[idx] <- P[idx.2[1]]
            pidxM <- c(pidxM,rep(P[idx.2[1]],length(idx)))
            idxM <- c(idxM,idx)
        }
        if (p %in% ii$parname) {
            parBelongsTo$reg <- c(parBelongsTo$reg,p)
            idx.2 <- which(x$par==p)
            v[idx] <- A[idx.2[1]]
            pidxM <- c(pidxM,rep(A[idx.2[1]],length(idx)))
            idxM <- c(idxM,idx)
        }
    }

    ## Ex-parameters
    idxE <- NULL
    pidxE <- parBelongsTo$epar
    ## named <- sapply(x$exfix, function(y) is.character(y) & !is.na(y))
    fixed <- sapply(x$exfix, function(y) is.numeric(y) & !is.na(y))
    epar <- rep(0,length(x$exfix))
    names(epar) <- names(x$expar)
    if (!(ii$npar.ex==0)) {
        idxE <- which(ii$e1==1)
        epar[idxE] <- pidxE
    }
    if (any(fixed))
        epar[fixed] <- unlist(x$exfix[fixed])
    for (p in ii$eparname) {
        idx <- which(x$exfix==p)
        if (!(p%in%c(ii$parname,ii$covparname,ii$mparname))) {
            if (length(idx)>1) {
                idxE <- c(idxE,idx[-1])
                pidxE <- c(pidxE,rep(epar[idx[1]],length(idx)-1))
            }
            parval[[p]] <- epar[idx[1]]
        }
        attributes(parval[[p]])$e.idx <- idx

        if (length(idx)>1)
            epar[idx[-1]] <- parval[[p]]
        if (p %in% setdiff(ii$covparname,c(ii$parname,ii$mparname))) {
            parBelongsTo$cov <- c(parBelongsTo$cov,p)
            idx.2 <- which(x$covpar==p)
            epar[idx] <- P[idx.2[1]]
            pidxE <- c(pidxE,rep(P[idx.2[1]],length(idx)))
            idxE <- c(idxE,idx)
        }
        if (p %in% setdiff(ii$parname,ii$mparname)) {
            parBelongsTo$reg <- c(parBelongsTo$reg,p)
            idx.2 <- which(x$par==p)
            epar[idx] <- A[idx.2[1]]
            pidxE <- c(pidxE,rep(A[idx.2[1]],length(idx)))
            idxE <- c(idxE,idx)
        }
        if (p %in% ii$mparname) {
            parBelongsTo$mean <- c(parBelongsTo$mean,p)
            idx.2 <- which(x$mean==p)
            epar[idx] <- v[idx.2[1]]
            pidxE <- c(pidxE,rep(v[idx.2[1]],length(idx)))
            idxE <- c(idxE,idx)
        }
    }
    ee <- cbind(idxE,pidxE); rownames(ee) <- names(x$expar)[ee[,1]]

    ## Constrained...
    constrain.par <- names(constrain(x))
    constrain.idx <- NULL

    if (length(constrain.par)>0) {
        constrain.idx <- list()
        for (p in constrain.par) {
            reg.tidx <- reg.idx <- cov.idx <- m.idx <- e.idx <- NULL
            myc <- constrain(x)[[p]]
            xargs <- manifest(x)[na.omit(match(attributes(myc)$args,manifest(x)))]
            if (length(xargs)>0) {
                parval[xargs] <- 0
            }
            if (p%in%ii$parname.all) {
                reg.idx <- which(x$par==p)
                reg.tidx <- which(t(x$par==p))
            }
            if (p%in%ii$covparname.all) {
                cov.idx <- which(x$covpar==p)
            }
            if (p%in%ii$mparname.all) {
                m.idx <- which(x$mean==p)
            }
            if (p%in%ii$eparname.all) {
                e.idx <- which(x$exfix==p)
            }
            constrain.idx[[p]] <- list(reg.idx=reg.idx,reg.tidx=reg.tidx,cov.idx=cov.idx,m.idx=m.idx,e.idx=e.idx)
        }
    }

    parBelongsTo <- lapply(parBelongsTo,function(x) sort(unique(x)))

    meanl <- idxM
    if (!is.null(meanl)) meanl <- cbind(meanl, pidxM)
    regl <- idxA
    if (!is.null(regl)) regl <- cbind(regl, pidxA)
    covl <- idxP
    if (!is.null(covl)) covl <- cbind(covl, pidxP)

    return(list(mean=meanl,
                reg=regl,
                cov=covl,
                epar=ee,
                parval=parval,
                constrain.idx=constrain.idx,
                parBelongsTo=parBelongsTo))

}

##' @export
matrices.lvm <- function(x,pars,meanpar=NULL,epars=NULL,data=NULL,...) {
    ii <- index(x)
    pp <- c(rep(NA,ii$npar.mean),pars,epars)
    ##v <- NULL
    v <- ii$v0
    if (!is.null(meanpar) && length(meanpar)>0) {
        pp[seq(ii$npar.mean)] <- meanpar
        v[ii$mean[,1]] <- meanpar[ii$mean[,2]]
    }

    A <- ii$A
    A[ii$reg[,1]] <- pp[ii$reg[,2]]
    P <- ii$P
    P[ii$cov[,1]] <- pp[ii$cov[,2]]
    e <- NULL

    if (length(x$expar)>0) {
        e <- rep(0,length(x$expar))
        fixed <- sapply(x$exfix, function(y) is.numeric(y) & !is.na(y))
        if (any(fixed))
            e[fixed] <- unlist(x$exfix[fixed])
        if (nrow(ii$epar)>0)
            e[ii$epar[,1]] <- pp[ii$epar[,2]]
        names(e) <- names(x$expar)
    }

    parval <- lapply(ii$parval,function(x) {
        res <- pp[x];
        attributes(res) <- attributes(x);
        res })
    ## Constrained...
    constrain.par <- names(constrain(x))
    constrain.idx <- NULL
    cname <- constrainpar <- c()
    if (length(constrain.par)>0 && is.numeric(c(pars,meanpar))) {
        constrain.idx <- list()
        for (p in constrain.par) {
            cname <- c(cname,p)
            myc <- constrain(x)[[p]]
            xargs <- manifest(x)[na.omit(match(attributes(myc)$args,manifest(x)))]
            if (length(xargs)>0) {
                if (!is.null(data)) {
                    parval[xargs] <- (data)[xargs]
                } else parval[xargs] <- 0
            }
            val <- unlist(c(parval,constrainpar,x$mean,e)[attributes(myc)$args])
            cpar <- myc(val);
            constrainpar <- c(constrainpar,list(cpar)); names(constrainpar) <- cname
            if (p%in%ii$parname.all) {
                if (!is.null(val))
                    A[ii$constrain.idx[[p]]$reg.idx] <- cpar
            }
            if (p%in%ii$covparname.all) {
                if (!is.null(val))
                    P[ii$constrain.idx[[p]]$cov.idx] <- cpar
            }
            if (p%in%ii$mparname.all) {
                if (!is.null(val))
                    v[ii$constrain.idx[[p]]$m.idx] <- cpar
            }
            if (p%in%ii$eparname.all) {
                if (!is.null(val))
                    e[ii$constrain.idx[[p]]$e.idx] <- cpar
            }
        }
    }

    return(list(A=A, P=P, v=v, e=e, parval=parval,
                constrain.idx=ii$constrain.idx, constrainpar=constrainpar))
}

##' @export
matrices.multigroup <- function(x, p, ...) {
    pp <- modelPar(x,p)
    res <- list()
    for (i in seq_len(x$ngroup))
        res <- c(res, list(matrices2(x$lvm[[i]],pp$p[[i]])))
    return(res)
}

matrices2 <- function(x,p,...) {
    m0 <- p[seq_len(index(x)$npar.mean)]
    p0 <- p[with(index(x),seq_len(npar)+npar.mean)]
    e0 <- p[with(index(x),seq_len(npar.ex)+npar.mean+npar)]
    matrices(x,p0,m0,e0,...)
}

matrices.lvm <- function(x,pars,meanpar=NULL,epars=NULL,data=NULL,...) {
    ii <- index(x)
    A <- ii$A ## Matrix with fixed parameters and ones where parameters are free
    M1 <- ii$M1 ## Index of free and _unique_ regression parameters
    P <- ii$P  ## Matrix with fixed variance parameters and ones where parameters are free
    P1 <- ii$P1 ## Index of free and _unique_ regression parameters

    constrain.par <- names(constrain(x))
    parval <- list()

    if (ii$npar.reg>0) {
        A[which(M1==1)] <- pars[seq_len(ii$npar.reg)]
        for (p in ii$parname) {
            idx <- which((x$par==p))
            newval <- A[idx[1]]
            attributes(newval)$reg.idx <- idx
            attributes(newval)$reg.tidx <- which(t(x$par==p))
            parval[[p]] <- newval
            if (length(idx)>1) {
                A[idx[-1]] <- parval[[p]]
            }
        } ## duplicate parameters
    }

    if (ii$npar.reg==0) {
        pars.var <- pars
    } else {
        pars.var <- pars[-seq_len(ii$npar.reg)]
    }

    diag(P)[ii$which.diag] <- pars.var[seq_along(ii$which.diag)]
    pars.off.diag <- pars.var
    if (length(ii$which.diag)>0)
        pars.off.diag <- pars.off.diag[-seq_along(ii$which.diag)]
    counter <- 0
    if (length(pars.off.diag)>0 & ncol(P)>1)
        for (i in seq_len(ncol(P1)-1))
            for (j in seq(i+1,nrow(P1))) {
                if (ii$P1[j,i]!=0) {
                    counter <- counter+1
                    P[j,i] <- pars.off.diag[counter]
                }
            }

    if (length(ii$covparname)>0)
        for (p in ii$covparname) {
            idx <- which(x$covpar==p)
            if (!(p%in%ii$parname)) {
                parval[[p]] <- P[idx[1]]
            }
            attributes(parval[[p]])$cov.idx <- idx
            if (length(idx)>1) {
                P[idx[-1]] <- parval[[p]]
            }
            if (ii$npar.reg>0 && p%in%ii$parname) {
                idx.reg <- which(x$par==p)
                P[idx] <- A[idx.reg[1]]
                atr <- attributes(parval[[p]])
                parval[[p]] <- A[idx.reg[1]] # CHECK
                attributes(parval[[p]]) <- atr
            }
        } ## duplicate parameters
    P[upper.tri(P)] <- t(P)[upper.tri(P)]  ## Symmetrize...


    v <- NULL
    {
        ## named <- sapply(x$mean, function(y) is.character(y) & !is.na(y))
        fixed <- sapply(x$mean, function(y) is.numeric(y) & !is.na(y))
        v <- rep(0,length(x$mean))
        names(v) <- colnames(P)
        if (!(is.null(meanpar) | ii$npar.mean==0))
            v[ii$v1==1] <- meanpar
        if (any(fixed))
            v[fixed] <- unlist(x$mean[fixed])

        for (p in ii$mparname) {
            idx <- which(x$mean==p)

            if (!(p%in%c(ii$parname,ii$covparname))) {
                parval[[p]] <- v[idx[1]]
            }
            attributes(parval[[p]])$m.idx <- idx

            if (length(idx)>1)
                v[idx[-1]] <- parval[[p]]
            if (p %in% ii$covparname & !(p %in% ii$parname)) {
                idx.2 <- which(x$covpar==p)
                v[idx] <- P[idx.2[1]]
            }
            if (p %in% ii$parname) {
                idx.2 <- which(x$par==p)
                v[idx] <- A[idx.2[1]]
            }
        }
    }


    ## Ex-parameters
    e <- NULL
    {
        ## named <- sapply(x$exfix, function(y) is.character(y) & !is.na(y))
        fixed <- sapply(x$exfix, function(y) is.numeric(y) & !is.na(y))

        e <- rep(0,length(x$exfix))

        names(e) <- names(x$expar)
        if (!(is.null(epars) | ii$npar.ex==0))
            e[which(ii$e1==1)] <- epars
        if (any(fixed))
            e[fixed] <- unlist(x$exfix[fixed])
        for (p in ii$eparname) {
            idx <- which(x$exfix==p)
            if (!(p%in%c(ii$parname,ii$covparname,ii$mparname))) {
                parval[[p]] <- e[idx[1]]
            }
            attributes(parval[[p]])$e.idx <- idx

            if (length(idx)>1)
                e[idx[-1]] <- parval[[p]]
            if (p %in% setdiff(ii$covparname,c(ii$parname,ii$mparname))) {
                idx.2 <- which(x$covpar==p)
                e[idx] <- P[idx.2[1]]
            }
            if (p %in% setdiff(ii$parname,ii$mparname)) {
                idx.2 <- which(x$par==p)
                e[idx] <- A[idx.2[1]]
            }
            if (p %in% ii$mparname) {
                idx.2 <- which(x$mean==p)
                e[idx] <- v[idx.2[1]]
            }
        }
    }

    ## Constrained...
    constrain.idx <- NULL
    cname <- constrainpar <- c()
    if (length(constrain.par)>0 && is.numeric(c(pars,meanpar,e))) {
        constrain.idx <- list()
        for (p in constrain.par) {
            cname <- c(cname,p)
            reg.tidx <- reg.idx <- cov.idx <- m.idx <- e.idx <- NULL
            myc <- constrain(x)[[p]]
            xargs <- manifest(x)[na.omit(match(attributes(myc)$args,manifest(x)))]
            if (length(xargs)>0) {
                if (!is.null(data)) {
                    parval[xargs] <- (data)[xargs]
                } else parval[xargs] <- 0
            }
            val <- rbind(unlist(c(parval,constrainpar,x$mean,e)[attributes(myc)$args]))
            cpar <- myc(val);
            constrainpar <- c(constrainpar,list(cpar)); names(constrainpar) <- cname
            if (p%in%ii$parname.all) {
                reg.idx <- which(x$par==p)
                reg.tidx <- which(t(x$par==p))
                if (!is.null(val))
                    A[reg.idx] <- cpar##myc(val)
            }
            if (p%in%ii$covparname.all) {
                cov.idx <- which(x$covpar==p)
                if (!is.null(val))
                    P[cov.idx] <- cpar##myc(val)
            }
            if (p%in%ii$mparname.all) {
                m.idx <- which(x$mean==p)
                if (!is.null(val))
                    v[m.idx] <- cpar##myc(val)
            }
            if (p%in%ii$eparname.all) {
                e.idx <- which(x$exfix==p)
                if (!is.null(val))
                    e[e.idx] <- cpar##myc(val)
            }
            constrain.idx[[p]] <- list(reg.idx=reg.idx,reg.tidx=reg.tidx,cov.idx=cov.idx,m.idx=m.idx,e.idx=e.idx)
        }
    }

    if (x$index$sparse & !is.character(class(pars)[1])) {
        A <- as(A,"sparseMatrix")
        P <- as(P,"sparseMatrix")
        v <- as(v,"sparseMatrix")
    }
    return(list(A=A, P=P, v=v, e=e, parval=parval, constrain.idx=constrain.idx, constrainpar=constrainpar))
}

##' @export
procformula <- function(object=NULL,value,exo=lava.options()$exogenous,...) {

    ## Split into reponse and covariates by ~ disregarding expressions in parantheses
    ## '(?!...)' Negative lookahead assertion
    regex <- "~(?![^\\(].*\\))"
    yx <- lapply(strsplit(as.character(value),regex,perl=TRUE),function(x) gsub(" ","",x))[-1]
    yx <- lapply(yx,function(x) gsub("\n","",x))
    iscovar <- FALSE
    if (length(yx)==1) {
        lhs <- NULL; xidx <- 1
    } else {
            lhs <- yx[1]; xidx <- 2
            if (yx[[xidx]][1]=="") {
                yx[[xidx]] <- yx[[xidx]][-1]
                iscovar <- TRUE
            }
    }
    ##Check for link function
    invlink <- NULL
    if (xidx==2) {
        if (length(grep("[a-zA-Z0-9_]*\\(.*\\)$",yx[[xidx]]))>0) { ## rhs of the form F(x+y)
            invlink <- strsplit(yx[[xidx]],"\\(.*\\)")[[1]][1]
                if (invlink%in%c("f","v","I","") ||
                    grepl("\\+",invlink))
                { ## Reserved for setting linear constraints
                    invlink <- NULL
                } else {
                    yx[[xidx]] <- gsub(paste0(invlink,"\\(|\\)$"),"",yx[[xidx]])
                }
            }
    }

    ## Handling constraints with negative coefficients
    ## while not tampering with formulas like y~f(x,-2)
    st <- yx[[xidx]]
    st <- gsub("\\-","\\+\\-",gsub("\\+\\-","\\-",st)) ## Convert - to +- (to allow for splitting on '+')
    ##gsub("[^,]\\-","\\+\\-",st) ## Convert back any - not starting with ','
    st <- gsub(",\\+",",",st) ## Remove + inside 'f' and 'v' constraints
    st <- gsub("^\\+","",st) ## Remove leading plus
    yx[[xidx]] <- st
    
    ## Match '+' but not when preceeded by ( ... )
    X <- strsplit(yx[[xidx]],"\\+(?![^\\(]*\\))", perl=TRUE)[[1]]
    
    ##regex <- "(?!(\\(*))[\\(\\)]"
    regex <- "[\\(\\)]"
    ## Keep squares brackets and |(...) statements
    ## Extract variables from expressions like
    ## f(x,b) -> x,b  and  2*x -> 2,cx
    ## but avoid to tamper with transformation expressions:
    ## a~(x*b)
    res <- lapply(X,decomp.specials,regex,pattern2="\\*",pattern.ignore="~",reverse=TRUE,perl=TRUE)
    ##OLD:
    ##res <- lapply(X,decomp.specials,pattern2="[*]",reverse=TRUE)
    xx <- unlist(lapply(res, function(x) x[1]))

    xxf <- lapply(as.list(xx),function(x) decomp.specials(x,NULL,pattern2="\\[|~",perl=TRUE))
    xs <- unlist(lapply(xxf,function(x) x[1]))

    ## Alter intercepts?    
    intpos <- which(vapply(xs,function(x) grepl("^[\\+\\-]*[\\.|0-9]+$",x), 0)==1)
    ## Match '(int)'
    intpos0 <- which(vapply(X,function(x) grepl("^\\([\\+\\-]*[\\.|0-9]+\\)$",x),0)==1)

    yy <- ys <- NULL
    if (length(lhs)>0) {
        yy <- decomp.specials(lhs)
        yyf <- lapply(yy,function(y) decomp.specials(y,NULL,pattern2="[",fixed=TRUE,perl=FALSE))
        ys <- unlist(lapply(yyf,function(x) x[1]))
    }
    
    notexo <- c()
    if (!is.null(object)) {
      if (length(lhs)>0) {
        object <- addvar(object,ys,reindex=FALSE,...)
        notexo <- ys
        ## Add link-transformation
        if (!is.null(invlink)) {
            if (invlink=="") {
                object <- transform(object,ys,NULL,post=FALSE)
                covariance(object,ys) <- NA
            } else {
                ff <- function(x) {};  body(ff) <- parse(text=paste0(invlink,"(x)"))
                object <- transform(object,ys,ff,post=FALSE)
                covariance(object,ys) <- 0
            }
        }
      }

      if (length(intpos>0)) {
          xs[intpos[1]] <- gsub("\\+","",xs[intpos[1]])
          if (xs[intpos[1]]==1 && (!length(intpos0)>0) ) {
              xs[intpos[1]] <- NA
          }
          intercept(object,ys) <- char2num(xs[intpos[1]])
          xs <- xs[-intpos]
          res[intpos] <- NULL
      }


        object <- addvar(object,xs,reindex=FALSE ,...)
        exolist <- c()


        for (i in seq_len(length(xs))) {

            ## Extract transformation statements: var~(expr)
            xf0 <- strsplit(xx[[i]],"~")[[1]]
            if (length(xf0)>1) {
                myexpr <- xf0[2]
                ftr <- toformula(y="",x=paste0("-1+I(",myexpr,")"))
                xtr <- all.vars(ftr)
                xf0 <- xf0[1]
                transform(object, y=xf0, x=xtr) <- function(x) {
                    structure(model.matrix(ftr,as.data.frame(x)),dimnames=list(NULL,xf0))
                }
            }

            xf <- unlist(strsplit(xf0,"[\\[\\]]",perl=TRUE))
            if (length(xf)>1) {
                xpar <- strsplit(xf[2],":")[[1]]
                if (length(xpar)>1) {
                    val <- ifelse(xpar[2]=="NA",NA,xpar[2])
                    valn <- char2num(val)
                    covariance(object,xs[i]) <- ifelse(is.na(valn),val,valn)
                }
                val <- ifelse(xpar[1]=="NA",NA,xpar[1])
                valn <- char2num(val)
                if (is.na(val) || val!=".") {
                    intercept(object,xs[i]) <- ifelse(is.na(valn),val,valn)
                    notexo <- c(notexo,xs[i])
                }
            } else { exolist <- c(exolist,xs[i]) }
        }

        for (i in seq_len(length(ys))) {
            y <- ys[i]
            yf <- unlist(strsplit(yy[i],"[\\[\\]]",perl=TRUE))
            if (length(yf)>1) {
                ypar <- strsplit(yf[2],":")[[1]]
                if (length(ypar)>1) {
                    val <- ifelse(ypar[2]=="NA",NA,ypar[2])
                    valn <- char2num(val)
                    covariance(object,y) <- ifelse(is.na(valn),val,valn)
                }
                val <- ifelse(ypar[1]=="NA",NA,ypar[1])
                valn <- char2num(val)
                if (is.na(val) || val!=".") {
                    intercept(object,y) <- ifelse(is.na(valn),val,valn)
                }
            }
        }

        curvar <- index(object)$var
        if (exo) {
            oldexo <- exogenous(object)
            newexo <- setdiff(exolist,c(notexo,curvar,ys))
            exogenous(object) <- union(newexo,setdiff(oldexo,notexo))
        }
    }

    return(list(object=object,
                yx=yx,
                X=X,
                ys=ys,
                xx=xx,
                xs=xs,
                yy=yy,
                ys=ys,
                res=res,
                notexo=notexo,
                intpos=intpos,
                invlink=invlink,
                lhs=lhs,
                iscovar=iscovar))
}

simulatehook_multiple_inputs <- function(x,data,...) {
    minp <- x$attributes$multiple.inputs
    if (length(minp)>0) {
        for (i in seq_along(minp)) {            
            outcome <- names(minp[i])
            inp <- minp[[i]]$input
            fun <- minp[[i]]$fun
            data[,outcome] <- fun(x, data, inp)
        }
    }
    return(data)
}

addhook("simulatehook_multiple_inputs","sim.hooks")

printhook_multiple_inputs <- function(x,...) {
    minp <- x$attributes$multiple.inputs
    if (length(minp)>0) {
        outcomes <- names(minp)
        for (i in seq_along(minp)) {
            cat(minp[[i]]$type, ":\n\n")
            st <- paste0(outcomes[i]," ~ ", paste0(minp[[i]]$input,collapse=" | "))
            cat("  ", st, "\n")
            cat("\n")
        }
    }
    return(NULL)
}

addhook("printhook_multiple_inputs","print.hooks")

###{{{ missingModel
missingModel <- function(model,data,var=endogenous(model),fix=FALSE,type=2,keep=NULL,weights=NULL,data2=NULL,cluster=NULL,...) {
  if (!inherits(model,"lvm")) stop("Needs a lvm-object")
  if (type==3) {
    var <- manifest(model)
  }

  data.mis <- is.na(data[,var,drop=FALSE])
  colnames(data.mis) <- var
  patterns <- unique(data.mis,MARGIN=1)

  mis.type <- apply(data.mis,1,
                  function(x) which(apply(patterns,1,function(y) identical(x,y))))
  pattern.allmis <- which(apply(patterns,1,all)) ## Remove entry with all missing

  data2.list.provided <- is.list(data2) & !is.data.frame(data2)
  weights.list.provided <- is.list(weights) & !is.data.frame(weights)
  models <- datasets <- datasets2 <- weightsdata <- clusters <- c()
  mymodel <- baptize(model)
  pattern.compl <- 0
  count <- 0
  A <- index(model)$A
  topendo <- endogenous(model,top=TRUE)
  exo <- exogenous(model)
  exclude <- c()

  warned <- FALSE
  for (i in setdiff(seq_len(nrow(patterns)),pattern.allmis)) {
    exoremove <- c()
    count <- count+1
    mypattern <- patterns[i,]
    m0 <- mymodel;
    if (any(mypattern)) {
      latent(m0) <- colnames(data.mis)[mypattern]
      if (type>1) {
          mytop <- intersect(topendo,colnames(data.mis)[mypattern])
        if (!is.null(mytop)) {
          rmvar(m0) <- mytop
          for (xx in exo) {
          ## If exogenous variable only have effect on missing variables,
          ##  then remove it from the model
              if (all(c(rownames(A)[A[xx,]==1])%in%mytop) &&
                  !(xx%in%m0$par)
                  ##&& !(xx%in%names(index(m0))$parval)
                  ) {
                  exoremove <- c(exoremove,xx)
                  rmvar(m0) <- xx
              }
          }
        }
      }
    } else
        pattern.compl <- count
    ## d0 <- data[mis.type==i,manifest(m0),drop=FALSE];
    d0 <- data[which(mis.type==i),c(manifest(m0),keep),drop=FALSE];
    w0 <- NULL
    if (!is.null(weights))
    if (!weights.list.provided) {
        if (NCOL(weights)==1) {
            w0 <- as.matrix(weights) [which(mis.type==i)];
        } else {
            w0.var <- intersect(manifest(m0),colnames(weights))
            w0 <- weights[which(mis.type==i),w0.var,drop=FALSE];
        }
    }
    if (!data2.list.provided) {
      w02.var <- intersect(manifest(m0),colnames(data2))
      w02 <- data2[which(mis.type==i),w02.var,drop=FALSE];
    } 

    clust0 <- cluster[which(mis.type==i)]
    ex0 <- exogenous(m0) <- setdiff(exo,exoremove)
    xmis <- which(apply(d0[,ex0,drop=FALSE],1,function(x) any(is.na(x))))

    if (length(xmis)>0) {
      misx <- ex0[apply(d0[xmis,ex0,drop=FALSE],2,function(x) any(is.na(x)))]
      if (!warned)
          warning("Missing exogenous variables: ", paste(misx,collapse=","),
                  ". Removing rows...")
      warned <- TRUE
      d0 <- d0[-xmis,,drop=FALSE]
      if (NCOL(weights)==1) {
          w0 <- w0[-xmis]
      } else {
          w0 <- w0[-xmis,,drop=FALSE]
      }
      clust0 <- clust0[-xmis]
      w02 <- w02[-xmis,,drop=FALSE]
    }
    if (length(misx <- intersect(ex0,latent(m0)))>0) {
      warning("Missing exogenous variables:", paste(misx,collapse=","),
              "! Remove manually!.")
    }
##    else
    {
      if( sum(unlist(index(m0)[c("npar","npar.mean")]))>0 ) {
        models <- c(models, list(m0))
        datasets <- c(datasets, list(d0))
        if (!weights.list.provided)
          weightsdata <- c(weightsdata, list(w0))
        if (!data2.list.provided)
          datasets2 <- c(datasets2, list(w02))
        clusters <- c(clusters, list(clust0))
      } else {
        exclude <- c(exclude,count)
      }
    }
  }

  rmset <- c()
  for (i in seq_len(length(datasets))) {
    if (nrow(datasets[[i]])==0) rmset <- c(rmset,i)
  }
  if (length(rmset)>0) {
    models[[rmset]] <- NULL
    datasets[[rmset]] <- NULL
    weightsdata[[rmset]] <- NULL
    datasets2[[rmset]] <- NULL
    clusters[[rmset]] <- NULL
    patterns <- patterns[-rmset,,drop=FALSE]
  }
  if (data2.list.provided) datasets2 <- data2
  if (weights.list.provided) weightsdata <- weights

  Patterns <- patterns
  if (length(exclude)>0)
    Patterns <- Patterns[-exclude,]
  pattern.allcomp<- which(apply(Patterns,1,function(x) all(!x))) ## Complete cases

  res <- list(models=models, datasets=datasets,
              weights=weightsdata,
              data2=datasets2,
              clusters=clusters,
              patterns=Patterns,
              pattern.compl=pattern.compl,
              pattern.allmis=pattern.allmis,
              pattern.allcomp=pattern.allcomp,
              mis.type=mis.type)
  return(res)
}

###}}}

###{{{ estimate.MAR.lvm

##' @export
estimate.MAR <- function(x,data,which=endogenous(x),fix,type=2,startcc=FALSE,control=list(),messages=lava.options()$messages,weights,data2,cluster,onlymodel=FALSE,estimator="gaussian",hessian=TRUE,keep=NULL,...) {
  cl <- match.call()

  Debug("estimate.MAR")
  redvar <- intersect(intersect(parlabels(x),latent(x)),colnames(data))
  if (length(redvar)>0 & (messages>0))
    warning(paste("Remove latent variable colnames from dataset",redvar))

  xfix <- setdiff(colnames(data)[(colnames(data)%in%parlabels(x,exo=TRUE))],latent(x))
  if (missing(fix))
    fix <- ifelse(length(xfix)>0,FALSE,TRUE)

  S <- diag(nrow=length(manifest(x)));
  mu <- rep(0,nrow(S));
  K <- length(exogenous(x))
  vnames <- index(x)$manifest
  names(mu) <- rownames(S) <- colnames(S) <- vnames
  if (K>0) {
    xx <- subset(Model(x),exogenous(x))
    exogenous(xx) <- NULL
    covfix(xx, vars(xx)) <- NA
    xx <- covariance(xx,exogenous(x),exogenous(x))
    datax <- data[,exogenous(x),drop=FALSE]
    exo.idx <- match(exogenous(x),manifest(x))

    mu0 <- colMeans(datax,na.rm=TRUE)
    cov0 <- cov(datax,use="pairwise.complete.obs")*(nrow(datax)-1)/nrow(datax)
    cov0upper <- cov0[upper.tri(cov0,diag=TRUE)]
    exogenous(xx) <- NULL
    coefpos <- matrices(xx,seq_len(K*(K-1)/2+K))$P
    ii <- coefpos[upper.tri(coefpos,diag=TRUE)]
    start <- c(mu0, cov0upper[order(ii)])
    S[exo.idx,exo.idx] <- cov0
    mu[exo.idx] <- mu0
    ##    message("\n")
  }

  x0 <- x
  x <- fixsome(x, measurement.fix=fix, exo.fix=TRUE, S=S, mu=mu, n=1)
  if (messages>1)
    message("Identifying missing patterns...")
  val <- missingModel(x,data,var=which,type=type,keep=c(keep,xfix),weights=weights,data2=data2,cluster=cluster,...)
  if (messages>1)
    message("\n")
  if (nrow(val$patterns)==1) {
      res <- estimate(x, data=data, fix=fix, weights=weights, data2=data2,
                     estimator=estimator, messages=messages, control=control, ...)
    return(res)
  }

  if (startcc & is.null(control$start)) {
    if (messages>1)
      message("Obtaining starting value...")
    start0 <- rep(1,sum(unlist(index(x)[c("npar","npar.mean")])))
    mystart <- tryCatch(
                        (estimate(x,data=na.omit(data),messages=0,
                                     weights=weights,data2=data2,estimator=estimator,quick=TRUE,...
                                      )),
                        error=function(e) rep(1,sum(unlist(index(x)[c("npar","npar.mean")])))
                        )
    control$start <- mystart
    if (messages>1)
      message("\n")
  }

  if (is.null(control$meanstructure))
    control$meanstructure <- TRUE
  mg0 <- with(val, suppressWarnings(multigroup(models,datasets,fix=FALSE,exo.fix=FALSE,missing=FALSE)))
  if (!is.null(names(control$start))) {
    parorder1 <- attributes(parpos(mg0,p=names(control$start)))$name
    paridx <- match(parorder1,names(control$start))
    ## newpos <- paridx[which(!is.na(paridx))]
    start0 <- control$start
    start0[which(!is.na(paridx))] <- control$start[na.omit(paridx)]
    names(start0)[which(!is.na(paridx))] <- names(control$start[na.omit(paridx)])
    control$start <- start0
  }

  if (onlymodel) return(list(mg=mg0,val=val,weights=val$weights,data2=val$data2,cluster=val$clusters))

  if (all(unlist(lapply(val$weights,is.null)))) val$weights <- NULL
  if (all(unlist(lapply(val$data2,is.null)))) val$data2 <- NULL
  if (all(unlist(lapply(val$clusters,is.null)))) val$clusters <- NULL

  e.mis <- estimate(mg0,control=control,messages=messages,
                    weights=val$weights,data2=val$data2,
                    cluster=val$clusters,estimator=estimator,...)

  cc <- coef(e.mis,type=1)
  mynames <- c()
  if (e.mis$model$npar.mean>0)
    mynames <- c(mynames,paste0("m",seq_len(e.mis$model$npar.mean)))
   if (e.mis$model$npar>0)
     mynames <- c(mynames,paste0("p",seq_len(e.mis$model$npar)))
  rownames(cc) <- mynames


  mycc <- val$pattern.allcomp ## Position of complete-case model
  nmis <- with(val, as.numeric(table(mis.type)[pattern.allmis])) ## Number of completely missing observations
  if (length(nmis)>0 & length(mycc)>0) ## Any individuals with all missing?
    if (val$pattern.allmis<mycc)
      mycc <- mycc-1

  if (length(xfix)>0) {
    nrow <- length(vars(x))
    xpos <- lapply(xfix,function(y) which(regfix(x)$labels==y))
    colpos <- lapply(xpos, function(y) ceiling(y/nrow))
    rowpos <- lapply(xpos, function(y) (y-1)%%nrow+1)
    for (i in seq_along(xfix))
      regfix(x, from=vars(x)[rowpos[[i]]],to=vars(x)[colpos[[i]]]) <-
        rep(colMeans(data[,xfix[i],drop=FALSE],na.rm=TRUE),length(rowpos[[i]]))
    x <- updatelvm(x,zeroones=TRUE,deriv=TRUE)
  }

  ord <- c()
  ordlist <- list()
  for (i in seq_len(nrow(val$patterns))) {
    ordlist <- c(ordlist, list(which(val$mis.type==i)))
    ord <- c(ord, ordlist[[i]])
  }

  res <- with(val, list(coef=cc,
                        patterns=patterns, table=table(mis.type),
                        mis.type=mis.type,
                        order=ord,
                        orderlist=ordlist,
                        nmis=nmis,
                        allmis=pattern.allmis,
                        cc=mycc,
                        ncc=as.numeric(table(mis.type)[pattern.allcomp]),
                        multigroup=e.mis$model,
                        estimate=e.mis,
                        model=x,
                        model0=x0,
                        vcov=e.mis$vcov, opt=e.mis$opt,
                        control=control,
                        data=list(model.frame=data),
                        estimator=estimator,
                        call=cl
                        ))
  class(res) <- c("lvm.missing","lvmfit")
  if (inherits(e.mis,"lvmfit.randomslope"))
    class(res) <- c(class(res),"lvmfit.randomslope")

  if (hessian & is.null(cluster)) {
    if (messages>1)
      message("Calculating asymptotic variance...\n")
    res$vcov <- solve(information(res$estimate,type="hessian"))
    cc[] <- coef(e.mis,type=1,vcov=res$vcov)
    res$coef <- cc
  }

  return(res)
}

###}}} estimate.MAR.lvm

##' Variance based on influence function
##'
##' Computes the empirical variance of the influence function, providing a
##' sandwich-type variance estimator.
##' @param x Influence function matrix (observations x parameters), typically
##'   obtained from [IC].
##' @param ... Additional arguments (currently not used)
##' @return Covariance matrix (p x p).
#' @export
var_ic <- function(x, ...) {
  N <- crossprod(!is.na(x))
  V <- var(x, use="pairwise.complete.obs")*(N-1)/N^2
  V[N==0] <- 0
  return(V)
}

ordinal_remove_hook <- function(x,var,...) {
    ordinal(x,K=0) <- var
    return(x)
}

color_ordinal <- function(x,subset=vars(x),...) {
    return(list(vars=intersect(subset,ordinal(x)),col="indianred1"))
}

ordinal_sim_hook <- function(x,data,p,modelpar,...) {
    ovar <- ordinal(x)

    for (i in seq_len(length(ovar))) {
        if (attributes(ovar)$liability[i]) {
            idx <- attributes(ovar)$idx[[ovar[i]]]
            if (length(idx)==0) {
                breaks <- c(-Inf,0,Inf)
            } else {
                breaks <- c(-Inf,ordreg_threshold(modelpar$e[idx]),Inf)
            }
            z <- cut(data[,ovar[i]],breaks=breaks)
            data[,ovar[i]] <- as.numeric(z)-1
        }
        K <- attributes(ovar)$K[i]
        lab <- attributes(ovar)$labels[ovar[i]][[1]]
        if (!is.null(lab))
            data[,ovar[i]] <- factor(data[,ovar[i]],
                                     levels=seq(K)-1,
                                     labels=lab)
    }
    return(data)
}

ordinal_estimate_hook <- function(x,data,weights,data2,estimator,...) {
    dots <- list(...)

    nestimator <- c("normal")
    nestimator2 <- c("tobit","tobitw","gaussian")

    ord <- ordinal(x)
    bin <- NULL
    if (is.null(estimator) && length(ord)>0) estimator <- nestimator[1]

    ## Binary outcomes -> censored regression
    if (is.null(dim(data))) return(NULL)
    if (is.null(estimator) || estimator%in%c(nestimator2,nestimator)) {
        for (i in setdiff(lava::endogenous(x),bin)) {
            if (is.character(data[,i]) | is.factor(data[,i])) { # Transform binary 'factor'
                y <- as.factor(data[,i])
                data[,i] <- as.numeric(y)-1
                estimator <- nestimator[1]
                ordinal(x,K=nlevels(y)) <- i
            }
        }
        ord <- ordinal(x)
        if (length(ord)>0 && !is.null(estimator) && estimator%in%nestimator2) {
          estimator <- nestimator[1]
        }
    }

    ## Transform 'Surv' objects
    data2 <- mynames <- NULL
    if (is.null(estimator) || estimator%in%nestimator[1] ) {
        for (i in setdiff(lava::endogenous(x),c(bin,ord))) {
            if (survival::is.Surv(data[,i])) {
                S <- data[,i]
                y1 <- S[,1]
                y2 <- y1
                if (attributes(S)$type=="left")  {
                    y1[S[,2]==0] <- -Inf
                }
                if (attributes(S)$type=="right") {
                    y2[S[,2]==0] <- Inf
                }
                if (attributes(S)$type=="interval2") {
                    y2 <- S[,2]
                }
                if (attributes(S)$type=="interval") {
                  rcens <- which(S[,3]==0L)
                  if (length(rcens)>0) y2[rcens] <- Inf
                  lcens <- which(S[,3]==2L)
                  if (length(lcens)>0) y1[lcens] <- -Inf
                  icens <- which(S[,3]==3L)
                  y2[icens] <- S[icens, 2]
                }
                if (!(attributes(S)$type%in%c("left","right","interval2","interval"))) stop("Surv type not supported.")
                mynames <- c(mynames,i)
                y2 <- cbind(y2)
                colnames(y2) <- i
                data2 <- cbind(data2,y2)
                data[,i] <- y1
                estimator <- "normal"
            }
        }
    }

    return(c(list(x=x,data=data,weights=weights,data2=data2,estimator=estimator),dots))
}


##' Define variables as ordinal
##'
##' Define variables as ordinal in latent variable model object
##' @export
##' @aliases ordinal ordinal<-
##' @param x Object
##' @param ... additional arguments to lower level functions
##' @param value variable (formula or character vector)
##' @examples
##' if (requireNamespace("mets")) {
##' m <- lvm(y + z ~ x + 1*u[0], latent=~u)
##' ordinal(m, K=3) <- ~y+z
##' d <- sim(m, 100, seed=1)
##' e <- estimate(m, d)
##' }
##' 
"ordinal<-" <- function(x,...,value) UseMethod("ordinal<-")

##' @export
"ordinal<-.lvm" <- function(x,...,value) {
  ordinal(x, value, ...)
}

##' @export
"ordinal" <- function(x,...) UseMethod("ordinal")

##' @export
print.ordinal.lvm <- function(x,...) {
  cat(rep("_",28),"\n",sep="")
  for (i in x) {
    val <- attr(x,"fix")[[i]]
    if (length(val)==0)
      cat(paste(i,"binary",sep=":"),"\n")
    else print(unlist(attr(x,"fix")[[i]]),quote=FALSE)
    cat(rep("_",28),"\n",sep="")
  }
}

##' @export
`ordinal.lvm` <- function(x,var=NULL,K=2, constrain, breaks=NULL, p, liability=TRUE, labels, exo=FALSE, ...) {
    if (inherits(var,"formula")) {
        var <- all.vars(var)
    }
    if (is.null(var)) {
        ordidx <- unlist(x$attributes$ordinal)
        KK <- unlist(x$attributes$nordinal)
        idx <- x$attributes$ordinalparname
        fix <- lapply(idx,function(z) x$exfix[z])
        liability <- x$attributes$liability
        labels <- x$attributes$labels
        if (length(ordidx)>0) {
            val <- names(ordidx)
            return(structure(val,K=KK,idx=idx,fix=fix,liability=liability,labels=labels,class="ordinal.lvm"))
        }
        else
            return(NULL)
    }
    if (K[1]==0L || is.null(K[1]) || (is.logical(K) & !K[1])) {
        idx <- na.omit(match(var,names(x$attributes$ordinal)))
        if (length(idx)>0) {
            pp <- unlist(x$attributes$ordinalparname[idx])
            if (!is.null(pp)) parameter(x,remove=TRUE) <- pp
            if (!is.null(x$attributes$ordinalparname))
                x$attributes$ordinalparname <- x$attributes$ordinalparname[-idx]
            x$attributes$ordinal <- x$attributes$ordinal[-idx]
            ##x$attributes$labels[var] <- NULL
            x$attributes$type <- x$attributes$type[-idx]
            x$attributes$constrainY <- x$attributes$constrainY[setdiff(names(x$attributes$constrainY),var)]
            x$attributes$liability <- x$attributes$liability[-idx]
            x$attributes$nordinal <- x$attributes$nordinal[-idx]
            x$attributes$normal <- x$attributes$normal[-idx]
            exo <- intersect(var,exogenous(x,latent=TRUE))
            if (length(exo)>0) {
                intercept(x,var) <- NA
                covariance(x,var) <- NA
                exogenous(x) <- union(exogenous(x),exo)
            }
        }
        return(x)
    }
    
    if (!missing(p)) breaks <- qnorm(cumsum(p))
    if (!is.null(breaks)) {
        breaks <- ordreg_ithreshold(breaks)
        K <- length(breaks)+1
    }
    if (!missing(labels)) K <- length(labels)
    if (length(var)>length(K)) K <- rep(K[1],length(var))
    if (length(var)==1 && !missing(constrain)) constrain <- list(constrain)
    if (length(var)>1) {
        if (!missing(labels) && !is.list(labels)) labels <- rep(list(labels),length(var))
        if (!missing(breaks) && !is.list(breaks)) breaks <- rep(list(breaks),length(var))
        if (!missing(constrain) && !is.list(constrain)) constrain <- rep(list(constrain),length(var))
    }

    addvar(x) <- var
    for (i in seq_len(length(var))) {
        if (K[i]>2 || (K[i]==2 && !liability)) {
            parname <- paste0(var[i],":",paste(seq(K[i]-1)-1,seq(K[i]-1),sep="|"))
            newpar <- if (is.null(breaks)) {
                rep(-1,K[i]-1)
            } else if (is.list(breaks)) breaks[[i]] else breaks
            if (length(newpar)<K[i]-1) stop("Wrong number of starting values")
            newfix <- if (missing(constrain))
                rep(list(NA),length(newpar)) else constrain[[i]]
            if (length(newfix)<K[i]-1) stop("Wrong number of constraints")
            names(newpar) <- names(newfix) <- parname

            parameter(x,newfix,start=newpar) <- names(newfix)
            ## pex <- parname%in%names(x$expar)
            ## if (any(pex <- parname%in%names(x$expar))) {
            ##     if (!all(pex)) stop("Cannot change number of categories! Re-specify model.")
            ##     x$attributes$iordinal[var] <- list(idx)
            ## }
            x$attributes$ordinalparname[var[i]] <- list(names(newfix))
        }
        x$attributes$type[var[i]] <- ifelse(K[i]>2,"categorical","binary")
        if (K[i]>2) intfix(x,var[i],NULL) <- 0
        if (!liability) {
            mytr <- function(y,p,idx,...) {
                breaks <- c(-Inf,ordreg_threshold(p[idx]),Inf)
                as.numeric(cut(y,breaks=breaks))-1
            }
            myalist <- substitute(alist(y=,p=,idx=pp),
                                  list(pp=x$attributes$ordinalparname[[var[i]]]))
            formals(mytr) <- eval(myalist)
            transform(x,var[i],post=FALSE) <- mytr

        }
    }
    x$attributes$liability[var] <- liability
    x$attributes$ordinal[var] <- TRUE
    if (!missing(labels)) {
        if (length(var)==1) labels <- list(labels)
        x$attributes$labels[var] <- labels
    }
    x$attributes$nordinal[var] <- K
    x$attributes$normal[var] <- FALSE
    covfix(x,var,NULL,exo=exo) <- 1
    if (is.null(index(x))) index(x) <- reindex(x)
    return(x)
}

## ----include=FALSE------------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
mets <- lava:::versioncheck('mets', 1)
fullVignette <- Sys.getenv("_R_FULL_VIGNETTE_") %in% c("1","TRUE")

## ----load, results="hide",message=FALSE,warning=FALSE-------------------------
library('lava')

## ----sim----------------------------------------------------------------------
f <- function(x) cos(1.25*x) + x - 0.25*x^2
m <- lvm(x1+x2+x3 ~ eta1, y1+y2+y3 ~ eta2, latent=~eta1+eta2)
regression(m) <- eta1+eta2 ~ z
functional(m, eta2~eta1) <- f

d <- sim(m, n=200, seed=42) # Default is all parameters are 1

## -----------------------------------------------------------------------------
plot(m, plot.engine="visNetwork")

## ----specifymodels------------------------------------------------------------
m1 <- lvm(x1+x2+x3 ~ eta1, eta1 ~ z, latent=~eta1)
m2 <- lvm(y1+y2+y3 ~ eta2, eta2 ~ z, latent=~eta2)

## -----------------------------------------------------------------------------
nonlinear(m2, type="quadratic") <- eta2 ~ eta1

## ----twostage1----------------------------------------------------------------
e1 <- twostage(m1, m2, data=d)
e1

## ----linear_mle---------------------------------------------------------------
e0 <- estimate(regression(m1%++%m2, eta2~eta1), d)
estimate(e0,keep="^eta2~[a-z]",regex=TRUE) ## Extract coef. matching reg.ex.

## ----pred1--------------------------------------------------------------------
newd <- expand.grid(eta1=seq(-4, 4, by=0.1), z=0)
pred1 <- predict(e1, newdata=newd, x=TRUE)
head(pred1)

## ----spline_twostage----------------------------------------------------------
kn <- seq(-3,3,length.out=5)
nonlinear(m2, type="spline", knots=kn) <- eta2 ~ eta1
e2 <- twostage(m1, m2, data=d)
e2

## ----spline_ci----------------------------------------------------------------
p <- cbind(eta1=newd$eta1,
	  estimate(e2,f=function(p) predict(e2,p=p,newdata=newd))$coefmat)
head(p)

## ----fig:pred2----------------------------------------------------------------
plot(I(eta2-z) ~ eta1, data=d, col=Col("black",0.5), pch=16,
     xlab=expression(eta[1]), ylab=expression(eta[2]), xlim=c(-4,4))
lines(Estimate ~ eta1, data=as.data.frame(p), col="darkblue", lwd=5)
confband(p[,1], lower=p[,4], upper=p[,5], polygon=TRUE,
	 border=NA, col=Col("darkblue",0.2))

## ----spline_several-----------------------------------------------------------
m2a <- nonlinear(m2, type="linear", eta2~eta1)
m2b <- nonlinear(m2, type="quadratic", eta2~eta1)
kn1 <- seq(-3,3,length.out=5)
kn2 <- seq(-3,3,length.out=8)
m2c <- nonlinear(m2, type="spline", knots=kn1, eta2~eta1)
m2d <- nonlinear(m2, type="spline", knots=kn2, eta2~eta1)

## ----cv_fit, cache=TRUE, eval=fullVignette------------------------------------
#  ## Scale models in stage 2 to allow for a fair RMSE comparison
#  d0 <- d
#  for (i in endogenous(m2))
#      d0[,i] <- scale(d0[,i],center=TRUE,scale=TRUE)
#  ## Repeated 5-fold cross-validation:
#  ff <- lapply(list(linear=m2a,quadratic=m2b,spline4=m2c,spline6=m2d),
#  	    function(m) function(data,...) twostage(m1,m,data=data,stderr=FALSE,control=list(start=coef(e0),contrain=TRUE)))
#  fit.cv <- cv(ff,data=d,K=5,rep=2,mc.cores=parallel::detectCores(),seed=1)

## ----results="hide", echo=FALSE-----------------------------------------------
## To save time building the vignettes on CRAN, we cache time consuming computations
if (fullVignette) {
  fit.cv$fit <- NULL
  saveRDS(fit.cv, "nonlinear_fitcv.rds")
} else {
  fit.cv <- readRDS("nonlinear_fitcv.rds")
}

## -----------------------------------------------------------------------------
summary(fit.cv)

## ----multifit-----------------------------------------------------------------
fit <- lapply(list(m2a,m2b,m2c,m2d),
	     function(x) {
		 e <- twostage(m1,x,data=d)
		 pr <- cbind(eta1=newd$eta1,predict(e,newdata=newd$eta1,x=TRUE))
		 return(list(estimate=e,predict=as.data.frame(pr)))
	     })

plot(I(eta2-z) ~ eta1, data=d, col=Col("black",0.5), pch=16,
     xlab=expression(eta[1]), ylab=expression(eta[2]), xlim=c(-4,4))
col <- c("orange","darkred","darkgreen","darkblue")
lty <- c(3,4,1,5)
for (i in seq_along(fit)) {
    with(fit[[i]]$pr, lines(eta2 ~ eta1, col=col[i], lwd=4, lty=lty[i]))
}
legend("bottomright",
      c("linear","quadratic","spline(df=4)","spline(df=6)"),
      col=col, lty=lty, lwd=3)

## ----twostageCV, cache=TRUE, eval=fullVignette--------------------------------
#  selmod <- twostageCV(m1, m2, data=d, df=2:4, nmix=1:2,
#  	    nfolds=2, rep=1, mc.cores=parallel::detectCores())

## ----results="hide", echo=FALSE-----------------------------------------------
## To save time building the vignettes on CRAN, we cache time consuming computations
if (fullVignette) {
  saveRDS(summary(selmod), "nonlinear_selmod.rds")
} else {
  selmod <- readRDS("nonlinear_selmod.rds")
}

## -----------------------------------------------------------------------------
selmod

## -----------------------------------------------------------------------------
d$g <- (d$z<0)*1 ## Group variable
mm1 <- regression(m1, ~g)  # Add grouping variable as exogenous variable (effect specified via 'predict.fun')
mm2 <- regression(m2, eta2~ u1+u2+u1:g+u2:g+z)
pred <- function(mu,var,data,...) {
    cbind("u1"=mu[,1],"u2"=mu[,1]^2+var[1],
	  "u1:g"=mu[,1]*data[,"g"],"u2:g"=(mu[,1]^2+var[1])*data[,"g"])
}
ee1 <- twostage(mm1, model2=mm2, data=d, predict.fun=pred)
estimate(ee1,keep="eta2~u",regex=TRUE)

## -----------------------------------------------------------------------------
summary(estimate(ee1,keep="(:g)", regex=TRUE))

## -----------------------------------------------------------------------------
m1 <- baptize(m1)  ## Label all parameters
intercept(m1, ~x1+eta1) <- list(0,NA) ## Set intercept of x1 to zero. Remove the label of eta1
regression(m1,x1~eta1) <- 1 ## Factor loading fixed to 1

## ----mixture1, cache=TRUE, eval=fullVignette----------------------------------
#  set.seed(1)
#  em0 <- mixture(m1, k=2, data=d)

## ----estmixture, cache=TRUE,warnings=FALSE,messages=FALSE,eval=FALSE----------
#  em0 <- NULL
#  ll <- c()
#  for (i in 1:5) {
#      set.seed(i)
#      em <- mixture(m1, k=2, data=d, control=list(trace=0))
#      ll <- c(ll,logLik(em))
#      if (is.null(em0) || logLik(em0)<tail(ll,1))
#  	em0 <- em
#  }

## ----results="hide", echo=FALSE-----------------------------------------------
## To save time building the vignettes on CRAN, we cache time consuming computations
if (fullVignette) {
  saveRDS(em0, "nonlinear_em0.rds")
} else {
  em0 <- readRDS("nonlinear_em0.rds")
}

## -----------------------------------------------------------------------------
summary(em0)

## ----eval=mets----------------------------------------------------------------
e0 <- estimate(m1,data=d)
AIC(e0,em0)

## ----eval=mets----------------------------------------------------------------
em2 <- twostage(em0,m2,data=d)
em2

## ----mixturefit, eval=mets----------------------------------------------------
plot(I(eta2-z) ~ eta1, data=d, col=Col("black",0.5), pch=16,
     xlab=expression(eta[1]), ylab=expression(eta[2]))

lines(Estimate ~ eta1, data=as.data.frame(p), col="darkblue", lwd=5)
confband(p[,1], lower=p[,4], upper=p[,5], polygon=TRUE,
	 border=NA, col=Col("darkblue",0.2))

pm <- cbind(eta1=newd$eta1,
	    estimate(em2, f=function(p) predict(e2,p=p,newdata=newd))$coefmat)
lines(Estimate ~ eta1, data=as.data.frame(pm), col="darkred", lwd=5)
confband(pm[,1], lower=pm[,4], upper=pm[,5], polygon=TRUE,
	 border=NA, col=Col("darkred",0.2))
legend("bottomright", c("Gaussian","Mixture"),
       col=c("darkblue","darkred"), lwd=2, bty="n")

##' Model searching
##'
##' Performs Wald or score tests
##'
##'
##' @aliases modelsearch
##' @param x \code{lvmfit}-object
##' @param k Number of parameters to test simultaneously. For \code{equivalence}
##' the number of additional associations to be added instead of \code{rel}.
##' @param dir Direction to do model search. "forward" := add
##' associations/arrows to model/graph (score tests), "backward" := remove
##' associations/arrows from model/graph (wald test)
##' @param type If equal to 'correlation' only consider score tests for covariance parameters. If equal to 'regression' go through direct effects only  (default 'all' is to do both)
##' @param ... Additional arguments to be passed to the low level functions
##' @return Matrix of test-statistics and p-values
##' @author Klaus K. Holst
##' @seealso \code{\link{compare}}, \code{\link{equivalence}}
##' @keywords htest
##' @examples
##'
##' m <- lvm();
##' regression(m) <- c(y1,y2,y3) ~ eta; latent(m) <- ~eta
##' regression(m) <- eta ~ x
##' m0 <- m; regression(m0) <- y2 ~ x
##' dd <- sim(m0,100)[,manifest(m0)]
##' e <- estimate(m,dd);
##' modelsearch(e,messages=0)
##' modelsearch(e,messages=0,type="cor")
##' @export
modelsearch <- function(x, k = 1, dir = "forward", type = "all", ...) {
  if (dir == "forward") {
    res <- forwardsearch(x, k, type = type, ...)
    return(res)
  }
  if (dir == "backstep") {
    res <- backwardeliminate(x, ...)
    return(res)
  }
  res <- backwardsearch(x, k, ...)
  return(res)
}

backwardeliminate <- function(x,
                              keep = NULL,
                              pthres = 0.05,
                              AIC = FALSE,
                              messages = 0,
                              missing = FALSE,
                              intercepts = FALSE,
                              maxsteps = Inf,
                              information = "E",
                              data,
                              ...) {
  if (inherits(x, "lvm")) {
    M <- x
  } else {
    M <- Model(x)
  }
  if (missing(data)) data <- model.frame(x)

  dots <- list(...)
  if (is.null(dots$control$start)) {
    p0 <- estimate(M, data, quick = TRUE, messages = messages, missing = FALSE, ...)
    dots$control <- c(dots$control, list(start = p0, information = "E"))
  }

  ff <- function() {
    ii <- grep("m", names(coef(M)))
    vv <- variances(M, mean = TRUE)
    args <- c(list(x = M, data = data, missing = missing, quick = TRUE, messages = messages), dots)
    cc <- do.call("estimate", args)
    if (is.numeric(cc)) {
      I0 <- information(M, p = cc, data = data, type = information)[-c(ii, vv), -c(ii, vv)]
      cc0 <- cc[-c(ii, vv)]
      res <- (pnorm(abs(cc0 / sqrt(diag(solve(I0)))), lower.tail = FALSE)) * 2
      attributes(res)$coef <- cc
    } else {
      coefs <- coef(cc)
      res <- (pnorm(abs(coefs / sqrt(diag(vcov(cc)))), lower.tail = FALSE)) * 2
      res <- res[-c(ii, vv)]
      attributes(res)$coef <- coefs
    }
    return(res)
  }

  done <- FALSE
  i <- 0
  while (!done & i < maxsteps) {
    p <- ff()
    ordp <- order(p, decreasing = TRUE)
    curp <- p[ordp[1]]
    if (curp < pthres) break
    dots$control$start <- attributes(p)$coef[-ordp[1]]
    if (messages) message("Removed: ", names(curp), " p-value: ", round(curp, 3))
    ## var1 <- unlist(strsplit(names(curp),lava.options()$symbol[1]))
    nn <- strsplit(names(curp), paste0(lava.options()$symbol, collapse = "|"))[[1]]
    cancel(M) <- nn
  }

  if (messages) message("")
  return(M)
}

backwardsearch <- function(x, k = 1, ...) {
  if (!inherits(x, "lvmfit")) stop("Expected an object of class 'lvmfit'.")
  p <- pars(x)
  cur <- Model(x)
  pp <- modelPar(cur, p)
  p1 <- pp$p
  Tests <- c()
  Vars <- list()

  parnotvar <- setdiff(seq_along(p1), variances(Model(x))) ## We don't want to perform tests on the boundary of the parameter space
  freecomb <- utils::combn(parnotvar, k)

  for (i in seq_len(ncol(freecomb)))
  {
    cc0 <- coef(cur, mean = FALSE, messages = 0, symbol = lava.options()$symbol)
    ii <- freecomb[, i]
    p0 <- p1
    p0[ii] <- 0
    R <- diag(nrow = length(p0))
    R <- matrix(R[ii, ], nrow = length(ii))
    I <- information(Model(x), p = p1, n = x$data$n, data = model.frame(x))
    if (!is.null(pp$meanpar)) {
      rmidx <- seq_along(pp$meanpar)
      I <- I[-rmidx, -rmidx]
    }
    iI <- solve(I)
    W <- t(rbind(R) %*% p1) %*% solve(R %*% iI %*% t(R)) %*% (cbind(R) %*% p1)
    Tests <- c(Tests, W)
    Vars <- c(Vars, list(cc0[ii]))
  }
  ord <- order(Tests, decreasing = TRUE)
  Tests <- cbind(Tests, pchisq(Tests, k, lower.tail = FALSE))
  colnames(Tests) <- c("Test Statistic", "P-value")
  res <- list(test = Tests[ord, , drop = FALSE], var = Vars[ord])
  PM <- matrix(ncol = 3, nrow = 0)
  for (i in seq_len(nrow(Tests))) {
    if (!is.na(res$test[i, 1])) {
      newrow <- c(formatC(res$test[i, 1]), formatC(res$test[i, 2]), paste(res$var[[i]], collapse = ", "))
      PM <- rbind(PM, newrow)
    }
  }
  colnames(PM) <- c("Wald: W", "P(W>w)", "Index")
  rownames(PM) <- rep("", nrow(PM))

  res <- list(res = PM, test = res$test)
  class(res) <- "modelsearch"
  res
}

forwardsearch <- function(x, k = 1, messages = lava.options()$messages, type = "all", exclude.var = NULL, ...) {
  if (!inherits(x, "lvmfit")) stop("Expected an object of class 'lvmfit'.")

  p <- pars(x, reorder = TRUE)
  cur <- Model(x)
  Y <- endogenous(x)
  X <- exogenous(x)
  V <- vars(x)
  q <- length(Y)
  qx <- length(X)
  npar.sat <- q + q * (q - 1) / 2 + q * qx
  npar.cur <- index(cur)$npar
  nfree <- npar.sat - npar.cur
  if (nfree < k) {
    message("Cannot free ", k, " variables from model.\n")
    return()
  }

  directional <- !(tolower(type) %in% c("cor", "correlation", "cov", "covariance"))
  all <- tolower(type) %in% c("all", "both")

  Tests <- c()
  Vars <- list()
  AP <- with(index(cur), A + t(A) + P)
  restricted <- c()
  ## idx1 <- seq_len(ncol(AP)-1)
  ## idx2 <- seq(i+1,nrow(AP))

  idx <- seq_len(ncol(AP))
  if (!is.null(exclude.var)) {
    if (is.character(exclude.var)) {
      exclude.var <- match(exclude.var, V)
    }
    idx <- setdiff(idx, exclude.var)
  }
  for (i0 in seq_len(length(idx) - 1)) {
    for (j0 in seq(i0 + 1, length(idx))) {
      i <- idx[i0]
      j <- idx[j0]
      if (AP[j, i] == 0) {
        restricted <- rbind(restricted, c(i, j))
      }
    }
  }

  if (is.null(restricted)) {
    return(NULL)
  }
  if (all) {
    ntest <- nrow(restricted)
    directional <- rep(FALSE, ntest)
    restricted <- rbind(restricted, restricted, restricted[, 2:1])
    directional <- c(directional, rep(TRUE, 2 * ntest))
  } else {
    if (directional) {
      restricted <- rbind(restricted, restricted[, 2:1])
    }
    directional <- rep(directional, nrow(restricted))
  }

  restrictedcomb <- utils::combn(seq_len(nrow(restricted)), k) # Combinations of k-additions to the model

  if (!inherits(model.frame(x), c("data.frame", "matrix"))) {
    n <- model.frame(x)$n
    S <- model.frame(x)$S
    mu <- model.frame(x)$mu
  } else {
    n <- nrow(model.frame(x))
    S <- (n - 1) / n * var(model.frame(x), na.rm = TRUE)
    mu <- colMeans(model.frame(x), na.rm = TRUE)
  }
  pb <- progressr::progressor(steps = ncol(restrictedcomb))
  if (messages > 0) {
    message("Calculating score test for ", ncol(restrictedcomb), " models:")
  }
  for (i in seq_len(ncol(restrictedcomb))) {
    if (messages > 0) pb()
    varlist <- c()
    altmodel <- cur ## HA: altmodel, H0: cur
    for (j in seq_len(k)) {
      myvar <- restricted[restrictedcomb[j, i], ]
      if (any(wx <- V[myvar] %in% X)) {
        altmodel <- regression(altmodel, V[myvar][which(!wx)], V[myvar][which(wx)])
      } else {
        if (directional[i]) {
          covariance(altmodel, pairwise = TRUE) <- V[myvar]
        }
        covariance(altmodel, pairwise = TRUE) <- V[myvar]
      }
      varlist <- rbind(varlist, V[myvar])
    }
    altmodel$parpos <- NULL
    altmodel <- updatelvm(altmodel, deriv = TRUE, zeroones = TRUE, mean = TRUE)
    cc <- coef(altmodel, mean = TRUE, messages = 0, symbol = lava.options()$symbol)
    cc0 <- coef(cur, mean = TRUE, messages = 0, symbol = lava.options()$symbol)
    p1 <- numeric(length(p) + k)
    ## Need to be sure we place 0 at the correct position
    for (ic in seq_along(cc)) {
      idx <- match(cc[ic], cc0)
      if (!is.na(idx)) {
        p1[ic] <- p[idx]
      }
    }
    if (x$estimator == "gaussian" && !inherits(x, "lvm.missing")) {
      Sc2 <- score(altmodel,
        p = p1, data = NULL,
        model = x$estimator, weights = Weights(x), S = S, mu = mu, n = n
      )
    } else {
      Sc2 <- score(altmodel,
        p = p1, data = model.frame(x),
        model = x$estimator, weights = Weights(x)
      )
    }
    I <- information(altmodel, p = p1, n = n, data = model.frame(x), weights = Weights(x), estimator = x$estimator) ## [-rmidx,-rmidx]

    iI <- try(Inverse(I), silent = TRUE)
    Q <- ifelse(inherits(iI, "try-error"), NA, ## Score test
      (Sc2) %*% iI %*% t(Sc2)
    )
    Tests <- c(Tests, Q)
    Vars <- c(Vars, list(varlist))
  }

  ##if (messages > 0) close(pb)
  ord <- order(Tests)
  Tests <- cbind(Tests, pchisq(Tests, k, lower.tail = FALSE))
  colnames(Tests) <- c("Test Statistic", "P-value")
  PM <- c()
  for (i in seq_len(nrow(Tests))) {
    if (!is.na(Tests[i, 1])) {
      vv <- apply(Vars[[i]], 1, function(x) paste(x, collapse = lava.options()$symbol[2 - directional[i]]))
      newrow <- c(
        formatC(Tests[i, 1]), formatC(Tests[i, 2]),
        paste(vv, collapse = ",")
      )
      PM <- rbind(PM, newrow)
    }
  }
  if (is.null(PM)) {
    message("Saturated model")
    return(invisible(NULL))
  }
  Tests <- Tests[ord, , drop = FALSE]
  Vars <- Vars[ord]
  PM <- PM[ord, , drop = FALSE]

  colnames(PM) <- c("Score: S", "P(S>s)", "Index")
  rownames(PM) <- rep("", nrow(PM))
  res <- list(res = PM, test = Tests, var = Vars, directional = directional)
  class(res) <- "modelsearch"
  return(res)
}

##' @export
print.modelsearch <- function(x, tail = nrow(x$res), adj = c("holm", "BH"), ...) {
  N <- nrow(x$res)
  if (!is.null(adj)) {
    ##    adjp <- rev(holm(as.numeric(x$test[,2])))
    adjp <- rbind(sapply(adj, function(i) p.adjust(x$test[, 2], method = i)))
    colnames(adjp) <- adj
    x$res <- cbind(x$res, rbind(formatC(adjp)))
  }
  print(x$res[seq(N - tail + 1, N), ], quote = FALSE, ...)
  invisible(x)
}

###{{{ print.lvm

##' @export
`print.lvm` <-
function(x, ..., print.transform=TRUE,print.exogenous=TRUE) {
  res <- NULL
  myhooks <- gethook("print.hooks")
  for (f in myhooks) {
      res <- do.call(f, list(x=x, ...))
  }
  if (is.null(res)) {
    k <- length(vars(x))
    L <- rep(FALSE, k); names(L) <- vars(x); L[latent(x)] <- TRUE
    cat("Latent Variable Model\n") ##;" \n\twith: ", k, " variables.\n", sep="");
    if (k==0) {
        cat("\nEmpty\n")
        return()
    }
    ff <- formula(x, char=TRUE, all=TRUE)
    R <- Rx <- Rt <- c()
    exo <- exogenous(x)
    for (f in ff) {
      oneline <- as.character(f);
      y <- strsplit(f, "~")[[1]][1]
      y <- trim(y)
      {
        col1 <- as.character(oneline)
        D <- attributes(distribution(x)[[y]])$family
        Tr <- x$attributes$transform[[y]]
        col2 <- tryCatch(x$attributes$type[[y]], error=function(...) NULL)
        if (is.null(col2) || is.na(col2)) {
            if (!is.na(x$covfix[y, y]) && x$covfix[y, y]==0L) {
                col2 <- "deterministic"
            } else {
                col2 <- "gaussian"
            }
        }
        if (!is.null(Tr)){
            col1 <- paste0(y, " ~ ", paste0(Tr$x, collapse="+"), sep="")
            Rt <- rbind(Rt, c(col1, ""))
        }
        if (!is.null(D$family)) {
            col2 <- paste0(D$family)
        }
        if (!is.null(D$link)) col2 <- paste0(col2, "(", D$link, ")")
        if (!is.null(D$par)) col2 <- paste0(col2, "(", paste(D$par, collapse=","), ")")
        if (is.list(distribution(x)[[y]]) && is.vector(distribution(x)[[y]][[1]])) col2 <- "fixed"
        if (L[y]) col2 <- paste0(col2, ", latent")
        if (y%in%exo) {
            Rx <- rbind(Rx,c(col1, col2))
        } else {
            if (is.null(Tr)) {
                R <- rbind(R,c(col1, col2))
            }
        }
      }
    }
    if (length(R)>0) {
        rownames(R) <- paste(" ", R[, 1], " "); colnames(R) <- rep("", ncol(R))
        print(R[, 2, drop=FALSE], quote=FALSE, ...)
    }
    if (print.exogenous && length(Rx)>0) {
        cat("\nExogenous variables:")
        rownames(Rx) <- paste(" ", gsub("~ 1", "", Rx[, 1]), "     "); colnames(Rx) <- rep("", ncol(Rx))
        print(Rx[, 2, drop=FALSE], quote=FALSE, ...)
    }
    if (print.transform && length(Rt)>0) {
        cat("\nTransformations:")
        rownames(Rt) <- paste(" ", gsub("~ 1", "", Rt[,1]), "     "); colnames(Rt) <- rep("", ncol(Rt))
        print(Rt[,2,drop=FALSE],quote=FALSE, ...)
    }
  }
  cat("\n")
  invisible(x)
}

###}}} print.lvm

###{{{ print.lvmfit

##' @export
`print.lvmfit` <-
function(x,type=2,labels=FALSE,...) {
    print(CoefMat(x,labels=labels,type=type,...),quote=FALSE,right=TRUE)
    minSV <- attr(vcov(x),"minSV")
    if (!is.null(minSV) && minSV<1e-12) {
        warning("Small singular value: ", format(minSV))
    }
    pseudo <- attr(vcov(x),"pseudo")
    if (!is.null(pseudo) && pseudo) warning("Singular covariance matrix. Pseudo-inverse used.")
    invisible(x)
}

###}}} print.lvmfit

###{{{ print.lvmfit.randomslope

##' @export
print.lvmfit.randomslope <- function(x,labels=FALSE,type=2,...) {
  print(CoefMat(x,labels=labels,type=type,...),quote=FALSE,right=TRUE)
  invisible(x)
}

###}}}

###{{{ print.multigroupfit

##' @export
print.multigroupfit <- function(x,groups=NULL,...)  {
  if (is.null(groups)) {
    if (x$model$missing) {
      modelclass <- attributes(x$model0)$modelclass
      nmis <- attributes(x$model0)$nmis
      orggroup <- unique(modelclass)
      groupn <- unlist(lapply(orggroup,function(i) sum(modelclass==i)))
      cumsumgroup <- cumsum(c(0,groupn))
      groups <- unlist(lapply(orggroup,function(i)
                              which.min(nmis[which(modelclass==i)])+cumsumgroup[i])) ##  groups with max. number of variables
      for (i in seq_len(length(groups))) {
        if (nmis[groups[i]]>0) warning("No complete cases in group ",i,". Showing results of group with max number of variables. All coefficients can be extracted with 'coef'. All missing pattern groups belonging to this sub-model can be extracted by calling: coef(..., groups=c(",paste(which(modelclass==i),collapse=","),"))")
      }
      if (!is.null(x$model$mnameses))
        x$model$names <- x$model$mnames
    } else {
      groups <- seq_len(length(x$model$lvm))
    }
  }
  res <- coef(x,type=2,groups=groups,...)
  counter <- 0
  dots <- list(...)
  dots$groups <- groups
  type <- if (is.null(dots$type)) {
    dots$type <- 2
##    dots$type <- ifelse("lvmfit.randomslope"%in%class(x),2,9)
  }
  myargs <- c(list(x=x), dots)
  myargs$groups <- groups
  CC <- do.call("CoefMat.multigroupfit",myargs)
  for (cc in res) {
    counter <- counter+1
    cat(rep("_",52),"\n",sep="")
    cat("Group ", counter, sep="")
    myname <- x$model$names[counter]
    if (!is.null(myname) && !is.na(myname))
      cat(": ",myname,sep="")
    if (!x$model$missing) cat(" (n=",nrow(Model(x)$data[[groups[counter]]]), ")", sep="")
    cat("\n")
    print(CC[[counter]],quote=FALSE,right=TRUE)
  }
  cat("\n")
  invisible(x)
}

###}}} print.multigroupfit

###{{{ print.multigroup

##' @export
print.multigroup <- function(x,...) {
  cat("\n")
  cat("Number of groups:", x$ngroup, "\n")
  cat("Number of free parameters (not counting mean-parameters):", x$npar,"\n")
##  cat("Parameter-vector:", unlist(x$parlist), "\n\n")
  cat("Number of free mean parameters:", length(grep("m",x$mean)),"\n")
##  cat("Mean-vector:", x$mean, "\n\n")
  invisible(x)
}

###}}} print.multigroup

###{{{ printmany

printmany <- function(A,B,nspace=1,name1=NULL,name2=NULL,digits=3,rownames=NULL,emptystr=" ",bothrows=!is.table(A),right=TRUE,print=TRUE,...) {
  A <- format(A, digits=digits, right=right, ...)
  B <- format(B, digits=digits, right=right, ...)
  nA <- nrow(A); nB <- nrow(B)
  if (nrow(A)<nrow(B)) {
    A <- rbind(A, matrix("", nrow=nB-nA, ncol=ncol(A)))
  }
  if (nrow(B)<nrow(A)) {
    B <- rbind(B, rep("", nrow=nA-nB, ncol=ncol(B)))
  }
  if (!is.null(rownames) & length(rownames)==nrow(A))
    rownames(A) <- rownames(B) <- rownames
  res <- cbind(A, matrix("", nrow=nrow(A), ncol=nspace));
  dnn <- dimnames(A)
  dnn[[2]] <- c(dnn[[2]],rep(emptystr,nspace))
  dimnames(res) <- dnn
  ##dimnames(res)[[2]] <- c(dimnames(res)[[2]],rep(emptystr,nspace))
  if (!is.null(name1)) {
      oldname <- colnames(res)
      res <- cbind(rep("",nrow(res)), rownames(res), res);
      cres <- name1
      if (!is.null(rownames(res))) cres <- c(cres,"")
      colnames(res) <- c(cres,oldname)
      rownames(res) <- rep("",nrow(res))
  }
  if (!is.null(name2)) {
    oldname <- colnames(res)
    res <- cbind(res,rep("",nrow(res))); colnames(res) <- c(oldname,name2)
  }
  if (!identical(rownames(A),rownames(B)) & bothrows)
      res <- cbind(res, rownames(B))
  res <- cbind(res, B)
  if (is.null(name2)) {
      dnn[[2]] <- c(dnn[[2]],dimnames(B)[[2]])
      dimnames(res) <- dnn
  }
  if (print) print(res, quote=FALSE, right=right, ...)
  invisible(res)
}

###}}} printmany

missingModel <- function(model,data,var=endogenous(model),fix=FALSE,type=2,keep=NULL,weights=NULL,data2=NULL,cluster=NULL,...) {
  if (!inherits(model,"lvm")) stop("Needs a lvm-object")
  if (type==3) {
    var <- manifest(model)
  }

  data.mis <- is.na(data[,var,drop=FALSE])
  colnames(data.mis) <- var
  patterns <- unique(data.mis,MARGIN=1)

  mis.type <- apply(data.mis,1,
                  function(x) which(apply(patterns,1,function(y) identical(x,y))))
  pattern.allmis <- which(apply(patterns,1,all)) ## Remove entry with all missing

  data2.list.provided <- is.list(data2) & !is.data.frame(data2)
  weights.list.provided <- is.list(weights) & !is.data.frame(weights)
  models <- datasets <- datasets2 <- weightsdata <- clusters <- c()
  mymodel <- baptize(model)
  pattern.compl <- 0
  count <- 0
  A <- index(model)$A
  topendo <- endogenous(model,top=TRUE)
  exo <- exogenous(model)
  exclude <- c()

  warned <- FALSE
  for (i in setdiff(seq_len(nrow(patterns)),pattern.allmis)) {
    exoremove <- c()
    count <- count+1
    mypattern <- patterns[i,]
    m0 <- mymodel;
    if (any(mypattern)) {
      latent(m0) <- colnames(data.mis)[mypattern]
      if (type>1) {
          mytop <- intersect(topendo,colnames(data.mis)[mypattern])
        if (!is.null(mytop)) {
          rmvar(m0) <- mytop
          for (xx in exo) {
          ## If exogenous variable only have effect on missing variables,
          ##  then remove it from the model
              if (all(c(rownames(A)[A[xx,]==1])%in%mytop) &&
                  !(xx%in%m0$par)
                  ##&& !(xx%in%names(index(m0))$parval)
                  ) {
                  exoremove <- c(exoremove,xx)
                  rmvar(m0) <- xx
              }
          }
        }
      }
    } else
        pattern.compl <- count
    ## d0 <- data[mis.type==i,manifest(m0),drop=FALSE];
    d0 <- data[which(mis.type==i),c(manifest(m0),keep),drop=FALSE];
    w0 <- NULL
    if (!is.null(weights))
    if (!weights.list.provided) {
        if (NCOL(weights)==1) {
            w0 <- as.matrix(weights) [which(mis.type==i)];
        } else {
            w0.var <- intersect(manifest(m0),colnames(weights))
            w0 <- weights[which(mis.type==i),w0.var,drop=FALSE];
        }
    }
    if (!data2.list.provided) {
      w02.var <- intersect(manifest(m0),colnames(data2))
      w02 <- data2[which(mis.type==i),w02.var,drop=FALSE];
    } 

    clust0 <- cluster[which(mis.type==i)]
    ex0 <- exogenous(m0) <- setdiff(exo,exoremove)
    xmis <- which(apply(d0[,ex0,drop=FALSE],1,function(x) any(is.na(x))))

    if (length(xmis)>0) {
      misx <- ex0[apply(d0[xmis,ex0,drop=FALSE],2,function(x) any(is.na(x)))]
      if (!warned)
          warning("Missing exogenous variables: ", paste(misx,collapse=","),
                  ". Removing rows...")
      warned <- TRUE
      d0 <- d0[-xmis,,drop=FALSE]
      if (NCOL(weights)==1) {
          w0 <- w0[-xmis]
      } else {
          w0 <- w0[-xmis,,drop=FALSE]
      }
      clust0 <- clust0[-xmis]
      w02 <- w02[-xmis,,drop=FALSE]
    }
    if (length(misx <- intersect(ex0,latent(m0)))>0) {
      warning("Missing exogenous variables:", paste(misx,collapse=","),
              "! Remove manually!.")
    }
##    else
    {
      if( sum(unlist(index(m0)[c("npar","npar.mean")]))>0 ) {
        models <- c(models, list(m0))
        datasets <- c(datasets, list(d0))
        if (!weights.list.provided)
          weightsdata <- c(weightsdata, list(w0))
        if (!data2.list.provided)
          datasets2 <- c(datasets2, list(w02))
        clusters <- c(clusters, list(clust0))
      } else {
        exclude <- c(exclude,count)
      }
    }
  }

  rmset <- c()
  for (i in seq_len(length(datasets))) {
    if (nrow(datasets[[i]])==0) rmset <- c(rmset,i)
  }
  if (length(rmset)>0) {
    models[[rmset]] <- NULL
    datasets[[rmset]] <- NULL
    weightsdata[[rmset]] <- NULL
    datasets2[[rmset]] <- NULL
    clusters[[rmset]] <- NULL
    patterns <- patterns[-rmset,,drop=FALSE]
  }
  if (data2.list.provided) datasets2 <- data2
  if (weights.list.provided) weightsdata <- weights

  Patterns <- patterns
  if (length(exclude)>0)
    Patterns <- Patterns[-exclude,]
  pattern.allcomp<- which(apply(Patterns,1,function(x) all(!x))) ## Complete cases

  res <- list(models=models, datasets=datasets,
              weights=weightsdata,
              data2=datasets2,
              clusters=clusters,
              patterns=Patterns,
              pattern.compl=pattern.compl,
              pattern.allmis=pattern.allmis,
              pattern.allcomp=pattern.allcomp,
              mis.type=mis.type)
  return(res)
}

##' @export
estimate.MAR <- function(x,data,which=endogenous(x),fix,type=2,startcc=FALSE,control=list(),messages=lava.options()$messages,weights,data2,cluster,onlymodel=FALSE,estimator="gaussian",hessian=TRUE,keep=NULL,...) {
  cl <- match.call()

  Debug("estimate.MAR")
  redvar <- intersect(intersect(parlabels(x),latent(x)),colnames(data))
  if (length(redvar)>0 & (messages>0))
    warning(paste("Remove latent variable colnames from dataset",redvar))

  xfix <- setdiff(colnames(data)[(colnames(data)%in%parlabels(x,exo=TRUE))],latent(x))
  if (missing(fix))
    fix <- ifelse(length(xfix)>0,FALSE,TRUE)

  S <- diag(nrow=length(manifest(x)));
  mu <- rep(0,nrow(S));
  K <- length(exogenous(x))
  vnames <- index(x)$manifest
  names(mu) <- rownames(S) <- colnames(S) <- vnames
  if (K>0) {
    xx <- subset(Model(x),exogenous(x))
    exogenous(xx) <- NULL
    covfix(xx, vars(xx)) <- NA
    xx <- covariance(xx,exogenous(x),exogenous(x))
    datax <- data[,exogenous(x),drop=FALSE]
    exo.idx <- match(exogenous(x),manifest(x))

    mu0 <- colMeans(datax,na.rm=TRUE)
    cov0 <- cov(datax,use="pairwise.complete.obs")*(nrow(datax)-1)/nrow(datax)
    cov0upper <- cov0[upper.tri(cov0,diag=TRUE)]
    exogenous(xx) <- NULL
    coefpos <- matrices(xx,seq_len(K*(K-1)/2+K))$P
    ii <- coefpos[upper.tri(coefpos,diag=TRUE)]
    start <- c(mu0, cov0upper[order(ii)])
    S[exo.idx,exo.idx] <- cov0
    mu[exo.idx] <- mu0
    ##    message("\n")
  }

  x0 <- x
  x <- fixsome(x, measurement.fix=fix, exo.fix=TRUE, S=S, mu=mu, n=1)
  if (messages>1)
    message("Identifying missing patterns...")
  val <- missingModel(x,data,var=which,type=type,keep=c(keep,xfix),weights=weights,data2=data2,cluster=cluster,...)
  if (messages>1)
    message("\n")
  if (nrow(val$patterns)==1) {
      res <- estimate(x, data=data, fix=fix, weights=weights, data2=data2,
                     estimator=estimator, messages=messages, control=control, ...)
    return(res)
  }

  if (startcc & is.null(control$start)) {
    if (messages>1)
      message("Obtaining starting value...")
    start0 <- rep(1,sum(unlist(index(x)[c("npar","npar.mean")])))
    mystart <- tryCatch(
                        (estimate(x,data=na.omit(data),messages=0,
                                     weights=weights,data2=data2,estimator=estimator,quick=TRUE,...
                                      )),
                        error=function(e) rep(1,sum(unlist(index(x)[c("npar","npar.mean")])))
                        )
    control$start <- mystart
    if (messages>1)
      message("\n")
  }

  if (is.null(control$meanstructure))
    control$meanstructure <- TRUE
  mg0 <- with(val, suppressWarnings(multigroup(models,datasets,fix=FALSE,exo.fix=FALSE,missing=FALSE)))
  if (!is.null(names(control$start))) {
    parorder1 <- attributes(parpos(mg0,p=names(control$start)))$name
    paridx <- match(parorder1,names(control$start))
    ## newpos <- paridx[which(!is.na(paridx))]
    start0 <- control$start
    start0[which(!is.na(paridx))] <- control$start[na.omit(paridx)]
    names(start0)[which(!is.na(paridx))] <- names(control$start[na.omit(paridx)])
    control$start <- start0
  }

  if (onlymodel) return(list(mg=mg0,val=val,weights=val$weights,data2=val$data2,cluster=val$clusters))

  if (all(unlist(lapply(val$weights,is.null)))) val$weights <- NULL
  if (all(unlist(lapply(val$data2,is.null)))) val$data2 <- NULL
  if (all(unlist(lapply(val$clusters,is.null)))) val$clusters <- NULL

  e.mis <- estimate(mg0,control=control,messages=messages,
                    weights=val$weights,data2=val$data2,
                    cluster=val$clusters,estimator=estimator,...)

  cc <- coef(e.mis,type=1)
  mynames <- c()
  if (e.mis$model$npar.mean>0)
    mynames <- c(mynames,paste0("m",seq_len(e.mis$model$npar.mean)))
   if (e.mis$model$npar>0)
     mynames <- c(mynames,paste0("p",seq_len(e.mis$model$npar)))
  rownames(cc) <- mynames


  mycc <- val$pattern.allcomp ## Position of complete-case model
  nmis <- with(val, as.numeric(table(mis.type)[pattern.allmis])) ## Number of completely missing observations
  if (length(nmis)>0 & length(mycc)>0) ## Any individuals with all missing?
    if (val$pattern.allmis<mycc)
      mycc <- mycc-1

  if (length(xfix)>0) {
    nrow <- length(vars(x))
    xpos <- lapply(xfix,function(y) which(regfix(x)$labels==y))
    colpos <- lapply(xpos, function(y) ceiling(y/nrow))
    rowpos <- lapply(xpos, function(y) (y-1)%%nrow+1)
    for (i in seq_along(xfix))
      regfix(x, from=vars(x)[rowpos[[i]]],to=vars(x)[colpos[[i]]]) <-
        rep(colMeans(data[,xfix[i],drop=FALSE],na.rm=TRUE),length(rowpos[[i]]))
    x <- updatelvm(x,zeroones=TRUE,deriv=TRUE)
  }

  ord <- c()
  ordlist <- list()
  for (i in seq_len(nrow(val$patterns))) {
    ordlist <- c(ordlist, list(which(val$mis.type==i)))
    ord <- c(ord, ordlist[[i]])
  }

  res <- with(val, list(coef=cc,
                        patterns=patterns, table=table(mis.type),
                        mis.type=mis.type,
                        order=ord,
                        orderlist=ordlist,
                        nmis=nmis,
                        allmis=pattern.allmis,
                        cc=mycc,
                        ncc=as.numeric(table(mis.type)[pattern.allcomp]),
                        multigroup=e.mis$model,
                        estimate=e.mis,
                        model=x,
                        model0=x0,
                        vcov=e.mis$vcov, opt=e.mis$opt,
                        control=control,
                        data=list(model.frame=data),
                        estimator=estimator,
                        call=cl
                        ))
  class(res) <- c("lvm.missing","lvmfit")
  if (inherits(e.mis,"lvmfit.randomslope"))
    class(res) <- c(class(res),"lvmfit.randomslope")

  if (hessian & is.null(cluster)) {
    if (messages>1)
      message("Calculating asymptotic variance...\n")
    res$vcov <- solve(information(res$estimate,type="hessian"))
    cc[] <- coef(e.mis,type=1,vcov=res$vcov)
    res$coef <- cc
  }

  return(res)
}

##' @export
predict.lvmfit <- function(object,x=NULL,y=NULL,data=model.frame(object),p=coef(object),...) {
  predict(Model(object),x=x,y=y,p=p,data=data,...)
}


##' @export
predict.lvm.missing <- function(object,x=NULL,y=NULL,data=model.frame(object),p=coef(object),...) {
    idx <- match(coef(Model(object)),names(coef(object)))
    xx <- exogenous(object)
    p <- p[idx]
    if (!is.null(x)) {
        if (inherits(x,"formula")) {
            xy <- getoutcome(x)
            if (length(xy)>0) {
                if (is.null(y)) y <- decomp.specials(xy)
            }
            x <- attributes(xy)$x
      }
      x <- intersect(x,endogenous(object))
      if (is.null(y))
          y <- setdiff(vars(object),c(x,xx))
    } 
    obs0 <- !is.na(data[,x,drop=FALSE])
    data[,xx][which(is.na(data[,xx]),arr.ind=TRUE)] <- 0
    pp <- predict.lvmfit(object,x=x,y=y,data=data,p=p,...)
    if (all(obs0)) return(pp)

    if (!requireNamespace("mets",quietly=TRUE)) stop("Requires 'mets'")
    obs <- mets::fast.pattern(obs0)
    res <- matrix(nrow=nrow(data),ncol=NCOL(pp))
    for (i in seq_len(nrow(obs$pattern))) {
        jj <- which(obs$pattern[i,]==1)
        ii <- which(obs$group==i-1)
        if (length(jj)==0) {
            res[ii,] <- NA
        } else {
            res[ii,] <- predict.lvmfit(object,...,p=p,x=x[jj],y=y,data=data[ii,,drop=FALSE])[,colnames(pp),drop=FALSE]
        }
    }
    attributes(res) <- attributes(pp)
    return(res)
}


##' Prediction in structural equation models
##'
##' Prediction in structural equation models
##' @param object Model object
##' @param x optional list of (endogenous) variables to condition on
##' @param y optional subset of variables to predict
##' @param residual If true the residuals are predicted
##' @param p Parameter vector
##' @param data Data to use in prediction
##' @param path Path prediction
##' @param quick If TRUE the conditional mean and variance given covariates are returned (and all other calculations skipped)
##' @param \dots Additional arguments to lower level function
##' @seealso predictlvm
##' @examples
##' m <- lvm(list(c(y1,y2,y3)~u,u~x)); latent(m) <- ~u
##' d <- sim(m,100)
##' e <- estimate(m,d)
##'
##' ## Conditional mean (and variance as attribute) given covariates
##' r <- predict(e)
##' ## Best linear unbiased predictor (BLUP)
##' r <- predict(e,vars(e))
##' ##  Conditional mean of y3 giving covariates and y1,y2
##' r <- predict(e,y3~y1+y2)
##' ##  Conditional mean  gives covariates and y1
##' r <- predict(e,~y1)
##' ##  Predicted residuals (conditional on all observed variables)
##' r <- predict(e,vars(e),residual=TRUE)
##'
##' @method predict lvm
##' @aliases predict.lvmfit
##' @export
predict.lvm <- function(object,x=NULL,y=NULL,residual=FALSE,p,data,path=FALSE,quick=is.null(x)&!(residual|path),...) {
  ## data = data.frame of exogenous variables

  if (!quick && !all(exogenous(object)%in%colnames(data))) stop("data.frame should contain exogenous variables")
  m <- moments(object,p,data=data,...)
  if (quick) { ## Only conditional moments given covariates
      ii <- index(object)
      P.x <- m$P; P.x[ii$exo.idx, ii$exo.idx] <- 0
      Cy.x <- (m$IAi%*% tcrossprod(P.x,m$IAi))[ii$endo.idx,ii$endo.idx,drop=FALSE]
      X <- ii$exogenous
      mu.0 <- m$v; mu.0[ii$exo.idx] <- 0
      if (length(X)>0) {
          mu.x <- matrix(0,ncol=nrow(data),nrow=length(mu.0))
          mu.x[ii$exo.idx,] <- t(data[,X,drop=FALSE])
          xi.x <- t(m$IAi[ii$endo.idx,,drop=FALSE]%*%(mu.0 + mu.x))
      } else {
          xi.x <- m$xi%x%rep(1,nrow(data))
          colnames(xi.x) <- ii$endogenous
          ##xi.x <- matrix(as.vector(m$IAi[ii$endo.obsidx,]%*%mu.0),ncol=nrow(data),nrow=length(mu.0))
          ##rownames(xi.x) <- names(mu.0)
      }
      return(structure(xi.x,cond.var=Cy.x,
                       p=m$p,
                       e=m$e))
  }

    
  X <- exogenous(object)
  Y <- setdiff(manifest(object), X)
  if (path) {
    X <- colnames(data)
    Y <- setdiff(Y,X)
    idx <- which(vars(object)%in%X)
    if (length(Y)==0) stop("New data set should only contain exogenous variables and a true subset of the endogenous variables for 'path' prediction.")
    A <- t(m$A)
    A[,idx] <- 0 ## i.e., A <- A%*%J
    IAi <- solve(diag(nrow=nrow(A))-t(A))
    mu.0 <- m$v;
    mu.0[X] <- 0
    mu.x <- matrix(0,ncol=nrow(data),nrow=length(mu.0))
    mu.x[idx,] <- t(data[,vars(object)[idx],drop=FALSE])
    pred <- t(IAi%*%(mu.0 + mu.x))
    return(pred)
    ##  Y <- endogenous(object,top=TRUE)
    ##  X <- setdiff(manifest(object),Y)
  }


  IAi <- m$IAi
  eta.idx <- match(latent(object),vars(object))
  obs.idx <- match(manifest(object),vars(object))
  X.idx.all <- match(X, vars(object))
  Y.idx.all <- match(Y, vars(object))

  ## Calculation of conditional variance given X=x
  P.x <- m$P; P.x[X.idx.all, X.idx.all] <- 0
  C.x <- (IAi%*% P.x %*%t(IAi))
  Cy.x <- C.x[Y.idx.all,Y.idx.all,drop=FALSE]
  ## Calculation of conditional mean given X=x
  mu.0 <- m$v; mu.0[X.idx.all] <- 0
  if (length(X)>0) {
    xs <- data[,X,drop=FALSE]
    mu.x <- apply(xs, 1, FUN=function(i) {res <- rep(0,length(mu.0)); res[X.idx.all] <- i; res})
    xi.x <- (IAi%*%(mu.0 + mu.x))
  } else {
    xi.x <- matrix(as.vector(IAi%*%mu.0),ncol=nrow(data),nrow=length(mu.0))
    rownames(xi.x) <- names(mu.0)
  }

  attr(xi.x,"cond.var") <- Cy.x
  if (path) return(t(xi.x))
  Ey.x <- xi.x[Y.idx.all,,drop=FALSE]
  Eeta.x <- xi.x[eta.idx,,drop=FALSE]
  Cy.epsilon <- P.x%*%t(IAi) ## Covariance y,residual
    ##Czeta.y <- Cy.epsilon[eta.idx,index(object)$endo.idx]
  A <- m$A
  IA <- diag(nrow=nrow(A))-t(A)

  y0 <- intersect(Y,colnames(data))
  ys <- data[,y0,drop=FALSE]
  y0.idx <- match(y0,Y)
  ry <- t(ys)-Ey.x[y0.idx,,drop=FALSE]

  if (!is.null(x)) {
      if (inherits(x,"formula")) {
          xy <- getoutcome(x)
          if (length(xy)>0) {
              if (is.null(y)) y <- decomp.specials(xy)
          }
          x <- attributes(xy)$x
      }
      if (length(x)==0) {
          if (!is.null(y)) {
              xi.x <- xi.x[y,,drop=FALSE]
              attr(xi.x,"cond.var") <- Cy.x[y,y,drop=FALSE]
          }
          return(t(xi.x))
      }
      x <- intersect(x,endogenous(object))
      if (is.null(y))
          y <- setdiff(vars(object),c(x,exogenous(object)))
      if (length(x)>0) {
          E.x <- xi.x[y,,drop=FALSE] + C.x[y,x]%*%solve(C.x[x,x])%*%ry[x,,drop=FALSE]
      } else {
          E.x <- xi.x[y,,drop=FALSE]
      }
      if (residual) {
          Vhat <- matrix(0, nrow(data), length(vars(object))); colnames(Vhat) <- vars(object)
          Vhat[,obs.idx] <- as.matrix(data[,manifest(object),drop=FALSE])
          Vhat[,y] <- t(E.x)
          return(t((IA%*%t(Vhat)-m$v)))
      }
      res <- t(E.x); colnames(res) <- y
      if (length(x)>0) {
          attr(res,"cond.var") <-
              C.x[y,y,drop=FALSE]-C.x[y,x,drop=FALSE]%*%solve(C.x[x,x,drop=FALSE])%*%C.x[x,y,drop=FALSE]
      } else {
          attr(res,"cond.var") <- C.x[y,y,drop=FALSE]
      }
      return(res)
  }

  ys <- data[,Y,drop=FALSE]
  ry <- t(ys)-Ey.x

  if (length(eta.idx)>0) {
    Ceta.x <- C.x[eta.idx,eta.idx]
    Lambda <- A[Y.idx.all,eta.idx,drop=FALSE] ##, ncol=length(eta.idx))
    Cetay.x <- Ceta.x%*%t(Lambda)
    KK <- Cetay.x %*% solve(Cy.x)
    Eeta.y <- Eeta.x + KK %*% ry

    Ceta.y <- Ceta.x - KK%*% t(Cetay.x)
  } else {
    Eeta.y <- NA
    Ceta.y <- NA
  }

  Vhat <- matrix(0, nrow(data), length(vars(object))); colnames(Vhat) <- vars(object)
  Vhat[,obs.idx] <- as.matrix(data[,manifest(object)])
  if (length(eta.idx)>0)
    Vhat[,latent(object)] <- t(Eeta.y)
  epsilonhat <- (t( IA%*%t(Vhat) - m$v ))[,c(endogenous(object),latent(object)),drop=FALSE]
  if (residual) {
    return(epsilonhat)
  }

  mydata <- matrix(0,ncol=ncol(A),nrow=nrow(data)); colnames(mydata) <- vars(object)
  mydata[,manifest(object)] <- as.matrix(data[,manifest(object)])
  for (i in latent(object))
    mydata[,i] <- m$v[i]

  res <- cbind(t(Ey.x)) ## Conditional mean

  attr(res, "cond.var") <- Cy.x
  attr(res, "blup") <- t(Eeta.y)
  attr(res, "var.blup") <- Ceta.y
  attr(res, "Ey.x") <- Ey.x
  attr(res, "eta.x") <- Eeta.x
  attr(res, "epsilon.y") <- epsilonhat
  attr(res, "p") <- m$p
  attr(res, "e") <- m$e
  class(res) <- c("lvm.predict","matrix")
  return(res)
}

##' @export
print.lvm.predict <- function(x,...) print(x[,])

##' Predict function for latent variable models
##'
##' Predictions of conditinoal mean and variance and calculation of
##' jacobian with respect to parameter vector.
##' @export
##' @param object Model object
##' @param formula Formula specifying which variables to predict and which to condition on
##' @param p Parameter vector
##' @param data Data.frame
##' @param ... Additional arguments to lower level functions
##' @seealso predict.lvm
##' @examples
##' m <- lvm(c(x1,x2,x3)~u1,u1~z,
##'          c(y1,y2,y3)~u2,u2~u1+z)
##' latent(m) <- ~u1+u2
##' d <- simulate(m,10,"u2,u2"=2,"u1,u1"=0.5,seed=123)
##' e <- estimate(m,d)
##' 
##' ## Conditional mean given covariates
##' predictlvm(e,c(x1,x2)~1)$mean
##' ## Conditional variance of u1,y1 given x1,x2
##' predictlvm(e,c(u1,y1)~x1+x2)$var
predictlvm <- function(object,formula,p=coef(object),data=model.frame(object),...) {
    model <- Model(object)
    if (!missing(formula)) {
        yx <- getoutcome(formula)
        y <- decomp.specials(yx)
        x <- attr(yx,"x")
        x <- setdiff(x,index(model)$exogenous)
    } else {
        y <- index(model)$latent
        x <- index(model)$endogenous
    }    
    endo <- with(index(model),setdiff(vars,exogenous))
    idxY <- match(y,endo)
    idxX <- match(x,endo)
    ny <- length(y)
    if (ny==0) return(NULL)
    m <- modelVar(model,p,conditional=TRUE,data=data,latent=TRUE)
    D <- deriv.lvm(model,p,conditional=TRUE,data=data,latent=TRUE)
    N <- nrow(data)
    ii0 <- seq(N)
    iiY <- sort(unlist(lapply(idxY,function(x) ii0+N*(x-1))))
    k <- ncol(m$xi)
    J <- matrix(seq(k^2),k)
    if (length(idxX)==0) { ## Return conditional mean and variance given covariates
        M <- m$xi[,idxY,drop=FALSE]
        dM <- D$dxi[iiY,,drop=FALSE]
        V <- m$C[idxY,idxY,drop=FALSE]
        dV <- D$dS[as.vector(J[idxY,idxY]),,drop=FALSE]
    } else {
        iiX <- sort(unlist(lapply(idxX,function(x) ii0+N*(x-1))))
        X <- as.matrix(data[,x,drop=FALSE])
        rX <- X-m$xi[,idxX,drop=FALSE]
        dX <- D$dxi[iiX,,drop=FALSE]
        ic <- solve(m$C[idxX,idxX,drop=FALSE])
        c2 <- m$C[idxY,idxX,drop=FALSE]
        B <- c2%*%ic
        ## Conditional variance
        V <- m$C[idxY,idxY,drop=FALSE]-B%*%t(c2)
        dV <- D$dS[as.vector(J[idxY,idxY]),,drop=FALSE] -
            (
                (B%x%diag(nrow=ny))%*%D$dS[as.vector(J[idxY,idxX]),,drop=FALSE] +
            -(B%x%B)%*%D$dS[as.vector(J[idxX,idxX]),,drop=FALSE] +
            (diag(nrow=ny)%x%B)%*%D$dS[as.vector(J[idxX,idxY]),,drop=FALSE]
            )
        ## Conditional mean
        M <- m$xi[,idxY,drop=FALSE]+rX%*%t(B)
        dB <- (ic%x%diag(nrow=ny))%*%D$dS[as.vector(J[idxY,idxX]),,drop=FALSE]+
            -(ic%x%B)%*%D$dS[as.vector(J[idxX,idxX]),,drop=FALSE]
        ## Find derivative of transposed matrix
        n0 <- as.vector(matrix(seq(prod(dim(B))),ncol=nrow(B),byrow=TRUE))
        dB. <- dB[n0,,drop=FALSE]
        dM <- D$dxi[iiY,,drop=FALSE] +
        ((diag(nrow=ny)%x%rX)%*%dB.) - kronprod(B,dX)
    }
    colnames(M) <- y
    dimnames(V) <- list(y,y)
    return(list(mean=M,mean.jacobian=dM,var=V,var.jacobian=dV))
}

#' @export
estimate <- function(x, ...) UseMethod("estimate")

#' Influence function based inference
#'
#' Primary tool for obtaining parameter estimates with robust (sandwich)
#' standard errors, applying the delta method, and testing linear hypotheses.
#' The function returns an object of class `estimate` which serves as a
#' general container for parameter estimates and their influence functions
#' (IFs). Three calling conventions are supported:
#'
#' - `estimate(x, ...)` -- extract estimates from a model object
#' - `estimate(coef=, IC=, ...)` -- construct from coefficients and IF matrix
#' - `estimate(coef=, vcov=, ...)` -- construct from coefficients and
#'   covariance matrix
#'
#' @param x model object (`glm`, `lvmfit`, ...) or an existing `estimate`
#'   object. When two model objects are supplied (e.g., `estimate(g, g0)`) a
#'   likelihood-ratio test is performed.
#' @param f transformation of model parameters. Accepts several input types: - A
#'   **function** `f(p)` or `f(p, data)`: applies the delta method. When `f`
#'   returns a named list the names are used as parameter labels. - A
#'   **matrix**: used as a contrast (linear combination) matrix. - A **numeric
#'   vector** of parameter indices: converted to a contrast that selects and
#'   differences those parameters. - A **list** of indices: each element selects
#'   one parameter. - **Character** expressions: supports wildcards (`"?"`,
#'   `"*"`) and arithmetic on parameter names (e.g., `"z" - "x"`, `2 * "z" - 3 *
#'   "x"`).
#' @param ... additional arguments to lower level functions
#' @param data `data.frame` used by `f` when the transformation depends on
#'   covariates (see `average`). Defaults to `model.frame(x)`.
#' @param id (optional) cluster identifier. Can be a vector of cluster IDs, a
#'   one-sided formula (evaluated in `data`), a single character column name, or
#'   a logical scalar (`TRUE` for one-to-one matching, `FALSE` for
#'   independence). When supplied, the IF is aggregated within clusters to
#'   produce cluster-robust standard errors.
#' @param coef (optional) named parameter vector. Used instead of `coef(x)` when
#'   constructing an `estimate` object without a model.
#' @param IC if `TRUE` (default) the influence function matrix is estimated and
#'   stored in the returned object (extract with the [IC] method). Can also be a
#'   user-supplied IF matrix (one row per observation, one column per
#'   parameter), which is used directly instead of estimating it from `x`.
#' @param vcov (optional) covariance matrix of parameter estimates, or a
#'   logical. If `TRUE`, [stats::vcov] is used to obtain the (model-based)
#'   covariance matrix from `x`, yielding non-robust standard errors. If a
#'   matrix is supplied it is used directly. When omitted or `FALSE`, robust
#'   standard errors are computed from the influence function.
#' @param stack if `TRUE` (default) the influence function contributions are
#'   summed within each cluster defined by `id`. Set to `FALSE` to keep the
#'   un-stacked (per-observation) decomposition.
#' @param average if `TRUE` the function computes the standardized
#'   (marginalized) estimate \eqn{\hat\Psi = P_n f(X; \hat\theta)}, i.e., the
#'   empirical mean of `f(p, data)` over all rows of `data`. The influence
#'   function accounts for both the empirical averaging and the parameter
#'   estimation uncertainty (see Details).
#' @param subset (optional) logical vector, expression evaluated in `data`, or
#'   column name. When used together with `average = TRUE`, the average is
#'   conditioned on the subpopulation where `subset` is `TRUE`, yielding a
#'   conditional marginalized estimate.
#' @param keep (optional) index of parameters to keep from final result. Accepts
#'   integer indices, character names, or (with `regex = TRUE`) perl-compatible
#'   regular expressions.
#' @param use (optional) index of parameters to use in calculations. The
#'   selected parameters are first extracted (via `keep`) and then the remaining
#'   arguments (`f`, `contrast`, etc.) are applied to this subset.
#' @param regex if `TRUE` use perl-compatible regular expressions for `keep` and
#'   `use` arguments
#' @param ignore.case ignore case in regular expressions
#' @param print (optional) custom print function for the resulting `estimate`
#'   object
#' @param labels (optional) character vector of coefficient names
#' @param label.width (optional) max display width of labels
#' @param contrast (deprecated, use summary method) contrast matrix for a final
#'   Wald test. When supplied together with `null`, tests \eqn{H_0: B\theta =
#'   b_0}.
#' @param null (deprecated, use summary method) null hypothesis vector \eqn{b_0}
#'   to test against (default 0)
#' @param level (deprecated, use summary method) level of confidence limits
#'   (default 0.95)
#' @param type (deprecated, use summary method) type of small-sample correction
#'   for cluster-robust variance. One of: - `"robust"` (default): no correction.
#'   - `"df"`: applies \eqn{n/(n-p)} correction (Mancl & DeRouen, 2001). -
#'   `"mbn"`: Morel-Bokossa-Neerchal (2003) correction. - `"hc3"`:
#'   leverage-adjusted HC3-type correction (blended with `var.adj`). - `"hc4"`:
#'   Cribari-Neto (2004) leverage-adjusted correction.
#' @param var.adj (deprecated, use summary method) blending parameter for the
#'   HC3 leverage adjustment (default 0.25). Controls the weight between
#'   observation-level empirical leverage and the average leverage \eqn{p/n}.
#' @param df (deprecated, use summary method) degrees of freedom for t-based
#'   inference (default: `NULL` for Gaussian approximation; when set, confidence
#'   intervals and p-values use the t-distribution with `df` degrees of
#'   freedom).
#' @param back.transform (deprecated, use summary method) function applied to
#'   the point estimates and confidence interval bounds *after* inference is
#'   performed on the original scale. Useful for variance-stabilizing
#'   transformations, e.g., compute CIs on the `atanh` (Fisher z) scale and
#'   back-transform with `tanh`.
#' @details
#'
#' # Influence functions and robust standard errors
#'
#' An estimator \eqn{\widehat{\theta}} is *regular and asymptotically
#' linear* (RAL) when it admits the iid decomposition
#' \deqn{\sqrt{n}(\widehat{\theta}-\theta) =
#' \frac{1}{\sqrt{n}}\sum_{i=1}^n \mathrm{IC}(Z_i; P) + o_p(1)}
#' where \eqn{\mathrm{IC}} is the unique *influence function* satisfying
#' \eqn{E\{\mathrm{IC}(Z; P)\} = 0}. By the central limit theorem
#' \deqn{\sqrt{n}(\widehat{\theta}-\theta)
#' \overset{d}{\longrightarrow}
#' N(0,\; \mathrm{Var}\{\mathrm{IC}(Z; P)\})}
#' and the asymptotic variance is consistently estimated by the empirical
#' variance of the plugin IF estimate, yielding robust (sandwich) standard
#' errors. The estimated IF can be extracted with the [IC] method.
#'
#' # Parameter transformations (delta method)
#'
#' When `f` is a function \eqn{\phi: R^p \to R^m}, the delta method is
#' applied:
#' \deqn{\sqrt{n}\{\phi(\widehat{\theta}) - \phi(\theta)\} =
#' \frac{1}{\sqrt{n}}\sum_{i=1}^n
#' \nabla\phi(\theta)\,\mathrm{IC}(Z_i; P) + o_p(1)}
#' Derivatives are computed numerically via [numDeriv::jacobian] unless the
#' function returns an attribute `"grad"` with the analytic Jacobian.
#'
#' Alternatively, `estimate` objects support direct arithmetic operations
#' (e.g., `a * b`, `exp(a)`, `a^b`) which apply the delta method with
#' *exact* (analytical) derivatives computed automatically. This influence
#' function calculus allows building complex transformations from simple
#' building blocks without numerical differentiation. See the last example
#' section ("influence function calculus") and
#' `vignette("influencefunction", package = "lava")` for details.
#'
#' # Averaging and marginalization
#'
#' When `average = TRUE` and `f(p, data)` depends on covariates, the
#' target parameter is the standardized (marginalized) estimate
#' \eqn{\Psi = E\{f(X;\theta)\}}. The IF for the averaged estimate
#' accounts for both the empirical averaging and parameter estimation
#' uncertainty:
#' \deqn{\mathrm{IC}_\Psi(Z; P) = f(X;\theta) - \Psi +
#' [E\nabla_\theta f(X;\theta)]\,\phi(Z; P)}
#' When `subset` is also specified, the average is conditioned on the
#' subpopulation, yielding a conditional marginalized estimate.
#'
#' # Cluster-robust standard errors
#'
#' When `id` is supplied, the per-observation IF contributions are summed
#' within clusters (when `stack = TRUE`), producing the cluster-level IF
#' \eqn{\widetilde{\mathrm{IC}}(Z_i; P) = \sum_{k=1}^{N_i}
#' \frac{n}{N}\mathrm{IC}(Z_{ik}; P)}.
#' The resulting variance estimate is equivalent to the GEE working
#' independence sandwich estimator.
#'
#' For full theoretical background and worked examples see
#' `vignette("influencefunction", package = "lava")`.
#'
#' @export
#' @export estimate.default
#' @examples
#'
#' ## Simulation from logistic regression model
#' m <- lvm(y~x+z);
#' distribution(m,y~x) <- binomial.lvm("logit")
#' d <- sim(m,1000)
#' g <- glm(y~z+x,data=d,family=binomial())
#' g0 <- glm(y~1,data=d,family=binomial())
#'
#' ## LRT
#' estimate(g, g0)
#'
#'
## Plain estimates (robust standard errors)
#' estimate(g)
#'
#' ## Testing contrasts
#' summary(estimate(g), null=0)
#' estimate(g, rbind(c(1,1,0), c(1,0,2)))
#' summary(estimate(g, rbind(c(1,1,0), c(1,0,2))), null=c(1,2))
#' estimate(g, 2:3) ## same as cbind(0,1,-1)
#' estimate(g, as.list(2:3)) ## same as rbind(c(0,1,0),c(0,0,1))
#' ## Alternative syntax
#' estimate(g, "z", "z"-"x", 2*"z"-3*"x")
#' estimate(g, "?")  ## Wildcards
#' estimate(g, "*Int*", "z")
#' summary(estimate(g, "1", "2"-"3"), null = c(0,1))
#' estimate(g, 2, 3)
#'
#' ## Usual (non-robust) confidence intervals
#' estimate(g, vcov=TRUE)
#' estimate(g, vcov=vcov(g))
#'
#' ## Transformations
#' estimate(g, function(p) p[1]+p[2])
#'
#' ## Multiple parameters
#' e <- estimate(g, function(p) c(p[1]+p[2], p[1]*p[2]))
#' e
#' vcov(e)
#'
#' ## Label new parameters
#' estimate(g, function(p) list("a1"=p[1]+p[2], "b1"=p[1]*p[2]))
#' #'
#' ## Multiple group
#' m <- lvm(y~x)
#' m <- baptize(m)
#' d2 <- d1 <- sim(m,50,seed=1)
#' e <- estimate(list(m,m),list(d1,d2))
#' estimate(e) ## Wrong
#' ee <- estimate(e, id=rep(seq(nrow(d1)), 2)) ## Clustered
#' ee
#' estimate(lm(y~x,d1))
#'
#' ## Marginalize / standardization
#' f <- function(p,data)
#'   list(p0=expit(p["(Intercept)"] + p["z"]*data[,"z"]),
#'        p1=expit(p["(Intercept)"] + p["x"] + p["z"]*data[,"z"]))
#' e <- estimate(g, f, average=TRUE)
#' e
#' estimate(e,diff)
#' estimate(e,cbind(1,1))
#'
#' ## Clusters and subset (conditional marginal effects)
#' d$id <- rep(seq(nrow(d)/4),each=4)
#' estimate(g,function(p,data)
#'          list(p0=expit(p[1] + p["z"]*data[,"z"])),
#'          subset=d$z>0, id=d$id, average=TRUE)
#'
#' ## More examples with clusters:
#' m <- lvm(c(y1,y2,y3)~u+x)
#' d <- sim(m,10)
#' l1 <- glm(y1~x,data=d)
#' l2 <- glm(y2~x,data=d)
#' l3 <- glm(y3~x,data=d)
#'
#' ## Some random id-numbers
#' id1 <- c(1,1,4,1,3,1,2,3,4,5)
#' id2 <- c(1,2,3,4,5,6,7,8,1,1)
#' id3 <- seq(10)
#'
#' ## Un-stacked and stacked i.i.d. decomposition
#' IC(estimate(l1,id=id1,stack=FALSE))
#' IC(estimate(l1,id=id1))
#'
#' ## Combined i.i.d. decomposition
#' e1 <- estimate(l1,id=id1)
#' e2 <- estimate(l2,id=id2)
#' e3 <- estimate(l3,id=id3)
#' (a2 <- merge(e1,e2,e3))
#'
#' ## If all models were estimated on the same data we could use the
#' ## syntax:
#' ## Reduce(merge,estimate(list(l1,l2,l3)))
#'
#' ## Same:
#' IC(a1 <- merge(l1,l2,l3,id=list(id1,id2,id3)))
#'
#' IC(merge(l1,l2,l3,id=TRUE)) # one-to-one (same clusters)
#' IC(merge(l1,l2,l3,id=FALSE)) # independence
#'
#'
#' # ------ influence function calculus -------
#' a <- estimate(coef = c("a" = 0.5), IC = scale(rnorm(10), scale=FALSE), id = 1:10)
#' b <- estimate(coef = c("b" = 0.8), IC = scale(rnorm(10), scale=FALSE), id = 1:10)
#'
#' e <- c(a, b) # merge
#' merge(a, b)
#' c(e1=a, b) # naming of par
#' labels(e, c("p1", "p2")) # renaming parameters
#' e["a"] # subset
#' subset(e, "a")
#'
#' # pipes
#' # c(a, b) |>
#' #  transform(function(x) x^2) |>
#' #  subset("a") |>
#' #  labels("sq")
#'
#' # Parameter transformation with automatic calculation of derivatives
#' a * b
#' (3 * cos(a) / sqrt(b) + 1) / a
#' expit(c(a,b))
#' c(sum=sum(e), sum2=a+b,
#'   prod=prod(e), prod2=a*b)
#' e %*% e # inner prod.
#' c(1, 2) %*% e
#' c(pow = a^b)
#' a^c(0.5, 2)
#' c(b=e["a"] * e["b"] / a, also.b=e["b"])
#'
#' B <- rbind(c(1,-1), c(1,0), c(0,1))
#' B %*% e
#' e == 1 # wald-test, null-hypothesis H0: b=1
#' e == c(1,2)
#' B %*% e == 1
#' @aliases estimate estimate.default
#' @aliases estimate.mlm
#' @seealso [estimate.array], [merge.estimate], [contr], [parsedesign],
#'   [pairwise.diff], [c.estimate], [summary.estimate],
#'   `coef.estimate`,
#'   `vcov.estimate`, `transform.estimate`, `labels.estimate`,
#' @return Object of class `estimate` with the following elements:
#'   \item{coef}{Named vector of parameter estimates.}
#'   \item{vcov}{Variance-covariance matrix.}
#'   \item{IC}{Influence function matrix (observations x parameters).}
#'   \item{coefmat}{Formatted coefficient table (estimate, std.err,
#'     confidence limits, p-value).}
#'   \item{id}{Cluster/id variable used.}
#'   \item{ncluster}{Number of clusters.}
#'   \item{n}{Number of observations.}
#'   \item{compare}{(When `null` or contrasts are specified) Wald test result.}
#' @method estimate default
#' @export
estimate.default <- function(x=NULL, f=NULL, ...,
                             data, id,
                             coef, IC=TRUE, vcov,
                             stack=TRUE,
                             average=FALSE, subset,
                             keep, use,
                             regex=FALSE, ignore.case=FALSE,
                             print=NULL, labels, label.width,
                             # deprecated argumets (removed in version 1.9.3)
                             # return object of class summary.estimate
                             contrast,
                             null,
                             level=NULL,
                             type=NULL,
                             var.adj=NULL,
                             df=NULL,
                             back.transform=NULL
                             ) {
  cl <- match.call(expand.dots = TRUE)
  cal <- match.call()

  if ("iid" %in% names(cl)) {
    stop("The 'iid' argument is obsolete. Please use the 'IC' argument")
  }
  if ("R" %in% names(cl) || "null.sim" %in% names(cl)) {
    stop("The 'R' and 'null.sim' arguments have been removed. ")
  }
  if ("robust" %in% names(cl)) {
    warning(
      "The 'robust' argument is deprecated and ignored. ",
      "Robust standard errors are now always computed. ",
      "Use the 'vcov' argument to compute model-based SEs."
    )
  }

  if (!missing(use)) {
    p0 <- c(
      "f", "subset", "average",
      "keep", "labels", "null"
    )
    cl0 <- cl
    cl0[c("use", p0)] <- NULL
    cl0$keep <- use
    cl$x <- eval(cl0, parent.frame())
    cl[c("vcov", "use")] <- NULL
    res <- eval(cl, parent.frame())
    res$call <- cal
    return(res)
  }
  expr <- suppressWarnings(inherits(try(f, silent=TRUE), "try-error"))
  if (!missing(coef)) {
    pp <- coef
  } else {
    pp <- suppressWarnings(try(stats::coef(x), silent = TRUE))
    if (inherits(x, "survreg") && length(pp) < NROW(x$var)) {
      pp <- c(pp, scale=x$scale)
    }
  }

  if (is.null(names(pp))) {
    names(pp) <- paste0("p", seq_along(pp))
  }

  if ((expr || is.character(f) || (is.numeric(f))
    && !is.matrix(f))) { ## || is.call(f)) {
    dots <- lapply(substitute(placeholder(...))[-1], function(x) x)
    args <- c(list(
      coef = names(pp),
      x = substitute(f),
      regex = regex
    ), dots)
    f <- do.call(parsedesign, args)
  }

  contrast.transform <- FALSE  # if TRUE parameter estimates should be
                               # transformed according to contrast matrix 'f'
  if (!is.null(f) && !is.function(f)) {
    if (!(is.matrix(f) || is.vector(f)))
      return(compare(x, f, ...)) ## LRT
    contrast.transform <- TRUE
  }

  if (lava.options()$cluster.index) {
    if (!requireNamespace("mets", quietly=TRUE)) stop("'mets' package required")
  }
  if (missing(data))
    data <- tryCatch(model.frame(x), error=function(...) NULL)
  nn <- NULL
  if (((is.logical(IC) && IC) || length(IC)>0) &&
      (missing(vcov) || is.null(vcov) ||
       (is.logical(vcov) && vcov[1]==FALSE && !is.na(vcov[1])))) {
    ## If user supplied vcov, then don't estimate IC
    if (!is.logical(IC)) {
      ic_theta <- cbind(IC)
      if (NCOL(ic_theta) != length(pp)) {
        warning("Wrong dimension of influence function IC")
      }
      if (lava.options()$check.ic) {
        check_ic_mean_zero(ic_theta)
      }
      IC <- TRUE
    } else {
      ic_theta <- IC(x)
    }
  } else {
    if (!is.null(x) && (missing(vcov) ||
                        (is.logical(vcov) && !is.na(vcov)[1])))
      suppressWarnings(vcov <- stats::vcov(x))
    ic_theta <- NULL
  }

  if (any(is.na(ic_theta))) {
    ## Rescale each column according to I(obs)/pr(obs)
    for (i in seq_len(NCOL(ic_theta))) {
      pr <- mean(!is.na(ic_theta[, i]))
      ic_theta[, i] <- ic_theta[, i]/pr
    }
    ic_theta[is.na(ic_theta)] <- 0
  }

  if (!missing(subset)) {
    e <- substitute(subset)
    expr <- suppressWarnings(inherits(try(subset, silent=TRUE), "try-error"))
    if (expr) subset <- eval(e, envir=data)
    if (is.character(subset)) subset <- data[, subset]
    if (is.numeric(subset)) subset <- subset > 0
  }
  idstack <- NULL
  ## Preserve id from 'estimate' object
  if (missing(id)) {
    if (inherits(x, "measurement.error")) {
      if (!is.null(x[["id"]])) id <- x[["id"]]
    } else if (inherits(x, "estimate") && !is.null(index(x))) {
      id <- index(x)
    }
  }
  if (!missing(id) && IC) {
    if (is.null(ic_theta)) stop("'IC' method needed")
    nprev <- nrow(ic_theta)
    if (inherits(id, "formula")) {
      id <- interaction(get_all_vars(id, data))
    }
    ## e <- substitute(id)
    ## expr <- suppressWarnings(inherits(try(id,silent=TRUE),"try-error"))
    ## if (expr) id <- eval(e,envir=data)
    ##if (!is.null(data)) id <- eval(e, data)
    if (is.logical(id) && length(id)==1) {
      id <- if(is.null(ic_theta)) seq_len(nrow(data)) else seq_len(nprev)
      stack <- FALSE
    }
    if (is.character(id) && length(id)==1)
      id <- data[, id, drop=TRUE]
    if (!is.null(ic_theta)) {
      if (length(id)!=nprev) {
        if (!is.null(x$na.action) &&
            (length(id)==length(x$na.action) + nprev)) {
          warning("Applying na.action")
          id <- id[-x$na.action]
        } else stop("Dimensions of i.i.d decomposition and 'id' does not agree")
      }
    } else {
      if (length(id)!=nrow(data)) {
        if (!is.null(x$na.action) &&
            (length(id)==length(x$na.action)+nrow(data))) {
          warning("Applying na.action")
          id <- id[-x$na.action]
        } else stop("Dimensions of IC and 'id' does not agree")
      }
    }
    if (stack) {
      N <- nrow(ic_theta)
      clidx <- NULL
      atr <- attributes(ic_theta)
      atr$dimnames <- NULL
      atr$dim <- NULL
      if (!lava.options()$cluster.index) {
        ic_theta <- matrix(unlist(by(ic_theta, id, colSums)),
                           byrow=TRUE, ncol=ncol(ic_theta))
        attributes(ic_theta)[names(atr)] <- atr
        idstack <- sort(unique(id))
      } else {
        clidx <- mets::cluster.index(id, mat=ic_theta, return.all=TRUE)
        ic_theta <- with(clidx, X)
        attributes(ic_theta)[names(atr)] <- atr
        idstack <- id[as.vector(clidx$firstclustid)+1]
      }
      ic_theta <- ic_theta*NROW(ic_theta)/length(id)
      if (is.null(attributes(ic_theta)$N)) {
        attributes(ic_theta)$N <- N
      }
    } else idstack <- id
  } else {
    if (!is.null(data)) idstack <- rownames(data)
  }
  if (!is.null(ic_theta) && (length(idstack)==nrow(ic_theta))) {
    rownames(ic_theta) <- idstack
  }
  if (!is.null(ic_theta) && (missing(vcov) || is.null(vcov))) {
    V <- var_ic(ic_theta)
  } else {
    if (!missing(vcov)) {
      if (length(vcov) == 1 && is.na(vcov)) {
        vcov <- matrix(NA, length(pp), length(pp))
      }
      V <- cbind(vcov)
    } else {
      suppressWarnings(V <- stats::vcov(x))
    }
  }

  if (contrast.transform) {
    B <- f
    if (is.vector(B) || is.list(B)) {
      B <- contr(f, names(pp), ...)
    }
    obj <- structure(list(coef=pp, vcov=V), class="estimate")
    cc <- compare(obj, contrast=B) # to construct new parameter names
    pp <- as.vector(B %*% pp)
    names(pp) <- strip_bracket(cc$cnames)
    if (!is.null(ic_theta)) {
      ic_theta <- ic_theta %*% t(B)
    }
    V <- B %*% V %*% t(B)
    f <- NULL
  }

  derivative <- NULL
  if (!is.null(f)) {
    form <- names(formals(f))
    dots <- ("..."%in%names(form))
    form0 <- setdiff(form, "...")
    parname <- "p"
    if (!is.null(form)) parname <- form[1] # unless .Primitive
    if (length(form0)==1 && !(form0%in%c("object", "data"))) {
      parname <- form0
    }
    if (!is.null(ic_theta)) {
      arglist <- c(list(object=x, data=data, p=vec(pp)), list(...))
      names(arglist)[3] <- parname
    } else {
      arglist <- c(list(object=x, p=vec(pp)), list(...))
      names(arglist)[2] <- parname
    }
    if (!dots) {
      arglist <- arglist[intersect(form0, names(arglist))]
    }
    newf <- NULL
    if (length(form)==0) {
      arglist <- list(vec(pp))
      newf <- function(...) do.call("f", list(...))
      val <- do.call("f", arglist)
    } else {
      val <- do.call("f", arglist)
      if (is.list(val)) {
        nn <- names(val)
        val <- do.call("cbind", val)
        newf <- function(...) do.call("cbind", f(...))
      }
    }
    k <- NCOL(val)
    N <- NROW(val)
    D <- attributes(val)$grad
    if (!is.null(D)) derivative <- D
    if (is.null(D)) {
      D <- numDeriv::jacobian(function(p, ...) {
        if (length(form)==0) arglist[[1]] <- p
        else arglist[[parname]] <- p
        if (is.null(newf))
          return(do.call("f", arglist))
        return(do.call("newf", arglist)) }, pp)
    }
    if (is.null(ic_theta)) {
      pp <- structure(as.vector(val), names=names(val))
      V <- D%*%V%*%t(D)
    } else {
      if (!average || (N<NROW(data))) {  ## transformation not depending on data
        pp <- structure(as.vector(val), names=names(val))
        ic_theta <- ic_theta%*%t(D)
        V <- var_ic(ic_theta)
      } else {
        if (k>1) { ## More than one parameter (and depends on data)
          if (!missing(subset)) { ## Conditional estimate
            val <- apply(val, 2, function(x) x*subset)
          }
          D0 <- matrix(nrow=k, ncol=length(pp))
          for (i in seq_len(k)) {
            D1 <- D[seq(N)+(i-1)*N, , drop=FALSE]
            if (!missing(subset)) ## Conditional estimate
              D1 <- apply(D1, 2, function(x) x*subset)
            D0[i, ] <- colMeans(D1)
          }
          D <- D0
          ic2 <- ic_theta%*%t(D)
        } else { ## Single parameter
          if (!missing(subset)) { ## Conditional estimate
            val <- val*subset
            D <- apply(rbind(D), 2, function(x) x*subset)
          }
          D <- colMeans(rbind(D))
          ic2 <- ic_theta%*%D
        }
        pp <- vec(colMeans(cbind(val)))
        ic1 <- (cbind(val)-rbind(pp)%x%cbind(rep(1, N)))
        if (NROW(ic_theta)==NROW(ic1)) {
          rownames(ic1) <- rownames(ic_theta)
        }
        if (!missing(id)) {
          if (!lava.options()$cluster.index)
            ic1 <- matrix(unlist(by(ic1, id, colSums)),
                          byrow=TRUE, ncol=ncol(ic1))
          else {
            ic1 <- mets::cluster.index(id, mat=ic1, return.all=FALSE)
          }
          ic1 <- ic1 * NROW(ic1) / length(id)
        }
        if (!missing(subset)) { ## Conditional estimate
          phat <- mean(subset)
          ic3 <- cbind(-1/phat^2 * (subset-phat)) ## check
          if (!missing(id)) {
            if (!lava.options()$cluster.index) {
              ic3 <- matrix(unlist(by(ic3, id, colSums)),
                            byrow=TRUE, ncol=ncol(ic3))
            } else {
              ic3 <- mets::cluster.index(id, mat=ic3, return.all=FALSE)
            }
          }
          ic3 <- ic3*NROW(ic3)/length(id)
          ic_theta <- (ic1+ic2)/phat + rbind(pp)%x%ic3
          pp <- pp/phat
          V <- var_ic(ic_theta)
        } else {
          if (nrow(ic1)!=nrow(ic2)) {
            message("Assuming independence between model iid decomposition and new data frame") #nolint
            V <- var_ic(ic1) + var_ic(ic2)
          } else {
            ic_theta <- ic1+ic2
            V <- var_ic(ic_theta)
          }
        }
      }
    }
  }

  df_mod <- NULL
  if (inherits(x, "lm") && family(x)$family == "gaussian"
      && !missing(vcov)) {
    # defaults to t-distribution when calculating p-values with model-based SEs
    df_mod <- x$df.residual
  }
  if (is.null(V)) {
    res <- cbind(pp, NA, NA, NA, NA)
  } else {
    if (length(pp)==1)
      res <- rbind(c(pp, diag(V)^0.5))
    else
      res <- cbind(pp, diag(V)^0.5)
  }
  res <- estimate_coefmat(res[, 1], res[, 2], df=df_mod, level=0.95, null=0)
  if (nrow(res)>0)
    if (!is.null(nn)) {
      rownames(res) <- nn
    } else {
      nn <- attributes(res)$varnames
      if (!is.null(nn))
        rownames(res) <- nn
      if (is.null(rownames(res)))
        rownames(res) <- paste0("p", seq_len(nrow(res)))
    }

  if (NROW(res)==0L) {
    coefs <- NULL
  } else {
    coefs <- res[, 1, drop=TRUE]
    names(coefs) <- rownames(res)
  }
  res <- structure(list(coef=coefs, coefmat=res, vcov=V,
                        IC=NULL, print=print, id=idstack, df=df_mod),
                   class="estimate")
  if (IC) {
    res$IC <- ic_theta
  }
  if (length(coefs)==0L) return(res)

  if (!missing(keep) && !is.null(keep)) {
    if (is.character(keep)) {
      if (regex) {
        nn <- rownames(res$coefmat)
        keep <- unlist(lapply(keep, function(x) {
          grep(x, nn,
               perl = TRUE,
               ignore.case = ignore.case
               )
        }))
      } else {
        keep <- match(keep, rownames(res$coefmat))
      }
    }
    res$coef <- res$coef[keep]
    res$coefmat <- res$coefmat[keep, , drop=FALSE]
    if (!is.null(res$IC)) res$IC <- res$IC[, keep, drop=FALSE]
    res$vcov <- res$vcov[keep, keep, drop=FALSE]
  }

  res <- labels.estimate(
    object = res, str = labels, label.width = label.width
  )
  res$call <- cal
  res$n <- nrow(data)
  res$ncluster <- if (!is.null(ic_theta)) nrow(ic_theta) else nrow(data)
  res$derivative <- derivative
  res <- structure(res, class="estimate")

  if (!missing(null) || !missing(contrast) ||
      !is.null(type) || !is.null(var.adj) ||
      !is.null(back.transform) || !is.null(level) ||
      !is.null(df)
      ) {
    .Deprecated(
      msg = paste0(
        "The 'null', 'contrast', 'type', 'back.transform', 'level' ",
        "and 'var.adj' arguments of estimate.default() are deprecated ",
        "and will be removed in version 1.9.3. Use ",
        "summary(estimate(...), null=, contrast=, type=, transform=,",
        "level=, df=, var.adj=) instead."
      )
    )
    args <- list(object=res)
    if (!missing(contrast)) args$contrast <- contrast
    if (!missing(null)) args$null <- null
    if (!is.null(level)) args$level <- level
    if (!is.null(type)) args$type <- type
    if (!is.null(var.adj)) args$var.adj <- var.adj
    if (!is.null(df)) args$df <- df
    if (!is.null(back.transform)) args$transform <- back.transform
    return(do.call(summary, args))
  }

  return(res)
}

#' @export
print.estimate <- function(x, type=0L, digits=4L, width=25L,
                           std.error=TRUE, p.value=TRUE,
                           sep=cli::symbol[["line"]],
                           sep.which,
                           sep.labels=NULL,
                           indent=" ", unique.names=TRUE,
                           na.print="", ...) {

  if (!is.null(x$print)) {
    x$print(x, digits=digits, width=width, ...)
    return(invisible(x))
  }
  if (type>0 && !is.null(x$call)) {
    cat("Call: ")
    print(x$call)
    print(cli::rule(width=min(cli::console_width(),60)))
  }
  if (type>0) {
    if (!is.null(x[["n"]]) && !is.null(x[["k"]])) {
      cat("n = ", x[["n"]], ", clusters = ", x[["k"]], "\n\n", sep="")
    } else {
      if (!is.null(x[["n"]])) {
        cat("n = ", x[["n"]], "\n\n", sep="")
      }
      if (!is.null(x[["k"]])) {
        cat("n = ", x[["k"]], "\n\n", sep="")
      }
    }
  }

  cc <- x$coefmat
  if (!is.null(rownames(cc)) && unique.names)
    rownames(cc) <- make.unique(
      unlist(lapply(rownames(cc),
                    function(x) toString(x, width=width)))
    )
  if (!std.error) cc <- cc[, -2, drop=FALSE]
  if (!p.value) cc <- cc[, -ncol(cc), drop=FALSE]

  sep.pos <- c()
  if (missing(sep.which) && !is.null(x$model.index)) {
    sep.which <- unlist(lapply(x$model.index,
                               function(x)
                                 tail(x, 1)))[-length(x$model.index)]
  }
  if (missing(sep.which)) sep.which <- NULL

  if (!is.null(sep.which)) {
    sep0 <- 0%in%sep.which
    if (sep0)
      sep.which <- setdiff(sep.which, 0)
    cc0 <- c()
    sep.which <- c(0, sep.which, nrow(cc))
    N <- length(sep.which)-1
    for (i in seq(N)) {
      if ((sep.which[i]+1)<=nrow(cc))
        cc0 <- rbind(cc0, cc[seq(sep.which[i]+1, sep.which[i+1]), , drop=FALSE])
      if (i<N) {
        cc0 <- rbind(cc0, NA)
        sep.pos <- c(sep.pos, nrow(cc0))
      }
    }
    if (sep0) {
      sep.pos <- c(1, sep.pos+1)
      cc0 <- rbind(NA, cc0)
    }
    cc <- cc0
  }
  if (!is.null(sep.labels)) {
    sep.labels <- rep(sep.labels, length.out=length(sep.pos))
    rownames(cc)[sep.pos] <- sep.labels
    rownames(cc)[-sep.pos] <- paste0(indent, rownames(cc)[-sep.pos])
  } else {
    if (length(sep.pos)>0)
      rownames(cc)[sep.pos] <- rep(paste0(rep(sep, max(nchar(rownames(cc)))),
                                          collapse=""), length(sep.pos))
  }
  print(cc, digits=digits, na.print=na.print, ...)

  if (!is.null(attributes(x)$extra)) {
    cat("\n")
    print(attributes(x)$extra)
  }

  if (!is.null(x$compare)) {
    print(cli::rule(width=min(cli::console_width(),60)))
    cat(x$compare$method[3], "\n")
    cat(paste(" ", x$compare$method[-(1:3)], collapse="\n"), "\n")
    if (length(x$compare$method)>=4) {
      out <- character()
      out <- with(x$compare, c(out, paste(names(statistic),
                                          "=", format(round(statistic, 4)))))
      out <- with(x$compare, c(out, paste(names(parameter),
                                          "=", format(round(parameter, 3)))))
      fp  <- with(x$compare, format.pval(p.value, digits = digits))
      out <- c(out, paste("p-value", if (substr(fp, 1L, 1L) == "<")
                                       fp else paste("=", fp)))
      cat(" ", strwrap(paste(out, collapse = ", ")), sep = "\n")
    }
  }
}

#' @export
vcov.estimate <- function(object, list=FALSE, ...) {
  res <- object$vcov
  nn <- names(coef(object, ...))
  if (list && !is.null(object$model.index)) {
    return(lapply(object$model.index, function(x) object$vcov[x, x]))
  }
  dimnames(res) <- list(nn, nn)
  res
}

#' @export
coef.estimate <- function(object,
                          mat=FALSE,
                          list=FALSE,
                          ...) {
  if (mat) return(object$coefmat)
  if (list && !is.null(object$model.index)) {
    return(lapply(object$model.index, function(x) object$coef[x]))
  }
  object$coef
}

#' @export
transform.estimate <- function(`_data`, ...) {
  estimate(`_data`, ...)
}

#' @export
labels.estimate <- function(object, str, label.width, ...) {
  if (!missing(str)) {
    names(object$coef) <- str
    if (!is.null(object$IC))
      colnames(object$IC) <- str
    if (!is.null(object$vcov))
      colnames(object$vcov) <- rownames(object$vcov) <- str
    rownames(object$coefmat) <- str
  }
  if (!missing(label.width)) {
    rownames(object$coefmat) <- make.unique(
      unlist(lapply(rownames(object$coefmat),
                    function(x) toString(x, width = label.width)))
    )
  }
  return(object)
}

#' @export
parameter.estimate <- function(x, ...) {
  return(x$coefmat)
}

#' @export
IC.estimate <- function(x, ...) {
  if (is.null(x$IC)) return(NULL)
  dimn <- dimnames(x$IC)
  if (!is.null(dimn)) {
    dimn[[2]] <- names(coef(x))
  } else {
    dimn <- list(NULL, names(coef(x)))
  }
  structure(x$IC, dimnames=dimn)
}

#' @export
model.frame.estimate <- function(formula, ...) {
  NULL
}

##' Initialize new latent variable model
##'
##' Function that constructs a new latent variable model object
##'
##' @aliases lvm print.lvm summary.lvm
##' @param x Vector of variable names. Optional but gives control of the
##' sequence of appearance of the variables. The argument can be given as a
##' character vector or formula, e.g. \code{~y1+y2} is equivalent to
##' \code{c("y1","y2")}. Alternatively the argument can be a formula specifying
##' a linear model.
##' @param \dots Additional arguments to be passed to the low level functions
##' @param latent (optional) Latent variables
##' @param messages Controls what messages are printed (0: none)
##' @return Returns an object of class \code{lvm}.
##' @author Klaus K. Holst
##' @seealso \code{\link{regression}}, \code{\link{covariance}},
##' \code{\link{intercept}}, ...
##' @keywords models regression
##' @export
##' @examples
##'
##' m <- lvm() # Empty model
##' m1 <- lvm(y~x) # Simple linear regression
##' m2 <- lvm(~y1+y2) # Model with two independent variables (argument)
##' m3 <- lvm(list(c(y1,y2,y3)~u,u~x+z)) # SEM with three items
##'
lvm <- function(x=NULL, ..., latent=NULL, messages=lava.options()$messages) {

  M <- C <- par <- fix <- numeric(); mu <- list()

  noderender <- list(
                  fill=c(),
                  shape=c(),
                  label=c()
                  )

  edgerender <- list(lty=c(),
                     lwd=c(),
                     col=c(),
                     textCol=c(),
                     est=c(),
                     arrowhead=c(),
                     dir=c(),
                     cex=c(),
                     futureinfo=list())
  graphrender <- list(recipEdges="distinct")

  graphdefault <- list(
                    "fill"="white",
                    "shape"="rectangle",
                    "label"=expression(NA),
                    "lty"=1,
                    "lwd"=1,
                    "col"="black",
                    "textCol"="black",
                    "est"=0,
                    "arrowhead"="open",
                    "dir"="forward",
                    "cex"=1.5,
                    "label"=expression(),
                    "futureinfo"=c())

  modelattr <- list(
      randomslope=list(),
      survival=list(),
      parameter=list(),
      categorical=list(),
      distribution=list(),
      nonlinear=list(),
      functional=list(),
      label=list())

  res <- list(M=M, par=par, cov=C, covpar=C, fix=fix, covfix=fix,latent=list(),
              mean=mu, index=NULL, exogenous=NA,
              constrain=list(), constrainY=list(),
              attributes=modelattr, noderender=noderender,
              edgerender=edgerender, graphrender=graphrender,
              graphdef=graphdefault)
  class(res) <- "lvm"

  myhooks <- gethook("init.hooks")
  for (f in myhooks) {
    res <- do.call(f, list(x=res))
  }

  myvar <- NULL
  if (!is.list(x)) x <- list(x,...)
  for (myvar in x) {
    if (inherits(myvar,"formula")) {
      regression(res,messages=messages) <- myvar
    }
    if (is.character(myvar)) {
      res <- addvar(res, myvar, messages=messages)  }
  }
  if (!is.null(myvar)) {
    index(res) <- reindex(res,zeroones=TRUE) }

  if (!is.null(latent)) {
      latent(res) <- latent
  }

  return(res)
}

impute0 <- function(object,rows,idx,na.action=na.omit,value,...) {
  if (missing(rows) && missing(idx)) {
    df <- na.action(object,...)
    rows <- attr(df,"na.action")
  }
  if (!missing(idx)) {
    obs1 <- setdiff(seq(length(object)),idx)[1]
  } else {
    obs1 <- setdiff(seq(NROW(object)),rows)[1]
  }
  if (missing(value)) {
    fobs <- object[obs1]
    if (is.logical(fobs)) value <- FALSE
    else if (is.character(fobs)) value <- fobs
    else if (is.factor(fobs)) value <- levels(fobs)[1]
    else value <- 0
  }
  if (!missing(idx)) {
    object[idx] <- value
    return(object)
  }
  if (is.matrix(object)) {
    object[rows,] <- value
  } else {
    object[rows] <- value
  }
  return(object)
}

##' @title Handle Missing Values in Objects
##' @description Returns the object with missing data replaced by zeros. This is
##'   sometimes useful for example when working with inverse probability
##'   weighting of the complete-case data.
##' @seealso [na.pass()], [na.omit()], [na.fail()]
##' @export
##' @param object data.frame, vector
##' @param na.action function used to identify missing values
##' @param row.wise when true entire row is set to zero whenever a missing data
##'   is observed in that row
##' @param value value to replace NAs with (default 0)
##' @param ... additional arguments to lower level functions
##' @examples
##' d <- data.frame(y=c(1,1,NA,2,NA,2), r=c(1,1,0,1,1,1))
##' na.pass0(d)
##' glm(y ~ 1, weights=d$r, data=d, na.action=na.pass0)
na.pass0 <- function(object,
                     na.action = na.omit,
                     row.wise = FALSE,
                     value = 0, ...) {
  if (!row.wise && NCOL(object)>1L) {
    nas <- is.na(object)
    if (!any(nas)) return(object)
    idx <- which(nas)
    ii <- arrayInd(idx, dim(object))
    if (length(ii) > 0) {
      object[ii] <- value
    }
    return(structure(object,na.action=structure(idx,class="pass0")))
  }
  df <- na.action(object,...)
  idx <- attr(df, "na.action")
  if (is.null(idx)) return(object)
  if (is.matrix(object) || is.vector(object)) {
    object <- impute0(object,idx=idx,value=value,...)
  } else {
    for (i in seq_len(NCOL(object))) {
      object[[i]] <- impute0(object[[i]],idx=idx,value=value,...)
    }
  }
  return(structure(object,na.action=structure(idx,class="pass0")))
}

##' @export
`%++%.lvm` <- function(x,y) merge(x,y)

##' @export
"+.lvm" <- function(x,...) {
  merge(x,...)
}

## ##' @export
## "+.lm" <- function(x,...) {
##   merge(x,...)
## }

##' @export
merge.lvm <- function(x,y,...) {
  objects <- list(x,y,...)
  if (length(objects)<2) return(x)
  m <- objects[[1]]
  for (i in seq(2,length(objects))) {
    m2 <- objects[[i]]
    if (length(latent(m2))>0)
      latent(m) <- latent(m2)
    if (length(m2$constrain)>0)
      m$constrain <- c(m$constrain,m2$constrain)
    M <- (index(m2)$M)
    P <- (index(m2)$P)
    nn <- vars(m2)
    for (j in seq_len(nrow(M))) {
      if (any(idx <- M[j,]!=0)) {
        val <- as.list(rep(NA,sum(idx==TRUE)))
        if (any(idx. <- !is.na(m2$par[j,idx])))
          val[idx.] <- m2$par[j,idx][idx.]
        if (any(idx. <- !is.na(m2$fix[j,idx])))
          val[idx.] <- m2$fix[j,idx][idx.]
        regression(m,to=nn[idx],from=nn[j],messages=0) <- val
      }

      P0 <- P[j,]; P0[seq_len(j-1)] <- 0
      idx <- P[j,]!=0 | m2$covfix[j,]==0
      idx[is.na(idx)] <- FALSE
      if (any(idx)) {
          val <- as.list(rep(NA,sum(idx==TRUE)))
          if (any(idx. <- !is.na(m2$covpar[j,idx])))
            val[idx.] <- m2$covpar[j,idx][idx.]
          if (any(idx. <- !is.na(m2$covfix[j,idx])))
            val[idx.] <- m2$covfix[j,idx][idx.]
          covariance(m,nn[idx],nn[j],messages=0) <- val
      }
    }
    intercept(m,nn) <- intercept(m2)
    m2x <- exogenous(m2)
    if (length(m2x)>0)
      exogenous(m) <- c(exogenous(m),m2x)
  }
  index(m) <- reindex(m)
  return(m)
}


##' @export
"+.estimate" <- function(x,...) {
  merge(x, ..., paired=TRUE)
}

##' @export
merge.estimate <- function(x,y,...,id,paired=FALSE,labels=NULL,keep=NULL,subset=NULL) {
    objects <- list(x, estimate(y), ...)
    if (length(nai <- names(objects)=="NA")>0)
    names(objects)[which(nai)] <- ""
    if (!missing(subset)) {
        coefs <- unlist(lapply(objects, function(x) coef(x)[subset]))
    } else {
        coefs <- unlist(lapply(objects,coef))
    }
    if (!is.null(labels)) {
        names(coefs) <- labels
    } else {
        names(coefs) <- make.unique(names(coefs))
    }
    if (!missing(id) && is.null(id)) { ## Independence between datasets in x,y,...
        nn <- unlist(lapply(objects,function(x) nrow(x$iid)))
        cnn <- c(0,cumsum(nn))
        id <- list()
        for (i in seq_along(nn)) id <- c(id,list(seq(nn[i])+cnn[i]))
    }
    if (missing(id)) {
        if (paired) { ## One-to-one dependence between observations in x,y,...
            id <- rep(list(seq(nrow(x$iid))),length(objects))
        } else {
            id <- lapply(objects,function(x) x$id)
        }
    } else {
        nn <- unlist(lapply(objects,function(x) NROW(iid(x))))
        if (length(id)==1 && is.logical(id)) {
            if (id) {
                if (any(nn[1]!=nn)) stop("Expected objects of the same size: ", paste(nn,collapse=","))
                id0 <- seq(nn[1]); id <- c()
                for (i in seq(length(nn))) id <- c(id,list(id0))
            } else {
                id <- c()
                N <- cumsum(c(0,nn))
                for (i in seq(length(nn))) id <- c(id,list(seq(nn[i])+N[i]))
            }
        }
        if (length(id)!=length(objects)) stop("Same number of id-elements as model objects expected")
        idlen <- unlist(lapply(id,length))
        if (!identical(idlen,nn)) stop("Wrong lengths of 'id': ", paste(idlen,collapse=","), "; ", paste(nn,collapse=","))
    }
    ##if (any(unlist(lapply(id,is.null)))) stop("Id needed for each model object")
    ##iid <- Reduce("cbind",lapply(objects,iid))
    ids <- iidall <- c(); count <- 0
    for (z in objects) {
        count <- count+1
        clidx <- NULL
        id0 <- id[[count]]
        iidz <- iid(z)
        if (is.null(id0)) {
            id0 <- rownames(iidz)
            if (is.null(id0)) stop("Need id for object number ", count)
        }
        if (!missing(subset)) iidz <- iidz[,subset,drop=FALSE]
        if (!lava.options()$cluster.index) {
            iid0 <- matrix(unlist(by(iidz,id0,colSums)),byrow=TRUE,ncol=ncol(iidz))
            ids <- c(ids, list(sort(unique(id0))))

        } else {
            if (!requireNamespace("mets",quietly=TRUE)) stop("'mets' package required")
            clidx <- mets::cluster.index(id0,mat=iidz,return.all=TRUE)
            iid0 <- clidx$X
            ids <- c(ids, list(id0[as.vector(clidx$firstclustid)+1]))
        }
        iidall <- c(iidall, list(iid0))
    }
    id <- unique(unlist(ids))
    iid0 <- matrix(0,nrow=length(id),ncol=length(coefs))
    model.index <- c()
    colpos <- 0
    for (i in seq(length(objects))) {
        relpos <- seq_along(coef(objects[[i]]))        
        if (!missing(subset)) relpos <- seq_along(subset)
        iid0[match(ids[[i]],id),relpos+colpos] <- iidall[[i]]
        model.index <- c(model.index,list(relpos+colpos))
        colpos <- colpos+tail(relpos,1)
    }
    rownames(iid0) <- id
    res <- estimate.default(NULL, coef=coefs, stack=FALSE, data=NULL, iid=iid0, id=id, keep=keep)
    res$model.index <- model.index
    return(res)
}


##' @export
merge.lm <- function(x,y,...) {
    args <- c(list(x,y),list(...))
    nn <- names(formals(merge.estimate)[-seq(3)])
    idx <- na.omit(match(nn,names(args)))
    models <- args; models[idx] <- NULL
    mm <- lapply(args,function(x) tryCatch(estimate(x),error=function(e) NULL))
    names(mm)[1:2] <- c("x","y")
    ii <- which(unlist(lapply(mm,is.null)))
    if (length(ii)>0) mm[ii] <- NULL
    do.call(merge,c(mm,args[idx]))
}

##' @export
merge.glm <- merge.lm

##' @export
merge.lvmfit <- merge.lm

##' @export
merge.multinomial <- function(x,...) {
    merge.estimate(x,...)
}

normal.threshold <- function(object,p=coef(object),...) {
    M <- moments(object,p=p)
    ord <- ordinal(Model(object))
    K <- attributes(ord)$K
    cK <- c(0,cumsum(K-1))
    breaks.orig <- list()
    for (i in seq(K)) {
        breaks.orig <- c(breaks.orig,list(M$e[seq(K[i]-1)+cK[i]]))
    }
    breaks <- lapply(breaks.orig, ordreg_threshold)
    names(breaks) <- names(K)
    ii <- match(names(K),vars(object))
    sigma <- M$Cfull[ii,ii]
    list(breaks=breaks,sigma=sigma,mean=M$v[ii],K=K)
}

prob.normal <- function(sigma,breaks,breaks2=breaks) {
    if (!requireNamespace("mets",quietly=TRUE)) stop("'mets' package required")
    if (ncol(sigma)!=2 || missing(breaks)) stop("Wrong input")
    P <- matrix(ncol=length(breaks2)-1, nrow=length(breaks)-1)
    for (i in seq(length(breaks)-1))
        for (j in seq(length(breaks2)-1))
            P[i,j] <- mets::pmvn(lower=c(breaks[i],breaks2[j]),upper=c(breaks[i+1],breaks2[j+1]),sigma=sigma)
    return(P)
}

assoc <- function(P,sigma,breaks,...) {
    if (missing(P)) P <- prob.normal(sigma,breaks,...)
    Agree <- sum(diag(P))
    marg.row <- rowSums(P)
    marg.col <- colSums(P)
    Chance <- sum(marg.row*marg.col)
    kap <- (Agree-Chance)/(1-Chance)
    gam <- goodmankruskal_gamma(P)$gamma
    inf <- information_assoc(P)
    res <- c(list(kappa=kap,gamma=gam),inf)
    if (!missing(sigma)) res <- c(res,rho=sigma[1,2])
    return(res)
}



##################################################
### Risk comparison
##################################################

## or:= riskcomp(x,scale=odds) // OR
##' @export
riskcomp <- function(x,...,scale,op="/",type=1,struct=FALSE) {
    val <- c(x,unlist(list(...)))
    if (!missing(scale)) val <- do.call(scale,list(val))
    if (!struct && length(val)==2) {
        if (type==2) {
            return(do.call(op,list(val[2],val[1])))
        } else if (type==1) {
            return(do.call(op,list(val[1],val[2])))
        }
        return(c(do.call(op,list(val[2],val[1])),
                 do.call(op,list(val[1],val[2]))))
    }
    outer(val,val,op)
    offdiag(outer(val,val,op) ,type=type)
}

##' @export
Ratio <- function(x,...) riskcomp(x,...,op="/")

##' @export
Diff <- function(x,...) riskcomp(x,...,op="-")


##################################################
## Odds ratio
##################################################

##' @export
odds <- function(x) x/(1-x)

logor <- function(x) {
    c(log(prod(diag(x))/prod(revdiag(x))),sum(1/x)^.5)
}


##' @export
OR <- function(x,tabulate=FALSE,log=FALSE,...) {
    if (!inherits(x,c("multinomial","table"))) {
        val <- riskcomp(x,...,scale=odds)
        if (log) val <- base::log(val)
        return(val)
    }
    if (inherits(x,"multinomial")) {
        M <- x
    } else {
        M <- multinomial(x)
    }
    pos <- M$position
    if (ncol(pos)!=2 & ncol(pos)!=2) stop("Only for 2x2 tables")
    orfun <- function(p,...) {
        list(logOR=sum(log(p[diag(pos)]))-sum(log(p[revdiag(pos)])))
    }
    estimate(M,orfun,back.transform=exp)
}



##################################################
## Information theoretical measures
##################################################


information_assoc <- function(P,base=exp(1),...) {
    P.row <- rowSums(P)
    P.col <- colSums(P)
    H.row <- H.col <- H <- 0
    for (j in seq_along(P.col))
        if (P.col[j]>0) H.col <- H.col - P.col[j]*log(P.col[j]+(P.col[j]==0),base=base)
    for (i in seq_along(P.row)) {
        if (P.row[i]>0) H.row <- H.row - P.row[i]*log(P.row[i]+(P.row[i]==0),base=base)
        for (j in seq_along(P.col)) {
            if (P[i,j]>0) H <- H - P[i,j]*log(P[i,j],base=base)
        }
    }
    I <- H.row+H.col-H
    return(list(MI=I,H=H,H.row=H.row,H.col=H.col,
                U.row=I/H.row,U.col=I/H.col,U.sym=2*I/(H.row+H.col)))
}


##' @export
information.data.frame <- function(x,...) {
    information(multinomial(x,marginal=TRUE),...)
}

##' @export
information.table <- function(x,...) {
    information(multinomial(x,marginal=TRUE),...)
}

##' @export
information.multinomial <- function(x,...) {
    estimate(x,function(p,object,...) {
        P <- object$position; P[] <- p[object$position]
        information_assoc(P)},...)
}


##################################################
## Independence tests
##################################################

independence <- function(x,...) {
    if (is.table(x) || is.data.frame(x) || is.matrix(x)) {
        x <- multinomial(x)
    }
    if (!inherits(x,"multinomial")) stop("Expected table, data.frame or multinomial object")
    if (length(x$levels)!=2) stop("Data from two categorical variables expected")
    f <- function(p) {
        P <- x$position; P[] <- p[x$position]
        n <- nrow(x$IC)
        k1 <- length(x$levels[[1]])
        k2 <- length(x$levels[[2]])
        A1 <- matrix(0,ncol=length(p),nrow=k1)
        for (i in seq(k1)) A1[i,x$position[i,]] <- 1
        A2 <- matrix(0,ncol=length(p),nrow=k2)
        for (i in seq(k2)) A2[i,x$position[,i]] <- 1
        P1 <- A1%*%p
        P2 <- A2%*%p
        I <- P1%*%t(P2)
        Q <- P-I
        #        Q <- sum(n*P*(log(I[1,1])-P1
        sum((P-I)^2)
        ##V <- sqrt(sum((P*n-I*n)^2/I/n) /(n*(min(k1,k2)-1)))
        V <- sqrt(sum((P-I)^2/I)   / ((min(k1,k2)-1)))
        return(V)
        sum(n*Q^2/I)^0.25
        return((sum((P-I)^2))^.5)
        ##        V
    }

    ## M <- P*n
    ## O2 <- colSums(M)
    ## O1 <- rowSums(M)
    ## M[1,1]-O1[1]*O2[1]/200
    ## M[2,2]-O1[2]*O2[2]/200
    ## sum((M-I*n)^2/(I*n))
    ## sum((P*n-I*n)^2/I/n)
    ## sum(Q)
    ## sum(Q^2)
    ## M <- P
    ## chisq.test(M,correct=FALSE)

    return(estimate(x,function(p) list(cramersV=f(p)),IC=TRUE,...))


    e <- estimate(x,f,IC=TRUE,print=function(x,...) {
        cat("\tTest for independence\n\n")
        cat("Test statistc:\t ", formatC(x$coefmat[1]/x$coefmat[2]),
            "\nP-value:\t ", x$coefmat[5],"\n\n")
        print(estimate(x))
    },...)
    return(list(p.value=e$coefmat[5]))
    ## Q <- sum((a$coefmat[1,1]/a$coefmat[1,2]))
    ## df <- nrow(a$coefmat)
    ## res <- list(##data.name=hypothesis,
    ##             statistic = Q, parameter = df,
    ##             p.value=pchisq(Q,df=1,lower.tail=FALSE),
    ##             method = "Test for independence")
    ## class(res) <- "htest"
    ## res
}

## independence(x)
## chisq.test(table(dd))

context("Inference")

test_that("Effects",{
    m <- lvm()
    regression(m) <- c(y1,y2,y3)~u; latent(m) <- ~u
    regression(m) <- c(z1,z2,z3)~v; latent(m) <- ~v
    regression(m) <- u~v
    regression(m) <- c(u,v,z3,y1)~x
    d <- sim(m,100,seed=1)
    start <- c(rep(0,6),rep(1,17))
    suppressWarnings(e <- estimate(m,d,control=list(iter.max=0,start=start)))
    f <- summary(ef <- effects(e,y1~x))$coef
    testthat::expect_true(all(f[,2]>0)) ## Std.err
    testthat::expect_equal(f["Total",1],3)
    testthat::expect_equal(f["Direct",1],1)
    f2 <- summary(effects(e,u~v))$coef
    testthat::expect_equal(f2["Total",1],1)
    testthat::expect_equal(f2["Direct",1],1)
    testthat::expect_equal(f2["Indirect",1],0)

    testthat::expect_output(print(ef),"Mediation proportion")
    testthat::expect_equivalent(confint(ef)["Direct",],
                                confint(e)["y1~x",])

    testthat::expect_equivalent(totaleffects(e,y1~x)[,1:4],f["Total",])

    ##g <- graph::updateGraph(plot(m,noplot=TRUE))
    ##testthat::expect_equivalent(path(g,y1~x),path(m,y1~x))
})

test_that("Profile confidence limits", {
    m <- lvm(y~b*x)
    constrain(m,b~psi) <- identity
    set.seed(1)
    d <- sim(m,100,seed=1)
    e <- estimate(m, d)
    ci0 <- confint(e,3)
    ci <- confint(e,3,profile=TRUE)
    testthat::expect_true(mean((ci0-ci)^2)<0.1)
})

test_that("IV-estimator", {
    m <- lvm(c(y1,y2,y3)~u); latent(m) <- ~u
    set.seed(1)
    d <- sim(m,100,seed=1)
    e0 <- estimate(m,d)
    e <- estimate(m,d,estimator="iv") ## := MLE
    testthat::expect_true(mean((coef(e)-coef(e0))^2)<1e-9)
})

test_that("glm-estimator", {
    m <- lvm(y~x+z)
    regression(m) <- x~z
    distribution(m,~y+z) <- binomial.lvm("logit")
    set.seed(1)
    d <- sim(m,1e3,seed=1)
    e <- estimate(m,d,estimator="glm")
    c1 <- coef(e,2)[c("y","y~x","y~z"),1:2]
    c2 <- estimate(glm(y~x+z,d,family=binomial))$coefmat[,1:2]
    testthat::expect_equivalent(c1,c2)
})

test_that("gaussian", {
    m <- lvm(y~x)
    d <- simulate(m,100,seed=1)
    S <- cov(d[,vars(m),drop=FALSE])
    mu <- colMeans(d[,vars(m),drop=FALSE])
    f <- function(p) lava:::gaussian_objective.lvm(p,x=m,S=S,mu=mu,n=nrow(d))
    g <- function(p) lava:::gaussian_score.lvm(p,x=m,n=nrow(d),data=d,indiv=TRUE)
    s1 <- numDeriv::grad(f,c(0,1,1))
    s2 <- g(c(0,1,1))
    testthat::expect_equal(s1,-colSums(s2),tolerance=0.1)

    e <- estimate(m, d, estimator="gaussian")
    e0 <- estimate(m, d, estimator="gaussian0") # numerical grad.
    expect_true(mean((coef(e)- coef(e0))^2)<1e-9)
    expect_equivalent(logLik(e), logLik(e0))
    e1 <- estimate(m, d, estimator="gaussian1") # numerical hessian
    expect_equivalent(logLik(e), logLik(e1))
    e2 <- estimate(m, d, estimator="gaussian2") # numerical hessian
    expect_equivalent(logLik(e), logLik(e2)) # BHHH
})

if (requireNamespace("mets",quietly = TRUE))
test_that("Association measures", {
    P <- matrix(c(0.25,0.25,0.25,0.25),2)
    a1 <- lava:::assoc(P)
    testthat::expect_equivalent(-log(0.25),a1$H)
    testthat::expect_true(with(a1, all(c(kappa,gamma,MI,U.sym)==0)))

    p <- lava:::prob.normal(sigma=diag(nrow=2),breaks=c(-Inf,0),breaks2=c(-Inf,0))[1]
    testthat::expect_equal(p[1],0.25)
    ## q <- qnorm(0.75)
    ## m <- ordinal(lvm(y~x),~y, K=3)#, breaks=c(-q,q))
    ## normal.threshold(m,p=c(0,1,2))
})

test_that("equivalence", {
    m <- lvm(c(y1,y2,y3)~u,u~x,y1~x)
    latent(m) <- ~u
    d <- sim(m,100,seed=12)
    cancel(m) <- y1~x
    regression(m) <- y2~x
    e <- estimate(m,d)
    ##eq <- equivalence(e,y1~x,k=1)
    dm <- capture.output(eq <- equivalence(e,y2~x,k=1))
    testthat::expect_output(print(eq),paste0("y1",lava.options()$symbol[2],"y3"))
    testthat::expect_true(all(c("y1","y3")%in%eq$equiv[[1]][1,]))
})

test_that("multiple testing", {
    testthat::expect_equivalent(lava:::holm(c(0.05/3,0.025,0.05)),rep(0.05,3))

    ci1 <- scheffe(l <- lm(1:5~c(0.5,0.7,1,1.3,1.5)))
    ci2 <- predict(l,interval="confidence")
    testthat::expect_equivalent(ci1[,1],ci2[,1])
    testthat::expect_true(all(ci1[,2]<ci2[,2]))
    testthat::expect_true(all(ci1[,3]>ci2[,3]))
})

test_that("modelsearch and GoF", {
    m <- lvm(c(y1,y2)~x)
    d <- sim(m,100,seed=1)
    e <- estimate(lvm(c(y1,y2)~1,y1~x),d)
    e0 <- estimate(lvm(c(y1,y2)~x,y1~~y2),d)

    s1 <- modelsearch(e,silent=TRUE,type="correlation")
    testthat::expect_true(nrow(s1$res)==2)
    s1b <- modelsearch(e,silent=TRUE,type="regression")
    testthat::expect_true(nrow(s1b$res)==4)
    s2 <- modelsearch(e0,silent=TRUE,dir="backward")
    testthat::expect_true(nrow(s2$res)==3)
    e00 <- estimate(e0,vcov=vcov(e0))$coefmat
    ii <- match(s2$res[,"Index"],rownames(e00))
    testthat::expect_equivalent(e00[ii,5],s2$test[,2])

    s3 <- modelsearch(e0,dir="backward",k=3)
    testthat::expect_true(nrow(s3$res)==1)

    ee <- modelsearch(e0,dir="backstep",messages=FALSE)
    testthat::expect_true(inherits(ee,"lvm"))

    ## TODO
    g <- gof(e,all=TRUE)
    r <- rsq(e)[[1]]
    testthat::expect_true(abs(summary(lm(y1~x,d))$r.square-r["y1"])<1e-5)
})

test_that("Bootstrap", {
    y <- rep(c(0,1),each=5)
    x <- 1:10
    e <- estimate(y~x,lvm=TRUE)
    B1 <- bootstrap(e,R=2,silent=TRUE,mc.cores=1,sd=TRUE)
    B2 <- bootstrap(e,R=2,silent=TRUE,bollenstine=TRUE,mc.cores=1)
    testthat::expect_false(B1$bollenstine)
    testthat::expect_true(B2$bollenstine)
    testthat::expect_true(nrow(B1$coef)==2)
    testthat::expect_output(print(B1),"Standard errors:")
    # if (requireNamespace("future",quietly=TRUE)) future::plan("sequential")
    dm <- capture.output(B3 <- bootstrap(e,R=2,fun=function(x) coef(x)[2]^2+10))
    testthat::expect_true(all(mean(B3$coef)>10))

    y <- rep(c(0,1),each=5)
    x <- 1:10
    m <- lvm(y~b*x)
    constrain(m,alpha~b) <- function(x) x^2
    e <- estimate(m,data.frame(y=y,x=x))
    b <- bootstrap(e,R=1,silent=TRUE)
    testthat::expect_output(print(b),"alpha")
})

test_that("Survreg", {
    m <- lvm(y0~x)
    transform(m,y~y0) <- function(x) pmin(x[,1],2)
    transform(m,status~y0) <- function(x) x<2
    d <- simulate(m,100,seed=1)
    require('survival')
    m <- survreg(Surv(y,status)~x,data=d,dist='gaussian')
    s <- score(m)
    testthat::expect_true(length(pars(m))==length(coef(m))+1)
    testthat::expect_true(abs(attr(score(m,pars(m)),'logLik')-logLik(m))<1e-9)
    testthat::expect_true(mean(colSums(s)^2)<1e-6)
    testthat::expect_equivalent(vcov(m), attr(s,'bread')/nrow(d))
})

test_that("diagtest", {
  M <- as.table(matrix(c(42,12,
                         35,28),ncol=2,byrow=TRUE,
                       dimnames=list(rater=c("no","yes"),gold=c("no","yes"))))
  d1 <- diagtest(M,exact=TRUE)
  d2 <- diagtest(M,exact=FALSE)
  e1 <- d1$coefmat[1:7,]
  e2 <- d2$coefmat[1:7,]
  expect_equivalent(e1[,1], e2[,1])
  expect_true(e1[1,1] == (sum(M[,2])/sum(M)))
})

test_that("Combine", { ## Combine model output
    data(serotonin)
    m1 <- lm(cau ~ age*gene1 + age*gene2,data=serotonin)
    m2 <- lm(cau ~ age + gene1,data=serotonin)

    cc <- Combine(list('model A'=m1,'model B'=m2),fun=function(x) c(R2=format(summary(x)$r.squared,digits=2)))
    testthat::expect_true(nrow(cc)==length(coef(m1))+1)
    testthat::expect_equivalent(colnames(cc),c('model A','model B'))
    testthat::expect_equivalent(cc['R2',2],format(summary(m2)$r.squared,digits=2))

})

test_that("zero-inflated binomial regression (zib)", {
    set.seed(1)
    n <- 1e3
    x <- runif(n,0,20)
    age <- runif(n,10,30)
    z0 <- rnorm(n,mean=-1+0.05*age)
    z <- cut(z0,breaks=c(-Inf,-1,0,1,Inf))
    p0 <- lava::expit(model.matrix(~z+age) %*% c(-.4, -.4, 0.2, 2, -0.05))
    y <- (runif(n)<lava:::tigol(-1+0.25*x-0*age))*1
    u <- runif(n)<p0
    y[u==0] <- 0
    d <- data.frame(y=y,x=x,u=u*1,z=z,age=age)

    ## Estimation
    e0 <- zibreg(y~x*z,~1+z+age,data=d)
    e <- zibreg(y~x,~1+z+age,data=d)
    lrt <- compare(e,e0)
    testthat::expect_equivalent(score(e,data=d),
                      colSums(score(e,indiv=TRUE)))
    testthat::expect_equivalent(logLik(e,data=d),
                      logLik(e))
    testthat::expect_equivalent(vcov(e), Inverse(information(e,type="obs",data=d)))

    testthat::expect_output(print(e), "Prevalence probabilities:")

    pdval <- PD(e0,intercept=c(1,3),slope=c(2,6))

    B <- rbind(c(1,0,0,0,20),
               c(1,1,0,0,20),
               c(1,0,1,0,20),
               c(1,0,0,1,20))
    prev <- summary(e,pr.contrast=B)$prevalence

    x <- seq(0,100,length.out=100)
    newdata <- expand.grid(x=x,age=20,z=levels(d$z))
    fit <- predict(e,newdata=newdata)
})

if (requireNamespace("nlme",quietly = TRUE) && requireNamespace("mets",quietly = TRUE))
test_that("Prediction with missing data, random intercept", {
    ## Random intercept model
    m <- lvm(c(y1,y2,y3)~u[0])
    latent(m) <- ~u
    regression(m) <- y1~x1
    regression(m) <- y2~x2
    regression(m) <- y3~x3

    d <- simulate(m,1e3,seed=1)
    dd <- reshape(d,varying=list(c('y1','y2','y3'),c('x1','x2','x3')),direction='long',v.names=c('y','x'))

    ##system.time(l <- lme4::lmer(y~x+(1|id), data=dd, REML=FALSE))
    system.time(l <- nlme::lme(y~x,random=~1|id, data=dd, method="ML"))
    m0 <- lvm(c(y1[m:v],y2[m:v],y3[m:v])~1*u[0])
    latent(m0) <- ~u
    regression(m0,y=c('y1','y2','y3'),x=c('x1','x2','x3')) <- rep('b',3)
    system.time(e <- estimate(m0,d))

    mytol <- 1e-6
    mse <- function(x,y=0) mean(na.omit(as.matrix(x)-as.matrix(y))^2)
    testthat::expect_true(mse(logLik(e),logLik(l))<mytol)
    testthat::expect_true(mse(nlme::fixef(l),coef(e)[1:2])<mytol)
    u1 <- nlme::ranef(l)##[[1]][,1]
    u2 <- predict(e,endogenous(e))
    testthat::expect_true(mse(u1,u2)<1e-9)

    ## Missing data
    idx <- sample(seq(nrow(dd)),nrow(dd)*0.5)
    dd0 <- dd[idx,,drop=FALSE]
    d0 <- mets::fast.reshape(subset(dd0,select=-u),id='id',num='time')

    system.time(e0 <- estimate(m0,d0,missing=TRUE))
    ##system.time(l0 <- lme4::lmer(y~x+(1|id), data=dd0, REML=FALSE))
    system.time(l0 <- nlme::lme(y~x,random=~1|id, data=dd0, method="ML"))
    testthat::expect_true(mse(logLik(e0),logLik(l0))<mytol)
    testthat::expect_true(mse(nlme::fixef(l0),coef(e0)[1:2])<mytol)
    u01 <- nlme::ranef(l0)##[[1]][,1]
    u02 <- predict(e0,endogenous(e0))
    expect_true(mse(u01,u02)<1e-9)

    s <- summary(e0)
    testthat::expect_output(print(s),paste0("data=",nrow(d0)))
    testthat::expect_true(inherits(e0$estimate,"multigroupfit"))
    testthat::expect_output(print(e0$estimate),"Group 1")
    testthat::expect_output(print(summary(e0$estimate)),paste0("observations = ",nrow(d0)))
})

## if (requireNamespace("lme4", quietly = TRUE) && requireNamespace("mets", quietly = TRUE)) {
if (requireNamespace("mets",quietly = TRUE))
test_that("Random slope model", {
    set.seed(1)
    m <- lvm()
    regression(m) <- y1 ~ 1*u + 1*s
    regression(m) <- y2 ~ 1*u + 2*s
    regression(m) <- y3 ~ 1*u + 3*s
    latent(m) <- ~u+s
    dw <- sim(m,20)

    dd <- mets::fast.reshape(dw)
    dd$num <- dd$num+runif(nrow(dd),0,0.2)
    dd0 <- dd[-c(1:2*3),]
    ##l <- lme4::lmer(y~ 1+num +(1+num|id),dd,REML=FALSE)
    l <- nlme::lme(y ~ 1+num, random=~1+num|id, data=dd, method="ML")
    sl <- lava:::varcomp(l)

    d <- mets::fast.reshape(dd,id="id")
    d0 <- mets::fast.reshape(dd0,id="id")

    m0 <- lvm(c(y1[0:v],y2[0:v],y3[0:v])~1*u)
    addvar(m0) <- ~num1+num2+num3
    covariance(m0) <- u~s
    latent(m0) <- ~s+u
    regression(m0) <- y1 ~ num1*s
    regression(m0) <- y2 ~ num2*s
    regression(m0) <- y3 ~ num3*s
    system.time(e <- estimate(m0,d,param="none",control=list(trace=0,constrain=FALSE)))

    testthat::expect_true(mean(sl$coef-coef(e)[c("u","s")])^2<1e-5)
    testthat::expect_true((logLik(l)-logLik(e))^2<1e-5)
    varcomp.nam <- c(paste0("u",lava.options()$symbol[2],"u"),
                paste0("s",lava.options()$symbol[2],"s"))
    testthat::expect_true(mean(diag(sl$varcomp)-coef(e)[varcomp.nam])^2<1e-5)

    ## missing
    testthat::expect_output(e0 <- estimate(m0,d0,missing=TRUE,param="none",
                                           control=list(method="NR",constrain=FALSE,start=coef(e),trace=1)),
                            "Iter=")
    ## l0 <- lmer(y ~ 1 + num + (1 + num | id), dd0, REML = FALSE)
    l0 <- nlme::lme(y~ 1+num, random=~1+num|id, data=dd0, method="ML")
    testthat::expect_true((logLik(l0)-logLik(e0))^2<1e-5)

    m1 <- lvm(c(y1[0:v],y2[0:v],y3[0:v])~1*u)
    addvar(m1) <- ~num1+num2+num3
    covariance(m1) <- u~s
    latent(m1) <- ~s+u
    regression(m1) <- y1 ~ b1*s
    regression(m1) <- y2 ~ b2*s
    regression(m1) <- y3 ~ b3*s
    constrain(m1,b1~num1) <- function(x) x
    constrain(m1,b2~num2) <- function(x) x
    constrain(m1,b3~num3) <- function(x) x
    system.time(e1 <- estimate(m1,d,param="none",p=coef(e)))
    testthat::expect_true((logLik(e1)-logLik(e))^2<1e-5)

    ## TODO
    ## missing
    ## system.time(e10 <- estimate(m1,d0,missing=TRUE,param="none",
    ##                             control=list(trace=coef(e0))))

})


test_that("Predictions, jacobians", {
    m <- lvm(c(x1,x2,x3)~u1,u1~z,
             c(y1,y2,y3)~u2,u2~u1+z)
    latent(m) <- ~u1+u2

    p <- c("u2,u2"=2,"u1,u1"=0.5)
    names(p) <- gsub(",",lava.options()$symbols[2],names(p))
    d <- simulate(m,50,p=p,seed=123)
    e <- estimate(m,d)

    object <- e
    f <- function(p,x=vars(object)) predict(object,x,p=p)
    testthat::expect_true(sum(abs(numDeriv::jacobian(f,coef(object))-predictlvm(object)$mean.jacobian))
                          <1e-6)
    testthat::expect_true(sum(abs(numDeriv::jacobian(function(p) predictlvm(object,p=p)$var,coef(object))-predictlvm(object)$var.jacobian))
                          <1e-6)

    testthat::expect_true(sum(abs(numDeriv::jacobian(function(p) f(p,x1~1),coef(object))-
                                  predictlvm(object,x1~1)$mean.jacobian))
                          <1e-6)

    testthat::expect_true(sum(abs(numDeriv::jacobian(function(p) f(p,u1~x1+x2+x3),coef(object))-
                                  predictlvm(object,u1~x1+x2+x3)$mean.jacobian))
                          <1e-6)
})


if (requireNamespace("mets",quietly = TRUE))
test_that("multinomial", {
    set.seed(1)
    breaks <- c(-Inf,-1,0,Inf)
    m <- lvm(); covariance(m,pairwise=TRUE) <- ~y1+y2+y3+y4
    d <- transform(sim(m,5e2),
                   z1=cut(y1,breaks=breaks),
                   z2=cut(y2,breaks=breaks),
                   z3=cut(y3,breaks=breaks),
                   z4=cut(y4,breaks=breaks))

    m <- multinomial(d[,5])
    lev <- levels(d[,5])
    e <- estimate(l <- lm(d[,5]==lev[1]~1))
    testthat::expect_true(abs(vcov(e)[1]-vcov(m)[1])<1e-9)

    a1 <- multinomial(d[,5:6],marginal=TRUE)
    K1 <- kappa(a1) ## Cohen's kappa
    P <- a1$P
    marg1 <- rowSums(P)
    marg2 <- colSums(P)
    testthat::expect_equivalent(K1$coef,sum(diag(P)-marg1*marg2)/(1-sum(marg1*marg2)))

    K2 <- kappa(d[,7:8])
    ## Testing difference K1-K2:
    e1 <- estimate(merge(K1,K2,id=TRUE),diff)
    e2 <- estimate(merge(K1,K2,id=NULL),diff)
    testthat::expect_true(vcov(e1)!=vcov(e2))
    testthat::expect_equivalent(vcov(e2),(vcov(K1)+vcov(K2)))

    g1 <- gkgamma(d[,5:6])
    g2 <- gkgamma(table(d[,5:6]))
    g3 <- gkgamma(multinomial(d[,5:6]))
    testthat::expect_equivalent(g1$coefmat,g2$coefmat)
    testthat::expect_equivalent(g3$coefmat,g2$coefmat)


    ## TODO
    indep <- lava:::independence(d[,5:6])
    inf <- information(d[,5:6])

    ## pcor
    if (requireNamespace("polycor",quietly=TRUE)) {
        system.time(rho <- pcor(d[,5],d[,6]))
        rho2 <- polycor::polychor(d[,5],d[,6],ML=TRUE,std.err=TRUE)
        testthat::expect_true(abs(rho$coef[1]-rho2$rho)^2<1e-5)
        testthat::expect_true(abs(rho$vcov[1]-rho2$var[1])^2<1e-5)
        testthat::expect_true(mean(score(rho))^2<1e-5)
    }
})


test_that("predict,residuals", {
    m <- lvm(c(y1,y2,y3)~u,u~x)
    latent(m) <- ~u
    set.seed(1)
    d <- sim(m,100,'y1~u'=1,'y3~u'=3)
    e <- estimate(m,d)

    l <- lm(y3~x,data=d)
    e3 <- estimate(y3~x,data=d,lvm=TRUE)
    testthat::expect_true(mean((residuals(l)-residuals(e3))^2)<1e-12)
    testthat::expect_true(mean(var(residuals(e3,std=TRUE))[1]*99/100-1)<1e-3)

    r <- residuals(e)
    testthat::expect_true(ncol(r)==3)
})


test_that("partialcor", {
    m <- lvm(c(y1,y2,y3)~x1+x2)
    covariance(m) <- c(y1,y2,y3)~y1+y2+y3
    set.seed(1)
    d <- sim(m,500)
    c1 <- partialcor(~x1+x2,d)
    e <- estimate(m,d)
    ec <- correlation(e)
    c2 <- parameter(ec)
    testthat::expect_true(mean(c1[,1]-c2[,1])^2<1e-9)
    ## CI, note difference var(z)=1/(n-k-3) vs var(z)=1/(n-3)
    testthat::expect_true(mean(c1[,4]-c2[,3])^2<1e-3)
    testthat::expect_true(mean(c1[,5]-c2[,4])^2<1e-3)
})

## test_that("partialgamma", {
## TODO
## })

test_that("correlation", {
  set.seed(1)
  d <- matrix(rnorm(500), ncol=2)
  e1 <- parameter(correlation(d, z=FALSE))
  e2 <- cor.test(d[,1],d[,2])
  expect_equivalent(e2$estimate, e1[1])
  ci <- e1[3:4]
  ci2 <- e2$conf.int
  expect_true(mean((ci-ci2)^2)< sqrt(1/nrow(d))) # df discrepancy
  e1 <- parameter(correlation(d, z=TRUE))
  ci <- e1[3:4]
  ci2 <- atanh(e2$estimate) + qnorm(0.975)*c(-1,1)*1/sqrt(nrow(d))
  expect_true(mean((ci-ci2)^2)< sqrt(1/nrow(d)))

  d <- matrix(rnorm(500), ncol=5)
  e1 <- parameter(correlation(d, z=FALSE))
  e2 <- parameter(correlation(d, z=TRUE))
  e3 <- parameter(correlation(d, z=TRUE, back.transform=FALSE))
  expect_equivalent(e1[,1], e2[,1]) # est same, but not pval
  expect_equivalent(e2[,5], e3[,5]) # p.vals same, but z vs r
})

if (requireNamespace("mets",quietly = TRUE))
test_that("Weighted",{
    m <- lvm(y~x)
    set.seed(1)
    d <- sim(m,10)
    d$w <- runif(nrow(d),0.1,1)
    e <- estimate(m,data=d)
    l <- lm(y~x,data=d)
    testthat::expect_true(mean((coef(e)[1:2]-coef(l))^2)<1e-12)

    w2 <- estimate(m,data=d,weights=d$w,control=list(trace=0))
    w <- estimate(m,data=d,weights=d$w,estimator="normal",control=list(trace=0))
    expect_equivalent(coef(w2), coef(w))
    lw <- lm(y~x,data=d, weights=d$w)
    testthat::expect_true(mean((coef(w)[1:2]-coef(lw))^2)<1e-12)
})


if (requireNamespace("mets",quietly = TRUE))
test_that("Tobit",{
    m0 <- lvm(t~x)
    distribution(m0,~w) <- uniform.lvm(0.1,1)
    d <- sim(m0,10,seed=1)
    d$status <- rep(c(TRUE,FALSE),each=nrow(d)/2)
    d$s <- with(d, survival::Surv(t,status))
    s <- survival::survreg(s~x,data=d,dist="gaussian",weights=d$w)
    m <- lvm(s~x)
    e <- estimate(m,data=d,estimator="normal",weights="w")
    testthat::expect_true(mean((coef(e)[1:2]-coef(s))^2)<1e-9)
})

context("Inference")

test_that("Effects",{
    m <- lvm()
    regression(m) <- c(y1,y2,y3)~u; latent(m) <- ~u
    regression(m) <- c(z1,z2,z3)~v; latent(m) <- ~v
    regression(m) <- u~v
    regression(m) <- c(u,v,z3,y1)~x
    d <- sim(m,100,seed=1)
    start <- c(rep(0,6),rep(1,17))
    suppressWarnings(e <- estimate(m,d,control=list(iter.max=0,start=start)))
    f <- summary(ef <- effects(e,y1~x))$coef
    testthat::expect_true(all(f[,2]>0)) ## Std.err
    testthat::expect_equal(f["Total",1],3)
    testthat::expect_equal(f["Direct",1],1)
    f2 <- summary(effects(e,u~v))$coef
    testthat::expect_equal(f2["Total",1],1)
    testthat::expect_equal(f2["Direct",1],1)
    testthat::expect_equal(f2["Indirect",1],0)

    testthat::expect_output(print(ef),"Mediation proportion")
    testthat::expect_equivalent(confint(ef)["Direct",],
                                confint(e)["y1~x",])

    testthat::expect_equivalent(totaleffects(e,y1~x)[,1:4],f["Total",])

    ##g <- graph::updateGraph(plot(m,noplot=TRUE))
    ##testthat::expect_equivalent(path(g,y1~x),path(m,y1~x))
})

test_that("Profile confidence limits", {
    m <- lvm(y~b*x)
    constrain(m,b~psi) <- identity
    set.seed(1)
    d <- sim(m,100,seed=1)
    e <- estimate(m, d)
    ci0 <- confint(e,3)
    ci <- confint(e,3,profile=TRUE)
    testthat::expect_true(mean((ci0-ci)^2)<0.1)
})

test_that("IV-estimator", {
    m <- lvm(c(y1,y2,y3)~u); latent(m) <- ~u
    set.seed(1)
    d <- sim(m,100,seed=1)
    e0 <- estimate(m,d)
    e <- estimate(m,d,estimator="iv") ## := MLE
    testthat::expect_true(mean((coef(e)-coef(e0))^2)<1e-9)
})

test_that("glm-estimator", {
    m <- lvm(y~x+z)
    regression(m) <- x~z
    distribution(m,~y+z) <- binomial.lvm("logit")
    set.seed(1)
    d <- sim(m,1e3,seed=1)
    e <- estimate(m,d,estimator="glm")
    c1 <- coef(e,2)[c("y","y~x","y~z"),1:2]
    c2 <- estimate(glm(y~x+z,d,family=binomial))$coefmat[,1:2]
    testthat::expect_equivalent(c1,c2)
    testthat::expect_true(mean(colSums(score(e))^2)<1e-9)
    testthat::expect_true(
                mean((colSums(score(e,p=coef(e)+1, indiv=TRUE))-
                score(e,p=coef(e)+1, indiv=FALSE))**2)<1e-9)
})

test_that("gaussian", {
    m <- lvm(y~x)
    d <- simulate(m,100,seed=1)
    S <- cov(d[,vars(m),drop=FALSE])
    mu <- colMeans(d[,vars(m),drop=FALSE])
    f <- function(p) lava:::gaussian_objective.lvm(p,x=m,S=S,mu=mu,n=nrow(d))
    g <- function(p) lava:::gaussian_score.lvm(p,x=m,n=nrow(d),data=d,indiv=TRUE)
    s1 <- numDeriv::grad(f,c(0,1,1))
    s2 <- g(c(0,1,1))
    testthat::expect_equal(s1,-colSums(s2),tolerance=0.1)

    e <- estimate(m, d, estimator="gaussian")
    e0 <- estimate(m, d, estimator="gaussian0") # numerical grad.
    expect_true(mean((coef(e)- coef(e0))^2)<1e-9)
    expect_equivalent(logLik(e), logLik(e0))
    e1 <- estimate(m, d, estimator="gaussian1") # numerical hessian
    expect_equivalent(logLik(e), logLik(e1))
    e2 <- estimate(m, d, estimator="gaussian2") # numerical hessian
    expect_equivalent(logLik(e), logLik(e2)) # BHHH
})

if (requireNamespace("mets",quietly = TRUE))
test_that("Association measures", {
    P <- matrix(c(0.25,0.25,0.25,0.25),2)
    a1 <- lava:::assoc(P)
    testthat::expect_equivalent(-log(0.25),a1$H)
    testthat::expect_true(with(a1, all(c(kappa,gamma,MI,U.sym)==0)))

    p <- lava:::prob.normal(sigma=diag(nrow=2),breaks=c(-Inf,0),breaks2=c(-Inf,0))[1]
    testthat::expect_equal(p[1],0.25)
    ## q <- qnorm(0.75)
    ## m <- ordinal(lvm(y~x),~y, K=3)#, breaks=c(-q,q))
    ## normal.threshold(m,p=c(0,1,2))
})

test_that("equivalence", {
    m <- lvm(c(y1,y2,y3)~u,u~x,y1~x)
    latent(m) <- ~u
    d <- sim(m,100,seed=12)
    cancel(m) <- y1~x
    regression(m) <- y2~x
    e <- estimate(m,d)
    ##eq <- equivalence(e,y1~x,k=1)
    dm <- capture.output(eq <- equivalence(e,y2~x,k=1))
    testthat::expect_output(print(eq),paste0("y1",lava.options()$symbol[2],"y3"))
    testthat::expect_true(all(c("y1","y3")%in%eq$equiv[[1]][1,]))
})

test_that("multiple testing", {
    testthat::expect_equivalent(lava:::holm(c(0.05/3,0.025,0.05)),rep(0.05,3))

    ci1 <- scheffe(l <- lm(1:5~c(0.5,0.7,1,1.3,1.5)))
    ci2 <- predict(l,interval="confidence")
    testthat::expect_equivalent(ci1[,1],ci2[,1])
    testthat::expect_true(all(ci1[,2]<ci2[,2]))
    testthat::expect_true(all(ci1[,3]>ci2[,3]))
})

test_that("modelsearch and GoF", {
    m <- lvm(c(y1,y2)~x)
    d <- sim(m,100,seed=1)
    e <- estimate(lvm(c(y1,y2)~1,y1~x),d)
    e0 <- estimate(lvm(c(y1,y2)~x,y1~~y2),d)

    s1 <- modelsearch(e,silent=TRUE,type="correlation")
    testthat::expect_true(nrow(s1$res)==2)
    s1b <- modelsearch(e,silent=TRUE,type="regression")
    testthat::expect_true(nrow(s1b$res)==4)
    s2 <- modelsearch(e0,silent=TRUE,dir="backward")
    testthat::expect_true(nrow(s2$res)==3)
    e00 <- estimate(e0,vcov=vcov(e0))$coefmat
    ii <- match(s2$res[,"Index"],rownames(e00))
    testthat::expect_equivalent(e00[ii,5],s2$test[,2])

    s3 <- modelsearch(e0,dir="backward",k=3)
    testthat::expect_true(nrow(s3$res)==1)

    ee <- modelsearch(e0,dir="backstep",messages=FALSE)
    testthat::expect_true(inherits(ee,"lvm"))

    ## TODO
    g <- gof(e,all=TRUE)
    r <- rsq(e)[[1]]
    testthat::expect_true(abs(summary(lm(y1~x,d))$r.square-r["y1"])<1e-5)
})

test_that("Bootstrap", {
    y <- rep(c(0,1),each=5)
    x <- 1:10
    e <- estimate(y~x,lvm=TRUE)
    B1 <- bootstrap(e,R=2,silent=TRUE,mc.cores=1,sd=TRUE)
    B2 <- bootstrap(e,R=2,silent=TRUE,bollenstine=TRUE,mc.cores=1)
    testthat::expect_false(B1$bollenstine)
    testthat::expect_true(B2$bollenstine)
    testthat::expect_true(nrow(B1$coef)==2)
    testthat::expect_output(print(B1),"Standard errors:")
    # if (requireNamespace("future",quietly=TRUE)) future::plan("sequential")
    dm <- capture.output(B3 <- bootstrap(e,R=2,fun=function(x) coef(x)[2]^2+10))
    testthat::expect_true(all(mean(B3$coef)>10))

    y <- rep(c(0,1),each=5)
    x <- 1:10
    m <- lvm(y~b*x)
    constrain(m,alpha~b) <- function(x) x^2
    e <- estimate(m,data.frame(y=y,x=x))
    b <- bootstrap(e,R=1,silent=TRUE)
    testthat::expect_output(print(b),"alpha")
})

test_that("Survreg", {
    m <- lvm(y0~x)
    transform(m,y~y0) <- function(x) pmin(x[,1],2)
    transform(m,status~y0) <- function(x) x<2
    d <- simulate(m,100,seed=1)
    require('survival')
    m <- survreg(Surv(y,status)~x,data=d,dist='gaussian')
    s <- score(m)
    testthat::expect_true(length(pars(m))==length(coef(m))+1)
    testthat::expect_true(abs(attr(score(m,pars(m)),'logLik')-logLik(m))<1e-9)
    testthat::expect_true(mean(colSums(s)^2)<1e-6)
    testthat::expect_equivalent(vcov(m), attr(s,'bread')/nrow(d))
})

test_that("diagtest", {
  M <- as.table(matrix(c(42,12,
                         35,28),ncol=2,byrow=TRUE,
                       dimnames=list(rater=c("no","yes"),gold=c("no","yes"))))
  d1 <- diagtest(M,exact=TRUE)
  d2 <- diagtest(M,exact=FALSE)
  e1 <- d1$coefmat[1:7,]
  e2 <- d2$coefmat[1:7,]
  expect_equivalent(e1[,1], e2[,1])
  expect_true(e1[1,1] == (sum(M[,2])/sum(M)))
})

test_that("Combine", { ## Combine model output
    data(serotonin)
    m1 <- lm(cau ~ age*gene1 + age*gene2,data=serotonin)
    m2 <- lm(cau ~ age + gene1,data=serotonin)

    cc <- Combine(list('model A'=m1,'model B'=m2),fun=function(x) c(R2=format(summary(x)$r.squared,digits=2)))
    testthat::expect_true(nrow(cc)==length(coef(m1))+1)
    testthat::expect_equivalent(colnames(cc),c('model A','model B'))
    testthat::expect_equivalent(cc['R2',2],format(summary(m2)$r.squared,digits=2))

})

test_that("zero-inflated binomial regression (zib)", {
    set.seed(1)
    n <- 1e3
    x <- runif(n,0,20)
    age <- runif(n,10,30)
    z0 <- rnorm(n,mean=-1+0.05*age)
    z <- cut(z0,breaks=c(-Inf,-1,0,1,Inf))
    p0 <- lava::expit(model.matrix(~z+age) %*% c(-.4, -.4, 0.2, 2, -0.05))
    y <- (runif(n)<lava:::tigol(-1+0.25*x-0*age))*1
    u <- runif(n)<p0
    y[u==0] <- 0
    d <- data.frame(y=y,x=x,u=u*1,z=z,age=age)

    ## Estimation
    e0 <- zibreg(y~x*z,~1+z+age,data=d)
    e <- zibreg(y~x,~1+z+age,data=d)
    lrt <- compare(e,e0)
    testthat::expect_equivalent(score(e,data=d),
                      colSums(score(e,indiv=TRUE)))
    testthat::expect_equivalent(logLik(e,data=d),
                      logLik(e))
    testthat::expect_equivalent(vcov(e), Inverse(information(e,type="obs",data=d)))

    testthat::expect_output(print(e), "Prevalence probabilities:")

    pdval <- PD(e0,intercept=c(1,3),slope=c(2,6))

    B <- rbind(c(1,0,0,0,20),
               c(1,1,0,0,20),
               c(1,0,1,0,20),
               c(1,0,0,1,20))
    prev <- summary(e,pr.contrast=B)$prevalence

    x <- seq(0,100,length.out=100)
    newdata <- expand.grid(x=x,age=20,z=levels(d$z))
    fit <- predict(e,newdata=newdata)
})

if (requireNamespace("nlme",quietly = TRUE) && requireNamespace("mets",quietly = TRUE))
test_that("Prediction with missing data, random intercept", {
    ## Random intercept model
    m <- lvm(c(y1,y2,y3)~u[0])
    latent(m) <- ~u
    regression(m) <- y1~x1
    regression(m) <- y2~x2
    regression(m) <- y3~x3

    d <- simulate(m,1e3,seed=1)
    dd <- reshape(d,varying=list(c('y1','y2','y3'),c('x1','x2','x3')),direction='long',v.names=c('y','x'))

    ##system.time(l <- lme4::lmer(y~x+(1|id), data=dd, REML=FALSE))
    system.time(l <- nlme::lme(y~x,random=~1|id, data=dd, method="ML"))
    m0 <- lvm(c(y1[m:v],y2[m:v],y3[m:v])~1*u[0])
    latent(m0) <- ~u
    regression(m0,y=c('y1','y2','y3'),x=c('x1','x2','x3')) <- rep('b',3)
    system.time(e <- estimate(m0,d))

    mytol <- 1e-6
    mse <- function(x,y=0) mean(na.omit(as.matrix(x)-as.matrix(y))^2)
    testthat::expect_true(mse(logLik(e),logLik(l))<mytol)
    testthat::expect_true(mse(nlme::fixef(l),coef(e)[1:2])<mytol)
    u1 <- nlme::ranef(l)##[[1]][,1]
    u2 <- predict(e,endogenous(e))
    testthat::expect_true(mse(u1,u2)<1e-9)

    ## Missing data
    idx <- sample(seq(nrow(dd)),nrow(dd)*0.5)
    dd0 <- dd[idx,,drop=FALSE]
    d0 <- mets::fast.reshape(subset(dd0,select=-u),id='id',num='time')

    system.time(e0 <- estimate(m0,d0,missing=TRUE))
    ##system.time(l0 <- lme4::lmer(y~x+(1|id), data=dd0, REML=FALSE))
    system.time(l0 <- nlme::lme(y~x,random=~1|id, data=dd0, method="ML"))
    testthat::expect_true(mse(logLik(e0),logLik(l0))<mytol)
    testthat::expect_true(mse(nlme::fixef(l0),coef(e0)[1:2])<mytol)
    u01 <- nlme::ranef(l0)##[[1]][,1]
    u02 <- predict(e0,endogenous(e0))
    expect_true(mse(u01,u02)<1e-9)

    s <- summary(e0)
    testthat::expect_output(print(s),paste0("data=",nrow(d0)))
    testthat::expect_true(inherits(e0$estimate,"multigroupfit"))
    testthat::expect_output(print(e0$estimate),"Group 1")
    testthat::expect_output(print(summary(e0$estimate)),paste0("observations = ",nrow(d0)))
})

## if (requireNamespace("lme4", quietly = TRUE) && requireNamespace("mets", quietly = TRUE)) {
if (requireNamespace("mets",quietly = TRUE))
test_that("Random slope model", {
    set.seed(1)
    m <- lvm()
    regression(m) <- y1 ~ 1*u + 1*s
    regression(m) <- y2 ~ 1*u + 2*s
    regression(m) <- y3 ~ 1*u + 3*s
    latent(m) <- ~u+s
    dw <- sim(m,20)

    dd <- mets::fast.reshape(dw)
    dd$num <- dd$num+runif(nrow(dd),0,0.2)
    dd0 <- dd[-c(1:2*3),]
    ##l <- lme4::lmer(y~ 1+num +(1+num|id),dd,REML=FALSE)
    l <- nlme::lme(y ~ 1+num, random=~1+num|id, data=dd, method="ML")
    sl <- lava:::varcomp(l)

    d <- mets::fast.reshape(dd,id="id")
    d0 <- mets::fast.reshape(dd0,id="id")

    m0 <- lvm(c(y1[0:v],y2[0:v],y3[0:v])~1*u)
    addvar(m0) <- ~num1+num2+num3
    covariance(m0) <- u~s
    latent(m0) <- ~s+u
    regression(m0) <- y1 ~ num1*s
    regression(m0) <- y2 ~ num2*s
    regression(m0) <- y3 ~ num3*s
    system.time(e <- estimate(m0,d,param="none",control=list(trace=0,constrain=FALSE)))

    testthat::expect_true(mean(sl$coef-coef(e)[c("u","s")])^2<1e-5)
    testthat::expect_true((logLik(l)-logLik(e))^2<1e-5)
    varcomp.nam <- c(paste0("u",lava.options()$symbol[2],"u"),
                paste0("s",lava.options()$symbol[2],"s"))
    testthat::expect_true(mean(diag(sl$varcomp)-coef(e)[varcomp.nam])^2<1e-5)

    ## missing
    testthat::expect_output(e0 <- estimate(m0,d0,missing=TRUE,param="none",
                                           control=list(method="NR",constrain=FALSE,start=coef(e),trace=1)),
                            "Iter=")
    ## l0 <- lmer(y ~ 1 + num + (1 + num | id), dd0, REML = FALSE)
    l0 <- nlme::lme(y~ 1+num, random=~1+num|id, data=dd0, method="ML")
    testthat::expect_true((logLik(l0)-logLik(e0))^2<1e-5)

    m1 <- lvm(c(y1[0:v],y2[0:v],y3[0:v])~1*u)
    addvar(m1) <- ~num1+num2+num3
    covariance(m1) <- u~s
    latent(m1) <- ~s+u
    regression(m1) <- y1 ~ b1*s
    regression(m1) <- y2 ~ b2*s
    regression(m1) <- y3 ~ b3*s
    constrain(m1,b1~num1) <- function(x) x
    constrain(m1,b2~num2) <- function(x) x
    constrain(m1,b3~num3) <- function(x) x
    system.time(e1 <- estimate(m1,d,param="none",p=coef(e)))
    testthat::expect_true((logLik(e1)-logLik(e))^2<1e-5)

    ## TODO
    ## missing
    ## system.time(e10 <- estimate(m1,d0,missing=TRUE,param="none",
    ##                             control=list(trace=coef(e0))))

})


test_that("Predictions, jacobians", {
    m <- lvm(c(x1,x2,x3)~u1,u1~z,
             c(y1,y2,y3)~u2,u2~u1+z)
    latent(m) <- ~u1+u2

    p <- c("u2,u2"=2,"u1,u1"=0.5)
    names(p) <- gsub(",",lava.options()$symbols[2],names(p))
    d <- simulate(m,50,p=p,seed=123)
    e <- estimate(m,d)

    object <- e
    f <- function(p,x=vars(object)) predict(object,x,p=p)
    testthat::expect_true(sum(abs(numDeriv::jacobian(f,coef(object))-predictlvm(object)$mean.jacobian))
                          <1e-6)
    testthat::expect_true(sum(abs(numDeriv::jacobian(function(p) predictlvm(object,p=p)$var,coef(object))-predictlvm(object)$var.jacobian))
                          <1e-6)

    testthat::expect_true(sum(abs(numDeriv::jacobian(function(p) f(p,x1~1),coef(object))-
                                  predictlvm(object,x1~1)$mean.jacobian))
                          <1e-6)

    testthat::expect_true(sum(abs(numDeriv::jacobian(function(p) f(p,u1~x1+x2+x3),coef(object))-
                                  predictlvm(object,u1~x1+x2+x3)$mean.jacobian))
                          <1e-6)
})


if (requireNamespace("mets",quietly = TRUE))
test_that("multinomial", {
    set.seed(1)
    breaks <- c(-Inf,-1,0,Inf)
    m <- lvm(); covariance(m,pairwise=TRUE) <- ~y1+y2+y3+y4
    d <- transform(sim(m,5e2),
                   z1=cut(y1,breaks=breaks),
                   z2=cut(y2,breaks=breaks),
                   z3=cut(y3,breaks=breaks),
                   z4=cut(y4,breaks=breaks))

    m <- multinomial(d[,5])
    lev <- levels(d[,5])
    e <- estimate(l <- lm(d[,5]==lev[1]~1))
    testthat::expect_true(abs(vcov(e)[1]-vcov(m)[1])<1e-9)

    a1 <- multinomial(d[,5:6],marginal=TRUE)
    K1 <- kappa(a1) ## Cohen's kappa
    P <- a1$P
    marg1 <- rowSums(P)
    marg2 <- colSums(P)
    testthat::expect_equivalent(K1$coef,sum(diag(P)-marg1*marg2)/(1-sum(marg1*marg2)))

    K2 <- kappa(d[,7:8])
    ## Testing difference K1-K2:
    e1 <- estimate(merge(K1,K2,id=TRUE),diff)
    e2 <- estimate(merge(K1,K2,id=NULL),diff)
    testthat::expect_true(vcov(e1)!=vcov(e2))
    testthat::expect_equivalent(vcov(e2),(vcov(K1)+vcov(K2)))

    g1 <- gkgamma(d[,5:6])
    g2 <- gkgamma(table(d[,5:6]))
    g3 <- gkgamma(multinomial(d[,5:6]))
    testthat::expect_equivalent(g1$coefmat,g2$coefmat)
    testthat::expect_equivalent(g3$coefmat,g2$coefmat)


    ## TODO
    indep <- lava:::independence(d[,5:6])
    inf <- information(d[,5:6])

    ## pcor
    if (requireNamespace("polycor",quietly=TRUE)) {
        system.time(rho <- pcor(d[,5],d[,6]))
        rho2 <- polycor::polychor(d[,5],d[,6],ML=TRUE,std.err=TRUE)
        testthat::expect_true(abs(rho$coef[1]-rho2$rho)^2<1e-5)
        testthat::expect_true(abs(rho$vcov[1]-rho2$var[1])^2<1e-5)
        testthat::expect_true(mean(score(rho))^2<1e-5)
    }
})


test_that("predict,residuals", {
    m <- lvm(c(y1,y2,y3)~u,u~x)
    latent(m) <- ~u
    set.seed(1)
    d <- sim(m,100,'y1~u'=1,'y3~u'=3)
    e <- estimate(m,d)

    l <- lm(y3~x,data=d)
    e3 <- estimate(y3~x,data=d,lvm=TRUE)
    testthat::expect_true(mean((residuals(l)-residuals(e3))^2)<1e-12)
    testthat::expect_true(mean(var(residuals(e3,std=TRUE))[1]*99/100-1)<1e-3)

    r <- residuals(e)
    testthat::expect_true(ncol(r)==3)
})


test_that("partialcor", {
    m <- lvm(c(y1,y2,y3)~x1+x2)
    covariance(m) <- c(y1,y2,y3)~y1+y2+y3
    set.seed(1)
    d <- sim(m,500)
    c1 <- partialcor(~x1+x2,d)
    e <- estimate(m,d)
    ec <- correlation(e)
    c2 <- parameter(ec)
    testthat::expect_true(mean(c1[,1]-c2[,1])^2<1e-9)
    ## CI, note difference var(z)=1/(n-k-3) vs var(z)=1/(n-3)
    testthat::expect_true(mean(c1[,4]-c2[,3])^2<1e-3)
    testthat::expect_true(mean(c1[,5]-c2[,4])^2<1e-3)
})

## test_that("partialgamma", {
## TODO
## })

test_that("correlation", {
  set.seed(1)
  d <- matrix(rnorm(500), ncol=2)
  e1 <- parameter(correlation(d, z=FALSE))
  e2 <- cor.test(d[,1],d[,2])
  expect_equivalent(e2$estimate, e1[1])
  ci <- e1[3:4]
  ci2 <- e2$conf.int
  expect_true(mean((ci-ci2)^2)< sqrt(1/nrow(d))) # df discrepancy
  e1 <- parameter(correlation(d, z=TRUE))
  ci <- e1[3:4]
  ci2 <- atanh(e2$estimate) + qnorm(0.975)*c(-1,1)*1/sqrt(nrow(d))
  expect_true(mean((ci-ci2)^2)< sqrt(1/nrow(d)))

  d <- matrix(rnorm(500), ncol=5)
  e1 <- parameter(correlation(d, z=FALSE))
  e2 <- parameter(correlation(d, z=TRUE))
  e3 <- parameter(correlation(d, z=TRUE, back.transform=FALSE))
  expect_equivalent(e1[,1], e2[,1]) # est same, but not pval
  expect_equivalent(e2[,5], e3[,5]) # p.vals same, but z vs r
})

if (requireNamespace("mets",quietly = TRUE))
test_that("Weighted",{
    m <- lvm(y~x)
    set.seed(1)
    d <- sim(m,10)
    d$w <- runif(nrow(d),0.1,1)
    e <- estimate(m,data=d)
    l <- lm(y~x,data=d)
    testthat::expect_true(mean((coef(e)[1:2]-coef(l))^2)<1e-12)

    w2 <- estimate(m,data=d,weights=d$w,control=list(trace=0))
    w <- estimate(m,data=d,weights=d$w,estimator="normal",control=list(trace=0))
    expect_equivalent(coef(w2), coef(w))
    lw <- lm(y~x,data=d, weights=d$w)
    testthat::expect_true(mean((coef(w)[1:2]-coef(lw))^2)<1e-12)
})


if (requireNamespace("mets",quietly = TRUE))
test_that("Tobit",{
    m0 <- lvm(t~x)
    distribution(m0,~w) <- uniform.lvm(0.1,1)
    d <- sim(m0,10,seed=1)
    d$status <- rep(c(TRUE,FALSE),each=nrow(d)/2)
    d$s <- with(d, survival::Surv(t,status))
    s <- survival::survreg(s~x,data=d,dist="gaussian",weights=d$w)
    m <- lvm(s~x)
    e <- estimate(m,data=d,estimator="normal",weights="w")
    testthat::expect_true(mean((coef(e)[1:2]-coef(s))^2)<1e-9)
})

#' @keywords internal
"_PACKAGE"

##' Estimation and simulation of latent variable models
##'
##' A general implementation of Structural Equation Models wth latent variables
##' (MLE, 2SLS, and composite likelihood estimators) with both continuous,
##' censored, and ordinal outcomes (Holst and Budtz-Joergensen (2013)
##' <doi:10.1007/s00180-012-0344-y>). Mixture latent variable models and
##' non-linear latent variable models (Holst and Budtz-Joergensen (2020)
##' <doi:10.1093/biostatistics/kxy082>). The package also provides methods for
##' graph exploration (d-separation, back-door criterion), simulation of general
##' non-linear latent variable models, and estimation of influence functions for
##' a broad range of statistical models.
##' @name lava-package
##' @importFrom graphics plot lines points abline points text layout
##'     par plot.new plot.window title rect locator segments image
##'     mtext box axis polygon matplot contour contour.default
##'     identify rug curve
##' @importFrom grDevices xy.coords col2rgb rgb colors rainbow
##'     topo.colors gray.colors palette colorRampPalette heat.colors
##' @importFrom utils stack combn read.csv getTxtProgressBar
##'     setTxtProgressBar txtProgressBar head tail modifyList
##'     getFromNamespace write.table methods data
##'     glob2rx
##' @importFrom stats density deriv effects lm family simulate vcov
##'     var cov cor coef model.frame model.weights as.formula
##'     model.matrix rnorm rchisq runif rlnorm pnorm qnorm na.omit AIC
##'     terms logLik qt pt update update.formula confint approxfun
##'     pchisq confint.default formula fft uniroot rbinom predict sd
##'     addmargins residuals dnorm quantile qf cov2cor qchisq
##'     get_all_vars p.adjust rpois rt rmultinom rgamma printCoefmat 
##'     glm nlminb na.pass na.omit influence weights density.default
##' @importFrom survival is.Surv
##' @importFrom future.apply future_mapply future_lapply
##' @importFrom methods new as
##' @aliases lava-package lava
##' @author Klaus K. Holst Maintainer: <klaus@@holst.it>
##' @keywords package
##' @examples
##' lava()
NULL

##' Longitudinal Bone Mineral Density Data
##'
##' Bone Mineral Density Data consisting of 112 girls randomized to receive
##' calcium og placebo. Longitudinal measurements of bone mineral density
##' (g/cm^2) measured approximately every 6th month in 3 years.
##' @name calcium
##' @docType data
##' @format A data.frame containing 560 (incomplete) observations. The 'person'
##' column defines the individual girls of the study with measurements at
##' visiting times 'visit', and age in years 'age' at the time of visit. The
##' bone mineral density variable is 'bmd' (g/cm^2).
##' @source Vonesh & Chinchilli (1997), Table 5.4.1 on page 228.
##' @keywords datasets
NULL

##' Longitudinal Bone Mineral Density Data (Wide format)
##'
##' Bone Mineral Density Data consisting of 112 girls randomized to receive
##' calcium og placebo. Longitudinal measurements of bone mineral density
##' (g/cm^2) measured approximately every 6th month in 3 years.
##' @name bmd
##' @docType data
##' @source Vonesh & Chinchilli (1997), Table 5.4.1 on page 228.
##' @format data.frame
##' @keywords datasets
##' @seealso calcium
NULL

##' Simulated data
##'
##' Simulated data
##' @name brisa
##' @docType data
##' @format data.frame
##' @source Simulated
##' @keywords datasets
NULL

##' Data
##'
##' Description
##' @name bmidata
##' @docType data
##' @format data.frame
##' @keywords datasets
NULL

##' 50 patients from Monash Medical Centre, Melbourne
##'
##' Diagnosis of depression (DSM-III-R MDD, Dysthymia, Adjustment Disorder with
##' Depressed Mood, Depression NOS), Beck Depression Inventory (BDI) (Beck et
##' al., 1961) General Health Questionnaire (GHQ) (Goldberg & Williams, 1988)
##'
##' @name deprdiag
##' @docType data
##' @format data.frame
##' @source Clarke, D. M., Smith, G. C., & Herrman, H. E. (1993). A comparative
##'   study of screening instruments for mental disorders in general hospital
##'   patients. International Journal Psychiatry in Medicine, 23, pp. 323-337.
##'
##' McKenzie et al. (1996). Comparing correlated Kappas by resampling: Is one
##' level of agreement significantly different from another? J. Psychiat. Res.
##' 30 (6), pp. 483-492.
##' @keywords datasets
NULL

##' Hubble data
##'
##' Velocity (v) and distance (D) measures of 36 Type Ia super-novae from the
##' Hubble Space Telescope
##' @name hubble
##' @docType data
##' @format data.frame
##' @source Freedman, W. L., et al. 2001, AstroPhysicalJournal, 553, 47.
##' @keywords datasets
NULL

##' Hubble data
##'
##' @name hubble2
##' @seealso hubble
##' @docType data
##' @format data.frame
##' @keywords datasets
NULL

##' Data
##'
##' Description
##' @name indoorenv
##' @docType data
##' @format data.frame
##' @source Simulated
##' @keywords datasets
NULL

##' Missing data example
##'
##' Simulated data generated from model
##' \deqn{E(Y_i\mid X) = X, \quad cov(Y_1,Y_2\mid X)=0.5}
##'
##' The list contains four data sets
##' 1) Complete data
##' 2) MCAR
##' 3) MAR
##' 4) MNAR (missing mechanism depends on variable V correlated with Y1,Y2)
##' @examples
##' data(missingdata)
##' e0 <- estimate(lvm(c(y1,y2)~b*x,y1~~y2),missingdata[[1]]) ## No missing
##' e1 <- estimate(lvm(c(y1,y2)~b*x,y1~~y2),missingdata[[2]]) ## CC (MCAR)
##' e2 <- estimate(lvm(c(y1,y2)~b*x,y1~~y2),missingdata[[2]],missing=TRUE) ## MCAR
##' e3 <- estimate(lvm(c(y1,y2)~b*x,y1~~y2),missingdata[[3]]) ## CC (MAR)
##' e4 <- estimate(lvm(c(y1,y2)~b*x,y1~~y2),missingdata[[3]],missing=TRUE) ## MAR
##' @name missingdata
##' @docType data
##' @format list of data.frames
##' @source Simulated
##' @keywords datasets
NULL

##' Example data (nonlinear model)
##'
##' @name nldata
##' @docType data
##' @format data.frame
##' @source Simulated
##' @keywords datasets
NULL

##' Example SEM data (nonlinear)
##'
##' Simulated data
##' @name nsem
##' @docType data
##' @format data.frame
##' @source Simulated
##' @keywords datasets
NULL

##' Example SEM data
##'
##' Simulated data
##' @name semdata
##' @docType data
##' @source Simulated
##' @format data.frame
##' @keywords datasets
NULL

##' Serotonin data
##'
##' This simulated data mimics a PET imaging study where the 5-HT2A
##' receptor and serotonin transporter (SERT) binding potential has
##' been quantified into 8 different regions. The 5-HT2A
##' cortical regions are considered high-binding regions
## 'which are a priori known to yield quite similar and highly correlated
##' measurements.  These measurements can be regarded as proxy measures of
##' the extra-cellular levels of serotonin in the brain
##' \tabular{rll}{
##'         day    \tab numeric \tab Scan day of the year \cr
##'         age    \tab numeric \tab Age at baseline scan \cr
##'         mem    \tab numeric \tab Memory performance score \cr
##'         depr   \tab numeric \tab Depression (mild) status 500 days after baseline \cr
##'         gene1  \tab numeric \tab Gene marker 1 (HTR2A) \cr
##'         gene2  \tab numeric \tab Gene marker 2 (HTTTLPR) \cr
##'         cau \tab numeric \tab SERT binding, Caudate Nucleus \cr
##'         th  \tab numeric \tab SERT binding, Thalamus \cr
##'         put \tab numeric \tab SERT binding, Putamen \cr
##'         mid \tab numeric \tab SERT binding, Midbrain \cr
##'         aci \tab numeric \tab 5-HT2A binding, Anterior cingulate gyrus \cr
##'         pci  \tab numeric \tab 5-HT2A binding, Posterior cingulate gyrus \cr
##'         sfc \tab numeric \tab 5-HT2A binding, Superior frontal cortex \cr
##'         par \tab numeric \tab 5-HT2A binding, Parietal cortex \cr
##' }
##' @name serotonin
##' @docType data
##' @format data.frame
##' @source Simulated
##' @keywords datasets
NULL

##' Twin menarche data
##'
##' Simulated data
##' \tabular{rll}{
##'         id    \tab numeric \tab Twin-pair id \cr
##'         zyg    \tab character \tab Zygosity (MZ or DZ) \cr
##'         twinnum    \tab numeric \tab Twin number (1 or 2) \cr
##'         agemena    \tab numeric \tab Age at menarche (or censoring) \cr
##'         status    \tab logical \tab Censoring status (observed:=T,censored:=F) \cr
##'         bw    \tab numeric  \tab Birth weight \cr
##'         msmoke    \tab numeric \tab Did mother smoke? (yes:=1,no:=0) \cr
##' }
##' @name twindata
##' @docType data
##' @format data.frame
##' @keywords datasets
##' @source Simulated
NULL

##' For internal use
##'
##' @title For internal use
##' @name startvalues
##' @author Klaus K. Holst
##' @keywords internal
##' @rdname internal
##' @export
##' @aliases startvalues startvalues0 startvalues1 startvalues2
##'   starter.multigroup
##'   modelPar modelVar matrices pars pars.lvm regfix
##'   pars.lvmfit pars.glm score.glm procdata.lvmfit mat.lvm
##'   matrices reorderdata graph2lvm igraph.lvm subgraph finalize
##'   randomslope randomslope<- lisrel
##'   variances offdiags describecoef parlabels rsq stdcoef CoefMat
##'   CoefMat.multigroupfit deriv updatelvm checkmultigroup profci
##'   estimate.MAR
##'   missingModel Identical gaussian_logLik.lvm
##'   addhook gethook multigroup
##'   Weights fixsome IV parameter Specials
##'   procformula getoutcome decomp.specials
##'   rmvn0 dmvn0 logit expit tigol
NULL

##' Performs a rotation in the plane
##'
##' @title Performs a rotation in the plane
##' @aliases rotate2 rot2D rot3D
##' @param x Matrix to be rotated (2 times n)
##' @param theta Rotation in radians
##' @return Returns a matrix of the same dimension as \code{x}
##' @author Klaus K. Holst
##' @export
##' @examples
##' rotate2(cbind(c(1,2),c(2,1)))
##' @keywords hplot
`rotate2` <-
function(x,theta=pi) {
  R <- matrix(c(cos(theta), -sin(theta), sin(theta), cos(theta)), byrow=TRUE, ncol=2)
  x%*%R
}

## clockwise rotation 2d:
##' @export
rot2D <- function(theta) {
  matrix(c(cos(theta),sin(theta),-sin(theta),cos(theta)),2)
}

##' @export
rot3D <- function(x=0,y=0,z=0) {
  Rx <- function() {
    R2 <- rot2D(x)
    R <- diag(3)
    R[2:3,2:3] <- R2
    return(R)
  }
  Ry <- function() {
    R2 <- rot2D(y)
    R <- diag(3)
    R[c(1,3),c(1,3)] <- R2
    return(R)
  }
  Rz <- function() {
    R2 <- rot2D(z)
    R <- diag(3)
    R[1:2,1:2] <- R2
    return(R)
  } 
  res <- diag(3)
  if (x!=0) res <- res%*%Rx()
  if (y!=0) res <- res%*%Ry()
  if (z!=0) res <- res%*%Rz()
  return(res)
}

context("Linear algebra functions")

test_that("Matrix operations:", {
    ## vec operator
    testthat::expect_equivalent(vec(diag(3)),c(1,0,0,0,1,0,0,0,1))
    testthat::expect_true(nrow(vec(diag(3),matrix=TRUE))==9)

    ## commutaion matrix
    A <- matrix(1:16 ,ncol=4)
    K <- commutation(A)
    testthat::expect_equivalent(K%*%as.vector(A),vec(t(A),matrix=TRUE))
})

test_that("blockdiag:", {
    ## Block diagonal
    A <- diag(3)+1
    B <- blockdiag(A,A,A,pad=NA)
    testthat::expect_equivalent(dim(B),c(9,9))
    testthat::expect_true(sum(is.na(B))==81-27)

    B <- matrix(1, ncol=2, nrow=2)
    res <- blockdiag(A, B, B)
    expect_equivalent(res[1:3, 1:3], A)
    expect_equivalent(res[4:5, 4:5], B)
    expect_equivalent(res[6:7, 6:7], B)
    expect_true(sum(res==0) == (7*7 - 3*3 - 2*2 - 2*2))
})

test_that("wrapvev", {
    testthat::expect_equivalent(wrapvec(5,2),c(3,4,5,1,2))
    testthat::expect_equivalent(wrapvec(seq(1:5),-1),c(5,1,2,3,4))
})

test_that("revdiag", {
    A <- revdiag(1:3)
    testthat::expect_equivalent(A,matrix(c(0,0,1,0,2,0,3,0,0),3))
    testthat::expect_equivalent(1:3,revdiag(A))
    revdiag(A) <- 4
    testthat::expect_equivalent(rep(4,3),revdiag(A))
    diag(A) <- 0
    offdiag(A) <- 5
    testthat::expect_true(sum(offdiag(A))==6*5)
})

test_that("Inv, matrix inverse", {
    A <- matrix(0,3,3)
    offdiag(A,type=3) <- 1:6
    B <- crossprod(A)

    testthat::expect_equivalent(solve(A),Inverse(A))
    testthat::expect_equivalent(det(B),attr(Inverse(B,chol=TRUE),"det"))
})

test_that("rotate2", {
  ##rotate2: default rotation by pi flips the point
  x <- cbind(c(1, 0), c(0, 1))  # 2x2 matrix (2 rows, 2 cols)
  result <- rotate2(x, theta = pi)
  expected <- cbind(c(-1, 0), c(0, -1))
  expect_equal(result, expected, tolerance = 1e-10)

  ##rotate2: rotation by 0 returns the same matrix
  x <- cbind(c(1, 2), c(3, 4))
  result <- rotate2(x, theta = 0)
  expect_equal(result, x, tolerance = 1e-10)

  ##rotate2: rotation by pi/2 rotates 90 degrees
  x <- rbind(c(1, 0))  # row vector (1x2)
  result <- rotate2(x, theta = pi / 2)
  expected <- rbind(c(0, -1))
  expect_equal(result, expected, tolerance = 1e-10)

  ##rotate2: two rotations of pi/2 equal one rotation of pi
  x <- cbind(c(1, 2), c(2, 1))
  result_double <- rotate2(rotate2(x, theta = pi / 2), theta = pi / 2)
  result_pi     <- rotate2(x, theta = pi)
  expect_equal(result_double, result_pi, tolerance = 1e-10)
})

test_that("rot2D", {
  ## returns a 2x2 matrix
  R <- rot2D(pi / 4)
  expect_equal(dim(R), c(2, 2))
  ## rot2D: rotation matrix is orthogonal (R^T * R = I)
  R <- rot2D(pi / 3)
  expect_equal(t(R) %*% R, diag(2), tolerance = 1e-10)
  ## rot2D: determinant of rotation matrix is 1
  R <- rot2D(pi / 6)
  expect_equal(det(R), 1, tolerance = 1e-10)
  ## rot2D: rotation by 0 returns identity matrix
  R <- rot2D(0)
  expect_equal(R, diag(2), tolerance = 1e-10)
})

test_that("rot3D", {
  ## returns a 3x3 matrix
  R <- rot3D(x = pi / 4, y = pi / 6, z = pi / 3)
  expect_equal(dim(R), c(3, 3))
  ## rot3D: no rotation returns identity matrix
  R <- rot3D()
  expect_equal(R, diag(3), tolerance = 1e-10)
  ## rot3D: rotation matrix is orthogonal (R^T * R = I)
  R <- rot3D(x = pi / 4, y = pi / 3, z = pi / 6)
  expect_equal(t(R) %*% R, diag(3), tolerance = 1e-10)
  ## rot3D: determinant of rotation matrix is 1
  R <- rot3D(x = pi / 5, y = pi / 7, z = pi / 9)
  expect_equal(det(R), 1, tolerance = 1e-10)
  ## rot3D: rotation around z-axis only leaves z-component unchanged
  R <- rot3D(z = pi / 2)
  v <- c(1, 0, 0)
  result <- R %*% v
  # x-axis rotated 90° around z should give (0, 1, 0)
  expect_equal(as.numeric(result), c(0, 1, 0), tolerance = 1e-10)
  ## rot3D: rotation around x-axis only leaves x-component unchanged
  R <- rot3D(x = pi / 2)
  v <- c(0, 1, 0)
  result <- R %*% v
  # y-axis rotated 90° around x should give (0, 0, 1)
  expect_equal(as.numeric(result), c(0, 0, 1), tolerance = 1e-10)
})

with_unique_warnings <- function(expr) {
  # utility function that prevents the same warning being cast more than once
  seen <- character(0)
  withCallingHandlers(expr, warning = function(w) {
    msg <- conditionMessage(w)
    if (msg %in% seen) {
      invokeRestart("muffleWarning")
    } else {
      seen <<- c(seen, msg)
    }
  })
}

# small sample corrections to variance estimates
small_sample_correction <- function(ic, type, var.adj) {
  V <- var_ic(ic)
  ## Small-sample corrections for clustered data
  K <- NROW(ic)
  N <- attributes(ic)$N
  if (is.null(N)) N <- K
  p <- NCOL(ic)
  adj0 <- K/(K-p) ## Mancl & DeRouen, 2001
  adj1 <- K/(K-1) ## Mancl & DeRouen, 2001
  adj2 <- (N-1)/(N-p)*(K/(K-1)) ## Morel,Bokossa & Neerchal, 2003
  if (tolower(type[1])=="mbn" && !is.null(attributes(ic)$bread)) {
    V0 <- V
    iI0 <- attributes(ic)$bread
    I0 <- Inverse(iI0)
    delta <- min(0.5, p / (K - p))
    phi <- max(1, tr(I0%*%V0)*adj2/p)
    V <- adj2*V0 + delta*phi*iI0
  }
  if (tolower(type[1])=="df") {
    V <- adj0*V
  }
  if (tolower(type[1])=="df1") {
    V <- adj1*V
  }
  if (tolower(type[1])=="df2") {
    V <- adj2*V
  }
  if (tolower(type[1])%in%c("hc3", "hc4")) {
    ic <- cbind(ic)
    S <- Inverse(crossprod(ic), tol=sqrt(.Machine$double.eps))
    h_emp <- rowSums((ic %*% S) * ic) # empirical h, lev.
    n <- nrow(ic)
    ## h_emp <- pmin(h_emp, 0.99) * (n-1)/n  # Truncate leverage to prevent division by zero
    if (tolower(type[1])=="hc3") {
      ## phi_norm <- sqrt(rowSums(ic^2))
      ## ex_kurt <- (mean((phi_norm - mean(phi_norm))^4) / var(phi_norm)^2) - 3
      ## alpha <- exp(-max(0, ex_kurt) / 25)
      v.alpha <- ifelse(missing(var.adj), 0.25, var.adj)
      h <- v.alpha * h_emp + (1 - v.alpha) * (ncol(ic) / n)
      adj <- 1 / (1 - h)
    } else {
      ## Cribari-Neto (2004)
      delta <- pmin(1.5, n/ncol(ic) * h_emp)
      adj <- 1 / (1 - h_emp)**delta
    }
    for (i in seq_len(NCOL(ic))) ic[, i] <- ic[, i] * adj
    V <- var_ic(ic)
  }
  return(V)
}

# construct coefficient matrix with confidence limits and two-sided p-values
estimate_coefmat <- function(est, se, df, level = 0.95, null = 0) {
  alpha <- 1 - level
  alpha.str <- paste(c(alpha/2, 1 -alpha/2)*100, "", sep="%")
  if (!is.null(df)) {
    za <- qt(1-alpha/2, df=df)
    pval <- 2*pt(abs((est-null) / se), df=df, lower.tail=FALSE)
  } else {
    za <- qnorm(1-alpha/2)
    pval <- 2*pnorm(abs((est-null)/ se), lower.tail=FALSE)
  }
  res <- cbind(est, se, est - za * se, est + za * se, pval)
  colnames(res) <- c("Estimate", "Std.Err", alpha.str, "P-value")
  return(res)
}

#' Summary of estimate objects
#'
#' Computes hypothesis tests, contrasts, and small-sample corrections for
#' an [estimate] object. The arguments `null`, `contrast`, `type`, and
#' `var.adj` were previously available on [estimate.default()] and have
#' been moved here.
#'
#' @param object an `estimate` object.
#' @param contrast (optional) contrast matrix for the final Wald test.
#' @param null (optional) null hypothesis to test.
#' @param type type of small-sample correction. Requires the estimate to have
#'   been computed with `IC=TRUE` (the default).
#' @param var.adj variance adjustment parameter for small-sample correction.
#'   Requires the estimate to have been computed with `IC=TRUE` (the default).
#' @param df degrees of freedom for t-based inference (default: `NULL` for
#'   Gaussian approximation; when set, confidence intervals and p-values use the
#'   t-distribution with `df` degrees of freedom)
#' @param level level of confidence limits (default 0.95)
#' @param transform (optional) function applied to the point estimates and
#'   confidence interval bounds *after* inference is performed on the original
#'   scale. Useful for variance-stabilizing transformations, e.g., compute CIs
#'   on the `atanh` (Fisher z) scale and back-transform with `tanh`.
#' @param print (optional) custom print function for the resulting
#'   `summary.estimate` object
#' @param ... additional arguments passed to [contr].
#' @seealso [estimate.default()]
#' @export
summary.estimate <- function(object,
                             contrast,
                             ...,
                             null = 0,
                             level = 0.95,
                             type,
                             var.adj = 0.25,
                             df,
                             transform = NULL,
                             print = NULL) {
  with_unique_warnings({
    p <- coef(object)
    if (missing(df)) df <- object$df
    if (missing(contrast)) contrast <- diag(1, nrow=length(p))
    correction <- !missing(type)
    if (correction) {
      if (is.null(object$IC)) {
        stop(
          "Small-sample corrections in summary() require an estimate ",
          "computed with IC=TRUE."
        )
      }
      V <- small_sample_correction(IC(object), type=type, var.adj=var.adj)
    } else {
      V <- vcov(object)
    }
    ## Preserve dimnames on vcov: the recall path through type/var.adj
    ## corrections builds V from the influence function and may drop
    ## dimnames. Restore them from the original coef names so the
    ## summary output is consistent with the deprecated estimate() path.
    if (!is.null(V) && is.null(dimnames(V)) &&
        !is.null(names(p)) &&
        nrow(V) == length(p)) {
      dimnames(V) <- list(names(p), names(p))
    }

    if (is.vector(contrast) || is.list(contrast)) {
      contrast <- contr(contrast, names(object$coef), ...)
    }

    cc0 <- estimate_coefmat(p, diag(V)**.5, df=df, level=level, null=null)
    rownames(cc0) <- rownames(parameter(object))
    waldtest <- NULL
    if (!is.null(contrast)) {
      waldtest <- compare(object, contrast=contrast, null=null, vcov=V)
    }
    class(object) <- "list"
    res <- c(object[c("coef", "coefmat", "vcov", "call",
                      "ncluster", "model.index")], list(compare=waldtest))
    res$coefmat <- cc0
    if (!is.null(transform)) {
      res$coefmat[, c(1, 3, 4)] <- do.call(transform,
                                           list(res$coefmat[, c(1, 3, 4)]))
      res$coefmat[, 2] <- NA
      res$vcov <- NULL
      res$coef <- res$coefmat[, 1, drop=TRUE]
    }
    res$print <- print
    class(res) <- c("summary.estimate", "estimate")
    return(res)
  })
}

#' @export
print.summary.estimate <- function(x, ...) {
  if (!is.null(x$print)) {
    x$print(x, ...)
    return(invisible(x))
  }
  print.estimate(x, type=2L, ...)
  if (!is.null(attributes(x)$extra)) {
    print(attributes(x)$extra)
  }
  return(invisible(x))
}

#' @export
coef.summary.estimate <- function(object, ...) {
  return(object$coef)
}

#' @export
confint.summary.estimate <- function(object, ...) {
  parameter(object)[, 3:4, drop=FALSE]
}

#' @export
parameter.summary.estimate <- function(x, ...) {
  return(x$coefmat)
}

#' @export
vcov.summary.estimate <- function(object, ...) {
  res <- object$vcov
  p <- coef(object)
  if (is.null(res)) {
    res <- matrix(NA, length(p), length(p))
    dimnames(res) <- list(names(p), names(p))
  }
  return(res)
}

#' Concatenate summary.estimate objects
#'
#' When all arguments are `summary.estimate` objects, they are merged into a
#' single `summary.estimate` object. When some arguments are not
#' `summary.estimate` objects but are named numeric scalars/vectors, an object
#' of class `summary.estimate` is returned with an additional `extra` attribute
#' that contains the provided numeric scalars/vectors. This is useful for
#' bundling auxiliary per-iteration information (e.g., convergence status)
#' alongside an estimate for use with [sim.default()].
#'
#' @param ... `summary.estimate` objects and/or named numeric values
#' @examples
#' e1 <- estimate(coef = 1, IC = scale(rnorm(10)), id = 1:10, labels = "a1")
#' e2 <- estimate(coef = 2, IC = scale(rnorm(10)), id = 1:10, labels = "a2")
#'
#' # concatenating two summary.estimate objects
#' c(e1, e2)
#'
#' # concatenating one summary.estimate object with one numerical variable
#' ss <- c(summary(e1), niter = 2)
#' print(ss)
#' attributes(ss)$extra
#' @export
c.summary.estimate <- function(...) {
  args <- list(...)
  if (length(args) == 1) return(args[[1]])

  mask <- sapply(args, function(x) inherits(x, "summary.estimate"))
  summary_objects <- args[mask]

  extras_existing <- unlist(
    lapply(summary_objects, function(x) attributes(x)$extra)
  )

  extras <- c(extras_existing, unlist(args[!mask]))

  .print <- function(x, digits = 4, ...) {
    cat("Concatenated summary.estimate objects: \n")
    print(cli::rule(width = min(cli::console_width(), 60)))
    print(x$coefmat, digits = digits)
    print(cli::rule(width = min(cli::console_width(), 60)))
    if (!is.null(attributes(x)$extra)) {
      print(attributes(x)$extra)
    }
  }

  res <- structure(list(
    coefmat = Reduce(rbind, lapply(summary_objects, function(x) parameter(x))),
    coef = as.vector(
      Reduce(rbind, lapply(summary_objects, function(x) coef(x)))
    ),
    vcov = cbind(Reduce(function(...) blockdiag(..., pad=NA),
                        lapply(summary_objects, function(x) vcov(x)))),
    objects = summary_objects,
    print = .print # consumed by print.summary.estimate
  ), class = c("summary.estimate", "estimate")
  )
  names(res$coef) <- rownames(res$coefmat)
  if (length(extras) > 0) attr(res, "extra") <- extras

  return(res)
}

##' Define range constraints of parameters
##'
##' @aliases Range.lvm
##' @title Define range constraints of parameters
##' @param a Lower bound
##' @param b Upper bound
##' @return function
##' @author Klaus K. Holst
##' @export
Range.lvm <- function(a=0,b=1) {
    if (b==Inf) {
        f <- function(x) {
            res <- a+exp(x)
            attributes(res)$grad <- exp
            res
        }
        return(f)
    }
    if (a==-Inf) {
        f <- function(x) {
            res <- -exp(x)+b
            attributes(res)$grad <- function(x) -exp(x)
            res
        }
        return(f)
    }
    f <- function(x) {
        res <- (a+b*exp(x))/(1+exp(x))
        attributes(res)$grad <- function(x) exp(x)*(b-a-a*b*exp(x))/(1+exp(x))^2
        res
    }
    return(f)
}

##' Add non-linear constraints to latent variable model
##'
##' Add non-linear constraints to latent variable model
##'
##' Add non-linear parameter constraints as well as non-linear associations
##' between covariates and latent or observed variables in the model (non-linear
##' regression).
##'
##' As an example we will specify the follow multiple regression model:
##'
##' \deqn{E(Y|X_1,X_2) = \alpha + \beta_1 X_1 + \beta_2 X_2} \deqn{V(Y|X_1,X_2)
##' = v}
##'
##' which is defined (with the appropiate parameter labels) as
##'
##' \code{m <- lvm(y ~ f(x,beta1) + f(x,beta2))}
##'
##' \code{intercept(m) <- y ~ f(alpha)}
##'
##' \code{covariance(m) <- y ~ f(v)}
##'
##' The somewhat strained parameter constraint \deqn{ v =
##' \frac{(beta1-beta2)^2}{alpha}}
##'
##' can then specified as
##'
##' \code{constrain(m,v ~ beta1 + beta2 + alpha) <- function(x)
##' (x[1]-x[2])^2/x[3] }
##'
##' A subset of the arguments \code{args} can be covariates in the model,
##' allowing the specification of non-linear regression models.  As an example
##' the non-linear regression model \deqn{ E(Y\mid X) = \nu + \Phi(\alpha +
##' \beta X)} where \eqn{\Phi} denotes the standard normal cumulative
##' distribution function, can be defined as
##'
##' \code{m <- lvm(y ~ f(x,0)) # No linear effect of x}
##'
##' Next we add three new parameters using the \code{parameter} assigment
##' function:
##'
##' \code{parameter(m) <- ~nu+alpha+beta}
##'
##' The intercept of \eqn{Y} is defined as \code{mu}
##'
##' \code{intercept(m) <- y ~ f(mu)}
##'
##' And finally the newly added intercept parameter \code{mu} is defined as the
##' appropiate non-linear function of \eqn{\alpha}, \eqn{\nu} and \eqn{\beta}:
##'
##' \code{constrain(m, mu ~ x + alpha + nu) <- function(x)
##' pnorm(x[1]*x[2])+x[3]}
##'
##' The \code{constraints} function can be used to show the estimated non-linear
##' parameter constraints of an estimated model object (\code{lvmfit} or
##' \code{multigroupfit}). Calling \code{constrain} with no additional arguments
##' beyound \code{x} will return a list of the functions and parameter names
##' defining the non-linear restrictions.
##'
##' The gradient function can optionally be added as an attribute \code{grad} to
##' the return value of the function defined by \code{value}. In this case the
##' analytical derivatives will be calculated via the chain rule when evaluating
##' the corresponding score function of the log-likelihood. If the gradient
##' attribute is omitted the chain rule will be applied on a numeric
##' approximation of the gradient.
##' @aliases constrain constrain<- constrain.default constrain<-.multigroup
##' constrain<-.default constraints parameter<-
##' @return A \code{lvm} object.
##' @author Klaus K. Holst
##' @seealso \code{\link{regression}}, \code{\link{intercept}},
##' \code{\link{covariance}}
##' @keywords models regression
##' @examples
##' ##############################
##' ### Non-linear parameter constraints 1
##' ##############################
##' m <- lvm(y ~ f(x1,gamma)+f(x2,beta))
##' covariance(m) <- y ~ f(v)
##' d <- sim(m,100)
##' m1 <- m; constrain(m1,beta ~ v) <- function(x) x^2
##' ## Define slope of x2 to be the square of the residual variance of y
##' ## Estimate both restricted and unrestricted model
##' e <- estimate(m,d,control=list(method="NR"))
##' e1 <- estimate(m1,d)
##' p1 <- coef(e1)
##' p1 <- c(p1[1:2],p1[3]^2,p1[3])
##' ## Likelihood of unrestricted model evaluated in MLE of restricted model
##' logLik(e,p1)
##' ## Likelihood of restricted model (MLE)
##' logLik(e1)
##'
##' ##############################
##' ### Non-linear regression
##' ##############################
##'
##' ## Simulate data
##' m <- lvm(c(y1,y2)~f(x,0)+f(eta,1))
##' latent(m) <- ~eta
##' covariance(m,~y1+y2) <- "v"
##' intercept(m,~y1+y2) <- "mu"
##' covariance(m,~eta) <- "zeta"
##' intercept(m,~eta) <- 0
##' set.seed(1)
##' d <- sim(m,100,p=c(v=0.01,zeta=0.01))[,manifest(m)]
##' d <- transform(d,
##'                y1=y1+2*pnorm(2*x),
##'                y2=y2+2*pnorm(2*x))
##'
##' ## Specify model and estimate parameters
##' constrain(m, mu ~ x + alpha + nu + gamma) <- function(x) x[4]*pnorm(x[3]+x[1]*x[2])
##' \donttest{ ## Reduce Ex.Timings
##' e <- estimate(m,d,control=list(trace=1,constrain=TRUE))
##' constraints(e,data=d)
##' ## Plot model-fit
##' plot(y1~x,d,pch=16); points(y2~x,d,pch=16,col="gray")
##' x0 <- seq(-4,4,length.out=100)
##' lines(x0,coef(e)["nu"] + coef(e)["gamma"]*pnorm(coef(e)["alpha"]*x0))
##' }
##'
##' ##############################
##' ### Multigroup model
##' ##############################
##' ### Define two models
##' m1 <- lvm(y ~ f(x,beta)+f(z,beta2))
##' m2 <- lvm(y ~ f(x,psi) + z)
##' ### And simulate data from them
##' d1 <- sim(m1,500)
##' d2 <- sim(m2,500)
##' ### Add 'non'-linear parameter constraint
##' constrain(m2,psi ~ beta2) <- function(x) x
##' ## Add parameter beta2 to model 2, now beta2 exists in both models
##' parameter(m2) <- ~ beta2
##' ee <- estimate(list(m1,m2),list(d1,d2),control=list(method="NR"))
##' summary(ee)
##'
##' m3 <- lvm(y ~ f(x,beta)+f(z,beta2))
##' m4 <- lvm(y ~ f(x,beta2) + z)
##' e2 <- estimate(list(m3,m4),list(d1,d2),control=list(method="NR"))
##' e2
##' @export
##' @usage
##'
##' \method{constrain}{default}(x,par,args,endogenous=TRUE,...) <- value
##'
##' \method{constrain}{multigroup}(x,par,k=1,...) <- value
##'
##' constraints(object,data=model.frame(object),vcov=object$vcov,level=0.95,
##'                         p=pars.default(object),k,idx,...)
##'
##' @param x \code{lvm}-object
##' @param par Name of new parameter. Alternatively a formula with lhs
##' specifying the new parameter and the rhs defining the names of the
##' parameters or variable names defining the new parameter (overruling the
##' \code{args} argument).
##' @param args Vector of variables names or parameter names that are used in
##' defining \code{par}
##' @param endogenous TRUE if variable is endogenous (sink node)
##' @param k For multigroup models this argument specifies which group to
##' add/extract the constraint
##' @param value Real function taking args as a vector argument
##' @param object \code{lvm}-object
##' @param data Data-row from which possible non-linear constraints should be
##' calculated
##' @param vcov Variance matrix of parameter estimates
##' @param level Level of confidence limits
##' @param p Parameter vector
##' @param idx Index indicating which constraints to extract
##' @param \dots Additional arguments to be passed to the low level functions
"constrain<-" <- function(x,...,value) UseMethod("constrain<-")

##' @export
"constrain" <- function(x,...) UseMethod("constrain")

##' @export
constrain.default <- function(x, par, fun, idx, level=0.95, vcov, estimate=FALSE, ...) {
    if (!missing(par)) {
        constrain(x, par, ...) <- fun
        return(x)
    }
    if (estimate) {
        return(constraints(x,...))
    }
    if (missing(fun)) {
        if (inherits(Model(x),"multigroup")) {
            res <- list()
            for (m in Model(x)$lvm) {
                if (length(constrain(m))>0)
                    res <- c(res, constrain(m))
            }
            return(res)
        }
        return(Model(x)$constrain)
    }
    if (is.numeric(x)) {
        b <- x
    } else {
        b <- pars(x)
    }
    if (missing(vcov)) {
        S <- stats::vcov(x)
    } else {
        S <- vcov
    }
    if (!missing(idx)) {
        b <- b[idx]; S <- S[idx,idx,drop=FALSE]
    }
    fb <- fun(b)
    pl <- 1-(1-level)/2
    D <- rbind(numDeriv::grad(fun,b))
    se <- (D%*%S%*%t(D))^0.5
    res <- c(fb,se,fb+c(-1,1)*qnorm(pl)*c(se))
    pstr <- paste0(format(c(round(1000-1000*pl),round(pl*1000))/10),"%")
    names(res) <- c("Estimate","Std.Err",pstr)
    res
}

##' @export
"constrain<-.multigroupfit" <- function(x,par,k=1,...,value) {
        constrain(Model(x)$lvm[[k]],par=par,...) <- value
        return(x)
    }

##' @export
"constrain<-.multigroup" <- function(x,par,k=1,...,value) {
        constrain(Model(x)$lvm[[k]],par=par,...) <- value
        return(x)
    }

##' @export
"constrain<-.default" <- function(x,par,args,endogenous=TRUE,...,value) {
    if (inherits(par,"formula")) {
        lhs <- getoutcome(par)
        xf <- attributes(terms(par))$term.labels
        par <- lhs
        if (length(par)==0) {
          par <- xf
          xf <- NULL
        }
        if (par%in%vars(x)) {
            if (is.na(x$mean[[par]])) {
                intercept(x,par) <- par
            } else {
                par <- x$mean[[par]]
            }
        }
        args <- xf
    }
    if (is.null(value) || suppressWarnings(is.na(value))) {
        if (!is.null(par)) {
            Model(x)$constrain[[par]] <- NULL
            Model(x)$constrainY[[par]] <- NULL
        } else {
            Model(x)$constrain[[args]] <- NULL
        }
        return(x)
    }
    for (i in args) {
        if (!(i%in%c(parlabels(Model(x)), vars(Model(x)),
                     names(constrain(x))))) {
            if (lava.options()$messages>1)
                message("\tAdding parameter '", i, "'\n",sep="")
            parameter(x, messages=0) <- i
        }
    }
    if (par%in%vars(x) && endogenous) {
        if (!"..."%in%names(formals(value)) && !is.primitive(value)) {
            formals(value) <- c(formals(value), alist(...=))
        }
        Model(x)$constrainY[[par]] <- list(fun=value, args=args)
    } else {
        ## Wrap around do.call, since functions are not really
        ## parsed as call-by-value in R, and hence setting
        ## attributes to e.g. value=cos, will be overwritten
        ## if value=cos is used again later with new args.
        Model(x)$constrain[[par]] <- function(x) do.call(value, list(x))
        attributes(Model(x)$constrain[[par]])$args <- args
        index(Model(x)) <- reindex(Model(x))
    }
    return(x)
}

##' @export
constraints <- function(object,data=model.frame(object),vcov=object$vcov,level=0.95,
                 p=pars.default(object),k,idx,...) {
    if (class(object)[1]=="multigroupfit") {
        if (!missing(k)) {
            if (class(data)[1]=="list") data <- data[[k]]
            parpos <- modelPar(object, seq_len(length(p)))$p[[k]]
            if (nrow(data)>1 & !missing(idx)) {
                res <- t(apply(data, 1, function(x) constraints(Model(object)$lvm[[k]],data=x,p=p[parpos],vcov=vcov[parpos,parpos],level=level)[idx,]))
                return(res)
            }
            return(constraints(Model(object)$lvm[[k]],data=data,p=p[parpos],vcov=vcov[parpos,parpos],level=level))
        }
        return(attributes(CoefMat.multigroupfit(object,data=data,vcov=vcov,...))$nlincon)
    }

    if (NROW(data)>1 & !missing(idx)) {
        res <- t(apply(data,1,function(x) constraints(object,data=x,p=p,vcov=vcov,level=level)[idx,],...))
        return(res)
    }

    if (length(index(object)$constrain.par)<1) return(NULL)
    parpos <- Model(object)$parpos
    if (is.null(parpos)) {
        parpos <- with(index(object),matrices2(Model(object),seq_len(npar+npar.mean+npar.ex)))
        parpos$A[index(object)$M0==0] <- 0
        parpos$P[index(object)$P0==0] <- 0
        parpos$v[index(object)$v1==0] <- 0
        parpos$e[index(object)$e1==0] <- 0
    }
    myidx <- unlist(lapply(parpos$parval, function(x) {
        if (!is.null(attributes(x)$reg.idx)) {
            return(parpos$A[attributes(x)$reg.idx[1]])
        } else if (!is.null(attributes(x)$cov.idx)) {
            return(parpos$P[attributes(x)$cov.idx[1]])
        } else if (!is.null(attributes(x)$m.idx)) {
            return(parpos$v[attributes(x)$m.idx[1]])
        } else if (!is.null(attributes(x)$e.idx))
            return(parpos$e[attributes(x)$e.idx[1]])
        else NA
    }))
    names(myidx) <- names(parpos$parval)
    mynames <- c()
    N <- length(index(object)$constrain.par)
    if (N>0)
        res <- c()
    count <- 0
    mydata <- rbind(numeric(length(manifest(object))))
    colnames(mydata) <- manifest(object)
    data <- rbind(data)
    iname <- intersect(colnames(mydata),colnames(data))
    mydata[1,iname] <- unlist(data[1,iname])
    for (pp in index(object)$constrain.par) {
        count <- count+1
        myc <- constrain(Model(object))[[pp]]
        mycoef <- numeric(6)
        val.idx <- myidx[attributes(myc)$args]
        val.idx0 <- na.omit(val.idx)
        M <- modelVar(Model(object),p=p,data=as.data.frame(mydata))
        vals <- with(M,c(parval,constrainpar))[attributes(myc)$args]
        fval <- try(myc(unlist(vals)),silent=TRUE)
        fmat <- inherits(fval,"try-error")
        if (fmat) {
            fval <- myc(rbind(unlist(vals)))
        }
        mycoef[1] <- fval
        myc0 <- function(theta) {
            theta0 <- unlist(vals);
            theta0[!is.na(val.idx)] <- theta
            if (fmat) {
                res <- myc(rbind(theta0))
            } else {
                res <- myc(theta0)
            }
            return(res)
        }
        vals0 <- unlist(vals)[!is.na(val.idx)]
        if (length(vals0)==0)
            mycoef[2] <- NA
        else {
            if (!is.null(attributes(fval)$grad)) {
                if (fmat) {
                    Gr <- cbind(attributes(fval)$grad(rbind(unlist(vals0))))
                } else {
                    Gr <- cbind(attributes(fval)$grad(unlist(vals0)))
                }
            } else {
                if (fmat) {
                    Gr <- cbind(as.numeric(numDeriv::jacobian(myc0, unlist(vals0))))
                } else {
                    Gr <- cbind(as.numeric(numDeriv::jacobian(myc0, rbind(unlist(vals0)))))
                }
            }
            V <- vcov[val.idx0,val.idx0]
            mycoef[2] <- (t(Gr)%*%V%*%Gr)^0.5
        }
        ## if (second) {
        ##   if (!is.null(attributes(fval)$hessian)) {
        ##     H <- attributes(fval)$hessian(unlist(vals))
        ##   } else {
        ##     H <- hessian(myc, unlist(vals))
        ##   }
        ##   HV <- H%*%vcov[val.idx,val.idx]
        ##   mycoef[1] <- mycoef[1] + 0.5*sum(diag(HV))
        ##   mycoef[2] <- mycoef[2] + 0.5*sum(diag(HV%*%HV))
        ## }
        mycoef[3] <- mycoef[1]/mycoef[2]
        mycoef[4] <- 2*(pnorm(abs(mycoef[3]),lower.tail=FALSE))
        mycoef[5:6] <- mycoef[1] + c(1,-1)*qnorm((1-level)/2)*mycoef[2]
        res <- rbind(res,mycoef)
        mynames <- c(mynames,pp)
        if (!is.null(attributes(fval)$inv)){
            res2 <- attributes(fval)$inv(mycoef[c(1,5,6)])
            res <- rbind(res, c(res2[1],NA,NA,NA,res2[2],res2[3]))
            mynames <- c(mynames,paste0("inv(",pp,")"))
        }
    }
    rownames(res) <- mynames
    colnames(res) <- c("Estimate","Std. Error", "Z value", "Pr(>|z|)", "2.5%", "97.5%")
    return(res)
}

context("Multiple Group")

test_that("Multiple group I", {
  m <- lvm(y~x)
  set.seed(1)
  d <- sim(m,1000)
  ## Just a stratified analysis
  e <- estimate(list("Group A"=m,"Group B"=m), list(d,d))
  expect_true(mean((coef(e)[c(1,3)]-coef(lm(y~x,d)))^2)<1e-9)
  expect_true(mean((coef(e)[c(2,5)]-coef(lm(y~x,d)))^2)<1e-9)
  ci <- confint(e)[c(1,3),]
  ci2 <- confint(lm(y~x,d))
  expect_true(mean((ci-ci2)^2)<1e-5)
})

test_that("Multiple group II", {
  m <- baptize(lvm(y~x))
  set.seed(1)
  d <- sim(m,100)
  ## Just a standard linear regression (single group)
  e <- estimate(list(m,m),list(d,d))
  testthat::expect_identical(coef(e,type=2)[[1]],coef(e,type=2)[[2]])
  testthat::expect_true(mean((coef(e,type=2)[[1]][1:2,1]-coef(lm(y~x,cbind(d,d))))^2)<1e-9)
})

context("Missing data")

test_that("Missing data analysis", {
  ## Random intercept model
  m <- lvm(c(y1,y2,y3)~x+u); latent(m) <- ~u
  set.seed(1)
  ## Missing on first two outcomes
  d <- makemissing(sim(m,200),p=0.3,cols=c("y1","y2"))  
  e <- estimate(m,d,missing=TRUE)
  testthat::expect_true("lvm.missing"%in%class(e))
  testthat::expect_true(sum(unlist(lapply(e$estimate$model$data,nrow)))==200)
  ## Convergence:
  g <- gof(e)
  testthat::expect_true(mean(score(e))<1e-3)
  testthat::expect_true(g$rankV==length(pars(e)))
})

test_that("Multiple group, missing data analysis", {
  m <- lvm(list(c(y1,y2,y3)~u,u~x)); latent(m) <- ~u
  m <- baptize(fixsome(m))
  regression(m,u~x) <- NA
  covariance(m,~u) <- NA
  set.seed(1)
  ## Missing on all outcomes
  d1 <- makemissing(sim(m,500),cols=c("y1","y2"),p=0.3)
  d2 <- makemissing(sim(m,500),cols=c("y1","y2"),p=0.3)
  e <- estimate(list(m,m),list(d1,d2),missing=TRUE)
  g <- gof(e)
  testthat::expect_true(g$n==1000)
  testthat::expect_true(mean(score(e))<1e-3)
  testthat::expect_true(g$rankV==length(pars(e)))
})

test_that("Multiple group, constraints", {
    m1 <- lvm(y ~ f(x,beta)+f(z,beta2))
    m2 <- lvm(y ~ f(x,psi) + z)
    ## And simulate data from them
    set.seed(1)
    d1 <- sim(m1,100)
    d2 <- sim(m2,100)
    ## Add 'non'-linear parameter constraint
    constrain(m2,psi ~ beta2) <- function(x) x
    ## Add parameter beta2 to model 2, now beta2 exists in both models
    parameter(m2) <- ~ beta2
    ee <- estimate(list(m1,m2),list(d1,d2))

    m <- lvm(y1 ~ x1 + beta2*z1)
    regression(m) <- y2 ~ beta2*x2 + z2
    d <- cbind(d1,d2); names(d) <- c(paste0(names(d1),1),paste0(names(d1),2))
    e <- estimate(m,d)

    b1 <- coef(e,2,labels=TRUE)["beta2",1]
    b2 <- constraints(ee)[1]
    testthat::expect_true(mean((b1-b2)^2)<1e-5)
    
    ## "Multiple group, constraints (non-linear in x)
    m <- lvm(y[m:v] ~ 1)
    addvar(m) <- ~x
    parameter(m) <- ~a+b
    constrain(m,m~a+b+x) <- function(z) z[1]+z[2]*z[3]
    ee <- estimate(list(m,m),list(d1[1:5,],d1[6:10,]))
    b1 <- coef(lm(y~x,d1[1:10,]))
    b2 <- coef(ee)[c("a@1","b@1")]
    testthat::expect_true(mean(b1-b2)^2<1e-4)

})

test_that("Multiple cluster", {
  m <- lvm(y~x)
  set.seed(1)
  d <- sim(m,100)
  ## Just a stratified analysis
  ## e <- estimate(list("Group A"=m,"Group B"=m), list(d,d))
  l <- lm(y ~ x, data=d)
  e <- estimate(list("Group A"=m,"Group B"=m), list(d,d),
                id=list(1:nrow(d),1:nrow(d)))
  e2 <- estimate(l, id=1:nrow(d))
  expect_true(mean((vcov(e2)-vcov(e)[c(1,3),c(1,3)])^2)<1e-9)
})

with_unique_warnings <- function(expr) {
  # utility function that prevents the same warning being cast more than once
  seen <- character(0)
  withCallingHandlers(expr, warning = function(w) {
    msg <- conditionMessage(w)
    if (msg %in% seen) {
      invokeRestart("muffleWarning")
    } else {
      seen <<- c(seen, msg)
    }
  })
}

# small sample corrections to variance estimates
small_sample_correction <- function(ic, type, var.adj) {
  V <- var_ic(ic)
  ## Small-sample corrections for clustered data
  K <- NROW(ic)
  N <- attributes(ic)$N
  if (is.null(N)) N <- K
  p <- NCOL(ic)
  adj0 <- K/(K-p) ## Mancl & DeRouen, 2001
  adj1 <- K/(K-1) ## Mancl & DeRouen, 2001
  adj2 <- (N-1)/(N-p)*(K/(K-1)) ## Morel,Bokossa & Neerchal, 2003
  if (tolower(type[1])=="mbn" && !is.null(attributes(ic)$bread)) {
    V0 <- V
    iI0 <- attributes(ic)$bread
    I0 <- Inverse(iI0)
    delta <- min(0.5, p / (K - p))
    phi <- max(1, tr(I0%*%V0)*adj2/p)
    V <- adj2*V0 + delta*phi*iI0
  }
  if (tolower(type[1])=="df") {
    V <- adj0*V
  }
  if (tolower(type[1])=="df1") {
    V <- adj1*V
  }
  if (tolower(type[1])=="df2") {
    V <- adj2*V
  }
  if (tolower(type[1])%in%c("hc3", "hc4")) {
    ic <- cbind(ic)
    S <- Inverse(crossprod(ic), tol=sqrt(.Machine$double.eps))
    h_emp <- rowSums((ic %*% S) * ic) # empirical h, lev.
    n <- nrow(ic)
    ## h_emp <- pmin(h_emp, 0.99) * (n-1)/n  # Truncate leverage to prevent division by zero
    if (tolower(type[1])=="hc3") {
      ## phi_norm <- sqrt(rowSums(ic^2))
      ## ex_kurt <- (mean((phi_norm - mean(phi_norm))^4) / var(phi_norm)^2) - 3
      ## alpha <- exp(-max(0, ex_kurt) / 25)
      v.alpha <- ifelse(missing(var.adj), 0.25, var.adj)
      h <- v.alpha * h_emp + (1 - v.alpha) * (ncol(ic) / n)
      adj <- 1 / (1 - h)
    } else {
      ## Cribari-Neto (2004)
      delta <- pmin(1.5, n/ncol(ic) * h_emp)
      adj <- 1 / (1 - h_emp)**delta
    }
    for (i in seq_len(NCOL(ic))) ic[, i] <- ic[, i] * adj
    V <- var_ic(ic)
  }
  return(V)
}

# construct coefficient matrix with confidence limits and two-sided p-values
estimate_coefmat <- function(est, se, df, level = 0.95, null = 0) {
  alpha <- 1 - level
  alpha.str <- paste(c(alpha/2, 1 -alpha/2)*100, "", sep="%")
  if (!is.null(df)) {
    za <- qt(1-alpha/2, df=df)
    pval <- 2*pt(abs((est-null) / se), df=df, lower.tail=FALSE)
  } else {
    za <- qnorm(1-alpha/2)
    pval <- 2*pnorm(abs((est-null)/ se), lower.tail=FALSE)
  }
  res <- cbind(est, se, est - za * se, est + za * se, pval)
  colnames(res) <- c("Estimate", "Std.Err", alpha.str, "P-value")
  return(res)
}

#' Summary of estimate objects
#'
#' Computes hypothesis tests, contrasts, and small-sample corrections for
#' an [estimate] object. The arguments `null`, `contrast`, `type`, and
#' `var.adj` were previously available on [estimate.default()] and have
#' been moved here.
#'
#' @param object an `estimate` object.
#' @param contrast (optional) contrast matrix for the final Wald test.
#' @param null (optional) null hypothesis to test.
#' @param type type of small-sample correction. Requires the estimate to have
#'   been computed with `IC=TRUE` (the default).
#' @param var.adj variance adjustment parameter for small-sample correction.
#'   Requires the estimate to have been computed with `IC=TRUE` (the default).
#' @param df degrees of freedom for t-based inference (default: `NULL` for
#'   Gaussian approximation; when set, confidence intervals and p-values use the
#'   t-distribution with `df` degrees of freedom)
#' @param level level of confidence limits (default 0.95)
#' @param transform (optional) function applied to the point estimates and
#'   confidence interval bounds *after* inference is performed on the original
#'   scale. Useful for variance-stabilizing transformations, e.g., compute CIs
#'   on the `atanh` (Fisher z) scale and back-transform with `tanh`.
#' @param print (optional) custom print function for the resulting
#'   `summary.estimate` object
#' @param ... additional arguments passed to [contr].
#' @seealso [estimate.default()]
#' @export
summary.estimate <- function(object,
                             contrast,
                             ...,
                             null = 0,
                             level = 0.95,
                             type,
                             var.adj = 0.25,
                             df,
                             transform = NULL,
                             print = NULL) {
  with_unique_warnings({
    p <- coef(object)
    if (missing(df)) df <- object$df
    if (missing(contrast)) contrast <- diag(1, nrow=length(p))
    correction <- !missing(type)
    if (correction) {
      if (is.null(object$IC)) {
        stop(
          "Small-sample corrections in summary() require an estimate ",
          "computed with IC=TRUE."
        )
      }
      V <- small_sample_correction(IC(object), type=type, var.adj=var.adj)
    } else {
      V <- vcov(object)
    }
    ## Preserve dimnames on vcov: the recall path through type/var.adj
    ## corrections builds V from the influence function and may drop
    ## dimnames. Restore them from the original coef names so the
    ## summary output is consistent with the deprecated estimate() path.
    if (!is.null(V) && is.null(dimnames(V)) &&
        !is.null(names(p)) &&
        nrow(V) == length(p)) {
      dimnames(V) <- list(names(p), names(p))
    }

    if (is.vector(contrast) || is.list(contrast)) {
      contrast <- contr(contrast, names(object$coef), ...)
    }

    cc0 <- estimate_coefmat(p, diag(V)**.5, df=df, level=level, null=null)
    rownames(cc0) <- rownames(parameter(object))
    waldtest <- compare(object, contrast=contrast, null=null, vcov=V)
    class(object) <- "list"
    res <- c(object[c("coef", "coefmat", "vcov", "call",
                      "ncluster", "model.index")], list(compare=waldtest))
    res$coefmat <- cc0
    if (!is.null(transform)) {
      res$coefmat[, c(1, 3, 4)] <- do.call(transform,
                                           list(res$coefmat[, c(1, 3, 4)]))
      res$coefmat[, 2] <- NA
      res$vcov <- NULL
      res$coef <- res$coefmat[, 1, drop=TRUE]
    }
    res$print <- print
    class(res) <- c("summary.estimate", "estimate")
    return(res)
  })
}

#' @export
print.summary.estimate <- function(x, ...) {
  if (!is.null(x$print)) {
    x$print(x, ...)
    return(invisible(x))
  }
  print.estimate(x, type=2L, ...)
  if (!is.null(attributes(x)$extra)) {
    print(attributes(x)$extra)
  }
  return(invisible(x))
}

#' @export
coef.summary.estimate <- function(object, ...) {
  return(object$coef)
}

#' @export
confint.summary.estimate <- function(object, ...) {
  parameter(object)[, 3:4, drop=FALSE]
}

#' @export
parameter.summary.estimate <- function(x, ...) {
  return(x$coefmat)
}

#' @export
vcov.summary.estimate <- function(object, ...) {
  res <- object$vcov
  p <- coef(object)
  if (is.null(res)) {
    res <- matrix(NA, length(p), length(p))
    dimnames(res) <- list(names(p), names(p))
  }
  return(res)
}

#' Concatenate summary.estimate objects
#'
#' When all arguments are `summary.estimate` objects, they are merged into a
#' single `summary.estimate` object. When some arguments are not
#' `summary.estimate` objects but are named numeric scalars/vectors, an object
#' of class `summary.estimate` is returned with an additional `extra` attribute
#' that contains the provided numeric scalars/vectors. This is useful for
#' bundling auxiliary per-iteration information (e.g., convergence status)
#' alongside an estimate for use with [sim.default()].
#'
#' @param ... `summary.estimate` objects and/or named numeric values
#' @examples
#' e1 <- estimate(coef = 1, IC = scale(rnorm(10)), id = 1:10, labels = "a1")
#' e2 <- estimate(coef = 2, IC = scale(rnorm(10)), id = 1:10, labels = "a2")
#'
#' # concatenating two summary.estimate objects
#' c(e1, e2)
#'
#' # concatenating one summary.estimate object with one numerical variable
#' ss <- c(summary(e1), niter = 2)
#' print(ss)
#' attributes(ss)$extra
#' @export
c.summary.estimate <- function(...) {
  args <- list(...)
  if (length(args) == 1) return(args[[1]])

  mask <- sapply(args, function(x) inherits(x, "summary.estimate"))
  summary_objects <- args[mask]

  extras_existing <- unlist(
    lapply(summary_objects, function(x) attributes(x)$extra)
  )

  extras <- c(extras_existing, unlist(args[!mask]))

  .print <- function(x, digits = 4, ...) {
    cat("Concatenated summary.estimate objects: \n")
    print(cli::rule(width = min(cli::console_width(), 60)))
    print(x$coefmat, digits = digits)
    print(cli::rule(width = min(cli::console_width(), 60)))
    if (!is.null(attributes(x)$extra)) {
      print(attributes(x)$extra)
    }
  }

  res <- structure(list(
    coefmat = Reduce(rbind, lapply(summary_objects, function(x) parameter(x))),
    coef = as.vector(
      Reduce(rbind, lapply(summary_objects, function(x) coef(x)))
    ),
    vcov = cbind(Reduce(function(...) blockdiag(..., pad=NA),
                        lapply(summary_objects, function(x) vcov(x)))),
    objects = summary_objects,
    print = .print # consumed by print.summary.estimate
  ), class = c("summary.estimate", "estimate")
  )
  names(res$coef) <- rownames(res$coefmat)
  if (length(extras) > 0) attr(res, "extra") <- extras

  return(res)
}

##' Define constant risk difference or relative risk association for binary exposure
##'
##' Set up model as defined in Richardson, Robins and Wang (2017).
##' @param x model
##' @param response response variable (character or formula)
##' @param exposure exposure variable (character or formula)
##' @param target.model variable defining the linear predictor for the target model
##' @param nuisance.model variable defining the linear predictor for the nuisance model
##' @param exposure.model model for exposure (default binomial logit link)
##' @param ... additional arguments to lower level functions
##' @aliases binomial.rd binomial.rr
##' @export
binomial.rd <- function(x,response,exposure,
                 target.model,nuisance.model,
                 exposure.model=binomial.lvm(),...) {
    binomial.rrw(x,response,exposure,target.model,nuisance.model,exposure.model,type="rd",...)
}

##' @export
binomial.rr <- function(x,response,exposure,
                 target.model,nuisance.model,
                 exposure.model=binomial.lvm(),...) {
    binomial.rrw(x,response,exposure,target.model,nuisance.model,exposure.model,type="rr",...)
}


binomial.rrw <- function(x, response, exposure,
                 target.model, nuisance.model,
                 exposure.model=binomial.lvm(), type="rd", ...) {
    if (inherits(response,"formula")) {
        vars <- all.vars(response)
        if (length(vars)==1L) {
            response <- vars
        } else {
            yf <- getoutcome(response, sep="|")
            exposure  <- attr(yf,"x")[[1]]
            if (length(attr(yf,"x"))>1)
                target.model  <- attr(yf,"x")[[2]]
            if (length(attr(yf,"x"))>2)
                nuisance.model  <- attr(yf,"x")[[3]]
            response <- yf[1]
        }
    }
    if (inherits(exposure,"formula")) exposure <- all.vars(exposure)
    if (inherits(target.model,"formula")) target.model <- all.vars(target.model)
    if (inherits(nuisance.model,"formula")) nuisance.model <- all.vars(nuisance.model)

    if (type=="rd") {
        val <- list(list(input=c(exposure,target.model,nuisance.model),
                         fun=simulate_binomial_rd,
                         type="Binomial regression (exposure | risk-difference | odds-product)"))
    } else {
        val <- list(list(input=c(exposure,target.model,nuisance.model),
                         fun=simulate_binomial_rr,
                         type="Binomial regression (exposure | relative-risk | odds-product)"))
    }
    if (is.null(distribution(x)[[exposure]]))
        distribution(x, exposure) <- binomial.lvm(link="logit")
    covariance(x,c(target.model,nuisance.model)) <- 0
    distribution(x,exposure) <- exposure.model
    names(val) <- response
    x$attributes$multiple.inputs <- val
    return(x)
}


simulate_binomial_rd <- function(x,data,inputs,...) {
    exposure <- data[,inputs[1]]
    lp1 <- data[,inputs[2]]
    lp2 <- data[,inputs[3]]
    rd <- tanh(lp1)
    op <- exp(lp2)
    pp <- RD_OP(rd,op)
    pp <- pp[,1]*(1-exposure) + pp[,2]*exposure
    y <- rbinom(NROW(data), 1, pp)
}

simulate_binomial_rr <- function(x,data,inputs,...) {
    exposure <- data[,inputs[1]]
    lp1 <- data[,inputs[2]]
    lp2 <- data[,inputs[3]]
    rr <- exp(lp1)
    op <- exp(lp2)
    pp <- RR_OP(rr,op)
    pp <- pp[,1]*(1-exposure) + pp[,2]*exposure
    y <- rbinom(NROW(data), 1, pp)
}


##' @export
Identical <- function(x,y=1,tolerance = .Machine$double.eps^0.5) {
    Mod(x-y)<tolerance
}

RD_OP <- function(rd,op) {
    a <- op-1
    b <- -op*(2-rd)-rd
    p0 <- (-b - sqrt(b^2 - 4*op*(1-rd)*a))/(2*a)
    op1 <- which(sapply(op,function(x) Identical(x,1)))
    if (length(op1)>0)
        p0[op1] = 0.5 * (1 - rd[op1])
    p1 <- p0 + rd
    cbind(p0,p1)
}


RR_OP <- function(rr,op) {
    b <- op*(1+rr)
    a <- rr*(1-op)
    p0 <- (-b + sqrt(b^2 + 4*a*op))/(2*a)
    op1 <- which(sapply(op,function(x) Identical(x,1)))
    if (length(op1)>0)
        p0[op1] = 1/(1+rr)
    p1 <- p0*rr
    cbind(p0,p1)
}

context("Binomial risk-difference / relative-risk models (binomial.rrw)")

## ---------------------------------------------------------------------------
## Identical()
## ---------------------------------------------------------------------------
test_that("Identical: numerical equality with default tolerance", {
    testthat::expect_true(Identical(1, 1))
    testthat::expect_true(Identical(1 + 1e-12, 1))
    testthat::expect_false(Identical(1 + 1e-3, 1))
    ## vector input
    res <- Identical(c(1, 2, 1 + 1e-12), 1)
    testthat::expect_equal(res, c(TRUE, FALSE, TRUE))
    ## custom tolerance
    testthat::expect_true(Identical(1.001, 1, tolerance = 1e-2))
})

## ---------------------------------------------------------------------------
## RD_OP: round-trip check
## ---------------------------------------------------------------------------
test_that("RD_OP recovers (p0, p1) from (rd, op)", {
    RD_OP <- lava:::RD_OP
    p0 <- c(0.2, 0.4, 0.6)
    p1 <- c(0.3, 0.5, 0.8)
    rd <- p1 - p0
    ## op is the odds *product*: (p0/(1-p0)) * (p1/(1-p1))
    op <- (p0 / (1 - p0)) * (p1 / (1 - p1))
    out <- RD_OP(rd, op)
    testthat::expect_equal(dim(out), c(3L, 2L))
    testthat::expect_equal(out[, 1], p0, tolerance = 1e-8)
    testthat::expect_equal(out[, 2], p1, tolerance = 1e-8)
})

test_that("RD_OP handles op == 1 (no association) branch", {
    RD_OP <- lava:::RD_OP
    rd <- c(0.1, -0.2)
    op <- c(1, 1)
    out <- RD_OP(rd, op)
    ## When op == 1: p0 = 0.5*(1 - rd), p1 = p0 + rd
    testthat::expect_equal(out[, 1], 0.5 * (1 - rd), tolerance = 1e-10)
    testthat::expect_equal(out[, 2], 0.5 * (1 - rd) + rd, tolerance = 1e-10)
    testthat::expect_true(all(is.finite(out)))
})

## ---------------------------------------------------------------------------
## RR_OP: round-trip check
## ---------------------------------------------------------------------------
test_that("RR_OP recovers (p0, p1) from (rr, op)", {
    RR_OP <- lava:::RR_OP
    p0 <- c(0.2, 0.3, 0.5)
    p1 <- c(0.3, 0.45, 0.6)
    rr <- p1 / p0
    ## op is the odds *product*: (p0/(1-p0)) * (p1/(1-p1))
    op <- (p0 / (1 - p0)) * (p1 / (1 - p1))
    out <- RR_OP(rr, op)
    testthat::expect_equal(dim(out), c(3L, 2L))
    testthat::expect_equal(out[, 1], p0, tolerance = 1e-8)
    testthat::expect_equal(out[, 2], p1, tolerance = 1e-8)
})

test_that("RR_OP handles op == 1 (no association) branch", {
    RR_OP <- lava:::RR_OP
    rr <- c(1.5, 2)
    op <- c(1, 1)
    out <- RR_OP(rr, op)
    ## When op == 1: p0 = 1/(1+rr), p1 = p0 * rr
    testthat::expect_equal(out[, 1], 1 / (1 + rr), tolerance = 1e-10)
    testthat::expect_equal(out[, 2], rr / (1 + rr), tolerance = 1e-10)
    testthat::expect_true(all(is.finite(out)))
})

## ---------------------------------------------------------------------------
## binomial.rd model modifier
## ---------------------------------------------------------------------------
test_that("binomial.rd installs multiple-input simulation node", {
    m <- lvm()
    regression(m) <- z ~ x
    regression(m) <- lp ~ x
    regression(m) <- op ~ x
    m <- binomial.rd(m, response = "y", exposure = "z",
                     target.model = "lp", nuisance.model = "op")
    mi <- m$attributes$multiple.inputs
    testthat::expect_true(!is.null(mi))
    testthat::expect_true("y" %in% names(mi))
    testthat::expect_equal(mi$y$input, c("z", "lp", "op"))
    testthat::expect_true(grepl("risk-difference", mi$y$type, fixed = TRUE))
    ## exposure should now have a binomial distribution attached
    testthat::expect_false(is.null(distribution(m)[["z"]]))
})

test_that("binomial.rd accepts formula interface", {
    m1 <- lvm()
    regression(m1) <- z ~ x
    regression(m1) <- lp ~ x
    regression(m1) <- op ~ x
    m1 <- binomial.rd(m1, response = "y", exposure = "z",
                      target.model = "lp", nuisance.model = "op")

    m2 <- lvm()
    regression(m2) <- z ~ x
    regression(m2) <- lp ~ x
    regression(m2) <- op ~ x
    m2 <- binomial.rd(m2, y ~ z | lp | op)

    testthat::expect_equal(m1$attributes$multiple.inputs$y$input,
                           m2$attributes$multiple.inputs$y$input)
    testthat::expect_equal(m1$attributes$multiple.inputs$y$type,
                           m2$attributes$multiple.inputs$y$type)
})

## ---------------------------------------------------------------------------
## binomial.rr model modifier
## ---------------------------------------------------------------------------
test_that("binomial.rr installs multiple-input simulation node", {
    m <- lvm()
    regression(m) <- z ~ x
    regression(m) <- lp ~ x
    regression(m) <- op ~ x
    m <- binomial.rr(m, response = "y", exposure = "z",
                     target.model = "lp", nuisance.model = "op")
    mi <- m$attributes$multiple.inputs
    testthat::expect_true("y" %in% names(mi))
    testthat::expect_equal(mi$y$input, c("z", "lp", "op"))
    testthat::expect_true(grepl("relative-risk", mi$y$type, fixed = TRUE))
})

## ---------------------------------------------------------------------------
## End-to-end simulation: numerical correctness
## ---------------------------------------------------------------------------
test_that("binomial.rd simulation produces target risk difference", {
    testthat::skip_on_cran()
    ## Build model where lp is constant (intercept-only) so the implied RD
    ## is tanh(lp_const).  We set lp via constrain().
    m <- lvm()
    regression(m) <- z ~ x
    distribution(m, ~x) <- normal.lvm()
    distribution(m, ~lp) <- normal.lvm(mean = 0, sd = 0)  ## degenerate -> 0
    distribution(m, ~op) <- normal.lvm(mean = 0, sd = 0)  ## op = exp(0) = 1
    intercept(m, ~lp) <- 0.4   ## constant linear predictor for RD
    intercept(m, ~op) <- 0     ## odds product = 1
    m <- binomial.rd(m, response = "y", exposure = "z",
                     target.model = "lp", nuisance.model = "op")
    set.seed(123)
    d <- sim(m, n = 8000)
    testthat::expect_true(all(d$y %in% c(0, 1)))
    rd_emp <- mean(d$y[d$z == 1]) - mean(d$y[d$z == 0])
    target <- tanh(0.4)
    ## SE roughly sqrt(p*(1-p)/n_per_group) ~ 0.01; allow generous 3x
    testthat::expect_lt(abs(rd_emp - target), 0.04)
})

test_that("binomial.rr simulation produces target relative risk", {
    testthat::skip_on_cran()
    m <- lvm()
    regression(m) <- z ~ x
    distribution(m, ~x) <- normal.lvm()
    distribution(m, ~lp) <- normal.lvm(mean = 0, sd = 0)
    distribution(m, ~op) <- normal.lvm(mean = 0, sd = 0)
    intercept(m, ~lp) <- log(1.5)  ## constant log-RR
    intercept(m, ~op) <- 0
    m <- binomial.rr(m, response = "y", exposure = "z",
                     target.model = "lp", nuisance.model = "op")
    set.seed(456)
    d <- sim(m, n = 8000)
    testthat::expect_true(all(d$y %in% c(0, 1)))
    p1 <- mean(d$y[d$z == 1])
    p0 <- mean(d$y[d$z == 0])
    rr_emp <- p1 / p0
    testthat::expect_lt(abs(log(rr_emp) - log(1.5)), 0.10)
})

## ---------------------------------------------------------------------------
## Bonus: riskcomp (lives in R/assoc.R but logically related)
## ---------------------------------------------------------------------------
test_that("riskcomp computes risk difference / risk ratio for two probabilities", {
    p <- c(0.4, 0.6)
    ## Default op="/", type=1: val[1]/val[2]
    testthat::expect_equal(riskcomp(p[1], p[2]), p[1] / p[2])
    ## Explicit Diff (op="-")
    testthat::expect_equal(Diff(p[1], p[2]), p[1] - p[2])
    ## Explicit Ratio (op="/")
    testthat::expect_equal(Ratio(p[1], p[2]), p[1] / p[2])
    ## type=2 swaps operand order
    testthat::expect_equal(riskcomp(p[1], p[2], type = 2), p[2] / p[1])
})

test_that("riskcomp with scale=odds gives odds ratio", {
    p <- c(0.4, 0.6)
    o <- p / (1 - p)
    testthat::expect_equal(riskcomp(p[1], p[2], scale = odds), o[1] / o[2])
    ## type=2 swaps order
    testthat::expect_equal(riskcomp(p[1], p[2], scale = odds, type = 2),
                           o[2] / o[1])
})

## ---------------------------------------------------------------------------
## Regression: simulate_binomial_* errors clearly on out-of-range probabilities
## ---------------------------------------------------------------------------
test_that("simulate_binomial_rd errors when implied probability is out of [0,1]", {
    sim_rd <- lava:::simulate_binomial_rd
    ## NaN linear predictor propagates to NaN probabilities -> guard triggers
    data <- data.frame(z = c(0, 1), lp = c(NaN, NaN), op = c(0, 0))
    testthat::expect_error(sim_rd(NULL, data, inputs = c("z", "lp", "op")),
                           regexp = "outside")
})

test_that("simulate_binomial_rr errors when implied probability is out of [0,1]", {
    sim_rr <- lava:::simulate_binomial_rr
    data <- data.frame(z = c(0, 1), lp = c(NaN, NaN), op = c(0, 0))
    testthat::expect_error(sim_rr(NULL, data, inputs = c("z", "lp", "op")),
                           regexp = "outside")
})

###{{{ estimate.multigroup

##' @export
`estimate.multigroup` <- function(x, control=list(),
                           estimator=NULL,
                           weights, weightsname,
                           data2,
                           id=NULL,
                           messages=lava.options()$messages,
                           quick=FALSE,
                           param,
                           cluster,
                           ...) {
    cl <- match.call()
    Optim <- list(
        iter.max=lava.options()$iter.max,
        trace=ifelse(lava.options()$debug,3,0),
        gamma=lava.options()$gamma,
        ngamma=lava.options()$ngamma,
        backtrace=TRUE,
        gamma2=1,
        lambda=0.05,
        abs.tol=1e-9,
        epsilon=1e-10,
        delta=1e-10,
        S.tol=1e-6,
        stabil=FALSE,
        start=NULL,
        constrain=lava.options()$constrain,
        method=NULL,
        starterfun=startvalues0,
        information="E",
        meanstructure=TRUE,
        sparse=FALSE,
        lbound=1e-9,
        reindex=FALSE,
        tol=lava.options()$tol)


    if (!missing(param)) {
        oldparam <- lava.options()$param
        lava.options(param=param)
        on.exit(lava.options(param=oldparam))
    }
    if (!missing(cluster)) id <- cluster

    defopt <- lava.options()[]
    defopt <- defopt[intersect(names(defopt),names(Optim))]
    Optim[names(defopt)] <- defopt

    if (length(control)>0) {
        Optim[names(control)] <- control
    }


    Debug("Start values...")
    if (!is.null(Optim$start) & length(Optim$start)==(x$npar+x$npar.mean)) {
        mystart <- Optim$start
    } else {
        if (messages>1) cat("Obtaining starting value...")
        if (is.null(control$starterfun) && lava.options()$param!="relative")
            Optim$starterfun <- startvalues0
        mystart <- with(Optim, starter.multigroup(x,meanstructure=meanstructure,starterfun=starterfun,messages=messages,fix=FALSE))
        if (!is.null(Optim$start)) {
            pname <- names(Optim$start)
            ppos <- parpos.multigroup(x,p=pname,mean=TRUE)
            if (any(!is.na(ppos)))
                mystart[ppos] <- Optim$start[na.omit(match(attributes(ppos)$name,pname))]
        }
        if (messages>1) cat("\n")
    }
    Debug(mystart)
    Debug("Constraints...")
    ## Setup optimization constraints
    lower <- rep(-Inf, x$npar);
    for (i in seq_len(x$ngroup)) {
        vpos <- sapply(x$parlist[[i]][variances(x$lvm[[i]],mean=FALSE)], function(y) char2num(substr(y,2,nchar(y))))
        if (length(vpos)>0)
            lower[vpos] <- Optim$lbound
    }
    if (Optim$meanstructure)
        lower <- c(rep(-Inf,x$npar.mean), lower)
    if (any(Optim$constrain)) {
        if (length(Optim$constrain)!=length(lower))
            constrained <- is.finite(lower)
        else
            constrained <- Optim$constrain
        constrained <- which(constrained)
        lower[] <- -Inf
        Optim$constrain <- TRUE
        mystart[constrained] <- log(mystart[constrained])
    }

    if (!missing(weights)) {
        if (is.character(weights)) {
            stweights <- weights
            weights <- list()
            for (i in seq_along(x$data)) {
                newweights <- as.matrix(x$data[[i]][,stweights])
                colnames(newweights) <- index(x$lvm[[i]])$endogenous[seq_len(ncol(newweights))]
                weights <- c(weights, list(newweights))
            }
        }
    } else {
        weights <- NULL
    }
    if (!missing(data2)) {
        if (is.character(data2)) {
            stdata2 <- data2
            data2 <- list()
            for (i in seq_along(x$data)) {
                newdata <- as.matrix(x$data[[i]][,stdata2,drop=FALSE])
                dropcol <- apply(newdata,2,function(x) any(is.na(x)))
                newdata <- newdata[,!dropcol,drop=FALSE]
                colnames(newdata) <- index(x$lvm[[i]])$endogenous[seq_len(ncol(newdata))]
                data2 <- c(data2, list(newdata))
            }
        }
    } else {
        data2 <- NULL
    }

    ### Run hooks (additional lava plugins)
    myhooks <- gethook()
    newweights <- list()
    newdata2 <- list()
    newoptim <- newestimator <- NULL
    for (f in myhooks) {
        for ( i in seq_len(x$ngroup)) {
            res <- do.call(f, list(x=x$lvm[[i]],data=x$data[[i]],weights=weights[[i]],data2=data2[[i]],estimator=estimator,optim=Optim))
            if (!is.null(res$x)) x$lvm[[i]] <- res$x
            if (!is.null(res$data)) x$data[[i]] <- res$data
            if (!is.null(res$weights)) newweights <- c(newweights,list(res$weights))
            if (!is.null(res$data2)) newdata2 <- c(newdata2,list(res$data2))
            if (!is.null(res$optim)) newoptim <- res$optim
            if (!is.null(res$estimator)) newestimator <- res$estimator
        }
        if (!is.null(newestimator)) estimator <- newestimator
        if (!is.null(newoptim)) Optim <- newoptim
        if (!is.null(res$weights))
            if (!any(unlist(lapply(newweights,is.null)))) {
                weights <- newweights
            }
        if (!is.null(res$data2))
            if (!any(unlist(lapply(newdata2,is.null)))) {
                data2 <- newdata2
            }
    }
    if (is.null(estimator)) {
        if (!missing(weights) && !is.null(weights)) {
            estimator <- "normal"
        } else estimator <- "gaussian"
    }

    checkestimator <- function(x,...) {
        ffname <- paste0(x,c("_objective","_gradient"),".lvm")
        exists(ffname[1])||exists(ffname[2])
    }
    if (!checkestimator(estimator)) { ## Try down/up-case version
        estimator <- tolower(estimator)
        if (!checkestimator(estimator)) {
            estimator <- toupper(estimator)
        }
    }

    Method <-  paste0(estimator, "_method", ".lvm")
    if (!exists(Method))
        Method <- "nlminb1"
    else
        Method <- get(Method)
    if (is.null(Optim$method)) {
        Optim$method <- Method
    }

    ## Check for random slopes
    Xfix <- FALSE
    Xconstrain <- FALSE
    xfix <- list()
    for (i in seq_len(x$ngroup)) {
        x0 <- x$lvm[[i]]
        data0 <- x$data[[i]]
        xfix0 <- colnames(data0)[(colnames(data0)%in%parlabels(x0,exo=TRUE))]
        xconstrain0 <- intersect(unlist(lapply(constrain(x0),function(z) attributes(z)$args)),manifest(x0))
        xfix <- c(xfix, list(xfix0))
        if (length(xfix0)>0) Xfix<-TRUE ## Yes, random slopes
        if (length(xconstrain0)>0) Xconstrain <- TRUE ## Yes, nonlinear regression
    }

    ## Non-linear parameter constraints involving observed variables? (e.g. nonlinear regression)
    XconstrStdOpt <- TRUE
    xconstrainM <- TRUE
    xconstrain <- c()
    if (Xconstrain)
        for (i in seq_len(x$ngroup)) {
            x0 <- x$lvm[[i]]
            data0 <- x$data[[i]]
            constr0 <- lapply(constrain(x0), function(z)(attributes(z)$args))
            xconstrain0 <- intersect(unlist(constr0), manifest(x0))
            xconstrain <- c(xconstrain, list(xconstrain0))
            if (length(xconstrain0)>0) {
                constrainM0 <- names(constr0)%in%unlist(x0$mean)
                for (i in seq_len(length(constr0))) {
                    if (!constrainM0[i]) {
                        if (xconstrain0%in%constr0[[i]]) {
                            xconstrainM <- FALSE
                        }
                    }
                }
                if (xconstrainM & ((is.null(control$method) || Optim$method=="nlminb0") & (lava.options()$test & estimator=="gaussian")) ) {
                    XconstrStdOpt <- FALSE
                    Optim$method <- "nlminb0"
                    if (is.null(control$constrain)) control$constrain <- TRUE
                }
            }
        }

    ## Define objective function and first and second derivatives
    ObjectiveFun  <- paste0(estimator, "_objective", ".lvm")
    GradFun  <- paste0(estimator, "_gradient", ".lvm")
    if (!exists(ObjectiveFun) & !exists(GradFun)) stop("Unknown estimator.")

    InformationFun <- paste0(estimator, "_hessian", ".lvm")

    parord <- modelPar(x,seq_len(with(x,npar+npar.mean)))$p
    mymodel <- x

    parkeep <- c()
    myclass <- c("multigroupfit","lvmfit")
    myfix <- list()

    if (Xfix |  (Xconstrain & XconstrStdOpt | !lava.options()$test)) { ## Model with random slopes:
        #############################################################

        if (Xfix) {
            myclass <- c(myclass,"lvmfit.randomslope")
            for (k in seq_len(x$ngroup)) {
                x1 <- x0 <- x$lvm[[k]]
                data0 <- x$data[[k]]

                nrow <- length(vars(x0))
                xpos <- lapply(xfix[[k]],function(y) which(regfix(x0)$labels==y))
                colpos <- lapply(xpos, function(y) ceiling(y/nrow))
                rowpos <- lapply(xpos, function(y) (y-1)%%nrow+1)
                myfix0 <- list(var=xfix[[k]], col=colpos, row=rowpos)
                myfix <- c(myfix, list(myfix0))

                for (i in seq_along(myfix0$var))
                    for (j in seq_along(myfix0$col[[i]]))
                        regfix(x0,
                               from=vars(x0)[myfix0$row[[i]][j]],to=vars(x0)[myfix0$col[[i]][j]]) <-
                            colMeans(data0[,myfix0$var[[i]],drop=FALSE],na.rm=TRUE)
                index(x0) <- reindex(x0,zeroones=TRUE,deriv=TRUE)
                x$lvm[[k]] <- x0
                parkeep <- c(parkeep, parord[[k]][coef(x1,mean=TRUE,fix=FALSE)%in%coef(x0,mean=TRUE,fix=FALSE)])
            }
            parkeep <- sort(unique(parkeep))
            ## Alter start-values:

            if (length(mystart)!=length(parkeep))
                mystart <- mystart[parkeep]
            lower <- lower[parkeep]
            x <- multigroup(x$lvm,x$data,fix=FALSE,exo.fix=FALSE)
        }

        parord <- modelPar(x,seq_along(mystart))$p
        mydata <- list()
        for (i in seq_len(x$ngroup)) {
            mydata <- c(mydata, list(as.matrix(x$data[[i]][,manifest(x$lvm[[i]])])))
        }

        myObj <- function(theta) {
            if (Optim$constrain)
                theta[constrained] <- exp(theta[constrained])
            pp <- modelPar(x,theta)$p
            res <- 0
            for (k in seq_len(x$ngroup)) {
                x0 <- x$lvm[[k]]
                data0 <- x$data[[k]]
                if (Xfix) {
                    xfix0 <- xfix[[k]]
                    myfix0 <- myfix[[k]]
                }
                p0 <- pp[[k]]
                myfun <- function(ii) {
                    if (Xfix)
                        for (i in seq_along(myfix0$var)) {
                            x0$fix[cbind(myfix0$row[[i]],myfix0$col[[i]])] <-
                                index(x0)$A[cbind(myfix0$row[[i]],myfix0$col[[i]])] <-
                                            data0[ii,xfix0[i]]
                        }
                    if (is.list(data2[[k]][ii,])) {
                        res <- do.call(ObjectiveFun, list(x=x0, p=p0,
                                                          data=data0[ii,manifest(x0),drop=FALSE],
                                                          n=1, S=NULL, weights=weights[[k]][ii,],
                                                          data2=data2[[k]]))

                    } else {
                        res <- do.call(ObjectiveFun, list(x=x0, p=p0,
                                                          data=data0[ii,manifest(x0),drop=FALSE],
                                                          n=1, S=NULL, weights=weights[[k]][ii,],
                                                          data2=data2[[k]][ii,]))
                    }
                    return(res)
                }
                res <- res + sum(sapply(seq_len(nrow(mydata[[k]])),myfun))
            }
            res
        }

        myGrad <- function(theta) {
            if (Optim$constrain) {
                theta[constrained] <- exp(theta[constrained])
            }
            pp <- modelPar(x,theta)$p
            D0 <- res <- rbind(numeric(length(mystart)))
            for (k in seq_len(x$ngroup)) {
                if (Xfix) {
                    myfix0 <- myfix[[k]]
                }
                x0 <- x$lvm[[k]]
                myfun <- function(ii) {
                    if (Xfix)
                        for (i in seq_along(myfix0$var)) {
                            x0$fix[cbind(myfix0$row[[i]],myfix0$col[[i]])] <-
                                index(x0)$A[cbind(myfix0$row[[i]],myfix0$col[[i]])] <-
                                            x$data[[k]][ii,xfix[[k]][i]]
                        }
                    if (is.list(data2[[k]][ii,])) {

                    } else {
                        val <- do.call(GradFun, list(x=x0, p=pp[[k]],
                                                     data=mydata[[k]][ii,,drop=FALSE], n=1,
                                                     S=NULL,
                                                     weights=weights[[k]][ii,],
                                                     data2=data2[[k]][ii,]))
                    }
                    return(val)
                }
                D <- D0; D[parord[[k]]] <- rowSums(sapply(seq_len(nrow(mydata[[k]])),myfun))
                res <- res+D
            }
            if (Optim$constrain) {
                res[constrained] <- res[constrained]*theta[constrained]
            }
            return(as.vector(res))
        }

        myInformation <- function(theta) {
            theta0 <- theta
            if (Optim$constrain) {
                theta[constrained] <- exp(theta[constrained])
            }
            pp <- modelPar(x,theta)$p
            I0 <- res <- matrix(0,length(theta),length(theta))
            grad <- grad0 <- numeric(length(theta))
            for (k in seq_len(x$ngroup)) {
                x0 <- x$lvm[[k]]
                if (Xfix) {
                    myfix0 <- myfix[[k]]
                }
                myfun <- function(ii) {
                    if (Xfix)
                        for (i in seq_along(myfix0$var)) {
                            x0$fix[cbind(myfix0$row[[i]],myfix0$col[[i]])] <- index(x0)$A[cbind(myfix0$row[[i]],myfix0$col[[i]])] <-
                                                                                          x$data[[k]][ii,xfix[[k]][i]]
                        }
                    I <- I0
                    J <- do.call(InformationFun,
                                 list(x=x0, p=pp[[k]],
                                      data=mydata[[k]][ii,], n=1,
                                      S=NULL,
                                      weights=weights[[k]][ii,],
                                      data2=data2[[k]][ii,],
                                      type=Optim$information
                                      )
                                 )
                    D <- grad0
                    if (!is.null(attributes(J)$grad)) {
                        D[ parord[[k]] ] <- attributes(J)$grad
                        attributes(I)$grad <- D
                    }
                    I[ parord[[k]], parord[[k]] ] <- J
                    return(I)
                }
                L <- lapply(seq_len(nrow(x$data[[k]])),function(x) myfun(x))
                if (!is.null(attributes(L[[1]])$grad))
                    grad <- grad + rowSums(matrix((unlist(lapply(L,function(x) attributes(x)$grad))),ncol=length(L)))
                res <- res + apply(array(unlist(L),dim=c(length(theta),length(theta),nrow(x$data[[k]]))),c(1,2),sum)
            }
            if (!is.null(attributes(L[[1]])$grad))
                attributes(res)$grad <- grad
            return(res)
        }
    } else { ## Model without random slopes:
        ###########################################################


        ## Non-linear parameter constraints involving observed variables? (e.g. nonlinear regression)
        yconstrain <- c()
        iconstrain <- c()
        xconstrain <- c()
        for (j in seq_len(x$ngroup)) {
            x0 <- x$lvm[[j]]
            data0 <- x$data[[j]]
            xconstrain0 <- c()
            for (i in seq_len(length(constrain(x0)))) {
                z <- constrain(x0)[[i]]
                xx <- intersect(attributes(z)$args,manifest(x0))
                if (length(xx)>0) {
                    warg <- setdiff(attributes(z)$args,xx)
                    wargidx <- which(attributes(z)$args%in%warg)
                    exoidx <- which(attributes(z)$args%in%xx)
                    parname <- names(constrain(x0))[i]
                    y <- names(which(unlist(lapply(intercept(x0),function(x) x==parname))))
                    el <- list(i,y,parname,xx,exoidx,warg,wargidx,z)
                    names(el) <- c("idx","endo","parname","exo","exoidx","warg","wargidx","func")
                    xconstrain0 <- c(xconstrain0,list(el))
                }
            }
            yconstrain0 <- unlist(lapply(xconstrain0,function(x) x$endo))
            iconstrain0 <- unlist(lapply(xconstrain0,function(x) x$idx))
            xconstrain <- c(xconstrain, list(xconstrain0))
            yconstrain <- c(yconstrain, list(yconstrain0))
            iconstrain <- c(iconstrain, list(iconstrain0))
        }

        MkOffset <- function(pp,x,data,xconstrain,grad=FALSE) {
            if (length(xconstrain)>0) {
                Mu <- matrix(0,nrow(data),length(vars(x))); colnames(Mu) <- vars(x)
                M <- modelVar(x,p=pp,data=data)
                M$parval <- c(M$parval,  x$mean[unlist(lapply(x$mean,is.numeric))])
                for (i in seq_len(length(xconstrain))) {
                    pp <- unlist(M$parval[xconstrain[[i]]$warg]);
                    myidx <- with(xconstrain[[i]],order(c(wargidx,exoidx)))
                    mu <- with(xconstrain[[i]],
                              apply(data[,exo,drop=FALSE],1,
                                    function(x) func(
                                             unlist(c(pp,x))[myidx])))
                    Mu[,xconstrain[[i]]$endo] <- mu
                }
                offsets <- Mu%*%t(M$IAi)[,endogenous(x)]
                return(offsets)
            }
            return(NULL)
        }


        myObj <- function(theta) {
            theta0 <- theta
            if (Optim$constrain) {
                theta[constrained] <- exp(theta[constrained])
            }
            pp <- modelPar(x,theta)$p
            res <- c()
            for (i in seq_len(x$ngroup)) {
                offset <- MkOffset(pp[[i]],x$lvm[[i]],x$data[[i]],xconstrain[[i]])
                x0 <- x$lvm[[i]]
                data0 <- x$data[[i]][,index(x$lvm[[i]])$manifest,drop=FALSE]
                S <- x$samplestat[[i]]$S
                mu <- x$samplestat[[i]]$mu
                n <- x$samplestat[[i]]$n
                if (!is.null(offset)) {
                    x0$constrain[iconstrain[[i]]] <- NULL
                    pd <- procdata.lvm(x0,data0[,endogenous(x0),drop=FALSE]-offset)
                    S[endogenous(x0),endogenous(x0)] <- pd$S
                    mu[endogenous(x0)] <- pd$mu
                    n <- pd$n
                    x0$mean[yconstrain[[i]]] <- 0
                }
                res <- c(res,
                         do.call(ObjectiveFun, list(x=x0, p=pp[[i]], data=data0, S=S, mu=mu, n=n, weights=weights[[i]], data2=data2[[i]], offset=offset)))

            }
            sum(res)
        }

        if (!exists(GradFun)) {
            myGrad <- NULL
        } else  {
            myGrad <- function(theta) {
                theta0 <- theta
                if (Optim$constrain) {
                    theta[constrained] <- exp(theta[constrained])
                }
                pp <- modelPar(x,theta)$p
                D0 <- res <- rbind(numeric(length(theta)))
                for (i in seq_len(x$ngroup)) {
                    repval <- with(x$samplestat[[i]],
                                   do.call(GradFun, list(x=x$lvm[[i]],p=pp[[i]],
                                                         data=x$data[[i]][,index(x$lvm[[i]])$manifest,drop=FALSE],
                                                         S=S,mu=mu,n=n,
                                                         weights=weights[[i]], data2=data2[[i]])))
                    D <- D0; D[ parord[[i]] ] <- repval
                    res <- res + D
                }
                if (Optim$constrain) {
                    res[constrained] <- res[constrained]*theta[constrained]
                }
                return(as.vector(res))
            }
        }

        myInformation <- function(theta) {
            theta0 <- theta
            if (Optim$constrain) {
                theta[constrained] <- exp(theta[constrained])
            }
            pp <- modelPar(x,theta)$p
            I0 <- res <- matrix(0,length(theta),length(theta))
            for (i in seq_len(x$ngroup)) {
                I <- I0;
                I[ parord[[i]], parord[[i]] ] <- with(x$samplestat[[i]], do.call(InformationFun, list(p=pp[[i]], x=x$lvm[[i]], data=x$data[[i]],
                                                                                                      S=S, mu=mu, n=n, weights=weights[[i]],
                                                                                                      data2=data2[[i]],
                                                                                                      type=Optim$information)))
                res <- res + I
            }
            D <- myGrad(theta0)
            if (Optim$constrain) {
                res[constrained,-constrained] <- apply(res[constrained,-constrained,drop=FALSE],2,function(x) x*theta[constrained]);
                res[-constrained,constrained] <- t(res[constrained,-constrained])
                if (sum(constrained)==1) {
                    res[constrained,constrained] <- res[constrained,constrained]*outer(theta[constrained],theta[constrained]) - (D[constrained])
                } else {
                    res[constrained,constrained] <- res[constrained,constrained]*outer(theta[constrained],theta[constrained]) - diag(D[constrained],nrow=length(constrained))
                }
            }
            attributes(res)$grad <- D
            return(res)
        }
    }

    ##############################################################


    if (!exists(InformationFun)) myInformation <- NULL
    else if (is.null(get(InformationFun))) myInformation <- NULL
    if (is.null(get(GradFun))) myGrad <- NULL

    if (messages>1) cat("Optimizing objective function...\n")
    if (lava.options()$debug) {
        print(lower)
        print(Optim$constrain)
        print(Optim$method)
    }
    opt <- do.call(Optim$method,
                   list(start=mystart, objective=myObj, gradient=myGrad, hessian=myInformation, lower=lower, control=Optim))

    opt$estimate <- opt$par
    if (Optim$constrain) {
        opt$estimate[constrained] <- exp(opt$estimate[constrained])
    }
    if (quick) return(list(opt=opt,vcov=NA))

    if (is.null(myGrad) | !XconstrStdOpt ) {
        ## if (!requireNamespace("numDeriv")) {
        ##   opt$gradient <- naiveGrad(myObj, opt$estimate)
        ## } else {
        opt$gradient <- numDeriv::grad(myObj, opt$par, method=lava.options()$Dmethod)
    } else {
        opt$gradient <- myGrad(opt$estimate)
    }

    if (!is.null(opt$convergence)) {
        if (opt$convergence!=0) warning("Lack of convergence. Increase number of iteration or change starting values.")
    } else if (!is.null(opt$gradient) && mean(opt$gradient)^2>1e-3) warning("Lack of convergence. Increase number of iteration or change starting values.")

    if (!XconstrStdOpt) {
        myInformation <- function(theta) information(x,p=theta)
    } else {
        if (is.null(myInformation)) {
            ##     if (!requireNamespace("numDeriv")) stop("I do not know how to calculate the asymptotic variance of this estimator.
            ## For numerical approximation please install the library 'numDeriv'.")
            if (!is.null(myGrad) & XconstrStdOpt)
                myInformation <- function(theta) numDeriv::jacobian(myGrad, theta, method=lava.options()$Dmethod)
            else {
                myInformation <- function(theta) numDeriv::hessian(myObj, theta)
            }
        }
    }
    I <- myInformation(opt$estimate)
    asVar <- tryCatch(Inverse(I),
                      error=function(e) matrix(NA, length(mystart), length(mystart)))

    res <- list(model=x, model0=mymodel, call=cl, opt=opt, meanstructure=Optim$meanstructure,
                vcov=asVar, estimator=estimator, weights=weights, data2=data2, cluster=id)
    class(res) <- myclass

    myhooks <- gethook("post.hooks")
    for (f in myhooks) {
        res0 <- do.call(f,list(x=res))
        if (!is.null(res0))
            res <- res0
    }

    return(res)
}

###}}}

###{{{ estimate.list

estimate_lvmlist <-
    function(x, data, messages=lava.options()$messages, fix, missing=FALSE,  ...) {
        if (base::missing(data)) {
            return(estimate(x[[1]],x[[2]],missing=missing,...))
        }
        nm <- length(x)
        if (nm==1) {
            return(estimate(x[[1]],data,missing=missing,...))
        }
        if (!all(unlist(lapply(x, function(y) inherits(y,"lvm")))))
            stop ("Expected a list of 'lvm' objects.")
        if (is.data.frame(data)) {
            warning("Only one dataset - going for standard analysis on each submodel.")
            res <- c()
            for (i in seq_len(nm)) {
                res <- c(res, list(estimate(x[[i]],data=data,messages=0,missing=missing, ...)))
            }
            return(res)
        }

        if (nm!=length(data)) stop("Supply dataset for each model")

        Xfix <- FALSE
        xfix <- list()
        for (i in seq_along(x)) {
            data0 <- data[[i]]
            xfix0 <- colnames(data0)[(colnames(data0)%in%parlabels(x[[i]],exo=TRUE))]
            xfix <- c(xfix, list(xfix0))
            if (length(xfix0)>0) { ## Yes, random slopes
                Xfix<-TRUE
            }
        }
        if (base::missing(fix)) {
            fix <- ifelse(Xfix,FALSE,TRUE)
        }


        mg <- multigroup(x,data,fix=fix,missing=missing,...)
        res <- estimate(mg,...)

        return(res)
    }

###}}}

##' @export
procformula <- function(object=NULL,value,exo=lava.options()$exogenous,...) {

    ## Split into reponse and covariates by ~ disregarding expressions in parantheses
    ## Use deparse on formula components directly to avoid reliance on
    ## as.character.formula which can be overwritten (e.g., by formula.tools)
    ## '(?!...)' Negative lookahead assertion
    regex <- "~(?![^\\(].*\\))"
    if (length(value) == 3) {
        lhs_str <- gsub(" ", "", paste(deparse(value[[2]]), collapse = ""))
        rhs_str <- paste(deparse(value[[3]]), collapse = "")
        rhs_parts <- gsub(" ", "", strsplit(rhs_str, regex, perl = TRUE)[[1]])
        yx <- list(lhs_str, rhs_parts)
    } else {
        rhs_str <- paste(deparse(value[[2]]), collapse = "")
        rhs_parts <- gsub(" ", "", strsplit(rhs_str, regex, perl = TRUE)[[1]])
        yx <- list(rhs_parts)
    }
    yx <- lapply(yx,function(x) gsub("\n","",x))
    iscovar <- FALSE
    if (length(yx)==1) {
        lhs <- NULL; xidx <- 1
    } else {
            lhs <- yx[1]; xidx <- 2
            if (yx[[xidx]][1]=="") {
                yx[[xidx]] <- yx[[xidx]][-1]
                iscovar <- TRUE
            }
    }
    ##Check for link function
    invlink <- NULL
    if (xidx==2) {
        if (length(grep("[a-zA-Z0-9_]*\\(.*\\)$",yx[[xidx]]))>0) { ## rhs of the form F(x+y)
            invlink <- strsplit(yx[[xidx]],"\\(.*\\)")[[1]][1]
                if (invlink%in%c("f","v","I","") ||
                    grepl("\\+",invlink))
                { ## Reserved for setting linear constraints
                    invlink <- NULL
                } else {
                    yx[[xidx]] <- gsub(paste0(invlink,"\\(|\\)$"),"",yx[[xidx]])
                }
            }
    }

    ## Handling constraints with negative coefficients
    ## while not tampering with formulas like y~f(x,-2)
    st <- yx[[xidx]]
    st <- gsub("\\-","\\+\\-",gsub("\\+\\-","\\-",st)) ## Convert - to +- (to allow for splitting on '+')
    ##gsub("[^,]\\-","\\+\\-",st) ## Convert back any - not starting with ','
    st <- gsub(",\\+",",",st) ## Remove + inside 'f' and 'v' constraints
    st <- gsub("^\\+","",st) ## Remove leading plus
    yx[[xidx]] <- st
    
    ## Match '+' but not when preceeded by ( ... )
    X <- strsplit(yx[[xidx]],"\\+(?![^\\(]*\\))", perl=TRUE)[[1]]
    
    ##regex <- "(?!(\\(*))[\\(\\)]"
    regex <- "[\\(\\)]"
    ## Keep squares brackets and |(...) statements
    ## Extract variables from expressions like
    ## f(x,b) -> x,b  and  2*x -> 2,cx
    ## but avoid to tamper with transformation expressions:
    ## a~(x*b)
    res <- lapply(X,decomp.specials,regex,pattern2="\\*",pattern.ignore="~",reverse=TRUE,perl=TRUE)
    ##OLD:
    ##res <- lapply(X,decomp.specials,pattern2="[*]",reverse=TRUE)
    xx <- unlist(lapply(res, function(x) x[1]))

    xxf <- lapply(as.list(xx),function(x) decomp.specials(x,NULL,pattern2="\\[|~",perl=TRUE))
    xs <- unlist(lapply(xxf,function(x) x[1]))

    ## Alter intercepts?    
    intpos <- which(vapply(xs,function(x) grepl("^[\\+\\-]*[\\.|0-9]+$",x), 0)==1)
    ## Match '(int)'
    intpos0 <- which(vapply(X,function(x) grepl("^\\([\\+\\-]*[\\.|0-9]+\\)$",x),0)==1)

    yy <- ys <- NULL
    if (length(lhs)>0) {
        yy <- decomp.specials(lhs)
        yyf <- lapply(yy,function(y) decomp.specials(y,NULL,pattern2="[",fixed=TRUE,perl=FALSE))
        ys <- unlist(lapply(yyf,function(x) x[1]))
    }
    
    notexo <- c()
    if (!is.null(object)) {
      if (length(lhs)>0) {
        object <- addvar(object,ys,reindex=FALSE,...)
        notexo <- ys
        ## Add link-transformation
        if (!is.null(invlink)) {
            if (invlink=="") {
                object <- transform(object,ys,NULL,post=FALSE)
                covariance(object,ys) <- NA
            } else {
                ff <- function(x) {};  body(ff) <- parse(text=paste0(invlink,"(x)"))
                object <- transform(object,ys,ff,post=FALSE)
                covariance(object,ys) <- 0
            }
        }
      }

      if (length(intpos>0)) {
          xs[intpos[1]] <- gsub("\\+","",xs[intpos[1]])
          if (xs[intpos[1]]==1 && (!length(intpos0)>0) ) {
              xs[intpos[1]] <- NA
          }
          intercept(object,ys) <- char2num(xs[intpos[1]])
          xs <- xs[-intpos]
          res[intpos] <- NULL
      }


        object <- addvar(object,xs,reindex=FALSE ,...)
        exolist <- c()


        for (i in seq_len(length(xs))) {

            ## Extract transformation statements: var~(expr)
            xf0 <- strsplit(xx[[i]],"~")[[1]]
            if (length(xf0)>1) {
                myexpr <- xf0[2]
                ftr <- toformula(y="",x=paste0("-1+I(",myexpr,")"))
                xtr <- all.vars(ftr)
                xf0 <- xf0[1]
                transform(object, y=xf0, x=xtr) <- function(x) {
                    structure(model.matrix(ftr,as.data.frame(x)),dimnames=list(NULL,xf0))
                }
            }

            xf <- unlist(strsplit(xf0,"[\\[\\]]",perl=TRUE))
            if (length(xf)>1) {
                xpar <- strsplit(xf[2],":")[[1]]
                if (length(xpar)>1) {
                    val <- ifelse(xpar[2]=="NA",NA,xpar[2])
                    valn <- char2num(val)
                    covariance(object,xs[i]) <- ifelse(is.na(valn),val,valn)
                }
                val <- ifelse(xpar[1]=="NA",NA,xpar[1])
                valn <- char2num(val)
                if (is.na(val) || val!=".") {
                    intercept(object,xs[i]) <- ifelse(is.na(valn),val,valn)
                    notexo <- c(notexo,xs[i])
                }
            } else { exolist <- c(exolist,xs[i]) }
        }

        for (i in seq_len(length(ys))) {
            y <- ys[i]
            yf <- unlist(strsplit(yy[i],"[\\[\\]]",perl=TRUE))
            if (length(yf)>1) {
                ypar <- strsplit(yf[2],":")[[1]]
                if (length(ypar)>1) {
                    val <- ifelse(ypar[2]=="NA",NA,ypar[2])
                    valn <- char2num(val)
                    covariance(object,y) <- ifelse(is.na(valn),val,valn)
                }
                val <- ifelse(ypar[1]=="NA",NA,ypar[1])
                valn <- char2num(val)
                if (is.na(val) || val!=".") {
                    intercept(object,y) <- ifelse(is.na(valn),val,valn)
                }
            }
        }

        curvar <- index(object)$var
        if (exo) {
            oldexo <- exogenous(object)
            newexo <- setdiff(exolist,c(notexo,curvar,ys))
            exogenous(object) <- union(newexo,setdiff(oldexo,notexo))
        }
    }

    return(list(object=object,
                yx=yx,
                X=X,
                ys=ys,
                xx=xx,
                xs=xs,
                yy=yy,
                ys=ys,
                res=res,
                notexo=notexo,
                intpos=intpos,
                invlink=invlink,
                lhs=lhs,
                iscovar=iscovar))
}

##' @export
estimate.formula <- function(x, data, weights, family=stats::gaussian,
                      ..., model="glm", lvm=FALSE) {

    cl <- match.call()
    if (lvm) {
        cl[[1]] <- as.call(parse(text="lava:::estimate0"))[[1]]
        return(eval(cl,envir=parent.frame()))
    }
    if (missing(data)) {
        data <- environment(x)
        cl$data <- quote(data)
    }
    mf <- match.call(expand.dots = FALSE)
    m <- match(c("x", "data", "weights", "subset",
                 "etastart", "mustart", "offset"), names(mf), 0L)
    mf <- mf[c(1L, m)]
    mf$drop.unused.levels <- TRUE
    mf[[1L]] <- quote(stats::model.frame)
    mf$na.action <- na.pass0
    names(mf)[which(names(mf)=="x")] <- "formula"
    mf <- eval(mf, parent.frame())
    idx <- attr(na.omit(mf),"na.action")
    weights <- as.vector(model.weights(mf))
    if (length(idx)>0) {
        if (is.null(weights)) weights <- rep(1,nrow(mf))
        if (length(idx)>0) weights[idx] <- 0
        cl$weights <- quote(weights)
        cl$na.action <- quote(na.pass0)
    }
    cl$data <- quote(data)
    argsModelObj <- names(formals(model))
    dots <- list(...)
    idx <- which(names(dots) %ni% argsModelObj)
    rmarg <- c("model","raw","lvm")
    if (length(idx)>0) rmarg <- c(names(dots)[idx],rmarg)
    cl[rmarg] <- NULL
    names(cl)[names(cl)=="x"] <- "formula"
    cl[[1]] <- as.name(model)
    fit <- eval(cl)
    if (length(idx)==0) return(fit)
    optarg <- NULL
    if (length(idx)>0) {
        optarg <- dots[idx]
    }
    do.call(estimate, c(list(fit),optarg))
}


estimate0 <- function(x,data=parent.frame(),pred.norm=c(),unstruct=FALSE,messages=0,id=NULL,distribution=NULL,estimator="gaussian",...) {
    formulaId <- union(Specials(x,c("cluster")),Specials(x,c("id")))
    formulaSt <- paste0("~.-cluster(",formulaId,")-id(",formulaId,")")
    if (!is.null(formulaId)) {
        id <- formulaId
        x <- update(x,as.formula(formulaSt))
    }
    if (!is.null(id))
        x <- update(x,as.formula(paste(".~.+",id)))
    mf <- model.frame(x,data)
    yvar <- names(mf)[1]
    y <- mf[,yvar]
    opt <- options(na.action="na.pass")
    mm <- model.matrix(x,data)
    options(opt)
    covars <- colnames(mm)
    covars <- unlist(lapply(covars, function(x) gsub("[^a-zA-Z0-9._]","",x)))
    colnames(mm) <- covars

    if (attr(terms(x),"intercept")==1) {
        covars <- covars[-1]
        it <- c()
    } else {
        it <- "0"
    }

    if (!is.null(id)) covars <- setdiff(covars,id)
    model <- lvm(toformula(yvar,c(it,covars)),messages=0)
    if (!is.null(distribution)) {
        lava::distribution(model,yvar) <- distribution
        estimator <- "glm"
    }
    mydata <- na.omit(as.data.frame(cbind(data.frame(y),mm))); names(mydata)[1] <- yvar
    exogenous(model) <- setdiff(covars,pred.norm)
    if (unstruct) {
        model <- covariance(model,pred.norm,pairwise=TRUE)
    }
    estimate(model,mydata,messages=messages,id=id,estimator=estimator,...)
}

##' Split data into folds
##'
##' @title Split data into folds
##' @param x Data or integer (size)
##' @param p Number of folds, or if a number between 0 and 1 is given two folds of size p and (1-p) will be returned
##' @param replace With or with-out replacement
##' @param return.index If TRUE index of folds are returned otherwise the actual data splits are returned (default)
##' @param k (Optional, only used when p=NULL) number of folds without shuffling
##' @param ... additional arguments to lower-level functions
##' @export
##' @aliases csplit foldr
##' @examples
##' foldr(5,2,rep=2)
##' csplit(10,3)
##' csplit(iris[1:10,]) ## Split in two sets 1:(n/2) and (n/2+1):n
##' csplit(iris[1:10,],0.5)
##' @author Klaus K. Holst
csplit <- function(x,p=NULL,replace=FALSE,return.index=FALSE,k=2,...) {
    if (length(x)==1 & is.numeric(x)) x <- seq(x)
    N <- NROW(x)
    if (is.null(p)) { ##
        K <- base::round(N/k)
        idx <- split(seq(N),sort(rep(seq(k),length.out=N,each=K)))
    } else {
        if (p<1) { ## two folds (size N*p and N*(1-p))
            idx1 <- base::sample(N,base::round(p*N),replace=replace)
            idx <- list(idx1,
                        base::sample(setdiff(seq(N),idx1),replace=replace))
        } else { ## Number of folds (equal size)
            idx <- split(sample(seq(N)), rep(seq(p), length=N))
        }
    }
    if (return.index)
        return(idx)
    if (!is.vector(x)) {
        return(lapply(idx,function(ii) x[ii,,drop=FALSE]))
    }
    return(lapply(idx,function(ii) x[ii]))
}

##' @export
foldr <- function(n,K=5,rep=10) {
    replicate(rep,split(sample(seq(n)), rep(seq(K), length=n)),simplify=FALSE)
}

##' Generic method for memberships from object (e.g. a graph)
##'
##' @title Extract children or parent elements of object
##' @export
##' @aliases children parents ancestors descendants roots sinks adjMat edgeList
##' @param object Object
##' @param \dots Additional arguments
##' @author Klaus K. Holst
"children" <- function(object,...) UseMethod("children")
##' @export
"parents" <- function(object,...) UseMethod("parents")
##' @export
"roots" <- function(object,...) UseMethod("roots")
##' @export
"sinks" <- function(object,...) UseMethod("sinks")
##' @export
"descendants" <- function(object,...) UseMethod("descendants")
##' @export
"ancestors" <- function(object,...) UseMethod("ancestors")
##' @export
"adjMat" <- function(object,...) UseMethod("adjMat")
##' @export
"edgeList" <- function(object,...) UseMethod("edgeList")

##' @export
adjMat.lvm <- function(object,...) t(object$M)

##' @export
adjMat.lvmfit <- function(object,...) adjMat(Model(object),...)

##' @export
edgeList.lvmfit <- function(object,...) edgeList(Model(object),...)

##' @export
edgeList.lvm <- function(object, labels=FALSE, list=FALSE, ...) {
    edgelist <- data.frame(from=NULL,to=NULL)
    A <- adjMat(object)
    for (i in 1:nrow(A)) {
        ii <- which(A[,i]>0)
        if (length(ii)>0)
            edgelist <- rbind(edgelist,data.frame(from=ii,to=i))
    }
    if (list) labels <- TRUE
    if (labels) edgelist <- as.data.frame(apply(edgelist,2,function(x) vars(object)[x]))
    if (list) {
      edgelist <- apply(edgelist, 1, function(x) paste0(x,collapse="~"))
    }
    edgelist
}

##' @export
parents.lvmfit <- function(object,...) parents(Model(object),...)

##' @export
children.lvmfit <- function(object,...) children(Model(object),...)

##' @export
descendants.lvmfit <- function(object,...) descendants(Model(object),...)

##' @export
ancestors.lvmfit <- function(object,...) ancestors(Model(object),...)

##' @export
roots.lvmfit <- function(object,...) roots(Model(object),...)

##' @export
sinks.lvmfit <- function(object,...) sinks(Model(object),...)



##' @export
parents.lvm <- function(object,var,...) {
    A <- index(object)$A
    if (missing(var)) {
        return(rownames(A))
    }
    if (inherits(var,"formula"))
        var <- all.vars(var)
    res <- lapply(var, function(v) rownames(A)[A[,v]!=0])
    res <- unique(unlist(res))
    if (length(res)==0) res <- NULL
    res
}

##' @export
children.lvm <- function(object,var,...) {
    A <- index(object)$A
    if (missing(var)) {
        return(rownames(A))
    }
    if (inherits(var,"formula"))
        var <- all.vars(var)
    res <- lapply(var, function(v) rownames(A)[A[v,]!=0])
    res <- unique(unlist(res))
    if (length(res)==0) res <- NULL
    res
}


##' @export
ancestors.lvm <- function(object,x,...) {
    if (inherits(x,"formula")) x <- all.vars(x)
    res <- c()
    left <- setdiff(vars(object),x)
    count <- 0
    child <- x
    while (length(x)>0) {
        count <- count+1
        x <- parents(object,child)
        child <- intersect(x,left)
        res <- union(res,child)
        left <- setdiff(left,child)
    }
    if (length(res)==0) res <- NULL
    return(res)
}

##' @export
descendants.lvm <- function(object,x,...) {
    if (inherits(x,"formula")) x <- all.vars(x)
    res <- c()
    left <- setdiff(vars(object),x)
    count <- 0
    parent <- x
    while (length(x)>0) {
        count <- count+1
        x <- children(object,parent)
        parent <- intersect(x,left)
        res <- union(res,parent)
        left <- setdiff(left,parent)
    }
    if (length(res)==0) res <- NULL
    return(res)
}

##' @export
roots.lvm <- function(object,...) {
    return(exogenous(object,index=FALSE,...))
}

##' @export
sinks.lvm <- function(object,...) {
    return(endogenous(object,top=TRUE,...))
}

context("Inference")

test_that("d-separation",{
    m <- lvm(x5 ~ x4+x3, x4~x3+x1, x3~x2, x2~x1)
    testthat::expect_true(dsep(m,x5~x1|x3+x4))
    testthat::expect_false(dsep(m,x5~x1|x2+x4))
    testthat::expect_true(dsep(m,x5~x1|x2+x3+x4))
    testthat::expect_false(dsep(m,~x1+x2+x3|x4))

    testthat::expect_true(setequal(ancestors(m,~x5),setdiff(vars(m),"x5")))    
    testthat::expect_true(setequal(ancestors(m,~x1),NULL))
    testthat::expect_true(setequal(descendants(m,~x5),NULL))
    testthat::expect_true(setequal(descendants(m,~x1),setdiff(vars(m),"x1")))
})

context("Graph routines")

test_that("d-separation",{
    m <- lvm(x5 ~ x4+x3, x4~x3+x1, x3~x2, x2~x1)
    testthat::expect_true(dsep(m,x5~x1|x3+x4))
    testthat::expect_false(dsep(m,x5~x1|x2+x4))
    testthat::expect_true(dsep(m,x5~x1|x2+x3+x4))
    testthat::expect_false(dsep(m,~x1+x2+x3|x4))

    testthat::expect_true(setequal(ancestors(m,~x5),setdiff(vars(m),"x5")))    
    testthat::expect_true(setequal(ancestors(m,~x1),NULL))
    testthat::expect_true(setequal(descendants(m,~x5),NULL))
    testthat::expect_true(setequal(descendants(m,~x1),setdiff(vars(m),"x1")))
})

context("Model specification")

test_that("Basic model building blocks", {
    m <- lvm(y[m]~x)
    covariance(m) <- y~z
    testthat::expect_true(covariance(m)$rel["z","y"]==1)
    testthat::expect_true(regression(m)$rel["x","y"]==1)

    ## Children parent,nodes
    testthat::expect_true(children(m,~x)=="y")
    testthat::expect_true(parents(m,~y)=="x")
    testthat::expect_equivalent(parents(m),vars(m))
    testthat::expect_equivalent(children(m),vars(m))

    ## Remove association
    cancel(m) <- y~z+x
    testthat::expect_true(covariance(m)$rel["z","y"]==0)
    testthat::expect_true(regression(m)$rel["x","y"]==0)

    ## Remove variable
    kill(m) <- ~x
    testthat::expect_equivalent(vars(m),c("y","z"))
    testthat::expect_true(intercept(m)["y"]=="m")

    m <- lvm(c(y1,y2,y3)~x)
    d <- sim(m,50)
    e <- estimate(m,d)
    ## Equivalence
    ##equivalence(e,silent=TRUE)


    ## formula
    f <- formula(m,all=TRUE)
    testthat::expect_true(length(f)==length(vars(m)))
    testthat::expect_true(all(unlist(lapply(f,function(x) inherits(x,"formula")))))

    ## Parametrization
    m <- lvm(c(y1,y2,y3)~u)
    latent(m) <- ~u
    m2 <- fixsome(m,param=NULL)
    testthat::expect_true(all(is.na(regression(m2)$values)))
    m2 <- fixsome(m,param="relative")
    testthat::expect_true(regression(m2)$values["u","y1"]==1)
    testthat::expect_true(intercept(m2)[["y1"]]==0)
    m2 <- fixsome(m,param="hybrid")
    testthat::expect_true(regression(m2)$values["u","y1"]==1)
    testthat::expect_true(intercept(m2)[["u"]]==0)
    m2 <- fixsome(m,param="absolute")
    testthat::expect_true(all(is.na(regression(m2)$values)))
    testthat::expect_true(intercept(m2)[["u"]]==0)
    testthat::expect_true(covariance(m2)$values["u","u"]==1)

    ## Merge
    m1 <- lvm(c(y1,y2,y3)~1*u1[m1:v1])
    latent(m1) <- ~u1
    m2 <- lvm(c(y1,y2,y3)~2*u2[m2:v2])
    latent(m2) <- ~u2
    mm <- m1%++%m2

    testthat::expect_true(covariance(mm)$labels["u1","u1"]=="v1")
    testthat::expect_true(intercept(mm)[["u2"]]=="m2")

    ## LISREL
    ## mm <- fixsome(mm)
    ## L <- lava:::lisrel(mm,rep(1,length(coef(mm))))
    ## testthat::expect_equivalent(L$B,matrix(0,2,2))
    ## testthat::expect_equivalent(L$Theta,diag(3))
    ## testthat::expect_equivalent(L$Psi,diag(2))
})

test_that("Linear constraints", {
    m <- lvm(c(y[m:v]~b*x))
    constrain(m,b~a) <- base::identity
    d <- sim(m,100,seed=1)
    l <- lm(y~x, d)
    e <- estimate(m, d)
    err <- sum((coef(l)-coef(e)[c('y','a')])^2)
    testthat::expect_true(err<1e-12)
})

if (requireNamespace("Rgraphviz",quietly = TRUE))
test_that("Graph attributes", {
    m <- lvm(y~x)
    suppressMessages(g1 <- graph::updateGraph(plot(m,noplot=TRUE)))
    m1 <- graph2lvm(g1)
    testthat::expect_equivalent(m1$M, m$M)

    col <- "blue"
    v <- "y"
    g1 <- lava:::addattr(g1, "fill", v, col)
    testthat::expect_true(col == graph::nodeRenderInfo(g1)$fill[[v]])
    nodecolor(m, v) <- "blue"

    g2 <- Graph(m, add=TRUE)
    testthat::expect_true(inherits(g2, "graph"))
    testthat::expect_true(col == graph::nodeRenderInfo(g2)$fill[[v]])
    testthat::expect_true(lava:::addattr(g2, "fill")[[v]] == "blue")
    graph::graphRenderInfo(g2)$rankdir <- "LR"
    Graph(m) <- g2
    testthat::expect_true(graph::graphRenderInfo(Graph(m))$rankdir=="LR")

    ## Labels
    labels(m) <- c(y = "Y")
    lava:::addattr(Graph(m, add=TRUE), "label")
    testthat::expect_true(lava:::addattr(finalize(m), "label")[["y"]]=="Y")
    labels(g2) <- c(y = "Y")
    testthat::expect_true(!is.null(graph::nodeRenderInfo(g2)$label["y"]))

    edgelabels(m, y~x) <- "a"
    testthat::expect_true(!is.null(edgelabels(finalize(m))))
})

test_that("Categorical variables", {
    m <- lvm()
    categorical(m,K=3,p=c(0.1,0.5)) <- ~x
    d1 <- simulate(m,10,seed=1)
    categorical(m,K=3) <- ~x
    d2 <- simulate(m,10,seed=1)
    testthat::expect_false(identical(d1,d2))

    regression(m,additive=FALSE,y~x) <- c(0,-5,5)
    d <- simulate(m,100,seed=1)
    l <- lm(y~factor(x),d)
    testthat::expect_true(sign(coef(l))[2]==-sign(coef(l))[3])

})

context("Model specification")

test_that("Basic model building blocks", {
    m <- lvm(y[m]~x)
    covariance(m) <- y~z
    testthat::expect_true(covariance(m)$rel["z","y"]==1)
    testthat::expect_true(regression(m)$rel["x","y"]==1)

    ## Children parent,nodes
    testthat::expect_true(children(m,~x)=="y")
    testthat::expect_true(parents(m,~y)=="x")
    testthat::expect_equivalent(parents(m),vars(m))
    testthat::expect_equivalent(children(m),vars(m))

    ## Remove association
    cancel(m) <- y~z+x
    testthat::expect_true(covariance(m)$rel["z","y"]==0)
    testthat::expect_true(regression(m)$rel["x","y"]==0)

    ## Remove variable
    kill(m) <- ~x
    testthat::expect_equivalent(vars(m),c("y","z"))
    testthat::expect_true(intercept(m)["y"]=="m")

    m <- lvm(c(y1,y2,y3)~x)
    d <- sim(m,50)
    e <- estimate(m,d)
    ## Equivalence
    ##equivalence(e,silent=TRUE)


    ## formula
    f <- formula(m,all=TRUE)
    testthat::expect_true(length(f)==length(vars(m)))
    testthat::expect_true(all(unlist(lapply(f,function(x) inherits(x,"formula")))))

    ## Parametrization
    m <- lvm(c(y1,y2,y3)~u)
    latent(m) <- ~u
    m2 <- fixsome(m,param=NULL)
    testthat::expect_true(all(is.na(regression(m2)$values)))
    m2 <- fixsome(m,param="relative")
    testthat::expect_true(regression(m2)$values["u","y1"]==1)
    testthat::expect_true(intercept(m2)[["y1"]]==0)
    m2 <- fixsome(m,param="hybrid")
    testthat::expect_true(regression(m2)$values["u","y1"]==1)
    testthat::expect_true(intercept(m2)[["u"]]==0)
    m2 <- fixsome(m,param="absolute")
    testthat::expect_true(all(is.na(regression(m2)$values)))
    testthat::expect_true(intercept(m2)[["u"]]==0)
    testthat::expect_true(covariance(m2)$values["u","u"]==1)

    ## Merge
    m1 <- lvm(c(y1,y2,y3)~1*u1[m1:v1])
    latent(m1) <- ~u1
    m2 <- lvm(c(y1,y2,y3)~2*u2[m2:v2])
    latent(m2) <- ~u2
    mm <- m1%++%m2

    testthat::expect_true(covariance(mm)$labels["u1","u1"]=="v1")
    testthat::expect_true(intercept(mm)[["u2"]]=="m2")

    ## LISREL
    ## mm <- fixsome(mm)
    ## L <- lava:::lisrel(mm,rep(1,length(coef(mm))))
    ## testthat::expect_equivalent(L$B,matrix(0,2,2))
    ## testthat::expect_equivalent(L$Theta,diag(3))
    ## testthat::expect_equivalent(L$Psi,diag(2))
})

test_that("Linear constraints", {
    m <- lvm(c(y[m:v]~b*x))
    constrain(m,b~a) <- base::identity
    d <- sim(m,100,seed=1)
    l <- lm(y~x, d)
    e <- estimate(m, d)
    err <- sum((coef(l)-coef(e)[c('y','a')])^2)
    testthat::expect_true(err<1e-12)
})

if (requireNamespace("Rgraphviz",quietly = TRUE))
test_that("Graph attributes", {
    m <- lvm(y~x)
    suppressMessages(g1 <- graph::updateGraph(plot(m,noplot=TRUE)))
    m1 <- graph2lvm(g1)
    testthat::expect_equivalent(m1$M, m$M)

    col <- "blue"
    v <- "y"
    g1 <- lava:::addattr(g1, "fill", v, col)
    testthat::expect_true(col == graph::nodeRenderInfo(g1)$fill[[v]])
    nodecolor(m, v) <- "blue"

    g2 <- Graph(m, add=TRUE)
    testthat::expect_true(inherits(g2, "graph"))
    testthat::expect_true(col == graph::nodeRenderInfo(g2)$fill[[v]])
    testthat::expect_true(lava:::addattr(g2, "fill")[[v]] == "blue")
    graph::graphRenderInfo(g2)$rankdir <- "LR"
    Graph(m) <- g2
    testthat::expect_true(graph::graphRenderInfo(Graph(m))$rankdir=="LR")

    ## Labels
    labels(m) <- c(y = "Y")
    lava:::addattr(Graph(m, add=TRUE), "label")
    testthat::expect_true(lava:::addattr(finalize(m), "label")[["y"]]=="Y")
    labels(g2) <- c(y = "Y")
    testthat::expect_true(!is.null(graph::nodeRenderInfo(g2)$label["y"]))

    edgelabels(m, y~x) <- "a"
    testthat::expect_true(!is.null(edgelabels(finalize(m))))
})

test_that("regression<- with character to/from matches formula style", {
    # Formula style: regression(m, y~x) <- "b"
    m1 <- lvm()
    regression(m1, y ~ x) <- "b"

    # Character positional: regression(m, "y", "x") <- "b"
    m2 <- lvm()
    regression(m2, "y", "x") <- "b"

    # Character named: regression(m, to="y", from="x") <- "b"
    m3 <- lvm()
    regression(m3, to = "y", from = "x") <- "b"

    testthat::expect_identical(m1$par, m2$par)
    testthat::expect_identical(m1$par, m3$par)
    testthat::expect_identical(m1$fix, m2$fix)
    testthat::expect_identical(m1$fix, m3$fix)

    # Numeric constraint
    m4 <- lvm()
    regression(m4, "y", "x") <- 2
    testthat::expect_equal(m4$fix["x", "y"], 2)

    # NA to clear constraint
    m5 <- lvm(y ~ x)
    regression(m5, "y", "x") <- "b"
    testthat::expect_equal(m5$par["x", "y"], "b")
    regression(m5, "y", "x") <- NA
    testthat::expect_true(is.na(m5$par["x", "y"]))
    testthat::expect_true(is.na(m5$fix["x", "y"]))
})

test_that("regression() with value= and character to/from", {
    # regression(m, to, from, value=) should apply the constraint
    m1 <- lvm()
    regression(m1, y ~ x) <- "b"

    m2 <- lvm()
    m2 <- regression(m2, to = "y", from = "x", value = "b")

    testthat::expect_identical(m1$par, m2$par)
    testthat::expect_identical(m1$fix, m2$fix)

    # Numeric value
    m3 <- lvm()
    m3 <- regression(m3, "y", "x", value = 2)
    testthat::expect_equal(m3$fix["x", "y"], 2)

    # Multiple predictors
    m4 <- lvm()
    regression(m4, y ~ x1 + x2) <- "b"

    m5 <- lvm()
    m5 <- regression(m5, to = "y", from = c("x1", "x2"), value = "b")

    testthat::expect_identical(m4$par, m5$par)
    testthat::expect_identical(m4$fix, m5$fix)
})

test_that("Categorical variables", {
    m <- lvm()
    categorical(m,K=3,p=c(0.1,0.5)) <- ~x
    d1 <- simulate(m,10,seed=1)
    categorical(m,K=3) <- ~x
    d2 <- simulate(m,10,seed=1)
    testthat::expect_false(identical(d1,d2))

    regression(m,additive=FALSE,y~x) <- c(0,-5,5)
    d <- simulate(m,100,seed=1)
    l <- lm(y~factor(x),d)
    testthat::expect_true(sign(coef(l))[2]==-sign(coef(l))[3])

})

### Return position of variance elements in the parameter vector (without mean parameters)
### Optimization constraints are needed on these parameters
##' @export
variances <- function(x,mean=FALSE) {
##  if (is.null(x$parpos))
##    x$parpos <- parpos(x)
  x$parpos <- parpos(Model(x),mean=TRUE)
  res <- diag(x$parpos$P)[which(diag(index(x)$P0)==1)]
  if (!mean) {
    return(res - index(x)$npar.mean)
  }
  return(res)
}
## And the off-diagonal (covariance) parameters
##' @export
offdiags <- function(x,mean=FALSE) {
  parpos <- parpos(x,mean=mean)
  pp <- parpos$P
  pp[lower.tri(pp)][(index(x)$P0)[lower.tri(pp)]==1]
}

##' @export
`print.lvm` <-
function(x, ..., print.transform=TRUE,print.exogenous=TRUE) {
  res <- NULL
  myhooks <- gethook("print.hooks")
  for (f in myhooks) {
      res <- do.call(f, list(x=x, ...))
  }
  if (is.null(res)) {
    k <- length(vars(x))
    L <- rep(FALSE, k); names(L) <- vars(x); L[latent(x)] <- TRUE
    cat("Latent Variable Model\n") ##;" \n\twith: ", k, " variables.\n", sep="");
    if (k==0) {
        cat("\nEmpty\n")
        return()
    }
    ff <- formula(x, char=TRUE, all=TRUE)
    R <- Rx <- Rt <- c()
    exo <- exogenous(x)
    for (f in ff) {
      oneline <- as.character(f);
      y <- strsplit(f, "~")[[1]][1]
      y <- trim(y)
      {
        col1 <- as.character(oneline)
        D <- attributes(distribution(x)[[y]])$family
        Tr <- x$attributes$transform[[y]]
        col2 <- tryCatch(x$attributes$type[[y]], error=function(...) NULL)
        if (is.null(col2) || is.na(col2)) {
            if (!is.na(x$covfix[y, y]) && x$covfix[y, y]==0L) {
                col2 <- "deterministic"
            } else {
                col2 <- "gaussian"
            }
        }
        if (!is.null(Tr)){
            col1 <- paste0(y, " ~ ", paste0(Tr$x, collapse="+"), sep="")
            Rt <- rbind(Rt, c(col1, ""))
        }
        if (!is.null(D$family)) {
            col2 <- paste0(D$family)
        }
        if (!is.null(D$link)) col2 <- paste0(col2, "(", D$link, ")")
        if (!is.null(D$par)) col2 <- paste0(col2, "(", paste(D$par, collapse=","), ")")
        if (is.list(distribution(x)[[y]]) && is.vector(distribution(x)[[y]][[1]])) col2 <- "fixed"
        if (L[y]) col2 <- paste0(col2, ", latent")
        if (y%in%exo) {
            Rx <- rbind(Rx,c(col1, col2))
        } else {
            if (is.null(Tr)) {
                R <- rbind(R,c(col1, col2))
            }
        }
      }
    }
    if (length(R)>0) {
        rownames(R) <- paste(" ", R[, 1], " "); colnames(R) <- rep("", ncol(R))
        print(R[, 2, drop=FALSE], quote=FALSE, ...)
    }
    if (print.exogenous && length(Rx)>0) {
        cat("\nExogenous variables:")
        rownames(Rx) <- paste(" ", gsub("~ 1", "", Rx[, 1]), "     "); colnames(Rx) <- rep("", ncol(Rx))
        print(Rx[, 2, drop=FALSE], quote=FALSE, ...)
    }
    if (print.transform && length(Rt)>0) {
        cat("\nTransformations:")
        rownames(Rt) <- paste(" ", gsub("~ 1", "", Rt[,1]), "     "); colnames(Rt) <- rep("", ncol(Rt))
        print(Rt[,2,drop=FALSE],quote=FALSE, ...)
    }
  }
  cat("\n")
  invisible(x)
}

##' @export
`print.lvmfit` <-
function(x,type=2,labels=FALSE,...) {
    print(CoefMat(x,labels=labels,type=type,...),quote=FALSE,right=TRUE)
    minSV <- attr(vcov(x),"minSV")
    if (!is.null(minSV) && minSV<1e-12) {
        warning("Small singular value: ", format(minSV))
    }
    pseudo <- attr(vcov(x),"pseudo")
    if (!is.null(pseudo) && pseudo) warning("Singular covariance matrix. Pseudo-inverse used.")
    invisible(x)
}

##' @export
print.lvmfit.randomslope <- function(x,labels=FALSE,type=2,...) {
  print(CoefMat(x,labels=labels,type=type,...),quote=FALSE,right=TRUE)
  invisible(x)
}

##' @export
print.multigroupfit <- function(x,groups=NULL,...)  {
  if (is.null(groups)) {
    if (x$model$missing) {
      modelclass <- attributes(x$model0)$modelclass
      nmis <- attributes(x$model0)$nmis
      orggroup <- unique(modelclass)
      groupn <- unlist(lapply(orggroup,function(i) sum(modelclass==i)))
      cumsumgroup <- cumsum(c(0,groupn))
      groups <- unlist(lapply(orggroup,function(i)
                              which.min(nmis[which(modelclass==i)])+cumsumgroup[i])) ##  groups with max. number of variables
      for (i in seq_len(length(groups))) {
        if (nmis[groups[i]]>0) warning("No complete cases in group ",i,". Showing results of group with max number of variables. All coefficients can be extracted with 'coef'. All missing pattern groups belonging to this sub-model can be extracted by calling: coef(..., groups=c(",paste(which(modelclass==i),collapse=","),"))")
      }
      if (!is.null(x$model$mnameses))
        x$model$names <- x$model$mnames
    } else {
      groups <- seq_len(length(x$model$lvm))
    }
  }
  res <- coef(x,type=2,groups=groups,...)
  counter <- 0
  dots <- list(...)
  dots$groups <- groups
  type <- if (is.null(dots$type)) {
    dots$type <- 2
##    dots$type <- ifelse("lvmfit.randomslope"%in%class(x),2,9)
  }
  myargs <- c(list(x=x), dots)
  myargs$groups <- groups
  CC <- do.call("CoefMat.multigroupfit",myargs)
  for (cc in res) {
    counter <- counter+1
    cat(rep("_",52),"\n",sep="")
    cat("Group ", counter, sep="")
    myname <- x$model$names[counter]
    if (!is.null(myname) && !is.na(myname))
      cat(": ",myname,sep="")
    if (!x$model$missing) cat(" (n=",nrow(Model(x)$data[[groups[counter]]]), ")", sep="")
    cat("\n")
    print(CC[[counter]],quote=FALSE,right=TRUE)
  }
  cat("\n")
  invisible(x)
}

##' @export
print.multigroup <- function(x,...) {
  cat("\n")
  cat("Number of groups:", x$ngroup, "\n")
  cat("Number of free parameters (not counting mean-parameters):", x$npar,"\n")
##  cat("Parameter-vector:", unlist(x$parlist), "\n\n")
  cat("Number of free mean parameters:", length(grep("m",x$mean)),"\n")
##  cat("Mean-vector:", x$mean, "\n\n")
  invisible(x)
}

printmany <- function(A,B,nspace=1,name1=NULL,name2=NULL,digits=3,rownames=NULL,emptystr=" ",bothrows=!is.table(A),right=TRUE,print=TRUE,...) {
  A <- format(A, digits=digits, right=right, ...)
  B <- format(B, digits=digits, right=right, ...)
  nA <- nrow(A); nB <- nrow(B)
  if (nrow(A)<nrow(B)) {
    A <- rbind(A, matrix("", nrow=nB-nA, ncol=ncol(A)))
  }
  if (nrow(B)<nrow(A)) {
    B <- rbind(B, rep("", nrow=nA-nB, ncol=ncol(B)))
  }
  if (!is.null(rownames) & length(rownames)==nrow(A))
    rownames(A) <- rownames(B) <- rownames
  res <- cbind(A, matrix("", nrow=nrow(A), ncol=nspace));
  dnn <- dimnames(A)
  dnn[[2]] <- c(dnn[[2]],rep(emptystr,nspace))
  dimnames(res) <- dnn
  ##dimnames(res)[[2]] <- c(dimnames(res)[[2]],rep(emptystr,nspace))
  if (!is.null(name1)) {
      oldname <- colnames(res)
      res <- cbind(rep("",nrow(res)), rownames(res), res);
      cres <- name1
      if (!is.null(rownames(res))) cres <- c(cres,"")
      colnames(res) <- c(cres,oldname)
      rownames(res) <- rep("",nrow(res))
  }
  if (!is.null(name2)) {
    oldname <- colnames(res)
    res <- cbind(res,rep("",nrow(res))); colnames(res) <- c(oldname,name2)
  }
  if (!identical(rownames(A),rownames(B)) & bothrows)
      res <- cbind(res, rownames(B))
  res <- cbind(res, B)
  if (is.null(name2)) {
      dnn[[2]] <- c(dnn[[2]],dimnames(B)[[2]])
      dimnames(res) <- dnn
  }
  if (print) print(res, quote=FALSE, right=right, ...)
  invisible(res)
}

###{{{ mixture

#' Estimate mixture latent variable model.
#'
#' Estimate mixture latent variable model
#'
#' Estimate parameters in a mixture of latent variable models via the EM
#' algorithm.
#'
#' The performance of the EM algorithm can be tuned via the \code{control}
#' argument, a list where a subset of the following members can be altered:
#'
#' \describe{ \item{start}{Optional starting values} \item{nstart}{Evaluate
#' \code{nstart} different starting values and run the EM-algorithm on the
#' parameters with largest likelihood} \item{tol}{Convergence tolerance of the
#' EM-algorithm.  The algorithm is stopped when the absolute change in
#' likelihood and parameter (2-norm) between successive iterations is less than
#' \code{tol}} \item{iter.max}{Maximum number of iterations of the
#' EM-algorithm} \item{gamma}{Scale-down (i.e. number between 0 and 1) of the
#' step-size of the Newton-Raphson algorithm in the M-step} \item{trace}{Trace
#' information on the EM-algorithm is printed on every \code{trace}th
#' iteration} }
#'
#' Note that the algorithm can be aborted any time (C-c) and still be saved
#' (via on.exit call).
#'
 #' @param x List of \code{lvm} objects. If only a single \code{lvm} object is
#' given, then a \code{k}-mixture of this model is fitted (free parameters
#' varying between mixture components).
#' @param data \code{data.frame}
#' @param k Number of mixture components
#' @param control Optimization parameters (see details)
#' #type Type of EM algorithm (standard, classification, stochastic)
#' @param vcov of asymptotic covariance matrix (NULL to omit)
#' @param names If TRUE returns the names of the parameters (for defining starting values)
#' @param ... Additional arguments parsed to lower-level functions
#' @author Klaus K. Holst
#' @seealso \code{mvnmix}
#' @keywords models regression
#' @examples
#'
#' \donttest{
#' m0 <- lvm(list(y~x+z,x~z))
#' distribution(m0,~z) <- binomial.lvm()
#' d <- sim(m0,2000,p=c("y~z"=2,"y~x"=1),seed=1)
#'
#' ## unmeasured confounder example
#' m <- baptize(lvm(y~x, x~1));
#' intercept(m,~x+y) <- NA
#'
#' if (requireNamespace('mets', quietly=TRUE)) {
#'   set.seed(42)
#'   M <- mixture(m,k=2,data=d,control=list(trace=1,tol=1e-6))
#'   summary(M)
#'   lm(y~x,d)
#'   estimate(M,"y~x")
#'   ## True slope := 1
#' }
#' }
#'
#' @export mixture
mixture <- function(x, data, k=length(x),
             control=list(),
             vcov="observed",
             names=FALSE,
             ...) {
    MODEL <- "normal"
    ##type=c("standard","CEM","SEM"),
    ## type <- tolower(type[1])
    ## if (type[1]!="standard") {
    ##     return(mixture0(x,data=data,k=k,control=control,type=type,...))
    ## }

    optim <- list(
        rerun=TRUE,
        ## EM (accelerated):
        K=1, ## K=1, first order approx, slower but some times more stable
        square=TRUE,
        step.max0=1,
        step.min0=1,
        mstep=4,
        kr=1,
        objfn.inc=2,
        keep.objfval=TRUE,
        convtype= "parameter",
        ## convtype = "objfn",
        maxiter=1500,
        tol=1e-5,
        trace=0,
        ## Starting values:
        start=NULL,
        startbounds=c(-2,2),
        startmean=FALSE,
        nstart=1,
        prob=NULL,
        ## Newton raphson:
        delta=1e-2,
        constrain=TRUE,
        stopc=2,
        lbound=1e-9,
        stabil=TRUE,
        gamma=0.5,
        gamma2=1,
        newton=10,
        lambda=0 # Stabilizing factor (avoid singularities of I)
    )

    if (!missing(control))
        optim[base::names(control)] <- control
    if ("iter.max"%in%base::names(optim)) optim$maxiter <- optim$iter.max

    if (k==1) {
        if (is.list(x))
            res <- estimate(x[[1]],data,...)
        else
            res <- estimate(x,data,...)
        return(res)
    }

    start0 <- NULL

    xx <- x
    if (inherits(x,"lvm")) {
        xx <- rep(list(x), k)
    }
    mg <- multigroup(xx, rep(list(data),k), fix=FALSE)
    ppos <- parpos(mg)
    parname <- attr(ppos,"name")
    naparname <- which(is.na(parname))
    parname[naparname]  <- mg$name[naparname]
    if (names) {
        return(parname)
    }

    if (class(x)[1]=="lvm") {
        index(x) <- reindex(x,zeroones=TRUE,deriv=TRUE)
        if ((is.null(optim$start) || length(optim$start)<length(parname))) {
            start0 <- tryCatch(estimate(x,data,quick=TRUE),
                               error=function(...) NULL)
        }
    }

    ## Bounds on variance parameters
    Npar <- with(mg, npar+npar.mean)
    ParPos <- modelPar(mg,1:Npar)$p
    lower <- rep(-Inf, mg$npar);
    offdiagpos <- c()
    varpos <- c()
    for (i in 1:k) {
        vpos <- sapply(mg$parlist[[i]][variances(mg$lvm[[i]])], function(y) as.numeric(substr(y,2,nchar(y))))
        offpos <- sapply(mg$parlist[[i]][offdiags(mg$lvm[[i]])], function(y) as.numeric(substr(y,2,nchar(y))))
        varpos <- c(varpos, vpos)
        offdiagpos <- c(offdiagpos,offpos)
        if (length(vpos)>0)
            lower[vpos] <- optim$lbound  ## Setup optimization constraints
    }
    lower <- c(rep(-Inf,mg$npar.mean), lower)
    constrained <- which(is.finite(lower))
    if (!any(constrained)) optim$constrain <- FALSE

    mymodel <- list(multigroup=mg,k=k,data=data,parpos=ParPos); class(mymodel) <- "lvm.mixture"

    selfstarter <- FALSE
    if (any(is.na(optim$start)) || length(optim$start)<length(ppos)) {
        constrLogLikS <- function(p) {
            if (optim$constrain) {
                p[constrained] <- exp(p[constrained])
            }
            -logLik(mymodel,c(p=p,rep(1/k,k-1)))
        }

        start <- rep(NA,Npar)
        if (!is.null(start0)) {
            start[ParPos[[1]]] <- start0[seq_along(ParPos[[1]])]
            ii <- which(is.na(start))
            start[ii] <- runif(length(ii),optim$startbounds[1],optim$startbounds[2])
        } else {
            selfstarter <- TRUE
            start <- runif(Npar,optim$startbounds[1],optim$startbounds[2]);
            if (length(offdiagpos)>0)
                start[mg$npar.mean + offdiagpos] <- 0
            if (optim$nstart>1) {
                myll <- constrLogLikS(start)
                for (i in 1:optim$nstart) {
                    newstart <- runif(Npar,optim$startbounds[1],optim$startbounds[2]);
                    newmyll <- constrLogLikS(newstart)
                    if (newmyll<myll) {
                        start <- newstart
                    }
                }
            }
        }
        if (!is.null(optim$start)) {
            iistart <- match(names(optim$start), parname)
            start[iistart] <- optim$start
        }
        optim$start <- start
    }
    if (length(optim$start)>Npar) {
        optim$prob <- optim$start[Npar+seq_len(k-1)]
        optim$start <- optim$start[seq_len(Npar)]
    }

    if (is.null(optim$prob))
        optim$prob <- rep(1/k,k-1)
    thetacur <- optim$start

    probcur <- with(optim, c(prob,1-sum(prob)))

    if (optim$constrain & !selfstarter) {
        thetacur[constrained] <- log(thetacur[constrained])
    }

    PosteriorProb <- function(pp,priorprob,constrain=FALSE) {
        if (!is.list(pp)) {
            if (constrain) {
                pp[constrained] <- exp(pp[constrained])
            }
            if (missing(priorprob)) priorprob <- pp[seq(Npar+1,length(pp))]
            pp <- lapply(ParPos,function(x) pp[x])
        }
        priorprob <- pmax(priorprob,1e-16)
        priorprob <- priorprob/sum(priorprob)
        k <- length(pp)
        logff <- sapply(seq(k), function(j) (logLik(mg$lvm[[j]],p=pp[[j]],data=data,indiv=TRUE,model=MODEL)))
        logplogff <- t(apply(logff,1, function(z) z+log(priorprob)))
        ## Log-sum-exp (see e.g. NR)
        zmax <- apply(logplogff,1,max)
        logsumpff <- log(rowSums(exp(logplogff-zmax)))+zmax
        gamma <- exp(apply(logplogff,2,function(y) y - logsumpff)) ## Posterior class probabilities
        return(gamma)
    }

    negLogLik <- function(p) {
        if (optim$constrain) {
            p[constrained] <- exp(p[constrained])
        }
        myp <- lapply(ParPos,function(x) p[x])
        K <- length(myp)
        prob <- p[seq(Npar+1,Npar+K-1)]; prob <- c(prob,1-sum(prob))
        logff <- sapply(1:length(myp), function(j) (logLik(mg$lvm[[j]],p=myp[[j]],data=data,indiv=TRUE,model=MODEL)))
        ## logff <- sapply(1:length(myp),
        ##              function(j) -normal_objective.lvm(mg$lvm[[j]],p=myp[[j]],data=data,indiv=TRUE))
        logplogff <- t(apply(logff,1, function(y) y+log(prob)))
        zmax <- apply(logplogff,1,max)
        logsumpff <- log(rowSums(exp(logplogff-zmax)))+zmax
        loglik <- sum(logsumpff)
        return(-loglik)
    }


    ObjEstep <- function(p,gamma,pr) {
        if (optim$constrain) {
            p[constrained] <- exp(p[constrained])
        }
        myp <- lapply(ParPos,function(x) p[x])
        loglik = 0;
        for (j in seq_along(myp))
            loglik = loglik + gamma[,j]*(logLik(mg$lvm[[j]],p=myp[[j]],data=data,indiv=TRUE,model=MODEL))
        ## logff <- sapply(1:length(myp), function(j) gamma[,j]*(logLik(mg$lvm[[j]],p=myp[[j]],data=data,indiv=TRUE,model=MODEL)))
        ## zmax <- apply(logff,1,max)
        ## ffz <- apply(logff,2,function(x) exp(x-zmax))
        ## logsff <- log(rowSums(ffz))+zmax
        ## loglik <- sum(logsff)
        return(-sum(loglik))
    }

    GradEstep <- function(p,gamma,pr) {
        if (optim$constrain) {
            p[constrained] <- exp(p[constrained])
        }
        myp <- lapply(ParPos,function(x) p[x])
        ## logff <- sapply(1:length(myp), function(j) log(gamma[,j])+(logLik(mg$lvm[[j]],p=myp[[j]],data=data,indiv=TRUE,model=MODEL)))
        ## Exp-sum normalization: exp(xi)/sum(exp(xi)) = exp(xi-b)*exp(b)/sum(exp(xi-b)exp(b))
        ## = exp(xi-b)/sum(exp(xi-b)),  b=max(xi)
        ## zmax <- apply(logff,1,max)
        ## ffz <- apply(logff,2,function(x) exp(x-zmax))
        ## sffz <- rowSums(ffz)
        D <- lapply(1:length(myp), function(j) {
            ## K <- ffz[,j]/sffz
            val <- score(mg$lvm[[j]],p=myp[[j]],data=data,indiv=TRUE,model=MODEL)
            apply(val,2,function(x) x*gamma[,j])
        })
        D0 <- matrix(0,nrow(data),length(p))
        for (j in 1:k) D0[,ParPos[[j]]] <- D0[,ParPos[[j]]]+D[[j]]
        S <- colSums(D0)
        if (optim$constrain) {
            S[constrained] <- S[constrained]*p[constrained]
        }
        return(-as.vector(S))
    }

   Information <- function(p,gamma,pr) {
        if (optim$constrain) {
            p[constrained] <- exp(p[constrained])
        }
        myp <- lapply(ParPos,function(x) p[x])
        D <- lapply(1:length(myp), function(j) {
            ## K <- ffz[,j]/sffz
            val <- score(mg$lvm[[j]],p=myp[[j]],data=data,indiv=TRUE,model=MODEL)
            apply(val,2,function(x) x*gamma[,j])
        })
        D0 <- matrix(0,nrow(data),length(p))
        for (j in 1:k) D0[,ParPos[[j]]] <- D0[,ParPos[[j]]]+D[[j]]
        if (optim$constrain) {
            for (j in constrained)
                D0[,j] <- D0[,j]*p[j]
        }
        S <- colSums(D0)
        structure(crossprod(D0), grad=S)
    }

    EMstep <- function(p,all=FALSE) {
        thetacur <- p[seq(Npar)]
        gamma <- PosteriorProb(p,constrain=optim$constrain)
        probcur <- colMeans(gamma)
        ## I <- function(p) {
        ##     I <- Information(p,gamma,probcur)
        ##     D <- attr(I, "grad")
        ##     res <- -Inverse(I)
        ##     res <- -I
        ##     attributes(res)$grad <- D
        ##     res
        ## }
        D <- function(p) GradEstep(p,gamma,probcur)
        ## if (optim$newton>0) {
        ##     newpar <- NR(thetacur, gradient=D)
        ## }
        ## if (mean(newpar$gradient^2)>optim$tol) {
        newpar <- nlminb(thetacur,function(p) ObjEstep(p,gamma,probcur), D)
        ## }
        thetacur <- newpar$par
        thetacur0 <- thetacur
        if (optim$constrain) {
            thetacur0[constrained] <- exp(thetacur[constrained])
        }
        p <- c(thetacur,probcur)
        if (all) {
            res <- list(p=p,gamma=gamma,
                        theta=rbind(thetacur0),
                        prob=rbind(probcur))
            return(res)
        }
        return(p)
    }
    em.idx <- match(c("K","method","square","step.min0","step.max0","mstep",
                      "objfn.inc","kr",
                      ##"keep.objfval","convtype",
                      "maxiter","tol","trace"),base::names(optim))
    em.control <- optim[na.omit(em.idx)]
    if (!is.null(em.control$trace)) em.control$trace <- em.control$trace>0

    p <- c(thetacur,probcur)
    opt <- SQUAREM::squarem(p,fixptfn=EMstep,##objfn=negLogLik,
                            control=em.control)
    ## opt2 <- nlminb(opt$par, function(p) negLogLik(p=p), control=list(trace=1))
    val <- EMstep(opt$par,all=TRUE)
    delta <- 1e-6
    if (any(val$prob<delta)) {
        val$prob[val$prob<delta] <- delta
        val$prob <- val$prob/sum(val$prob)
    }

    val <- c(val, list(member=apply(val$gamma,1,which.max),
                       k=k,
                       data=data,
                       parpos=ParPos,
                       multigroup=mg,
                       model=mg$lvm,
                       parname=parname,
                       logLik=NA,
                       opt=opt))
    class(val) <- "lvm.mixture"

    np <- length(coef(val))
    if (is.null(vcov)) {
        val$vcov <- matrix(NA,np,np)
    } else {
        I <- suppressWarnings(information.lvm.mixture(val,type=vcov))
        val$vcov <- tryCatch(Inverse(I),error=function(...) matrix(NA,np,np))
    }
    return(val)
}

###}}} mixture


###{{{ logLik, score, information

##' @export
score.lvm.mixture <- function(x,p=coef(x,full=TRUE),prob,indiv=FALSE,model="normal",...) {
    myp <- lapply(x$parpos,function(x) p[x])
    if (missing(prob)) {
        prob <- tail(p,x$k-1)
        p <- p[seq_len(length(p)-x$k+1)]
    }
    if (length(prob)<x$k)
        prob <- c(prob,1-sum(prob))
    logff <- sapply(seq(x$k), function(j) (logLik(x$multigroup$lvm[[j]],p=myp[[j]],data=x$data,indiv=TRUE,model=model)))
    logplogff <- t(apply(logff,1, function(y) y+log(prob)))
    zmax <- apply(logplogff,1,max)
    logsumpff <- log(rowSums(exp(logplogff-zmax)))+zmax
    aji <- apply(logplogff,2,function(x) exp(x-logsumpff))

    scoref <- lapply(score(x$multigroup,p=p,indiv=TRUE,model=model),
                     function(x) { x[which(is.na(x))] <- 0; x })

    Stheta <- matrix(0,ncol=ncol(scoref[[1]]),nrow=nrow(scoref[[1]]))
    Spi <- matrix(0,ncol=x$k-1,nrow=nrow(Stheta))
    for (j in 1:x$k) {
        Stheta <- Stheta + apply(scoref[[j]],2,function(x) x*aji[,j])
        if (j<x$k)
            Spi[,j] <- aji[,j]/prob[j] - aji[,x$k]/prob[x$k]
    }
    S <- cbind(Stheta,Spi)
    if (!indiv)
        return(colSums(S))
    return(S)
}

##' @export
information.lvm.mixture <- function(x,p=coef(x,full=TRUE),...,type="observed") {
    if (tolower(type)%in%c("obs","observed","hessian")) {
        S <- function(p=p,...) score.lvm.mixture(x,p=p,indiv=FALSE,...)
        I <- -numDeriv::jacobian(S,p)
        return(I)
    }
    S <- score.lvm.mixture(x,p=p,indiv=TRUE,...)
    res <- t(S)%*%S
    attributes(res)$grad <- colSums(S)
    return(res)
}

##' @export
logLik.lvm.mixture <- function(object,p=coef(object,full=TRUE),prob,model="normal",...) {
    if (is.null(object$parpos)) {
        object$parpos <- modelPar(object$multigroup,seq_along(p))$p
    }
    myp <- lapply(object$parpos, function(x) p[x])
    if (missing(prob))
        prob <- tail(p,object$k-1)
    if (length(prob)<object$k)
        prob <- c(prob,1-sum(prob))
    logff <- sapply(seq(object$k), function(j) (logLik(object$multigroup$lvm[[j]],p=myp[[j]],data=object$data,indiv=TRUE,model=model)))
    logplogff <- t(apply(logff,1, function(y) y+log(prob)))
    ## Log-sum-exp (see e.g. NR)
    zmax <- apply(logplogff,1,max)
    logsumpff <- log(rowSums(exp(logplogff-zmax)))+zmax
    loglik <- sum(logsumpff)
    npar <- length(p)
    nobs <- nrow(object$data)
    attr(loglik, "nall") <- nobs
    attr(loglik, "nobs") <- nobs
    attr(loglik, "df") <- npar
    class(loglik) <- "logLik"
    return(loglik)
}

###}}} logLik, score, information

###{{{ vcov

##' @export
vcov.lvm.mixture <- function(object,...) {
    return(object$vcov)
}

###}}}

###{{{ summary/print

##' @export
summary.lvm.mixture <- function(object,type=0,labels=0,...) {
    ppos <- parpos(object$multigroup)
    parname <- attr(ppos,"name")
    naparname <- which(is.na(parname))
    parname[naparname]  <- object$multigroup$name[naparname]
    ParPos <- modelPar(object$multigroup,seq_along(coef(object)))$p
    mm <- object$multigroup$lvm
    p <- coef(object,list=TRUE)
    p0 <- coef(object,prob=FALSE)
    myp <- modelPar(object$multigroup,1:length(p0))$p
    coefs <- list()
    ncluster <- c()
    Coefs <- matrix(NA,ncol=4,nrow=length(parname))
    colnames(Coefs) <- c("Estimate","Std. Error", "Z value", "Pr(>|z|)")
    Types <- rep("other",length(parname))
    Variable  <- rep(NA,length(parname))
    From  <- rep(NA,length(parname))
    Latent <- c()
    for (i in 1:length(mm)) {
        cc.idx <- order(coef(mm[[i]],p=seq_along(p[[i]]),type=2)[,1])

        cc <- coef(mm[[i]],p=p[[i]],vcov=vcov(object)[myp[[i]],myp[[i]]],data=NULL,labels=labels,type=2)
        Latent <- union(Latent,attr(cc,"latent"))
        Coefs[ParPos[[i]],] <- cc[cc.idx,,drop=FALSE]
        Types[ParPos[[i]]] <- attr(cc,"type")[cc.idx]
        Variable[ParPos[[i]]] <- attr(cc, "var")[cc.idx]
        From[ParPos[[i]]] <- attr(cc, "from")[cc.idx]
        cc <- CoefMat(mm[[i]],p=p[[i]],vcov=vcov(object)[myp[[i]],myp[[i]]],data=NULL,labels=labels)
        coefs <- c(coefs, list(cc))
        ncluster <- c(ncluster,sum(object$member==i))
    }
    rownames(Coefs) <- parname

    s2 <- tryCatch(sum(score(object)^2), error=function(...) NA)
    AIC <- tryCatch(AIC(object), error=function(...) NA)
    res <- list(coef=coefs, coefmat=Coefs, coeftype=Types,
                type=type, var=Variable, from=From, latent=Latent,
                ncluster=ncluster, prob=tail(object$prob,1),
                AIC=AIC, s2=s2)
    class(res) <- "summary.lvm.mixture"
    return(res)
}

##' @export
print.summary.lvm.mixture <- function(x,...) {
    if (x$type>0) {
        cc <- x$coefmat
        attr(cc,"type") <- x$coeftype
        attr(cc,"latent") <- x$latent
        attr(cc,"var") <- x$var
        attr(cc,"from") <- x$from
        cat("Mixing parameters:\n")
        cat("  ", paste(as.vector(formatC(x$prob))),"\n")
        print(CoefMat(cc), quote=FALSE)
        return(invisible())
    }
    for (i in 1:length(x$coef)) {
        cat("Cluster ",i," (n=",x$ncluster[i],", Prior=", formatC(x$prob[i]),"):\n",sep="")
        cat(rep("-",50),"\n",sep="")
        print(x$coef[[i]], quote=FALSE)
        if (i<length(x$coef)) cat("\n")
    }
    cat(rep("-",50),"\n",sep="")
    if (!is.na(x$AIC)) cat("AIC=",x$AIC,"\n")
    if (!is.na(x$s2)) cat("||score||^2=",x$s2,"\n")
    invisible()
}

##' @export
print.lvm.mixture <- function(x,...) {
    ## for (i in 1:x$k) {
    ##     cat("Cluster ",i," (pr=",x$prob[i],"):\n",sep="")
    ##     cat(rep("-",50),"\n",sep="")
    ##     print(coef(x,prob=FALSE)[x$parpos[[i]]], quote=FALSE)
    ##     cat("\n")
    ## }
    print(summary(x,type=1,...))
}

###}}}

###{{{ plot

##' @export
plot.lvm.mixture <- function(x,type="l",...) {
    matplot(x$theta,type=type,...)
}

###}}} plot

###{{{ coef

##' @export
coef.lvm.mixture <- function(object,iter,list=FALSE,full=TRUE,prob=FALSE,class=FALSE,label=TRUE,...) {
    N <- nrow(object$theta)
    res <- object$theta
    ## nn <- attr(parpos(object$multigroup),"name")
    ## if (length(ii <- which(is.na(nn)))>0 && label) {
    ##     nn[ii] <- paste0("p",seq_along(ii))
    ## }
    nn <- object$parname
    colnames(res) <- nn
    if (class) return(object$gammas)
    if (list) {
        res <- list()
        for (i in 1:object$k) {
            nn <- coef(object$multigroup$lvm[[i]])
            cc <- coef(object)[object$parpos[[i]]]
            base::names(cc) <- nn
            res <- c(res, list(cc))
        }
        return(res)
    }
    if (full) {
        pp <- object$prob[,seq(ncol(object$prob)-1),drop=FALSE]
        colnames(pp) <- paste0("pr",seq(ncol(pp)))
        res <- cbind(res,pp)
    }
    if (prob) {
        res <- object$prob
    }
    if (missing(iter))
        return(res[N,,drop=TRUE])
    else
        return(res[iter,])
}

###}}} coef

##' @export
model.frame.lvm.mixture <- function(formula,...) {
    return(formula$data)
}

##' @export
IC.lvm.mixture <- function(x,...) {
  U <- score(x,indiv=TRUE)
  bread <- vcov(x)*NROW(U)
  structure(t(bread%*%t(U)),bread=bread)
}

##' @export
manifest.lvm.mixture <- function(x,...) {
    manifest(x$multigroup,...)
}

## Return position of variance elements in the parameter vector (without mean parameters)
## Optimization constraints are needed on these parameters
##' @export
variances <- function(x,mean=FALSE) {
##  if (is.null(x$parpos))
##    x$parpos <- parpos(x)
  x$parpos <- parpos(Model(x),mean=TRUE)
  res <- diag(x$parpos$P)[which(diag(index(x)$P0)==1)]
  if (!mean) {
    return(res - index(x)$npar.mean)
  }
  return(res)
}
## And the off-diagonal (covariance) parameters
##' @export
offdiags <- function(x,mean=FALSE) {
  parpos <- parpos(x,mean=mean)
  pp <- parpos$P
  pp[lower.tri(pp)][(index(x)$P0)[lower.tri(pp)]==1]
}

#' @keywords internal
"_PACKAGE"

##' Estimation and simulation of latent variable models
##'
##' A general implementation of Structural Equation Models wth latent variables
##' (MLE, 2SLS, and composite likelihood estimators) with both continuous,
##' censored, and ordinal outcomes (Holst and Budtz-Joergensen (2013)
##' <doi:10.1007/s00180-012-0344-y>). Mixture latent variable models and
##' non-linear latent variable models (Holst and Budtz-Joergensen (2020)
##' <doi:10.1093/biostatistics/kxy082>). The package also provides methods for
##' graph exploration (d-separation, back-door criterion), simulation of general
##' non-linear latent variable models, and estimation of influence functions for
##' a broad range of statistical models.
##' @name lava-package
##' @importFrom graphics plot lines points abline points text layout par plot.new plot.window title rect locator segments image mtext box axis polygon matplot contour contour.default identify rug curve
##' @importFrom grDevices xy.coords col2rgb rgb colors rainbow topo.colors gray.colors palette colorRampPalette heat.colors
##' @importFrom utils stack combn read.csv getTxtProgressBar setTxtProgressBar txtProgressBar head tail modifyList getFromNamespace write.table methods data glob2rx
##' @importFrom stats density deriv effects lm family simulate vcov var cov cor coef model.frame model.weights as.formula model.matrix rnorm rchisq runif rlnorm pnorm qnorm na.omit AIC terms logLik qt pt update update.formula confint approxfun pchisq confint.default formula fft uniroot rbinom predict sd addmargins residuals dnorm quantile qf cov2cor qchisq get_all_vars p.adjust rpois rt rmultinom rgamma printCoefmat glm nlminb na.pass na.omit influence weights density.default
##' @importFrom survival is.Surv
##' @importFrom future.apply future_mapply future_lapply
##' @importFrom methods new as
##' @aliases lava-package lava
##' @author Klaus K. Holst Maintainer: <klaus@@holst.it>
##' @keywords package
##' @examples
##' lava()
NULL

##' Longitudinal Bone Mineral Density Data
##'
##' Bone Mineral Density Data consisting of 112 girls randomized to receive
##' calcium og placebo. Longitudinal measurements of bone mineral density
##' (g/cm^2) measured approximately every 6th month in 3 years.
##' @name calcium
##' @docType data
##' @format A data.frame containing 560 (incomplete) observations. The 'person'
##' column defines the individual girls of the study with measurements at
##' visiting times 'visit', and age in years 'age' at the time of visit. The
##' bone mineral density variable is 'bmd' (g/cm^2).
##' @source Vonesh & Chinchilli (1997), Table 5.4.1 on page 228.
##' @keywords datasets
NULL

##' Longitudinal Bone Mineral Density Data (Wide format)
##'
##' Bone Mineral Density Data consisting of 112 girls randomized to receive
##' calcium og placebo. Longitudinal measurements of bone mineral density
##' (g/cm^2) measured approximately every 6th month in 3 years.
##' @name bmd
##' @docType data
##' @source Vonesh & Chinchilli (1997), Table 5.4.1 on page 228.
##' @format data.frame
##' @keywords datasets
##' @seealso calcium
NULL

##' Simulated data
##'
##' Simulated data
##' @name brisa
##' @docType data
##' @format data.frame
##' @source Simulated
##' @keywords datasets
NULL

##' Data
##'
##' Description
##' @name bmidata
##' @docType data
##' @format data.frame
##' @keywords datasets
NULL

##' 50 patients from Monash Medical Centre, Melbourne
##'
##' Diagnosis of depression (DSM-III-R MDD, Dysthymia, Adjustment Disorder with
##' Depressed Mood, Depression NOS), Beck Depression Inventory (BDI) (Beck et
##' al., 1961) General Health Questionnaire (GHQ) (Goldberg & Williams, 1988)
##'
##' @name deprdiag
##' @docType data
##' @format data.frame
##' @source Clarke, D. M., Smith, G. C., & Herrman, H. E. (1993). A comparative
##'   study of screening instruments for mental disorders in general hospital
##'   patients. International Journal Psychiatry in Medicine, 23, pp. 323-337.
##'
##' McKenzie et al. (1996). Comparing correlated Kappas by resampling: Is one
##' level of agreement significantly different from another? J. Psychiat. Res.
##' 30 (6), pp. 483-492.
##' @keywords datasets
NULL

##' Hubble data
##'
##' Velocity (v) and distance (D) measures of 36 Type Ia super-novae from the
##' Hubble Space Telescope
##' @name hubble
##' @docType data
##' @format data.frame
##' @source Freedman, W. L., et al. 2001, AstroPhysicalJournal, 553, 47.
##' @keywords datasets
NULL

##' Hubble data
##'
##' @name hubble2
##' @seealso hubble
##' @docType data
##' @format data.frame
##' @keywords datasets
NULL

##' Data
##'
##' Description
##' @name indoorenv
##' @docType data
##' @format data.frame
##' @source Simulated
##' @keywords datasets
NULL

##' Missing data example
##'
##' Simulated data generated from model
##' \deqn{E(Y_i\mid X) = X, \quad cov(Y_1,Y_2\mid X)=0.5}
##'
##' The list contains four data sets
##' 1) Complete data
##' 2) MCAR
##' 3) MAR
##' 4) MNAR (missing mechanism depends on variable V correlated with Y1,Y2)
##' @examples
##' data(missingdata)
##' e0 <- estimate(lvm(c(y1,y2)~b*x,y1~~y2),missingdata[[1]]) ## No missing
##' e1 <- estimate(lvm(c(y1,y2)~b*x,y1~~y2),missingdata[[2]]) ## CC (MCAR)
##' e2 <- estimate(lvm(c(y1,y2)~b*x,y1~~y2),missingdata[[2]],missing=TRUE) ## MCAR
##' e3 <- estimate(lvm(c(y1,y2)~b*x,y1~~y2),missingdata[[3]]) ## CC (MAR)
##' e4 <- estimate(lvm(c(y1,y2)~b*x,y1~~y2),missingdata[[3]],missing=TRUE) ## MAR
##' @name missingdata
##' @docType data
##' @format list of data.frames
##' @source Simulated
##' @keywords datasets
NULL

##' Example data (nonlinear model)
##'
##' @name nldata
##' @docType data
##' @format data.frame
##' @source Simulated
##' @keywords datasets
NULL

##' Example SEM data (nonlinear)
##'
##' Simulated data
##' @name nsem
##' @docType data
##' @format data.frame
##' @source Simulated
##' @keywords datasets
NULL

##' Example SEM data
##'
##' Simulated data
##' @name semdata
##' @docType data
##' @source Simulated
##' @format data.frame
##' @keywords datasets
NULL

##' Serotonin data
##'
##' This simulated data mimics a PET imaging study where the 5-HT2A
##' receptor and serotonin transporter (SERT) binding potential has
##' been quantified into 8 different regions. The 5-HT2A
##' cortical regions are considered high-binding regions
## 'which are a priori known to yield quite similar and highly correlated
##' measurements.  These measurements can be regarded as proxy measures of
##' the extra-cellular levels of serotonin in the brain
##' \tabular{rll}{
##'         day    \tab numeric \tab Scan day of the year \cr
##'         age    \tab numeric \tab Age at baseline scan \cr
##'         mem    \tab numeric \tab Memory performance score \cr
##'         depr   \tab numeric \tab Depression (mild) status 500 days after baseline \cr
##'         gene1  \tab numeric \tab Gene marker 1 (HTR2A) \cr
##'         gene2  \tab numeric \tab Gene marker 2 (HTTTLPR) \cr
##'         cau \tab numeric \tab SERT binding, Caudate Nucleus \cr
##'         th  \tab numeric \tab SERT binding, Thalamus \cr
##'         put \tab numeric \tab SERT binding, Putamen \cr
##'         mid \tab numeric \tab SERT binding, Midbrain \cr
##'         aci \tab numeric \tab 5-HT2A binding, Anterior cingulate gyrus \cr
##'         pci  \tab numeric \tab 5-HT2A binding, Posterior cingulate gyrus \cr
##'         sfc \tab numeric \tab 5-HT2A binding, Superior frontal cortex \cr
##'         par \tab numeric \tab 5-HT2A binding, Parietal cortex \cr
##' }
##' @name serotonin
##' @docType data
##' @format data.frame
##' @source Simulated
##' @keywords datasets
NULL

##' Twin menarche data
##'
##' Simulated data
##' \tabular{rll}{
##'         id    \tab numeric \tab Twin-pair id \cr
##'         zyg    \tab character \tab Zygosity (MZ or DZ) \cr
##'         twinnum    \tab numeric \tab Twin number (1 or 2) \cr
##'         agemena    \tab numeric \tab Age at menarche (or censoring) \cr
##'         status    \tab logical \tab Censoring status (observed:=T,censored:=F) \cr
##'         bw    \tab numeric  \tab Birth weight \cr
##'         msmoke    \tab numeric \tab Did mother smoke? (yes:=1,no:=0) \cr
##' }
##' @name twindata
##' @docType data
##' @format data.frame
##' @keywords datasets
##' @source Simulated
NULL

##' For internal use
##'
##' @title For internal use
##' @name startvalues
##' @author Klaus K. Holst
##' @keywords internal
##' @rdname internal
##' @export
##' @aliases startvalues startvalues0 startvalues1 startvalues2
##' @aliases starter.multigroup
##' @aliases modelPar modelVar matrices pars pars.lvm regfix
##' @aliases pars.lvmfit pars.glm score.glm procdata.lvmfit mat.lvm
##' @aliases matrices reorderdata graph2lvm igraph.lvm subgraph finalize
##' @aliases randomslope randomslope<- lisrel
##' @aliases variances offdiags describecoef parlabels rsq stdcoef CoefMat
##' @aliases CoefMat.multigroupfit deriv updatelvm checkmultigroup profci
##' @aliases estimate.MAR
##' @aliases missingModel Identical gaussian_logLik.lvm
##' @aliases addhook gethook multigroup
##' @aliases Weights fixsome IV parameter Specials
##' @aliases procformula getoutcome decomp.specials
##' @aliases rmvn0 dmvn0 logit expit tigol
NULL

##' @export
predict.lvm.mixture <- function(object,p=coef(object,full=TRUE),model="normal",predict.fun=NULL,...) {
    p0 <- coef(object,full=FALSE)
    pp <- p[seq_along(p0)]
    pr <- p[length(p0)+seq(length(p)-length(p0))];
    if (length(pr)<object$k) pr <- c(pr,1-sum(pr))
    myp <- modelPar(object$multigroup,p=pp)$p
    logff <- sapply(seq(object$k), function(j) (logLik(object$multigroup$lvm[[j]],p=myp[[j]],data=object$data,indiv=TRUE,model=model)))
    logplogff <- t(apply(logff,1, function(y) y+log(pr)))
    zmax <- apply(logplogff,1,max)
    logsumpff <- log(rowSums(exp(logplogff-zmax)))+zmax
    gamma <- exp(apply(logplogff,2,function(y) y - logsumpff)) ## Posterior class probabilities, conditional mean
    ##Vgamma <- gamma-gamma^2 ## conditional variance
    M <- 0; V <- 0
    x <- lava::vars(object$model)
    for (i in seq(object$k)) {
        m <- Model(object$multigroup)[[i]]
        P <- predict(m,data=object$data,p=myp[[i]],x=x)
        if (!is.null(predict.fun)) {
            M <- M+gamma[,i]*predict.fun(P,as.vector(attributes(P)$cond.var), ...)
        } else {
            M <- M+gamma[,i]*P
            V <- V+gamma[,i]*as.vector(attributes(P)$cond.var) + gamma[,i]*P^2
        }
    }
    ## Var(X) = E[(X-m)^2] = EX^2 - m^2 
    ##        = sum(pi*Xi^2) - [sum(pi*mi)]^2 = sum(pi*sigmai^2) + sum(pi*mi^2) - [sum(pi*mi)]^2
    V <- V-M^2    
    return(M)
    ## if (!is.null(predict.fun)) return(M)
    ## structure(M,cond.var=V)
}

strip_bracket <- function(s) {
  if (length(s) > 1L) {
    return(unlist(lapply(s, function(x) strip_bracket(x))))
  }
  # Check if it starts and ends with brackets
  if (!grepl("^\\[.*\\]$", s)) return(s)

  # Calculate cumulative bracket depth
  chars <- strsplit(s, "")[[1]]
  depth <- 0
  is_wrapped <- TRUE

  for (i in 1:(length(chars) - 1)) {
    if (chars[i] == "[") depth <- depth + 1
    if (chars[i] == "]") depth <- depth - 1

    # If depth hits 0 before the end, it's side-by-side like [a] + [b]
    if (depth == 0 && i < length(chars)) {
      is_wrapped <- FALSE
      break
    }
  }

  if (is_wrapped) {
    return(sub("^\\[(.*)\\]$", "\\1", s))
  } else {
    return(s)
  }
}

substArg <- function(x,env,...) {
  if (!missing(env)) {
    a <- with(env,substitute(x))
  } else {
    a <- substitute(x)
  }
  myclass <- tryCatch(class(eval(a)),error=function(e) NULL)
  if (is.null(myclass) || myclass=="name") {
      res <- unlist(sapply(as.character(a),
                           function(z) {
                               trimmed <- gsub(" ","",z,fixed=TRUE)
                               val <- strsplit(trimmed,"+",fixed=TRUE)
                               if (val[1]=="") val <- NULL
                               val
                           })); attributes(res)$names <- NULL
      return(res)
  }
  return(eval(a))
}

## g <- function(zz,...) {
##   env=new.env(); assign("x",substitute(zz),env)
##   substArg(zz,env=env)
## }
## h <- function(x,...) {
##   env=new.env(); assign("x",substitute(x),env)
##   substArg(x,env=TRUE)
## }


###{{{ procrandomslope

procrandomslope <- function(object,data=object$data,...) {
  Xfix <- FALSE
  xfix <- myfix <- list()
  xx <- object
  for (i in seq_len(object$ngroup)) {
    x0 <- object$lvm[[i]]
    data0 <- data[[i]]
    xfix0 <- colnames(data0)[(colnames(data0)%in%parlabels(x0,exo=TRUE))]
    xfix <- c(xfix, list(xfix0))
    if (length(xfix0)>0) { ## Yes, random slopes
      Xfix<-TRUE
    }
    xx$lvm[[i]] <- x0
  }
  if (Xfix) {
    for (k in seq_len(object$ngroup)) {
      x0 <- object$lvm[[k]]
      data0 <- data[[k]]
      nrow <- length(vars(x0))
      xpos <- lapply(xfix[[k]],function(y) which(regfix(x0)$labels==y))
      colpos <- lapply(xpos, function(y) ceiling(y/nrow))
      rowpos <- lapply(xpos, function(y) (y-1)%%nrow+1)
      myfix0 <- list(var=xfix[[k]], col=colpos, row=rowpos)
      myfix <- c(myfix, list(myfix0))
      for (i in seq_along(myfix0$var))
        for (j in seq_along(myfix0$col[[i]]))
          regfix(x0,
                 from=vars(x0)[myfix0$row[[i]][j]],to=vars(x0)[myfix0$col[[i]][j]]) <-
                   colMeans(data0[,myfix0$var[[i]],drop=FALSE],na.rm=TRUE)
      index(x0) <- reindex(x0,zeroones=TRUE,deriv=TRUE)
      object$lvm[[k]] <- x0
    }
    object <- multigroup(object$lvm,data,fix=FALSE,exo.fix=FALSE)
  }
  return(list(model=object,fix=myfix))
}

###}}} procrandomslope

###{{{ kronprod

## ' Calculate matrix product with kronecker product
## '
## ' \deqn{(A\crossprod B) Y}
## ' @title Calculate matrix product with kronecker product
## ' @param A
## ' @param B
## ' @param Y
## ' @author Klaus K. Holst
kronprod <- function(A,B,Y) {
    if (missing(Y)) {
        ## Assume 'B'=Identity, (A otimes B)Y
        k <- nrow(B)/ncol(A)
        res <- rbind(apply(B,2,function(x) matrix(x,nrow=k)%*%t(A)))
        return(res)
    }
    rbind(apply(Y,2,function(x) B%*%matrix(x,nrow=ncol(B))%*%t(A)))
}

###}}} kronprod

###{{{ izero

izero <- function(i,n) { ## n-1 zeros and 1 at ith entry
  x <- rep(0,n); x[i] <- 1
  x
}

###}}}

###{{{ Debug

`Debug` <-
  function(msg, cond=lava.options()$debug) {
    if (cond)
      print(paste(msg, collapse=" "))
  }

###}}}

###{{{ categorical2dummy

categorical2dummy <- function(x,data,messages=0,...) {
  x0 <- x
  X <- intersect(index(x)$exogenous,colnames(data))
  catX <- c()
  for (i in X) {
    if (!is.numeric(data[,i])) catX <- c(catX,i)
  }
  if (length(catX)==0) return(list(x=x,data=data))
  f <- as.formula(paste("~ 1+", paste(catX,collapse="+")))
  opt <- options(na.action="na.pass")
  M <- model.matrix(f,data)

  options(opt)
  Mnames <- colnames(M)
  Mpos <- attributes(M)$assign
  A <- index(x)$A
  F <- regfix(x)
  count <- 0
  for (i in catX) {
    count <- count+1
    mnames <- Mnames[Mpos==count]
    kill(x0) <- i
    Y <- colnames(A)[A[i,]==1]
    if (length(mnames)==1) {
      fix <- as.list(F$labels[i,])
      fixval <- F$values[i,]
      fix[which(!is.na(fixval))] <- fixval[na.omit(fixval)]
      regression(x0,to=Y,from=mnames,messages=messages) <- fix[Y]
    } else {
      x0 <- regression(x0,to=Y,from=mnames,messages=messages)
    }
  }
  index(x0) <- reindex(x0,zeroones=TRUE,deriv=TRUE)
  return(list(x=x0,data=cbind(data,M)))
}

###}}}

###{{{ procdata.lvm

`procdata.lvm` <-
  function(x,data,categorical=FALSE,
    na.method=ifelse(any(is.na(data[,intersect(colnames(data),manifest(x))])),"complete.obs","pairwise.complete.obs"),
    missing=FALSE
    ) {
    if (is.numeric(data) & !is.list(data)) {
      data <- rbind(data)
    }
     if (is.data.frame(data) | is.matrix(data)) {
      nn <- colnames(data)
      data <- as.data.frame(data); colnames(data) <- nn; rownames(data) <- NULL
      obs <- setdiff(intersect(vars(x), colnames(data)),latent(x))
      Debug(obs)
      mydata <- subset(data, select=obs)
      if (NROW(mydata)==0) stop("No observations")
      for (i in seq_len(ncol(mydata))) {
        if (inherits(mydata[,i],"Surv"))
          mydata[,i] <- mydata[,i][,1]
        if (is.character(mydata[,i]) | is.factor(mydata[,i]))
          mydata[,i] <- as.numeric(as.factor(mydata[,i]))-1
      }

      S <- NULL
      n <- nrow(mydata)
      if (n==1) {
        S <- diag(nrow=ncol(mydata)); colnames(S) <- rownames(S) <- obs
      }
      if (na.method=="complete.obs" && !missing) {
        mydata0 <- na.omit(mydata)
        n <- nrow(mydata0)
        mu <- colMeans(mydata0)
        if (is.null(S) && n>2)
            S <- (n-1)/n*cov(mydata0) ## MLE variance matrix of observed variables
        rm(mydata0)
      }
      nS <- is.null(S) || any(is.na(S))
      if (na.method=="pairwise.complete.obs" || nS) {
          mu <- colMeans(mydata,na.rm=TRUE)
          if (nS) {
              n <- nrow(mydata)
              S <- (n-1)/n*cov(mydata,use="pairwise.complete.obs")
              S[is.na(S)] <- 1e-3
          }
      }
    }
    else
      if (is.list(data)) {
        if ("cov"%in%names(data)) data$S <- data$cov
        if ("var"%in%names(data)) data$S <- data$var
        if ("mean"%in%names(data)) data$mu <- data$mean
        n <- data$n
        S <- reorderdata.lvm(x,data$S)
        mu <- reorderdata.lvm(x,data$mu)
        ##      if (is.null(n)) stop("n was not specified");
      }
      else
        stop("Unexpected type of data!");
    if (nrow(S)!=ncol(S)) stop("Wrong type of data!");
    return(list(S=S,mu=mu,n=n))
  }

###}}}

###{{{ reorderdata.lvm

`reorderdata.lvm` <-
  function(x, data) {
    if (is.vector(data)) {
      nn <- names(data)
      ii <- na.omit(match(index(x)$manifest, nn))
      data[ii,drop=FALSE]
    } else {
      nn <- colnames(data)
      ii <- na.omit(match(index(x)$manifest, nn))
      data[ii,ii,drop=FALSE]
    }
  }

###}}}

###{{{ symmetrize

`symmetrize` <-
function(M, upper=TRUE) {
  if (length(M)==1) return(M)
  if (!is.matrix(M) | ncol(M)!=nrow(M)) stop("Only implemented for square matrices.")
  if (upper) {
    for (i in seq_len(ncol(M)-1))
      for (j in seq(i+1,nrow(M)))
        M[i,j] <- M[j,i]
    return(M)
  } else {
    for (i in seq_len(ncol(M)))
      for (j in seq_len(nrow(M)))
        if (M[i,j]==0)
          M[i,j] <- M[j,i]
        else
          M[j,i] <- M[i,j]
    return(M)
  }
}

###}}}

###{{{ naiveGrad

naiveGrad <- function(f, x, h=1e-9) {
  nabla <- numeric(length(x))
  for (i in seq_along(x)) {
    xh <- x; xh[i] <- x[i]+h
    nabla[i] <- (f(xh)-f(x))/h
  }
  return(nabla)
}

###}}}

###{{{ CondMom

# conditional on Compl(idx)
CondMom <- function(mu,S,idx,X) {
  idxY <- idx

  idxX <- setdiff(seq_len(ncol(S)),idxY)
  SXX <- S[idxX,idxX,drop=FALSE];
  SYY <- S[idxY,idxY,drop=FALSE]
  SYX <- S[idxY,idxX,drop=FALSE]
  iSXX <- solve(SXX)
  condvar <- SYY-SYX%*%iSXX%*%t(SYX)
  if (missing(mu)) return(condvar)

  muY <- mu[,idxY,drop=FALSE]
  muX <- mu[,idxX,drop=FALSE]
  if (is.matrix(mu))
    Z <- t(X-muX)
  else
    Z <- apply(X,1,function(xx) xx-muX)
  SZ  <- t(SYX%*%iSXX%*%Z)
##  condmean <- matrix(
  if (is.matrix(mu))
    condmean <- SZ+muY
  else
    condmean <- t(apply(SZ,1,function(x) muY+x))
##  ,ncol=ncol(SZ),nrow=nrow(SZ))
  return(list(mean=condmean,var=condvar))
}

###}}} CondMom

###{{{ Depth-First/acc (accessible)

DFS <- function(M,v,explored=c()) {
  explored <- union(explored,v)
  incident <- M[v,]
  for (v1 in setdiff(which(incident==1),explored)) {
    explored <- DFS(M,v1,explored)
  }
  return(explored)
}

acc <- function(M,v) {
  if (is.character(v)) v <- which(colnames(M)==v)
  colnames(M)[setdiff(DFS(M,v),v)]
}

###}}} Depth-First/acc (accessible)


npar.lvm <- function(x) {
  return(index(x)$npar+ index(x)$npar.mean+index(x)$npar.ex)

}

##' @export
as.numeric.list <- function(x, ...) {
  lapply(x, function(y) {
    ifelse(is.na(as.numeric(y)), y, as.numeric(y))
  })
}

edge2pair <- function(e) {
  sapply(e, function(x) strsplit(x, "~"))
}

numberdup <- function(xx) { ## Convert to numbered list
  dup.xx <- duplicated(xx)
  ## dups <- xx[dup.xx]
  xx.new <- numeric(length(xx))
  count <- 0
  for (i in seq_along(xx)) {
    if (!dup.xx[i]) {
      count <- count+1
      xx.new[i] <- count
    } else {
      xx.new[i] <- xx.new[match(xx[i],xx)[1]]
    }
  }
  return(xx.new)
}

extractvar <- function(f) {
    yy <- getoutcome(f)
    xx <- attributes(terms(f))$term.labels
    myvars <- all.vars(f)
    return(list(y=yy,x=xx,all=myvars))
}

##' @export
getoutcome <- function(formula,sep,...) {
  aa <- attributes(terms(formula,...))
  if (aa$response==0) {
    res <- NULL
  } else {
    res <- paste(deparse(formula[[2]]),collapse="")
  }
  if (!missing(sep) && length(aa$term.labels)>0) {
      attributes(res)$x <- lapply(strsplit(aa$term.labels,"\\|")[[1]],
                                  function(x) as.formula(paste0("~",x)))
  } else {
      attributes(res)$x <- aa$term.labels
  }
  return(res)
}


##' @export
Specials <- function(f,spec,split2="+",...) {
  tt <- terms(f,spec)
  pos <- attributes(tt)$specials[[spec]]
  if (is.null(pos)) return(NULL)
  x <- rownames(attributes(tt)$factors)[pos]
  st <- gsub(" ","",x)
  res <- unlist(strsplit(st,"[()]"))[2]
  if (is.null(split2)) return(res)
  unlist(strsplit(res,split2,fixed=TRUE))
}


##' @export
decomp.specials <- function(x,pattern="[()]",pattern2=NULL, pattern.ignore=NULL, sep="[,\\+]",perl=TRUE,reverse=FALSE,...) {
  st <- gsub(" |^\\(|)$","",x) # Remove white space and leading/trailing parantheses
  if (!is.null(pattern.ignore)) {
      if (grepl(pattern.ignore,st,perl=perl,...)) return(st)
  }
  if (!is.null(pattern)) {
    st <- rev(unlist(strsplit(st,pattern,perl=perl,...)))[1]
  }
  if (!is.null(pattern2)) {
    st <- (unlist(strsplit(st,pattern2,perl=perl,...)))
    if (reverse) st <- rev(st)
  }
  unlist(strsplit(st,sep,perl=perl,...))
}

##' @export
`%++%.lvm` <- function(x, y) merge(x, y)

##' @export
"+.lvm" <- function(x, ...) {
  merge(x, ...)
}

##' @export
merge.lvm <- function(x, y, ...) {
  objects <- list(x, y, ...)
  if (length(objects)<2) return(x)
  m <- objects[[1]]
  for (i in seq(2, length(objects))) {
    m2 <- objects[[i]]
    if (length(latent(m2))>0)
      latent(m) <- latent(m2)
    if (length(m2$constrain)>0)
      m$constrain <- c(m$constrain,m2$constrain)
    M <- (index(m2)$M)
    P <- (index(m2)$P)
    nn <- vars(m2)
    for (j in seq_len(nrow(M))) {
      if (any(idx <- M[j,]!=0)) {
        val <- as.list(rep(NA,sum(idx==TRUE)))
        if (any(idx. <- !is.na(m2$par[j,idx])))
          val[idx.] <- m2$par[j,idx][idx.]
        if (any(idx. <- !is.na(m2$fix[j,idx])))
          val[idx.] <- m2$fix[j,idx][idx.]
        regression(m,to=nn[idx],from=nn[j],messages=0) <- val
      }

      P0 <- P[j,]; P0[seq_len(j-1)] <- 0
      idx <- P[j,]!=0 | m2$covfix[j,]==0
      idx[is.na(idx)] <- FALSE
      if (any(idx)) {
          val <- as.list(rep(NA,sum(idx==TRUE)))
          if (any(idx. <- !is.na(m2$covpar[j,idx])))
            val[idx.] <- m2$covpar[j,idx][idx.]
          if (any(idx. <- !is.na(m2$covfix[j,idx])))
            val[idx.] <- m2$covfix[j,idx][idx.]
          covariance(m,nn[idx],nn[j],messages=0) <- val
      }
    }
    intercept(m,nn) <- intercept(m2)
    m2x <- exogenous(m2)
    if (length(m2x)>0)
      exogenous(m) <- c(exogenous(m),m2x)
  }
  index(m) <- reindex(m)
  return(m)
}

##' @export
merge.list <- function(x, ...) {
  do.call(merge, c(x, list(...)))
}

##' @export
merge.lm <- function(x,y,...) {
  args <- c(list(x, y), list(...))
  nn <- names(formals(merge.estimate)[-seq(3)])
  idx <- na.omit(match(nn, names(args)))
  models <- args
  models[idx] <- NULL
  mm <- lapply(models, function(x) tryCatch(estimate(x),error=function(e) NULL))
  names(mm)[1:2] <- c("x", "y")
  ii <- which(unlist(lapply(mm, is.null)))
  if (length(ii) > 0) mm[ii] <- NULL
  do.call(merge,c(mm,args[idx]))
}

##' @export
merge.glm <- merge.lm

##' @export
merge.lvmfit <- merge.lm

##' @export
merge.multinomial <- function(x,...) {
    merge.estimate(x,...)
}

#' @export
estimate.data.frame <- function(x, ...) {
  estimate(as.matrix(x), ...)
}

IC_quantile <- function(x, estimate, probs=0.5, type = 7, ...) {
  x <- na.omit(x)
  f0 <- density(x, ...)
  ## U <- function(est) (tau - (x <= est))
  if (missing(estimate)) {
    estimate <- quantile(x, probs=probs, type = type, ...)
  }
  res <- c()
  for (i in seq_len(length(estimate))) {
    res <- cbind(res,
    (probs[i] - (x <= estimate[i]))/with(f0, approx(x, y, estimate[i]))$y
    )
  }
  res
}

#' Estimate parameters and influence function.
#'
#' Estimate parameters for the sample mean, variance, and quantiles
#' @export
#' @aliases estimate.array estimate.data.frame
#' @param x numeric matrix
#' @param type target parameter ("mean", "variance", "quantile"). The type of
#'   quantile (see `stats::quantile`) can be chosen with "quantile1",
#'   "quantile2", ...
#' @param probs numeric vector of probabilities (for type="quantile")
#' @param ... Additional arguments to lower level functions (i.e.,
#'   stats::density.default when type="quantile")
#' @return Object of class `estimate` (see [estimate.default]).
estimate.array <- function(x, type = "mean", probs = 0.5, ...) {
  cl <- match.call()
  if (missing(x) || is.null(x)) {
    return(estimate(NULL, ...))
  }
  dots <- list(...)
  density.args <- dots[]
  cc <- apply(x, 2, function(y) mean(y, na.rm = TRUE))
  ic <- apply(x, 2, function(y) y - mean(y, na.rm = TRUE))
  if (!("id" %in% names(dots))) {
    dots[["id"]] <- rownames(x)
  }
  if (tolower(type) %in% c("var", "variance")) {
    cc <- apply(x, 2,
                function(y) mean((y - mean(y, na.rm = TRUE))^2,
                                 na.rm = TRUE))
    ic <- ic^2
    for (i in seq_len(NCOL(ic))) {
      ic[, i] <- ic[, i] - cc[i]
    }
  }
  if (grepl("^quantile", tolower(type))) {
    m <- regexpr("\\d+$", type)
    quantile_type <- ifelse(m > 0,
                            as.integer(regmatches(type, m)), 7)
    density.args <- list()
    dargs <- names(formals(density.default))
    didx <- which(dargs %in% names(dots))
    if (length(didx)>0) {
      density.args <- dots[dargs[didx]]
      dots[dargs[didx]] <- NULL
    }
    cc <- unlist(apply(x, 2, function(y)
      quantile(y, probs=probs, na.rm = TRUE, type = quantile_type),
      simplify=FALSE))
    ic <- c()
    for (i in seq_len(NCOL(x))) {
      ic <- cbind(ic, do.call(IC_quantile,
                              c(list(x[, i], probs=probs, type = quantile_type),
                                density.args)))
    }
  }
  if (any(c("vcov", "IC") %in% names(list(...)))) {
    return(estimate(NULL, coef = cc, ...))
  }
  res <- do.call(estimate, c(list(NULL, coef = cc, IC = ic), dots))
  res$call <- cl
  return(res)
}

#' @export
estimate.data.frame <- function(x, ...) {
  estimate(as.matrix(x), ...)
}

IC_quantile <- function(x, estimate, probs=0.5, ...) {
  x <- na.omit(x)
  f0 <- density(x, ...)
  ## U <- function(est) (tau - (x <= est))
  if (missing(estimate)) {
    estimate <- quantile(x, probs=probs)
  }
  res <- c()
  for (i in seq_len(length(estimate))) {
    res <- cbind(res,
    (probs[i] - (x <= estimate[i]))/with(f0, approx(x, y, estimate[i]))$y
    )
  }
  res
}

#' Estimate parameters and influence function.
#'
#' Estimate parameters for the sample mean, variance, and quantiles
#' @export
#' @aliases estimate.array estimate.data.frame
#' @param x numeric matrix
#' @param type target parameter ("mean", "variance", "quantile")
#' @param probs numeric vector of probabilities (for type="quantile")
#' @param ... Additional arguments to lower level functions (i.e.,
#'   stats::density.default when type="quantile")
#' @return Object of class `estimate` (see [estimate.default]).
estimate.array <- function(x, type="mean", probs=0.5, ...) {
  cl <- match.call()
  if (missing(x) || is.null(x)) {
    return(estimate(NULL, ...))
  }
  dots <- list(...)
  density.args <- dots[]
  cc <- apply(x, 2, function(y) mean(y, na.rm = TRUE))
  ic <- apply(x, 2, function(y) y - mean(y, na.rm = TRUE))
  if (tolower(type) %in% c("var", "variance")) {
    cc <- apply(x, 2, function(y) mean((y - mean(y, na.rm=TRUE))^2, na.rm = TRUE))
    ic <- ic^2
    for (i in seq_len(NCOL(ic))) {
      ic[, i] <- ic[, i] - cc[i]
    }
  }
  if (tolower(type) %in% c("quantile")) {
    density.args <- list()
    dargs <- names(formals(density.default))
    didx <- which(dargs %in% names(dots))
    if (length(didx)>0) {
      density.args <- dots[dargs[didx]]
      dots[dargs[didx]] <- NULL
    }
    cc <- unlist(apply(x, 2, function(y)
      quantile(y, probs=probs, na.rm = TRUE),
      simplify=FALSE))
    ic <- c()
    for (i in seq_len(NCOL(x))) {
      ic <- cbind(ic, do.call(IC_quantile,
                        c(list(x[, i], probs=probs), density.args)))
    }
  }
  if (any(c("vcov", "IC") %in% names(list(...)))) {
    return(estimate(NULL, coef = cc, ...))
  }
  res <- do.call(estimate, c(list(NULL, coef = cc, IC = ic), dots))
  res$call <- cl
  return(res)
}

##' Define range constraints of parameters
##'
##' @aliases Range.lvm
##' @title Define range constraints of parameters
##' @param a Lower bound
##' @param b Upper bound
##' @return function
##' @author Klaus K. Holst
##' @export
Range.lvm <- function(a=0,b=1) {
    if (b==Inf) {
        f <- function(x) {
            res <- a+exp(x)
            attributes(res)$grad <- exp
            res
        }
        return(f)
    }
    if (a==-Inf) {
        f <- function(x) {
            res <- -exp(x)+b
            attributes(res)$grad <- function(x) -exp(x)
            res
        }
        return(f)
    }
    f <- function(x) {
        res <- (a+b*exp(x))/(1+exp(x))
        attributes(res)$grad <- function(x) exp(x)*(b-a-a*b*exp(x))/(1+exp(x))^2
        res
    }
    return(f)
}

##' Add non-linear constraints to latent variable model
##'
##' Add non-linear constraints to latent variable model
##'
##' Add non-linear parameter constraints as well as non-linear associations
##' between covariates and latent or observed variables in the model (non-linear
##' regression).
##'
##' As an example we will specify the follow multiple regression model:
##'
##' \deqn{E(Y|X_1,X_2) = \alpha + \beta_1 X_1 + \beta_2 X_2} \deqn{V(Y|X_1,X_2)
##' = v}
##'
##' which is defined (with the appropiate parameter labels) as
##'
##' \code{m <- lvm(y ~ f(x,beta1) + f(x,beta2))}
##'
##' \code{intercept(m) <- y ~ f(alpha)}
##'
##' \code{covariance(m) <- y ~ f(v)}
##'
##' The somewhat strained parameter constraint \deqn{ v =
##' \frac{(beta1-beta2)^2}{alpha}}
##'
##' can then specified as
##'
##' \code{constrain(m,v ~ beta1 + beta2 + alpha) <- function(x)
##' (x[1]-x[2])^2/x[3] }
##'
##' A subset of the arguments \code{args} can be covariates in the model,
##' allowing the specification of non-linear regression models.  As an example
##' the non-linear regression model \deqn{ E(Y\mid X) = \nu + \Phi(\alpha +
##' \beta X)} where \eqn{\Phi} denotes the standard normal cumulative
##' distribution function, can be defined as
##'
##' \code{m <- lvm(y ~ f(x,0)) # No linear effect of x}
##'
##' Next we add three new parameters using the \code{parameter} assigment
##' function:
##'
##' \code{parameter(m) <- ~nu+alpha+beta}
##'
##' The intercept of \eqn{Y} is defined as \code{mu}
##'
##' \code{intercept(m) <- y ~ f(mu)}
##'
##' And finally the newly added intercept parameter \code{mu} is defined as the
##' appropiate non-linear function of \eqn{\alpha}, \eqn{\nu} and \eqn{\beta}:
##'
##' \code{constrain(m, mu ~ x + alpha + nu) <- function(x)
##' pnorm(x[1]*x[2])+x[3]}
##'
##' The \code{constraints} function can be used to show the estimated non-linear
##' parameter constraints of an estimated model object (\code{lvmfit} or
##' \code{multigroupfit}). Calling \code{constrain} with no additional arguments
##' beyound \code{x} will return a list of the functions and parameter names
##' defining the non-linear restrictions.
##'
##' The gradient function can optionally be added as an attribute \code{grad} to
##' the return value of the function defined by \code{value}. In this case the
##' analytical derivatives will be calculated via the chain rule when evaluating
##' the corresponding score function of the log-likelihood. If the gradient
##' attribute is omitted the chain rule will be applied on a numeric
##' approximation of the gradient.
##' @aliases constrain constrain<- constrain.default constrain<-.multigroup
##' @aliases constrain<-.default constraints parameter<-
##' @return A \code{lvm} object.
##' @author Klaus K. Holst
##' @seealso \code{\link{regression}}, \code{\link{intercept}},
##' \code{\link{covariance}}
##' @keywords models regression
##' @examples
##' ## Non-linear parameter constraints 1
##' m <- lvm(y ~ f(x1,gamma)+f(x2,beta))
##' covariance(m) <- y ~ f(v)
##' d <- sim(m,100)
##' m1 <- m; constrain(m1,beta ~ v) <- function(x) x^2
##' ## Define slope of x2 to be the square of the residual variance of y
##' ## Estimate both restricted and unrestricted model
##' e <- estimate(m,d,control=list(method="NR"))
##' e1 <- estimate(m1,d)
##' p1 <- coef(e1)
##' p1 <- c(p1[1:2],p1[3]^2,p1[3])
##' ## Likelihood of unrestricted model evaluated in MLE of restricted model
##' logLik(e,p1)
##' ## Likelihood of restricted model (MLE)
##' logLik(e1)
##'
##' ## Non-linear regression
##'
##' ## Simulate data
##' m <- lvm(c(y1,y2)~f(x,0)+f(eta,1))
##' latent(m) <- ~eta
##' covariance(m,~y1+y2) <- "v"
##' intercept(m,~y1+y2) <- "mu"
##' covariance(m,~eta) <- "zeta"
##' intercept(m,~eta) <- 0
##' set.seed(1)
##' d <- sim(m,100,p=c(v=0.01,zeta=0.01))[,manifest(m)]
##' d <- transform(d,
##'                y1=y1+2*pnorm(2*x),
##'                y2=y2+2*pnorm(2*x))
##'
##' ## Specify model and estimate parameters
##' constrain(m, mu ~ x + alpha + nu + gamma) <- function(x) x[4]*pnorm(x[3]+x[1]*x[2])
##' \dontrun{ ## Reduce Ex.Timings
##' e <- estimate(m,d,control=list(trace=1,constrain=TRUE))
##' constraints(e,data=d)
##' ## Plot model-fit
##' plot(y1~x,d,pch=16); points(y2~x,d,pch=16,col="gray")
##' x0 <- seq(-4,4,length.out=100)
##' lines(x0,coef(e)["nu"] + coef(e)["gamma"]*pnorm(coef(e)["alpha"]*x0))
##' }
##'
##' ## Multigroup model
##' m1 <- lvm(y ~ f(x,beta)+f(z,beta2))
##' m2 <- lvm(y ~ f(x,psi) + z)
##' # simulate data from the two models
##' d1 <- sim(m1,500)
##' d2 <- sim(m2,500)
##' # Add 'non'-linear parameter constraint
##' constrain(m2,psi ~ beta2) <- function(x) x
##' ## Add parameter beta2 to model 2, now beta2 exists in both models
##' parameter(m2) <- ~ beta2
##' ee <- estimate(list(m1,m2),list(d1,d2),control=list(method="NR"))
##' summary(ee)
##'
##' m3 <- lvm(y ~ f(x,beta)+f(z,beta2))
##' m4 <- lvm(y ~ f(x,beta2) + z)
##' e2 <- estimate(list(m3,m4),list(d1,d2),control=list(method="NR"))
##' e2
##' @export
##' @usage
##'
##' \method{constrain}{default}(x,par,args,endogenous=TRUE,...) <- value
##'
##' \method{constrain}{multigroup}(x,par,k=1,...) <- value
##'
##' constraints(object,data=model.frame(object),vcov=object$vcov,level=0.95,
##'                         p=pars.default(object),k,idx,...)
##'
##' @param x \code{lvm}-object
##' @param par Name of new parameter. Alternatively a formula with lhs
##' specifying the new parameter and the rhs defining the names of the
##' parameters or variable names defining the new parameter (overruling the
##' \code{args} argument).
##' @param args Vector of variables names or parameter names that are used in
##' defining \code{par}
##' @param endogenous TRUE if variable is endogenous (sink node)
##' @param k For multigroup models this argument specifies which group to
##' add/extract the constraint
##' @param value Real function taking args as a vector argument
##' @param object \code{lvm}-object
##' @param data Data-row from which possible non-linear constraints should be
##' calculated
##' @param vcov Variance matrix of parameter estimates
##' @param level Level of confidence limits
##' @param p Parameter vector
##' @param idx Index indicating which constraints to extract
##' @param \dots Additional arguments to be passed to the low level functions
"constrain<-" <- function(x,...,value) UseMethod("constrain<-")

##' @export
"constrain" <- function(x,...) UseMethod("constrain")

##' @export
constrain.default <- function(x, par, fun, idx, level=0.95, vcov, estimate=FALSE, ...) {
    if (!missing(par)) {
        constrain(x, par, ...) <- fun
        return(x)
    }
    if (estimate) {
        return(constraints(x,...))
    }
    if (missing(fun)) {
        if (inherits(Model(x),"multigroup")) {
            res <- list()
            for (m in Model(x)$lvm) {
                if (length(constrain(m))>0)
                    res <- c(res, constrain(m))
            }
            return(res)
        }
        return(Model(x)$constrain)
    }
    if (is.numeric(x)) {
        b <- x
    } else {
        b <- pars(x)
    }
    if (missing(vcov)) {
        S <- stats::vcov(x)
    } else {
        S <- vcov
    }
    if (!missing(idx)) {
        b <- b[idx]; S <- S[idx,idx,drop=FALSE]
    }
    fb <- fun(b)
    pl <- 1-(1-level)/2
    D <- rbind(numDeriv::grad(fun,b))
    se <- (D%*%S%*%t(D))^0.5
    res <- c(fb,se,fb+c(-1,1)*qnorm(pl)*c(se))
    pstr <- paste0(format(c(round(1000-1000*pl),round(pl*1000))/10),"%")
    names(res) <- c("Estimate","Std.Err",pstr)
    res
}

##' @export
"constrain<-.multigroupfit" <- function(x,par,k=1,...,value) {
        constrain(Model(x)$lvm[[k]],par=par,...) <- value
        return(x)
    }

##' @export
"constrain<-.multigroup" <- function(x,par,k=1,...,value) {
        constrain(Model(x)$lvm[[k]],par=par,...) <- value
        return(x)
    }

##' @export
"constrain<-.default" <- function(x,par,args,endogenous=TRUE,...,value) {
    if (inherits(par,"formula")) {
        lhs <- getoutcome(par)
        xf <- attributes(terms(par))$term.labels
        par <- lhs
        if (length(par)==0) {
          par <- xf
          xf <- NULL
        }
        if (par%in%vars(x)) {
            if (is.na(x$mean[[par]])) {
                intercept(x,par) <- par
            } else {
                par <- x$mean[[par]]
            }
        }
        args <- xf
    }
    if (is.null(value) || suppressWarnings(is.na(value))) {
        if (!is.null(par)) {
            Model(x)$constrain[[par]] <- NULL
            Model(x)$constrainY[[par]] <- NULL
        } else {
            Model(x)$constrain[[args]] <- NULL
        }
        return(x)
    }
    for (i in args) {
        if (!(i%in%c(parlabels(Model(x)), vars(Model(x)),
                     names(constrain(x))))) {
            if (lava.options()$messages>1)
                message("\tAdding parameter '", i, "'\n",sep="")
            parameter(x, messages=0) <- i
        }
    }
    if (par%in%vars(x) && endogenous) {
        if (!"..."%in%names(formals(value)) && !is.primitive(value)) {
            formals(value) <- c(formals(value), alist(...=))
        }
        Model(x)$constrainY[[par]] <- list(fun=value, args=args)
    } else {
        ## Wrap around do.call, since functions are not really
        ## parsed as call-by-value in R, and hence setting
        ## attributes to e.g. value=cos, will be overwritten
        ## if value=cos is used again later with new args.
        Model(x)$constrain[[par]] <- function(x) do.call(value, list(x))
        attributes(Model(x)$constrain[[par]])$args <- args
        index(Model(x)) <- reindex(Model(x))
    }
    return(x)
}

##' @export
constraints <- function(object,data=model.frame(object),vcov=object$vcov,level=0.95,
                 p=pars.default(object),k,idx,...) {
    if (class(object)[1]=="multigroupfit") {
        if (!missing(k)) {
            if (class(data)[1]=="list") data <- data[[k]]
            parpos <- modelPar(object, seq_len(length(p)))$p[[k]]
            if (nrow(data)>1 & !missing(idx)) {
                res <- t(apply(data, 1, function(x) constraints(Model(object)$lvm[[k]],data=x,p=p[parpos],vcov=vcov[parpos,parpos],level=level)[idx,]))
                return(res)
            }
            return(constraints(Model(object)$lvm[[k]],data=data,p=p[parpos],vcov=vcov[parpos,parpos],level=level))
        }
        return(attributes(CoefMat.multigroupfit(object,data=data,vcov=vcov,...))$nlincon)
    }

    if (NROW(data)>1 & !missing(idx)) {
        res <- t(apply(data,1,function(x) constraints(object,data=x,p=p,vcov=vcov,level=level)[idx,],...))
        return(res)
    }

    if (length(index(object)$constrain.par)<1) return(NULL)
    parpos <- Model(object)$parpos
    if (is.null(parpos)) {
        parpos <- with(index(object),matrices2(Model(object),seq_len(npar+npar.mean+npar.ex)))
        parpos$A[index(object)$M0==0] <- 0
        parpos$P[index(object)$P0==0] <- 0
        parpos$v[index(object)$v1==0] <- 0
        parpos$e[index(object)$e1==0] <- 0
    }
    myidx <- unlist(lapply(parpos$parval, function(x) {
        if (!is.null(attributes(x)$reg.idx)) {
            return(parpos$A[attributes(x)$reg.idx[1]])
        } else if (!is.null(attributes(x)$cov.idx)) {
            return(parpos$P[attributes(x)$cov.idx[1]])
        } else if (!is.null(attributes(x)$m.idx)) {
            return(parpos$v[attributes(x)$m.idx[1]])
        } else if (!is.null(attributes(x)$e.idx))
            return(parpos$e[attributes(x)$e.idx[1]])
        else NA
    }))
    names(myidx) <- names(parpos$parval)
    mynames <- c()
    N <- length(index(object)$constrain.par)
    if (N>0)
        res <- c()
    count <- 0
    mydata <- rbind(numeric(length(manifest(object))))
    colnames(mydata) <- manifest(object)
    data <- rbind(data)
    iname <- intersect(colnames(mydata),colnames(data))
    mydata[1,iname] <- unlist(data[1,iname])
    for (pp in index(object)$constrain.par) {
        count <- count+1
        myc <- constrain(Model(object))[[pp]]
        mycoef <- numeric(6)
        val.idx <- myidx[attributes(myc)$args]
        val.idx0 <- na.omit(val.idx)
        M <- modelVar(Model(object),p=p,data=as.data.frame(mydata))
        vals <- with(M,c(parval,constrainpar))[attributes(myc)$args]
        fval <- try(myc(unlist(vals)),silent=TRUE)
        fmat <- inherits(fval,"try-error")
        if (fmat) {
            fval <- myc(rbind(unlist(vals)))
        }
        mycoef[1] <- fval
        myc0 <- function(theta) {
            theta0 <- unlist(vals);
            theta0[!is.na(val.idx)] <- theta
            if (fmat) {
                res <- myc(rbind(theta0))
            } else {
                res <- myc(theta0)
            }
            return(res)
        }
        vals0 <- unlist(vals)[!is.na(val.idx)]
        if (length(vals0)==0)
            mycoef[2] <- NA
        else {
            if (!is.null(attributes(fval)$grad)) {
                if (fmat) {
                    Gr <- cbind(attributes(fval)$grad(rbind(unlist(vals0))))
                } else {
                    Gr <- cbind(attributes(fval)$grad(unlist(vals0)))
                }
            } else {
                if (fmat) {
                    Gr <- cbind(as.numeric(numDeriv::jacobian(myc0, unlist(vals0))))
                } else {
                    Gr <- cbind(as.numeric(numDeriv::jacobian(myc0, rbind(unlist(vals0)))))
                }
            }
            V <- vcov[val.idx0,val.idx0]
            mycoef[2] <- (t(Gr)%*%V%*%Gr)^0.5
        }
        ## if (second) {
        ##   if (!is.null(attributes(fval)$hessian)) {
        ##     H <- attributes(fval)$hessian(unlist(vals))
        ##   } else {
        ##     H <- hessian(myc, unlist(vals))
        ##   }
        ##   HV <- H%*%vcov[val.idx,val.idx]
        ##   mycoef[1] <- mycoef[1] + 0.5*sum(diag(HV))
        ##   mycoef[2] <- mycoef[2] + 0.5*sum(diag(HV%*%HV))
        ## }
        mycoef[3] <- mycoef[1]/mycoef[2]
        mycoef[4] <- 2*(pnorm(abs(mycoef[3]),lower.tail=FALSE))
        mycoef[5:6] <- mycoef[1] + c(1,-1)*qnorm((1-level)/2)*mycoef[2]
        res <- rbind(res,mycoef)
        mynames <- c(mynames,pp)
        if (!is.null(attributes(fval)$inv)){
            res2 <- attributes(fval)$inv(mycoef[c(1,5,6)])
            res <- rbind(res, c(res2[1],NA,NA,NA,res2[2],res2[3]))
            mynames <- c(mynames,paste0("inv(",pp,")"))
        }
    }
    rownames(res) <- mynames
    colnames(res) <- c("Estimate","Std. Error", "Z value", "Pr(>|z|)", "2.5%", "97.5%")
    return(res)
}

##' Define constant risk difference or relative risk association for binary exposure
##'
##' Set up model as defined in Richardson, Robins and Wang (2017).
##' @param x model
##' @param response response variable (character or formula)
##' @param exposure exposure variable (character or formula)
##' @param target.model variable defining the linear predictor for the target model
##' @param nuisance.model variable defining the linear predictor for the nuisance model
##' @param exposure.model model for exposure (default binomial logit link)
##' @param ... additional arguments to lower level functions
##' @aliases binomial.rd binomial.rr
##' @export
##' @examples
##' ## ---------------------------------------------------------------
##' ## binomial.rd: constant risk-difference model
##' ##   P(Y=1|Z=1) - P(Y=1|Z=0) = tanh(lp)
##' ## ---------------------------------------------------------------
##' m <- lvm()
##' regression(m) <- z ~ x
##' regression(m) <- lp ~ x
##' regression(m) <- op ~ x
##' intercept(m, ~lp) <- 0.4   ## constant linear predictor for RD
##' intercept(m, ~op) <- 0     ## odds product = exp(0) = 1
##' distribution(m, ~lp) <- dist_gaussian(sd = 0)
##' distribution(m, ~op) <- dist_gaussian(sd = 0)
##' m <- binomial.rd(m, response = "y", exposure = "z",
##'                  target.model = "lp", nuisance.model = "op")
##' set.seed(1)
##' d <- sim(m, n = 2000)
##' ## Empirical risk difference should be close to tanh(0.4)
##' mean(d$y[d$z == 1]) - mean(d$y[d$z == 0])
##' tanh(0.4)
##'
##' ## Formula interface: response ~ exposure | target | nuisance
##' m2 <- lvm()
##' regression(m2) <- z ~ x
##' regression(m2) <- lp ~ x
##' regression(m2) <- op ~ x
##' m2 <- binomial.rd(m2, y ~ z | lp | op)
##'
##' ## ---------------------------------------------------------------
##' ## binomial.rr: constant relative-risk model
##' ##   log(P(Y=1|Z=1) / P(Y=1|Z=0)) = lp
##' ## ---------------------------------------------------------------
##' m <- lvm()
##' regression(m) <- z ~ x
##' regression(m) <- lp ~ x
##' regression(m) <- op ~ x
##' intercept(m, ~lp) <- log(1.5)   ## constant log relative-risk
##' intercept(m, ~op) <- 0          ## odds product = 1
##' distribution(m, ~lp) <- dist_gaussian(sd = 0)
##' distribution(m, ~op) <- dist_gaussian(sd = 0)
##' m <- binomial.rr(m, response = "y", exposure = "z",
##'                  target.model = "lp", nuisance.model = "op")
##' set.seed(1)
##' d <- sim(m, n = 2000)
##' ## Empirical log-RR should be close to log(1.5)
##' log(mean(d$y[d$z == 1]) / mean(d$y[d$z == 0]))
##' log(1.5)
binomial.rd <- function(x,response,exposure,
                 target.model,nuisance.model,
                 exposure.model=dist_bernoulli(),...) {
    binomial.rrw(x,response,exposure,target.model,nuisance.model,exposure.model,type="rd",...)
}

##' @export
binomial.rr <- function(x,response,exposure,
                 target.model,nuisance.model,
                 exposure.model=dist_bernoulli(),...) {
    binomial.rrw(x,response,exposure,target.model,nuisance.model,exposure.model,type="rr",...)
}


binomial.rrw <- function(x, response, exposure,
                 target.model, nuisance.model,
                 exposure.model=dist_bernoulli(), type="rd", ...) {
    if (inherits(response,"formula")) {
        vars <- all.vars(response)
        if (length(vars)==1L) {
            response <- vars
        } else {
            yf <- getoutcome(response, sep="|")
            exposure  <- attr(yf,"x")[[1]]
            if (length(attr(yf,"x"))>1)
                target.model  <- attr(yf,"x")[[2]]
            if (length(attr(yf,"x"))>2)
                nuisance.model  <- attr(yf,"x")[[3]]
            response <- yf[1]
        }
    }
    if (inherits(exposure,"formula")) exposure <- all.vars(exposure)
    if (inherits(target.model,"formula")) target.model <- all.vars(target.model)
    if (inherits(nuisance.model,"formula")) nuisance.model <- all.vars(nuisance.model)

    if (type=="rd") {
        val <- list(list(input=c(exposure,target.model,nuisance.model),
                         fun=simulate_binomial_rd,
                         type="Binomial regression (exposure | risk-difference | odds-product)"))
    } else {
        val <- list(list(input=c(exposure,target.model,nuisance.model),
                         fun=simulate_binomial_rr,
                         type="Binomial regression (exposure | relative-risk | odds-product)"))
    }
    if (is.null(distribution(x)[[exposure]]))
        distribution(x, exposure) <- dist_bernoulli(link="logit")
    covariance(x,c(target.model,nuisance.model)) <- 0
    distribution(x,exposure) <- exposure.model
    names(val) <- response
    x$attributes$multiple.inputs <- val
    return(x)
}


simulate_binomial_rd <- function(x,data,inputs,...) {
    exposure <- data[,inputs[1]]
    lp1 <- data[,inputs[2]]
    lp2 <- data[,inputs[3]]
    rd <- tanh(lp1)
    op <- exp(lp2)
    pp <- RD_OP(rd,op)
    pp <- pp[,1]*(1-exposure) + pp[,2]*exposure
    if (anyNA(pp) || any(pp < 0) || any(pp > 1))
        stop("simulate_binomial_rd: implied probabilities outside [0,1]; check (rd, op) inputs")
    return(rbinom(NROW(data), 1, pp))
}

simulate_binomial_rr <- function(x,data,inputs,...) {
    exposure <- data[,inputs[1]]
    lp1 <- data[,inputs[2]]
    lp2 <- data[,inputs[3]]
    rr <- exp(lp1)
    op <- exp(lp2)
    pp <- RR_OP(rr,op)
    pp <- pp[,1]*(1-exposure) + pp[,2]*exposure
    if (anyNA(pp) || any(pp < 0) || any(pp > 1))
        stop("simulate_binomial_rr: implied probabilities outside [0,1]; check (rr, op) inputs")
    return(rbinom(NROW(data), 1, pp))
}

Identical <- function(x,y=1,tolerance = .Machine$double.eps^0.5) {
    Mod(x-y)<tolerance
}

RD_OP <- function(rd,op) {
    a <- op-1
    b <- -op*(2-rd)-rd
    p0 <- (-b - sqrt(b^2 - 4*op*(1-rd)*a))/(2*a)
    op1 <- which(sapply(op,function(x) Identical(x,1)))
    if (length(op1)>0)
        p0[op1] = 0.5 * (1 - rd[op1])
    p1 <- p0 + rd
    cbind(p0,p1)
}


RR_OP <- function(rr,op) {
    b <- op*(1+rr)
    a <- rr*(1-op)
    p0 <- (-b + sqrt(b^2 + 4*a*op))/(2*a)
    op1 <- which(sapply(op,function(x) Identical(x,1)))
    if (length(op1)>0)
        p0[op1] = 1/(1+rr)
    p1 <- p0*rr
    cbind(p0,p1)
}

##' Convert PDF file to print quality png (default 300 dpi)
##'
##' Access to ghostscript program 'gs' is needed
##' @title Convert pdf to raster format
##' @param files Vector of (pdf-)filenames to process
##' @param dpi DPI
##' @param resolution Resolution of raster image file
##' @param gs Optional ghostscript command
##' @param gsopt Optional ghostscript arguments
##' @param resize Optional resize arguments (mogrify)
##' @param format Raster format (e.g. png, jpg, tif, ...)
##' @param \dots Additional arguments
##' @seealso \code{dev.copy2pdf}, \code{printdev}
##' @export
##' @author Klaus K. Holst
##' @keywords iplot
pdfconvert <- function(files, dpi=300, resolution=1024, gs, gsopt, resize, format="png", ...) {
    if (missing(gsopt))
        gsopt <- "-dSAFTER -dBATCH -dNOPAUSE -sDEVICE=png16m -dGraphicsAlphaBits=4 -dTextAlphaBits=4"
    if (missing(gs)) {
        gs <- names(which(Sys.which(c("gs", "gswin32c", "gswin64c")) != ""))
    }
    cmd1 <- paste0(gs," -r",dpi," -dBackgroundColor='16#ffffff'")
    if (missing(resize)) {
        resize <- paste0("mogrify -resize ", resolution)
    }
    for (f in files) {
        f0 <- strsplit(f,".pdf")[1]
        f.out <- paste(f0,format,sep=".")
        f.pdf <- paste(f0,"pdf",sep=".")
        mycmd1 <- paste0(cmd1, " ", gsopt, " -sOutputFile=", f.out, " > /dev/null ", f.pdf)
        mycmd2 <- paste0(resize, " ", f.out)
        cat(f.pdf)
        system(mycmd1)
        cat(" -> ")
        system(mycmd2)
        cat(f.out, "\n")
  }
}

##' @export
`estimate.multigroup` <- function(x, control=list(),
                           estimator=NULL,
                           weights, weightsname,
                           data2,
                           id=NULL,
                           messages=lava.options()$messages,
                           quick=FALSE,
                           param,
                           cluster,
                           ...) {
    cl <- match.call()
    Optim <- list(
        iter.max=lava.options()$iter.max,
        trace=ifelse(lava.options()$debug,3,0),
        gamma=lava.options()$gamma,
        ngamma=lava.options()$ngamma,
        backtrace=TRUE,
        gamma2=1,
        lambda=0.05,
        abs.tol=1e-9,
        epsilon=1e-10,
        delta=1e-10,
        S.tol=1e-6,
        stabil=FALSE,
        start=NULL,
        constrain=lava.options()$constrain,
        method=NULL,
        starterfun=startvalues0,
        information="E",
        meanstructure=TRUE,
        sparse=FALSE,
        lbound=1e-9,
        reindex=FALSE,
        tol=lava.options()$tol)


    if (!missing(param)) {
        oldparam <- lava.options()$param
        lava.options(param=param)
        on.exit(lava.options(param=oldparam))
    }
    if (!missing(cluster)) id <- cluster

    defopt <- lava.options()[]
    defopt <- defopt[intersect(names(defopt),names(Optim))]
    Optim[names(defopt)] <- defopt

    if (length(control)>0) {
        Optim[names(control)] <- control
    }


    Debug("Start values...")
    if (!is.null(Optim$start) & length(Optim$start)==(x$npar+x$npar.mean)) {
        mystart <- Optim$start
    } else {
        if (messages>1) cat("Obtaining starting value...")
        if (is.null(control$starterfun) && lava.options()$param!="relative")
            Optim$starterfun <- startvalues0
        mystart <- with(Optim, starter.multigroup(x,meanstructure=meanstructure,starterfun=starterfun,messages=messages,fix=FALSE))
        if (!is.null(Optim$start)) {
            pname <- names(Optim$start)
            ppos <- parpos.multigroup(x,p=pname,mean=TRUE)
            if (any(!is.na(ppos)))
                mystart[ppos] <- Optim$start[na.omit(match(attributes(ppos)$name,pname))]
        }
        if (messages>1) cat("\n")
    }
    Debug(mystart)
    Debug("Constraints...")
    ## Setup optimization constraints
    lower <- rep(-Inf, x$npar);
    for (i in seq_len(x$ngroup)) {
        vpos <- sapply(x$parlist[[i]][variances(x$lvm[[i]],mean=FALSE)], function(y) char2num(substr(y,2,nchar(y))))
        if (length(vpos)>0)
            lower[vpos] <- Optim$lbound
    }
    if (Optim$meanstructure)
        lower <- c(rep(-Inf,x$npar.mean), lower)
    if (any(Optim$constrain)) {
        if (length(Optim$constrain)!=length(lower))
            constrained <- is.finite(lower)
        else
            constrained <- Optim$constrain
        constrained <- which(constrained)
        lower[] <- -Inf
        Optim$constrain <- TRUE
        mystart[constrained] <- log(mystart[constrained])
    }

    if (!missing(weights)) {
        if (is.character(weights)) {
            stweights <- weights
            weights <- list()
            for (i in seq_along(x$data)) {
                newweights <- as.matrix(x$data[[i]][,stweights])
                colnames(newweights) <- index(x$lvm[[i]])$endogenous[seq_len(ncol(newweights))]
                weights <- c(weights, list(newweights))
            }
        }
    } else {
        weights <- NULL
    }
    if (!missing(data2)) {
        if (is.character(data2)) {
            stdata2 <- data2
            data2 <- list()
            for (i in seq_along(x$data)) {
                newdata <- as.matrix(x$data[[i]][,stdata2,drop=FALSE])
                dropcol <- apply(newdata,2,function(x) any(is.na(x)))
                newdata <- newdata[,!dropcol,drop=FALSE]
                colnames(newdata) <- index(x$lvm[[i]])$endogenous[seq_len(ncol(newdata))]
                data2 <- c(data2, list(newdata))
            }
        }
    } else {
        data2 <- NULL
    }

    ## Run hooks (additional lava plugins)
    myhooks <- gethook()
    newweights <- list()
    newdata2 <- list()
    newoptim <- newestimator <- NULL
    for (f in myhooks) {
        for ( i in seq_len(x$ngroup)) {
            res <- do.call(f, list(x=x$lvm[[i]],data=x$data[[i]],weights=weights[[i]],data2=data2[[i]],estimator=estimator,optim=Optim))
            if (!is.null(res$x)) x$lvm[[i]] <- res$x
            if (!is.null(res$data)) x$data[[i]] <- res$data
            if (!is.null(res$weights)) newweights <- c(newweights,list(res$weights))
            if (!is.null(res$data2)) newdata2 <- c(newdata2,list(res$data2))
            if (!is.null(res$optim)) newoptim <- res$optim
            if (!is.null(res$estimator)) newestimator <- res$estimator
        }
        if (!is.null(newestimator)) estimator <- newestimator
        if (!is.null(newoptim)) Optim <- newoptim
        if (!is.null(res$weights))
            if (!any(unlist(lapply(newweights,is.null)))) {
                weights <- newweights
            }
        if (!is.null(res$data2))
            if (!any(unlist(lapply(newdata2,is.null)))) {
                data2 <- newdata2
            }
    }
    if (is.null(estimator)) {
        if (!missing(weights) && !is.null(weights)) {
            estimator <- "normal"
        } else estimator <- "gaussian"
    }

    checkestimator <- function(x,...) {
        ffname <- paste0(x,c("_objective","_gradient"),".lvm")
        exists(ffname[1])||exists(ffname[2])
    }
    if (!checkestimator(estimator)) { ## Try down/up-case version
        estimator <- tolower(estimator)
        if (!checkestimator(estimator)) {
            estimator <- toupper(estimator)
        }
    }

    Method <-  paste0(estimator, "_method", ".lvm")
    if (!exists(Method))
        Method <- "nlminb1"
    else
        Method <- get(Method)
    if (is.null(Optim$method)) {
        Optim$method <- Method
    }

    ## Check for random slopes
    Xfix <- FALSE
    Xconstrain <- FALSE
    xfix <- list()
    for (i in seq_len(x$ngroup)) {
        x0 <- x$lvm[[i]]
        data0 <- x$data[[i]]
        xfix0 <- colnames(data0)[(colnames(data0)%in%parlabels(x0,exo=TRUE))]
        xconstrain0 <- intersect(unlist(lapply(constrain(x0),function(z) attributes(z)$args)),manifest(x0))
        xfix <- c(xfix, list(xfix0))
        if (length(xfix0)>0) Xfix<-TRUE ## Yes, random slopes
        if (length(xconstrain0)>0) Xconstrain <- TRUE ## Yes, nonlinear regression
    }

    ## Non-linear parameter constraints involving observed variables? (e.g. nonlinear regression)
    XconstrStdOpt <- TRUE
    xconstrainM <- TRUE
    xconstrain <- c()
    if (Xconstrain)
        for (i in seq_len(x$ngroup)) {
            x0 <- x$lvm[[i]]
            data0 <- x$data[[i]]
            constr0 <- lapply(constrain(x0), function(z)(attributes(z)$args))
            xconstrain0 <- intersect(unlist(constr0), manifest(x0))
            xconstrain <- c(xconstrain, list(xconstrain0))
            if (length(xconstrain0)>0) {
                constrainM0 <- names(constr0)%in%unlist(x0$mean)
                for (i in seq_len(length(constr0))) {
                    if (!constrainM0[i]) {
                        if (xconstrain0%in%constr0[[i]]) {
                            xconstrainM <- FALSE
                        }
                    }
                }
                if (xconstrainM & ((is.null(control$method) || Optim$method=="nlminb0") & (lava.options()$test & estimator=="gaussian")) ) {
                    XconstrStdOpt <- FALSE
                    Optim$method <- "nlminb0"
                    if (is.null(control$constrain)) control$constrain <- TRUE
                }
            }
        }

    ## Define objective function and first and second derivatives
    ObjectiveFun  <- paste0(estimator, "_objective", ".lvm")
    GradFun  <- paste0(estimator, "_gradient", ".lvm")
    if (!exists(ObjectiveFun) & !exists(GradFun)) stop("Unknown estimator.")

    InformationFun <- paste0(estimator, "_hessian", ".lvm")

    parord <- modelPar(x,seq_len(with(x,npar+npar.mean)))$p
    mymodel <- x

    parkeep <- c()
    myclass <- c("multigroupfit","lvmfit")
    myfix <- list()

    if (Xfix |  (Xconstrain & XconstrStdOpt | !lava.options()$test)) { ## Model with random slopes:
        if (Xfix) {
            myclass <- c(myclass,"lvmfit.randomslope")
            for (k in seq_len(x$ngroup)) {
                x1 <- x0 <- x$lvm[[k]]
                data0 <- x$data[[k]]

                nrow <- length(vars(x0))
                xpos <- lapply(xfix[[k]],function(y) which(regfix(x0)$labels==y))
                colpos <- lapply(xpos, function(y) ceiling(y/nrow))
                rowpos <- lapply(xpos, function(y) (y-1)%%nrow+1)
                myfix0 <- list(var=xfix[[k]], col=colpos, row=rowpos)
                myfix <- c(myfix, list(myfix0))

                for (i in seq_along(myfix0$var))
                    for (j in seq_along(myfix0$col[[i]]))
                        regfix(x0,
                               from=vars(x0)[myfix0$row[[i]][j]],to=vars(x0)[myfix0$col[[i]][j]]) <-
                            colMeans(data0[,myfix0$var[[i]],drop=FALSE],na.rm=TRUE)
                index(x0) <- reindex(x0,zeroones=TRUE,deriv=TRUE)
                x$lvm[[k]] <- x0
                parkeep <- c(parkeep, parord[[k]][coef(x1,mean=TRUE,fix=FALSE)%in%coef(x0,mean=TRUE,fix=FALSE)])
            }
            parkeep <- sort(unique(parkeep))
            ## Alter start-values:

            if (length(mystart)!=length(parkeep))
                mystart <- mystart[parkeep]
            lower <- lower[parkeep]
            x <- multigroup(x$lvm,x$data,fix=FALSE,exo.fix=FALSE)
        }

        parord <- modelPar(x,seq_along(mystart))$p
        mydata <- list()
        for (i in seq_len(x$ngroup)) {
            mydata <- c(mydata, list(as.matrix(x$data[[i]][,manifest(x$lvm[[i]])])))
        }

        myObj <- function(theta) {
            if (Optim$constrain)
                theta[constrained] <- exp(theta[constrained])
            pp <- modelPar(x,theta)$p
            res <- 0
            for (k in seq_len(x$ngroup)) {
                x0 <- x$lvm[[k]]
                data0 <- x$data[[k]]
                if (Xfix) {
                    xfix0 <- xfix[[k]]
                    myfix0 <- myfix[[k]]
                }
                p0 <- pp[[k]]
                myfun <- function(ii) {
                    if (Xfix)
                        for (i in seq_along(myfix0$var)) {
                            x0$fix[cbind(myfix0$row[[i]],myfix0$col[[i]])] <-
                                index(x0)$A[cbind(myfix0$row[[i]],myfix0$col[[i]])] <-
                                            data0[ii,xfix0[i]]
                        }
                    if (is.list(data2[[k]][ii,])) {
                        res <- do.call(ObjectiveFun, list(x=x0, p=p0,
                                                          data=data0[ii,manifest(x0),drop=FALSE],
                                                          n=1, S=NULL, weights=weights[[k]][ii,],
                                                          data2=data2[[k]]))

                    } else {
                        res <- do.call(ObjectiveFun, list(x=x0, p=p0,
                                                          data=data0[ii,manifest(x0),drop=FALSE],
                                                          n=1, S=NULL, weights=weights[[k]][ii,],
                                                          data2=data2[[k]][ii,]))
                    }
                    return(res)
                }
                res <- res + sum(sapply(seq_len(nrow(mydata[[k]])),myfun))
            }
            res
        }

        myGrad <- function(theta) {
            if (Optim$constrain) {
                theta[constrained] <- exp(theta[constrained])
            }
            pp <- modelPar(x,theta)$p
            D0 <- res <- rbind(numeric(length(mystart)))
            for (k in seq_len(x$ngroup)) {
                if (Xfix) {
                    myfix0 <- myfix[[k]]
                }
                x0 <- x$lvm[[k]]
                myfun <- function(ii) {
                    if (Xfix)
                        for (i in seq_along(myfix0$var)) {
                            x0$fix[cbind(myfix0$row[[i]],myfix0$col[[i]])] <-
                                index(x0)$A[cbind(myfix0$row[[i]],myfix0$col[[i]])] <-
                                            x$data[[k]][ii,xfix[[k]][i]]
                        }
                    if (is.list(data2[[k]][ii,])) {

                    } else {
                        val <- do.call(GradFun, list(x=x0, p=pp[[k]],
                                                     data=mydata[[k]][ii,,drop=FALSE], n=1,
                                                     S=NULL,
                                                     weights=weights[[k]][ii,],
                                                     data2=data2[[k]][ii,]))
                    }
                    return(val)
                }
                D <- D0; D[parord[[k]]] <- rowSums(sapply(seq_len(nrow(mydata[[k]])),myfun))
                res <- res+D
            }
            if (Optim$constrain) {
                res[constrained] <- res[constrained]*theta[constrained]
            }
            return(as.vector(res))
        }

        myInformation <- function(theta) {
            theta0 <- theta
            if (Optim$constrain) {
                theta[constrained] <- exp(theta[constrained])
            }
            pp <- modelPar(x,theta)$p
            I0 <- res <- matrix(0,length(theta),length(theta))
            grad <- grad0 <- numeric(length(theta))
            for (k in seq_len(x$ngroup)) {
                x0 <- x$lvm[[k]]
                if (Xfix) {
                    myfix0 <- myfix[[k]]
                }
                myfun <- function(ii) {
                    if (Xfix)
                        for (i in seq_along(myfix0$var)) {
                            x0$fix[cbind(myfix0$row[[i]],myfix0$col[[i]])] <- index(x0)$A[cbind(myfix0$row[[i]],myfix0$col[[i]])] <-
                                                                                          x$data[[k]][ii,xfix[[k]][i]]
                        }
                    I <- I0
                    J <- do.call(InformationFun,
                                 list(x=x0, p=pp[[k]],
                                      data=mydata[[k]][ii,], n=1,
                                      S=NULL,
                                      weights=weights[[k]][ii,],
                                      data2=data2[[k]][ii,],
                                      type=Optim$information
                                      )
                                 )
                    D <- grad0
                    if (!is.null(attributes(J)$grad)) {
                        D[ parord[[k]] ] <- attributes(J)$grad
                        attributes(I)$grad <- D
                    }
                    I[ parord[[k]], parord[[k]] ] <- J
                    return(I)
                }
                L <- lapply(seq_len(nrow(x$data[[k]])),function(x) myfun(x))
                if (!is.null(attributes(L[[1]])$grad))
                    grad <- grad + rowSums(matrix((unlist(lapply(L,function(x) attributes(x)$grad))),ncol=length(L)))
                res <- res + apply(array(unlist(L),dim=c(length(theta),length(theta),nrow(x$data[[k]]))),c(1,2),sum)
            }
            if (!is.null(attributes(L[[1]])$grad))
                attributes(res)$grad <- grad
            return(res)
        }
    } else { ## Model without random slopes:
        ## Non-linear parameter constraints involving observed variables? (e.g. nonlinear regression)
        yconstrain <- c()
        iconstrain <- c()
        xconstrain <- c()
        for (j in seq_len(x$ngroup)) {
            x0 <- x$lvm[[j]]
            data0 <- x$data[[j]]
            xconstrain0 <- c()
            for (i in seq_len(length(constrain(x0)))) {
                z <- constrain(x0)[[i]]
                xx <- intersect(attributes(z)$args,manifest(x0))
                if (length(xx)>0) {
                    warg <- setdiff(attributes(z)$args,xx)
                    wargidx <- which(attributes(z)$args%in%warg)
                    exoidx <- which(attributes(z)$args%in%xx)
                    parname <- names(constrain(x0))[i]
                    y <- names(which(unlist(lapply(intercept(x0),function(x) x==parname))))
                    el <- list(i,y,parname,xx,exoidx,warg,wargidx,z)
                    names(el) <- c("idx","endo","parname","exo","exoidx","warg","wargidx","func")
                    xconstrain0 <- c(xconstrain0,list(el))
                }
            }
            yconstrain0 <- unlist(lapply(xconstrain0,function(x) x$endo))
            iconstrain0 <- unlist(lapply(xconstrain0,function(x) x$idx))
            xconstrain <- c(xconstrain, list(xconstrain0))
            yconstrain <- c(yconstrain, list(yconstrain0))
            iconstrain <- c(iconstrain, list(iconstrain0))
        }

        MkOffset <- function(pp,x,data,xconstrain,grad=FALSE) {
            if (length(xconstrain)>0) {
                Mu <- matrix(0,nrow(data),length(vars(x))); colnames(Mu) <- vars(x)
                M <- modelVar(x,p=pp,data=data)
                M$parval <- c(M$parval,  x$mean[unlist(lapply(x$mean,is.numeric))])
                for (i in seq_len(length(xconstrain))) {
                    pp <- unlist(M$parval[xconstrain[[i]]$warg]);
                    myidx <- with(xconstrain[[i]],order(c(wargidx,exoidx)))
                    mu <- with(xconstrain[[i]],
                              apply(data[,exo,drop=FALSE],1,
                                    function(x) func(
                                             unlist(c(pp,x))[myidx])))
                    Mu[,xconstrain[[i]]$endo] <- mu
                }
                offsets <- Mu%*%t(M$IAi)[,endogenous(x)]
                return(offsets)
            }
            return(NULL)
        }


        myObj <- function(theta) {
            theta0 <- theta
            if (Optim$constrain) {
                theta[constrained] <- exp(theta[constrained])
            }
            pp <- modelPar(x,theta)$p
            res <- c()
            for (i in seq_len(x$ngroup)) {
                offset <- MkOffset(pp[[i]],x$lvm[[i]],x$data[[i]],xconstrain[[i]])
                x0 <- x$lvm[[i]]
                data0 <- x$data[[i]][,index(x$lvm[[i]])$manifest,drop=FALSE]
                S <- x$samplestat[[i]]$S
                mu <- x$samplestat[[i]]$mu
                n <- x$samplestat[[i]]$n
                if (!is.null(offset)) {
                    x0$constrain[iconstrain[[i]]] <- NULL
                    pd <- procdata.lvm(x0,data0[,endogenous(x0),drop=FALSE]-offset)
                    S[endogenous(x0),endogenous(x0)] <- pd$S
                    mu[endogenous(x0)] <- pd$mu
                    n <- pd$n
                    x0$mean[yconstrain[[i]]] <- 0
                }
                res <- c(res,
                         do.call(ObjectiveFun, list(x=x0, p=pp[[i]], data=data0, S=S, mu=mu, n=n, weights=weights[[i]], data2=data2[[i]], offset=offset)))

            }
            sum(res)
        }

        if (!exists(GradFun)) {
            myGrad <- NULL
        } else  {
            myGrad <- function(theta) {
                theta0 <- theta
                if (Optim$constrain) {
                    theta[constrained] <- exp(theta[constrained])
                }
                pp <- modelPar(x,theta)$p
                D0 <- res <- rbind(numeric(length(theta)))
                for (i in seq_len(x$ngroup)) {
                    repval <- with(x$samplestat[[i]],
                                   do.call(GradFun, list(x=x$lvm[[i]],p=pp[[i]],
                                                         data=x$data[[i]][,index(x$lvm[[i]])$manifest,drop=FALSE],
                                                         S=S,mu=mu,n=n,
                                                         weights=weights[[i]], data2=data2[[i]])))
                    D <- D0; D[ parord[[i]] ] <- repval
                    res <- res + D
                }
                if (Optim$constrain) {
                    res[constrained] <- res[constrained]*theta[constrained]
                }
                return(as.vector(res))
            }
        }

        myInformation <- function(theta) {
            theta0 <- theta
            if (Optim$constrain) {
                theta[constrained] <- exp(theta[constrained])
            }
            pp <- modelPar(x,theta)$p
            I0 <- res <- matrix(0,length(theta),length(theta))
            for (i in seq_len(x$ngroup)) {
                I <- I0;
                I[ parord[[i]], parord[[i]] ] <- with(x$samplestat[[i]], do.call(InformationFun, list(p=pp[[i]], x=x$lvm[[i]], data=x$data[[i]],
                                                                                                      S=S, mu=mu, n=n, weights=weights[[i]],
                                                                                                      data2=data2[[i]],
                                                                                                      type=Optim$information)))
                res <- res + I
            }
            D <- myGrad(theta0)
            if (Optim$constrain) {
                res[constrained,-constrained] <- apply(res[constrained,-constrained,drop=FALSE],2,function(x) x*theta[constrained]);
                res[-constrained,constrained] <- t(res[constrained,-constrained])
                if (sum(constrained)==1) {
                    res[constrained,constrained] <- res[constrained,constrained]*outer(theta[constrained],theta[constrained]) - (D[constrained])
                } else {
                    res[constrained,constrained] <- res[constrained,constrained]*outer(theta[constrained],theta[constrained]) - diag(D[constrained],nrow=length(constrained))
                }
            }
            attributes(res)$grad <- D
            return(res)
        }
    }

    if (!exists(InformationFun)) myInformation <- NULL
    else if (is.null(get(InformationFun))) myInformation <- NULL
    if (is.null(get(GradFun))) myGrad <- NULL

    if (messages>1) cat("Optimizing objective function...\n")
    if (lava.options()$debug) {
        print(lower)
        print(Optim$constrain)
        print(Optim$method)
    }
    opt <- do.call(Optim$method,
                   list(start=mystart, objective=myObj, gradient=myGrad, hessian=myInformation, lower=lower, control=Optim))

    opt$estimate <- opt$par
    if (Optim$constrain) {
        opt$estimate[constrained] <- exp(opt$estimate[constrained])
    }
    if (quick) return(list(opt=opt,vcov=NA))

    if (is.null(myGrad) | !XconstrStdOpt ) {
        ## if (!requireNamespace("numDeriv")) {
        ##   opt$gradient <- naiveGrad(myObj, opt$estimate)
        ## } else {
        opt$gradient <- numDeriv::grad(myObj, opt$par, method=lava.options()$Dmethod)
    } else {
        opt$gradient <- myGrad(opt$estimate)
    }

    if (!is.null(opt$convergence)) {
        if (opt$convergence!=0) warning("Lack of convergence. Increase number of iteration or change starting values.")
    } else if (!is.null(opt$gradient) && mean(opt$gradient)^2>1e-3) warning("Lack of convergence. Increase number of iteration or change starting values.")

    if (!XconstrStdOpt) {
        myInformation <- function(theta) information(x,p=theta)
    } else {
        if (is.null(myInformation)) {
            ##     if (!requireNamespace("numDeriv")) stop("I do not know how to calculate the asymptotic variance of this estimator.
            ## For numerical approximation please install the library 'numDeriv'.")
            if (!is.null(myGrad) & XconstrStdOpt)
                myInformation <- function(theta) numDeriv::jacobian(myGrad, theta, method=lava.options()$Dmethod)
            else {
                myInformation <- function(theta) numDeriv::hessian(myObj, theta)
            }
        }
    }
    I <- myInformation(opt$estimate)
    asVar <- tryCatch(Inverse(I),
                      error=function(e) matrix(NA, length(mystart), length(mystart)))

    res <- list(model=x, model0=mymodel, call=cl, opt=opt, meanstructure=Optim$meanstructure,
                vcov=asVar, estimator=estimator, weights=weights, data2=data2, cluster=id)
    class(res) <- myclass

    myhooks <- gethook("post.hooks")
    for (f in myhooks) {
        res0 <- do.call(f,list(x=res))
        if (!is.null(res0))
            res <- res0
    }

    return(res)
}

estimate_lvmlist <-
    function(x, data, messages=lava.options()$messages, fix, missing=FALSE,  ...) {
        if (base::missing(data)) {
            return(estimate(x[[1]],x[[2]],missing=missing,...))
        }
        nm <- length(x)
        if (nm==1) {
            return(estimate(x[[1]],data,missing=missing,...))
        }
        if (!all(unlist(lapply(x, function(y) inherits(y,"lvm")))))
            stop ("Expected a list of 'lvm' objects.")
        if (is.data.frame(data)) {
            warning("Only one dataset - going for standard analysis on each submodel.")
            res <- c()
            for (i in seq_len(nm)) {
                res <- c(res, list(estimate(x[[i]],data=data,messages=0,missing=missing, ...)))
            }
            return(res)
        }

        if (nm!=length(data)) stop("Supply dataset for each model")

        Xfix <- FALSE
        xfix <- list()
        for (i in seq_along(x)) {
            data0 <- data[[i]]
            xfix0 <- colnames(data0)[(colnames(data0)%in%parlabels(x[[i]],exo=TRUE))]
            xfix <- c(xfix, list(xfix0))
            if (length(xfix0)>0) { ## Yes, random slopes
                Xfix<-TRUE
            }
        }
        if (base::missing(fix)) {
            fix <- ifelse(Xfix,FALSE,TRUE)
        }


        mg <- multigroup(x,data,fix=fix,missing=missing,...)
        res <- estimate(mg,...)

        return(res)
    }

##' Estimate method for formulas
##'
##' Fit a GLM model specified by a formula and apply [estimate.default]
##' for robust inference.
##' @param x formula specifying the model
##' @param data data.frame
##' @param weights optional weights
##' @param family GLM family (default `gaussian`)
##' @param ... Additional arguments to [estimate.default]
##' @param model character string specifying the model type (default `"glm"`)
##' @param lvm logical, if TRUE use `estimate.lvm` instead
##' @return Object of class `estimate` (see [estimate.default]).
##' @export
estimate.formula <- function(x, data, weights, family=stats::gaussian,
                      ..., model="glm", lvm=FALSE) {

    cl <- match.call()
    if (lvm) {
        cl[[1]] <- as.call(parse(text="estimate0"))[[1]]
        return(eval(cl,envir=parent.frame()))
    }
    if (missing(data)) {
        data <- environment(x)
        cl$data <- quote(data)
    }
    mf <- match.call(expand.dots = FALSE)
    m <- match(c("x", "data", "weights", "subset",
                 "etastart", "mustart", "offset"), names(mf), 0L)
    mf <- mf[c(1L, m)]
    mf$drop.unused.levels <- TRUE
    mf[[1L]] <- quote(stats::model.frame)
    mf$na.action <- na.pass0
    names(mf)[which(names(mf)=="x")] <- "formula"
    mf <- eval(mf, parent.frame())
    idx <- attr(na.omit(mf),"na.action")
    weights <- as.vector(model.weights(mf))
    if (length(idx)>0) {
        if (is.null(weights)) weights <- rep(1,nrow(mf))
        if (length(idx)>0) weights[idx] <- 0
        cl$weights <- quote(weights)
        cl$na.action <- quote(na.pass0)
    }
    cl$data <- quote(data)
    argsModelObj <- names(formals(model))
    dots <- list(...)
    idx <- which(names(dots) %ni% argsModelObj)
    rmarg <- c("model","raw","lvm")
    if (length(idx)>0) rmarg <- c(names(dots)[idx],rmarg)
    cl[rmarg] <- NULL
    names(cl)[names(cl)=="x"] <- "formula"
    cl[[1]] <- as.name(model)
    fit <- eval(cl)
    if (length(idx)==0) return(fit)
    optarg <- NULL
    if (length(idx)>0) {
        optarg <- dots[idx]
    }
    do.call(estimate, c(list(fit),optarg))
}


estimate0 <- function(x,data=parent.frame(),pred.norm=c(),unstruct=FALSE,messages=0,id=NULL,distribution=NULL,estimator="gaussian",...) {
    formulaId <- union(Specials(x,c("cluster")),Specials(x,c("id")))
    formulaSt <- paste0("~.-cluster(",formulaId,")-id(",formulaId,")")
    if (!is.null(formulaId)) {
        id <- formulaId
        x <- update(x,as.formula(formulaSt))
    }
    if (!is.null(id))
        x <- update(x,as.formula(paste(".~.+",id)))
    mf <- model.frame(x,data)
    yvar <- names(mf)[1]
    y <- mf[,yvar]
    opt <- options(na.action="na.pass")
    mm <- model.matrix(x,data)
    options(opt)
    covars <- colnames(mm)
    covars <- unlist(lapply(covars, function(x) gsub("[^a-zA-Z0-9._]","",x)))
    colnames(mm) <- covars

    if (attr(terms(x),"intercept")==1) {
        covars <- covars[-1]
        it <- c()
    } else {
        it <- "0"
    }

    if (!is.null(id)) covars <- setdiff(covars,id)
    model <- lvm(toformula(yvar,c(it,covars)),messages=0)
    if (!is.null(distribution)) {
        lava::distribution(model,yvar) <- distribution
        estimator <- "glm"
    }
    mydata <- na.omit(as.data.frame(cbind(data.frame(y),mm))); names(mydata)[1] <- yvar
    exogenous(model) <- setdiff(covars,pred.norm)
    if (unstruct) {
        model <- covariance(model,pred.norm,pairwise=TRUE)
    }
    estimate(model,mydata,messages=messages,id=id,estimator=estimator,...)
}

##' Generic method for memberships from object (e.g. a graph)
##'
##' @title Extract children or parent elements of object
##' @export
##' @aliases children parents ancestors descendants roots sinks adjMat edgeList
##' @param object Object
##' @param \dots Additional arguments
##' @author Klaus K. Holst
"children" <- function(object,...) UseMethod("children")
##' @export
"parents" <- function(object,...) UseMethod("parents")
##' @export
"roots" <- function(object,...) UseMethod("roots")
##' @export
"sinks" <- function(object,...) UseMethod("sinks")
##' @export
"descendants" <- function(object,...) UseMethod("descendants")
##' @export
"ancestors" <- function(object,...) UseMethod("ancestors")
##' @export
"adjMat" <- function(object,...) UseMethod("adjMat")
##' @export
"edgeList" <- function(object,...) UseMethod("edgeList")

##' @export
adjMat.lvm <- function(object,...) t(object$M)

##' @export
adjMat.lvmfit <- function(object,...) adjMat(Model(object),...)

##' @export
edgeList.lvmfit <- function(object,...) edgeList(Model(object),...)

##' @export
edgeList.lvm <- function(object, labels=FALSE, list=FALSE, ...) {
    edgelist <- data.frame(from=NULL,to=NULL)
    A <- adjMat(object)
    for (i in seq_len(nrow(A))) {
        ii <- which(A[,i]>0)
        if (length(ii)>0)
            edgelist <- rbind(edgelist,data.frame(from=ii,to=i))
    }
    if (list) labels <- TRUE
    if (labels) edgelist <- as.data.frame(apply(edgelist,2,function(x) vars(object)[x]))
    if (list) {
      edgelist <- apply(edgelist, 1, function(x) paste0(x,collapse="~"))
    }
    edgelist
}

##' @export
parents.lvmfit <- function(object,...) parents(Model(object),...)

##' @export
children.lvmfit <- function(object,...) children(Model(object),...)

##' @export
descendants.lvmfit <- function(object,...) descendants(Model(object),...)

##' @export
ancestors.lvmfit <- function(object,...) ancestors(Model(object),...)

##' @export
roots.lvmfit <- function(object,...) roots(Model(object),...)

##' @export
sinks.lvmfit <- function(object,...) sinks(Model(object),...)



##' @export
parents.lvm <- function(object,var,...) {
    A <- index(object)$A
    if (missing(var)) {
        return(rownames(A))
    }
    if (inherits(var,"formula"))
        var <- all.vars(var)
    res <- lapply(var, function(v) rownames(A)[A[,v]!=0])
    res <- unique(unlist(res))
    if (length(res)==0) res <- NULL
    res
}

##' @export
children.lvm <- function(object,var,...) {
    A <- index(object)$A
    if (missing(var)) {
        return(rownames(A))
    }
    if (inherits(var,"formula"))
        var <- all.vars(var)
    res <- lapply(var, function(v) rownames(A)[A[v,]!=0])
    res <- unique(unlist(res))
    if (length(res)==0) res <- NULL
    res
}


##' @export
ancestors.lvm <- function(object,x,...) {
    if (inherits(x,"formula")) x <- all.vars(x)
    res <- c()
    left <- setdiff(vars(object),x)
    count <- 0
    child <- x
    while (length(x)>0) {
        count <- count+1
        x <- parents(object,child)
        child <- intersect(x,left)
        res <- union(res,child)
        left <- setdiff(left,child)
    }
    if (length(res)==0) res <- NULL
    return(res)
}

##' @export
descendants.lvm <- function(object,x,...) {
    if (inherits(x,"formula")) x <- all.vars(x)
    res <- c()
    left <- setdiff(vars(object),x)
    count <- 0
    parent <- x
    while (length(x)>0) {
        count <- count+1
        x <- children(object,parent)
        parent <- intersect(x,left)
        res <- union(res,parent)
        left <- setdiff(left,parent)
    }
    if (length(res)==0) res <- NULL
    return(res)
}

##' @export
roots.lvm <- function(object,...) {
    return(exogenous(object,index=FALSE,...))
}

##' @export
sinks.lvm <- function(object,...) {
    return(endogenous(object,top=TRUE,...))
}

##' Identify candidates of equivalent models
##'
##' Identifies candidates of equivalent models
##'
##'
##' @param x \code{lvmfit}-object
##' @param rel Formula or character-vector specifying two variables to omit from
##' the model and subsequently search for possible equivalent models
##' @param tol Define two models as empirical equivalent if the absolute
##' difference in score test is less than \code{tol}
##' @param k Number of parameters to test simultaneously. For \code{equivalence}
##' the number of additional associations to be added instead of \code{rel}.
##' @param omitrel if \code{k} greater than 1, this boolean defines wether to
##' omit candidates containing \code{rel} from the output
##' @param \dots Additional arguments to be passed to the lower-level functions
##' @author Klaus K. Holst
##' @seealso \code{\link{compare}}, \code{\link{modelsearch}}
##' @export
equivalence <- function(x,rel,tol=1e-3,k=1,omitrel=TRUE,...) {
    if (missing(rel)) stop("Specify association 'rel' (formula or character vector)")
    if (inherits(rel,"formula")) {
        myvars <- all.vars(rel)
    } else {
        myvars <- rel
    }
    if (length(myvars)!=2) stop("Two variables only")
    x0 <- Model(x)
    cancel(x0) <- rel
    e0 <- estimate(x0,data=model.frame(x),weights=Weights(x),estimator=x$estimator,...)
    if (k!=1) {
        p0 <- coef(x)
        p0[] <- 0
        p0[match(names(coef(e0)),names(p0))] <- coef(e0)
        S0 <- score(x,p=p0)[,,drop=TRUE];
        I0 <- information(x,p=p0)
        T0 <- rbind(S0)%*%solve(I0)%*%cbind(S0); names(T0) <- "Q"
    }
    s <- modelsearch(e0,k=k,...)
    relname <- c(paste(myvars,collapse=lava.options()$symbol[2]),
                 paste(rev(myvars),collapse=lava.options()$symbol[2]))
    relidx <- NULL
    if (k==1) {
        relidx <- na.omit(match(relname,s$res[,"Index"]))
        T0 <- s$test[relidx,1]
    }
    T <- s$test[,1]
    Equiv <- setdiff(which(abs(T-T0)<tol),relidx)
    Improve <- which((T-T0)>tol)
    if (omitrel) { ## Don't save models including 'rel'
        keep <- c()
        if (length(Equiv)>0) {
            for (i in seq_len(length(Equiv))) {
                newvars <- s$var[[Equiv[i]]]
                if (!any(apply(newvars,1,function(z) all(z%in%myvars)))) keep <- c(keep,Equiv[i])
            }
            Equiv <- keep
        }
        keep <- c()
        if (length(Improve)>0) {
            for (i in seq_len(length(Improve))) {
                newvars <- s$var[[Improve[i]]]
                if (!any(apply(newvars,1,function(z) all(z%in%myvars)))) keep <- c(keep,Improve[i])
            }
            Improve <- keep
        }
    }
    eqvar <- ivar <- NULL
    models <- list()
    if (length(Equiv)>0){
        for (i in seq_len(length(Equiv))) {
            xnew <- x0
            newvars <- s$var[[Equiv[i]]]
            for (j in seq_len(nrow(newvars))) {
                exo.idx <- which(newvars[j,]%in%index(x0)$exogenous)
                if (length(exo.idx)>0) {
                    xnew <- regression(xnew,from=newvars[j,exo.idx],to=newvars[j,setdiff(1:2,exo.idx)])
                } else {
                    covariance(xnew) <- newvars
                }
            }
            models <- c(models,list(xnew))
        }
        eqvar <- s$var[Equiv]
    }
    if (length(Improve)>0)   {
        for (i in seq_len(length(Improve))) {
            xnew <- x0
            newvars <- s$var[[Improve[i]]]
            for (j in seq_len(nrow(newvars))) {
                exo.idx <- which(newvars[j,]%in%index(x0)$exogenous)
                if (length(exo.idx)>0) {
                    xnew <- regression(xnew,from=newvars[j,exo.idx],to=newvars[j,setdiff(1:2,exo.idx)])
                } else {
                    covariance(xnew) <- newvars
                }
            }
            models <- c(models,list(xnew))
        }
        ivar <- s$var[Improve]
    }
    res <- list(equiv=eqvar, improve=ivar, scoretest=s, models=models, I=Improve, E=Equiv, T0=T0, vars=myvars)
    class(res) <- "equivalence"
    return(res)
}

##' @export
print.equivalence <- function(x,...) {
    cat("  0)\t ",paste0(x$vars,collapse=lava.options()$symbol[2]),"  (",formatC(x$T0),")\n")
    cat("Empirical equivalent models:\n")
    if (length(x$E)==0)
        cat("\t none\n")
    else
        for (i in seq_len(length(x$E))) {
            cat("  ",i,")\t ",  x$scoretest$res[x$E[i],"Index"],
                "  (",x$scoretest$res[x$E[i],1],")",
                "\n",sep="")
        }
    cat("Candidates for model improvement:\n")
    if (length(x$I)==0)
        cat("\t none\n")
    else
        for (i in seq_len(length(x$I))) {
            cat("  ",i,")\t ",  x$scoretest$res[x$I[i],"Index"],
                "  (",x$scoretest$res[x$I[i],1],")",
                "\n",sep="")
        }
    invisible(x)
}

holm <- function(p) {
    k <- length(p)
    w <- 1/k
    ii <- order(p)
    po <- p[ii]
    qs <- min(1,po[1]/w)
    for (i in 2:k) {
        qs <- c(qs, min(1, max(qs[i-1],po[i]*(1-w*(i-1))/w)))
    }
    return(qs)
}

##' @export
categorical <- function(x,formula,K,beta,p,liability=FALSE,regr.only=FALSE,exo=TRUE,...) {

    if (is.character(formula)) {
        regr <- FALSE
        X <- formula
    } else {
        y <- getoutcome(formula)
        X <- attributes(y)$x
        regr <- TRUE
        if (length(y)==0) regr <- FALSE
        if (length(attributes(y)$x)==0) {
            X <- y; regr <- FALSE
        }
    }
    if (!missing(p)) {
        if (!missing(K)) {
            if (!(K==length(p) || K==length(p)+1)) stop("Wrong dimension of 'p'")
            if (length(K)==length(p)) {
                if (!identical(sum(p),1.0)) stop("Not a probability vector")
                p <- p[-length(p)]
            }
        }
        if (is.numeric(p) && sum(p)>1) warning("'p' sum > 1")
        if (is.logical(all.equal(1.0,sum(p)))) p <- p[-length(p)]
    }
    if (missing(K)) {
        if (!is.null(list(...)$labels)) K <- length(list(...)$labels)
        if (!missing(beta)) K <- length(beta)
        if (!missing(p)) K <- length(p)+1
    }
    if (!regr.only) {
        if (missing(p)) p <- rep(1/K,K-1)
        pname <- names(p)
        if (is.null(pname)) pname <- rep(NA,K-1)
        ordinal(x,K=K,liability=liability,p=p,constrain=pname,exo=exo,...) <- X
        if (!regr) return(x)
    }

    if (missing(beta)) beta <- rep(0,K)
    fname <- paste(gsub(" ","",deparse(formula)),seq(K)-1,sep=":")
    fpar <- names(beta)
    if (is.null(fpar)) fpar <- fname

    parameter(x,fpar,start=beta) <- fname
    val <- paste0("function(x,p,...) p[\"",fpar[1],"\"]*(x==0)")
    for (i in seq(K-1)) {
        val <- paste0(val,"+p[\"",fpar[i+1],"\"]*(x==",i,")")
    }
    functional(x,formula) <- eval(parse(text=val))
    return(x)
}

##' @export
'categorical<-' <- function(x,...,value) categorical(x,value,...)

##' Add Confidence limits bar to plot
##'
##' @title Add Confidence limits bar to plot
##' @param x Position (x-coordinate if vert=TRUE, y-coordinate otherwise)
##' @param lower Lower limit (if NULL no limits is added, and only the
##' center is drawn (if not NULL))
##' @param upper Upper limit
##' @param center Center point
##' @param line If FALSE do not add line between upper and lower bound
##' @param delta Length of limit bars
##' @param centermark Length of center bar
##' @param pch Center symbol (if missing a line is drawn)
##' @param blank If TRUE a white ball is plotted before the center is
##' added to the plot
##' @param vert If TRUE a vertical bar is plotted. Otherwise a horizontal
##' bar is used
##' @param polygon If TRUE polygons are added between 'lower' and 'upper'.
##' @param step Type of polygon (step-function or piecewise linear)
##' @param ... Additional low level arguments (e.g. col, lwd, lty,...)
##' @seealso \code{confband}
##' @export
##' @keywords iplot
##' @aliases confband forestplot plot_region
##' @author Klaus K. Holst
##' @examples
##' plot(0,0,type="n",xlab="",ylab="")
##' confband(0.5,-0.5,0.5,0,col="darkblue")
##' confband(0.8,-0.5,0.5,0,col="darkred",vert=FALSE,pch=1,cex=1.5)
##'
##' set.seed(1)
##' K <- 20
##' est <- rnorm(K)
##' se <- runif(K,0.2,0.4)
##' x <- cbind(est,est-2*se,est+2*se,runif(K,0.5,2))
##' x[c(3:4,10:12),] <- NA
##' rownames(x) <- unlist(lapply(letters[seq(K)],function(x) paste(rep(x,4),collapse="")))
##' rownames(x)[which(is.na(est))] <- ""
##' signif <- sign(x[,2])==sign(x[,3])
##' forestplot(x,text.right=FALSE)
##' forestplot(x[,-4],sep=c(2,15),col=signif+1,box1=TRUE,delta=0.2,pch=16,cex=1.5)
##' forestplot(x,vert=TRUE,text=FALSE)
##' forestplot(x,vert=TRUE,text=FALSE,pch=NA)
##' ##forestplot(x,vert=TRUE,text.vert=FALSE)
##' ##forestplot(val,vert=TRUE,add=TRUE)
##'
##' z <- seq(10)
##' zu <- c(z[-1],10)
##' plot(z,type="n")
##' confband(z,zu,rep(0,length(z)),col=Col("darkblue"),polygon=TRUE,step=TRUE)
##' confband(z,zu,zu-2,col=Col("darkred"),polygon=TRUE,step=TRUE)
##'
##' z <- seq(0,1,length.out=100)
##' plot(z,z,type="n")
##' confband(z,z,z^2,polygon="TRUE",col=Col("darkblue"))
##'
##' set.seed(1)
##' k <- 10
##' x <- seq(k)
##' est <- rnorm(k)
##' sd <- runif(k)
##' val <- cbind(x,est,est-sd,est+sd)
##' par(mfrow=c(1,2))
##' plot(0,type="n",xlim=c(0,k+1),ylim=range(val[,-1]),axes=FALSE,xlab="",ylab="")
##' axis(2)
##' confband(val[,1],val[,3],val[,4],val[,2],pch=16,cex=2)
##' plot(0,type="n",ylim=c(0,k+1),xlim=range(val[,-1]),axes=FALSE,xlab="",ylab="")
##' axis(1)
##' confband(val[,1],val[,3],val[,4],val[,2],pch=16,cex=2,vert=FALSE)
##'
##' x <- seq(0, 3, length.out=20)
##' y <- cos(x)
##' yl <- y - 1
##' yu <- y + 1
##' plot_region(x, y, yl, yu)
##' plot_region(x, y, yl, yu, type='s', col="darkblue", add=TRUE)
##'
confband <- function(x, lower, upper, center=NULL,
                     line=TRUE, delta=0.07,
                     centermark=0.03,
                     pch, blank=TRUE, vert=TRUE,
                     polygon=FALSE, step=FALSE,...) {
    if (polygon) {
        if (step) {
            x1 <- rep(x,each=2)[-1]
            y1 <- rep(lower, each=2);  y1 <- y1[-length(y1)]
            x2 <- rep(rev(x),each=2); x2 <- x2[-length(x2)]
            y2 <- rep(rev(upper), each=2)[-1]
            xx <- c(x1,x2)
            if (!is.null(center))
                center <- rep(center, each=2)[-1]
            yy <- c(y1,y2)
        } else {
            xx <- c(x, rev(x))
            yy <- c(lower, rev(upper))
        }
        polygon(xx,yy,...)
        if (line && !is.null(center)) {
          mlines <- function(x, y, ..., border, fillOddEven) {
            if (step) {
              lines(x, y, type="s", ...)
            } else {
              lines(x, y, ,...)
            }
          }
          mlines(xx[seq(length(xx)/2)],center,...)
        }
        return(invisible(NULL))
    }
    if (vert) {
        if (line && !missing(lower) && !missing(upper))
            segments(x,lower,x,upper,...)
        if (!missing(lower))
            segments(x-delta,lower,x+delta,lower,...)
        if (!missing(upper))
            segments(x-delta,upper,x+delta,upper,...)
        if (!is.null(center)) {
            if (!missing(pch)) {
                if (blank) {
                  points(x,center,pch=16,col="white")
                }
                points(x,center,pch=pch,...)
            } else {
                segments(x-centermark,center,x+centermark,center,...)
            }
        }
    } else {
        if (line && !missing(lower) && !missing(upper))
            segments(lower,x,upper,x,...)
        if (!missing(lower))
            segments(lower,x-delta,lower,x+delta,...)
        if (!missing(upper))
            segments(upper,x-delta,upper,x+delta,...)

        if (!is.null(center)) {
            if (!missing(pch)) {
                if (blank)
                    points(center,x,pch=16,col="white")
                points(center,x,pch=pch,...)
            } else {
                segments(center,x-centermark,center,x+centermark,...)
            }
        }
    }
    if (missing(lower)) lower <- NULL
    if (missing(upper)) upper <- NULL
    invisible(c(x,lower,upper,center))
}


##' @export
forestplot <- function(x,lower,upper,line=0,labels,
                text=TRUE,text.right=text,text.fixed=NULL,text.vert=TRUE,
                adj=NULL,
                delta=0,axes=TRUE,cex=1,pch=15,
                xlab="",ylab="",sep,air,
                xlim,ylim,mar,box1=FALSE,box2=FALSE,
                vert=FALSE,cex.axis=1,cex.estimate=0.8,
                add=FALSE,
                reset.par=FALSE,
                ...) {
    if (is.matrix(x)) {
        lower <- x[,2]; upper <- x[,3]
        if (ncol(x)>3) cex <- x[,4]
        x <- x[,1]
    }
    if (missing(mar) && !add) {
        if (vert) {
            mar <- c(8,4,1,1)
        } else {
            mar <- c(4,8,1,1)
        }
    }
    if (missing(labels)) labels <- names(x)
    K <- length(x)
    onelayout <- FALSE
    if (!add) {
        def.par <- par(no.readonly=TRUE)
        if (reset.par) on.exit(par(def.par))
        if (text.right) {
            if (vert) {
                layout(rbind(1,2),heights=c(0.25,0.75))
            } else {
                layout(cbind(2,1),widths=c(0.75,0.25))
            }
        } else {
            onelayout <- TRUE
            layout(1)
        }
    }
    if (vert) {
        if (missing(ylim)) {
            if (missing(air)) air <- max(upper-lower,na.rm=TRUE)*0.5
            ylim <- range(c(x,lower-air,upper+air),na.rm=TRUE)
        }
        if (missing(xlim)) xlim <- c(0,K) + 0.5
    } else {
        if (missing(ylim)) ylim <- c(0,K) + 0.5
        if (missing(xlim)) {
            if (missing(air)) air <- max(upper-lower,na.rm=TRUE)*0.5
            xlim <- range(c(x,lower-air,upper+air),na.rm=TRUE)
        }
    }
    args0 <- list(...)
    formatCargsn <- names(formals(args(formatC)))[-1]
    nn <- setdiff(names(args0),formatCargsn)
    plotargs <- args0[nn]

    mainplot <- function(...) {
        par(mar=mar) ## bottom,left,top,right
        do.call("plot",c(list(x=0,type="n",axes=FALSE,xlab=xlab,ylab=ylab,xlim=xlim,ylim=ylim),plotargs))
        if (box1) box()
        if (axes) {
            if (vert) {
                axis(2,cex.axis=cex.axis)
            } else {
                axis(1,cex.axis=cex.axis)
            }
        }
    }
    if (onelayout && !add) mainplot()

    if (text) {
        xpos <- upper
        if (text.right && !add) {
            if (vert) {
                par(mar=c(0,mar[2],0,mar[4]))
            } else {
                par(mar=c(mar[1],0,mar[3],0))
            }
            plot.new()
            if (vert) {
                plot.window(xlim=xlim,ylim=c(0,0.5))
            } else {
                plot.window(ylim=ylim,xlim=c(0,0.5))
            }
            if (box2) box()
            xpos[] <- 0
        }
        if (!is.null(text.fixed)) {
            if (is.logical(text.fixed) && text.fixed) text.fixed <- max(xpos)
            xpos <- rep(text.fixed,length.out=K)
        }
        nn <- intersect(names(args0),formatCargsn)
        args <- args0[nn]
        for (i in seq_len(K)) {
            st <- c(do.call(formatC,c(list(x=x[i]),args)),
                    paste0("(",
                           do.call(formatC,c(list(x=lower[i]),args)),"; ",
                           do.call(formatC,c(list(x=upper[i]),args)),")"))
            if (text.vert) {
                st <- paste0(" ",st[1]," ",st[2],collapse="")
                st <- paste(" ", st)
            }
            if (vert) {
                if (!is.na(x[i])) {
                    if (!text.vert) {
                        if (text.right) xpos[i] <- xpos[i]+0.025
                        graphics::text(i,xpos[i],paste(st,collapse="\n"),xpd=TRUE,
                                       offset=3, cex=cex.estimate, adj=adj)
                    } else {
                        if (!is.na(x[i])) {
                          graphics::text(i,xpos[i],st,xpd=TRUE, srt=90, offset=0, pos=4, cex=cex.estimate, adj=adj)
                        }
                    }
                }
            } else {
                if (!is.na(x[i])) {
                  graphics::text(xpos[i],i,trim(st),xpd=TRUE,pos=4,cex=cex.estimate, adj=adj)
                }
            }
        }
    }

    if (!onelayout && !add) mainplot()

    if (!is.null(line)) {
        if (vert) {
            abline(h=line,lty=2,col="lightgray")
        } else {
            abline(v=line,lty=2,col="lightgray")
        }
    }
    if (!missing(sep)) {
        if (vert) {
            abline(v=sep+.5,col="gray")
        } else {
            abline(h=sep+.5,col="gray")
        }
    }
    do.call("confband",
            c(list(x=seq(K),lower=lower,upper=upper,x,
                   pch=pch,cex=cex,vert=vert,blank=FALSE),
              plotargs))
    if (!add) {
        if (is.null(adj)) adj <- NA
        if (vert) {
            mtext(labels,1,at=seq(K),las=2,line=1,cex=cex.axis, adj=adj)
        } else {
            mtext(labels,2,at=seq(K),las=2,line=1,cex=cex.axis, adj=adj)
        }
    }
}


##' @export
plot_region <- function(x, y=NULL, lower=NULL, upper=NULL,
                        line=TRUE, add=FALSE,
                        border=NA, col=1, alpha=0.5,
                        type="l", ...) {
  if (is.matrix(x) && ncol(x)==4) {
    if (ncol(x)>=4) {
      y <- x[, 2]
      lower <- x[, 3]
      upper <- x[, 4]
      x <- x[, 1]
    }
    if (ncol(x)==3) {
      lower <- x[, 2]
      upper <- x[, 3]
      x <- x[, 1]
      line <- FALSE
    }
  } else {
    if (is.matrix(y)) {
      if (ncol(y)==2) {
        lower <- y[, 1]
        upper <- y[, 2]
      } else if (ncol(y)>2) {
        lower <- y[, 2]
        upper <- y[, 3]
        y <- y[, 1]
      }
    }
  }
  if (!add) {
    yy <- range(c(y,lower,upper))
    plot(range(x), range(yy), ..., type="n")
  }
  confband(x, lower, upper, y, line=FALSE, polygon=TRUE,
           step=(type=="s"), col=Col(col, alpha),
           border=border, ...)
  if (line) {
    lines(x, y, type=type, col=col, ...)
  }
}

context("Graphics functions")
      
testthat::test_that("color", {
    cur <- palette()
    old <- lava:::mypal()
    testthat::expect_equivalent(col2rgb(cur),col2rgb(old))
    testthat::expect_equivalent(col2rgb(palette()),col2rgb(lava:::mypal(set=FALSE)))

    testthat::expect_equivalent(Col("red",0.5),rgb(1,0,0,0.5))
    testthat::expect_equivalent(Col(c("red","blue"),0.5),rgb(c(1,0),c(0,0),c(0,1),0.5))
    testthat::expect_equivalent(Col(c("red","blue"),c(0.2,0.5)),rgb(c(1,0),c(0,0),c(0,1),c(0.2,0.5)))
    testthat::expect_equivalent(Col(rgb(1,0,0),0.5),rgb(1,0,0,0.5))

    plot(0,xlim=c(0,1),ylim=c(0,1),type="n",ann=FALSE,axes=FALSE)
    devc1 <- devcoords()
    par(mar=c(0,0,0,0))
    plot(0,xlim=c(0,1),ylim=c(0,1),type="n",ann=FALSE,axes=FALSE)
    devc2 <- devcoords()
    figx <- c("fig.x1","fig.x2","fig.y1","fig.y2")
    devx <- c("dev.x1","dev.x2","dev.y1","dev.y2")
    testthat::expect_equivalent(devc1[figx],devc2[devx])    
    
})

if (requireNamespace("visualTest",quietly=TRUE)) {

    gropen <- function(resolution=200,...) {
        tmpfile <- tempfile(fileext=".png")
        png(file=tmpfile,width=200,height=200)
        res <- dev.cur()
        return(structure(tmpfile,dev=res))
    }
    grcompare <- function(file1,file2,...) {
        res <- visualTest::isSimilar(file1,file2,...)
        unlink(c(file1,file2))
        return(res)
    }

    
    testthat::test_that("plotConf", {
        set.seed(1)
        x <- rnorm(50)
        y <- rnorm(50,x)
        z <- rbinom(50,1,0.5)
        d <- data.frame(y,z,x)
        l <- lm(y~x*z)
        
        d1 <- gropen()
        par(mar=c(0,0,0,0))
        plotConf(l,var1="x",var2="z",col=c("black","blue"),alpha=0.5,legend=FALSE)
        dev.off()
        
        newd <- data.frame(x=seq(min(x),max(x),length.out=100))
        l0 <- lm(y~x,subset(d,z==0))
        ci0 <- predict(l0,newdata=newd,interval="confidence")
        l1 <- lm(y~x,subset(d,z==1))
        ci1 <- predict(l1,newdata=newd,interval="confidence")

        d2 <- gropen()
        par(mar=c(0,0,0,0))
        plot(y~x,col=c("black","blue")[z+1],pch=16,ylim=c(min(ci0,ci1,y),max(ci0,ci1,y)))
        lines(newd$x,ci0[,1],col="black",lwd=2)
        lines(newd$x,ci1[,1],col="blue",lwd=2)
        confband(newd$x,lower=ci0[,2],upper=ci0[,3],polygon=TRUE,col=Col("black",0.5),border=FALSE)
        confband(newd$x,lower=ci1[,2],upper=ci1[,3],polygon=TRUE,col=Col("blue",0.5),border=FALSE)
        points(y~x,col=c("black","blue")[z+1],pch=16)
        dev.off()
        
        testthat::expect_true(grcompare(d1,d2,threshold=5))

        
        d1 <- gropen()
        par(mar=c(0,0,0,0))
        l <- lm(y~z)
        plotConf(l,var2="z",var1=NULL,jitter=0,col="black",alpha=0.5,xlim=c(.5,2.5),ylim=range(y))
        dev.off()

        d2 <- gropen()
        par(mar=c(0,0,0,0))
        plot(y~I(z+1),ylim=range(y),xlim=c(0.5,2.5),pch=16,col=Col("black",0.5)) 
        l0 <- lm(y~-1+factor(z))
        confband(1:2,lower=confint(l0)[,1],upper=confint(l0)[,2],lwd=3,
                 center=coef(l0))
        dev.off()

        testthat::expect_true(grcompare(d1,d2,threshold=10))
                
  })


    testthat::test_that("forestplot", {
        set.seed(1)
        K <- 20
        est <- rnorm(K); est[c(3:4,10:12)] <- NA        
        se <- runif(K,0.2,0.4)        
        x <- cbind(est,est-2*se,est+2*se,runif(K,0.5,2))        
        rownames(x) <- unlist(lapply(letters[seq(K)],function(x) paste(rep(x,4),collapse="")))
        rownames(x)[which(is.na(est))] <- ""
        signif <- sign(x[,2])==sign(x[,3])        
        forestplot(x)
        ## TODO
    })

    test_that("plot.sim", {
        onerun2 <- function(a,b,...) {
            return(cbind(a=a,b=b,c=a-1,d=a+1))
        }
        R <- data.frame(a=1:2,b=3:4)
        val2 <- sim(onerun2,R=R,type=0)
        plot(val2)
        plot(val2,plot.type="single")
        density(val2)
        ## TODO
    })

    test_that("spaghetti", {
        K <- 5
        y <- "y"%++%seq(K)
        m <- lvm()
        regression(m,y=y,x=~u) <- 1
        regression(m,y=y,x=~s) <- seq(K)-1
        regression(m,y=y,x=~x) <- "b"
        d <- sim(m,5)
        dd <- mets::fast.reshape(d);
        dd$num <- dd$num+rnorm(nrow(dd),sd=0.5) ## Unbalance
        spaghetti(y~num,dd,id="id",lty=1,col=Col(1,.4),trend=TRUE,trend.col="darkblue")
        ## TODO
    })
    
    test_that("ksmooth", {
        ## TODO
    })

    test_that("plot.lvm", {
        ## TODO
        m <- lvm(y~1*u[0:1],u~1*x)
        latent(m) <- ~u
        plot(m)
        d <- sim(m,20,seed=1)
        e <- estimate(m,d)
        plot(e)
        plot(lava:::beautify(m))
        g <- igraph.lvm(m)
        testthat::expect_true(inherits(g,"igraph"))
    })
    
    test_that("images", {
        ## TODO
    })

    test_that("labels,edgelabels", {
        ## TODO
    })

    test_that("colorbar", {
        ## TODO
    })

    test_that("fplot", {
        ## TODO
    })

    test_that("interactive", {
        ## TODO
    })

    test_that("pdfconvert", {
        ## TODO
    })

    test_that("plot.estimate", {
        ## TODO
    })

    test_that("logo", {
        lava(seed=1)
    })

    
    
    
}

context("Graphics functions")
library("vdiffr")

test_that("color, devcoords", {
  cur <- palette()
  old <- lava:::mypal()
  testthat::expect_equivalent(col2rgb(cur),col2rgb(old))
  testthat::expect_equivalent(col2rgb(palette()),col2rgb(lava:::mypal(set=FALSE)))

  testthat::expect_equivalent(Col("red",0.5),rgb(1,0,0,0.5))
  testthat::expect_equivalent(Col(c("red","blue"),0.5),rgb(c(1,0),c(0,0),c(0,1),0.5))
  testthat::expect_equivalent(Col(c("red","blue"),c(0.2,0.5)),rgb(c(1,0),c(0,0),c(0,1),c(0.2,0.5)))
  testthat::expect_equivalent(Col(rgb(1,0,0),0.5),rgb(1,0,0,0.5))

  plot(0,xlim=c(0,1),ylim=c(0,1),type="n",ann=FALSE,axes=FALSE)
  devc1 <- devcoords()
  par(mar=c(0,0,0,0))
  plot(0,xlim=c(0,1),ylim=c(0,1),type="n",ann=FALSE,axes=FALSE)
  devc2 <- devcoords()
  figx <- c("fig.x1","fig.x2","fig.y1","fig.y2")
  devx <- c("dev.x1","dev.x2","dev.y1","dev.y2")
  testthat::expect_equivalent(devc1[figx],devc2[devx])
})

test_that("plotConf", {
  skip_on_cran()

  set.seed(1)
  x <- rnorm(50)
  y <- rnorm(50,x)
  z <- rbinom(50,1,0.5)
  d <- data.frame(y,z,x)
  l <- lm(y~x*z)
  val <- expect_doppelganger("plotConf1", {
    val <- plotConf(l,var1="x",var2="z",col=c("black","blue"),alpha=0.5,legend=FALSE)
  })
  ##par(mar=c(0,0,0,0))
  ## newd <- data.frame(x=seq(min(x),max(x),length.out=100))
  ## l0 <- lm(y~x,subset(d,z==0))
  ## ci0 <- predict(l0,newdata=newd,interval="confidence")
  ## l1 <- lm(y~x,subset(d,z==1))
  ## ci1 <- predict(l1,newdata=newd,interval="confidence")
  ## plot(y~x,col=c("black","blue")[z+1],pch=16,ylim=c(min(ci0,ci1,y),max(ci0,ci1,y)))
  ## lines(newd$x,ci0[,1],col="black",lwd=2)
  ## lines(newd$x,ci1[,1],col="blue",lwd=2)
  ## confband(newd$x,lower=ci0[,2],upper=ci0[,3],polygon=TRUE,col=Col("black",0.5),border=FALSE)
  ## confband(newd$x,lower=ci1[,2],upper=ci1[,3],polygon=TRUE,col=Col("blue",0.5),border=FALSE)
  ## points(y~x,col=c("black","blue")[z+1],pch=16)
  #testthat::expect_true(grcompare(d1,d2,threshold=5))
  l <- lm(y~z)
  val <- expect_doppelganger("plotConf2", {
    val <- plotConf(l,var2="z",var1=NULL,jitter=0,col="black",alpha=0.5,xlim=c(.5,2.5),ylim=range(y))
  })
  ## par(mar=c(0,0,0,0))
  ## plot(y~I(z+1),ylim=range(y),xlim=c(0.5,2.5),pch=16,col=Col("black",0.5))
  ## l0 <- lm(y~-1+factor(z))
  ## confband(1:2,lower=confint(l0)[,1],upper=confint(l0)[,2],lwd=3,
  ##          center=coef(l0))
  })

test_that("forestplot", {
  skip_on_cran()

  set.seed(1)
  K <- 20
  est <- rnorm(K); est[c(3:4,10:12)] <- NA
  se <- runif(K,0.2,0.4)
  x <- cbind(est,est-2*se,est+2*se,runif(K,0.5,2))
  rownames(x) <- unlist(lapply(letters[seq(K)],function(x) paste(rep(x,4),collapse="")))
  rownames(x)[which(is.na(est))] <- ""
  signif <- sign(x[,2])==sign(x[,3])
  val <- expect_doppelganger("forestplot1", {
    val <- forestplot(x)
  })
})

test_that("plot.sim", {
  skip_on_cran()
  skip_on_ci() # Skips the test on GitHub Actions

  onerun2 <- function(a,b,...) {
    return(cbind(a=a,b=b,c=a-1,d=a+1))
  }
  R <- data.frame(a=1:2,b=3:4)
  val2 <- sim(onerun2,R=R,type=0)
  val <- expect_doppelganger("plot.sim-1", {
    val <- plot(val2, alpha=1)
  })
  val <- expect_doppelganger("plot.sim-2", {
    val <- plot(val2,plot.type="single",alpha=1)
  })
  val <- expect_doppelganger("density.sim-1", {
    val <- density(val2, alpha=1)
  })
})

test_that("plot.estimate", {
  skip_on_cran()

  set.seed(1)
  e1 <- estimate(coef=1, IC=1:10, id=1:10)
  e2 <- estimate(coef=1.1, IC=0:9, id=1:10)
  val <- expect_doppelganger("plot.estimate-1", {
     val <- plot(c(e1, e2), null=1)
  })
})

test_that("spaghetti", {
  skip_on_cran()

  K <- 5
  y <- "y"%++%seq(K)
  m <- lvm()
  regression(m,y=y,x=~u) <- 1
  regression(m,y=y,x=~s) <- seq(K)-1
  regression(m,y=y,x=~x) <- "b"
  d <- sim(m,5,seed=1)
  dd <- mets::fast.reshape(d);
  dd$num <- dd$num+rnorm(nrow(dd),sd=0.5) ## Unbalance
  val <- expect_doppelganger("spaghetti-1", {
    val <- spaghetti(y~num,dd,id="id",lty=1,col=Col(1,.4),trend=TRUE,trend.col="darkblue")
  })
})

test_that("plot.lvm", {
  skip_on_cran()

  ## TODO
  m <- lvm(y~1*u[0:1],u~1*x)
  latent(m) <- ~u
  ## expect_doppelganger("plot.lvm-1", {
  ##   plot(m)
  ## })
  ## d <- sim(m,20,seed=1)
  ## e <- estimate(m,d)
  ## expect_doppelganger("plot.lvmfit-1", {
  ## plot(e)
  ## })
  ## expect_doppelganger("plot.lvmfit-2", {
  ##   plot(lava:::beautify(m))
  ## })
  g <- plot(lava:::beautify(m), noplot=TRUE)
  testthat::expect_true(inherits(g, "graphNEL"))
  g <- igraph.lvm(m)
  testthat::expect_true(inherits(g,"igraph"))
})

test_that("ksmooth", {
  skip_on_cran()

  ## TODO
})

test_that("images", {
  skip_on_cran()

  ## TODO
})

test_that("labels,edgelabels", {
  skip_on_cran()

  ## TODO
})

test_that("colorbar", {
  skip_on_cran()

  ## TODO
})

test_that("fplot", {
  skip_on_cran()

  ## TODO
})

test_that("interactive", {
  skip_on_cran()

  ## TODO
})

test_that("pdfconvert", {
  skip_on_cran()

  ## TODO
})

test_that("logo", {
  skip_on_cran()

  val <- expect_doppelganger("logo", {
    lava:::lava(w=10, seed=42)
  })
})

# ---- Merge, subset ------------------------------------------------------

##' @export
merge.estimate <- function(x,y,...,
                           id,
                           paired=FALSE,
                           labels=NULL,
                           keep=NULL,
                           subset=NULL,
                           regex=FALSE,
                           ignore.case=FALSE) {
    if (missing(y)) {
      objects <- c(list(x), list(...))
    } else {
      objects <- c(list(x), list(estimate(y)), list(...))
    }
    trans <- unlist(lapply(
      objects, function(x) !is.null(x[["back.transform"]])
    ))
    if (any(trans)) warning("back-transformation ignored (`back.transform`)")
    if (length(nai <- names(objects)=="NA")>0)
    names(objects)[which(nai)] <- ""
    if (!missing(subset)) {
      if (regex) {
        # TODO
      }
      coefs <- unlist(lapply(objects, function(x) coef(x,messages=0)[subset]))
    } else {
      coefs <- unlist(lapply(objects, function(x) coef(x, messages=0)))
    }
    if (!is.null(labels)) {
      names(coefs) <- labels
    } else {
      names(coefs) <- make.unique(names(coefs))
    }
    if (regex) {
      if (!is.null(keep)) {
        cc <- names(coefs)
        keep <- unlist(lapply(keep, function(x) {
          cc[grepl(x, cc, perl = TRUE, ignore.case=ignore.case)]
        }))
      }
    }
    hasIC <- unlist(lapply(objects, function(x) !is.null(IC(x))))
    if (sum(!hasIC) > 0L) { # Some objects do not have influence function
      ## if (!missing(id) && !is.null(id)) {
      ##   warning(
      ##     "Argument 'id' is only applicable to objects with an influence function. ",
      ##     "It will be used only for objects with IC; ",
      ##     "cross-terms involving objects without IC will be NA."
      ##   )
      ## }
      npar       <- unlist(lapply(objects, function(x) length(coef(x, messages=0))))
      col_ends   <- cumsum(npar)
      col_starts <- col_ends - npar + 1L
      V <- matrix(NA, nrow=sum(npar), ncol=sum(npar))

      ic_idx <- which(hasIC)
      if (length(ic_idx) > 0L) {
        if (length(ic_idx) > 1L) {
          ic_args <- c(objects[ic_idx],
                       list(paired = paired),
                       if (!missing(id)) list(id = id[ic_idx]))
          m_ic  <- do.call(merge, ic_args)
        } else m_ic <- objects[[ic_idx]]
        ic_cols <- unlist(Map(`:`, col_starts[ic_idx], col_ends[ic_idx]))
        V[ic_cols, ic_cols] <- vcov(m_ic)
      }
      # Fill diagonal blocks for non-IC objects
      for (k in which(!hasIC)) {
        pos <- col_starts[k]:col_ends[k]
        V[pos, pos] <- suppressMessages(vcov(objects[[k]]))
      }
      return(estimate(coef=coefs, vcov=V, keep=keep))
    }

    if (!missing(id) && is.null(id)) { ## Independence between datasets in x,y,...
        nn <- unlist(lapply(
          objects,
          function(x) nrow(x$IC)
        ))
        cnn <- c(0, cumsum(nn))
        id <- list()
        for (i in seq_along(nn)) {
          id <- c(id, list(seq(nn[i]) + cnn[i]))
        }
    }
    if (missing(id)) {
      if (paired) { ## One-to-one dependence between observations in x,y,...
        id <- lapply(objects, function(x) {
          seq_len(NROW(x$IC))
        })
        } else {
            id <- lapply(objects, function(x) x$id)
        }
    } else {
        nn <- unlist(lapply(objects,function(x) NROW(IC(x))))
        if (length(id)==1 && is.logical(id)) {
            if (id) {
                if (any(nn[1]!=nn)) stop("Expected objects of the same size: ", paste(nn,collapse=","))
                id0 <- seq(nn[1]); id <- c()
                for (i in seq(length(nn))) id <- c(id,list(id0))
            } else {
                id <- c()
                N <- cumsum(c(0,nn))
                for (i in seq(length(nn))) id <- c(id,list(seq(nn[i])+N[i]))
            }
        }
        if (length(id)!=length(objects)) stop("Same number of id-elements as model objects expected")
        idlen <- unlist(lapply(id,length))
        if (!identical(idlen,nn)) stop("Wrong lengths of 'id': ", paste(idlen,collapse=","), "; ", paste(nn,collapse=","))
    }
    ids <- ic_all <- c(); count <- 0
    for (z in objects) {
        count <- count+1
        clidx <- NULL
        id0 <- id[[count]]
        icz <- IC(z)
        if (is.null(id0)) {
            id0 <- rownames(icz)
            if (is.null(id0)) stop("Need id for object number ", count)
        }
        if (!missing(subset)) icz <- icz[,subset,drop=FALSE]
        if (!lava.options()$cluster.index) {
            ic0 <- matrix(unlist(by(icz,id0,colSums)),byrow=TRUE,ncol=ncol(icz))
            ids <- c(ids, list(sort(unique(id0))))
        } else {
            if (!requireNamespace("mets",quietly=TRUE)) stop("'mets' package required")
            clidx <- mets::cluster.index(id0,mat=icz,return.all=TRUE)
            ic0 <- clidx$X
            ids <- c(ids, list(id0[as.vector(clidx$firstclustid)+1]))
        }
        ic0 <- ic0*NROW(ic0)/length(id0)
        ic_all <- c(ic_all, list(ic0))
    }
    id <- unique(unlist(ids))
    ic0 <- matrix(NA, nrow=length(id),ncol=length(coefs))
    model.index <- c()
    colpos <- 0
    for (i in seq(length(objects))) {
        relpos <- seq_along(coef(objects[[i]], messages=0))
        if (!missing(subset)) relpos <- seq_along(subset)
        ic0[match(ids[[i]], id), relpos + colpos] <- ic_all[[i]]
        ## midx <- objects[[i]]$model.index
        ## if (!is.null(midx)) {
        ##   midx <- lapply(midx, function(x) {
        ##     intersect(x, relpos) + colpos
        ##   })
        ## } else {
          midx <- list(relpos + colpos)
        ## }
        model.index <- c(model.index, midx)
        colpos <- colpos+tail(relpos,1)
    }
    rownames(ic0) <- id
    ## Rescale each column according to I(obs)/pr(obs)
    for (i in seq(NCOL(ic0))) {
      pr <- mean(!is.na(ic0[,i]))
      ic0[,i] <- ic0[,i]/pr
    }
    ic0[is.na(ic0)] <- 0
    res <- estimate.default(
      coef = coefs, stack = FALSE, data = NULL,
      IC = ic0, id = id, keep = keep
      )
    if (is.null(keep)) {
      res$model.index <- model.index
    }
    return(res)
}

##' @export
"%++%.estimate" <- function(x, ...) {
  merge(x, ...)
}

##' @export
"c.estimate" <- function(...) {
  args <- list(...)
  is_estimate <- unlist(lapply(args, function(x)
    inherits(x, c("estimate", "glm", "targeted", "phreg"))))
  if (!all(is_estimate)) { # fallback to default concatenation
    cl <- class(args[[1]])
    class(args[[1]]) <- "list"
    return(do.call(c, args))
  }
  lab <- names(args)
  names(args) <- NULL
  res <- do.call(merge, args)
  newlabels <- names(coef(res))
  if (!is.null(lab)) {
    idx <- which(lab != "")
    newlabels[idx] <- lab[idx]
    return(labels(res, newlabels))
  }
  return(res)
}

##' @export
subset.estimate <- function(x, keep, ...) {
  estimate(x, keep = keep, ...)
}

##' @export
"[.estimate" <- function(x, i, ...) {
  subset(x, i, ...)
}

##' @export
with.estimate <- function(data, expr, ...) {
    # Recursively walk the expression tree and replace symbols
  # that match names in `data` with data["symbol"] calls
  replace_syms <- function(e) {
    # Base case: if it's a symbol, check if it matches a name in data
    if (is.symbol(e)) {
      nm <- as.character(e)
      if (nm %in% names(coef(data))) {
        # Replace symbol with data["nm"] call
        return(call("[", quote(data), nm))
      }
      return(e)
    }
    # Recursive case: walk the call tree
    if (is.call(e)) {
      return(as.call(lapply(e, replace_syms)))
    }
    # Literals (numbers, strings, etc.) — return as-is
    return(e)
  }
  # Substitute and transform the expression
  expr_sub  <- substitute(expr)
  expr_new  <- replace_syms(expr_sub)
  # Create a local environment where `data` exists,
  # with the parent frame as the enclosing environment
  eval_env <- new.env(parent = parent.frame())
  eval_env$data <- data
  # Evaluate in calling environment so non-estimate symbols still resolve
  eval(expr_new, envir = eval_env)
}

# ---- Trigonometric Functions --------------------------------------------

##' @export
sin.estimate <- function(x, ...) {
  estimate(x, function(p) {
    y <- sin(p)
    structure(y, grad = diag(cos(p), nrow = length(p)))
  }, ...)
}

##' @export
cos.estimate <- function(x, ...) {
  estimate(x, function(p) {
    y <- cos(p)
    structure(y, grad = diag(-sin(p), nrow = length(p)))
  }, ...)
}

##' @export
tan.estimate <- function(x, ...) {
  estimate(x, function(p) {
    y <- tan(p)
    structure(y, grad = diag(1 / cos(p)^2, nrow = length(p)))
  }, ...)
}

# ---- Inverse Trigonometric Functions ------------------------------------

##' @export
asin.estimate <- function(x, ...) {
  estimate(x, function(p) {
    y <- asin(p)
    structure(y, grad = diag(1 / sqrt(1 - p^2), nrow = length(p)))
  }, ...)
}

##' @export
acos.estimate <- function(x, ...) {
  estimate(x, function(p) {
    y <- acos(p)
    structure(y, grad = diag(-1 / sqrt(1 - p^2), nrow = length(p)))
  }, ...)
}

##' @export
atan.estimate <- function(x, ...) {
  estimate(x, function(p) {
    y <- atan(p)
    structure(y, grad = diag(1 / (1 + p^2), nrow = length(p)))
  }, ...)
}

# ---- Hyperbolic Functions -----------------------------------------------

##' @export
sinh.estimate <- function(x, ...) {
  estimate(x, function(p) {
    y <- sinh(p)
    structure(y, grad = diag(cosh(p), nrow = length(p)))
  }, ...)
}

##' @export
cosh.estimate <- function(x, ...) {
  estimate(x, function(p) {
    y <- cosh(p)
    structure(y, grad = diag(sinh(p), nrow = length(p)))
  }, ...)
}

##' @export
tanh.estimate <- function(x, ...) {
  estimate(x, function(p) {
    y <- tanh(p)
    structure(y, grad = diag(1 / cosh(p)^2, nrow = length(p)))
  }, ...)
}

# ---- Inverse Hyperbolic Functions ---------------------------------------

##' @export
asinh.estimate <- function(x, ...) {
  estimate(x, function(p) {
    y <- asinh(p)
    structure(y, grad = diag(1 / sqrt(p^2 + 1), nrow = length(p)))
  }, ...)
}

##' @export
acosh.estimate <- function(x, ...) {
  estimate(x, function(p) {
    y <- acosh(p)
    structure(y, grad = diag(1 / sqrt(p^2 - 1), nrow = length(p)))
  }, ...)
}

##' @export
atanh.estimate <- function(x, ...) {
  estimate(x, function(p) {
    y <- atanh(p)
    structure(y, grad = diag(1 / (1 - p^2), nrow = length(p)))
  }, ...)
}

# ---- Other Common Functions ---------------------------------------------

##' @export
log1p.estimate <- function(x, ...) {
  # log(1 + p) — more numerically stable than log(p+1) for small p
  estimate(x, function(p) {
    y <- log1p(p)
    structure(y, grad = diag(1 / (1 + p), nrow = length(p)))
  }, ...)
}

##' @export
expm1.estimate <- function(x, ...) {
  # exp(p) - 1 — more numerically stable than exp(p)-1 for small p
  estimate(x, function(p) {
    y <- expm1(p)
    structure(y, grad = diag(exp(p), nrow = length(p)))
  }, ...)
}

##' @export
log.estimate <- function(x, base = exp(1), ...) {
  estimate(x, function(p) {
    y <- log(p, base = base)
    structure(y,
              grad = diag(1 / (p * log(base)), nrow=length(p)))
  }, ...)
}

##' @export
exp.estimate <- function(x, ...) {
  estimate(x, function(p) {
    y <- exp(p)
    structure(y, grad = diag(y, nrow=length(p)))
  }, ...)
}

##' @export
sqrt.estimate <- function(x, ...) {
  estimate(x^.5, ...)
}

##' @export
sum.estimate <- function(x, ...) {
  estimate(x,
           function(p)
             structure(sum(p),
                       grad = matrix(1, nrow = 1, ncol = length(p))),
           ...)
}

 ##' @export
"%*%.estimate" <- function(x, y, ...) {
  if (is.matrix(x)) {
    return(estimate(y, contrast=x, ...))
  } else if (is.matrix(y)) {
    return(estimate(x, contrast=t(y), ...))
  }
  sum(x * y)
}

prod_except <- function(p) {
  n     <- length(p)
  left  <- c(1, cumprod(p[-n]))
  right <- rev(cumprod(rev(p[-1])))
  right <- c(right, 1)
  left * right
}

##' @export
prod.estimate <- function(x, ...) {
  estimate(x, function(p) {
    y <- prod(p)
    # grad of prod(p) is: d/dp_i = prod(p) / p_i
    structure(y, grad = matrix(prod_except(p), nrow = 1, ncol = length(p)))
  }, ...)
}

# ---- +,-,*,/ ------------------------------------------------------------

operator_estimate <- function(x, y, op, ...) {
  x_const <- is.numeric(x)
  y_const <- is.numeric(y)
  np1 <- ifelse(x_const, length(x), length(coef(x)))
  np2 <- ifelse(y_const, length(y), length(coef(y)))

  if (np1 != 1L && np2 != 1L && np1 != np2) {
    stop("expecting equal length objects or one of them to be a scalar")
  }
  if (y_const || x_const) { # x estimate, y numeric
    e <- if (y_const) x else y
    return(
      estimate(e, function(p) {
        if (y_const) {
          op(p, y, y_const=TRUE)
        } else {
          op(x, p, x_const=TRUE)
        }
      }, ...)
    )
  }
  e <- merge(x, y) # x estimate, y estimate
  estimate(e, function(p) {
    p1 <- p[seq_len(np1)]
    p2 <- p[seq_len(np2)+np1]
    res <- op(p1, p2)
    return(res)
  }, ...)
}

operator_grad <- function(x, y, x_const, y_const, dx, dy) {
  nx <- length(x)
  ny <- length(y)
  n <- max(nx, ny)
  grad <-
    if (y_const) { # y is constant: df/dx only
      if (nx > 1L || ny == 1L) { # ny = nx or ny = 1
        diag(dx, ncol=nx, nrow=nx)
      } else { # ny > 1, nx = 1
        matrix(dx, nrow = n, ncol = 1L)
      }
    } else if (x_const) { # x is constant: df/dy only
      if (ny > 1L || nx == 1L) { # ny = nx or nx = 1
        diag(dy, ncol=ny, nrow=ny)
      } else { # ny > 1, nx = 1
        matrix(dy, nrow = n, ncol = 1L)
      }
    } else {
      # both estimates: df/d(c(x,y)) = [dx | dy]
      D <- matrix(0, nrow = n, ncol = nx + ny)
      if (nx == 1L) {
        D[, 1] <- dx
        D[, 2:ncol(D)] <- diag(dy, ny, ny)
        D
      } else if (ny == 1L) {
        cbind(diag(dx, nrow=n, ncol=n), dy)
      } else {
        cbind(diag(dx, nrow = n, ncol=n), diag(dy, nrow = n, ncol=n))
      }
    }
  return(grad)
}

##' @export
"+.estimate" <- function(e1, e2, ...) {
  operator_estimate(
    e1, e2,
    function(x, y, x_const=FALSE, y_const=FALSE) {
      structure(
        x + y,
        grad = operator_grad(x, y, x_const, y_const,
                             dx = 1, dy = 1)
      )
      }, ...)
}

##' @export
"-.estimate" <- function(e1, e2, ...) {
  if (missing(e2)) return(-1*e1)
  operator_estimate(
    e1, e2,
    function(x, y, x_const=FALSE, y_const=FALSE) {
      structure(
        x - y,
        grad = operator_grad(x, y, x_const, y_const,
                             dx=1, dy=-1)
      )
      }, ...)
}

##' @export
"*.estimate" <- function(e1, e2, ...) {
  operator_estimate(
    e1, e2,
    function(x, y, x_const=FALSE, y_const=FALSE) {
      structure(
        x * y,
        grad = operator_grad(x, y, x_const, y_const,
                             dx=y, dy=x)
      )
      }, ...)
}

##' @export
"/.estimate" <- function(e1, e2, ...) {
  operator_estimate(
    e1, e2,
    function(x, y, x_const=FALSE, y_const=FALSE) {
      structure(
        x / y,
        grad = operator_grad(x, y, x_const, y_const,
                             dx = 1 / y, dy = -x / y^2)
      )
      }, ...)
}

##' @export
"^.estimate" <- function(e1, e2, ...) {
  operator_estimate(
    e1, e2,
    function(x, y, x_const=FALSE, y_const=FALSE) {
      nx <- length(x)
      ny <- length(y)
      n <- max(nx, ny)
      f <- x^y
      structure(
        f,
        grad = operator_grad(x, y, x_const, y_const,
                             dx = y*x^(y-1), dy = log(x)*f)
      )
    }, ...)
}

# ---- == ---- ------------------------------------------------------------

##' @export
"==.estimate" <- function(e1, e2) {
  if (!(is.numeric(e1) || is.numeric(e2))) stop("numeric comperator needed")
  null <- if (is.numeric(e1)) e1 else e2
  e <- if (is.numeric(e1)) e2 else e1
  estimate(e, null=null)
}

##' Simulate model
##'
##' Simulate data from a general SEM model including non-linear effects and
##' general link and distribution of variables.
##'
##' @aliases sim.lvmfit sim.lvm
##' simulate.lvmfit simulate.lvm
##' transform<- transform<-.lvm transform.lvm
##' functional functional<-  functional.lvm functional<-.lvm
##' distribution distribution distribution<- distribution.lvm distribution<-.lvm
##' heavytail heavytail<-
##' weibull.lvm
##' binomial.lvm
##' poisson.lvm
##' uniform.lvm
##' multinomial.lvm
##' beta.lvm
##' normal.lvm mvn.lvm
##' lognormal.lvm
##' gaussian.lvm
##' GM2.lvm
##' GM3.lvm
##' probit.lvm
##' logit.lvm
##' pareto.lvm
##' student.lvm
##' chisq.lvm
##' coxGompertz.lvm
##' coxWeibull.lvm
##' coxExponential.lvm
##' aalenExponential.lvm
##' Gamma.lvm gamma.lvm
##' loggamma.lvm
##' categorical categorical<-
##' threshold.lvm
##' ones.lvm Binary.lvm binary.lvm
##' Sequence.lvm
##' none.lvm constant.lvm id.lvm
##' @usage
##' \method{sim}{lvm}(x, n = NULL, p = NULL, normal = FALSE, cond = FALSE,
##' sigma = 1, rho = 0.5, X = NULL, unlink=FALSE, latent=TRUE,
##' use.labels = TRUE, seed=NULL, ...)
##' @param x Model object
##' @param n Number of simulated values/individuals
##' @param p Parameter value (optional)
##' @param normal Logical indicating whether to simulate data from a
##' multivariate normal distribution conditional on exogenous variables hence
##' ignoring functional/distribution definition
##' @param cond for internal use
##' @param sigma Default residual variance (1)
##' @param rho Default covariance parameter (0.5)
##' @param X Optional matrix of fixed values of variables (manipulation)
##' @param unlink Return Inverse link transformed data
##' @param latent Include latent variables (default TRUE)
##' @param use.labels convert categorical variables to factors before applying transformation
##' @param seed Random seed
##' @param \dots Additional arguments to be passed to the low level functions
##' @author Klaus K. Holst
##' @keywords models datagen regression
##' @export
##' @examples
##' ##################################################
##' ## Logistic regression
##' ##################################################
##' m <- lvm(y~x+z)
##' regression(m) <- x~z
##' distribution(m,~y+z) <- binomial.lvm("logit")
##' d <- sim(m,1e3)
##' head(d)
##'
##' e <- estimate(m,d,estimator="glm")
##' e
##' ## Simulate a few observation from estimated model
##' sim(e,n=5)
##'
##' ##################################################
##' ## Poisson
##' ##################################################
##' distribution(m,~y) <- poisson.lvm()
##' d <- sim(m,1e4,p=c(y=-1,"y~x"=2,z=1))
##' head(d)
##' estimate(m,d,estimator="glm")
##' mean(d$z); lava:::expit(1)
##'
##' summary(lm(y~x,sim(lvm(y[1:2]~4*x),1e3)))
##'
##' ##################################################
##' ### Gamma distribution
##' ##################################################
##' m <- lvm(y~x)
##' distribution(m,~y+x) <- list(Gamma.lvm(shape=2),binomial.lvm())
##' intercept(m,~y) <- 0.5
##' d <- sim(m,1e4)
##' summary(g <- glm(y~x,family=Gamma(),data=d))
##' \dontrun{MASS::gamma.shape(g)}
##'
##' args(lava::Gamma.lvm)
##' distribution(m,~y) <- Gamma.lvm(shape=2,log=TRUE)
##' sim(m,10,p=c(y=0.5))[,"y"]
##'
##' ##################################################
##' ### Beta
##' ##################################################
##' m <- lvm()
##' distribution(m,~y) <- beta.lvm(alpha=2,beta=1)
##' var(sim(m,100,"y,y"=2))
##' distribution(m,~y) <- beta.lvm(alpha=2,beta=1,scale=FALSE)
##' var(sim(m,100))
##'
##' ##################################################
##' ### Transform
##' ##################################################
##' m <- lvm()
##' transform(m,xz~x+z) <- function(x) x[1]*(x[2]>0)
##' regression(m) <- y~x+z+xz
##' d <- sim(m,1e3)
##' summary(lm(y~x+z + x*I(z>0),d))
##'
##' ##################################################
##' ### Non-random variables
##' ##################################################
##' m <- lvm()
##' distribution(m,~x+z+v+w) <- list(Sequence.lvm(0,5),## Seq. 0 to 5 by 1/n
##'                                Binary.lvm(),       ## Vector of ones
##'                                Binary.lvm(0.5),    ##  0.5n 0, 0.5n 1
##'                                Binary.lvm(interval=list(c(0.3,0.5),c(0.8,1))))
##' sim(m,10)
##'
##' ##################################################
##' ### Cox model
##' ### piecewise constant hazard
##' ################################################
##' m <- lvm(t~x)
##' rates <- c(1,0.5); cuts <- c(0,5)
##' ## Constant rate: 1 in [0,5), 0.5 in [5,Inf)
##' distribution(m,~t) <- coxExponential.lvm(rate=rates,timecut=cuts)
##'
##' \dontrun{
##'     d <- sim(m,2e4,p=c("t~x"=0.1)); d$status <- TRUE
##'     plot(timereg::aalen(survival::Surv(t,status)~x,data=d,
##'                         resample.iid=0,robust=0),spec=1)
##'     L <- approxfun(c(cuts,max(d$t)),f=1,
##'                    cumsum(c(0,rates*diff(c(cuts,max(d$t))))),
##'                    method="linear")
##'     curve(L,0,100,add=TRUE,col="blue")
##' }
##'
##' ##################################################
##' ### Cox model
##' ### piecewise constant hazard, gamma frailty
##' ##################################################
##' m <- lvm(y~x+z)
##' rates <- c(0.3,0.5); cuts <- c(0,5)
##' distribution(m,~y+z) <- list(coxExponential.lvm(rate=rates,timecut=cuts),
##'                              loggamma.lvm(rate=1,shape=1))
##' \dontrun{
##'     d <- sim(m,2e4,p=c("y~x"=0,"y~z"=0)); d$status <- TRUE
##'     plot(timereg::aalen(survival::Surv(y,status)~x,data=d,
##'                         resample.iid=0,robust=0),spec=1)
##'     L <- approxfun(c(cuts,max(d$y)),f=1,
##'                    cumsum(c(0,rates*diff(c(cuts,max(d$y))))),
##'                    method="linear")
##'     curve(L,0,100,add=TRUE,col="blue")
##' }
##' ## Equivalent via transform (here with Aalens additive hazard model)
##' m <- lvm(y~x)
##' distribution(m,~y) <- aalenExponential.lvm(rate=rates,timecut=cuts)
##' distribution(m,~z) <- Gamma.lvm(rate=1,shape=1)
##' transform(m,t~y+z) <- prod
##' sim(m,10)
##' ## Shared frailty
##' m <- lvm(c(t1,t2)~x+z)
##' rates <- c(1,0.5); cuts <- c(0,5)
##' distribution(m,~y) <- aalenExponential.lvm(rate=rates,timecut=cuts)
##' distribution(m,~z) <- loggamma.lvm(rate=1,shape=1)
##' \dontrun{
##' mets::fast.reshape(sim(m,100),varying="t")
##' }
##'
##' ##################################################
##' ### General multivariate distributions
##' ##################################################
##' \dontrun{
##' m <- lvm()
##' distribution(m,~y1+y2,oratio=4) <- VGAM::rbiplackcop
##' ksmooth2(sim(m,1e4),rgl=FALSE,theta=-20,phi=25)
##'
##' m <- lvm()
##' distribution(m,~z1+z2,"or1") <- VGAM::rbiplackcop
##' distribution(m,~y1+y2,"or2") <- VGAM::rbiplackcop
##' sim(m,10,p=c(or1=0.1,or2=4))
##' }
##'
##' m <- lvm()
##' distribution(m,~y1+y2+y3,TRUE) <- function(n,...) rmvn0(n,sigma=diag(3)+1)
##' var(sim(m,100))
##'
##' ## Syntax also useful for univariate generators, e.g.
##' m <- lvm(y~x+z)
##' distribution(m,~y,TRUE) <- function(n) rnorm(n,mean=1000)
##' sim(m,5)
##' distribution(m,~y,"m1",0) <- rnorm
##' sim(m,5)
##' sim(m,5,p=c(m1=100))
##'
##' ##################################################
##' ### Regression design in other parameters
##' ##################################################
##' ## Variance heterogeneity
##' m <- lvm(y~x)
##' distribution(m,~y) <- function(n,mean,x) rnorm(n,mean,exp(x)^.5)
##' if (interactive()) plot(y~x,sim(m,1e3))
##' ## Alternaively, calculate the standard error directly
##' addvar(m) <- ~sd ## If 'sd' should be part of the resulting data.frame
##' constrain(m,sd~x) <- function(x) exp(x)^.5
##' distribution(m,~y) <- function(n,mean,sd) rnorm(n,mean,sd)
##' if (interactive()) plot(y~x,sim(m,1e3))
##'
##' ## Regression on variance parameter
##' m <- lvm()
##' regression(m) <- y~x
##' regression(m) <- v~x
##' ##distribution(m,~v) <- 0 # No stochastic term
##' ## Alternative:
##' ## regression(m) <- v[NA:0]~x
##' distribution(m,~y) <- function(n,mean,v) rnorm(n,mean,exp(v)^.5)
##' if (interactive()) plot(y~x,sim(m,1e3))
##'
##' ## Regression on shape parameter in Weibull model
##' m <- lvm()
##' regression(m) <- y ~ z+v
##' regression(m) <- s ~ exp(0.6*x-0.5*z)
##' distribution(m,~x+z) <- binomial.lvm()
##' distribution(m,~cens) <- coxWeibull.lvm(scale=1)
##' distribution(m,~y) <- coxWeibull.lvm(scale=0.1,shape=~s)
##' eventTime(m) <- time ~ min(y=1,cens=0)
##'
##' if (interactive()) {
##'     d <- sim(m,1e3)
##'     require(survival)
##'     (cc <- coxph(Surv(time,status)~v+strata(x,z),data=d))
##'     plot(survfit(cc) ,col=1:4,mark.time=FALSE)
##' }
##'
##' ##################################################
##' ### Categorical predictor
##' ##################################################
##' m <- lvm()
##' ## categorical(m,K=3) <- "v"
##' categorical(m,labels=c("A","B","C")) <- "v"
##'
##' regression(m,additive=FALSE) <- y~v
##' \dontrun{
##' plot(y~v,sim(m,1000,p=c("y~v:2"=3)))
##' }
##'
##' m <- lvm()
##' categorical(m,labels=c("A","B","C"),p=c(0.5,0.3)) <- "v"
##' regression(m,additive=FALSE,beta=c(0,2,-1)) <- y~v
##' ## equivalent to:
##' ## regression(m,y~v,additive=FALSE) <- c(0,2,-1)
##' regression(m,additive=FALSE,beta=c(0,4,-1)) <- z~v
##' table(sim(m,1e4)$v)
##' glm(y~v, data=sim(m,1e4))
##' glm(y~v, data=sim(m,1e4,p=c("y~v:1"=3)))
##'
##' transform(m,v2~v) <- function(x) x=='A'
##' sim(m,10)
##'
##' ##################################################
##' ### Pre-calculate object
##' ##################################################
##' m <- lvm(y~x)
##' m2 <- sim(m,'y~x'=2)
##' sim(m,10,'y~x'=2)
##' sim(m2,10) ## Faster
##' @export
sim.lvm <- function(x,n=NULL,p=NULL,normal=FALSE,cond=FALSE,sigma=1,rho=.5,
            X=NULL,unlink=FALSE,latent=TRUE,use.labels=TRUE,seed=NULL,...) {

    if (!exists(".Random.seed", envir = .GlobalEnv, inherits = FALSE))
        runif(1)
    if (is.null(seed))
        RNGstate <- get(".Random.seed", envir = .GlobalEnv)
    else {
        R.seed <- get(".Random.seed", envir = .GlobalEnv)
        set.seed(seed)
        RNGstate <- structure(seed, kind = as.list(RNGkind()))
        on.exit(assign(".Random.seed", R.seed, envir = .GlobalEnv))
    }


    v.env <- c("A","M","P","PP","PPdiag","xx","vv","mdist","mdistnam","mii",
              "nn","mu","xf","xfix","X",
              "vartrans","multitrans","multitrans.idx",
              "X.idx","ii.mvn","xconstrain.idx","xconstrain",
              "xconstrain.par","covparnames","exo_constrainY")

    setup <- is.null(n) && is.null(X) ## Save environment (variables v.env) and return sim object
    loadconfig <- !is.null(x$sim.env) && !setup && (length(list(...))==0 && length(p)==0)
    Yfix <- NULL

    if (loadconfig) {
        for (v in setdiff(v.env,"X")) assign(v, x$sim.env[[v]])
        if (is.null(X)) X <-  x$sim.env[['X']]
    } else {
        if (!is.null(n) && n<1) return(NULL)
        p <- c(p,unlist(list(...)))
        xx <- exogenous(x)

        if (!is.null(X)) {
            if (is.null(n))
                n <- nrow(X)
            if (!is.null(colnames(X))) {
                yfix <- setdiff(colnames(X),xx)
                if (length(yfix)>0) Yfix <- X[,yfix,drop=FALSE]
                xx0 <- intersect(xx,colnames(X))
                if (length(xx0)>0)
                    X <- as.matrix(X[,xx0,drop=FALSE])
            }
        } else {
            if (!is.null(p)) {
                i1 <- unique(na.omit(c(match(names(p),xx),
                                       match(names(p),paste0(xx,lava.options()$symbol[2],xx)))))
                covariance(x) <- xx[i1]
            }
        }
        ##  index(x) <- reindex(x)
        vv <- vars(x)
        nn <- setdiff(vv,parameter(x))
        mu <- unlist(lapply(x$mean, function(l) ifelse(is.na(l)|is.character(l),0,l)))
        xf <- intersect(unique(parlabels(x)),xx)
        xfix <- c(randomslope(x),xf); if (length(xfix)>0) normal <- FALSE
        ## Match parameter names
        if ((!is.null(names(p)) && all(!is.na(names(p)))) || length(p)!=(index(x)$npar+index(x)$npar.mean+index(x)$npar.ex) | is.null(names(p))) {
            nullp <- is.null(p)
            p0 <- p
            ep <- NULL
            ei <- which(index(x)$e1==1)
            if (length(ei)>0)
                ep <- unlist(x$expar)[ei]
            p <- c(rep(1, index(x)$npar+index(x)$npar.mean),ep)
            p[seq_len(index(x)$npar.mean)] <- 0
            p[index(x)$npar.mean + variances(x)] <- sigma
            p[index(x)$npar.mean + offdiags(x)] <- rho
            if (!nullp) {
                c1 <- coef(x,mean=TRUE,fix=FALSE)
                c2 <- coef(x,mean=TRUE,fix=FALSE,labels=TRUE)
                idx1 <- na.omit(match(names(p0),c1))
                idx11 <- na.omit(match(names(p0),c2))
                idx2 <- na.omit(which(names(p0)%in%c1))
                idx22 <- na.omit(which(names(p0)%in%c2))
                if (length(idx1)>0 && !any(is.na(idx1)))
                    p[idx1] <- p0[idx2]
                if (length(idx11)>0 && !any(is.na(idx11)))
                    p[idx11] <- p0[idx22]
            }
        }
        M <- modelVar(x,p,data=NULL)
        A <- M$A; P <- M$P
        if (!is.null(M$v)) mu <- M$v

        ## Square root of residual variance matrix
        PP <- with(svd(P), v%*%diag(sqrt(d),nrow=length(d))%*%t(u))
        ## Multivariate distributions
        mdist <- distribution(x,multivariate=TRUE)$var
        mdistnam <- names(mdist)
        mii <- match(mdistnam,vars(x))
        if (length(distribution(x))>0 ) {
            ii <- match(names(distribution(x)),vv)
            ii.mvn <- setdiff(seq(ncol(P)),c(ii,mii))
        } else {
            ii.mvn <- seq(ncol(P))
        }
        PPdiag <- sum(abs(offdiag(PP[ii.mvn,ii.mvn,drop=FALSE])^2))<1e-20
    }

    if (!setup) {
        E <- matrix(0,ncol=ncol(P),nrow=n)
        if (length(ii.mvn)>0) {
            ## Error term for conditional normal distributed variables
            if (PPdiag) {
                for (i in ii.mvn) E[,i] <- rnorm(n,sd=PP[i,i])
            } else {
                E[,ii.mvn] <-  matrix(rnorm(length(ii.mvn)*n),ncol=length(ii.mvn))%*%PP[ii.mvn,ii.mvn,drop=FALSE]
            }
        }

        if (length(mdistnam)>0) {
            fun <- distribution(x,multivariate=TRUE)$fun
            for (i in seq_along(fun)) {
                mv <- names(which(unlist(mdist)==i))
                ii <- match(mv,vv)
                E[,ii] <- distribution(x,multivariate=TRUE)$fun[[i]](n,p=p,object=x) # ,...)
            }
        }

        ## Simulate exogenous variables (covariates)
        res <- matrix(0,ncol=length(nn),nrow=n)
        colnames(res) <- nn
    }

    if (!loadconfig) {
        vartrans <- names(x$attributes$transform)
        multitrans <- multitrans.idx <- NULL
        if (length(x$attributes$multitransform)>0) {
            multitrans <- unlist(lapply(x$attributes$multitransform,function(z) z$y))
            for (i in (seq_along(x$attributes$multitransform))) {
                multitrans.idx <- c(multitrans.idx,rep(i,length(x$attributes$multitransform[[i]]$y)))
            }
        }
        xx <- unique(c(exogenous(x, latent=FALSE, index=TRUE),xfix))
        xx <- setdiff(xx,vartrans)

        X.idx <- match(xx,vv)
    }

    if (!setup) {
        res[,X.idx] <- t(mu[X.idx]+t(E[,X.idx]))
        if (is.null(X) || NCOL(X)<length(xx)) {
            if (!is.null(xx) && length(xx)>0)
                for (i in seq_along(xx)) {
                    mu.x <- mu[X.idx[i]]
                    dist.x <- distribution(x,xx[i])[[1]]
                    if (is.list(dist.x) && is.function(dist.x[[1]])) dist.x <- dist.x[[1]]
                    if (is.list(dist.x)) {
                        dist.x <- dist.x[[1]]
                        if (length(dist.x)==1) dist.x <- rep(dist.x,n)
                    }
                    if (is.function(dist.x)) {
                        res[,X.idx[i]] <- dist.x(n=n,mu=mu.x,var=P[X.idx[i],X.idx[i]])
                    } else {
                        if (is.null(dist.x) || any(is.na(dist.x))) {
                        } else {
                            if (length(dist.x)!=n) stop("'",vv[X.idx[i]], "' fixed at length ", length(dist.x)," != ",n,".")
                            res[,X.idx[i]] <- dist.x ## Deterministic
                        }
                    }
                }
        }
        if (!is.null(X)) {
            ii <- match(colnames(X),vv)
            res0 <- res
            for (i in seq(ncol(X)))
                res[,ii[i]] <- X[,i]
        }
    }

    simuled <- c(xx)
    resunlink <- NULL
    if (unlink) {
        resunlink <- res
    }

    if ( normal | ( is.null(distribution(x)) & is.null(functional(x)) & is.null(constrain(x))) ) {
        if(cond) { ## Simulate from conditional distribution of Y given X
            mypar <- pars(x,A,P,mu)
            Ey.x <- predict(x, mypar, data.frame(res))
            Vy.x <- attributes(Ey.x)$cond.var
            PP <- with(svd(Vy.x), v%*%diag(sqrt(d),nrow=length(d))%*%t(u))
            yy <- Ey.x + matrix(n*ncol(Vy.x),ncol=ncol(Vy.x))%*%PP
            res <- cbind(yy, res[,xx]); colnames(res) <- c(colnames(Vy.x),xx)
            return(res)
        }
        ## Simulate from sim. distribution (Y,X) (mv-normal)
        I <- diag(nrow=length(nn))
        IAi <- Inverse(I-t(A))
        colnames(E) <- vv
        dd <- t(apply(heavytail_sim_hook(x,E),1,function(x) x+mu))
        res <- dd%*%t(IAi)
        colnames(res) <- vv
    } else {

        if (!loadconfig) {
            xc <- index(x)$vars
            xconstrain.idx <- unlist(lapply(lapply(constrain(x),function(z) attributes(z)$args),function(z) length(intersect(z,xc))>0))
            xconstrain <- intersect(unlist(lapply(constrain(x),function(z) attributes(z)$args)),xc)
            xconstrain.par <- names(xconstrain.idx)[xconstrain.idx]
            covparnames <- unique(as.vector(covariance(x)$labels))
            exo_constrainY <- intersect(exogenous(x),names(x$constrainY))
        }

        if (setup) {
            sim.env <- c()
            sim.env[v.env] <- list(NULL)
            for (v in v.env) if (!is.null(get(v))) sim.env[[v]] <- get(v)
            x$sim.env <- sim.env
            return(x)
        }

      if (length(xconstrain)>0) {
        for (i in which(xconstrain.idx)) {
          if (!(names(xconstrain.idx[i]) %in% nn)) { ## A parameter and not a variable
            ff <- constrain(x)[[i]]
            myargs <- attributes(ff)$args
            D <- matrix(0,n,length(myargs))
            for (j in seq_len(ncol(D))) {
              if (myargs[j]%in%xconstrain)
                D[,j] <- res[,myargs[j]]
              else
                D[,j] <- M$parval[[myargs[j]]]
            }
            val <- try(val <- ff(D), silent=TRUE)
            if (inherits(val,"try-error") || NROW(val)<n) apply(D,1,ff)
            res <- cbind(res, val); colnames(res)[ncol(res)] <- names(xconstrain.idx)[i]
          }
        }
      }

        if (any(xconstrain.par%in%covparnames)) {
            mu0 <- rep(0,ncol(P))
            P0 <- P
            E <- t(sapply(seq_len(n),function(idx) {
              for (i in intersect(xconstrain.par,covparnames)) {
                  P0[covariance(x)$labels==i] <- res[idx,i]
                }
                PP <- with(svd(P0), v%*%diag(sqrt(d),nrow=length(d))%*%t(u))
                return(mu0+rbind(rnorm(ncol(P0)))%*%PP)
            }))
        }

        colnames(E) <- vv
        E <- heavytail_sim_hook(x,E)

        ## Non-linear regression components
        xconstrain <- c()
        for (i in seq_len(length(constrain(x)))) {
            z <- constrain(x)[[i]]
            xx <- intersect(attributes(z)$args,manifest(x))
            if (length(xx)>0) {
                warg <- setdiff(attributes(z)$args,xx)
                wargidx <- which(attributes(z)$args%in%warg)
                exoidx <- which(attributes(z)$args%in%xx)
                parname <- names(constrain(x))[i]
                y <- names(which(unlist(lapply(intercept(x),function(x) x==parname))))
                el <- list(i,y,parname,xx,exoidx,warg,wargidx,z)
                names(el) <- c("idx","endo","parname","exo","exoidx","warg","wargidx","func")
                xconstrain <- c(xconstrain,list(el))
            }
        }
        yconstrain <- unlist(lapply(xconstrain,function(x) x$endo))
      for (i in exo_constrainY) {
            cc <- x$constrainY[[i]]
            args <- cc$args
            args <- if (is.null(args) || length(args)==0) res[,i] else res[,args]
            res[,i] <- cc$fun(args,p) # ,...)
        }

        res <- data.frame(res)
        if (length(vartrans)>0) {
            parvals <- parpos(x)$parval
            parvalsnam <- setdiff(names(parvals),xx)
            if (length(parvalsnam)>0) {
                Parvals <- p[unlist(parvals)];
                res <- cbind(res,
                             cbind(rep(1,nrow(res)))%x%rbind(Parvals))
                colnames(res)[seq(length(Parvals))+ncol(res)-length(Parvals)] <-
                    names(parvals)
            }
        }

        leftovers <- c()
        itercount <- 0
        while (length(simuled)<length(nn)) {
            leftoversPrev <- leftovers
            leftovers <- setdiff(nn,simuled)
            if (!is.null(leftoversPrev) && length(leftoversPrev)==length(leftovers)) {
                if (itercount>0)
                stop("Infinite loop (feedback).")
                itercount <- itercount+1
            }
            for (i in leftovers) {
                if (length(Yfix)>0 && i %in% colnames(Yfix)) {
                    if (NROW(Yfix) == 1) {
                        res[,i] <- rep(Yfix[,i],length.out=nrow(res))
                    } else                     
                        res[,i] <- rep(Yfix[,i,drop=TRUE],length.out=nrow(res))
                    simuled <- c(simuled,i)
                    next
                }
                if (i%in%vartrans) {
                    xtrans <- x$attributes$transform[[i]]$x
                    if (all(xtrans%in%c(simuled,names(parvals))))  {
                        xtr <- res[,xtrans,drop=FALSE]
                        if (use.labels) {
                            lb <- x$attributes$labels
                            lb.idx <- na.omit(match(names(lb),xtrans))
                            ## For categorical variables turn them into factors so we can
                            ## use the actual labels in function calls/transform
                            if (length(lb.idx)>0) {
                                xtr <- as.data.frame(xtr)
                                for (lb0 in lb.idx) {
                                    lab <- lb[[names(xtr)[lb0]]]
                                    xtr[,lb0] <- factor(xtr[,lb0],levels=seq_along(lab)-1,labels=lab)
                                }
                            }
                        }
                        suppressWarnings(yy <- with(x$attributes$transform[[i]], fun(xtr))) ##fun(res[,xtrans])))
                        if (NROW(yy) != NROW(res)) { ## apply row-wise
                            res[,i] <- with(x$attributes$transform[[i]], ##apply(res[,xtrans,drop=FALSE],1,fun))
                                           apply(xtr,1,fun))
                        } else {
                            colnames(yy) <- NULL
                            res[,i] <- yy
                        }
                        simuled <- c(simuled,i)
                    }
                } else if (i%in%multitrans) {
                    idx0 <- match(i,multitrans)
                    idx <- multitrans.idx[idx0]
                    mtval <- x$attributes$multitransform[[idx]]
                    if (all(mtval$x%in%simuled)) {
                        res[,mtval$y] <- mtval$fun(res[,mtval$x])
                        simuled <- c(simuled,mtval$y)
                        break;
                    }
                } else {
                    ipos <- which(yconstrain%in%i)
                    if (length(ipos)==0 || all(xconstrain[[ipos]]$exo%in%simuled)) {
                        pos <- match(i,vv)
                        relations <- colnames(A)[A[,pos]!=0]
                        simvars <- x$attributes$simvar[[i]]
                        dist.i <- distribution(x,i)[[1]] ## User-specified distribution function
                        dist.xx <- NULL
                        if (is.function(dist.i)) {
                            dist.args0 <- names(formals(dist.i))
                            dist.args <- setdiff(dist.args0,c("n","mean","mu","var","..."))
                            dist.xx <- intersect(names(res),dist.args) ## Variables influencing distribution
                        }
                        if (all(c(relations,simvars,dist.xx)%in%simuled)) { ## Only depending on already simulated variables
                            if (x$mean[[pos]]%in%xconstrain.par && length(ipos)==0) {
                                mu.i <- res[,x$mean[[pos]] ]
                            } else {
                                mu.i <- mu[pos]
                            }
                            if (length(ipos)>0) {
                                pp <- unlist(M$parval[xconstrain[[ipos]]$warg])
                                myidx <- with(xconstrain[[ipos]],order(c(wargidx,exoidx)))
                                ## myidx <- with(xconstrain[[ipos]],
                                ##              match(attr(func,"args"), c(warg,exo)))
                                X <- with(xconstrain[[ipos]],
                                          if (length(pp)>0)
                                              cbind(rbind(pp)%x%cbind(rep(1,nrow(res))),
                                                    res[,exo,drop=FALSE])
                                          else res[,exo,drop=FALSE])
                                yy <- try(with(xconstrain[[ipos]],
                                               func(X[,myidx,drop=FALSE])),silent=TRUE)
                                if (NROW(yy) != NROW(res)) { ## apply row-wise
                                    mu.i <- #mu.i +
                                        with(xconstrain[[ipos]],
                                             apply(res[,exo,drop=FALSE],1,
                                                   function(x) func(
                                                            unlist(c(pp,x))[myidx])))
                                } else {
                                    mu.i <- ##mu.i+
                                        yy
                                }
                            }
                            for (From in relations) {
                                f <- functional(x,i,From)[[1]]
                                if (!is.function(f))
                                    f <- function(x,...) x
                                reglab <- regfix(x)$labels[From,pos]
                                if (reglab%in%c(xfix,xconstrain.par)) {
                                    if (is.function(f)) {
                                        if (length(formals(f))>1) {
                                            mu.i <- mu.i + res[,reglab]*f(res[,From],p)
                                        } else {
                                            mu.i <- mu.i + res[,reglab]*f(res[,From])
                                        }
                                    } else  mu.i <- mu.i + res[,reglab]*res[,From]
                                }
                                else {
                                    if (is.function(f)) {
                                        if (length(formals(f))>1) {
                                            mu.i <- mu.i + A[From,pos]*f(res[,From],p)
                                        } else {
                                            mu.i <- mu.i + A[From,pos]*f(res[,From])
                                        }
                                    } else  mu.i <- mu.i + A[From,pos]*res[,From]
                                }
                            }
                            if (!is.function(dist.i)) {
                                res[,pos] <- mu.i + E[,pos]
                                if (unlink)
                                    resunlink[,pos] <- res[,pos]
                            }
                            else {
                                if (length(simvars)>0) { ## Depends on mu and also on other variables (e.g. time-depending effect)
                                    if (length(mu.i)==1) mu.i <- rep(mu.i,n)
                                    mu.i <- cbind("m0"=mu.i,res[,simvars,drop=FALSE])
                                }
                                new.args <- list(n=n)
                                mu.arg <- intersect(c("mean","mu"),dist.args0)
                                if (length(mu.arg)>0) {
                                    new.args <- c(new.args,list(mu.i))
                                    names(new.args)[length(new.args)] <- mu.arg[1]
                                }
                                var.arg <- intersect(c("var"),dist.args0)
                                if (length(var.arg)>0) {
                                    new.args <- c(new.args,list(P[pos,pos]))
                                    names(new.args)[length(new.args)] <- var.arg[1]
                                }
                                for (jj in dist.xx) {
                                    new.args <- c(new.args,list(res[,jj,drop=TRUE]))
                                    names(new.args)[length(new.args)] <- jj
                                }
                                res[,pos] <- do.call(dist.i,new.args)
                                if (unlink)
                                    resunlink[,pos] <- mu.i
                            }

                            if (length(x$constrainY)>0 && i%in%names(x$constrainY)) {
                                cc <- x$constrainY[[i]]
                                args <- cc$args
                                args <- if (length(args)==0)
                                           res[,pos]
                                       else {
                                           ii <- intersect(names(M$parval),args)
                                           args0 <- args
                                           args <- res[,intersect(args0,colnames(res)),drop=FALSE]

                                           if (length(ii)>0) {
                                               pp <- rbind(unlist(M$parval[ii]))%x%cbind(rep(1,n))
                                               colnames(pp) <- ii
                                               args <- cbind(res,pp)[,args0,drop=FALSE]
                                           }
                                           args
                                       }
                                res[,pos] <- cc$fun(args,p) # ,...)
                            }
                            simuled <- c(simuled,i)
                        }
                    }
                }
            }
        }
        res <- res[,nn,drop=FALSE]
    }

    res <- as.data.frame(res)
    myhooks <- gethook("sim.hooks")
    for (f in myhooks) {
        res <- do.call(f, list(x=x,data=res,p=p,modelpar=M))
    }
    if (unlink) res <- resunlink

    res <- as.data.frame(res)
    self <- x$attributes$selftransform
    for (v in names(self)) {
        res[,v] <- self[[v]](res[,v])
    }
    if (!latent && length(latent(x))>0) return(subset(res[,-which(colnames(res)%in%latent(x))]))
    return(res)
}

##' @export
sim.lvmfit <- function(x, n = nrow(model.frame(x)), p = pars(x), xfix = TRUE, ...) {
  m <- Model(x)
  if ((nrow(model.frame(x)) == n) & xfix) {
    X <- exogenous(x)
    mydata <- model.frame(x)
    for (pred in X) {
      distribution(m, pred) <- list(mydata[, pred])
    }
  }
  sim(m, n = n, p = p, ...)
}

##' @export
simulate.lvm <- function(object, nsim, seed = NULL, ...) {
  sim(object, nsim, seed = seed, ...)
}

##' @export
simulate.lvmfit <- function(object, nsim, seed = NULL, ...) {
  sim(object, nsim, seed = seed, ...)
}

#' Estimate mixture latent variable model.
#'
#' Estimate mixture latent variable model
#'
#' Estimate parameters in a mixture of latent variable models via the EM
#' algorithm.
#'
#' The performance of the EM algorithm can be tuned via the \code{control}
#' argument, a list where a subset of the following members can be altered:
#'
#' \describe{ \item{start}{Optional starting values} \item{nstart}{Evaluate
#' \code{nstart} different starting values and run the EM-algorithm on the
#' parameters with largest likelihood} \item{tol}{Convergence tolerance of the
#' EM-algorithm.  The algorithm is stopped when the absolute change in
#' likelihood and parameter (2-norm) between successive iterations is less than
#' \code{tol}} \item{iter.max}{Maximum number of iterations of the
#' EM-algorithm} \item{gamma}{Scale-down (i.e. number between 0 and 1) of the
#' step-size of the Newton-Raphson algorithm in the M-step} \item{trace}{Trace
#' information on the EM-algorithm is printed on every \code{trace}th
#' iteration} }
#'
#' Note that the algorithm can be aborted any time (C-c) and still be saved
#' (via on.exit call).
#'
 #' @param x List of \code{lvm} objects. If only a single \code{lvm} object is
#' given, then a \code{k}-mixture of this model is fitted (free parameters
#' varying between mixture components).
#' @param data \code{data.frame}
#' @param k Number of mixture components
#' @param control Optimization parameters (see details)
#' #type Type of EM algorithm (standard, classification, stochastic)
#' @param vcov of asymptotic covariance matrix (NULL to omit)
#' @param names If TRUE returns the names of the parameters (for defining starting values)
#' @param ... Additional arguments parsed to lower-level functions
#' @author Klaus K. Holst
#' @seealso \code{mvnmix}
#' @keywords models regression
#' @examples
#'
#' \dontrun{ ## Reduce Ex.timings
#' m0 <- lvm(list(y~x+z,x~z))
#' distribution(m0,~z) <- dist_bernoulli()
#' d <- sim(m0,2000,p=c("y~z"=2,"y~x"=1),seed=1)
#'
#' ## unmeasured confounder example
#' m <- baptize(lvm(y~x, x~1));
#' intercept(m,~x+y) <- NA
#'
#' if (requireNamespace('mets', quietly=TRUE)) {
#'   set.seed(42)
#'   M <- mixture(m,k=2,data=d,control=list(trace=1,tol=1e-6))
#'   summary(M)
#'   lm(y~x,d)
#'   estimate(M,"y~x")
#'   ## True slope := 1
#' }
#' }
#'
#' @export mixture
mixture <- function(x, data, k=length(x),
             control=list(),
             vcov="observed",
             names=FALSE,
             ...) {
    MODEL <- "normal"
    ##type=c("standard","CEM","SEM"),
    ## type <- tolower(type[1])
    ## if (type[1]!="standard") {
    ##     return(mixture0(x,data=data,k=k,control=control,type=type,...))
    ## }

    optim <- list(
        rerun=TRUE,
        ## EM (accelerated):
        K=1, ## K=1, first order approx, slower but some times more stable
        square=TRUE,
        step.max0=1,
        step.min0=1,
        mstep=4,
        kr=1,
        objfn.inc=2,
        keep.objfval=TRUE,
        convtype= "parameter",
        ## convtype = "objfn",
        maxiter=1500,
        tol=1e-5,
        trace=0,
        ## Starting values:
        start=NULL,
        startbounds=c(-2,2),
        startmean=FALSE,
        nstart=1,
        prob=NULL,
        ## Newton raphson:
        delta=1e-2,
        constrain=TRUE,
        stopc=2,
        lbound=1e-9,
        stabil=TRUE,
        gamma=0.5,
        gamma2=1,
        newton=10,
        lambda=0 # Stabilizing factor (avoid singularities of I)
    )

    if (!missing(control))
        optim[base::names(control)] <- control
    if ("iter.max"%in%base::names(optim)) optim$maxiter <- optim$iter.max

    if (k==1) {
        if (is.list(x))
            res <- estimate(x[[1]],data,...)
        else
            res <- estimate(x,data,...)
        return(res)
    }

    start0 <- NULL

    xx <- x
    if (inherits(x,"lvm")) {
        xx <- rep(list(x), k)
    }
    mg <- multigroup(xx, rep(list(data),k), fix=FALSE)
    ppos <- parpos(mg)
    parname <- attr(ppos,"name")
    naparname <- which(is.na(parname))
    parname[naparname]  <- mg$name[naparname]
    if (names) {
        return(parname)
    }

    if (class(x)[1]=="lvm") {
        index(x) <- reindex(x,zeroones=TRUE,deriv=TRUE)
        if ((is.null(optim$start) || length(optim$start)<length(parname))) {
            start0 <- tryCatch(estimate(x,data,quick=TRUE),
                               error=function(...) NULL)
        }
    }

    ## Bounds on variance parameters
    Npar <- with(mg, npar+npar.mean)
    ParPos <- modelPar(mg,1:Npar)$p
    lower <- rep(-Inf, mg$npar);
    offdiagpos <- c()
    varpos <- c()
    for (i in 1:k) {
        vpos <- sapply(mg$parlist[[i]][variances(mg$lvm[[i]])], function(y) as.numeric(substr(y,2,nchar(y))))
        offpos <- sapply(mg$parlist[[i]][offdiags(mg$lvm[[i]])], function(y) as.numeric(substr(y,2,nchar(y))))
        varpos <- c(varpos, vpos)
        offdiagpos <- c(offdiagpos,offpos)
        if (length(vpos)>0)
            lower[vpos] <- optim$lbound  ## Setup optimization constraints
    }
    lower <- c(rep(-Inf,mg$npar.mean), lower)
    constrained <- which(is.finite(lower))
    if (!any(constrained)) optim$constrain <- FALSE

    mymodel <- list(multigroup=mg,k=k,data=data,parpos=ParPos); class(mymodel) <- "lvm.mixture"

    selfstarter <- FALSE
    if (any(is.na(optim$start)) || length(optim$start)<length(ppos)) {
        constrLogLikS <- function(p) {
            if (optim$constrain) {
                p[constrained] <- exp(p[constrained])
            }
            -logLik(mymodel,c(p=p,rep(1/k,k-1)))
        }

        start <- rep(NA,Npar)
        if (!is.null(start0)) {
            start[ParPos[[1]]] <- start0[seq_along(ParPos[[1]])]
            ii <- which(is.na(start))
            start[ii] <- runif(length(ii),optim$startbounds[1],optim$startbounds[2])
        } else {
            selfstarter <- TRUE
            start <- runif(Npar,optim$startbounds[1],optim$startbounds[2]);
            if (length(offdiagpos)>0)
                start[mg$npar.mean + offdiagpos] <- 0
            if (optim$nstart>1) {
                myll <- constrLogLikS(start)
                for (i in 1:optim$nstart) {
                    newstart <- runif(Npar,optim$startbounds[1],optim$startbounds[2]);
                    newmyll <- constrLogLikS(newstart)
                    if (newmyll<myll) {
                        start <- newstart
                    }
                }
            }
        }
        if (!is.null(optim$start)) {
            iistart <- match(names(optim$start), parname)
            start[iistart] <- optim$start
        }
        optim$start <- start
    }
    if (length(optim$start)>Npar) {
        optim$prob <- optim$start[Npar+seq_len(k-1)]
        optim$start <- optim$start[seq_len(Npar)]
    }

    if (is.null(optim$prob))
        optim$prob <- rep(1/k,k-1)
    thetacur <- optim$start

    probcur <- with(optim, c(prob,1-sum(prob)))

    if (optim$constrain & !selfstarter) {
        thetacur[constrained] <- log(thetacur[constrained])
    }

    PosteriorProb <- function(pp,priorprob,constrain=FALSE) {
        if (!is.list(pp)) {
            if (constrain) {
                pp[constrained] <- exp(pp[constrained])
            }
            if (missing(priorprob)) priorprob <- pp[seq(Npar+1,length(pp))]
            pp <- lapply(ParPos,function(x) pp[x])
        }
        priorprob <- pmax(priorprob,1e-16)
        priorprob <- priorprob/sum(priorprob)
        k <- length(pp)
        logff <- sapply(seq(k), function(j) (logLik(mg$lvm[[j]],p=pp[[j]],data=data,indiv=TRUE,model=MODEL)))
        logplogff <- t(apply(logff,1, function(z) z+log(priorprob)))
        ## Log-sum-exp (see e.g. NR)
        zmax <- apply(logplogff,1,max)
        logsumpff <- log(rowSums(exp(logplogff-zmax)))+zmax
        gamma <- exp(apply(logplogff,2,function(y) y - logsumpff)) ## Posterior class probabilities
        return(gamma)
    }

    negLogLik <- function(p) {
        if (optim$constrain) {
            p[constrained] <- exp(p[constrained])
        }
        myp <- lapply(ParPos,function(x) p[x])
        K <- length(myp)
        prob <- p[seq(Npar+1,Npar+K-1)]; prob <- c(prob,1-sum(prob))
        logff <- sapply(seq_along(myp), function(j) (logLik(mg$lvm[[j]],p=myp[[j]],data=data,indiv=TRUE,model=MODEL)))
        ## logff <- sapply(1:length(myp),
        ##              function(j) -normal_objective.lvm(mg$lvm[[j]],p=myp[[j]],data=data,indiv=TRUE))
        logplogff <- t(apply(logff,1, function(y) y+log(prob)))
        zmax <- apply(logplogff,1,max)
        logsumpff <- log(rowSums(exp(logplogff-zmax)))+zmax
        loglik <- sum(logsumpff)
        return(-loglik)
    }


    ObjEstep <- function(p,gamma,pr) {
        if (optim$constrain) {
            p[constrained] <- exp(p[constrained])
        }
        myp <- lapply(ParPos,function(x) p[x])
        loglik = 0;
        for (j in seq_along(myp))
            loglik = loglik + gamma[,j]*(logLik(mg$lvm[[j]],p=myp[[j]],data=data,indiv=TRUE,model=MODEL))
        ## logff <- sapply(1:length(myp), function(j) gamma[,j]*(logLik(mg$lvm[[j]],p=myp[[j]],data=data,indiv=TRUE,model=MODEL)))
        ## zmax <- apply(logff,1,max)
        ## ffz <- apply(logff,2,function(x) exp(x-zmax))
        ## logsff <- log(rowSums(ffz))+zmax
        ## loglik <- sum(logsff)
        return(-sum(loglik))
    }

    GradEstep <- function(p,gamma,pr) {
        if (optim$constrain) {
            p[constrained] <- exp(p[constrained])
        }
        myp <- lapply(ParPos,function(x) p[x])
        ## logff <- sapply(1:length(myp), function(j) log(gamma[,j])+(logLik(mg$lvm[[j]],p=myp[[j]],data=data,indiv=TRUE,model=MODEL)))
        ## Exp-sum normalization: exp(xi)/sum(exp(xi)) = exp(xi-b)*exp(b)/sum(exp(xi-b)exp(b))
        ## = exp(xi-b)/sum(exp(xi-b)),  b=max(xi)
        ## zmax <- apply(logff,1,max)
        ## ffz <- apply(logff,2,function(x) exp(x-zmax))
        ## sffz <- rowSums(ffz)
        D <- lapply(seq_along(myp), function(j) {
            ## K <- ffz[,j]/sffz
            val <- score(mg$lvm[[j]],p=myp[[j]],data=data,indiv=TRUE,model=MODEL)
            apply(val,2,function(x) x*gamma[,j])
        })
        D0 <- matrix(0,nrow(data),length(p))
        for (j in 1:k) D0[,ParPos[[j]]] <- D0[,ParPos[[j]]]+D[[j]]
        S <- colSums(D0)
        if (optim$constrain) {
            S[constrained] <- S[constrained]*p[constrained]
        }
        return(-as.vector(S))
    }

   Information <- function(p,gamma,pr) {
        if (optim$constrain) {
            p[constrained] <- exp(p[constrained])
        }
        myp <- lapply(ParPos,function(x) p[x])
        D <- lapply(seq_along(myp), function(j) {
            ## K <- ffz[,j]/sffz
            val <- score(mg$lvm[[j]],p=myp[[j]],data=data,indiv=TRUE,model=MODEL)
            apply(val,2,function(x) x*gamma[,j])
        })
        D0 <- matrix(0,nrow(data),length(p))
        for (j in 1:k) D0[,ParPos[[j]]] <- D0[,ParPos[[j]]]+D[[j]]
        if (optim$constrain) {
            for (j in constrained)
                D0[,j] <- D0[,j]*p[j]
        }
        S <- colSums(D0)
        structure(crossprod(D0), grad=S)
    }

    EMstep <- function(p,all=FALSE) {
        thetacur <- p[seq(Npar)]
        gamma <- PosteriorProb(p,constrain=optim$constrain)
        probcur <- colMeans(gamma)
        ## I <- function(p) {
        ##     I <- Information(p,gamma,probcur)
        ##     D <- attr(I, "grad")
        ##     res <- -Inverse(I)
        ##     res <- -I
        ##     attributes(res)$grad <- D
        ##     res
        ## }
        D <- function(p) GradEstep(p,gamma,probcur)
        ## if (optim$newton>0) {
        ##     newpar <- NR(thetacur, gradient=D)
        ## }
        ## if (mean(newpar$gradient^2)>optim$tol) {
        newpar <- nlminb(thetacur,function(p) ObjEstep(p,gamma,probcur), D)
        ## }
        thetacur <- newpar$par
        thetacur0 <- thetacur
        if (optim$constrain) {
            thetacur0[constrained] <- exp(thetacur[constrained])
        }
        p <- c(thetacur,probcur)
        if (all) {
            res <- list(p=p,gamma=gamma,
                        theta=rbind(thetacur0),
                        prob=rbind(probcur))
            return(res)
        }
        return(p)
    }
    em.idx <- match(c("K","method","square","step.min0","step.max0","mstep",
                      "objfn.inc","kr",
                      ##"keep.objfval","convtype",
                      "maxiter","tol","trace"),base::names(optim))
    em.control <- optim[na.omit(em.idx)]
    if (!is.null(em.control$trace)) em.control$trace <- em.control$trace>0

    p <- c(thetacur,probcur)
    opt <- SQUAREM::squarem(p,fixptfn=EMstep,##objfn=negLogLik,
                            control=em.control)
    ## opt2 <- nlminb(opt$par, function(p) negLogLik(p=p), control=list(trace=1))
    val <- EMstep(opt$par,all=TRUE)
    delta <- 1e-6
    if (any(val$prob<delta)) {
        val$prob[val$prob<delta] <- delta
        val$prob <- val$prob/sum(val$prob)
    }

    val <- c(val, list(member=apply(val$gamma,1,which.max),
                       k=k,
                       data=data,
                       parpos=ParPos,
                       multigroup=mg,
                       model=mg$lvm,
                       parname=parname,
                       logLik=NA,
                       opt=opt))
    class(val) <- "lvm.mixture"

    np <- length(coef(val))
    if (is.null(vcov)) {
        val$vcov <- matrix(NA,np,np)
    } else {
        I <- suppressWarnings(information.lvm.mixture(val,type=vcov))
        val$vcov <- tryCatch(Inverse(I),error=function(...) matrix(NA,np,np))
    }
    return(val)
}

##' @export
score.lvm.mixture <- function(x,p=coef(x,full=TRUE),prob,indiv=FALSE,model="normal",...) {
    myp <- lapply(x$parpos,function(x) p[x])
    if (missing(prob)) {
        prob <- tail(p,x$k-1)
        p <- p[seq_len(length(p)-x$k+1)]
    }
    if (length(prob)<x$k)
        prob <- c(prob,1-sum(prob))
    logff <- sapply(seq(x$k), function(j) (logLik(x$multigroup$lvm[[j]],p=myp[[j]],data=x$data,indiv=TRUE,model=model)))
    logplogff <- t(apply(logff,1, function(y) y+log(prob)))
    zmax <- apply(logplogff,1,max)
    logsumpff <- log(rowSums(exp(logplogff-zmax)))+zmax
    aji <- apply(logplogff,2,function(x) exp(x-logsumpff))

    scoref <- lapply(score(x$multigroup,p=p,indiv=TRUE,model=model),
                     function(x) { x[which(is.na(x))] <- 0; x })

    Stheta <- matrix(0,ncol=ncol(scoref[[1]]),nrow=nrow(scoref[[1]]))
    Spi <- matrix(0,ncol=x$k-1,nrow=nrow(Stheta))
    for (j in 1:x$k) {
        Stheta <- Stheta + apply(scoref[[j]],2,function(x) x*aji[,j])
        if (j<x$k)
            Spi[,j] <- aji[,j]/prob[j] - aji[,x$k]/prob[x$k]
    }
    S <- cbind(Stheta,Spi)
    if (!indiv)
        return(colSums(S))
    return(S)
}

##' @export
information.lvm.mixture <- function(x,p=coef(x,full=TRUE),...,type="observed") {
    if (tolower(type)%in%c("obs","observed","hessian")) {
        S <- function(p=p,...) score.lvm.mixture(x,p=p,indiv=FALSE,...)
        I <- -numDeriv::jacobian(S,p)
        return(I)
    }
    S <- score.lvm.mixture(x,p=p,indiv=TRUE,...)
    res <- t(S)%*%S
    attributes(res)$grad <- colSums(S)
    return(res)
}

##' @export
logLik.lvm.mixture <- function(object,p=coef(object,full=TRUE),prob,model="normal",...) {
    if (is.null(object$parpos)) {
        object$parpos <- modelPar(object$multigroup,seq_along(p))$p
    }
    myp <- lapply(object$parpos, function(x) p[x])
    if (missing(prob))
        prob <- tail(p,object$k-1)
    if (length(prob)<object$k)
        prob <- c(prob,1-sum(prob))
    logff <- sapply(seq(object$k), function(j) (logLik(object$multigroup$lvm[[j]],p=myp[[j]],data=object$data,indiv=TRUE,model=model)))
    logplogff <- t(apply(logff,1, function(y) y+log(prob)))
    ## Log-sum-exp (see e.g. NR)
    zmax <- apply(logplogff,1,max)
    logsumpff <- log(rowSums(exp(logplogff-zmax)))+zmax
    loglik <- sum(logsumpff)
    npar <- length(p)
    nobs <- nrow(object$data)
    attr(loglik, "nall") <- nobs
    attr(loglik, "nobs") <- nobs
    attr(loglik, "df") <- npar
    class(loglik) <- "logLik"
    return(loglik)
}

##' @export
vcov.lvm.mixture <- function(object,...) {
    return(object$vcov)
}

##' @export
summary.lvm.mixture <- function(object,type=0,labels=0,...) {
    ppos <- parpos(object$multigroup)
    parname <- attr(ppos,"name")
    naparname <- which(is.na(parname))
    parname[naparname]  <- object$multigroup$name[naparname]
    ParPos <- modelPar(object$multigroup,seq_along(coef(object)))$p
    mm <- object$multigroup$lvm
    p <- coef(object,list=TRUE)
    p0 <- coef(object,prob=FALSE)
    myp <- modelPar(object$multigroup,seq_along(p0))$p
    coefs <- list()
    ncluster <- c()
    Coefs <- matrix(NA,ncol=4,nrow=length(parname))
    colnames(Coefs) <- c("Estimate","Std. Error", "Z value", "Pr(>|z|)")
    Types <- rep("other",length(parname))
    Variable  <- rep(NA,length(parname))
    From  <- rep(NA,length(parname))
    Latent <- c()
    for (i in seq_along(mm)) {
        cc.idx <- order(coef(mm[[i]],p=seq_along(p[[i]]),type=2)[,1])

        cc <- coef(mm[[i]],p=p[[i]],vcov=vcov(object)[myp[[i]],myp[[i]]],data=NULL,labels=labels,type=2)
        Latent <- union(Latent,attr(cc,"latent"))
        Coefs[ParPos[[i]],] <- cc[cc.idx,,drop=FALSE]
        Types[ParPos[[i]]] <- attr(cc,"type")[cc.idx]
        Variable[ParPos[[i]]] <- attr(cc, "var")[cc.idx]
        From[ParPos[[i]]] <- attr(cc, "from")[cc.idx]
        cc <- CoefMat(mm[[i]],p=p[[i]],vcov=vcov(object)[myp[[i]],myp[[i]]],data=NULL,labels=labels)
        coefs <- c(coefs, list(cc))
        ncluster <- c(ncluster,sum(object$member==i))
    }
    rownames(Coefs) <- parname

    s2 <- tryCatch(sum(score(object)^2), error=function(...) NA)
    AIC <- tryCatch(AIC(object), error=function(...) NA)
    res <- list(coef=coefs, coefmat=Coefs, coeftype=Types,
                type=type, var=Variable, from=From, latent=Latent,
                ncluster=ncluster, prob=tail(object$prob,1),
                AIC=AIC, s2=s2)
    class(res) <- "summary.lvm.mixture"
    return(res)
}

##' @export
print.summary.lvm.mixture <- function(x,...) {
    if (x$type>0) {
        cc <- x$coefmat
        attr(cc,"type") <- x$coeftype
        attr(cc,"latent") <- x$latent
        attr(cc,"var") <- x$var
        attr(cc,"from") <- x$from
        cat("Mixing parameters:\n")
        cat("  ", paste(as.vector(formatC(x$prob))),"\n")
        print(CoefMat(cc), quote=FALSE)
        return(invisible())
    }
    for (i in seq_along(x$coef)) {
        cat("Cluster ",i," (n=",x$ncluster[i],", Prior=", formatC(x$prob[i]),"):\n",sep="")
        cat(rep("-",50),"\n",sep="")
        print(x$coef[[i]], quote=FALSE)
        if (i<length(x$coef)) cat("\n")
    }
    cat(rep("-",50),"\n",sep="")
    if (!is.na(x$AIC)) cat("AIC=",x$AIC,"\n")
    if (!is.na(x$s2)) cat("||score||^2=",x$s2,"\n")
    invisible()
}

##' @export
print.lvm.mixture <- function(x,...) {
    ## for (i in 1:x$k) {
    ##     cat("Cluster ",i," (pr=",x$prob[i],"):\n",sep="")
    ##     cat(rep("-",50),"\n",sep="")
    ##     print(coef(x,prob=FALSE)[x$parpos[[i]]], quote=FALSE)
    ##     cat("\n")
    ## }
    print(summary(x,type=1,...))
}

##' @export
plot.lvm.mixture <- function(x,type="l",...) {
    matplot(x$theta,type=type,...)
}

##' @export
coef.lvm.mixture <- function(object,iter,list=FALSE,full=TRUE,prob=FALSE,class=FALSE,label=TRUE,...) {
    N <- nrow(object$theta)
    res <- object$theta
    ## nn <- attr(parpos(object$multigroup),"name")
    ## if (length(ii <- which(is.na(nn)))>0 && label) {
    ##     nn[ii] <- paste0("p",seq_along(ii))
    ## }
    nn <- object$parname
    colnames(res) <- nn
    if (class) return(object$gammas)
    if (list) {
        res <- list()
        for (i in 1:object$k) {
            nn <- coef(object$multigroup$lvm[[i]])
            cc <- coef(object)[object$parpos[[i]]]
            base::names(cc) <- nn
            res <- c(res, list(cc))
        }
        return(res)
    }
    if (full) {
        pp <- object$prob[,seq(ncol(object$prob)-1),drop=FALSE]
        colnames(pp) <- paste0("pr",seq(ncol(pp)))
        res <- cbind(res,pp)
    }
    if (prob) {
        res <- object$prob
    }
    if (missing(iter))
        return(res[N,,drop=TRUE])
    else
        return(res[iter,])
}

##' @export
model.frame.lvm.mixture <- function(formula,...) {
    return(formula$data)
}

##' @export
IC.lvm.mixture <- function(x,...) {
  U <- score(x,indiv=TRUE)
  bread <- vcov(x)*NROW(U)
  structure(t(bread%*%t(U)),bread=bread)
}

##' @export
manifest.lvm.mixture <- function(x,...) {
    manifest(x$multigroup,...)
}

##' @export
`measurement` <-
    function(x, ...) {
        M <- x$M
        latent.idx <- match(latent(x),vars(x))
        obs.idx <- match(manifest(x),vars(x))
        if (length(latent.idx)==0)
            return(NULL)

        measurementmodels <- c()
        for (i in seq_along(latent.idx)) {
            ii <- latent.idx[i]

            relation <- M[ii,obs.idx]==1
            byNodes <- names(relation)[relation]
            newnodes <- c(latent(x)[i],byNodes)
            lvm1 <- subset(x,newnodes)
            measurementmodels <- c(measurementmodels, list(lvm1))
        }

        measurementmodels
    }

##' Trim string of (leading/trailing/all) white spaces
##' @title Trim string of (leading/trailing/all) white spaces
##' @param x String
##' @param all Trim all whitespaces?
##' @param \dots additional arguments to lower level functions
##' @author Klaus K. Holst
##' @export
trim <- function(x,all=FALSE,...) {
    ## y <- gsub("^ .", "", x) # remove leading white space
    ## y <- gsub(". $", "", x) # remove trailing white space
    if (!all) return(gsub("^\\s+|\\s+$", "", x))
    return(gsub("\\s","",x))
}

##' Create contrast matrix
##'
##' Create contrast matrix typically for use with 'estimate' (Wald tests).
##' @export
##' @param p index of non-zero entries (see example)
##' @param n Total number of parameters (if omitted the max number in p will be used)
##' @param diff If FALSE all non-zero entries are +1, otherwise the second non-zero element in each row will be -1.
##' @param ... Additional arguments to lower level functions
##' @aliases contr parsedesign
##' @examples
##' contr(2,n=5)
##' contr(as.list(2:4),n=5)
##' contr(list(1,2,4),n=5)
##' contr(c(2,3,4),n=5)
##' contr(list(c(1,3),c(2,4)),n=5)
##' contr(list(c(1,3),c(2,4),5))
##'
##' parsedesign(c("aa","b","c"),"?","?",diff=c(FALSE,TRUE))
contr <- function(p,n,diff=TRUE,...) {
    if (missing(n)) n <- max(unlist(p))
    if (is.character(p)) {
        return(parsedesign(n,p,...))
    }
    if (is.list(p)) {
        return(Reduce(rbind,lapply(p, function(x) do.call(contr, list(x,n,diff[1L])))))
    }
    if (is.character(n)) n <- length(n)
    if (!is.numeric(n)) {
        try(n <- length(coef(n)),silent=TRUE)
    }
    B <- matrix(0,ncol=n,nrow=max(1L,length(p)-1L))
    B[,p[1]] <- 1L
    if (length(p)>1L)
        B[cbind(seq(nrow(B)),p[-1])] <- ifelse(diff[1L],-1,1)
    return(B)
}

lisrel <- function(model,p,X=NULL,muX=NULL,varX=NULL,...) {
    mom <- moments(model,p)
    A <- t(index(model)$M)

    eta.idx <- match(latent(model),vars(model))
    exo.idx <- match(exogenous(model),vars(model))
    y <- setdiff(manifest(model), exogenous(model))
    y.idx <- match(y, vars(model))

    ##  Jy <- Jx <- Jeta <- I <- diag(length(vars(model)))
    ##  if (length(eta.idx)>0)
    ##    J[eta.idx,eta.idx] <- 0; J <- J[-eta.idx,]
    ##  Jeta[obs.idx,obs.idx] <- 0; Jeta <- J[-obs.idx,]

    A <- t(mom$A)
    Lambda <- A[y.idx,eta.idx,drop=FALSE]
    K <- A[y.idx,exo.idx,drop=FALSE]
    B <- A[eta.idx,eta.idx,drop=FALSE]
    I <- diag(nrow=nrow(B))
    Gamma <- A[eta.idx,exo.idx,drop=FALSE]
    V <- mom$P
    Psi <- V[eta.idx,eta.idx] ## Residual variance

    Theta <- V[y.idx,y.idx] ## -
    IBi <- if (ncol(I)>0) solve(I-B) else I
    LIBi <- Lambda%*%IBi
    Phi <- LIBi%*%Gamma + K

    Veta.x <- IBi%*%Psi%*%IBi  ## Variance of eta given x
    COVetay.x <- Veta.x%*%t(Lambda) ## Covariance of eta,y given x
    ##  Vy.x <- Lambda%*%COVetay.x + Theta ## Omega
    Vy.x <- LIBi%*%Psi%*%t(LIBi) + Theta

    if (!is.null(X)) {
        Ey.x <- t(apply(as.matrix(X)%*% t(LIBi%*%Gamma + K),1,function(x) x + mom$v[y.idx]))
    } else Ey.x <- NULL

    CV <- COVetay.x%*%Vy.x
    ##  Sigma <- Vy.x + Phi%*%varX%*%t(Phi)

    return(list(mu=mom$v,
                Lambda=Lambda, K=K, B=B, I=I, Gamma=Gamma, Psi=Psi, Theta=Theta, IBi=IBi, LIBi=LIBi, Phi=Phi,
                Vy.x=Vy.x, Veta.x=Veta.x, COVetay.x=COVetay.x, CV=CV, Ey.x=Ey.x))
}

twostagelvm <- function(object, model2,
                formula=NULL, model.object=FALSE, predict.fun=NULL,
                type="quadratic",...) {
    if (!inherits(model2,c("lvm"))) stop("Expected lava object ('lvm',...)")
    if (!is.null(formula)) {
        model2 <- nonlinear(model2, formula, type=type)
    }
    nonlin <- NULL
    val <- nonlinear(model2)
    if (is.null(formula) && length(val)==0 && length(nonlinear(object))>0) {
        val <- nonlinear(object)
    }
    xnam <- c()
    if (length(val)>0) {
        predict.fun <- NULL
        for (i in seq_along(val)) {
            if (!all(val[[i]]$newx%in%xnam)) {
                xnam <- union(xnam,val[[i]]$newx)
                predict.fun <- c(predict.fun, list(val[[i]]$pred))
            }
            model2$attributes$nonlinear <- NULL
            if (inherits(object,"lvmfit")) {
                object$model$attributes$nonlinear <- NULL
            }
            model2 <- regression(model2, to=names(val)[i], from=val[[i]]$newx)
        }
        nonlin <- val
    }
    if (model.object) {
        model <- Model(object) %++% model2
        cl <- match.call(expand.dots=TRUE)
        cl[[1]] <- twostage
        cl$object <- object
        cl$model2 <- model2
        cl$predict.fun <- predict.fun
        cl["model.object"] <- NULL
        return(structure(list(model=model, nonlinear=nonlin, call=cl), class="twostage.lvm"))
    }
    res <- c(list(object=object, model2=model2), list(...))
    res$predict.fun <- predict.fun
    res$nonlinear <- val
    return(res)
}

uhat <- function(p=coef(model1), model1, data=model.frame(model1), nlobj) {
    if (!is.function(nlobj)) {
        predict.fun <- lapply(nlobj, function(x) x[["pred"]])
    } else { predict.fun <- nlobj }
    if (inherits(model1, "lvm.mixture")) {
        if (is.list(predict.fun)) {
            unams <- lapply(nlobj,function(x) x$newx)
            unam <- unique(unlist(unams))
            res <- matrix(0, NROW(data), ncol=length(unam))
            colnames(res) <- unam
            for (i in seq_along(predict.fun)) {
                res[, unams[[i]]] <-
                    predict(model1, p=p, data=data, predict.fun=predict.fun[[i]])
            }
            return(res)
        } else {
            Pr <- cbind(predict(model1, p=p, data=data, predict.fun=predict.fun))
            return(Pr)
        }
        ##P <- list(mean=Pr, var=attr(Pr,"cond.var"))
    }  else {
        P <- predictlvm(model1, p=p, data=data)
    }
    if (is.list(predict.fun)) {
        unams <- lapply(nlobj,function(x) x$newx)
        unam <- unique(unlist(unams))
        args <- list(P$mean, P$var, data)
        res <- matrix(0, NROW(data), ncol=length(unam))
        colnames(res) <- unam
        for (i in seq_along(predict.fun)) {
            res[, unams[[i]]] <- do.call(predict.fun[[i]], args)
        }
        return(res)
    }
    return(cbind(predict.fun(P$mean, P$var, model.frame(model1))))
}

##' Two-stage estimator
##'
##' Generic function.
##'
##' @seealso twostage.lvm twostage.lvmfit twostage.lvm.mixture twostage.estimate
##' @export
##' @param object Model object
##' @param ... Additional arguments to lower level functions
"twostage" <- function(object, ...) UseMethod("twostage")

##' Two-stage estimator (non-linear SEM)
##'
##' Two-stage estimator for non-linear structural equation models
##' @export
##' @param object Stage 1 measurement model
##' @param model2 Stage 2 SEM
##' @param data data.frame
##' @param predict.fun Prediction of latent variable
##' @param id1 Optional id-variable (stage 1 model)
##' @param id2 Optional id-variable (stage 2 model)
##' @param all If TRUE return additional output (naive estimates)
##' @param formula optional formula specifying non-linear relation
##' @param std.err If FALSE calculations of standard errors will be skipped
##' @param ... Additional arguments to lower level functions
##' @aliases twostage.lvmfit twostage.lvm twostage.lvm.mixture twostage.estimate nonlinear nonlinear<-
##' @examples
##' m <- lvm(c(x1,x2,x3)~f1,f1~z,
##'          c(y1,y2,y3)~f2,f2~f1+z)
##' latent(m) <- ~f1+f2
##' d <- simulate(m,100,p=c("f2,f2"=2,"f1,f1"=0.5),seed=1)
##'
##' ## Full MLE
##' ee <- estimate(m,d)
##'
##' ## Manual two-stage
##' \dontrun{
##' m1 <- lvm(c(x1,x2,x3)~f1,f1~z); latent(m1) <- ~f1
##' e1 <- estimate(m1,d)
##' pp1 <- predict(e1,f1~x1+x2+x3)
##'
##' d$u1 <- pp1[,]
##' d$u2 <- pp1[,]^2+attr(pp1,"cond.var")[1]
##' m2 <- lvm(c(y1,y2,y3)~eta,c(y1,eta)~u1+u2+z); latent(m2) <- ~eta
##' e2 <- estimate(m2,d)
##' }
##'
##' ## Two-stage
##' m1 <- lvm(c(x1,x2,x3)~f1,f1~z); latent(m1) <- ~f1
##' m2 <- lvm(c(y1,y2,y3)~eta,c(y1,eta)~u1+u2+z); latent(m2) <- ~eta
##' pred <- function(mu,var,data,...)
##'     cbind("u1"=mu[,1],"u2"=mu[,1]^2+var[1])
##' (mm <- twostage(m1,model2=m2,data=d,predict.fun=pred))
##'
##' if (interactive()) {
##'     pf <- function(p) p["eta"]+p["eta~u1"]*u + p["eta~u2"]*u^2
##'     plot(mm,f=pf,data=data.frame(u=seq(-2,2,length.out=100)),lwd=2)
##' }
##'
##' \donttest{ ## Reduce test timing
##' ## Splines
##' f <- function(x) cos(2*x)+x+-0.25*x^2
##' m <- lvm(x1+x2+x3~eta1, y1+y2+y3~eta2, latent=~eta1+eta2)
##' functional(m, eta2~eta1) <- f
##' d <- sim(m,500,seed=1,latent=TRUE)
##' m1 <- lvm(x1+x2+x3~eta1,latent=~eta1)
##' m2 <- lvm(y1+y2+y3~eta2,latent=~eta2)
##' mm <- twostage(m1,m2,formula=eta2~eta1,type="spline")
##' if (interactive()) plot(mm)
##'
##' nonlinear(m2,type="quadratic") <- eta2~eta1
##' a <- twostage(m1,m2,data=d)
##' if (interactive()) plot(a)
##'
##' kn <- c(-1,0,1)
##' nonlinear(m2,type="spline",knots=kn) <- eta2~eta1
##' a <- twostage(m1,m2,data=d)
##' x <- seq(-3,3,by=0.1)
##' y <- predict(a, newdata=data.frame(eta1=x))
##'
##' if (interactive()) {
##'   plot(eta2~eta1, data=d)
##'   lines(x,y, col="red", lwd=5)
##'
##'   p <- estimate(a,f=function(p) predict(a,p=p,newdata=x))$coefmat
##'   plot(eta2~eta1, data=d)
##'   lines(x,p[,1], col="red", lwd=5)
##'   confband(x,lower=p[,3],upper=p[,4],center=p[,1], polygon=TRUE, col=Col(2,0.2))
##'
##'   l1 <- lm(eta2~splines::ns(eta1,knots=kn),data=d)
##'   p1 <- predict(l1,newdata=data.frame(eta1=x),interval="confidence")
##'   lines(x,p1[,1],col="green",lwd=5)
##'   confband(x,lower=p1[,2],upper=p1[,3],center=p1[,1], polygon=TRUE, col=Col(3,0.2))
##' }
##' } ## Reduce test timing
##'
##' \dontrun{ ## Reduce timing
##'  ## Cross-validation example
##'  ma <- lvm(c(x1,x2,x3)~u,latent=~u)
##'  ms <- functional(ma, y~u, value=function(x) -.4*x^2)
##'  d <- sim(ms,500)#,seed=1)
##'  ea <- estimate(ma,d)
##'
##'  mb <- lvm()
##'  mb1 <- nonlinear(mb,type="linear",y~u)
##'  mb2 <- nonlinear(mb,type="quadratic",y~u)
##'  mb3 <- nonlinear(mb,type="spline",knots=c(-3,-1,0,1,3),y~u)
##'  mb4 <- nonlinear(mb,type="spline",knots=c(-3,-2,-1,0,1,2,3),y~u)
##'  ff <- lapply(list(mb1,mb2,mb3,mb4),
##'	      function(m) function(data,...) twostage(ma,m,data=data,st.derr=FALSE))
##'  a <- cv(ff,data=d,rep=1)
##'  a
##'}
twostage.lvmfit <- function(object, model2, data=NULL,
                    predict.fun=NULL,
                    id1=NULL, id2=NULL, all=FALSE,
                    formula=NULL, std.err=TRUE,
                    ...) {
    if (!is.null(predict.fun)) {
        object$attributes$nonlinear <- list()
        model2$attributes$nonlinear <- list()
    }
    val <- twostagelvm(object=object,model2=model2,predict.fun=predict.fun,
                      id1=id1, id2=id2, all=all, formula=formula, ...)
    object <- val$object
    model2 <- val$model2
    predict.fun <- val$predict.fun
    p1 <- coef(object)
    if (length(val$nonlinear)==0) {
        val$nonlinear <- predict.fun
    }
    pp <- uhat(p1,object,nlobj=val$nonlinear)
    newd <- data
    newd[,colnames(pp)] <- pp

    model2 <- estimate(model2,data=newd,...)
    p2 <- coef(model2)
    if (std.err) {
        if (is.null(id1)) id1 <- seq(nrow(model.frame(object)))
        if (is.null(id2)) id2 <- seq(nrow(model.frame(model2)))
        model1 <- object
        if (!inherits(object,"estimate")) {
            model1 <- estimate(NULL, coef=p1, id=id1, IC=IC(object))
        }

        e2 <- estimate(model2, id=id2)
        U <- function(alpha = p1, beta = p2) {
            pp <- uhat(alpha, object, nlobj = val$nonlinear)
            newd <- model.frame(model2)
            newd[, colnames(pp)] <- pp
            score(model2, p = beta, data = newd)
        }
        stacked <- stack(model1,e2,U=U)
    } else {
        e2 <- estimate(coef=p2,vcov=NA)
    }
    coef <- model2$coef
    res <- model2
    res$estimator <- "generic"

    if (std.err) {
        res[names(stacked)] <- stacked
        cc <- stacked$coefmat[,c(1,2)];
        cc <- cbind(cc,cc[,1]/cc[,2],stacked$coefmat[,5])
        coef[,] <- cc
        res$coef <- coef
        res$vcov <- vcov(stacked)
        if (all) {
            res$naive <- model2
            res$naive.robust <- e2
        }
    } else {
        res$coef[,-1] <- NA
    }
    res$fun <- predict.fun
    res$estimate1 <- object
    res$estimate2 <- model2
    res$nonlinear <- val$nonlinear
    structure(res,class=c("twostage.lvmfit","measurement.error","lvmfit","estimate"))
}

##' @export
estimate.twostage.lvm <- function(x,data,...) {
    if (missing(data)) stop("'data' needed")
    m1 <- x$call$object
    m2 <- x$call$model2
    nl <- x$nonlinear
    if (!inherits(m1,"lvmfit")) {
        args <- c(list(x=m1, data=data), list(...))
        args <- args[intersect(names(as.list(base::args(estimate.lvm))),names(args))]
        m1 <- do.call(estimate, args)
    }
    m2$attributes$nonlinear <- nl
    twostage(object=m1,model2=m2,data=data,predict.fun=nl[[1]]$pred,...)
}

##' @export
twostage.twostage.lvm <- function(object,...) estimate.twostage.lvm(object,...)

##' @export
twostage.lvm <- function(object,model2,data=NULL, ...) {
    if (is.null(data)) {
        return(twostagelvm(object=object, model2=model2, model.object=TRUE, ...))
    }
    args <- c(list(x=object, data=data), list(...))
    args <- args[intersect(names(as.list(base::args(estimate.lvm))),names(args))]
    e1 <- do.call(estimate, args)
    twostage(object=e1,model2=model2,data=data, ...)
}

##' @export
twostage.lvm.mixture <- twostage.lvmfit

##' @export
twostage.estimate <- twostage.lvmfit

##' @export
print.twostage.lvm <- function(x,...) {
    print(cli::rule())
    cat("Model 1:\n")
    print(Model(x$call$object))
    print(cli::rule())
    cat("Model 2:\n")
    print(Model(x$call$model2))
}

##' @export
plot.twostage.lvm <- function(x,...) {
    model <- x$model
    m1 <- Model(x$call$object)
    m2 <- x$call$model2
    nl <- nonlinear(x)
    model <- regression(model, to=nl[[1]]$newx, from=nl[[1]]$x)
    elist <- edgeList(m1)
    vlist <- vars(m1)
    model <- beautify(model)
    for (i in seq_len(nrow(elist))) {
        e <- toformula(y=vlist[elist[i,2]],x=vlist[elist[i,1]])
        edgelabels(model, e, cex=0.7) <- 1
    }
    elist <- edgeList(m2)
    vlist <- vars(m2)
    for (i in seq_len(nrow(elist))) {
        e <- toformula(vlist[elist[i,2]],vlist[elist[i,1]])
        edgelabels(model, e, cex=0.7) <- 2
    }
    nodecolor(model, nl[[1]]$newx) <- "gray"
    for (xx in nl[[1]]$newx) {
        e <- toformula(y=names(nl)[1],x=xx)
        edgelabels(model,e,col="gray", cex=0.7, lty=1) <- 2
    }
    for (xx in nl[[1]]$newx) {
        e <- toformula(y=xx,x=nl[[1]]$x)
        edgelabels(model,e,col="gray", cex=0.7, lty=2) <- ""
    }
    plot(model, ...)
}


##' @export
predict.twostage.lvmfit <- function(object,
                            newdata,
                            variable=names(nonlinear(object)),
                            p=coef(object),
                            type=c("model2","latent"),
                            x=NULL,
                            ...) {
    if (missing(newdata)) stop("provide data for prediction")
    nl <- nonlinear(object)
    unam <- unique(unlist(lapply(nl,function(x) x$x)))
    if (is.vector(newdata) || all(unam%in%colnames(newdata)))
        type <- "latent"


    if (tolower(type[1])%ni%c("latent")) {
        p1 <- coef(object$estimate1)
        pred1 <- uhat(p1, data=newdata, object$estimate1, nlobj=nl)
        if (tolower(type[1])==c("model1"))
            return(pred1)
        newdata <- as.data.frame(newdata)
        newdata[,colnames(pred1)] <- pred1
        pred <- predict(object$estimate2,x=x,p=p,data=newdata,...)
        attr(pred,"p") <- NULL
        attr(pred,"e") <- NULL
        attr(pred,"cond.var") <- NULL
        return(pred)
    }

    ## Association between predicted latent variables and child nodes:
    if (is.numeric(variable)) {
        variable <- names(nonlinear(object))[variable]
    }
    nl <- nl[variable]
    res <- matrix(nrow=NROW(newdata),ncol=length(nl))
    colnames(res) <- names(nl)
    ##unam <- unique(unlist(lapply(nl, function(x) x$newx)))
    ##colnames(res) <- unam
    if (!is.null(x)) {
        newd <- newdata
        m0 <- Model(object$estimate2)
        p0 <- p
    }
    for (i in seq_along(nl)) {
        pnam <- c(variable,paste0(variable,"~",nl[[i]]$newx))
        pidx <- match(pnam,names(coef(object)))
        b <- p[pidx]
        F <- nl[[i]]$f
        if (is.vector(newdata)) {
            res[,i] <- F(b,newdata)
        } else {
            res[,i] <- F(b,newdata[,nl[[i]]$x])
            if (!is.null(x)) {
                newd[,nl[[i]]$newx] <- 0
                latent(m0,clear=TRUE) <- names(nl)[i]
                regression(m0, y=names(nl)[i], x=paste0(names(nl)[i],".offset")) <- 1
                p0[pidx] <- 0
            }
        }
    }

    if (!is.null(x) && !is.vector(newdata)) {
        latentres <- res
        colnames(latentres) <- paste0(colnames(res),".offset",collapse="")
        newd <- cbind(newd,latentres)
        res <- predict(m0,data=newd,p=p0,x=x,...)
        attr(res,"cond.var") <- NULL

    }
    return(res)
}

##' Cross-validated two-stage estimator
##'
##' Cross-validated two-stage estimator for non-linear SEM
##' @export
##' @param model1 model 1 (exposure measurement error model)
##' @param model2 model 2
##' @param data data.frame
##' @param control1 optimization parameters for model 1
##' @param control2 optimization parameters for model 1
##' @param knots.boundary boundary points for natural cubic spline basis
##' @param nmix number of mixture components
##' @param df spline degrees of freedom
##' @param fix automatically fix parameters for identification (TRUE)
##' @param std.err calculation of standard errors (TRUE)
##' @param nfolds Number of folds (cross-validation)
##' @param rep Number of repeats of cross-validation
##' @param messages print information (>0)
##' @param ... additional arguments to lower
##' @examples
##' \donttest{ ## Reduce Ex.Timings##'
##' m1 <- lvm( x1+x2+x3 ~ u, latent= ~u)
##' m2 <- lvm( y ~ 1 )
##' m <- functional(merge(m1,m2), y ~ u, value=function(x) sin(x)+x)
##' distribution(m, ~u1) <- uniform.lvm(-6,6)
##' d <- sim(m,n=500,seed=1)
##' nonlinear(m2) <- y~u1
##' if (requireNamespace('mets', quietly=TRUE)) {
##'   set.seed(1)
##'   val <- twostageCV(m1, m2, data=d, std.err=FALSE, df=2:6, nmix=1:2,
##'                   nfolds=2)
##'   val
##' }
##' }
twostageCV <- function(model1, model2, data, control1=list(trace=0), control2=list(trace=0),
                knots.boundary, nmix=1:4, df=1:9, fix=TRUE, std.err=TRUE,
                nfolds=5, rep=1, messages=0, ...) {

    F <- nonlinear(model2)
    if (length(F)==0) {
        stop("Specify at least one nonlinear association in 'model2' (using the 'nonlinear' method)")
    }
    form <- as.formula(paste(names(F)[1], "~", x=F[[1]]$x))

    op <- options(warn=-1)
    if (fix) {
        model1 <- baptize(fixsome(model1, param="relative"))
        intercept(model1, latent(model1)) <- NA
    }
    e1a <- estimate(model1, data=data, control=control1)
    if (missing(knots.boundary))
        knots.boundary <- range(predict(e1a,vars(e1a)))
    ## Starting values for mixture models
    plab <- parlabels(model1)
    pfree <- setdiff(coef(model1),plab)
    pfree.idx <- match(pfree,coef(model1)) ## Index of unlabeled parameters
    intpos <- setdiff(parpos(model1)$v,0)
    pfree.int <- intersect(pfree.idx,intpos) ## Free intercept parameters
    p0 <- coef(e1a)
    pint <- p0[setdiff(intpos,pfree.int)]
    startf <- function(n) {
        u0 <- seq(knots.boundary[1],knots.boundary[2],length.out=n); names(u0) <- paste0("p",seq_along(u0))
        c(pint,u0,p0[-intpos])
    }
    control1$start <- NULL
    ee <- list(e1a)

    nmix <- setdiff(nmix, 1)
    fit <- function(k, ...) {
        if (messages > 0) cat("Fitting mixture model with", k, "components\n")
        mixture(model1, k = k, data = data, control = c(control1, list(start = startf(k))))
    }
    future.args <- list(...)
    if (is.null(future.args$future.seed)) {
      future.args$future.seed <- TRUE
    }
    args <-  c(list(FUN = fit, c(as.list(nmix)), future.args))
    val <- do.call(future_lapply, args)
    ee <- c(ee, val)
    AIC1 <- unlist(lapply(ee,AIC))
    names(AIC1) <- c(1,nmix)
    ii <- which.min(AIC1)
    ## Exposure measurement model
    e1 <- ee[[ii]] ## Selected model by AIC

    MM <- list(nonlinear(model2, form, type="linear"))
    df <- setdiff(df, 1)
    Knots <- list()
    for (i in df) {
        knots <- seq(knots.boundary[1],knots.boundary[2],length.out=i+1)
        Knots  <- c(Knots, list(knots))
        MM <- c(MM,
               list(nonlinear(model2, form, type="spline", knots=knots)))
    }
    if (!inherits(e1, "lvm.mixture")) {
        f0 <- function(data) list(e0=estimate(model1,data=data,control=c(control1,list(start=coef(e1)))))
    } else {
        f0 <- function(data) list(e0=mixture(model1,data=data,k=e1$k,control=c(control1,list(start=coef(e1)))))
    }

    ff <- lapply(MM,
                 function(m) function(data,e0,...)  twostage(e0,m,data=data,std.derr=FALSE))
    a <- cv(ff, data=data, K=nfolds, rep=rep, shared=f0)
    M <- MM[[which.min(a$coef)]]
    e2 <- twostage(e1, M, data, control=control2, std.err=std.err)
    options(op)
    res <- list(AIC1=cbind(AIC1), model1=e1, model2=e2, cv=a$coef, knots=c(list(NA), Knots),
                nfolds=nfolds, rep=rep)
    structure(res, class="twostageCV")
}

##' @export
print.twostageCV <- function(x,...) {
    print(cli::rule())
    i1 <- which.min(x$AIC1)
    nmix <- rownames(x$AIC1)[i1]
    cat("Selected mixture model: ",nmix," component", ifelse(i1>1, "s",""),"\n", sep="")
    print(x$AIC1)
    print(cli::rule())
    i2 <- which.min(x$cv)
    splinedf <- unlist(lapply(x$knots,function(x) if (any(is.na(x))) return(1) else length(x)-1))
    cat("Selected spline model degrees of freedom: ", splinedf[i2] ,"\n", sep="")
    knots <- rbind(x$knots[[i2]])
    if (is.na(knots[1])) knots <- "none"
    cat("Knots:", paste(formatC(knots,...) , collapse=" "), "\n\n")
    rmse <- x$cv
    rownames(rmse) <- paste0("df:",splinedf)
    colnames(rmse) <- paste0("RMSE(nfolds=",x$nfolds,", rep=",x$rep,")")
    print(rmse)
    print(cli::rule())
    cat("\n")
    if (!inherits(x$model2, "lvmfit")) {
        print(x$model2$coefmat, quote=FALSE)
    } else {
        print(CoefMat(x$model2,...),quote=FALSE)
    }
}

##' @export
coef.twostageCV <- function(object,...) {
    coef(Model(object),...)
}

##' @export
vcov.twostageCV <- function(object,...) {
    vcov(Model(object),...)
}

##' @export
IC.twostageCV <- function(x,...) {
    IC(Model(x),...)
}

##' @export
Model.twostageCV <- function(x,...) {
    x$model2
}

##' @export
summary.twostageCV <- function(object,...) {
    res <- with(object, list(model1=summary(model1),
                                AIC1=AIC1, cv=cv, knots=knots,
                                model2=summary(model2)))
    structure(res, class="summary.twostageCV")
}

##' @export
print.summary.twostageCV <- function(x,...) {
    print.twostageCV(x, ...)
}

##' @export
predict.twostageCV <- function(object,... ) {
    predict(Model(object),...)
}

##' Apply a Function to a Data Frame Split by Factors
##'
##' Simple wrapper of the 'by' function
##' @title Apply a Function to a Data Frame Split by Factors
##' @param x Data frame
##' @param INDICES Indices (vector or list of indices, vector of column names, or formula of column names)
##' @param FUN A function to be applied to data frame subsets of 'data'.
##' @param COLUMNS (Optional) subset of columns of x to work on
##' @param array if TRUE an array/matrix is always returned
##' @param ... Additional arguments to lower-level functions
##' @author Klaus K. Holst
##' @export
##' @examples
##' By(datasets::CO2,~Treatment+Type,colMeans,~conc)
##' By(datasets::CO2,~Treatment+Type,colMeans,~conc+uptake)
By <- function(x,INDICES,FUN,COLUMNS,array=FALSE,...) {
    if (inherits(INDICES,"formula")) {
        INDICES <- as.list(model.frame(INDICES,x))
    } else {
        if (is.character(INDICES) && length(INDICES)!=nrow(x)) {
            INDICES <- as.list(x[,INDICES,drop=FALSE])
        }
    }
    if (!missing(COLUMNS)) {
        if (inherits(COLUMNS,"formula")) {
            x <- model.frame(COLUMNS,x)
        } else {
            x <- x[,COLUMNS,drop=FALSE]
        }
    }
    a <- by(x, INDICES, FUN=FUN, ...)
    if (NCOL(x)==1 && !array) {
        ##DimElem <- length(a[rep(1,length(dim(a)))][[1]])
        a <- a[]
        attr(a,"call") <- NULL
    }
    return(a)
}

strip_bracket <- function(s) {
  if (length(s) > 1L) {
    return(unlist(lapply(s, function(x) strip_bracket(x))))
  }
  # Check if it starts and ends with brackets
  if (!grepl("^\\[.*\\]$", s)) return(s)

  # Calculate cumulative bracket depth
  chars <- strsplit(s, "")[[1]]
  depth <- 0
  is_wrapped <- TRUE

  for (i in 1:(length(chars) - 1)) {
    if (chars[i] == "[") depth <- depth + 1
    if (chars[i] == "]") depth <- depth - 1

    # If depth hits 0 before the end, it's side-by-side like [a] + [b]
    if (depth == 0 && i < length(chars)) {
      is_wrapped <- FALSE
      break
    }
  }

  if (is_wrapped) {
    return(sub("^\\[(.*)\\]$", "\\1", s))
  } else {
    return(s)
  }
}

substArg <- function(x,env,...) {
  if (!missing(env)) {
    a <- with(env,substitute(x))
  } else {
    a <- substitute(x)
  }
  myclass <- tryCatch(class(eval(a)),error=function(e) NULL)
  if (is.null(myclass) || myclass=="name") {
      res <- unlist(sapply(as.character(a),
                           function(z) {
                               trimmed <- gsub(" ","",z,fixed=TRUE)
                               val <- strsplit(trimmed,"+",fixed=TRUE)
                               if (val[1]=="") val <- NULL
                               val
                           })); attributes(res)$names <- NULL
      return(res)
  }
  return(eval(a))
}

procrandomslope <- function(object,data=object$data,...) {
  Xfix <- FALSE
  xfix <- myfix <- list()
  xx <- object
  for (i in seq_len(object$ngroup)) {
    x0 <- object$lvm[[i]]
    data0 <- data[[i]]
    xfix0 <- colnames(data0)[(colnames(data0)%in%parlabels(x0,exo=TRUE))]
    xfix <- c(xfix, list(xfix0))
    if (length(xfix0)>0) { ## Yes, random slopes
      Xfix<-TRUE
    }
    xx$lvm[[i]] <- x0
  }
  if (Xfix) {
    for (k in seq_len(object$ngroup)) {
      x0 <- object$lvm[[k]]
      data0 <- data[[k]]
      nrow <- length(vars(x0))
      xpos <- lapply(xfix[[k]],function(y) which(regfix(x0)$labels==y))
      colpos <- lapply(xpos, function(y) ceiling(y/nrow))
      rowpos <- lapply(xpos, function(y) (y-1)%%nrow+1)
      myfix0 <- list(var=xfix[[k]], col=colpos, row=rowpos)
      myfix <- c(myfix, list(myfix0))
      for (i in seq_along(myfix0$var))
        for (j in seq_along(myfix0$col[[i]]))
          regfix(x0,
                 from=vars(x0)[myfix0$row[[i]][j]],to=vars(x0)[myfix0$col[[i]][j]]) <-
                   colMeans(data0[,myfix0$var[[i]],drop=FALSE],na.rm=TRUE)
      index(x0) <- reindex(x0,zeroones=TRUE,deriv=TRUE)
      object$lvm[[k]] <- x0
    }
    object <- multigroup(object$lvm,data,fix=FALSE,exo.fix=FALSE)
  }
  return(list(model=object,fix=myfix))
}

## ' Calculate matrix product with kronecker product
## '
## ' \deqn{(A\crossprod B) Y}
## ' @title Calculate matrix product with kronecker product
## ' @param A
## ' @param B
## ' @param Y
## ' @author Klaus K. Holst
kronprod <- function(A,B,Y) {
    if (missing(Y)) {
        ## Assume 'B'=Identity, (A otimes B)Y
        k <- nrow(B)/ncol(A)
        res <- rbind(apply(B,2,function(x) matrix(x,nrow=k)%*%t(A)))
        return(res)
    }
    rbind(apply(Y,2,function(x) B%*%matrix(x,nrow=ncol(B))%*%t(A)))
}


izero <- function(i,n) { ## n-1 zeros and 1 at ith entry
  x <- rep(0,n); x[i] <- 1
  x
}

`Debug` <-
  function(msg, cond=lava.options()$debug) {
    if (cond)
      print(paste(msg, collapse=" "))
  }


categorical2dummy <- function(x,data,messages=0,...) {
  x0 <- x
  X <- intersect(index(x)$exogenous,colnames(data))
  catX <- c()
  for (i in X) {
    if (!is.numeric(data[,i])) catX <- c(catX,i)
  }
  if (length(catX)==0) return(list(x=x,data=data))
  f <- as.formula(paste("~ 1+", paste(catX,collapse="+")))
  opt <- options(na.action="na.pass")
  M <- model.matrix(f,data)

  options(opt)
  Mnames <- colnames(M)
  Mpos <- attributes(M)$assign
  A <- index(x)$A
  F <- regfix(x)
  count <- 0
  for (i in catX) {
    count <- count+1
    mnames <- Mnames[Mpos==count]
    kill(x0) <- i
    Y <- colnames(A)[A[i,]==1]
    if (length(mnames)==1) {
      fix <- as.list(F$labels[i,])
      fixval <- F$values[i,]
      fix[which(!is.na(fixval))] <- fixval[na.omit(fixval)]
      regression(x0,to=Y,from=mnames,messages=messages) <- fix[Y]
    } else {
      x0 <- regression(x0,to=Y,from=mnames,messages=messages)
    }
  }
  index(x0) <- reindex(x0,zeroones=TRUE,deriv=TRUE)
  return(list(x=x0,data=cbind(data,M)))
}

`procdata.lvm` <-
  function(x,data,categorical=FALSE,
    na.method=ifelse(any(is.na(data[,intersect(colnames(data),manifest(x))])),"complete.obs","pairwise.complete.obs"),
    missing=FALSE
    ) {
    if (is.numeric(data) & !is.list(data)) {
      data <- rbind(data)
    }
     if (is.data.frame(data) | is.matrix(data)) {
      nn <- colnames(data)
      data <- as.data.frame(data); colnames(data) <- nn; rownames(data) <- NULL
      obs <- setdiff(intersect(vars(x), colnames(data)),latent(x))
      Debug(obs)
      mydata <- subset(data, select=obs)
      if (NROW(mydata)==0) stop("No observations")
      for (i in seq_len(ncol(mydata))) {
        if (inherits(mydata[,i],"Surv"))
          mydata[,i] <- mydata[,i][,1]
        if (is.character(mydata[,i]) | is.factor(mydata[,i]))
          mydata[,i] <- as.numeric(as.factor(mydata[,i]))-1
      }

      S <- NULL
      n <- nrow(mydata)
      if (n==1) {
        S <- diag(nrow=ncol(mydata)); colnames(S) <- rownames(S) <- obs
      }
      if (na.method=="complete.obs" && !missing) {
        mydata0 <- na.omit(mydata)
        n <- nrow(mydata0)
        mu <- colMeans(mydata0)
        if (is.null(S) && n>2)
            S <- (n-1)/n*cov(mydata0) ## MLE variance matrix of observed variables
        rm(mydata0)
      }
      nS <- is.null(S) || any(is.na(S))
      if (na.method=="pairwise.complete.obs" || nS) {
          mu <- colMeans(mydata,na.rm=TRUE)
          if (nS) {
              n <- nrow(mydata)
              S <- (n-1)/n*cov(mydata,use="pairwise.complete.obs")
              S[is.na(S)] <- 1e-3
          }
      }
    }
    else
      if (is.list(data)) {
        if ("cov"%in%names(data)) data$S <- data$cov
        if ("var"%in%names(data)) data$S <- data$var
        if ("mean"%in%names(data)) data$mu <- data$mean
        n <- data$n
        S <- reorderdata.lvm(x,data$S)
        mu <- reorderdata.lvm(x,data$mu)
        ##      if (is.null(n)) stop("n was not specified");
      }
      else
        stop("Unexpected type of data!");
    if (nrow(S)!=ncol(S)) stop("Wrong type of data!");
    return(list(S=S,mu=mu,n=n))
  }

`reorderdata.lvm` <-
  function(x, data) {
    if (is.vector(data)) {
      nn <- names(data)
      ii <- na.omit(match(index(x)$manifest, nn))
      data[ii,drop=FALSE]
    } else {
      nn <- colnames(data)
      ii <- na.omit(match(index(x)$manifest, nn))
      data[ii,ii,drop=FALSE]
    }
  }

`symmetrize` <-
function(M, upper=TRUE) {
  if (length(M)==1) return(M)
  if (!is.matrix(M) | ncol(M)!=nrow(M)) stop("Only implemented for square matrices.")
  if (upper) {
    for (i in seq_len(ncol(M)-1))
      for (j in seq(i+1,nrow(M)))
        M[i,j] <- M[j,i]
    return(M)
  } else {
    for (i in seq_len(ncol(M)))
      for (j in seq_len(nrow(M)))
        if (M[i,j]==0)
          M[i,j] <- M[j,i]
        else
          M[j,i] <- M[i,j]
    return(M)
  }
}

naiveGrad <- function(f, x, h=1e-9) {
  nabla <- numeric(length(x))
  for (i in seq_along(x)) {
    xh <- x; xh[i] <- x[i]+h
    nabla[i] <- (f(xh)-f(x))/h
  }
  return(nabla)
}

# conditional on Compl(idx)
CondMom <- function(mu,S,idx,X) {
  idxY <- idx

  idxX <- setdiff(seq_len(ncol(S)),idxY)
  SXX <- S[idxX,idxX,drop=FALSE];
  SYY <- S[idxY,idxY,drop=FALSE]
  SYX <- S[idxY,idxX,drop=FALSE]
  iSXX <- solve(SXX)
  condvar <- SYY-SYX%*%iSXX%*%t(SYX)
  if (missing(mu)) return(condvar)

  muY <- mu[,idxY,drop=FALSE]
  muX <- mu[,idxX,drop=FALSE]
  if (is.matrix(mu))
    Z <- t(X-muX)
  else
    Z <- apply(X,1,function(xx) xx-muX)
  SZ  <- t(SYX%*%iSXX%*%Z)
##  condmean <- matrix(
  if (is.matrix(mu))
    condmean <- SZ+muY
  else
    condmean <- t(apply(SZ,1,function(x) muY+x))
##  ,ncol=ncol(SZ),nrow=nrow(SZ))
  return(list(mean=condmean,var=condvar))
}

## Depth-First/acc (accessible)
DFS <- function(M,v,explored=c()) {
  explored <- union(explored,v)
  incident <- M[v,]
  for (v1 in setdiff(which(incident==1),explored)) {
    explored <- DFS(M,v1,explored)
  }
  return(explored)
}

## Depth-First/acc (accessible)
acc <- function(M,v) {
  if (is.character(v)) v <- which(colnames(M)==v)
  colnames(M)[setdiff(DFS(M,v),v)]
}

npar.lvm <- function(x) {
  return(index(x)$npar+ index(x)$npar.mean+index(x)$npar.ex)

}

##' @export
as.numeric.list <- function(x, ...) {
  lapply(x, function(y) {
    ifelse(is.na(as.numeric(y)), y, as.numeric(y))
  })
}

edge2pair <- function(e) {
  sapply(e, function(x) strsplit(x, "~"))
}

numberdup <- function(xx) { ## Convert to numbered list
  dup.xx <- duplicated(xx)
  ## dups <- xx[dup.xx]
  xx.new <- numeric(length(xx))
  count <- 0
  for (i in seq_along(xx)) {
    if (!dup.xx[i]) {
      count <- count+1
      xx.new[i] <- count
    } else {
      xx.new[i] <- xx.new[match(xx[i],xx)[1]]
    }
  }
  return(xx.new)
}

extractvar <- function(f) {
    yy <- getoutcome(f)
    xx <- attributes(terms(f))$term.labels
    myvars <- all.vars(f)
    return(list(y=yy,x=xx,all=myvars))
}

##' @export
getoutcome <- function(formula,sep,...) {
  aa <- attributes(terms(formula,...))
  if (aa$response==0) {
    res <- NULL
  } else {
    res <- paste(deparse(formula[[2]]),collapse="")
  }
  if (!missing(sep) && length(aa$term.labels)>0) {
      attributes(res)$x <- lapply(strsplit(aa$term.labels,"\\|")[[1]],
                                  function(x) as.formula(paste0("~",x)))
  } else {
      attributes(res)$x <- aa$term.labels
  }
  return(res)
}

##' @export
Specials <- function(f,spec,split2="+",...) {
  tt <- terms(f,spec)
  pos <- attributes(tt)$specials[[spec]]
  if (is.null(pos)) return(NULL)
  x <- rownames(attributes(tt)$factors)[pos]
  st <- gsub(" ","",x)
  res <- unlist(strsplit(st,"[()]"))[2]
  if (is.null(split2)) return(res)
  unlist(strsplit(res,split2,fixed=TRUE))
}

##' @export
decomp.specials <- function(x,pattern="[()]",pattern2=NULL, pattern.ignore=NULL, sep="[,\\+]",perl=TRUE,reverse=FALSE,...) {
  st <- gsub(" |^\\(|)$","",x) # Remove white space and leading/trailing parantheses
  if (!is.null(pattern.ignore)) {
      if (grepl(pattern.ignore,st,perl=perl,...)) return(st)
  }
  if (!is.null(pattern)) {
    st <- rev(unlist(strsplit(st,pattern,perl=perl,...)))[1]
  }
  if (!is.null(pattern2)) {
    st <- (unlist(strsplit(st,pattern2,perl=perl,...)))
    if (reverse) st <- rev(st)
  }
  unlist(strsplit(st,sep,perl=perl,...))
}

## Check that IC columns have empirical mean zero (relative to RMS)
check_ic_mean_zero <- function(ic, tol = lava.options()$check.ic.tol) {
  ic <- cbind(ic)
  n <- nrow(ic)
  if (is.null(n) || n < 2L) return(invisible(NULL))
  m <- colMeans(ic, na.rm = TRUE)
  rms <- sqrt(colMeans(ic^2, na.rm = TRUE))
  ratio <- abs(m) / rms
  flag <- any(ratio > tol, na.rm=TRUE)
  if (flag) {
    warning(
      sprintf(
        "IC does not have mean zero (max |mean|/rms = %.2g). %s",
        max(ratio),
        "Using lava.options(check.ic = FALSE) disables the warning globally."
      ))
  }
  return(invisible(flag))
}

center_ic <- function(x) scale(x, scale=FALSE, center=TRUE)

##' Performs Likelihood-ratio, Wald and score tests
##' @title Statistical tests
##' @aliases compare test_wald
##' @export
##' @param object \code{lvmfit}-object
##' @param \dots Additional arguments to low-level functions
##' @return Matrix of test-statistics and p-values
##' @author Klaus K. Holst
##' @seealso \code{\link{modelsearch}}, \code{\link{equivalence}}
##' @keywords htest
##' @examples
##' m <- lvm();
##' regression(m) <- c(y1,y2,y3) ~ eta; latent(m) <- ~eta
##' regression(m) <- eta ~ x
##' m2 <- regression(m, c(y3,eta) ~ x)
##' set.seed(1)
##' d <- sim(m,1000)
##' e <- estimate(m,d)
##' e2 <- estimate(m2,d)
##'
##' compare(e)
##'
##' compare(e,e2) ## LRT, H0: y3<-x=0
##' compare(e,scoretest=y3~x) ## Score-test, H0: y3~x=0
##' compare(e2,par=c("y3~x")) ## Wald-test, H0: y3~x=0
##'
##' B <- diag(2); colnames(B) <- c("y2~eta","y3~eta")
##' compare(e2,contrast=B,null=c(1,1))
##'
##' B <- rep(0,length(coef(e2))); B[1:3] <- 1
##' compare(e2,contrast=B)
##'
##' compare(e,scoretest=list(y3~x,y2~x))
compare <- function(object,...) UseMethod("compare")

##' @export
compare.default <- function(object,...,
                            par,contrast,null,
                            scoretest,Sigma,level=.95,df=NULL) {
    if (!missing(par) || (!missing(contrast) && is.character(contrast))) {
        if (!missing(contrast) && is.character(contrast)) par <- contrast
        contrast <- rep(0,length(coef(object)))
        myidx <- parpos(Model(object),p=par)
        contrast[myidx] <- 1
        contrast <- diag(contrast,nrow=length(contrast))[which(contrast!=0),,drop=FALSE]
        if (!missing(null) && length(null)>1) null <- null[attributes(myidx)$ord]
    }
    ### Wald test
    if (!missing(contrast)) {
        B <- contrast
        p <- coef(object)
        pname <- names(p)
        B <- rbind(B);
        colnames(B) <- if (is.vector(contrast)) names(contrast) else colnames(contrast)
        if (missing(Sigma)) {
            Sigma <- vcov(object)
        }
        if (ncol(B)<length(p)) {
            nn <- colnames(B)
            myidx <- parpos(Model(object),p=nn)
            B0 <- matrix(0,nrow=nrow(B),ncol=length(coef(object)))
            B0[,myidx] <- B[,attributes(myidx)$ord]
            B <- B0
        }
        if (missing(null)) null <- rep(0,nrow(B))
        if (length(null)==1) null <- rep(null,nrow(B))
        Bp <- B%*%p
        V <- B%*%Sigma%*%t(B)
        ct <- cbind(Bp,diag(V)^.5)
        p <- 1-(1-level)/2
        qp <- if(!is.null(df)) qt(p,df=df) else qnorm(p)
        ct <- cbind(ct,ct[,1] + qp*cbind(-1,1)%x%ct[,2])
        colnames(ct) <- c("Estimate","Std.Err",paste0(c(1-p,p)*100,"%"))
        rownames(ct) <- rep("",nrow(ct))
        Q <- t(Bp-null)%*%Inverse(V)%*%(Bp-null)
        if (any(is.nan(V)) ||any(is.na(V))) {
          df <- 0
        } else {
          df <- tryCatch(qr(V)$rank, error=function(...) 0)
        }
        names(df) <- "df"
        attributes(Q) <- NULL; names(Q) <- "chisq";
        pQ <- ifelse(df==0,NA,pchisq(Q,df,lower.tail=FALSE))

        method = "- Wald test -";
        cnames <- c()
        if (!is.null(pname)) {
            msg <- c()
            for (i in seq_len(nrow(B))) {
                Bidx <- which(B[i,]!=0)
                Bval <- abs(B[i,Bidx]); Bval[Bval==1] <- ""
                sgn  <- rep(" + ",length(Bval)); sgn[sign(B[i,Bidx])==-1] <- " - ";
                if (sgn[1] == " + ") {
                  sgn[1] <- ""
                } else {
                    sgn[1] <- "-"
                }
                cnames <- c(cnames,
                            paste0(sgn,Bval,
                                   paste0("[",pname[Bidx],"]"),collapse=""))
                msg <- c(msg,paste0(cnames[i]," = ",null[i]))
            }
            method <- c(method,"","Null Hypothesis:",msg)
            ##method <- c(method,"","Observed:",paste(formatC(as.vector(Bp)),collapse=" "))
        }
        rownames(ct) <- cnames
        res <- list(##data.name=hypothesis,
            statistic = Q, parameter = df,
            p.value=pQ, method = method, estimate=ct, vcov=V, coef=ct[,1],
            null=null, cnames=cnames
        )
        class(res) <- "htest"
        attributes(res)$B <- B
        return(res)
    }

    ### Score test
    if (!missing(scoretest)) {
        altmodel <- Model(object)
        if (inherits(scoretest,"formula")) scoretest <- list(scoretest)
        for (i in scoretest) {
            regression(altmodel) <- i
        }
        p0 <- numeric(length(coef(altmodel)))
        idx <-  match(coef(Model(object)),coef(altmodel))
        p0[idx] <- coef(object)
        Sc2 <- score(altmodel,p=p0,data=model.frame(object),weigth=Weights(altmodel),
                     estimator=object$estimator,...)
        I <- information(altmodel,p=p0,n=object$data$n,
                         data=model.frame(object),weigth=Weights(object),
                         estimator=object$estimator,...
                         )
        iI <- try(solve(I), silent=TRUE)
        Q <- ifelse (inherits(iI, "try-error"), NA, ## Score test
                     ## rbind(Sc)%*%iI%*%cbind(Sc)
        (Sc2)%*%iI%*%t(Sc2)
        )
        attributes(Q) <- NULL; names(Q) <- "chisq"
        df <- length(p0)-length(coef(object)); names(df) <- "df"
        pQ <- ifelse(df==0,NA,pchisq(Q,df,lower.tail=FALSE))
        res <- list(data.name=as.character(scoretest),
                    statistic = Q, parameter = df,
                    p.value=pQ, method = "- Score test -")
        class(res) <- "htest"
        return(res)
    }

    ### Likelihood ratio test
    objects <- list(object,...)
    if (length(objects)<2) {
        if (!(inherits(object,"lvmfit"))) {
            cc <- rbind(logLik(object),AIC(object))
            rownames(cc) <- c("logLik","AIC")
            colnames(cc) <-  ""
            return(cc)
        }
        L0 <- logLik(object)
        L1 <- satmodel(object,logLik=TRUE)
        df <- attributes(L1)$df-attributes(L0)$df; names(df) <- "df"
        Q <- abs(2*(L0-L1));
        attributes(Q) <- NULL; names(Q) <- "chisq";
        pQ <- ifelse(df==0,NA,pchisq(Q,df,lower.tail=FALSE))

        values <- c(L0,L1); names(values) <- c("log likelihood (model)", "log likelihood (saturated model)")
        res <- list(statistic = Q, parameter = df,
                    p.value=pQ, method = "- Likelihood ratio test -",
                    estimate = values)
        class(res) <- "htest"
        return(res)
    }
    if (length(objects)==2)
        return(comparepair(objects[[1]],objects[[2]]))
    res <- list()
    for (i in seq_len(length(objects)-1)) {
        res <- c(res, list(comparepair(objects[[i]],objects[[i+1]])))
    }
    return(res)
}


comparepair <- function(x1,x2) {
    l1 <- do.call("logLik",list(x1),envir=parent.frame(2))
    l2 <- do.call("logLik",list(x2),envir=parent.frame(2))
    df1 <- attributes(l1)$df;  df2 <- attributes(l2)$df;
    if (is.null(df1)) {
        df1 <- length(do.call("coef",list(x1),envir=parent.frame(2)))
        df2 <- length(do.call("coef",list(x2),envir=parent.frame(2)))
    }
    Q <- abs(2*(l1-l2))
    names(Q) <- "chisq"
    df <- abs(df1-df2); names(df) <- "df"
    p <- pchisq(Q,df=df,lower.tail=FALSE)
    values <- c(l1,l2); names(values) <- c("log likelihood (model 1)", "log likelihood (model 2)")

    res <- list(statistic = Q, parameter = df,
                p.value= p, method = "- Likelihood ratio test -",
                estimate = values)
    class(res) <- "htest"
    return(res)
}

##' Missing value generator
##'
##' This function adds a binary variable to a given \code{lvm} model
##' and also a variable which is equal to the original variable where
##' the binary variable is equal to zero
##'
##' @title Missing value generator
##' @param object  \code{lvm}-object.
##' @param formula The right hand side specifies the name of a latent
##' variable which is not always observed. The left hand side
##' specifies the name of a new variable which is equal to the latent
##' variable but has missing values.  If given as a string then this
##' is used as the name of the latent (full-data) name, and the
##' observed data name is 'missing.data'
##' @param Rformula Missing data mechanism with left hand side
##' specifying the name of the observed data indicator (may also just
##' be given as a character instead of a formula)
##' @param missing.name Name of observed data variable (only used if
##' 'formula' was given as a character specifying the name of the
##' full-data variable)
##' @param suffix If missing.name is missing, then the name of the
##' oberved data variable will be the name of the full-data variable +
##' the suffix
##' @param ... Passed to binomial.lvm.
##' @return lvm object
##' @aliases Missing, Missing<-
##' @examples
##' library(lava)
##' set.seed(17)
##' m <- lvm(y0~x01+x02+x03)
##' m <- Missing(m,formula=x1~x01,Rformula=R1~0.3*x02+-0.7*x01,p=0.4)
##' sim(m,10)
##'
##'
##' m <- lvm(y~1)
##' m <- Missing(m,"y","r")
##' ## same as
##' ## m <- Missing(m,y~1,r~1)
##' sim(m,10)
##'
##' ## same as
##' m <- lvm(y~1)
##' Missing(m,"y") <- r~x
##' sim(m,10)
##'
##' m <- lvm(y~1)
##' m <- Missing(m,"y","r",suffix=".")
##' ## same as
##' ## m <- Missing(m,"y","r",missing.name="y.")
##' ## same as
##' ## m <- Missing(m,y.~y,"r")
##' sim(m,10)
##'
##' @export
##' @author Thomas A. Gerds <tag@@biostat.ku.dk>
Missing <- function(object,formula,Rformula,missing.name,suffix="0",...){
    if (is.character(Rformula)) {
        indicatorname <- Rformula
        Rformula <- toformula(Rformula,1)
    } else {
        indicatorname <- all.vars(Rformula)[1]
    }
    if (length(all.vars(formula))==1) formula <- all.vars(formula)
    if (is.character(formula)) {
        if (missing(missing.name)) missing.name <- paste0(formula,suffix)
        formula <- toformula(missing.name,formula)
    }
    newf <- update(formula,paste(".~.+",indicatorname))
    if (is.null(distribution(object,indicatorname)[[1]]) || length(list(...))>0) {
        distribution(object,indicatorname) <- binomial.lvm(...)
    }
    transform(object,newf) <- function(u){
        out <- u[,1]
        out[u[,2]==0] <- NA
        out
    }
    regression(object) <- Rformula
    object
}

##' @export
"Missing<-" <- function(object,formula,...,value) {
    Missing(object,formula,value,...)
}

##' Split data into folds
##'
##' @title Split data into folds
##' @param x Data or integer (size)
##' @param p Number of folds, or if a number between 0 and 1 is given two folds of size p and (1-p) will be returned
##' @param replace With or with-out replacement
##' @param return.index If TRUE index of folds are returned otherwise the actual data splits are returned (default)
##' @param k (Optional, only used when p=NULL) number of folds without shuffling
##' @param ... additional arguments to lower-level functions
##' @export
##' @aliases csplit foldr
##' @examples
##' foldr(5,2,rep=2)
##' csplit(10,3)
##' csplit(iris[1:10,]) ## Split in two sets 1:(n/2) and (n/2+1):n
##' csplit(iris[1:10,],0.5)
##' @author Klaus K. Holst
csplit <- function(x, p=NULL, replace=FALSE,
                   return.index=FALSE, k=2, ...) {
  if (length(x)==1 && is.numeric(x))
    x <- seq(x)
  N <- NROW(x)
  if (is.null(p)) { ##
    K <- base::round(N/k)
    idx <- split(seq(N), sort(rep(seq(k), length.out=N, each=K)))
  } else {
    if (p<1) { ## two folds (size N*p and N*(1-p))

      idx1 <- base::sample(N, base::round(p*N), replace=replace)
      idx <- list(idx1,
                  base::sample(setdiff(seq(N), idx1), replace=replace))
    } else { ## Number of folds (equal size)
      idx <- split(sample(seq(N)), rep(seq(p), length=N))
    }
  }
  if (return.index)
    return(idx)
  if (!is.vector(x)) {
    return(lapply(idx, function(ii) x[ii, , drop=FALSE]))
  }
  return(lapply(idx, function(ii) x[ii]))
}

##' @export
foldr <- function(n, K=5, rep=1, list=TRUE) {
  res <- replicate(rep, split(sample(seq(n)),
                              rep(seq(K), length=n)),
                   simplify=FALSE)
  if (!list) {
    ids <- rep(seq_len(length(res[[1]])),
           unlist(lapply(res[[1]], length)))
    res <- lapply(res, function(x) {
      ids[order(unlist(x))]
    })
    if (length(res)==1)
      res <- res[[1]]
  }
  return(res)
}

##' Add regression association to latent variable model
##'
##' Define regression association between variables in a \code{lvm}-object and
##' define linear constraints between model equations.
##'
##'
##' The \code{regression} function is used to specify linear associations
##' between variables of a latent variable model, and offers formula syntax
##' resembling the model specification of e.g. \code{lm}.
##'
##' For instance, to add the following linear regression model, to the
##' \code{lvm}-object, \code{m}:
##' \deqn{ E(Y|X_1,X_2) = \beta_1 X_1 + \beta_2 X_2}
##' We can write
##'
##' \code{regression(m) <- y ~ x1 + x2}
##'
##' Multivariate models can be specified by successive calls with
##' \code{regression}, but multivariate formulas are also supported, e.g.
##'
##' \code{regression(m) <- c(y1,y2) ~ x1 + x2}
##'
##' defines
##' \deqn{ E(Y_i|X_1,X_2) = \beta_{1i} X_1 + \beta_{2i} X_2 }
##'
##' The special function, \code{f}, can be used in the model specification to
##' specify linear constraints. E.g. to fix \eqn{\beta_1=\beta_2}
##' , we could write
##'
##' \code{regression(m) <- y ~ f(x1,beta) + f(x2,beta)}
##'
##' The second argument of \code{f} can also be a number (e.g. defining an
##' offset) or be set to \code{NA} in order to clear any previously defined
##' linear constraints.
##'
##' Alternatively, a more straight forward notation can be used:
##'
##' \code{regression(m) <- y ~ beta*x1 + beta*x2}
##'
##' All the parameter values of the linear constraints can be given as the right
##' handside expression of the assigment function \code{regression<-} if the
##' first (and possibly second) argument is defined as well. E.g:
##'
##' \code{regression(m,y1~x1+x2) <- list("a1","b1")}
##'
##' defines \eqn{E(Y_1|X_1,X_2) = a1 X_1 + b1 X_2}. The rhs argument can be a
##' mixture of character and numeric values (and NA's to remove constraints).
##'
##' The function \code{regression} (called without additional arguments) can be
##' used to inspect the linear constraints of a \code{lvm}-object.
##'
##' @aliases regression regression<- regression<-.lvm regression.lvm
##' @param object \code{lvm}-object.
##' @param value A formula specifying the linear constraints or if
##' \code{to=NULL} a \code{list} of parameter values.
##' @param to Character vector of outcome(s) or formula object.
##' @param from Character vector of predictor(s).
##' @param fn Real function defining the functional form of predictors (for
##' simulation only).
##' @param messages Controls which messages are turned on/off (0: all off)
##' @param additive If FALSE and predictor is categorical a non-additive effect is assumed
##' @param y Alias for 'to'
##' @param x Alias for 'from'
##' @param quick Faster implementation without parameter constraints
##' @param \dots Additional arguments to be passed to the low level functions
##' @usage
##' \method{regression}{lvm}(object = lvm(), to, from, fn = NA,
##' messages = lava.options()$messages, additive=TRUE, y, x, value, ...)
##' \method{regression}{lvm}(object, to=NULL, quick=FALSE, ...) <- value
##' @return A \code{lvm}-object
##' @note Variables will be added to the model if not already present.
##' @author Klaus K. Holst
##' @seealso \code{\link{intercept<-}}, \code{\link{covariance<-}},
##' \code{\link{constrain<-}}, \code{\link{parameter<-}},
##' \code{\link{latent<-}}, \code{\link{cancel<-}}, \code{\link{kill<-}}
##' @keywords models regression
##' @examples
##'
##' m <- lvm() ## Initialize empty lvm-object
##' ### E(y1|z,v) = beta1*z + beta2*v
##' regression(m) <- y1 ~ z + v
##' ### E(y2|x,z,v) = beta*x + beta*z + 2*v + beta3*u
##' regression(m) <- y2 ~ f(x,beta) + f(z,beta)  + f(v,2) + u
##' ### Clear restriction on association between y and
##' ### fix slope coefficient of u to beta
##' regression(m, y2 ~ v+u) <- list(NA,"beta")
##'
##' regression(m) ## Examine current linear parameter constraints
##'
##' ## ## A multivariate model, E(yi|x1,x2) = beta[1i]*x1 + beta[2i]*x2:
##' m2 <- lvm(c(y1,y2) ~ x1+x2)
##'
##' @export
"regression<-" <- function(object,...,value) UseMethod("regression<-")

##' @export
regression.formula <- function(object,...) regression(lvm(),object,...)

##' @export
"regression<-.lvm" <- function(object, to=NULL, quick=FALSE, ..., value) {
    dots <- list(...)
    if (length(dots$additive)>0 && !dots$additive && !inherits(value,"formula")) {
        regression(object,beta=value,...) <- to
        return(object)
    }
    if (!is.null(to) || !is.null(dots$y)) {
        regfix(object, to=to, ...) <- value
        return(object)
    } else  {
        if (is.list(value)) {
            for (v in value) {
                regression(object,...) <- v
            }
            return(object)
        }

        if (inherits(value,"formula")) {
            fff <- procformula(object,value,...)
            object <- fff$object
            lhs <- fff$lhs
            xs <- fff$xs
            ys <- fff$ys
            res <- fff$res
            X <- fff$X

            
        if (fff$iscovar) {
            ## return(covariance(object,var1=decomp.specials(lhs[[1]]),var2=X))
            covariance(object) <- toformula(decomp.specials(lhs[[1]]),X)
            return(object)
        }
        if (!is.null(lhs) && nchar(lhs[[1]])>2 && substr(lhs[[1]],1,2)=="v(") {
            v <- update(value,paste(decomp.specials(lhs),"~."))
            covariance(object,...) <- v
            return(object)
        }

        if (length(lhs)==0) {
            index(object) <- reindex(object)
            return(object)
        }

        for (i in seq_len(length(ys))) {
        y <- ys[i]
        for (j in seq_len(length(xs))) {
          if (length(res[[j]])>1) {
            regfix(object, to=y[1], from=xs[j],...) <- res[[j]][2]
          } else {
            object <- regression(object,to=y[1],from=xs[j],...)
          }
        }
      }
      object$parpos <- NULL
      return(object)
    }
    if (!is.list(value) | length(value)>2) stop("Value should contain names of outcome (to) and predictors (from)")
    if (all(c("to","from")%in%names(value))) {

      xval <- value$x; yval <- value$y
    } else {
      yval <- value[[1]]; xval <- value[[2]]
    }
    regression(object, to=yval, from=xval,...)
  }
}

##' @export
`regression` <-
  function(object,to,from,...) UseMethod("regression")

##' @export
`regression.lvm` <-
    function(object=lvm(),to,from,fn=NA,messages=lava.options()$messages,
             additive=TRUE, y,x,value,...) {
        if (!missing(y)) {
            if (inherits(y,"formula")) y <- all.vars(y)
            to <- y
        }
        if (!missing(x)) {
            if (inherits(x,"formula")) x <- all.vars(x)
            from <- x
        }
        if (!additive) {
            if (!inherits(to,"formula")) to <- toformula(to,from)
            x <- attributes(getoutcome(to))$x
            K <- object$attributes$nordinal[x]
            if (is.null(K) || is.na(K)) {
                K <- list(...)$K
                if (is.null(K)) stop("Supply number of categories, K (or use method 'categorical' before calling 'regression').")
                object <- categorical(object,x,...)
            }
            dots <- list(...);
            dots$K <- K
            dots$x <- object
            dots$formula <- to
            dots$regr.only <- TRUE
            object <- do.call("categorical",dots)
            return(object)
        }


        if (missing(to)) {
            return(regfix(object))
        }
        if ((missing(from) || is.null(from)) &
            is.character(to)) {
          to <- as.formula(paste(to, "~1"))
        }
        if (inherits(to,"formula")) {
          if (!missing(value)) {
                regression(object,to,messages=messages,...) <- value
            } else {
                regression(object,messages=messages,...) <- to
            }
            object$parpos <- NULL
            return(object)
        }
        if (is.list(to)) {
            for (t in to)
                regression(object,messages=messages,...) <- t
            object$parpos <- NULL
            return(object)
        }

        sx <- strsplit(from,"@")
        xx <- sapply(sx, FUN=function(i) i[1])
        ps <- sapply(sx, FUN=function(i) i[2])
        sx <- strsplit(xx,"$",fixed=TRUE)
        xs <- sapply(sx, FUN=function(i) i[1])
        fix <- char2num(sapply(sx, FUN=function(i) i[2]))
        allv <- index(object)$vars

        object <- addvar(object, c(to,xs), messages=messages, reindex=FALSE)

        for (i in to)
            for (j in xs) {
                object$M[j,i] <- 1
                if (!is.na(fn))
                    functional(object,j,i) <- fn
            }

        if (lava.options()$exogenous) {
            newexo <- setdiff(xs,c(to,allv))
            exo <- exogenous(object)
            if (length(newexo)>0)
                exo <- unique(c(exo,newexo))
            exogenous(object) <- setdiff(exo,to)
        }

        if (lava.options()$debug) {
            print(object$fix)
        }
        object$fix[xs,to] <- fix
        object$par[xs,to] <- ps
        object$parpos <- NULL

        index(object) <- reindex(object)
        return(object)
    }

##' Add covariance structure to Latent Variable Model
##'
##' Define covariances between residual terms in a \code{lvm}-object.
##'
##' The \code{covariance} function is used to specify correlation structure
##' between residual terms of a latent variable model, using a formula syntax.
##'
##' For instance, a multivariate model with three response variables,
##'
##' \deqn{Y_1 = \mu_1 + \epsilon_1}
##'
##' \deqn{Y_2 = \mu_2 + \epsilon_2}
##'
##' \deqn{Y_3 = \mu_3 + \epsilon_3}
##'
##' can be specified as
##'
##' \code{m <- lvm(~y1+y2+y3)}
##'
##' Pr. default the two variables are assumed to be independent. To add a
##' covariance parameter \eqn{r = cov(\epsilon_1,\epsilon_2)}, we execute the
##' following code
##'
##' \code{covariance(m) <- y1 ~ f(y2,r)}
##'
##' The special function \code{f} and its second argument could be omitted thus
##' assigning an unique parameter the covariance between \code{y1} and
##' \code{y2}.
##'
##' Similarily the marginal variance of the two response variables can be fixed
##' to be identical (\eqn{var(Y_i)=v}) via
##'
##' \code{covariance(m) <- c(y1,y2,y3) ~ f(v)}
##'
##' To specify a completely unstructured covariance structure, we can call
##'
##' \code{covariance(m) <- ~y1+y2+y3}
##'
##' All the parameter values of the linear constraints can be given as the right
##' handside expression of the assigment function \code{covariance<-} if the
##' first (and possibly second) argument is defined as well. E.g:
##'
##' \code{covariance(m,y1~y1+y2) <- list("a1","b1")}
##'
##' \code{covariance(m,~y2+y3) <- list("a2",2)}
##'
##' Defines
##'
##' \deqn{var(\epsilon_1) = a1}
##'
##' \deqn{var(\epsilon_2) = a2}
##'
##' \deqn{var(\epsilon_3) = 2}
##'
##' \deqn{cov(\epsilon_1,\epsilon_2) = b1}
##'
##' Parameter constraints can be cleared by fixing the relevant parameters to
##' \code{NA} (see also the \code{regression} method).
##'
##' The function \code{covariance} (called without additional arguments) can be
##' used to inspect the covariance constraints of a \code{lvm}-object.
##'
#
##'
##' @aliases covariance covariance<- covariance.lvm covariance<-.lvm
##' variance variance<- variance.lvm variance<-.lvm
##' @param object \code{lvm}-object
##' @param var1 Vector of variables names (or formula)
##' @param var2 Vector of variables names (or formula) defining pairwise
##' covariance between \code{var1} and \code{var2})
##' @param constrain Define non-linear parameter constraints to ensure positive definite structure
##' @param pairwise If TRUE and \code{var2} is omitted then pairwise correlation is added between all variables in \code{var1}
##' @param \dots Additional arguments to be passed to the low level functions
##' @param value List of parameter values or (if \code{var1} is unspecified)
##' @usage
##' \method{covariance}{lvm}(object, var1=NULL, var2=NULL, constrain=FALSE, pairwise=FALSE,...) <- value
##' @return A \code{lvm}-object
##' @author Klaus K. Holst
##' @seealso \code{\link{regression<-}}, \code{\link{intercept<-}},
##' \code{\link{constrain<-}} \code{\link{parameter<-}}, \code{\link{latent<-}},
##' \code{\link{cancel<-}}, \code{\link{kill<-}}
##' @keywords models regression
##' @export
##' @examples
##'
##' m <- lvm()
##' ### Define covariance between residuals terms of y1 and y2
##' covariance(m) <- y1~y2
##' covariance(m) <- c(y1,y2)~f(v) ## Same marginal variance
##' covariance(m) ## Examine covariance structure
##'
##'
`covariance` <- function(object,...) UseMethod("covariance")

##' @export
"variance<-" <- function(object,...,value) UseMethod("covariance<-")

##' @export
`variance` <- function(object,...) UseMethod("variance")

##' @export
"variance.lvm" <- function(object,...) covariance(object,...)

##' @export
"variance.formula" <- function(object,...) covariance(lvm(),object,...)

##' @export
"covariance.formula" <- function(object,...) covariance(lvm(),object,...)


##' @export
"variance<-.lvm" <- function(object,...,value) {
    covariance(object,...) <- value
    return(object)
}

##' @export
"covariance<-" <- function(object,...,value) UseMethod("covariance<-")

##' @export
"covariance<-.lvm" <- function(object, var1=NULL, var2=NULL, constrain=FALSE, pairwise=FALSE, ..., value) {

    if (!is.null(var1)) {
        if (inherits(var1,"formula")) {
            lhs <- getoutcome(var1)
            xf <- attributes(terms(var1))$term.labels
            xx <- unlist(lapply(xf, function(x) x[1]))
            if (length(lhs)==0) {
                covfix(object,var1,var2,pairwise=pairwise,...) <- value
                object$parpos <- NULL
                return(object)
            }
            else {
                yy <- decomp.specials(lhs)
            }
        } else {
            yy <- var1; xx <- var2
        }
        covfix(object,var1=yy,var2=xx,pairwise=pairwise,...) <- value
        object$parpos <- NULL
        return(object)
    }

    if (is.list(value)) {
        for (v in value) {
            covariance(object,pairwise=pairwise,constrain=constrain,...) <- v
        }
        return(object)
    }

    if (inherits(value,"formula")) {
        lhs <- getoutcome(value)
        if (length(lhs)==0) {
            return(covariance(object,all.vars(value),constrain=constrain,pairwise=pairwise,...))
        }
        yy <- decomp.specials(lhs)

        tt <- terms(value, specials=c("f","v"))
        xf <- attributes(terms(tt))$term.labels
        res <- lapply(xf,decomp.specials)
        nx <- length(xf)

        if (nx==1) {
            if(is.null(attr(tt,"specials")$f) | length(res[[1]])<2) {
                if(is.null(attr(tt,"specials")$v) & is.null(attr(tt,"specials")$f))
                {
                    for (i in yy)
                        for (j in res[[1]])
                            object <- covariance(object, c(i,j), pairwise=TRUE, constrain=constrain, ...)
                } else {
                    covfix(object,var1=yy,var2=NULL) <- res[[1]]
                }
            } else {
                covfix(object,var1=yy,var2=res[[1]][1]) <- res[[1]][2]
            }
            object$parpos <- NULL
            return(object)
        }

        xx <- unlist(lapply(res, function(z) z[1]))
        for (y in yy)
            for (i in seq_along(xx)) {
                if (length(res[[i]])>1) {
                    covfix(object, var1=y, var2=res[[i]][1]) <- res[[i]][2]
                } else if ((i+1)%in%attr(tt,"specials")$f | (i+1)%in%attr(tt,"specials")$v) {
                    covfix(object, var1=y, var2=NULL) <- res[[i]]
                } else {
                    object <- covariance(object,c(y,xx[i]),pairwise=TRUE,...)
                }
            }

        object$parpos <- NULL
        return(object)
    }
    else covariance(object,value,pairwise=pairwise,...)
}

##' @export
`covariance.lvm` <-
    function(object,var1=NULL,var2=NULL,exo=FALSE,pairwise=FALSE,constrain=FALSE,value,...) {

        if (!missing(value)) {
            covariance(object,var1=var1,var2,exo=exo,pariwise=pairwise,constrain=constrain,...) <- value
            return(object)
        }
        if (!is.null(var1)) {
            if (inherits(var1,"formula")) {
                covariance(object,constrain=constrain,
                           pairwise=pairwise,exo=exo,...) <- var1
                return(object)
            }
            allvars <- var1
            if (!missing(var2)) {
                if (inherits(var2,"formula"))
                    var2 <- all.vars(var2)
                allvars <- c(allvars,var2)
            }
            object <- addvar(object, allvars, messages=0, reindex=FALSE)

            if (constrain) {
                if (length(allvars)!=2) stop("Constraints only implemented for pairs")
                return(covarianceconst(object,allvars[1],allvars[2],...))
            }


            xorg <- exogenous(object)
            exoset <- setdiff(xorg,allvars)
            if (!exo & length(exoset)<length(xorg)) {
                exogenous(object) <- exoset
            }

            if (!missing(var2)) {
                for (i in seq_len(length(var1))) {
                    c1 <- var1[i]
                    for (j in seq_len(length(var2))) {
                        c2 <- var2[j]
                        object$cov[c1,c2] <- object$cov[c2,c1] <- 1
                        object$parpos <- NULL
                        index(object) <- reindex(object)
                    }
                }
            }
            else {
                if (pairwise) {
                    for (i in seq_len(length(var1))) {
                        c1 <- var1[i]
                        for (j in seq_len(length(var1))) {
                            c2 <- var1[j]
                            object$cov[c1,c2] <- object$cov[c2,c1] <- 1
                            object$parpos <- NULL
                            index(object) <- reindex(object)
                        }
                    }
                }
            }
            return(object)
        }
        else
            return(covfix(object))
    }

covarianceconst <- function(object,var1,var2,cname=NA,rname=NA,vname=NA,v2name=vname,lname=NA,l2name=lname,...) {
    if (inherits(var1,"formula")) {
        var1 <- getoutcome(var1)
        var2 <- attributes(var1)$x
    }
    curpar <- parlabels(object)
    if (is.na(cname)) {
        cname <- object$covpar[var1,var2]
    }

    if (is.na(v2name)) {
        v2name <- object$covpar[var2,var2]
    }
    if (is.na(vname)) {
        vname <- object$covpar[var1,var1]
    }

    nvarname <- c("rname","cname","vname","v2name","lname","l2name")[is.na(c(rname,cname,vname,v2name,lname,l2name))]

    nprefix <- sapply(nvarname, function(x) substr(x,1,1))
    for (i in seq_len(length(nvarname))) {
        count <- 0
        repeat {
            count <- count+1
            curname <- paste0(nprefix[i],count)
            if (!(curname%in%curpar)) break;
        }
        curpar <- c(curname,curpar)
        assign(nvarname[i],curname)
    }
    covariance(object,c(var1,var2)) <- c(vname,v2name)
    ff <- function(x) exp(x)
    constrain(object,vname,lname) <- ff
    if (vname!=v2name)
        constrain(object,v2name,l2name) <- ff
    covariance(object,var1,var2) <- cname
    cpar <- unique(c(lname,l2name,rname))
    constrain(object,cname,cpar) <- function(x) {
        prod(exp(x[seq(length(cpar)-1)]))^(1/(length(cpar)-1))*tanh(x[length(cpar)])
    }
    return(structure(object,rname=rname,cname=cname))
}

##' Generic method for adding variables to model object
##'
##' @title Add variable to (model) object
##' @param x Model object
##' @param \dots Additional arguments
##' @author Klaus K. Holst
##' @aliases addvar<-
##' @export
`addvar` <-
    function(x,...) UseMethod("addvar")

##' @export
`addvar<-` <-
    function(x,...,value) UseMethod("addvar<-")


##' @export
`addvar<-.lvm` <- function(x,...,value) {
    if (inherits(value,"formula")) {
        regression(x,...) <- value
        return(x)
    }
    addvar(x, var=value, ...)
}

##' @export
`addvar.lvm` <-
    function(x, var, messages=lava.options()$messages,reindex=TRUE,...) {
        new <- setdiff(var,c(vars(x),parameter(x)))
        k <- length(new)
        Debug(new)
        if (k>0) {
            if (lava.options()$sparse) {
                requireNamespace("Matrix",quietly=TRUE)
                newNA <- newM <- Matrix::Matrix(0,k,k)
                newNAc <- newNA; diag(newNAc) <- NA
                newcov <- Matrix::Diagonal(k)
            } else {
                newM <- matrix(0,k,k)
                newcov <- diag(k)
            }
            newNA <- matrix(NA,k,k)
            colnames(newM) <- rownames(newM) <-
                colnames(newcov) <- rownames(newcov) <-
                colnames(newNA) <- rownames(newNA) <- new
            newmean <- as.list(rep(NA,k))
            N <- nrow(x$cov)
            if (is.null(N)) {
                N <- 0
                x$M <- newM
                x$cov <- newcov; x$covfix <- x$fix <- x$par <- x$covpar <- newNA
                x$mean <- newmean
            } else {
                if (lava.options()$sparse) {
                    x$M <- Matrix::bdiag(x$M, newM) ## Add regression labels.R
                    x$cov <- Matrix::bdiag(x$cov, newcov) ## Add covariance
                    x$par <- Matrix::bdiag(x$par, newNA) ## Add regression labels
                    x$covpar <- Matrix::bdiag(x$covpar, newNA) ## Add covariance labels
                    x$fix <- Matrix::bdiag(x$fix, newNA)
                    x$covfix <- Matrix::bdiag(x$covfix,  newNA)

                } else {
                    x$M <- blockdiag(x$M, newM, pad=0) ## Add regression labels
                    x$cov <- blockdiag(x$cov, newcov, pad=0) ## Add covariance
                    x$par <- blockdiag(x$par, newNA, pad=NA) ## Add regression labels
                    x$covpar <- blockdiag(x$covpar, newNA, pad=NA) ## Add covariance labels
                    x$fix <- blockdiag(x$fix, newNA, pad=NA) ##
                    x$covfix <- blockdiag(x$covfix,  newNA, pad=NA) ##
                }
                x$mean <- c(x$mean, newmean)
            }
            names(x$mean)[N+seq_len(k)] <-
                colnames(x$M)[N+seq_len(k)] <- rownames(x$M)[N+seq_len(k)] <-
                colnames(x$covfix)[N+seq_len(k)] <- rownames(x$covfix)[N+seq_len(k)] <-
                colnames(x$fix)[N+seq_len(k)] <- rownames(x$fix)[N+seq_len(k)] <-
                colnames(x$covpar)[N+seq_len(k)] <- rownames(x$covpar)[N+seq_len(k)] <-
                colnames(x$par)[N+seq_len(k)] <- rownames(x$par)[N+seq_len(k)] <-
                colnames(x$cov)[N+seq_len(k)] <- rownames(x$cov)[N+seq_len(k)] <- new

            if (messages>1) {
                if (k==1)
                    message("\tAdded '", new, "' to model.\n", sep="")
                else
                    message("\tAdded ",paste(paste("'",new,"'",sep=""),collapse=",")," to model.\n", sep="")
            }
            exogenous(x) <- c(new,exogenous(x))
        }
        if (reindex)
            index(x) <- reindex(x)
        return(x)
    }

##' Convert to/from NA
##'
##' Convert vector to/from NA
##' @aliases NA2x x2NA
##' @param s The input vector (of arbitrary class)
##' @param x The elements to transform into \code{NA} resp. what to transform
##' \code{NA} into.
##' @return A vector with same dimension and class as \code{s}.
##' @author Klaus K. Holst
##' @keywords manip
##' @examples##' 
##' x2NA(1:10, 1:5)
##' NA2x(x2NA(c(1:10),5),5)##'
##' @export
NA2x <- function(s,x=0) { sapply(s, function(y) ifelse(is.na(y),x,y) ) }
##' @export
x2NA <- function(s,x=0) { sapply(s, function(y) ifelse(y%in%x,NA,y) ) }

##' @export
coxWeibull.lvm <- dist_cox_weibull

##' @export
coxExponential.lvm <- dist_cox_exponential

##' @export
aalenExponential.lvm <- dist_aalen_exponential

##' @export
coxGompertz.lvm <- dist_cox_gompertz

##' @export
normal.lvm <- dist_gaussian

##' @export
gaussian.lvm <- normal.lvm

##' @export
lognormal.lvm <- dist_lognormal

##' @export
poisson.lvm <- dist_poisson

##' @export
pareto.lvm <- dist_pareto

##' @export
threshold.lvm <- dist_threshold

##' @export
multinomial.lvm <- dist_multinomial

##' @export
binomial.lvm <- dist_bernoulli

##' @export
logit.lvm <- binomial.lvm("logit")

##' @export
probit.lvm <- binomial.lvm("probit")

##' @export
Gamma.lvm <- dist_gamma

##' @export
loggamma.lvm <- dist_loggamma

##' @export
chisq.lvm <- dist_chisq

##' @export
student.lvm <- dist_t

##' @export
uniform.lvm <- dist_uniform

##' @export
weibull.lvm <- dist_weibull

##' @export
id.lvm <- dist_seqint

##' @export
Sequence.lvm <- dist_seq

##' @export
none.lvm <- dist_none

##' @export
constant.lvm <- dist_const

##' @export
ones.lvm <- dist_ones

##' @export
binary.lvm <- ones.lvm

##' @export
beta.lvm <- dist_beta

##' @export
mvn.lvm <- dist_mvn

##' @export
GM2.lvm <- dist_gaussian_mixture2

##' @export
GM3.lvm <- dist_gaussian_mixture3

##' @export
`nodecolor<-` <-
function(object,var,...,value) UseMethod("nodecolor<-")

##' @export
`nodecolor<-.lvm` <-
  function(object, var=vars(object), border, labcol, shape, lwd, ..., value) {
    if (length(var)>0 & length(value)>0) {
      if (inherits(var,"formula")) var <- all.vars(var)
      object$noderender$fill[var] <- value
      if (!missing(border))
        object$noderender$col[var] <- border
      if (!missing(shape))
        object$noderender$shape[var] <- shape
      if (!missing(labcol))
        object$noderender$textCol[var] <- labcol
      if (!missing(lwd))
        object$noderender$lwd[var] <- lwd
    }
    return(object)
  }

##' @export
`nodecolor<-.default` <-
  function(object, var=vars(object), border, labcol, shape, lwd, ..., value) {
    if (length(var)>0 & length(value)>0) {
      if (inherits(var,"formula")) var <- all.vars(var)
      object <- addattr(object,attr="fill",var=var,val=value)
      if (!missing(border))
        object <- addattr(object,attr="col",var=var,val=border)
      if (!missing(shape))
        object <- addattr(object,attr="shape",var=var,val=shape)
      if (!missing(labcol))
        object <- addattr(object,attr="textCol",var=var,val=labcol)
      if (!missing(lwd))
        object <- addattr(object,attr="lwd",var=var,val=lwd)
    }
    return(object)
  }

##' @export
"intervention<-" <- function(object, ..., value)
  UseMethod("intervention<-")

##' @export
`intervention` <-
  function(object, ...) UseMethod("intervention")

##' Define intervention
##'
##' Define intervention in a `lvm` object
##' @param object lvm object
##' @param to String defining variable or formula
##' @param value function defining intervention
##' @param dist Distribution
##' @param ... Additional arguments to lower level functions
##' @aliases intervention<- intervention intervention.lvm intervention<-.lvm
##' @seealso regression lvm sim
##' @examples
##' m <- lvm(y ~ a + x, a ~ x)
##' distribution(m, ~a+y) <- binomial.lvm()
##' mm <- intervention(m, "a", value=3)
##' sim(mm, 10)
##' mm <- intervention(m, a~x, function(x) (x>0)*1)
##' sim(mm, 10)
##' @export
intervention.lvm <- function(object, to, value, dist=none.lvm(), ...) {
  if (!is.numeric(value))
    regression(object, to, ...) <- value
  y <- to
  if (inherits(to, "formula")) {
    y <- getoutcome(to)
    if (length(y)==0)
      y <- attr(y, "x")
  }
  parents <- parents(object, y)
  if (length(parents)>0)
    cancel(object) <- toformula(y, parents)
  if (is.numeric(value)) {
    distribution(object, y) <- constant.lvm(value)
  } else {
    distribution(object, y) <- dist
  }
  return(object)
}

##' @export
"intervention<-.lvm" <- function(object, to, ..., value) {
  object <- intervention(object, to, value, ...)
  return(object)
}

##' Generates missing entries in data.frame/matrix
##'
##' @title Create random missing data
##' @param data data.frame
##' @param p Fraction of missing data in each column
##' @param cols Which columns (name or index) to alter
##' @param rowwise Should missing occur row-wise (either none or all selected columns are missing)
##' @param nafun (Optional) function to be applied on data.frame before return (e.g. \code{na.omit} to return complete-cases only)
##' @param seed Random seed
##' @return data.frame
##' @author Klaus K. Holst
##' @keywords utilities
##' @export
makemissing <- function(data,p=0.2,cols=seq_len(ncol(data)),rowwise=FALSE,nafun=function(x) x, seed=NULL) {
    if (!exists(".Random.seed", envir = .GlobalEnv, inherits = FALSE))
        runif(1)
    if (is.null(seed))
        RNGstate <- get(".Random.seed", envir = .GlobalEnv)
    else {
        R.seed <- get(".Random.seed", envir = .GlobalEnv)
        set.seed(seed)
        RNGstate <- structure(seed, kind = as.list(RNGkind()))
        on.exit(assign(".Random.seed", R.seed, envir = .GlobalEnv))
    }    
    p <- rep(p,length.out=length(cols))
    if (!rowwise)
        for (i in seq_along(cols)) {
            data[rbinom(nrow(data),1,p[i])==1,cols[i]] <- NA
        }
    else
        data[which(rbinom(nrow(data),1,p)==1),cols] <- NA
    return(nafun(data))
}

##' @export
describecoef <- function(x,par,from,to,mean=TRUE) {
    p <- coef(x, mean=mean)
    if (!missing(from)) {
        st1 <- paste0(to,lava.options()$symbol[1],from)
        st2 <- paste0(to,lava.options()$symbol[2],from)
        st3 <- paste0(from,lava.options()$symbol[2],to)
        pos <- na.omit(match(unique(c(st1,st2,st3)),p))
        attributes(pos) <- NULL
        return(pos)
    }
    res <- strsplit(p,lava.options()$symbol[2])
    var.idx <- which(unlist(lapply(res,length))>1) ## Variance parameters
    rest.idx <- setdiff(seq_along(p),var.idx)
    res[rest.idx] <- strsplit(p[rest.idx],lava.options()$symbol[1])
    mean.idx <- which(unlist(lapply(res,length))==1) ## Mean parameters
    reg.idx <- setdiff(rest.idx,mean.idx)
    names(res)[mean.idx] <- paste0("m",seq_along(mean.idx))
    for (i in var.idx)
        attr(res[[i]],"type") <- "cov"
    for (i in mean.idx)
        attr(res[[i]],"type") <- "mean"
    for (i in reg.idx)
        attr(res[[i]],"type") <- "reg"
    if (missing(par))
        return(res)
    return(res[par])
}

##' @export
colsel <- function(locate,...) {
    ytop    <- rep(seq(1/26,1,by=1/26),each=26)[1:657]
    ybottom <- rep(seq(0,1-1/26,by=1/26),each=26)[1:657]
    xleft   <- rep(seq(0,1-1/26,by=1/26),times=26)[1:657]
    xright  <- rep(seq(1/26,1,by=1/26),times=26)[1:657]
    pall    <- round(col2rgb(colors())/256)
    pall    <- colSums(pall) ; pall2 <- character(0)
    pall2[pall>0]   <- "black"
    pall2[pall==0]  <- "white"

    par(mar=c(0,0,1,0))

    plot.new()
    title(main="Palette of colors()")
    rect(xleft,ybottom,xright,ytop,col=colors())
    text(x=xleft+((1/26)/2)
        ,y=ytop-((1/26)/2)
        ,labels = 1:657
             ,cex=0.55
        ,col=pall2)

    if (missing(locate)) return(invisible(NULL))
    colmat <- matrix(c(1:657,rep(NA,26^2-657)),byrow=T,ncol=26,nrow=26)
    cols <- NA
    for(i in seq_len(locate))
    {
        h    <- locator(1)
        if(any(h$x<0,h$y<0,h$x>1,h$y>1)) stop("locator out of bounds!")
        else {
            cc        <- floor(h$x/(1/26))+1
            rr        <- floor(h$y/(1/26))+1
            cols[i]    <- colors()[colmat[rr,cc]]
        }
    }
    return(cols)
}


##' Extension of the \code{identify} function
##'
##' For the usual 'X11' device the identification process is
##' terminated by pressing any mouse button other than the first. For
##' the 'quartz' device the process is terminated by pressing either
##' the pop-up menu equivalent (usually second mouse button or
##' 'Ctrl'-click) or the 'ESC' key.
##' @title Identify points on plot
##' @usage
##' \method{click}{default}(x, y=NULL, label=TRUE, n=length(x), pch=19, col="orange", cex=3, ...)
##' idplot(x, y ,..., id=list(), return.data=FALSE)
##' @aliases idplot click.default click colsel
##' @param x X coordinates
##' @param y Y coordinates
##' @param label Should labels be added?
##' @param n Max number of inputs to expect
##' @param pch Symbol
##' @param col Colour
##' @param cex Size
##' @param id List of arguments parsed to \code{click} function
##' @param return.data Boolean indicating if selected points should be returned
##' @param \dots Additional arguments parsed to \code{plot} function
##' @author Klaus K. Holst
##' @seealso \code{\link{idplot}}, \code{identify}
##' @examples
##' if (interactive()) {
##'     n <- 10; x <- seq(n); y <- runif(n)
##'     plot(y ~ x); click(x,y)
##'
##'     data(iris)
##'     l <- lm(Sepal.Length ~ Sepal.Width*Species,iris)
##'     res <- plotConf(l,var2="Species")## ylim=c(6,8), xlim=c(2.5,3.3))
##'     with(res, click(x,y))
##'
##'     with(iris, idplot(Sepal.Length,Petal.Length))
##' }
##' @keywords iplot
##' @export
click <- function(x,...){
  UseMethod("click")
}

##' @export
click.default <-
function(x, y=NULL, label=TRUE, n=length(x),
         pch=19, col="orange", cex=3, ...) {
  xy <- xy.coords(x, y); x <- xy$x; y <- xy$y
  sel <- rep(FALSE, length(x)); res <- integer(0)
  while(sum(sel) < n) {
    ans <- identify(x[!sel], y[!sel], n=1, plot=FALSE, ...)
    if(!length(ans)) break
    ans <- which(!sel)[ans]
    points(x[ans], y[ans], pch = pch, col=col, cex=cex)
    if (label)
      text(x[ans], y[ans], ans)
    sel[ans] <- TRUE
    res <- c(res, ans)
  }
  res
}

##' @export
idplot <- function(x, y=NULL, ..., id=list(), return.data=FALSE) {
  plot(x,y,...)
  id$x <- x; id$y <- y
  idx <- do.call("click", id)
  if (return.data) {
    return(with(id, cbind(x,y)[idx,]))
  }
  return(idx)
}

##' Spaghetti plot for longitudinal data
##'
##' @title Spaghetti plot
##' @param formula Formula (response ~ time)
##' @param data data.frame
##' @param id Id variable
##' @param group group variable
##' @param type Type (line 'l', stair 's', ...)
##' @param lty Line type
##' @param pch Colour
##' @param col Colour
##' @param alpha transparency (0-1)
##' @param lwd Line width
##' @param level Confidence level
##' @param trend.formula Formula for trendline
##' @param tau Quantile to estimate (trend)
##' @param trend.lty Trend line type
##' @param trend.join Trend polygon
##' @param trend.delta Length of limit bars
##' @param trend Add trend line
##' @param trend.col Colour of trend line
##' @param trend.alpha Transparency
##' @param trend.lwd Trend line width
##' @param trend.jitter Jitter amount
##' @param legend Legend
##' @param by make separate plot for each level in 'by' (formula, name of column, or vector)
##' @param xlab Label of X-axis
##' @param ylab Label of Y-axis
##' @param add Add to existing device
##' @param ... Additional arguments to lower level arguments
##' @author Klaus K. Holst
##' @export
##' @examples
##' if (interactive() & requireNamespace("mets")) {
##' K <- 5
##' y <- "y"%++%seq(K)
##' m <- lvm()
##' regression(m,y=y,x=~u) <- 1
##' regression(m,y=y,x=~s) <- seq(K)-1
##' regression(m,y=y,x=~x) <- "b"
##' N <- 50
##' d <- sim(m,N); d$z <- rbinom(N,1,0.5)
##' dd <- mets::fast.reshape(d); dd$num <- dd$num+3
##' spaghetti(y~num,dd,id="id",lty=1,col=Col(1,.4),
##'           trend.formula=~factor(num),trend=TRUE,trend.col="darkblue")
##' dd$num <- dd$num+rnorm(nrow(dd),sd=0.5) ## Unbalance
##' spaghetti(y~num,dd,id="id",lty=1,col=Col(1,.4),
##'           trend=TRUE,trend.col="darkblue")
##' spaghetti(y~num,dd,id="id",lty=1,col=Col(1,.4),
##'            trend.formula=~num+I(num^2),trend=TRUE,trend.col="darkblue")
##' }
spaghetti <- function(formula,data=NULL,id="id",group=NULL,
              type="o",lty=1,pch=NA,col=1:10,alpha=0.3,lwd=1,
              level=0.95,
              trend.formula=formula,tau=NULL,
              trend.lty=1,trend.join=TRUE,trend.delta=0.2,
              trend=!is.null(tau),trend.col=col,
              trend.alpha=0.2,trend.lwd=3,
              trend.jitter=0,
              legend=NULL, by=NULL,
              xlab="Time",ylab="",add=FALSE,...) {

    if (!lava.options()$cluster.index) stop("mets not available? Check 'lava.options()cluster.index'.")
    if (!is.null(by)) {
        if (is.character(by) && length(by==1)) {
            by <- data[,by]
        } else if (inherits(by,"formula")) {
            ##by <- model.matrix(update(by,~-1+.), model.frame(~.,data,na.action=na.pass))
            by <- model.frame(by,data,na.action=na.pass)
        }
        cl <- match.call(expand.dots=TRUE)
        cl$by <- NULL
        datasets <- split(data,by)
        res <- c()
        for (d in datasets) {
            cl$data <- d
            res <- c(res, eval(cl,parent.frame()))
        }
        return(invisible(res))
    }
    if (!is.null(group)) {
        if (is.character(group) && length(group==1)) {
            M <- data[,group]
        } else if (inherits(group,"formula")) {
            ##M <- model.matrix(update(group,~-1+.),data)
            M <- model.frame(group,data,na.action=na.pass)
        } else {
            M <- group
        }
        if (!add) plot(formula,data=data,xlab=xlab,ylab=ylab,...,type="n")
        dd <- split(data,M)
        K <- length(dd)
        if (length(type)<K)        type <- rep(type,K)
        if (length(col)<K)         col <- rep(col,K)
        if (length(pch)<K)         pch <- rep(pch,K)
        if (length(lty)<K)         lty <- rep(lty,K)
        if (length(lwd)<K)         lwd <- rep(lwd,K)
        if (length(alpha)<K)       alpha <- rep(alpha,K)
        if (length(trend)<K)       trend <- rep(trend,K)
        if (length(trend.col)<K)   trend.col <- rep(trend.col,K)
        if (length(trend.lty)<K)   trend.lty <- rep(trend.lty,K)
        if (length(trend.alpha)<K) trend.alpha <- rep(trend.alpha,K)
        if (length(trend.lwd)<K)   trend.lwd <- rep(trend.lwd,K)
        for (i in seq_len(K)) {
            spaghetti(formula,data=dd[[i]],id=id,type=type[i],
                     lty=lty[i],pch=pch[i],col=col[i],lwd=lwd[i],
                     alpha=alpha[i],
                     group=NULL,
                     trend=trend[i],tau=tau,
                     trend.col=trend.col[i],
                     trend.alpha=trend.alpha[i],
                     trend.lwd=trend.lwd[i],
                     trend.lty=trend.lty[i],
                     trend.delta=trend.delta,
                     trend.formula=trend.formula,
                     add=TRUE,...)
        }
        if (!is.null(legend)) {
            graphics::legend(legend,names(dd),lwd=lwd,col=col,lty=lty)
        }
        return(invisible(NULL))
    }

    if (inherits(id,"formula")) id <- all.vars(id)
    if (inherits(group,"formula")) group <- all.vars(group)
    if (is.character(id) && length(id)==1) Id <- data[,id]
    y <- getoutcome(formula)
    x <- attributes(y)$x
    Idx <- function(vname,widenames) {
        idx <- which(unlist(lapply(widenames,function(x)
            length(grep(paste0(vname,"[0-9]+$"),x))>0)))
        nn <- widenames[idx]
        ord <- order(char2num(unlist(lapply(nn,function(x) gsub(vname,"",x)))))
        idx[ord]
    }

    if (!requireNamespace("mets",quietly=TRUE)) stop("'mets' package required")

    if (length(x)==0) {
        data <- data[,c(id,y),drop=FALSE]
        wide <- mets::fast.reshape(data,id=id,varying=y,...)
        yidx <- Idx(y,names(wide))
        Y <- wide[,yidx,drop=FALSE]
        X <- NULL
        matplot(t(Y),type=type,lty=lty,pch=pch,lwd=lwd,col=Col(col[1],alpha[1]),xlab=xlab,ylab=ylab,...)
    } else {        
        data <- data[,c(id,x,y),drop=FALSE]
        wide <- mets::fast.reshape(data[order(data[,id],data[,x]),],id=id,varying=c(y,x),...)
        yidx <- Idx(y,names(wide))
        xidx <- Idx(x,names(wide))
        Y <- wide[,yidx,drop=FALSE]
        X <- wide[,xidx,drop=FALSE]
        matplot(t(X),t(Y),type=type,pch=pch,lty=lty,lwd=lwd,col=Col(col[1],alpha[1]),xlab=xlab,ylab=ylab,add=add,...)
        if (trend) {
            if (is.numeric(trend.formula)) {
                trend.formula <- sort(trend.formula)
                tf <- toformula(y,"1")
                res <- c()
                if (!is.null(tau)) {
                    if (length(trend.alpha)<length(tau)) 	trend.alpha <- rep(trend.alpha,length(tau))
                    if (length(trend.lty)<length(tau)) 	trend.lty <- rep(trend.lty,length(tau))
                    if (length(trend.col)<length(tau)) 	trend.col <- rep(trend.col,length(tau))
                    if (length(trend.lwd)<length(tau)) 	trend.lwd <- rep(trend.lwd,length(tau))
                }
                for (i in trend.formula) {
                    data0 <- data[data[,x]==i,,drop=FALSE]
                    newdata <- data.frame(i); names(newdata) <- x
                    if (!is.null(tau)) {
                        ##if (!require(quantreg)) stop("Install 'quantreg'")
                        suppressWarnings(r1 <- quantreg::rq(tf,data=data0,tau=tau))
                        pr <- predict(r1,newdata=newdata,level=level)
                        res <- rbind(res,pr)
                    } else {
                        l1 <- lm(tf,data0)
                        pr <- predict(l1,newdata=newdata,interval="confidence",level=level)
                        res <- rbind(res,pr)
                    }
                }
                if (!is.null(tau)) {
                    for (j in seq_len(ncol(res))) {
                            if (trend.join) lines(trend.formula,res[,j],col=trend.col[j],lwd=trend.lwd[j],lty=trend.lty[j],...)
                            if (trend.delta>0) confband(trend.formula,res[,j],line=FALSE,col=trend.col[j],lty=trend.lty[j],lwd=trend.lwd[j],delta=trend.delta,...)
                        }
                } else {
                    confband(trend.formula,res[,2],res[,3],res[,1],col=Col(trend.col,trend.alpha),lty=trend.lty,lwd=trend.lwd,polygon=trend.join,...)
                }
            } else {
                tf <- getoutcome(trend.formula)
                if (is.list(tf)) {
                    trend.formula <- update(trend.formula,toformula(y,"."))
                }
                if (!is.null(tau)) {
                    ##if (!require(quantreg)) stop("Install 'quantreg'")
                    suppressWarnings(r1 <- quantreg::rq(trend.formula,data=data,tau=tau))
                    newdata <- data.frame(seq(min(X,na.rm=TRUE),max(X,na.rm=TRUE),length.out=100))
                    names(newdata) <- x
                    pr <- predict(r1,newdata=newdata,interval="confidence",level=level)
                    ##confband(xx,pr[,3],pr[,2],polygon=TRUE,col=Col(trend.col,trend.alpha),border=FALSE)
                    for (i in seq_along(tau))
                        lines(newdata[,1],pr[,i],col=trend.col,lwd=trend.lwd,lty=trend.lty)
                } else {
                    l1. <- lm(trend.formula,data)
                    l1 <- estimate(l1.,id=data[,id],level=level)
                    xy <- plotConf(l1.,vcov=vcov(l1),data=data,partres=FALSE,plot=FALSE,level=level,...)
                    xx <- xy$x
                    pr <- xy$predict$fit
                    if (is.factor(xx)) {
                        xx <- char2num(as.character(xx))
                        if (trend.jitter>0) xx <- jitter(xx,trend.jitter)                        
                        confband(xx,pr[,3],pr[,2],pr[,1],col=trend.col,lwd=2)
                    } else {
                        confband(xx,pr[,3],pr[,2],polygon=TRUE,col=Col(trend.col,trend.alpha),border=FALSE)
                        lines(xx,pr[,1],col=trend.col,lwd=trend.lwd,lty=trend.lty)
                    }
                }
            }
        }
    }
    return(invisible(list(Y,X)))
}

##' Plot/estimate surface
##'
##' @export
##' @aliases ksmooth2 surface
##' @param x formula or data
##' @param data data.frame
##' @param h bandwidth
##' @param xlab X label
##' @param ylab Y label
##' @param zlab Z label
##' @param gridsize grid size of kernel smoother
##' @param ... Additional arguments to graphics routine (persp3d or persp)
##' @examples
##' if (requireNamespace("KernSmooth")) {##'
##' ksmooth2(rmvn0(1e4,sigma=diag(2)*.5+.5),c(-3.5,3.5),h=1,
##'         rgl=FALSE,theta=30)
##' ##'
##' if (interactive()) {
##'     ksmooth2(rmvn0(1e4,sigma=diag(2)*.5+.5),c(-3.5,3.5),h=1)
##'     ksmooth2(function(x,y) x^2+y^2, c(-20,20))
##'     ksmooth2(function(x,y) x^2+y^2, xlim=c(-5,5), ylim=c(0,10))
##'
##'     f <- function(x,y) 1-sqrt(x^2+y^2)
##'     surface(f,xlim=c(-1,1),alpha=0.9,aspect=c(1,1,0.75))
##'     surface(f,xlim=c(-1,1),clut=heat.colors(128))
##'     ##play3d(spin3d(axis=c(0,0,1), rpm=8), duration=5)
##' }
##'
##' if (interactive()) {
##'     surface(function(x) dmvn0(x,sigma=diag(2)),c(-3,3),lit=FALSE,smooth=FALSE,box=FALSE,alpha=0.8)
##'     surface(function(x) dmvn0(x,sigma=diag(2)),c(-3,3),box=FALSE,specular="black")##'
##' }
##'
##' if (!inherits(try(find.package("fields"),silent=TRUE),"try-error")) {
##'     f <- function(x,y) 1-sqrt(x^2+y^2)
##'     ksmooth2(f,c(-1,1),rgl=FALSE,image=fields::image.plot)
##' }
##'
##' }
ksmooth2 <- function(x,data,h=NULL,xlab=NULL,ylab=NULL,zlab="",gridsize=rep(51L,2),...) {
    if (is.function(x)) {
        args <- c(list(f=x,h=h,xlab=xlab,ylab=ylab,zlab=zlab),list(...))
        if (is.null(args$xlim) && !missing(data)) {
            if (is.list(data)) {
                args$xlim <- data[[1]]
                args$ylim <- data[[2]]
            } else args$xlim <- data
        }
        return(do.call(surface,args))
    }
    if (inherits(x,"formula")) {
        x <- model.frame(x,data)
    }
    if (length(gridsize)==1) gridsize <- rep(gridsize,2)
    
    if (is.null(h)) h <- apply(as.matrix(x),2,sd)*nrow(x)^(-1/5)
    est <- KernSmooth::bkde2D(x, bandwidth=h, gridsize=gridsize)
    if (is.null(xlab)) xlab <- names(x)[1]
    if (is.null(ylab)) ylab <- names(x)[2]
    surface(est$fhat, x=est$x1, y=est$x2, est$fhat,
            xlab=xlab, ylab=ylab, zlab=zlab, ...)
    return(invisible(est))
}


##' @export
surface <- function(f,xlim=c(0,1),ylim=xlim,n=rep(100,2),col,clut="gold",clut.center,x,y,rgl=TRUE,expand=0.5,nlevels=10,col.contour="black",contour=TRUE,persp=TRUE,image="image",...) {
    if (missing(x)) {
        if (length(n)==1) n <- rep(n,2)
        x <- seq(xlim[1],xlim[2],length.out=n[1])
        y <- seq(ylim[1],ylim[2],length.out=n[2])
    }
    if (is.function(f)) {
        xy <- as.matrix(expand.grid(x,y))
        if (inherits(try(f(c(x[1],y[1])),silent=TRUE),"try-error")) {
            f <- matrix(f(xy[,1],xy[,2]),nrow=length(x),ncol=length(y))
        } else {
            val <- f(xy)
            if (length(val)<NROW(xy)) {
                f <- matrix(apply(xy,1,f),nrow=length(x),ncol=length(y))
            } else {
                f <- matrix(val,nrow=length(x),ncol=length(y))
            }
        }
    }
    zrg <- range(f)
    zlen <- diff(zrg)
    if (length(clut)==1) {
        ncolour <- 128
        clut <- switch(clut,
                       topo=topo.colors(ncolour),
                       red=colorRampPalette(c("yellow","red"),bias=1)(ncolour),
                       blue=colorRampPalette(c("white","blue"),bias=1)(ncolour),
                       gold=colorRampPalette(c("white","goldenrod1"),bias=1)(ncolour),
                       france=colorRampPalette(c("blue","white","red"))(ncolour),
                       rainbow=rainbow(n=128,start=0,end=1),
                       heat=heat.colors(ncolour),
                       heatrev=rev(heat.colors(ncolour)),
                       colorRampPalette(c("white","goldenrod1"),bias=1)(ncolour)
                       )
    }
    ncolour <- length(clut)
    if (!rgl) {
        if (missing(col)) {
            nc <- ncol(f); nr <- nrow(f)
            facet <- f[-1,-1]+f[-1,-nc]+f[-nr,-1]+f[-nr,-nc]
            facetcol <- cut(facet, ncolour)
            col <- clut[facetcol]
        }
        dots <- list(...)
        parargs <- list()
        if (persp) {
            parargs$mar <- c(0.2,0,0,0)
            if (contour || !is.null(image)) parargs$mfrow=c(2,1)
        }
        op <- do.call(par,parargs)
        if (persp) graphics::persp(x=x,y=x,z=f, col=col, expand=expand, ...)
        if (contour | !is.null(image)) {
            if (persp) par(mar=c(3,3,0.5,3)) ## (bottom, left, top, right)
            if (!is.null(image)) {                
                do.call(image,list(x=x,y=y,z=f,col=clut,xlim=range(x),ylim=range(y),xlab="",ylab=""))
                box()
            }
            if (contour) {
                args <- c(list(x=x,y=y,z=f,nlevels=nlevels,col=col.contour,add=!is.null(image)),dots)
                args <- args[names(formals(contour.default))]
                do.call("contour",args)
            }
        }
        suppressWarnings(par(op))
    } else {
        if (missing(col)) {
            if (!missing(clut.center)) {
                zmax <- max(abs(zrg))
                zrg <- c(-zmax,zmax)
                zlen <- 2*zmax
            }
            col <- clut[round((length(clut)-1)*(f-zrg[1])/zlen)+1]
        }
        rgl::persp3d(x, y, f, col=col,...)
    }
}

##' Extract all possible paths from one variable to another connected component
##' in a latent variable model. In an estimated model the effect size is
##' decomposed into direct, indirect and total effects including approximate
##' standard errors.
##'
##' @title Extract pathways in model graph
##' @export
##' @aliases path effects path.lvm effects.lvmfit
##' totaleffects
##' @seealso \code{children}, \code{parents}
##' @return If \code{object} is of class \code{lvmfit} a list with the following
##' elements is returned \item{idx}{ A list where each element defines a
##' possible pathway via a integer vector indicating the index of the visited
##' nodes. } \item{V }{ A List of covariance matrices for each path. }
##' \item{coef }{A list of parameters estimates for each path} \item{path }{A
##' list where each element defines a possible pathway via a character vector
##' naming the visited nodes in order.  } \item{edges }{Description of 'comp2'}
##'
##' If \code{object} is of class \code{lvm} only the \code{path} element will be
##' returned.
##'
##' The \code{effects} method returns an object of class \code{effects}.
##' @note For a \code{lvmfit}-object the parameters estimates and their
##' corresponding covariance matrix are also returned.  The
##' \code{effects}-function additionally calculates the total and indirect
##' effects with approximate standard errors
##' @author Klaus K. Holst
##' @keywords methods models graphs
##' @examples
##'
##' m <- lvm(c(y1,y2,y3)~eta)
##' regression(m) <- y2~x1
##' latent(m) <- ~eta
##' regression(m) <- eta~x1+x2
##' d <- sim(m,500)
##' e <- estimate(m,d)
##'
##' path(Model(e),y2~x1)
##' parents(Model(e), ~y2)
##' children(Model(e), ~x2)
##' children(Model(e), ~x2+eta)
##' effects(e,y2~x1)
##' ## All simple paths (undirected)
##' path(m,y1~x1,all=TRUE)
##'
##' @usage
##' \method{path}{lvm} (object, to = NULL, from, all=FALSE, ...)
##' \method{effects}{lvmfit} (object, to, from, ...)
##' @param object Model object (\code{lvm})
##' @param to Outcome variable (string). Alternatively a formula specifying
##' response and predictor in which case the argument \code{from} is ignored.
##' @param from Response variable (string), not necessarily directly affected by
##' \code{to}.
##' @param all If TRUE all simple paths (in undirected graph) is returned
##' on/off.
##' @param \dots Additional arguments to be passed to the low level functions
##' @export
path <- function(object,...) UseMethod("path")

##' @export
path.lvmfit <- function(object,to=NULL,from,...) {
    mypath <- pathM(Model(object)$M,to,from,...)
    cc <- coef(object,type=9,labels=FALSE) ## All parameters (fixed and variable)
    cc0 <- coef(object,type=2) ## Estimated parameters
    i1 <- na.omit(match(rownames(cc),rownames(cc0)))
    idx.cc0 <-  which(rownames(cc)%in%rownames(cc0)); ## Position of estimated parameters among all parameters
    S <- matrix(0,nrow(cc),nrow(cc)); rownames(S) <- colnames(S) <- rownames(cc)
    V <- object$vcov
    S[idx.cc0,idx.cc0] <- V[i1,i1]  ## "Covariance matrix" of all parameters

    idx <- list()
    coefs <- list()
    V <- list()
    for (i in seq_along(mypath)) {
        xx <- mypath[[i]]
        ii <- c()
        for (j in seq_len(length(xx)-1)) {
            st <- paste0(xx[j+1], lava.options()$symbol[1], xx[j])
            ii <- c(ii, match(st,rownames(cc)))
        }
        idx <- c(idx, list(ii))
        V <- c(V, list(S[ii,ii]))
        coefs <- c(coefs, list(cc[ii]))
    }

    edges <- list()
    for (i in seq_along(mypath)) {
        p0 <- mypath[[i]]
        ee <- c()
        for (i in seq_len(length(p0)-1)) {
            ee <- c(ee, paste(p0[i],p0[i+1],sep="~"))
        }
        edges <- c(edges, list(ee))
    }
    res <- list(idx=idx,V=V,coef=coefs, path=mypath, edges=edges)
    return(res)
}

##' @export
path.lvm <- function(object,to=NULL,from,all=FALSE,...) {
    pathM(object$M,to=to,from=from,all=all,...)
}

##' @export
path.graphNEL <- function(object,to,from,...) {
    if (inherits(to,"formula")) {
        fvar <- extractvar(to)
        if (length(fvar$x)==1 & length(fvar$y)==1)
            return(path(object,to=fvar$y,from=fvar$x))
        res <- list()
        for (y in fvar$y) {
            for (x in fvar$x) {
                cat("x=",x, " y=",y, "\n")
                res <- c(res, list(path(object,to=y,from=x)))
            }
        }
        return(res)
    }
    ff <- function(g,from=1,to=NULL,res=list()) {
        M <- graph::edgeMatrix(g)
        i1 <- which(M[1,]==from)
        for (i in i1) {
            e <- M[,i]; newto <- e[2];
            if (is.null(to) || M[2,i]==to) {
                res <- c(res, list(M[,i]))
            }
            newpath <- ff(g,from=newto,to=to,list())
            if (length(newpath)>0)
                for (j in seq_along(newpath)) {
                    if (is.null(to) || (tail(newpath[[j]],1)==to))
                        res <- c(res, list(c(M[,i],newpath[[j]][-1])))
                }
        }
        return(res)
    }
    idxfrom <- ifelse(is.numeric(from),from,which(from==graph::nodes(object)))
    ##M <- as(object,"matrix")
    ##reachable <- acc(M,graph::nodes(object)[idxfrom])
    reachable <- graph::acc(object,graph::nodes(object)[idxfrom])[[1]]

    if (is.null(to)) {
        idxto <- reachable
    } else {
        idxto <- ifelse(is.numeric(to),to,which(to==graph::nodes(object)))
    }

    if (!(graph::nodes(object)[idxto] %in% names(reachable)))
        ##    return(structure(list(),to=to[1],from=from[1]))
        return(NULL)
    ##    stop("No directional relationship between variables")

    mypaths <- ff(object,idxfrom,idxto)
    res <- list()
    for (i in seq_along(mypaths)) {
        res <- c(res, list(graph::nodes(object)[mypaths[[i]]]))
    }
    return(res)
}


pathM <- function(M,to,from,all=FALSE,...) {
    nn <- colnames(M)
    if (inherits(to,"formula")) {
        fvar <- extractvar(to)
        if (length(fvar$x)==1 & length(fvar$y)==1)
            return(pathM(M,to=fvar$y,from=fvar$x,all=all))
        res <- list()
        for (y in fvar$y) {
            for (x in fvar$x) {
                cat("x=",x, " y=",y, "\n")
                res <- c(res, list(pathM(M,to=y,from=x,all=all)))
            }
        }
        return(res)
    }
    if (all) { ## Get all simple paths
        res <- simplePaths(to,from,from,M,list())
        return(res)
    }


    ff <- function(g,from=1,to=NULL,res=list()) {
        i1 <- which(M[from,]==1)
        for (i in i1) {
            ##      e <- M[,i]; newto <- e[2];
            if (is.null(to) || i==to) {
                res <- c(res, list(c(from,i)))
            }
            newpath <- ff(g,from=i,to=to,list())
            if (length(newpath)>0)
                for (j in seq_along(newpath)) {
                    if (is.null(to) || (tail(newpath[[j]],1)==to))
                        res <- c(res, list(c(c(from,i),newpath[[j]][-1])))
                }
        }
        return(res)
    }
    idxfrom <- ifelse(is.numeric(from),from,which(from==nn))
    reachable <- acc(M,nn[idxfrom])

    if (is.null(to)) {
        idxto <- reachable
    } else {
        idxto <- ifelse(is.numeric(to),to,which(to==nn))
    }

    if (!(nn[idxto] %in% reachable))
        return(NULL)
    ##    stop("No directional relationship between variables")

    mypaths <- ff(M,idxfrom,idxto)
    res <- list()
    for (i in seq_along(mypaths)) {
        res <- c(res, list(nn[mypaths[[i]]]))
    }
    return(res)
}



## Find all simple paths (no cycles) in an undirected graph
simplePaths <- function(target,currentpath,visited,adjmat,allpaths) {
    lastnode <- currentpath[length(currentpath)]
    A <- (adjmat+t(adjmat))>0
    if (lastnode==target) {
        allpaths <- c(allpaths,list(currentpath))
    } else {
        for (neighbour in rownames(adjmat)[which(A[,lastnode])]) {
            if (!(neighbour%in%visited)) {
                currentpath <- c(currentpath,neighbour)
                visited <- c(visited,neighbour)
                allpaths <- simplePaths(target,currentpath,visited,adjmat,allpaths)
                visited <- setdiff(visited,neighbour)
                currentpath <- currentpath[-length(currentpath)]
            }
        }
    }
    return(allpaths)
}

##' Extract i.i.d. decomposition (influence function) from model object
##'
##' Extract i.i.d. decomposition (influence function) from model object
##' @export
##' @usage
##'
##' iid(x,...)
##'
##' \method{iid}{default}(x,bread,id=NULL,folds=0,maxsize=(folds>0)*1e6,...)
##'
##' @aliases iid.default
##' @param x model object
##' @param id (optional) id/cluster variable
##' @param bread (optional) Inverse of derivative of mean score function
##' @param folds (optional) Calculate aggregated iid decomposition (0:=disabled)
##' @param maxsize (optional) Data is split in groups of size up to 'maxsize' (0:=disabled)
##' @param ... additional arguments
##' @examples
##' m <- lvm(y~x+z)
##' distribution(m, ~y+z) <- binomial.lvm("logit")
##' d <- sim(m,1e3)
##' g <- glm(y~x+z,data=d,family=binomial)
##' crossprod(iid(g))
##'
iid <- function(x,...) UseMethod("iid")

##' @export
iid.default <- function(x,bread,id=NULL,folds=0,maxsize=(folds>0)*1e6,...) {
    if (!any(paste("score",class(x),sep=".") %in% methods("score"))) {
        warning("Not available for this class")
        return(NULL)
    }

    if (folds>0 || maxsize>0 || (!missing(id) && lava.options()$cluster.index)) {
        if (!requireNamespace("mets",quietly=TRUE)) stop("Requires 'mets'")
    }
    
    if (folds>0) {
        U <- Reduce("rbind",mets::divide.conquer(function(data) score(x,data=data,...),
                                                 id=id,
                                                 data=data,size=round(nrow(data)/folds)))
    } else {
        U <- score(x,indiv=TRUE,...)
    }
    pp <- pars(x)
    if (!missing(bread) && is.null(bread)) {
        bread <- suppressWarnings(vcov(x))
    }
    if (missing(bread)) bread <- attributes(U)$bread
    if (is.null(bread)) {
        bread <- attributes(x)$bread
        if (is.null(bread)) bread <- x$bread
        if (is.null(bread)) {
            if (maxsize>0) {
                ff <- function(p) colSums(Reduce("rbind",mets::divide.conquer(function(data) score(x,data=data,p=p,...),
                                                                       data=data,size=maxsize)))
                I <- -numDeriv::jacobian(ff,pp,method=lava.options()$Dmethod)
            } else {
                I <- -numDeriv::jacobian(function(p) score(x,p=p,indiv=FALSE,...),pp,method=lava.options()$Dmethod)
            }
            bread <- Inverse(I)
        }
    }
    iid0 <- U%*%bread
    if (!missing(id)) {
        N <- nrow(iid0)
        if (!lava.options()$cluster.index) {
            iid0 <- matrix(unlist(by(iid0,id,colSums)),byrow=TRUE,ncol=ncol(bread))
        } else {
            iid0 <- mets::cluster.index(id,mat=iid0,return.all=FALSE)
        }
        attributes(iid0)$N <- N
    }
    colnames(iid0) <- colnames(U)
    return(structure(iid0,bread=bread))
}


##' @export
iid.multigroupfit <- function(x,...) iid.default(x,combine=TRUE,...)

##' @export
iid.matrix <- function(x,...) {
    p <- NCOL(x)
    n <- NROW(x)
    mu <- colMeans(x,na.rm=TRUE); S <- var(x,use="pairwise.complete.obs")*(n-1)/n
    iid1 <- t(t(x)-mu)
    iid2 <- matrix(ncol=(p+1)*p/2,nrow=n)
    pos <- 0
    nn <- c()
    cc <- mu
    for (i in seq(p))
        for (j in seq(i,p)) {
            pos <- pos+1
            cc <- c(cc,S[i,j])
            iid2[,pos] <- (iid1[,i]*iid1[,j])-cc[length(cc)]
            nn <- c(nn,paste(colnames(x)[c(i,j)],collapse=lava.options()$symbols[2]))
        }
    colnames(iid1) <- colnames(x); colnames(iid2) <- nn
    names(cc) <- c(colnames(iid1),colnames(iid2))
    iid1[is.na(iid1)] <- 0
    iid2[is.na(iid2)] <- 0
    structure(cbind(iid1/n,iid2/n),
              coef=cc,
              mean=mu, var=S)
}

##' @export
iid.numeric <- function(x,...) {
    n <- length(x)
    mu <- mean(x); S <- var(x)*(n-1)/n
    iid1 <- t(t(x)-mu)
    structure(cbind(mean=iid1/n,var=(iid1^2-S)/n),coef=c(mean=mu,var=S),mean=mu,var=S)
}


##' @export
iid.data.frame <- function(x,...) {
    if (!all(apply(x[1,,drop=FALSE],2,function(x) inherits(x,c("numeric","integer")))))
        stop("Don't know how to handle data.frames of this type")
    iid(as.matrix(x))
}

##' @export
parlabels <- function(x,exo=FALSE) {
  res <- c(unlist(intfix(x)[unlist(lapply(intfix(x), function(y) !is.na(y) & !is.numeric(y)))]),
           regfix(x)$labels[!is.na(regfix(x)$labels)],
           covfix(x)$labels[!is.na(covfix(x)$labels)])
  if (!is.null(x$exfix))
    res <- c(res,
           unlist(x$exfix[!is.na(x$exfix) & !is.numeric(x$exfix)]))
  if (exo)
    res <- intersect(res,index(Model(x))$exogenous)
  return(res)
}

###{{{ multigroup

##' @export
multigroup <- function(models, datasets, fix, exo.fix=TRUE, keep=NULL, missing=FALSE, ...) {
    nm <- length(models)
    if (nm!=length(datasets)) stop("Supply dataset for each model")
    if (nm<2) stop("Two or more groups neeeded")
    mynames <- names(models)

    ## Check for random slopes
    xfix <- list()
    for (i in seq_len(nm)) {
        x0 <- models[[i]]
        data0 <- datasets[[i]]
        xfix0 <- colnames(data0)[(colnames(data0)%in%parlabels(x0,exo=TRUE))]
        xfix <- c(xfix, list(xfix0))
    }
    if (missing(fix)) {
        fix <- !any(unlist(lapply(xfix, function(x) length(x)>0)))
    }

    for (i in seq_len(nm)) {
        x0 <- models[[i]]
        data0 <- datasets[[i]]
        if (length(exogenous(x0)>0)) {
            catx <- categorical2dummy(x0,data0)
            models[[i]] <- catx$x; datasets[[i]] <- catx$data
        }
        if (!lava.options()$exogenous) exogenous(models[[i]]) <- NULL
    }

    models.orig <- NULL
    ######################
    ### MLE with MAR mechanism
    ######################
    if (missing) {

        reservedpars <- c()
        mynpar <- c()
        for (i in seq_len(nm)) {
            ## Fix some parameters (predictors,latent variables,...)

            d0 <- datasets[[i]][1,,drop=FALSE]; d0[,] <- 1
            if (fix)
                models[[i]] <- fixsome(models[[i]], exo.fix=exo.fix, measurement.fix=fix, data=d0)
            ## Find named/labelled parameters
            rpar <- unique(parlabels(models[[i]]))
            reservedpars <- c(reservedpars, rpar)
            mynpar <- c(mynpar, with(index(models[[1]]), npar+npar.mean+npar.ex))
        }; reservedpars <- unique(reservedpars)
        nonamepar <- sum(mynpar)
        ## Find unique parameter-names for all parameters
        newpars <- c()
        i <- 0
        pos <- 1
        while(pos<=nonamepar) {
            i <- i+1
            newname <- paste0("par",i)
            if (!(newname%in%reservedpars)) {
                newpars <- c(newpars,newname)
                pos <- pos+1
            }
        }

        pos <- 0
        models0 <- list()
        datasets0 <- list()
        complidx <- c()
        nmodels <- 0
        modelclass <- c()
        nmis <- c()
        for (i in seq_len(nm)) {
            myvars <- unlist(intersect(colnames(datasets[[i]]),c(vars(models[[i]]),xfix[[i]],keep)))
            mydata <- datasets[[i]][,myvars]
            if (any(is.na(mydata))) {
                if (i>1) pos <- pos+mynpar[i-1]
                models[[i]] <- baptize(models[[i]],newpars[pos+seq_len(mynpar[i])] ,overwrite=FALSE)
                val <- missingModel(models[[i]],mydata,fix=FALSE,keep=keep,...)
                nmodels <- c(nmodels,length(val$models))
                complidx <- c(complidx,val$pattern.allcomp+nmodels[i]+1)
                nmis0 <- rowSums(val$patterns);
                allmis <- which(nmis0==ncol(val$patterns))
                if (length(allmis)>0) nmis0 <- nmis0[-allmis]
                nmis <- c(nmis,nmis0)
                datasets0 <- c(datasets0, val$datasets)
                models0 <- c(models0, val$models)
                modelclass <- c(modelclass,rep(i,length(val$models)))
            } else {
                datasets0 <- c(datasets0, list(mydata))
                models0 <- c(models0, list(models[[i]]))
                modelclass <- c(modelclass,i)
                nmis <- c(nmis,0)
            }
        }

        models.orig <- models

        suppressWarnings(
            val <- multigroup(models0,datasets0,fix=FALSE,missing=FALSE,exo.fix=TRUE,...)
        )
        val$models.orig <- models.orig; val$missing <- TRUE
        val$complete <- complidx-1
        val$mnames <- mynames
        attributes(val)$modelclass <- modelclass
        attributes(val)$nmis <- nmis
        return(val)
    }


    ######################
    ### Usual analysis:
    ######################
    warned <- FALSE
    for (i in seq_len(nm)) {
        if (inherits(datasets[[i]],c("data.frame","matrix"))) {
            myvars <- intersect(colnames(datasets[[i]]),c(vars(models[[i]]),xfix[[i]],keep))
            if (any(is.na(datasets[[i]][,myvars]))) {
                if (!warned)
                    warning(paste0("Missing data encountered. Going for complete-case analysis"))
                warned  <- TRUE
                datasets[[i]] <- na.omit(datasets[[i]][,myvars,drop=FALSE])
            }
        }
    }

    exo <- exogenous(models)
    means <- lvms <- As <- Ps <- ps <- exs <- datas <- samplestat <- list()
    for (i in seq_len(nm)) {

        if (!is.null(exogenous(models[[i]]))) {
            if (any(is.na(exogenous(models[[i]])))) {
                exogenous(models[[i]]) <- exo
            }
        }

        mydata <- datasets[[i]]
        mymodel <- fixsome(models[[i]], data=mydata, measurement.fix=fix, exo.fix=exo.fix)
        mymodel <- updatelvm(mymodel,zeroones=TRUE,deriv=TRUE)

        P <- index(mymodel)$P1; P[P==0] <- NA
        P[!is.na(P) & !is.na(mymodel$covpar)] <- mymodel$covpar[!is.na(P) & !is.na(mymodel$covpar)]

        A <- index(mymodel)$M1; A[A==0] <- NA
        A[!is.na(A) & !is.na(mymodel$par)] <- mymodel$par[!is.na(A) & !is.na(mymodel$par)]

        mu <- unlist(mymodel$mean)[which(index(mymodel)$v1==1)]
        #ex <- names(mymodel$expar)[which(index(mymodel)$e1==1)]
        ex <- mymodel$exfix
        if (length(ex)>0) {
            if (any(is.na(ex))) ex[is.na(ex)] <- mymodel$expar[is.na(ex)]
            ex <- ex[which(index(mymodel)$e1==1)]
        }

        p <- pars(mymodel, A, P, e=ex)
        p[p=="1"] <- NA

        means <- c(means, list(mu))
        lvms <- c(lvms, list(mymodel))
        datas <- c(datas, list(mydata))
        samplestat <- c(samplestat, list(procdata.lvm(models[[i]],data=mydata)))
        As <- c(As, list(A))
        Ps <- c(Ps, list(P))
        ps <- c(ps, list(p))
        exs <- c(exs, list(ex))
    };

    ######
    pp <- unlist(ps)
    parname <- unique(pp[!is.na(pp)])
    pidx <- is.na(char2num(parname))
    parname <- unique(unlist(pp[!is.na(pp)]));
    nfree <- sum(is.na(pp)) + length(parname)

    if (nfree>0) {
        pp0 <- lapply(ps, is.na)
        usedname <- cbind(parname, rep(NA,length(parname)))
        counter <- 1
        pres <- pres0 <- pp0
        for (i in seq_len(length(pp0))) {
            if (length(pp0[[i]]>0))
                for (j in seq_len(length(pp0[[i]]))) {
                    pidx <- match(ps[[i]][j],parname)
                    if (pp0[[i]][j]) {
                        pres[[i]][j] <- paste0("p",counter)
                        pres0[[i]][j] <- counter
                        counter <- counter+1
                    } else if (!is.na(pidx)) {
                        if (!is.na(usedname[pidx,2])) {
                            pres[[i]][j] <- usedname[pidx,2]
                            pres0[[i]][j] <- char2num(substr(pres[[i]][j],2,nchar(pres[[i]][j])))
                        } else {
                            val <- paste0("p",counter)
                            pres[[i]][j] <- val
                            pres0[[i]][j] <- counter
                            usedname[pidx,2] <- val
                            counter <- counter+1
                        }
                    } else {
                        pres[[i]][j] <- NA
                    }
                }
        }
        mypar <- paste0("p",seq_len(nfree))
        myparPos <- pres0
        myparpos <- pres
        myparlist <- lapply(pres, function(x) x[!is.na(x)])
    } else {
        myparPos <- NULL
        mypar <- NULL
        myparpos <- NULL
        myparlist <- NULL
    }

    ### Mean parameter

    mm <- unlist(means)
    meanparname <- unique(mm[!is.na(mm)])
    midx <- is.na(char2num(meanparname));
    meanparname <- meanparname[midx]
    any.mean <- sum(is.na(mm)) + length(meanparname)
    nfree.mean <- sum(is.na(mm)) + length(setdiff(meanparname,parname))
    ## mean.fixed <- na.omit(match(parname,mm))
    mean.omit <- lapply(means,function(x) na.omit(match(parname,x)))

    if (any.mean>0) {
        mm0 <- lapply(means, is.na)
        usedname <- cbind(meanparname, rep(NA,length(meanparname)))
        counter <- 1
        res0 <- res <- mm0
        for (i in seq_len(length(mm0))) {
            if (length(mm0[[i]])>0)
                for (j in seq_len(length(mm0[[i]]))) {
                    midx <- match(means[[i]][j],meanparname)
                    if (mm0[[i]][j]) {
                        res[[i]][j] <- paste0("m",counter)
                        res0[[i]][j] <- counter
                        counter <- counter+1
                    } else if (!is.na(midx)) {
                        pidx <- match(meanparname[midx],pp)
                        if (!is.na(pidx)) {
                            res[[i]][j] <- unlist(myparlist)[pidx]
                            res0[[i]][j] <- char2num(substr(res[[i]][j],2,nchar(res[[i]][j]))) +
                                nfree.mean
                        } else {
                            if (!is.na(usedname[midx,2])) {
                                res[[i]][j] <- usedname[midx,2]
                                res0[[i]][j] <- char2num(substr(res[[i]][j],2,nchar(res[[i]][j])))
                            } else {
                                val <- paste0("m",counter)
                                res[[i]][j] <- val
                                res0[[i]][j] <- counter
                                usedname[midx,2] <- val
                                counter <- counter+1
                            }
                        }
                    } else {
                        res[[i]][j] <- NA
                    }
                }
        }
        mymeanPos <- res0
        mymeanpos <- res
        mymeanlist <- lapply(res, function(x) x[!is.na(x)])
        mymean <- unique(unlist(mymeanlist))
    } else {
        mymeanPos <- NULL
        mymean <- NULL
        mymeanpos <- NULL
        mymeanlist <- NULL
    }

    ### Extra parameters

    m0 <- p0 <- c()
    coefs <- coefsm <- mm0 <- mm <- pp0 <- pp <- c()
    for (i in seq_len(length(myparPos))) {
        mi <- mymeanPos[[i]]
        pi <- myparPos[[i]]
        p1 <- setdiff(pi,p0)
        p0 <- c(p0,p1)
        ##    pp0 <- c(pp0,list(match(p1,pi)+nfree.mean))
        pp0 <- c(pp0,list(match(p1,pi)))
        if (length(mean.omit[[i]])>0) mi <- mi[-mean.omit[[i]]]
        m1 <- setdiff(mi,m0)
        m0 <- c(m0,m1)
        mm0 <- c(mm0,list(match(m1,mi)))
        pp <- c(pp,list(c(m1,p1+nfree.mean)))
        if (length(p1)>0)
            coefs <- c(coefs,paste(coef(lvms[[i]],fix=FALSE,mean=FALSE)[pp0[[i]]],i,sep="@"))
            #coefs <- c(coefs,paste(i,coef(lvms[[i]],fix=FALSE,mean=FALSE)[pp0[[i]]],sep="@"))
        if (length(m1)>0) {
            coefsm0 <- paste(coef(lvms[[i]],fix=FALSE,mean=TRUE)[mm0[[i]]],i,sep="@")
            #coefsm0 <- paste(i,coef(lvms[[i]],fix=FALSE,mean=TRUE)[mm0[[i]]],sep="@")
            coefsm <- c(coefsm,coefsm0)
        }
    }
    coefs <- c(coefsm,coefs)

    res <- list(npar=nfree, npar.mean=nfree.mean,
                ngroup=length(lvms), names=mynames,
                lvm=lvms, data=datas, samplestat=samplestat,
                A=As, P=Ps, expar=exs,
                meanpar=names(mu), name=coefs, coef=pp, coef.idx=pp0,
                par=mypar, parlist=myparlist,  parpos=myparpos,
                mean=mymean, meanlist=mymeanlist, meanpos=mymeanpos,
                parposN=myparPos,
                meanposN=mymeanPos,
                models.orig=models.orig, missing=missing
                )
    class(res) <- "multigroup"
    checkmultigroup(res)
    return(res)
}

###}}}

###{{{ checkmultigroup

checkmultigroup <- function(x) {
    ## Check validity:
    for (i in seq_len(x$ngroup)) {
        if (nrow(x$data[[i]])<2) {
            warning("With only one observation in the group, all parameters should be inherited from another a group!")
        }
    }
}

###}}} checkmultigroup

##' Add Confidence limits bar to plot
##'
##' @title Add Confidence limits bar to plot
##' @param x Position (x-coordinate if vert=TRUE, y-coordinate otherwise)
##' @param lower Lower limit (if NULL no limits is added, and only the
##' center is drawn (if not NULL))
##' @param upper Upper limit
##' @param center Center point
##' @param line If FALSE do not add line between upper and lower bound
##' @param delta Length of limit bars
##' @param centermark Length of center bar
##' @param pch Center symbol (if missing a line is drawn)
##' @param blank If TRUE a white ball is plotted before the center is
##' added to the plot
##' @param vert If TRUE a vertical bar is plotted. Otherwise a horizontal
##' bar is used
##' @param polygon If TRUE polygons are added between 'lower' and 'upper'
##' @param alpha transparency of fill-color of polygon (alpha<1)
##' @param step Type of polygon (step-function or piecewise linear)
##' @param ... Additional low level arguments (e.g. col, lwd, lty,...)
##' @seealso \code{confband}
##' @export
##' @keywords iplot
##' @aliases confband forestplot plot_region
##' @author Klaus K. Holst
##' @examples
##' plot(0,0,type="n",xlab="",ylab="")
##' confband(0.5,-0.5,0.5,0,col="darkblue")
##' confband(0.8,-0.5,0.5,0,col="darkred",vert=FALSE,pch=1,cex=1.5)
##'
##' set.seed(1)
##' K <- 20
##' est <- rnorm(K)
##' se <- runif(K,0.2,0.4)
##' x <- cbind(est,est-2*se,est+2*se,runif(K,0.5,2))
##' x[c(3:4,10:12),] <- NA
##' rownames(x) <- unlist(lapply(letters[seq(K)],function(x) paste(rep(x,4),collapse="")))
##' rownames(x)[which(is.na(est))] <- ""
##' signif <- sign(x[,2])==sign(x[,3])
##' forestplot(x,text.right=FALSE)
##' forestplot(x[,-4],sep=c(2,15),col=signif+1,box1=TRUE,delta=0.2,pch=16,cex=1.5)
##' forestplot(x,vert=TRUE,text=FALSE)
##' forestplot(x,vert=TRUE,text=FALSE,pch=NA)
##' ##forestplot(x,vert=TRUE,text.vert=FALSE)
##' ##forestplot(val,vert=TRUE,add=TRUE)
##'
##' z <- seq(10)
##' zu <- c(z[-1],10)
##' plot(z,type="n")
##' confband(z,zu,rep(0,length(z)),col=Col("darkblue"),polygon=TRUE,step=TRUE)
##' confband(z,zu,zu-2,col=Col("darkred"),polygon=TRUE,step=TRUE)
##'
##' z <- seq(0,1,length.out=100)
##' plot(z,z,type="n")
##' confband(z,z,z^2,polygon=TRUE,col="darkred", alpha=0.1)
##'
##' set.seed(1)
##' k <- 10
##' x <- seq(k)
##' est <- rnorm(k)
##' sd <- runif(k)
##' val <- cbind(x,est,est-sd,est+sd)
##' par(mfrow=c(1,2))
##' plot(0,type="n",xlim=c(0,k+1),ylim=range(val[,-1]),axes=FALSE,xlab="",ylab="")
##' axis(2)
##' confband(val[,1],val[,3],val[,4],val[,2],pch=16,cex=2)
##' plot(0,type="n",ylim=c(0,k+1),xlim=range(val[,-1]),axes=FALSE,xlab="",ylab="")
##' axis(1)
##' confband(val[,1],val[,3],val[,4],val[,2],pch=16,cex=2,vert=FALSE)
##'
##' x <- seq(0, 3, length.out=20)
##' y <- cos(x)
##' yl <- y - 1
##' yu <- y + 1
##' plot_region(x, y, yl, yu)
##' plot_region(x, y, yl, yu, type='s', col="darkblue", add=TRUE)
##'
confband <- function(x, lower, upper, center=NULL,
                     line=TRUE, delta=0.07,
                     centermark=0.03,
                     pch, blank=TRUE, vert=TRUE,
                     polygon=FALSE, alpha=1.0,
                     step=FALSE,...) {
  if (polygon) {
    if (step) {
      x1 <- rep(x,each=2)[-1]
      y1 <- rep(lower, each=2);  y1 <- y1[-length(y1)]
      x2 <- rep(rev(x),each=2); x2 <- x2[-length(x2)]
      y2 <- rep(rev(upper), each=2)[-1]
      xx <- c(x1,x2)
      if (!is.null(center))
        center <- rep(center, each=2)[-1]
      yy <- c(y1,y2)
    } else {
      xx <- c(x, rev(x))
      yy <- c(lower, rev(upper))
    }
    dots <- list(...)
    dots$x <- xx
    dots$y <- yy
    if (alpha<1.0 && !is.null(dots$col)) {
      dots$col <- Col(dots$col, alpha)
    }
    do.call(graphics::polygon, dots)
    if (line && !is.null(center)) {
      mlines <- function(x, y, ...,
                         angle, density,
                         border, fillOddEven) {
        if (step) {
          lines(x, y, type="s", ...)
        } else {
          lines(x, y, ,...)
        }
      }
      mlines(xx[seq(length(xx)/2)],center,...)
    }
    return(invisible(NULL))
  }
  if (vert) {
    if (line && !missing(lower) && !missing(upper))
      segments(x,lower,x,upper,...)
    if (!missing(lower))
      segments(x-delta,lower,x+delta,lower,...)
    if (!missing(upper))
      segments(x-delta,upper,x+delta,upper,...)
    if (!is.null(center)) {
      if (!missing(pch)) {
        if (blank) {
          points(x,center,pch=16,col="white")
        }
        points(x,center,pch=pch,...)
      } else {
        segments(x-centermark,center,x+centermark,center,...)
      }
    }
  } else {
    if (line && !missing(lower) && !missing(upper))
      segments(lower,x,upper,x,...)
    if (!missing(lower))
      segments(lower,x-delta,lower,x+delta,...)
    if (!missing(upper))
      segments(upper,x-delta,upper,x+delta,...)

    if (!is.null(center)) {
      if (!missing(pch)) {
        if (blank)
          points(center,x,pch=16,col="white")
        points(center,x,pch=pch,...)
      } else {
        segments(center,x-centermark,center,x+centermark,...)
      }
    }
  }
  if (missing(lower)) lower <- NULL
  if (missing(upper)) upper <- NULL
  invisible(c(x,lower,upper,center))
}


##' @export
forestplot <- function(x,lower,upper,line=0,labels,
                text=TRUE,text.right=text,text.fixed=NULL,text.vert=TRUE,
                adj=NULL,
                delta=0,axes=TRUE,cex=1,pch=15,
                xlab="",ylab="",sep,air,
                xlim,ylim,mar,box1=FALSE,box2=FALSE,
                vert=FALSE,cex.axis=1,cex.estimate=0.8,
                add=FALSE,
                reset.par=FALSE,
                ...) {
    if (is.matrix(x)) {
        lower <- x[,2]; upper <- x[,3]
        if (ncol(x)>3) cex <- x[,4]
        x <- x[,1]
    }
    if (missing(mar) && !add) {
        if (vert) {
            mar <- c(8,4,1,1)
        } else {
            mar <- c(4,8,1,1)
        }
    }
    if (missing(labels)) labels <- names(x)
    K <- length(x)
    onelayout <- FALSE
    if (!add) {
        def.par <- par(no.readonly=TRUE)
        if (reset.par) on.exit(par(def.par))
        if (text.right) {
            if (vert) {
                layout(rbind(1,2),heights=c(0.25,0.75))
            } else {
                layout(cbind(2,1),widths=c(0.75,0.25))
            }
        } else {
            onelayout <- TRUE
            layout(1)
        }
    }
    if (vert) {
        if (missing(ylim)) {
            if (missing(air)) air <- max(upper-lower,na.rm=TRUE)*0.5
            ylim <- range(c(x,lower-air,upper+air),na.rm=TRUE)
        }
        if (missing(xlim)) xlim <- c(0,K) + 0.5
    } else {
        if (missing(ylim)) ylim <- c(0,K) + 0.5
        if (missing(xlim)) {
            if (missing(air)) air <- max(upper-lower,na.rm=TRUE)*0.5
            xlim <- range(c(x,lower-air,upper+air),na.rm=TRUE)
        }
    }
    args0 <- list(...)
    formatCargsn <- names(formals(args(formatC)))[-1]
    nn <- setdiff(names(args0),formatCargsn)
    plotargs <- args0[nn]

    mainplot <- function(...) {
        par(mar=mar) ## bottom,left,top,right
        do.call("plot",c(list(x=0,type="n",axes=FALSE,xlab=xlab,ylab=ylab,xlim=xlim,ylim=ylim),plotargs))
        if (box1) box()
        if (axes) {
            if (vert) {
                axis(2,cex.axis=cex.axis)
            } else {
                axis(1,cex.axis=cex.axis)
            }
        }
    }
    if (onelayout && !add) mainplot()

    if (text) {
        xpos <- upper
        if (text.right && !add) {
            if (vert) {
                par(mar=c(0,mar[2],0,mar[4]))
            } else {
                par(mar=c(mar[1],0,mar[3],0))
            }
            plot.new()
            if (vert) {
                plot.window(xlim=xlim,ylim=c(0,0.5))
            } else {
                plot.window(ylim=ylim,xlim=c(0,0.5))
            }
            if (box2) box()
            xpos[] <- 0
        }
        if (!is.null(text.fixed)) {
            if (is.logical(text.fixed) && text.fixed) text.fixed <- max(xpos)
            xpos <- rep(text.fixed,length.out=K)
        }
        nn <- intersect(names(args0),formatCargsn)
        args <- args0[nn]
        for (i in seq_len(K)) {
            st <- c(do.call(formatC,c(list(x=x[i]),args)),
                    paste0("(",
                           do.call(formatC,c(list(x=lower[i]),args)),"; ",
                           do.call(formatC,c(list(x=upper[i]),args)),")"))
            if (text.vert) {
                st <- paste0(" ",st[1]," ",st[2],collapse="")
                st <- paste(" ", st)
            }
            if (vert) {
                if (!is.na(x[i])) {
                    if (!text.vert) {
                        if (text.right) xpos[i] <- xpos[i]+0.025
                        graphics::text(i,xpos[i],paste(st,collapse="\n"),xpd=TRUE,
                                       offset=3, cex=cex.estimate, adj=adj)
                    } else {
                        if (!is.na(x[i])) {
                          graphics::text(i,xpos[i],st,xpd=TRUE, srt=90, offset=0, pos=4, cex=cex.estimate, adj=adj)
                        }
                    }
                }
            } else {
                if (!is.na(x[i])) {
                  graphics::text(xpos[i],i,trim(st),xpd=TRUE,pos=4,cex=cex.estimate, adj=adj)
                }
            }
        }
    }

    if (!onelayout && !add) mainplot()

    if (!is.null(line)) {
        if (vert) {
            abline(h=line,lty=2,col="lightgray")
        } else {
            abline(v=line,lty=2,col="lightgray")
        }
    }
    if (!missing(sep)) {
        if (vert) {
            abline(v=sep+.5,col="gray")
        } else {
            abline(h=sep+.5,col="gray")
        }
    }
    do.call("confband",
            c(list(x=seq(K),lower=lower,upper=upper,x,
                   pch=pch,cex=cex,vert=vert,blank=FALSE),
              plotargs))
    if (!add) {
        if (is.null(adj)) adj <- NA
        if (vert) {
            mtext(labels,1,at=seq(K),las=2,line=1,cex=cex.axis, adj=adj)
        } else {
            mtext(labels,2,at=seq(K),las=2,line=1,cex=cex.axis, adj=adj)
        }
    }
}


##' @export
plot_region <- function(x, y=NULL, lower=NULL, upper=NULL,
                        line=TRUE, add=FALSE,
                        border=NA, col=1, alpha=0.5,
                        type="l", ...) {
  if (is.matrix(x) && ncol(x)==4) {
    if (ncol(x)>=4) {
      y <- x[, 2]
      lower <- x[, 3]
      upper <- x[, 4]
      x <- x[, 1]
    }
    if (ncol(x)==3) {
      lower <- x[, 2]
      upper <- x[, 3]
      x <- x[, 1]
      line <- FALSE
    }
  } else {
    if (is.matrix(y)) {
      if (ncol(y)==2) {
        lower <- y[, 1]
        upper <- y[, 2]
      } else if (ncol(y)>2) {
        lower <- y[, 2]
        upper <- y[, 3]
        y <- y[, 1]
      }
    }
  }
  if (!add) {
    yy <- range(c(y,lower,upper))
    plot(range(x), range(yy), ..., type="n")
  }
  confband(x, lower, upper, y, line=FALSE, polygon=TRUE,
           step=(type=="s"), col=Col(col, alpha),
           border=border, ...)
  if (line) {
    lines(x, y, type=type, col=col, ...)
  }
}

rmse1 <- function(fit, data, response=NULL, ...) {
    yhat <- predict(fit, newdata = data, ...)
    if (is.null(response)) response <- endogenous(fit)
    y <- data[, response]
    c(RMSE = mean(as.matrix(y - yhat)^2)^.5)
}

cv <- function(modelList, data, K=5, rep=1, perf, seed=NULL, shared=NULL, ...) {
    if (is.vector(data)) data <- cbind(data)
    if (missing(perf)) perf <- rmse1
    if (!is.list(modelList)) modelList <- list(modelList)
    nam <- names(modelList)
    if (is.null(nam)) nam <- paste0("model", seq_along(modelList))
    args0 <- list(...)
    args <- args0
    if (!is.null(shared)) {
        sharedres <- shared(data,...)
        args <- c(args, sharedres)
    }
    ## Models run on full data:
    if (is.function(modelList[[1]])) {
        fit0 <- lapply(modelList, function(f) do.call(f,c(list(data),args)))
    } else {
        fit0 <- modelList
    }
    ## In-sample predictive performance:
    perf0 <- lapply(fit0, function(fit) do.call(perf,c(list(fit,data=data),args)))
    namPerf <- names(perf0[[1]])
    names(fit0) <- names(perf0) <- nam
    n <- NROW(data)
    M <- length(perf0)      # Number of models
    P <- length(perf0[[1]]) # Number of performance measures
    if (!is.null(seed)) {
        if (!exists(".Random.seed", envir = .GlobalEnv, inherits = FALSE))
            runif(1)
        R.seed <- get(".Random.seed", envir = .GlobalEnv)
        set.seed(seed)
        RNGstate <- structure(seed, kind = as.list(RNGkind()))
        on.exit(assign(".Random.seed", R.seed, envir = .GlobalEnv))
    }
    if (K==0) {
        rep <- 1
        K <- 1
        folds <- list(csplit(seq(n)))
    } else {
        folds <- foldr(n,K,rep)
    }
    arg <- expand.grid(R=seq(rep),K=seq(K)) #,M=seq_along(modelList))
    dim <- c(rep,K,M,P)
    PerfArr <- array(0,dim)
    dimnames(PerfArr) <- list(NULL,NULL,nam,namPerf)

    pb <- progressr::progressor(along=seq(nrow(arg)))
    ff <- function(i) {
        R <- arg[i,1]
        k <- arg[i,2]
        fold <- folds[[R]]
        dtest <- data[fold[[k]],,drop=FALSE]
        dtrain <- data[unlist(fold[-k]),,drop=FALSE]
        args <- args0
        if (!is.null(shared)) {
            sharedres <- shared(dtrain,...)
            args <- c(args, sharedres)
        }
        if (is.function(modelList[[1]])) {
            fits <- lapply(modelList, function(f) do.call(f,c(list(dtrain),args)))
        } else {
            fits <- lapply(modelList, function(m) do.call(update,c(list(m,data=dtrain),args)))
        }
        perfs <- lapply(fits, function(fit) do.call(perf,c(list(fit,data=dtest),args)))
        pb()
        do.call(rbind,perfs)
    }
    val <- future_mapply(ff, seq(nrow(arg)), SIMPLIFY=FALSE, future.seed=TRUE)
    for (i in seq(nrow(arg))) {
        R <- arg[i, 1]
        k <- arg[i, 2]
        PerfArr[R, k, ,] <- val[[i]]
    }

    cc <- apply(PerfArr, 3:4, function(x) mean(x))
    if (length(nam)==nrow(cc)) rownames(cc) <- nam

    structure(list(cv=PerfArr,
                   coef = cc,
                   call=match.call(),
                   names=nam,
                   rep=rep, folds=K,
                   fit=fit0))
}

###{{{ gaussian

gaussian_method.lvm <- "nlminb2"
`gaussian_objective.lvm` <-
  function(x,p,data,S,mu,n,...) {
    mp <- modelVar(x,p=p,data=data,...)
    C <- mp$C ## Model specific covariance matrix
    xi <- mp$xi ## Model specific mean-vector
    if (!lava.options()$allow.negative.variance && any(diag(mp$P)<0)) return(NaN)

    iC <- Inverse(C,det=TRUE, symmetric = TRUE)
    detC <- attributes(iC)$det
    if (n<2) {
      z <- as.numeric(data-xi)
      val <- log(detC) + tcrossprod(z,crossprod(z,iC))[1]
      return(0.5*val)
    }
    if (!is.null(mu)){
      W <- suppressMessages(crossprod(rbind(mu-xi)))
      T <- S+W
    } else {
      T <- S
    }
    res <- n/2*log(detC) + n/2*tr(T%*%iC) ## Objective function (Full Information ML)
    ## if (any(attr(iC,"lambda")<1e-16)) res <- res-1e2
    return(res)
  }

`gaussian_hessian.lvm` <- function(x,p,n,...) {
  dots <- list(...); dots$weights <- NULL
  do.call("information", c(list(x=x,p=p,n=n),dots))
}

gaussian_gradient.lvm <-  function(x,p,data,S,mu,n,...) {
  dots <- list(...); dots$weights <- NULL
  if (n>2) data <- NULL
  val <- -gaussian_score.lvm(x,p=p,S=S,mu=mu,n=n,data=data,reindex=FALSE,...)
  if (!is.null(nrow(val))) {
    val <- colSums(val)
  }
  val
}

gaussian_score.lvm <- function(x, data, p, S, n, mu=NULL, weights=NULL, debug=FALSE, reindex=FALSE, mean=TRUE, constrain=TRUE, indiv=FALSE,...) {

  if (!is.null(data)) {
    if ((nrow(data)<2 | !is.null(weights))| indiv)
    {
      mp <- modelVar(x,p,data=data[1,])
      iC <- Inverse(mp$C,det=FALSE, symmetric = TRUE)
      MeanPar <- attributes(mp)$meanpar
      D <- with(attributes(mp), deriv.lvm(x, meanpar=MeanPar, p=pars, mom=mp, mu=NULL)) ##, all=length(constrain(x))>0))
      myvars <- (index(x)$manifest)
      data <- data[,myvars,drop=FALSE]
      score <- matrix(ncol=length(p),nrow=NROW(data))
      score0 <- -1/2*as.vector(iC)%*%D$dS
      if (!is.null(weights)) {
        W0 <- diag(nrow=length(myvars))
        widx <- match(colnames(weights),myvars)
      }

      for (i in seq_len(NROW(data))) {
        z <- as.numeric(data[i,])
        u <- z-as.numeric(mp$xi)
        if (!is.null(weights)) {
          W <- W0; diag(W)[widx] <- as.numeric(weights[i,])
          score[i,] <-
            as.numeric(crossprod(u,iC%*%W)%*%D$dxi +
                       -1/2*(as.vector((iC
                         - iC %*% crossprod(rbind(u))
                         %*% iC)%*%W)) %*% D$dS
                       )
        } else {
          score[i, ] <-
            as.numeric(score0 + crossprod(u, iC)%*%D$dxi +
                       1/2*as.vector(iC%*%crossprod(rbind(u))%*%iC)%*%D$dS)
        }
      }; colnames(score) <- names(p)
      return(score)
    }
  }

  ### Here the emperical mean and variance of the population are sufficient statistics:
  if (missing(S)) {
    data0 <- na.omit(data[,manifest(x),drop=FALSE])
    n <- NROW(data0)
    S <- cov(data0)*(n-1)/n
    mu <- colMeans(data0)
  }
  mp <- modelVar(x,p)
  C <- mp$C
  xi <- mp$xi
  iC <- Inverse(C,det=FALSE, symmetric = TRUE)
  Debug("Sufficient stats.",debug)
  if (!is.null(mu) & !is.null(xi)) {
    W <- crossprod(rbind(mu-xi))
    T <- S+W
  } else {
    T <- S
  }
  D <- deriv.lvm(x, meanpar=attributes(mp)$meanpar, mom=mp, p=p, mu=mu, mean=mean)
  vec.iC <- as.vector(iC)
  if (lava.options()$devel) {
    Grad <- numeric(length(p))
    imean <- with(index(x)$parBelongsTo,mean)
    Grad[-imean] <- n/2*crossprod(D$dS[,-imean], as.vector(iC%*%T%*%iC)-vec.iC)
  } else {
    Grad <- n/2*crossprod(D$dS, as.vector(iC%*%T%*%iC)-vec.iC)
  }
  if (!is.null(mu) & !is.null(xi)) {
    if (!(lava.options()$devel)) {
      Grad <- Grad - (n/2*crossprod(D$dT,vec.iC))
    } else {
      Grad[with(index(x)$parBelongsTo,c(mean,reg))] <- Grad[with(index(x)$parBelongsTo,c(mean,reg))] - (n/2*crossprod(D$dT,vec.iC))
    }
  }
  res <- as.numeric(Grad)
  return(rbind(res))
}

###}}} gaussian

###{{{ gaussian variants

## Maximum Likelihood with numerical gradient + hessian
gaussian0_objective.lvm <- gaussian_objective.lvm

gaussian1_objective.lvm <- gaussian_objective.lvm
gaussian1_gradient.lvm <- function(...) gaussian_gradient.lvm(...)
gaussian1_hessian.lvm <- function(x,p,...) {
    myg2 <- function(p1) gaussian_gradient.lvm(x,p=p1,...)
    myg3 <- function(p1) numDeriv::jacobian(myg2,p1)
    return(myg3(p))
    ## myg <- function(p1) gaussian_objective.lvm(x,p=p1,...)
    ## numDeriv::hessian(myg,p)
}

## BHHH
gaussian2_method.lvm <- "NR"
gaussian2_objective.lvm <- gaussian_objective.lvm
gaussian2_gradient.lvm <- gaussian_gradient.lvm
gaussian2_hessian.lvm <- function(x,p,n,data,...) {
    S <- -score(x,p=p,n=n,data=data,indiv=TRUE,...)
    I <- t(S)%*%S
    attributes(I)$grad <- colSums(S)
    return(I)
}

###}}}

# ---- Merge, subset ------------------------------------------------------

#' Merge estimate objects
#'
#' @param x Object of class `estimate`
#' @param y Object of class `estimate`
#' @param ... Additional `estimate` objects or arguments
#' @param id Optional cluster variable
#' @param paired If TRUE a paired (matched) analysis is performed
#' @param labels Optional character vector of labels for the merged estimates
#' @param keep Optional character vector of parameter names to keep
#' @param subset Optional character vector of parameter names to subset
#' @param regex If TRUE, `keep` and `subset` are treated as regular expressions
#' @param sep Separator used for labeling
#' @param drop.ic If TRUE, drop the influence function from the result
#' @param ignore.case If TRUE, case is ignored in `keep`/`subset` matching
#' @param sort if true the returned influence function will be sorted according
#'   to the id variables (lexigraphically)
#' @seealso [c.estimate()]
#' @return Object of class `estimate` (see [estimate.default]).
#' @export
merge.estimate <- function(x, y,
                           ...,
                           id,
                           paired = FALSE,
                           labels = NULL,
                           keep = NULL,
                           subset = NULL,
                           regex = FALSE,
                           sep = FALSE,
                           drop.ic = FALSE,
                           ignore.case = FALSE,
                           sort = FALSE) {
    if (missing(y)) {
      objects <- c(list(x), list(...))
    } else {
      objects <- c(list(x), list(y), list(...))
    }
    if (drop.ic) {
      for (i in seq_along(objects))
      if (inherits(objects[[i]], "estimate")) {
        objects[[i]]$IC <- NULL
      }
    }
    if (length(nai <- names(objects)=="NA")>0)
    names(objects)[which(nai)] <- ""
    if (!missing(subset)) {
      if (regex) {
        warning("regular expression not supported by `merge` operation")
      }
      coefs <- unlist(lapply(objects, function(x) coef(x,messages=0)[subset]))
    } else {
      coefs <- unlist(lapply(objects, function(x) coef(x, messages=0)))
    }
    if (!is.null(labels)) {
      names(coefs) <- labels
    } else {
      names(coefs) <- make.unique(names(coefs))
    }
    if (regex) {
      if (!is.null(keep)) {
        cc <- names(coefs)
        keep <- unlist(lapply(keep, function(x) {
          cc[grepl(x, cc, perl = TRUE, ignore.case=ignore.case)]
        }))
      }
    }
    hasIC <- unlist(lapply(objects, function(x) !is.null(IC(x))))
    if (sum(!hasIC) > 0L) { # Some objects do not have influence function
      npar <- unlist(lapply(objects, function(x) length(coef(x, messages=0))))
      col_ends <- cumsum(npar)
      col_starts <- col_ends - npar + 1L
      V <- matrix(NA, nrow=sum(npar), ncol=sum(npar))

      ic_idx <- which(hasIC)
      if (length(ic_idx) > 0L) {
        if (length(ic_idx) > 1L) {
          ic_args <- c(objects[ic_idx],
                       list(paired = paired),
                       if (!missing(id)) list(id = id[ic_idx]))
          m_ic  <- do.call(merge, ic_args)
        } else m_ic <- objects[[ic_idx]]
        ic_cols <- unlist(Map(`:`, col_starts[ic_idx], col_ends[ic_idx]))
        V[ic_cols, ic_cols] <- vcov(m_ic)
      }
      # Fill diagonal blocks for non-IC objects
      for (k in which(!hasIC)) {
        pos <- col_starts[k]:col_ends[k]
        V[pos, pos] <- suppressMessages(vcov(objects[[k]]))
      }
      return(estimate(coef=coefs, vcov=V, keep=keep))
    }
    if (!missing(id) && is.null(id)) { # Independence betw. datasets in x,y,...
        nn <- unlist(lapply(
          objects,
          function(x) nrow(x$IC)
        ))
        cnn <- c(0, cumsum(nn))
        id <- list()
        for (i in seq_along(nn)) {
          id <- c(id, list(seq(nn[i]) + cnn[i]))
        }
    }
    if (missing(id)) {
      if (paired) { ## One-to-one dependence between observations in x,y,...
        id <- lapply(objects, function(x) {
          seq_len(NROW(x$IC))
        })
        } else {
            id <- lapply(objects, function(x) x$id)
        }
    } else {
        nn <- unlist(lapply(objects,function(x) NROW(IC(x))))
        if (length(id)==1 && is.logical(id)) {
            if (id) {
              if (any(nn[1]!=nn)) {
                stop("Expected objects of the same size: ",
                     paste(nn,collapse=","))
              }
              id0 <- seq(nn[1]); id <- c()
              for (i in seq(length(nn))) id <- c(id,list(id0))
            } else {
              id <- c()
              N <- cumsum(c(0,nn))
              for (i in seq(length(nn))) id <- c(id,list(seq(nn[i])+N[i]))
            }
        }
        if (length(id)!=length(objects)) {
          stop("Same number of id-elements as model objects expected")
        }
        idlen <- unlist(lapply(id,length))
        if (!identical(idlen,nn)) {
          stop("Wrong lengths of 'id': ",
               paste(idlen,collapse=","), "; ", paste(nn,collapse=","))
        }
    }
    ids <- ic_all <- c(); count <- 0
    first_id <- id[[1]]
    for (z in objects) {
        count <- count+1
        clidx <- NULL
        id0 <- id[[count]]
        icz <- IC(z)
        if (is.null(id0)) {
            id0 <- rownames(icz)
            if (is.null(id0)) stop("Need id for object number ", count)
        }
        if (!missing(subset)) icz <- icz[,subset,drop=FALSE]
        if (lava.options()$check.ic) {
          check_ic_mean_zero(icz)
        }
        if (!lava.options()$cluster.index) {
          ic0 <- matrix(
            unlist(by(icz,id0,colSums)),
            byrow = TRUE, ncol = ncol(icz))
            ids <- c(ids, list(base::sort(unique(id0))))
        } else {
          if (!requireNamespace("mets", quietly = TRUE)) {
            stop("'mets' package required")
          }
          clidx <- mets::cluster.index(id0, mat = icz, return.all = TRUE)
          ic0 <- clidx$X
          ids <- c(ids, list(id0[as.vector(clidx$firstclustid)+1]))
        }
        # order observations such that the order
        ic0 <- ic0 * NROW(ic0) / length(id0)
        ic_all <- c(ic_all, list(ic0))
    }
    id <- unique(unlist(ids))
    ic0 <- matrix(NA, nrow=length(id),ncol=length(coefs))
    model.index <- c()
    colpos <- 0
    for (i in seq(length(objects))) {
        relpos <- seq_along(coef(objects[[i]], messages=0))
        if (!missing(subset)) relpos <- seq_along(subset)
        ic0[match(ids[[i]], id), relpos + colpos] <- ic_all[[i]]
        midx <- list(relpos + colpos)
        ## }
        model.index <- c(model.index, midx)
        colpos <- colpos+tail(relpos,1)
    }
    rownames(ic0) <- id
    if (!sort) { # Reorder to original order of the first IC
      ord <- match(id, unique(c(first_id, id)))
      ord <- match(unique(c(first_id, id)), id)
      ic0 <- ic0[ord, , drop=FALSE]
      id <- id[ord]
    }
    ## Rescale each column according to I(obs)/pr(obs)
    for (i in seq(NCOL(ic0))) {
      pr <- mean(!is.na(ic0[,i]))
      ic0[,i] <- ic0[,i]/pr
    }
    ic0[is.na(ic0)] <- 0
    res <- estimate.default(
      coef = coefs, stack = FALSE, data = NULL,
      IC = ic0, id = id, keep = keep
      )
    if (is.null(keep) && sep) {
      res$model.index <- model.index
    }
    return(res)
}

#' @export
"%++%.estimate" <- function(x, ...) {
  merge(x, ...)
}

#' Concatenate estimate objects
#'
#' When all arguments are `estimate` objects, they are merged into a single
#' `estimate` object. When some arguments are not `estimate` objects but are
#' named numeric scalars/vectors, the merged `estimate` object is returned
#' with an `"extra"` attribute containing the provided values. This is useful
#' for bundling auxiliary per-iteration information (e.g., convergence status)
#' alongside an estimate for use with [sim.default()].
#'
#' @param ... `estimate` objects and/or named numeric values
#' @param as.list if TRUE the returned object will be of class `list` and not
#'   an `estimate` object.
#' @return An `estimate` object. If extra named numeric values were provided,
#'   the result carries an `"extra"` attribute (a named numeric vector).
#' @seealso [sim.default()] [merge.estimate()]
#' @details arguments `drop.ic`, `paired`, `sep` are passed to [merge.estimate]
#' @examples
#' e <- estimate(coef = c(a = 1, b = 2), vcov = diag(2) * 0.1)
#' # Bundle estimate with extra information
#' res <- c(e, converged = 1, niter = 10)
#' attr(res, "extra")
#' @export
c.estimate <- function(..., as.list = FALSE) {
  args <- list(...)
  if (as.list) { # fallback to default concatenation
    class(args[[1]]) <- "list"
    return(do.call(c, args))
  }
  # Handle names robustly
  arg_names <- names(args) %||% character(length(args))
  is_estimate <- vapply(args, inherits, logical(1L), "estimate")
  merge_args <- c("drop.ic", "paired", "sep")

  not_merge_arg <- which(arg_names %ni% merge_args)
  est_idx <- which(is_estimate)
  extra_idx <- setdiff(not_merge_arg, est_idx)
  lab <- arg_names[est_idx]
  # Collect extras from existing estimate objects (propagate "extra" attribute)
  extras_existing <- unlist(
    lapply(args[est_idx], function(x) attr(x, "extra"))
  )
  # Extract extra (non-estimate, non-merge) arguments
  extra <- c(extras_existing,
             if (length(extra_idx) > 0L) unlist(args[extra_idx]))
  # Blank non-merge names (estimate labels handled later, extras removed)
  arg_names[not_merge_arg] <- ""
  names(args) <- arg_names
  args[extra_idx] <- NULL
  # Merge estimate objects and apply potential merge_args
  res <- do.call(merge, args)
  # Add new labels
  newlabels <- names(coef(res))
  if (!is.null(lab)) {
    idx <- which(lab != "")
    newlabels[idx] <- lab[idx]
    res <- labels(res, newlabels)
  }
  # Attach extra as attribute (same pattern as c.summary.estimate)
  if (length(extra) > 0L) attr(res, "extra") <- extra
  res
}

#' @export
subset.estimate <- function(x, keep, ...) {
  estimate(x, keep = keep, ...)
}

#' @export
"[.estimate" <- function(x, i, ...) {
  subset(x, i, ...)
}

#' @export
with.estimate <- function(data, expr, ...) {
    # Recursively walk the expression tree and replace symbols
  # that match names in `data` with data["symbol"] calls
  replace_syms <- function(e) {
    # Base case: if it's a symbol, check if it matches a name in data
    if (is.symbol(e)) {
      nm <- as.character(e)
      if (nm %in% names(coef(data))) {
        # Replace symbol with data["nm"] call
        return(call("[", quote(data), nm))
      }
      return(e)
    }
    # Recursive case: walk the call tree
    if (is.call(e)) {
      return(as.call(lapply(e, replace_syms)))
    }
    # Literals (numbers, strings, etc.) — return as-is
    return(e)
  }
  # Substitute and transform the expression
  expr_sub  <- substitute(expr)
  expr_new  <- replace_syms(expr_sub)
  # Create a local environment where `data` exists,
  # with the parent frame as the enclosing environment
  eval_env <- new.env(parent = parent.frame())
  eval_env$data <- data
  # Evaluate in calling environment so non-estimate symbols still resolve
  eval(expr_new, envir = eval_env)
}

# ---- id / cluster  ------------------------------------------------------

#' @export
index.estimate <- function(x, ...) {
  return(x[["id"]])
}

#' @export
`index<-.estimate` <- function(x, ..., value) {
  estimate(x, id=value, ...)
}

# ---- Trigonometric Functions --------------------------------------------

#' @export
sin.estimate <- function(x, ...) {
  estimate(x, function(p) {
    y <- sin(p)
    structure(y, grad = diag(cos(p), nrow = length(p)))
  }, ...)
}

#' @export
cos.estimate <- function(x, ...) {
  estimate(x, function(p) {
    y <- cos(p)
    structure(y, grad = diag(-sin(p), nrow = length(p)))
  }, ...)
}

#' @export
tan.estimate <- function(x, ...) {
  estimate(x, function(p) {
    y <- tan(p)
    structure(y, grad = diag(1 / cos(p)^2, nrow = length(p)))
  }, ...)
}

# ---- Inverse Trigonometric Functions ------------------------------------

#' @export
asin.estimate <- function(x, ...) {
  estimate(x, function(p) {
    y <- asin(p)
    structure(y, grad = diag(1 / sqrt(1 - p^2), nrow = length(p)))
  }, ...)
}

#' @export
acos.estimate <- function(x, ...) {
  estimate(x, function(p) {
    y <- acos(p)
    structure(y, grad = diag(-1 / sqrt(1 - p^2), nrow = length(p)))
  }, ...)
}

#' @export
atan.estimate <- function(x, ...) {
  estimate(x, function(p) {
    y <- atan(p)
    structure(y, grad = diag(1 / (1 + p^2), nrow = length(p)))
  }, ...)
}

# ---- Hyperbolic Functions -----------------------------------------------

#' @export
sinh.estimate <- function(x, ...) {
  estimate(x, function(p) {
    y <- sinh(p)
    structure(y, grad = diag(cosh(p), nrow = length(p)))
  }, ...)
}

#' @export
cosh.estimate <- function(x, ...) {
  estimate(x, function(p) {
    y <- cosh(p)
    structure(y, grad = diag(sinh(p), nrow = length(p)))
  }, ...)
}

#' @export
tanh.estimate <- function(x, ...) {
  estimate(x, function(p) {
    y <- tanh(p)
    structure(y, grad = diag(1 / cosh(p)^2, nrow = length(p)))
  }, ...)
}

# ---- Inverse Hyperbolic Functions ---------------------------------------

#' @export
asinh.estimate <- function(x, ...) {
  estimate(x, function(p) {
    y <- asinh(p)
    structure(y, grad = diag(1 / sqrt(p^2 + 1), nrow = length(p)))
  }, ...)
}

#' @export
acosh.estimate <- function(x, ...) {
  estimate(x, function(p) {
    y <- acosh(p)
    structure(y, grad = diag(1 / sqrt(p^2 - 1), nrow = length(p)))
  }, ...)
}

#' @export
atanh.estimate <- function(x, ...) {
  estimate(x, function(p) {
    y <- atanh(p)
    structure(y, grad = diag(1 / (1 - p^2), nrow = length(p)))
  }, ...)
}

# ---- Other Common Functions ---------------------------------------------

#' @export
log1p.estimate <- function(x, ...) {
  # log(1 + p) — more numerically stable than log(p+1) for small p
  estimate(x, function(p) {
    y <- log1p(p)
    structure(y, grad = diag(1 / (1 + p), nrow = length(p)))
  }, ...)
}

#' @export
expm1.estimate <- function(x, ...) {
  # exp(p) - 1 — more numerically stable than exp(p)-1 for small p
  estimate(x, function(p) {
    y <- expm1(p)
    structure(y, grad = diag(exp(p), nrow = length(p)))
  }, ...)
}

#' @export
log.estimate <- function(x, base = exp(1), ...) {
  estimate(x, function(p) {
    y <- log(p, base = base)
    structure(y,
              grad = diag(1 / (p * log(base)), nrow=length(p)))
  }, ...)
}

#' @export
exp.estimate <- function(x, ...) {
  estimate(x, function(p) {
    y <- exp(p)
    structure(y, grad = diag(y, nrow=length(p)))
  }, ...)
}

#' @export
sqrt.estimate <- function(x, ...) {
  estimate(x^.5, ...)
}

#' @export
sum.estimate <- function(x, ...) {
  estimate(x,
           function(p)
             structure(sum(p),
                       grad = matrix(1, nrow = 1, ncol = length(p))),
           ...)
}

 #' @export
"%*%.estimate" <- function(x, y, ...) {
  if (is.matrix(x)) {
    return(estimate(y, f=x, ...))
  } else if (is.matrix(y)) {
    return(estimate(x, f=t(y), ...))
  }
  sum(x * y)
}

prod_except <- function(p) {
  n     <- length(p)
  left  <- c(1, cumprod(p[-n]))
  right <- rev(cumprod(rev(p[-1])))
  right <- c(right, 1)
  left * right
}

#' @export
prod.estimate <- function(x, ...) {
  estimate(x, function(p) {
    y <- prod(p)
    # grad of prod(p) is: d/dp_i = prod(p) / p_i
    structure(y, grad = matrix(prod_except(p), nrow = 1, ncol = length(p)))
  }, ...)
}

# ---- +,-,*,/ ------------------------------------------------------------

operator_estimate <- function(x, y, op, ...) {
  x_const <- is.numeric(x)
  y_const <- is.numeric(y)
  np1 <- ifelse(x_const, length(x), length(coef(x)))
  np2 <- ifelse(y_const, length(y), length(coef(y)))

  if (np1 != 1L && np2 != 1L && np1 != np2) {
    stop("expecting equal length objects or one of them to be a scalar")
  }
  if (y_const || x_const) { # x estimate, y numeric
    e <- if (y_const) x else y
    return(
      estimate(e, function(p) {
        if (y_const) {
          op(p, y, y_const=TRUE)
        } else {
          op(x, p, x_const=TRUE)
        }
      }, ...)
    )
  }
  e <- merge(x, y) # x estimate, y estimate
  estimate(e, function(p) {
    p1 <- p[seq_len(np1)]
    p2 <- p[seq_len(np2)+np1]
    res <- op(p1, p2)
    return(res)
  }, ...)
}

operator_grad <- function(x, y, x_const, y_const, dx, dy) {
  nx <- length(x)
  ny <- length(y)
  n <- max(nx, ny)
  grad <-
    if (y_const) { # y is constant: df/dx only
      if (nx > 1L || ny == 1L) { # ny = nx or ny = 1
        diag(dx, ncol=nx, nrow=nx)
      } else { # ny > 1, nx = 1
        matrix(dx, nrow = n, ncol = 1L)
      }
    } else if (x_const) { # x is constant: df/dy only
      if (ny > 1L || nx == 1L) { # ny = nx or nx = 1
        diag(dy, ncol=ny, nrow=ny)
      } else { # ny > 1, nx = 1
        matrix(dy, nrow = n, ncol = 1L)
      }
    } else {
      # both estimates: df/d(c(x,y)) = [dx | dy]
      D <- matrix(0, nrow = n, ncol = nx + ny)
      if (nx == 1L) {
        D[, 1] <- dx
        D[, 2:ncol(D)] <- diag(dy, ny, ny)
        D
      } else if (ny == 1L) {
        cbind(diag(dx, nrow=n, ncol=n), dy)
      } else {
        cbind(diag(dx, nrow = n, ncol=n), diag(dy, nrow = n, ncol=n))
      }
    }
  return(grad)
}

#' @export
"+.estimate" <- function(e1, e2, ...) {
  operator_estimate(
    e1, e2,
    function(x, y, x_const=FALSE, y_const=FALSE) {
      structure(
        x + y,
        grad = operator_grad(x, y, x_const, y_const,
                             dx = 1, dy = 1)
      )
      }, ...)
}

#' @export
"-.estimate" <- function(e1, e2, ...) {
  if (missing(e2)) return(-1*e1)
  operator_estimate(
    e1, e2,
    function(x, y, x_const=FALSE, y_const=FALSE) {
      structure(
        x - y,
        grad = operator_grad(x, y, x_const, y_const,
                             dx=1, dy=-1)
      )
      }, ...)
}

#' @export
"*.estimate" <- function(e1, e2, ...) {
  operator_estimate(
    e1, e2,
    function(x, y, x_const=FALSE, y_const=FALSE) {
      structure(
        x * y,
        grad = operator_grad(x, y, x_const, y_const,
                             dx=y, dy=x)
      )
      }, ...)
}

#' @export
"/.estimate" <- function(e1, e2, ...) {
  operator_estimate(
    e1, e2,
    function(x, y, x_const=FALSE, y_const=FALSE) {
      structure(
        x / y,
        grad = operator_grad(x, y, x_const, y_const,
                             dx = 1 / y, dy = -x / y^2)
      )
      }, ...)
}

#' @export
"^.estimate" <- function(e1, e2, ...) {
  operator_estimate(
    e1, e2,
    function(x, y, x_const=FALSE, y_const=FALSE) {
      nx <- length(x)
      ny <- length(y)
      n <- max(nx, ny)
      f <- x^y
      structure(
        f,
        grad = operator_grad(x, y, x_const, y_const,
                             dx = y*x^(y-1), dy = log(x)*f)
      )
    }, ...)
}

# ---- == / hypothesis ----------------------------------------------------

#' @export
"==.estimate" <- function(e1, e2) {
  if (!(is.numeric(e1) || is.numeric(e2))) stop("numeric comparator needed")
  null <- if (is.numeric(e1)) e1 else e2
  e <- if (is.numeric(e1)) e2 else e1
  summary(e, null = null)
}

##' Create contrast matrix
##'
##' Create contrast matrix typically for use with 'estimate' (Wald tests).
##' @export
##' @param p index of non-zero entries (see example)
##' @param n Total number of parameters (if omitted the max number in p will be used)
##' @param diff If FALSE all non-zero entries are +1, otherwise the second non-zero element in each row will be -1.
##' @param ... Additional arguments to lower level functions
##' @aliases contr parsedesign pairwise.diff
##' @examples
##' contr(2,n=5)
##' contr(as.list(2:4),n=5)
##' contr(list(1,2,4),n=5)
##' contr(c(2,3,4),n=5)
##' contr(list(c(1,3),c(2,4)),n=5)
##' contr(list(c(1,3),c(2,4),5))
##'
##' parsedesign(c("aa","b","c"),"?","?",diff=c(FALSE,TRUE))
##'
##' ## All pairs comparisons:
##' pdiff <- function(n) lava::contr(lapply(seq(n-1), function(x) seq(x, n)))
##' pdiff(4)
contr <- function(p, n, diff = TRUE, ...) {
  if (missing(n)) n <- max(unlist(p))
  if (is.character(p)) {
    return(parsedesign(n, p, ...))
  }
  if (is.list(p)) {
    return(Reduce(rbind, lapply(p, function(x) {
      do.call(contr, list(x, n, diff[1L]))
    })))
  }
  if (is.character(n)) n <- length(n)
  if (!is.numeric(n)) {
    try(n <- length(coef(n)), silent = TRUE)
  }
  B <- matrix(0, ncol = n, nrow = max(1L, length(p) - 1L))
  B[, p[1]] <- 1L
  if (length(p) > 1L) {
    B[cbind(seq_len(nrow(B)), p[-1])] <- ifelse(diff[1L], -1, 1)
  }
  return(B)
}

##' @export
pairwise.diff <- function(n) {
  pdiff <- function(n) lava::contr(lapply(seq(n-1), function(x) seq(x, n)))
  pdiff(n)
}

##' Simulate model
##'
##' Simulate data from a general SEM model including non-linear effects and
##' general link and distribution of variables.
##'
##' @aliases sim.lvmfit sim.lvm
##' @aliases simulate.lvmfit simulate.lvm
##' @aliases transform<- transform<-.lvm transform.lvm
##' @aliases functional functional<-  functional.lvm functional<-.lvm
##' @aliases distribution distribution<- distribution.lvm distribution<-.lvm
##' @aliases heavytail heavytail<-
##' @aliases categorical categorical<-
##' @aliases dist_weibull weibull.lvm
##' @aliases dist_bernoulli binomial.lvm probit.lvm logit.lvm
##' @aliases dist_poisson poisson.lvm
##' @aliases dist_uniform uniform.lvm
##' @aliases dist_multinomial multinomial.lvm
##' @aliases dist_beta beta.lvm
##' @aliases dist_gaussian normal.lvm gaussian.lvm
##' @aliases dist_mvn mvn.lvm
##' @aliases dist_lognormal lognormal.lvm
##' @aliases dist_gaussian_mixture2 GM2.lvm
##' @aliases dist_gaussian_mixture3 GM3.lvm
##' @aliases dist_pareto pareto.lvm
##' @aliases dist_t student.lvm
##' @aliases dist_chisq chisq.lvm
##' @aliases dist_cox_gompertz coxGompertz.lvm
##' @aliases dist_cox_weibull coxWeibull.lvm
##' @aliases dist_cox_exponential coxExponential.lvm
##' @aliases dist_aalen_exponential aalenExponential.lvm
##' @aliases dist_gamma Gamma.lvm gamma.lvm
##' @aliases dist_loggamma loggamma.lvm
##' @aliases dist_threshold threshold.lvm
##' @aliases dist_ones ones.lvm binary.lvm
##' @aliases dist_seq Sequence.lvm
##' @aliases dist_seqint id.lvm
##' @aliases dist_none none.lvm
##' @aliases dist_const constant.lvm
##' @usage
##' \method{sim}{lvm}(x, n = NULL, p = NULL, normal = FALSE, cond = FALSE,
##' sigma = 1, rho = 0.5, X = NULL, unlink=FALSE, latent=TRUE,
##' use.labels = TRUE, seed=NULL, ...)
##' @param x Model object
##' @param n Number of simulated values/individuals
##' @param p Parameter value (optional)
##' @param normal Logical indicating whether to simulate data from a
##' multivariate normal distribution conditional on exogenous variables hence
##' ignoring functional/distribution definition
##' @param cond for internal use
##' @param sigma Default residual variance (1)
##' @param rho Default covariance parameter (0.5)
##' @param X Optional matrix of fixed values of variables (manipulation)
##' @param unlink Return Inverse link transformed data
##' @param latent Include latent variables (default TRUE)
##' @param use.labels convert categorical variables to factors before applying transformation
##' @param seed Random seed
##' @param \dots Additional arguments to be passed to the low level functions
##' @author Klaus K. Holst
##' @keywords models datagen regression
##' @export
##' @examples
##' ##################################################
##' ## Logistic regression
##' ##################################################
##' m <- lvm(y~x+z)
##' regression(m) <- x~z
##' distribution(m,~y+z) <- dist_bernoulli("logit")
##' d <- sim(m,1e3)
##' head(d)
##'
##' e <- estimate(m,d,estimator="glm")
##' e
##' ## Simulate a few observation from estimated model
##' sim(e,n=5)
##'
##' ##################################################
##' ## Poisson
##' ##################################################
##' distribution(m,~y) <- dist_poisson()
##' d <- sim(m,1e4,p=c(y=-1,"y~x"=2,z=1))
##' head(d)
##' estimate(m,d,estimator="glm")
##' mean(d$z); lava:::expit(1)
##'
##' summary(lm(y~x,sim(lvm(y[1:2]~4*x),1e3)))
##'
##' ##################################################
##' ### Gamma distribution
##' ##################################################
##' m <- lvm(y~x)
##' distribution(m,~y+x) <- list(dist_gamma(shape=2), dist_bernoulli())
##' intercept(m,~y) <- 0.5
##' d <- sim(m,1e4)
##' summary(g <- glm(y~x,family=Gamma(),data=d))
##' \dontrun{MASS::gamma.shape(g)}
##'
##' args(lava::dist_gamma)
##' distribution(m,~y) <- dist_gamma(shape=2,log=TRUE)
##' sim(m,10,p=c(y=0.5))[,"y"]
##'
##' ##################################################
##' ### Beta
##' ##################################################
##' m <- lvm()
##' distribution(m,~y) <- dist_beta(alpha=2,beta=1)
##' var(sim(m,100,"y,y"=2))
##' distribution(m,~y) <- dist_beta(alpha=2,beta=1,scale=FALSE)
##' var(sim(m,100))
##'
##' ##################################################
##' ### Transform
##' ##################################################
##' m <- lvm()
##' transform(m,xz~x+z) <- function(x) x[1]*(x[2]>0)
##' regression(m) <- y~x+z+xz
##' d <- sim(m,1e3)
##' summary(lm(y~x+z + x*I(z>0),d))
##'
##' ##################################################
##' ### Non-random variables
##' ##################################################
##' m <- lvm()
##' distribution(m,~x+z+v+w) <- list(
##'   dist_seqint(0,5),## Seq. 0 to 5 by 1/n
##'   dist_ones(),       ## Vector of ones
##'   dist_ones(0.5),    ##  0.5n 0, 0.5n 1
##'   dist_ones(interval=list(c(0.3,0.5),c(0.8,1)))
##' )
##' sim(m,10)
##'
##' ##################################################
##' ### Cox model
##' ### piecewise constant hazard
##' ################################################
##' m <- lvm(t~x)
##' rates <- c(1,0.5); cuts <- c(0,5)
##' ## Constant rate: 1 in [0,5), 0.5 in [5,Inf)
##' distribution(m,~t) <- dist_cox_exponential(rate=rates,timecut=cuts)
##'
##' \dontrun{
##'     d <- sim(m,2e4,p=c("t~x"=0.1)); d$status <- TRUE
##'     plot(timereg::aalen(survival::Surv(t,status)~x,data=d,
##'                         resample.iid=0,robust=0),spec=1)
##'     L <- approxfun(c(cuts,max(d$t)),f=1,
##'                    cumsum(c(0,rates*diff(c(cuts,max(d$t))))),
##'                    method="linear")
##'     curve(L,0,100,add=TRUE,col="blue")
##' }
##'
##' ##################################################
##' ### Cox model
##' ### piecewise constant hazard, gamma frailty
##' ##################################################
##' m <- lvm(y~x+z)
##' rates <- c(0.3,0.5); cuts <- c(0,5)
##' distribution(m,~y+z) <- list(dist_cox_exponential(rate=rates,timecut=cuts),
##'                              dist_loggamma(rate=1,shape=1))
##' \dontrun{
##'     d <- sim(m,2e4,p=c("y~x"=0,"y~z"=0)); d$status <- TRUE
##'     plot(timereg::aalen(survival::Surv(y,status)~x,data=d,
##'                         resample.iid=0,robust=0),spec=1)
##'     L <- approxfun(c(cuts,max(d$y)),f=1,
##'                    cumsum(c(0,rates*diff(c(cuts,max(d$y))))),
##'                    method="linear")
##'     curve(L,0,100,add=TRUE,col="blue")
##' }
##' ## Equivalent via transform (here with Aalens additive hazard model)
##' m <- lvm(y~x)
##' distribution(m,~y) <- dist_aalen_exponential(rate=rates, timecut=cuts)
##' distribution(m,~z) <- dist_gamma(rate=1, shape=1)
##' transform(m,t~y+z) <- prod
##' sim(m,10)
##' ## Shared frailty
##' m <- lvm(c(t1,t2)~x+z)
##' rates <- c(1,0.5); cuts <- c(0,5)
##' distribution(m,~y) <- dist_aalen_exponential(rate=rates, timecut=cuts)
##' distribution(m,~z) <- dist_loggamma(rate=1,shape=1)
##' \dontrun{
##' mets::fast.reshape(sim(m,100),varying="t")
##' }
##'
##' ##################################################
##' ### General multivariate distributions
##' ##################################################
##' \dontrun{
##' m <- lvm()
##' distribution(m,~y1+y2,oratio=4) <- VGAM::rbiplackcop
##' ksmooth2(sim(m,1e4),rgl=FALSE,theta=-20,phi=25)
##'
##' m <- lvm()
##' distribution(m,~z1+z2,"or1") <- VGAM::rbiplackcop
##' distribution(m,~y1+y2,"or2") <- VGAM::rbiplackcop
##' sim(m,10,p=c(or1=0.1,or2=4))
##' }
##'
##' m <- lvm()
##' distribution(m,~y1+y2+y3,TRUE) <- function(n,...) rmvn0(n,sigma=diag(3)+1)
##' var(sim(m,100))
##'
##' ## Syntax also useful for univariate generators, e.g.
##' m <- lvm(y~x+z)
##' distribution(m,~y,TRUE) <- function(n) rnorm(n,mean=1000)
##' sim(m,5)
##' distribution(m,~y,"m1",0) <- rnorm
##' sim(m,5)
##' sim(m,5,p=c(m1=100))
##'
##' ##################################################
##' ### Regression design in other parameters
##' ##################################################
##' ## Variance heterogeneity
##' m <- lvm(y~x)
##' distribution(m,~y) <- function(n,mean,x) rnorm(n,mean,exp(x)^.5)
##' if (interactive()) plot(y~x,sim(m,1e3))
##' ## Alternaively, calculate the standard error directly
##' addvar(m) <- ~sd ## If 'sd' should be part of the resulting data.frame
##' constrain(m,sd~x) <- function(x) exp(x)^.5
##' distribution(m,~y) <- function(n,mean,sd) rnorm(n,mean,sd)
##' if (interactive()) plot(y~x,sim(m,1e3))
##'
##' ## Regression on variance parameter
##' m <- lvm()
##' regression(m) <- y~x
##' regression(m) <- v~x
##' ##distribution(m,~v) <- 0 # No stochastic term
##' ## Alternative:
##' ## regression(m) <- v[NA:0]~x
##' distribution(m,~y) <- function(n,mean,v) rnorm(n,mean,exp(v)^.5)
##' if (interactive()) plot(y~x,sim(m,1e3))
##'
##' ## Regression on shape parameter in Weibull model
##' m <- lvm()
##' regression(m) <- y ~ z+v
##' regression(m) <- s ~ exp(0.6*x-0.5*z)
##' distribution(m,~x+z) <- dist_bernoulli()
##' distribution(m,~cens) <- dist_cox_weibull(scale=1)
##' distribution(m,~y) <- dist_cox_weibull(scale=0.1,shape=~s)
##' eventTime(m) <- time ~ min(y=1,cens=0)
##'
##' if (interactive()) {
##'     d <- sim(m,1e3)
##'     require(survival)
##'     (cc <- coxph(Surv(time,status)~v+strata(x,z),data=d))
##'     plot(survfit(cc) ,col=1:4,mark.time=FALSE)
##' }
##'
##' ##################################################
##' ### Categorical predictor
##' ##################################################
##' m <- lvm()
##' ## categorical(m,K=3) <- "v"
##' categorical(m,labels=c("A","B","C")) <- "v"
##'
##' regression(m,additive=FALSE) <- y~v
##' \dontrun{
##' plot(y~v,sim(m,1000,p=c("y~v:2"=3)))
##' }
##'
##' m <- lvm()
##' categorical(m,labels=c("A","B","C"),p=c(0.5,0.3)) <- "v"
##' regression(m,additive=FALSE,beta=c(0,2,-1)) <- y~v
##' ## equivalent to:
##' ## regression(m,y~v,additive=FALSE) <- c(0,2,-1)
##' regression(m,additive=FALSE,beta=c(0,4,-1)) <- z~v
##' table(sim(m,1e4)$v)
##' glm(y~v, data=sim(m,1e4))
##' glm(y~v, data=sim(m,1e4,p=c("y~v:1"=3)))
##'
##' transform(m,v2~v) <- function(x) x=='A'
##' sim(m,10)
##'
##' ##################################################
##' ### Pre-calculate object
##' ##################################################
##' m <- lvm(y~x)
##' m2 <- sim(m,'y~x'=2)
##' sim(m,10,'y~x'=2)
##' sim(m2,10) ## Faster
##' @export
sim.lvm <- function(x,n=NULL,p=NULL,normal=FALSE,cond=FALSE,sigma=1,rho=.5,
            X=NULL,unlink=FALSE,latent=TRUE,use.labels=TRUE,seed=NULL,...) {

    if (!exists(".Random.seed", envir = .GlobalEnv, inherits = FALSE))
        runif(1)
    if (is.null(seed))
        RNGstate <- get(".Random.seed", envir = .GlobalEnv)
    else {
        R.seed <- get(".Random.seed", envir = .GlobalEnv)
        set.seed(seed)
        RNGstate <- structure(seed, kind = as.list(RNGkind()))
        on.exit(assign(".Random.seed", R.seed, envir = .GlobalEnv))
    }


    v.env <- c("A","M","P","PP","PPdiag","xx","vv","mdist","mdistnam","mii",
              "nn","mu","xf","xfix","X",
              "vartrans","multitrans","multitrans.idx",
              "X.idx","ii.mvn","xconstrain.idx","xconstrain",
              "xconstrain.par","covparnames","exo_constrainY")

    setup <- is.null(n) && is.null(X) ## Save environment (variables v.env) and return sim object
    loadconfig <- !is.null(x$sim.env) && !setup && (length(list(...))==0 && length(p)==0)
    Yfix <- NULL

    if (loadconfig) {
        for (v in setdiff(v.env,"X")) assign(v, x$sim.env[[v]])
        if (is.null(X)) X <-  x$sim.env[['X']]
    } else {
        if (!is.null(n) && n<1) return(NULL)
        p <- c(p,unlist(list(...)))
        xx <- exogenous(x)

        if (!is.null(X)) {
            if (is.null(n))
                n <- nrow(X)
            if (!is.null(colnames(X))) {
                yfix <- setdiff(colnames(X),xx)
                if (length(yfix)>0) Yfix <- X[,yfix,drop=FALSE]
                xx0 <- intersect(xx,colnames(X))
                if (length(xx0)>0)
                    X <- as.matrix(X[,xx0,drop=FALSE])
            }
        } else {
            if (!is.null(p)) {
                i1 <- unique(na.omit(c(match(names(p),xx),
                                       match(names(p),paste0(xx,lava.options()$symbol[2],xx)))))
                covariance(x) <- xx[i1]
            }
        }
        ##  index(x) <- reindex(x)
        vv <- vars(x)
        nn <- setdiff(vv,parameter(x))
        mu <- unlist(lapply(x$mean, function(l) ifelse(is.na(l)|is.character(l),0,l)))
        xf <- intersect(unique(parlabels(x)),xx)
        xfix <- c(randomslope(x),xf); if (length(xfix)>0) normal <- FALSE
        ## Match parameter names
        if ((!is.null(names(p)) && all(!is.na(names(p)))) || length(p)!=(index(x)$npar+index(x)$npar.mean+index(x)$npar.ex) | is.null(names(p))) {
            nullp <- is.null(p)
            p0 <- p
            ep <- NULL
            ei <- which(index(x)$e1==1)
            if (length(ei)>0)
                ep <- unlist(x$expar)[ei]
            p <- c(rep(1, index(x)$npar+index(x)$npar.mean),ep)
            p[seq_len(index(x)$npar.mean)] <- 0
            p[index(x)$npar.mean + variances(x)] <- sigma
            p[index(x)$npar.mean + offdiags(x)] <- rho
            if (!nullp) {
                c1 <- coef(x,mean=TRUE,fix=FALSE)
                c2 <- coef(x,mean=TRUE,fix=FALSE,labels=TRUE)
                idx1 <- na.omit(match(names(p0),c1))
                idx11 <- na.omit(match(names(p0),c2))
                idx2 <- na.omit(which(names(p0)%in%c1))
                idx22 <- na.omit(which(names(p0)%in%c2))
                if (length(idx1)>0 && !any(is.na(idx1)))
                    p[idx1] <- p0[idx2]
                if (length(idx11)>0 && !any(is.na(idx11)))
                    p[idx11] <- p0[idx22]
            }
        }
        M <- modelVar(x,p,data=NULL)
        A <- M$A; P <- M$P
        if (!is.null(M$v)) mu <- M$v

        ## Square root of residual variance matrix
        PP <- with(svd(P), v%*%diag(sqrt(d),nrow=length(d))%*%t(u))
        ## Multivariate distributions
        mdist <- distribution(x,multivariate=TRUE)$var
        mdistnam <- names(mdist)
        mii <- match(mdistnam,vars(x))
        if (length(distribution(x))>0 ) {
            ii <- match(names(distribution(x)),vv)
            ii.mvn <- setdiff(seq(ncol(P)),c(ii,mii))
        } else {
            ii.mvn <- seq(ncol(P))
        }
        PPdiag <- sum(abs(offdiag(PP[ii.mvn,ii.mvn,drop=FALSE])^2))<1e-20
    }

    if (!setup) {
        E <- matrix(0,ncol=ncol(P),nrow=n)
        if (length(ii.mvn)>0) {
            ## Error term for conditional normal distributed variables
            if (PPdiag) {
                for (i in ii.mvn) E[,i] <- rnorm(n,sd=PP[i,i])
            } else {
                E[,ii.mvn] <-  matrix(rnorm(length(ii.mvn)*n),ncol=length(ii.mvn))%*%PP[ii.mvn,ii.mvn,drop=FALSE]
            }
        }

        if (length(mdistnam)>0) {
            fun <- distribution(x,multivariate=TRUE)$fun
            for (i in seq_along(fun)) {
                mv <- names(which(unlist(mdist)==i))
                ii <- match(mv,vv)
                E[,ii] <- distribution(x,multivariate=TRUE)$fun[[i]](n,p=p,object=x) # ,...)
            }
        }

        ## Simulate exogenous variables (covariates)
        res <- matrix(0,ncol=length(nn),nrow=n)
        colnames(res) <- nn
    }

    if (!loadconfig) {
        vartrans <- names(x$attributes$transform)
        multitrans <- multitrans.idx <- NULL
        if (length(x$attributes$multitransform)>0) {
            multitrans <- unlist(lapply(x$attributes$multitransform,function(z) z$y))
            for (i in (seq_along(x$attributes$multitransform))) {
                multitrans.idx <- c(multitrans.idx,rep(i,length(x$attributes$multitransform[[i]]$y)))
            }
        }
        xx <- unique(c(exogenous(x, latent=FALSE, index=TRUE),xfix))
        xx <- setdiff(xx,vartrans)

        X.idx <- match(xx,vv)
    }

    if (!setup) {
        res[,X.idx] <- t(mu[X.idx]+t(E[,X.idx]))
        if (is.null(X) || NCOL(X)<length(xx)) {
            if (!is.null(xx) && length(xx)>0)
                for (i in seq_along(xx)) {
                    mu.x <- mu[X.idx[i]]
                    dist.x <- distribution(x,xx[i])[[1]]
                    if (is.list(dist.x) && is.function(dist.x[[1]])) dist.x <- dist.x[[1]]
                    if (is.list(dist.x)) {
                        dist.x <- dist.x[[1]]
                        if (length(dist.x)==1) dist.x <- rep(dist.x,n)
                    }
                    if (is.function(dist.x)) {
                        res[,X.idx[i]] <- dist.x(n=n,mu=mu.x,var=P[X.idx[i],X.idx[i]])
                    } else {
                        if (is.null(dist.x) || any(is.na(dist.x))) {
                        } else {
                            if (length(dist.x)!=n) stop("'",vv[X.idx[i]], "' fixed at length ", length(dist.x)," != ",n,".")
                            res[,X.idx[i]] <- dist.x ## Deterministic
                        }
                    }
                }
        }
        if (!is.null(X)) {
            ii <- match(colnames(X),vv)
            res0 <- res
            for (i in seq(ncol(X)))
                res[,ii[i]] <- X[,i]
        }
    }

    simuled <- c(xx)
    resunlink <- NULL
    if (unlink) {
        resunlink <- res
    }

    if ( normal | ( is.null(distribution(x)) & is.null(functional(x)) & is.null(constrain(x))) ) {
        if(cond) { ## Simulate from conditional distribution of Y given X
            mypar <- pars(x,A,P,mu)
            Ey.x <- predict(x, mypar, data.frame(res))
            Vy.x <- attributes(Ey.x)$cond.var
            PP <- with(svd(Vy.x), v%*%diag(sqrt(d),nrow=length(d))%*%t(u))
            yy <- Ey.x + matrix(n*ncol(Vy.x),ncol=ncol(Vy.x))%*%PP
            res <- cbind(yy, res[,xx]); colnames(res) <- c(colnames(Vy.x),xx)
            return(res)
        }
        ## Simulate from sim. distribution (Y,X) (mv-normal)
        I <- diag(nrow=length(nn))
        IAi <- Inverse(I-t(A))
        colnames(E) <- vv
        dd <- t(apply(heavytail_sim_hook(x,E),1,function(x) x+mu))
        res <- dd%*%t(IAi)
        colnames(res) <- vv
    } else {

        if (!loadconfig) {
            xc <- index(x)$vars
            xconstrain.idx <- unlist(lapply(lapply(constrain(x),function(z) attributes(z)$args),function(z) length(intersect(z,xc))>0))
            xconstrain <- intersect(unlist(lapply(constrain(x),function(z) attributes(z)$args)),xc)
            xconstrain.par <- names(xconstrain.idx)[xconstrain.idx]
            covparnames <- unique(as.vector(covariance(x)$labels))
            exo_constrainY <- intersect(exogenous(x),names(x$constrainY))
        }

        if (setup) {
            sim.env <- c()
            sim.env[v.env] <- list(NULL)
            for (v in v.env) if (!is.null(get(v))) sim.env[[v]] <- get(v)
            x$sim.env <- sim.env
            return(x)
        }

      if (length(xconstrain)>0) {
        for (i in which(xconstrain.idx)) {
          if (!(names(xconstrain.idx[i]) %in% nn)) { ## A parameter and not a variable
            ff <- constrain(x)[[i]]
            myargs <- attributes(ff)$args
            D <- matrix(0,n,length(myargs))
            for (j in seq_len(ncol(D))) {
              if (myargs[j]%in%xconstrain)
                D[,j] <- res[,myargs[j]]
              else
                D[,j] <- M$parval[[myargs[j]]]
            }
            val <- try(val <- ff(D), silent=TRUE)
            if (inherits(val,"try-error") || NROW(val)<n) apply(D,1,ff)
            res <- cbind(res, val); colnames(res)[ncol(res)] <- names(xconstrain.idx)[i]
          }
        }
      }

        if (any(xconstrain.par%in%covparnames)) {
            mu0 <- rep(0,ncol(P))
            P0 <- P
            E <- t(sapply(seq_len(n),function(idx) {
              for (i in intersect(xconstrain.par,covparnames)) {
                  P0[covariance(x)$labels==i] <- res[idx,i]
                }
                PP <- with(svd(P0), v%*%diag(sqrt(d),nrow=length(d))%*%t(u))
                return(mu0+rbind(rnorm(ncol(P0)))%*%PP)
            }))
        }

        colnames(E) <- vv
        E <- heavytail_sim_hook(x,E)

        ## Non-linear regression components
        xconstrain <- c()
        for (i in seq_len(length(constrain(x)))) {
            z <- constrain(x)[[i]]
            xx <- intersect(attributes(z)$args,manifest(x))
            if (length(xx)>0) {
                warg <- setdiff(attributes(z)$args,xx)
                wargidx <- which(attributes(z)$args%in%warg)
                exoidx <- which(attributes(z)$args%in%xx)
                parname <- names(constrain(x))[i]
                y <- names(which(unlist(lapply(intercept(x),function(x) x==parname))))
                el <- list(i,y,parname,xx,exoidx,warg,wargidx,z)
                names(el) <- c("idx","endo","parname","exo","exoidx","warg","wargidx","func")
                xconstrain <- c(xconstrain,list(el))
            }
        }
        yconstrain <- unlist(lapply(xconstrain,function(x) x$endo))
      for (i in exo_constrainY) {
            cc <- x$constrainY[[i]]
            args <- cc$args
            args <- if (is.null(args) || length(args)==0) res[,i] else res[,args]
            res[,i] <- cc$fun(args,p) # ,...)
        }

        res <- data.frame(res)
        if (length(vartrans)>0) {
            parvals <- parpos(x)$parval
            parvalsnam <- setdiff(names(parvals),xx)
            if (length(parvalsnam)>0) {
                Parvals <- p[unlist(parvals)];
                res <- cbind(res,
                             cbind(rep(1,nrow(res)))%x%rbind(Parvals))
                colnames(res)[seq(length(Parvals))+ncol(res)-length(Parvals)] <-
                    names(parvals)
            }
        }

        leftovers <- c()
        itercount <- 0
        while (length(simuled)<length(nn)) {
            leftoversPrev <- leftovers
            leftovers <- setdiff(nn,simuled)
            if (!is.null(leftoversPrev) && length(leftoversPrev)==length(leftovers)) {
                if (itercount>0)
                stop("Infinite loop (feedback).")
                itercount <- itercount+1
            }
            for (i in leftovers) {
                if (length(Yfix)>0 && i %in% colnames(Yfix)) {
                    if (NROW(Yfix) == 1) {
                        res[,i] <- rep(Yfix[,i],length.out=nrow(res))
                    } else                     
                        res[,i] <- rep(Yfix[,i,drop=TRUE],length.out=nrow(res))
                    simuled <- c(simuled,i)
                    next
                }
                if (i%in%vartrans) {
                    xtrans <- x$attributes$transform[[i]]$x
                    if (all(xtrans%in%c(simuled,names(parvals))))  {
                        xtr <- res[,xtrans,drop=FALSE]
                        if (use.labels) {
                            lb <- x$attributes$labels
                            lb.idx <- na.omit(match(names(lb),xtrans))
                            ## For categorical variables turn them into factors so we can
                            ## use the actual labels in function calls/transform
                            if (length(lb.idx)>0) {
                                xtr <- as.data.frame(xtr)
                                for (lb0 in lb.idx) {
                                    lab <- lb[[names(xtr)[lb0]]]
                                    xtr[,lb0] <- factor(xtr[,lb0],levels=seq_along(lab)-1,labels=lab)
                                }
                            }
                        }
                        suppressWarnings(yy <- with(x$attributes$transform[[i]], fun(xtr))) ##fun(res[,xtrans])))
                        if (NROW(yy) != NROW(res)) { ## apply row-wise
                            res[,i] <- with(x$attributes$transform[[i]], ##apply(res[,xtrans,drop=FALSE],1,fun))
                                           apply(xtr,1,fun))
                        } else {
                            colnames(yy) <- NULL
                            res[,i] <- yy
                        }
                        simuled <- c(simuled,i)
                    }
                } else if (i%in%multitrans) {
                    idx0 <- match(i,multitrans)
                    idx <- multitrans.idx[idx0]
                    mtval <- x$attributes$multitransform[[idx]]
                    if (all(mtval$x%in%simuled)) {
                        res[,mtval$y] <- mtval$fun(res[,mtval$x])
                        simuled <- c(simuled,mtval$y)
                        break;
                    }
                } else {
                    ipos <- which(yconstrain%in%i)
                    if (length(ipos)==0 || all(xconstrain[[ipos]]$exo%in%simuled)) {
                        pos <- match(i,vv)
                        relations <- colnames(A)[A[,pos]!=0]
                        simvars <- x$attributes$simvar[[i]]
                        dist.i <- distribution(x,i)[[1]] ## User-specified distribution function
                        dist.xx <- NULL
                        if (is.function(dist.i)) {
                            dist.args0 <- names(formals(dist.i))
                            dist.args <- setdiff(dist.args0,c("n","mean","mu","var","..."))
                            dist.xx <- intersect(names(res),dist.args) ## Variables influencing distribution
                        }
                        if (all(c(relations,simvars,dist.xx)%in%simuled)) { ## Only depending on already simulated variables
                            if (x$mean[[pos]]%in%xconstrain.par && length(ipos)==0) {
                                mu.i <- res[,x$mean[[pos]] ]
                            } else {
                                mu.i <- mu[pos]
                            }
                            if (length(ipos)>0) {
                                pp <- unlist(M$parval[xconstrain[[ipos]]$warg])
                                myidx <- with(xconstrain[[ipos]],order(c(wargidx,exoidx)))
                                ## myidx <- with(xconstrain[[ipos]],
                                ##              match(attr(func,"args"), c(warg,exo)))
                                X <- with(xconstrain[[ipos]],
                                          if (length(pp)>0)
                                              cbind(rbind(pp)%x%cbind(rep(1,nrow(res))),
                                                    res[,exo,drop=FALSE])
                                          else res[,exo,drop=FALSE])
                                yy <- try(with(xconstrain[[ipos]],
                                               func(X[,myidx,drop=FALSE])),silent=TRUE)
                                if (NROW(yy) != NROW(res)) { ## apply row-wise
                                    mu.i <- #mu.i +
                                        with(xconstrain[[ipos]],
                                             apply(res[,exo,drop=FALSE],1,
                                                   function(x) func(
                                                            unlist(c(pp,x))[myidx])))
                                } else {
                                    mu.i <- ##mu.i+
                                        yy
                                }
                            }
                            for (From in relations) {
                                f <- functional(x,i,From)[[1]]
                                if (!is.function(f))
                                    f <- function(x,...) x
                                reglab <- regfix(x)$labels[From,pos]
                                if (reglab%in%c(xfix,xconstrain.par)) {
                                    if (is.function(f)) {
                                        if (length(formals(f))>1) {
                                            mu.i <- mu.i + res[,reglab]*f(res[,From],p)
                                        } else {
                                            mu.i <- mu.i + res[,reglab]*f(res[,From])
                                        }
                                    } else  mu.i <- mu.i + res[,reglab]*res[,From]
                                }
                                else {
                                    if (is.function(f)) {
                                        if (length(formals(f))>1) {
                                            mu.i <- mu.i + A[From,pos]*f(res[,From],p)
                                        } else {
                                            mu.i <- mu.i + A[From,pos]*f(res[,From])
                                        }
                                    } else  mu.i <- mu.i + A[From,pos]*res[,From]
                                }
                            }
                            if (!is.function(dist.i)) {
                                res[,pos] <- mu.i + E[,pos]
                                if (unlink)
                                    resunlink[,pos] <- res[,pos]
                            }
                            else {
                                if (length(simvars)>0) { ## Depends on mu and also on other variables (e.g. time-depending effect)
                                    if (length(mu.i)==1) mu.i <- rep(mu.i,n)
                                    mu.i <- cbind("m0"=mu.i,res[,simvars,drop=FALSE])
                                }
                                new.args <- list(n=n)
                                mu.arg <- intersect(c("mean","mu"),dist.args0)
                                if (length(mu.arg)>0) {
                                    new.args <- c(new.args,list(mu.i))
                                    names(new.args)[length(new.args)] <- mu.arg[1]
                                }
                                var.arg <- intersect(c("var"),dist.args0)
                                if (length(var.arg)>0) {
                                    new.args <- c(new.args,list(P[pos,pos]))
                                    names(new.args)[length(new.args)] <- var.arg[1]
                                }
                                for (jj in dist.xx) {
                                    new.args <- c(new.args,list(res[,jj,drop=TRUE]))
                                    names(new.args)[length(new.args)] <- jj
                                }
                                res[,pos] <- do.call(dist.i,new.args)
                                if (unlink)
                                    resunlink[,pos] <- mu.i
                            }

                            if (length(x$constrainY)>0 && i%in%names(x$constrainY)) {
                                cc <- x$constrainY[[i]]
                                args <- cc$args
                                args <- if (length(args)==0)
                                           res[,pos]
                                       else {
                                           ii <- intersect(names(M$parval),args)
                                           args0 <- args
                                           args <- res[,intersect(args0,colnames(res)),drop=FALSE]

                                           if (length(ii)>0) {
                                               pp <- rbind(unlist(M$parval[ii]))%x%cbind(rep(1,n))
                                               colnames(pp) <- ii
                                               args <- cbind(res,pp)[,args0,drop=FALSE]
                                           }
                                           args
                                       }
                                res[,pos] <- cc$fun(args,p) # ,...)
                            }
                            simuled <- c(simuled,i)
                        }
                    }
                }
            }
        }
        res <- res[,nn,drop=FALSE]
    }

    res <- as.data.frame(res)
    myhooks <- gethook("sim.hooks")
    for (f in myhooks) {
        res <- do.call(f, list(x=x,data=res,p=p,modelpar=M))
    }
    if (unlink) res <- resunlink

    res <- as.data.frame(res)
    self <- x$attributes$selftransform
    for (v in names(self)) {
        res[,v] <- self[[v]](res[,v])
    }
    if (!latent && length(latent(x))>0) return(subset(res[,-which(colnames(res)%in%latent(x))]))
    return(res)
}

##' @export
sim.lvmfit <- function(x, n = nrow(model.frame(x)), p = pars(x), xfix = TRUE, ...) {
  m <- Model(x)
  if ((nrow(model.frame(x)) == n) & xfix) {
    X <- exogenous(x)
    mydata <- model.frame(x)
    for (pred in X) {
      distribution(m, pred) <- list(mydata[, pred])
    }
  }
  sim(m, n = n, p = p, ...)
}

##' @export
simulate.lvm <- function(object, nsim, seed = NULL, ...) {
  sim(object, nsim, seed = seed, ...)
}

##' @export
simulate.lvmfit <- function(object, nsim, seed = NULL, ...) {
  sim(object, nsim, seed = seed, ...)
}

##' Generic cancel method
##'
##' @title Generic cancel method
##' @param x Object
##' @param \dots Additioal arguments
##' @author Klaus K. Holst
##' @aliases cancel<-
##' @export
"cancel" <- function(x,...) UseMethod("cancel")

##' @export
"cancel<-" <- function(x,...,value) UseMethod("cancel<-")

##' @export
"cancel<-.lvm" <- function(x, ..., value) {
  cancel(x,value,...)
}


##' @export
cancel.lvm <- function(x,value,...) {
  if (inherits(value,"formula")) {
    lhs <- getoutcome(value)
    if (length(lhs)==0) yy <- NULL else yy <- decomp.specials(lhs)
    xf <- attributes(terms(value))$term.labels
    if(identical(all.vars(value),xf))
      return(cancel(x,xf))
    res <- lapply(xf,decomp.specials)
    xx <- unlist(lapply(res, function(z) z[1]))
    for (i in yy) {
      for (j in xx)
        cancel(x) <- c(i,j)
      }
    index(x) <- reindex(x)
  return(x)
  }

  for (v1 in value)
    for (v2 in value)
      if (v1!=v2)
        {
          if (all(c(v1,v2)%in%vars(x))) {
            x$M[v1,v2] <- 0
            x$par[v1,v2] <- x$fix[v1,v2] <-
              x$covpar[v1,v2] <- x$covfix[v1,v2] <- NA
            x$cov[v1,v2] <- 0
          }
        }
  x$parpos <- NULL
  index(x) <- reindex(x)
  return(x)
}

##' Returns device-coordinates and plot-region
##'
##' @title Returns device-coordinates and plot-region
##' @return A \code{list} with elements
##'  \item{dev.x1}{Device: Left x-coordinate}
##'  \item{dev.x2}{Device: Right x-coordinate}
##'  \item{dev.y1}{Device Bottom y-coordinate}
##'  \item{dev.y2}{Device Top y-coordinate}
##'  \item{fig.x1}{Plot: Left x-coordinate}
##'  \item{fig.x2}{Plot: Right x-coordinate}
##'  \item{fig.y1}{Plot: Bottom y-coordinate}
##'  \item{fig.y2}{Plot: Top y-coordinate}
##' @author Klaus K. Holst
##' @export
##' @keywords hplot
`devcoords` <-
    function() {
        cc <- par("usr") ## extremes of coordinates of plotting region (x1,x2,y1,y2)
        plotinch <- par("pin") ## Plot dimensions (width,height) in inches
        margininch <- par("mai") ## Margin sizes in inches (bottom, left, top ,right)
        plotlenX <- cc[2]-cc[1]
        unitinchX <- plotlenX/plotinch[1]
        plotlenY <- cc[4]-cc[3]
        unitinchY <- plotlenY/plotinch[2]
        deviceXleft <- cc[1]-unitinchX*margininch[2]
        deviceXright <- cc[2]+unitinchX*margininch[4]
        deviceYtop <- cc[4]+unitinchY*margininch[3]
        deviceYbottom <- cc[3]-unitinchY*margininch[1]
        return(list(dev.x1=deviceXleft, dev.x2=deviceXright, dev.y1=deviceYbottom, dev.y2=deviceYtop, fig.x1=cc[1], fig.x2=cc[2], fig.y1=cc[3], fig.y2=cc[4]))
    }

##' @export
totaleffects <- function(object,...,value) UseMethod("totaleffects")

##' @export
totaleffects.lvmfit <- function(object,to,...,level=0.95) {
    p <- (1-level)/2
    q <- qnorm(p)
    res <- c()
    if (inherits(to,"formula")) {
        if (substr(deparse(to[3]),1,1)==".") {
            trim <- function(x) sapply(x,function(z) gsub(" ","",z,fixed=TRUE))
            to <- trim(strsplit(deparse(to),"~")[[1]][1])
        } else {
            to <- list(to)
        }
    }
    if (is.null(list(...)$from) & is.character(to)[1]) {
        to <- lapply(paste(to,setdiff(vars(object),to),sep="~"),FUN=as.formula)
    }
    ef <- function(tt) {
        f <- effects(object,tt,...)
        rbind(with(f$totalef,c(est,sd,est/sd,2*(pnorm(abs(est/sd),lower.tail=FALSE)),est+q*sd,est-q*sd)))
    }
    if (is.list(to)) {
        for (tt in to) {
            res <- rbind(res,ef(tt))
        }
    }
    else
        res <- ef(to)
    colnames(res) <- c("Estimate","Std.Err","z value","Pr(>|z|)",
                       paste0(c(1-p,p)*100,"%"))
    rownames(res) <- to
    res
}

##' @export
effects.lvmfit <- function(object,to,from,...) {
    if (missing(to)) {
        return(summary(object))
    }
    P <- path(object,to=to,from=from,...)
    if (is.null(P$path)) {
        if (inherits(to,"formula")) {
            f <- extractvar(to)
            to <- f$y; from <- f$x
        }
    } else {
        from <- P$path[[1]][1]
        to <- tail(P$path[[1]],1)
    }
    cc <- coef(object,type=9,labels=FALSE) ## All parameters (fixed and variable)
    cc0 <- cbind(coef(object)) ## Estimated parameters
    i1 <- na.omit(match(rownames(cc),rownames(cc0)))
    idx.cc0 <-  which(rownames(cc)%in%rownames(cc0)); ## Position of estimated parameters among all parameters
    S <- matrix(0,nrow(cc),nrow(cc)); rownames(S) <- colnames(S) <- rownames(cc)
    V <- object$vcov
    npar.mean <- index(object)$npar.mean
    S[idx.cc0,idx.cc0] <- V[i1,i1] ## "Covariance matrix" of all parameters

    cclab <- rownames(coef(object,type=9,labels=TRUE)) ## Identify equivalence constraints
    cctab <- table(cclab)
    equiv <- which(cctab>1)
    for (i in seq_len(length(equiv))) {
        orgpos <- which(cclab==(names(equiv)[i]))
        pos <- orgpos[-1]
        for (p in pos)
            S[p,-orgpos[1]] <- S[-orgpos[1],p] <- S[orgpos[1],-p]
    }

    idx.orig <- unique(unlist(P$idx))
    coefs.all <- cc[idx.orig]

    S.all <- S[idx.orig,idx.orig]
    idx.all <- numberdup(unlist(P$idx))
    pos <- 1; idx.list <- P$idx; for (i in seq_len(length(idx.list))) {
                                     K <- length(idx.list[[i]])
                                     idx.list[[i]] <- idx.all[pos:(pos+K-1)]; pos <- pos+K
                                 }
    margef <- list()
    if (length(coefs.all)==1 && is.na(coefs.all)) {
        totalef <- list(est=0,sd=0)
        margef <- c(margef,list(est=0,sd=NA))
    } else {
        totalef <- prodsumdelta(coefs.all, idx.list, S.all,...)
        for (i in seq_len(length(idx.list))) {
            margef <- c(margef, list(prodsumdelta(coefs.all, idx.list[i], S.all,...)))
        }
    }

    directidx <- which(lapply(P$path,length)==2)

    inef.list <- idx.list
    if (length(directidx)==0) {
        directef <- list(est=0, sd=NA)
    } else {
        inef.list <- inef.list[-directidx]
        directef <- margef[[directidx]]
    }
    if (length(inef.list)==0) {
        totalinef <- list(est=0,sd=NA,grad=NA,hess=NA)
    } else {
        totalinef <- prodsumdelta(coefs.all, inef.list, S.all,...)
    }

    nn <- c("total","direct","indirect")
    for (i in seq_len(length(margef))) {
        if (length(P$path[[i]])>2) {
            nn <- c(nn,paste(rev(P$path[[i]]),collapse=lava.options()$symbol[1]))
        }
    }
    b <- c(totalef$est,directef$est,totalinef$est,totalinef$b)
    names(b) <- nn
    D <- t(cbind(totalef$grad,directef$grad,totalinef$grad,totalinef$D))
    V <- D%*%S.all%*%t(D)
    val <- list(coef=b, vcov=V, grad=D, paths=P$path, totalef=totalef, directef=directef, totalinef=totalinef, margef=margef, from=from, to=to)
    class(val) <- "effects"
    val
}

##' @export
print.effects <- function(x,digits=4,...) {
    s <- summary(x,...)
    print(s$coef,digits=digits,...)
    cat("\n")
    print(s$medprop$coefmat[,c(1,3,4),drop=FALSE],digits=digits,...)
    return(invisible(x))
}

##' @export
coef.effects <- function(object,...) {
    object$coef
}

##' @export
vcov.effects <- function(object,...) {
    object$vcov
}

##' @export
summary.effects <- function(object,...) {
    totalef <- with(object$totalef, cbind(est,sd[1]))
    directef <- with(object$directef, cbind(est,sd[1]))
    totindirectef <- with(object$totalinef, cbind(est,sd[1]))
    rownames(totalef) <- "Total"
    rownames(directef) <- "Direct"
    rownames(totindirectef) <- "Indirect"
    nn <- indirectef <- c()
    K <- seq_len(length(object$margef))
    for (i in K) {
        if (length(object$paths[[i]])>2) {
            nn <- c(nn,paste(rev(object$paths[[i]]),collapse=lava.options()$symbol[1]))
            indirectef <- rbind(indirectef, with(object$margef[[i]], c(est,sd)))
        }
    }; rownames(indirectef) <- nn
    mycoef <- rbind(totalef,directef,totindirectef,indirectef)
    mycoef <- cbind(mycoef,mycoef[,1]/mycoef[,2])
    mycoef <- cbind(mycoef,2*(pnorm(abs(mycoef[,3]),lower.tail=FALSE)))
    colnames(mycoef) <- c("Estimate","Std.Err","z value","Pr(>|z|)")
    medprop <- NULL
    if (totindirectef[1]!=0) {
        if (abs(coef(object)[2])<1e-12) {
            medprop <- estimate(NULL,coef=c("Mediation proportion"=1),vcov=matrix(NA))
        } else {
            medprop <- estimate(object, function(x) list("Mediation proportion"=1-x[2]/x[1]))
            ##medprop <- estimate(object, function(x) list("Mediation proportion"=logit(x[3]/x[1])),back.transform=expit)
        }
    }
    list(coef=mycoef,medprop=medprop)
}


##' @export
confint.effects <- function(object,parm,level=0.95,...) {
    mycoef <- summary(object)$coef
    p <- 1-(1-level)/2
    res <- mycoef[,1] +  + qnorm(p)*cbind(-1,1)%x%mycoef[,2]
    colnames(res) <- paste0(c(1-p,p)*100,"%")
    rownames(res) <- rownames(mycoef)
    res
}


prodtrans <- function(betas) {
    k <- length(betas)
    res <- prod(betas)
    nabla <- numeric(k)
    for (i in seq_len(k))
        nabla[i] <- prod(betas[-i])

    H <- matrix(0,k,k)
    if (k>1)
        for (i in seq_len(k-1))
            for (j in (i+1):k)
                H[j,i] <- H[i,j] <- prod(c(1,betas[-c(i,j)]))
    attr(res,"gradient") <- nabla
    attr(res,"hessian") <- H
    return(res)
}
prodsumdelta <- function(betas,prodidx,S,order=1) { ## Delta-method
    k <- length(prodidx)
    p <- length(betas)
    if (p==1) {
        return(list(est=betas, sd=sqrt(S), grad=0, beta=betas, D=0, hess=0))
    }
    val <- 0; grad <- numeric(p)
    D <- matrix(0,nrow=p,ncol=k)
    beta <- numeric(k)
    H <- matrix(0,p,p)
    for (i in seq_len(k)) {
        ii <- prodidx[[i]]
        myterm <- prodtrans(betas[ii]);
        if (order>1) {
            H0 <- attributes(myterm)$hessian
            Sigma <- S[ii,ii]
            print(sum(diag(Sigma%*%H0))/2)
            val <- val + (myterm + sum(diag(Sigma%*%H0))/2)
        } else {
            val <- val + myterm
            beta[i] <- myterm
        }
        D[ii,i] <- attributes(myterm)$gradient
        grad[ii] <- grad[ii] + attributes(myterm)$gradient
    }; grad <- matrix(grad,ncol=1)
    return(list(est=val, sd=sqrt(t(grad)%*%S%*%grad), grad=grad, b=beta, D=D, hess=H))
}

##' Stack estimating equations
##'
##' Stack estimating equations (two-stage estimator)
##' @param x Model 1
##' @param model2 Model 2
##' @param D1u Derivative of score of model 2 w.r.t. parameter vector of model 1
##' @param inv.D2u Inverse of deri
##' @param propensity propensity score (vector or function)
##' @param dpropensity derivative of propensity score wrt parameters of model 1
##' @param U Optional score function (model 2) as function of all parameters
##' @param keep1 If FALSE only parameters of model 2 is returned
##' @param propensity.arg Arguments to propensity function
##' @param estimate.arg Arguments to 'estimate'
##' @param na.action Method for dealing with missing data in propensity score
##' @param ... Additional arguments to lower level functions
##' @aliases stack.estimate
##' @export
##' @examples
##' m <- lvm(z0~x)
##' Missing(m, z ~ z0) <- r~x
##' distribution(m,~x) <- binomial.lvm()
##' p <- c(r=-1,'r~x'=0.5,'z0~x'=2)
##' beta <- p[3]/2
##' d <- sim(m,500,p=p)
##' m1 <- estimate(r~x,data=d,family=binomial)
##' d$w <- d$r/predict(m1,type="response")
##' m2 <- estimate(z~1, weights=w, data=d)
##' (e <- stack(m1,m2,propensity=TRUE))
stack.estimate <- function(x,model2,D1u,inv.D2u,
                   propensity,dpropensity,U,
                   keep1=FALSE,
                   propensity.arg,estimate.arg,
                   na.action=na.pass, ...) {
    iid2 <- iid(model2)
    if (missing(inv.D2u)) {
        inv.D2u <- -attributes(iid2)$bread
    }    
    if (is.null(inv.D2u)) stop("Need derivative of second stage score")
    if (!missing(U)) {
        D1u <- numDeriv::jacobian(U,coef(x))
        D1u <- colSums(D1u)
    }
    if (!missing(propensity) && is.logical(propensity) && propensity) {
        if (!inherits(x,"glm")) stop("'x' must be of 'glm' class")
        f <- family(x)
        ginv <- f$linkinv
        dginv <- f$mu.eta ## D[linkinv]        
        dpropensity <- function(p, data=model.frame(x), ...) {
            X <- model.matrix(formula(x),data, ...)
            p1 <- dginv(X%*%p)
            apply(X,2,function(x) x*p1)
        }
        propensity <- function(p, data=model.frame(x), ...) {
            X <- model.matrix(formula(x),data, ...)
            ginv(X%*%p)            
        }        
    }    
    if (!missing(propensity) && is.function(propensity)) {
        op <- options(na.action=na.action)
        wfun <- propensity
        if (!missing(propensity.arg)) {
            wfun <- function(p,...)
                do.call(propensity, c(list(p),propensity.arg))
        }
        if (missing(dpropensity)) {
            dpropensity <- numDeriv::jacobian(wfun,coef(x),...)
        } else if (is.function(dpropensity)) {
            if (!missing(propensity.arg)) {
                dwfun <- function(p,...)
                    do.call(dpropensity, c(list(p),propensity.arg))
                dpropensity <- dwfun(coef(x),...)
            } else {
                dpropensity <- dpropensity(coef(x),...)
            }
        }
        propensity <- wfun(coef(x),...)
        mis <- which(is.na(propensity))
        propensity[mis] <- 0
        dpropensity[mis,] <- 0
        options(op)
    }
    if (!missing(dpropensity)) {
        D2u <- Inverse(inv.D2u)
        u2 <- -iid2%*%D2u ## Score of stage two equation derived from estimated influence function
        ## Derivative of score wrt first set of parameters (propensity score model)
        D1u <- crossprod(apply(u2,2,function(x) -na.pass0(x/propensity)),dpropensity)
    }
    ii <- iid(merge(x,model2))
    iid1. <- ii[,seq_along(coef(x)),drop=FALSE]
    iid2. <- ii[,length(coef(x))+seq_along(coef(model2)),drop=FALSE]
    iid3 <- -t(inv.D2u%*%(D1u%*%t(iid1.)))
    val <- cbind(iid2. + iid3)
    if (!keep1) return(estimate(coef=coef(model2), iid=val, ...))
    estimate(coef=c(coef(x), coef(model2)), iid=cbind(iid1., val),...)
}

##' @export
stack.glm <- function(x,model2,...) {
    stack.estimate(x,model2,...)
}

##' @export
stack.lvmfit <- function(x,model2,...) {
    stack(estimate(x),model2,...)
}

##' Extract model
##'
##' Extract or replace model object
##'
##'
##' @aliases Model Model<-
##' @usage
##'
##' Model(x, ...)
##'
##' Model(x, ...) <- value
##'
##' @param x Fitted model
##' @param value New model object (e.g. \code{lvm} or \code{multigroup})
##' @param \dots Additional arguments to be passed to the low level functions
##' @return Returns a model object (e.g. \code{lvm} or \code{multigroup})
##' @author Klaus K. Holst
##' @seealso \code{\link{Graph}}
##' @keywords models
##' @examples
##'
##' m <- lvm(y~x)
##' e <- estimate(m, sim(m,100))
##' Model(e)
##'
##' @export
`Model` <- function(x,...) UseMethod("Model")


##' @export
`Model.default` <- function(x,...) x

##' @export
`Model.lvm` <- function(x,...) x

##' @export
`Model.lvmfit` <- function(x,...) x$model

##' @export
`Model.multigroup` <- function(x,...) x$lvm

##' @export
`Model.multigroupfit` <- function(x,...) x$model

##' @export
"Model<-" <- function(x,...,value) UseMethod("Model<-")

##' @export
"Model<-.lvm" <- function(x,...,value) { x <- value; return(x) }
##' @export
"Model<-.lvmfit" <- function(x,...,value) { x$model <- value; return(x) }
##' @export
"Model<-.multigroup" <- function(x,...,value) { x$lvm <- value; return(x) }
##' @export
"Model<-.multigroupfit" <- function(x,...,value) { x$model <- value; return(x) }

sumsplit <- function(x,...) {
    plus <- strsplit(x,"\\+")[[1]]
    spl <- unlist(lapply(plus, function(x) {
        val <- strsplit(x,"\\-")[[1]]
        val[-1] <- paste0("-",val[-1])
        setdiff(val,"")
    }))
    res <- c()
    for (st in spl) {
        st <- gsub(" ","",st)
        st0 <- gsub("^[-0-9\\*]*","",st)
        val <- gsub("\\*","",regmatches(st,gregexpr("^[-0-9\\*]*",st))[[1]])
        if (val=="") val <- "1"
        val <- switch(val,"-"=-1,val)
        res <- c(res,val,st0)    
    }
    return(res)
}

##' @export
parsedesign <- function(coef,x,...,regex=FALSE,diff=TRUE) {
    if (!is.vector(coef)) coef <- stats::coef(coef)
    if (is.numeric(coef) && !is.null(names(coef))) coef <- names(coef)    
    dots <- lapply(substitute(list(...)),function(x) x)[-1]
    expr <- suppressWarnings(inherits(try(x,silent=TRUE),"try-error"))
    if (expr) {
        ee <- c(deparse(substitute(x)), unlist(lapply(dots, deparse)))
    } else {
        ee <- c(deparse(x), sapply(dots, function(x) deparse(x)))
    }
    if (!expr && is.numeric(x)) {
        return(do.call(contr, list(c(list(x),list(...)),n=length(coef),diff=diff)))
    }
    res <- c()
    diff <- rep(diff,length.out=length(ee))
    count <- 0
    for (e in ee) {
        count <- count+1
        diff0 <- FALSE
        Val <- rbind(rep(0,length(coef)))
        if (grepl('\"',e)) {        
            diff0 <- diff[count] && grepl("^c\\(",e)
            e0 <- gsub(" |\\)$|^c\\(","",e)
            ff <- strsplit(e0,'\"')[[1L]]
        } else {
            ff <- sumsplit(e)
        }
        for (i in seq(length(ff)/2)) {
            val0 <- gsub("[*()]","",ff[2*(i-1)+1])            
            val <- char2num(val0)
            if (is.na(val)) {
                val <- switch(val0,"-"=-1,1)
            }
            par0 <- ff[2*i]
            par0int <- as.integer(char2num(par0))
            if (!regex) par0 <- glob2rx(par0)
            if (is.na(par0int)) par0int <- grep(par0,coef)
            if (length(par0int)>1) {
                diff0 <- diff[count]
                for (k in seq_along(par0int)) {
                    if (par0int[k]<=length(Val)) {
                        if (diff[count]) {
                            Val[par0int[k]] <- val
                        } else {
                            Val0 <- Val; Val0[] <- 0
                            Val0[par0int[k]] <- val
                            res <- rbind(res,Val0)
                        }
                    }
                }
            } else {
                if (length(par0int)>0 && par0int<=length(Val)) Val[par0int] <- val
            }
        }
        if (diff0) {
            n <- sum(Val!=0)
            if (n>1) {
                Val0 <- Val
                ii <- which(Val0!=0)
                Val <- matrix(0,nrow=n-1,ncol=length(Val))
                for (i in seq(n-1)) {
                    Val[i,ii[c(1,i+1)]] <- Val0[ii[c(1,i+1)]]*c(1,-1)
                }
            }
        }
        if (any(Val!=0)) res <- rbind(res,Val)
    }
    res
}

##' @export
multigroup <- function(models, datasets, fix, exo.fix=TRUE, keep=NULL, missing=FALSE, ...) {
    nm <- length(models)
    if (nm!=length(datasets)) stop("Supply dataset for each model")
    if (nm<2) stop("Two or more groups neeeded")
    mynames <- names(models)

    ## Check for random slopes
    xfix <- list()
    for (i in seq_len(nm)) {
        x0 <- models[[i]]
        data0 <- datasets[[i]]
        xfix0 <- colnames(data0)[(colnames(data0)%in%parlabels(x0,exo=TRUE))]
        xfix <- c(xfix, list(xfix0))
    }
    if (missing(fix)) {
        fix <- !any(unlist(lapply(xfix, function(x) length(x)>0)))
    }

    for (i in seq_len(nm)) {
        x0 <- models[[i]]
        data0 <- datasets[[i]]
        if (length(exogenous(x0)>0)) {
            catx <- categorical2dummy(x0,data0)
            models[[i]] <- catx$x; datasets[[i]] <- catx$data
        }
        if (!lava.options()$exogenous) exogenous(models[[i]]) <- NULL
    }

    models.orig <- NULL
    ######################
    ### MLE with MAR mechanism
    ######################
    if (missing) {

        reservedpars <- c()
        mynpar <- c()
        for (i in seq_len(nm)) {
            ## Fix some parameters (predictors,latent variables,...)

            d0 <- datasets[[i]][1,,drop=FALSE]; d0[,] <- 1
            if (fix)
                models[[i]] <- fixsome(models[[i]], exo.fix=exo.fix, measurement.fix=fix, data=d0)
            ## Find named/labelled parameters
            rpar <- unique(parlabels(models[[i]]))
            reservedpars <- c(reservedpars, rpar)
            mynpar <- c(mynpar, with(index(models[[1]]), npar+npar.mean+npar.ex))
        }; reservedpars <- unique(reservedpars)
        nonamepar <- sum(mynpar)
        ## Find unique parameter-names for all parameters
        newpars <- c()
        i <- 0
        pos <- 1
        while(pos<=nonamepar) {
            i <- i+1
            newname <- paste0("par",i)
            if (!(newname%in%reservedpars)) {
                newpars <- c(newpars,newname)
                pos <- pos+1
            }
        }

        pos <- 0
        models0 <- list()
        datasets0 <- list()
        complidx <- c()
        nmodels <- 0
        modelclass <- c()
        nmis <- c()
        for (i in seq_len(nm)) {
            myvars <- unlist(intersect(colnames(datasets[[i]]),c(vars(models[[i]]),xfix[[i]],keep)))
            mydata <- datasets[[i]][,myvars]
            if (any(is.na(mydata))) {
                if (i>1) pos <- pos+mynpar[i-1]
                models[[i]] <- baptize(models[[i]],newpars[pos+seq_len(mynpar[i])] ,overwrite=FALSE)
                val <- missingModel(models[[i]],mydata,fix=FALSE,keep=keep,...)
                nmodels <- c(nmodels,length(val$models))
                complidx <- c(complidx,val$pattern.allcomp+nmodels[i]+1)
                nmis0 <- rowSums(val$patterns);
                allmis <- which(nmis0==ncol(val$patterns))
                if (length(allmis)>0) nmis0 <- nmis0[-allmis]
                nmis <- c(nmis,nmis0)
                datasets0 <- c(datasets0, val$datasets)
                models0 <- c(models0, val$models)
                modelclass <- c(modelclass,rep(i,length(val$models)))
            } else {
                datasets0 <- c(datasets0, list(mydata))
                models0 <- c(models0, list(models[[i]]))
                modelclass <- c(modelclass,i)
                nmis <- c(nmis,0)
            }
        }

        models.orig <- models

        suppressWarnings(
            val <- multigroup(models0,datasets0,fix=FALSE,missing=FALSE,exo.fix=TRUE,...)
        )
        val$models.orig <- models.orig; val$missing <- TRUE
        val$complete <- complidx-1
        val$mnames <- mynames
        attributes(val)$modelclass <- modelclass
        attributes(val)$nmis <- nmis
        return(val)
    }


    ######################
    ### Usual analysis:
    ######################
    warned <- FALSE
    for (i in seq_len(nm)) {
        if (inherits(datasets[[i]],c("data.frame","matrix"))) {
            myvars <- intersect(colnames(datasets[[i]]),c(vars(models[[i]]),xfix[[i]],keep))
            if (any(is.na(datasets[[i]][,myvars]))) {
                if (!warned)
                    warning(paste0("Missing data encountered. Going for complete-case analysis"))
                warned  <- TRUE
                datasets[[i]] <- na.omit(datasets[[i]][,myvars,drop=FALSE])
            }
        }
    }

    exo <- exogenous(models)
    means <- lvms <- As <- Ps <- ps <- exs <- datas <- samplestat <- list()
    for (i in seq_len(nm)) {

        if (!is.null(exogenous(models[[i]]))) {
            if (any(is.na(exogenous(models[[i]])))) {
                exogenous(models[[i]]) <- exo
            }
        }

        mydata <- datasets[[i]]
        mymodel <- fixsome(models[[i]], data=mydata, measurement.fix=fix, exo.fix=exo.fix)
        mymodel <- updatelvm(mymodel,zeroones=TRUE,deriv=TRUE)

        P <- index(mymodel)$P1; P[P==0] <- NA
        P[!is.na(P) & !is.na(mymodel$covpar)] <- mymodel$covpar[!is.na(P) & !is.na(mymodel$covpar)]

        A <- index(mymodel)$M1; A[A==0] <- NA
        A[!is.na(A) & !is.na(mymodel$par)] <- mymodel$par[!is.na(A) & !is.na(mymodel$par)]

        mu <- unlist(mymodel$mean)[which(index(mymodel)$v1==1)]
        #ex <- names(mymodel$expar)[which(index(mymodel)$e1==1)]
        ex <- mymodel$exfix
        if (length(ex)>0) {
            if (any(is.na(ex))) ex[is.na(ex)] <- mymodel$expar[is.na(ex)]
            ex <- ex[which(index(mymodel)$e1==1)]
        }

        p <- pars(mymodel, A, P, e=ex)
        p[p=="1"] <- NA

        means <- c(means, list(mu))
        lvms <- c(lvms, list(mymodel))
        datas <- c(datas, list(mydata))
        samplestat <- c(samplestat, list(procdata.lvm(models[[i]],data=mydata)))
        As <- c(As, list(A))
        Ps <- c(Ps, list(P))
        ps <- c(ps, list(p))
        exs <- c(exs, list(ex))
    };

    ######
    pp <- unlist(ps)
    parname <- unique(pp[!is.na(pp)])
    pidx <- is.na(char2num(parname))
    parname <- unique(unlist(pp[!is.na(pp)]));
    nfree <- sum(is.na(pp)) + length(parname)

    if (nfree>0) {
        pp0 <- lapply(ps, is.na)
        usedname <- cbind(parname, rep(NA,length(parname)))
        counter <- 1
        pres <- pres0 <- pp0
        for (i in seq_len(length(pp0))) {
            if (length(pp0[[i]]>0))
                for (j in seq_len(length(pp0[[i]]))) {
                    pidx <- match(ps[[i]][j],parname)
                    if (pp0[[i]][j]) {
                        pres[[i]][j] <- paste0("p",counter)
                        pres0[[i]][j] <- counter
                        counter <- counter+1
                    } else if (!is.na(pidx)) {
                        if (!is.na(usedname[pidx,2])) {
                            pres[[i]][j] <- usedname[pidx,2]
                            pres0[[i]][j] <- char2num(substr(pres[[i]][j],2,nchar(pres[[i]][j])))
                        } else {
                            val <- paste0("p",counter)
                            pres[[i]][j] <- val
                            pres0[[i]][j] <- counter
                            usedname[pidx,2] <- val
                            counter <- counter+1
                        }
                    } else {
                        pres[[i]][j] <- NA
                    }
                }
        }
        mypar <- paste0("p",seq_len(nfree))
        myparPos <- pres0
        myparpos <- pres
        myparlist <- lapply(pres, function(x) x[!is.na(x)])
    } else {
        myparPos <- NULL
        mypar <- NULL
        myparpos <- NULL
        myparlist <- NULL
    }

    ### Mean parameter

    mm <- unlist(means)
    meanparname <- unique(mm[!is.na(mm)])
    midx <- is.na(char2num(meanparname));
    meanparname <- meanparname[midx]
    any.mean <- sum(is.na(mm)) + length(meanparname)
    nfree.mean <- sum(is.na(mm)) + length(setdiff(meanparname,parname))
    ## mean.fixed <- na.omit(match(parname,mm))
    mean.omit <- lapply(means,function(x) na.omit(match(parname,x)))

    if (any.mean>0) {
        mm0 <- lapply(means, is.na)
        usedname <- cbind(meanparname, rep(NA,length(meanparname)))
        counter <- 1
        res0 <- res <- mm0
        for (i in seq_len(length(mm0))) {
            if (length(mm0[[i]])>0)
                for (j in seq_len(length(mm0[[i]]))) {
                    midx <- match(means[[i]][j],meanparname)
                    if (mm0[[i]][j]) {
                        res[[i]][j] <- paste0("m",counter)
                        res0[[i]][j] <- counter
                        counter <- counter+1
                    } else if (!is.na(midx)) {
                        pidx <- match(meanparname[midx],pp)
                        if (!is.na(pidx)) {
                            res[[i]][j] <- unlist(myparlist)[pidx]
                            res0[[i]][j] <- char2num(substr(res[[i]][j],2,nchar(res[[i]][j]))) +
                                nfree.mean
                        } else {
                            if (!is.na(usedname[midx,2])) {
                                res[[i]][j] <- usedname[midx,2]
                                res0[[i]][j] <- char2num(substr(res[[i]][j],2,nchar(res[[i]][j])))
                            } else {
                                val <- paste0("m",counter)
                                res[[i]][j] <- val
                                res0[[i]][j] <- counter
                                usedname[midx,2] <- val
                                counter <- counter+1
                            }
                        }
                    } else {
                        res[[i]][j] <- NA
                    }
                }
        }
        mymeanPos <- res0
        mymeanpos <- res
        mymeanlist <- lapply(res, function(x) x[!is.na(x)])
        mymean <- unique(unlist(mymeanlist))
    } else {
        mymeanPos <- NULL
        mymean <- NULL
        mymeanpos <- NULL
        mymeanlist <- NULL
    }

    ### Extra parameters

    m0 <- p0 <- c()
    coefs <- coefsm <- mm0 <- mm <- pp0 <- pp <- c()
    for (i in seq_len(length(myparPos))) {
        mi <- mymeanPos[[i]]
        pi <- myparPos[[i]]
        p1 <- setdiff(pi,p0)
        p0 <- c(p0,p1)
        ##    pp0 <- c(pp0,list(match(p1,pi)+nfree.mean))
        pp0 <- c(pp0,list(match(p1,pi)))
        if (length(mean.omit[[i]])>0) mi <- mi[-mean.omit[[i]]]
        m1 <- setdiff(mi,m0)
        m0 <- c(m0,m1)
        mm0 <- c(mm0,list(match(m1,mi)))
        pp <- c(pp,list(c(m1,p1+nfree.mean)))
        if (length(p1)>0)
            coefs <- c(coefs,paste(coef(lvms[[i]],fix=FALSE,mean=FALSE)[pp0[[i]]],i,sep="@"))
            #coefs <- c(coefs,paste(i,coef(lvms[[i]],fix=FALSE,mean=FALSE)[pp0[[i]]],sep="@"))
        if (length(m1)>0) {
            coefsm0 <- paste(coef(lvms[[i]],fix=FALSE,mean=TRUE)[mm0[[i]]],i,sep="@")
            #coefsm0 <- paste(i,coef(lvms[[i]],fix=FALSE,mean=TRUE)[mm0[[i]]],sep="@")
            coefsm <- c(coefsm,coefsm0)
        }
    }
    coefs <- c(coefsm,coefs)

    res <- list(npar=nfree, npar.mean=nfree.mean,
                ngroup=length(lvms), names=mynames,
                lvm=lvms, data=datas, samplestat=samplestat,
                A=As, P=Ps, expar=exs,
                meanpar=names(mu), name=coefs, coef=pp, coef.idx=pp0,
                par=mypar, parlist=myparlist,  parpos=myparpos,
                mean=mymean, meanlist=mymeanlist, meanpos=mymeanpos,
                parposN=myparPos,
                meanposN=mymeanPos,
                models.orig=models.orig, missing=missing
                )
    class(res) <- "multigroup"
    checkmultigroup(res)
    return(res)
}

checkmultigroup <- function(x) {
    ## Check validity:
    for (i in seq_len(x$ngroup)) {
        if (nrow(x$data[[i]])<2) {
            warning("With only one observation in the group, all parameters should be inherited from another a group!")
        }
    }
}

twostagelvm <- function(object, model2,
                formula=NULL, model.object=FALSE, predict.fun=NULL,
                type="quadratic",...) {
    if (!inherits(model2,c("lvm"))) stop("Expected lava object ('lvm',...)")
    if (!is.null(formula)) {
        model2 <- nonlinear(model2, formula, type=type)
    }
    nonlin <- NULL
    val <- nonlinear(model2)
    if (is.null(formula) && length(val)==0 && length(nonlinear(object))>0) {
        val <- nonlinear(object)
    }
    xnam <- c()
    if (length(val)>0) {
        predict.fun <- NULL
        for (i in seq_along(val)) {
            if (!all(val[[i]]$newx%in%xnam)) {
                xnam <- union(xnam,val[[i]]$newx)
                predict.fun <- c(predict.fun, list(val[[i]]$pred))
            }
            model2$attributes$nonlinear <- NULL
            if (inherits(object,"lvmfit")) {
                object$model$attributes$nonlinear <- NULL
            }
            model2 <- regression(model2, to=names(val)[i], from=val[[i]]$newx)
        }
        nonlin <- val
    }
    if (model.object) {
        model <- Model(object) %++% model2
        cl <- match.call(expand.dots=TRUE)
        cl[[1]] <- twostage
        cl$object <- object
        cl$model2 <- model2
        cl$predict.fun <- predict.fun
        cl["model.object"] <- NULL
        return(structure(list(model=model, nonlinear=nonlin, call=cl), class="twostage.lvm"))
    }
    res <- c(list(object=object, model2=model2), list(...))
    res$predict.fun <- predict.fun
    res$nonlinear <- val
    return(res)
}

uhat <- function(p=coef(model1), model1, data=model.frame(model1), nlobj) {
    if (!is.function(nlobj)) {
        predict.fun <- lapply(nlobj, function(x) x[["pred"]])
    } else { predict.fun <- nlobj }
    if (inherits(model1, "lvm.mixture")) {
        if (is.list(predict.fun)) {
            unams <- lapply(nlobj,function(x) x$newx)
            unam <- unique(unlist(unams))
            res <- matrix(0, NROW(data), ncol=length(unam))
            colnames(res) <- unam
            for (i in seq_along(predict.fun)) {
                res[, unams[[i]]] <-
                    predict(model1, p=p, data=data, predict.fun=predict.fun[[i]])
            }
            return(res)
        } else {
            Pr <- cbind(predict(model1, p=p, data=data, predict.fun=predict.fun))
            return(Pr)
        }
        ##P <- list(mean=Pr, var=attr(Pr,"cond.var"))
    }  else {
        P <- predictlvm(model1, p=p, data=data)
    }
    if (is.list(predict.fun)) {
        unams <- lapply(nlobj,function(x) x$newx)
        unam <- unique(unlist(unams))
        args <- list(P$mean, P$var, data)
        res <- matrix(0, NROW(data), ncol=length(unam))
        colnames(res) <- unam
        for (i in seq_along(predict.fun)) {
            res[, unams[[i]]] <- do.call(predict.fun[[i]], args)
        }
        return(res)
    }
    return(cbind(predict.fun(P$mean, P$var, model.frame(model1))))
}

##' Two-stage estimator
##'
##' Generic function.
##'
##' @seealso twostage.lvm twostage.lvmfit twostage.lvm.mixture twostage.estimate
##' @export
##' @param object Model object
##' @param ... Additional arguments to lower level functions
"twostage" <- function(object, ...) UseMethod("twostage")

##' Two-stage estimator (non-linear SEM)
##'
##' Two-stage estimator for non-linear structural equation models
##' @export
##' @param object Stage 1 measurement model
##' @param model2 Stage 2 SEM
##' @param data data.frame
##' @param predict.fun Prediction of latent variable
##' @param id1 Optional id-variable (stage 1 model)
##' @param id2 Optional id-variable (stage 2 model)
##' @param all If TRUE return additional output (naive estimates)
##' @param formula optional formula specifying non-linear relation
##' @param std.err If FALSE calculations of standard errors will be skipped
##' @param ... Additional arguments to lower level functions
##' @aliases twostage.lvmfit twostage.lvm twostage.lvm.mixture twostage.estimate nonlinear nonlinear<-
##' @examples
##' m <- lvm(c(x1,x2,x3)~f1,f1~z,
##'          c(y1,y2,y3)~f2,f2~f1+z)
##' latent(m) <- ~f1+f2
##' d <- simulate(m,100,p=c("f2,f2"=2,"f1,f1"=0.5),seed=1)
##'
##' ## Full MLE
##' ee <- estimate(m,d)
##'
##' ## Manual two-stage
##' \dontrun{
##' m1 <- lvm(c(x1,x2,x3)~f1,f1~z); latent(m1) <- ~f1
##' e1 <- estimate(m1,d)
##' pp1 <- predict(e1,f1~x1+x2+x3)
##'
##' d$u1 <- pp1[,]
##' d$u2 <- pp1[,]^2+attr(pp1,"cond.var")[1]
##' m2 <- lvm(c(y1,y2,y3)~eta,c(y1,eta)~u1+u2+z); latent(m2) <- ~eta
##' e2 <- estimate(m2,d)
##' }
##'
##' ## Two-stage
##' m1 <- lvm(c(x1,x2,x3)~f1,f1~z); latent(m1) <- ~f1
##' m2 <- lvm(c(y1,y2,y3)~eta,c(y1,eta)~u1+u2+z); latent(m2) <- ~eta
##' pred <- function(mu,var,data,...)
##'     cbind("u1"=mu[,1],"u2"=mu[,1]^2+var[1])
##' (mm <- twostage(m1,model2=m2,data=d,predict.fun=pred))
##'
##' if (interactive()) {
##'     pf <- function(p) p["eta"]+p["eta~u1"]*u + p["eta~u2"]*u^2
##'     plot(mm,f=pf,data=data.frame(u=seq(-2,2,length.out=100)),lwd=2)
##' }
##'
##' \dontrun{ ## Reduce test timing
##' ## Splines
##' f <- function(x) cos(2*x)+x+-0.25*x^2
##' m <- lvm(x1+x2+x3~eta1, y1+y2+y3~eta2, latent=~eta1+eta2)
##' functional(m, eta2~eta1) <- f
##' d <- sim(m,500,seed=1,latent=TRUE)
##' m1 <- lvm(x1+x2+x3~eta1,latent=~eta1)
##' m2 <- lvm(y1+y2+y3~eta2,latent=~eta2)
##' mm <- twostage(m1,m2,formula=eta2~eta1,type="spline")
##' if (interactive()) plot(mm)
##'
##' nonlinear(m2,type="quadratic") <- eta2~eta1
##' a <- twostage(m1,m2,data=d)
##' if (interactive()) plot(a)
##'
##' kn <- c(-1,0,1)
##' nonlinear(m2,type="spline",knots=kn) <- eta2~eta1
##' a <- twostage(m1,m2,data=d)
##' x <- seq(-3,3,by=0.1)
##' y <- predict(a, newdata=data.frame(eta1=x))
##'
##' if (interactive()) {
##'   plot(eta2~eta1, data=d)
##'   lines(x,y, col="red", lwd=5)
##'
##'   p <- estimate(a,f=function(p) predict(a,p=p,newdata=x))$coefmat
##'   plot(eta2~eta1, data=d)
##'   lines(x,p[,1], col="red", lwd=5)
##'   confband(x,lower=p[,3],upper=p[,4],center=p[,1], polygon=TRUE, col=Col(2,0.2))
##'
##'   l1 <- lm(eta2~splines::ns(eta1,knots=kn),data=d)
##'   p1 <- predict(l1,newdata=data.frame(eta1=x),interval="confidence")
##'   lines(x,p1[,1],col="green",lwd=5)
##'   confband(x,lower=p1[,2],upper=p1[,3],center=p1[,1], polygon=TRUE, col=Col(3,0.2))
##' }
##' } ## Reduce test timing
##'
##' \dontrun{ ## Reduce timing
##'  ## Cross-validation example
##'  ma <- lvm(c(x1,x2,x3)~u,latent=~u)
##'  ms <- functional(ma, y~u, value=function(x) -.4*x^2)
##'  d <- sim(ms,500)#,seed=1)
##'  ea <- estimate(ma,d)
##'
##'  mb <- lvm()
##'  mb1 <- nonlinear(mb,type="linear",y~u)
##'  mb2 <- nonlinear(mb,type="quadratic",y~u)
##'  mb3 <- nonlinear(mb,type="spline",knots=c(-3,-1,0,1,3),y~u)
##'  mb4 <- nonlinear(mb,type="spline",knots=c(-3,-2,-1,0,1,2,3),y~u)
##'  ff <- lapply(list(mb1,mb2,mb3,mb4),
##'	      function(m) function(data,...) twostage(ma,m,data=data,st.derr=FALSE))
##'  a <- cv(ff,data=d,rep=1)
##'  a
##'}
twostage.lvmfit <- function(object, model2, data=NULL,
                    predict.fun=NULL,
                    id1=NULL, id2=NULL, all=FALSE,
                    formula=NULL, std.err=TRUE,
                    ...) {
    if (!is.null(predict.fun)) {
        object$attributes$nonlinear <- list()
        model2$attributes$nonlinear <- list()
    }
    val <- twostagelvm(object=object,model2=model2,predict.fun=predict.fun,
                      id1=id1, id2=id2, all=all, formula=formula, ...)
    object <- val$object
    model2 <- val$model2
    predict.fun <- val$predict.fun
    p1 <- coef(object)
    if (length(val$nonlinear)==0) {
        val$nonlinear <- predict.fun
    }
    pp <- uhat(p1,object,nlobj=val$nonlinear)
    newd <- data
    newd[,colnames(pp)] <- pp

    model2 <- estimate(model2,data=newd,...)
    p2 <- coef(model2)
    if (std.err) {
        if (is.null(id1)) id1 <- seq(nrow(model.frame(object)))
        if (is.null(id2)) id2 <- seq(nrow(model.frame(model2)))
        model1 <- object
        if (!inherits(object,"estimate")) {
            model1 <- estimate(NULL, coef = p1, id = id1,
                               IC = scale(IC(object), center=TRUE, scale=FALSE)
                               )
        }
        e2 <- estimate(model2, id=id2)
        U <- function(alpha = p1, beta = p2) {
            pp <- uhat(alpha, object, nlobj = val$nonlinear)
            newd <- model.frame(model2)
            newd[, colnames(pp)] <- pp
            score(model2, p = beta, data = newd)
        }
        stacked <- stack(model1,e2,U=U)
    } else {
        e2 <- estimate(coef=p2,vcov=NA)
    }
    coef <- model2$coef
    res <- model2
    res$estimator <- "generic"

    if (std.err) {
        res[names(stacked)] <- stacked
        cc <- stacked$coefmat[,c(1,2)];
        cc <- cbind(cc,cc[,1]/cc[,2],stacked$coefmat[,5])
        coef[,] <- cc
        res$coef <- coef
        res$vcov <- vcov(stacked)
        if (all) {
            res$naive <- model2
            res$naive.robust <- e2
        }
    } else {
        res$coef[,-1] <- NA
    }
    res$fun <- predict.fun
    res$estimate1 <- object
    res$estimate2 <- model2
    res$nonlinear <- val$nonlinear
    structure(
      res, class=c("twostage.lvmfit",
                   "measurement.error",
                   "lvmfit",
                   "estimate")
    )
}

##' @export
estimate.twostage.lvm <- function(x,data,...) {
    if (missing(data)) stop("'data' needed")
    m1 <- x$call$object
    m2 <- x$call$model2
    nl <- x$nonlinear
    if (!inherits(m1,"lvmfit")) {
        args <- c(list(x=m1, data=data), list(...))
        args <- args[intersect(names(as.list(base::args(estimate.lvm))),names(args))]
        m1 <- do.call(estimate, args)
    }
    m2$attributes$nonlinear <- nl
    twostage(object=m1,model2=m2,data=data,predict.fun=nl[[1]]$pred,...)
}

##' @export
twostage.twostage.lvm <- function(object,...) estimate.twostage.lvm(object,...)

##' @export
twostage.lvm <- function(object,model2,data=NULL, ...) {
    if (is.null(data)) {
        return(twostagelvm(object=object, model2=model2, model.object=TRUE, ...))
    }
    args <- c(list(x=object, data=data), list(...))
    args <- args[intersect(names(as.list(base::args(estimate.lvm))),names(args))]
    e1 <- do.call(estimate, args)
    twostage(object=e1,model2=model2,data=data, ...)
}

##' @export
twostage.lvm.mixture <- twostage.lvmfit

##' @export
twostage.estimate <- twostage.lvmfit

##' @export
print.twostage.lvm <- function(x,...) {
    print(cli::rule())
    cat("Model 1:\n")
    print(Model(x$call$object))
    print(cli::rule())
    cat("Model 2:\n")
    print(Model(x$call$model2))
}

##' @export
plot.twostage.lvm <- function(x,...) {
    model <- x$model
    m1 <- Model(x$call$object)
    m2 <- x$call$model2
    nl <- nonlinear(x)
    model <- regression(model, to=nl[[1]]$newx, from=nl[[1]]$x)
    elist <- edgeList(m1)
    vlist <- vars(m1)
    model <- beautify(model)
    for (i in seq_len(nrow(elist))) {
        e <- toformula(y=vlist[elist[i,2]],x=vlist[elist[i,1]])
        edgelabels(model, e, cex=0.7) <- 1
    }
    elist <- edgeList(m2)
    vlist <- vars(m2)
    for (i in seq_len(nrow(elist))) {
        e <- toformula(vlist[elist[i,2]],vlist[elist[i,1]])
        edgelabels(model, e, cex=0.7) <- 2
    }
    nodecolor(model, nl[[1]]$newx) <- "gray"
    for (xx in nl[[1]]$newx) {
        e <- toformula(y=names(nl)[1],x=xx)
        edgelabels(model,e,col="gray", cex=0.7, lty=1) <- 2
    }
    for (xx in nl[[1]]$newx) {
        e <- toformula(y=xx,x=nl[[1]]$x)
        edgelabels(model,e,col="gray", cex=0.7, lty=2) <- ""
    }
    plot(model, ...)
}

##' @export
predict.twostage.lvmfit <- function(object,
                            newdata,
                            variable=names(nonlinear(object)),
                            p=coef(object),
                            type=c("model2","latent"),
                            x=NULL,
                            ...) {
    if (missing(newdata)) stop("provide data for prediction")
    nl <- nonlinear(object)
    unam <- unique(unlist(lapply(nl,function(x) x$x)))
    if (is.vector(newdata) || all(unam%in%colnames(newdata)))
        type <- "latent"


    if (tolower(type[1])%ni%c("latent")) {
        p1 <- coef(object$estimate1)
        pred1 <- uhat(p1, data=newdata, object$estimate1, nlobj=nl)
        if (tolower(type[1])==c("model1"))
            return(pred1)
        newdata <- as.data.frame(newdata)
        newdata[,colnames(pred1)] <- pred1
        pred <- predict(object$estimate2,x=x,p=p,data=newdata,...)
        attr(pred,"p") <- NULL
        attr(pred,"e") <- NULL
        attr(pred,"cond.var") <- NULL
        return(pred)
    }

    ## Association between predicted latent variables and child nodes:
    if (is.numeric(variable)) {
        variable <- names(nonlinear(object))[variable]
    }
    nl <- nl[variable]
    res <- matrix(nrow=NROW(newdata),ncol=length(nl))
    colnames(res) <- names(nl)
    ##unam <- unique(unlist(lapply(nl, function(x) x$newx)))
    ##colnames(res) <- unam
    if (!is.null(x)) {
        newd <- newdata
        m0 <- Model(object$estimate2)
        p0 <- p
    }
    for (i in seq_along(nl)) {
        pnam <- c(variable,paste0(variable,"~",nl[[i]]$newx))
        pidx <- match(pnam,names(coef(object)))
        b <- p[pidx]
        F <- nl[[i]]$f
        if (is.vector(newdata)) {
            res[,i] <- F(b,newdata)
        } else {
            res[,i] <- F(b,newdata[,nl[[i]]$x])
            if (!is.null(x)) {
                newd[,nl[[i]]$newx] <- 0
                latent(m0,clear=TRUE) <- names(nl)[i]
                regression(m0, y=names(nl)[i], x=paste0(names(nl)[i],".offset")) <- 1
                p0[pidx] <- 0
            }
        }
    }

    if (!is.null(x) && !is.vector(newdata)) {
        latentres <- res
        colnames(latentres) <- paste0(colnames(res),".offset",collapse="")
        newd <- cbind(newd,latentres)
        res <- predict(m0,data=newd,p=p0,x=x,...)
        attr(res,"cond.var") <- NULL

    }
    return(res)
}

##' Cross-validated two-stage estimator
##'
##' Cross-validated two-stage estimator for non-linear SEM
##' @export
##' @param model1 model 1 (exposure measurement error model)
##' @param model2 model 2
##' @param data data.frame
##' @param control1 optimization parameters for model 1
##' @param control2 optimization parameters for model 1
##' @param knots.boundary boundary points for natural cubic spline basis
##' @param nmix number of mixture components
##' @param df spline degrees of freedom
##' @param fix automatically fix parameters for identification (TRUE)
##' @param std.err calculation of standard errors (TRUE)
##' @param nfolds Number of folds (cross-validation)
##' @param rep Number of repeats of cross-validation
##' @param messages print information (>0)
##' @param ... additional arguments to lower
##' @examples
##' \dontrun{ ## Reduce Ex.Timings
##' m1 <- lvm( x1+x2+x3 ~ u, latent= ~u)
##' m2 <- lvm( y ~ 1 )
##' m <- functional(merge(m1,m2), y ~ u, value=function(x) sin(x)+x)
##' distribution(m, ~u1) <- dist_uniform(-6, 6)
##' d <- sim(m,n=500,seed=1)
##' nonlinear(m2) <- y~u1
##' if (requireNamespace('mets', quietly=TRUE)) {
##'   set.seed(1)
##'   val <- twostageCV(m1, m2, data=d, std.err=FALSE, df=2:6, nmix=1:2,
##'                   nfolds=2)
##'   val
##' }
##' } ## Reduce Ex.Timings
twostageCV <- function(model1, model2, data, control1=list(trace=0), control2=list(trace=0),
                knots.boundary, nmix=1:4, df=1:9, fix=TRUE, std.err=TRUE,
                nfolds=5, rep=1, messages=0, ...) {

    F <- nonlinear(model2)
    if (length(F)==0) {
        stop("Specify at least one nonlinear association in 'model2' (using the 'nonlinear' method)")
    }
    form <- as.formula(paste(names(F)[1], "~", x=F[[1]]$x))

    op <- options(warn=-1)
    if (fix) {
        model1 <- baptize(fixsome(model1, param="relative"))
        intercept(model1, latent(model1)) <- NA
    }
    e1a <- estimate(model1, data=data, control=control1)
    if (missing(knots.boundary))
        knots.boundary <- range(predict(e1a,vars(e1a)))
    ## Starting values for mixture models
    plab <- parlabels(model1)
    pfree <- setdiff(coef(model1),plab)
    pfree.idx <- match(pfree,coef(model1)) ## Index of unlabeled parameters
    intpos <- setdiff(parpos(model1)$v,0)
    pfree.int <- intersect(pfree.idx,intpos) ## Free intercept parameters
    p0 <- coef(e1a)
    pint <- p0[setdiff(intpos,pfree.int)]
    startf <- function(n) {
        u0 <- seq(knots.boundary[1],knots.boundary[2],length.out=n); names(u0) <- paste0("p",seq_along(u0))
        c(pint,u0,p0[-intpos])
    }
    control1$start <- NULL
    ee <- list(e1a)

    nmix <- setdiff(nmix, 1)
    fit <- function(k, ...) {
        if (messages > 0) cat("Fitting mixture model with", k, "components\n")
        mixture(model1, k = k, data = data, control = c(control1, list(start = startf(k))))
    }
    future.args <- list(...)
    if (is.null(future.args$future.seed)) {
      future.args$future.seed <- TRUE
    }
    args <-  c(list(FUN = fit, c(as.list(nmix)), future.args))
    val <- do.call(future_lapply, args)
    ee <- c(ee, val)
    AIC1 <- unlist(lapply(ee,AIC))
    names(AIC1) <- c(1,nmix)
    ii <- which.min(AIC1)
    ## Exposure measurement model
    e1 <- ee[[ii]] ## Selected model by AIC

    MM <- list(nonlinear(model2, form, type="linear"))
    df <- setdiff(df, 1)
    Knots <- list()
    for (i in df) {
        knots <- seq(knots.boundary[1],knots.boundary[2],length.out=i+1)
        Knots  <- c(Knots, list(knots))
        MM <- c(MM,
               list(nonlinear(model2, form, type="spline", knots=knots)))
    }
    if (!inherits(e1, "lvm.mixture")) {
        f0 <- function(data) list(e0=estimate(model1,data=data,control=c(control1,list(start=coef(e1)))))
    } else {
        f0 <- function(data) list(e0=mixture(model1,data=data,k=e1$k,control=c(control1,list(start=coef(e1)))))
    }

    ff <- lapply(MM,
                 function(m) function(data,e0,...)  twostage(e0,m,data=data,std.derr=FALSE))
    a <- cv(ff, data=data, K=nfolds, rep=rep, shared=f0)
    M <- MM[[which.min(a$coef)]]
    e2 <- twostage(e1, M, data, control=control2, std.err=std.err)
    options(op)
    res <- list(AIC1=cbind(AIC1), model1=e1, model2=e2, cv=a$coef, knots=c(list(NA), Knots),
                nfolds=nfolds, rep=rep)
    structure(res, class="twostageCV")
}

##' @export
print.twostageCV <- function(x,...) {
    print(cli::rule())
    i1 <- which.min(x$AIC1)
    nmix <- rownames(x$AIC1)[i1]
    cat("Selected mixture model: ",nmix," component", ifelse(i1>1, "s",""),"\n", sep="")
    print(x$AIC1)
    print(cli::rule())
    i2 <- which.min(x$cv)
    splinedf <- unlist(lapply(x$knots,function(x) if (any(is.na(x))) return(1) else length(x)-1))
    cat("Selected spline model degrees of freedom: ", splinedf[i2] ,"\n", sep="")
    knots <- rbind(x$knots[[i2]])
    if (is.na(knots[1])) knots <- "none"
    cat("Knots:", paste(formatC(knots,...) , collapse=" "), "\n\n")
    rmse <- x$cv
    rownames(rmse) <- paste0("df:",splinedf)
    colnames(rmse) <- paste0("RMSE(nfolds=",x$nfolds,", rep=",x$rep,")")
    print(rmse)
    print(cli::rule())
    cat("\n")
    if (!inherits(x$model2, "lvmfit")) {
        print(x$model2$coefmat, quote=FALSE)
    } else {
        print(CoefMat(x$model2,...),quote=FALSE)
    }
}

##' @export
coef.twostageCV <- function(object,...) {
    coef(Model(object),...)
}

##' @export
vcov.twostageCV <- function(object,...) {
    vcov(Model(object),...)
}

##' @export
IC.twostageCV <- function(x,...) {
    IC(Model(x),...)
}

##' @export
Model.twostageCV <- function(x,...) {
    x$model2
}

##' @export
summary.twostageCV <- function(object,...) {
    res <- with(object, list(model1=summary(model1),
                                AIC1=AIC1, cv=cv, knots=knots,
                                model2=summary(model2)))
    structure(res, class="summary.twostageCV")
}

##' @export
print.summary.twostageCV <- function(x,...) {
    print.twostageCV(x, ...)
}

##' @export
predict.twostageCV <- function(object,... ) {
    predict(Model(object),...)
}

##'Appending \code{Surv} objects
##'
##'\code{rbind} method for \code{Surv} objects
##'
##' 
##'@param ... \code{Surv} objects
##'@return \code{Surv} object
##'@author Klaus K. Holst
##'@keywords utilities
##'@examples
##'
##'y <- yl <- yr <- rnorm(10)
##'yl[1:5] <- NA; yr[6:10] <- NA
##'S1 <- survival::Surv(yl,yr,type="interval2")
##'S2 <- survival::Surv(y,y>0,type="right")
##'S3 <- survival::Surv(y,y<0,type="left")
##'
##'rbind(S1,S1)
##'rbind(S2,S2)
##'rbind(S3,S3)
##'
##' @export
rbind.Surv <- function(...) 
{
  dots <- list(...)
  type <- attributes(dots[[1]])$type
  ncol <- dim(dots[[1]])[2]
  nrow <- unlist(lapply(dots,nrow))
  cnrow <- c(0,cumsum(nrow))
  M <- matrix(ncol=ncol,nrow=sum(nrow))
  for (i in 1:length(dots)) {
    M[(cnrow[i]+1):cnrow[i+1],] <- dots[[i]]
  }
  x <- c(); for (i in 1:ncol(M)) x <- c(x,list(M[,i]))
  x <- c(x,list(type=type))
  do.call(survival::Surv, x)
} 

##' @export
`heavytail` <- function(x,...) UseMethod("heavytail")
##' @export
"heavytail<-" <- function(x,...,value) UseMethod("heavytail<-")

##' @export
"heavytail<-.lvm" <- function(x,...,value) {
    if (inherits(value,"formula")) {
        return(heavytail(x,all.vars(value),...))
    }
    heavytail(x, value, ...)
}

##' @export
`heavytail.lvm` <-
    function(x,var=NULL,df=1,...) {
        if (is.null(var)) {
            htidx <- x$attributes$heavytail
            if (length(htidx)>0 && any(htidx!=0)) {
                res <- htidx[htidx>0]
                attributes(res)$couple <- unlist(x$attributes$heavytail.couple)[htidx>0]
                return(res)
            }
            return(NULL)
        }
        couples <- attributes(heavytail(x))$couple
        newval <- 1
        if (length(couples)>0) newval <- max(couples)+1
        x$attributes$heavytail.couple[var] <- newval
        x$attributes$heavytail[var] <- df
        return(x)
    }

heavytail_init_hook <- function(x,...) {
    x$attributes$heavytail <- list()
    x$attributes$heavytail.couple <- list()
    return(x)
}

heavytail_sim_hook <- function(x,data,...) {
    n <- nrow(data)
    hvar <- heavytail(x)
    if (length(hvar)==0) return(data)
    couples <- unique(attributes(hvar)$couple)
    h.type <- list()
    for (j in couples)
        h.type <- c(h.type, list( hvar[(which(attributes(hvar)$couple==j))]))
    for (i in seq_along(couples)) {
        df <- hvar[[i]][1]
        Z <- rchisq(n,df=df)/df
        for (v in names(h.type[[i]])) {
            data[,v] <- data[,v]/sqrt(Z)
        }
    }
    return(data)
}

toTheta <- function(mu,Sigma,p) {
  theta <- c(as.vector(t(mu)), as.vector(t(Sigma)), p[-nrow(mu)])
  return(theta)
}
toPar <- function(theta, D, k) {
  mus <- Sigmas <- c()
  for (j in 1:k) {
    muj.idx <- (1+((j-1)*D)):(j*D)
    mus <- rbind(mus, theta[muj.idx])
    Sigmaj.start <- k*D+1 + ((j-1)*D^2)
    Sigmaj.idx <- Sigmaj.start + 1:D^2-1
    Sigmas <- rbind(Sigmas, theta[Sigmaj.idx])
  };  ps <- tail(theta,k-1); ps <- c(ps,1-sum(ps))
  return(list(mu=mus, Sigma=Sigmas, p=ps))
}
getMeanVar <- function(object,k,iter,...) {
  if (missing(iter))
    pp <- with(object,toPar(pars,D,k))
  else
    pp <- with(object,toPar(thetas[iter,],D,k))
  res <- list()
  for (i in 1:object$k) {
    mu <- pp$mu[i,]
    V <- matrix(pp$Sigma[i,],ncol=object$D);
    res <- c(res, list(list(mean=mu, var=V)))
  }
  if (missing(k))
    return(res)
  else
    return(res[[k]])
}



#' Estimate mixture latent variable model
#'
#' Estimate mixture latent variable model
#'
#' Estimate parameters in a mixture of latent variable models via the EM
#' algorithm.
#'
#' @param data \code{data.frame}
#' @param k Number of mixture components
#' @param theta Optional starting values
#' @param steps Maximum number of iterations
#' @param tol Convergence tolerance of EM algorithm
#' @param lambda Regularisation parameter. Added to diagonal of covariance matrix (to avoid
#' singularities)
#' @param mu Initial centres (if unspecified random centres will be chosen)
#' @param silent Turn on/off output messages
#' @param extra Extra debug information
#' @param n.start Number of restarts
#' @param init Function to choose initial centres
#' @param ... Additional arguments parsed to lower-level functions
#' @return A \code{mixture} object
#' @author Klaus K. Holst
#' @seealso \code{mixture}
#' @keywords models regression
#' @examples
#'
#' data(faithful)
#' set.seed(1)
#' M1 <- mvnmix(faithful[,"waiting",drop=FALSE],k=2)
#' M2 <- mvnmix(faithful,k=2)
#' if (interactive()) {
#'     par(mfrow=c(2,1))
#'     plot(M1,col=c("orange","blue"),ylim=c(0,0.05))
#'     plot(M2,col=c("orange","blue"))
#' }
#'
#' @export mvnmix
mvnmix <- function(data, k=2, theta, steps=500,
            tol=1e-16, lambda=0,
            mu=NULL,
            silent=TRUE, extra=FALSE,
            n.start=1,
            init="kmpp", ...
            )  {

    if (k<2) stop("Only one cluster")
    ## theta = (mu1, ..., muk, Sigma1, ..., Sigmak, p1, ..., p[k-1])
    if (is.vector(data)) data <- matrix(data,ncol=1)
    if (is.data.frame(data)) data <- as.matrix(data)
    i <- 0
    E <- tol
    D <- ncol(data)
    yunique <- unique(data)
    if (n.start>1) extra <- FALSE

    logllmax <- -Inf
    for (ii in seq(n.start)) {
      if (ii>1) mu <- NULL
      if (missing(theta)) {
          mus <- c()
          if (!is.null(mu)) {
              mus <- mu
          } else {
              if (!exists(init)) { ## Random select centres
                  idx <- sample(NROW(data),k)
              } else {
                  idx <- do.call(init, list(data, k))
              }
              mus <- unlist(lapply(idx, function(i) cbind(data)[i,,drop=TRUE]))
          }
          Sigmas <- rep(as.vector(cov(data)),k)
          ps <- rep(1/k,k-1)
          theta <- c(mus,Sigmas,ps)
      }

      theta0 <- theta
      if (!silent)
          cat(i,":\t", paste(formatC(theta0),collapse=" "),"\n")
      thetas <- members <- c()
      while ((i<steps) & (E>=tol)) {
          if (extra)
              thetas <- rbind(thetas, theta)
          pp <- toPar(theta,D,k)
          mus <- pp$mu; Sigmas <- pp$Sigma; ps <- pp$p
          ## E(xpectation step)
          lphis <- c()
          for (j in 1:k) {
              C <- matrix(Sigmas[j,],ncol=D); diag(C) <- diag(C)+lambda ## Assure C is not singular
              lphis <- cbind(lphis, lava::dmvn0(data,mus[j,],C,log=TRUE))
          }
          gammas <- c()
          ## denom <- t(ps%*%t(phis))
          for (j in 1:k) {
              gammas <- cbind(gammas, log(ps[j]) + lphis[,j])
          }
          ## llmax <- apply(gammas,1,max)
          ## for (j in 1:k) {
          ##     gammas[,j] <- gammas[,j]-llmax
          ## }
          gammas <- exp(gammas) #
          denom <- rowSums(gammas)
          for (j in 1:k) gammas[,j] <- gammas[,j]/denom # Posterior
          sqrtgammas <- sqrt(gammas)
          ## M(aximization step)
          mus.new <- c()
          Sigmas.new <- c()
          for (j in 1:k) {
              mu.new <- colSums(gammas[,j]*data)/sum(gammas[,j])
              mus.new <- rbind(mus.new, mu.new)
              wcy <- sqrtgammas[,j]*t(t(data)-mus.new[j,])
              Sigma.new <- t(wcy)%*%wcy/sum(gammas[,j])
              Sigmas.new <- rbind(Sigmas.new, as.vector(Sigma.new))
          }; ps.new <- colMeans(gammas)
          theta.old <- theta
          if (extra)
              members <- cbind(members,
                               apply(gammas,1,function(x) order(x,decreasing=TRUE)[1]))
          theta <- toTheta(mus.new,Sigmas.new,ps.new)
          E <- sum((theta-theta.old)^2)
          i <- i+1
          iter <- i
          if (!silent)
              cat(i,":\t", paste(formatC(theta),collapse=" "),
                  ",\t\te=",formatC(E), "\n",sep="")
      }
      if (n.start>1) {
          logll <- sum(log(denom))
          if (logll>logllmax) {
              logllmax <- logll
              theta.keep <- theta
              gammas.keep <- gammas
              E.keep <- E
          }
      }
    }
    if (n.start>1) {
        theta <- theta.keep
        gammas <- gammas.keep
        E <- E.keep
    }

        myvars <- colnames(data)
    if (is.null(myvars)) myvars <- colnames(data) <- paste("y",1:NCOL(data),sep="")
    data <- as.data.frame(data)
    m <- lvm(myvars,silent=TRUE); m <- covariance(m,myvars,pairwise=TRUE)
    models <- datas <- c()
    for (i in 1:k) {
        models <- c(models, list(m))
        datas <- c(datas, list(data))
    }

    membership <- apply(gammas,1,function(x) order(x,decreasing=TRUE)[1])
    res <- list(pars=theta, thetas=thetas , gammas=gammas, member=membership,
                members=members, k=k, D=D, data=data, E=E,
                prob=rbind(colMeans(gammas)),
              iter=iter,
              models=models,
              multigroup=multigroup(models,datas)
              )
    class(res) <- c("mvn.mixture","lvm.mixture")

    parpos <- c()
    npar1 <- D+D*(D-1)/2
    for (i in 1:k)
        parpos <- c(parpos, list(c(seq_len(D)+(i-1)*D, k*D + seq_len(npar1)+
                                                       (i-1)*(npar1))))

    theta <- c(unlist(lapply(getMeanVar(res),function(x) x$mean)),
             unlist(lapply(getMeanVar(res),function(x) c(diag(x$var),unlist(x$var[upper.tri(x$var)])))))
    res$theta <- rbind(theta)
    res$parpos <- parpos
    res$opt <- list(estimate=theta)
    if (requireNamespace('mets', quietly=TRUE))
        res$vcov <- solve(information(res,type="E"))
    return(res)
}


##' @export
print.mvn.mixture <- function(x,...) {
  par <- toPar(x$pars,x$D,x$k)
  space <- paste(rep(" ",12),collapse="")
  for (i in 1:x$k) {
    cat("Cluster ",i," (p=",formatC(par$p[i]),"):\n",sep="")
    cat(rep("-",50),"\n",sep="")
    cat("\tcenter = \n ",space,paste(formatC(par$mu[i,]),collapse=" "),sep="")
    cat("\n\tvariance = \t");
    V <- matrix(formatC(par$Sigma[i,],flag=" "),ncol=x$D);
    colnames(V) <- rep("",x$D); rownames(V) <- rep(space,x$D)
    print(V, quote=FALSE)
    cat("\n")
  }
  invisible(par)
}

##' @export
plot.mvn.mixture <- function(x, label=2,iter,col,alpha=0.5,nonpar=TRUE,...) {
  opts <- list(...)
  ##  cols <- opts$col; if(is.null(cols)) cols <- 1:gmfit$k
  if (missing(col)) col <- 1:x$k
  lwd <- opts$lwd; if (is.null(lwd)) lwd <- 2
  cex <- opts$cex; if(is.null(cex)) cex <- 0.9
  y <- as.matrix(x$data)
  if (is.vector(y)) y <- matrix(y,ncol=1)
  pp <- getMeanVar(x,iter=iter)
  D <- ncol(y)
  pi <- colSums(x$gammas)/nrow(x$gammas)

  if (D==1) {
    if (nonpar)
      plot(density(as.vector(y)), main="", ...)
    else
      plot(density(y), main="", type="n", col="lightgray", ...)
    if (!is.null(label)) {
      for (i in 1:x$k) {
        rug(y[x$member==i], col=col[i])
      }
    }
    else
      rug(y)
    cc <- par("usr")
    {
      mycurve <- function(xx) {
        a <- 0;
        for (i in 1:(x$k))
          a <- a+pi[i]*dnorm(xx,pp[[i]]$mean,sqrt(pp[[i]]$var[1]))
        a
      }
      curve(mycurve, from=cc[1], to=cc[2], add=TRUE, lwd=lwd,...)
    }
  }
  if (D==2) {
      if (!requireNamespace("ellipse")) stop("ellipse required")
    plot(y, type="n", ...)

    for (i in 1:x$k) {
      C1 <- with(pp[[i]], ellipse::ellipse(var, centre=mean))
      lines(C1, col=col[i], lwd=lwd)
    }

    if (!is.null(label)) {
      for (i in 1:x$k) {
          if (label==1 | missing(iter)) {
          pot <- y[which(x$member==i),]
        }
        else {
          pot <- y[which(x$members[,iter]==i),]
        }
        points(pot, cex=cex, pch=16, col=Col(col[i],alpha))
      }
    }
    else
      points(y, cex=cex)
  }
  if (D==3) {
    if (!requireNamespace("rgl")) stop("rgl required")
    rgl::plot3d(y, type="n", box=FALSE)
    for (i in 1:x$k) {
        pot <- y[which(x$member==i),]
        rgl::plot3d(pot, type="s", radius=0.1, col=col[i], add=TRUE)
        ee <- rgl::ellipse3d(pp[[i]]$var,centre=pp[[i]]$mean)
        rgl::plot3d(ee, col=col[i], alpha=alpha, add = TRUE)
    }
  }
}

##' @export
sim.mvn.mixture <- function(x,n,...) {
    pars <- getMeanVar(x)
    K <- length(pars)
    p <- tail(coef(x),K-1); p <- c(p,1-sum(p))
    ng <- as.vector(rmultinom(1,n,p))
    res <- c()
    for (i in seq(K)) {
        res <- rbind(res,
                     mets::rmvn(ng[i],pars[[i]]$mean,pars[[i]]$var))
    }
    return(res)
}

##' Performs Likelihood-ratio, Wald and score tests
##' @title Statistical tests
##' @aliases compare test_wald
##' @export
##' @param object \code{lvmfit}-object
##' @param \dots Additional arguments to low-level functions
##' @return Matrix of test-statistics and p-values
##' @author Klaus K. Holst
##' @seealso \code{\link{modelsearch}}, \code{\link{equivalence}}
##' @keywords htest
##' @examples
##' m <- lvm();
##' regression(m) <- c(y1,y2,y3) ~ eta; latent(m) <- ~eta
##' regression(m) <- eta ~ x
##' m2 <- regression(m, c(y3,eta) ~ x)
##' set.seed(1)
##' d <- sim(m,1000)
##' e <- estimate(m,d)
##' e2 <- estimate(m2,d)
##'
##' compare(e)
##'
##' compare(e,e2) ## LRT, H0: y3<-x=0
##' compare(e,scoretest=y3~x) ## Score-test, H0: y3~x=0
##' compare(e2,par=c("y3~x")) ## Wald-test, H0: y3~x=0
##'
##' B <- diag(2); colnames(B) <- c("y2~eta","y3~eta")
##' compare(e2,contrast=B,null=c(1,1))
##'
##' B <- rep(0,length(coef(e2))); B[1:3] <- 1
##' compare(e2,contrast=B)
##'
##' compare(e,scoretest=list(y3~x,y2~x))
compare <- function(object,...) UseMethod("compare")

##' @export
compare.default <- function(object,...,
                            par,contrast,null,
                            scoretest,Sigma,level=.95,df=NULL) {
    if (!missing(par) || (!missing(contrast) && is.character(contrast))) {
        if (!missing(contrast) && is.character(contrast)) par <- contrast
        contrast <- rep(0,length(coef(object)))
        myidx <- parpos(Model(object),p=par)
        contrast[myidx] <- 1
        contrast <- diag(contrast,nrow=length(contrast))[which(contrast!=0),,drop=FALSE]
        if (!missing(null) && length(null)>1) null <- null[attributes(myidx)$ord]
    }
    ## Wald test
    if (!missing(contrast)) {
        B <- contrast
        p <- coef(object)
        pname <- names(p)
        B <- rbind(B);
        colnames(B) <- if (is.vector(contrast)) names(contrast) else colnames(contrast)
        if (missing(Sigma)) {
            Sigma <- vcov(object)
        }
        if (ncol(B)<length(p)) {
            nn <- colnames(B)
            myidx <- parpos(Model(object),p=nn)
            B0 <- matrix(0,nrow=nrow(B),ncol=length(coef(object)))
            B0[,myidx] <- B[,attributes(myidx)$ord]
            B <- B0
        }
        if (missing(null)) null <- rep(0,nrow(B))
        if (length(null)==1) null <- rep(null,nrow(B))
        Bp <- B%*%p
        V <- B%*%Sigma%*%t(B)
        ct <- cbind(Bp,diag(V)^.5)
        p <- 1-(1-level)/2
        qp <- if(!is.null(df)) qt(p,df=df) else qnorm(p)
        ct <- cbind(ct,ct[,1] + qp*cbind(-1,1)%x%ct[,2])
        colnames(ct) <- c("Estimate","Std.Err",paste0(c(1-p,p)*100,"%"))
        rownames(ct) <- rep("",nrow(ct))
        Q <- t(Bp-null)%*%Inverse(V)%*%(Bp-null)
        if (any(is.nan(V)) ||any(is.na(V))) {
          df <- 0
        } else {
          df <- tryCatch(qr(V)$rank, error=function(...) 0)
        }
        names(df) <- "df"
        attributes(Q) <- NULL; names(Q) <- "chisq";
        pQ <- ifelse(df==0,NA,pchisq(Q,df,lower.tail=FALSE))

        method = "- Wald test -";
        cnames <- c()
        if (!is.null(pname)) {
            msg <- c()
            for (i in seq_len(nrow(B))) {
                Bidx <- which(B[i,]!=0)
                Bval <- abs(B[i,Bidx]); Bval[Bval==1] <- ""
                sgn  <- rep(" + ",length(Bval)); sgn[sign(B[i,Bidx])==-1] <- " - ";
                if (sgn[1] == " + ") {
                  sgn[1] <- ""
                } else {
                    sgn[1] <- "-"
                }
                cnames <- c(cnames,
                            paste0(sgn,Bval,
                                   paste0("[",pname[Bidx],"]"),collapse=""))
                msg <- c(msg,paste0(cnames[i]," = ",null[i]))
            }
            method <- c(method,"","Null Hypothesis:",msg)
            ##method <- c(method,"","Observed:",paste(formatC(as.vector(Bp)),collapse=" "))
        }
        rownames(ct) <- cnames
        res <- list(##data.name=hypothesis,
            statistic = Q, parameter = df,
            p.value=pQ, method = method, estimate=ct, vcov=V, coef=ct[,1],
            null=null, cnames=cnames
        )
        class(res) <- "htest"
        attributes(res)$B <- B
        return(res)
    }

    ## Score test
    if (!missing(scoretest)) {
        altmodel <- Model(object)
        if (inherits(scoretest,"formula")) scoretest <- list(scoretest)
        for (i in scoretest) {
            regression(altmodel) <- i
        }
        p0 <- numeric(length(coef(altmodel)))
        idx <-  match(coef(Model(object)),coef(altmodel))
        p0[idx] <- coef(object)
        Sc2 <- score(altmodel,p=p0,data=model.frame(object),weigth=Weights(altmodel),
                     estimator=object$estimator,...)
        I <- information(altmodel,p=p0,n=object$data$n,
                         data=model.frame(object),weigth=Weights(object),
                         estimator=object$estimator,...
                         )
        iI <- try(solve(I), silent=TRUE)
        Q <- ifelse (inherits(iI, "try-error"), NA, ## Score test
                     ## rbind(Sc)%*%iI%*%cbind(Sc)
        (Sc2)%*%iI%*%t(Sc2)
        )
        attributes(Q) <- NULL; names(Q) <- "chisq"
        df <- length(p0)-length(coef(object)); names(df) <- "df"
        pQ <- ifelse(df==0,NA,pchisq(Q,df,lower.tail=FALSE))
        res <- list(data.name=as.character(scoretest),
                    statistic = Q, parameter = df,
                    p.value=pQ, method = "- Score test -")
        class(res) <- "htest"
        return(res)
    }

    ## Likelihood ratio test
    objects <- list(object,...)
    if (length(objects)<2) {
        if (!(inherits(object,"lvmfit"))) {
            cc <- rbind(logLik(object),AIC(object))
            rownames(cc) <- c("logLik","AIC")
            colnames(cc) <-  ""
            return(cc)
        }
        L0 <- logLik(object)
        L1 <- satmodel(object,logLik=TRUE)
        df <- attributes(L1)$df-attributes(L0)$df; names(df) <- "df"
        Q <- abs(2*(L0-L1));
        attributes(Q) <- NULL; names(Q) <- "chisq";
        pQ <- ifelse(df==0,NA,pchisq(Q,df,lower.tail=FALSE))

        values <- c(L0,L1); names(values) <- c("log likelihood (model)", "log likelihood (saturated model)")
        res <- list(statistic = Q, parameter = df,
                    p.value=pQ, method = "- Likelihood ratio test -",
                    estimate = values)
        class(res) <- "htest"
        return(res)
    }
    if (length(objects)==2)
        return(comparepair(objects[[1]],objects[[2]]))
    res <- list()
    for (i in seq_len(length(objects)-1)) {
        res <- c(res, list(comparepair(objects[[i]],objects[[i+1]])))
    }
    return(res)
}


comparepair <- function(x1,x2) {
    l1 <- do.call("logLik",list(x1),envir=parent.frame(2))
    l2 <- do.call("logLik",list(x2),envir=parent.frame(2))
    df1 <- attributes(l1)$df;  df2 <- attributes(l2)$df;
    if (is.null(df1)) {
        df1 <- length(do.call("coef",list(x1),envir=parent.frame(2)))
        df2 <- length(do.call("coef",list(x2),envir=parent.frame(2)))
    }
    Q <- abs(2*(l1-l2))
    names(Q) <- "chisq"
    df <- abs(df1-df2); names(df) <- "df"
    p <- pchisq(Q,df=df,lower.tail=FALSE)
    values <- c(l1,l2); names(values) <- c("log likelihood (model 1)", "log likelihood (model 2)")

    res <- list(statistic = Q, parameter = df,
                p.value= p, method = "- Likelihood ratio test -",
                estimate = values)
    class(res) <- "htest"
    return(res)
}

##' Add regression association to latent variable model
##'
##' Define regression association between variables in a \code{lvm}-object and
##' define linear constraints between model equations.
##'
##'
##' The \code{regression} function is used to specify linear associations
##' between variables of a latent variable model, and offers formula syntax
##' resembling the model specification of e.g. \code{lm}.
##'
##' For instance, to add the following linear regression model, to the
##' \code{lvm}-object, \code{m}:
##' \deqn{ E(Y|X_1,X_2) = \beta_1 X_1 + \beta_2 X_2}
##' We can write
##'
##' \code{regression(m) <- y ~ x1 + x2}
##'
##' Multivariate models can be specified by successive calls with
##' \code{regression}, but multivariate formulas are also supported, e.g.
##'
##' \code{regression(m) <- c(y1,y2) ~ x1 + x2}
##'
##' defines
##' \deqn{ E(Y_i|X_1,X_2) = \beta_{1i} X_1 + \beta_{2i} X_2 }
##'
##' The special function, \code{f}, can be used in the model specification to
##' specify linear constraints. E.g. to fix \eqn{\beta_1=\beta_2}
##' , we could write
##'
##' \code{regression(m) <- y ~ f(x1,beta) + f(x2,beta)}
##'
##' The second argument of \code{f} can also be a number (e.g. defining an
##' offset) or be set to \code{NA} in order to clear any previously defined
##' linear constraints.
##'
##' Alternatively, a more straight forward notation can be used:
##'
##' \code{regression(m) <- y ~ beta*x1 + beta*x2}
##'
##' All the parameter values of the linear constraints can be given as the right
##' handside expression of the assigment function \code{regression<-} if the
##' first (and possibly second) argument is defined as well. E.g:
##'
##' \code{regression(m,y1~x1+x2) <- list("a1","b1")}
##'
##' defines \eqn{E(Y_1|X_1,X_2) = a1 X_1 + b1 X_2}. The rhs argument can be a
##' mixture of character and numeric values (and NA's to remove constraints).
##'
##' The function \code{regression} (called without additional arguments) can be
##' used to inspect the linear constraints of a \code{lvm}-object.
##'
##' @aliases regression regression<- regression<-.lvm regression.lvm
##' @param object \code{lvm}-object.
##' @param value A formula specifying the linear constraints or if
##' \code{to=NULL} a \code{list} of parameter values.
##' @param to Character vector of outcome(s) or formula object.
##' @param from Character vector of predictor(s).
##' @param fn Real function defining the functional form of predictors (for
##' simulation only).
##' @param messages Controls which messages are turned on/off (0: all off)
##' @param additive If FALSE and predictor is categorical a non-additive effect is assumed
##' @param y Alias for 'to'
##' @param x Alias for 'from'
##' @param \dots Additional arguments to be passed to the low level functions
##' @usage
##' \method{regression}{lvm}(object = lvm(), to, from, fn = NA,
##' messages = lava.options()$messages, additive=TRUE, y, x, value, ...)
##' \method{regression}{lvm}(object, to=NULL, from=NULL, ...) <- value
##' @return A \code{lvm}-object
##' @note Variables will be added to the model if not already present.
##' @author Klaus K. Holst
##' @seealso \code{\link{intercept<-}}, \code{\link{covariance<-}},
##' \code{\link{constrain<-}}, \code{\link{parameter<-}},
##' \code{\link{latent<-}}, \code{\link{cancel<-}}, \code{\link{kill<-}}
##' @keywords models regression
##' @examples
##'
##' m <- lvm() ## Initialize empty lvm-object
##' ## E(y1|z,v) = beta1*z + beta2*v
##' regression(m) <- y1 ~ z + v
##' ## E(y2|x,z,v) = beta*x + beta*z + 2*v + beta3*u
##' regression(m) <- y2 ~ f(x,beta) + f(z,beta)  + f(v,2) + u
##' ## Clear restriction on association between y and
##' ## fix slope coefficient of u to beta
##' regression(m, y2 ~ v+u) <- list(NA,"beta")
##'
##' regression(m) ## Examine current linear parameter constraints
##'
##' ## A multivariate model, E(yi|x1,x2) = beta[1i]*x1 + beta[2i]*x2:
##' m2 <- lvm(c(y1,y2) ~ x1+x2)
##'
##' @export
"regression<-" <- function(object,...,value) UseMethod("regression<-")

##' @export
regression.formula <- function(object,...) regression(lvm(),object,...)

##' @export
"regression<-.lvm" <- function(object, to=NULL, from=NULL, ..., value) {
    dots <- list(...)
    if (length(dots$additive)>0 && !dots$additive && !inherits(value,"formula")) {
        regression(object,beta=value,...) <- to
        return(object)
    }
    if (!is.null(to) || !is.null(dots$y)) {
        regfix(object, to=to, from=from, ...) <- value
        return(object)
    } else  {
        if (is.list(value)) {
            for (v in value) {
                regression(object,...) <- v
            }
            return(object)
        }

        if (inherits(value,"formula")) {
            fff <- procformula(object,value,...)
            object <- fff$object
            lhs <- fff$lhs
            xs <- fff$xs
            ys <- fff$ys
            res <- fff$res
            X <- fff$X

            
        if (fff$iscovar) {
            ## return(covariance(object,var1=decomp.specials(lhs[[1]]),var2=X))
            covariance(object) <- toformula(decomp.specials(lhs[[1]]),X)
            return(object)
        }
        if (!is.null(lhs) && nchar(lhs[[1]])>2 && substr(lhs[[1]],1,2)=="v(") {
            v <- update(value,paste(decomp.specials(lhs),"~."))
            covariance(object,...) <- v
            return(object)
        }

        if (length(lhs)==0) {
            index(object) <- reindex(object)
            return(object)
        }

        for (i in seq_len(length(ys))) {
        y <- ys[i]
        for (j in seq_len(length(xs))) {
          if (length(res[[j]])>1) {
            regfix(object, to=y[1], from=xs[j],...) <- res[[j]][2]
          } else {
            object <- regression(object,to=y[1],from=xs[j],...)
          }
        }
      }
      object$parpos <- NULL
      return(object)
    }
    if (!is.list(value) | length(value)>2) stop("Value should contain names of outcome (to) and predictors (from)")
    if (all(c("to","from")%in%names(value))) {

      xval <- value$x; yval <- value$y
    } else {
      yval <- value[[1]]; xval <- value[[2]]
    }
    regression(object, to=yval, from=xval,...)
  }
}

##' @export
`regression` <-
  function(object,to,from,...) UseMethod("regression")

##' @export
`regression.lvm` <-
    function(object=lvm(),to,from,fn=NA,messages=lava.options()$messages,
             additive=TRUE, y,x,value,...) {
        if (!missing(y)) {
            if (inherits(y,"formula")) y <- all.vars(y)
            to <- y
        }
        if (!missing(x)) {
            if (inherits(x,"formula")) x <- all.vars(x)
            from <- x
        }
        if (!additive) {
            if (!inherits(to,"formula")) to <- toformula(to,from)
            x <- attributes(getoutcome(to))$x
            K <- object$attributes$nordinal[x]
            if (is.null(K) || is.na(K)) {
                K <- list(...)$K
                if (is.null(K)) stop("Supply number of categories, K (or use method 'categorical' before calling 'regression').")
                object <- categorical(object,x,...)
            }
            dots <- list(...);
            dots$K <- K
            dots$x <- object
            dots$formula <- to
            dots$regr.only <- TRUE
            object <- do.call("categorical",dots)
            return(object)
        }


        if (missing(to)) {
            return(regfix(object))
        }
        if ((missing(from) || is.null(from)) &
            is.character(to)) {
          to <- as.formula(paste(to, "~1"))
        }
        if (inherits(to,"formula")) {
          if (!missing(value)) {
                regression(object,to,messages=messages,...) <- value
            } else {
                regression(object,messages=messages,...) <- to
            }
            object$parpos <- NULL
            return(object)
        }
        if (!missing(value) && is.character(to) && !is.null(from)) {
            regfix(object, to=to, from=from, ...) <- value
            return(object)
        }
        if (is.list(to)) {
            for (t in to)
                regression(object,messages=messages,...) <- t
            object$parpos <- NULL
            return(object)
        }

        sx <- strsplit(from,"@")
        xx <- sapply(sx, FUN=function(i) i[1])
        ps <- sapply(sx, FUN=function(i) i[2])
        sx <- strsplit(xx,"$",fixed=TRUE)
        xs <- sapply(sx, FUN=function(i) i[1])
        fix <- char2num(sapply(sx, FUN=function(i) i[2]))
        allv <- index(object)$vars

        object <- addvar(object, c(to,xs), messages=messages, reindex=FALSE)

        for (i in to)
            for (j in xs) {
                object$M[j,i] <- 1
                if (!is.na(fn))
                    functional(object,j,i) <- fn
            }

        if (lava.options()$exogenous) {
            newexo <- setdiff(xs,c(to,allv))
            exo <- exogenous(object)
            if (length(newexo)>0)
                exo <- unique(c(exo,newexo))
            exogenous(object) <- setdiff(exo,to)
        }

        if (lava.options()$debug) {
            print(object$fix)
        }
        object$fix[xs,to] <- fix
        object$par[xs,to] <- ps
        object$parpos <- NULL

        index(object) <- reindex(object)
        return(object)
    }

##' @export
"nonlinear<-" <- function(object,...,value) UseMethod("nonlinear<-")

##' @export
"nonlinear" <- function(object,...) UseMethod("nonlinear")

naturalcubicspline <- function(x, knots=stats::median(x,na.rm=TRUE), boundary=range(x,na.rm=TRUE)) {
    ## C2 functions, piecewise cubic
    breaks <- c(boundary[1],knots,boundary[2])
    K <- length(breaks)
    g <- function(x,tau) (x-tau)^3*((x-tau)>0)
    gg <- matrix(0,nrow=length(x),ncol=K)
    for (i in seq(K)) {
        gg[,i] <- g(x,breaks[i])
    }
    B <- matrix(0,nrow=length(x),ncol=K-2)
    for (i in seq(K-2)) {
        B[,i] <- gg[,i] -
            (breaks[K]-breaks[i])/(breaks[K]-breaks[K-1])*gg[,K-1] +
            (breaks[K-1]-breaks[i])/(breaks[K]-breaks[K-1])*gg[,K]
    }
    cbind(x,B)
}

ncspred <- function(mu, var, knots=c(-5,0,5)) {
    breaks <- knots
    K <- length(breaks)

    v <- as.vector(var)
    k <- sqrt(v/(2*pi))
    g <- function(x,tau) {
        x0 <- (x-tau)
        x2 <- x0^2
        p0 <- 1-pnorm(-x0/sqrt(v)) # P(x>tau|...)
        k*(2*v + x2)*exp(-(x0/(sqrt(2*v)))^2) +
            x0*(x2+3*v)*p0
    }
    n <- NROW(mu)
    gg <- matrix(0,nrow=n,ncol=K)
    for (i in seq(K)) {
        gg[,i] <- g(mu,breaks[i])
    }
    B <- matrix(0,nrow=n,ncol=K-2)
    for (i in seq(K-2)) {
        B[,i] <- gg[,i] -
            (breaks[K]-breaks[i])/(breaks[K]-breaks[K-1])*gg[,K-1] +
            (breaks[K-1]-breaks[i])/(breaks[K]-breaks[K-1])*gg[,K]
    }
    cbind(mu,B)
}


##' @export
nonlinear.lvm <- function(object, to, from=NULL, type=c("quadratic"), knots=c(-5,0,5), names, ...) {
    if (missing(to)) {
        return(object$attributes$nonlinear)
    }
    if (inherits(to,"formula")) {
        yy <- decomp.specials(getoutcome(to))
        myvars <- all.vars(to)
        from <- setdiff(myvars,yy)
        to <- yy
    }
    if (length(to)>1) stop("Supply only one response variable")
    if (length(from)>1) stop("Supply only one explanatory variable")
    object <- cancel(object, c(from,to))
    variance(object) <- to
    f <- pred <- NULL

    if (tolower(type)[1]%in%c("ncs","spline","naturalspline","cubicspline","natural cubic spline")) {
        if (is.null(knots)) stop("Need cut-points ('knots')")
        if (length(knots)<3) {
            warning("Supply at least three knots (one interior and boundaries)")
            ## Fall-back to linear
            type <- "linear"
        }
        if (missing(names)) names <- paste0(from,"_",seq(length(knots)-1))
        f <- function(p,x) {
            B <- cbind(1,naturalcubicspline(x,knots=knots[-c(1,length(knots))],boundary=knots[c(1,length(knots))]))
            colnames(B) <- c("(Intercept)",names)
            as.vector(B%*%p)
        }
        pred <- function(mu,var,...) {
            B <- ncspred(mu,var,knots=knots)
            structure(B,dimnames=list(NULL,names))
        }
    }

    
    if (tolower(type)[1]=="linear") {
        if (missing(names)) names <- from
        f <- function(p,x) p[1] + p[2]*x
        pred <- function(mu,var,...) {            
            structure(cbind(mu[,1]),dimnames=list(NULL,names))
        }
    }
    
    if (tolower(type)[1]=="quadratic") {
        if (missing(names)) names <- paste0(from,"_",1:2)
        f <- function(p,x) p[1] + p[2]*x + p[3]*(x*x)
        pred <- function(mu,var,...) {
            structure(cbind(mu[,1],mu[,1]^2+var[1]),dimnames=list(NULL,names))
        }
    }

    if (tolower(type)[1]%in%c("piecewise","piecewise linear","linear")) {
        if (is.null(knots)) stop("Need cut-points ('knots')")
    }

    if (tolower(type)[1]%in%c("exp","exponential")) {
        if (missing(names)) names <- paste0(from,"_",1:2)
        f <- function(p,x) p[1] + p[2]*x + p[3]*exp(x)
        pred <- function(mu,var,...) {
            structure(cbind(mu[,1], exp(0.5*var[1] + mu[,1])),dimnames=list(NULL,names))
        }
    }

    if (tolower(type)[1]%in%c("exp0")) {
        if (missing(names)) names <- paste0(from,"_",1)
        f <- function(p,x) p[1] + p[2]*exp(x)
        pred <- function(mu,var,...) {
            structure(cbind(exp(0.5*var[1] + mu[,1])),dimnames=list(NULL,names))
        }
    }

    object$attributes$nonlinear[[to]] <- list(x=from, p=length(names)+1, newx=names, f=f, pred=pred, type=tolower(type[1]))
    return(object)
}

##' @export
nonlinear.lvmfit <- function(object, to, ...) {
    if (missing(to)) {
        return(Model(object)$attributes$nonlinear)
    }
    Model(object) <- nonlinear(Model(object),to=to,...)
    return(object)
}

##' @export
nonlinear.twostage.lvm <- function(object, ...) {
    return(object$nonlinear)
}

##' @export
nonlinear.lvmfit <- function(object, ...) {
    return(object$nonlinear)
}

##' @export
`nonlinear<-.lvm` <- function(object, ..., type="quadratic", value) {
    nonlinear(object,to=value,type=type,...)
}

##' Estimate method for GLM objects
##'
##' Applies [estimate.default] to obtain robust inference for a fitted
##' `glm` object.
##' @param x Fitted `glm` object
##' @param ... Additional arguments to [estimate.default]
##' @return Object of class `estimate` (see [estimate.default]).
##' @export
estimate.glm <- function(x, ...) {
  estimate.default(x, ...)
}

##' @export
IC.mlm <- function(x, ...) {
  cc <- coef(x)
  r <- residuals(x)
  X <- model.matrix(x)
  w <- weights(x)
  if (!is.null(w)) {
    r <- apply(r, 2, function(x) x * w)
  }
  q <- NCOL(cc)
  ics <- lapply(1:q, function(i) {
    apply(X, 2, function(x) x * r[, i])
  })
  res <- Reduce("cbind", ics)
  colnames(res) <- names(pars(x))
  return(res)
}

##' @export
pars.mlm <- function(x, ...) {
  cc <- coef(x)
  q <- NCOL(cc)
  nn <- unlist(lapply(
    1:q, function(i)
      paste0(colnames(cc)[i], ":", rownames(cc))
  ))
  coefs <- unlist(lapply(1:q, function(x) cc[, x, drop=TRUE]))
  names(coefs) <- nn
  coefs
}

##' @export
estimate.mlm <- function(x, ...) {
  estimate.default(x, coef=pars(x), ...)
}

##' @export
##' @export density.sim
density.sim <- function(x, ..., plot.type="single") {
    plot.sim(x, ..., scatter.plot=FALSE, plot.type=plot.type)
}

##' Plot method for simulation 'sim' objects
##'
##' Density and scatter plots
##' @examples
##' n <- 1000
##' val <- cbind(est1=rnorm(n,sd=1),est2=rnorm(n,sd=0.2),est3=rnorm(n,1,sd=0.5),
##'              sd1=runif(n,0.8,1.2),sd2=runif(n,0.1,0.3),sd3=runif(n,0.25,0.75))
##' 
##' plot.sim(val,estimate=c(1,2),true=c(0,0),se=c(4,5),equal=TRUE,scatter.plot=TRUE)
##' plot.sim(val,estimate=c(1,3),true=c(0,1),se=c(4,6),xlim=c(-3,3),
##' 	scatter.ylim=c(-3,3),scatter.plot=TRUE)
##' plot.sim(val,estimate=c(1,2),true=c(0,0),se=c(4,5),equal=TRUE,
##' 	plot.type="single",scatter.plot=TRUE)
##' plot.sim(val,estimate=c(1),se=c(4,5,6),plot.type="single",scatter.plot=TRUE)
##' plot.sim(val,estimate=c(1,2,3),equal=TRUE,scatter.plot=TRUE)
##' plot.sim(val,estimate=c(1,2,3),equal=TRUE,byrow=TRUE,scatter.plot=TRUE)
##' plot.sim(val,estimate=c(1,2,3),plot.type="single",scatter.plot=TRUE)
##' plot.sim(val,estimate=1,se=c(3,4,5),plot.type="single",scatter.plot=TRUE)
##' 
##' density.sim(val,estimate=c(1,2,3),density=c(0,10,10), lwd=2, angle=c(0,45,-45),cex.legend=1.3)
##' @aliases density.sim plot.sim
##' @export
##' @export plot.sim
##' @param x sim object
##' @param estimate columns with estimates
##' @param se columns with standard error estimates
##' @param true (optional) vector of true parameter values
##' @param names (optional) names of estimates
##' @param auto.layout Auto layout (default TRUE)
##' @param byrow Add new plots to layout by row
##' @param type plot type
##' @param ask if TRUE user is asked for input, before a new figure is drawn
##' @param col colour (for each estimate)
##' @param pch plot symbol
##' @param cex point size
##' @param lty line type
##' @param lwd line width
##' @param legend legend 
##' @param legendpos legend position
##' @param cex.legend size of legend text
##' @param plot.type 'single' or 'multiple' (default)
##' @param polygon if TRUE fill the density estimates with colour
##' @param density if non-zero add shading lines to polygon
##' @param angle shading lines angle of polygon
##' @param cex.axis Font size on axis
##' @param alpha Semi-transparent level (1: non-transparent, 0: full)
##' @param main Main title
##' @param cex.main Size of title font
##' @param equal Same x-axis and y-axis for all plots
##' @param delta Controls the amount of space around axis limits
##' @param ylim y-axis limits
##' @param xlim x-axis limits
##' @param ylab y axis label
##' @param xlab x axis label
##' @param rug if TRUE add rug representation of data to x-axis
##' @param rug.alpha rug semi-transparency level
##' @param line.col line colour (running mean, only for scatter plots)
##' @param line.lwd line width (running mean, only for scatter plots)
##' @param line.lty line type (running mean, only for scatter plots)
##' @param line.alpha line transparency 
##' @param scatter.ylab y label for density plots
##' @param scatter.ylim y-axis limits for density plots
##' @param scatter.xlim x-axis limits for density plots
##' @param scatter.alpha semi-transparency of scatter plot
##' @param scatter.col scatter plot colour
##' @param border border colour of density estimates
##' @param true.lty true parameter estimate line type
##' @param true.col true parameter colour
##' @param true.lwd true parameter line width
##' @param density.plot if TRUE add density plot
##' @param scatter.plot if TRUE add scatter plot 
##' @param running.mean if TRUE add running average estimate to scatter plot
##' @param ... additional arguments to lower level functions
plot.sim <- function(x,estimate,se=NULL,true=NULL,
             names=NULL,
             auto.layout=TRUE,
             byrow=FALSE,
             type="p",
             ask=grDevices::dev.interactive(),             
             col=c("gray60","orange","darkblue","seagreen","darkred"),
             pch=16,cex=0.5,lty=1,lwd=0.3,
             legend,
             legendpos="topleft",
             cex.legend=0.8,
             plot.type=c("multiple","single"),
             polygon=TRUE,
             density=NULL,
             angle=-45,
             cex.axis=0.8,
             alpha=0.2,
             main,
             cex.main=1,
             equal=FALSE,
             delta=1.15,
             ylim=NULL,
             xlim=NULL,
             ylab="",
             xlab="",
             rug=FALSE,
             rug.alpha=0.5,
             line.col=scatter.col,             
             line.lwd=1,
             line.lty=1,
             line.alpha=1,
             scatter.ylab="Estimate",
             scatter.ylim=NULL,
             scatter.xlim=NULL,
             scatter.alpha=0.5,
             scatter.col=col,
             border=col,
             true.lty=2,true.col="gray70",true.lwd=1.2,
             density.plot=TRUE,
             scatter.plot=FALSE,
             running.mean=scatter.plot,
             ...) {

    if (missing(estimate)) {
        estimate <- seq(ncol(x))
    }
    if (is.null(estimate)) {
        av <- apply(x[,drop=FALSE],2,function(z) cumsum(z)/seq(length(z)))
        graphics::matplot(x,type="p",pch=pch, cex=cex, col=col,...)
        graphics::matlines(av,type="l",col=col,lty=lty,...)
        if (!is.null(true)) abline(h=true,lty=true.lty,...)
        if (missing(legend)) legend <- colnames(x)
        if (!is.null(legend))
            graphics::legend(legendpos,legend=legend,bg="white",
                             col=scatter.col,lty=lty,pch=pch,...)
        return(invisible(NULL))
    }
    if (is.character(estimate)) {
        estimate <- match(estimate,colnames(x))
    }

    K <- length(estimate)
    est <- tru <- c()
    if (length(se)>0) {
        if (K==1 && !is.list(se))
            se <- list(se)
        else se <- as.list(se)
    } else {
        est <- estimate; tru <- true
    }

    for (i in seq_along(estimate)) {
        est <- c(est,list(rep(estimate[i],length(se[[i]]))))
        if (!is.null(true)) tru <- c(tru,list(rep(true[i],length(se[[i]]))))
    }

    if (length(se)>0) {
      for (i in seq_along(se)) {
            if (is.character(se[[i]])) se[[i]] <- match(se[[i]],colnames(x))
        }

    }
  ss <- summary.sim(x,estimate=unlist(est),se=unlist(se),true=unlist(tru),names=names)

    oldpar <- NULL
    on.exit({
        par(oldpar)
        return(invisible(ss))
    })

    single <- tolower(plot.type[1])=="single"

    nc <- (scatter.plot || running.mean) + density.plot
    nr <- min(6,K)
    if (single) nr <- 1
    
    if (auto.layout && (nc>1 || nr>1)) {
        oma.multi = c(2, 0, 2, 0)
        mar.multi = c(1.5, 4.1, 1, 1)
        opt <- list(mar=mar.multi, oma=oma.multi,
                      cex.axis=cex.axis,las=1,
                      ask=FALSE)
        if (byrow) {
            opt <- c(opt, list(mfrow=c(nr,nc)))
        } else {
            opt <- c(opt, list(mfcol=c(nc,nr)))
        }
        oldpar <- do.call(par, opt)
    }
  dys <- c()
  maxdy <- 0

  if (density.plot)
    for (i in seq(K)) {
            ii <- estimate[i]
            y <- as.vector(x[,ii,drop=TRUE])
            dy <- stats::density(y,...)
            dys <- c(dys,list(dy))
            maxdy <- max(maxdy,dy$y)
        }

    if (equal || single) {
        if (is.null(scatter.ylim)) {
            rg <- range(x[,estimate])
            rg <- rg+c(-1,1)*abs(diff(rg)*(delta-1))
            scatter.ylim <-  rep(list(rg),K)
        }
        if (density.plot) {
            if (is.null(ylim)) ylim <- rep(list(c(0,maxdy*delta)),K)
            if (is.null(xlim)) xlim <- scatter.ylim
        }
    }

    if (!is.null(ylim)) {
        if (!is.list(ylim)) ylim <- list(ylim)
        ylim <- rep(ylim,length.out=K)
    }
    ylab <- rep(ylab,length.out=K)
    if (!is.null(scatter.ylim)) {
        if (!is.list(scatter.ylim)) scatter.ylim <- list(scatter.ylim)
        scatter.ylim <- rep(scatter.ylim,length.out=K)
    }
    if (!is.null(xlim)) {
        if (!is.list(xlim)) xlim <- list(xlim)
        xlim <- rep(xlim,length.out=K)
    }
    if (missing(main)) {
        main <- NULL
        if (!missing(names)) main <- names
        else if (K>1 && !single) main <- colnames(ss)
    }
    if (!is.null(main)) main <- rep(main,length.out=K)
    if (missing(lty)) {
        lty <- rep(1,K)
        if (single || !polygon) {
            lty <- 1:20
        }
    }

    my.scatter.sim <- function(i,add=FALSE,colors,...) {
        ii <- estimate[i]
        if (!missing(colors)) {
            scatter.col <- line.col <- true.col <- colors[1]
        }
        y <- as.vector(x[,ii,drop=TRUE])
        args <- list(y,ylab=scatter.ylab[i],
                     col=Col(scatter.col[1], scatter.alpha),
                     cex=cex,pch=pch,type=type)
        if (!is.null(scatter.ylim)) args <- c(args,list(ylim=scatter.ylim[[i]]))
        if (scatter.plot) {
            if (!add) {
                do.call(graphics::plot,args)
            } else {
                do.call(graphics::points,args)
            }
        }
        if (running.mean) {
            lines(cumsum(y)/seq_along(y),col=line.col[1],lwd=line.lwd,lty=line.lty)
            if (!is.null(true))
                abline(h=true[i],lty=true.lty,col=true.col[1],lwd=true.lwd)
        }
    }

    my.density.sim <- function(i,add=FALSE,
                               colors,
                               alphas=alpha,
                               auto.legend=TRUE,
                               densities=NULL,
                               angles=angle,
                               ...) {
        ii <- estimate[i]
        y <- as.vector(x[,ii,drop=TRUE])
        if (!missing(colors)) {
            col <- border <- colors
            col <- true.col <- colors
        }
        if (density.plot) {
            dy <- stats::density(y)
            if (is.null(ylim)) {
                density.ylim0 <- c(0,max(dy$y)*delta)
            } else {
                density.ylim0 <- ylim[[i]]
            }
            if (is.null(xlim)) {
                density.xlim0 <- range(dy$x)
            } else {
                density.xlim0 <- xlim[[i]]
            }
            if (!add) graphics::plot(0, 0, type="n",
                                     main="", ylab=ylab, xlab=xlab,
                                     ylim=density.ylim0, xlim=density.xlim0)
            if (polygon) {
              pcol <- col[1]
              if (alphas[1]<1 && alphas[1]>0) {
                pcol <- Col(pcol, alpha=alphas[1])
              }
              with(dy, graphics::polygon(c(x,rev(x)),
                                         c(y,rep(0,length(y))),
                                         col=pcol,
                                         border=NA,
                                         density=densities[1],angle=angles[1]))
                if (!is.null(border)) with(dy, lines(x,y,col=border[1],lty=lty[1],lwd=lwd[1]))
            } else {
                graphics::lines(dy,main="",lty=lty[1],col=col[1],lwd=lwd[1])
            }
            if (rug) graphics::rug(y,col=Col(col[1],rug.alpha[1]))
            if (!is.null(main) && !(running.mean || scatter.plot)) {
                title(main[i],cex.main=cex.main)
            }
            if (!is.null(true)) {
                abline(v=true[i],lty=true.lty,col=true.col,lwd=true.lwd)
            }
            if (!is.null(se) && !any(is.na(se[[i]]))) {
                se.pos <- match(se[[i]],unlist(se))
                ns <- length(se.pos)+1
                se.alpha <- rep(alphas,length.out=ns)[-1]
                se.border <- rep(border,length.out=ns)[-1]
                se.col <- rep(col,length.out=ns)[-1]
                se.lty <- rep(lty,length.out=ns)[-1]
                se.lwd <- rep(lwd,length.out=ns)[-1]
                xx <- dy$x
                for (j in seq_along(se.pos)) {
                    if (polygon) {
                        yy <- dnorm(xx,mean=ss["Mean",se.pos[j]],sd=ss["SE",se.pos[j]])
                        if (se.alpha[j]>0) {
                          graphics::polygon(c(xx,rev(xx)),c(yy,rep(0,length(yy))),col=Col(se.col[j],alpha=se.alpha[j]),border=NA,density=densities[j],angle=angles[j])
                        }
                        if (!is.null(border)) lines(xx,yy,col=se.border[j],lty=se.lty[j],lwd=se.lwd[j])
                    } else {
                        graphics::curve(dnorm(x,mean=ss["Mean",se.pos[j]],sd=ss["SE",se.pos[j]]),lwd=se.lwd[j],lty=se.lty[j],col=se.col[j],add=TRUE)
                    }
                 }
                if (auto.legend) legend <- c("Kernel",colnames(ss)[se.pos])
                if (!is.null(legend)) {
                    if (polygon) {
                        dcol <- c(col[1],se.col)
                        fill <- Col(dcol,alpha)
                        fill[which(alpha==0)] <- NA
                        border[which(alpha==0)] <- NA
                        dcol[which(alpha!=0)] <- NA
                        graphics::legend(legendpos,legend,
                                         fill=fill,
                                         lty=1,
                                         col=dcol,
                                         border=border,
                                         cex=cex.legend)
                    } else {
                        graphics::legend(legendpos,legend,
                                         col=c(col[1],se.col),
                                         lty=c(lty[1],se.lty),
                                         lwd=c(lwd[1],se.lwd),
                                         cex=cex.legend)
                    }
                }
            }

        }
    }

    if (single) {
        nk <- unlist(lapply(se,length))
        col <- rep(col,length.out=K)
        for (i in seq(K)) {
            my.scatter.sim(i,add=(i>1),colors=col[i])
        }
        if (!is.null(main) && !byrow) {
            title(main[1],cex.main=cex.main)
        }
        if (missing(legend)) legend <- colnames(x)[estimate]
        legendold <- legend
        legend <- NULL
        alpha <- rep(alpha,length.out=K)
        density <- rep(density,length.out=K)
        angle <- rep(angle,length.out=K)
        for (i in seq_len(K)) {
            alphas <- if (K==1L) alpha else alpha[i]
            densities <- density[i]
            if (!is.null(densities) && densities<1) densities <- NULL
            if (length(se)>0) alphas <- c(alphas,rep(0,nk[i]))
            my.density.sim(i, add=(i>1),colors=col[i],alphas=alphas,
                           densities=densities,
                           angles=angle[i],
                           auto.legend=FALSE)
        }
        if (!is.null(legendold)) {
            legend <- rep(legendold,length.out=K)
            lcol <- Col(col,alpha)
            graphics::legend(legendpos,legend,
                             fill=lcol,border=col,cex=cex.legend)
        }

    } else {
        for (i in seq(K)) {
            my.scatter.sim(i)
            if (!is.null(main) && !byrow && scatter.plot) {
                title(main[i],cex.main=cex.main)
            }
            my.density.sim(i,auto.legend=missing(legend),
                           densities = rep(density, K),
                           angles = rep(angle, K)
                           )
            if (i==1 && ask && K>1) par(ask=ask)
        }
    }

}

##' Add covariance structure to Latent Variable Model
##'
##' Define covariances between residual terms in a \code{lvm}-object.
##'
##' The \code{covariance} function is used to specify correlation structure
##' between residual terms of a latent variable model, using a formula syntax.
##'
##' For instance, a multivariate model with three response variables,
##'
##' \deqn{Y_1 = \mu_1 + \epsilon_1}
##'
##' \deqn{Y_2 = \mu_2 + \epsilon_2}
##'
##' \deqn{Y_3 = \mu_3 + \epsilon_3}
##'
##' can be specified as
##'
##' \code{m <- lvm(~y1+y2+y3)}
##'
##' Pr. default the two variables are assumed to be independent. To add a
##' covariance parameter \eqn{r = cov(\epsilon_1,\epsilon_2)}, we execute the
##' following code
##'
##' \code{covariance(m) <- y1 ~ f(y2,r)}
##'
##' The special function \code{f} and its second argument could be omitted thus
##' assigning an unique parameter the covariance between \code{y1} and
##' \code{y2}.
##'
##' Similarily the marginal variance of the two response variables can be fixed
##' to be identical (\eqn{var(Y_i)=v}) via
##'
##' \code{covariance(m) <- c(y1,y2,y3) ~ f(v)}
##'
##' To specify a completely unstructured covariance structure, we can call
##'
##' \code{covariance(m) <- ~y1+y2+y3}
##'
##' All the parameter values of the linear constraints can be given as the right
##' handside expression of the assigment function \code{covariance<-} if the
##' first (and possibly second) argument is defined as well. E.g:
##'
##' \code{covariance(m,y1~y1+y2) <- list("a1","b1")}
##'
##' \code{covariance(m,~y2+y3) <- list("a2",2)}
##'
##' Defines
##'
##' \deqn{var(\epsilon_1) = a1}
##'
##' \deqn{var(\epsilon_2) = a2}
##'
##' \deqn{var(\epsilon_3) = 2}
##'
##' \deqn{cov(\epsilon_1,\epsilon_2) = b1}
##'
##' Parameter constraints can be cleared by fixing the relevant parameters to
##' \code{NA} (see also the \code{regression} method).
##'
##' The function \code{covariance} (called without additional arguments) can be
##' used to inspect the covariance constraints of a \code{lvm}-object.
##'
#
##'
##' @aliases covariance covariance<- covariance.lvm covariance<-.lvm
##' @aliases variance variance<- variance.lvm variance<-.lvm
##' @param object \code{lvm}-object
##' @param var1 Vector of variables names (or formula)
##' @param var2 Vector of variables names (or formula) defining pairwise
##' covariance between \code{var1} and \code{var2})
##' @param constrain Define non-linear parameter constraints to ensure positive definite structure
##' @param pairwise If TRUE and \code{var2} is omitted then pairwise correlation is added between all variables in \code{var1}
##' @param \dots Additional arguments to be passed to the low level functions
##' @param value List of parameter values or (if \code{var1} is unspecified)
##' @usage
##' \method{covariance}{lvm}(object, var1=NULL, var2=NULL, constrain=FALSE, pairwise=FALSE,...) <- value
##' @return A \code{lvm}-object
##' @author Klaus K. Holst
##' @seealso \code{\link{regression<-}}, \code{\link{intercept<-}},
##' \code{\link{constrain<-}} \code{\link{parameter<-}}, \code{\link{latent<-}},
##' \code{\link{cancel<-}}, \code{\link{kill<-}}
##' @keywords models regression
##' @export
##' @examples
##'
##' m <- lvm()
##' ## Define covariance between residuals terms of y1 and y2
##' covariance(m) <- y1~y2
##' covariance(m) <- c(y1,y2)~f(v) ## Same marginal variance
##' covariance(m) ## Examine covariance structure
##'
##'
`covariance` <- function(object,...) UseMethod("covariance")

##' @export
"variance<-" <- function(object,...,value) UseMethod("covariance<-")

##' @export
`variance` <- function(object,...) UseMethod("variance")

##' @export
"variance.lvm" <- function(object,...) covariance(object,...)

##' @export
"variance.formula" <- function(object,...) covariance(lvm(),object,...)

##' @export
"covariance.formula" <- function(object,...) covariance(lvm(),object,...)


##' @export
"variance<-.lvm" <- function(object,...,value) {
    covariance(object,...) <- value
    return(object)
}

##' @export
"covariance<-" <- function(object,...,value) UseMethod("covariance<-")

##' @export
"covariance<-.lvm" <- function(object, var1=NULL, var2=NULL, constrain=FALSE, pairwise=FALSE, ..., value) {

    if (!is.null(var1)) {
        if (inherits(var1,"formula")) {
            lhs <- getoutcome(var1)
            xf <- attributes(terms(var1))$term.labels
            xx <- unlist(lapply(xf, function(x) x[1]))
            if (length(lhs)==0) {
                covfix(object,var1,var2,pairwise=pairwise,...) <- value
                object$parpos <- NULL
                return(object)
            }
            else {
                yy <- decomp.specials(lhs)
            }
        } else {
            yy <- var1; xx <- var2
        }
        covfix(object,var1=yy,var2=xx,pairwise=pairwise,...) <- value
        object$parpos <- NULL
        return(object)
    }

    if (is.list(value)) {
        for (v in value) {
            covariance(object,pairwise=pairwise,constrain=constrain,...) <- v
        }
        return(object)
    }

    if (inherits(value,"formula")) {
        lhs <- getoutcome(value)
        if (length(lhs)==0) {
            return(covariance(object,all.vars(value),constrain=constrain,pairwise=pairwise,...))
        }
        yy <- decomp.specials(lhs)

        tt <- terms(value, specials=c("f","v"))
        xf <- attributes(terms(tt))$term.labels
        res <- lapply(xf,decomp.specials)
        nx <- length(xf)

        if (nx==1) {
            if(is.null(attr(tt,"specials")$f) | length(res[[1]])<2) {
                if(is.null(attr(tt,"specials")$v) & is.null(attr(tt,"specials")$f))
                {
                    for (i in yy)
                        for (j in res[[1]])
                            object <- covariance(object, c(i,j), pairwise=TRUE, constrain=constrain, ...)
                } else {
                    covfix(object,var1=yy,var2=NULL) <- res[[1]]
                }
            } else {
                covfix(object,var1=yy,var2=res[[1]][1]) <- res[[1]][2]
            }
            object$parpos <- NULL
            return(object)
        }

        xx <- unlist(lapply(res, function(z) z[1]))
        for (y in yy)
            for (i in seq_along(xx)) {
                if (length(res[[i]])>1) {
                    covfix(object, var1=y, var2=res[[i]][1]) <- res[[i]][2]
                } else if ((i+1)%in%attr(tt,"specials")$f | (i+1)%in%attr(tt,"specials")$v) {
                    covfix(object, var1=y, var2=NULL) <- res[[i]]
                } else {
                    object <- covariance(object,c(y,xx[i]),pairwise=TRUE,...)
                }
            }

        object$parpos <- NULL
        return(object)
    }
    else covariance(object,value,pairwise=pairwise,...)
}

##' @export
`covariance.lvm` <-
    function(object,var1=NULL,var2=NULL,exo=FALSE,pairwise=FALSE,constrain=FALSE,value,...) {

        if (!missing(value)) {
            covariance(object,var1=var1,var2,exo=exo,pariwise=pairwise,constrain=constrain,...) <- value
            return(object)
        }
        if (!is.null(var1)) {
            if (inherits(var1,"formula")) {
                covariance(object,constrain=constrain,
                           pairwise=pairwise,exo=exo,...) <- var1
                return(object)
            }
            allvars <- var1
            if (!missing(var2)) {
                if (inherits(var2,"formula"))
                    var2 <- all.vars(var2)
                allvars <- c(allvars,var2)
            }
            object <- addvar(object, allvars, messages=0, reindex=FALSE)

            if (constrain) {
                if (length(allvars)!=2) stop("Constraints only implemented for pairs")
                return(covarianceconst(object,allvars[1],allvars[2],...))
            }


            xorg <- exogenous(object)
            exoset <- setdiff(xorg,allvars)
            if (!exo & length(exoset)<length(xorg)) {
                exogenous(object) <- exoset
            }

            if (!missing(var2)) {
                for (i in seq_len(length(var1))) {
                    c1 <- var1[i]
                    for (j in seq_len(length(var2))) {
                        c2 <- var2[j]
                        object$cov[c1,c2] <- object$cov[c2,c1] <- 1
                        object$parpos <- NULL
                        index(object) <- reindex(object)
                    }
                }
            }
            else {
                if (pairwise) {
                    for (i in seq_len(length(var1))) {
                        c1 <- var1[i]
                        for (j in seq_len(length(var1))) {
                            c2 <- var1[j]
                            object$cov[c1,c2] <- object$cov[c2,c1] <- 1
                            object$parpos <- NULL
                            index(object) <- reindex(object)
                        }
                    }
                }
            }
            return(object)
        }
        else
            return(covfix(object))
    }

covarianceconst <- function(object,var1,var2,cname=NA,rname=NA,vname=NA,v2name=vname,lname=NA,l2name=lname,...) {
    if (inherits(var1,"formula")) {
        var1 <- getoutcome(var1)
        var2 <- attributes(var1)$x
    }
    curpar <- parlabels(object)
    if (is.na(cname)) {
        cname <- object$covpar[var1,var2]
    }

    if (is.na(v2name)) {
        v2name <- object$covpar[var2,var2]
    }
    if (is.na(vname)) {
        vname <- object$covpar[var1,var1]
    }

    nvarname <- c("rname","cname","vname","v2name","lname","l2name")[is.na(c(rname,cname,vname,v2name,lname,l2name))]

    nprefix <- sapply(nvarname, function(x) substr(x,1,1))
    for (i in seq_len(length(nvarname))) {
        count <- 0
        repeat {
            count <- count+1
            curname <- paste0(nprefix[i],count)
            if (!(curname%in%curpar)) break;
        }
        curpar <- c(curname,curpar)
        assign(nvarname[i],curname)
    }
    covariance(object,c(var1,var2)) <- c(vname,v2name)
    ff <- function(x) exp(x)
    constrain(object,vname,lname) <- ff
    if (vname!=v2name)
        constrain(object,v2name,l2name) <- ff
    covariance(object,var1,var2) <- cname
    cpar <- unique(c(lname,l2name,rname))
    constrain(object,cname,cpar) <- function(x) {
        prod(exp(x[seq(length(cpar)-1)]))^(1/(length(cpar)-1))*tanh(x[length(cpar)])
    }
    return(structure(object,rname=rname,cname=cname))
}

##' @export
`dsep` <-
    function(object,...) UseMethod("dsep")

##' Check d-separation criterion
##'
##' Check for conditional independence (d-separation)
##' @export
##' @aliases dsep dsep.lvm
##' @param object lvm object
##' @param x Variables for which to check for conditional independence
##' @param cond Conditioning set
##' @param return.graph If TRUE the moralized ancestral graph with the
##'     conditioning set removed is returned
##' @param ... Additional arguments to lower level functions
##' @details The argument `x` can be given as a formula, e.g. `x~y|z+v`
##'     or `~x+y|z+v` With everything on the rhs of the bar defining the
##'     variables on which to condition on.
##' @examples
##' m <- lvm(x5 ~ x4+x3, x4~x3+x1, x3~x2, x2~x1)
##' if (interactive()) {
##' plot(m,layoutType='neato')
##' }
##' dsep(m,x5~x1|x2+x4)
##' dsep(m,x5~x1|x3+x4)
##' dsep(m,~x1+x2+x3|x4)
##'
dsep.lvm <- function(object,x,cond=NULL,return.graph=FALSE,...) {
    if (inherits(x,"formula")) {
        xf <- getoutcome(x,sep="|")
        xx <- attr(xf,"x")
        if (length(xx)==0) stop("Not a valid formula")
        x <- c(xf,all.vars(xx[[1]]))
        if (length(xx)>1) {
            cond <- all.vars(xx[[2]])
        }
    }
    if (inherits(cond,"formula")) {
        cond <- all.vars(cond)
    }
    nod <- vars(object)
    x <- intersect(x,nod)
    cond <- intersect(cond,nod)
    V <- c(x,cond)
    ## Ancenstral graph
    keep <- c(V,ancestors(object,V))
    del <- setdiff(nod,keep)
    if (length(del)>0) object <- rmvar(object,del)
    ## moralized graph
    man <- object
    for (v in V) {
        pa <- parents(object,v)
        if (length(pa)>1)
            man$M[pa,pa] <- 1
    }
    man.sel <- rmvar(man,cond)
    ii <- match(x,vars(man.sel))
    A <- with(man.sel, (t(M)+M)>0)
    dsep <- c()
    for (i in ii) {
        conn <- DFS(A,i)
        i0 <- setdiff(ii,i)
        dsep <- c(dsep,!any(i0%in%conn))
    }
    res <- all(dsep)
    attr(man.sel,"dsep") <- res
    if (return.graph) return(man.sel)
    return(res)
}

##' Adds curly brackets to plot
##'
##' @title Adds curly brackets to plot
##' @param x center of the x axis of the curly brackets (or start end coordinates (x1,x2))
##' @param y center of the y axis of the curly brackets (or start end coordinates (y1,y2))
##' @param len Length of the curly brackets
##' @param theta angle (in radians) of the curly brackets orientation
##' @param wid Width of the curly brackets
##' @param shape shape (curvature)
##' @param col color (passed to lines/grid.lines)
##' @param lwd line width (passed to lines/grid.lines)
##' @param lty line type (passed to lines/grid.lines)
##' @param grid If TRUE use grid graphics (compatability with ggplot2)
##' @param npoints Number of points used in curves
##' @param text Label
##' @param offset Label offset (x,y)
##' @export
##' @examples
##' if (interactive()) {
##' plot(0,0,type="n",axes=FALSE,xlab="",ylab="")
##' curly(x=c(1,0),y=c(0,1),lwd=2,text="a")
##' curly(x=c(1,0),y=c(0,1),lwd=2,text="b",theta=pi)
##' curly(x=-0.5,y=0,shape=1,theta=pi,text="c")
##' curly(x=0,y=0,shape=1,theta=0,text="d")
##' curly(x=0.5,y=0,len=0.2,theta=pi/2,col="blue",lty=2)
##' curly(x=0.5,y=-0.5,len=0.2,theta=-pi/2,col="red",shape=1e3,text="e")
##' }
curly <- function(x,y,len=1,theta=0,
           wid,shape=1,
           col=1,lwd=1,lty=1,
           grid=FALSE,npoints=50,text=NULL,offset=c(0.05,0)) {
    if (length(x)==2 || length(y)==2) {
        x <- rep(x,length.out=2)
        y <- rep(y,length.out=2)
        v <- c(x[1]-x[2],y[1]-y[2])
        v0 <- c(1,0)-v
        len <- sum(v^2)^.5
        innerprod <- sum(v0)
        theta <- acos(innerprod/len)+theta
        len <- len/2
        x <- (x[1]-x[2])/2
        y <- (y[2]-y[1])/2
    }
    ii <- seq(0, pi/2, length.out=npoints)
    if (missing(wid)) {
        wid <- with(devcoords(),
        (fig.y2-fig.y1)/50)
    }
    x1 <- c(wid*(sin(ii)-1),
            c(0,0),
            wid*(1 - sin(rev(ii))),
            wid*(1 - sin(ii)),
            c(0,0),
            wid*(sin(rev(ii)) - 1))
    y1 <- c(-cos(ii),
            c(0,shape),
            shape+(cos(rev(ii))),
            shape+(2 - cos(ii)),
            c(shape+2, 2*shape+2),
            2*shape+2+cos(rev(ii)))

    x1 <- x1 + x + wid
    idx.max <- which.max(x1)
    max.y <- max(y1)
    y1 <- y1+1-(max.y+1)/2
    min.y <- min(y1)
    y1 <- y1*len/min.y+y
    ## Rotation
    x2 <- cos(theta) * (x1 - x) - sin(theta) * (y1 - y) + x
    y2 <- cos(theta) * (y1 - y) + sin(theta) * (x1 - x) + y
    x0 <- x1[idx.max]+offset[1]
    y0 <- y1[idx.max]+offset[2]
    xm <- cos(theta) * (x0 - x) - sin(theta) * (y0 - y) + x
    ym <- cos(theta) * (y0 - y) + sin(theta) * (x0 - x) + y
    if(grid){
        grid::grid.lines(grid::unit(x2,"npc"), grid::unit(y2,"npc"),
                         gp=grid::gpar(col=col,lwd=lwd,lty=lty))
    }
    else{
        points(x2,y2,type='l',col=col,lwd=lwd,lty=lty,xpd=TRUE)
    }
    theta <- acos(abs(cos(theta)))
    deg <- ((theta-pi/2)*180/pi)
    if (!is.null(text)) {
        text(xm,ym,text,srt=deg)
    }
}

img <- function(x,idx,col=list(gray.colors(10,1,0.2)),
         ylab="Item",xlab="Subject",lab=TRUE,
         border=1,rowcol=FALSE,plotfun=NULL,
         axis1=TRUE,axis2=TRUE,yaxs="r",xaxs="r",cex.axis=0.4,...) {
    x0 <- seq(nrow(x))
    y0 <- seq(ncol(x))
    image(x=x0,y=y0,as.matrix(x),col=col[[1]],axes=FALSE,ylab=ylab,xlab=xlab,xaxs=xaxs,yaxs=yaxs,...)
    if (axis1) {
        axis(1,at=seq(nrow(x)),lwd=0.5,cex.axis=cex.axis,las=3)
        if (lab) suppressWarnings(title("",xlab=xlab))
    }
    if (axis2) {
        axis(2,at=seq(ncol(x)),lwd=0.5,cex.axis=cex.axis,las=1)
        if (lab) suppressWarnings(title("",ylab=ylab))
    }
    if (!is.null(plotfun)) {
        plotfun(...)
    }
    if (!missing(idx)) {
        if (rowcol) {
            for (i in seq_len(length(idx)))
                image(x=x0,y=idx[[i]],as.matrix(x[,idx[[i]]]),col=col[[i]],add=TRUE,xaxs=xaxs,yaxs=yaxs,...)
        } else
            for (i in seq_len(length(idx)))
                image(x=idx[[i]],y=y0,as.matrix(x[idx[[i]],]),col=col[[i]],add=TRUE,xaxs=xaxs,yaxs=yaxs,...)
    }
}



##' Visualize categorical by group variable
##'
##' @title Organize several image calls (for visualizing categorical data)
##' @param x data.frame or matrix
##' @param group group variable
##' @param ncol number of columns in layout
##' @param byrow organize by row if TRUE
##' @param colorbar Add color bar
##' @param colorbar.space Space around color bar
##' @param label.offset label offset
##' @param order order
##' @param colorbar.border Add border around color bar
##' @param main Main title
##' @param rowcol switch rows and columns
##' @param plotfun Alternative plot function (instead of 'image')
##' @param axis1 Axis 1
##' @param axis2 Axis 2
##' @param mar Margins
##' @param col Colours
##' @param ... Additional arguments to lower level graphics functions
##' @author Klaus Holst
##' @examples
##' X <- matrix(rbinom(400,3,0.5),20)
##' group <- rep(1:4,each=5)
##' images(X,colorbar=0,zlim=c(0,3))
##' images(X,group=group,zlim=c(0,3))
##' \dontrun{
##' images(X,group=group,col=list(RColorBrewer::brewer.pal(4,"Purples"),
##'                                RColorBrewer::brewer.pal(4,"Greys"),
##'                                RColorBrewer::brewer.pal(4,"YlGn"),
##'                                RColorBrewer::brewer.pal(4,"PuBuGn")),colorbar=2,zlim=c(0,3))
##' }
##' images(list(X,X,X,X),group=group,zlim=c(0,3))
##' images(list(X,X,X,X),ncol=1,group=group,zlim=c(0,3))
##' images(list(X,X),group,axis2=c(FALSE,FALSE),axis1=c(FALSE,FALSE),
##'       mar=list(c(0,0,0,0),c(0,0,0,0)),yaxs="i",xaxs="i",zlim=c(0,3))
##' @export
images <- function(x,group,ncol=2,byrow=TRUE,colorbar=1,colorbar.space=0.1,label.offset=0.02,
            order=TRUE,colorbar.border=0,main,rowcol=FALSE,plotfun=NULL,
            axis1,axis2,mar,
            col=list(c("#EFF3FF", "#BDD7E7", "#6BAED6", "#2171B5"),
                     c("#FEE5D9", "#FCAE91", "#FB6A4A", "#CB181D"),
                     c("#EDF8E9", "#BAE4B3", "#74C476", "#238B45"),
                     c("#FEEDDE", "#FDBE85", "#FD8D3C", "#D94701")),
            ...) {
    if (is.data.frame(x) || is.matrix(x)) x <- list(x)
    K <- length(x)
    lout <- matrix(seq(K),ncol=ncol,byrow=byrow)
    hei <- rep(1,nrow(lout))/nrow(lout)
    wid <- rep(1,ncol)/ncol
    if (colorbar==1) {
        wid <- c(rep(1,ncol)/ncol*(1-colorbar.space),colorbar.space)
        lout <- cbind(lout,K+1)
    }
    if (colorbar==2) {
        hei <- c(rep(1,nrow(lout))/nrow(lout)*(1-colorbar.space),colorbar.space)
        lout <- rbind(lout,K+1)
    }
    if (missing(group)) {
        group <- rep(1,nrow(x[[1]]))
    }
    if (missing(main)) main <- rep("",K)
    if (!is.list(col)) col <- list(col)
    group <- factor(group)
    idxs <- lapply(levels(group), function(x) which(group==x))
    layout(lout,widths=wid,heights=hei)
    ##if (missing(mar)) par(mar=c(4,4,3,0))
    if (missing(axis2)) axis2 <- c(TRUE,rep(FALSE,K-1))
    if (missing(axis1)) axis1 <- rep(TRUE,K)
    for (i in seq(length(x))) {
        ##        if (!missing(mar)) par(mar=mar[[i]])
        img(x[[i]],idxs,col,axis2=axis2[i],axis1=axis1[i],main=main[i],rowcol=rowcol,plotfun=plotfun[[i]],...)
        ##        if (missing(mar)) par(mar=c(4,2,3,2))
    }
    G <- nlevels(group)
    M <- length(col[[1]])
    if (colorbar==1) {
        par(mar=c(0,0,0,2))
        plot.new(); plot.window(xlim=c(0,1),ylim=c(0,1))
        for (i in seq(G)) {
            lava::colorbar(col[[i]],values=seq(M)-1,direction="horizontal",
                           y.range=c(1-i/(G+1),1-i/(G+1)+label.offset),
                           border=colorbar.border,x.range=c(0,1),srt=0,cex=0.6)
            text(0.5,1-i/(G+1)-label.offset, levels(group)[i])
        }
    }
    if (colorbar==2) {
        par(mar=c(0,0,0,0))
        plot.new(); plot.window(xlim=c(0,1),ylim=c(0,1))
        for (i in seq(G)) {
            xr <- c(1-i/(G+1),1-i/(G+1)+.1)-.1/2
            lava::colorbar(col[[i]],values=seq(M)-1,direction="horizontal",
                           x.range=xr,
                           border=colorbar.border,y.range=c(0.3,0.5),srt=0,cex=0.6)
            text(mean(xr),.1, levels(group)[i])
        }
    }
}

##' Extract i.i.d. decomposition from model object
##'
##' This function extracts
##' @param x Model object
##' @param ... Additional arguments (see the man-page of the IC method)
##' @export
iid <- function(x, ...) UseMethod("iid")

##' @export
iid.default <- function(x, ...) {
  res <- IC(x, ...)
  if (!is.null(attr(res, "bread"))) {
    attr(res, "bread") <- attr(res, "bread")/NROW(res)
  }
  return(res/NROW(res))
}

##' Run SAS code like in the following:
##'
##' ODS CSVALL BODY="myest.csv";
##' proc nlmixed data=aj qpoints=2 dampstep=0.5;
##' ...
##' run;
##' ODS CSVALL Close;
##'
##' and read results into R with:
##'
##' \code{getsas("myest.csv","Parameter Estimates")}
##'
##' @title Read SAS output
##' @param infile file (csv file generated by ODS)
##' @param entry Name of entry to capture
##' @param \dots additional arguments to lower level functions
##' @author Klaus K. Holst
##' @export
getSAS <- function(infile,entry="Parameter Estimates",...) {
    con <- file(infile, blocking = FALSE)
    inp <- readLines(con)
    close(con)
    linestart <- 1; lineend <- length(inp)
    idx <- sapply(inp,function(x) length(grep(entry, x))>0)
    if (sum(idx)==1) {
        linestart <- which(idx)
        for (i in seq(linestart,length(inp))) {
            lineend <- i-1
            if (inp[i]=="") break;
        }
    } else {
        stop("No match or duplicate entries!")
    }
    subinp <- inp[(linestart+1):(lineend)]
    con <- textConnection(subinp)
    res <- read.csv(con,header=TRUE)
    close(con)
    return(res)
}

##' @export
"intervention<-" <- function(object, ..., value)
  UseMethod("intervention<-")

##' @export
`intervention` <-
  function(object, ...) UseMethod("intervention")

##' Define intervention
##'
##' Define intervention in a `lvm` object
##' @param object lvm object
##' @param to String defining variable or formula
##' @param value function defining intervention
##' @param dist Distribution
##' @param ... Additional arguments to lower level functions
##' @aliases intervention<- intervention intervention.lvm intervention<-.lvm
##' @seealso regression lvm sim
##' @examples
##' m <- lvm(y ~ a + x, a ~ x)
##' distribution(m, ~a+y) <- dist_bernoulli()
##' mm <- intervention(m, "a", value=3)
##' sim(mm, 10)
##' mm <- intervention(m, a~x, function(x) (x>0)*1)
##' sim(mm, 10)
##' @export
intervention.lvm <- function(object, to, value, dist=dist_none(), ...) {
  if (!is.numeric(value))
    regression(object, to, ...) <- value
  y <- to
  if (inherits(to, "formula")) {
    y <- getoutcome(to)
    if (length(y)==0)
      y <- attr(y, "x")
  }
  parents <- parents(object, y)
  if (length(parents)>0)
    cancel(object) <- toformula(y, parents)
  if (is.numeric(value)) {
    distribution(object, y) <- dist_const(value)
  } else {
    distribution(object, y) <- dist
  }
  return(object)
}

##' @export
"intervention<-.lvm" <- function(object, to, ..., value) {
  object <- intervention(object, to, value, ...)
  return(object)
}

##' @export
##' @export density.sim
density.sim <- function(x, ..., add=FALSE, plot.type="single") {
    plot.sim(x, ...,
             add=add,
             scatter.plot=FALSE,
             plot.type=plot.type)
}

##' Plot method for simulation 'sim' objects
##'
##' Density and scatter plots
##' @examples
##' n <- 1000
##' val <- cbind(est1=rnorm(n,sd=1),est2=rnorm(n,sd=0.2),est3=rnorm(n,1,sd=0.5),
##'              sd1=runif(n,0.8,1.2),sd2=runif(n,0.1,0.3),sd3=runif(n,0.25,0.75))
##' 
##' plot.sim(val,estimate=c(1,2),true=c(0,0),se=c(4,5),equal=TRUE,scatter.plot=TRUE)
##' plot.sim(val,estimate=c(1,3),true=c(0,1),se=c(4,6),xlim=c(-3,3),
##' 	scatter.ylim=c(-3,3),scatter.plot=TRUE)
##' plot.sim(val,estimate=c(1,2),true=c(0,0),se=c(4,5),equal=TRUE,
##' 	plot.type="single",scatter.plot=TRUE)
##' plot.sim(val,estimate=c(1),se=c(4,5,6),plot.type="single",scatter.plot=TRUE)
##' plot.sim(val,estimate=c(1,2,3),equal=TRUE,scatter.plot=TRUE)
##' plot.sim(val,estimate=c(1,2,3),equal=TRUE,byrow=TRUE,scatter.plot=TRUE)
##' plot.sim(val,estimate=c(1,2,3),plot.type="single",scatter.plot=TRUE)
##' plot.sim(val,estimate=1,se=c(3,4,5),plot.type="single",scatter.plot=TRUE)
##' 
##' density.sim(val,estimate=c(1,2,3),density=c(0,10,10), lwd=2, angle=c(0,45,-45),cex.legend=1.3)
##' @aliases density.sim plot.sim
##' @export
##' @export plot.sim
##' @param x sim object
##' @param estimate columns with estimates
##' @param se columns with standard error estimates
##' @param true (optional) vector of true parameter values
##' @param names (optional) names of estimates
##' @param auto.layout Auto layout (default TRUE)
##' @param byrow Add new plots to layout by row
##' @param type plot type
##' @param ask if TRUE user is asked for input, before a new figure is drawn
##' @param col colour (for each estimate)
##' @param pch plot symbol
##' @param cex point size
##' @param lty line type
##' @param lwd line width
##' @param legend legend 
##' @param legendpos legend position
##' @param cex.legend size of legend text
##' @param plot.type 'single' or 'multiple' (default)
##' @param polygon if TRUE fill the density estimates with colour
##' @param density if non-zero add shading lines to polygon
##' @param angle shading lines angle of polygon
##' @param cex.axis Font size on axis
##' @param alpha Semi-transparent level (1: non-transparent, 0: full)
##' @param main Main title
##' @param cex.main Size of title font
##' @param equal Same x-axis and y-axis for all plots
##' @param delta Controls the amount of space around axis limits
##' @param ylim y-axis limits
##' @param xlim x-axis limits
##' @param ylab y axis label
##' @param xlab x axis label
##' @param rug if TRUE add rug representation of data to x-axis
##' @param rug.alpha rug semi-transparency level
##' @param line.col line colour (running mean, only for scatter plots)
##' @param line.lwd line width (running mean, only for scatter plots)
##' @param line.lty line type (running mean, only for scatter plots)
##' @param line.alpha line transparency 
##' @param scatter.ylab y label for density plots
##' @param scatter.ylim y-axis limits for density plots
##' @param scatter.xlim x-axis limits for density plots
##' @param scatter.alpha semi-transparency of scatter plot
##' @param scatter.col scatter plot colour
##' @param border border colour of density estimates
##' @param true.lty true parameter estimate line type
##' @param true.col true parameter colour
##' @param true.lwd true parameter line width
##' @param density.plot if TRUE add density plot
##' @param scatter.plot if TRUE add scatter plot 
##' @param running.mean if TRUE add running average estimate to scatter plot
##' @param add if TRUE add to existing plot
##' @param ... additional arguments to lower level functions
plot.sim <- function(x,estimate,se=NULL,true=NULL,
             names=NULL,
             auto.layout=TRUE,
             byrow=FALSE,
             type="p",
             ask=grDevices::dev.interactive(),             
             col=c("gray60","orange","darkblue","seagreen","darkred"),
             pch=16,cex=0.5,lty=1,lwd=0.3,
             legend,
             legendpos="topleft",
             cex.legend=0.8,
             plot.type=c("multiple","single"),
             polygon=TRUE,
             density=NULL,
             angle=-45,
             cex.axis=0.8,
             alpha=0.2,
             main,
             cex.main=1,
             equal=FALSE,
             delta=1.15,
             ylim=NULL,
             xlim=NULL,
             ylab="",
             xlab="",
             rug=FALSE,
             rug.alpha=0.5,
             line.col=scatter.col,             
             line.lwd=1,
             line.lty=1,
             line.alpha=1,
             scatter.ylab="Estimate",
             scatter.ylim=NULL,
             scatter.xlim=NULL,
             scatter.alpha=0.5,
             scatter.col=col,
             border=col,
             true.lty=2,true.col="gray70",true.lwd=1.2,
             density.plot=TRUE,
             scatter.plot=FALSE,
             running.mean=scatter.plot,
             add = FALSE,
             ...) {

    if (missing(estimate)) {
        estimate <- seq(ncol(x))
    }
    if (is.null(estimate)) {
        av <- apply(x[,drop=FALSE],2,function(z) cumsum(z)/seq(length(z)))
        graphics::matplot(x,type="p",pch=pch, cex=cex, col=col,...)
        graphics::matlines(av,type="l",col=col,lty=lty,...)
        if (!is.null(true)) abline(h=true,lty=true.lty,...)
        if (missing(legend)) legend <- colnames(x)
        if (!is.null(legend))
            graphics::legend(legendpos,legend=legend,bg="white",
                             col=scatter.col,lty=lty,pch=pch,...)
        return(invisible(NULL))
    }
    if (is.character(estimate)) {
        estimate <- match(estimate,colnames(x))
    }

    K <- length(estimate)
    est <- tru <- c()
    if (length(se)>0) {
        if (K==1 && !is.list(se))
            se <- list(se)
        else se <- as.list(se)
    } else {
        est <- estimate; tru <- true
    }

    for (i in seq_along(estimate)) {
        est <- c(est,list(rep(estimate[i],length(se[[i]]))))
        if (!is.null(true)) tru <- c(tru,list(rep(true[i],length(se[[i]]))))
    }

    if (length(se)>0) {
      for (i in seq_along(se)) {
            if (is.character(se[[i]])) se[[i]] <- match(se[[i]],colnames(x))
        }

    }
  ss <- summary.sim(x,estimate=unlist(est),se=unlist(se),true=unlist(tru),names=names, confint=FALSE)

    oldpar <- NULL
    on.exit({
        par(oldpar)
        return(invisible(ss))
    })

    single <- tolower(plot.type[1])=="single"

    nc <- (scatter.plot || running.mean) + density.plot
    nr <- min(6,K)
    if (single) nr <- 1
    
    if (auto.layout && (nc>1 || nr>1)) {
        oma.multi = c(2, 0, 2, 0)
        mar.multi = c(1.5, 4.1, 1, 1)
        opt <- list(mar=mar.multi, oma=oma.multi,
                      cex.axis=cex.axis,las=1,
                      ask=FALSE)
        if (byrow) {
            opt <- c(opt, list(mfrow=c(nr,nc)))
        } else {
            opt <- c(opt, list(mfcol=c(nc,nr)))
        }
        oldpar <- do.call(par, opt)
    }
  dys <- c()
  maxdy <- 0

  if (density.plot)
    for (i in seq(K)) {
            ii <- estimate[i]
            y <- as.vector(x[,ii,drop=TRUE])
            dy <- stats::density(y,...)
            dys <- c(dys,list(dy))
            maxdy <- max(maxdy,dy$y)
        }

    if (equal || single) {
        if (is.null(scatter.ylim)) {
            rg <- range(x[,estimate])
            rg <- rg+c(-1,1)*abs(diff(rg)*(delta-1))
            scatter.ylim <-  rep(list(rg),K)
        }
        if (density.plot) {
            if (is.null(ylim)) ylim <- rep(list(c(0,maxdy*delta)),K)
            if (is.null(xlim)) xlim <- scatter.ylim
        }
    }

    if (!is.null(ylim)) {
        if (!is.list(ylim)) ylim <- list(ylim)
        ylim <- rep(ylim,length.out=K)
    }
    ylab <- rep(ylab,length.out=K)
    if (!is.null(scatter.ylim)) {
        if (!is.list(scatter.ylim)) scatter.ylim <- list(scatter.ylim)
        scatter.ylim <- rep(scatter.ylim,length.out=K)
    }
    if (!is.null(xlim)) {
        if (!is.list(xlim)) xlim <- list(xlim)
        xlim <- rep(xlim,length.out=K)
    }
    if (missing(main)) {
        main <- NULL
        ## if (!missing(names)) main <- names
        ## else if (K>1 && !single) main <- colnames(ss)
    }
    if (!is.null(main)) main <- rep(main,length.out=K)
    if (missing(lty)) {
        lty <- rep(1,K)
        if (single || !polygon) {
            lty <- 1:20
        }
    }

    my.scatter.sim <- function(i,add=FALSE,colors,...) {
        ii <- estimate[i]
        if (!missing(colors)) {
            scatter.col <- line.col <- true.col <- colors[1]
        }
        y <- as.vector(x[,ii,drop=TRUE])
        args <- list(y,ylab=scatter.ylab[i],
                     col=Col(scatter.col[1], scatter.alpha),
                     cex=cex,pch=pch,type=type,...)
        if (!is.null(scatter.ylim)) args <- c(args,list(ylim=scatter.ylim[[i]]))
        if (!is.null(scatter.xlim)) args <- c(args, list(xlim=scatter.xlim[[i]]))
        if (scatter.plot) {
            if (!add) {
                do.call(graphics::plot,args)
            } else {
                do.call(graphics::points,args)
            }
        }
        if (running.mean) {
            lines(cumsum(y)/seq_along(y),col=line.col[1],lwd=line.lwd,lty=line.lty)
            if (!is.null(true))
                abline(h=true[i],lty=true.lty,col=true.col[1],lwd=true.lwd)
        }
    }

    my.density.sim <- function(i,add=FALSE,
                               colors,
                               alphas=alpha,
                               auto.legend=TRUE,
                               densities=NULL,
                               angles=angle,
                               ...) {
        ii <- estimate[i]
        y <- as.vector(x[,ii,drop=TRUE])
        if (!missing(colors)) {
            col <- border <- colors
            col <- true.col <- colors
        }
        if (density.plot) {
            dy <- stats::density(y)
            if (is.null(ylim)) {
                density.ylim0 <- c(0,max(dy$y)*delta)
            } else {
                density.ylim0 <- ylim[[i]]
            }
            if (is.null(xlim)) {
                density.xlim0 <- range(dy$x)
            } else {
                density.xlim0 <- xlim[[i]]
            }
            if (!add) graphics::plot(0, 0, type="n",
                                     main="", ylab=ylab, xlab=xlab,
                                     ylim=density.ylim0, xlim=density.xlim0)
            if (polygon) {
              pcol <- col[1]
              if (alphas[1]<1 && alphas[1]>0) {
                pcol <- Col(pcol, alpha=alphas[1])
              }
              with(dy, graphics::polygon(c(x,rev(x)),
                                         c(y,rep(0,length(y))),
                                         col=pcol,
                                         border=NA, #col[1],
                                         density=densities[1],
                                         angle=angles[1]))
                if (!is.null(border)) with(dy, lines(x,y,col=border[1],lty=lty[1],lwd=lwd[1]))
            } else {
                graphics::lines(dy,main="",lty=lty[1],col=col[1],lwd=lwd[1])
            }
            if (rug) graphics::rug(y,col=Col(col[1],rug.alpha[1]))
            if (!is.null(main) && !(running.mean || scatter.plot)) {
                title(main[i],cex.main=cex.main)
            }
            if (!is.null(true)) {
                abline(v=true[i],lty=true.lty,col=true.col,lwd=true.lwd)
            }
            if (!is.null(se) && !any(is.na(se[[i]]))) {
                se.pos <- match(se[[i]],unlist(se))
                ns <- length(se.pos)+1
                se.alpha <- rep(alphas,length.out=ns)[-1]
                se.border <- rep(border,length.out=ns)[-1]
                se.col <- rep(col,length.out=ns)[-1]
                se.lty <- rep(lty,length.out=ns)[-1]
                se.lwd <- rep(lwd,length.out=ns)[-1]
                xx <- dy$x
                for (j in seq_along(se.pos)) {
                  if (polygon) {
                    yy <- dnorm(xx, mean=ss["Mean",se.pos[j]], sd=ss["SE",se.pos[j]])
                    ## Mirror the same fill logic as the empirical polygon
                    se.fill <- if (is.null(densities[j+1]) || is.na(densities[j+1])) {
                                 Col(se.col[j], alpha=se.alpha[j])
                               } else {
                                 se.col[j]
                               }
                    graphics::polygon(c(xx,rev(xx)), c(yy,rep(0,length(yy))),
                                      col     = se.fill,
                                      border  = NA, #se.col[j],
                                      density = densities[j+1],
                                      angle   = angles[j+1])
                    if (!is.null(border)) lines(xx, yy, col=se.col[j],
                                                lty=se.lty[j], lwd=se.lwd[j])
                  } else {
                    graphics::curve(dnorm(x, mean=ss["Mean",se.pos[j]], sd=ss["SE",se.pos[j]]),
                                    lwd=se.lwd[j], lty=se.lty[j], col=se.col[j], add=TRUE)
                  }
                }
                if (auto.legend) legend <- c("Empirical",colnames(ss)[se.pos])
                if (!is.null(legend)) {
                    if (polygon) {
                        dcol <- c(col[1],se.col)
                        fill <- Col(dcol,alpha)
                        fill[which(alpha==0)] <- NA
                        border[which(alpha==0)] <- NA
                        dcol[which(alpha!=0)] <- NA
                        graphics::legend(legendpos,legend,
                                         fill=fill,
                                         ## lty=1,
                                         col=dcol,
                                         border=border,
                                         cex=cex.legend)
                    } else {
                        graphics::legend(legendpos,legend,
                                         col=c(col[1],se.col),
                                         lty=c(lty[1],se.lty),
                                         lwd=c(lwd[1],se.lwd),
                                         cex=cex.legend)
                    }
                }
            }

        }
    }

    if (single) {
        nk <- unlist(lapply(se,length))
        col <- rep(col,length.out=K)
        for (i in seq(K)) {
            my.scatter.sim(i,add=(i>1),colors=col[i])
        }
        if (!is.null(main) && !byrow) {
            title(main[1],cex.main=cex.main)
        }
        if (missing(legend)) legend <- colnames(x)[estimate]
        legendold <- legend
        legend <- NULL
        alpha <- rep(alpha,length.out=K)
        density <- if (!is.null(density)) rep(density, length.out=K) else NULL
        angle <- rep(angle,length.out=K)
        for (i in seq_len(K)) {
          alphas <- if (K==1L) alpha else alpha[i]
          ## if (length(se)>0) alphas <- c(alphas, rep(0, nk[i]))
          my.density.sim(i, add=if(i>1) TRUE else add, # first panel respects add
                   colors    = col[i],
                   alphas    = alphas,
                   densities = density,
                   angles    = angle,
                   auto.legend=FALSE, ...)
        }
        if (!is.null(legendold)) {
            legend <- rep(legendold,length.out=K)
            lcol <- Col(col,alpha)
            graphics::legend(legendpos,legend,
                             fill=lcol,border=col,cex=cex.legend)
        }

    } else {
        for (i in seq(K)) {
            my.scatter.sim(i)
            if (!is.null(main) && !byrow && scatter.plot) {
                title(main[i],cex.main=cex.main)
            }
            nse_i <- if (length(se) > 0) length(se[[i]]) else 0
            my.density.sim(i, add=add,
                           auto.legend=missing(legend),
                           densities = density,
                           angles = angle, ...
                           )
            if (i==1 && ask && K>1) par(ask=ask)
        }
    }

}

##' @export
colsel <- function(locate,...) {
    ytop    <- rep(seq(1/26,1,by=1/26),each=26)[1:657]
    ybottom <- rep(seq(0,1-1/26,by=1/26),each=26)[1:657]
    xleft   <- rep(seq(0,1-1/26,by=1/26),times=26)[1:657]
    xright  <- rep(seq(1/26,1,by=1/26),times=26)[1:657]
    pall    <- round(col2rgb(colors())/256)
    pall    <- colSums(pall) ; pall2 <- character(0)
    pall2[pall>0]   <- "black"
    pall2[pall==0]  <- "white"

    par(mar=c(0,0,1,0))

    plot.new()
    title(main="Palette of colors()")
    rect(xleft,ybottom,xright,ytop,col=colors())
    text(x=xleft+((1/26)/2)
        ,y=ytop-((1/26)/2)
        ,labels = 1:657
             ,cex=0.55
        ,col=pall2)

    if (missing(locate)) return(invisible(NULL))
    colmat <- matrix(c(1:657,rep(NA,26^2-657)),byrow=TRUE,ncol=26,nrow=26)
    cols <- NA
    for(i in seq_len(locate))
    {
        h    <- locator(1)
        if(any(h$x<0,h$y<0,h$x>1,h$y>1)) stop("locator out of bounds!")
        else {
            cc        <- floor(h$x/(1/26))+1
            rr        <- floor(h$y/(1/26))+1
            cols[i]    <- colors()[colmat[rr,cc]]
        }
    }
    return(cols)
}


##' Extension of the \code{identify} function
##'
##' For the usual 'X11' device the identification process is
##' terminated by pressing any mouse button other than the first. For
##' the 'quartz' device the process is terminated by pressing either
##' the pop-up menu equivalent (usually second mouse button or
##' 'Ctrl'-click) or the 'ESC' key.
##' @title Identify points on plot
##' @usage
##' \method{click}{default}(x, y=NULL, label=TRUE, n=length(x), pch=19, col="orange", cex=3, ...)
##' idplot(x, y ,..., id=list(), return.data=FALSE)
##' @aliases idplot click.default click colsel
##' @param x X coordinates
##' @param y Y coordinates
##' @param label Should labels be added?
##' @param n Max number of inputs to expect
##' @param pch Symbol
##' @param col Colour
##' @param cex Size
##' @param id List of arguments parsed to \code{click} function
##' @param return.data Boolean indicating if selected points should be returned
##' @param \dots Additional arguments parsed to \code{plot} function
##' @author Klaus K. Holst
##' @seealso \code{\link{idplot}}, \code{identify}
##' @examples
##' if (interactive()) {
##'     n <- 10; x <- seq(n); y <- runif(n)
##'     plot(y ~ x); click(x,y)
##'
##'     data(iris)
##'     l <- lm(Sepal.Length ~ Sepal.Width*Species,iris)
##'     res <- plotConf(l,var2="Species")## ylim=c(6,8), xlim=c(2.5,3.3))
##'     with(res, click(x,y))
##'
##'     with(iris, idplot(Sepal.Length,Petal.Length))
##' }
##' @keywords iplot
##' @export
click <- function(x,...){
  UseMethod("click")
}

##' @export
click.default <-
function(x, y=NULL, label=TRUE, n=length(x),
         pch=19, col="orange", cex=3, ...) {
  xy <- xy.coords(x, y); x <- xy$x; y <- xy$y
  sel <- rep(FALSE, length(x)); res <- integer(0)
  while(sum(sel) < n) {
    ans <- identify(x[!sel], y[!sel], n=1, plot=FALSE, ...)
    if(!length(ans)) break
    ans <- which(!sel)[ans]
    points(x[ans], y[ans], pch = pch, col=col, cex=cex)
    if (label)
      text(x[ans], y[ans], ans)
    sel[ans] <- TRUE
    res <- c(res, ans)
  }
  res
}

##' @export
idplot <- function(x, y=NULL, ..., id=list(), return.data=FALSE) {
  plot(x,y,...)
  id$x <- x; id$y <- y
  idx <- do.call("click", id)
  if (return.data) {
    return(with(id, cbind(x,y)[idx,]))
  }
  return(idx)
}

##' Spaghetti plot for longitudinal data
##'
##' @title Spaghetti plot
##' @param formula Formula (response ~ time)
##' @param data data.frame
##' @param id Id variable
##' @param group group variable
##' @param type Type (line 'l', stair 's', ...)
##' @param lty Line type
##' @param pch Colour
##' @param col Colour
##' @param alpha transparency (0-1)
##' @param lwd Line width
##' @param level Confidence level
##' @param trend.formula Formula for trendline
##' @param tau Quantile to estimate (trend)
##' @param trend.lty Trend line type
##' @param trend.join Trend polygon
##' @param trend.delta Length of limit bars
##' @param trend Add trend line
##' @param trend.col Colour of trend line
##' @param trend.alpha Transparency
##' @param trend.lwd Trend line width
##' @param trend.jitter Jitter amount
##' @param legend Legend
##' @param by make separate plot for each level in 'by' (formula, name of column, or vector)
##' @param xlab Label of X-axis
##' @param ylab Label of Y-axis
##' @param add Add to existing device
##' @param ... Additional arguments to lower level arguments
##' @author Klaus K. Holst
##' @export
##' @examples
##' if (interactive() & requireNamespace("mets")) {
##' K <- 5
##' y <- "y"%++%seq(K)
##' m <- lvm()
##' regression(m,y=y,x=~u) <- 1
##' regression(m,y=y,x=~s) <- seq(K)-1
##' regression(m,y=y,x=~x) <- "b"
##' N <- 50
##' d <- sim(m,N); d$z <- rbinom(N,1,0.5)
##' dd <- mets::fast.reshape(d); dd$num <- dd$num+3
##' spaghetti(y~num,dd,id="id",lty=1,col=Col(1,.4),
##'           trend.formula=~factor(num),trend=TRUE,trend.col="darkblue")
##' dd$num <- dd$num+rnorm(nrow(dd),sd=0.5) ## Unbalance
##' spaghetti(y~num,dd,id="id",lty=1,col=Col(1,.4),
##'           trend=TRUE,trend.col="darkblue")
##' spaghetti(y~num,dd,id="id",lty=1,col=Col(1,.4),
##'            trend.formula=~num+I(num^2),trend=TRUE,trend.col="darkblue")
##' }
spaghetti <- function(formula,data=NULL,id="id",group=NULL,
              type="o",lty=1,pch=NA,col=1:10,alpha=0.3,lwd=1,
              level=0.95,
              trend.formula=formula,tau=NULL,
              trend.lty=1,trend.join=TRUE,trend.delta=0.2,
              trend=!is.null(tau),trend.col=col,
              trend.alpha=0.2,trend.lwd=3,
              trend.jitter=0,
              legend=NULL, by=NULL,
              xlab="Time",ylab="",add=FALSE,...) {

    if (!lava.options()$cluster.index) stop("mets not available? Check 'lava.options()cluster.index'.")
    if (!is.null(by)) {
        if (is.character(by) && length(by==1)) {
            by <- data[,by]
        } else if (inherits(by,"formula")) {
            ##by <- model.matrix(update(by,~-1+.), model.frame(~.,data,na.action=na.pass))
            by <- model.frame(by,data,na.action=na.pass)
        }
        cl <- match.call(expand.dots=TRUE)
        cl$by <- NULL
        datasets <- split(data,by)
        res <- c()
        for (d in datasets) {
            cl$data <- d
            res <- c(res, eval(cl,parent.frame()))
        }
        return(invisible(res))
    }
    if (!is.null(group)) {
        if (is.character(group) && length(group==1)) {
            M <- data[,group]
        } else if (inherits(group,"formula")) {
            ##M <- model.matrix(update(group,~-1+.),data)
            M <- model.frame(group,data,na.action=na.pass)
        } else {
            M <- group
        }
        if (!add) plot(formula,data=data,xlab=xlab,ylab=ylab,...,type="n")
        dd <- split(data,M)
        K <- length(dd)
        if (length(type)<K)        type <- rep(type,K)
        if (length(col)<K)         col <- rep(col,K)
        if (length(pch)<K)         pch <- rep(pch,K)
        if (length(lty)<K)         lty <- rep(lty,K)
        if (length(lwd)<K)         lwd <- rep(lwd,K)
        if (length(alpha)<K)       alpha <- rep(alpha,K)
        if (length(trend)<K)       trend <- rep(trend,K)
        if (length(trend.col)<K)   trend.col <- rep(trend.col,K)
        if (length(trend.lty)<K)   trend.lty <- rep(trend.lty,K)
        if (length(trend.alpha)<K) trend.alpha <- rep(trend.alpha,K)
        if (length(trend.lwd)<K)   trend.lwd <- rep(trend.lwd,K)
        for (i in seq_len(K)) {
            spaghetti(formula,data=dd[[i]],id=id,type=type[i],
                     lty=lty[i],pch=pch[i],col=col[i],lwd=lwd[i],
                     alpha=alpha[i],
                     group=NULL,
                     trend=trend[i],tau=tau,
                     trend.col=trend.col[i],
                     trend.alpha=trend.alpha[i],
                     trend.lwd=trend.lwd[i],
                     trend.lty=trend.lty[i],
                     trend.delta=trend.delta,
                     trend.formula=trend.formula,
                     add=TRUE,...)
        }
        if (!is.null(legend)) {
            graphics::legend(legend,names(dd),lwd=lwd,col=col,lty=lty)
        }
        return(invisible(NULL))
    }

    if (inherits(id,"formula")) id <- all.vars(id)
    if (inherits(group,"formula")) group <- all.vars(group)
    if (is.character(id) && length(id)==1) Id <- data[,id]
    y <- getoutcome(formula)
    x <- attributes(y)$x
    Idx <- function(vname,widenames) {
        idx <- which(unlist(lapply(widenames,function(x)
            length(grep(paste0(vname,"[0-9]+$"),x))>0)))
        nn <- widenames[idx]
        ord <- order(char2num(unlist(lapply(nn,function(x) gsub(vname,"",x)))))
        idx[ord]
    }

    if (!requireNamespace("mets",quietly=TRUE)) stop("'mets' package required")

    if (length(x)==0) {
        data <- data[,c(id,y),drop=FALSE]
        wide <- mets::fast.reshape(data,id=id,varying=y,...)
        yidx <- Idx(y,names(wide))
        Y <- wide[,yidx,drop=FALSE]
        X <- NULL
        matplot(t(Y),type=type,lty=lty,pch=pch,lwd=lwd,col=Col(col[1],alpha[1]),xlab=xlab,ylab=ylab,...)
    } else {        
        data <- data[,c(id,x,y),drop=FALSE]
        wide <- mets::fast.reshape(data[order(data[,id],data[,x]),],id=id,varying=c(y,x),...)
        yidx <- Idx(y,names(wide))
        xidx <- Idx(x,names(wide))
        Y <- wide[,yidx,drop=FALSE]
        X <- wide[,xidx,drop=FALSE]
        matplot(t(X),t(Y),type=type,pch=pch,lty=lty,lwd=lwd,col=Col(col[1],alpha[1]),xlab=xlab,ylab=ylab,add=add,...)
        if (trend) {
            if (is.numeric(trend.formula)) {
                trend.formula <- sort(trend.formula)
                tf <- toformula(y,"1")
                res <- c()
                if (!is.null(tau)) {
                    if (length(trend.alpha)<length(tau)) 	trend.alpha <- rep(trend.alpha,length(tau))
                    if (length(trend.lty)<length(tau)) 	trend.lty <- rep(trend.lty,length(tau))
                    if (length(trend.col)<length(tau)) 	trend.col <- rep(trend.col,length(tau))
                    if (length(trend.lwd)<length(tau)) 	trend.lwd <- rep(trend.lwd,length(tau))
                }
                for (i in trend.formula) {
                    data0 <- data[data[,x]==i,,drop=FALSE]
                    newdata <- data.frame(i); names(newdata) <- x
                    if (!is.null(tau)) {
                        ##if (!require(quantreg)) stop("Install 'quantreg'")
                        suppressWarnings(r1 <- quantreg::rq(tf,data=data0,tau=tau))
                        pr <- predict(r1,newdata=newdata,level=level)
                        res <- rbind(res,pr)
                    } else {
                        l1 <- lm(tf,data0)
                        pr <- predict(l1,newdata=newdata,interval="confidence",level=level)
                        res <- rbind(res,pr)
                    }
                }
                if (!is.null(tau)) {
                    for (j in seq_len(ncol(res))) {
                            if (trend.join) lines(trend.formula,res[,j],col=trend.col[j],lwd=trend.lwd[j],lty=trend.lty[j],...)
                            if (trend.delta>0) confband(trend.formula,res[,j],line=FALSE,col=trend.col[j],lty=trend.lty[j],lwd=trend.lwd[j],delta=trend.delta,...)
                        }
                } else {
                    confband(trend.formula,res[,2],res[,3],res[,1],col=Col(trend.col,trend.alpha),lty=trend.lty,lwd=trend.lwd,polygon=trend.join,...)
                }
            } else {
                tf <- getoutcome(trend.formula)
                if (is.list(tf)) {
                    trend.formula <- update(trend.formula,toformula(y,"."))
                }
                if (!is.null(tau)) {
                    ##if (!require(quantreg)) stop("Install 'quantreg'")
                    suppressWarnings(r1 <- quantreg::rq(trend.formula,data=data,tau=tau))
                    newdata <- data.frame(seq(min(X,na.rm=TRUE),max(X,na.rm=TRUE),length.out=100))
                    names(newdata) <- x
                    pr <- predict(r1,newdata=newdata,interval="confidence",level=level)
                    ##confband(xx,pr[,3],pr[,2],polygon=TRUE,col=Col(trend.col,trend.alpha),border=FALSE)
                    for (i in seq_along(tau))
                        lines(newdata[,1],pr[,i],col=trend.col,lwd=trend.lwd,lty=trend.lty)
                } else {
                    l1. <- lm(trend.formula,data)
                    l1 <- summary(estimate(l1.,id=data[,id]),level=level)
                    xy <- plotConf(l1.,vcov=vcov(l1),data=data,partres=FALSE,plot=FALSE,level=level,...)
                    xx <- xy$x
                    pr <- xy$predict$fit
                    if (is.factor(xx)) {
                        xx <- char2num(as.character(xx))
                        if (trend.jitter>0) xx <- jitter(xx,trend.jitter)                        
                        confband(xx,pr[,3],pr[,2],pr[,1],col=trend.col,lwd=2)
                    } else {
                        confband(xx,pr[,3],pr[,2],polygon=TRUE,col=Col(trend.col,trend.alpha),border=FALSE)
                        lines(xx,pr[,1],col=trend.col,lwd=trend.lwd,lty=trend.lty)
                    }
                }
            }
        }
    }
    return(invisible(list(Y,X)))
}

##' Extract all possible paths from one variable to another connected component
##' in a latent variable model. In an estimated model the effect size is
##' decomposed into direct, indirect and total effects including approximate
##' standard errors.
##'
##' @title Extract pathways in model graph
##' @export
##' @aliases path effects path.lvm effects.lvmfit
##' @aliases totaleffects
##' @seealso \code{children}, \code{parents}
##' @return If \code{object} is of class \code{lvmfit} a list with the following
##' elements is returned \item{idx}{ A list where each element defines a
##' possible pathway via a integer vector indicating the index of the visited
##' nodes. } \item{V }{ A List of covariance matrices for each path. }
##' \item{coef }{A list of parameters estimates for each path} \item{path }{A
##' list where each element defines a possible pathway via a character vector
##' naming the visited nodes in order.  } \item{edges }{Description of 'comp2'}
##'
##' If \code{object} is of class \code{lvm} only the \code{path} element will be
##' returned.
##'
##' The \code{effects} method returns an object of class \code{effects}.
##' @note For a \code{lvmfit}-object the parameters estimates and their
##' corresponding covariance matrix are also returned.  The
##' \code{effects}-function additionally calculates the total and indirect
##' effects with approximate standard errors
##' @author Klaus K. Holst
##' @keywords methods models graphs
##' @examples
##'
##' m <- lvm(c(y1,y2,y3)~eta)
##' regression(m) <- y2~x1
##' latent(m) <- ~eta
##' regression(m) <- eta~x1+x2
##' d <- sim(m,500)
##' e <- estimate(m,d)
##'
##' path(Model(e),y2~x1)
##' parents(Model(e), ~y2)
##' children(Model(e), ~x2)
##' children(Model(e), ~x2+eta)
##' effects(e,y2~x1)
##' ## All simple paths (undirected)
##' path(m,y1~x1,all=TRUE)
##'
##' @usage
##' \method{path}{lvm} (object, to = NULL, from, all=FALSE, ...)
##' \method{effects}{lvmfit} (object, to, from, ...)
##' @param object Model object (\code{lvm})
##' @param to Outcome variable (string). Alternatively a formula specifying
##' response and predictor in which case the argument \code{from} is ignored.
##' @param from Response variable (string), not necessarily directly affected by
##' \code{to}.
##' @param all If TRUE all simple paths (in undirected graph) is returned
##' on/off.
##' @param \dots Additional arguments to be passed to the low level functions
##' @export
path <- function(object,...) UseMethod("path")

##' @export
path.lvmfit <- function(object,to=NULL,from,...) {
    mypath <- pathM(Model(object)$M,to,from,...)
    cc <- coef(object,type=9,labels=FALSE) ## All parameters (fixed and variable)
    cc0 <- coef(object,type=2) ## Estimated parameters
    i1 <- na.omit(match(rownames(cc),rownames(cc0)))
    idx.cc0 <-  which(rownames(cc)%in%rownames(cc0)); ## Position of estimated parameters among all parameters
    S <- matrix(0,nrow(cc),nrow(cc)); rownames(S) <- colnames(S) <- rownames(cc)
    V <- object$vcov
    S[idx.cc0,idx.cc0] <- V[i1,i1]  ## "Covariance matrix" of all parameters

    idx <- list()
    coefs <- list()
    V <- list()
    for (i in seq_along(mypath)) {
        xx <- mypath[[i]]
        ii <- c()
        for (j in seq_len(length(xx)-1)) {
            st <- paste0(xx[j+1], lava.options()$symbol[1], xx[j])
            ii <- c(ii, match(st,rownames(cc)))
        }
        idx <- c(idx, list(ii))
        V <- c(V, list(S[ii,ii]))
        coefs <- c(coefs, list(cc[ii]))
    }

    edges <- list()
    for (i in seq_along(mypath)) {
        p0 <- mypath[[i]]
        ee <- c()
        for (i in seq_len(length(p0)-1)) {
            ee <- c(ee, paste(p0[i],p0[i+1],sep="~"))
        }
        edges <- c(edges, list(ee))
    }
    res <- list(idx=idx,V=V,coef=coefs, path=mypath, edges=edges)
    return(res)
}

##' @export
path.lvm <- function(object,to=NULL,from,all=FALSE,...) {
    pathM(object$M,to=to,from=from,all=all,...)
}

##' @export
path.graphNEL <- function(object,to,from,...) {
    if (inherits(to,"formula")) {
        fvar <- extractvar(to)
        if (length(fvar$x)==1 & length(fvar$y)==1)
            return(path(object,to=fvar$y,from=fvar$x))
        res <- list()
        for (y in fvar$y) {
            for (x in fvar$x) {
                cat("x=",x, " y=",y, "\n")
                res <- c(res, list(path(object,to=y,from=x)))
            }
        }
        return(res)
    }
    ff <- function(g,from=1,to=NULL,res=list()) {
        M <- graph::edgeMatrix(g)
        i1 <- which(M[1,]==from)
        for (i in i1) {
            e <- M[,i]; newto <- e[2];
            if (is.null(to) || M[2,i]==to) {
                res <- c(res, list(M[,i]))
            }
            newpath <- ff(g,from=newto,to=to,list())
            if (length(newpath)>0)
                for (j in seq_along(newpath)) {
                    if (is.null(to) || (tail(newpath[[j]],1)==to))
                        res <- c(res, list(c(M[,i],newpath[[j]][-1])))
                }
        }
        return(res)
    }
    idxfrom <- ifelse(is.numeric(from),from,which(from==graph::nodes(object)))
    ##M <- as(object,"matrix")
    ##reachable <- acc(M,graph::nodes(object)[idxfrom])
    reachable <- graph::acc(object,graph::nodes(object)[idxfrom])[[1]]

    if (is.null(to)) {
        idxto <- reachable
    } else {
        idxto <- ifelse(is.numeric(to),to,which(to==graph::nodes(object)))
    }

    if (!(graph::nodes(object)[idxto] %in% names(reachable)))
        ##    return(structure(list(),to=to[1],from=from[1]))
        return(NULL)
    ##    stop("No directional relationship between variables")

    mypaths <- ff(object,idxfrom,idxto)
    res <- list()
    for (i in seq_along(mypaths)) {
        res <- c(res, list(graph::nodes(object)[mypaths[[i]]]))
    }
    return(res)
}


pathM <- function(M,to,from,all=FALSE,...) {
    nn <- colnames(M)
    if (inherits(to,"formula")) {
        fvar <- extractvar(to)
        if (length(fvar$x)==1 & length(fvar$y)==1)
            return(pathM(M,to=fvar$y,from=fvar$x,all=all))
        res <- list()
        for (y in fvar$y) {
            for (x in fvar$x) {
                cat("x=",x, " y=",y, "\n")
                res <- c(res, list(pathM(M,to=y,from=x,all=all)))
            }
        }
        return(res)
    }
    if (all) { ## Get all simple paths
        res <- simplePaths(to,from,from,M,list())
        return(res)
    }


    ff <- function(g,from=1,to=NULL,res=list()) {
        i1 <- which(M[from,]==1)
        for (i in i1) {
            ##      e <- M[,i]; newto <- e[2];
            if (is.null(to) || i==to) {
                res <- c(res, list(c(from,i)))
            }
            newpath <- ff(g,from=i,to=to,list())
            if (length(newpath)>0)
                for (j in seq_along(newpath)) {
                    if (is.null(to) || (tail(newpath[[j]],1)==to))
                        res <- c(res, list(c(c(from,i),newpath[[j]][-1])))
                }
        }
        return(res)
    }
    idxfrom <- ifelse(is.numeric(from),from,which(from==nn))
    reachable <- acc(M,nn[idxfrom])

    if (is.null(to)) {
        idxto <- reachable
    } else {
        idxto <- ifelse(is.numeric(to),to,which(to==nn))
    }

    if (!(nn[idxto] %in% reachable))
        return(NULL)
    ##    stop("No directional relationship between variables")

    mypaths <- ff(M,idxfrom,idxto)
    res <- list()
    for (i in seq_along(mypaths)) {
        res <- c(res, list(nn[mypaths[[i]]]))
    }
    return(res)
}



## Find all simple paths (no cycles) in an undirected graph
simplePaths <- function(target,currentpath,visited,adjmat,allpaths) {
    lastnode <- currentpath[length(currentpath)]
    A <- (adjmat+t(adjmat))>0
    if (lastnode==target) {
        allpaths <- c(allpaths,list(currentpath))
    } else {
        for (neighbour in rownames(adjmat)[which(A[,lastnode])]) {
            if (!(neighbour%in%visited)) {
                currentpath <- c(currentpath,neighbour)
                visited <- c(visited,neighbour)
                allpaths <- simplePaths(target,currentpath,visited,adjmat,allpaths)
                visited <- setdiff(visited,neighbour)
                currentpath <- currentpath[-length(currentpath)]
            }
        }
    }
    return(allpaths)
}

##' Generic method for calculating bootstrap statistics
##'
##' @title Generic bootstrap method
##' @param x Model object
##' @param \dots Additional arguments
##' @seealso \code{bootstrap.lvm} \code{bootstrap.lvmfit}
##' @author Klaus K. Holst
##' @export
bootstrap <- function(x,...) UseMethod("bootstrap")

##' Calculate bootstrap estimates of a lvm object
##'
##' Draws non-parametric bootstrap samples
##'
##' @param x \code{lvm}-object.
##' @param R Number of bootstrap samples
##' @param data The data to resample from
##' @param fun Optional function of the (bootstrapped) model-fit defining the
##' statistic of interest
##' @param control Options to the optimization routine
##' @param p Parameter vector of the null model for the parametric bootstrap
##' @param parametric If TRUE a parametric bootstrap is calculated. If FALSE a
##' non-parametric (row-sampling) bootstrap is computed.
##' @param bollenstine Bollen-Stine transformation (non-parametric bootstrap) for bootstrap hypothesis testing.
##' @param constraints Logical indicating whether non-linear parameter
##' constraints should be included in the bootstrap procedure
##' @param sd Logical indicating whether standard error estimates should be
##' included in the bootstrap procedure
##' @param mc.cores Optional number of cores for parallel computing. If omitted future.apply will be used (see future::plan)
##' @param future.args arguments to future.apply::future_lapply
##' @param estimator String definining estimator, e.g. 'gaussian' (see
##' \code{estimator})
##' @param weights Optional weights matrix used by \code{estimator}
##' @param \dots Additional arguments, e.g. choice of estimator.
##' @aliases bootstrap.lvmfit
##' @usage
##'
##' \method{bootstrap}{lvm}(x,R=100,data,fun=NULL,control=list(),
##'                           p, parametric=FALSE, bollenstine=FALSE,
##'                           constraints=TRUE,sd=FALSE, mc.cores,
##'                           future.args=list(future.seed=TRUE),
##'                           ...)
##'
##' \method{bootstrap}{lvmfit}(x,R=100,data=model.frame(x),
##'                              control=list(start=coef(x)),
##'                              p=coef(x), parametric=FALSE, bollenstine=FALSE,
##'                              estimator=x$estimator,weights=Weights(x),...)
##'
##' @return A \code{bootstrap.lvm} object.
##' @author Klaus K. Holst
##' @seealso \code{\link{confint.lvmfit}}
##' @keywords models regression
##' @examples
##' m <- lvm(y~x)
##' d <- sim(m,100)
##' e <- estimate(lvm(y~x), data=d)
##' \donttest{ ## Reduce Ex.Timings
##' B <- bootstrap(e,R=50,mc.cores=1)
##' B
##' }
##' @export
bootstrap.lvm <- function(x, R = 100, data, fun = NULL, control = list(),
                          p, parametric = FALSE, bollenstine = FALSE,
                          constraints = TRUE, sd = FALSE, mc.cores,
                          future.args=list(future.seed=TRUE),
                          ...) {
  coefs <- sds <- c()
  on.exit(list(coef = coefs[-1, ], sd = sds[-1, ], coef0 = coefs[1, ], sd0 = sds[1, ], model = x))
  pb <- progressr::progressor(steps = R)
  pmis <- missing(p)
  bootfun <- function(i) {
    if (i == 0) {
      d0 <- data
    } else {
      if (!parametric | pmis) {
        d0 <- data[sample(seq_len(nrow(data)), replace = TRUE), ]
      } else {
        d0 <- sim(x, p = p, n = nrow(data))
      }
    }
    suppressWarnings(e0 <- estimate(x, data = d0, control = control, messages = 0, index = FALSE, ...))
    pb()
    if (!is.null(fun)) {
      coefs <- fun(e0)
      newsd <- NULL
    } else {
      coefs <- coef(e0)
      newsd <- c()
      if (sd) {
        newsd <- e0$coef[, 2]
      }
      if (constraints & length(constrain(x)) > 0) {
        cc <- constraints(e0, ...)
        coefs <- c(coefs, cc[, 1])
        names(coefs)[seq(length(coefs) - length(cc[, 1]) + 1, length(coefs))] <- rownames(cc)
        if (sd) {
          newsd <- c(newsd, cc[, 2])
        }
      }
    }
    return(list(coefs = coefs, sds = newsd))
  }
  if (bollenstine) {
    e0 <- estimate(x, data = data, control = control, messages = 0, index = FALSE, ...)
    mm <- modelVar(e0)
    mu <- mm$xi
    Y <- t(t(data[, manifest(e0)]) - as.vector(mu))
    Sigma <- mm$C
    S <- (ncol(Y) - 1) / ncol(Y) * var(Y)
    sSigma <- with(eigen(Sigma), vectors %*% diag(sqrt(values), ncol = ncol(vectors)) %*% t(vectors))
    isS <- with(eigen(S), vectors %*% diag(1 / sqrt(values), ncol = ncol(vectors)) %*% t(vectors))
    data <- as.matrix(Y) %*% (isS %*% sSigma)
    colnames(data) <- manifest(e0)
  }

  i <- 0
  if (!missing(mc.cores)) {
    res <- parallel::mclapply(0:R, bootfun, mc.cores=mc.cores)
  } else {
    res <- do.call(future_lapply, c(list(0:R, bootfun), future.args))
  }

  coefs <- matrix(unlist(lapply(res, function(x) x$coefs)), nrow = R + 1, byrow = TRUE)
  nn <- names(res[[1]]$coefs)
  if (!is.null(nn)) colnames(coefs) <- nn
  sds <- NULL
  if (sd) {
    sds <- matrix(unlist(lapply(res, function(x) x$sds)), nrow = R + 1, byrow = TRUE)
  }

  if (!is.null(fun)) {
    rownames(coefs) <- c()
    res <- list(coef = coefs[-1, , drop = FALSE], coef0 = coefs[1, ], model = x)
  } else {
    colnames(coefs) <- names(res[[1]]$coefs)
    rownames(coefs) <- c()
    if (sd) colnames(sds) <- colnames(coefs)
    res <- list(coef = coefs[-1, , drop = FALSE], sd = sds[-1, , drop = FALSE], coef0 = coefs[1, ], sd0 = sds[1, ], model = x, bollenstine = bollenstine)
  }
  class(res) <- "bootstrap.lvm"
  return(res)
}

##' @export
bootstrap.lvmfit <- function(x, R = 100, data = model.frame(x),
                             control = list(start = coef(x)),
                             p = coef(x), parametric = FALSE, bollenstine = FALSE,
                             estimator = x$estimator, weights = Weights(x), ...) {
  bootstrap.lvm(Model(x), R = R, data = data, control = control, estimator = estimator, weights = weights, parametric = parametric, bollenstine = bollenstine, p = p, ...)
}

##' @export
"print.bootstrap.lvm" <- function(x, idx, level = 0.95, ...) {
  cat("Non-parametric bootstrap statistics (R=", nrow(x$coef), "):\n\n", sep = "")
  uplow <- (c(0, 1) + c(1, -1) * (1 - level) / 2)
  nn <- paste(uplow * 100, "%")
  c1 <- t(apply(x$coef, 2, function(x) c(mean(x), sd(x), quantile(x, uplow))))

  c1 <- cbind(x$coef0, c1[, 1] - x$coef0, c1[, -1, drop = FALSE])
  colnames(c1) <- c("Estimate", "Bias", "Std.Err", nn)
  if (missing(idx)) {
    print(format(c1, ...), quote = FALSE)
  } else {
    print(format(c1[idx, , drop = FALSE], ...), quote = FALSE)
  }
  if (length(x$sd) > 0) {
    c2 <- t(apply(x$sd, 2, function(x) c(mean(x), sd(x), quantile(x, c(0.025, 0.975)))))
    c2 <- cbind(c2[, 1], c2[, 1] - x$sd0, c2[, -1])
    colnames(c2) <- c("Estimate", "Bias", "Std.Err", "2.5%", "97.5%")
    cat("\nStandard errors:\n")
    if (missing(idx)) {
      print(format(c2, ...), quote = FALSE)
    } else {
      print(format(c2[idx, , drop = FALSE], ...), quote = FALSE)
    }
  }
  cat("\n")
  invisible(x)
}

mgstart <- function(x,p) {
    if (is.list(p)) {
        npar <- with(x, npar+npar.mean+length(unlist(expar)))
        pos <- modelPar(x,seq(npar))$p
        start <- matrix(NA,ncol=npar,nrow=x$ngroup)
        startl <- lapply(pos, function(x) rep(NA,length(x)))
        for (i in seq_len(x$ngroup)) {
            p0 <- p[[i]]
            pos0 <- pos[[i]]            
            if (!is.null(names(p0))) {
                ii <- parpos(Model(x)[[i]],names(p0))
                startl[[i]][ii] <- p0[names(ii)]
            } else {
                ii <- seq(min(length(p0),length(startl[[i]])))
                startl[[i]][ii] <- p0[ii]
            }
            start[i,pos0] <- startl[[i]]
        }
        start0 <- apply(start,2,function(x) mean(x,na.rm=TRUE))
    }
    return(start0)
}


###{{{ starter.multigroup

##' @export
starter.multigroup <- function(x, starterfun=startvalues2, meanstructure=TRUE,messages=0,...) {
  ## Initial values:
  W <- c() ## Weight-vector
  s <- list()
  for (i in seq_len(x$ngroup)) {
    mydata <- x$data[[i]][,manifest(x$lvm[[i]]),drop=FALSE]
    W <- c(W, nrow(mydata))
    if (nrow(mydata)<3) {
      ii <- index(x$lvm[[i]])
      nn <- ifelse(meanstructure, ii$npar+ii$npar.mean, ii$npar)
      s0 <- rep(1,nn)
    }
    else {
      S <- x$samplestat[[i]]$S
      mu <- if (meanstructure) x$samplestat[[i]]$mu else NULL;
      ##      S <- cov(mydata); mu <- if (meanstructure) colMeans(mydata) else NULL;
      s0 <- do.call(starterfun, list(x$lvm[[i]], S=S, mu=mu,messages=messages))
    }
    s <- c(s, list(s0))
  }

  Wtotal <- sum(W); W <- W/Wtotal

  pg <- vector("list", x$npar); for (i in seq_len(length(pg))) pg[[i]] <- rep(0,x$ngroup)
  meang <- vector("list", x$npar.mean); for (i in seq_len(length(meang))) meang[[i]] <- rep(0,x$ngroup)

  for (i in seq_len(x$ngroup)) {
    pp <- modelPar(x$lvm[[i]],s[[i]])
    pos <- sapply(x$parlist[[i]], function(y) char2num(substr(y,2,nchar(y))))
    for (j in seq_len(length(pos)))
      pg[[ pos[j] ]][i] <-  pp$p[j]

    pos <- sapply(x$meanlist[[i]], function(y) char2num(substr(y,2,nchar(y))))
    ptype <- sapply(x$meanlist[[i]], function(y) substr(y,1,1)=="m")
    if (!(any(ptype)))
      pos <- NULL
    else
      pos <- pos[ptype]
    if (length(pos)>0)
    for (j in seq_len(length(pos))) {
      meang[[ pos[j] ]][i] <-  pp$meanpar[j]
    }
  }

  ## Weighted average
  wp <- unlist(lapply(pg, function(y) {
    ppos <- !is.na(y)
    myweight <- W[ppos]/sum(W[ppos])
    sum(y[ppos]*myweight)
  }))
  wmean <- unlist(lapply(meang, function(y) {
    ppos <- !is.na(y)
    myweight <- W[ppos]/sum(W[ppos])
    sum(y[ppos]*myweight)
  }))
  res <- c(wmean,wp)
  res[!is.finite(res) | is.nan(res) | is.na(res) | is.complex(res)] <- .5
  return(as.numeric(res))
}

###}}}

###{{{ startmean

startmean <- function(x,p,mu) {
  if (is.null(mu))
    return(p)
  meanpar <- numeric(index(x)$npar.mean)
  mymeans <- vars(x)[index(x)$v1==1]
  midx <- na.omit(match(names(mu),mymeans))
  meanpar[midx] <- mu[midx]
  AP <- matrices(x,p,meanpar)
  nu <- numeric(length(vars(x)))
  nu[vars(x)%in%manifest(x)] <- mu
  meanstart <- ((diag(nrow=nrow(AP$A))-t(AP$A))%*%nu)[index(x)$v1==1]
  names(meanstart) <- vars(x)[index(x)$v1==1]
  return( c(meanstart, p) )
}

###}}}

###{{{ startvalues3

`startvalues3` <-
function(x, S, debug=FALSE, tol=1e-6,...) {
  S <- reorderdata.lvm(x,S)
  if (nrow(S)!=length(manifest(x))) stop("Number of observed variables in data and models does not agree.")
  A <- t(index(x)$M) ## Adjacency matrix
  m <- nrow(A) ## Number of variables
  A0 <- t(index(x)$M0) ## Adjacency matrix (without fixed parameters)

  obs.idx <- index(x)$obs.idx;
  ##obs.idx <- as.vector(J%*%(seq_len(m)));
  latent.idx <- setdiff(seq_len(m), obs.idx)
  lat <- colnames(A)[latent.idx]

  exo.idx <- index(x)$exo.idx ## match(exogenous(x),vars(x))
  exo.idxObs <- index(x)$exo.obsidx ##match(exogenous(x),manifest(x))

  AP0 <- moments(x, rep(0,index(x)$npar))
  newP <- t(AP0$P)
  newA <- t(AP0$A)
  fixed <- t(x$fix)

  for (i in latent.idx) {
    fix.idx <- colnames(fixed)[which(!is.na(t(fixed[,i])))[1]]
    lambda0 <- newA[fix.idx,i]
    rel.idx <- which(A0[,i]==1)
    rel.all <- which(A[,i]==1)
    rel.pos <-  colnames(A)[rel.all]
    ## Estimation of lambda (latent -> endogenous)
    for (j in rel.idx) {
      lambda <- lambda0*S[fix.idx, j]/S[fix.idx,fix.idx]
      newA[j,i] <- lambda
    }
    lambdas <- newA[rel.pos,i]


    ## Estimation of beta (covariate -> latent)
    exo2latent <- which(A0[i,exo.idx]==1)
    exo.pos <- colnames(S)[exo.idxObs[exo2latent]]
    varX.eta <- S[exo.pos, exo.pos]
    InvvarX.eta <- Inverse(varX.eta,tol=1e-3)
    rel.pos <- setdiff(rel.pos,lat)
    covXY <- S[exo.pos, rel.pos,drop=FALSE]
    beta <- 0
    for (j in seq_len(length(rel.pos)))
      beta <- beta + 1/lambdas[j]*InvvarX.eta %*% covXY[,j]
    beta <- beta/length(rel.pos)

    for (k in seq_len(length(exo.pos))) {
      if (A0[i,exo.pos[k]]==1) {
        newA[i,exo.pos[k]] <- beta[k]
      }
    }

    beta.eta <- matrix(newA[i,exo.pos], ncol=1)

    ## Estimation of  zeta^2 (variance of latent variable)
    betavar <- matrix(beta.eta,nrow=1)%*%varX.eta%*%beta.eta

    zetas <- c()
    for (r1 in seq_len(length(rel.pos)-1))
      for (r2 in seq(r1+1,length(rel.pos))) {
        zetas <- c(zetas, S[rel.pos[r1], rel.pos[r2]]/ (lambdas[r1]*lambdas[r2]) - betavar)
      }
    zeta <- mean(zetas)

    newP[i,i] <- zeta
    for (j in rel.all) {
      pos <- colnames(newA)[j]
      vary <- S[pos,pos] - newA[pos,i]^2*(zeta+betavar)
      newP[pos,pos] <- ifelse(vary<0.25,0.25,vary)
    }

  }
  Debug(list("start=",start), debug)
  start <- pars(x, A=t(newA), P=newP)
  return(start)
}

###}}} startvalues3

###{{{ startvalues2


## Estimate sub-models (measurement models)
##' @export
`startvalues2` <-
  function(x, S, mu=NULL, debug=FALSE, messages=0,...) {
    if (messages>0) cat("Obtaining start values...\n")
    S <- reorderdata.lvm(x,S)
    ss <- startvalues(x,S)
    Debug(list("ss=",ss),debug);
    g <- measurement(x,messages=0)
    keep <- c()
    if (length(g)>1) {
      for (i in seq_len(length(g))) {
        if (length(endogenous(g[[i]]))>2)
          keep <- c(keep,i)
      }
      g <- g[keep]
    }
    if (length(g)<2)
      return(startmean(x,ss,mu=mu))
    ## if (messages>0) cat("Fitting marginal measurement models...\n")
    op <- options(warn=-1)
    e <- lapply(g, function(y) {
        estimate(y, data=list(S=S[manifest(y),manifest(y),drop=FALSE], mu=mu[manifest(y)], n=100), control=list(meanstructure=FALSE, starterfun="startvalues", estimator="Simple", method="nlminb1"), optcontrol=list(), debug=FALSE, messages=0)
    })
    for (l in e) {
      ##    a <- coef(l$estimate)[,1]
      a <- coef(l)
      for (i in seq_len(length(a))) {
        pos <- match(names(a)[i],names(ss))
        if (!is.na(pos))
          ss[pos] <- a[i]
    }
    }
    options(op)
    startmean(x,ss,mu=mu)
  }

###}}} startvalues2

###{{{ startvalues0

##' @export
startvalues1 <- function(x,S,mu=NULL,tol=1e-6,delta=1e-6,...) {
    p0 <- startvalues(x,S,mu,...)
    p0[index(x)$npar.mean+variances(x)] <- 0.1
    p0[index(x)$npar.mean+offdiags(x)] <- 0
    p0
}

startvalues00 <- function(x,S,mu=NULL,tol=1e-6,delta=1e-6,...) {
    p0 <- startvalues(x,S,mu,...)
    p0 <- numeric(length(p0))
    P0 <- x$cov*1
    ##P0[!is.na(x$covfix)] <-
    ##P0 <- x$covfix; P0[is.na(P0)] <- 0
    ##diag(P0)[index(x)$endo.idx] <- diag(S)[index(x)$endo.obsidx]/2
    ##lu <- min(diag(P0)[index(x)$endo.idx])/2
    ##    diag(P0)[] <- 0.1
    ## diag(P0)[index(x)$endo.idx] <- 1
    diag(P0)[index(x)$eta.idx] <- 0.1 ##mean(diag(S)[index(x)$endo.idx])/2
    ee <- eigen(P0)
    tol <- 1e-6
    ii <- ee$values
    ii[ee$values<tol] <- tol
    P0 <- ee$vectors%*%diag(ii,nrow=length(ii))%*%t(ee$vectors)
    A0 <- P0
    pp <- pars(x,A=t(A0),P=P0,v=rep(0,index(x)$npar.mean))
    idx <- index(x)$npar.mean + c(offdiags(x),variances(x))
    p0[idx] <- pp[idx]
    return(p0)
}


##' @export
startvalues0 <- function(x,S,mu=NULL,tol=1e-6,delta=1e-6,...) {
    p0 <- startvalues(x,S,mu,...)
    A <- t(index(x)$M) ## Adjacency matrix
    P0 <- A0 <- matrix(0,nrow(A),ncol(A))
    A0[,index(x)$eta.idx] <- A[,index(x)$eta.idx]
    diag(P0)[index(x)$endo.idx] <- diag(S)[index(x)$endo.obsidx]/3
    ##lu <- min(diag(P0)[index(x)$endo.idx])
    lu <- 0.9
    diag(P0)[index(x)$eta.idx] <- lu##mean(diag(S)[index(x)$endo.idx])/2
    pp <- pars(x,A=t(A0),P=P0,v=rep(1,length(index(x)$vars)))
    pp[pp==1] <- p0[pp==1]
    ## nu <- numeric(length(vars(x)))
    ## if (!is.null(mu)) {
    ##     nu[vars(x)%in%manifest(x)] <- mu
    ##     (diag(nrow(A0))-t(A0))%*%nu
    ##     meanstart <- solve(diag(nrow(A0))+t(A0))%*%nu
    ##     meanstart <- meanstart[which(is.na(x$mean))]
    ##     if (length(meanstart)>0)
    ##         pp[seq(length(meanstart))] <- meanstart
    ## }
    names(pp) <- coef(x, messages=0, fixed=FALSE, mean=TRUE)[seq_len(length(pp))]
    pp[!is.finite(pp) | is.nan(pp) | is.na(pp)] <- 0.01
    return(pp)
}

###}}} startvalues0

###{{{ startvalues

## McDonald & Hartmann, 1992
##' @export
startvalues <-
function(x, S, mu=NULL, debug=FALSE, messages=lava.options()$messages, tol=1e-6, delta=1e-6,...) {
  ## As proposed by McDonald & Hartmann, 1992.
  ## Implementation based on John Fox's implementation in the 'sem' R-package
  S <- reorderdata.lvm(x,S)
  if (nrow(S)!=length(manifest(x))) stop("Number of observed variables in data and models does not agree.")
  J <- index(x)$J ## Manifest selection
  P0 <- index(x)$P0 ## covariance 'adjacency'
  A <- t(index(x)$M) ## Adjacency matrix
  n <- nrow(S) ## Number of manifest variables
  m <- nrow(A) ## Number of variables
  A0 <- t(index(x)$M0) ## Adjacency matrix (without fixed parameters)
  obs.idx <- as.vector(J%*%(seq_len(m)));  latent.idx <- setdiff(seq_len(m), obs.idx)
  s <- sqrt(diag(S))
  suppressWarnings(R <- (cov2cor(S))) ## S/outer(s,s)
  C <- P0
  Debug(list("obs.idx", obs.idx), debug)
  C[obs.idx,obs.idx] <- R
  ## Estimates of covariance between latent and manifest variables
  Debug((C), debug)
  for (i in latent.idx) {
    inRelation <- A[obs.idx,i]==1
    for (j in seq_len(length(obs.idx))) {
      Debug((j), debug)
      C[obs.idx[j],i] <- C[i,obs.idx[j]] <- if (any(inRelation)) {
        numerator <- sum(R[j, which(inRelation)])
        denominator <- sqrt(sum(R[which(inRelation), which(inRelation)]))
        numerator/denominator ## as proposed by McDonald & Hartmann
      } else {
        runif(1, .3, .5) ## No arrows => small random covariance
      }
    }
  }
  ## Estimates of covariance between latent variables
  for (i in latent.idx) {
    for (j in latent.idx) {
      C[i,j] <- C[j,i] <-
        if (i==j) {
          1
        } else {
          inRelation.i <- A[obs.idx, i]==1
          inRelation.j <- A[obs.idx, j]==1
          if ((any(inRelation.i)) | (any(inRelation.j))) {
            numerator <- sum(R[which(inRelation.i), which(inRelation.j)])
            denominator <- sqrt( sum(R[which(inRelation.i), which(inRelation.i)])
                                * sum(R[which(inRelation.j), which(inRelation.j)]))
            numerator/(denominator+0.01) ## Avoid division by zero
          } else {
            runif(1, .3, .5)
          }
        }
    }
  }
  if (debug) {
    print("C="); print(C);
  }
  Ahat <- matrix(0,m,m)
  C[is.nan(C)] <- 0
  for (j in seq_len(m)) { ## OLS-estimates
    relation <- A[j,]==1
    if (!any(relation)) next
    Ahat[j, relation] <- tryCatch(Inverse(C[relation,relation] + diag(nrow=sum(relation))*delta,tol=1e-3) %*% C[relation,j], error=function(...) 0)
  }
  Ahat[obs.idx,] <- Ahat[obs.idx,]*matrix(s, n, m)
  Ahat[,obs.idx] <- Ahat[,obs.idx]/matrix(s, m, n, byrow=TRUE)
  Chat <- C
  Chat[obs.idx,] <- Chat[obs.idx,]*matrix(s,n,m)  ##
  Chat[,obs.idx] <- Chat[,obs.idx]*matrix(s,m,n,byrow=TRUE)  ##
  Phat <- (diag(m)-Ahat)%*%Chat%*%t(diag(m)-Ahat)
  ##diag(Phat) <- abs(diag(Phat))
  ## Guarantee PD-matrix:
  Phat[is.nan(Phat) | is.na(Phat)] <- 0
  diag(Phat)[diag(Phat)==0] <- 1
  eig <- eigen(Phat)
  L <- abs(eig$values); L[L<1e-3] <- 1e-3
  Phat <- eig$vectors%*%diag(L,ncol=ncol(eig$vectors))%*%t(eig$vectors)

  Debug(list("start=",start), debug)
  start <- pars(x, A=t(Ahat*A0), P=(Phat*P0))
  names(start) <- coef(x, messages=0, fixed=FALSE, mean=FALSE)[seq_len(length(start))]
  res <- startmean(x,start,mu)
  res[!is.finite(res) | is.nan(res) | is.na(res)] <- 1
  res
}

###}}} startvalues

##' @title Generalized matrix inverse
##' @export
##' @param X nxn matrix
##' @param tol tolerance for pseudo inverse
##' @param det logical, if true the determinant is returned
##' @param names preserve dimnames
##' @param chol use Cholesky decomposition for calculating inverse otherwise SVD
##' @param symmetric set to true if matrix is symmetric
Inverse <- function(X,tol=lava.options()$itol,det=TRUE,names=!chol,chol=FALSE,symmetric=FALSE) {
    n <- NROW(X)
    if (n==1L) {
        res <- 1/X
        if (det) attributes(res)$det <- X
        if (chol) attributes(res)$chol <- X
        return(res)
    }
    if (chol) {
        L <- chol(X)
        res <- chol2inv(L)
        if (det) attributes(res)$det <- prod(diag(L)^2)
        if (chol) attributes(res)$chol <- X
    } else {
        if(symmetric){
            decomp <- eigen(X, symmetric = TRUE)
            D <- decomp$values
            U <- decomp$vectors
            V <- decomp$vectors
        }else{
            X.svd <- svd(X)
            U <- X.svd$u
            V <- X.svd$v
            D <- X.svd$d
        }
        id0 <- numeric(n)
        idx <- which(abs(D)>tol)
        id0[idx] <- 1/D[idx]
        res <- V%*%diag(id0,nrow=length(id0))%*%t(U)

        if (det)
            attributes(res)$det <- prod(D[D>tol])
        attributes(res)$pseudo <- (length(idx)<n)
        attributes(res)$minSV <- min(D)

    }
    if (names && !is.null(colnames(X))) dimnames(res) <- list(colnames(X),colnames(X))
    return(res)
}

###{{{ modelVar

##' @export
`modelVar` <-
  function(x,p,...) UseMethod("modelVar")

##' @export
modelVar.lvmfit <- function(x, p=pars(x), ...) modelVar(Model(x),p=p,...)

##' @export
modelVar.lvm <- function(x,p,data,...) {
  pp <- modelPar(x,p)
  res <- moments(x, p=p, data=data,...)
  attr(res, "pars") <- pp$p
  attr(res, "meanpar") <- pp$meanpar
  attr(res, "epar") <- pp$epar
  res
}
###}}} modelVar

gaussian_method.lvm <- "nlminb2"
`gaussian_objective.lvm` <-
  function(x,p,data,S,mu,n,...) {
    mp <- modelVar(x,p=p,data=data,...)
    C <- mp$C ## Model specific covariance matrix
    xi <- mp$xi ## Model specific mean-vector
    if (!lava.options()$allow.negative.variance && any(diag(mp$P)<0)) return(NaN)

    iC <- Inverse(C,det=TRUE, symmetric = TRUE)
    detC <- attributes(iC)$det
    if (n<2) {
      z <- as.numeric(data-xi)
      val <- log(detC) + tcrossprod(z,crossprod(z,iC))[1]
      return(0.5*val)
    }
    if (!is.null(mu)){
      W <- suppressMessages(crossprod(rbind(mu-xi)))
      T <- S+W
    } else {
      T <- S
    }
    res <- n/2*log(detC) + n/2*tr(T%*%iC) ## Objective function (Full Information ML)
    ## if (any(attr(iC,"lambda")<1e-16)) res <- res-1e2
    return(res)
  }

`gaussian_hessian.lvm` <- function(x,p,n,...) {
  dots <- list(...); dots$weights <- NULL
  do.call("information", c(list(x=x,p=p,n=n),dots))
}

gaussian_gradient.lvm <-  function(x,p,data,S,mu,n,...) {
  dots <- list(...); dots$weights <- NULL
  if (n>2) data <- NULL
  val <- -gaussian_score.lvm(x,p=p,S=S,mu=mu,n=n,data=data,reindex=FALSE,...)
  if (!is.null(nrow(val))) {
    val <- colSums(val)
  }
  val
}

gaussian_score.lvm <- function(x, data, p, S, n, mu=NULL, weights=NULL, debug=FALSE, reindex=FALSE, mean=TRUE, constrain=TRUE, indiv=FALSE,...) {

  if (!is.null(data)) {
    if ((nrow(data)<2 | !is.null(weights))| indiv)
    {
      mp <- modelVar(x,p,data=data[1,])
      iC <- Inverse(mp$C,det=FALSE, symmetric = TRUE)
      MeanPar <- attributes(mp)$meanpar
      D <- with(attributes(mp), deriv.lvm(x, meanpar=MeanPar, p=pars, mom=mp, mu=NULL)) ##, all=length(constrain(x))>0))
      myvars <- (index(x)$manifest)
      data <- data[,myvars,drop=FALSE]
      score <- matrix(ncol=length(p),nrow=NROW(data))
      score0 <- -1/2*as.vector(iC)%*%D$dS
      if (!is.null(weights)) {
        W0 <- diag(nrow=length(myvars))
        widx <- match(colnames(weights),myvars)
      }

      for (i in seq_len(NROW(data))) {
        z <- as.numeric(data[i,])
        u <- z-as.numeric(mp$xi)
        if (!is.null(weights)) {
          W <- W0; diag(W)[widx] <- as.numeric(weights[i,])
          score[i,] <-
            as.numeric(crossprod(u,iC%*%W)%*%D$dxi +
                       -1/2*(as.vector((iC
                         - iC %*% crossprod(rbind(u))
                         %*% iC)%*%W)) %*% D$dS
                       )
        } else {
          score[i, ] <-
            as.numeric(score0 + crossprod(u, iC)%*%D$dxi +
                       1/2*as.vector(iC%*%crossprod(rbind(u))%*%iC)%*%D$dS)
        }
      }; colnames(score) <- names(p)
      return(score)
    }
  }

  ## Here the emperical mean and variance of the population are sufficient statistics:
  if (missing(S)) {
    data0 <- na.omit(data[,manifest(x),drop=FALSE])
    n <- NROW(data0)
    S <- cov(data0)*(n-1)/n
    mu <- colMeans(data0)
  }
  mp <- modelVar(x,p)
  C <- mp$C
  xi <- mp$xi
  iC <- Inverse(C,det=FALSE, symmetric = TRUE)
  Debug("Sufficient stats.",debug)
  if (!is.null(mu) & !is.null(xi)) {
    W <- crossprod(rbind(mu-xi))
    T <- S+W
  } else {
    T <- S
  }
  D <- deriv.lvm(x, meanpar=attributes(mp)$meanpar, mom=mp, p=p, mu=mu, mean=mean)
  vec.iC <- as.vector(iC)
  if (lava.options()$devel) {
    Grad <- numeric(length(p))
    imean <- with(index(x)$parBelongsTo,mean)
    Grad[-imean] <- n/2*crossprod(D$dS[,-imean], as.vector(iC%*%T%*%iC)-vec.iC)
  } else {
    Grad <- n/2*crossprod(D$dS, as.vector(iC%*%T%*%iC)-vec.iC)
  }
  if (!is.null(mu) & !is.null(xi)) {
    if (!(lava.options()$devel)) {
      Grad <- Grad - (n/2*crossprod(D$dT,vec.iC))
    } else {
      Grad[with(index(x)$parBelongsTo,c(mean,reg))] <- Grad[with(index(x)$parBelongsTo,c(mean,reg))] - (n/2*crossprod(D$dT,vec.iC))
    }
  }
  res <- as.numeric(Grad)
  return(rbind(res))
}

## Maximum Likelihood with numerical gradient + hessian
gaussian0_objective.lvm <- gaussian_objective.lvm

gaussian1_objective.lvm <- gaussian_objective.lvm
gaussian1_gradient.lvm <- function(...) gaussian_gradient.lvm(...)
gaussian1_hessian.lvm <- function(x,p,...) {
    myg2 <- function(p1) gaussian_gradient.lvm(x,p=p1,...)
    myg3 <- function(p1) numDeriv::jacobian(myg2,p1)
    return(myg3(p))
    ## myg <- function(p1) gaussian_objective.lvm(x,p=p1,...)
    ## numDeriv::hessian(myg,p)
}

## BHHH
gaussian2_method.lvm <- "NR"
gaussian2_objective.lvm <- gaussian_objective.lvm
gaussian2_gradient.lvm <- gaussian_gradient.lvm
gaussian2_hessian.lvm <- function(x,p,n,data,...) {
    S <- -score(x,p=p,n=n,data=data,indiv=TRUE,...)
    I <- t(S)%*%S
    attributes(I)$grad <- colSums(S)
    return(I)
}

rchisqsum <- function(n,lambda) {
    p <- length(lambda)
    X2 <- matrix(rnorm(n*p)^2,ncol=p) ## Chi-squared (df=1)
    res <- numeric(n)
    for (i in seq(p)) {
        res <- res + X2[,i]*lambda[i]
    }
    return(res)
}

pchisqsum <- function(x, lambda=1, B=1e6, seed=NULL) {
    if (!is.null(seed)) set.seed(seed)
    y <- rchisqsum(B,lambda)
    mean(y<=x)
}

context("Weighted sum of chi-squared (chisqsum)")

## rchisqsum / pchisqsum are internal -> accessed via lava:::

test_that("rchisqsum returns numeric vector of correct length and non-negative", {
    set.seed(1)
    y <- lava:::rchisqsum(100, lambda = c(1, 2, 0.5))
    testthat::expect_type(y, "double")
    testthat::expect_length(y, 100)
    testthat::expect_true(all(y >= 0))
})

test_that("rchisqsum mean and variance match theoretical values", {
    testthat::skip_on_cran()
    set.seed(42)
    lambda <- c(0.5, 1, 2)
    y <- lava:::rchisqsum(2e5, lambda = lambda)
    ## E[sum lambda_i Z_i^2] = sum(lambda)
    testthat::expect_lt(abs(mean(y) - sum(lambda)), 0.05)
    ## Var = 2 * sum(lambda^2)
    testthat::expect_lt(abs(var(y) - 2 * sum(lambda^2)), 0.5)
})

test_that("rchisqsum reduces to chi-squared(df) when lambda = rep(1, df)", {
    testthat::skip_on_cran()
    set.seed(7)
    df <- 5
    y <- lava:::rchisqsum(2e5, lambda = rep(1, df))
    testthat::expect_lt(abs(mean(y) - df), 0.05)
    testthat::expect_lt(abs(var(y) - 2 * df), 0.5)
})

test_that("rchisqsum with single lambda=1 matches chi-squared(1)", {
    testthat::skip_on_cran()
    set.seed(11)
    y <- lava:::rchisqsum(2e5, lambda = 1)
    testthat::expect_lt(abs(mean(y) - 1), 0.03)
    testthat::expect_lt(abs(var(y) - 2), 0.1)
})

test_that("pchisqsum returns probability in [0,1]", {
    set.seed(1)
    p <- lava:::pchisqsum(1, lambda = 1, B = 5e3, seed = 1)
    testthat::expect_true(p >= 0 && p <= 1)
})

test_that("pchisqsum is monotone non-decreasing in x", {
    testthat::skip_on_cran()
    lambda <- c(1, 2)
    xs <- c(0.5, 1, 2, 5, 10)
    ps <- vapply(xs,
                 function(xx) lava:::pchisqsum(xx, lambda = lambda,
                                                B = 2e4, seed = 1),
                 numeric(1))
    testthat::expect_true(all(diff(ps) >= 0))
})

test_that("pchisqsum boundary behaviour at extremes", {
    testthat::skip_on_cran()
    p_lo <- lava:::pchisqsum(-1, lambda = 1, B = 1e4, seed = 1)
    p_hi <- lava:::pchisqsum(1e6, lambda = 1, B = 1e4, seed = 1)
    testthat::expect_equal(p_lo, 0)
    testthat::expect_equal(p_hi, 1)
})

test_that("pchisqsum is reproducible with seed argument", {
    p1 <- lava:::pchisqsum(2, lambda = c(1, 0.5), B = 5e3, seed = 99)
    p2 <- lava:::pchisqsum(2, lambda = c(1, 0.5), B = 5e3, seed = 99)
    testthat::expect_identical(p1, p2)
    p3 <- lava:::pchisqsum(2, lambda = c(1, 0.5), B = 5e3, seed = 100)
    testthat::expect_false(isTRUE(all.equal(p1, p3)))
})

test_that("pchisqsum approximates pchisq when lambda = 1", {
    testthat::skip_on_cran()
    xs <- c(0.5, 1, 2, 4)
    for (x in xs) {
        p_mc <- lava:::pchisqsum(x, lambda = 1, B = 2e5, seed = 13)
        p_exact <- pchisq(x, df = 1)
        testthat::expect_lt(abs(p_mc - p_exact), 0.01)
    }
})

## ---------------------------------------------------------------------------
## Regression: pchisqsum vectorized over x
## ---------------------------------------------------------------------------
test_that("pchisqsum returns one probability per element of vector x", {
    xs <- c(0.5, 1, 2, 4)
    p_vec <- lava:::pchisqsum(xs, lambda = c(1, 0.5), B = 5e3, seed = 7)
    testthat::expect_length(p_vec, length(xs))
    testthat::expect_true(all(p_vec >= 0 & p_vec <= 1))
    testthat::expect_true(all(diff(p_vec) >= 0))
    ## scalar call agrees with vectorized call (same seed, same B)
    p1 <- lava:::pchisqsum(xs[2], lambda = c(1, 0.5), B = 5e3, seed = 7)
    testthat::expect_equal(p_vec[2], p1)
})

subgraph <- function(g,from,to,Tree=new("graphNEL",node=c(to,from),edgemode="directed"),...) {
    adjnodes <- graph::adj(g,from)[[1]]
    if (length(adjnodes)==0)
        return(Tree)
    for (v in adjnodes) {
        if (v==to) {
            Tree <- graph::addEdge(from, v, Tree)
        }
        re1 <- graph::acc(g,v)[[1]] ## Reachable nodes from v
        if ((to %in% names(re1)[re1>0])) {
            if (!(v %in% graph::nodes(Tree)))
                Tree <- graph::addNode(v,Tree)
            Tree <- graph::addEdge(from, v, Tree)
            Tree <- path(g,v,to,Tree)
        }
    }
    return(Tree)
}

##' Trace operator
##'
##' Calculates the trace of a square matrix.
##' @param x Square numeric matrix
##' @param \dots Additional arguments to lower level functions
##' @return \code{numeric}
##' @author Klaus K. Holst
##' @seealso \code{\link{crossprod}}, \code{\link{tcrossprod}}
##' @keywords math algebra
##' @examples
##'
##' tr(diag(1:5))
##' @export
"tr" <- function(x,...) UseMethod("tr")

##' @export
`tr.matrix` <-
function(x,na.rm=FALSE,...) {
  if (length(x)==1)
    return(x)
  n <- nrow(x)
  if (!n)
    stop("0 x 0 matrix")
  if (n != ncol(x))
    stop("non-square matrix")
  if (!na.rm && any(!is.finite(x)))
    stop("infinite or missing values")
  return(sum(diag(x),na.rm=na.rm))
}

##' Composite Likelihood for probit latent variable models
##'
##' Estimate parameters in a probit latent variable model via a composite
##' likelihood decomposition.
##' @param x \code{lvm}-object
##' @param data data.frame
##' @param k Size of composite groups
##' @param type Determines number of groups. With \code{type="nearest"} (default)
##' only neighboring items will be grouped, e.g. for \code{k=2}
##' (y1,y2),(y2,y3),... With \code{type="all"} all combinations of size \code{k}
##' are included
##' @param pairlist A list of indices specifying the composite groups. Optional
##' argument which overrides \code{k} and \code{type} but gives complete
##' flexibility in the specification of the composite likelihood
##' @param messages Control amount of messages printed
##' @param estimator Model (pseudo-likelihood) to use for the pairs/groups
##' @param quick If TRUE the parameter estimates are calculated but all additional
##' information such as standard errors are skipped
##' @param \dots Additional arguments parsed on to lower-level functions
##' @return An object of class \code{estimate.complik} inheriting methods from \code{lvm}
##' @author Klaus K. Holst
##' @seealso estimate
##' @keywords models regression
##' @export
##' @examples
##' m <- lvm(c(y1,y2,y3)~b*x+1*u[0],latent=~u)
##' ordinal(m,K=2) <- ~y1+y2+y3
##' d <- sim(m,50,seed=1)
##' if (requireNamespace("mets", quietly=TRUE)) {
##'    e1 <- complik(m,d,control=list(trace=1),type="all")
##' }
complik <- function(x, data, k=2, type=c("all","nearest"), pairlist,
             messages=0, estimator="normal", quick=FALSE, ...) {

    y <- setdiff(endogenous(x),latent(x))
    binsurv <- rep(FALSE,length(y))
    for (i in 1:length(y)) {
        z <- try(data[,y[i]],silent=TRUE)
        ## binsurv[i] <- is.Surv(z) | (is.factor(z) && length(levels(z))==2)
        if (!inherits(z,"try-error"))
            binsurv[i] <- inherits(z,"Surv") | (is.factor(z))
    }

    ord <- ordinal(x)
    binsurv <- unique(c(y[binsurv],ord))
    if (!missing(pairlist)) {
        binsurvpos <- which(colnames(data)%in%endogenous(x))
    } else {
        binsurvpos <- which(colnames(data)%in%binsurv)
    }
    if (missing(pairlist)) {
        if (type[1]=="all") {
            mypar <- combn(length(binsurv),k) ## all pairs (or multiplets), k=2: k*(k-1)/2
        } else {
            mypar <- sapply(0:(length(binsurv)-k), function(x) x+1:k)
        }
    } else {
        mypar <- pairlist
    }

    if (is.matrix(mypar)) {
        mypar0 <- mypar; mypar <- c()
        for (i in seq(ncol(mypar0)))
            mypar <- c(mypar, list(mypar0[,i]))
    }
    nblocks <- length(mypar)

    mydata0 <- data[,,drop=FALSE]
    mydata <-  as.data.frame(matrix(NA, nblocks*nrow(data), ncol=ncol(data)))
    names(mydata) <- names(mydata0)
    for (i in 1:ncol(mydata)) {
        if (is.factor(data[,i])) {
            mydata[,i] <- factor(mydata[,i],levels=levels(mydata0[,i]))
        }
        if (survival::is.Surv(data[,i])) {
            S <- data[,i]
            for (j in 2:nblocks) S <- rbind(S,data[,i])
            S[] <- NA
            mydata[,i] <- S
        }
    }
    for (ii in 1:nblocks) {
        data0 <- data;
        for (i in binsurvpos[-mypar[[ii]]]) {
            if (survival::is.Surv(data[,i])) {
                S <- data0[,i]
                S[] <- NA
                data0[,i] <- S
            } else {
                data0[,i] <- NA
                if (is.factor(data[,i])) data0[,i] <- factor(data0[,i],levels=levels(data[,i]))
            }
        }
        mydata[(1:nrow(data))+(ii-1)*nrow(data),] <- data0
    }
    suppressWarnings(e0 <- estimate(x,data=mydata,estimator=estimator,missing=TRUE,messages=messages,
                                    hessian=!quick,
                                    ...))
    e0$n <- nrow(data)
    e0$eigenvalues <- rep(1,nrow(e0$coef))
    e0$coef[,2:4] <- NA
    if (!quick) {
        S <- score(e0,indiv=TRUE)
        nd <- nrow(data)
        block1 <- which((1:nd)%in%(rownames(S)))
        blocks <- sapply(1:nblocks, function(x) 1:length(block1)+length(block1)*(x-1))
        if (nblocks==1) {
            Siid <- S
        } else {
            Siid <- matrix(0,nrow=length(block1),ncol=ncol(S))
            for (j in 1:ncol(blocks)) {
                Siid <- Siid+S[blocks[,j],]
            }
        }
        ##B <- solve(information(e0, type="hessian"))
        D <- numDeriv::jacobian(function(p) score(e0, p=p), coef(e0), method=lava.options()$Dmethod)
        B <- -Inverse(D)*NROW(Siid)
        A <- crossprod(Siid)/NROW(Siid)
        e0$bread <- B
        e0$meat <- A
        e0$iidscore <- Siid
        e0$blocks <- blocks
        BA <- B%*%A
        eig <- eigen(BA)$values
        e0$eigenvalues <- eig
        e0$vcov <- (BA%*%B)/NROW(Siid)
        cc <- e0$coef;
        cc[,2] <- sqrt(diag(e0$vcov))
        cc[,3] <- cc[,1]/cc[,2]; cc[,4] <- 2*(1-pnorm(abs(cc[,3])))
        e0$coef <- cc
    }
    class(e0) <- c("estimate.complik",class(e0))
    return(e0)
}

##' @export
score.estimate.complik <- function(x,indiv=FALSE,...) {
    if (!indiv)
        return(colSums(x$iidscore))
    x$iidscore
}

##' @export
IC.estimate.complik <- function(x,...) {
    IC.default(x,bread=x$bread,...)
}

##' @export
logLik.estimate.complik <- function(object,...) {
    with(object, structure(opt$objective,
              nall=n,
              nobs=n,
              df=sum(eigenvalues),
              class="logLik"))
}

cluster_post_hook <- function(x,...) {
    if (class(x)[1]=="multigroupfit") {
        if (is.null(x$cluster)) return(NULL)
        if (any(unlist(lapply(x$cluster,is.null)))) return(NULL)
        allclusters <- unlist(x$cluster)
        uclust <- unique(allclusters)
        K <- length(uclust)
        G <- x$model$ngroup
        S0 <- lapply(score(x,indiv=TRUE), function(x) { x[which(is.na(x))] <- 0; x })
        S <- matrix(0,length(pars(x)),nrow=K)
        for (i in uclust) {
            for (j in seq_len(G)) {
                idx <- which(x$cluster[[j]]==i)
                if (length(idx)>0)
                    S[i,] <- S[i,] + colSums(S0[[j]][idx,,drop=FALSE])
            }
        }
        J <- crossprod(S)
        I <- information(x,type="hessian",...)
        iI <- Inverse(I)
        asVar <- iI%*%J%*%iI
        x$vcov <- asVar
        return(x)
    }

    ## lvmfit:
    if (!is.null(x$cluster)) {
        uclust <- unique(x$cluster)
        K <- length(uclust)
        S <- score(x,indiv=TRUE) #,...)
        I <- information(x,type="hessian") #,...)
        iI <- Inverse(I)

        S0 <- matrix(0,ncol=ncol(S),nrow=K)
        count <- 0
        for (i in uclust) {
            count <- count+1
            S0[count,] <- colSums(S[which(x$cluster==i),,drop=FALSE])
        };
        adj1 <- K/(K-1)
        ## p <- ncol(S)
        ## adj1 <- K/(K-p) ## Mancl & DeRouen, 2001

        J <- adj1*crossprod(S0)
        col3 <- sqrt(diag(iI)); ## Naive se
        nn <- c("Estimate","Robust SE", "Naive SE", "P-value")
        asVar <- iI%*%J%*%iI
    } else {
        asVar <- x$vcov
    }
    diag(asVar)[diag(asVar)==0] <- NA
    mycoef <- x$opt$estimate
    x$vcov <- asVar
    SD <- sqrt(diag(asVar))
    Z <- mycoef/SD
    pval <- 2*(pnorm(abs(Z),lower.tail=FALSE))
    if (is.null(x$cluster)) {
        col3 <- Z
        nn <-  c("Estimate","Std. Error", "Z-value", "P-value")
    }
    newcoef <- cbind(mycoef, SD, col3, pval);
    nparall <- index(x)$npar + ifelse(x$control$meanstructure, index(x)$npar.mean,0)
    if (!is.null(x$expar)) {
        nparall <- nparall+length(x$expar)
    }
    mycoef <- matrix(NA,nrow=nparall,ncol=4)
    mycoef[x$pp.idx,] <- newcoef
    colnames(mycoef) <- nn
    mynames <- c()
    if (x$control$meanstructure) {
        mynames <- vars(x)[index(x)$v1==1]
    }
    if (index(x)$npar>0) {
        mynames <- c(mynames,paste0("p",seq_len(index(x)$npar)))
    }
    if (!is.null(x$expar)) {
        mynames <- c(mynames,names(x$expar))
    }

    rownames(mycoef) <- mynames
    x$coef <- mycoef
    return(x)
}

##' @export
"transform<-" <- function(`_data`,...,value) UseMethod("transform<-")

##' @export
"transform<-.lvm" <- function(`_data`,formula=NULL,...,value) {
    transform(`_data`,formula,value,...)
}

##' @export
print.transform.lvm <- function(x,...) {
    for (i in seq_along(x)) {
        cat("Variable: ", names(x)[i],"\n",sep="")
        cat("Transformation: (",paste0(x[[i]]$x,collapse=","),") -> ",sep="")
        print(x[[i]]$fun)
        cat("\n")
    }
    invisible(x)
}

##' @export
"transform.lvm" <- function(`_data`,formula,value,post=TRUE,y,x,...) {
    if (missing(formula)) {
        if (length(tr <- `_data`$attributes$transform)==0) {
            return(NULL)
        }        
        return(structure(`_data`$attributes$transform,class="transform.lvm"))
    }
    
    if (!missing(y) && !missing(x)) {
        xx <- x
    } else {
        if (is.character(formula)) {
            y <- NULL; xx <- formula
        } else {
            y <- getoutcome(formula)
            xx <- attributes(y)$x
        }
    }
    if (length(xx)==0) { xx <- y; y <- NULL }
    if (length(y)==0) {
        if (post) {
            `_data`$constrainY[xx] <- NULL
            `_data`$constrain[xx] <- NULL
            if (is.null(`_data`$attributes$selftransform))
                `_data`$attributes$selftransform <- list()
            `_data`$attributes$selftransform[[xx]] <- value
            return(`_data`)
        }
        `_data`$attributes$selftransform[xx] <- NULL
        constrain(`_data`,xx,y,...) <- value
        return(`_data`)
    }
    
    
    `_data`$attributes$selftransform[y] <- NULL
    addvar(`_data`) <- y
    intercept(`_data`,y) <- 0; covariance(`_data`,y) <- 0
    if (is.null(`_data`$attributes$transform))
        `_data`$attributes$transform <- list()
    if (is.null(value)) `_data`$attributes$transform[y] <- NULL
    else {
        if (length(y)>1) {
            if (is.null(`_data`$attributes$multitransform))
                `_data`$attributes$multitransform <- list()
            `_data`$attributes$multitransform
            for (yi in y) {
                `_data`$attributes$transform[yi] <- NULL
            }
            rmidx <- c()
            for (i in seq_along(`_data`$attributes$multitransform)) {
                l <- `_data`$attributes$multitransform[[i]]
                if (any(y%in%letters)) rmidx <- c(rmidx,i)
            }
            if (length(rmidx)>0) `_data`$attributes$transform[rmidx] <- NULL            
            `_data`$attributes$multitransform <- c(`_data`$attributes$multitransform,                                                   
                                                   list(list(fun=value,y=y,x=xx)))
        } else {
            `_data`$attributes$transform[[y]] <- list(fun=value,x=xx)
        }
    }
    return(`_data`)
}


addhook("plothook_transform","plot.post.hooks")

plothook_transform <- function(x,...) {
    trans <- x$attributes$transform
    transnames <- names(trans)
    for (v in transnames) {
        xx <- trans[[v]][["x"]]
        if (length(xx)>0) {
            x <- regression(x,x=xx,y=v)
            edgelabels(x,from=xx,to=v,col="gray70") <- ""
        }        
    }
    return(x)
}

##' @export
totaleffects <- function(object,...,value) UseMethod("totaleffects")

##' @export
totaleffects.lvmfit <- function(object,to,...,level=0.95) {
    p <- (1-level)/2
    q <- qnorm(p)
    res <- c()
    if (inherits(to,"formula")) {
        if (substr(deparse(to[3]),1,1)==".") {
            trim <- function(x) sapply(x,function(z) gsub(" ","",z,fixed=TRUE))
            to <- trim(strsplit(deparse(to),"~")[[1]][1])
        } else {
            to <- list(to)
        }
    }
    if (is.null(list(...)$from) & is.character(to)[1]) {
        to <- lapply(paste(to,setdiff(vars(object),to),sep="~"),FUN=as.formula)
    }
    ef <- function(tt) {
        f <- effects(object,tt,...)
        rbind(with(f$totalef,c(est,sd,est/sd,2*(pnorm(abs(est/sd),lower.tail=FALSE)),est+q*sd,est-q*sd)))
    }
    if (is.list(to)) {
        for (tt in to) {
            res <- rbind(res,ef(tt))
        }
    }
    else
        res <- ef(to)
    colnames(res) <- c("Estimate","Std.Err","z value","Pr(>|z|)",
                       paste0(c(1-p,p)*100,"%"))
    rownames(res) <- to
    res
}

##' @export
effects.lvmfit <- function(object,to,from,...) {
    if (missing(to)) {
        return(summary(object))
    }
    P <- path(object,to=to,from=from,...)
    if (is.null(P$path)) {
        if (inherits(to,"formula")) {
            f <- extractvar(to)
            to <- f$y; from <- f$x
        }
    } else {
        from <- P$path[[1]][1]
        to <- tail(P$path[[1]],1)
    }
    cc <- coef(object,type=9,labels=FALSE) ## All parameters (fixed and variable)
    cc0 <- cbind(coef(object)) ## Estimated parameters
    i1 <- na.omit(match(rownames(cc),rownames(cc0)))
    idx.cc0 <-  which(rownames(cc)%in%rownames(cc0)); ## Position of estimated parameters among all parameters
    S <- matrix(0,nrow(cc),nrow(cc)); rownames(S) <- colnames(S) <- rownames(cc)
    V <- object$vcov
    npar.mean <- index(object)$npar.mean
    S[idx.cc0,idx.cc0] <- V[i1,i1] ## "Covariance matrix" of all parameters

    cclab <- rownames(coef(object,type=9,labels=TRUE)) ## Identify equivalence constraints
    cctab <- table(cclab)
    equiv <- which(cctab>1)
    for (i in seq_len(length(equiv))) {
        orgpos <- which(cclab==(names(equiv)[i]))
        pos <- orgpos[-1]
        for (p in pos)
            S[p,-orgpos[1]] <- S[-orgpos[1],p] <- S[orgpos[1],-p]
    }

    idx.orig <- unique(unlist(P$idx))
    coefs.all <- cc[idx.orig]

    S.all <- S[idx.orig,idx.orig]
    idx.all <- numberdup(unlist(P$idx))
    pos <- 1; idx.list <- P$idx; for (i in seq_len(length(idx.list))) {
                                     K <- length(idx.list[[i]])
                                     idx.list[[i]] <- idx.all[pos:(pos+K-1)]; pos <- pos+K
                                 }
    margef <- list()
    if (length(coefs.all)==1 && is.na(coefs.all)) {
        totalef <- list(est=0,sd=0)
        margef <- c(margef,list(est=0,sd=NA))
    } else {
        totalef <- prodsumdelta(coefs.all, idx.list, S.all,...)
        for (i in seq_len(length(idx.list))) {
            margef <- c(margef, list(prodsumdelta(coefs.all, idx.list[i], S.all,...)))
        }
    }

    directidx <- which(lapply(P$path,length)==2)

    inef.list <- idx.list
    if (length(directidx)==0) {
        directef <- list(est=0, sd=NA)
    } else {
        inef.list <- inef.list[-directidx]
        directef <- margef[[directidx]]
    }
    if (length(inef.list)==0) {
        totalinef <- list(est=0,sd=NA,grad=NA,hess=NA)
    } else {
        totalinef <- prodsumdelta(coefs.all, inef.list, S.all,...)
    }

    nn <- c("total","direct","indirect")
    for (i in seq_len(length(margef))) {
        if (length(P$path[[i]])>2) {
            nn <- c(nn,paste(rev(P$path[[i]]),collapse=lava.options()$symbol[1]))
        }
    }
    b <- c(totalef$est,directef$est,totalinef$est,totalinef$b)
    names(b) <- nn
    D <- t(cbind(totalef$grad,directef$grad,totalinef$grad,totalinef$D))
    V <- D%*%S.all%*%t(D)
    val <- list(coef=b, vcov=V, grad=D, paths=P$path, totalef=totalef, directef=directef, totalinef=totalinef, margef=margef, from=from, to=to)
    class(val) <- "effects"
    val
}

##' @export
print.effects <- function(x,digits=4,...) {
    s <- summary(x,...)
    print(s$coef,digits=digits,...)
    cat("\n")
    print(s$medprop$coefmat[,c(1,3,4),drop=FALSE],digits=digits,...)
    return(invisible(x))
}

##' @export
coef.effects <- function(object,...) {
    object$coef
}

##' @export
vcov.effects <- function(object,...) {
    object$vcov
}

##' @export
summary.effects <- function(object,...) {
    totalef <- with(object$totalef, cbind(est,sd[1]))
    directef <- with(object$directef, cbind(est,sd[1]))
    totindirectef <- with(object$totalinef, cbind(est,sd[1]))
    rownames(totalef) <- "Total"
    rownames(directef) <- "Direct"
    rownames(totindirectef) <- "Indirect"
    nn <- indirectef <- c()
    K <- seq_len(length(object$margef))
    for (i in K) {
        if (length(object$paths[[i]])>2) {
            nn <- c(nn,paste(rev(object$paths[[i]]),collapse=lava.options()$symbol[1]))
            indirectef <- rbind(indirectef, with(object$margef[[i]], c(est,sd)))
        }
    }; rownames(indirectef) <- nn
    mycoef <- rbind(totalef,directef,totindirectef,indirectef)
    mycoef <- cbind(mycoef,mycoef[,1]/mycoef[,2])
    mycoef <- cbind(mycoef,2*(pnorm(abs(mycoef[,3]),lower.tail=FALSE)))
    colnames(mycoef) <- c("Estimate","Std.Err","z value","Pr(>|z|)")
    medprop <- NULL
    if (totindirectef[1]!=0) {
        if (abs(coef(object)[2])<1e-12) {
          medprop <- estimate(NULL,coef=c("Mediation proportion"=1),vcov=matrix(NA))
        } else {
          # set IC=FALSE to prevent warning from estimnate.default because
          # effect object has no IC function
          medprop <- estimate(
            object, function(x) list("Mediation proportion"=1-x[2]/x[1]),
            IC = FALSE
          )
        }
    }
    list(coef=mycoef,medprop=medprop)
}


##' @export
confint.effects <- function(object,parm,level=0.95,...) {
    mycoef <- summary(object)$coef
    p <- 1-(1-level)/2
    res <- mycoef[,1] +  + qnorm(p)*cbind(-1,1)%x%mycoef[,2]
    colnames(res) <- paste0(c(1-p,p)*100,"%")
    rownames(res) <- rownames(mycoef)
    res
}


prodtrans <- function(betas) {
    k <- length(betas)
    res <- prod(betas)
    nabla <- numeric(k)
    for (i in seq_len(k))
        nabla[i] <- prod(betas[-i])

    H <- matrix(0,k,k)
    if (k>1)
        for (i in seq_len(k-1))
            for (j in (i+1):k)
                H[j,i] <- H[i,j] <- prod(c(1,betas[-c(i,j)]))
    attr(res,"gradient") <- nabla
    attr(res,"hessian") <- H
    return(res)
}
prodsumdelta <- function(betas,prodidx,S,order=1) { ## Delta-method
    k <- length(prodidx)
    p <- length(betas)
    if (p==1) {
        return(list(est=betas, sd=sqrt(S), grad=0, beta=betas, D=0, hess=0))
    }
    val <- 0; grad <- numeric(p)
    D <- matrix(0,nrow=p,ncol=k)
    beta <- numeric(k)
    H <- matrix(0,p,p)
    for (i in seq_len(k)) {
        ii <- prodidx[[i]]
        myterm <- prodtrans(betas[ii]);
        if (order>1) {
            H0 <- attributes(myterm)$hessian
            Sigma <- S[ii,ii]
            print(sum(diag(Sigma%*%H0))/2)
            val <- val + (myterm + sum(diag(Sigma%*%H0))/2)
        } else {
            val <- val + myterm
            beta[i] <- myterm
        }
        D[ii,i] <- attributes(myterm)$gradient
        grad[ii] <- grad[ii] + attributes(myterm)$gradient
    }; grad <- matrix(grad,ncol=1)
    return(list(est=val, sd=sqrt(t(grad)%*%S%*%grad), grad=grad, b=beta, D=D, hess=H))
}

nlminb2 <- function(start,objective,gradient,hessian,...) {
  nlminbcontrols <- c("eval.max","iter.max","trace","abs.tol","rel.tol","x.tol","step.min")
  dots <- list(...)
  control <- list(...)$control
  control <- control[names(control)%in%nlminbcontrols]
  dots$control <- control
  if (length(dots$trace)>0 && dots$trace>0) cat("\n")
  mypar <- c(list(start=start,objective=objective,gradient=gradient,hessian=hessian),dots)
  mypar["debug"] <- NULL
  do.call("nlminb", mypar)
}

nlminb1 <- function(start,objective,gradient,hessian,...) {
  nlminb2(start,objective,gradient=gradient,hessian=NULL,...)
}

nlminb0 <- function(start,objective,gradient,hessian,...) {
  nlminb2(start,objective,gradient=NULL,hessian=NULL,...)
}

################################################################################

estimatingfunction <- function(start,objective,gradient,hessian,NR=FALSE,...) {
  myobj <- function(x,...) {
    S <- gradient(x,...)
    crossprod(S)[1]
  }
  if (!missing(hessian) && !is.null(hessian)) {
    mygrad <- function(x) {
      H <- hessian(x)
      S <- gradient(x)
      2*S%*%H
    }
  } else {
    hessian <- function(x) numDeriv::jacobian(gradient,x,method=lava.options()$Dmethod)
    mygrad <- function(x) {
      H <- hessian(x)
      S <- gradient(x)
      2*S%*%H
    }
  }
  if (NR) {
    op <- lava::NR(start,gradient=gradient,hessian=hessian,...)
  } else {
    op <- nlminb2(start,myobj,mygrad,hessian=NULL,...)
  }
  return(op)
}

estimatingfunction0 <- function(...,hessian=NULL) estimatingfunction(...,hessian=hessian)

################################################################################
## Newton-Raphson/Scoring
################################################################################

##' @title Newton-Raphson method
##' 
##' @param start Starting value
##' @param objective Optional objective function (used for selecting step length)
##' @param gradient gradient
##' @param hessian hessian (if NULL a numerical derivative is used)
##' @param control optimization arguments (see details)
##' @param args Optional list of arguments parsed to objective, gradient and hessian
##' @param ... additional arguments parsed to lower level functions
##' @details
##' \code{control} should be a list with one or more of the following components:
##' \itemize{
##' \item{trace} integer for which output is printed each 'trace'th iteration
##' \item{iter.max} number of iterations
##' \item{stepsize}: Step size (default 1)
##' \item{nstepsize}: Increase stepsize every nstepsize iteration (from stepsize to 1)
##' \item{tol}: Convergence criterion (gradient)
##' \item{epsilon}: threshold used in pseudo-inverse
##' \item{backtrack}: In each iteration reduce stepsize unless solution is improved according to criterion (gradient, armijo, curvature, wolfe)
##' }
##' @export
##' @examples
##' # Objective function with gradient and hessian as attributes
##' f <- function(z) {
##'    x <- z[1]; y <- z[2]
##'    val <- x^2 + x*y^2 + x + y
##'    structure(val, gradient=c(2*x+y^2+1, 2*y*x+1),
##'              hessian=rbind(c(2,2*y),c(2*y,2*x)))
##' }
##' NR(c(0,0),f)
##' 
##' # Parsing arguments to the function and
##' g <- function(x,y) (x*y+1)^2
##' NR(0, gradient=g, args=list(y=2), control=list(trace=1,tol=1e-20))
##' 
##' 
NR <- function(start,objective=NULL,gradient=NULL,hessian=NULL,control,args=NULL,...) {
  control0 <- list(trace=0,
                   stepsize=1,
                   lambda=0,
                   ngamma=0,
                   gamma2=0,
                   backtrack=TRUE,
                   iter.max=200,
                   tol=1e-6,
                   stabil=FALSE,
                   epsilon=1e-9)
  if (!missing(control)) {
      control0[names(control)] <- control
      # Backward compatibility:
      if (!is.null(control0$gammma)) control0$stepsize <- control0$gamma
  }

  ## conditions to select the step length
  if(control0$backtrack[1] == "armijo"){
    control0$backtrack <- c(1e-4,0) # page 33
  }
  if(control0$backtrack[1] == "curvature"){
    control0$backtrack <- c(0,0.9) # page 34
  }
  if(control0$backtrack[1] == "wolfe"){
      control0$backtrack <- c(1e-4,0.9)
  }
  if(!is.logical(control0$backtrack) || length(control0$backtrack)!=1){
    if(length(control0$backtrack) != 2){
      stop("control$backtrack must have length two if not TRUE or FALSE \n")
    }
    if(any(!is.numeric(control0$backtrack)) || any(abs(control0$backtrack)>1)){
      stop("elements in control$backtrack must be in [0,1] \n")
    }
    if(control0$backtrack[2]==0){
      control0$backtrack[2] <- +Inf # no Wolfe condition
    }
  }
  obj <- objective
  grad <- gradient
  hess <- hessian
  if (!is.null(args)) {
      if (!is.list(args)) args <- list(args)
      if (!is.null(objective))
          obj <- function(p) do.call(objective, c(list(p),args))
      if (!is.null(gradient))
          grad <- function(p) do.call(gradient, c(list(p),args))
      if (!is.null(hessian))
          hess <- function(p) do.call(hessian, c(list(p),args))
  }
  if (control0$trace>0) {
       cat("\nIter=0\t")
      if (!is.null(obj))
          cat("Objective=",obj(as.double(start)))
      cat(";\t\n \tp=", paste0(formatC(start), collapse=" "),"\n")
  }
  gradFun = !is.null(grad)
  if (!gradFun & is.null(hess)) {
    hess <- function(p) {
      ff <- obj(p)
      res <- attributes(ff)$hess
      if (is.function(res)) {
          res <- res(p)
          attributes(res)$grad <- as.vector(attributes(ff)$grad(p))
      } else {      
          attributes(res)$grad <- as.vector(attributes(ff)$grad)
      }
      return(res)
    }
    grad <- function(p) {
        if (control0$trace>0) print("Numerical gradient")
        numDeriv::jacobian(obj,p)
    }
  }
  oneiter <- function(p.orig,Dprev,return.mat=FALSE,iter=1) {
      D <- I <- NULL # Place-holders for gradient and negative hessian
      if (!is.logical(control0$backtrack) && !is.null(obj)) { # Back-tracking based on objective function evaluations
          objective.origin <- obj(p.orig)
          D <- attributes(objective.origin)$grad
          I <- attributes(objective.origin)$hess
          if (!is.null(I)) I <- -I
      }
      if (is.null(D) || is.null(I)) {
          if (!is.null(hess)) {
              H <- hess(p.orig)
          }
          if (is.null(hess) || is.null(H)) {
              if (control0$trace>0) print("Numerical Hessian")
              I <- -numDeriv::jacobian(grad,p.orig,method=lava.options()$Dmethod)
          } else {
              I <- -H
          }
          D <- attributes(I)$grad
          if (is.null(D)) {
              D <- grad(p.orig)
          }
      }
      
    if (return.mat) return(list(D=D,I=I))
    if (control0$stabil) {
      if (control0$lambda!=0) {
        if (control0$lambda<0) {
          sigma <- (t(D)%*%(D))[1]
        } else {
          sigma <- control0$lambda
        }
        sigma <- min(sigma,10)
        I <- I+control0$gamma2*sigma*diag(nrow=nrow(I))
      } else {
        sigma <- ((D)%*%t(D))
        I <- I+control0$gamma2*(sigma)
      }
    }
    iI <- Inverse(I, symmetric=TRUE, tol=control0$epsilon)
    Delta <- control0$stepsize*tryCatch(solve(I, cbind(as.vector(D))),
                            error=function(...) { ## Fall back to Pseudo-Inverse using SVD:
                                iI%*%cbind(as.vector(D))})
    Lambda <- 1 ## Initial step-size
    if (identical(control0$backtrack, TRUE)) {
      mD0 <- mean(Dprev^2)
      mD <- mean(D^2)
      p <- p.orig + as.vector(Lambda*Delta)
      while (mD>=mD0) {
        if (gradFun) {
          D = grad(p)
        } else {
          DI <- oneiter(p,return.mat=TRUE)
          D = DI$D
        }
        mD = mean(D^2)
        if (is.nan(mD)) mD=mD0
        Lambda <- Lambda/2
        if (Lambda<1e-4) break;
        p <- p.orig + as.vector(Lambda*Delta)
      }

    } else if(identical(control0$backtrack, FALSE)) {
      p <- p.orig + Lambda*Delta
    } else {  # objective(p.orig) - obj(p) <= mu*Lambda*grad(p.orig)*Delta
        ## curvature
        c_D.origin_Delta <- control0$backtrack * c(rbind(D) %*% Delta)
        p <- p.orig + as.vector(Lambda*Delta)

        mD0 <- c(objective.origin + Lambda * c_D.origin_Delta[1], abs(c_D.origin_Delta[2]))#
        objective.new <- obj(p)
        grad.new <- attributes(objective.new)$grad
        if (is.null(grad.new)) {
            grad.new <- grad(p)
        }
        mD <- c(objective.new, abs(grad.new %*% Delta))
        count <- 0
        while (any(mD>mD0) || any(is.nan(mD))) {
            count <- count+1
            Lambda <- Lambda/2
            if (Lambda<1e-4) break;
            p <- p.orig + Lambda*Delta
            objective.new <- obj(p)
            grad.new <- attributes(objective.new)$grad
            if(!is.infinite(mD0[1])){
                mD0[1] <- objective.origin + Lambda * c_D.origin_Delta[1]#
                mD[1] <- objective.new
            }
            if(!is.infinite(mD0[2])){
                if (is.null(grad.new)) {
                    grad.new <- grad(p)
                }
                mD[2] <- abs(grad.new %*% Delta)
            }
        }
    }

    return(list(p=p,D=D,iI=iI))
  }

  count <- count2 <- 0
  thetacur <- start
  stepsizecount <- 0
  Dprev <- rep(Inf,length(start))
  for (jj in seq_len(control0$iter.max)) {
    stepsizecount <- stepsizecount+1
    count <-  count+1
    count2 <- count2+1
    newpar <- oneiter(thetacur,Dprev,iter=jj)
    Dprev <- newpar$D
    thetacur <- newpar$p
    if (!is.null(control0$nstepsize) && control0$nstepsize>0) {
      if (control0$nstepsize<=stepsizecount) {
        control0$stepsize <- sqrt(control0$stepsize)
        stepsizecount <- 0
      }
    }
    if (count2==control0$trace) {
        cat("Iter=", count,"\t",sep="")
        if (!is.null(obj))
            cat("Objective=",obj(as.double(newpar$p)))
        cat(";\n\tD=", paste0(formatC(newpar$D),
                               collapse = " "), "\n")
        cat("\tp=", paste0(formatC(thetacur), collapse = " "),
            "\n")
        count2 <- 0
    }
    if (mean(newpar$D^2)^.5<control0$tol) break;
  }

  res <- list(par=as.vector(thetacur), iterations=count, method="NR",
              gradient=newpar$D, iH=newpar$iI)
  return(res)
}

context("Optimization")

test_that("Optimization", {
    m <- lvm(y~x+z)
    d <- simulate(m,20,seed=1)
    e1 <- estimate(m,d,control=list(method="nlminb0"))
    e2 <- estimate(m,d,control=list(method="NR"))
    testthat::expect_equivalent(round(coef(e1),3),round(coef(e2),3))

    f <- function(x) x^2*log(x) # x>0
    df <- function(x) 2*x*log(x) + x
    df2 <- function(x) 2*log(x) + 3
    op <- NR(5,f,df,df2,control=list(tol=1e-40)) ## Find root
    testthat::expect_equivalent(round(op$par,digits=7),.6065307)
    op2 <- lava:::estimatingfunction0(5,gradient=df)
    op3 <- lava:::estimatingfunction(5,gradient=df,hessian=df2,control=list(tol=1e-40))
    testthat::expect_equivalent(op$par,op2$par)
    testthat::expect_equivalent(op$par,op3$par)
})

##' Generic method for calculating bootstrap statistics
##'
##' @title Generic bootstrap method
##' @param x Model object
##' @param \dots Additional arguments
##' @seealso \code{bootstrap.lvm} \code{bootstrap.lvmfit}
##' @author Klaus K. Holst
##' @export
bootstrap <- function(x,...) UseMethod("bootstrap")

##' Calculate bootstrap estimates of a lvm object
##'
##' Draws non-parametric bootstrap samples
##'
##' @param x \code{lvm}-object.
##' @param R Number of bootstrap samples
##' @param data The data to resample from
##' @param fun Optional function of the (bootstrapped) model-fit defining the
##' statistic of interest
##' @param control Options to the optimization routine
##' @param p Parameter vector of the null model for the parametric bootstrap
##' @param parametric If TRUE a parametric bootstrap is calculated. If FALSE a
##' non-parametric (row-sampling) bootstrap is computed.
##' @param bollenstine Bollen-Stine transformation (non-parametric bootstrap) for bootstrap hypothesis testing.
##' @param constraints Logical indicating whether non-linear parameter
##' constraints should be included in the bootstrap procedure
##' @param sd Logical indicating whether standard error estimates should be
##' included in the bootstrap procedure
##' @param mc.cores Optional number of cores for parallel computing. If omitted future.apply will be used (see future::plan)
##' @param future.args arguments to future.apply::future_lapply
##' @param estimator String definining estimator, e.g. 'gaussian' (see
##' \code{estimator})
##' @param weights Optional weights matrix used by \code{estimator}
##' @param \dots Additional arguments, e.g. choice of estimator.
##' @aliases bootstrap.lvmfit
##' @usage
##'
##' \method{bootstrap}{lvm}(x,R=100,data,fun=NULL,control=list(),
##'                           p, parametric=FALSE, bollenstine=FALSE,
##'                           constraints=TRUE,sd=FALSE, mc.cores,
##'                           future.args=list(future.seed=TRUE),
##'                           ...)
##'
##' \method{bootstrap}{lvmfit}(x,R=100,data=model.frame(x),
##'                              control=list(start=coef(x)),
##'                              p=coef(x), parametric=FALSE, bollenstine=FALSE,
##'                              estimator=x$estimator,weights=Weights(x),...)
##'
##' @return A \code{bootstrap.lvm} object.
##' @author Klaus K. Holst
##' @seealso \code{\link{confint.lvmfit}}
##' @keywords models regression
##' @examples
##' m <- lvm(y~x)
##' d <- sim(m,100)
##' e <- estimate(lvm(y~x), data=d)
##' \donttest{ ## Reduce Ex.Timings
##' B <- bootstrap(e,R=50,mc.cores=1)
##' B
##' }
##' @export
bootstrap.lvm <- function(x, R = 100, data, fun = NULL, control = list(),
                          p, parametric = FALSE, bollenstine = FALSE,
                          constraints = TRUE, sd = FALSE, mc.cores,
                          future.args=list(future.seed=TRUE),
                          ...) {
  coefs <- sds <- c()
  on.exit(list(coef = coefs[-1, ], sd = sds[-1, ], coef0 = coefs[1, ], sd0 = sds[1, ], model = x))
  pb <- progressr::progressor(steps = R+1)
  pmis <- missing(p)
  bootfun <- function(i) {
    if (i == 0) {
      d0 <- data
    } else {
      if (!parametric | pmis) {
        d0 <- data[sample(seq_len(nrow(data)), replace = TRUE), ]
      } else {
        d0 <- sim(x, p = p, n = nrow(data))
      }
    }
    suppressWarnings(e0 <- estimate(x, data = d0, control = control, messages = 0, index = FALSE, ...))
    pb()
    if (!is.null(fun)) {
      coefs <- fun(e0)
      newsd <- NULL
    } else {
      coefs <- coef(e0)
      newsd <- c()
      if (sd) {
        newsd <- e0$coef[, 2]
      }
      if (constraints & length(constrain(x)) > 0) {
        cc <- constraints(e0, ...)
        coefs <- c(coefs, cc[, 1])
        names(coefs)[seq(length(coefs) - length(cc[, 1]) + 1, length(coefs))] <- rownames(cc)
        if (sd) {
          newsd <- c(newsd, cc[, 2])
        }
      }
    }
    return(list(coefs = coefs, sds = newsd))
  }
  if (bollenstine) {
    e0 <- estimate(x, data = data, control = control, messages = 0, index = FALSE, ...)
    mm <- modelVar(e0)
    mu <- mm$xi
    Y <- t(t(data[, manifest(e0)]) - as.vector(mu))
    Sigma <- mm$C
    S <- (ncol(Y) - 1) / ncol(Y) * var(Y)
    sSigma <- with(eigen(Sigma), vectors %*% diag(sqrt(values), ncol = ncol(vectors)) %*% t(vectors))
    isS <- with(eigen(S), vectors %*% diag(1 / sqrt(values), ncol = ncol(vectors)) %*% t(vectors))
    data <- as.matrix(Y) %*% (isS %*% sSigma)
    colnames(data) <- manifest(e0)
  }

  i <- 0
  if (!missing(mc.cores)) {
    res <- parallel::mclapply(0:R, bootfun, mc.cores=mc.cores)
  } else {
    res <- do.call(future_lapply, c(list(0:R, bootfun), future.args))
  }

  coefs <- matrix(unlist(lapply(res, function(x) x$coefs)), nrow = R + 1, byrow = TRUE)
  nn <- names(res[[1]]$coefs)
  if (!is.null(nn)) colnames(coefs) <- nn
  sds <- NULL
  if (sd) {
    sds <- matrix(unlist(lapply(res, function(x) x$sds)), nrow = R + 1, byrow = TRUE)
  }

  if (!is.null(fun)) {
    rownames(coefs) <- c()
    res <- list(coef = coefs[-1, , drop = FALSE], coef0 = coefs[1, ], model = x)
  } else {
    colnames(coefs) <- names(res[[1]]$coefs)
    rownames(coefs) <- c()
    if (sd) colnames(sds) <- colnames(coefs)
    res <- list(coef = coefs[-1, , drop = FALSE], sd = sds[-1, , drop = FALSE], coef0 = coefs[1, ], sd0 = sds[1, ], model = x, bollenstine = bollenstine)
  }
  class(res) <- "bootstrap.lvm"
  return(res)
}

##' @export
bootstrap.lvmfit <- function(x, R = 100, data = model.frame(x),
                             control = list(start = coef(x)),
                             p = coef(x), parametric = FALSE, bollenstine = FALSE,
                             estimator = x$estimator, weights = Weights(x), ...) {
  bootstrap.lvm(Model(x), R = R, data = data, control = control, estimator = estimator, weights = weights, parametric = parametric, bollenstine = bollenstine, p = p, ...)
}

##' @export
"print.bootstrap.lvm" <- function(x, idx, level = 0.95, ...) {
  cat("Non-parametric bootstrap statistics (R=", nrow(x$coef), "):\n\n", sep = "")
  uplow <- (c(0, 1) + c(1, -1) * (1 - level) / 2)
  nn <- paste(uplow * 100, "%")
  c1 <- t(apply(x$coef, 2, function(x) c(mean(x), sd(x), quantile(x, uplow))))

  c1 <- cbind(x$coef0, c1[, 1] - x$coef0, c1[, -1, drop = FALSE])
  colnames(c1) <- c("Estimate", "Bias", "Std.Err", nn)
  if (missing(idx)) {
    print(format(c1, ...), quote = FALSE)
  } else {
    print(format(c1[idx, , drop = FALSE], ...), quote = FALSE)
  }
  if (length(x$sd) > 0) {
    c2 <- t(apply(x$sd, 2, function(x) c(mean(x), sd(x), quantile(x, c(0.025, 0.975)))))
    c2 <- cbind(c2[, 1], c2[, 1] - x$sd0, c2[, -1])
    colnames(c2) <- c("Estimate", "Bias", "Std.Err", "2.5%", "97.5%")
    cat("\nStandard errors:\n")
    if (missing(idx)) {
      print(format(c2, ...), quote = FALSE)
    } else {
      print(format(c2[idx, , drop = FALSE], ...), quote = FALSE)
    }
  }
  cat("\n")
  invisible(x)
}

##'Appending \code{Surv} objects
##'
##'\code{rbind} method for \code{Surv} objects
##'
##' 
##'@param ... \code{Surv} objects
##'@return \code{Surv} object
##'@author Klaus K. Holst
##'@keywords utilities
##'@examples
##'
##'y <- yl <- yr <- rnorm(10)
##'yl[1:5] <- NA; yr[6:10] <- NA
##'S1 <- survival::Surv(yl,yr,type="interval2")
##'S2 <- survival::Surv(y,y>0,type="right")
##'S3 <- survival::Surv(y,y<0,type="left")
##'
##'rbind(S1,S1)
##'rbind(S2,S2)
##'rbind(S3,S3)
##'
##' @export
rbind.Surv <- function(...) 
{
  dots <- list(...)
  type <- attributes(dots[[1]])$type
  ncol <- dim(dots[[1]])[2]
  nrow <- unlist(lapply(dots,nrow))
  cnrow <- c(0,cumsum(nrow))
  M <- matrix(ncol=ncol,nrow=sum(nrow))
  for (i in seq_along(dots)) {
    M[(cnrow[i]+1):cnrow[i+1],] <- dots[[i]]
  }
  x <- c(); for (i in seq_len(ncol(M))) x <- c(x,list(M[,i]))
  x <- c(x,list(type=type))
  do.call(survival::Surv, x)
} 

rchisqsum <- function(n,lambda) {
    p <- length(lambda)
    X2 <- matrix(rnorm(n*p)^2,ncol=p) ## Chi-squared (df=1)
    res <- numeric(n)
    for (i in seq(p)) {
        res <- res + X2[,i]*lambda[i]
    }
    return(res)
}

pchisqsum <- function(x, lambda=1, B=1e6, seed=NULL) {
    if (!is.null(seed)) set.seed(seed)
    y <- rchisqsum(B,lambda)
    if (length(x) == 1L) return(mean(y <= x))
    vapply(x, function(xi) mean(y <= xi), numeric(1))
}

##' @export
pars.survreg <- function(x,...) {
    c(coef(x),scale=x$scale)    
}


##' @export
score.survreg <- function(x,p,scale=TRUE,logscale=FALSE,indiv.logLik=FALSE,...) {    
    npar <- NROW(x$var)
    m <- model.frame(x)
    X <- model.matrix(terms(x), m)
    hasscale <- npar>length(x$coefficients)
    if (!missing(p)) {
        if (hasscale) sigma <- tail(p,1)
        p <- p[seq(length(p)-1)]
        x$linear.predictors <- as.vector(X%*%p)
        x$coefficients <- p
        x$scale <- sigma
    }
    derivatives <- residuals(x, type = "matrix")
    w <- model.weights(m)
    if (is.null(w)) w <- 1
    dldLP <- w*derivatives[,"dg"] ## Derivative wrt linear-predictor p=Xbeta
    S <- apply(X,2,function(x) x*dldLP)
    if (!is.null(x$naive.var)) {
        V <- x$naive.var
    } else {
        V <- x$var
    }
    if (hasscale && scale) {
        ds <- cbind("logsigma"=derivatives[,"ds"])
        if (!logscale) {
            ds <- ds/x$scale
            names(ds) <- "sigma"            
        }
        S <- cbind(S,ds)
    }
    if (hasscale && !scale) {
        V <- V[-npar,-npar,drop=FALSE]
    }
    attributes(S)$logLik <- 
                    if (indiv.logLik) derivatives[,"g"]
                    else sum(derivatives[,"g"])    
    attributes(S)$bread <- V*NROW(S)
    return(S)
}

##' @export
estimate.list <- function(x,...) {
    if (inherits(x[[1]],"lvm")) return(estimate_lvmlist(x,...))
    res <- lapply(x,function(x) estimate(x,...))
    class(res) <- c("estimate.list","list")
    res
}

##' @export
coef.estimate.list <- function(object,...) {
    lapply(object,coef)
}

##' @export
fixsome <- function(x, exo.fix=TRUE, measurement.fix=TRUE,
             S, mu, n, data, x0=FALSE, na.method="complete.obs",
             param=lava.options()$param,...) {

    if (is.null(param)) {
        param <- "none"
    } else {
        paramval <- c("hybrid","relative","none","absolute")
        param <- agrep(param,paramval,max.distance=0,value=TRUE)
    }

    if (is.character(measurement.fix)) {
        param <- measurement.fix
        measurement.fix <- TRUE
    }
    var.missing <- c()
    if (!missing(data) | !missing(S)) {

        if (!missing(data)) {
            dd <- procdata.lvm(x,data=data,na.method=na.method)
        } else {
            dd <- procdata.lvm(x, list(S=S,mu=mu,n=n))
        }
        S <- dd$S; mu <- dd$mu; n <- dd$n
        var.missing <- setdiff(index(x)$manifest,colnames(S))
    } else { S <- NULL; mu <- NULL }

    if (measurement.fix & param!="none") {
        if (length(var.missing)>0) {## Convert to latent:
            new.lat <- setdiff(var.missing,latent(x))
            if (length(new.lat)>0)
                x <- latent(x, new.lat)
        }
        etas <- latent(x)
        ys <- endogenous(x)
        M <- x$M

        for (e in etas) { ## Makes sure that at least one arrow from latent variable is fixed (identification)
            ys. <- names(which(M[e,ys]==1))
            if (length(ys.)>0) {
                if (tolower(param)=="absolute") {
                    if (is.na(intercept(x)[[e]])) intercept(x,e) <- 0
                    if (is.na(x$covfix[e,e]) & is.na(x$covpar[e,e])) covariance(x,e) <- 1
                } else {
                    if (param=="hybrid") {
                        if (is.na(intercept(x)[[e]])) intercept(x,e) <- 0
                        if (all(is.na(x$fix[e, ]==1)) &
                            is.na(x$covpar[e,e]) & is.na(x$covfix[e,e]))
                            regfix(x,from=e,to=ys.[1]) <- 1
                    } else { ## relative
                        if (all(is.na(x$fix[e, ]==1)) &
                            is.na(x$covpar[e,e]) & is.na(x$covfix[e,e]))
                            regfix(x,from=e,to=ys.[1]) <- 1
                        if (!any(unlist(lapply(intercept(x)[ys.],is.numeric))) & is.na(intercept(x)[[e]])) {
                            if (tryCatch(any(idx <- !is.na(x$fix[e,ys.])),error=function(x) FALSE)) {
                                intercept(x, ys.[which(idx)[1]]) <- 0
                            } else {
                                intercept(x,ys.[1]) <- 0
                            }
                        }
                    }
                }
            }
        }
    }

    if (is.null(S)) x0 <- TRUE
    if (exo.fix) {
        if (x0) {
            S0 <- diag(nrow=length(index(x)$manifest))
            mu0 <- rep(0,nrow(S0))
        }
        else {
            S0 <- S
            S0[is.na(S0)] <- 0
            mu0 <- mu
            e0 <- eigen(S0)
            thres <- lava.options()$itol^(1/2)
            if (any(e0$values<thres)) {
                ## Projection to nearest positive definite matrix
                ii <- e0$values
                ii[ii<thres] <- thres
                S0 <- e0$vectors%*%diag(ii,nrow=length(ii))%*%t(e0$vectors)
            }

        }
        exo.idx <- index(x)$exo.obsidx;
        exo_all.idx <- index(x)$exo.idx
        if (length(exo.idx)>0) {
            for (i in seq_along(exo.idx))
                for (j in seq_along(exo.idx)) {
                    i. <- exo_all.idx[i]; j. <- exo_all.idx[j]
                    myval <- S0[exo.idx[i],exo.idx[j]];
                    if (i.==j. & myval==0) {
                        ##warning("Overparametrized model. Problem with '"%++%index(x)$vars[j.]%++%"'")
                        myval <- 1
                    }
                    else if (is.na(myval) || is.nan(myval)) myval <- 0
                    x$covfix[i.,j.] <- x$covfix[j.,i.] <- myval
                }
            x$mean[exo_all.idx] <- mu0[exo.idx]
        }
    }

    index(x) <- reindex(x)
    return(x)
}

##' @export
`modelPar` <-
  function(x,p,...) UseMethod("modelPar")

###{{{ modelPar.lvmfit
##' @export
modelPar.lvmfit <- function(x, p=pars(x), ...) modelPar(Model(x),p=p,...)

###}}} modelPar.lvmfit

###{{{ modelPar

##' @export
modelPar.lvm <- function(x,p, ...) {
  npar <- index(x)$npar
  npar.mean <- index(x)$npar.mean
  if (length(p)!=npar & length(p)<(npar+npar.mean)) stop("Wrong dimension of parameter vector!")
  p2 <- NULL
  if (length(p)!=npar) { ## if meanstructure
    meanpar <- p[seq_len(npar.mean)]
    p. <- p
    if (length(meanpar)>0) {
        p. <- p[-seq_len(npar.mean)]
    } else meanpar <- NULL
    p <- p.[seq_len(npar)]
    if (npar>0) {
        p2 <- p.[-seq_len(npar)]
    } else p2 <- p.
  } else {
    meanpar <- NULL
    p2 <- NULL
  }
  return(list(p=p,meanpar=meanpar,p2=p2))
}

###}}} modelpar.lvm

###{{{ modelPar.multigroupfit

##' @export
modelPar.multigroupfit <- function(x,p=pars(x),...) {
  modelPar(Model(x),p,...)
}
###}}}

###{{{ modelPar.multigroup

##' @export
modelPar.multigroup <- function(x,p, ...) {
  if (length(p)==x$npar) {
    pp <- lapply(x$parposN,function(z) p[z])
    res <- list(p=pp, par=pp, mean=NULL)
    return(res)
  }
  ppos <- x$parposN
  pp <- lapply(ppos,function(z) p[z+x$npar.mean])

  if (length(pp)==0) pp <- lapply(seq_len(x$ngroup),function(x) logical())
  mm <- lapply(x$meanposN,function(x) p[x])
  if (is.null(mm)) mm <- lapply(seq_len(x$ngroup),logical())
  pm <- mm
  for (i in seq_len(length(pm))) pm[[i]] <- c(pm[[i]],pp[[i]])
  res <- list(p=pm,par=pp,mean=mm)
  return(res)
}

###}}}

modelPar2.multigroup <-
  function(x,p, ...) {
  npar <- x$npar
  npar.mean <- x$npar.mean
  k <- x$ngroup
  if (length(p)!=npar & length(p)!=(npar+npar.mean)) stop("Wrong dimension of parameter vector!")
  if (length(p)!=npar) { ## if meanstructure
      meanpar <- p[seq_len(npar.mean)]
      p. <- p[-seq_len(npar.mean)]
    } else {
      meanpar <- NULL
      p. <- p
    }


  parlist <- list(); for (i in seq_len(k)) parlist[[i]] <- numeric(length(x$parlist[[i]]))
  if (!is.null(meanpar)) {
    meanlist <- list(); for (i in seq_len(k)) meanlist[[i]] <- numeric(length(x$meanlist[[i]]))
  }

  if (length(p.)>0)
  for (i in seq_along(p.)) {
    for (j in seq_len(k)) {
      idx <- match(paste0("p",i), x$parlist[[j]])
      if (!is.na(idx))
        parlist[[j]][idx] <- p.[i]
      if (!is.null(meanpar)) {
        midx <- match(paste0("p",i), x$meanlist[[j]])
        if (!is.na(midx))
          meanlist[[j]][midx] <- p.[i]
      }
    }
  }

  if (!is.null(meanpar)) {
    for (i in seq_along(meanpar)) {
      for (j in seq_len(k)) {
        idx <- match(paste0("m",i), x$meanlist[[j]])
        if (!is.na(idx))
          meanlist[[j]][idx] <- meanpar[i]
      }
    }
  } else {
    meanlist <- NULL
  }
  p0 <- parlist
  for (i in seq_along(p0))
    p0[[i]] <- c(meanlist[[i]],parlist[[i]])
  return(list(p=p0, par=parlist, mean=meanlist))
}

##' Wait for user input (keyboard or mouse)
##'
##' @title Wait for user input (keyboard or mouse)
##' @aliases waitclick
##' @author Klaus K. Holst
##' @export
##' @keywords iplot
wait <- function() {
    cat(gettext("\nPress <Enter> to continue..."))
    res <- try(scan("", what=0, quiet=TRUE, nlines=1), silent=TRUE)
}
waitclick <- function() if(is.null(locator(1))) invisible(NULL)

glm_estimate_hook <- function(x,estimator,...) {
    yy <- c()
    if (length(estimator)>0 && estimator=="glm") {
        for (y in endogenous(x)) {
            fam <- attributes(distribution(x)[[y]])$family
            if (is.null(fam)) fam <- stats::gaussian()
            if (!(tolower(fam$family)%in%
                  c("gaussian","gamma","inverse.gaussian","weibull"))) {
                yy <- c(yy,y)
            }
        }
        if (length(yy)>0) covariance(x,yy) <- 1
    }
    return(c(list(x=x,estimator=estimator,...)))
}

GLMest <- function(m,data,control=list(),...) {
    yvar <- endogenous(m)
    res <- c()
    count <- 0
    V <- NULL
    mymsg <- c()
    ics <- c()
    breads <- c()

    et <- eventTime(m)
    yvar.et <- rep(NA,length(yvar))
    names(yvar.et) <- yvar
    if (!is.null(et)) {
        for (i in seq_along(et)) {
            ## if (!survival::is.Surv(data[,et[[i]]$names[1]]))
            ##         data[,et[[i]]$names[1]] <- with(et[[i]],
            ##         survival::Surv(data[,names[1]],data[,names[2]]))
            yvar <- setdiff(yvar,c(et[[i]]$latentTimes[-1],et[[i]]$names))
            yvar.et[et[[i]]$latentTimes[1]] <- et[[i]]$names[1]
        }
    }

    for (y in yvar) {
        count <- count+1
        xx <- parents(m, y)
        fam <- attributes(distribution(m)[[y]])$family
        if (is.null(fam)) fam <- stats::gaussian()
        if (!is.null(fam$link)) {
            mymsg <- c(mymsg, with(fam, paste0(family,"(",link,")")))
        } else {
            mymsg <- c(mymsg, with(fam, paste0(family)))
        }
        if (length(xx)==0) xx <- 1
        nn0 <- paste(y,xx,sep=lava.options()$symbol[1])
        y0 <- y

        f <- as.formula(paste0(y, "~",
          paste(xx, collapse = "+")
        ))
        isSurv <- inherits(data[1, y], "Surv")
        if (isSurv) {
          g <- survival::survreg(f,data=data,dist=fam$family,...)
        } else {
          g <- glm(f,family=fam,data=data,...)
        }
        p <- pars(g)
        ii <- IC(g)
        V0 <- attr(ii, "bread")
        ics <- cbind(ics,ii)
        y <- y0
        names(p)[1] <- y
        if (length(p)>1) {
          xx0 <- setdiff(xx, 1)
          if (length(xx0) > 0) {
            nn0 <- paste(y, xx0, sep = lava.options()$symbol[1])
            names(p)[seq_along(nn0) + 1] <- nn0
          }
          if (length(p)>length(xx0)+1) names(p)[length(p)] <- paste(y,y,sep=lava.options()$symbol[2])
        }
        if (tolower(fam$family)%in%c("gaussian","gamma","inverse.gaussian") && !isSurv) {
            ics <- cbind(ics,0)
            null <- matrix(0); dimnames(null) <- list("scale","scale")
            V0 <- blockdiag(V0,null,pad=0)
        }
        breads <- c(breads, list(V0))
        res <- c(res, list(p));
    }
    coefs <- unlist(res)
    idx <- na.omit(match(coef(m),names(coefs)))
    coefs <- coefs[idx]
    V <- var_ic(ics[, idx])
    mymsg <- noquote(cbind(mymsg))
    colnames(mymsg) <- "Family(Link)"; rownames(mymsg) <- paste(yvar,":")
    list(estimate=coefs, vcov=V, breads=breads, IC=ics[,idx],
         summary.message=function(...)  { mymsg }, dispname="Dispersion:")
}

GLMscore <- function(x,p,data,indiv=TRUE,logLik=FALSE,...) {
    yvar <- endogenous(x)
    S <- pnames <- c()
    count <- 0
    breads <- c()
    L <- 0
    for (y in yvar) {
        count <- count+1
        xx <- parents(x,y)
        pname <- c(y,paste0(y,sep=lava.options()$symbol[1],xx),paste(y,y,sep=lava.options()$symbol[2]))
        pidx <- na.omit(match(pname,coef(x)))
        fam <- attributes(distribution(x)[[y]])$family
        if (is.null(fam)) fam <- stats::gaussian()
        if (length(xx)==0) xx <- 1
        f <- as.formula(paste0(y,"~",paste(xx,collapse="+")))
        isSurv <- inherits(data[1,y],"Surv")
        if (inherits(data[,y],"Surv")) {
            g <- survival::survreg(f,data=data,dist=fam$family)
        } else {
            g <- glm(f,family=fam,data=data)
        }
        p0 <- p[pidx]
        if (!isSurv) L0 <- logL.glm(g,p=p0,indiv=TRUE,...)
        if (tolower(fam$family)%in%c("gaussian","gamma","inverse.gaussian") && !isSurv) {
            p0 <- p0[-length(p0)]
            S0 <- score(g,p=p0,indiv=TRUE,pearson=TRUE,...)
            V0 <- attr(S0,"bread")
            r <- attr(S0,"pearson")
            ## dispersion <- mean(r^2)
            S0 <- cbind(S0,scale=0)
            null <- matrix(0); dimnames(null) <- list("scale","scale")
            V0 <- blockdiag(V0,null,pad=0)
        } else {
            S0 <- score(g,p=p0,indiv=TRUE,...)
            if (isSurv) L0 <- attr(S0,"logLik")
            V0 <- attr(S0,"bread")
        }
        L <- L+sum(L0)
        breads <- c(breads,list(V0))
        S <- c(S,list(S0))
        pnames <- c(pnames, list(pname));
    }
    coefs <- unlist(pnames)
    idx <- na.omit(match(coefs,coef(x)))
    idx <- order(idx)
    V <- Reduce(blockdiag,breads)[idx,idx]
    S1 <- Reduce(cbind,S)[,idx,drop=FALSE]
    colnames(S1) <- coef(x)
    attributes(S1)$bread <- V
    attributes(S1)$logLik <- structure(L,nobs=nrow(data),nall=nrow(data),df=length(p),class="logLik")
    if (!indiv) S1 <- colSums(S1)
    return(S1)
}


##' @export
score.lm <- function(x, p=coef(x), data, indiv=FALSE,
              y, X, offset=NULL, weights=NULL, dispersion=TRUE, ...) {
    if (missing(data)) {
        X <- model.matrix(x)
        y <- model.frame(x)[,1]
    } else {
        X <- model.matrix(formula(x),data=data)
        y <- model.frame(formula(x),data=data)[,1]
    }
    if(any(is.na(p))) warning("Over-parameterized model")
    Xbeta <- X%*%p
    if (is.null(offset)) offset <- x$offset
    if (!is.null(offset)) Xbeta <- Xbeta+offset
    r <- y-Xbeta
    if (is.null(weights)) weights <- x$weights
    if (!is.null(weights)) {
        sigma2 <- sum(r^2*weights)/(length(r)-length(p))
        r <- r*weights
    } else {
        sigma2 <- sum(r^2)/(length(r)-length(p))
    }
    if (!dispersion) sigma2 <- 1
    ##sigma2 <- suppressWarnings(summary(x)$sigma^2)
    A <- as.vector(r)/sigma2
    S <- apply(X,2,function(x) x*A)
    if (!indiv) return(colSums(S))
    suppressWarnings(attributes(S)$bread <- vcov(x)*NROW(S))
    return(S)
}

##' @export
score.glm <- function(x,p=coef(x),data,indiv=FALSE,pearson=FALSE,
               y,X,link,dispersion,offset=NULL,weights=NULL,...) {

    if (inherits(x,"glm")) {
        link <- family(x)
        if (missing(data)) {
            X <- model.matrix(x)
            y <- model.frame(x)[,1]
        } else {
            X <- model.matrix(formula(x),data=data)
            y <- model.frame(formula(x),data=data)[,1]
        }
        offset <- x$offset
    } else {
        if (missing(link)) stop("Family needed")
        if (missing(data)) stop("data needed")
        X <- model.matrix(formula(x),data=data)
        y <- model.frame(formula(x),data=data)[,1]
    }
    if (is.character(y) || is.factor(y)) {
        y <- as.numeric(as.factor(y)) - 1
    }
    ## g <- link$linkfun
    ginv <- link$linkinv
    dginv <- link$mu.eta ## D[linkinv]
    ##dg <- function(x) 1/dginv(g(x)) ## Dh^-1 = 1/(h'(h^-1(x)))
    if (inherits(x, "negbin")) {
        Dcanlink <- function(x) 1/x
    } else {
        canonf <- do.call(link$family,list())
        ## caninvlink <- canonf$linkinv
        canlink <- canonf$linkfun
        Dcaninvlink <- canonf$mu.eta
        Dcanlink <- function(x) 1 / Dcaninvlink(canlink(x))
    }
    ##gmu <- function(x) g(caninvlink(x))
    ##invgmu <- function(z) canlink(ginv(z))
    h <- function(z) Dcanlink(ginv(z))*dginv(z)
    if(any(is.na(p))) stop("Over-parameterized model")
    Xbeta <- X%*%p
    if (!is.null(offset)) Xbeta <- Xbeta+offset
    if (missing(data) && !is.null(x$offset) && is.null(offset) ) Xbeta <- Xbeta+x$offset
    pi <- ginv(Xbeta)
    r <- y-pi
    if (!is.null(x$prior.weights) || !is.null(weights)) {
        if (is.null(weights)) weights <- x$prior.weights
    } else {
        weights <- !is.na(r)
    }
    a.phi <- 1
    r <- r*weights
    rpearson <- as.vector(r)/link$variance(pi)^.5
    if (length(p) > length(coef(x))) {
        a.phi <- p[length(coef(x)) + 1]
    } else { ## if (tolower(family(x)$family)%in%c("gaussian","gamma","inverse.gaussian"))
        suppressWarnings(disp <- summary(x)$dispersion)
        if (!is.null(disp)) a.phi <- disp
        if (inherits(x, "negbin")) {
            a.phi <- link$variance(pi) / pi
        }
        ## a.phi <- sum(rpearson^2)*x$df.residual/x$df.residual^2
    }
    A <- as.vector(h(Xbeta) * r) / a.phi
    S <- apply(X, 2, function(x) x * A)
    if (!indiv) return(colSums(S))
    if (pearson) attr(S,"pearson") <- rpearson
    suppressWarnings(attributes(S)$bread <- vcov(x)*NROW(S))
    if (x$family$family=="quasi" && x$family$link=="identity" && x$family$varfun=="constant")
        attributes(S)$bread <- -Inverse(information.glm(x)*NROW(S))
    return(S)
}

##' @export
pars.glm <- function(x,...) {
    if (tolower(family(x)$family)%in%c("gaussian","gamma","inverse.gaussian")) {
        res <- c(coef(x),suppressWarnings(summary(x)$dispersion))
        names(res)[length(res)] <- "Dispersion"
        return(res)
    }
    return(coef(x))
}

logL.glm <- function(x,p=pars.glm(x),data,indiv=FALSE,...) {
    if (!missing(data)) {
        x <- update(x,data=data,...)
    }
    f <- family(x)
    ginv <- f$linkinv
    X <- model.matrix(x)
    n <- nrow(X)
    ##disp <- 1;
    p0 <- p
    if (tolower(family(x)$family)%in%c("gaussian","gamma","inverse.gaussian")) {
        if (length(p)==ncol(X)) {
            ##disp <- suppressWarnings((summary(x)$dispersion))
        } else {
            ##disp <- tail(p,1)
            p0 <- p[-length(p)]
        }
    }
    if(any(is.na(p))) {
        warning("Over-parametrized model")
    }
    Xbeta <- X%*%p0
    if (!is.null(x$offset)) Xbeta <- Xbeta+x$offset
    y <- model.frame(x)[,1]
    mu <- ginv(Xbeta)
    w <- x$prior.weights
    dev <-  f$dev.resids(y,mu,w)
    if (indiv) {

    }
    loglik <- length(p)-(f$aic(y,n,mu,w,sum(dev))/2+x$rank)
    structure(loglik,nobs=n,df=length(p),class="logLik")
}

##' @export
IC.glm <- function(x,...) {
    IC.default(x,...)
}

hessian.glm <- function(x,p=coef(x),...) {
    numDeriv::jacobian(function(theta) score.glm(x,p=theta,indiv=FALSE,...),p)
}

##' @export
information.glm <- function(x,...) hessian.glm(x,...)

robustvar <- function(x,id=NULL,...) {
    U <- score(x,indiv=TRUE)
    II <- unique(id)
    K <- length(II)
    J <- 0
    if (is.null(id)) {
        J <- crossprod(U)
    } else {
        for (ii in II) {
            J <- J+tcrossprod(colSums(U[which(id==ii),,drop=FALSE]))
        }
        J <- K/(K-1)*J
    }
    iI <- vcov(x)
    V <- iI%*%J%*%iI
    return(V)
}

glm_logLik.lvm <- function(object,...) {
    attr(GLMscore(object,...),"logLik")
}

glm_method.lvm <- NULL
glm_objective.lvm <- function(x,p,data,...) {
    GLMest(x,data,...)
}
glm_gradient.lvm <- function(x,p,data,...) {
    -GLMscore(x,p,data,...)
}

glm_variance.lvm <- function(x,p,data,opt,...) {
    opt$vcov
}

context("glm components")

test_that("score.glm can handle a overparameterized model (NA coef)", {
  set.seed(1)
  n <- 100
  data <- data.frame(
    id = seq_len(n),
    a = rbinom(n, 1, 0.5),
    x = rnorm(n)
  )
  data$x_dup <- data$x  # exact collinearity with `x`
  data$y <- 1 + data$a + data$x + rnorm(n)

  m_ref <- glm(y ~ a * x, data = data)

  score_ref <- score.glm(m_ref, indiv = FALSE)

  m <- glm(y ~ a * x + x_dup, data = data)
  idx_m_ref <- names(coef(m)) != "x_dup"
  score_na <- expect_warning(
    score.glm(m, indiv = FALSE),
    "Over-parameterized"
  )

  expect_equal(score_ref, score_na[!is.na(coef(m))])

  expect_true(score_na["x_dup"] == 0)

  score_ref <- score.glm(m_ref, indiv = TRUE)
  score_na <- suppressWarnings(score.glm(m, indiv = TRUE))

  expect_equal(score_ref, score_na[,!is.na(coef(m))],
               check.attributes = FALSE)

  expect_equal(
    attr(score_ref, "bread"),
    attr(score_na, "bread")[idx_m_ref,
                            idx_m_ref]
  )

  expect_true(
    all(attr(score_na, "bread")[names(coef(m)) == "x_dup", ] == 0) &&
      all(attr(score_na, "bread")[, names(coef(m)) == "x_dup"] == 0)
  )

  ## check IC
  expect_equal(
    IC(m_ref),
    suppressWarnings(IC(m))[, idx_m_ref],
    check.attributes = FALSE
  )

  ## check hessian
  suppressWarnings(I <- information(m))
  expect_equal(
    I[idx_m_ref, idx_m_ref],
    information(m_ref),
    check.attributes = FALSE
  )
  expect_equal(
    Inverse(I)[idx_m_ref, idx_m_ref],
    vcov(m_ref),
    check.attributes = FALSE
  )

  ## logL
  ll <- suppressWarnings(lava:::logL.glm(m))
  expect_equal(ll, logLik(m_ref),
    check.attributes = FALSE
  )

  ## user should get a warning with overparametrized models
  expect_warning(estimate(m))

})

##' vec operator
##'
##' Convert array into vector
##' @title vec operator
##' @param x Array
##' @param matrix If TRUE a row vector (matrix) is returned
##' @param sep Seperator
##' @param ... Additional arguments
##' @author Klaus Holst
##' @export
vec <- function(x,matrix=FALSE,sep=".",...) {
    if (is.vector(x) && !is.list(x)) {
        res <- x
    } else if (is.list(x)) {
        res <- stats::setNames(unlist(x),names(x))
    } else {
        if (is.matrix(x) && is.null(rownames(x))) {
            nn <- colnames(x)
        } else {
            nn <- apply(expand.grid(dimnames(x)),1,function(x) paste(x,collapse=sep))
        }
        res <- as.vector(x); names(res) <- nn
    }
    if (matrix) return(cbind(res))
    return(res)
}

##' Returns device-coordinates and plot-region
##'
##' @title Returns device-coordinates and plot-region
##' @return A `list` with elements
##'  * `dev.x1`: device left x-coordinate
##'  * `dev.x2`: device right x-coordinate
##'  * `dev.y1`: device bottom y-coordinate
##'  * `dev.y2`: device top y-coordinate
##'  * `fig.x1`: plot left x-coordinate
##'  * `fig.x2`: plot right x-coordinate
##'  * `fig.y1`: plot bottom y-coordinate
##'  * `fig.y2`: plot top y-coordinate
##' @author Klaus K. Holst
##' @export
##' @keywords hplot
`devcoords` <- function() {
  cc <- par("usr") ## extremes of coordinates of plotting region (x1,x2,y1,y2)
  plotinch <- par("pin") ## Plot dimensions (width,height) in inches
  margininch <- par("mai") ## Margin sizes in inches (bottom, left, top ,right)
  plotlenX <- cc[2]-cc[1]
  unitinchX <- plotlenX/plotinch[1]
  plotlenY <- cc[4]-cc[3]
  unitinchY <- plotlenY/plotinch[2]
  deviceXleft <- cc[1]-unitinchX*margininch[2]
  deviceXright <- cc[2]+unitinchX*margininch[4]
  deviceYtop <- cc[4]+unitinchY*margininch[3]
  deviceYbottom <- cc[3]-unitinchY*margininch[1]
  return(list(dev.x1=deviceXleft, dev.x2=deviceXright,
              dev.y1=deviceYbottom, dev.y2=deviceYtop,
              fig.x1=cc[1], fig.x2=cc[2],
              fig.y1=cc[3], fig.y2=cc[4]))
}

sumsplit <- function(x, ...) {
    plus <- strsplit(x, "\\+")[[1]]
    spl <- unlist(lapply(plus, function(x) {
        val <- strsplit(x, "\\-")[[1]]
        val[-1] <- paste0("-", val[-1])
        setdiff(val, "")
    }))
    res <- c()
    for (st in spl) {
        st <- gsub(" ", "", st)
        st0 <- gsub("^[-0-9\\*]*", "", st)
        val <- gsub(
          "\\*", "",
          regmatches(st, gregexpr("^[-0-9\\*]*", st))[[1]]
        )
        if (val == "") val <- "1"
        val <- switch(val, "-"=-1, val)
        res <- c(res, val, st0)
    }
    return(res)
}

##' @export
parsedesign <- function(coef, x, ..., regex=FALSE, diff=TRUE) {
    if (!is.vector(coef)) coef <- stats::coef(coef)
    if (is.numeric(coef) && !is.null(names(coef))) coef <- names(coef)
    dots <- lapply(substitute(list(...)), function(x) x)[-1]
    expr <- suppressWarnings(inherits(try(x, silent=TRUE), "try-error"))
    if (expr) {
        ee <- c(deparse(substitute(x)), unlist(lapply(dots, deparse)))
    } else {
        ee <- c(deparse(x), sapply(dots, function(x) deparse(x)))
    }
    if (!expr && is.numeric(x)) {
        return(do.call(
          contr,
          list(c(list(x), list(...)),
               n = length(coef), diff = diff)
        ))
    }
    res <- c()
    diff <- rep(diff, length.out=length(ee))
    count <- 0
    for (e in ee) {
        count <- count+1
        diff0 <- FALSE
        Val <- rbind(rep(0, length(coef)))
        if (grepl('\"', e)) {
            diff0 <- diff[count] && grepl("^c\\(", e)
            e0 <- gsub(" |\\)$|^c\\(", "", e)
            ff <- strsplit(e0, '\"')[[1L]]
        } else {
            ff <- sumsplit(e)
        }
        for (i in seq(length(ff)/2)) {
            val0 <- gsub("[*()]", "", ff[2*(i-1)+1])
            val <- char2num(val0)
            if (is.na(val)) {
                val <- switch(val0, "-"=-1, 1)
            }
            par0 <- ff[2*i]
            par0int <- as.integer(char2num(par0))
            if (!regex) par0 <- glob2rx(par0)
            if (is.na(par0int)) par0int <- grep(par0, coef)
            if (length(par0int)>1) {
                diff0 <- diff[count]
                for (k in seq_along(par0int)) {
                    if (par0int[k]<=length(Val)) {
                        if (diff[count]) {
                            Val[par0int[k]] <- val
                        } else {
                            Val0 <- Val
                            Val0[] <- 0
                            Val0[par0int[k]] <- val
                            res <- rbind(res, Val0)
                        }
                    }
                }
            } else {
              if (length(par0int) > 0 && par0int <= length(Val)) {
                Val[par0int] <- val
                }
            }
        }
        if (diff0) {
            n <- sum(Val!=0)
            if (n>1) {
                Val0 <- Val
                ii <- which(Val0!=0)
                Val <- matrix(0, nrow=n-1, ncol=length(Val))
                for (i in seq(n-1)) {
                    Val[i, ii[c(1, i+1)]] <- Val0[ii[c(1, i+1)]]*c(1, -1)
                }
            }
        }
        if (any(Val!=0)) res <- rbind(res, Val)
    }
    res
}

###{{{ plot.lvm

##' Plot path diagram
##'
##' Plot the path diagram of a SEM
##'
##'
##' @aliases plot.lvmfit
##' @param x Model object
##' @param diag Logical argument indicating whether to visualize
##'     variance parameters (i.e. diagonal of variance matrix)
##' @param cor Logical argument indicating whether to visualize
##'     correlation parameters
##' @param labels Logical argument indiciating whether to add labels
##'     to plot (Unnamed parameters will be labeled p1,p2,...)
##' @param intercept Logical argument indiciating whether to add
##'     intercept labels
##' @param addcolor Logical argument indiciating whether to add colors
##'     to plot (overrides \code{nodecolor} calls)
##' @param plain if TRUE strip plot of colors and boxes
##' @param cex Fontsize of node labels
##' @param fontsize1 Fontsize of edge labels
##' @param noplot if TRUE then return \code{graphNEL} object only
##' @param graph Graph attributes (Rgraphviz)
##' @param attrs Attributes (Rgraphviz)
##' @param unexpr if TRUE remove expressions from labels
##' @param addstyle Logical argument indicating whether additional
##'     style should automatically be added to the plot (e.g. dashed
##'     lines to double-headed arrows)
##' @param plot.engine default 'Rgraphviz' if available, otherwise
##'     visNetwork,igraph
##' @param init Reinitialize graph (for internal use)
##' @param layout Graph layout (see Rgraphviz or igraph manual)
##' @param edgecolor if TRUE plot style with colored edges
##' @param graph.proc Function that post-process the graph object
##'     (default: subscripts are automatically added to labels of the
##'     nodes)
##' @param ... Additional arguments to be passed to the low level
##'     functions
##' @author Klaus K. Holst
##' @keywords hplot regression
##' @examples
##'
##' if (interactive()) {
##' m <- lvm(c(y1,y2) ~ eta)
##' regression(m) <- eta ~ z+x2
##' regression(m) <- c(eta,z) ~ x1
##' latent(m) <- ~eta
##' labels(m) <- c(y1=expression(y[scriptscriptstyle(1)]),
##' y2=expression(y[scriptscriptstyle(2)]),
##' x1=expression(x[scriptscriptstyle(1)]),
##' x2=expression(x[scriptscriptstyle(2)]),
##' eta=expression(eta))
##' edgelabels(m, eta ~ z+x1+x2, cex=2, lwd=3,
##'            col=c("orange","lightblue","lightblue")) <- expression(rho,phi,psi)
##' nodecolor(m, vars(m), border="white", labcol="darkblue") <- NA
##' nodecolor(m, ~y1+y2+z, labcol=c("white","white","black")) <- NA
##' plot(m,cex=1.5)
##'
##' d <- sim(m,100)
##' e <- estimate(m,d)
##' plot(e)
##'
##' m <- lvm(c(y1,y2) ~ eta)
##' regression(m) <- eta ~ z+x2
##' regression(m) <- c(eta,z) ~ x1
##' latent(m) <- ~eta
##' plot(lava:::beautify(m,edgecol=FALSE))
##' }
##' @export
##' @method plot lvm
`plot.lvm` <-
  function(x,diag=FALSE,cor=TRUE,labels=FALSE,intercept=FALSE,addcolor=TRUE,plain=FALSE,cex,fontsize1=10,noplot=FALSE,graph=list(rankdir="BT"),
         attrs=list(graph=graph),
         unexpr=FALSE,
         addstyle=TRUE,plot.engine=lava.options()$plot.engine,init=TRUE,
         layout=lava.options()$layout,
         edgecolor=lava.options()$edgecolor,
         graph.proc=lava.options()$graph.proc,

         ...) {
    if (is.null(vars(x))) {
      message("Nothing to plot: model has no variables.")
      return(NULL)
    }
  index(x) <- reindex(x)
  myhooks <- gethook("plot.post.hooks")
  for (f in myhooks) {
    x <- do.call(f, list(x=x,...))
  }

    plot.engine <- tolower(plot.engine)
    if (plot.engine=="rgraphviz" && (!(requireNamespace("graph",quietly=TRUE)) || !(requireNamespace("Rgraphviz",quietly=TRUE)))) {
        plot.engine <- "visnetwork"
    }
    if (plot.engine=="visnetwork" && (!(requireNamespace("visNetwork",quietly=TRUE)))) {
        plot.engine <- "igraph"
    }
  if (plot.engine=="igraph") {
    if (!requireNamespace("igraph",quietly=TRUE)) {
      message("package 'Rgraphviz','igraph' or 'visNetwork' not available")
      return(NULL)
    }
    L <- igraph::layout.sugiyama(g <- igraph.lvm(x,...))$layout
    if (noplot) {
        return(igraph::as_graphnel(g))
    }
    dots <- list(...)
    if (is.character(layout))
      plot(g,layout=L,...)
    else plot(g,layout=layout,...)
    return(invisible(g))
  }
  if (plot.engine=="visnetwork" && !noplot) {
        g <- vis.lvm(x,labels=labels,...)
        return(g)
  }
  if (init) {
        if (!is.null(graph.proc)) {
            x <- do.call(graph.proc, list(x,edgecol=edgecolor,...))
        }
    g <- finalize(x,diag=diag,cor=cor,addcolor=addcolor,intercept=intercept,plain=plain,cex=cex,fontsize1=fontsize1,unexpr=unexpr,addstyle=addstyle)
  } else {
    g <- Graph(x)
  }
  if  (labels) {
    AP <- matrices(x,paste0("p",seq_len(index(x)$npar)))
    mylab <- AP$P; mylab[AP$A!="0"] <- AP$A[AP$A!="0"]
    mylab[!is.na(x$par)] <- x$par[!is.na(x$par)]
    mylab[!is.na(x$covpar)] <- x$covpar[!is.na(x$covpar)]
    g <- edgelabels(g, lab=mylab)
  }
  if (lava.options()$debug) {
    plot(g)
  } else {
    ## graphRenderInfo(g)$recipEdges <- "distinct"
    .savedOpt <- options(warn=-1) ## Temporarily disable warnings as renderGraph comes with a stupid warning when labels are given as "expression"
    dots <- list(...)
    dots$attrs <- attrs
    dots$x <- g
    dots$recipEdges <- "distinct"
    if (attributes(g)$feedback) dots$recipEdges <- c("combine")
    if (is.null(dots$layoutType)) dots$layoutType <- layout[1]
    if (all(index(x)$A==0))
      dots$layoutType <- "circo"

    if (requireNamespace("Rgraphviz",quietly=TRUE))
        g <- do.call(getFromNamespace("layoutGraph","Rgraphviz"), dots)
    ## Temporary work around:
    graph::nodeRenderInfo(g)$fill <- graph::nodeRenderInfo(dots$x)$fill
    graph::nodeRenderInfo(g)$col <- graph::nodeRenderInfo(dots$x)$col
    graph::edgeRenderInfo(g)$col <- graph::edgeRenderInfo(dots$x)$col
    if (noplot) return(suppressMessages(graph::updateGraph(g)))
    res <- tryCatch(Rgraphviz::renderGraph(g),error=function(e) NULL)
    ## Redo nodes to avoid edges overlapping node borders
    ##par(new=TRUE)
    ##res <- tryCatch(Rgraphviz::renderGraph(g,drawEdges=NULL,new=FALSE),error=function(e) NULL)
    ##
    options(.savedOpt)
  }

  myhooks <- gethook("plot.hooks")
  for (f in myhooks) {
    do.call(f, list(x=x,...))
  }

  invisible(g)
}

###}}} plot.lvm

###{{{ vis.lvm

vis.lvm <- function(m,randomSeed=1,width="100%",height="700px",labels=FALSE,cor=TRUE,...) {
    if (!requireNamespace("visNetwork",quietly=TRUE)) stop("'visNetwork' required")
    types <- rep("endogenous",length(vars(m)))
    types[index(m)$eta.idx] <- "latent"
    types[index(m)$exo.idx] <- "exogenous"
    col <- lava.options()$node.color
    colors <- rep(col[2],length(types))
    colors[index(m)$eta.idx] <- col[3]
    colors[index(m)$exo.idx] <- col[1]
    trf <- transform(m)
    if (length(trf)>0) {
        colors[which(index(m)$vars%in%names(trf))] <- col[4]
    }
    shapes <- rep("box",length(types))
    shapes[index(m)$eta.idx] <- "circle"
    nodes <- data.frame(id=seq_along(types),
                        label=vars(m),
                        color=colors,
                        shape=shapes,
                        shadow=TRUE,
                        size=rep(1.0,length(types)),
                        group=types)
    edges <- cbind(edgeList(m))#,shadow=TRUE)

    AP <- matrices(m,paste0("p",seq_len(index(m)$npar)))
    if (labels) {
        mylab <- AP$A;
        mylab[!is.na(m$par)] <- m$par[!is.na(m$par)]
        lab <- c()
        for (i in seq(nrow(edges))) {
            lab <- c(lab,t(mylab)[edges[i,1],edges[i,2]])
        }
        edges <- cbind(edges,label=lab)
    }
    if (length(edges)>0)
        edges <- cbind(edges,dashes=FALSE,arrows="from")


    if (cor) {
        mylab <- AP$P
        mylab[!is.na(m$covpar)] <- m$covpar[!is.na(m$covpar)]
        coredges <- data.frame(from=numeric(),to=numeric(),label=character())
        for (i in seq_len(nrow(mylab)-1)) {
            for (j in seq(i+1,nrow(mylab))) {
                if (mylab[i,j]!="0") {
                    coredges <- rbind(coredges,
                                      data.frame(from=i,to=j,label=mylab[i,j]))
                }
            }
        }
        if (nrow(coredges)>0) {
            if (!labels) coredges <- coredges[,1:2,drop=FALSE]
            coredges <- cbind(coredges,dashes=TRUE,arrows="false")
            edges <- rbind(edges,coredges)
        }
    }

    if (length(edges)>0) edges$physics <- TRUE
    v <- visNetwork::visNetwork(nodes,edges,width=width,height=height,...)
    v <- visNetwork::visEdges(v, arrows=list(from=list(enabled=TRUE, scaleFactor = 0.5)),
                              scaling = list(min = 2, max = 2))
    v <- visNetwork::visLayout(v,randomSeed=randomSeed)
    v
}

###}}} vis.lvm

###{{{ plot.lvmfit

##' @export
`plot.lvmfit` <-
    function(x,diag=TRUE,cor=TRUE,type,noplot=FALSE,fontsize1=5,f,graph.proc=lava.options()$graph.proc,...) {
        if (!missing(f)) {
            return(plot.estimate(x,f=f,...))
        }
    .savedOpt <- options(warn=-1) ## Temporarily disable warnings as renderGraph comes with a warning when labels are given as "expression"
    if (!requireNamespace("graph",quietly=TRUE)) {
      plot(Model(x),...)
      return(invisible(x))
    }
    g <- Graph(x)
    newgraph <- FALSE
    if (is.null(g)) {
        newgraph <- TRUE
        if (!is.null(graph.proc)) {
            Model(x) <- beautify(Model(x),edgecol=FALSE,...)
        }
        Graph(x) <- finalize(Model(x), diag=TRUE, cor=FALSE, fontsize1=fontsize1, ...)
    }
    if(noplot) return(Graph(x))
    if (newgraph) {
      if (missing(type))
        type <- "est"
      x <- edgelabels(x, type=type, diag=diag, cor=cor, fontsize1=fontsize1, ...)
    } else {
      if (!missing(type)) {
        x <- edgelabels(x, type=type, diag=diag, cor=cor, fontsize1=fontsize1, ...)
      }
    }
    g <- Graph(x)
    var <- rownames(covariance(Model(x))$rel)
    if (!cor) {
       delta <- 1
      for (r in seq_len(nrow(covariance(Model(x))$rel)-delta) ) {
        for (s in seq(r+delta,ncol(covariance(Model(x))$rel)) ) {
          if (covariance(Model(x))$rel[r,s]==1) {
            g <- graph::removeEdge(var[r],var[s], g)
            g <- graph::removeEdge(var[s],var[r], g)
          }
        }
      }
    }
    if (!diag) {
      for (r in seq_len(nrow(covariance(Model(x))$rel)) ) {
        if (graph::isAdjacent(g,var[r],var[r]))
          g <- graph::removeEdge(var[r],var[r],g)
      }
    }
    m <- Model(x); Graph(m) <- g
    g <- plot(m, diag=diag, cor=cor, fontsize1=fontsize1, init=FALSE, ...)
    options(.savedOpt)
    invisible(g)
  }

###}}} plot.lvmfit

###{{{ plot.multigroup

##' @export
plot.multigroup <- function(x,diag=TRUE,labels=TRUE,...) {
  k <- x$ngroup
  for (i in seq_len(k))
    plot(x$lvm[[i]],diag=diag,labels=labels, ...)
}

##' @export
plot.multigroupfit <- function(x,...) {
  plot(Model(x),...)
}

###}}}

###{{{ igraph.lvm

##' @export
igraph.lvm <- function(x,layout=igraph::layout.kamada.kawai,...) {
  requireNamespace("igraph",quietly=TRUE)
  oC <- covariance(x)$rel
  for (i in seq_len(nrow(oC)-1))
    for (j in seq(i+1,nrow(oC))) {
      if (oC[i,j]!=0) {
        x <- regression(x,vars(x)[i],vars(x)[j])
        x <- regression(x,vars(x)[j],vars(x)[i])
      }
    }
  g <- igraph::graph.adjacency(x$M,mode="directed")
  igraph::V(g)$color <- "lightblue"
  igraph::V(g)$label <- vars(x)
  igraph::V(g)$shape <- "rectangle"
  for (i in match(latent(x),igraph::V(g)$name)) {
      igraph::V(g)$shape[i] <- "circle"
      igraph::V(g)$color[i] <- "green"
  }
  endo <- index(x)$endogenous
  for (i in match(endo,igraph::V(g)$name)) {
      igraph::V(g)$color[i] <- "orange"
  }
  igraph::E(g)$label <- as.list(rep("",length(igraph::E(g))))
  oE <- edgelabels(x)
  for (i in seq_along(igraph::E(g))) {
    st <- as.character(oE[i])
    if (length(st)>0)
      igraph::E(g)$label[[i]] <- st
  }
  g$layout <- layout(g)
  return(g)
}

###}}} igraph.lvm


beautify <- function(x,col=lava.options()$node.color,border=rep("black",3),labcol=rep("darkblue",3),edgecol=TRUE,...) {
    if (is.null(x$noderender$fill)) notcolored <- vars(x)
    else notcolored <- vars(x)[is.na(x$noderender$fill)]
    x0 <- intersect(notcolored,exogenous(x))
    if (length(x0)>0)
        nodecolor(x, x0, border=border[1], labcol=labcol[1]) <- col[1]
    x0 <- intersect(notcolored,endogenous(x))
    if (length(x0)>0)
        nodecolor(x, x0, border=border[1], labcol=labcol[1]) <- col[2]
    x0 <- intersect(notcolored,latent(x))
    if (length(x0)>0)
        nodecolor(x, x0, border=border[1], labcol=labcol[1]) <- col[3]

    trimmed <- gsub("[[:digit:]]*$","",vars(x))
    keep <- num <- c()
    for (i in seq_len(length(vars(x)))) {
        lb <- labels(x)[vars(x)[i]]
        if (is.null(try(eval(lb),silent=TRUE))) {
            keep <- c(keep,i)
            num <- c(num,gsub(trimmed[i],"",vars(x)[i]))
        }
    }
    if (length(keep)>0) {
        trimmed <- trimmed[keep]
        trim <- gsub(" ",",",trimmed)
        lab <- paste0('"',vars(x)[keep],'"',"=",paste0("expression(",trim,"[scriptscriptstyle(",num,")])"),collapse=",")
        labels(x) <- eval(parse(text=paste("c(",lab,")")))
    }
    if (!edgecol) return(x)
    iex <- index(x)$exo.idx
    ien <- index(x)$endo.idx
    ila <- index(x)$eta.idx
    for (i in iex) {
        for (j in which(x$M[i,]==1)) {
            elab <- edgelabels(x,to=vars(x)[j],from=rev(vars(x)[i]))
            elab2 <- try(eval(elab),silent=TRUE)
            if (is.null(elab2)) elab2 <- ""
            edgelabels(x, to=vars(x)[j], from=rev(vars(x)[i]), cex=2, lwd=3,col=col[1]) <- elab2
        }
    }
    for (i in ien) {
        for (j in which(x$M[i,]==1)) {
            elab <- edgelabels(x,to=vars(x)[j],from=rev(vars(x)[i]))
            elab2 <- try(eval(elab),silent=TRUE)
            if (is.null(elab2)) elab2 <- ""
            edgelabels(x, to=vars(x)[j], from=rev(vars(x)[i]), cex=2, lwd=3,col=col[2]) <- elab2
        }
    }
    for (i in ila) {
        for (j in which(x$M[i,]==1)) {
            elab <- edgelabels(x,to=vars(x)[j],from=rev(vars(x)[i]))
            elab2 <- try(eval(elab),silent=TRUE)
            if (is.null(elab2)) elab2 <- ""
            if (is.null(try(eval(elab),silent=TRUE))) elab <- ""
            edgelabels(x, to=vars(x)[j], from=rev(vars(x)[i]), cex=2, lwd=3,col=col[3]) <- elab2
        }
    }
    x

}

##' @export
`finalize` <-
    function(x,...) UseMethod("finalize")

##' @export
`finalize.lvm` <-
    function(x, diag=FALSE, cor=FALSE, addcolor=TRUE, intercept=FALSE, plain=FALSE, cex, fontsize1=10, cols=lava.options()$node.color, unexpr=FALSE, addstyle=TRUE, ...) {

        g <- as(new("graphAM",adjMat=x$M,"directed"),"graphNEL")
        graph::nodeRenderInfo(g)$fill <- NA
        graph::nodeRenderInfo(g)$label <- NA
        graph::nodeRenderInfo(g)$label[vars(x)] <- vars(x)
        graph::nodeRenderInfo(g)$shape <- x$graphdef$shape

        Lab <- NULL
        for (i in seq_len(length(x$noderender))) {
            nn <- unlist(x$noderender[[i]])
            if (length(nn)>0) {
                R <- list(as.list(x$noderender[[i]])); names(R) <- names(x$noderender)[i]
                if (names(x$noderender)[i]!="label")
                    graph::nodeRenderInfo(g) <- x$noderender[i]
                else Lab <- R[[1]]
            }
        }

        if (!is.null(Lab)) { ## Ugly hack to allow mathematical annotation
            nn <- names(graph::nodeRenderInfo(g)$label)
            LL <- as.list(graph::nodeRenderInfo(g)$label)
            LL[names(Lab)] <- Lab
            if (any(unlist(lapply(LL,function(x) is.expression(x) || is.name(x) || is.call(x))))) {
                graph::nodeRenderInfo(g) <- list(label=as.expression(LL))
            } else graph::nodeRenderInfo(g) <- list(label=LL)
            names(graph::nodeRenderInfo(g)$label) <- nn
            ii <- which(names(graph::nodeRenderInfo(g)$label)=="")
            if (length(ii)>0)
                graph::nodeRenderInfo(g)$label <- graph::nodeRenderInfo(g)$label[-ii]
        }

        graph::edgeDataDefaults(g)$futureinfo <- x$edgerender$futureinfo
        graph::edgeRenderInfo(g)$lty <- x$graphdef$lty
        graph::edgeRenderInfo(g)$lwd <- x$graphdef$lty
        graph::edgeRenderInfo(g)$col <- x$graphdef$col
        graph::edgeRenderInfo(g)$textCol <- x$graphdef$textCol
        graph::edgeRenderInfo(g)$arrowhead <- x$graphdef$arrowhead
        graph::edgeRenderInfo(g)$dir <- x$graphdef$dir
        graph::edgeRenderInfo(g)$arrowtail <- "none"
        graph::edgeRenderInfo(g)$cex <- x$graphdef$cex
        graph::edgeRenderInfo(g)$label <- x$graphdef$label
        for (i in seq_len(length(x$edgerender))) {
            ee <- x$edgerender[[i]]
            if (length(ee)>0 && names(x$edgerender)[i]!="futureinfo") {
                graph::edgeRenderInfo(g)[names(x$edgerender)[i]][names(ee)] <- ee
            }
        }

        opt <- options(warn=-1)
        var <- rownames(covariance(x)$rel)


        if (unexpr) {
            mylab <- as.character(graph::edgeRenderInfo(g)$label); names(mylab) <- names(graph::edgeRenderInfo(g)$label)
            g@renderInfo@edges$label <- as.list(mylab)
        }


        if (intercept) {
            ##  mu <- intfix(x)
            ##  nNA <- sum(is.na(mu))
            ##   if (nNA>0)
            ##      mu[is.na(mu)] <- paste("m",seq_len(nNA))
            ##    mu <- unlist(mu)
            ##    x <- addNode(mu,x)
            ##    for (i in seq_along(mu)) {
            ##    print(mu[i])
            ##      x <- addEdge(var[i], var[i], x)
            ##    }
            ##    x <- addattr(x,attr="shape",var=mu,val="none")
        }

        allEdges <- graph::edgeNames(g)
        regEdges <- c()
        feedback <- c()
        A <- index(x)$A
        if (index(x)$npar.reg>0)
            for (i in seq_len(nrow(A)-1))
                for (j in (i+1):(ncol(A))) {
                    if(A[i,j]==1 & A[j,i]==1) feedback <- c(feedback,
                                                          paste0(var[i],"~",var[j]),
                                                          paste0(var[j],"~",var[i]))
                    if (A[j,i]==0 & x$M[j,i]!=0) {
                        g <- graph::removeEdge(var[j],var[i],g)
                    }
                    if (A[i,j]==1) regEdges <- c(regEdges,paste0(var[i],"~",var[j]))
                    if (A[j,i]==1) regEdges <- c(regEdges,paste0(var[j],"~",var[i]))
                }


        varEdges <- corEdges <- c()
        delta <- ifelse(diag,0,1)
        if (cor | diag) {
            for (r in seq_len(nrow(covariance(x)$rel)-delta) ) {
                for (s in (r+delta):ncol(covariance(x)$rel) ) {
                    if (cor | r==s)
                        if (covariance(x)$rel[r,s]==1 & (!any(c(var[r],var[s])%in%exogenous(x)))) {
                            newedges <- c()
                            if (A[r,s]!=1) {
                                g <- graph::addEdge(var[r],var[s], g)
                                newedges <- paste0(var[r],"~",var[s])
                            } else {
                                if (A[s,r]!=1) {
                                    g <- graph::addEdge(var[s],var[r], g)
                                    newedges <- c(newedges,paste0(var[s],"~",var[r]))
                                }
                            }
                            if (r==s)
                                varEdges <- c(varEdges,
                                              newedges
                                              )
                            if (r!=s)
                                corEdges <- c(corEdges,newedges)
                        }
                }
            }
        }

        if (length(x$edgerender$futureinfo)>0) {
            estr <- names(x$edgerender$futureinfo$label)
            estr <- estr[which(unlist(lapply(estr,nchar))>0)]
            revestr <- sapply(estr, function(y) paste(rev(unlist(strsplit(y,"~"))),collapse="~"))
            revidx <- which(revestr%in%graph::edgeNames(g))
            count <- 0
            for (i in estr) {
                count <- count+1
                for (f in names(x$edgerender$futureinfo)) {
                    if (count%in%revidx) {
                        g@renderInfo@edges[[f]][[revestr[count]]] <- x$edgerender$futureinfo[[f]][[i]]
                    } else {
                        g@renderInfo@edges[[f]][[i]] <- x$edgerender$futureinfo[[f]][[i]]
                    }
                }
            }
        }
        allEdges <- unique(c(regEdges,corEdges,varEdges))
        corEdges <- setdiff(corEdges,regEdges)

        for (e in allEdges) {
            dir <- "forward"; lty <- 1; arrowtail <- "none"
            if (e %in% feedback) {
                dir <- "none"; lty <- 1; arrowtail <- "closed"
            }
            if (e %in% varEdges) {
                dir <- "none"; lty <- 2; arrowtail <- "none"
            }
            if (e %in% corEdges) {
                dir <- "none"; lty <- 2; arrowtail <- "closed"
            }
            arrowhead <- "closed"
            estr <- e
            for (f in c("col","cex","textCol","lwd","lty")) {
                if (!(estr%in%names(graph::edgeRenderInfo(g)[[f]]))
                    || is.na(graph::edgeRenderInfo(g)[[f]][[estr]]))
                    g <- addattr(g,f,var=estr,
                                 val=x$graphdef[[f]],
                                 fun="graph::edgeRenderInfo")
            }

            if (addstyle) {
                g <- addattr(g,"lty",var=estr,val=lty,fun="graph::edgeRenderInfo")
                g <- addattr(g,"direction",var=estr,val=dir,fun="graph::edgeRenderInfo")
                g <- addattr(g,"dir",var=estr,val=dir,fun="graph::edgeRenderInfo")
                g <- addattr(g,"arrowhead",var=estr,val=arrowhead,fun="graph::edgeRenderInfo")
                g <- addattr(g,"arrowtail",var=estr,val=arrowtail,fun="graph::edgeRenderInfo")
                g <- addattr(g,attr="fontsize",var=estr,val=fontsize1,fun="graph::edgeRenderInfo")
            }
            if (is.null(graph::edgeRenderInfo(g)$label))
                graph::edgeRenderInfo(g)$label <- expression()

            if (!missing(cex))
                if (!is.null(cex))
                    graph::nodeRenderInfo(g)$cex <- cex
        }
        if (plain) {
            g <- addattr(g,attr="shape",var=vars(x),val="none")
        } else {
            if (addcolor) {
                if (is.null(x$noderender$fill)) notcolored <- vars(x)
                else notcolored <- vars(x)[is.na(x$noderender$fill)]
                nodecolor(g, intersect(notcolored,exogenous(x))) <- cols[1]
                nodecolor(g, intersect(notcolored,endogenous(x))) <- cols[2]
                nodecolor(g, intersect(notcolored,latent(x))) <- cols[3]
                if (!is.null(trv <- x$attributes$transform)) {
                    nodecolor (g, names(trv)) <- cols[4]
                }
                ##        nodecolor(x, intersect(notcolored,survival(x))) <- cols[4]
                myhooks <- gethook("color.hooks")
                count <- 3
                for (f in myhooks) {
                    count <- count+1
                    res <- do.call(f, list(x=x,subset=notcolored))
                    if (length(cols)>=count) {
                        nodecolor(g,res$vars) <- cols[count]
                    } else {
                        nodecolor(g, res$vars) <- res$col
                    }
                }
            }
        }
        options(opt)
        attributes(g)$feedback <- (length(feedback)>0)
        return(g)
    }

##' @export
`pars` <-
  function(x,...) UseMethod("pars")

##' @export
pars.default <- function(x,...) {
  if (!is.null(x$opt$estimate))
      return(x$opt$estimate)
  if (!is.null(x$opt$par))
      return(x$opt$par)
  if (!is.null(x$coef))
    return(x$coef)
  return(coef(x))
}

##' @export
pars.lvm.missing <- function(x,reorder=FALSE,...) {
    res <- pars.default(x)
    if (reorder) {
        idx <- match(coef(Model(x)),names(coef(x)))
        return(res[idx])
    }
    return(res)
}



###{{{ pars.multigroupfit
## pars.multigroupfit <- function(x,...) {
##   res <- pars.default(x)
##   lapply(ee$model$lvm,coef))
##   coef()
##}
###}}}

###{{{ pars.lvm

##' @export
pars.lvm <- function(x, A, P, v, e, ...) {
  parres <- A[index(x)$M1==1]
  diagcorfree <- diag(P)[diag(index(x)$P1)==1]
  parres <- c(parres, diagcorfree)

  if (ncol(A)>1)
  for (i in seq_len(ncol(index(x)$P1)-1))
    for (j in seq(i+1,nrow(index(x)$P1))) {
      if (index(x)$P1[j,i]!=0) {
        parres <- c(parres, P[j,i])
      }
    }
  if (length(parres)>0)
  names(parres) <- paste0("p",seq_len(length(parres)))
  if (!missing(v)) {
    parres <- c( v[which(index(x)$v1==1)], parres)
  }
  if (!missing(e)) {
    parres <- c( parres, e[which(index(x)$e1==1)] )
  }
  return(parres)
}

###}}} pars.lvm

mgstart <- function(x,p) {
  if (is.list(p)) {
    npar <- with(x, npar+npar.mean+length(unlist(expar)))
    pos <- modelPar(x,seq(npar))$p
    start <- matrix(NA,ncol=npar,nrow=x$ngroup)
    startl <- lapply(pos, function(x) rep(NA,length(x)))
    for (i in seq_len(x$ngroup)) {
      p0 <- p[[i]]
      pos0 <- pos[[i]]
      if (!is.null(names(p0))) {
        ii <- parpos(Model(x)[[i]],names(p0))
        startl[[i]][ii] <- p0[names(ii)]
      } else {
        ii <- seq(min(length(p0),length(startl[[i]])))
        startl[[i]][ii] <- p0[ii]
      }
      start[i,pos0] <- startl[[i]]
    }
    start0 <- apply(start,2,function(x) mean(x,na.rm=TRUE))
  }
  return(start0)
}

##' @export
starter.multigroup <- function(x, starterfun=startvalues2, meanstructure=TRUE,messages=0,...) {
  ## Initial values:
  W <- c() ## Weight-vector
  s <- list()
  for (i in seq_len(x$ngroup)) {
         mydata <- x$data[[i]][,manifest(x$lvm[[i]]),drop=FALSE]
    W <- c(W, nrow(mydata))
    if (nrow(mydata)<3) {
         ii <- index(x$lvm[[i]])
      nn <- ifelse(meanstructure, ii$npar+ii$npar.mean, ii$npar)
      s0 <- rep(1,nn)
    }
    else {
      S <- x$samplestat[[i]]$S
      mu <- if (meanstructure) x$samplestat[[i]]$mu else NULL;
      ##      S <- cov(mydata); mu <- if (meanstructure) colMeans(mydata) else NULL;
      s0 <- do.call(starterfun, list(x$lvm[[i]], S=S, mu=mu,messages=messages))
    }
    s <- c(s, list(s0))
  }

  Wtotal <- sum(W); W <- W/Wtotal

  pg <- vector("list", x$npar); for (i in seq_len(length(pg))) pg[[i]] <- rep(0,x$ngroup)
  meang <- vector("list", x$npar.mean); for (i in seq_len(length(meang))) meang[[i]] <- rep(0,x$ngroup)

  for (i in seq_len(x$ngroup)) {
         pp <- modelPar(x$lvm[[i]],s[[i]])
    pos <- sapply(x$parlist[[i]], function(y) char2num(substr(y,2,nchar(y))))
    for (j in seq_len(length(pos)))
         pg[[ pos[j] ]][i] <-  pp$p[j]

    pos <- sapply(x$meanlist[[i]], function(y) char2num(substr(y,2,nchar(y))))
    ptype <- sapply(x$meanlist[[i]], function(y) substr(y,1,1)=="m")
    if (!(any(ptype)))
         pos <- NULL
    else
         pos <- pos[ptype]
    if (length(pos)>0)
         for (j in seq_len(length(pos))) {
         meang[[ pos[j] ]][i] <-  pp$meanpar[j]
    }
  }

  ## Weighted average
  wp <- unlist(lapply(pg, function(y) {
    ppos <- !is.na(y)
    myweight <- W[ppos]/sum(W[ppos])
    sum(y[ppos]*myweight)
  }))
  wmean <- unlist(lapply(meang, function(y) {
    ppos <- !is.na(y)
    myweight <- W[ppos]/sum(W[ppos])
    sum(y[ppos]*myweight)
  }))
  res <- c(wmean,wp)
  res[!is.finite(res) | is.nan(res) | is.na(res) | is.complex(res)] <- .5
  return(as.numeric(res))
}

startmean <- function(x,p,mu) {
  if (is.null(mu))
         return(p)
  meanpar <- numeric(index(x)$npar.mean)
  mymeans <- vars(x)[index(x)$v1==1]
  midx <- na.omit(match(names(mu),mymeans))
  meanpar[midx] <- mu[midx]
  AP <- matrices(x,p,meanpar)
  nu <- numeric(length(vars(x)))
  nu[vars(x)%in%manifest(x)] <- mu
  meanstart <- ((diag(nrow=nrow(AP$A))-t(AP$A))%*%nu)[index(x)$v1==1]
  names(meanstart) <- vars(x)[index(x)$v1==1]
  return( c(meanstart, p) )
}

## Estimate sub-models (measurement models)
##' @export
`startvalues2` <- function(x, S, mu=NULL, debug=FALSE, messages=0,...) {
  if (messages>0) cat("Obtaining start values...\n")
  S <- reorderdata.lvm(x,S)
  ss <- startvalues(x,S)
  Debug(list("ss=",ss),debug);
  g <- measurement(x,messages=0)
  keep <- c()
  if (length(g)>1) {
    for (i in seq_len(length(g))) {
      if (length(endogenous(g[[i]]))>2)
        keep <- c(keep,i)
    }
    g <- g[keep]
  }
  if (length(g)<2)
    return(startmean(x,ss,mu=mu))
  ## if (messages>0) cat("Fitting marginal measurement models...\n")
  op <- options(warn=-1)
  e <- lapply(g, function(y) {
    estimate(y, data=list(S=S[manifest(y),manifest(y),drop=FALSE], mu=mu[manifest(y)], n=100), control=list(meanstructure=FALSE, starterfun="startvalues", estimator="Simple", method="nlminb1"), optcontrol=list(), debug=FALSE, messages=0)
  })
  for (l in e) {
    ##    a <- coef(l$estimate)[,1]
    a <- coef(l)
    for (i in seq_len(length(a))) {
      pos <- match(names(a)[i],names(ss))
      if (!is.na(pos))
        ss[pos] <- a[i]
    }
  }
  options(op)
  startmean(x,ss,mu=mu)
}

##' @export
startvalues1 <- function(x,S,mu=NULL,tol=1e-6,delta=1e-6,...) {
  p0 <- startvalues(x,S,mu,...)
  p0[index(x)$npar.mean+variances(x)] <- 0.1
  p0[index(x)$npar.mean+offdiags(x)] <- 0
  p0
}

##' @export
startvalues0 <- function(x,S,mu=NULL,tol=1e-6,delta=1e-6,...) {
  p0 <- startvalues(x,S,mu,...)
  A <- t(index(x)$M) ## Adjacency matrix
  P0 <- A0 <- matrix(0,nrow(A),ncol(A))
  A0[,index(x)$eta.idx] <- A[,index(x)$eta.idx]
  diag(P0)[index(x)$endo.idx] <- diag(S)[index(x)$endo.obsidx]/3
  lu <- 0.9
  diag(P0)[index(x)$eta.idx] <- lu##mean(diag(S)[index(x)$endo.idx])/2
  pp <- pars(x,A=t(A0),P=P0,v=rep(1,length(index(x)$vars)))
  pp[pp==1] <- p0[pp==1]
  names(pp) <- coef(x, messages=0, fixed=FALSE, mean=TRUE)[seq_len(length(pp))]
  pp[!is.finite(pp) | is.nan(pp) | is.na(pp)] <- 0.01
  return(pp)
}

##' @export
startvalues <- ## McDonald & Hartmann, 1992
function(x, S, mu=NULL, debug=FALSE, messages=lava.options()$messages, tol=1e-6, delta=1e-6,...) {
  ## As proposed by McDonald & Hartmann, 1992.
  ## Implementation based on John Fox's implementation in the 'sem' R-package
  S <- reorderdata.lvm(x,S)
  if (nrow(S)!=length(manifest(x))) stop("Number of observed variables in data and models does not agree.")
  J <- index(x)$J ## Manifest selection
  P0 <- index(x)$P0 ## covariance 'adjacency'
  A <- t(index(x)$M) ## Adjacency matrix
  n <- nrow(S) ## Number of manifest variables
  m <- nrow(A) ## Number of variables
  A0 <- t(index(x)$M0) ## Adjacency matrix (without fixed parameters)
  obs.idx <- as.vector(J%*%(seq_len(m)));  latent.idx <- setdiff(seq_len(m), obs.idx)
  s <- sqrt(diag(S))
  suppressWarnings(R <- (cov2cor(S))) ## S/outer(s,s)
  C <- P0
  Debug(list("obs.idx", obs.idx), debug)
  C[obs.idx,obs.idx] <- R
  ## Estimates of covariance between latent and manifest variables
  Debug((C), debug)
  for (i in latent.idx) {
    inRelation <- A[obs.idx,i]==1
    for (j in seq_len(length(obs.idx))) {
      Debug((j), debug)
      C[obs.idx[j],i] <- C[i,obs.idx[j]] <- if (any(inRelation)) {
        numerator <- sum(R[j, which(inRelation)])
        denominator <- sqrt(sum(R[which(inRelation), which(inRelation)]))
        numerator/denominator ## as proposed by McDonald & Hartmann
      } else {
        runif(1, .3, .5) ## No arrows => small random covariance
      }
    }
  }
  ## Estimates of covariance between latent variables
  for (i in latent.idx) {
    for (j in latent.idx) {
      C[i,j] <- C[j,i] <-
        if (i==j) {
          1
        } else {
          inRelation.i <- A[obs.idx, i]==1
          inRelation.j <- A[obs.idx, j]==1
          if ((any(inRelation.i)) | (any(inRelation.j))) {
            numerator <- sum(R[which(inRelation.i), which(inRelation.j)])
            denominator <- sqrt( sum(R[which(inRelation.i), which(inRelation.i)])
                                * sum(R[which(inRelation.j), which(inRelation.j)]))
            numerator/(denominator+0.01) ## Avoid division by zero
          } else {
            runif(1, .3, .5)
          }
        }
    }
  }
  if (debug) {
    print("C="); print(C);
  }
  Ahat <- matrix(0,m,m)
  C[is.nan(C)] <- 0
  for (j in seq_len(m)) { ## OLS-estimates
    relation <- A[j,]==1
    if (!any(relation)) next
    Ahat[j, relation] <- tryCatch(Inverse(C[relation,relation] + diag(nrow=sum(relation))*delta,tol=1e-3) %*% C[relation,j], error=function(...) 0)
  }
  Ahat[obs.idx,] <- Ahat[obs.idx,]*matrix(s, n, m)
  Ahat[,obs.idx] <- Ahat[,obs.idx]/matrix(s, m, n, byrow=TRUE)
  Chat <- C
  Chat[obs.idx,] <- Chat[obs.idx,]*matrix(s,n,m)  ##
  Chat[,obs.idx] <- Chat[,obs.idx]*matrix(s,m,n,byrow=TRUE)  ##
  Phat <- (diag(m)-Ahat)%*%Chat%*%t(diag(m)-Ahat)
  ##diag(Phat) <- abs(diag(Phat))
  ## Guarantee PD-matrix:
  Phat[is.nan(Phat) | is.na(Phat)] <- 0
  diag(Phat)[diag(Phat)==0] <- 1
  eig <- eigen(Phat)
  L <- abs(eig$values); L[L<1e-3] <- 1e-3
  Phat <- eig$vectors%*%diag(L,ncol=ncol(eig$vectors))%*%t(eig$vectors)

  Debug(list("start=",start), debug)
  start <- pars(x, A=t(Ahat*A0), P=(Phat*P0))
  names(start) <- coef(x, messages=0, fixed=FALSE, mean=FALSE)[seq_len(length(start))]
  res <- startmean(x,start,mu)
  res[!is.finite(res) | is.nan(res) | is.na(res)] <- 1
  res
}

##' Extract variable names from latent variable model
##'
##' Extract exogenous variables (predictors), endogenous variables (outcomes),
##' latent variables (random effects), manifest (observed) variables from a
##' \code{lvm} object.
##'
##' \code{vars} returns all variables of the \code{lvm}-object including
##' manifest and latent variables. Similarily \code{manifest} and \code{latent}
##' returns the observered resp. latent variables of the model.
##' \code{exogenous} returns all manifest variables without parents, e.g.
##' covariates in the model, however the argument \code{latent=TRUE} can be used
##' to also include latent variables without parents in the result. Pr. default
##' \code{lava} will not include the parameters of the exogenous variables in
##' the optimisation routine during estimation (likelihood of the remaining
##' observered variables conditional on the covariates), however this behaviour
##' can be altered via the assignment function \code{exogenous<-} telling
##' \code{lava} which subset of (valid) variables to condition on.  Finally
##' \code{latent} returns a vector with the names of the latent variables in
##' \code{x}. The assigment function \code{latent<-} can be used to change the
##' latent status of variables in the model.
##'
##' @aliases vars vars.lvm vars.lvmfit latent latent<- latent.lvm latent<-.lvm
##' latent.lvmfit latent.multigroup manifest manifest.lvm manifest.lvmfit
##' manifest.multigroup exogenous exogenous<- exogenous.lvm exogenous<-.lvm
##' exogenous.lvmfit exogenous.multigroup endogenous endogenous.lvm
##' endogenous.lvmfit endogenous.multigroup
##' @usage
##'
##' vars(x,...)
##'
##' endogenous(x,...)
##'
##' exogenous(x,...)
##'
##' manifest(x,...)
##'
##' latent(x,...)
##'
##' \method{exogenous}{lvm}(x, xfree = TRUE,...) <- value
##'
##' \method{exogenous}{lvm}(x,variable,latent=FALSE,index=TRUE,...)
##'
##' \method{latent}{lvm}(x,clear=FALSE,...) <- value
##'
##' @param x \code{lvm}-object
##' @param variable list of variables to alter
##' @param latent Logical defining whether latent variables without parents
##' should be included in the result
##' @param index For internal use only
##' @param clear Logical indicating whether to add or remove latent variable
##' status
##' @param xfree For internal use only
##' @param value Formula or character vector of variable names.
##' @param \dots Additional arguments to be passed to the low level functions
##' @return Vector of variable names.
##' @author Klaus K. Holst
##' @seealso \code{\link{endogenous}}, \code{\link{manifest}},
##' \code{\link{latent}}, \code{\link{exogenous}}, \code{\link{vars}}
##' @keywords models regression
##' @examples
##'
##' g <- lvm(eta1 ~ x1+x2)
##' regression(g) <- c(y1,y2,y3) ~ eta1
##' latent(g) <- ~eta1
##' endogenous(g)
##' exogenous(g)
##' identical(latent(g), setdiff(vars(g),manifest(g)))
##'
##' @export
`vars` <-
function(x,...) UseMethod("vars")

##' @export
`vars.graph` <-
  function(x,...) {
    graph::nodes(x)
  }

##' @export
`vars.lvm` <-
  function(x,...) {
    colnames(x$M)
  }

##' @export
`vars.lvmfit` <-
  function(x,...) {
    vars(Model(x),...)
  }

##' @export
vars.list <- function(x,...) {
  varlist <- c()
  for (i in seq_along(x)) {
    varlist <- c(varlist, vars(x[[i]]))
  }
  varlist <- unique(varlist)
  return(varlist)
}

##' @export
`vars.lm` <-
  function(x,...) {
    c(endogenous(x),exogenous(x))
  }

###{{{ logLik.lvm

##' @export
logLik.lvm <- function(object,p,data,model="gaussian",indiv=FALSE,S,mu,n,debug=FALSE,weights=NULL,data2=NULL,...) {
    cl <- match.call()
    xfix <- colnames(data)[(colnames(data)%in%parlabels(object,exo=TRUE))]

    constr <- lapply(constrain(object), function(z)(attributes(z)$args))
    xconstrain <- intersect(unlist(constr), manifest(object))
    xconstrainM <- TRUE
    if (length(xconstrain)>0) {
        constrainM <- names(constr)%in%unlist(object$mean)
        for (i in seq_len(length(constr))) {
            if (!constrainM[i]) {
                if (any(constr[[i]]%in%xconstrain)) xconstrainM <- FALSE
            }
        }
    }

    Debug(xfix,debug)
    if (missing(n)) {
        n <- nrow(data)
        if (is.null(n)) n <- data$n
    }
    lname <- paste0(model,"_logLik.lvm")
    logLikFun <- get(lname)


    if (length(xfix)>0 | (length(xconstrain)>0 & !xconstrainM & !lava.options()$test & model!="gaussian")) { ##### Random slopes!
        x0 <- object
        if (length(xfix)>0) {
            Debug("random slopes...",debug)
            nrow <- length(vars(object))
            xpos <- lapply(xfix,function(y) which(regfix(object)$labels==y))
            colpos <- lapply(xpos, function(y) ceiling(y/nrow))
            rowpos <- lapply(xpos, function(y) (y-1)%%nrow+1)
            myfix <- list(var=xfix, col=colpos, row=rowpos)
            for (i in seq_along(myfix$var))
                for (j in seq_along(myfix$col[[i]])) {
                    regfix(x0, from=vars(x0)[myfix$row[[i]][j]],to=vars(x0)[myfix$col[[i]][j]]) <-
                        data[1,myfix$var[[i]]]
                }
            index(x0) <- reindex(x0,zeroones=TRUE,deriv=TRUE)
        }
        myfun <- function(ii) {
            if (length(xfix)>0)
                for (i in seq_along(myfix$var)) {
                    index(x0)$A[cbind(myfix$row[[i]],myfix$col[[i]])] <- data[ii,myfix$var[[i]]]
                }
            return(logLikFun(x0,data=data[ii,,drop=FALSE], p=p,weights=weights[ii,,drop=FALSE],data2=data2[ii,,drop=FALSE],
                             model=model,debug=debug,indiv=indiv,...))
        }
        loglik <- sapply(seq_len(nrow(data)),myfun)
        if (!indiv) {
            loglik <- sum(loglik)
            n <- nrow(data)
            attr(loglik, "nall") <- n
            attr(loglik, "nobs") <- n
            attr(loglik, "df") <- length(p)
            class(loglik) <- "logLik"
        }
        return(loglik)
    }

    if (xconstrainM) {
        xconstrain <- c()
        for (i in seq_len(length(constrain(object)))) {
            z <- constrain(object)[[i]]
            xx <- intersect(attributes(z)$args,manifest(object))
            if (length(xx)>0) {
                warg <- setdiff(attributes(z)$args,xx)
                wargidx <- which(attributes(z)$args%in%warg)
                exoidx <- which(attributes(z)$args%in%xx)
                parname <- names(constrain(object))[i]
                y <- names(which(unlist(lapply(intercept(object),function(x) x==parname))))
                el <- list(i,y,parname,xx,exoidx,warg,wargidx,z)
                names(el) <- c("idx","endo","parname","exo","exoidx","warg","wargidx","func")
                xconstrain <- c(xconstrain,list(el))
            }
        }
        if (length(xconstrain)>0) {
            yconstrain <- unlist(lapply(xconstrain,function(x) x$endo))
            iconstrain <- unlist(lapply(xconstrain,function(x) x$idx))

            Mu <- matrix(0,nrow(data),length(vars(object))); colnames(Mu) <- vars(object)
            M <- modelVar(object,p=p,data=data)
            M$parval <- c(M$parval,  object$mean[unlist(lapply(object$mean,is.numeric))])
            for (i in seq_len(length(xconstrain))) {
                pp <- unlist(M$parval[xconstrain[[i]]$warg]);
                myidx <- with(xconstrain[[i]],order(c(wargidx,exoidx)))
                mu <- with(xconstrain[[i]],
                          apply(data[,exo,drop=FALSE],1,
                                function(x) {
                                    func(unlist(c(pp,x))[myidx])
                                }))
                Mu[,xconstrain[[i]]$endo] <- mu
            }
            offsets <- Mu%*%t(M$IAi)[,endogenous(object),drop=FALSE]
            object$constrain[iconstrain] <- NULL
            object$mean[yconstrain] <- 0
            loglik <- do.call(lname, c(list(object=object,p=p,data=data,indiv=indiv,weights=weights,data2=data2,offset=offsets),list(...)))
        } else {
            cl[[1]] <- logLikFun
            loglik <- eval.parent(cl)
        }
    } else {
        loglik <- 0
        if (length(xconstrain)>0 && NROW(data)>1) {
            for (ii in seq(nrow(data))) {
                cl$data <- data[ii,]
                cl$weights <- weights[ii,]
                cl$data2 <- data2[ii,]
                loglik <- loglik+eval.parent(cl)
            }
        } else {
            cl[[1]] <- logLikFun
            loglik <- eval.parent(cl)
        }
    }

    if (is.null(attr(loglik,"nall")))
        attr(loglik, "nall") <- n
    if (is.null(attr(loglik,"nobs")))
        attr(loglik, "nobs") <- n##-length(p)
    if (is.null(attr(loglik,"df")))
        attr(loglik, "df") <- length(p)
    class(loglik) <- "logLik"
    return(loglik)
}

###}}}

###{{{ gaussian_loglik

##' @export
gaussian_logLik.lvm <- function(object,p,data,
                         type=c("cond","sim","exo","sat","cond2"),
                         weights=NULL, indiv=FALSE, S, mu, n, offset=NULL, debug=FALSE, meanstructure=TRUE,...) {
    exo.idx <- with(index(object), exo.obsidx)
    endo.idx <- with(index(object), endo.obsidx)
    if (type[1]=="exo") {
        if (length(exo.idx)==0 || any(is.na(exo.idx)))
            return(0)
    }

    cl <- match.call()
    if (type[1]=="cond") {
        cl$type <- "sim"
        L0 <- eval.parent(cl)
        cl$type <- "exo"
        L1 <- eval.parent(cl)
        loglik <- L0-L1
        return(loglik)
    }

    if (missing(n)) {
        if (is.vector(data)) n <- 1
        else n <- nrow(data)
    }
    k <- length(index(object)$manifest)

    if (!is.null(offset) && type[1]!="exo") {
        data[,colnames(offset)] <- data[,colnames(offset)]-offset
    }

    if (type[1]=="sat") {
        if (missing(S)) {
            d0 <- procdata.lvm(object,data=data)
            S <- d0$S; mu <- d0$mu; n <- d0$n

        }
        if (missing(p)) p <- rep(1,length(coef(object)))
        L1 <- logLik(object,p,data,type="exo",meanstructure=meanstructure)
        ##    Sigma <- (n-1)/n*S ## ML = 1/n * sum((xi-Ex)^2)
        Sigma <- S
        loglik <- -(n*k)/2*log(2*base::pi) -n/2*(log(det(Sigma)) + k) - L1
        P <- length(endo.idx)
        k <- length(exo.idx)
        npar <- P*(1+(P-1)/2)
        if (meanstructure) npar <- npar+ (P*k + P)
        attr(loglik, "nall") <- n
        attr(loglik, "nobs") <- n
        attr(loglik, "df") <- npar
        class(loglik) <- "logLik"
        return(loglik)
    }
    myidx <- switch(type[1],
                    sim =  seq_along(index(object)$manifest),
                    cond = { endo.idx },
                    cond2 = { endo.idx },
                    exo =  { exo.idx } )

    mom <- moments(object, p, conditional=(type[1]=="cond2"), data=data)

    if (!lava.options()$allow.negative.variance && any(diag(mom$P)<0)) return(NaN)

    C <- mom$C
    xi <- mom$xi
    if (type[1]=="exo") {
        C <- C[exo.idx,exo.idx,drop=FALSE]
        xi <- xi[exo.idx,drop=FALSE]
    }
    Debug(list("C=",C),debug)
    k <- nrow(C)

    iC <- Inverse(C,det=TRUE, symmetric = TRUE)
    detC <- attributes(iC)$det

    if (!is.null(weights)) {
        weights <- cbind(weights)
        K <- length(exo.idx)+length(endo.idx)
        if (ncol(weights)!=1 & ncol(weights)!=K) {
            w.temp <- weights
            weights <- matrix(1,nrow=nrow(weights),ncol=K)
            weights[,endo.idx] <- w.temp
        }
        if (type=="exo")
            weights <- NULL
    }

    notdatalist <- (!is.list(data) | is.data.frame(data))
    if (missing(n))
        if (!missing(data)) n <- NROW(data)
    if (!missing(n))
        if (notdatalist & (n<2 | indiv | !is.null(weights))) {
            if (n==1)
                data <- rbind(data)
            res <- numeric(n)
            data <- data[,index(object)$manifest,drop=FALSE]
            loglik <- 0;
            for (i in seq_len(n)) {
                ti <- as.numeric(data[i,myidx])
                if (meanstructure) {
                    ti <- cbind(ti-as.numeric(xi))
                }
                if (!is.null(weights)) {
                    W <- diag(weights[i,],nrow=length(weights[i,]))
                    val <- -k/2*log(2*base::pi) -1/2*log(detC) - 1/2*(t(ti)%*%W)%*%iC%*%(ti)
                } else {
                    val <- -k/2*log(2*base::pi) -1/2*log(detC) - 1/2*t(ti)%*%iC%*%(ti)
                }
                if (indiv)
                    res[i] <- val
                loglik <- loglik + val
            }
            if (indiv)
                return(res)
        } else {
            if (missing(S)) {
                d0 <- procdata.lvm(object,data=data)
                S <- d0$S; mu <- d0$mu; n <- d0$n
            }
            S <- S[myidx,myidx,drop=FALSE]
            mu <- mu[myidx,drop=FALSE]
            T <- S
            if (meanstructure) {
                W <- crossprod(rbind(mu-xi))
                T <- S+W
            }
            loglik <- -(n*k)/2*log(2*base::pi) -n/2*(log(detC) + tr(T%*%iC))
        }
    return(loglik)
}

###}}}

###{{{ logLik.lvmfit

##' @export
logLik.lvmfit <- function(object,
                   p=coef(object),
                   data=model.frame(object),
                   model=object$estimator,
                   weights=Weights(object),
                   data2=object$data$data2,
                   ...) {

    logLikFun <- paste0(model,"_logLik.lvm")
    if (!exists(logLikFun)) {
        model <- "gaussian"
    }
    l <- logLik.lvm(object$model0,p,data,model=model,weights=weights,
                    data2=data2,
                    ...)
    return(l)
}

###}}} logLik.lvmfit

###{{{ logLik.lvm.missing

##' @export
logLik.lvm.missing <- function(object,
                        p=pars(object), model=object$estimator,
                        weights=Weights(object$estimate),
                        ...) {
    logLik(object$estimate$model0, p=p, model=model, weights=weights, ...)
}

###}}}

###{{{ logLik.multigroup

##' @export
logLik.multigroup <- function(object,p,data=object$data,weights=NULL,type=c("cond","sim","exo","sat"),...) {
    res <- procrandomslope(object)
    pp <- with(res, modelPar(model,p)$p)

    if (type[1]=="sat") {
        n <- 0
        df <- 0
        loglik <- 0
        for (i in seq_len(object$ngroup)) {
            m <- Model(object)[[i]]
            L <- logLik(m,p=pp[[i]],data=object$data[[i]],type="sat")
            df <- df + attributes(L)$df
            loglik <- loglik + L
            n <- n + object$samplestat[[i]]$n
        }
        attr(loglik, "nall") <- n
        attr(loglik, "nobs") <- n##-df
        attr(loglik, "df") <- df
        class(loglik) <- "logLik"
        return(loglik)
    }

    n <- 0
    loglik <- 0; for (i in seq_len(object$ngroup)) {
                     n <- n + object$samplestat[[i]]$n
                     val <- logLik(object$lvm[[i]],pp[[i]],data[[i]],weights=weights[[i]],type=type,...)
                     loglik <- loglik + val
                 }
    attr(loglik, "nall") <- n
    attr(loglik, "nobs") <- n##-length(p)
    attr(loglik, "df") <- length(p)
    class(loglik) <- "logLik"
    return(loglik)
}

###}}} logLik.multigroup

###{{{ logLik.multigroupfit

##' @export
logLik.multigroupfit <- function(object,
                          p=pars(object), weights=Weights(object), model=object$estimator, ...) {
    logLik(object$model0,p=p,weights=weights,model=model,...)
}
###}}} logLik.multigroup

##' @export
`graph2lvm` <-
    function(g, debug=FALSE, messages=0) {
        res <- lvm(graph::nodes(g), debug=debug, messages=messages)
        M <- t(as(g, Class="matrix"))
        for (i in seq_len(nrow(M))) {
            if (any(M[,i]==1)) {
                res <- regression(res, rownames(M)[M[,i]==1], rownames(M)[i], messages=messages)
            }
        }
        res
    }

##' Calculate Wald og Likelihood based (profile likelihood) confidence intervals
##'
##' Calculates either Wald confidence limits: \deqn{\hat{\theta} \pm
##' z_{\alpha/2}*\hat\sigma_{\hat\theta}} or profile likelihood confidence
##' limits, defined as the set of value \eqn{\tau}:
##' \deqn{logLik(\hat\theta_{\tau},\tau)-logLik(\hat\theta)< q_{\alpha}/2}
##'
##' where \eqn{q_{\alpha}} is the \eqn{\alpha} fractile of the \eqn{\chi^2_1}
##' distribution, and \eqn{\hat\theta_{\tau}} are obtained by maximizing the
##' log-likelihood with tau being fixed.
##'
##' @title Calculate confidence limits for parameters
##' @param object \code{lvm}-object.
##' @param parm Index of which parameters to calculate confidence limits for.
##' @param level Confidence level
##' @param profile Logical expression defining whether to calculate confidence
##' limits via the profile log likelihood
##' @param curve if FALSE and profile is TRUE, confidence limits are
##' returned. Otherwise, the profile curve is returned.
##' @param n Number of points to evaluate profile log-likelihood in
##' over the interval defined by \code{interval}
##' @param interval Interval over which the profiling is done
##' @param lower If FALSE the lower limit will not be estimated (profile intervals only)
##' @param upper If FALSE the upper limit will not be estimated (profile intervals only)
##' @param \dots Additional arguments to be passed to the low level functions
##' @return A 2xp matrix with columns of lower and upper confidence limits
##' @author Klaus K. Holst
##' @seealso \code{\link{bootstrap}{lvm}}
##' @keywords models regression
##' @examples
##'
##' m <- lvm(y~x)
##' d <- sim(m,100)
##' e <- estimate(lvm(y~x), d)
##' confint(e,3,profile=TRUE)
##' confint(e,3)
##' \donttest{ ## Reduce Ex.timings
##' B <- bootstrap(e,R=50)
##' B
##' }
##' @aliases confint.multigroupfit
##' @export
##' @method confint lvmfit
confint.lvmfit <- function(object,parm=seq_len(length(coef(object))),level=0.95,profile=FALSE,curve=FALSE,n=20,interval=NULL,lower=TRUE,upper=TRUE,...) {
    if (is.character(parm)) {
        parm <- parpos(Model(object),p=parm)
        parm <- parm[attributes(parm)$ord]
    }
    if (!profile) {
        return(confint.default(object,parm=parm,level=level,...))
    }
    res <- c()
    for (i in parm) {
        res <- rbind(res, profci.lvmfit(object,parm=i,level=level,profile=profile,n=n,curve=curve,interval=interval,lower=lower,upper=upper,...))
        if (curve) return(res)
    }
    rownames(res) <- names(coef(object))[parm]
    colnames(res) <- paste((c(0,1) + c(1,-1)*(1-level)/2)*100,"%")
    return(res)
}


##' @export
confint.multigroupfit <- function(object,parm=seq_along(pars(object)),level=0.95,
                           ...) {
  p <- 1-(1-level)/2
  res <- cbind(pars(object),pars(object)) + qnorm(p)*cbind(-1,1)%x%diag(vcov(object))^0.5
  colnames(res) <- paste0(c(1-p,p)*100,"%")
  rownames(res) <- names(coef(object))
  res[parm,,drop=FALSE]
}

##' @export
`modelPar` <-
  function(x,p,...) UseMethod("modelPar")

##' @export
modelPar.lvmfit <- function(x, p=pars(x), ...) modelPar(Model(x),p=p,...)

##' @export
modelPar.lvm <- function(x,p, ...) {
  npar <- index(x)$npar
  npar.mean <- index(x)$npar.mean
  if (length(p)!=npar & length(p)<(npar+npar.mean)) stop("Wrong dimension of parameter vector!")
  p2 <- NULL
  if (length(p)!=npar) { ## if meanstructure
    meanpar <- p[seq_len(npar.mean)]
    p. <- p
    if (length(meanpar)>0) {
        p. <- p[-seq_len(npar.mean)]
    } else meanpar <- NULL
    p <- p.[seq_len(npar)]
    if (npar>0) {
        p2 <- p.[-seq_len(npar)]
    } else p2 <- p.
  } else {
    meanpar <- NULL
    p2 <- NULL
  }
  return(list(p=p,meanpar=meanpar,p2=p2))
}

##' @export
modelPar.multigroupfit <- function(x,p=pars(x),...) {
  modelPar(Model(x),p,...)
}

##' @export
modelPar.multigroup <- function(x,p, ...) {
  if (length(p)==x$npar) {
    pp <- lapply(x$parposN,function(z) p[z])
    res <- list(p=pp, par=pp, mean=NULL)
    return(res)
  }
  ppos <- x$parposN
  pp <- lapply(ppos,function(z) p[z+x$npar.mean])

  if (length(pp)==0) pp <- lapply(seq_len(x$ngroup),function(x) logical())
  mm <- lapply(x$meanposN,function(x) p[x])
  if (is.null(mm)) mm <- lapply(seq_len(x$ngroup),logical())
  pm <- mm
  for (i in seq_len(length(pm))) pm[[i]] <- c(pm[[i]],pp[[i]])
  res <- list(p=pm,par=pp,mean=mm)
  return(res)
}

modelPar2.multigroup <-
  function(x,p, ...) {
  npar <- x$npar
  npar.mean <- x$npar.mean
  k <- x$ngroup
  if (length(p)!=npar & length(p)!=(npar+npar.mean)) stop("Wrong dimension of parameter vector!")
  if (length(p)!=npar) { ## if meanstructure
      meanpar <- p[seq_len(npar.mean)]
      p. <- p[-seq_len(npar.mean)]
    } else {
      meanpar <- NULL
      p. <- p
    }


  parlist <- list(); for (i in seq_len(k)) parlist[[i]] <- numeric(length(x$parlist[[i]]))
  if (!is.null(meanpar)) {
    meanlist <- list(); for (i in seq_len(k)) meanlist[[i]] <- numeric(length(x$meanlist[[i]]))
  }

  if (length(p.)>0)
  for (i in seq_along(p.)) {
    for (j in seq_len(k)) {
      idx <- match(paste0("p",i), x$parlist[[j]])
      if (!is.na(idx))
        parlist[[j]][idx] <- p.[i]
      if (!is.null(meanpar)) {
        midx <- match(paste0("p",i), x$meanlist[[j]])
        if (!is.na(midx))
          meanlist[[j]][midx] <- p.[i]
      }
    }
  }

  if (!is.null(meanpar)) {
    for (i in seq_along(meanpar)) {
      for (j in seq_len(k)) {
        idx <- match(paste0("m",i), x$meanlist[[j]])
        if (!is.na(idx))
          meanlist[[j]][idx] <- meanpar[i]
      }
    }
  } else {
    meanlist <- NULL
  }
  p0 <- parlist
  for (i in seq_along(p0))
    p0[[i]] <- c(meanlist[[i]],parlist[[i]])
  return(list(p=p0, par=parlist, mean=meanlist))
}

##' @export
`modelVar` <-
  function(x,p,...) UseMethod("modelVar")

##' @export
modelVar.lvmfit <- function(x, p=pars(x), ...) modelVar(Model(x),p=p,...)

##' @export
modelVar.lvm <- function(x,p,data,...) {
  pp <- modelPar(x,p)
  res <- moments(x, p=p, data=data,...)
  attr(res, "pars") <- pp$p
  attr(res, "meanpar") <- pp$meanpar
  attr(res, "epar") <- pp$epar
  res
}

glm_estimate_hook <- function(x,estimator,...) {
    yy <- c()
    if (length(estimator)>0 && estimator=="glm") {
        for (y in endogenous(x)) {
            fam <- attributes(distribution(x)[[y]])$family
            if (is.null(fam)) fam <- stats::gaussian()
            if (!(tolower(fam$family)%in%
                  c("gaussian","gamma","inverse.gaussian","weibull"))) {
                yy <- c(yy,y)
            }
        }
        if (length(yy)>0) covariance(x,yy) <- 1
    }
    return(c(list(x=x,estimator=estimator,...)))
}

GLMest <- function(m,data,control=list(),...) {
    yvar <- endogenous(m)
    res <- c()
    count <- 0
    V <- NULL
    mymsg <- c()
    ics <- c()
    breads <- c()

    et <- eventTime(m)
    yvar.et <- rep(NA,length(yvar))
    names(yvar.et) <- yvar
    if (!is.null(et)) {
        for (i in seq_along(et)) {
            ## if (!survival::is.Surv(data[,et[[i]]$names[1]]))
            ##         data[,et[[i]]$names[1]] <- with(et[[i]],
            ##         survival::Surv(data[,names[1]],data[,names[2]]))
            yvar <- setdiff(yvar,c(et[[i]]$latentTimes[-1],et[[i]]$names))
            yvar.et[et[[i]]$latentTimes[1]] <- et[[i]]$names[1]
        }
    }

    for (y in yvar) {
        count <- count+1
        xx <- parents(m, y)
        fam <- attributes(distribution(m)[[y]])$family
        if (is.null(fam)) fam <- stats::gaussian()
        if (!is.null(fam$link)) {
            mymsg <- c(mymsg, with(fam, paste0(family,"(",link,")")))
        } else {
            mymsg <- c(mymsg, with(fam, paste0(family)))
        }
        if (length(xx)==0) xx <- 1
        nn0 <- paste(y,xx,sep=lava.options()$symbol[1])
        y0 <- y

        f <- as.formula(paste0(y, "~",
          paste(xx, collapse = "+")
        ))
        isSurv <- inherits(data[1, y], "Surv")
        if (isSurv) {
          g <- survival::survreg(f,data=data,dist=fam$family,...)
        } else {
          g <- glm(f,family=fam,data=data,...)
        }
        p <- pars(g)
        ii <- IC(g)
        V0 <- attr(ii, "bread")
        ics <- cbind(ics,ii)
        y <- y0
        names(p)[1] <- y
        if (length(p)>1) {
          xx0 <- setdiff(xx, 1)
          if (length(xx0) > 0) {
            nn0 <- paste(y, xx0, sep = lava.options()$symbol[1])
            names(p)[seq_along(nn0) + 1] <- nn0
          }
          if (length(p)>length(xx0)+1) names(p)[length(p)] <- paste(y,y,sep=lava.options()$symbol[2])
        }
        if (tolower(fam$family)%in%c("gaussian","gamma","inverse.gaussian") && !isSurv) {
            ics <- cbind(ics,0)
            null <- matrix(0); dimnames(null) <- list("scale","scale")
            V0 <- blockdiag(V0,null,pad=0)
        }
        breads <- c(breads, list(V0))
        res <- c(res, list(p));
    }
    coefs <- unlist(res)
    idx <- na.omit(match(coef(m),names(coefs)))
    coefs <- coefs[idx]
    V <- var_ic(ics[, idx])
    mymsg <- noquote(cbind(mymsg))
    colnames(mymsg) <- "Family(Link)"; rownames(mymsg) <- paste(yvar,":")
    list(estimate=coefs, vcov=V, breads=breads, IC=ics[,idx],
         summary.message=function(...)  { mymsg }, dispname="Dispersion:")
}

GLMscore <- function(x,p,data,indiv=TRUE,logLik=FALSE,...) {
    yvar <- endogenous(x)
    S <- pnames <- c()
    count <- 0
    breads <- c()
    L <- 0
    for (y in yvar) {
        count <- count+1
        xx <- parents(x,y)
        pname <- c(y,paste0(y,sep=lava.options()$symbol[1],xx),paste(y,y,sep=lava.options()$symbol[2]))
        pidx <- na.omit(match(pname,coef(x)))
        fam <- attributes(distribution(x)[[y]])$family
        if (is.null(fam)) fam <- stats::gaussian()
        if (length(xx)==0) xx <- 1
        f <- as.formula(paste0(y,"~",paste(xx,collapse="+")))
        isSurv <- inherits(data[1,y],"Surv")
        if (inherits(data[,y],"Surv")) {
            g <- survival::survreg(f,data=data,dist=fam$family)
        } else {
            g <- glm(f,family=fam,data=data)
        }
        p0 <- p[pidx]
        if (!isSurv) L0 <- logL.glm(g,p=p0,indiv=TRUE,...)
        if (tolower(fam$family)%in%c("gaussian","gamma","inverse.gaussian") && !isSurv) {
            p0 <- p0[-length(p0)]
            S0 <- score(g,p=p0,indiv=TRUE,pearson=TRUE,...)
            V0 <- attr(S0,"bread")
            r <- attr(S0,"pearson")
            ## dispersion <- mean(r^2)
            S0 <- cbind(S0,scale=0)
            null <- matrix(0); dimnames(null) <- list("scale","scale")
            V0 <- blockdiag(V0,null,pad=0)
        } else {
            S0 <- score(g,p=p0,indiv=TRUE,...)
            if (isSurv) L0 <- attr(S0,"logLik")
            V0 <- attr(S0,"bread")
        }
        L <- L+sum(L0)
        breads <- c(breads,list(V0))
        S <- c(S,list(S0))
        pnames <- c(pnames, list(pname));
    }
    coefs <- unlist(pnames)
    idx <- na.omit(match(coefs,coef(x)))
    idx <- order(idx)
    V <- Reduce(blockdiag,breads)[idx,idx]
    S1 <- Reduce(cbind,S)[,idx,drop=FALSE]
    colnames(S1) <- coef(x)
    attributes(S1)$bread <- V
    attributes(S1)$logLik <- structure(L,nobs=nrow(data),nall=nrow(data),df=length(p),class="logLik")
    if (!indiv) S1 <- colSums(S1)
    return(S1)
}

##' @export
score.lm <- function(x, p=coef(x), data, indiv=FALSE,
              y, X, offset=NULL, weights=NULL, dispersion=TRUE, ...) {
    if (missing(data)) {
        X <- model.matrix(x)
        y <- model.frame(x)[,1]
    } else {
        X <- model.matrix(formula(x),data=data)
        y <- model.frame(formula(x),data=data)[,1]
    }
    if(any(is.na(p))) warning("Over-parameterized model")
    Xbeta <- X%*%p
    if (is.null(offset)) offset <- x$offset
    if (!is.null(offset)) Xbeta <- Xbeta+offset
    r <- y-Xbeta
    if (is.null(weights)) weights <- x$weights
    if (!is.null(weights)) {
        sigma2 <- sum(r^2*weights)/(length(r)-length(p))
        r <- r*weights
    } else {
        sigma2 <- sum(r^2)/(length(r)-length(p))
    }
    if (!dispersion) sigma2 <- 1
    ##sigma2 <- suppressWarnings(summary(x)$sigma^2)
    A <- as.vector(r)/sigma2
    S <- apply(X,2,function(x) x*A)
    if (!indiv) return(colSums(S))
    suppressWarnings(attributes(S)$bread <- vcov(x)*NROW(S))
    return(S)
}

##' @export
score.glm <- function(x,p=coef(x),data,indiv=FALSE,pearson=FALSE,
               y,X,link,dispersion,offset=NULL,weights=NULL,...) {

    if (inherits(x,"glm")) {
        link <- family(x)
        if (missing(data)) {
            X <- model.matrix(x)
            y <- model.frame(x)[,1]
        } else {
            X <- model.matrix(formula(x),data=data)
            y <- model.frame(formula(x),data=data)[,1]
        }
        offset <- x$offset
    } else {
        if (missing(link)) stop("Family needed")
        if (missing(data)) stop("data needed")
        X <- model.matrix(formula(x),data=data)
        y <- model.frame(formula(x),data=data)[,1]
    }
    if (is.character(y) || is.factor(y)) {
        y <- as.numeric(as.factor(y)) - 1
    }
    ## g <- link$linkfun
    ginv <- link$linkinv
    dginv <- link$mu.eta ## D[linkinv]
    ##dg <- function(x) 1/dginv(g(x)) ## Dh^-1 = 1/(h'(h^-1(x)))
    if (inherits(x, "negbin")) {
        Dcanlink <- function(x) 1/x
    } else {
        canonf <- do.call(link$family,list())
        ## caninvlink <- canonf$linkinv
        canlink <- canonf$linkfun
        Dcaninvlink <- canonf$mu.eta
        Dcanlink <- function(x) 1 / Dcaninvlink(canlink(x))
    }
    ##gmu <- function(x) g(caninvlink(x))
    ##invgmu <- function(z) canlink(ginv(z))
    h <- function(z) Dcanlink(ginv(z))*dginv(z)
    if(any(is.na(p))) stop("Over-parameterized model")
    Xbeta <- X%*%p
    if (!is.null(offset)) Xbeta <- Xbeta+offset
    if (missing(data) && !is.null(x$offset) && is.null(offset) ) Xbeta <- Xbeta+x$offset
    pi <- ginv(Xbeta)
    r <- y-pi
    if (!is.null(x$prior.weights) || !is.null(weights)) {
        if (is.null(weights)) weights <- x$prior.weights
    } else {
        weights <- !is.na(r)
    }
    a.phi <- 1
    r <- r*weights
    rpearson <- as.vector(r)/link$variance(pi)^.5
    if (length(p) > length(coef(x))) {
        a.phi <- p[length(coef(x)) + 1]
    } else { ## if (tolower(family(x)$family)%in%c("gaussian","gamma","inverse.gaussian"))
        suppressWarnings(disp <- summary(x)$dispersion)
        if (!is.null(disp)) a.phi <- disp
        if (inherits(x, "negbin")) {
            a.phi <- link$variance(pi) / pi
        }
        ## a.phi <- sum(rpearson^2)*x$df.residual/x$df.residual^2
    }
    A <- as.vector(h(Xbeta) * r) / a.phi
    S <- apply(X, 2, function(x) x * A)
    if (!indiv) return(colSums(S))
    if (pearson) attr(S,"pearson") <- rpearson
    suppressWarnings(attributes(S)$bread <- vcov(x)*NROW(S))
    if (x$family$family=="quasi" && x$family$link=="identity" && x$family$varfun=="constant")
        attributes(S)$bread <- -Inverse(information.glm(x)*NROW(S))
    return(S)
}

##' @export
pars.glm <- function(x,...) {
    if (tolower(family(x)$family)%in%c("gaussian","gamma","inverse.gaussian")) {
        res <- c(coef(x),suppressWarnings(summary(x)$dispersion))
        names(res)[length(res)] <- "Dispersion"
        return(res)
    }
    return(coef(x))
}

logL.glm <- function(x,p=pars.glm(x),data,indiv=FALSE,...) {
    if (!missing(data)) {
        x <- update(x,data=data,...)
    }
    f <- family(x)
    ginv <- f$linkinv
    X <- model.matrix(x)
    n <- nrow(X)
    ##disp <- 1;
    p0 <- p
    if (tolower(family(x)$family)%in%c("gaussian","gamma","inverse.gaussian")) {
        if (length(p)==ncol(X)) {
            ##disp <- suppressWarnings((summary(x)$dispersion))
        } else {
            ##disp <- tail(p,1)
            p0 <- p[-length(p)]
        }
    }
    if(any(is.na(p))) {
        warning("Over-parametrized model")
    }
    Xbeta <- X%*%p0
    if (!is.null(x$offset)) Xbeta <- Xbeta+x$offset
    y <- model.frame(x)[,1]
    mu <- ginv(Xbeta)
    w <- x$prior.weights
    dev <-  f$dev.resids(y,mu,w)
    if (indiv) {

    }
    loglik <- length(p)-(f$aic(y,n,mu,w,sum(dev))/2+x$rank)
    structure(loglik,nobs=n,df=length(p),class="logLik")
}

##' @export
IC.glm <- function(x,...) {
    IC.default(x,...)
}

hessian.glm <- function(x,p=coef(x),...) {
    numDeriv::jacobian(function(theta) score.glm(x,p=theta,indiv=FALSE,...),p)
}

##' @export
information.glm <- function(x,...) hessian.glm(x,...)

glm_logLik.lvm <- function(object,...) {
    attr(GLMscore(object,...),"logLik")
}

glm_method.lvm <- NULL
glm_objective.lvm <- function(x,p,data,...) {
    GLMest(x,data,...)
}
glm_gradient.lvm <- function(x,p,data,...) {
    -GLMscore(x,p,data,...)
}

glm_variance.lvm <- function(x,p,data,opt,...) {
    opt$vcov
}

##' @export
rsq <- function(x,stderr=FALSE) {

    if (stderr) {
        v <- endogenous(x)
        vpar <- paste(v,v,sep=lava.options()$symbol[2])
        ic.v <- IC(model.frame(x)[,v])
        ic.mod <- IC(x)
        coef0 <- c(attributes(ic.v)$coef[vpar],
                   coef(x)[vpar])
        ic0 <- cbind(ic.v[,vpar],ic.mod[,vpar])
        p <- length(v)
        idx <- seq_len(p);
        ee <- estimate(NULL,data=NULL,
                       function(x) {
                           res <- (x[idx]-x[idx+p])/x[idx]
                           names(res) <- v
                           as.list(res)
                       },
                       print=function(x,...) {
                           cat("\nR-squared:\n\n")
                           print(x$coefmat)
                       },
                       coef=coef0, IC=ic0)

        res <- ee
        ##res <- list(ee)
        ## for (lat in latent(x)) {

        ##     v <- intersect(children(x,lat),endogenous(x))
        ##     vpar <- paste(v,v,sep=lava.options()$symbol[2])
        ##     lpar <- paste(lat,lat,sep=lava.options()$symbol[2])
        ##     rpar <- paste(v,lat,sep=lava.options()$symbol[1])
        ##     fix <- c(x$model$fix[lat,v,drop=TRUE],x$model$covfix[lat,lat])
        ##     pp <- coef(x)
        ##     idx <- x$model$parpos$A[lat,v]
        ##     idx2 <- x$model$parpos$P[lat,lat]
        ##     p0 <- c(idx,idx2)
        ##     p1 <- setdiff(unique(p0),0)
        ##     p2 <- match(p0,p1)

        ##     k <- length(v)
        ##     coef0 <- c(pp[p1],attributes(ic.v)$coef[vpar])
        ##     ic0 <- cbind(ic.mod[,p1],ic.v[,vpar])
        ##     ee <- estimate(NULL,data=NULL,
        ##                    function(p) {
        ##                        p. <- p[p2]
        ##                        p.[is.na(p.)] <- fix[is.na(p.)]
        ##                        res <- p.[seq_len(k)]^2*p.[k+1]/tail(p,k)
        ##                        names(res) <- v
        ##                        as.list(res)
        ##                    },
        ##                    print=function(x,...) {
        ##                        cat("\nVariance explained by '", lat,"':\n\n",sep="")
        ##                        print(x$coefmat)
        ##                    },coef=coef0,IC=ic0)
        ##     res <- c(res,list(ee))
        ## }

        return(res)
    }



    v <- c(endogenous(x),setdiff(latent(x),parameter(Model(x))))
    res <- coef(x,9,std="yx")
    idx <- with(attributes(res),
                which(type=="variance" & (var==from)))
    nam <- attributes(res)$var[idx]
    res <- 1-res[idx,5]
    names(res) <- nam
    res <- list("R-squared"=res)
    ## M <- moments(x,coef(x))
    ## v <- setdiff(vars(x),exogenous(x))
    ## vvar <- M$Cfull[cbind(v,v)]
    ## rsq <- (vvar-M$P[cbind(v,v)])/vvar

    if (length(latent(x))>0) {
        M <- moments(x,coef(x))
        nn <- names(res)
        for (lat in latent(x)) {
            v <- intersect(children(x,lat),endogenous(x))
            varl <- M$Cfull[lat,lat]
            varv <- M$Cfull[cbind(v,v)]
            fix <- c(x$model$fix[lat,v,drop=TRUE])
            pp <- coef(x)
            if (inherits(x,"lvm.missing")) {
                mp <- match(coef(x$model),names(coef(x)))
                pp <- pp[mp]
            }
            idx1 <- x$model$parpos$A[lat,v]
            ##idx2 <- x$model$parpos$P[lat,lat]
            ##idx3 <- x$model$parpos$P[cbind(v,v)]
            p0 <- c(idx1)
            p1 <- setdiff(unique(p0),0)
            p2 <- match(p0,p1)
            p <- pp[p1]
            p. <- p[p2]
            p.[is.na(p.)] <- fix[is.na(p.)]
            k <- length(v)
            val <- (p.^2*varl)/varv; names(val) <- v
            res <- c(res,list(val))
            nn <- c(nn,paste0("Variance explained by '",lat,"'"))
        }
        names(res) <- nn
    }
    res
}

satmodel <- function(object,logLik=TRUE,data=model.frame(object),
              control=list(trace=1),
              weights=Weights(object),estimator=object$estimator,
              missing=inherits(object,"lvm.missing"),
              regr=FALSE,
              ...) {
    if (object$estimator=="gaussian" & logLik & !missing) {
        if (class(object)[1]%in%c("multigroupfit","multigroup")) {

            ll <- structure(0,nall=0,nobs=0,df=0,class="logLik")
            for (i in seq_len(Model(object)$ngroup)) {
                l0 <- logLik(Model(Model(object))[[i]],data=model.frame(object)[[i]],type="sat")

                ll <- ll+l0
                for (atr in c("nall","nobs","df"))
                    attributes(ll)[[atr]] <- attributes(ll)[[atr]]+attributes(l0)[[atr]]
            }

        }
        return(logLik(object, type="sat"))
    }
    covar <- exogenous(object)
    y <- endogenous(object)
    m0 <- Model(object)
    if (length(covar)>0)
        suppressWarnings(m0 <- regression(m0,y,covar))
    if (length(latent(m0))>0)
        kill(m0) <- latent(m0)
    cancel(m0) <- y
    if (!regr)
        suppressWarnings(covariance(m0) <- y)
    else {
        if (length(y)>1) {
            for (i in seq_len(length(y)-1))
                for (j in seq(i+1,length(y))) {
                    m0 <- regression(m0,y[i],y[j])
                }
        }
        exogenous(m0) <- covar
    }
    if (is.null(control$start)) {
        mystart <- rep(0,with(index(m0), npar.mean+npar))
        mystart[variances(m0,mean=TRUE)] <- 1
        control$start <- mystart
    }
    message("Calculating MLE of saturated model:\n")
    e0 <- estimate(m0,data=data,weights=weights,estimator=estimator,messages=0,control=control,missing=missing,...)
    if (logLik)
        return(logLik(e0))
    return(e0)
}

condition <- function(x) {
    a <- svd(x)
    return(max(a$d)/min(a$d))
}


##' Extract model summaries and GOF statistics for model object
##'
##' Calculates various GOF statistics for model object including global
##' chi-squared test statistic and AIC. Extract model-specific mean and variance
##' structure, residuals and various predicitions.
##'
##'
##' @aliases gof gof.lvmfit moments moments.lvm information information.lvmfit
##' score score.lvmfit logLik.lvmfit
##' @param object Model object
##' @param x Model object
##' @param p Parameter vector used to calculate statistics
##' @param data Data.frame to use
##' @param latent If TRUE predictions of latent variables are included in output
##' @param data2 Optional second data.frame (only for censored observations)
##' @param weights Optional weight matrix
##' @param n Number of observations
##' @param conditional If TRUE the conditional moments given the covariates are
##' calculated. Otherwise the joint moments are calculated
##' @param model String defining estimator, e.g. "gaussian" (see
##' \code{estimate})
##' @param debug Debugging only
##' @param chisq Boolean indicating whether to calculate chi-squared
##' goodness-of-fit (always TRUE for estimator='gaussian')
##' @param level Level of confidence limits for RMSEA
##' @param rmsea.threshold Which probability to calculate, Pr(RMSEA<rmsea.treshold)
##' @param all Calculate all (ad hoc) FIT indices: TLI, CFI, NFI, SRMR, ...
##' @param \dots Additional arguments to be passed to the low level functions
##' @usage
##'
##' gof(object, ...)
##'
##' \method{gof}{lvmfit}(object, chisq=FALSE, level=0.90, rmsea.threshold=0.05,all=FALSE,...)
##'
##' moments(x,...)
##'
##' \method{moments}{lvm}(x, p, debug=FALSE, conditional=FALSE, data=NULL, latent=FALSE, ...)
##'
##' \method{logLik}{lvmfit}(object, p=coef(object),
##'                       data=model.frame(object),
##'                       model=object$estimator,
##'                       weights=Weights(object),
##'                       data2=object$data$data2,
##'                           ...)
##'
##' \method{score}{lvmfit}(x, data=model.frame(x), p=pars(x), model=x$estimator,
##'                    weights=Weights(x), data2=x$data$data2, ...)
##'
##' \method{information}{lvmfit}(x,p=pars(x),n=x$data$n,data=model.frame(x),
##'                    model=x$estimator,weights=Weights(x), data2=x$data$data2, ...)
##'
##' @return A \code{htest}-object.
##' @author Klaus K. Holst
##' @keywords methods models
##' @export
##' @examples
##' m <- lvm(list(y~v1+v2+v3+v4,c(v1,v2,v3,v4)~x))
##' set.seed(1)
##' dd <- sim(m,1000)
##' e <- estimate(m, dd)
##' gof(e,all=TRUE,rmsea.threshold=0.05,level=0.9)
##'
##'
##' set.seed(1)
##' m <- lvm(list(c(y1,y2,y3)~u,y1~x)); latent(m) <- ~u
##' regression(m,c(y2,y3)~u) <- "b"
##' d <- sim(m,1000)
##' e <- estimate(m,d)
##' rsq(e)
##' ##'
##' rr <- rsq(e,TRUE)
##' rr
##' estimate(rr,contrast=rbind(c(1,-1,0),c(1,0,-1),c(0,1,-1)))
##'
`gof` <-
    function(object,...) UseMethod("gof")

##' @export
gof.lvmfit <- function(object,chisq=FALSE,level=0.90,rmsea.threshold=0.05,all=FALSE,...) {
    n <- object$data$n
    if (class(object)[1]=="multigroupfit") n <- sum(unlist(lapply(object$model$data,nrow)))
    loglik <- logLik(object,...)

    df <- attributes(loglik)$df
    nobs <- attributes(loglik)$nall*length(endogenous(object))
    myAIC <- -2*(loglik - df); attributes(myAIC) <- NULL
    myBIC <- -2*loglik + df*log(nobs); attributes(myBIC) <- NULL

    xconstrain <- intersect(unlist(lapply(constrain(object),function(z) attributes(z)$args)),manifest(object))

    l2D <- mean(object$opt$grad^2)
    S <- vcov(object)
    rnkV <- tryCatch(qr(S)$rank,error=function(...) 0)
    minSV <- attr(S,"minSV")
    condnum <- tryCatch(condition(vcov(object)),error=function(...) NULL)

    if (((object$estimator=="gaussian" & class(object)[1]!="lvm.missing") | chisq) & length(xconstrain)==0 ) {
        res <- list(fit=compare(object), n=n, logLik=loglik, BIC=myBIC, AIC=myAIC)
        q <- res$fit$statistic
        qdf <- res$fit$parameter
        if (all) {
            m0 <- lvm(manifest(object)); exogenous(m0) <- NULL
            e0 <- estimate(m0,model.frame(object))
            g0 <- gof(e0)
            qbaseline <- g0$fit$statistic
            qdfbaseline <- g0$fit$parameter
            CFI <- ((qbaseline-qdfbaseline) - (q-qdf))/(qbaseline-qdfbaseline)
            NFI <- (qbaseline-q)/qbaseline
            TLI <- (qbaseline/qdfbaseline-q/qdf)/(qbaseline/qdfbaseline-1)
            S <- object$data$S
            mu <- object$data$mu
            C <- modelVar(object)$C
            xi <- as.vector(modelVar(object)$xi)
            if (is.null(S)) S <- cov(model.frame(object))
            if (is.null(mu)) mu <- colMeans(model.frame(object))
            L <- diag(S)^0.5
            R <- (diag(1/L))%*%(S-C)%*%(diag(1/L))
            R2 <- (mu-xi)/L
            SRMR <- mean(c(R[upper.tri(R,diag=TRUE)],R2)^2)^0.5
            res <- c(res,list(CFI=CFI,NFI=NFI,TLI=TLI,C=C,S=S,SRMR=SRMR))
            ## idx <- index(object)$endo.idx
            ## if (length(latent(object))>0) {
            ##   SRMR.endo <- mean(c(R[idx,idx][upper.tri(R[idx,idx],diag=TRUE)],R2[idx])^2)^0.5
            ##   res <- c(res,list("SRMR(endogenous)"=SRMR.endo))
            ## }
        }
        if (rnkV==ncol(vcov(object)) && (!is.null(minSV) && minSV>1e-12)) {

            rmseafun <- function(...) {
                epsilon <- function(lambda) sapply(lambda,function(x)
                    ifelse(x>0 & qdf>0,sqrt(x/(qdf*(n))),0)) ## n-1,n vs. n-df
                opf <- function(l,p) suppressWarnings(p-pchisq(q,df=qdf,ncp=l))
                alpha <- (1-level)/2
                RMSEA <- epsilon(q-qdf)
                B <- max(q-qdf,0)
                lo <- hi <- list(root=0)
                if (RMSEA>0 && opf(0,p=1-alpha)<0) {
                    hi <- uniroot(function(x) opf(x,p=1-alpha),c(0,B))
                }
                if (opf(B,p=alpha)<0) {
                    lo <- uniroot(function(x) opf(x,p=alpha),c(B,n))
                }
                ci <- c(epsilon(c(hi$root,lo$root)))
                RMSEA <- c(RMSEA=RMSEA,ci);
                names(RMSEA) <- c("RMSEA",paste0(100*c(alpha,(1-alpha)),"%"))
                pval <- pchisq(q,qdf,(n*qdf*rmsea.threshold^2),lower.tail=FALSE)
                res <- list(aa=((q-qdf)/(2*qdf)^0.5),RMSEA=RMSEA, level=level, rmsea.threshold=rmsea.threshold, pval.rmsea=pval)
                return(res)
            }
            rmseaval <- tryCatch(rmseafun(),error=function(e) NULL)
            res <- c(res,rmseaval)
        }
    } else {
        res <- list(n=n, logLik=loglik, BIC=myBIC, AIC=myAIC)
    }

    res <- c(res, L2score=l2D, rankV=rnkV, cond=condnum, k=nrow(vcov(object)))
    class(res) <- "gof.lvmfit"
    return(res)
}

##' @export
print.gof.lvmfit <- function(x,optim=TRUE,...) {
    if (!is.null(x$n)) {
        with(x,
             cat("\n Number of observations =", n, "\n"))
    }
    if (is.null(x$fit)) {
        with(x,
             cat(" Log-Likelihood =", logLik, "\n"))
    }
    with(x,  cat(" BIC =", BIC, "\n",
                 "AIC =", AIC, "\n"))
    if (!is.null(x$fit))
        with(x,
             cat(" log-Likelihood of model =", fit$estimate[1], "\n\n",
                 "log-Likelihood of saturated model =", fit$estimate[2], "\n",
                 "Chi-squared statistic: q =", fit$statistic,
                 ", df =", fit$parameter,
                 "\n  P(Q>q) =", fit$p.value, "\n"))
    if (!is.null(x$RMSEA)) {
        rr <- round(x$RMSEA*10000)/10000
        rmsea <- paste0(rr[1]," (",rr[2],";",rr[3],")")
        cat("\n RMSEA (",x$level*100,"% CI): ", rmsea,"\n",sep="")
        cat("  P(RMSEA<",x$rmsea.threshold,")=",  x$pval.rmsea,"\n",sep="")
    }
    for (i in c("TLI","CFI","NFI","SRMR","SRMR(endogenous)"))
        if (!is.null(x[[i]])) cat("", i,"=",x[[i]],"\n")

    if (optim) {
        cat("\nrank(Information) = ",x$rankV," (p=", x$k,")\n",sep="")
        cat("condition(Information) = ",x$cond,"\n",sep="")
        cat("mean(score^2) =",x$L2score,"\n")
    }

    invisible(x)
}



## gof.multigroupfit <- function(object,...) {
##   L0 <- logLik(object); df0 <- attributes(L0)$df
##   L1 <- logLik(object,type="sat"); df1 <- attributes(L1)$df

##   df <- df1-df0; names(df) <- "df"
##   Q <- -2*(L0-L1); attributes(Q) <- NULL; names(Q) <- "chisq";
##   pQ <- pchisq(Q,df,lower.tail=FALSE)
##   values <- c(L0,L1); names(values) <- c("log likelihood (model)", "log likelihood (saturated model)")
##   res <- list(statistic = Q, parameter = df,
##               p.value=pQ, method = "Likelihood ratio test",
##               estimate = values)
##   class(res) <- "htest"
##   return(res)
## }

##' Calculate partial correlations
##'
##' Calculate partial correlation coefficients and confidence limits via Fishers
##' z-transform
##'
##'
##' @param formula formula speciying the covariates and optionally the outcomes
##' to calculate partial correlation for
##' @param data data.frame
##' @param level Level of confidence limits
##' @param ... Additional arguments to lower level functions
##' @return A coefficient matrix
##' @author Klaus K. Holst
##' @keywords models regression
##' @examples
##'
##' m <- lvm(c(y1,y2,y3)~x1+x2)
##' covariance(m) <- c(y1,y2,y3)~y1+y2+y3
##' d <- sim(m,500)
##' partialcor(~x1+x2,d)
##'
##' @export
partialcor <- function(formula,data,level=0.95,...) {
  y <-  getoutcome(formula)
  if (length(y)==0) {
    preds <- all.vars(formula)
    yy <- setdiff(names(data),preds)
  } else {
    yy <- decomp.specials(y)  
    preds <- attr(y,"x")
  }  
  if (length(yy)<2)
      return(NULL)
  res <- c()
  for (i in seq_len(length(yy)-1))
      for (j in seq(i+1,length(yy))) {
          f <- as.formula(paste("cbind(",yy[i],",",yy[j],")~", paste(preds,collapse="+")))
          res <- rbind(res, partialcorpair(f,data,level=level))
          rownames(res)[nrow(res)] <- paste(yy[i],yy[j],sep="~")
      }
  return(res)
}


partialcorpair <- function(formula,data,level=0.95,...) {
  l <- lm(formula,data)
  k <- ncol(model.matrix(l))
  n <- nrow(model.matrix(l))
  r <- residuals(l)
  rho <- cor(r)[1,2]
  zrho <- atanh(rho)
  var.z <- 1/(n-k-3)
  ci.z <- zrho + c(-1,1)*qnorm(1-(1-level)/2)*sqrt(var.z)
  ci.rho <- tanh(ci.z)
  z <- 1/sqrt(var.z)*zrho
  p.z <- 2*(pnorm(-abs(z))) # p-value using z-transform for H_0: rho=0.
  return(c(cor=rho,z=z,pval=p.z,lowerCI=ci.rho[1],upperCI=ci.rho[2]))
}

##' Finds elements in vector or column-names in data.frame/matrix
##' 
##' Pattern matching in a vector or column names of a data.frame or matrix.
##' @param x vector, matrix or data.frame.
##' @param pattern regular expression to search for
##' @param subset If TRUE returns subset of data.frame/matrix otherwise just the matching column names
##' @param ignore.case Default ignore case
##' @param ... Additional arguments to 'grep'
##' @return A data.frame with 2 columns with the indices in the first and the
##' matching names in the second.
##' @author Klaus K. Holst
##' @seealso \code{\link{grep}}, and \code{\link{agrep}} for approximate string
##' matching,
##' @keywords misc utilities
##' @examples
##' data(iris)
##' head(Grep(iris,"(len)|(sp)"))
##' @export
`Grep` <-
    function(x, pattern, subset=TRUE, ignore.case = TRUE,...) {
        if (is.data.frame(x))
            nn <- names(x)
        else if (is.matrix(x))
            nn <- colnames(x)
        else nn <- x
        ii <- grep(pattern,nn,ignore.case=ignore.case,...)
        if (subset) {
            if (is.matrix(x) || is.data.frame(x))
                return(x[,ii,drop=FALSE])
            else return(x[ii])
        }  
        res <- data.frame(index=ii,name=nn[ii]);
        res
    }

`addattr` <- function(x,...) UseMethod("addattr")

`addattr.lvmfit` <- function(x,...) addattr(Model(x),...)

`addattr.lvm` <- function(x, attr, var=NULL, val=TRUE, fun=graph::nodeRenderInfo,debug=FALSE,...) {
    if (!is.null(var)) {
        Graph(x) <- addattr(Graph(x), attr=attr, var=var, val=val, fun=fun, debug=debug)
        return(x)
    } else {
        addattr(Graph(x), attr=attr, var=var, val=val, fun=fun)
    }
}

`addattr.graphNEL` <- function(x, attr, var=NULL, val=TRUE,fun="graph::nodeRenderInfo",debug=FALSE,...) {
    if (is.null(var)) {
        ff <- strsplit(fun,"::")[[1]]
        if (length(ff)>1) {
            ff <- getFromNamespace(ff[2],ff[1])
        }
        f <- do.call(ff,list(x))
        if (is.null(val) || !is.logical(f[[attr]]))
            attrvar <- f[[attr]]
        else
            attrvar <- names(f[[attr]])[which(val==f[[attr]])]
        return(attrvar)
    }
    if (is.character(val))
        myexpr <- paste0("list(",attr,"=c(", paste0("\"",var,"\"=\"",val,"\"" , collapse=", "), "))")
    else
        myexpr <- paste0("list(",attr,"=c(", paste0("\"",var,"\"=",val, collapse=", "), "))")
    Debug(list("str=",myexpr),debug)
    eval(parse(text=paste0(fun,"(x) <- ",myexpr)))
    return(x)
}

##' Composite Likelihood for probit latent variable models
##'
##' Estimate parameters in a probit latent variable model via a composite
##' likelihood decomposition.
##' @param x \code{lvm}-object
##' @param data data.frame
##' @param k Size of composite groups
##' @param type Determines number of groups. With \code{type="nearest"} (default)
##' only neighboring items will be grouped, e.g. for \code{k=2}
##' (y1,y2),(y2,y3),... With \code{type="all"} all combinations of size \code{k}
##' are included
##' @param pairlist A list of indices specifying the composite groups. Optional
##' argument which overrides \code{k} and \code{type} but gives complete
##' flexibility in the specification of the composite likelihood
##' @param messages Control amount of messages printed
##' @param estimator Model (pseudo-likelihood) to use for the pairs/groups
##' @param quick If TRUE the parameter estimates are calculated but all additional
##' information such as standard errors are skipped
##' @param \dots Additional arguments parsed on to lower-level functions
##' @return An object of class \code{estimate.complik} inheriting methods from \code{lvm}
##' @author Klaus K. Holst
##' @seealso estimate
##' @keywords models regression
##' @export
##' @examples
##' m <- lvm(c(y1,y2,y3)~b*x+1*u[0],latent=~u)
##' ordinal(m,K=2) <- ~y1+y2+y3
##' d <- sim(m,50,seed=1)
##' if (requireNamespace("mets", quietly=TRUE)) {
##'    e1 <- complik(m,d,control=list(trace=1),type="all")
##' }
complik <- function(x, data, k=2, type=c("all","nearest"), pairlist,
             messages=0, estimator="normal", quick=FALSE, ...) {

    y <- setdiff(endogenous(x),latent(x))
    binsurv <- rep(FALSE,length(y))
    for (i in seq_along(y)) {
        z <- try(data[,y[i]],silent=TRUE)
        ## binsurv[i] <- is.Surv(z) | (is.factor(z) && length(levels(z))==2)
        if (!inherits(z,"try-error"))
            binsurv[i] <- inherits(z,"Surv") | (is.factor(z))
    }

    ord <- ordinal(x)
    binsurv <- unique(c(y[binsurv],ord))
    if (!missing(pairlist)) {
        binsurvpos <- which(colnames(data)%in%endogenous(x))
    } else {
        binsurvpos <- which(colnames(data)%in%binsurv)
    }
    if (missing(pairlist)) {
        if (type[1]=="all") {
            mypar <- combn(length(binsurv),k) ## all pairs (or multiplets), k=2: k*(k-1)/2
        } else {
            mypar <- sapply(0:(length(binsurv)-k), function(x) x+1:k)
        }
    } else {
        mypar <- pairlist
    }

    if (is.matrix(mypar)) {
        mypar0 <- mypar; mypar <- c()
        for (i in seq(ncol(mypar0)))
            mypar <- c(mypar, list(mypar0[,i]))
    }
    nblocks <- length(mypar)

    mydata0 <- data[,,drop=FALSE]
    mydata <-  as.data.frame(matrix(NA, nblocks*nrow(data), ncol=ncol(data)))
    names(mydata) <- names(mydata0)
    for (i in seq_len(ncol(mydata))) {
        if (is.factor(data[,i])) {
            mydata[,i] <- factor(mydata[,i],levels=levels(mydata0[,i]))
        }
        if (survival::is.Surv(data[,i])) {
            S <- data[,i]
            for (j in 2:nblocks) S <- rbind(S,data[,i])
            S[] <- NA
            mydata[,i] <- S
        }
    }
    for (ii in 1:nblocks) {
        data0 <- data;
        for (i in binsurvpos[-mypar[[ii]]]) {
            if (survival::is.Surv(data[,i])) {
                S <- data0[,i]
                S[] <- NA
                data0[,i] <- S
            } else {
                data0[,i] <- NA
                if (is.factor(data[,i])) data0[,i] <- factor(data0[,i],levels=levels(data[,i]))
            }
        }
        mydata[seq_len(nrow(data))+(ii-1)*nrow(data),] <- data0
    }
    suppressWarnings(e0 <- estimate(x,data=mydata,estimator=estimator,missing=TRUE,messages=messages,
                                    hessian=!quick,
                                    ...))
    e0$n <- nrow(data)
    e0$eigenvalues <- rep(1,nrow(e0$coef))
    e0$coef[,2:4] <- NA
    if (!quick) {
        S <- score(e0,indiv=TRUE)
        nd <- nrow(data)
        block1 <- which((1:nd)%in%(rownames(S)))
        blocks <- sapply(seq_len(nblocks), function(x) seq_along(block1)+length(block1)*(x-1))
        if (nblocks==1) {
            Siid <- S
        } else {
            Siid <- matrix(0,nrow=length(block1),ncol=ncol(S))
            for (j in seq_len(ncol(blocks))) {
                Siid <- Siid+S[blocks[,j],]
            }
        }
        ##B <- solve(information(e0, type="hessian"))
        D <- numDeriv::jacobian(function(p) score(e0, p=p), coef(e0), method=lava.options()$Dmethod)
        B <- -Inverse(D)*NROW(Siid)
        A <- crossprod(Siid)/NROW(Siid)
        e0$bread <- B
        e0$meat <- A
        e0$iidscore <- Siid
        e0$blocks <- blocks
        BA <- B%*%A
        eig <- eigen(BA)$values
        e0$eigenvalues <- eig
        e0$vcov <- (BA%*%B)/NROW(Siid)
        cc <- e0$coef;
        cc[,2] <- sqrt(diag(e0$vcov))
        cc[,3] <- cc[,1]/cc[,2]; cc[,4] <- 2*(1-pnorm(abs(cc[,3])))
        e0$coef <- cc
    }
    class(e0) <- c("estimate.complik",class(e0))
    return(e0)
}

##' @export
score.estimate.complik <- function(x,indiv=FALSE,...) {
    if (!indiv)
        return(colSums(x$iidscore))
    x$iidscore
}

##' @export
IC.estimate.complik <- function(x,...) {
    IC.default(x,bread=x$bread,...)
}

##' @export
logLik.estimate.complik <- function(object,...) {
    with(object, structure(opt$objective,
              nall=n,
              nobs=n,
              df=sum(eigenvalues),
              class="logLik"))
}

##################################################
## Cohen's kappa
##################################################

##' @export
kappa.multinomial <- function(z,all=FALSE,...) {
    pp <- length(coef(z))
    if ((length(z$levels)!=2) || !(identical(z$levels[[1]],z$levels[[2]])))
        stop("Expected square table and same factor levels in rows and columns")
    k <- length(z$levels[[1]])
    zeros <- rbind(rep(0,pp))
    A0 <- zeros; A0[diag(z$position)] <- 1
    A <- matrix(0,ncol=pp,nrow=2*k)
    for (i in seq(k)) A[i,z$position[i,]] <- 1
    for (i in seq(k)) A[i+k,z$position[,i]] <- 1
    b <- estimate(z,function(p) as.vector(rbind(A0,A)%*%p),IC=TRUE)
    b2 <- estimate(b,function(p) c(p[1],sum(p[seq(k)+1]*p[seq(k)+k+1])),IC=TRUE)
    if (!all) {
        return(estimate(b2,function(p) list(kappa=(p[1]-p[2])/(1-p[2])),IC=TRUE,...))
    }
    estimate(b2,function(p) list(kappa=(p[1]-p[2])/(1-p[2]),agree=p[1], independence=p[2]),IC=TRUE,...)
}

##' @export
kappa.table <- function(z,...) {
    kappa(multinomial(Expand(z)),...)
}

##' @export
kappa.data.frame <- function(z,...) {
    kappa(multinomial(z),...)
}

##' Generic correlation method
##'
##' @title Generic method for extracting correlation coefficients of model object
##' @param x Object
##' @param \dots Additional arguments
##' @author Klaus K. Holst
##' @export
"correlation" <- function(x,...) UseMethod("correlation")

##' @export
correlation.lvmfit <- function(x,z=TRUE, IC=FALSE, back.transform=TRUE,...) {
    pp <- matrices2(Model(x), with(index(x),seq_len(npar+npar.mean+npar.ex)))$P
    pos <- pp[lower.tri(pp)][(index(x)$P0)[lower.tri(pp)]==1]
    if (length(pos)<1) return(NULL)
    pp0 <- pp
    pp0[index(x)$P0!=1] <- 0; pp0[lower.tri(pp0)] <- 0
    mynames <- vars(x)
    n <- nrow(pp0)
    ff <-  function(p) {
        res <- numeric(length(pos))
        nn <- character(length(pos))
        for (i in seq_along(pos)) {
            p0 <- pos[i]
            idx <- which(pp0==p0)
            rowpos <- (idx-1)%%n + 1
            colpos <- ceiling(idx/n)
            coefpos <- c(p0,pp0[rbind(c(rowpos,rowpos),c(colpos,colpos))])
            pval <- pp[rbind(c(rowpos,rowpos),c(colpos,colpos))]
            phi.v1.v2 <- numeric(3);
            newval <- p[coefpos]
            phi.v1.v2[coefpos!=0] <- newval
            phi.v1.v2[coefpos==0] <- pval[tail(coefpos==0,2)]
            rho <- atanh(phi.v1.v2[1]/sqrt(prod(phi.v1.v2[-1])))
            res[i] <- rho
            nn[i] <- paste(mynames[c(rowpos,colpos)],collapse="~")
        }
        structure(res,names=nn)
    }
    V <- NULL
    if (!IC) V <- vcov(x)
    if (back.transform) {
        back.transform <- tanh
    } else {
        back.transform <- NULL
    }
    estimate(x,coef=coef(x), vcov=V, f=ff, back.transform=back.transform,
             IC=IC, ...)
}


##' @export
correlation.matrix <- function(x,z=TRUE,back.transform=TRUE,mreg=FALSE,return.all=FALSE,...) {
    if (mreg) {
        m <- lvm()
        covariance(m,pairwise=TRUE) <- colnames(x)
        try(e <- estimate(m,as.data.frame(x),...),silent=TRUE)
        res <- correlation(e,...)
        if (return.all) {
            return(list(model=m,estimate=e,correlation=res))
        }
        return(res)
    }
    if (ncol(x)==2) {
        ii <- IC(x)
        ee <- estimate(coef=attributes(ii)$coef[3:5], IC=ii[,3:5])
        if (z) {
            if (back.transform) {
                ee <- estimate(ee, function(x) atanh(x[2]/sqrt(x[1]*x[3])), back.transform=tanh)
            } else {
                ee <- estimate(ee, function(x) atanh(x[2]/sqrt(x[1]*x[3])))
            }
        } else {
            ee <-  estimate(ee, function(x) x[2]/sqrt(x[1]*x[3]))
        }
        return(ee)
    }

    e <- c()
    R <- diag(nrow=ncol(x))
    dimnames(R) <- list(colnames(x),colnames(x))
    for (i in seq(ncol(x)-1))
        for (j in seq(i+1,ncol(x))) {
            e <- c(e,list(correlation(x[,c(i,j)],z=z,back.transform=FALSE,...)))
            R[j,i] <- coef(e[[length(e)]])
            if (z) R[j,i] <- tanh(R[j,i])
        }
    R <- R[-1,-ncol(R),drop=FALSE]
    res <- do.call(merge, c(e, paired=TRUE))
    if (z && back.transform) {
        res <- estimate(res,back.transform=tanh, print=function(x,digits=1,...) {
            print(x$coefmat[,-2,drop=FALSE],...)
            cat("\n")
            print(offdiag(R,type=4),digits=digits,...)
        })
    }
    return(res)
}

##' @export
correlation.data.frame <- function(x,...) {
    correlation(as.matrix(x),...)
}

nlminb2 <- function(start,objective,gradient,hessian,...) {
  nlminbcontrols <- c("eval.max","iter.max","trace","abs.tol","rel.tol","x.tol","step.min")
  dots <- list(...)
  control <- list(...)$control
  control <- control[names(control)%in%nlminbcontrols]
  dots$control <- control
  if (length(dots$trace)>0 && dots$trace>0) cat("\n")
  mypar <- c(list(start=start,objective=objective,gradient=gradient,hessian=hessian),dots)
  mypar["debug"] <- NULL
  do.call("nlminb", mypar)
}

nlminb1 <- function(start,objective,gradient,hessian,...) {
  nlminb2(start,objective,gradient=gradient,hessian=NULL,...)
}

nlminb0 <- function(start,objective,gradient,hessian,...) {
  nlminb2(start,objective,gradient=NULL,hessian=NULL,...)
}

estimatingfunction <- function(start,objective,gradient,hessian,NR=FALSE,...) {
  myobj <- function(x,...) {
    S <- gradient(x,...)
    crossprod(S)[1]
  }
  if (!missing(hessian) && !is.null(hessian)) {
    mygrad <- function(x) {
      H <- hessian(x)
      S <- gradient(x)
      2*S%*%H
    }
  } else {
    hessian <- function(x) numDeriv::jacobian(gradient,x,method=lava.options()$Dmethod)
    mygrad <- function(x) {
      H <- hessian(x)
      S <- gradient(x)
      2*S%*%H
    }
  }
  if (NR) {
    op <- lava::NR(start,gradient=gradient,hessian=hessian,...)
  } else {
    op <- nlminb2(start,myobj,mygrad,hessian=NULL,...)
  }
  return(op)
}

estimatingfunction0 <- function(...,hessian=NULL) estimatingfunction(...,hessian=hessian)

##' @title Newton-Raphson method
##' 
##' @param start Starting value
##' @param objective Optional objective function (used for selecting step length)
##' @param gradient gradient
##' @param hessian hessian (if NULL a numerical derivative is used)
##' @param control optimization arguments (see details)
##' @param args Optional list of arguments parsed to objective, gradient and hessian
##' @param ... additional arguments parsed to lower level functions
##' @details
##' \code{control} should be a list with one or more of the following components:
##' \itemize{
##' \item{trace} integer for which output is printed each 'trace'th iteration
##' \item{iter.max} number of iterations
##' \item{stepsize}: Step size (default 1)
##' \item{nstepsize}: Increase stepsize every nstepsize iteration (from stepsize to 1)
##' \item{tol}: Convergence criterion (gradient)
##' \item{epsilon}: threshold used in pseudo-inverse
##' \item{backtrack}: In each iteration reduce stepsize unless solution is improved according to criterion (gradient, armijo, curvature, wolfe)
##' }
##' @export
##' @examples
##' # Objective function with gradient and hessian as attributes
##' f <- function(z) {
##'    x <- z[1]; y <- z[2]
##'    val <- x^2 + x*y^2 + x + y
##'    structure(val, gradient=c(2*x+y^2+1, 2*y*x+1),
##'              hessian=rbind(c(2,2*y),c(2*y,2*x)))
##' }
##' NR(c(0,0),f)
##' 
##' # Parsing arguments to the function and
##' g <- function(x,y) (x*y+1)^2
##' NR(0, gradient=g, args=list(y=2), control=list(trace=1,tol=1e-20))
##' 
##' 
NR <- function(start,objective=NULL,gradient=NULL,hessian=NULL,control,args=NULL,...) {
  control0 <- list(trace=0,
                   stepsize=1,
                   lambda=0,
                   ngamma=0,
                   gamma2=0,
                   backtrack=TRUE,
                   iter.max=200,
                   tol=1e-6,
                   stabil=FALSE,
                   epsilon=1e-9)
  if (!missing(control)) {
      control0[names(control)] <- control
      # Backward compatibility:
      if (!is.null(control0$gammma)) control0$stepsize <- control0$gamma
  }

  ## conditions to select the step length
  if(control0$backtrack[1] == "armijo"){
    control0$backtrack <- c(1e-4,0) # page 33
  }
  if(control0$backtrack[1] == "curvature"){
    control0$backtrack <- c(0,0.9) # page 34
  }
  if(control0$backtrack[1] == "wolfe"){
      control0$backtrack <- c(1e-4,0.9)
  }
  if(!is.logical(control0$backtrack) || length(control0$backtrack)!=1){
    if(length(control0$backtrack) != 2){
      stop("control$backtrack must have length two if not TRUE or FALSE \n")
    }
    if(any(!is.numeric(control0$backtrack)) || any(abs(control0$backtrack)>1)){
      stop("elements in control$backtrack must be in [0,1] \n")
    }
    if(control0$backtrack[2]==0){
      control0$backtrack[2] <- +Inf # no Wolfe condition
    }
  }
  obj <- objective
  grad <- gradient
  hess <- hessian
  if (!is.null(args)) {
      if (!is.list(args)) args <- list(args)
      if (!is.null(objective))
          obj <- function(p) do.call(objective, c(list(p),args))
      if (!is.null(gradient))
          grad <- function(p) do.call(gradient, c(list(p),args))
      if (!is.null(hessian))
          hess <- function(p) do.call(hessian, c(list(p),args))
  }
  if (control0$trace>0) {
       cat("\nIter=0\t")
      if (!is.null(obj))
          cat("Objective=",obj(as.double(start)))
      cat(";\t\n \tp=", paste0(formatC(start), collapse=" "),"\n")
  }
  gradFun = !is.null(grad)
  if (!gradFun & is.null(hess)) {
    hess <- function(p) {
      ff <- obj(p)
      res <- attributes(ff)$hess
      if (is.function(res)) {
          res <- res(p)
          attributes(res)$grad <- as.vector(attributes(ff)$grad(p))
      } else {      
          attributes(res)$grad <- as.vector(attributes(ff)$grad)
      }
      return(res)
    }
    grad <- function(p) {
        if (control0$trace>0) print("Numerical gradient")
        numDeriv::jacobian(obj,p)
    }
  }
  oneiter <- function(p.orig,Dprev,return.mat=FALSE,iter=1) {
      D <- I <- NULL # Place-holders for gradient and negative hessian
      if (!is.logical(control0$backtrack) && !is.null(obj)) { # Back-tracking based on objective function evaluations
          objective.origin <- obj(p.orig)
          D <- attributes(objective.origin)$grad
          I <- attributes(objective.origin)$hess
          if (!is.null(I)) I <- -I
      }
      if (is.null(D) || is.null(I)) {
          if (!is.null(hess)) {
              H <- hess(p.orig)
          }
          if (is.null(hess) || is.null(H)) {
              if (control0$trace>0) print("Numerical Hessian")
              I <- -numDeriv::jacobian(grad,p.orig,method=lava.options()$Dmethod)
          } else {
              I <- -H
          }
          D <- attributes(I)$grad
          if (is.null(D)) {
              D <- grad(p.orig)
          }
      }
      
    if (return.mat) return(list(D=D,I=I))
    if (control0$stabil) {
      if (control0$lambda!=0) {
        if (control0$lambda<0) {
          sigma <- (t(D)%*%(D))[1]
        } else {
          sigma <- control0$lambda
        }
        sigma <- min(sigma,10)
        I <- I+control0$gamma2*sigma*diag(nrow=nrow(I))
      } else {
        sigma <- ((D)%*%t(D))
        I <- I+control0$gamma2*(sigma)
      }
    }
    iI <- Inverse(I, symmetric=TRUE, tol=control0$epsilon)
    Delta <- control0$stepsize*tryCatch(solve(I, cbind(as.vector(D))),
                            error=function(...) { ## Fall back to Pseudo-Inverse using SVD:
                                iI%*%cbind(as.vector(D))})
    Lambda <- 1 ## Initial step-size
    if (identical(control0$backtrack, TRUE)) {
      mD0 <- mean(Dprev^2)
      mD <- mean(D^2)
      p <- p.orig + as.vector(Lambda*Delta)
      while (mD>=mD0) {
        if (gradFun) {
          D = grad(p)
        } else {
          DI <- oneiter(p,return.mat=TRUE)
          D = DI$D
        }
        mD = mean(D^2)
        if (is.nan(mD)) mD=mD0
        Lambda <- Lambda/2
        if (Lambda<1e-4) break;
        p <- p.orig + as.vector(Lambda*Delta)
      }

    } else if(identical(control0$backtrack, FALSE)) {
      p <- p.orig + Lambda*Delta
    } else {  # objective(p.orig) - obj(p) <= mu*Lambda*grad(p.orig)*Delta
        ## curvature
        c_D.origin_Delta <- control0$backtrack * c(rbind(D) %*% Delta)
        p <- p.orig + as.vector(Lambda*Delta)

        mD0 <- c(objective.origin + Lambda * c_D.origin_Delta[1], abs(c_D.origin_Delta[2]))#
        objective.new <- obj(p)
        grad.new <- attributes(objective.new)$grad
        if (is.null(grad.new)) {
            grad.new <- grad(p)
        }
        mD <- c(objective.new, abs(grad.new %*% Delta))
        count <- 0
        while (any(mD>mD0) || any(is.nan(mD))) {
            count <- count+1
            Lambda <- Lambda/2
            if (Lambda<1e-4) break;
            p <- p.orig + Lambda*Delta
            objective.new <- obj(p)
            grad.new <- attributes(objective.new)$grad
            if(!is.infinite(mD0[1])){
                mD0[1] <- objective.origin + Lambda * c_D.origin_Delta[1]#
                mD[1] <- objective.new
            }
            if(!is.infinite(mD0[2])){
                if (is.null(grad.new)) {
                    grad.new <- grad(p)
                }
                mD[2] <- abs(grad.new %*% Delta)
            }
        }
    }

    return(list(p=p,D=D,iI=iI))
  }

  count <- count2 <- 0
  thetacur <- start
  stepsizecount <- 0
  Dprev <- rep(Inf,length(start))
  for (jj in seq_len(control0$iter.max)) {
    stepsizecount <- stepsizecount+1
    count <-  count+1
    count2 <- count2+1
    newpar <- oneiter(thetacur,Dprev,iter=jj)
    Dprev <- newpar$D
    thetacur <- newpar$p
    if (!is.null(control0$nstepsize) && control0$nstepsize>0) {
      if (control0$nstepsize<=stepsizecount) {
        control0$stepsize <- sqrt(control0$stepsize)
        stepsizecount <- 0
      }
    }
    if (count2==control0$trace) {
        cat("Iter=", count,"\t",sep="")
        if (!is.null(obj))
            cat("Objective=",obj(as.double(newpar$p)))
        cat(";\n\tD=", paste0(formatC(newpar$D),
                               collapse = " "), "\n")
        cat("\tp=", paste0(formatC(thetacur), collapse = " "),
            "\n")
        count2 <- 0
    }
    if (mean(newpar$D^2)^.5<control0$tol) break;
  }

  res <- list(par=as.vector(thetacur), iterations=count, method="NR",
              gradient=newpar$D, iH=newpar$iI)
  return(res)
}

##' Create a Data Frame from All Combinations of Factors
##'
##' Simple wrapper of the 'expand.grid' function.  If x is a table
##' then a data frame is returned with one row pr individual
##' observation.
##' @title Create a Data Frame from All Combinations of Factors
##' @param _data Data.frame
##' @param ... vectors, factors or a list containing these
##' @author Klaus K. Holst
##' @export
##' @aliases Expand
##' @examples
##' dd <- Expand(iris, Sepal.Length=2:8, Species=c("virginica","setosa"))
##' summary(dd)
##'
##' T <- with(warpbreaks, table(wool, tension))
##' Expand(T)
Expand <- function(`_data`, ...) {
    if (missing(`_data`)) {
        return(expand.grid(...))
    }
    if (inherits(`_data`, "table")) {
        M <- as.data.frame(`_data`)
        idx <- rep(seq(nrow(M)), M[, ncol(M)])
        return(M[idx, -ncol(M), drop=FALSE])
    }
    if (!inherits(`_data`, "data.frame")) {
        return(expand.grid(`_data`, ...))
    }
    dots <- list(...)
    nn <- names(dots)
    for (n in nn) {
        y <- dots[[n]]
        if (is.factor(`_data`[1, n])) {
            dots[[n]] <- factor(y, levels=levels(`_data`[1, n]))
        }
    }
    do.call("expand.grid", dots)
}

##' Wrap vector
##'
##' Wrap vector
##' @param x Vector or integer 
##' @param delta Shift
##' @param ... Additional parameters
##' @export
##' @examples
##' wrapvec(5,2)
wrapvec <- function(x,delta=0L,...) {
    if (length(x)==1 && floor(x)==x && x>0) {
        x <- seq(x)
    }
    if (delta==0L) return(x)
    x[(seq_along(x)+delta-1L)%%length(x)+1L]
}

###{{{ Objective

IV_method.lvm <- NULL
IV_objective.lvm <- function(x,p,data,...) {
    IV2(x,data,...)
}
IV_variance.lvm <- function(x,p,data,opt,...) {
    opt$vcov
}

IV0_method.lvm <- NULL
IV0_objective.lvm <- function(x,p,data,...) {
    IV2(x,data,type="non-robust",...)
}
IV0_variance.lvm <- function(x,p,data,opt,...) {
    opt$vcov
}

IV1_method.lvm <- NULL
IV1_objective.lvm <- function(x,p,data,...) {
    IV(x,data)
}
IV1_variance.lvm <- function(x,p,data,opt,...) {
    opt$vcov
}

###}}} Objective

CondVar <- function(S,idx) {
    idx2 <- setdiff(seq_len(ncol(S)),idx)
    S11 <- S[idx2,idx2];
    S22 <- S[idx,idx]
    S12 <- S[idx2,idx]
    S11-S12%*%solve(S22)%*%t(S12)
}

varest <- function(x,data) {
    p <- IV(x,data)$estimate
    idx <- match(names(p),coef(x,mean=TRUE))
    x0 <- parfix(Model(x),idx,p)
    index(x0) <- reindex(x0,zeroones=TRUE,deriv=TRUE)

    A <- t(index(x)$A)
    Afix <- A; Afix[t(index(x)$M0)==1] <- 0
    A[A!=0] <- 1
    k <- nrow(A)
    I <- diag(nrow=k)
    Ap <- modelVar(x)$A ## Estimated parameter matrix

    indicators <- setdiff(vars(x)[rowSums(A)==1],exogenous(x))
    responses <- endogenous(x,top=TRUE)
    y.indicators <- responses[rowSums(A[responses,])==1]
    Sigma <- var(data[,manifest(x)])

    var.eta <- c()
    for (eta in latent(x)) {
        reachable <- acc(x$M,eta)
        ys <- intersect(names(reachable),y.indicators)
        lambdas <- c()
        for (y in ys) {
            pp <- path(Model(x), from=eta, to=y)
            lambda1 <- 0
            for (i in seq_along(pp)) {
                lambda <- 1
                for (j in seq_len(length(pp[[i]])-1))
                    lambda <- lambda*Ap[pp[[i]][j],pp[[i]][j+1]]
                lambda1 <- lambda1+lambda
            }
            lambdas <- c(lambdas,lambda1)
        }
        val <- outer(1/lambdas,1/lambdas)*Sigma[ys,ys]
        var.eta <- c(var.eta, mean(val[upper.tri(val)]))
    }

    S <- rep(0,k); S[match(manifest(x),vars(x))] <- diag(Sigma); S[match(latent(x),vars(x))] <- var.eta; names(S) <- vars(x)
    I <- diag(nrow=k)
    IA <- (I-t(Ap))
    IA%*%cbind(S)%*%t(IA)

}


## Instrumental Variable Estimator / 2SLS
##' @export
IV <- function(m,data,R2thres=0,type="robust", ...) {
    if (length(constrain(m))>0) stop("Nonlinear constrains not supported!")
    if (inherits(m,"lvmfit")) {
        m <- Model(m)
    }
    R2 <- cor(data[,manifest(m)])^2

    A <- t(index(m)$A)
    Afix <- A; Afix[t(index(m)$M0)==1] <- 0
    A[A!=0] <- 1
    P <- index(m)$P
    k <- nrow(A)
    I <- diag(nrow=k)
    B <- rbind(I,solve(I-A))
    VV <- B%*%P%*%t(B)
    u.var <- index(m)$vars
    lat.idx <- with(index(m), which(vars%in%latent))
    if (length(lat.idx)==0) stop("Estimator only defined for models with latent variable")
    y.var <- endogenous(m)
    y.idx <- which(index(m)$vars%in%y.var)
    x.idx <- which(vars(m)%in%exogenous(m))

    ## Set of Indicator variables:
    indicators <- c()
    for (i in seq_len(nrow(A))) {
        ifix <- (Afix[i,]==1)
        if ((sum(ifix)==1) &  all(A[i,-which(ifix)]==0))
            indicators <- c(indicators, i)
    }
    y.indicators <- intersect(indicators,y.idx)

    y.scale <- list()
    for (eta in lat.idx) {
        pred.eta <- intersect(y.idx, which(Afix[,eta]==1)) ## Candidates for
        ## eta = y-epsilon
        if (length(pred.eta)<1)
            pred.eta <- intersect(lat.idx, which(Afix[,eta]==1))
        myidx <- c()
        for (y in pred.eta) {
            y.pred <- setdiff(eta,which(A[y,]==1)) ## No other variables predicting y?
            if (length(y.pred)==0)
                myidx <- c(myidx,y)
        }
        y.scale <- c(y.scale, list(myidx))
    }

    if (any(unlist(lapply(y.scale, function(x) length(x)))<1)) stop("At least one scale-measurement pr. latent variable")

    vv <- setdiff(seq_len(k),c(unlist(y.scale),x.idx))

    Ecor <- list()
    eta.surrogate <- c()
    latno <- 0
    for (e in lat.idx) {
        latno <- latno+1
        y0 <- y.scale[[latno]][1]
        if (!(y0%in%lat.idx)) {
            eta.surrogate <- c(eta.surrogate,vars(m)[y0])
            Ecor <- c(Ecor,list(y0))
        }
        else {
            v0 <- vars(m)[-c(e,indicators)]
            ##m.sub <- subset(m,vars(m)[c(e,indicators)])
            m.sub <- rmvar(m,v0)
            i <- 0
            while (i<length(y.indicators)) {
                i <- i+1
                pp <- path(m.sub,from=vars(m)[e],to=vars(m)[y.indicators[i]])[[1]]
                if (!is.null(pp)) {
                    Ecor <- c(Ecor,
                              list(which(vars(m)%in%pp[-1])))
                    eta.surrogate <- c(eta.surrogate, tail(pp,1))
                }
            }
        }
    };
    names(eta.surrogate) <- latent(m)

    coefname <- coef(m,mean=TRUE)
    P0 <- P
    V. <- list()
    Z. <- list()
    Y. <- list()
    count <- 0
    ff <- list()
    instruments <- c()
    parname <- c()
    for (v in vv) {
        pred <- which(A[v,]==1)
        if (sum(pred)>0) {
            Debug(vars(m)[v])
            pred.lat <- intersect(pred,lat.idx) # Any latent predictors?
            lpos <- match(v,lat.idx)
            lppos <- match(pred.lat,lat.idx)
            ecor <- c(v,unlist(Ecor[lppos]))
            if (!is.na(lpos)) {
                v0 <- match(eta.surrogate[lpos],vars(m))
                ecor <- c(ecor,Ecor[[lpos]])
            } else {
                v0 <- v
            }

            ecor <- unique(c(v0,ecor))
            XX <- vars(m)[A[v,]==1]
            intpred <- exogenous(m)
            newf <- c()
            if (length(pred.lat)>0) {
                intpred <- vars(m)
                for (i in seq_along(pred.lat)) {
                    uncor <- which(colSums(VV[ecor,k+seq_len(k),drop=FALSE])==0)
                    uncor <- setdiff(uncor,c(lat.idx))
                    mypred <- vars(m)[uncor]
                    XX[XX==vars(m)[pred.lat[i]]] <- eta.surrogate[lppos[i]]
                    intpred <- intersect(intpred,mypred)
                    f <- toformula(eta.surrogate[lppos[i]],mypred)
                    ff <- c(ff,
                            f)
                    f2 <- list(f)
                    names(f2) <- vars(m)[i]
                    newf <- c(newf,f2)
                }
            }

            intpred <- intersect(intpred,manifest(m))
            R2max <- apply(R2[XX,intpred,drop=FALSE],2,max)
            if (any(R2max<R2thres)) intpred <- intpred[R2max>=R2thres]
            newf <- list(intpred); names(newf) <- vars(m)[v]
            instruments <- c(instruments, newf)
            covariates <- unique(c(setdiff(colnames(A)[A[v,]==1],latent(m)),intpred))
            if (length(covariates)==0) stop("No instruments")
            Z <- model.matrix(toformula("",c("1",XX)),data)
            Y <- as.matrix(data[,vars(m)[v0]])
            V <- model.matrix(toformula("",c("1",unique(covariates))),data)
            count <- count+1
            V. <- c(V.,list(V))
            Z. <- c(Z.,list(Z))
            Y. <- c(Y.,list(Y))
            XX <- vars(m)[A[v,]==1 & Afix[v,]!=1]
            parname <- c(parname, c(vars(m)[v0],paste(vars(m)[v],XX,sep=lava.options()$symbol[1])))
        } else {
            if (vars(m)[v]%in%latent(m)) {
                lpos <- match(v,lat.idx)
                v0 <- match(eta.surrogate[lpos],vars(m))
                Y <- matrix(data[,vars(m)[v0]],ncol=1)
                Y. <- c(Y.,list(Y))
                V. <- c(V.,list(cbind(rep(1,nrow(Y)))))
                Z. <- c(Z.,list(cbind(rep(1,nrow(Y)))))
                parname <- c(parname, names(eta.surrogate)[lpos])
            }
        }
    }

    LS <- function(X) {
        with(svd(X), v%*%diag(1/d,nrow=length(d))%*%t(u))
    }
    ##projection <- function(X) X%*%LS(X)
    P0 <- lapply(V.,LS)
    Zhat <- list(); for (i in seq_along(Z.)) Zhat <- c(Zhat, list(V.[[i]]%*%(P0[[i]]%*%Z.[[i]])))
    ZhatLS <- lapply(Zhat,LS)
    theta <- list(); for (i in seq_along(Y.)) theta <- c(theta, list(ZhatLS[[i]]%*%Y.[[i]]))
    u <- c()
    for (i in seq_along(Y.))
        u <- cbind(u, Y.[[i]]-Z.[[i]]%*%theta[[i]])

    covu <- crossprod(u)/nrow(u)

    theta.npar <- unlist(lapply(theta,length))
    theta.ncum <- c(0,cumsum(theta.npar))
    vartheta <- matrix(0,ncol=sum(theta.npar),nrow=sum(theta.npar))
    for (i in seq_along(theta)) {
        for (j in seq(i,length(theta))) {
            idx1 <- seq_len(theta.npar[i]) + theta.ncum[i]
            idx2 <- seq_len(theta.npar[j]) + theta.ncum[j]
            if (type=="robust") {
                zi <- ZhatLS[[i]]
                for (k in seq(nrow(zi))) zi[k,] <- zi[k,]*u[,i]
                zj <- ZhatLS[[j]]
                for (k in seq(nrow(zj))) zj[k,] <- zj[k,]*u[,j]
                uZZ <- zi%*%t(zj)
            } else {
                uZZ <- covu[i,j]* (ZhatLS[[i]]%*%t(ZhatLS[[j]]))
            }
            vartheta[idx1,idx2] <- uZZ
            if (i!=j) {
                vartheta[idx2,idx1] <- t(uZZ)
            }
        }
    }

    parname[which(parname%in%eta.surrogate)] <- names(eta.surrogate)[which(eta.surrogate%in%parname)]

    coef <- cbind(unlist(theta),diag(vartheta)^0.5); rownames(coef) <- parname; colnames(coef) <- c("Estimate","Std.Err")
    res <- list(estimate=coef[,1], vcov=vartheta)
    attributes(res)$surrogates <- eta.surrogate
    attributes(res)$instruments <- instruments
    return(res)
}

IV2 <- function(m,data,control=list(),...) {
    if (is.null(control$R2thres)) control$R2thres <- 0
    res <- IV(m,data,R2thres=control$R2thres,...)
    p <- res$estimate
    idx <- match(names(p),coef(m,mean=TRUE))
    x0 <- parfix(m,idx,p)
    index(x0) <- reindex(x0,zeroones=TRUE,deriv=TRUE)
    idx0 <- order(idx)
    p0 <- p[idx0]
    V0 <- res$vcov[idx0,idx0]
    if (is.null(control$variance) || control$variance) {
        suppressWarnings(e0 <- estimate(x0,data,...,messages=0,quick=TRUE))
        p0 <- c(p0,e0)
        V0 <- V0%++%matrix(0,ncol=length(e0),nrow=length(e0))
    }
    R2 <- noquote(formatC(cor(data[,manifest(m)])^2))
    colnames(R2) <- rownames(R2) <- manifest(m)
    l1 <- noquote(rbind(paste(latent(m),collapse=","),
                        paste(attributes(res)$surrogates,collapse=","),
                        ""))
    rownames(l1) <- c("Latent variables","Surrogate variables:","")
    colnames(l1) <- ""
    ii <- attributes(res)$instruments
    I <- noquote(matrix(NA,ncol=2,nrow=length(ii)))
    rownames(I) <- rep("",nrow(I))
    colnames(I) <- c("Response","Instruments")
    for (i in seq_along(ii)) {
        I[i,] <- c(names(ii)[i],paste(ii[[i]],collapse=","))
    }
    mymsg <- list(l1,I);
    list(estimate=p0,vcov=V0,summary.message=function(...)  {
        mymsg })
}

##' Create/extract 'reverse'-diagonal matrix or off-diagonal elements
##' @title Create/extract 'reverse'-diagonal matrix or off-diagonal elements
##' @aliases revdiag revdiag<- offdiag offdiag<-
##' @usage
##' revdiag(x,...)
##' offdiag(x,type=0,...)
##'
##' revdiag(x,...) <- value
##' offdiag(x,type=0,...) <- value
##' @param x vector
##' @param value For the assignment function the values to put in the diagonal
##' @param type 0: upper and lower triangular, 1: upper triangular, 2: lower triangular, 3: upper triangular + diagonal, 4: lower triangular + diagonal
##' @param \dots additional arguments to lower level functions
##' @author Klaus K. Holst
##' @export
revdiag <- function(x,...) {
    if (NCOL(x)==1) {
      res <- matrix(0,length(x),length(x))
      revdiag(res) <- x
      return(res)
    }
    n <- max(ncol(x),nrow(x))
    x[cbind(rev(seq(n)),seq(n))]
  }

##' @export
"revdiag<-" <- function(x,...,value) {
  n <- max(ncol(x),nrow(x))
  x[cbind(rev(seq(n)),seq(n))] <- value
  x
}


##' @export
offdiag <- function(x,type=0,...) {
    ##if (NCOL(x)==1) return(NULL)
    if (type%in%c(1,3)) {
        ii <- which(upper.tri(x,diag=(type==3)))
    } else if (type%in%c(2,4)) {
        ii <- which(lower.tri(x,diag=(type==4)))
    } else {
        ii <- c(which(lower.tri(x,diag=FALSE)),which(upper.tri(x,diag=FALSE)))
    }
    res <- x[ii]
    class(res) <- c("offdiag",class(res))
    attributes(res) <-
        c(attributes(res),list(type=type,dimension=dim(x),index=ii,nam=dimnames(x)))
    return(res)
  }

##' @export
"offdiag<-" <- function(x,type=0,...,value) {
    if (type%in%c(1,3)) {
        ii <- which(upper.tri(x,diag=(type==3)))
    } else if (type%in%c(2,4)) {
        ii <- which(lower.tri(x,diag=(type==4)))
    } else {
        ii <- c(which(lower.tri(x,diag=FALSE)),which(upper.tri(x,diag=FALSE)))
    }
    x[ii] <- value
    return(x)
}

##' @export
print.offdiag <- function(x,...) {
    ## type <- attr(x,"type")
    nn <- attr(x,"dimension")
    M <- matrix(NA,nn[1],nn[2])
    M[attr(x,"index")] <- x
    dimnames(M) <- attr(x,"nam")
    print(M,na.print="",...)
}

##' @export
`manifest` <-
    function(x,...) UseMethod("manifest")

##' @export
`manifest.lvm` <-
    function(x,...) {
        if (length(vars(x))>0)
            setdiff(vars(x),latent(x))
        else
            NULL
    }

##' @export
`manifest.lvmfit` <-
    function(x,...) {
        manifest(Model(x))
    }

##' @export
manifest.list <- function(x,...) {
    manifestlist <- c()
    for (i in seq_along(x)) {
        manifestlist <- c(manifestlist, manifest(x[[i]]))
    }
    ## endolist <- unique(manifestlist)
    return(manifestlist)
}

##' @export
`manifest.multigroup` <-
    function(x,...) {
        manifest(Model(x))
    }

##' Estimate method for lists
##'
##' @param x list of model objects or `lvm` objects
##' @param ... Additional arguments to lower level functions
##' @return Object of class `estimate.list` (a list of `estimate` objects),
##'   or a fitted multi-group model if input is a list of `lvm` objects.
#' @export
estimate.list <- function(x,...) {
  if (inherits(x[[1]],"lvm")) return(estimate_lvmlist(x,...))
  res <- lapply(x,function(x) estimate(x,...))
  class(res) <- c("estimate.list","list")
  res
}

#' @export
coef.estimate.list <- function(object,...) {
  lapply(object, coef)
}

#' @export
vcov.estimate.list <- function(object,...) {
  lapply(object, vcov)
}

#' @export
IC.estimate.list <- function(x,...) {
  lapply(x, vcov)
}

#' @export
merge.estimate.list <- function(x,...) {
  Reduce(merge, x)
}

##' Estimate probabilities in contingency table
##'
##' @title Estimate probabilities in contingency table
##' @aliases multinomial kappa.multinomial kappa.table gkgamma
##' @param x Formula (or matrix or data.frame with observations, 1 or 2 columns)
##' @param data Optional data.frame
##' @param marginal If TRUE the marginals are estimated
##' @param transform Optional transformation of parameters (e.g., logit)
##' @param vcov Calculate asymptotic variance (default TRUE)
##' @param IC Return ic decomposition (default TRUE)
##' @param ... Additional arguments to lower-level functions
##' @export
##' @examples
##' set.seed(1)
##' breaks <- c(-Inf,-1,0,Inf)
##' m <- lvm(); covariance(m,pairwise=TRUE) <- ~y1+y2+y3+y4
##' d <- transform(sim(m,5e2),
##'               z1=cut(y1,breaks=breaks),
##'               z2=cut(y2,breaks=breaks),
##'               z3=cut(y3,breaks=breaks),
##'               z4=cut(y4,breaks=breaks))
##' 
##' multinomial(d[,5])
##' (a1 <- multinomial(d[,5:6]))
##' (K1 <- kappa(a1)) ## Cohen's kappa
##' 
##' K2 <- kappa(d[,7:8])
##' ## Testing difference K1-K2:
##' estimate(merge(K1,K2,id=TRUE),diff)
##' 
##' estimate(merge(K1,K2,id=FALSE),diff) ## Wrong std.err ignoring dependence
##' sqrt(vcov(K1)+vcov(K2))
##' 
##' ## Average of the two kappas:
##' estimate(merge(K1,K2,id=TRUE),function(x) mean(x))
##' estimate(merge(K1,K2,id=FALSE),function(x) mean(x)) ## Independence
##' ##'
##' ## Goodman-Kruskal's gamma
##' m2 <- lvm(); covariance(m2) <- y1~y2
##' breaks1 <- c(-Inf,-1,0,Inf)
##' breaks2 <- c(-Inf,0,Inf)
##' d2 <- transform(sim(m2,5e2),
##'               z1=cut(y1,breaks=breaks1),
##'               z2=cut(y2,breaks=breaks2))
##' 
##' (g1 <- gkgamma(d2[,3:4]))
##' ## same as
##' \dontrun{
##' gkgamma(table(d2[,3:4]))
##' gkgamma(multinomial(d2[,3:4]))
##' }
##' 
##' ##partial gamma
##' d2$x <- rbinom(nrow(d2),2,0.5)
##' gkgamma(z1~z2|x,data=d2)
##' @author Klaus K. Holst
multinomial <- function(x,data=parent.frame(),marginal=FALSE,transform,vcov=TRUE,IC=TRUE,...) {
    formula <- NULL
    if (inherits(x,"formula")) {
        trm <- terms(x)
        if (length(attr(trm,"term.labels"))>1) {
            x <- update(x,as.formula(paste0(".~ interaction(",
                                           paste0(attr(trm,"term.labels"),collapse=","),")")))
            trm <- terms(x)
            
        }
        formula <- x
        x <- as.matrix(model.frame(trm,data))
        if (ncol(x)>1)
            x <- x[,c(seq(ncol(x)-1)+1,1),drop=FALSE]
    } else {
        trm <- NULL
    }
    if (!vcov) IC <- FALSE
    if (is.table(x) && IC) x <- lava::Expand(x)
    if (NCOL(x)==1) {
        if (!is.table(x)) {
            x <- as.factor(x)
            lev <- levels(x)
            k <- length(lev)
            n <- length(x)
            P <- table(x)/n
        } else {
            n <- sum(x)
            P <- x/n
            lev <- names(x)
            k <- length(lev)
        }
        if (IC) {
            IC <- matrix(0,n,k)
            for (i in seq(k)) {
                IC[,i] <- (1*(x==lev[i])-P[i])/n
            };
            varcov <- crossprod(IC)
        } else {
            IC <- varcov <- NULL
            if (vcov) {
                varcov <- tcrossprod(cbind(P))/n
                diag(varcov) <- P*(1-P)/n
            }
        }
        coefs <- as.vector(P); names(coefs) <- paste0("p",seq(k))
        res <- list(call=match.call(), coef=coefs,P=P,
                    vcov=varcov,IC=IC*NROW(IC),
                    position=seq(k),levels=list(lev),data=x, terms=trm)
        class(res) <- "multinomial"
        return(res)
    }

    if (!is.table(x)) {
        if (NCOL(x)!=2L) stop("Matrix or data.frame with one or two columns expected")
        x <- as.data.frame(x)
        x[,1] <- as.factor(x[,1])
        x[,2] <- as.factor(x[,2])
        lev1 <- levels(x[,1])
        lev2 <- levels(x[,2])
        k1 <- length(lev1)
        k2 <- length(lev2)
        M <- table(x)
        n <- sum(M)
    } else {
        lev1 <- rownames(x)
        lev2 <- colnames(x)
        k1 <- length(lev1)
        k2 <- length(lev2)
        M <- x
        n <- sum(x)
    }
    Pos <- P <- M/n
    if (IC) {
        IC <- matrix(0,n,k1*k2)
        for (j in seq(k2)) {
            for (i in seq(k1)) {
                pos <- (j-1)*k1+i
                IC[,pos] <- (x[,1]==lev1[i])*(x[,2]==lev2[j])-P[i,j]
                Pos[i,j] <- pos
            }
        }; IC <- IC/n
    } else {
        IC <- varcov <- NULL
    }
    
    coefs <- as.vector(P);
    names(coefs) <-  as.vector(outer(seq(k1),seq(k2),function(...) paste0("p",...)))
    position1 <- position2 <- NULL
    if (marginal) {
        p1 <- rowSums(P)
        p2 <- colSums(P)
        names(p1) <- paste0("p",seq(k1),".")
        names(p2) <- paste0("p",".",seq(k2))
        coefs <- c(coefs,p1,p2)
        position1 <- length(P)+seq(k1)
        position2 <- length(P)+k1+seq(k2)
        if (!is.null(IC)) {
            ic1 <- apply(Pos,1,function(x) rowSums(IC[,x]))
            ic2 <- apply(Pos,2,function(x) rowSums(IC[,x]))
            IC <- cbind(IC,ic1,ic2)
            colnames(IC) <- names(coefs)
        }
    }
    if (!missing(transform) && !is.null(IC)) {
        f <- function(p) do.call(transform,list(p))
        D <- diag(numDeriv::grad(f,coefs),ncol=length(coefs))
        coefs <- f(coefs)
        IC <- IC%*%t(D)
    }
    if (vcov && !is.null(IC)) varcov <- crossprod(IC)
    res <- list(call=match.call(),
               formula=formula,
               coef=coefs,P=P,vcov=varcov,IC=IC*NROW(IC), position=Pos,
               call=match.call(), levels=list(lev1,lev2), data=x,
               position1=position1,position2=position2, ## Position of marginals)
               terms=trm
                )
    class(res) <- "multinomial"
    if (length(list(...))>0) {
        res <- structure(estimate(res,...),class=c("multinomial","estimate"))
    }
    return(res)
}

##' @export
model.frame.multinomial <- function(formula,...) {
    formula$data
}

##' @export
IC.multinomial <- function(x,...) {
    x$IC
}

##' @export
coef.multinomial <- function(object,...) {
    object$coef
}

##' @export
vcov.multinomial <- function(object,...) {
    object$vcov
}

##' @export
predict.multinomial <- function(object,newdata,type=c("prob","map"),...) {    
    if (missing(newdata) || is.null(newdata)) newdata <- object$data
    if (!is.null(object$formula) && is.data.frame(newdata)) {
        trm <- terms(object$formula)
        newdata <- model.frame(trm,newdata)[,-1]
    }
    px <- rowSums(object$P)
    idx <- match(trim(as.character(newdata)),trim(rownames(object$P)))
    pcond <- object$P
    for (i in seq(nrow(pcond))) pcond[i,] <- pcond[i,]/px[i]
    pr <- pcond[idx,,drop=FALSE]
    if (tolower(type[1])%in%c("map","class")) {
        pr <- colnames(pr)[apply(pr,1,which.max)]
    }
    return(pr)
}

##' @export
print.multinomial <- function(x, ...) {
    cat("Call: "); print(x$call)
    cat("\nJoint probabilities:\n")
    print(x$P, quote=FALSE)
    if (length(dim(x$P))>1) {
        cat("\nConditional probabilities:\n")
        print(predict(x, newdata=rownames(x$P)), quote=FALSE)
    }
    cat("\n")
    print(estimate(NULL, coef=coef(x), vcov=vcov(x)))
    ## stderr <- diag(vcov(x))^.5
    ## StdErr <- x$position
    ## StdErr[] <- stderr[StdErr]
    ## cat("\nStd.Err:\n")
    ## print(StdErr,quote=FALSE)
    ## cat("\nPosition:\n")
    ## print(x$position,quote=FALSE)
}

##' @export
"randomslope<-" <- function(x,...,value) UseMethod("randomslope<-")

##' @export
"randomslope<-.lvm" <- function(x, ..., value) {
  randomslope(x, covar=value, ...)
}

##' @export
`randomslope` <-
function(x,...) UseMethod("randomslope")

##' @export
`randomslope.lvm` <-
function(x,covar,random=NULL,response=NULL,param,postfix,clear=FALSE,zero=TRUE,...) {
  if (missing(covar)) {
    rsidx <- unlist(x$attributes$randomslope)
    if (length(rsidx)>0)
      return(names(rsidx)[rsidx])
    else
      return(NULL)
  }
  if (inherits(covar,"formula")) {
    covar <- all.vars(covar)
  }
  if (clear) {
    ##    x <- addattr(x,attr="shape",var=var,val="rectangle")
    x$attributes$randomslope[covar] <- FALSE
  } else {
    if (!is.null(random) & !is.null(response)) {
      if (inherits(random,"formula")) {
        random <- all.vars(random)
      }
      if (inherits(response,"formula")) {
        response <- all.vars(response)
      }
      if (length(covar)!=length(response)) stop("Vectors should be of the same length!")
      if (!(random%in%latent(x))) {
        addvar(x) <- random
        latent(x) <- random
      }
      if (missing(param) || !is.null(param)) {
        if (!missing(postfix))
          newlatent <-  paste0(random,postfix)
        else
          newlatent <-  paste(random,covar,sep=".")
        covariance(x,random) <- 1
        for (i in seq_along(covar)) {
          if (missing(param)) {
            x <- regression(x,to=newlatent[i],from=random)
          } else {
            if (inherits(param,"formula")) {
              param <- all.vars(param)
            }
            if (length(param)!=length(newlatent))
              param <- rep(param,length(newlatent))
            regfix(x,to=newlatent[i], from=random) <- param[i]
          }
          regfix(x,to=response[i],from=newlatent[i]) <- covar[i]
          latent(x) <- newlatent[i]
          covariance(x,newlatent[i]) <- 0
        }
      } else {
        for (i in seq_along(covar)) {
          regfix(x,to=response[i],from=random) <- covar[i]
        }
      }
    } else {
      x$attributes$randomslope[covar] <- TRUE
    }
  }
  index(x) <- reindex(x)
  return(x)
}

##' @export
`randomslope.lvmfit` <-
  function(x,...) {
    randomslope(Model(x),...)
  }

`addattr` <- function(x,...) UseMethod("addattr")

#' @exportS3Method
`addattr.lvmfit` <- function(x,...) addattr(Model(x),...)

#' @exportS3Method
`addattr.lvm` <- function(x, attr, var=NULL, val=TRUE, fun=graph::nodeRenderInfo,debug=FALSE,...) {
    if (!is.null(var)) {
        Graph(x) <- addattr(Graph(x), attr=attr, var=var, val=val, fun=fun, debug=debug)
        return(x)
    } else {
        addattr(Graph(x), attr=attr, var=var, val=val, fun=fun)
    }
}

#' @exportS3Method
`addattr.graphNEL` <- function(x, attr, var=NULL, val=TRUE,fun="graph::nodeRenderInfo",debug=FALSE,...) {
    if (is.null(var)) {
        ff <- strsplit(fun,"::")[[1]]
        if (length(ff)>1) {
            ff <- getFromNamespace(ff[2],ff[1])
        }
        f <- do.call(ff,list(x))
        if (is.null(val) || !is.logical(f[[attr]]))
            attrvar <- f[[attr]]
        else
            attrvar <- names(f[[attr]])[which(val==f[[attr]])]
        return(attrvar)
    }
    if (is.character(val))
        myexpr <- paste0("list(",attr,"=c(", paste0("\"",var,"\"=\"",val,"\"" , collapse=", "), "))")
    else
        myexpr <- paste0("list(",attr,"=c(", paste0("\"",var,"\"=",val, collapse=", "), "))")
    Debug(list("str=",myexpr),debug)
    eval(parse(text=paste0(fun,"(x) <- ",myexpr)))
    return(x)
}

## Cohen's kappa

##' @export
kappa.multinomial <- function(z,all=FALSE,...) {
    pp <- length(coef(z))
    if ((length(z$levels)!=2) || !(identical(z$levels[[1]],z$levels[[2]])))
        stop("Expected square table and same factor levels in rows and columns")
    k <- length(z$levels[[1]])
    zeros <- rbind(rep(0,pp))
    A0 <- zeros; A0[diag(z$position)] <- 1
    A <- matrix(0,ncol=pp,nrow=2*k)
    for (i in seq(k)) A[i,z$position[i,]] <- 1
    for (i in seq(k)) A[i+k,z$position[,i]] <- 1
    b <- estimate(z,function(p) as.vector(rbind(A0,A)%*%p),IC=TRUE)
    b2 <- estimate(b,function(p) c(p[1],sum(p[seq(k)+1]*p[seq(k)+k+1])),IC=TRUE)
    if (!all) {
        return(estimate(b2,function(p) list(kappa=(p[1]-p[2])/(1-p[2])),IC=TRUE,...))
    }
    estimate(b2,function(p) list(kappa=(p[1]-p[2])/(1-p[2]),agree=p[1], independence=p[2]),IC=TRUE,...)
}

##' @export
kappa.table <- function(z,...) {
    kappa(multinomial(Expand(z)),...)
}

##' @export
kappa.data.frame <- function(z,...) {
    kappa(multinomial(z),...)
}

##' Plot regression line (with interactions) and partial residuals.
##'
##' @title Plot regression lines
##' @param model Model object (e.g. [lm()])
##' @param var1 predictor (Continuous or factor)
##' @param var2 Factor that interacts with `var1`
##' @param data data.frame to use for prediction (model.frame is used as default)
##' @param ci.lty Line type for confidence limits
##' @param ci Boolean indicating wether to draw pointwise 95% confidence limits
##' @param level Level of confidence limits (default 95%)
##' @param pch Point type for partial residuals
##' @param lty Line type for estimated regression lines
##' @param lwd Line width for regression lines
##' @param npoints Number of points used to plot curves
##' @param xlim Range of x axis
##' @param col Color (for each level in `var2`)
##' @param colpt Color of partial residual points
##' @param alpha Alpha level
##' @param cex Point size
##' @param delta For categorical `var1`
##' @param centermark For categorical `var1`
##' @param jitter For categorical `var1`
##' @param cidiff For categorical `var1`
##' @param mean For categorical `var1`
##' @param legend Boolean (add legend)
##' @param trans Transform estimates (e.g. exponential)
##' @param partres Boolean indicating whether to plot partial residuals
##' @param partse .
##' @param labels Optional labels of `var2`
##' @param vcov Optional variance estimates
##' @param predictfun Optional predict-function used to calculate confidence limits and predictions
##' @param plot If FALSE return only predictions and confidence bands
##' @param new If FALSE add to current plot
##' @param \dots additional arguments to lower level functions
##' @return list with following members:
##' * `x` Variable on the x-axis `var1`
##' * `y` Variable on the y-axis (partial residuals)
##' * `predict` Matrix with confidence limits and predicted startvalues.R
##' @author Klaus K. Holst
##' @seealso [termplot()]
##' @export
##' @examples
##' n <- 100
##' x0 <- rnorm(n)
##' x1 <- seq(-3,3, length.out=n)
##' x2 <- factor(rep(c(1,2),each=n/2), labels=c("A","B"))
##' y <- 5 + 2*x0 + 0.5*x1 + -1*(x2=="B")*x1 + 0.5*(x2=="B") + rnorm(n, sd=0.25)
##' dd <- data.frame(y=y, x1=x1, x2=x2)
##' lm0 <- lm(y ~ x0 + x1*x2, dd)
##' plotConf(lm0, var1="x1", var2="x2")
##' abline(a=5,b=0.5,col="red")
##' abline(a=5.5,b=-0.5,col="red")
##' ### points(5+0.5*x1 -1*(x2=="B")*x1 + 0.5*(x2=="B") ~ x1, cex=2)
##'
##' data(iris)
##' l <- lm(Sepal.Length ~ Sepal.Width*Species,iris)
##' plotConf(l,var2="Species")
##' plotConf(l,var1="Sepal.Width",var2="Species")
##' 
##' \dontrun{
##' ## lme4 model
##' dd$Id <- rbinom(n, size = 3, prob = 0.3)
##' lmer0 <- lme4::lmer(y ~ x0 + x1*x2 + (1|Id), dd)
##' plotConf(lmer0, var1="x1", var2="x2")
##' }
##' @keywords hplot regression
plotConf <- function(model,
                     var1=NULL,
                     var2=NULL,
                     data=NULL,
                     ci.lty=0,
                     ci=TRUE, level=0.95,
                     pch=16, lty=1, lwd=2,
                     npoints=100,
                     xlim,
                     col=NULL,
                     colpt,
                     alpha=0.5,
                     cex=1,
                     delta=0.07,
                     centermark=0.03,
                     jitter=0.2,
                     cidiff=FALSE,
                     mean=TRUE,
                     legend=ifelse(is.null(var1),FALSE,"topright"),
                     trans=function(x) {x},
                     partres=inherits(model,"lm"),
                     partse=FALSE,
                     labels,
                     vcov,
                     predictfun,
                     plot=TRUE,
                     new=TRUE,
                     ...) {


    if (inherits(model,"formula")) model <- lm(model,data=data,...)
    if (inherits(model,"lmerMod")) {
        intercept <- lme4::fixef(model)["(Intercept)"]
    } else {
        intercept <- coef(model)["(Intercept)"]
    }
    if (is.na(intercept)) intercept <- 0

    if (is.null(data)) {
        curdata <- get_all_vars(model,data=model.frame(model))
    } else {
        curdata <- get_all_vars(formula(model), data=data)
    }
    
    if (inherits(model,"lmerMod")) {
      curdata0 <- model.frame(model, data = data, fixed.only = FALSE)
    } else {
      curdata0 <- model.frame(model,data) ## Checking for factors
    }
    
    
    if (is.null(var1) && is.null(var2)) {
        var1 <- colnames(curdata)[2]
        var10 <- colnames(curdata0)[2]
    } else var10 <- var1
        
    responseorig <- colnames(curdata)[1]
    if (inherits(curdata0[,var10],c("character","factor"))) {
        curdata <- curdata0
        var2 <- var10; var1 <- NULL
    }
    dots <- list(...)


    response <- all.vars(formula(model))[1]
    cname <- colnames(curdata)[-1]
    if (!is.factor(curdata[,var2]) & !is.null(var2)) {
        curdata[,var2] <- as.factor(curdata[,var2])
        colnames(curdata)[1] <- response
        model <- update(model,as.formula(paste(response,"~.")),data=curdata)
    }
    thelevels <- levels(curdata[,var2])
    if (missing(labels)) labels <- thelevels

    k <- ifelse(is.null(var2),1,length(thelevels))
    if (is.null(col)) {
        col <- c("black","darkblue","darkred","goldenrod","mediumpurple",
                 "seagreen","aquamarine3","violetred1","salmon1",
                 "lightgoldenrod1","darkorange2","firebrick1","violetred1", "gold")
    }

    if (missing(xlim)) {
        if (!is.null(var1))
            xlim <- range(curdata[,var1])
        else
            xlim <- c(0,length(thelevels))+0.5
    }
    dots$xlim <- xlim

    if (is.null(var1) & !is.null(var2)) {
        ##npoints <- 1
        x <- unique(curdata[,var2])
        npoints <- 1#length(x)
    } else {
        x <- seq(xlim[1], xlim[2], length.out=npoints)
    }
    xx <- c()

    newdata <- data.frame(id=seq(npoints))
    partdata <- curdata
    var1.idx <- 0
    ii <- 1
    for (nn in cname) {
        ii <- ii+1
        v <- curdata[,nn]
        if (!is.null(var1) && nn==var1) {
            var1.idx <- ii
            newdata <- cbind(newdata, rep(x, k))
            partdata[,nn] <- 0
        } else {
            if (is.factor(v)) {
                if (nn%in%var2) {
                    newdata <- cbind(newdata, factor(rep(levels(v), each=npoints),levels=thelevels))
                    partdata[,nn] <- factor(rep(levels(v)[1], nrow(partdata)),levels=levels(v))
                } else {
                    newdata <- cbind(newdata, factor(rep(levels(v)[1], k*npoints),
                                                     levels=levels(v)))
                }
            } else {
                if (is.logical(v))
                    newdata <- cbind(newdata,FALSE)
                else
                    newdata <- cbind(newdata,0)
            }
        }
    };

    colnames(newdata) <- c("_id", cname)

    partdata[,response] <- newdata[,response] <- 0
    atr <- c("terms")
    attributes(newdata)[atr] <- attributes(curdata)[atr]
    attributes(partdata)[atr] <- attributes(partdata)[atr]
    Y <- model.frame(model)[,1]
    if(inherits(Y,"Surv")) Y <- Y[,1]
    XX <- model.matrix(formula(terms(model)),data=newdata)
    if (inherits(model,"lmerMod")) {
        bb <- lme4::fixef(model)
    } else {
        bb <- coef(model)
    }
    if (!missing(vcov)) SS <- vcov else {
        if (inherits(model,"geeglm")) {
            SS <- (summary(model)$cov.unscaled)
        } else {
            SS <- as.matrix(stats::vcov(model))
        }
    }
    bidx <- which(apply(XX,2,function(x) !all(x==0)))
    notbidx <- setdiff(seq(length(bb)),bidx)
    bb0 <- bb; bb0[notbidx] <- 0
    myse <- apply(XX[,bidx,drop=FALSE],1,function(x) rbind(x)%*%SS[bidx,bidx,drop=FALSE]%*%cbind(x))^.5
    ci.all <- list(fit=XX%*%bb0,se.fit=myse)
    z <- qnorm(1-(1-level)/2)
    ci.all$fit <- cbind(ci.all$fit,ci.all$fit-z*ci.all$se.fit,ci.all$fit+z*ci.all$se.fit)

    if (!missing(predictfun)) {
        R <- Y-predict(model, newdata=partdata)
        ci.all <- predict(model, newdata=newdata, se.fit=TRUE, interval = "confidence", level=level,...)
    } else {
        XX0 <- model.matrix(formula(terms(model)),data=partdata)
        R <- Y-XX0%*%bb
    }
    if (inherits(model,"lmerMod")) {
        uz <- as.matrix(unlist(lme4::ranef(model))%*%do.call(rbind,lme4::getME(model,"Ztlist")))[1,]
        R <- R-uz
    }

    pr <- trans(intercept + R)
    if (is.na(intercept)) {
        intercept <- 0
        if (!is.null(var2)) {
            intercept <- coef(model)[paste0(var2,thelevels)][as.numeric(curdata[,var2])]
        }
    }

    if (is.null(dots$ylim)) {
        if (partres) {
            if (cidiff)
                dots$ylim <- range(pr)
            else
                dots$ylim <- range(trans(c(ci.all$fit)),pr)
        }
        else
            dots$ylim <- trans(range(ci.all$fit))
    }
    if (is.null(dots$ylab))
        dots$ylab <- responseorig

    if (is.null(var1)) {
        dots$axes=FALSE
        if (is.null(dots$xlab))
            dots$xlab <- ""
    } else  {
        if (is.null(dots$xlab))
            dots$xlab <- var1
    }

    if (!plot) return(list(x=x, y=pr, predict=ci.all, predict.newdata=newdata))

    plot.list <- c(x=0,y=0,type="n",dots)
    if (new) {
        do.call(graphics::plot, plot.list)
        if (is.null(var1)) {
            box()
            axis(2)
            axis(1,at=seq(length(thelevels)),labels)
        }
    }

    col.trans <- Col(col,alpha)

    Wrap <- function(k,n) { (seq_len(k)-1)%%n +1 }
    col.i <- Wrap(k,length(col));  col.k <- col[col.i];
    lty.k <- lty[Wrap(k,length(lty))]
    pch.k <- pch[Wrap(k,length(pch))]

    if (!is.null(var1)) {
        for (i in seq_len(k)) {
            ci0 <- trans(ci.all$fit[(npoints*(i-1)+1):(i*npoints),])
            y <- ci0[,1]; yu <- ci0[,3]; yl <- ci0[,2]
            lines(y ~ x, col=col.k[i], lwd=lwd, lty=lty.k[i])

            if (ci) {
                lines(yl ~ x, lwd=1, col=col.k[i], lty=ci.lty)
                lines(yu ~ x, lwd=1, col=col.k[i], lty=ci.lty)
                xx <- c(x, rev(x))
                yy <- c(yl, rev(yu))
                polygon(xx,yy, col=col.trans[col.i[i]], lty=0)
            }

        }
    }
    ii <- as.numeric(curdata[,var2])

    if (is.null(var1)) {
        xx <- curdata[,var2]
        x <- jitter(as.numeric(xx),jitter)
        if (missing(colpt)) colpt <- Col(col[1],alpha)
        if (partres>0)
            points(pr ~ x,pch=pch[1], col=colpt[1], cex=cex, ...)
        mycoef <- bb[paste0(var2,thelevels)][-1]
        if (inherits(model,c("lm","glm")))
            myconf <- confint(model)[paste0(var2,thelevels)[-1],,drop=FALSE]
        else {
            myconf <- matrix(mycoef,ncol=2,nrow=length(mycoef))
            myconf <- myconf + qnorm(0.975)*cbind((diag(as.matrix(SS))[-1])^0.5)%x%cbind(-1,1)
        }
        for (pos in seq(k)) {
            if (cidiff) {
                if (pos>1) {
                    ci0 <- trans(intercept+myconf[pos-1,])
                    yl <- ci0[1]; yu <- ci0[2]; y <- trans(intercept+mycoef[pos-1])
                } else {
                    yu <- yl <- NULL; y <- trans(intercept)
                }
            } else if (partse) {
                y0 <- pr[xx==levels(xx)[pos]]
                ci0 <- confint(lm(y0~1))
                yl <- ci0[1]; yu <- ci0[2]; y <- trans(mean(y0))
            } else {
                ci0 <- rbind(trans(ci.all$fit[(npoints*(pos-1)+1):(pos*npoints),]))
                y <- ci0[,1]; yu <- ci0[,3]; yl <- ci0[,2]
            }
            if (!mean) y <- NULL
            confband(pos,yl,yu,delta=delta,center=y,centermark=centermark,col=col[1],lwd=lwd[1],lty=lty[1],cex=cex)
        }
    } else {
        if (partres) {
            xx <- curdata[,var1]
            if (!missing(colpt)) {
                points(pr ~ xx, col=colpt, cex=cex, pch=pch[1],...)
            } else {
                if (!is.null(var2))
                    points(pr ~ xx, col=col.k[ii], pch=pch.k[ii], cex=cex, ...)
                else
                    points(pr ~ xx, col=col[1], pch=pch[1], cex=cex,...)
            }
        }
    }

    if (k>1 && legend!=FALSE) {
        if (length(lty)>1)
            legend(legend, legend=thelevels, col=col.k, pch=pch.k, bg="white", lty=lty.k,cex=cex)
        else
            legend(legend, legend=thelevels, col=col.k, pch=pch.k, bg="white",cex=cex)
    }

    invisible(list(x=xx, y=pr, predict=ci.all, predict.newdata=newdata))
}

##' Backdoor criterion
##'
##' Check backdoor criterion of a lvm object
##' @param object lvm object
##' @param f formula. Conditioning, z, set can be given as y~x|z
##' @param cond Vector of variables to conditon on
##' @param ... Additional arguments to lower level functions
##' @param return.graph Return moral ancestral graph with z and effects from x removed
##' @examples
##' m <- lvm(y~c2,c2~c1,x~c1,m1~x,y~m1, v1~c3, x~c3,v1~y,
##'          x~z1, z2~z1, z2~z3, y~z3+z2+g1+g2+g3)
##' ll <- backdoor(m, y~x)
##' backdoor(m, y~x|c1+z1+g1)
##' @export
backdoor <- function(object, f, cond, ..., return.graph=FALSE) {
    y <- getoutcome(f, sep = "|")
    x <- attr(y, "x")
    if (length(x) > 1) {
        cond <- all.vars(x[[2]])
    }
    x <- all.vars(x[[1]])
    nod <- vars(object)
    des <- descendants(object, x)
    ch <- children(object, x)
    g0 <- cancel(object, toformula(x, ch))

    if (!base::missing(cond)) {
        val <- dsep(g0, c(y, x), cond = cond) && !any(cond %in% des)
        if (return.graph) {
            res <- dsep(g0, c(y, x), cond = cond, return.graph=TRUE)
            attr(res,"result") <- val
            return(res)
        }
        return(val)
    }
    cset <- base::setdiff(nod, c(des, x, y)) ## possible conditioning set
    pp <- path(g0,from=x,to=y,all=TRUE) ## All backdoor paths
    M <- adjMat(g0)
    Collider <- function(vec) {
        M[vec[2],vec[1]] & M[vec[2],vec[3]]
    }
    blockList <- collideList <- c()
    for (i in seq_along(pp)) {
        p0 <- pp[[i]]
        blocks <- c()
        collide <- c()
        for (j in seq(length(p0)-2)) {
            if (Collider(p0[0:2 + j])) {
                collide <- c(collide,p0[1+j])
            } else {
                blocks <- c(blocks,p0[1+j])
            }
        }
        blockList <- c(blockList,list(blocks))
        collideList <- c(collideList,list(collide))
    }
    res <- list(blockList)
    ## Paths with colliders:
    col <- unlist(lapply(collideList,function(x) !is.null(x)))
    if (length(col)>0) col <- which(col)
    ## List of variables which are not on path between x and y:
    optional <- setdiff(cset,c(unlist(collideList),unlist(blockList)))
    callrecurs <- function(col,res=list()) {
        if (length(col)==0) return(res)
        blockList0 <- blockList
        blockList0[col] <- NULL
        blockList0 <- lapply(blockList0, function(x) setdiff(x,unlist(collideList[col])))
        if (!any(unlist(lapply(blockList0,is.null)))) {
            res <- c(res, list(blockList0))
        }
        for (i in seq_along(col)) {
            col0 <- col[-i]
            if (length(col0)>0)
                res <- callrecurs(col0,res)
        }
        return(res)
    }
    if (length(col)>0)
        res <- c(res,callrecurs(col))
    ## Any element can be included from 'optional' For a given element
    ## in 'include' at least one element in each member of the list
    ## must be included in the conditioning set.
    return(list(optional=optional, include=res))
}

##' For matrices a block-diagonal matrix is created. For all other
##' data types he operator is a wrapper of \code{paste}.
##'
##' Concatenation operator
##' @aliases %++%
##' @rdname op_concat
##' @usage x \%++\% y
##' @title Concatenation operator
##' @param x First object
##' @param y Second object of same class
##' @author Klaus K. Holst
##' @keywords utilities misc
##' @seealso \code{blockdiag}, \code{\link{paste}}, \code{\link{cat}},
##' @examples
##' ## Block diagonal
##' matrix(rnorm(25),5)%++%matrix(rnorm(25),5)
##' ## String concatenation
##' "Hello "%++%" World"
##' ## Function composition
##' f <- log %++% exp
##' f(2)
##' @export
`%++%` <- function(x,y) UseMethod("%++%",y)

## ##' @export
## `%+%` <- function(x,y) UseMethod("%+%",y)

##' @export
`%++%.default` <- function(x,y) paste0(x,y)

##' @export
`%++%.character` <- function(x,y) paste0(x,y)

##' @export
`%++%.matrix` <- function(x,y) blockdiag(x,y)

##' @export
`%++%.function` <- function(x,y) function(...) x(y(...))


notin <- Negate(get("%in%"))
##' Matching operator (x not in y) oposed to the \code{\%in\%}-operator (x in y)
##'
##' Matching operator
##' @rdname op_match
##' @aliases %ni% %in.open% %in.closed%
##' @usage x \%ni\% y
##' @param x vector
##' @param y vector of same type as \code{x}
##' @return A logical vector.
##' @author Klaus K. Holst
##' @seealso \code{\link{match}}
##' @keywords utilities misc
##' @examples
##'
##' 1:10 %ni% c(1,5,10)
##'
##' @export
"%ni%" <- function(x,y) notin(x,y)

## function(x,y) {
##   is.na(match(x,y))
## }

##' @export
"%in.open%" <- function(x, y) {
  if (length(y) == 1) y <- c(y, y)
  if (length(y) != 2 || !is.numeric(y)) stop("rhs should be a range (numeric vector of length 2)")
  x > y[1] & x < y[2]
}

##' @export
"%in.closed%" <- function(x,y) {
  if (length(y) == 1) y <- c(y, y)
  if (length(y) != 2 || !is.numeric(y)) stop("rhs should be a range (numeric vector of length 2)")
  x >= y[1] & x <= y[2]
}

glm_estimate_hook <- function(x,estimator,...) {
    yy <- c()
    if (length(estimator)>0 && estimator=="glm") {
        for (y in endogenous(x)) {
            fam <- attributes(distribution(x)[[y]])$family
            if (is.null(fam)) fam <- stats::gaussian()
            if (!(tolower(fam$family)%in%
                  c("gaussian","gamma","inverse.gaussian","weibull"))) {
                yy <- c(yy,y)
            }
        }
        if (length(yy)>0) covariance(x,yy) <- 1
    }
    return(c(list(x=x,estimator=estimator,...)))
}

GLMest <- function(m,data,control=list(),...) {
    yvar <- endogenous(m)
    res <- c()
    count <- 0
    V <- NULL
    mymsg <- c()
    ics <- c()
    breads <- c()

    et <- eventTime(m)
    yvar.et <- rep(NA, length(yvar))
    names(yvar.et) <- yvar
    if (!is.null(et)) {
        for (i in seq_along(et)) {
            ## if (!survival::is.Surv(data[,et[[i]]$names[1]]))
            ##         data[,et[[i]]$names[1]] <- with(et[[i]],
            ##         survival::Surv(data[,names[1]],data[,names[2]]))
            yvar <- setdiff(yvar,c(et[[i]]$latentTimes[-1],et[[i]]$names))
            yvar.et[et[[i]]$latentTimes[1]] <- et[[i]]$names[1]
        }
    }

    for (y in yvar) {
        count <- count+1
        xx <- parents(m, y)
        fam <- attributes(distribution(m)[[y]])$family
        if (is.null(fam)) fam <- stats::gaussian()
        if (!is.null(fam$link)) {
            mymsg <- c(mymsg, with(fam, paste0(family,"(",link,")")))
        } else {
            mymsg <- c(mymsg, with(fam, paste0(family)))
        }
        if (length(xx)==0) xx <- 1
        nn0 <- paste(y,xx,sep=lava.options()$symbol[1])
        y0 <- y

        f <- as.formula(paste0(y, "~",
          paste(xx, collapse = "+")
        ))
        isSurv <- inherits(data[1, y], "Surv")
        if (isSurv) {
          g <- survival::survreg(f,data=data,dist=fam$family,...)
        } else {
          g <- glm(f,family=fam,data=data,...)
        }
        p <- pars(g)
        ii <- IC(g)
        V0 <- attr(ii, "bread")
        ics <- cbind(ics,ii)
        y <- y0
        names(p)[1] <- y
        if (length(p)>1) {
          xx0 <- setdiff(xx, 1)
          if (length(xx0) > 0) {
            nn0 <- paste(y, xx0, sep = lava.options()$symbol[1])
            names(p)[seq_along(nn0) + 1] <- nn0
          }
          if (length(p)>length(xx0)+1) names(p)[length(p)] <- paste(y,y,sep=lava.options()$symbol[2])
        }
        if (tolower(fam$family)%in%c("gaussian","gamma","inverse.gaussian") && !isSurv) {
            ics <- cbind(ics,0)
            null <- matrix(0); dimnames(null) <- list("scale","scale")
            V0 <- blockdiag(V0,null,pad=0)
        }
        breads <- c(breads, list(V0))
        res <- c(res, list(p));
    }
    coefs <- unlist(res)
    idx <- na.omit(match(coef(m),names(coefs)))
    coefs <- coefs[idx]
    V <- var_ic(ics[, idx])
    mymsg <- noquote(cbind(mymsg))
    colnames(mymsg) <- "Family(Link)"; rownames(mymsg) <- paste(yvar,":")
    list(estimate=coefs, vcov=V, breads=breads, IC=ics[,idx],
         summary.message=function(...)  { mymsg }, dispname="Dispersion:")
}

GLMscore <- function(x,p,data,indiv=TRUE,logLik=FALSE,...) {
  yvar <- endogenous(x)
  S <- pnames <- c()
  count <- 0
  breads <- c()
  L <- 0
  for (y in yvar) {
    count <- count+1
    xx <- parents(x,y)
    pname <- c(y,paste0(y,sep=lava.options()$symbol[1],xx),paste(y,y,sep=lava.options()$symbol[2]))
    pidx <- na.omit(match(pname,coef(x)))
    fam <- attributes(distribution(x)[[y]])$family
    if (is.null(fam)) fam <- stats::gaussian()
    if (length(xx)==0) xx <- 1
    f <- as.formula(paste0(y,"~",paste(xx,collapse="+")))
    isSurv <- inherits(data[1,y],"Surv")
    if (inherits(data[,y],"Surv")) {
      g <- survival::survreg(f,data=data,dist=fam$family)
    } else {
      g <- glm(f,family=fam,data=data)
    }
    p0 <- p[pidx]
    if (!isSurv) L0 <- logL.glm(g,p=p0,indiv=TRUE,...)
    if (tolower(fam$family)%in%c("gaussian","gamma","inverse.gaussian") && !isSurv) {
      p0 <- p0[-length(p0)]
      S0 <- score(g,p=p0,indiv=TRUE,pearson=TRUE,...)
      V0 <- attr(S0,"bread")
      r <- attr(S0,"pearson")
      ## dispersion <- mean(r^2)
      S0 <- cbind(S0,scale=0)
      null <- matrix(0); dimnames(null) <- list("scale","scale")
      V0 <- blockdiag(V0,null,pad=0)
    } else {
      S0 <- score(g,p=p0,indiv=TRUE,...)
      if (isSurv) L0 <- attr(S0,"logLik")
      V0 <- attr(S0,"bread")
    }
    L <- L+sum(L0)
    breads <- c(breads,list(V0))
    S <- c(S,list(S0))
    pnames <- c(pnames, list(pname));
  }
  coefs <- unlist(pnames)
  idx <- na.omit(match(coefs,coef(x)))
  idx <- order(idx)
  V <- Reduce(blockdiag,breads)[idx,idx]
  S1 <- Reduce(cbind,S)[,idx,drop=FALSE]
  colnames(S1) <- coef(x)
  attributes(S1)$bread <- V
  attributes(S1)$logLik <- structure(L,nobs=nrow(data),nall=nrow(data),df=length(p),class="logLik")
  if (!indiv) S1 <- colSums(S1)
  return(S1)
}

##' @export
score.lm <- function(x, p=coef(x), data, indiv=FALSE,
              y, X, offset=NULL, weights=NULL, dispersion=TRUE, ...) {
    if (missing(data)) {
        X <- model.matrix(x)
        y <- model.frame(x)[,1]
    } else {
        X <- model.matrix(formula(x),data=data)
        y <- model.frame(formula(x),data=data)[,1]
    }
    if(any(is.na(p))) warning("Over-parameterized model")
    Xbeta <- X%*%p
    if (is.null(offset)) offset <- x$offset
    if (!is.null(offset)) Xbeta <- Xbeta+offset
    r <- y-Xbeta
    if (is.null(weights)) weights <- x$weights
    if (!is.null(weights)) {
        sigma2 <- sum(r^2*weights)/(length(r)-length(p))
        r <- r*weights
    } else {
        sigma2 <- sum(r^2)/(length(r)-length(p))
    }
    if (!dispersion) sigma2 <- 1
    ##sigma2 <- suppressWarnings(summary(x)$sigma^2)
    A <- as.vector(r)/sigma2
    S <- apply(X,2,function(x) x*A)
    if (!indiv) return(colSums(S))
    suppressWarnings(attributes(S)$bread <- vcov(x)*NROW(S))
    return(S)
}

##' Predict from a GLM with modified coefficients
##'
##' Compute predictions from a fitted [stats::glm] object, optionally
##' substituting new parameter values. Unlike [stats::predict.glm], this
##' function allows the user to supply an arbitrary coefficient vector `p`,
##' which is useful for computing predictions at counterfactual parameter
##' values (e.g., during optimization or simulation). The returned value
##' also includes a `"grad"` attribute containing the Jacobian of predictions
##' with respect to the coefficients.
##'
##' @param object A fitted `glm` object.
##' @param p Numeric vector of coefficients (defaults to `coef(x)`).
##' @param data Optional data frame for computing the model matrix and
##'   response. If missing, the original model matrix from the fitted
##'   object is used.
##' @param offset Optional offset vector. If `NULL` (default), the
##'   offset stored in the fitted model (`x$offset`) is used.
##' @param type Character; `"response"` (default) returns predictions on
##'   the response scale via the inverse link function, `"link"` returns
##'   predictions on the linear predictor scale.
##' @param ... Additional arguments (currently unused).
##' @return Numeric vector of predictions with a `"grad"` attribute
##'   containing the gradient (Jacobian) of predictions with respect to
##'   the coefficients. When `type = "link"`, the gradient is simply the
##'   model matrix `X`.
##' @seealso [stats::predict.glm], [score.glm]
##' @examples
##' m <- glm(mpg ~ hp + wt, data = mtcars)
##' p0 <- coef(m)
##' # Predictions at fitted coefficients match predict.glm
##' all.equal(as.numeric(predict_glm(m)), fitted(m))
##' # Predictions with modified coefficients
##' predict_glm(m, p = p0 * 1.1)
##' @export
predict_glm <- function(object, p=coef(object), data, offset = NULL,
                        type=c("response", "link"), ...) {
  x <- object
  if (!inherits(x, "glm")) stop("need glm object")


  ## Getting the model matrix
  if (missing(data)) {
    X <- stats::model.matrix.lm(x)
  } else {
    X <- stats::model.matrix.lm(formula(x),
                         data = data,
                         na.action = na.pass)
  }

  ## Setting NA parameters to zero
  if(any(is.na(p))) {
    warning("Over-parameterized model (setting NA's to zero)")
    p[is.na(p)] <- 0
  }

  ## Calculating the prediction on the link scale
  Xbeta <- X%*%p

  ## Handling offset
  if (is.null(offset)) {
    offset <- x$offset
  }
  if (!is.null(offset)) {
    Xbeta <- Xbeta + offset
  }

  ## If type == "link"
  if (tolower(type[1]) == "link") {
    return(structure(Xbeta, grad = X))
  }

  ## If type == "response"
  link <- family(x)
  ginv <- link$linkinv
  dginv <- link$mu.eta
  pr <- ginv(Xbeta)
  z <- dginv(Xbeta)
  gr <- apply(X, 2, function(x) x*z)
  return(structure(pr, grad=gr))
}

##' @export
score.glm <- function(x, p=coef(x), data, indiv=FALSE,
                      pearson=FALSE, y, X,
                      link, dispersion, offset=NULL, weights=NULL, ...) {
    if (inherits(x,"glm")) {
        link <- family(x)
        if (missing(data)) {
            X <- model.matrix(x)
            y <- model.frame(x)[,1]
        } else {
            X <- model.matrix(formula(x),data=data)
            y <- model.frame(formula(x),data=data)[,1]
        }
        if (is.null(offset)) offset <- x$offset
    } else {
        if (missing(link)) stop("Family needed")
        if (missing(data)) stop("data needed")
        X <- model.matrix(formula(x),data=data)
        y <- model.frame(formula(x),data=data)[,1]
    }
    if (is.character(y) || is.factor(y)) {
        y <- as.numeric(as.factor(y)) - 1
    }
    ## g <- link$linkfun
    ginv <- link$linkinv
    dginv <- link$mu.eta ## D[linkinv]
    ##dg <- function(x) 1/dginv(g(x)) ## Dh^-1 = 1/(h'(h^-1(x)))
    if (inherits(x, "negbin")) {
        Dcanlink <- function(x) 1/x
    } else {
        canonf <- do.call(link$family,list())
        ## caninvlink <- canonf$linkinv
        canlink <- canonf$linkfun
        Dcaninvlink <- canonf$mu.eta
        Dcanlink <- function(x) 1 / Dcaninvlink(canlink(x))
    }
    ##gmu <- function(x) g(caninvlink(x))
    ##invgmu <- function(z) canlink(ginv(z))
    h <- function(z) Dcanlink(ginv(z))*dginv(z)
    pna <- any(is.na(p))
    if(pna) {
      warning("Over-parameterized model (NA parameters). Ignoring NA parameters")
      idx <- is.na(p)
      ## setting paramter and model.matrix to zero for NA parameters
      p[idx] <- 0
      X[, idx] <- 0
    }
    Xbeta <- X%*%p
    if (!is.null(offset)) Xbeta <- Xbeta+offset
    pi <- ginv(Xbeta)
    r <- y-pi
    if (!is.null(x$prior.weights) || !is.null(weights)) {
        if (is.null(weights)) weights <- x$prior.weights
    } else {
        weights <- !is.na(r)
    }
    a.phi <- 1
    r <- r*weights
    rpearson <- as.vector(r)/link$variance(pi)^.5
    if (length(p) > length(coef(x))) {
        a.phi <- p[length(coef(x)) + 1]
    } else { ## if (tolower(family(x)$family)%in%c("gaussian","gamma","inverse.gaussian"))
        suppressWarnings(disp <- summary(x)$dispersion)
        if (!is.null(disp)) a.phi <- disp
        if (inherits(x, "negbin")) {
            a.phi <- link$variance(pi) / pi
        }
        ## a.phi <- sum(rpearson^2)*x$df.residual/x$df.residual^2
    }
    A <- as.vector(h(Xbeta) * r) / a.phi
    S <- apply(X, 2, function(x) x * A)
    if (!indiv) return(colSums(S))
    if (pearson) attr(S,"pearson") <- rpearson
    suppressWarnings(attributes(S)$bread <- vcov(x)*NROW(S))
    if (x$family$family=="quasi" && x$family$link=="identity" && x$family$varfun=="constant")
      attributes(S)$bread <-
                    suppressWarnings(
                      Inverse(information.glm(x)*NROW(S))
                    )
    if (pna) {
      attributes(S)$bread[idx, ] <- 0
      attributes(S)$bread[ , idx] <- 0
    }
    return(S)
}

##' @export
pars.glm <- function(x,...) {
    if (tolower(family(x)$family)%in%c("gaussian","gamma","inverse.gaussian")) {
        res <- c(coef(x),suppressWarnings(summary(x)$dispersion))
        names(res)[length(res)] <- "Dispersion"
        return(res)
    }
    return(coef(x))
}

logL.glm <- function(x,p=pars.glm(x),data,indiv=FALSE,offset=NULL,...) {
  if (!missing(data)) {
    x <- update(x, data=data, ...)
  }
  f <- family(x)
  ginv <- f$linkinv
  X <- model.matrix(x)
  n <- nrow(X)
  pna <- is.na(p)
  if(any(pna)) {
    warning("Over-parameterized model (NA parameters). Ignoring NA parameters")
    p[which(pna)] <- 0
  }
  ##disp <- 1;
  p0 <- p
  if (tolower(family(x)$family)%in%c("gaussian","gamma","inverse.gaussian")) {
    if (length(p)==ncol(X)) {
      ##disp <- suppressWarnings((summary(x)$dispersion))
    } else {
      ##disp <- tail(p,1)
      p0 <- p[-length(p)]
    }
  }
  Xbeta <- X%*%p0
  if (is.null(offset)) offset <- x$offset
  if (!is.null(offset)) Xbeta <- Xbeta+offset
  y <- model.frame(x)[,1]
  mu <- ginv(Xbeta)
  w <- x$prior.weights
  dev <-  f$dev.resids(y,mu,w)
   loglik <- sum(!pna)-(f$aic(y,n,mu,w,sum(dev))/2+x$rank)
  structure(loglik,nobs=n,df=length(p),class="logLik")
}

##' @export
IC.glm <- function(x,...) {
  IC.default(x,...)
}

hessian.glm <- function(x, p=coef(x), ...) {
  pna <- any(is.na(p))
  idx <- is.na(p)
  if(pna) {
    warning("Over-parameterized model (NA parameters). Ignoring NA parameters")
    p[idx] <- 0
  }
  res <- numDeriv::jacobian(function(theta) {
    score.glm(x, p=theta, indiv=FALSE, ...)
  }, p)
  res[idx, ] <- res[, idx] <- 0
  return(res)
}

##' @export
information.glm <- function(x, ...) -hessian.glm(x, ...)

glm_logLik.lvm <- function(object, ...) {
  attr(GLMscore(object, ...), "logLik")
}

glm_method.lvm <- NULL
glm_objective.lvm <- function(x,p,data,...) {
  GLMest(x, data, ...)
}
glm_gradient.lvm <- function(x,p,data,...) {
  -GLMscore(x, p, data, ...)
}

glm_variance.lvm <- function(x, p, data, opt, ...) {
  opt$vcov
}

context("predict_glm")

set.seed(42)
n <- 100
d <- data.frame(y = rpois(100, lambda = 3),
                x1 = rnorm(100), x2 = rnorm(100))

test_that("predict_glm matches predict.glm for poisson(log)", {
  m <- glm(y ~ x1 + x2, data = d, family = poisson(link = "log"))
  pred_resp <- as.numeric(predict_glm(m, type = "response"))
  pred_link <- as.numeric(predict_glm(m, type = "link"))
  expect_equal(pred_resp, as.numeric(predict(m, type = "response")))
  expect_equal(pred_link, as.numeric(predict(m, type = "link")))

  ## with modified parameter
  p0 <- coef(m) * 0
  pred_mod <- as.numeric(predict_glm(m, p = p0, type = "response"))
  expect_equal(pred_mod, exp(rep(0, nrow(d))))

  # with data
  pr <- as.numeric(predict_glm(m, p = p0, data=head(d, 5L)))
  expect_true(length(pr)==5L)
  expect_true(all(pr==1L))

  # with user offset
  pr <- as.numeric(predict_glm(m, p = p0, data=head(d, 5L), offset=1L))
  expect_true(all(pr==exp(1)))
})

test_that("predict_glm matches predict.glm with offset (poisson)", {
  m <- glm(y ~ x1 + offset(x2), data = d, family = poisson())
  pred_resp <- as.numeric(predict_glm(m, type = "response"))
  pred_link <- as.numeric(predict_glm(m, type = "link"))
  expect_equal(pred_resp, as.numeric(predict(m, type = "response")))
  expect_equal(pred_link, as.numeric(predict(m, type = "link")))
})

##' Plot path diagram
##'
##' Plot the path diagram of a SEM
##'
##'
##' @aliases plot.lvmfit
##' @param x Model object
##' @param diag Logical argument indicating whether to visualize
##'     variance parameters (i.e. diagonal of variance matrix)
##' @param cor Logical argument indicating whether to visualize
##'     correlation parameters
##' @param labels Logical argument indiciating whether to add labels
##'     to plot (Unnamed parameters will be labeled p1,p2,...)
##' @param intercept Logical argument indiciating whether to add
##'     intercept labels
##' @param addcolor Logical argument indiciating whether to add colors
##'     to plot (overrides \code{nodecolor} calls)
##' @param plain if TRUE strip plot of colors and boxes
##' @param cex Fontsize of node labels
##' @param fontsize1 Fontsize of edge labels
##' @param noplot if TRUE then return \code{graphNEL} object only
##' @param graph Graph attributes (Rgraphviz)
##' @param attrs Attributes (Rgraphviz)
##' @param unexpr if TRUE remove expressions from labels
##' @param addstyle Logical argument indicating whether additional
##'     style should automatically be added to the plot (e.g. dashed
##'     lines to double-headed arrows)
##' @param plot.engine default 'Rgraphviz' if available, otherwise
##'     visNetwork,igraph
##' @param init Reinitialize graph (for internal use)
##' @param layout Graph layout (see Rgraphviz or igraph manual)
##' @param edgecolor if TRUE plot style with colored edges
##' @param graph.proc Function that post-process the graph object
##'     (default: subscripts are automatically added to labels of the
##'     nodes)
##' @param ... Additional arguments to be passed to the low level
##'     functions
##' @author Klaus K. Holst
##' @keywords hplot regression
##' @examples
##'
##' if (interactive()) {
##' m <- lvm(c(y1,y2) ~ eta)
##' regression(m) <- eta ~ z+x2
##' regression(m) <- c(eta,z) ~ x1
##' latent(m) <- ~eta
##' labels(m) <- c(y1=expression(y[scriptscriptstyle(1)]),
##' y2=expression(y[scriptscriptstyle(2)]),
##' x1=expression(x[scriptscriptstyle(1)]),
##' x2=expression(x[scriptscriptstyle(2)]),
##' eta=expression(eta))
##' edgelabels(m, eta ~ z+x1+x2, cex=2, lwd=3,
##'            col=c("orange","lightblue","lightblue")) <- expression(rho,phi,psi)
##' nodecolor(m, vars(m), border="white", labcol="darkblue") <- NA
##' nodecolor(m, ~y1+y2+z, labcol=c("white","white","black")) <- NA
##' plot(m,cex=1.5)
##'
##' d <- sim(m,100)
##' e <- estimate(m,d)
##' plot(e)
##'
##' m <- lvm(c(y1,y2) ~ eta)
##' regression(m) <- eta ~ z+x2
##' regression(m) <- c(eta,z) ~ x1
##' latent(m) <- ~eta
##' plot(lava:::beautify(m,edgecol=FALSE))
##' }
##' @export
##' @method plot lvm
`plot.lvm` <-
  function(x,diag=FALSE,cor=TRUE,labels=FALSE,intercept=FALSE,addcolor=TRUE,plain=FALSE,cex,fontsize1=10,noplot=FALSE,graph=list(rankdir="BT"),
         attrs=list(graph=graph),
         unexpr=FALSE,
         addstyle=TRUE,plot.engine=lava.options()$plot.engine,init=TRUE,
         layout=lava.options()$layout,
         edgecolor=lava.options()$edgecolor,
         graph.proc=lava.options()$graph.proc,

         ...) {
    if (is.null(vars(x))) {
      message("Nothing to plot: model has no variables.")
      return(NULL)
    }
  index(x) <- reindex(x)
  myhooks <- gethook("plot.post.hooks")
  for (f in myhooks) {
    x <- do.call(f, list(x=x,...))
  }

    plot.engine <- tolower(plot.engine)
    if (plot.engine=="rgraphviz" && (!(requireNamespace("graph",quietly=TRUE)) || !(requireNamespace("Rgraphviz",quietly=TRUE)))) {
        plot.engine <- "visnetwork"
    }
    if (plot.engine=="visnetwork" && (!(requireNamespace("visNetwork",quietly=TRUE)))) {
        plot.engine <- "igraph"
    }
  if (plot.engine=="igraph") {
    if (!requireNamespace("igraph",quietly=TRUE)) {
      message("package 'Rgraphviz','igraph' or 'visNetwork' not available")
      return(NULL)
    }
    L <- igraph::layout.sugiyama(g <- igraph.lvm(x,...))$layout
    if (noplot) {
        return(igraph::as_graphnel(g))
    }
    dots <- list(...)
    if (is.character(layout))
      plot(g,layout=L,...)
    else plot(g,layout=layout,...)
    return(invisible(g))
  }
  if (plot.engine=="visnetwork" && !noplot) {
        g <- vis.lvm(x,labels=labels,...)
        return(g)
  }
  if (init) {
        if (!is.null(graph.proc)) {
            x <- do.call(graph.proc, list(x,edgecol=edgecolor,...))
        }
    g <- finalize(x,diag=diag,cor=cor,addcolor=addcolor,intercept=intercept,plain=plain,cex=cex,fontsize1=fontsize1,unexpr=unexpr,addstyle=addstyle)
  } else {
    g <- Graph(x)
  }
  if  (labels) {
    AP <- matrices(x,paste0("p",seq_len(index(x)$npar)))
    mylab <- AP$P; mylab[AP$A!="0"] <- AP$A[AP$A!="0"]
    mylab[!is.na(x$par)] <- x$par[!is.na(x$par)]
    mylab[!is.na(x$covpar)] <- x$covpar[!is.na(x$covpar)]
    g <- edgelabels(g, lab=mylab)
  }
  if (lava.options()$debug) {
    plot(g)
  } else {
    ## graphRenderInfo(g)$recipEdges <- "distinct"
    .savedOpt <- options(warn=-1) ## Temporarily disable warnings as renderGraph comes with a stupid warning when labels are given as "expression"
    dots <- list(...)
    dots$attrs <- attrs
    dots$x <- g
    dots$recipEdges <- "distinct"
    if (attributes(g)$feedback) dots$recipEdges <- c("combine")
    if (is.null(dots$layoutType)) dots$layoutType <- layout[1]
    if (all(index(x)$A==0))
      dots$layoutType <- "circo"

    if (requireNamespace("Rgraphviz",quietly=TRUE))
        g <- do.call(getFromNamespace("layoutGraph","Rgraphviz"), dots)
    ## Temporary work around:
    graph::nodeRenderInfo(g)$fill <- graph::nodeRenderInfo(dots$x)$fill
    graph::nodeRenderInfo(g)$col <- graph::nodeRenderInfo(dots$x)$col
    graph::edgeRenderInfo(g)$col <- graph::edgeRenderInfo(dots$x)$col
    if (noplot) return(suppressMessages(graph::updateGraph(g)))
    res <- tryCatch(Rgraphviz::renderGraph(g),error=function(e) NULL)
    ## Redo nodes to avoid edges overlapping node borders
    ##par(new=TRUE)
    ##res <- tryCatch(Rgraphviz::renderGraph(g,drawEdges=NULL,new=FALSE),error=function(e) NULL)
    ##
    options(.savedOpt)
  }

  myhooks <- gethook("plot.hooks")
  for (f in myhooks) {
    do.call(f, list(x=x,...))
  }

  invisible(g)
}

vis.lvm <- function(m,randomSeed=1,width="100%",height="700px",labels=FALSE,cor=TRUE,...) {
    if (!requireNamespace("visNetwork",quietly=TRUE)) stop("'visNetwork' required")
    types <- rep("endogenous",length(vars(m)))
    types[index(m)$eta.idx] <- "latent"
    types[index(m)$exo.idx] <- "exogenous"
    col <- lava.options()$node.color
    colors <- rep(col[2],length(types))
    colors[index(m)$eta.idx] <- col[3]
    colors[index(m)$exo.idx] <- col[1]
    trf <- transform(m)
    if (length(trf)>0) {
        colors[which(index(m)$vars%in%names(trf))] <- col[4]
    }
    shapes <- rep("box",length(types))
    shapes[index(m)$eta.idx] <- "circle"
    nodes <- data.frame(id=seq_along(types),
                        label=vars(m),
                        color=colors,
                        shape=shapes,
                        shadow=TRUE,
                        size=rep(1.0,length(types)),
                        group=types)
    edges <- cbind(edgeList(m))#,shadow=TRUE)

    AP <- matrices(m,paste0("p",seq_len(index(m)$npar)))
    if (labels) {
        mylab <- AP$A;
        mylab[!is.na(m$par)] <- m$par[!is.na(m$par)]
        lab <- c()
        for (i in seq(nrow(edges))) {
            lab <- c(lab,t(mylab)[edges[i,1],edges[i,2]])
        }
        edges <- cbind(edges,label=lab)
    }
    if (length(edges)>0)
        edges <- cbind(edges,dashes=FALSE,arrows="from")


    if (cor) {
        mylab <- AP$P
        mylab[!is.na(m$covpar)] <- m$covpar[!is.na(m$covpar)]
        coredges <- data.frame(from=numeric(),to=numeric(),label=character())
        for (i in seq_len(nrow(mylab)-1)) {
            for (j in seq(i+1,nrow(mylab))) {
                if (mylab[i,j]!="0") {
                    coredges <- rbind(coredges,
                                      data.frame(from=i,to=j,label=mylab[i,j]))
                }
            }
        }
        if (nrow(coredges)>0) {
            if (!labels) coredges <- coredges[,1:2,drop=FALSE]
            coredges <- cbind(coredges,dashes=TRUE,arrows="false")
            edges <- rbind(edges,coredges)
        }
    }

    if (length(edges)>0) edges$physics <- TRUE
    v <- visNetwork::visNetwork(nodes,edges,width=width,height=height,...)
    v <- visNetwork::visEdges(v, arrows=list(from=list(enabled=TRUE, scaleFactor = 0.5)),
                              scaling = list(min = 2, max = 2))
    v <- visNetwork::visLayout(v,randomSeed=randomSeed)
    v
}

##' @export
`plot.lvmfit` <-
    function(x,diag=TRUE,cor=TRUE,type,noplot=FALSE,fontsize1=5,f,graph.proc=lava.options()$graph.proc,...) {
        if (!missing(f)) {
            return(plot.estimate(x,f=f,...))
        }
    .savedOpt <- options(warn=-1) ## Temporarily disable warnings as renderGraph comes with a warning when labels are given as "expression"
    if (!requireNamespace("graph",quietly=TRUE)) {
      plot(Model(x),...)
      return(invisible(x))
    }
    g <- Graph(x)
    newgraph <- FALSE
    if (is.null(g)) {
        newgraph <- TRUE
        if (!is.null(graph.proc)) {
            Model(x) <- beautify(Model(x),edgecol=FALSE,...)
        }
        Graph(x) <- finalize(Model(x), diag=TRUE, cor=FALSE, fontsize1=fontsize1, ...)
    }
    if(noplot) return(Graph(x))
    if (newgraph) {
      if (missing(type))
        type <- "est"
      x <- edgelabels(x, type=type, diag=diag, cor=cor, fontsize1=fontsize1, ...)
    } else {
      if (!missing(type)) {
        x <- edgelabels(x, type=type, diag=diag, cor=cor, fontsize1=fontsize1, ...)
      }
    }
    g <- Graph(x)
    var <- rownames(covariance(Model(x))$rel)
    if (!cor) {
       delta <- 1
      for (r in seq_len(nrow(covariance(Model(x))$rel)-delta) ) {
        for (s in seq(r+delta,ncol(covariance(Model(x))$rel)) ) {
          if (covariance(Model(x))$rel[r,s]==1) {
            g <- graph::removeEdge(var[r],var[s], g)
            g <- graph::removeEdge(var[s],var[r], g)
          }
        }
      }
    }
    if (!diag) {
      for (r in seq_len(nrow(covariance(Model(x))$rel)) ) {
        if (graph::isAdjacent(g,var[r],var[r]))
          g <- graph::removeEdge(var[r],var[r],g)
      }
    }
    m <- Model(x); Graph(m) <- g
    g <- plot(m, diag=diag, cor=cor, fontsize1=fontsize1, init=FALSE, ...)
    options(.savedOpt)
    invisible(g)
  }

##' @export
plot.multigroup <- function(x,diag=TRUE,labels=TRUE,...) {
  k <- x$ngroup
  for (i in seq_len(k))
    plot(x$lvm[[i]],diag=diag,labels=labels, ...)
}

##' @export
plot.multigroupfit <- function(x,...) {
  plot(Model(x),...)
}

##' @export
igraph.lvm <- function(x,layout=igraph::layout.kamada.kawai,...) {
  requireNamespace("igraph",quietly=TRUE)
  oC <- covariance(x)$rel
  for (i in seq_len(nrow(oC)-1))
    for (j in seq(i+1,nrow(oC))) {
      if (oC[i,j]!=0) {
        x <- regression(x,vars(x)[i],vars(x)[j])
        x <- regression(x,vars(x)[j],vars(x)[i])
      }
    }
  g <- igraph::graph.adjacency(x$M,mode="directed")
  igraph::V(g)$color <- "lightblue"
  igraph::V(g)$label <- vars(x)
  igraph::V(g)$shape <- "rectangle"
  for (i in match(latent(x),igraph::V(g)$name)) {
      igraph::V(g)$shape[i] <- "circle"
      igraph::V(g)$color[i] <- "green"
  }
  endo <- index(x)$endogenous
  for (i in match(endo,igraph::V(g)$name)) {
      igraph::V(g)$color[i] <- "orange"
  }
  igraph::E(g)$label <- as.list(rep("",length(igraph::E(g))))
  oE <- edgelabels(x)
  for (i in seq_along(igraph::E(g))) {
    st <- as.character(oE[i])
    if (length(st)>0)
      igraph::E(g)$label[[i]] <- st
  }
  g$layout <- layout(g)
  return(g)
}

beautify <- function(x,col=lava.options()$node.color,border=rep("black",3),labcol=rep("darkblue",3),edgecol=TRUE,...) {
    if (is.null(x$noderender$fill)) notcolored <- vars(x)
    else notcolored <- vars(x)[is.na(x$noderender$fill)]
    x0 <- intersect(notcolored,exogenous(x))
    if (length(x0)>0)
        nodecolor(x, x0, border=border[1], labcol=labcol[1]) <- col[1]
    x0 <- intersect(notcolored,endogenous(x))
    if (length(x0)>0)
        nodecolor(x, x0, border=border[1], labcol=labcol[1]) <- col[2]
    x0 <- intersect(notcolored,latent(x))
    if (length(x0)>0)
        nodecolor(x, x0, border=border[1], labcol=labcol[1]) <- col[3]

    trimmed <- gsub("[[:digit:]]*$","",vars(x))
    keep <- num <- c()
    for (i in seq_len(length(vars(x)))) {
        lb <- labels(x)[vars(x)[i]]
        if (is.null(try(eval(lb),silent=TRUE))) {
            keep <- c(keep,i)
            num <- c(num,gsub(trimmed[i],"",vars(x)[i]))
        }
    }
    if (length(keep)>0) {
        trimmed <- trimmed[keep]
        trim <- gsub(" ",",",trimmed)
        lab <- paste0('"',vars(x)[keep],'"',"=",paste0("expression(",trim,"[scriptscriptstyle(",num,")])"),collapse=",")
        labels(x) <- eval(parse(text=paste("c(",lab,")")))
    }
    if (!edgecol) return(x)
    iex <- index(x)$exo.idx
    ien <- index(x)$endo.idx
    ila <- index(x)$eta.idx
    for (i in iex) {
        for (j in which(x$M[i,]==1)) {
            elab <- edgelabels(x,to=vars(x)[j],from=rev(vars(x)[i]))
            elab2 <- try(eval(elab),silent=TRUE)
            if (is.null(elab2)) elab2 <- ""
            edgelabels(x, to=vars(x)[j], from=rev(vars(x)[i]), cex=2, lwd=3,col=col[1]) <- elab2
        }
    }
    for (i in ien) {
        for (j in which(x$M[i,]==1)) {
            elab <- edgelabels(x,to=vars(x)[j],from=rev(vars(x)[i]))
            elab2 <- try(eval(elab),silent=TRUE)
            if (is.null(elab2)) elab2 <- ""
            edgelabels(x, to=vars(x)[j], from=rev(vars(x)[i]), cex=2, lwd=3,col=col[2]) <- elab2
        }
    }
    for (i in ila) {
        for (j in which(x$M[i,]==1)) {
            elab <- edgelabels(x,to=vars(x)[j],from=rev(vars(x)[i]))
            elab2 <- try(eval(elab),silent=TRUE)
            if (is.null(elab2)) elab2 <- ""
            if (is.null(try(eval(elab),silent=TRUE))) elab <- ""
            edgelabels(x, to=vars(x)[j], from=rev(vars(x)[i]), cex=2, lwd=3,col=col[3]) <- elab2
        }
    }
    x

}

IV_method.lvm <- NULL
IV_objective.lvm <- function(x,p,data,...) {
    IV2(x,data,...)
}
IV_variance.lvm <- function(x,p,data,opt,...) {
    opt$vcov
}

IV0_method.lvm <- NULL
IV0_objective.lvm <- function(x,p,data,...) {
    IV2(x,data,type="non-robust",...)
}
IV0_variance.lvm <- function(x,p,data,opt,...) {
    opt$vcov
}

IV1_method.lvm <- NULL
IV1_objective.lvm <- function(x,p,data,...) {
    IV(x,data)
}
IV1_variance.lvm <- function(x,p,data,opt,...) {
    opt$vcov
}

## Instrumental Variable Estimator / 2SLS
##' @export
IV <- function(m,data,R2thres=0,type="robust", ...) {
    if (length(constrain(m))>0) stop("Nonlinear constrains not supported!")
    if (inherits(m,"lvmfit")) {
        m <- Model(m)
    }
    R2 <- cor(data[,manifest(m)])^2

    A <- t(index(m)$A)
    Afix <- A; Afix[t(index(m)$M0)==1] <- 0
    A[A!=0] <- 1
    P <- index(m)$P
    k <- nrow(A)
    I <- diag(nrow=k)
    B <- rbind(I,solve(I-A))
    VV <- B%*%P%*%t(B)
    u.var <- index(m)$vars
    lat.idx <- with(index(m), which(vars%in%latent))
    if (length(lat.idx)==0) stop("Estimator only defined for models with latent variable")
    y.var <- endogenous(m)
    y.idx <- which(index(m)$vars%in%y.var)
    x.idx <- which(vars(m)%in%exogenous(m))

    ## Set of Indicator variables:
    indicators <- c()
    for (i in seq_len(nrow(A))) {
        ifix <- (Afix[i,]==1)
        if ((sum(ifix)==1) &  all(A[i,-which(ifix)]==0))
            indicators <- c(indicators, i)
    }
    y.indicators <- intersect(indicators,y.idx)

    y.scale <- list()
    for (eta in lat.idx) {
        pred.eta <- intersect(y.idx, which(Afix[,eta]==1)) ## Candidates for
        ## eta = y-epsilon
        if (length(pred.eta)<1)
            pred.eta <- intersect(lat.idx, which(Afix[,eta]==1))
        myidx <- c()
        for (y in pred.eta) {
            y.pred <- setdiff(eta,which(A[y,]==1)) ## No other variables predicting y?
            if (length(y.pred)==0)
                myidx <- c(myidx,y)
        }
        y.scale <- c(y.scale, list(myidx))
    }

    if (any(unlist(lapply(y.scale, function(x) length(x)))<1)) stop("At least one scale-measurement pr. latent variable")

    vv <- setdiff(seq_len(k),c(unlist(y.scale),x.idx))

    Ecor <- list()
    eta.surrogate <- c()
    latno <- 0
    for (e in lat.idx) {
        latno <- latno+1
        y0 <- y.scale[[latno]][1]
        if (!(y0%in%lat.idx)) {
            eta.surrogate <- c(eta.surrogate,vars(m)[y0])
            Ecor <- c(Ecor,list(y0))
        }
        else {
            v0 <- vars(m)[-c(e,indicators)]
            ##m.sub <- subset(m,vars(m)[c(e,indicators)])
            m.sub <- rmvar(m,v0)
            i <- 0
            while (i<length(y.indicators)) {
                i <- i+1
                pp <- path(m.sub,from=vars(m)[e],to=vars(m)[y.indicators[i]])[[1]]
                if (!is.null(pp)) {
                    Ecor <- c(Ecor,
                              list(which(vars(m)%in%pp[-1])))
                    eta.surrogate <- c(eta.surrogate, tail(pp,1))
                }
            }
        }
    };
    names(eta.surrogate) <- latent(m)

    coefname <- coef(m,mean=TRUE)
    P0 <- P
    V. <- list()
    Z. <- list()
    Y. <- list()
    count <- 0
    ff <- list()
    instruments <- c()
    parname <- c()
    for (v in vv) {
        pred <- which(A[v,]==1)
        if (sum(pred)>0) {
            Debug(vars(m)[v])
            pred.lat <- intersect(pred,lat.idx) # Any latent predictors?
            lpos <- match(v,lat.idx)
            lppos <- match(pred.lat,lat.idx)
            ecor <- c(v,unlist(Ecor[lppos]))
            if (!is.na(lpos)) {
                v0 <- match(eta.surrogate[lpos],vars(m))
                ecor <- c(ecor,Ecor[[lpos]])
            } else {
                v0 <- v
            }

            ecor <- unique(c(v0,ecor))
            XX <- vars(m)[A[v,]==1]
            intpred <- exogenous(m)
            newf <- c()
            if (length(pred.lat)>0) {
                intpred <- vars(m)
                for (i in seq_along(pred.lat)) {
                    uncor <- which(colSums(VV[ecor,k+seq_len(k),drop=FALSE])==0)
                    uncor <- setdiff(uncor,c(lat.idx))
                    mypred <- vars(m)[uncor]
                    XX[XX==vars(m)[pred.lat[i]]] <- eta.surrogate[lppos[i]]
                    intpred <- intersect(intpred,mypred)
                    f <- toformula(eta.surrogate[lppos[i]],mypred)
                    ff <- c(ff,
                            f)
                    f2 <- list(f)
                    names(f2) <- vars(m)[i]
                    newf <- c(newf,f2)
                }
            }

            intpred <- intersect(intpred,manifest(m))
            R2max <- apply(R2[XX,intpred,drop=FALSE],2,max)
            if (any(R2max<R2thres)) intpred <- intpred[R2max>=R2thres]
            newf <- list(intpred); names(newf) <- vars(m)[v]
            instruments <- c(instruments, newf)
            covariates <- unique(c(setdiff(colnames(A)[A[v,]==1],latent(m)),intpred))
            if (length(covariates)==0) stop("No instruments")
            Z <- model.matrix(toformula("",c("1",XX)),data)
            Y <- as.matrix(data[,vars(m)[v0]])
            V <- model.matrix(toformula("",c("1",unique(covariates))),data)
            count <- count+1
            V. <- c(V.,list(V))
            Z. <- c(Z.,list(Z))
            Y. <- c(Y.,list(Y))
            XX <- vars(m)[A[v,]==1 & Afix[v,]!=1]
            parname <- c(parname, c(vars(m)[v0],paste(vars(m)[v],XX,sep=lava.options()$symbol[1])))
        } else {
            if (vars(m)[v]%in%latent(m)) {
                lpos <- match(v,lat.idx)
                v0 <- match(eta.surrogate[lpos],vars(m))
                Y <- matrix(data[,vars(m)[v0]],ncol=1)
                Y. <- c(Y.,list(Y))
                V. <- c(V.,list(cbind(rep(1,nrow(Y)))))
                Z. <- c(Z.,list(cbind(rep(1,nrow(Y)))))
                parname <- c(parname, names(eta.surrogate)[lpos])
            }
        }
    }

    LS <- function(X) {
        with(svd(X), v%*%diag(1/d,nrow=length(d))%*%t(u))
    }
    ##projection <- function(X) X%*%LS(X)
    P0 <- lapply(V.,LS)
    Zhat <- list(); for (i in seq_along(Z.)) Zhat <- c(Zhat, list(V.[[i]]%*%(P0[[i]]%*%Z.[[i]])))
    ZhatLS <- lapply(Zhat,LS)
    theta <- list(); for (i in seq_along(Y.)) theta <- c(theta, list(ZhatLS[[i]]%*%Y.[[i]]))
    u <- c()
    for (i in seq_along(Y.))
        u <- cbind(u, Y.[[i]]-Z.[[i]]%*%theta[[i]])

    covu <- crossprod(u)/nrow(u)

    theta.npar <- unlist(lapply(theta,length))
    theta.ncum <- c(0,cumsum(theta.npar))
    vartheta <- matrix(0,ncol=sum(theta.npar),nrow=sum(theta.npar))
    for (i in seq_along(theta)) {
        for (j in seq(i,length(theta))) {
            idx1 <- seq_len(theta.npar[i]) + theta.ncum[i]
            idx2 <- seq_len(theta.npar[j]) + theta.ncum[j]
            if (type=="robust") {
                zi <- ZhatLS[[i]]
                for (k in seq(nrow(zi))) zi[k,] <- zi[k,]*u[,i]
                zj <- ZhatLS[[j]]
                for (k in seq(nrow(zj))) zj[k,] <- zj[k,]*u[,j]
                uZZ <- zi%*%t(zj)
            } else {
                uZZ <- covu[i,j]* (ZhatLS[[i]]%*%t(ZhatLS[[j]]))
            }
            vartheta[idx1,idx2] <- uZZ
            if (i!=j) {
                vartheta[idx2,idx1] <- t(uZZ)
            }
        }
    }

    parname[which(parname%in%eta.surrogate)] <- names(eta.surrogate)[which(eta.surrogate%in%parname)]

    coef <- cbind(unlist(theta),diag(vartheta)^0.5); rownames(coef) <- parname; colnames(coef) <- c("Estimate","Std.Err")
    res <- list(estimate=coef[,1], vcov=vartheta)
    attributes(res)$surrogates <- eta.surrogate
    attributes(res)$instruments <- instruments
    return(res)
}

IV2 <- function(m,data,control=list(),...) {
    if (is.null(control$R2thres)) control$R2thres <- 0
    res <- IV(m,data,R2thres=control$R2thres,...)
    p <- res$estimate
    idx <- match(names(p),coef(m,mean=TRUE))
    x0 <- parfix(m,idx,p)
    index(x0) <- reindex(x0,zeroones=TRUE,deriv=TRUE)
    idx0 <- order(idx)
    p0 <- p[idx0]
    V0 <- res$vcov[idx0,idx0]
    if (is.null(control$variance) || control$variance) {
        suppressWarnings(e0 <- estimate(x0,data,...,messages=0,quick=TRUE))
        p0 <- c(p0,e0)
        V0 <- V0%++%matrix(0,ncol=length(e0),nrow=length(e0))
    }
    R2 <- noquote(formatC(cor(data[,manifest(m)])^2))
    colnames(R2) <- rownames(R2) <- manifest(m)
    l1 <- noquote(rbind(paste(latent(m),collapse=","),
                        paste(attributes(res)$surrogates,collapse=","),
                        ""))
    rownames(l1) <- c("Latent variables","Surrogate variables:","")
    colnames(l1) <- ""
    ii <- attributes(res)$instruments
    I <- noquote(matrix(NA,ncol=2,nrow=length(ii)))
    rownames(I) <- rep("",nrow(I))
    colnames(I) <- c("Response","Instruments")
    for (i in seq_along(ii)) {
        I[i,] <- c(names(ii)[i],paste(ii[[i]],collapse=","))
    }
    mymsg <- list(l1,I);
    list(estimate=p0,vcov=V0,summary.message=function(...)  {
        mymsg })
}

##' @export
`pars` <-
  function(x,...) UseMethod("pars")

##' @export
pars.default <- function(x,...) {
  if (!is.null(x$opt$estimate))
      return(x$opt$estimate)
  if (!is.null(x$opt$par))
      return(x$opt$par)
  if (!is.null(x$coef))
    return(x$coef)
  return(coef(x))
}

##' @export
pars.lvm.missing <- function(x,reorder=FALSE,...) {
    res <- pars.default(x)
    if (reorder) {
        idx <- match(coef(Model(x)),names(coef(x)))
        return(res[idx])
    }
    return(res)
}

##' @export
pars.lvm <- function(x, A, P, v, e, ...) {
  parres <- A[index(x)$M1==1]
  diagcorfree <- diag(P)[diag(index(x)$P1)==1]
  parres <- c(parres, diagcorfree)

  if (ncol(A)>1)
  for (i in seq_len(ncol(index(x)$P1)-1))
    for (j in seq(i+1,nrow(index(x)$P1))) {
      if (index(x)$P1[j,i]!=0) {
        parres <- c(parres, P[j,i])
      }
    }
  if (length(parres)>0)
  names(parres) <- paste0("p",seq_len(length(parres)))
  if (!missing(v)) {
    parres <- c( v[which(index(x)$v1==1)], parres)
  }
  if (!missing(e)) {
    parres <- c( parres, e[which(index(x)$e1==1)] )
  }
  return(parres)
}

introotpn <- function(p) {
    ## Find integer root of x^2-x-2*p=0
    n <- 0.5*(1+sqrt(1+8*p))
    if (floor(n)!=n) n <- NA
    return(n)
}
rho2sigma <- function(rho) {
    if (length(rho)==1) return(diag(2)*(1-rho)+rho)
    p <- introotpn(length(rho))
    if (is.na(p)) stop("Unexpected length of correlation coefficients (p=n*(n-1)/2).")
    sigma <- diag(nrow=p)
    offdiag(sigma,type=2) <- rho
    offdiag(sigma,type=3) <- offdiag(t(sigma),type=3)
    return(sigma)
}

##' @export
rmvn0 <- function(n,mu,sigma,rho,...) {
    if (!missing(rho)) sigma <- rho2sigma(rho)
    if (!missing(mu) && missing(sigma)) sigma <- diag(nrow=length(mu))
    if (missing(sigma)) sigma <- matrix(1)
    if (is.vector(sigma)) sigma <- diag(sigma,ncol=length(sigma))
    if (missing(mu)) mu <- rep(0,ncol(sigma))
    PP <- with(svd(sigma), v%*%diag(sqrt(d),ncol=length(d))%*%t(u))
    res <- matrix(rnorm(ncol(sigma)*n),ncol=ncol(sigma))%*%PP
    if (NROW(mu)==nrow(res) && NCOL(mu)==ncol(res)) return(res+mu)
    return(res+cbind(rep(1,n))%*%mu)
}

##' @export
dmvn0 <- function(x,mu,sigma,rho,log=FALSE,nan.zero=TRUE,norm=TRUE,...) {
    if (!missing(rho)) sigma <- rho2sigma(rho)
    if (!missing(mu) && missing(sigma)) sigma <- diag(nrow=length(mu))
    if (missing(sigma)) sigma <- matrix(1)
    if (is.vector(sigma)) sigma <- diag(sigma,ncol=length(sigma))
    if (missing(mu)) mu <- rep(0,ncol(sigma))

    if (length(sigma)==1) {
        k <- 1
        isigma <- structure(cbind(1/sigma),det=as.vector(sigma))

    } else {
        k <- ncol(sigma)
        isigma <- Inverse(sigma)
    }
    if (!missing(mu)) {
        if (NROW(mu)==NROW(x) && NCOL(mu)==NCOL(x)) {
            x <- x-mu
        } else {
            x <- t(t(x)-mu)
        }
    }
    logval <- -0.5*(base::log(2*base::pi)*k+
                    base::log(attributes(isigma)$det)+
                    rowSums((x%*%isigma)*x))
    if (nan.zero) logval[is.nan(logval)] <- -Inf
    if (log) return(logval)
    return(exp(logval))
}


normal_method.lvm <- "nlminb0"

normal_objective.lvm <- function(x,p,data,weights=NULL,data2=NULL,indiv=FALSE,...) {
    if (!requireNamespace("mets",quietly=TRUE)) stop("'mets' package required")
    if (!exists(".Random.seed", envir = .GlobalEnv, inherits = FALSE)) runif(1)
    save.seed <- get(".Random.seed", envir = .GlobalEnv)
    on.exit(assign(".Random.seed", save.seed, envir = .GlobalEnv))
    set.seed(1)
    ii <- index(x)
    x.idx <- ii$exo.idx
    y <- ii$endogenous
    ord <- lava::ordinal(x)
    atr <- attributes(ord)
    ord <- intersect(y,ord)
    attributes(ord) <- atr

    status <- rep(0,length(y))
    status[match(ord,y)] <- 2

    Table <- (length(y)==length(ord)) && (length(x.idx)==0)
    if (Table) {
        pat <- mets::fast.pattern(data[,y,drop=FALSE],categories=max(data[,y,drop=FALSE])+1)
        data <- pat$pattern
        colnames(data) <- y
    }

    mu <- predict(x,data=data,p=p)
    S <- attributes(mu)$cond.var
    class(mu) <- "matrix"
    thres <- matrix(0,nrow=length(y),max(1,attributes(ord)$K-1)); rownames(thres) <- y
    for (i in seq_len(length(attributes(ord)$fix))) {
        nn <- names(attributes(ord)$idx)[i]
        ii <- attributes(ord)$idx[[nn]]
        val <- (attributes(mu)$e[ii])
        thres[nn,seq_len(length(val))] <-
            cumsum(c(val[1],exp(val[-1])))
    }

    yl <- yu <- as.matrix(data[,y,drop=FALSE])
    if (!inherits(yl[1,1],c("numeric","integer","logical")) ||
        !inherits(yu[1,1],c("numeric","integer","logical")))
        stop("Unexpected data (normal_objective)")

    if (!is.null(data2)) {
        yu[,colnames(data2)] <- data2
        status[match(colnames(data2),y)] <- 1
    }

    l <- mets::loglikMVN(yl,yu,status,mu,S,thres)

    if (!is.null(weights)) {
        ##if (is.matrix(weights)) weights <- weights[,1]
        l <- l*weights
    }

    if (Table) {
        l <- l[pat$group+1]
    }
    if (indiv) return(-l)
    return(-sum(l))
}


normal_logLik.lvm <- function(object,p,data,data2=NULL,...) {
    res <- -normal_objective.lvm(x=object,p=p,data=data,data2=data2,...)
    return(res)
}

normal_gradient.lvm <- function(x,p,data,weights=NULL,data2=NULL,indiv=FALSE,...) {
    if (!exists(".Random.seed", envir = .GlobalEnv, inherits = FALSE)) runif(1)
    save.seed <- get(".Random.seed", envir = .GlobalEnv)
    on.exit(assign(".Random.seed", save.seed, envir = .GlobalEnv))
    if (!requireNamespace("mets",quietly=TRUE)) stop("'mets' package required")
    if  (is.null(ordinal(x)) && is.null(data2) && is.null(weights)) {
        D <- deriv.lvm(x,p=p)
        M <- moments(x,p)
        Y <- as.matrix(data[,manifest(x),drop=FALSE])
        mu <- M$xi%x%rep(1,nrow(Y))
        ss <- -mets::scoreMVN(Y,mu,M$C,D$dxi,D$dS)
        if (!indiv) return(colSums(ss))
        return(ss)
    }
    if (indiv) {
        return(numDeriv::jacobian(function(p0) normal_objective.lvm(x,p=p0,data=data,weights=weights,data2=data2,indiv=TRUE,...),p,method=lava.options()$Dmethod))
    }
    numDeriv::grad(function(p0) normal_objective.lvm(x,p=p0,data=data,weights=weights,data2=data2,...),p,method=lava.options()$Dmethod)
}

normal_hessian.lvm <- function(x,p,outer=FALSE,data2=NULL,...) {
    if (!requireNamespace("mets",quietly=TRUE)) stop("'mets' package required")
    dots <- list(...); dots$weights <- NULL
    if (!exists(".Random.seed", envir = .GlobalEnv, inherits = FALSE)) runif(1)
    save.seed <- get(".Random.seed", envir = .GlobalEnv)
    on.exit(assign(".Random.seed", save.seed, envir = .GlobalEnv))
    if (!outer) {
        f <- function(p) {
            set.seed(1)
        do.call("normal_objective.lvm", c(list(x,p=p,indiv=FALSE,data2=data2),dots))
        }
        g <- function(p) {
            set.seed(1)
            do.call("normal_gradient.lvm", c(list(x,p=p,indiv=FALSE,data2=data2),dots))
        }
        if (is.null(ordinal(x)) && is.null(data2))
            return(numDeriv::jacobian(g,p))
        else {
            return(numDeriv::hessian(f,p))
        }
    }
    ## Else calculate outer product of the score (empirical variance of score)
    S <- normal_gradient.lvm(x,p=p,indiv=TRUE,...)
    J <- t(S)%*%S
    attributes(J)$grad <- colSums(S)
    return(J)

}

##normal_gradient.lvm <- normal_hessian.lvm <- NULL

##' Extract variable names from latent variable model
##'
##' Extract exogenous variables (predictors), endogenous variables (outcomes),
##' latent variables (random effects), manifest (observed) variables from a
##' \code{lvm} object.
##'
##' \code{vars} returns all variables of the \code{lvm}-object including
##' manifest and latent variables. Similarily \code{manifest} and \code{latent}
##' returns the observered resp. latent variables of the model.
##' \code{exogenous} returns all manifest variables without parents, e.g.
##' covariates in the model, however the argument \code{latent=TRUE} can be used
##' to also include latent variables without parents in the result. Pr. default
##' \code{lava} will not include the parameters of the exogenous variables in
##' the optimisation routine during estimation (likelihood of the remaining
##' observered variables conditional on the covariates), however this behaviour
##' can be altered via the assignment function \code{exogenous<-} telling
##' \code{lava} which subset of (valid) variables to condition on.  Finally
##' \code{latent} returns a vector with the names of the latent variables in
##' \code{x}. The assigment function \code{latent<-} can be used to change the
##' latent status of variables in the model.
##'
##' @aliases vars vars.lvm vars.lvmfit latent latent<- latent.lvm latent<-.lvm
##' @aliases latent.lvmfit latent.multigroup manifest manifest.lvm manifest.lvmfit
##' @aliases manifest.multigroup exogenous exogenous<- exogenous.lvm exogenous<-.lvm
##' @aliases exogenous.lvmfit exogenous.multigroup endogenous endogenous.lvm
##' @aliases endogenous.lvmfit endogenous.multigroup
##' @usage
##'
##' vars(x,...)
##'
##' endogenous(x,...)
##'
##' exogenous(x,...)
##'
##' manifest(x,...)
##'
##' latent(x,...)
##'
##' \method{exogenous}{lvm}(x, xfree = TRUE,...) <- value
##'
##' \method{exogenous}{lvm}(x,variable,latent=FALSE,index=TRUE,...)
##'
##' \method{latent}{lvm}(x,clear=FALSE,...) <- value
##'
##' @param x \code{lvm}-object
##' @param variable list of variables to alter
##' @param latent Logical defining whether latent variables without parents
##' should be included in the result
##' @param index For internal use only
##' @param clear Logical indicating whether to add or remove latent variable
##' status
##' @param xfree For internal use only
##' @param value Formula or character vector of variable names.
##' @param \dots Additional arguments to be passed to the low level functions
##' @return Vector of variable names.
##' @author Klaus K. Holst
##' @seealso \code{\link{endogenous}}, \code{\link{manifest}},
##' \code{\link{latent}}, \code{\link{exogenous}}, \code{\link{vars}}
##' @keywords models regression
##' @examples
##'
##' g <- lvm(eta1 ~ x1+x2)
##' regression(g) <- c(y1,y2,y3) ~ eta1
##' latent(g) <- ~eta1
##' endogenous(g)
##' exogenous(g)
##' identical(latent(g), setdiff(vars(g),manifest(g)))
##'
##' @export
`vars` <-
function(x,...) UseMethod("vars")

##' @export
`vars.graph` <-
  function(x,...) {
    graph::nodes(x)
  }

##' @export
`vars.lvm` <-
  function(x,...) {
    colnames(x$M)
  }

##' @export
`vars.lvmfit` <-
  function(x,...) {
    vars(Model(x),...)
  }

##' @export
vars.list <- function(x,...) {
  varlist <- c()
  for (i in seq_along(x)) {
    varlist <- c(varlist, vars(x[[i]]))
  }
  varlist <- unique(varlist)
  return(varlist)
}

##' @export
`vars.lm` <-
  function(x,...) {
    c(endogenous(x),exogenous(x))
  }

##' @export
vcov.lvmfit <- function(object,...) {
    res <- object$vcov
    if (inherits(object,"lvm.missing")) {
        resnames <- names(coef(object))

    } else {
        resnames <- coef(Model(object),fix=FALSE, mean=object$control$meanstructure)
    }
    colnames(res) <- rownames(res) <- resnames
    return(res)
}

##' @export
vcov.multigroupfit <- function(object,...) {
    res <- object$vcov
    colnames(res) <- rownames(res) <- object$model$name
    return(res)
}

##' Combine matrices to block diagonal structure
##' @title Combine matrices to block diagonal structure
##' @param x matrix
##' @param \dots additional matrices
##' @param pad value outside block-diagonal
##' @author Klaus K. Holst
##' @export
##' @examples
##' A <- diag(3)+1
##' blockdiag(A,A,A,pad=NA)
blockdiag <- function(x,...,pad=0) {
  if (is.list(x)) xx <- x else xx <- list(x,...)
  rows <- unlist(lapply(xx,nrow))
  crows <- c(0,cumsum(rows))
  cols <- unlist(lapply(xx,ncol))
  ccols <- c(0,cumsum(cols))
  res <- matrix(pad,nrow=sum(rows),ncol=sum(cols))
  for (i in seq_len(length(xx))) {
    idx1 <- seq_len(rows[i])+crows[i]; idx2 <- seq_len(cols[i])+ccols[i]
    res[idx1,idx2] <- xx[[i]]
  }
  colnames(res) <- unlist(lapply(xx,colnames)); rownames(res) <- unlist(lapply(xx,rownames))
  return(res)
}

##' Dose response calculation for binomial regression models
##'
##' @title Dose response calculation for binomial regression models
##' @param model Model object or vector of parameter estimates
##' @param intercept Index of intercept parameters
##' @param slope Index of intercept parameters
##' @param prob Index of mixture parameters (only relevant for
##' \code{zibreg} models)
##' @param x Optional weights
##' length(x)=length(intercept)+length(slope)+length(prob)
##' @param level Probability at which level to calculate dose
##' @param ci.level Level of confidence limits
##' @param vcov Optional estimate of variance matrix of parameter
##' estimates
##' @param family Optional distributional family argument
##' @param EB Optional ratio of treatment effect and adverse effects
##' used to find optimal dose (regret-function argument)
##' @author Klaus K. Holst
##' @export
PD <- function(model,intercept=1,slope=2,prob=NULL,x,level=0.5,
        ci.level=0.95,vcov,family, EB=NULL) {
    if (is.vector(model)) {
        beta <- model
        if (missing(vcov)) stop("vcov argument needed")
        if (missing(family)) stop("family argument needed")
    } else beta <- coef(model)
    if (missing(vcov)) vcov <- stats::vcov(model)
    if (missing(family)) family <- stats::family(model)
    N <- length(intercept)+length(slope)+length(prob)
    if (length(intercept)<length(beta)) {
        B.intercept <- rep(0,length(beta));
        if (!missing(x)) {
            if (length(x)!=N) stop("x should be of same length as the total length of 'intercept','slope','prob'")
            B.intercept[intercept] <- x[seq_len(length(intercept))]
        } else B.intercept[intercept] <- 1
    } else {
        B.intercept <- intercept
    }
    if (length(slope)<length(beta)) {
        B.slope <- rep(0,length(beta));
        if (!missing(x))
            B.slope[slope] <- x[length(intercept)+seq_len(length(slope))]
        else
            B.slope[slope] <- 1
    } else {
        B.slope <- slope
    }
    if (!is.null(prob)) {
        if (length(prob)<length(beta)) {
            B.prob <- rep(0,length(beta));
            if (!missing(x))
                B.prob[prob] <- x[length(intercept)+length(slope)+seq_len(length(prob))]
            else
                B.prob[prob] <- 1
        } else {
            B.prob <- prob
        }
    }
    if (is.null(prob)) B.prob <- NULL
    B <- rbind(B.intercept,B.slope,B.prob)
    S <- B%*%vcov%*%t(B)
    b <- as.vector(B%*%beta)

    f <- function(b) {
        mylevel <- level
        if (!is.null(EB)) {
            if (is.null(prob)) stop("Index of mixture-probability parameters needed")
            pi0 <- family$linkinv(b[3])
            mylevel <- 1-(1-pi0)/pi0*(EB)/(1-EB)
        }
        return(structure((family$linkfun(mylevel)-b[1])/b[2],level=mylevel))
    }

    xx <- f(b)
    Dxx <- -1/b[2]*rbind(1,xx)
    if (!is.null(EB))
        Dxx <- numDeriv::grad(f,b)
    se <- diag(t(Dxx)%*%S%*%Dxx)^0.5
    res <- cbind(Estimate=xx,"Std.Err"=se)
    alpha <- 1-ci.level
    alpha.str <- paste(c(alpha/2,1-alpha/2)*100,"",sep="%")
    res <- cbind(res,res[,1]-qnorm(1-alpha/2)*res[,2],res[,1]+qnorm(1-alpha/2)*res[,2])
    colnames(res)[3:4] <- alpha.str
    rownames(res) <- paste0(round(1000*attributes(xx)$level)/10,"%")
    structure(res,b=b)
}


TN.zibreg <- function(object,data=model.frame(object),p=coef(object),intercept=1,slope=2,alpha=0.95,...) {
    pp <- predict(object,link=FALSE,p=p,newdata=data)
    X <- attributes(pp)$grad$beta
    Z <- attributes(pp)$grad$gamma
    db1 <- db2 <- matrix(0,nrow(X),ncol(X))
    db1[,intercept] <- X[,intercept]
    db2[,slope[1]] <- 1; db2[,slope[-1]] <- X[,slope[-1]]
    b1 <- as.vector(db1%*%p[object$beta.idx])
    b2 <- as.vector(db2%*%p[object$beta.idx])
    ginv <-  object$family$linkinv
    dginv <- object$family$mu.eta ## D[linkinv]
    g <- object$family$linkfun
    dg <- function(x) 1/dginv(g(x)) ## Dh^-1 = 1/(h'(h^-1(x)))
    pi0 <- ginv(pp[,2])
    A2 <- dginv(pp[,2])
    dpi0 <- rbind(apply(Z,2,function(z) A2*z))
    h <- function(pi0) (alpha+pi0-1)/(alpha*pi0)
    dh <- function(pi0) (1-alpha)/(alpha*pi0^2)
    lev <- h(pi0)
    eta <- g(lev)
    detad2 <- rbind(apply(dpi0,2,function(z) dg(lev)*dh(pi0)*z))
    val <- (eta-b1)/b2
    dvald1 <- -(db1+db2*val)/b2
    return(structure(val,grad=cbind(dvald1,detad2/b2),varnames="theta"))
    ##  structure(g(coef(object)),grad=grad(g,coef(object)))
}



##' Regression model for binomial data with unkown group of immortals (zero-inflated binomial regression)
##'
##' @title Regression model for binomial data with unkown group of immortals
##' @param formula Formula specifying
##' @param formula.p Formula for model of disease prevalence
##' @param data data frame
##' @param family Distribution family (see the help page \code{family})
##' @param offset Optional offset
##' @param start Optional starting values
##' @param var Type of variance (robust, expected, hessian, outer)
##' @param ... Additional arguments to lower level functions
##' @author Klaus K. Holst
##' @export
##' @examples
##'
##' ## Simulation
##' n <- 2e3
##' x <- runif(n,0,20)
##' age <- runif(n,10,30)
##' z0 <- rnorm(n,mean=-1+0.05*age)
##' z <- cut(z0,breaks=c(-Inf,-1,0,1,Inf))
##' p0 <- lava:::expit(model.matrix(~z+age) %*% c(-.4, -.4, 0.2, 2, -0.05))
##' y <- (runif(n)<lava:::tigol(-1+0.25*x-0*age))*1
##' u <- runif(n)<p0
##' y[u==0] <- 0
##' d <- data.frame(y=y,x=x,u=u*1,z=z,age=age)
##' head(d)
##'
##' ## Estimation
##' e0 <- zibreg(y~x*z,~1+z+age,data=d)
##' e <- zibreg(y~x,~1+z+age,data=d)
##' compare(e,e0)
##' e
##' PD(e0,intercept=c(1,3),slope=c(2,6))
##'
##' B <- rbind(c(1,0,0,0,20),
##'            c(1,1,0,0,20),
##'            c(1,0,1,0,20),
##'            c(1,0,0,1,20))
##' prev <- summary(e,pr.contrast=B)$prevalence
##'
##' x <- seq(0,100,length.out=100)
##' newdata <- expand.grid(x=x,age=20,z=levels(d$z))
##' fit <- predict(e,newdata=newdata)
##' plot(0,0,type="n",xlim=c(0,101),ylim=c(0,1),xlab="x",ylab="Probability(Event)")
##' count <- 0
##' for (i in levels(newdata$z)) {
##'   count <- count+1
##'   lines(x,fit[which(newdata$z==i)],col="darkblue",lty=count)
##' }
##' abline(h=prev[3:4,1],lty=3:4,col="gray")
##' abline(h=prev[3:4,2],lty=3:4,col="lightgray")
##' abline(h=prev[3:4,3],lty=3:4,col="lightgray")
##' legend("topleft",levels(d$z),col="darkblue",lty=seq_len(length(levels(d$z))))
zibreg <- function(formula,formula.p=~1,data,family=stats::binomial(),offset=NULL,start,var="hessian",...) {
    md <- cbind(model.frame(formula,data),model.frame(formula.p,data))
    y <- md[,1]
    X <- model.matrix(formula,data)
    Z <- model.matrix(formula.p,data)
    beta.idx <- seq(ncol(X)); gamma.idx <- seq(ncol(Z))+ncol(X)
    if (missing(start)) start <- rep(0,ncol(X)+ncol(Z))
    op <- nlminb(start,function(x)
        -zibreg_logL(x[beta.idx],x[gamma.idx],y,X,Z),
        gradient=function(x)
            -zibreg_score(x[beta.idx],x[gamma.idx],y,X,Z),...)
    beta <- op$par[beta.idx]; gamma <- op$par[gamma.idx]
    cc <- c(beta,gamma)
    names(cc) <- c(colnames(X),paste0("pr:",colnames(Z)))
    bread <- Inverse(zibreg_information(beta,gamma,y,X,Z,offset,type="hessian",...))
    if (tolower(var[1])%in%c("robust","sandwich")) {
        meat <- zibreg_information(beta,gamma,y,X,Z,offset,family,type="outer",...)
        V <- bread%*%meat%*%bread
    } else {
        V <- bread
    }
    n <- length(y)
    colnames(V) <- rownames(V) <- names(cc)
    res <- list(coef=cc,opt=op,beta=beta,gamma=gamma,
                beta.idx=beta.idx,gamma.idx=gamma.idx,bread=bread*n,
                formula=formula,formula.p=formula.p, y=y, X=X, Z=Z, offset=offset, vcov=V, model.frame=md,family=family)
    class(res) <- "zibreg"
    res$fitted.values <- predict(res)
    return(res)
}

##' @export
vcov.zibreg <- function(object,...) object$vcov

##' @export
coef.zibreg <- function(object,...) object$coef

##' @export
family.zibreg <- function(object,...) object$family

##' @export
predict.zibreg <- function(object,p=coef(object),gamma,newdata,link=TRUE,subdist=FALSE,...) {
    newf <- as.formula(paste("~",as.character(object$formula)[3]))
    if (missing(newdata)) {
        X <- object$X; Z <- object$Z
    } else {
        X <- model.matrix(newf,newdata)
        Z <- model.matrix(object$formula.p,newdata)
    }
    if (length(p)==length(object$beta)+length(object$gamma)) {
        gamma <- p[object$gamma.idx]
        p <- p[object$beta.idx]
    }
    ginv <- object$family$linkinv
    dginv <- object$family$mu.eta ## D[linkinv]
    Xbeta <- as.vector(X%*%p)
    Zgamma <- as.vector(Z%*%gamma)
    if (!link) {
        res <- cbind(beta=Xbeta,gamma=Zgamma)
        return(structure(res,grad=list(beta=X,gamma=Z)))
    }
    Pred <- ginv(Xbeta)
    p0 <- ginv(Zgamma)
    A1 <- dginv(Xbeta)
    A2 <- dginv(Zgamma)
    if (subdist) {
        dgamma <- apply(Z,2,function(z) A2*z)
        dbeta <- apply(X,2,function(x) A1*x)
        res <- cbind(subdist=Pred,pr=p0)
        return(structure(res,grad=list(subdist=dbeta,pr=dgamma)))
    }
    Pred <- p0*Pred
    A1 <- p0*A1
    A2 <- Pred*dginv(Zgamma)
    dgamma <- apply(Z,2,function(z) A2*z)
    dbeta <- apply(X,2,function(x) A1*x)
    attributes(Pred)$grad <- cbind(dbeta,p0*dgamma)
    return(Pred)
}

##' @export
residuals.zibreg <- function(object,newdata,...) {
    if (missing(newdata)) {
        y <- object$y
    } else {
        y <- model.frame(object$formula,newdata)[,1]
    }
    y-predict(object,newdata=newdata,...)
}

##' @export
summary.zibreg <- function(object,level=0.95,pr.contrast,...) {
    alpha <- 1-level
    alpha.str <- paste(c(alpha/2,1-alpha/2)*100,"",sep="%")
    cc <- cbind(coef(object),diag(vcov(object))^0.5)
    pval <- 2*(pnorm(abs(cc[,1]/cc[,2]),lower.tail=FALSE))
    qq <- qnorm(1-alpha/2)
    cc <- cbind(cc[,1],cc[,1]-qq*cc[,2],cc[,1]+qq*cc[,2],pval)
    colnames(cc) <- c("Estimate",alpha.str,"P-value")
    pr.names <- unlist(lapply(rownames(cc)[object$gamma.idx],
                              function(x) substr(x,4,nchar(x))))
    if (missing(pr.contrast)) {
        withIntercept <- pr.names[1]=="(Intercept)"
        pr.contrast <- diag(length(object$gamma.idx))
        if (withIntercept) pr.contrast[,1] <- 1
    }
    pr.cc <- cbind(pr.contrast%*%cc[object$gamma.idx,1],
                   diag((pr.contrast)%*%vcov(object)[object$gamma.idx,object$gamma.idx]%*%t(pr.contrast))^0.5)
    pr.cc <- object$family$linkinv(cbind(pr.cc[,1],pr.cc[,1]-qq*pr.cc[,2],pr.cc[,1]+qq*pr.cc[,2]))
    colnames(pr.cc) <- colnames(cc)[1:3]
    ## B <- cbind(0,cbind(0,pr.contrast))
    ## print(compare(object,contrast=B))
    pr.rnames <- c()
    for (i in seq_len(nrow(pr.contrast))) {
        Bidx <- which(pr.contrast[i,]!=0)
        Bval <- pr.contrast[i,Bidx]; Bval[Bval==1] <- ""
        pr.rnames <- c(pr.rnames,
                       paste0(paste0(Bval,paste0("{",pr.names[Bidx],"}"),collapse=" + ")))
    }
    rownames(pr.cc) <- pr.rnames

    return(structure(list(coef=cc, prevalence=pr.cc),class="summary.zibreg"))
}


##' @export
print.summary.zibreg <- function(x,...) {
    print(x$coef,...)
    cat("\nPrevalence probabilities:\n")
    print(x$prevalence,...)
}

##' @export
print.zibreg <- function(x,...) {
    print(summary(x,...))
}

##' @export
logLik.zibreg <- function(object,beta=object$beta,gamma=object$gamma,data,offset=object$offset,indiv=FALSE,...) {
    if (!missing(data)) {
        y <- model.frame(object$formula,data)[,1]
        X <- model.matrix(object$formula,data)
        Z <- model.matrix(object$formula.p,data)
        return(zibreg_logL(beta,gamma,y,X,Z,offset,object$family,indiv=indiv,...))
    }
    zibreg_logL(beta,gamma,object$y,object$X,object$Z,offset,object$family,indiv=indiv,...)
}
zibreg_logL <- function(beta,gamma,y,X,Z,offset=NULL,family=stats::binomial(),indiv=FALSE,...) {
    ginv <- family$linkinv
    n <- nrow(X)
    Xbeta <- as.vector(X%*%beta)
    Zgamma <- as.vector(Z%*%gamma)
    p0 <- ginv(Zgamma)
    if (!is.null(offset)) Xbeta <- Xbeta+offset
    Pr <- p0*ginv(Xbeta)
    loglik <- y*log(Pr)+(1-y)*log(1-Pr)
    if (indiv) return(loglik)
    loglik <- sum(loglik)
    structure(loglik,nobs=n,df=length(beta)+length(gamma),class="logLik")
}

##' @export
score.zibreg <- function(x,beta=x$beta,gamma=x$gamma,data,offset=x$offset,indiv=FALSE,...) {
    if (!missing(data)) {
        y <- model.frame(x$formula,data)[,1]
        X <- model.matrix(x$formula,data)
        Z <- model.matrix(x$formula.p,data)
        s <- zibreg_score(beta,gamma,y,X,Z,offset,x$family,indiv=indiv,...)
    } else {
        s <- zibreg_score(beta,gamma,x$y,x$X,x$Z,offset,x$family,indiv=indiv,...)
    }
    if (indiv) colnames(s) <- names(x$coef) else names(s) <- names(x$coef)
    return(s)
}

zibreg_score <- function(beta,gamma,y,X,Z,offset=NULL,family=stats::binomial(),indiv=FALSE,...) {
    ginv <- family$linkinv
    dginv <- family$mu.eta ## D[linkinv]
    Xbeta <- as.vector(X%*%beta)
    Zgamma <- as.vector(Z%*%gamma)
    p0 <- ginv(Zgamma)
    if (!is.null(offset)) Xbeta <- Xbeta+offset
    Pr <- p0*ginv(Xbeta)
    A0 <- (y/Pr  - (1-y)/(1-Pr))
    A1 <- A0*p0*dginv(Xbeta)
    A2 <- A0*ginv(Xbeta)*dginv(Zgamma)
    dbeta <- apply(X,2,function(x) A1*x)
    dgamma <- apply(Z,2,function(z) A2*z)
    ss <- cbind(dbeta,dgamma)
    if (indiv) return(ss)
    colSums(ss)
}

##' @export
information.zibreg <- function(x,beta=x$beta,gamma=x$gamma,data,offset=x$offset,type=c("robust","outer","obs"),...) {
    if (!missing(data)) {
        y <- model.frame(x$formula,data)[,1]
        X <- model.matrix(x$formula,data)
        Z <- model.matrix(x$formula.p,data)
        I <- zibreg_information(beta,gamma,y,X,Z,offset,x$family,type=type,...)
    } else {
        I <- zibreg_information(beta,gamma,x$y,x$X,x$Z,offset,x$family,type=type,...)
    }
    colnames(I) <- rownames(I) <- names(x$coef)
    return(I)
}

zibreg_information <- function(beta,gamma,y,X,Z,offset=NULL,family=stats::binomial(),type=c("outer","obs","robust"),...) {
    if (tolower(type[1])%in%c("obs","hessian")) {
        beta.idx <- seq(ncol(X)); gamma.idx <- seq(ncol(Z))+ncol(X)
        I <- -numDeriv::jacobian(function(x)
            zibreg_score(x[beta.idx],x[gamma.idx],y,X,Z,offset,family,...),c(beta,gamma))
        return(I)
    }
    if (tolower(type[1])%in%c("robust","sandwich")) {
        I <- zibreg_information(beta,gamma,y,X,Z,offset,family,type="obs")
        J <- zibreg_information(beta,gamma,y,X,Z,offset,family,type="outer")
        return(J%*%Inverse(I)%*%J)
    }
    S <- zibreg_score(beta,gamma,y,X,Z,offset,family,indiv=TRUE,...)
    crossprod(S)
}

##' @export
logLik.lvm <- function(object,p,data,model="gaussian",indiv=FALSE,S,mu,n,debug=FALSE,weights=NULL,data2=NULL,...) {
    cl <- match.call()
    xfix <- colnames(data)[(colnames(data)%in%parlabels(object,exo=TRUE))]

    constr <- lapply(constrain(object), function(z)(attributes(z)$args))
    xconstrain <- intersect(unlist(constr), manifest(object))
    xconstrainM <- TRUE
    if (length(xconstrain)>0) {
        constrainM <- names(constr)%in%unlist(object$mean)
        for (i in seq_len(length(constr))) {
            if (!constrainM[i]) {
                if (any(constr[[i]]%in%xconstrain)) xconstrainM <- FALSE
            }
        }
    }

    Debug(xfix,debug)
    if (missing(n)) {
        n <- nrow(data)
        if (is.null(n)) n <- data$n
    }
    lname <- paste0(model,"_logLik.lvm")
    logLikFun <- get(lname)


    if (length(xfix)>0 | (length(xconstrain)>0 & !xconstrainM & !lava.options()$test & model!="gaussian")) { ## Random slopes!
        x0 <- object
        if (length(xfix)>0) {
            Debug("random slopes...",debug)
            nrow <- length(vars(object))
            xpos <- lapply(xfix,function(y) which(regfix(object)$labels==y))
            colpos <- lapply(xpos, function(y) ceiling(y/nrow))
            rowpos <- lapply(xpos, function(y) (y-1)%%nrow+1)
            myfix <- list(var=xfix, col=colpos, row=rowpos)
            for (i in seq_along(myfix$var))
                for (j in seq_along(myfix$col[[i]])) {
                    regfix(x0, from=vars(x0)[myfix$row[[i]][j]],to=vars(x0)[myfix$col[[i]][j]]) <-
                        data[1,myfix$var[[i]]]
                }
            index(x0) <- reindex(x0,zeroones=TRUE,deriv=TRUE)
        }
        myfun <- function(ii) {
            if (length(xfix)>0)
                for (i in seq_along(myfix$var)) {
                    index(x0)$A[cbind(myfix$row[[i]],myfix$col[[i]])] <- data[ii,myfix$var[[i]]]
                }
            return(logLikFun(x0,data=data[ii,,drop=FALSE], p=p,weights=weights[ii,,drop=FALSE],data2=data2[ii,,drop=FALSE],
                             model=model,debug=debug,indiv=indiv,...))
        }
        loglik <- sapply(seq_len(nrow(data)),myfun)
        if (!indiv) {
            loglik <- sum(loglik)
            n <- nrow(data)
            attr(loglik, "nall") <- n
            attr(loglik, "nobs") <- n
            attr(loglik, "df") <- length(p)
            class(loglik) <- "logLik"
        }
        return(loglik)
    }

    if (xconstrainM) {
        xconstrain <- c()
        for (i in seq_len(length(constrain(object)))) {
            z <- constrain(object)[[i]]
            xx <- intersect(attributes(z)$args,manifest(object))
            if (length(xx)>0) {
                warg <- setdiff(attributes(z)$args,xx)
                wargidx <- which(attributes(z)$args%in%warg)
                exoidx <- which(attributes(z)$args%in%xx)
                parname <- names(constrain(object))[i]
                y <- names(which(unlist(lapply(intercept(object),function(x) x==parname))))
                el <- list(i,y,parname,xx,exoidx,warg,wargidx,z)
                names(el) <- c("idx","endo","parname","exo","exoidx","warg","wargidx","func")
                xconstrain <- c(xconstrain,list(el))
            }
        }
        if (length(xconstrain)>0) {
            yconstrain <- unlist(lapply(xconstrain,function(x) x$endo))
            iconstrain <- unlist(lapply(xconstrain,function(x) x$idx))

            Mu <- matrix(0,nrow(data),length(vars(object))); colnames(Mu) <- vars(object)
            M <- modelVar(object,p=p,data=data)
            M$parval <- c(M$parval,  object$mean[unlist(lapply(object$mean,is.numeric))])
            for (i in seq_len(length(xconstrain))) {
                pp <- unlist(M$parval[xconstrain[[i]]$warg]);
                myidx <- with(xconstrain[[i]],order(c(wargidx,exoidx)))
                mu <- with(xconstrain[[i]],
                          apply(data[,exo,drop=FALSE],1,
                                function(x) {
                                    func(unlist(c(pp,x))[myidx])
                                }))
                Mu[,xconstrain[[i]]$endo] <- mu
            }
            offsets <- Mu%*%t(M$IAi)[,endogenous(object),drop=FALSE]
            object$constrain[iconstrain] <- NULL
            object$mean[yconstrain] <- 0
            loglik <- do.call(lname, c(list(object=object,p=p,data=data,indiv=indiv,weights=weights,data2=data2,offset=offsets),list(...)))
        } else {
            cl[[1]] <- logLikFun
            loglik <- eval.parent(cl)
        }
    } else {
        loglik <- 0
        if (length(xconstrain)>0 && NROW(data)>1) {
            for (ii in seq(nrow(data))) {
                cl$data <- data[ii,]
                cl$weights <- weights[ii,]
                cl$data2 <- data2[ii,]
                loglik <- loglik+eval.parent(cl)
            }
        } else {
            cl[[1]] <- logLikFun
            loglik <- eval.parent(cl)
        }
    }

    if (is.null(attr(loglik,"nall")))
        attr(loglik, "nall") <- n
    if (is.null(attr(loglik,"nobs")))
        attr(loglik, "nobs") <- n##-length(p)
    if (is.null(attr(loglik,"df")))
        attr(loglik, "df") <- length(p)
    class(loglik) <- "logLik"
    return(loglik)
}

##' @export
gaussian_logLik.lvm <- function(object,p,data,
                         type=c("cond","sim","exo","sat","cond2"),
                         weights=NULL, indiv=FALSE, S, mu, n, offset=NULL, debug=FALSE, meanstructure=TRUE,...) {
    exo.idx <- with(index(object), exo.obsidx)
    endo.idx <- with(index(object), endo.obsidx)
    if (type[1]=="exo") {
        if (length(exo.idx)==0 || any(is.na(exo.idx)))
            return(0)
    }

    cl <- match.call()
    if (type[1]=="cond") {
        cl$type <- "sim"
        L0 <- eval.parent(cl)
        cl$type <- "exo"
        L1 <- eval.parent(cl)
        loglik <- L0-L1
        return(loglik)
    }

    if (missing(n)) {
        if (is.vector(data)) n <- 1
        else n <- nrow(data)
    }
    k <- length(index(object)$manifest)

    if (!is.null(offset) && type[1]!="exo") {
        data[,colnames(offset)] <- data[,colnames(offset)]-offset
    }

    if (type[1]=="sat") {
        if (missing(S)) {
            d0 <- procdata.lvm(object,data=data)
            S <- d0$S; mu <- d0$mu; n <- d0$n

        }
        if (missing(p)) p <- rep(1,length(coef(object)))
        L1 <- logLik(object,p,data,type="exo",meanstructure=meanstructure)
        ##    Sigma <- (n-1)/n*S ## ML = 1/n * sum((xi-Ex)^2)
        Sigma <- S
        loglik <- -(n*k)/2*log(2*base::pi) -n/2*(log(det(Sigma)) + k) - L1
        P <- length(endo.idx)
        k <- length(exo.idx)
        npar <- P*(1+(P-1)/2)
        if (meanstructure) npar <- npar+ (P*k + P)
        attr(loglik, "nall") <- n
        attr(loglik, "nobs") <- n
        attr(loglik, "df") <- npar
        class(loglik) <- "logLik"
        return(loglik)
    }
    myidx <- switch(type[1],
                    sim =  seq_along(index(object)$manifest),
                    cond = { endo.idx },
                    cond2 = { endo.idx },
                    exo =  { exo.idx } )

    mom <- moments(object, p, conditional=(type[1]=="cond2"), data=data)

    if (!lava.options()$allow.negative.variance && any(diag(mom$P)<0)) return(NaN)

    C <- mom$C
    xi <- mom$xi
    if (type[1]=="exo") {
        C <- C[exo.idx,exo.idx,drop=FALSE]
        xi <- xi[exo.idx,drop=FALSE]
    }
    Debug(list("C=",C),debug)
    k <- nrow(C)

    iC <- Inverse(C,det=TRUE, symmetric = TRUE)
    detC <- attributes(iC)$det

    if (!is.null(weights)) {
        weights <- cbind(weights)
        K <- length(exo.idx)+length(endo.idx)
        if (ncol(weights)!=1 & ncol(weights)!=K) {
            w.temp <- weights
            weights <- matrix(1,nrow=nrow(weights),ncol=K)
            weights[,endo.idx] <- w.temp
        }
        if (type=="exo")
            weights <- NULL
    }

    notdatalist <- (!is.list(data) | is.data.frame(data))
    if (missing(n))
        if (!missing(data)) n <- NROW(data)
    if (!missing(n))
        if (notdatalist & (n<2 | indiv | !is.null(weights))) {
            if (n==1)
                data <- rbind(data)
            res <- numeric(n)
            data <- data[,index(object)$manifest,drop=FALSE]
            loglik <- 0;
            for (i in seq_len(n)) {
                ti <- as.numeric(data[i,myidx])
                if (meanstructure) {
                    ti <- cbind(ti-as.numeric(xi))
                }
                if (!is.null(weights)) {
                    W <- diag(weights[i,],nrow=length(weights[i,]))
                    val <- -k/2*log(2*base::pi) -1/2*log(detC) - 1/2*(t(ti)%*%W)%*%iC%*%(ti)
                } else {
                    val <- -k/2*log(2*base::pi) -1/2*log(detC) - 1/2*t(ti)%*%iC%*%(ti)
                }
                if (indiv)
                    res[i] <- val
                loglik <- loglik + val
            }
            if (indiv)
                return(res)
        } else {
            if (missing(S)) {
                d0 <- procdata.lvm(object,data=data)
                S <- d0$S; mu <- d0$mu; n <- d0$n
            }
            S <- S[myidx,myidx,drop=FALSE]
            mu <- mu[myidx,drop=FALSE]
            T <- S
            if (meanstructure) {
                W <- crossprod(rbind(mu-xi))
                T <- S+W
            }
            loglik <- -(n*k)/2*log(2*base::pi) -n/2*(log(detC) + tr(T%*%iC))
        }
    return(loglik)
}

##' @export
logLik.lvmfit <- function(object,
                   p=coef(object),
                   data=model.frame(object),
                   model=object$estimator,
                   weights=Weights(object),
                   data2=object$data$data2,
                   ...) {

    logLikFun <- paste0(model,"_logLik.lvm")
    if (!exists(logLikFun)) {
        model <- "gaussian"
    }
    l <- logLik.lvm(object$model0,p,data,model=model,weights=weights,
                    data2=data2,
                    ...)
    return(l)
}

##' @export
logLik.lvm.missing <- function(object,
                        p=pars(object), model=object$estimator,
                        weights=Weights(object$estimate),
                        ...) {
    logLik(object$estimate$model0, p=p, model=model, weights=weights, ...)
}

##' @export
logLik.multigroup <- function(object,p,data=object$data,weights=NULL,type=c("cond","sim","exo","sat"),...) {
    res <- procrandomslope(object)
    pp <- with(res, modelPar(model,p)$p)

    if (type[1]=="sat") {
        n <- 0
        df <- 0
        loglik <- 0
        for (i in seq_len(object$ngroup)) {
            m <- Model(object)[[i]]
            L <- logLik(m,p=pp[[i]],data=object$data[[i]],type="sat")
            df <- df + attributes(L)$df
            loglik <- loglik + L
            n <- n + object$samplestat[[i]]$n
        }
        attr(loglik, "nall") <- n
        attr(loglik, "nobs") <- n##-df
        attr(loglik, "df") <- df
        class(loglik) <- "logLik"
        return(loglik)
    }

    n <- 0
    loglik <- 0; for (i in seq_len(object$ngroup)) {
                     n <- n + object$samplestat[[i]]$n
                     val <- logLik(object$lvm[[i]],pp[[i]],data[[i]],weights=weights[[i]],type=type,...)
                     loglik <- loglik + val
                 }
    attr(loglik, "nall") <- n
    attr(loglik, "nobs") <- n##-length(p)
    attr(loglik, "df") <- length(p)
    class(loglik) <- "logLik"
    return(loglik)
}

##' @export
logLik.multigroupfit <- function(object,
                          p=pars(object), weights=Weights(object), model=object$estimator, ...) {
    logLik(object$model0,p=p,weights=weights,model=model,...)
}

##' Calculate Wald og Likelihood based (profile likelihood) confidence intervals
##'
##' Calculates either Wald confidence limits: \deqn{\hat{\theta} \pm
##' z_{\alpha/2}*\hat\sigma_{\hat\theta}} or profile likelihood confidence
##' limits, defined as the set of value \eqn{\tau}:
##' \deqn{logLik(\hat\theta_{\tau},\tau)-logLik(\hat\theta)< q_{\alpha}/2}
##'
##' where \eqn{q_{\alpha}} is the \eqn{\alpha} fractile of the \eqn{\chi^2_1}
##' distribution, and \eqn{\hat\theta_{\tau}} are obtained by maximizing the
##' log-likelihood with tau being fixed.
##'
##' @title Calculate confidence limits for parameters
##' @param object \code{lvm}-object.
##' @param parm Index of which parameters to calculate confidence limits for.
##' @param level Confidence level
##' @param profile Logical expression defining whether to calculate confidence
##' limits via the profile log likelihood
##' @param curve if FALSE and profile is TRUE, confidence limits are
##' returned. Otherwise, the profile curve is returned.
##' @param n Number of points to evaluate profile log-likelihood in
##' over the interval defined by \code{interval}
##' @param interval Interval over which the profiling is done
##' @param lower If FALSE the lower limit will not be estimated (profile intervals only)
##' @param upper If FALSE the upper limit will not be estimated (profile intervals only)
##' @param \dots Additional arguments to be passed to the low level functions
##' @return A 2xp matrix with columns of lower and upper confidence limits
##' @author Klaus K. Holst
##' @seealso \code{\link{bootstrap}{lvm}}
##' @keywords models regression
##' @examples
##'
##' m <- lvm(y~x)
##' d <- sim(m,100)
##' e <- estimate(lvm(y~x), d)
##' confint(e,3,profile=TRUE)
##' confint(e,3)
##' \dontrun{ ## Reduce Ex.timings
##' B <- bootstrap(e,R=50)
##' B
##' }
##' @aliases confint.multigroupfit
##' @export
##' @method confint lvmfit
confint.lvmfit <- function(object,parm=seq_len(length(coef(object))),level=0.95,profile=FALSE,curve=FALSE,n=20,interval=NULL,lower=TRUE,upper=TRUE,...) {
    if (is.character(parm)) {
        parm <- parpos(Model(object),p=parm)
        parm <- parm[attributes(parm)$ord]
    }
    if (!profile) {
        return(confint.default(object,parm=parm,level=level,...))
    }
    res <- c()
    for (i in parm) {
        res <- rbind(res, profci.lvmfit(object,parm=i,level=level,profile=profile,n=n,curve=curve,interval=interval,lower=lower,upper=upper,...))
        if (curve) return(res)
    }
    rownames(res) <- names(coef(object))[parm]
    colnames(res) <- paste((c(0,1) + c(1,-1)*(1-level)/2)*100,"%")
    return(res)
}


##' @export
confint.multigroupfit <- function(object,parm=seq_along(pars(object)),level=0.95,
                           ...) {
  p <- 1-(1-level)/2
  res <- cbind(pars(object),pars(object)) + qnorm(p)*cbind(-1,1)%x%diag(vcov(object))^0.5
  colnames(res) <- paste0(c(1-p,p)*100,"%")
  rownames(res) <- names(coef(object))
  res[parm,,drop=FALSE]
}

##' @export
rsq <- function(x,stderr=FALSE) {

    if (stderr) {
        v <- endogenous(x)
        vpar <- paste(v,v,sep=lava.options()$symbol[2])
        ic.v <- IC(model.frame(x)[,v])
        ic.mod <- IC(x)
        coef0 <- c(attributes(ic.v)$coef[vpar],
                   coef(x)[vpar])
        ic0 <- cbind(ic.v[,vpar],ic.mod[,vpar])
        p <- length(v)
        idx <- seq_len(p);
        ee <- estimate(NULL,data=NULL,
                       function(x) {
                           res <- (x[idx]-x[idx+p])/x[idx]
                           names(res) <- v
                           as.list(res)
                       },
                       print=function(x,...) {
                           cat("\nR-squared:\n\n")
                           print(x$coefmat)
                       },
                       coef=coef0, IC=ic0)

        res <- ee
        ##res <- list(ee)
        ## for (lat in latent(x)) {

        ##     v <- intersect(children(x,lat),endogenous(x))
        ##     vpar <- paste(v,v,sep=lava.options()$symbol[2])
        ##     lpar <- paste(lat,lat,sep=lava.options()$symbol[2])
        ##     rpar <- paste(v,lat,sep=lava.options()$symbol[1])
        ##     fix <- c(x$model$fix[lat,v,drop=TRUE],x$model$covfix[lat,lat])
        ##     pp <- coef(x)
        ##     idx <- x$model$parpos$A[lat,v]
        ##     idx2 <- x$model$parpos$P[lat,lat]
        ##     p0 <- c(idx,idx2)
        ##     p1 <- setdiff(unique(p0),0)
        ##     p2 <- match(p0,p1)

        ##     k <- length(v)
        ##     coef0 <- c(pp[p1],attributes(ic.v)$coef[vpar])
        ##     ic0 <- cbind(ic.mod[,p1],ic.v[,vpar])
        ##     ee <- estimate(NULL,data=NULL,
        ##                    function(p) {
        ##                        p. <- p[p2]
        ##                        p.[is.na(p.)] <- fix[is.na(p.)]
        ##                        res <- p.[seq_len(k)]^2*p.[k+1]/tail(p,k)
        ##                        names(res) <- v
        ##                        as.list(res)
        ##                    },
        ##                    print=function(x,...) {
        ##                        cat("\nVariance explained by '", lat,"':\n\n",sep="")
        ##                        print(x$coefmat)
        ##                    },coef=coef0,IC=ic0)
        ##     res <- c(res,list(ee))
        ## }

        return(res)
    }



    v <- c(endogenous(x),setdiff(latent(x),parameter(Model(x))))
    res <- coef(x,9,std="yx")
    idx <- with(attributes(res),
                which(type=="variance" & (var==from)))
    nam <- attributes(res)$var[idx]
    res <- 1-res[idx,5]
    names(res) <- nam
    res <- list("R-squared"=res)
    ## M <- moments(x,coef(x))
    ## v <- setdiff(vars(x),exogenous(x))
    ## vvar <- M$Cfull[cbind(v,v)]
    ## rsq <- (vvar-M$P[cbind(v,v)])/vvar

    if (length(latent(x))>0) {
        M <- moments(x,coef(x))
        nn <- names(res)
        for (lat in latent(x)) {
            v <- intersect(children(x,lat),endogenous(x))
            varl <- M$Cfull[lat,lat]
            varv <- M$Cfull[cbind(v,v)]
            fix <- c(x$model$fix[lat,v,drop=TRUE])
            pp <- coef(x)
            if (inherits(x,"lvm.missing")) {
                mp <- match(coef(x$model),names(coef(x)))
                pp <- pp[mp]
            }
            idx1 <- x$model$parpos$A[lat,v]
            ##idx2 <- x$model$parpos$P[lat,lat]
            ##idx3 <- x$model$parpos$P[cbind(v,v)]
            p0 <- c(idx1)
            p1 <- setdiff(unique(p0),0)
            p2 <- match(p0,p1)
            p <- pp[p1]
            p. <- p[p2]
            p.[is.na(p.)] <- fix[is.na(p.)]
            k <- length(v)
            val <- (p.^2*varl)/varv; names(val) <- v
            res <- c(res,list(val))
            nn <- c(nn,paste0("Variance explained by '",lat,"'"))
        }
        names(res) <- nn
    }
    res
}

satmodel <- function(object,logLik=TRUE,data=model.frame(object),
              control=list(trace=1),
              weights=Weights(object),estimator=object$estimator,
              missing=inherits(object,"lvm.missing"),
              regr=FALSE,
              ...) {
    if (object$estimator=="gaussian" & logLik & !missing) {
        if (inherits(object, c("multigroupfit","multigroup"))) {

            ll <- structure(0,nall=0,nobs=0,df=0,class="logLik")
            for (i in seq_len(Model(object)$ngroup)) {
                l0 <- logLik(Model(Model(object))[[i]],data=model.frame(object)[[i]],type="sat")

                ll <- ll+l0
                for (atr in c("nall","nobs","df"))
                    attributes(ll)[[atr]] <- attributes(ll)[[atr]]+attributes(l0)[[atr]]
            }

        }
        return(logLik(object, type="sat"))
    }
    covar <- exogenous(object)
    y <- endogenous(object)
    m0 <- Model(object)
    if (length(covar)>0)
        suppressWarnings(m0 <- regression(m0,y,covar))
    if (length(latent(m0))>0)
        kill(m0) <- latent(m0)
    cancel(m0) <- y
    if (!regr)
        suppressWarnings(covariance(m0) <- y)
    else {
        if (length(y)>1) {
            for (i in seq_len(length(y)-1))
                for (j in seq(i+1,length(y))) {
                    m0 <- regression(m0,y[i],y[j])
                }
        }
        exogenous(m0) <- covar
    }
    if (is.null(control$start)) {
        mystart <- rep(0,with(index(m0), npar.mean+npar))
        mystart[variances(m0,mean=TRUE)] <- 1
        control$start <- mystart
    }
    message("Calculating MLE of saturated model:\n")
    e0 <- estimate(m0,data=data,weights=weights,estimator=estimator,messages=0,control=control,missing=missing,...)
    if (logLik)
        return(logLik(e0))
    return(e0)
}

condition <- function(x) {
    a <- svd(x)
    return(max(a$d)/min(a$d))
}


##' Extract model summaries and GOF statistics for model object
##'
##' Calculates various GOF statistics for model object including global
##' chi-squared test statistic and AIC. Extract model-specific mean and variance
##' structure, residuals and various predicitions.
##'
##'
##' @aliases gof gof.lvmfit moments moments.lvm information.lvmfit
##' @aliases score.lvmfit logLik.lvmfit
##' @seealso [information], [score]
##' @param object Model object
##' @param x Model object
##' @param p Parameter vector used to calculate statistics
##' @param data Data.frame to use
##' @param latent If TRUE predictions of latent variables are included in output
##' @param data2 Optional second data.frame (only for censored observations)
##' @param weights Optional weight matrix
##' @param n Number of observations
##' @param conditional If TRUE the conditional moments given the covariates are
##' calculated. Otherwise the joint moments are calculated
##' @param model String defining estimator, e.g. "gaussian" (see
##' \code{estimate})
##' @param debug Debugging only
##' @param chisq Boolean indicating whether to calculate chi-squared
##' goodness-of-fit (always TRUE for estimator='gaussian')
##' @param level Level of confidence limits for RMSEA
##' @param rmsea.threshold Which probability to calculate, Pr(RMSEA<rmsea.treshold)
##' @param all Calculate all (ad hoc) FIT indices: TLI, CFI, NFI, SRMR, ...
##' @param \dots Additional arguments to be passed to the low level functions
##' @usage
##'
##' gof(object, ...)
##'
##' \method{gof}{lvmfit}(object, chisq=FALSE, level=0.90, rmsea.threshold=0.05,all=FALSE,...)
##'
##' moments(x,...)
##'
##' \method{moments}{lvm}(x, p, debug=FALSE, conditional=FALSE, data=NULL, latent=FALSE, ...)
##'
##' \method{logLik}{lvmfit}(object, p=coef(object),
##'                       data=model.frame(object),
##'                       model=object$estimator,
##'                       weights=Weights(object),
##'                       data2=object$data$data2,
##'                           ...)
##'
##' \method{score}{lvmfit}(x, data=model.frame(x), p=pars(x), model=x$estimator,
##'                    weights=Weights(x), data2=x$data$data2, ...)
##'
##' \method{information}{lvmfit}(x,p=pars(x),n=x$data$n,data=model.frame(x),
##'                    model=x$estimator,weights=Weights(x), data2=x$data$data2, ...)
##'
##' @return A \code{htest}-object.
##' @author Klaus K. Holst
##' @keywords methods models
##' @export
##' @examples
##' m <- lvm(list(y~v1+v2+v3+v4,c(v1,v2,v3,v4)~x))
##' set.seed(1)
##' dd <- sim(m,1000)
##' e <- estimate(m, dd)
##' gof(e,all=TRUE,rmsea.threshold=0.05,level=0.9)
##'
##'
##' set.seed(1)
##' m <- lvm(list(c(y1,y2,y3)~u,y1~x)); latent(m) <- ~u
##' regression(m,c(y2,y3)~u) <- "b"
##' d <- sim(m,1000)
##' e <- estimate(m,d)
##' rsq(e)
##' ##'
##' rr <- rsq(e,TRUE)
##' rr
##' estimate(rr,contrast=rbind(c(1,-1,0),c(1,0,-1),c(0,1,-1)))
##'
`gof` <-
    function(object,...) UseMethod("gof")

##' @export
gof.lvmfit <- function(object,chisq=FALSE,level=0.90,rmsea.threshold=0.05,all=FALSE,...) {
    n <- object$data$n
    if (inherits(object, "multigroupfit")) n <- sum(unlist(lapply(object$model$data,nrow)))
    loglik <- logLik(object,...)

    df <- attributes(loglik)$df
    nobs <- attributes(loglik)$nall*length(endogenous(object))
    myAIC <- -2*(loglik - df); attributes(myAIC) <- NULL
    myBIC <- -2*loglik + df*log(nobs); attributes(myBIC) <- NULL

    xconstrain <- intersect(unlist(lapply(constrain(object),function(z) attributes(z)$args)),manifest(object))

    l2D <- mean(object$opt$grad^2)
    S <- vcov(object)
    rnkV <- tryCatch(qr(S)$rank,error=function(...) 0)
    minSV <- attr(S,"minSV")
    condnum <- tryCatch(condition(vcov(object)),error=function(...) NULL)

    if (((object$estimator=="gaussian" & !inherits(object, "lvm.missing")) | chisq) & length(xconstrain)==0 ) {
        res <- list(fit=compare(object), n=n, logLik=loglik, BIC=myBIC, AIC=myAIC)
        q <- res$fit$statistic
        qdf <- res$fit$parameter
        if (all) {
            m0 <- lvm(manifest(object)); exogenous(m0) <- NULL
            e0 <- estimate(m0,model.frame(object))
            g0 <- gof(e0)
            qbaseline <- g0$fit$statistic
            qdfbaseline <- g0$fit$parameter
            CFI <- ((qbaseline-qdfbaseline) - (q-qdf))/(qbaseline-qdfbaseline)
            NFI <- (qbaseline-q)/qbaseline
            TLI <- (qbaseline/qdfbaseline-q/qdf)/(qbaseline/qdfbaseline-1)
            S <- object$data$S
            mu <- object$data$mu
            C <- modelVar(object)$C
            xi <- as.vector(modelVar(object)$xi)
            if (is.null(S)) S <- cov(model.frame(object))
            if (is.null(mu)) mu <- colMeans(model.frame(object))
            L <- diag(S)^0.5
            R <- (diag(1/L))%*%(S-C)%*%(diag(1/L))
            R2 <- (mu-xi)/L
            SRMR <- mean(c(R[upper.tri(R,diag=TRUE)],R2)^2)^0.5
            res <- c(res,list(CFI=CFI,NFI=NFI,TLI=TLI,C=C,S=S,SRMR=SRMR))
            ## idx <- index(object)$endo.idx
            ## if (length(latent(object))>0) {
            ##   SRMR.endo <- mean(c(R[idx,idx][upper.tri(R[idx,idx],diag=TRUE)],R2[idx])^2)^0.5
            ##   res <- c(res,list("SRMR(endogenous)"=SRMR.endo))
            ## }
        }
        if (rnkV==ncol(vcov(object)) && (!is.null(minSV) && minSV>1e-12)) {

            rmseafun <- function(...) {
                epsilon <- function(lambda) sapply(lambda,function(x)
                    ifelse(x>0 & qdf>0,sqrt(x/(qdf*(n))),0)) ## n-1,n vs. n-df
                opf <- function(l,p) suppressWarnings(p-pchisq(q,df=qdf,ncp=l))
                alpha <- (1-level)/2
                RMSEA <- epsilon(q-qdf)
                B <- max(q-qdf,0)
                lo <- hi <- list(root=0)
                if (RMSEA>0 && opf(0,p=1-alpha)<0) {
                    hi <- uniroot(function(x) opf(x,p=1-alpha),c(0,B))
                }
                if (opf(B,p=alpha)<0) {
                    lo <- uniroot(function(x) opf(x,p=alpha),c(B,n))
                }
                ci <- c(epsilon(c(hi$root,lo$root)))
                RMSEA <- c(RMSEA=RMSEA,ci);
                names(RMSEA) <- c("RMSEA",paste0(100*c(alpha,(1-alpha)),"%"))
                pval <- pchisq(q,qdf,(n*qdf*rmsea.threshold^2),lower.tail=FALSE)
                res <- list(aa=((q-qdf)/(2*qdf)^0.5),RMSEA=RMSEA, level=level, rmsea.threshold=rmsea.threshold, pval.rmsea=pval)
                return(res)
            }
            rmseaval <- tryCatch(rmseafun(),error=function(e) NULL)
            res <- c(res,rmseaval)
        }
    } else {
        res <- list(n=n, logLik=loglik, BIC=myBIC, AIC=myAIC)
    }

    res <- c(res, L2score=l2D, rankV=rnkV, cond=condnum, k=nrow(vcov(object)))
    class(res) <- "gof.lvmfit"
    return(res)
}

##' @export
print.gof.lvmfit <- function(x,optim=TRUE,...) {
    if (!is.null(x$n)) {
        with(x,
             cat("\n Number of observations =", n, "\n"))
    }
    if (is.null(x$fit)) {
        with(x,
             cat(" Log-Likelihood =", logLik, "\n"))
    }
    with(x,  cat(" BIC =", BIC, "\n",
                 "AIC =", AIC, "\n"))
    if (!is.null(x$fit))
        with(x,
             cat(" log-Likelihood of model =", fit$estimate[1], "\n\n",
                 "log-Likelihood of saturated model =", fit$estimate[2], "\n",
                 "Chi-squared statistic: q =", fit$statistic,
                 ", df =", fit$parameter,
                 "\n  P(Q>q) =", fit$p.value, "\n"))
    if (!is.null(x$RMSEA)) {
        rr <- round(x$RMSEA*10000)/10000
        rmsea <- paste0(rr[1]," (",rr[2],";",rr[3],")")
        cat("\n RMSEA (",x$level*100,"% CI): ", rmsea,"\n",sep="")
        cat("  P(RMSEA<",x$rmsea.threshold,")=",  x$pval.rmsea,"\n",sep="")
    }
    for (i in c("TLI","CFI","NFI","SRMR","SRMR(endogenous)"))
        if (!is.null(x[[i]])) cat("", i,"=",x[[i]],"\n")

    if (optim) {
        cat("\nrank(Information) = ",x$rankV," (p=", x$k,")\n",sep="")
        cat("condition(Information) = ",x$cond,"\n",sep="")
        cat("mean(score^2) =",x$L2score,"\n")
    }

    invisible(x)
}



## gof.multigroupfit <- function(object,...) {
##   L0 <- logLik(object); df0 <- attributes(L0)$df
##   L1 <- logLik(object,type="sat"); df1 <- attributes(L1)$df

##   df <- df1-df0; names(df) <- "df"
##   Q <- -2*(L0-L1); attributes(Q) <- NULL; names(Q) <- "chisq";
##   pQ <- pchisq(Q,df,lower.tail=FALSE)
##   values <- c(L0,L1); names(values) <- c("log likelihood (model)", "log likelihood (saturated model)")
##   res <- list(statistic = Q, parameter = df,
##               p.value=pQ, method = "Likelihood ratio test",
##               estimate = values)
##   class(res) <- "htest"
##   return(res)
## }

gfilter <- function(x,sigma=1) {
  gridfn <- function(fn,width,height,center=TRUE) {
    jx <- seq_len(height)
    jy <- seq_len(width)
    if (center) {
      jx <- jx/height-0.5
      jy <- jy/width-0.5
    }
    outer(jx, jy, FUN=fn)
  }
  width <- ncol(x); height <- nrow(x)
  oscunits <- gridfn(function(x,y) ((-1)^(x+y)),height=height,width=width,center=FALSE)
  x0 <- x*oscunits ## translate origo to center of image
  X <- fft(x0)
  d <- gridfn(function(x,y) (x^2+y^2),height=height,width=width,center=TRUE)
  Gn <- exp(-2*(base::pi*sigma)^2*d) # frequency response
  H <- X*Gn
  res <- Re(fft(H,inverse=TRUE))/(width*height)*oscunits
  return(res)
}

##' @export
lava <- function(seed,w=128,h=w,bw=4,sigma=5000,bg=20000,numcol=128,col=grDevices::heat.colors(numcol),...) {
  if (!missing(seed))
    set.seed(seed)
  x <- matrix(rnorm(w*h,bg,sigma),nrow=h, ncol=w)
  x0 <- gfilter(x,sigma=bw)
  y <- (x0-min(x0)+1)^1.2
  opt <- graphics::par(mai=c(0,0,0,0))
  graphics::image(y,axes=FALSE,col=col)
  graphics::par(opt)
  invisible(y)
}

mypal <- function(set=TRUE,...) {
    oldpal <- palette()
    col <- c("black","darkblue","darkred","goldenrod","mediumpurple",
             "seagreen","aquamarine3","violetred1","salmon1",
             "lightgoldenrod1","darkorange2","firebrick1","violetred1", "gold")
    if (!set) return(col)
    palette(col)
    invisible(oldpal)
}


##' This function transforms a standard color (e.g. "red") into an
##' transparent RGB-color (i.e. alpha-blend<1).
##'
##' This only works for certain graphics devices (Cairo-X11 (x11 as of R>=2.7),
##' quartz, pdf, ...).
##' @title Generate a transparent RGB color
##' @param col color (numeric or character)
##' @param alpha degree of transparency (0-1)
##' @param locate Choose colour (with mouse)
##' @return A character vector with elements of 7 or 9 characters, `#` followed
##'   by the red, blue, green and optionally alpha values in hexadecimal (after
##'   rescaling to '0 ... 255').
##' @author Klaus K. Holst
##' @examples
##' plot(runif(1000),cex=runif(1000,0,4),
##'      col=Col(c("darkblue","orange"),0.5),pch=16)
##' @keywords color
##' @export
Col <- function(col,alpha=0.2,locate=0) {
    if (locate>0) return(colsel(locate))

    mapply(function(x,alpha)
        do.call(rgb,as.list(c(col2rgb(x)/255,alpha))),
        col,alpha)
}

##' Extract subset of latent variable model
##'
##' Extract measurement models or user-specified subset of model
##'
##'
##' @aliases measurement
##' @param x \code{lvm}-object.
##' @param vars Character vector or formula specifying variables to include in
##' subset.
##' @param \dots Additional arguments to be passed to the low level functions
##' @return A \code{lvm}-object.
##' @author Klaus K. Holst
##' @keywords models regression
##' @examples
##'
##' m <- lvm(c(y1,y2)~x1+x2)
##' subset(m,~y1+x1)
##'
##' @export
##' @method subset lvm
subset.lvm <- function(x, vars, ...) {
    if (missing(vars)) return(x)
    if (inherits(vars,"formula")) vars <- all.vars(vars)
    if (!all(vars%in%vars(x))) stop("Not a subset of model")
    latentvars <- intersect(vars,latent(x))
    ##  g0 <- subGraph(vars, Graph(x))
    ##  res <- graph2lvm(g0)
    res <- lvm(vars)
    M <- t(x$M[vars,vars,drop=FALSE])
    for (i in seq_len(nrow(M))) {
        if (any(M[,i]==1)) {
            res <- regression(res, y=rownames(M)[M[,i]==1], x=rownames(M)[i], ...)
        }
    }
    if (length(latentvars)>0)
        latent(res) <- latentvars
    res$cov[vars,vars] <- x$cov[vars,vars]
    ## Fixed parameters:
    res$par[vars,vars] <- x$par[vars,vars]
    res$fix[vars,vars] <- x$fix[vars,vars]
    res$covpar[vars,vars] <- x$covpar[vars,vars]
    res$covfix[vars,vars] <- x$covfix[vars,vars]
    res$mean[vars] <- x$mean[vars]
    res$attributes <- x$attributes
    for (i in seq_along(x$attributes)) {
        val <- x$attributes[[i]]
        if (length(val)>0) {
            val <- val[intersect(vars,names(val))]
            res$attributes[[i]] <- val
        }
    }
    index(res) <- reindex(res)
    return(res)
}

##' @export
"functional<-" <- function(x,...,value) UseMethod("functional<-")

##' @export
"functional<-.lvm" <- function(x,to,from,...,value) {
    if (inherits(to,"formula")) {
        yy <- decomp.specials(getoutcome(to))
        myvars <- all.vars(to)
        xx <- setdiff(myvars,yy)
        if (length(yy)*length(xx)>length(value) & length(value)!=1) stop("Wrong number of values")
        count <- 0
        for (y in yy) {
            count <- count+1
            for (i in seq_along(xx)) {
                suppressWarnings(x <- regression(x,to=y,from=xx[i],messages=0))
                count <- count+1
                if (length(value)==1) {
                    functional(x, to=y, from=xx[i],...) <- value
                } else
                    functional(x, to=y, from=xx[i],...) <- value[[count]]
            }
        }
        return(x)
    }

    if (missing(from) | missing(to))
        return(x)

    edges <- paste(from,to,sep="~")
    x$attributes$functional[[edges]] <- value
    return(x)
}

##' @export
"functional" <- function(x,...) UseMethod("functional")

##' @export
functional.lvm <- function(x,to,from,value,...) {
    if (!missing(value)) {
        functional(x,to,from,...) <- value
        return(x)
    }
    if (missing(from))
        return(x$attributes$functional)

    edges <- paste(from,to,sep="~")
    x$attributes$functional[edges]
}

##' Plot regression line (with interactions) and partial residuals.
##'
##' @title Plot regression lines
##' @param model Model object (e.g. [lm()])
##' @param var1 predictor (Continuous or factor)
##' @param var2 Factor that interacts with `var1`
##' @param data data.frame to use for prediction (model.frame is used as default)
##' @param ci.lty Line type for confidence limits
##' @param ci Boolean indicating wether to draw pointwise 95% confidence limits
##' @param level Level of confidence limits (default 95%)
##' @param pch Point type for partial residuals
##' @param lty Line type for estimated regression lines
##' @param lwd Line width for regression lines
##' @param npoints Number of points used to plot curves
##' @param xlim Range of x axis
##' @param col Color (for each level in `var2`)
##' @param colpt Color of partial residual points
##' @param alpha Alpha level
##' @param cex Point size
##' @param delta For categorical `var1`
##' @param centermark For categorical `var1`
##' @param jitter For categorical `var1`
##' @param cidiff For categorical `var1`
##' @param mean For categorical `var1`
##' @param legend Boolean (add legend)
##' @param trans Transform estimates (e.g. exponential)
##' @param partres Boolean indicating whether to plot partial residuals
##' @param partse .
##' @param labels Optional labels of `var2`
##' @param vcov Optional variance estimates
##' @param predictfun Optional predict-function used to calculate confidence limits and predictions
##' @param plot If FALSE return only predictions and confidence bands
##' @param new If FALSE add to current plot
##' @param \dots additional arguments to lower level functions
##' @return list with following members:
##' * `x` Variable on the x-axis `var1`
##' * `y` Variable on the y-axis (partial residuals)
##' * `predict` Matrix with confidence limits and predicted startvalues.R
##' @author Klaus K. Holst
##' @seealso [termplot()]
##' @export
##' @examples
##' n <- 100
##' x0 <- rnorm(n)
##' x1 <- seq(-3,3, length.out=n)
##' x2 <- factor(rep(c(1,2),each=n/2), labels=c("A","B"))
##' y <- 5 + 2*x0 + 0.5*x1 + -1*(x2=="B")*x1 + 0.5*(x2=="B") + rnorm(n, sd=0.25)
##' dd <- data.frame(y=y, x1=x1, x2=x2)
##' lm0 <- lm(y ~ x0 + x1*x2, dd)
##' plotConf(lm0, var1="x1", var2="x2")
##' abline(a=5,b=0.5,col="red")
##' abline(a=5.5,b=-0.5,col="red")
##' ## points(5+0.5*x1 -1*(x2=="B")*x1 + 0.5*(x2=="B") ~ x1, cex=2)
##'
##' data(iris)
##' l <- lm(Sepal.Length ~ Sepal.Width*Species,iris)
##' plotConf(l,var2="Species")
##' plotConf(l,var1="Sepal.Width",var2="Species")
##' 
##' \dontrun{
##' ## lme4 model
##' dd$Id <- rbinom(n, size = 3, prob = 0.3)
##' lmer0 <- lme4::lmer(y ~ x0 + x1*x2 + (1|Id), dd)
##' plotConf(lmer0, var1="x1", var2="x2")
##' }
##' @keywords hplot regression
plotConf <- function(model,
                     var1=NULL,
                     var2=NULL,
                     data=NULL,
                     ci.lty=0,
                     ci=TRUE, level=0.95,
                     pch=16, lty=1, lwd=2,
                     npoints=100,
                     xlim,
                     col=NULL,
                     colpt,
                     alpha=0.5,
                     cex=1,
                     delta=0.07,
                     centermark=0.03,
                     jitter=0.2,
                     cidiff=FALSE,
                     mean=TRUE,
                     legend=ifelse(is.null(var1),FALSE,"topright"),
                     trans=function(x) {x},
                     partres=inherits(model,"lm"),
                     partse=FALSE,
                     labels,
                     vcov,
                     predictfun,
                     plot=TRUE,
                     new=TRUE,
                     ...) {


    if (inherits(model,"formula")) model <- lm(model,data=data,...)
    if (inherits(model,"lmerMod")) {
        intercept <- lme4::fixef(model)["(Intercept)"]
    } else {
        intercept <- coef(model)["(Intercept)"]
    }
    if (is.na(intercept)) intercept <- 0

    if (is.null(data)) {
        curdata <- get_all_vars(model,data=model.frame(model))
    } else {
        curdata <- get_all_vars(formula(model), data=data)
    }
    
    if (inherits(model,"lmerMod")) {
      curdata0 <- model.frame(model, data = data, fixed.only = FALSE)
    } else {
      curdata0 <- model.frame(model,data) ## Checking for factors
    }
    
    
    if (is.null(var1) && is.null(var2)) {
        var1 <- colnames(curdata)[2]
        var10 <- colnames(curdata0)[2]
    } else var10 <- var1
        
    responseorig <- colnames(curdata)[1]
    if (inherits(curdata0[,var10],c("character","factor"))) {
        curdata <- curdata0
        var2 <- var10; var1 <- NULL
    }
    dots <- list(...)


    response <- all.vars(formula(model))[1]
    cname <- colnames(curdata)[-1]
    if (!is.factor(curdata[,var2]) & !is.null(var2)) {
        curdata[,var2] <- as.factor(curdata[,var2])
        colnames(curdata)[1] <- response
        model <- update(model,as.formula(paste(response,"~.")),data=curdata)
    }
    thelevels <- levels(curdata[,var2])
    if (missing(labels)) labels <- thelevels

    k <- ifelse(is.null(var2),1,length(thelevels))
    if (is.null(col)) {
        col <- c("black","darkblue","darkred","goldenrod","mediumpurple",
                 "seagreen","aquamarine3","violetred1","salmon1",
                 "lightgoldenrod1","darkorange2","firebrick1","violetred1", "gold")
    }

    if (missing(xlim)) {
        if (!is.null(var1))
            xlim <- range(curdata[,var1])
        else
            xlim <- c(0,length(thelevels))+0.5
    }
    dots$xlim <- xlim

    if (is.null(var1) & !is.null(var2)) {
        ##npoints <- 1
        x <- unique(curdata[,var2])
        npoints <- 1#length(x)
    } else {
        x <- seq(xlim[1], xlim[2], length.out=npoints)
    }
    xx <- c()

    newdata <- data.frame(id=seq(npoints))
    partdata <- curdata
    var1.idx <- 0
    ii <- 1
    for (nn in cname) {
        ii <- ii+1
        v <- curdata[,nn]
        if (!is.null(var1) && nn==var1) {
            var1.idx <- ii
            newdata <- cbind(newdata, rep(x, k))
            partdata[,nn] <- 0
        } else {
            if (is.factor(v)) {
                if (nn%in%var2) {
                    newdata <- cbind(newdata, factor(rep(levels(v), each=npoints),levels=thelevels))
                    partdata[,nn] <- factor(rep(levels(v)[1], nrow(partdata)),levels=levels(v))
                } else {
                    newdata <- cbind(newdata, factor(rep(levels(v)[1], k*npoints),
                                                     levels=levels(v)))
                }
            } else {
                if (is.logical(v))
                    newdata <- cbind(newdata,FALSE)
                else
                    newdata <- cbind(newdata,0)
            }
        }
    };

    colnames(newdata) <- c("_id", cname)

    partdata[,response] <- newdata[,response] <- 0
    atr <- c("terms")
    attributes(newdata)[atr] <- attributes(curdata)[atr]
    attributes(partdata)[atr] <- attributes(partdata)[atr]
    Y <- model.frame(model)[,1]
    if(inherits(Y,"Surv")) Y <- Y[,1]
    XX <- model.matrix(formula(terms(model)),data=newdata)
    if (inherits(model,"lmerMod")) {
        bb <- lme4::fixef(model)
    } else {
        bb <- coef(model)
    }
    if (!missing(vcov)) SS <- vcov else {
        if (inherits(model,"geeglm")) {
            SS <- (summary(model)$cov.unscaled)
        } else {
            SS <- as.matrix(stats::vcov(model))
        }
    }
    bidx <- which(apply(XX,2,function(x) !all(x==0)))
    notbidx <- setdiff(seq(length(bb)),bidx)
    bb0 <- bb; bb0[notbidx] <- 0
    myse <- apply(XX[,bidx,drop=FALSE],1,function(x) rbind(x)%*%SS[bidx,bidx,drop=FALSE]%*%cbind(x))^.5
    ci.all <- list(fit=XX%*%bb0,se.fit=myse)
    z <- qnorm(1-(1-level)/2)
    ci.all$fit <- cbind(ci.all$fit,ci.all$fit-z*ci.all$se.fit,ci.all$fit+z*ci.all$se.fit)

    if (!missing(predictfun)) {
        R <- Y-predict(model, newdata=partdata)
        ci.all <- predict(model, newdata=newdata, se.fit=TRUE, interval = "confidence", level=level,...)
    } else {
        XX0 <- model.matrix(formula(terms(model)),data=partdata)
        R <- Y-XX0%*%bb
    }
    if (inherits(model,"lmerMod")) {
        uz <- as.matrix(unlist(lme4::ranef(model))%*%do.call(rbind,lme4::getME(model,"Ztlist")))[1,]
        R <- R-uz
    }

    pr <- trans(intercept + R)
    if (is.na(intercept)) {
        intercept <- 0
        if (!is.null(var2)) {
            intercept <- coef(model)[paste0(var2,thelevels)][as.numeric(curdata[,var2])]
        }
    }

    if (is.null(dots$ylim)) {
        if (partres) {
            if (cidiff)
                dots$ylim <- range(pr)
            else
                dots$ylim <- range(trans(c(ci.all$fit)),pr)
        }
        else
            dots$ylim <- trans(range(ci.all$fit))
    }
    if (is.null(dots$ylab))
        dots$ylab <- responseorig

    if (is.null(var1)) {
        dots$axes=FALSE
        if (is.null(dots$xlab))
            dots$xlab <- ""
    } else  {
        if (is.null(dots$xlab))
            dots$xlab <- var1
    }

    if (!plot) return(list(x=x, y=pr, predict=ci.all, predict.newdata=newdata))

    plot.list <- c(x=0,y=0,type="n",dots)
    if (new) {
        do.call(graphics::plot, plot.list)
        if (is.null(var1)) {
            box()
            axis(2)
            axis(1,at=seq(length(thelevels)),labels)
        }
    }

    col.trans <- Col(col,alpha)

    Wrap <- function(k,n) { (seq_len(k)-1)%%n +1 }
    col.i <- Wrap(k,length(col));  col.k <- col[col.i];
    lty.k <- lty[Wrap(k,length(lty))]
    pch.k <- pch[Wrap(k,length(pch))]

    if (!is.null(var1)) {
        for (i in seq_len(k)) {
            ci0 <- trans(ci.all$fit[(npoints*(i-1)+1):(i*npoints),])
            y <- ci0[,1]; yu <- ci0[,3]; yl <- ci0[,2]
            lines(y ~ x, col=col.k[i], lwd=lwd, lty=lty.k[i])

            if (ci) {
                lines(yl ~ x, lwd=1, col=col.k[i], lty=ci.lty)
                lines(yu ~ x, lwd=1, col=col.k[i], lty=ci.lty)
                xx <- c(x, rev(x))
                yy <- c(yl, rev(yu))
                polygon(xx,yy, col=col.trans[col.i[i]], lty=0)
            }

        }
    }
    ii <- as.numeric(curdata[,var2])

    if (is.null(var1)) {
        xx <- curdata[,var2]
        x <- jitter(as.numeric(xx),jitter)
        if (missing(colpt)) colpt <- Col(col[1],alpha)
        if (partres>0)
            points(pr ~ x,pch=pch[1], col=colpt[1], cex=cex, ...)
        mycoef <- bb[paste0(var2,thelevels)][-1]
        if (inherits(model,c("lm","glm")))
            myconf <- confint(model)[paste0(var2,thelevels)[-1],,drop=FALSE]
        else {
            myconf <- matrix(mycoef,ncol=2,nrow=length(mycoef))
            myconf <- myconf + qnorm(0.975)*cbind((diag(as.matrix(SS))[-1])^0.5)%x%cbind(-1,1)
        }
        for (pos in seq(k)) {
            if (cidiff) {
                if (pos>1) {
                    ci0 <- trans(intercept+myconf[pos-1,])
                    yl <- ci0[1]; yu <- ci0[2]; y <- trans(intercept+mycoef[pos-1])
                } else {
                    yu <- yl <- NULL; y <- trans(intercept)
                }
            } else if (partse) {
                y0 <- pr[xx==levels(xx)[pos]]
                ci0 <- confint(lm(y0~1))
                yl <- ci0[1]; yu <- ci0[2]; y <- trans(mean(y0))
            } else {
                ci0 <- rbind(trans(ci.all$fit[(npoints*(pos-1)+1):(pos*npoints),]))
                y <- ci0[,1]; yu <- ci0[,3]; yl <- ci0[,2]
            }
            if (!mean) y <- NULL
            confband(pos,yl,yu,delta=delta,center=y,centermark=centermark,col=col[1],lwd=lwd[1],lty=lty[1],cex=cex)
        }
    } else {
        if (partres) {
            xx <- curdata[,var1]
            if (!missing(colpt)) {
                points(pr ~ xx, col=colpt, cex=cex, pch=pch[1],...)
            } else {
                if (!is.null(var2))
                    points(pr ~ xx, col=col.k[ii], pch=pch.k[ii], cex=cex, ...)
                else
                    points(pr ~ xx, col=col[1], pch=pch[1], cex=cex,...)
            }
        }
    }

    if (k>1 && legend!=FALSE) {
        if (length(lty)>1)
            legend(legend, legend=thelevels, col=col.k, pch=pch.k, bg="white", lty=lty.k,cex=cex)
        else
            legend(legend, legend=thelevels, col=col.k, pch=pch.k, bg="white",cex=cex)
    }

    invisible(list(x=xx, y=pr, predict=ci.all, predict.newdata=newdata))
}

excoef <- function(x,digits=2,p.digits=3,format=FALSE,fun,se=FALSE,ci=TRUE,pvalue=TRUE,...) {
    cc <- coef(summary(x))
    res <- round(cbind(cc[,1:3,drop=FALSE],confint(x)),max(1,digits))
    pvalround <- round(cc[,4], max(1, p.digits))
    if (format) {
        res <- base::format(res,digits=digits,...)
        pval <- format(pvalround,p.digits=p.digits,...)
    } else {
        pval <- format(pvalround)
    }
    pval <- paste0("p=",pvalround)
    pval[which(pvalround<10^(-p.digits))] <- paste0("p<0.",paste(rep("0",p.digits-1),collapse=""),"1")
    res <- cbind(res,pval)
    nc <- apply(res,2,function(x) max(nchar(x))-nchar(x))
    res2 <- c()
    for (i in seq(nrow(res))) {
        row <- paste0(if(res[i,1]>=0) " " else "", res[i,1], paste(rep(" ",nc[i,1]),collapse=""), if (res[i,1]<0) " ",
                      if (se) paste0(" (", res[i,2], ")", paste(rep(" ",nc[i,2]), collapse="")),
                      if (ci) paste0(" [", res[i,4], ";", res[i,5], "]", paste(rep(" ",nc[i,4]+nc[i,5]),collapse="")),
                      if (pvalue) paste0(" ", res[i,6]))
        res2 <- rbind(res2," "=row)
    }
    names(res2) <- names(coef(x))
    if (!missing(fun)) {
        res2 <- c(res2,fun(x))
    }
    res2
}

##' Report estimates across different models
##'
##' @title Report estimates across different models
##' @param x list of model objects
##' @param ... additional arguments to lower-level functions
##' @author Klaus K. Holst
##' @examples
##' data(serotonin)
##' m1 <- lm(cau ~ age*gene1 + age*gene2,data=serotonin)
##' m2 <- lm(cau ~ age + gene1,data=serotonin)
##' m3 <- lm(cau ~ age*gene2,data=serotonin)
##'
##' Combine(list(A=m1,B=m2,C=m3),fun=function(x)
##'      c("_____"="",R2=" "%++%format(summary(x)$r.squared,digits=2)))
##' @export
Combine <- function(x,...) {
    ll <- lapply(x,excoef,...)
    nn <- lapply(ll,names)
    n0 <- unique(unlist(nn,use.names=FALSE))
    res <- matrix(NA,ncol=length(ll),nrow=length(n0))
    colnames(res) <- seq(length(ll))
    rownames(res) <- n0
    for (i in seq(length(ll))) {
        res[match(names(ll[[i]]),n0),i] <- ll[[i]]
    }
    colnames(res) <- names(ll)
    class(res) <- c("Combine","matrix")
    return(res)
}

##' @export
print.Combine <- function(x,...) {
    print(as.table(x),...)
}

#!/bin/sh

set -e

[ -n "$1" ] && exec "$@"

# Set default variables
URL=${URL:-http://localhost/}
LOGLEVEL=${LOGLEVEL:-DEBUG}
LOGFILE=${LOGFILE:--}

# Import variables
[ -e /etc/lava-dispatcher/lava-worker ] && . /etc/lava-dispatcher/lava-worker

for f in $(find /root/entrypoint.d/ -type f); do
    case "$f" in
        *.sh)
            echo "$0: running ${f}"
            "${f}"
            ;;
        *)
        echo "$0: ignoring ${f}"
        ;;
    esac
done

exec /usr/bin/lava-worker --level "$LOGLEVEL" --log-file "$LOGFILE" --url "$URL" $TOKEN $WORKER_NAME $WS_URL $HTTP_TIMEOUT

#!/bin/sh

set -e

[ -n "$1" ] && exec "$@"

# Set default variables
URL=${URL:-http://localhost/}
LOGLEVEL=${LOGLEVEL:-DEBUG}
LOGFILE=${LOGFILE:--}

# Add lava-docker-worker directories to PATH
for d in "/usr/share/lava-docker-worker" "/usr/local/share/lava-docker-worker"; do
    if [ -d "${d}/bin" ]; then
        export PATH="${d}/bin:${PATH}"
    fi
done

# Import variables
[ -e /etc/lava-dispatcher/lava-worker ] && . /etc/lava-dispatcher/lava-worker

for f in $(find /root/entrypoint.d/ -type f); do
    case "$f" in
        *.sh)
            echo "$0: running ${f}"
            "${f}"
            ;;
        *)
        echo "$0: ignoring ${f}"
        ;;
    esac
done

exec /usr/bin/lava-worker --level "$LOGLEVEL" --log-file "$LOGFILE" --url "$URL" $TOKEN $WORKER_NAME $WS_URL $HTTP_TIMEOUT $JOB_LOG_INTERVAL

#!/bin/bash

set -e

SUITE=unstable
DIR="../build-area/"
DEBUILD_OPTS=" -sa"

while getopts ":a:o:s:B" opt; do
  case $opt in
    a)
      ARCH=$OPTARG
      set +e
      chk=`dpkg-architecture -a$ARCH > /dev/null 2>&1 ; echo $?`
      set -e
      if [ $chk != 0 ]; then
          echo "Did not recognise ${ARCH} as a Debian architecture name. Exit."
          exit 1
      fi
      # preserve the .dsc
      DEBUILD_OPTS=" -a${ARCH}"
      ;;
    B)
      # only build arch:any binary packages
      DEBUILD_OPTS=" -B"
      ;;
    s)
      SUITE=$OPTARG
      ;;
    o)
      DIR=`readlink -f $OPTARG`
      ;;
    \?)
      echo "[LAVA-DEV] Invalid option: -$OPTARG" >&2
      echo
      echo "Usage: [-a <architecture> -o <directory> -s <suite> -B]"
      echo
      echo "If architecture is a known Debian architecture, build"
      echo "a binary-only package for this architecture. e.g. armhf or arm64"
      echo "Builds a sourceful native package locally, using debuild."
      echo "Specify the build directory using -o - default ../build-area"
      echo "Use -B to only build the architecture specific binary package"
      echo "e.g. when building lava-dispatcher on arm64."
      exit 1
      ;;
  esac
done

# lintian is disabled because it will complain about the
# change to a native build.
DEBUILD_OPTS=" --no-lintian -uc -us ${DEBUILD_OPTS}"

if [ ! -d ${DIR} ]; then
  mkdir ${DIR}
fi

if [ -x ./lava_common/version.py ]; then
  # native version for developer build
  VERSION=`python3 ./lava_common/version.py |tr - .`
else
  echo "[LAVA-DEV] Unable to find ./lava_common/version.py"
  exit 1
fi

RELEASE=9999
if [ "${SUITE}" != "unstable" -a "${SUITE}" != "sid" ]; then
  RELEASE="$(distro-info --release --$(distro-info --alias=${SUITE}))"
fi

dpkg-checkbuilddeps
LOCAL=`git diff | wc -l`
if [ ${LOCAL} != 0 ]; then
    echo "[LAVA-DEV] You have uncommitted changes in your source tree:"
    git status
    exit 3
fi

if [ -d './dist/' ]; then
    rm -f ./dist/*
fi

if [ -d .git ]; then
  LOG=`git log -n1 --pretty=format:"Last change %h by %an, %ar. %s%n" --no-merges`
fi
NAME=`dpkg-parsechangelog |grep Source|cut -d" " -f2`

BUILDDIR="${DIR}/lava-${VERSION}"
mkdir -p "${BUILDDIR}"
git archive HEAD | tar xf - -C "${BUILDDIR}"
cd "${BUILDDIR}"

export GIT_COMMITTER_NAME="lava-dev debian build script"
export GIT_COMMITTER_EMAIL="lava-dev@lavasoftware.org"
export GIT_AUTHOR_NAME="lava-dev debian build script"
export GIT_AUTHOR_EMAIL="lava-dev@lavasoftware.org"
export DEBEMAIL=lava-dev@lavasoftware.org
export DEBFULLNAME=lava-dev debian build script

# Save the version string
echo "${VERSION}" > lava_common/VERSION

# convert to a native package to include local changes.
echo "3.0 (native)" > debian/source/format
BUILD_SUITE="${SUITE}"
dch -b -v "${VERSION}+${RELEASE}+${SUITE}" -D ${BUILD_SUITE} "Local developer native build for ${BUILD_SUITE}"
if [ -n "${LOG}" ]; then
  dch -a "${LOG}"
fi

debuild ${DEBUILD_OPTS}
cd -

CHANGES="${DIR}/${NAME}_${VERSION}*.changes"
echo
echo ${LOG}
echo
echo Use "zless /usr/share/doc/${NAME}/changelog.Debian.gz"
echo to view the changelog, once packages are installed.
echo
if [ -x /usr/bin/dcmd ]; then
    dcmd ls ${CHANGES}
fi

#!/bin/bash

set -eu -o pipefail

LAVA_GUNICORN_WORKER_CLASS="${LAVA_GUNICORN_WORKER_CLASS:-eventlet}"
LAVA_GUNICORN_WORKER_NUM="${LAVA_GUNICORN_WORKER_NUM:-4}"
LAVA_GUNICORN_WORKER_CONNECTIONS="${LAVA_GUNICORN_WORKER_CONNECTIONS:-23}"  # 23*4 == 92
LAVA_GUNICORN_BIND_ADDR="${LAVA_GUNICORN_BIND_ADDR:-0.0.0.0}"
LAVA_GUNICORN_BIND_PORT="${LAVA_GUNICORN_BIND_PORT:-8000}"


exec /usr/bin/gunicorn3 lava_server.wsgi \
    -u lavaserver -g lavaserver \
    --preload \
    --worker-class "$LAVA_GUNICORN_WORKER_CLASS" \
    --workers "$LAVA_GUNICORN_WORKER_NUM" \
    --worker-connections "$LAVA_GUNICORN_WORKER_CONNECTIONS" \
    --bind "${LAVA_GUNICORN_BIND_ADDR}:${LAVA_GUNICORN_BIND_PORT}"

#!/bin/sh

set -e

if [ -z "$3" ]; then
	echo "Usage: <package> <chroot> <suite>"
	echo "A suitable schroot must already have been prepared"
	echo "and the Debian unstable LAVA dependencies made available"
	echo "inside that chroot."
	exit 1
fi

chroot=$2
suite=$3
PWD=`pwd`
NAME=${1}
if [ -x ./version.py ]; then
  VERSION=`python ./version.py`
else
  VERSION=`python setup.py --version`
fi
if [ -d './dist/' ]; then
    rm -f ./dist/*
fi
python setup.py sdist
if [ -d .git ]; then
  LOG=`git log -n1 --pretty=format:"Last change %h by %an, %ar. %s%n" --no-merges`
fi
DIR=`mktemp -d`
if [ -f './dist/${NAME}-${VERSION}.tar.gz' ]; then
  mv -v ./dist/${NAME}-${VERSION}.tar.gz ${DIR}/${NAME}_${VERSION}.orig.tar.gz
else
  echo "WARNING: broken setuptools tarball - Debian bug #786977"
  mv -v ./dist/${NAME}*.tar.gz ${DIR}/${NAME}_${VERSION}.orig.tar.gz
fi
cd ${DIR}
git clone https://github.com/Linaro/pkg-${NAME}.git
tar -xzf ${NAME}_${VERSION}.orig.tar.gz
if [ ! -d ${DIR}/${NAME}-${VERSION} ]; then
  mv -v ${DIR}/${NAME}-* ${DIR}/${NAME}-${VERSION}
fi
cd ${DIR}/pkg-${NAME}/
dpkg-checkbuilddeps
git archive master debian | tar -x -C ../${NAME}-${VERSION}
cd ${DIR}/${NAME}-${VERSION}
dch --force-distribution -v ${VERSION}-1 -D ${suite} "Local developer build"
if [ -n "${LOG}" ]; then
  dch -a ${LOG}
fi
sbuild -A -s -c ${chroot} -d ${suite}
cd ${DIR}
rm -rf ${DIR}/pkg-${NAME}
rm -rf ${DIR}/${NAME}-${VERSION}
echo
echo ${LOG}
echo
echo Use "zless /usr/share/doc/${NAME}/changelog.Debian.gz"
echo to view the changelog, once packages are installed.
echo
if [ -x /usr/bin/dcmd ]; then
    dcmd ls ${DIR}/${NAME}_${VERSION}*.changes
else
    echo ${DIR}
    ls ${DIR}
fi

#!/bin/bash

set -eu -o pipefail

DATABASE="$(python3 -c 'import os; from urllib.parse import urlparse; url = urlparse(os.environ["DATABASE_URL"]); print("{}:{}".format(url.hostname, url.port), end="")')"

/usr/bin/wait-for-it -t 30 "$DATABASE"

lava-server manage migrate --noinput --fake-initial
lava-server manage drop_materialized_views
lava-server manage refresh_queries --all

#!/bin/bash

set -e

SUITE=unstable
DIR="../build-area/"
DEBUILD_OPTS=("-sa")

while getopts ":a:o:s:B" opt; do
  case $opt in
    a)
      ARCH=$OPTARG
      set +e
      chk=$(dpkg-architecture --host-arch "$ARCH" > /dev/null 2>&1 ; echo $?)
      set -e
      if [ "$chk" != 0 ]; then
          echo "Did not recognise ${ARCH} as a Debian architecture name. Exit."
          exit 1
      fi
      # preserve the .dsc
      DEBUILD_OPTS=("--host-arch" "${ARCH}")
      ;;
    B)
      # only build arch:any binary packages
      DEBUILD_OPTS=("-B")
      ;;
    s)
      SUITE=$OPTARG
      ;;
    o)
      DIR=$(readlink -f "$OPTARG")
      ;;
    \?)
      echo "[LAVA-DEV] Invalid option: -$OPTARG" >&2
      echo
      echo "Usage: [-a <architecture> -o <directory> -s <suite> -B]"
      echo
      echo "If architecture is a known Debian architecture, build"
      echo "a binary-only package for this architecture. e.g. armhf or arm64"
      echo "Builds a sourceful native package locally, using debuild."
      echo "Specify the build directory using -o - default ../build-area"
      echo "Use -B to only build the architecture specific binary package"
      echo "e.g. when building lava-dispatcher on arm64."
      exit 1
      ;;
  esac
done

# lintian is disabled because it will complain about the
# change to a native build.
DEBUILD_OPTS=("--no-lintian" "-uc" "-us" "${DEBUILD_OPTS[@]}")

if [ ! -d "${DIR}" ]; then
  mkdir "${DIR}"
fi

if [ -x ./lava_common/version.py ]; then
  # native version for developer build
  VERSION="$(python3 ./lava_common/version.py | tr - .)"
else
  echo "[LAVA-DEV] Unable to find ./lava_common/version.py"
  exit 1
fi

RELEASE=9999
if [ "${SUITE}" != "unstable" ] && [ "${SUITE}" != "sid" ]; then
  RELEASE="$(distro-info --release --series="${SUITE}")"
fi

dpkg-checkbuilddeps
LOCAL=$(git diff | wc -l)
if [ "${LOCAL}" != 0 ]; then
    echo "[LAVA-DEV] You have uncommitted changes in your source tree:"
    git status
    exit 3
fi

if [ -d './dist/' ]; then
    rm -f ./dist/*
fi

if [ -d .git ]; then
  LOG=$(git log -n1 --pretty=format:"Last change %h by %an on %aD (%s)%n" --no-merges)
  DATE=$(git log -n1 --pretty=format:%aD --no-merges)
fi
NAME=$(dpkg-parsechangelog | grep Source | cut -d" " -f2)

BUILDDIR="${DIR}/lava-${VERSION}"
mkdir -p "${BUILDDIR}"
git archive HEAD | tar xf - -C "${BUILDDIR}"
cd "${BUILDDIR}"

export GIT_COMMITTER_NAME="lava-dev debian build script"
export GIT_COMMITTER_EMAIL="lava-dev@lavasoftware.org"
export GIT_AUTHOR_NAME="lava-dev debian build script"
export GIT_AUTHOR_EMAIL="lava-dev@lavasoftware.org"
export DEBEMAIL=lava-dev@lavasoftware.org
export DEBFULLNAME=lava-dev debian build script

# Save the version string
echo "${VERSION}" > lava_common/VERSION

# convert to a native package to include local changes.
echo "3.0 (native)" > debian/source/format
BUILD_SUITE="${SUITE}"
dch --force-distribution -b -v "${VERSION}+${RELEASE}+${SUITE}" -D "${BUILD_SUITE}" "Local developer native build for ${BUILD_SUITE}"
if [ -n "${LOG}" ]; then
  dch -a "${LOG}"
fi
if [ -n "${DATE}" ]; then
  # Use the git commit timestamp
  sed -i -e "0,/^ --/ { s/^ --.*/ -- $DEBFULLNAME <$DEBEMAIL>  $DATE/ }" \
    debian/changelog
fi

debuild "${DEBUILD_OPTS[@]}"
cd -

CHANGES="${DIR}/${NAME}_${VERSION}*.changes"
echo
echo "${LOG}"
echo
echo Use "zless /usr/share/doc/${NAME}/changelog.Debian.gz"
echo to view the changelog, once packages are installed.
echo
if [ -x /usr/bin/dcmd ]; then
    dcmd ls "${CHANGES}"
fi

#!/usr/bin/env bash

set -e

[ -n "$1" ] && exec "$@"

#############
# Variables #
#############
GUNICORN_PID=0
LAVA_CELERY_WORKER_PID=0
LAVA_COORDINATOR_PID=0
LAVA_PUBLISHER_PID=0
LAVA_SCHEDULER_PID=0


##################
# Signal handler #
##################
handler() {
    tail_pid="${!}"
    echo "Killing:"
    echo "* lava-celery-worker \$$LAVA_CELERY_WORKER_PID"
    [ "$LAVA_CELERY_WORKER_PID" != "0" ] && kill $LAVA_CELERY_WORKER_PID
    echo "* lava-coordinator \$$LAVA_COORDINATOR_PID"
    [ "$LAVA_COORDINATOR_PID" != "0" ] && kill $LAVA_COORDINATOR_PID
    echo "* lava-publisher \$$LAVA_PUBLISHER_PID"
    [ "$LAVA_PUBLISHER_PID" != "0" ] && kill $LAVA_PUBLISHER_PID
    echo "* lava-scheduler \$$LAVA_SCHEDULER_PID"
    [ "$LAVA_SCHEDULER_PID" != "0" ] && kill $LAVA_SCHEDULER_PID
    echo "* gunicorn \$$GUNICORN_PID"
    [ "$GUNICORN_PID" != "0" ] && kill $GUNICORN_PID
    echo "* apache2"
    apache2ctl stop

    echo "Waiting for:"
    echo "* lava-celery-worker"
    [ "$LAVA_CELERY_WORKER_PID" != "0" ] && wait $LAVA_CELERY_WORKER_PID || true
    echo "* lava-coordinator"
    [ "$LAVA_COORDINATOR_PID" != "0" ] && wait $LAVA_COORDINATOR_PID || true
    echo "* lava-publisher"
    [ "$LAVA_PUBLISHER_PID" != "0" ] && wait $LAVA_PUBLISHER_PID || true
    echo "* lava-scheduler"
    [ "$LAVA_SCHEDULER_PID" != "0" ] && wait $LAVA_SCHEDULER_PID || true
    echo "* gunicorn"
    [ "$GUNICORN_PID" != "0" ] && wait $GUNICORN_PID || true

    echo "Killing postgresql"
    /etc/init.d/postgresql stop

    echo "Killing log reader"
    [ -n "$tail_pid" ] && kill "$tail_pid"
    exit 0
}


#####################
# Check file owners #
#####################
check_owners() {
    files=$(find /etc/lava-server/dispatcher-config/ /var/lib/lava-server/default/media/ -not -user lavaserver)
    if [ "$files" ]; then
        echo "The following files should belong to 'lavaserver' user:"
        echo "$files"
        exit 1
    fi
    files=$(find /etc/lava-server/dispatcher-config/ /var/lib/lava-server/default/media/ -not -group lavaserver)
    if [ "$files" ]; then
        echo "The following files should belong to 'lavaserver' group:"
        echo "$files"
        exit 1
    fi
}


#################
# Start helpers #
#################
start_apache2() {
    export APACHE_CONFDIR=/etc/apache2
    export APACHE_ENVVARS=/etc/apache2/envvars
    if [ "$CAN_EXEC" = "1" ]; then
        exec apache2ctl -DFOREGROUND
    else
        apache2ctl -DFOREGROUND &
    fi
}


start_lava_celery_worker() {
    USER="lavaserver"
    GROUP="lavaserver"
    LOGLEVEL="INFO"
    LOGFILE="/var/log/lava-server/lava-celery-worker.log"
    CONCURRENCY=""
    AUTOSCALE=""
    ARGS=""
    [ -e /etc/default/lava-celery-worker ] && . /etc/default/lava-celery-worker
    [ -e /etc/lava-server/lava-celery-worker ] && . /etc/lava-server/lava-celery-worker
    if [ "$CAN_EXEC" = "1" ]; then
        exec /usr/bin/python3 -m celery -A lava_server worker --uid $USER --gid $GROUP --loglevel $LOGLEVEL $CONCURRENCY $AUTOSCALE $ARGS
    else
        setsid /usr/bin/python3 -m celery -A lava_server worker --uid $USER --gid $GROUP --loglevel $LOGLEVEL --logfile $LOGFILE $CONCURRENCY $AUTOSCALE $ARGS &
        LAVA_CELERY_WORKER_PID=$!
    fi
}

start_lava_coordinator() {
    LOGLEVEL="DEBUG"
    LOGFILE="/var/log/lava-coordinator.log"
    [ -e /etc/default/lava-coordinator ] && . /etc/default/lava-coordinator
    [ -e /etc/lava-coordinator/lava-coordinator ] && . /etc/lava-coordinator/lava-coordinator
    if [ "$CAN_EXEC" = "1" ]; then
        exec /usr/bin/lava-coordinator --logfile - --loglevel "$LOGLEVEL"
    else
        setsid /usr/bin/lava-coordinator --logfile "$LOGFILE" --loglevel "$LOGLEVEL" &
        LAVA_COORDINATOR_PID=$!
    fi
}


start_lava_publisher() {
    LOGLEVEL="DEBUG"
    HOST="*"
    PORT=8001
    [ -e /etc/default/lava-publisher ] && . /etc/default/lava-publisher
    [ -e /etc/lava-server/lava-publisher ] && . /etc/lava-server/lava-publisher
    if [ "$CAN_EXEC" = "1" ]; then
        exec /usr/bin/lava-server manage lava-publisher --log-file - --level "$LOGLEVEL" --host "$HOST" --port "$PORT"
    else
        setsid /usr/bin/lava-server manage lava-publisher --level "$LOGLEVEL" --host "$HOST" --port "$PORT" &
        LAVA_PUBLISHER_PID=$!
    fi
}


start_lava_scheduler() {
    LOGLEVEL="DEBUG"
    [ -e /etc/default/lava-scheduler ] && . /etc/default/lava-scheduler
    [ -e /etc/lava-server/lava-scheduler ] && . /etc/lava-server/lava-scheduler
    if [ "$CAN_EXEC" = "1" ]; then
        exec /usr/bin/lava-server manage lava-scheduler --log-file - --level "$LOGLEVEL" $EVENT_URL $IPV6
    else
        setsid /usr/bin/lava-server manage lava-scheduler --level "$LOGLEVEL" $EVENT_URL $IPV6 &
        LAVA_SCHEDULER_PID=$!
    fi
}


start_lava_server_gunicorn() {
    LOGLEVEL="DEBUG"
    TIMEOUT=""
    WORKER_CLASS="eventlet"
    WORKERS="4"
    LOGFILE="/var/log/lava-server/gunicorn.log"
    [ -e /etc/default/lava-server-gunicorn ] && . /etc/default/lava-server-gunicorn
    [ -e /etc/lava-server/lava-server-gunicorn ] && . /etc/lava-server/lava-server-gunicorn
    if [ "$CAN_EXEC" = "1" ]; then
        exec /usr/bin/gunicorn3 lava_server.wsgi --log-level "$LOGLEVEL" --log-file - -u lavaserver -g lavaserver --worker-class "$WORKER_CLASS" --workers "$WORKERS" --worker-tmp-dir /dev/shm $RELOAD $BIND $TIMEOUT
    else
        setsid /usr/bin/gunicorn3 lava_server.wsgi --log-level "$LOGLEVEL" --log-file "$LOGFILE" -u lavaserver -g lavaserver --worker-class "$WORKER_CLASS" --workers "$WORKERS" --worker-tmp-dir /dev/shm $RELOAD $BIND $TIMEOUT &
        GUNICORN_PID=$!
    fi
}


########
# Main #
########

# setup handlers
trap 'handler' INT QUIT TERM

# List of services to start
# By default lava-celery-worker is not started because no broker is included
SERVICES=${SERVICES-"apache2 lava-coordinator lava-publisher lava-scheduler gunicorn postgresql"}
echo "$SERVICES" | grep -q apache2 && APACHE2=1 || APACHE2=0
echo "$SERVICES" | grep -q lava-celery-worker && LAVA_CELERY_WORKER=1 || LAVA_CELERY_WORKER=0
echo "$SERVICES" | grep -q lava-coordinator && LAVA_COORDINATOR=1 || LAVA_COORDINATOR=0
echo "$SERVICES" | grep -q lava-publisher && LAVA_PUBLISHER=1 || LAVA_PUBLISHER=0
echo "$SERVICES" | grep -q lava-scheduler && LAVA_SCHEDULER=1 || LAVA_SCHEDULER=0
echo "$SERVICES" | grep -q gunicorn && GUNICORN=1 || GUNICORN=0
echo "$SERVICES" | grep -q postgresql && POSTGRESQL=1 || POSTGRESQL=0

# Is the database needed?
NEED_DB=$((LAVA_SCHEDULER+GUNICORN+POSTGRESQL))
# Migrate if LAVA_DB_MIGRATE is undefined and lava-scheduler is running in this
# container.
[ "$LAVA_SCHEDULER" = "1" ] && MIGRATE_DEFAULT="yes" || MIGRATE_DEFAULT="no"
LAVA_DB_MIGRATE=${LAVA_DB_MIGRATE:-$MIGRATE_DEFAULT}
# Should we use "exec"?
CAN_EXEC=$((APACHE2+LAVA_CELERY_WORKER+LAVA_COORDINATOR+LAVA_PUBLISHER+LAVA_SCHEDULER+GUNICORN+POSTGRESQL))
# Should we check for file owners?
LAVA_CHECK_OWNERS=${LAVA_CHECK_OWNERS:-1}

/usr/share/lava-server/postinst.py --config
echo "done"
echo

# Start requested services
if [ "$POSTGRESQL" = "1" ]
then
    echo "Starting postgresql"
    /etc/init.d/postgresql start
    echo "done"
    echo
    echo "Setup the database"
    /usr/share/lava-server/postinst.py --db
    echo "done"
    echo
fi

if [ "$NEED_DB" != "0" ]
then
    lava-server manage wait database
    if [ "$LAVA_DB_MIGRATE" = "yes" ]
    then
        echo "Applying migrations"
        lava-server manage migrate --no-input
    else
        echo "Waiting for migrations"
        lava-server manage wait migrations
    fi
    echo "done"
    echo
fi

if [ "$LAVA_SITE" != "" ]
then
    echo "Setting lava site: $LAVA_SITE"
    lava-server manage site update --name "$LAVA_SITE" --domain "$LAVA_SITE"
    echo "done"
    echo
fi

if [ "$LAVA_ADMIN_USERNAME" != "" ] && [ "$LAVA_ADMIN_PASSWORD" != "" ]
then
    echo "Creating lava admin: $LAVA_ADMIN_USERNAME"
    lava-server manage users details "$LAVA_ADMIN_USERNAME" || lava-server manage users add --passwd "$LAVA_ADMIN_PASSWORD" --staff --superuser "$LAVA_ADMIN_USERNAME"
    echo "done"
    echo
fi

# Run user scripts. The database is running and migrations has been run.
for f in $(find /root/entrypoint.d/ -type f); do
    case "$f" in
        *.sh)
            echo "$0: running ${f}"
            "${f}"
            ;;
        *)
        echo "$0: ignoring ${f}"
        ;;
    esac
done

if [ "$GUNICORN" = "1" ]
then
    echo "Checking file permissions"
    if [ "$LAVA_CHECK_OWNERS" = "1" ]
    then
        check_owners
        echo "done"
    else
        echo "skipped"
    fi
    echo
    echo "Starting gunicorn3"
    start_lava_server_gunicorn
    echo "done"
    echo
fi

if [ "$APACHE2" = "1" ]
then
    echo "Starting apache2"
    start_apache2
    echo "done"
    echo
fi

if [ "$LAVA_CELERY_WORKER" = "1" ]
then
    echo "Starting lava-celery-worker"
    start_lava_celery_worker
    echo "done"
    echo
fi


if [ "$LAVA_COORDINATOR" = "1" ]
then
    echo "Starting lava-coordinator"
    start_lava_coordinator
    echo "done"
    echo
fi

if [ "$LAVA_PUBLISHER" = "1" ]
then
    echo "Starting lava-publisher"
    start_lava_publisher
    echo "done"
    echo
fi

if [ "$LAVA_SCHEDULER" = "1" ]
then
    echo "Starting lava-scheduler"
    start_lava_scheduler
    echo "done"
    echo
fi


################
# Wait forever #
################
cd /var/log/lava-server
while true
do
  tail -F django.log gunicorn.log lava-celery-worker.log lava-publisher.log lava-scheduler.log /var/log/lava-coordinator.log /var/log/apache2/lava-server.log & wait ${!}
done

#!/usr/bin/env bash

set -e

[ -n "$1" ] && exec "$@"

#############
# Variables #
#############
GUNICORN_PID=0
LAVA_CELERY_WORKER_PID=0
LAVA_COORDINATOR_PID=0
LAVA_PUBLISHER_PID=0
LAVA_SCHEDULER_PID=0


##################
# Signal handler #
##################
handler() {
    tail_pid="${!}"
    echo "Killing:"
    echo "* lava-celery-worker \$$LAVA_CELERY_WORKER_PID"
    [ "$LAVA_CELERY_WORKER_PID" != "0" ] && kill $LAVA_CELERY_WORKER_PID
    echo "* lava-coordinator \$$LAVA_COORDINATOR_PID"
    [ "$LAVA_COORDINATOR_PID" != "0" ] && kill $LAVA_COORDINATOR_PID
    echo "* lava-publisher \$$LAVA_PUBLISHER_PID"
    [ "$LAVA_PUBLISHER_PID" != "0" ] && kill $LAVA_PUBLISHER_PID
    echo "* lava-scheduler \$$LAVA_SCHEDULER_PID"
    [ "$LAVA_SCHEDULER_PID" != "0" ] && kill $LAVA_SCHEDULER_PID
    echo "* gunicorn \$$GUNICORN_PID"
    [ "$GUNICORN_PID" != "0" ] && kill $GUNICORN_PID
    echo "* apache2"
    apache2ctl stop

    echo "Waiting for:"
    echo "* lava-celery-worker"
    [ "$LAVA_CELERY_WORKER_PID" != "0" ] && wait $LAVA_CELERY_WORKER_PID || true
    echo "* lava-coordinator"
    [ "$LAVA_COORDINATOR_PID" != "0" ] && wait $LAVA_COORDINATOR_PID || true
    echo "* lava-publisher"
    [ "$LAVA_PUBLISHER_PID" != "0" ] && wait $LAVA_PUBLISHER_PID || true
    echo "* lava-scheduler"
    [ "$LAVA_SCHEDULER_PID" != "0" ] && wait $LAVA_SCHEDULER_PID || true
    echo "* gunicorn"
    [ "$GUNICORN_PID" != "0" ] && wait $GUNICORN_PID || true

    echo "Killing postgresql"
    /etc/init.d/postgresql stop

    echo "Killing log reader"
    [ -n "$tail_pid" ] && kill "$tail_pid"
    exit 0
}


#####################
# Check file owners #
#####################
check_owners() {
    files=$(find /etc/lava-server/dispatcher-config/ /var/lib/lava-server/default/media/ -not -user lavaserver)
    if [ "$files" ]; then
        echo "The following files should belong to 'lavaserver' user:"
        echo "$files"
        exit 1
    fi
    files=$(find /etc/lava-server/dispatcher-config/ /var/lib/lava-server/default/media/ -not -group lavaserver)
    if [ "$files" ]; then
        echo "The following files should belong to 'lavaserver' group:"
        echo "$files"
        exit 1
    fi
}


#################
# Start helpers #
#################
start_apache2() {
    export APACHE_CONFDIR=/etc/apache2
    export APACHE_ENVVARS=/etc/apache2/envvars
    if [ "$CAN_EXEC" = "1" ]; then
        exec apache2ctl -DFOREGROUND
    else
        apache2ctl -DFOREGROUND &
    fi
}


start_lava_celery_worker() {
    USER="lavaserver"
    GROUP="lavaserver"
    LOGLEVEL=${CELERY_LOGLEVEL:-INFO}
    LOGFILE="/var/log/lava-server/lava-celery-worker.log"
    CONCURRENCY=${CELERY_CONCURRENCY:-}
    AUTOSCALE=${CELERY_AUTOSCALE:-}
    ARGS=${CELERY_ARGS:-}
    [ -e /etc/default/lava-celery-worker ] && . /etc/default/lava-celery-worker
    [ -e /etc/lava-server/lava-celery-worker ] && . /etc/lava-server/lava-celery-worker
    if [ "$CAN_EXEC" = "1" ]; then
        exec /usr/bin/python3 -m celery -A lava_server worker --uid $USER --gid $GROUP --loglevel $LOGLEVEL $CONCURRENCY $AUTOSCALE $ARGS
    else
        setsid /usr/bin/python3 -m celery -A lava_server worker --uid $USER --gid $GROUP --loglevel $LOGLEVEL --logfile $LOGFILE $CONCURRENCY $AUTOSCALE $ARGS &
        LAVA_CELERY_WORKER_PID=$!
    fi
}

start_lava_coordinator() {
    LOGLEVEL=${COORDINATOR_LOGLEVEL:-DEBUG}
    LOGFILE="/var/log/lava-coordinator.log"
    [ -e /etc/default/lava-coordinator ] && . /etc/default/lava-coordinator
    [ -e /etc/lava-coordinator/lava-coordinator ] && . /etc/lava-coordinator/lava-coordinator
    if [ "$CAN_EXEC" = "1" ]; then
        exec /usr/bin/lava-coordinator --logfile - --loglevel "$LOGLEVEL"
    else
        setsid /usr/bin/lava-coordinator --logfile "$LOGFILE" --loglevel "$LOGLEVEL" &
        LAVA_COORDINATOR_PID=$!
    fi
}


start_lava_publisher() {
    LOGLEVEL=${PUBLISHER_LOGLEVEL:-DEBUG}
    HOST=${PUBLISHER_HOST:-*}
    PORT=${PUBLISHER_PORT:-8001}

    [ -e /etc/default/lava-publisher ] && . /etc/default/lava-publisher
    [ -e /etc/lava-server/lava-publisher ] && . /etc/lava-server/lava-publisher
    if [ "$CAN_EXEC" = "1" ]; then
        exec /usr/bin/lava-server manage lava-publisher --log-file - --level "$LOGLEVEL" --host "$HOST" --port "$PORT"
    else
        setsid /usr/bin/lava-server manage lava-publisher --level "$LOGLEVEL" --host "$HOST" --port "$PORT" &
        LAVA_PUBLISHER_PID=$!
    fi
}


start_lava_scheduler() {
    LOGLEVEL=${SCHEDULER_LOGLEVEL:-DEBUG}
    EVENT_URL=${SCHEDULER_EVENT_URL:-}
    IPV6=${SCHEDULER_IPV6:-}

    [ -e /etc/default/lava-scheduler ] && . /etc/default/lava-scheduler
    [ -e /etc/lava-server/lava-scheduler ] && . /etc/lava-server/lava-scheduler
    if [ "$CAN_EXEC" = "1" ]; then
        exec /usr/bin/lava-server manage lava-scheduler --log-file - --level "$LOGLEVEL" $EVENT_URL $IPV6
    else
        setsid /usr/bin/lava-server manage lava-scheduler --level "$LOGLEVEL" $EVENT_URL $IPV6 &
        LAVA_SCHEDULER_PID=$!
    fi
}


start_lava_server_gunicorn() {
    BIND=${GUNICORN_BIND:-}
    LOGLEVEL=${GUNICORN_LOGLEVEL:-DEBUG}
    TIMEOUT=${GUNICORN_TIMEOUT:-}
    WORKER_CLASS=${GUNICORN_WORKER_CLASS:-gthread}
    WORKERS=${GUNICORN_WORKERS:-4}
    LOGFILE="/var/log/lava-server/gunicorn.log"
    EXTRA_ARGS=${GUNICORN_EXTRA_ARGS:-}

    [ -e /etc/default/lava-server-gunicorn ] && . /etc/default/lava-server-gunicorn
    [ -e /etc/lava-server/lava-server-gunicorn ] && . /etc/lava-server/lava-server-gunicorn
    if [ "$CAN_EXEC" = "1" ]; then
        exec /usr/bin/gunicorn3 lava_server.wsgi --log-level "$LOGLEVEL" --log-file - -u lavaserver -g lavaserver --preload --worker-class "$WORKER_CLASS" --workers "$WORKERS" --worker-tmp-dir /dev/shm $RELOAD $BIND $TIMEOUT $EXTRA_ARGS
    else
        setsid /usr/bin/gunicorn3 lava_server.wsgi --log-level "$LOGLEVEL" --log-file "$LOGFILE" -u lavaserver -g lavaserver --preload --worker-class "$WORKER_CLASS" --workers "$WORKERS" --worker-tmp-dir /dev/shm $RELOAD $BIND $TIMEOUT $EXTRA_ARGS &
        GUNICORN_PID=$!
    fi
}


########
# Main #
########

# setup handlers
trap 'handler' INT QUIT TERM

# List of services to start
# By default lava-celery-worker is not started because no broker is included
SERVICES=${SERVICES-"apache2 lava-coordinator lava-publisher lava-scheduler gunicorn postgresql"}
echo "$SERVICES" | grep -q apache2 && APACHE2=1 || APACHE2=0
echo "$SERVICES" | grep -q lava-celery-worker && LAVA_CELERY_WORKER=1 || LAVA_CELERY_WORKER=0
echo "$SERVICES" | grep -q lava-coordinator && LAVA_COORDINATOR=1 || LAVA_COORDINATOR=0
echo "$SERVICES" | grep -q lava-publisher && LAVA_PUBLISHER=1 || LAVA_PUBLISHER=0
echo "$SERVICES" | grep -q lava-scheduler && LAVA_SCHEDULER=1 || LAVA_SCHEDULER=0
echo "$SERVICES" | grep -q gunicorn && GUNICORN=1 || GUNICORN=0
echo "$SERVICES" | grep -q postgresql && POSTGRESQL=1 || POSTGRESQL=0

# Is the database needed?
NEED_DB=$((LAVA_SCHEDULER+GUNICORN+POSTGRESQL))
# Migrate if LAVA_DB_MIGRATE is undefined and lava-scheduler is running in this
# container.
[ "$LAVA_SCHEDULER" = "1" ] && MIGRATE_DEFAULT="yes" || MIGRATE_DEFAULT="no"
LAVA_DB_MIGRATE=${LAVA_DB_MIGRATE:-$MIGRATE_DEFAULT}
# Should we use "exec"?
CAN_EXEC=$((APACHE2+LAVA_CELERY_WORKER+LAVA_COORDINATOR+LAVA_PUBLISHER+LAVA_SCHEDULER+GUNICORN+POSTGRESQL))
# Should we check for file owners?
LAVA_CHECK_OWNERS=${LAVA_CHECK_OWNERS:-1}

/usr/share/lava-server/postinst.py --config
echo "done"
echo

# Start requested services
if [ "$POSTGRESQL" = "1" ]
then
    echo "Starting postgresql"
    /etc/init.d/postgresql start
    echo "done"
    echo
    echo "Setup the database"
    /usr/share/lava-server/postinst.py --config --db
    echo "done"
    echo
fi

if [ "$NEED_DB" != "0" ]
then
    lava-server manage wait database
    if [ "$LAVA_DB_MIGRATE" = "yes" ]
    then
        echo "Applying migrations"
        lava-server manage migrate --no-input
    else
        echo "Waiting for migrations"
        lava-server manage wait migrations
    fi
    echo "done"
    echo
fi

if [ "$LAVA_SITE" != "" ]
then
    echo "Setting lava site: $LAVA_SITE"
    lava-server manage site update --name "$LAVA_SITE" --domain "$LAVA_SITE"
    echo "done"
    echo
fi

if [ "$LAVA_ADMIN_USERNAME" != "" ] && [ "$LAVA_ADMIN_PASSWORD" != "" ]
then
    echo "Creating lava admin: $LAVA_ADMIN_USERNAME"
    lava-server manage users details "$LAVA_ADMIN_USERNAME" || lava-server manage users add --passwd "$LAVA_ADMIN_PASSWORD" --staff --superuser "$LAVA_ADMIN_USERNAME"
    echo "done"
    echo
fi

if [ "$LAVA_ADMIN_TOKEN" != "" ] && [ "$LAVA_ADMIN_USERNAME" != "" ]
then
    echo "Creating lava token for: ${LAVA_ADMIN_USERNAME} with secret ${LAVA_ADMIN_TOKEN}"
    lava-server manage tokens add --secret "$LAVA_ADMIN_TOKEN" --user "$LAVA_ADMIN_USERNAME" || true
    echo "done"
    echo
fi

# Run user scripts. The database is running and migrations has been run.
for f in $(find /root/entrypoint.d/ -type f); do
    case "$f" in
        *.sh)
            echo "$0: running ${f}"
            "${f}"
            ;;
        *)
        echo "$0: ignoring ${f}"
        ;;
    esac
done

if [ "$GUNICORN" = "1" ]
then
    echo "Checking file permissions"
    if [ "$LAVA_CHECK_OWNERS" = "1" ]
    then
        check_owners
        echo "done"
    else
        echo "skipped"
    fi
    echo
    echo "Starting gunicorn3"
    start_lava_server_gunicorn
    echo "done"
    echo
fi

if [ "$APACHE2" = "1" ]
then
    echo "Starting apache2"
    start_apache2
    echo "done"
    echo
fi

if [ "$LAVA_CELERY_WORKER" = "1" ]
then
    echo "Starting lava-celery-worker"
    start_lava_celery_worker
    echo "done"
    echo
fi


if [ "$LAVA_COORDINATOR" = "1" ]
then
    echo "Starting lava-coordinator"
    start_lava_coordinator
    echo "done"
    echo
fi

if [ "$LAVA_PUBLISHER" = "1" ]
then
    echo "Starting lava-publisher"
    start_lava_publisher
    echo "done"
    echo
fi

if [ "$LAVA_SCHEDULER" = "1" ]
then
    echo "Starting lava-scheduler"
    start_lava_scheduler
    echo "done"
    echo
fi


################
# Wait forever #
################
cd /var/log/lava-server
while true
do
  tail -F django.log gunicorn.log lava-celery-worker.log lava-publisher.log lava-scheduler.log /var/log/lava-coordinator.log /var/log/apache2/lava-server.log & wait ${!}
done

###default parameters from test definition###
######
###test parameters from job submission###
######
set -e
set -x
export TESTRUN_ID=0_smoke-tests-basic
cd /lava-4212/0/tests/0_smoke-tests-basic
UUID=`cat uuid`
set +x
echo "<LAVA_SIGNAL_STARTRUN $TESTRUN_ID $UUID>"
set -x
lava-test-case linux-posix-pwd --shell pwd
lava-test-case linux-posix-uname --shell uname -a
lava-test-case linux-posix-vmstat --shell vmstat
lava-test-case linux-posix-ifconfig --shell ifconfig -a
lava-test-case linux-posix-lscpu --shell lscpu
lava-test-case linux-posix-lsb_release --shell lsb_release -a
set +x
echo "<LAVA_SIGNAL_ENDRUN $TESTRUN_ID $UUID>"

---
title: The Art of Influence
subtitle: A Practical Guide to Working with Influence Functions
author: Klaus Kähler Holst
Rdate: "`r Sys.Date()`"
output:
  # html_document:
  rmarkdown::html_vignette:
    # toc_float: true
    fig_width: 6
    fig_height: 4
    fig_caption: true
    toc: true
    toc_depth: 2
    mathjax: "https://cdnjs.cloudflare.com/ajax/libs/mathjax/2.7.5/MathJax.js?config=TeX-AMS_CHTML.js"
vignette: >
  %\VignetteIndexEntry{The Art of Influence}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
bibliography: ref.bib
---

```{r  setup,include=FALSE }
library("lava")
knitr::opts_chunk$set(
  collapse = TRUE,
  cache = TRUE,
  comment = "#>",
  dev = "svg",
  fig.ext = "svg"
  )
has_mets <- lava:::versioncheck('mets', 1)
has_geepack <- lava:::versioncheck('geepack', 1)
```

\newcommand{\arctanh}{\operatorname{arctanh}}
\newcommand{\indep}{\!\perp\!\!\!\!\perp\!}
\newcommand{\pr}{\mathbb{P}}
\newcommand{\E}{\mathbb{E}}
\newcommand{\var}{\mathbb{V}\!\text{ar}}
\newcommand{\cov}{\mathbb{C}\!\text{ov}}
\newcommand{\cor}{\mathbb{C}\!\text{or}}
\newcommand{\IC}{\operatorname{IC}}
\newcommand{\one}{\mathbf{1}}
<!-- \newcommand{\Dto}{\rightsquigarrow} -->
\newcommand{\Dto}{\overset{\mathcal{D}}{\longrightarrow}}
\newcommand{\bm}[1]{\mathbf{#1}}
\newcommand{\Pz}{P_{0}}
\newcommand{\op}{o_{P}}

# Influence functions

Estimators that have parametric convergence rates can often be fully
characterized by their *influence function* (IF), also referred to as an influence
curve or canonical gradient [@bickel_effic_adapt_estim_semip_model; @vaart_1998_asymp]. The IF allows for the direct estimation of properties of the estimator, including its asymptotic variance. Moreover,
estimates of the IF enable the simple combination and transformation of
estimators into new ones.
This vignette describes how to estimate and manipulate IFs using the R-package
`lava` [@holst_budtzjorgensen_2013].


Formally, let $Z_1,\ldots,Z_n$ be iid $k$-dimensional stochastic variables, $Z_i=(Y_{i},A_{i},W_{i})\sim \Pz$, and $\widehat{\theta}$ a
consistent estimator for the parameter $\theta\in\mathbb{R}^p$. When
$\widehat{\theta}$ is a *regular and asymptotic linear* (RAL) estimator, it has
a unique iid decomposition
\begin{align*} \sqrt{n}(\widehat{\theta}-\theta) =
\frac{1}{\sqrt{n}}\sum_{i=1}^n \IC(Z_i; \Pz) + \op(1),
\end{align*}
where the
function $\IC$ is the unique *Influence Function* s.t. $\mathbb{E}\{\IC(Z_{i};
\Pz)\}=0$ and $\var\{\IC(Z_{i}; \Pz)^{2}\}<\infty$ [@tsiatis2006semiparametric;
@vaart_1998_asymp]. The influence
function thus fully characterizes the asymptotic behaviour of the estimator and by the central limit
theorem it follows that the estimator converges weakly to a Gaussian distribution
$$
\sqrt{n}(\widehat{\theta}-\theta) \Dto
\mathcal{N}(0, \var\{\IC(Z; \Pz)\}),
$$
where the empirical variance of the plugin estimator, $\pr_{n}\IC(Z;
\widehat{P})^{\otimes 2} = \frac{1}{n}\sum_{i=1}^n \IC(Z_{i};
\widehat{P})\IC(Z_{i};
\widehat{P})^{\top}$ can be used to obtain a consistent estimate of the
asymptotic variance. Note, in practice the estimate $\widehat{P}$ used in
the plugin-estimate, needs only to capture the parts of the distribution of
$Z$ that is necessary to evaluate the IF. In some cases this nuisance
parameter can be estimated using flexible machine learning components and in
other (parametric) cases be derived directly from $\widehat{\theta}$.

The IFs are easily derived for the parameters of many parametric statistical
models as illustrated in the [next example sections](#example-generalized-linear-model). More generally, the IF can
also be derived for a smooth target parameter $\Psi:
\mathcal{P}\to\mathbb{R}$ where $\mathcal{P}$ is a family of probability
distributions forming the statistical model, which often can be left completely
non-parametric.
Formally, the parameter must be *pathwise differentiable* see [@vaart_1998_asymp]
in the sense that there exists linear bounded function  $\dot\Psi\colon
L_{2}(P_{0})\to\mathbb{R}$ such that
$[\Psi(P_{t}) - \Psi(P_{0}))]t^{-1} \to \dot\Psi(P_{0})(g)$
as $t\to 0$ for any parametric submodel $P_t$
with score model $g(z)= \partial/(\partial t) \log (p_t)(z)|_{t=0}$.  Riesz's
representation theorem then tells us that the directional derivative has a
unique representer, $\phi_{P_{0}}$ lying in the closure of the submodel score
space (the *tangent space*), s.t.
\begin{align*}
    \dot\Psi(P_0)(g) = \langle\phi_{P_0}, g\rangle =
    \int \phi_{P_0}(Z)g(X)\,dP_0
\end{align*}
The unique representer is exactly the IF which can be found by solving the above
integral equation.
For more details on how to derive influence
functions, we refer to [@targetedlearning_2011; @hines2022].

As an example we might be interested in the target parameter $\Psi(P) =
\E_P(Z)$ which can be shown to have the unique (and thereby efficient)
influence function $Z\mapsto Z-\E_P(Z)$ under the non-parametric model.
Another target parameter could be $\Psi_{a}(P) = \E_{P}[\E_{P}(Y\mid A=a, W)]$ which is often
a key interest in causal inference and which has the IF
\begin{align*}
\IC(Y,A,W; P) = \frac{\one(A=a)}{\pr(A=a\mid W)}(Y-\E_{P}[Y\mid A=a,W]) + \E_{P}[Y\mid
A=a,W] - \Psi_{a}(P)
\end{align*}
See section on [average treatment effects](#average-treatment-effects).

# Working with influence functions with `lava::estimate`

The main functions for working with influence functions are

- `estimate` which prepares a model object and estimates the IF and
  corresponding robust standard errors. Can also be used to transform model
  parameters by application of the Delta Theorem.
- `merge`  method for combining estimates via their estimated IFs
- `IC`, `coef`, `vcov` methods to extract the estimated IF, coefficients, and
  asymptotic covariance matrix
- `subset` for extracting subset of parameters and IFs
- `transform` method for transforming parameter estimates
- `labels` for renaming parameters

The `estimate` function is the primary tool for obtaining parameter estimates
and related information. It returns an object of the class type `estimate`,
which is a general container for holding information about estimated parameters.
The estimate function takes as input either a model object (the first argument
`x`), or a parameter vector and corresponding influence function (IF) matrix
specified using the `coef` and `IC` arguments. If the primary goal is to apply the
delta method or test linear hypotheses, it is also possible to provide the
asymptotic variance estimate via the `vcov` argument, without specifying the IC
matrix.

```{r estimate.syntax, eval=FALSE}
estimate(x=, ...)
estimate(coef=, IC=, ...)
estimate(coef=, vcov=, ...)
```

A typical call could look like
```{r estimate.examples, eval=FALSE}
merge(subset(estimate(x), 1), estimate(coef=p, IC=ic, id=id)) |>
  transform(function(x) c(exp(x), exp(x[1]))) |> # parameter transformation
  labels(c("a", "b")) # rename parameters
```

Direct calculations is also possible, if `a` and `b` are estimate objects we can
obtain new transformed parameter estimates with the syntax
```{r estimate.direct, eval=FALSE}
c(exp(b)^0.5, exp(b * a) / (1 + exp(-b)))
```
with the necessary derivatives exactly calculated automatically.
We refer to section
[IF building blocks](#if-building-blocks-transformations-and-the-delta-theorem) for more details.


# Examples

To illustrate the methods we consider data arising from the model $Y_{ij} \sim
Bernoulli\{\operatorname{expit}(X_{ij} + A_{i} + W_{i})\}, A_{i} \sim
Bernoulli\{\operatorname{expit}(W_{i})\}$
with independent covariates
$X_{ij}\sim\mathcal{N}(0,1), W_{i}\sim\mathcal{N}(0,1)$.
```{r sim_model}
m <- lvm() |>
  regression(y1 ~ x1 + a + w) |>
  regression(y2 ~ x2 + a + w) |>
  regression(y3 ~ x3 + a + w) |>
  regression(y4 ~ x4 + a + w) |>
  regression(a ~ w) |>
  distribution(~ y1 + y2 + y3 + y4 + a, value = binomial.lvm()) |>
  distribution(~id, value = Sequence.lvm(integer = TRUE))
```

We simulate from the model where $Y_3$ is only observed for half of the subjects
```{r simulate}
n <- 4e2
dw <- sim(m, n, seed = 1) |>
  transform(y3 = y3 * ifelse(id > n / 2, NA, 1))
Print(dw)
## Data in long format
dl <- reshape(dw,
        varying = list(paste0("y",1:4),
                       paste0("x",1:4)),
        v.names=c("y", "x"), direction="long") |>
  na.omit()
dl <- dl[order(dl$id), ]
## dl <- mets::fast.reshape(dw, varying = c("y", "x")) |> na.omit()
Print(dl)
```

## Example: population mean

Here we first consider the problem of estimating the IF of the mean. For a
general transformation $f: \mathbb{R}^k\to\mathbb{R}^p$ we have that
$$
\sqrt{n}\{\mathbb{P}_{n}f(X) - \E[f(X)]\} = \frac{1}{\sqrt{n}}\sum_{i=1}^{n}
f(X_{i}) - \E[f(X)]
$$
<!-- $$
<!-- \sqrt{n}\{\mathbb{P}_{n}f(X) - \pr(Y_{1}=1)\} = \frac{1}{\sqrt{n}}\sum_{i=1}^{n} -->
<!-- \one(Y_{1i}=1) - \pr(Y_{1}=1) -->
<!-- $$  -->
and hence for the problem of estimating the proportion of the binary outcome
$Y_1$, the IF is given by $\one(Y_{1}=1) - \pr(Y_{1}=1)$.

To estimate this parameter and its IF we will use the `estimate` function
```{r inp1}
inp <- as.matrix(dw[, c("y1", "y2")])
e <- estimate(inp[, 1, drop = FALSE], type="mean")
class(e)
e
```
The reported standard errors from the `estimate` method
are the robust standard errors obtained from the IF.
The variance estimate and the parameters can be extracted with the `vcov`
and `coef` methods.
The IF itself can be extracted with the `IC` (or  `influence`) method:
```{r ic1}
IC(e) |> Print()
```

It is also possible to simultaneously estimate the proportions of each of the two
binary outcomes
```{r inp2}
estimate(inp)
```
or alternatively the input can be a model object, here a `mlm` object:
```{r mlm}
e <- lm(cbind(y1, y2) ~ 1, data = dw) |>
  estimate()
IC(e) |> head()
```

Different methods are available for inspecting an `estimate`  object
```{r estimatemethods}
summary(e)
## extract parameter coefficients
coef(e)
## ## Asymptotic (robust) variance estimate
vcov(e)
## Matrix with estimates and confidence limits
estimate(e, null=0, level = 0.99) |> parameter()
## Influence curve
IC(e) |> head()
## Join estimates
ee <- merge(e, e)
ee
## Forest plots
plot(ee, null=0.5, digits=2)
```

<!-- The IF for both the empirical mean and variance can also be estimated directly with the `IC` method: -->
<!-- ```{r} -->
<!-- with(dw, IC(y1)) |> head() -->
<!-- ## point estimates stores in the attributes -->
<!-- with(dw, IC(y1)) |> attributes() -->
<!-- ``` -->


## Example: generalized linear model

For a $Z$-estimator defined by the score equation $E[U(Z; \theta)] = 0$, the
IF is given by \begin{align*} IC(Z; \theta) =
-\E\Big\{\frac{\partial}{\partial\theta^\top}U(\theta; Z)\Big\}^{-1}U(Z; \theta)
\end{align*}
In particular, for a maximum likelihood estimator the score, $U$, is given by the partial derivative of
the log-likelihood function.

As an example, we can obtain the estimates with robust standard errors for a
logistic regression model:
```{r glm}
g <- glm(y1 ~ a + x1, data = dw, family = binomial)
estimate(g)
```
We can compare that to the usual (non-robust) standard errors by supplying
the model-based covariance matrix explicitly:
```{r glm.std}
estimate(g, vcov = vcov(g))
# estimate(g, vcov = TRUE) alternative syntax to obtain model-based SEs
```

The IF can be extracted from the `estimate` object or directly from the
model object
```{r ifglm}
IC(g) |> head()
```


The same estimates can be obtained with a
*cumulative link regression* model which also generalizes to
ordinal outcomes. Here we consider the proportional odds model given by
\begin{align*}
\log\left(\frac{\pr(Y\leq j\mid x)}{1-\pr(Y\leq j\mid x)}\right) = \alpha_{j} + \beta^{t}x, \quad j=1,\ldots,J
\end{align*}
```{r ordreg}
ordreg(y1 ~ a + x1, dw, family=binomial(logit)) |> estimate()
```

Note that the `sandwich::estfun` function from the `sandwich` library [@r_sandwich] can also estimate the IF
for different parametric models, but does not provide the tools for combining
and transforming these.

## Example: right-censored outcomess

To illustrate the methods on survival data we will use the Mayo Clinic Primary
Biliary Cholangitis Data [@therneau00surv]

```{r mets}
library("survival")
data(pbc, package="survival")
```

The Cox proportional hazards model can be fitted with the `mets::phreg` method
which can estimate the IF for both the partial likelihood parameters and the
baseline hazard. Here we fit a survival model with right-censored event times
```{r phreg, eval=has_mets }
fit.phreg <- mets::phreg(Surv(time, status > 0) ~ age + sex, data = pbc)
fit.phreg
IC(fit.phreg) |> head()
```

The IF for the baseline cumulative hazard at a specific time point
\begin{align*}
\Lambda_0(t) = \int_0^t \lambda_0(u)\,du,
\end{align*}
where $\lambda_0(t)$ is the baseline hazard, can be estimated in similar way:
```{r phreg-baseline, eval=has_mets }
baseline <- function(object, time, ...) {
  ic <- mets::IC(object, baseline = TRUE, time = time, ...)
  est <- mets::predictCumhaz(object$cumhaz, new.time = time)[1, 2]
  estimate(NULL, coef = est, IC = ic, labels = paste0("chaz:", time))
}
tt <- 2000
baseline(fit.phreg, tt)
```

The `estimate` and `IF` methods are also available for parametric survival
models via `survival::survreg`, here a Weibull model:
```{r survreg, eval=has_mets }
survival::survreg(Surv(time, status > 0) ~ age + sex, data = pbc, dist="weibull") |>
  estimate()
```

<!-- glm poisson -->
<!-- ```{r lifetable} -->
<!-- pbc2 <- mets::lifetable(Surv(time, status > 0) ~ sex, data = pbc, breaks = seq(0, 5000, length.out = 10)) -->
<!-- head(pbc2) -->
<!-- glm(events ~ offset(log(atrisk)) + sex + factor(int.end), data = pbc2, family=poisson(log)) -->
<!-- ``` -->


## Example: random effects model / structural equation model

General structural equation models (SEMs) can be estimated with `lava::lvm`.
Here we fit a random effects probit model
$$
\pr(Y_{ij} = 1 \mid U_{i}, W_{ij})=\Phi(\mu_{j} + \beta_{j} W_{ij} + U_{i}), \quad U_{i}\sim\mathcal{N}(0,\sigma_{u}^{2}),\quad j=1,2
$$
to the simulated dataset
```{r semfit, eval=has_mets }
sem <- lvm(y1 + y2 ~ 1 * u + w) |>
  latent(~ u) |>
  ordinal(K=2, ~ y1 + y2)
semfit <- estimate(sem, data = dw)

## Robust standard errors
estimate(semfit)
```

## Example: quantile

Let $\beta$ denote the $\tau$th quantile of $X$, with IF
\begin{align*}
\IC(x; \Pz) = \tau - \one(x\leq \beta)f_{0}(\beta)^{-1}
\end{align*}

where $f_{0}$ is the density function of $X$.

To calculate the variance estimate, an estimate of the density is thus needed which can be
obtained by a kernel estimate. Alternatively,
the resampling method of [@zenglin2008] can be applied.
Here we use a kernel smoother (additional arguments to the `estimate` function
are parsed on to `stats::density.default`) to estimate the quantiles and IF for
the 25%, 50%, and 75% quantiles of $W$ and $X_1$
```{r quantiles}
eq <- estimate(dw[, c("w", "x1")], type = "quantile", probs = c(0.25, 0.5, 0.75))
eq
IC(eq) |> head()
```


# Combining influence functions

A key benefit of working with the IFs of estimators is that this allows
for transforming or combining different estimates while easily deriving
the resulting IF and thereby asymptotic distribution of the new estimator.

Let $\widehat{\theta}_{1}, \ldots, \widehat{\theta}_{M}$ be $M$
different estimators with decompositions
\begin{align*}
\sqrt{n}(\widehat{\theta}_{m}-\theta_{m}) = \frac{1}{\sqrt{n}}\sum_{i=1}^{n}
\IC_m(Z_i; \Pz) + \op(1)
\end{align*}
based on iid data $Z_1,\ldots,Z_n$. It then follows immediately
[@vaart_1998_asymp Theorem 18.10[vi]] that the
joint distribution of
$\widehat{\theta} - {\theta} =
(\widehat{\theta}_{1}^{\top},\ldots,\widehat{\theta}_{M}^{\top})^\top -
({\theta}_{1}^{\top},\ldots,{\theta}_{M}^{\top})^\top$
is given by
\begin{align*}
\sqrt{n}(\widehat{\theta}-\theta) &= \frac{1}{\sqrt{n}}\sum_{i=1}^{n}
\underbrace{[\IC_{1}(Z_i; \Pz)^\top,\ldots,\IC_{M}(Z_i; \Pz)^\top]^{\top}}_{\overline{\IC}(Z_i; \Pz)} + \op(1) \\
&\Dto \mathcal{N}(0,\Sigma)
\end{align*}
by the CLT, and under regulatory conditions
$\mathbb{P}_{n}\overline{\IC}(Z_i; \widehat{P})^{\otimes 2} \overset{P}{\longrightarrow}\Sigma$
as $n\to\infty$.

To illustrate this we consider two marginal logistic regression models fitted
separately for $Y_1$ and $Y_2$ and combine the estimates and IFs using the
`merge` method
```{r glmmarg}
g1 <- glm(y1 ~ a, family=binomial, data=dw)
g2 <- glm(y2 ~ a, family=binomial, data=dw)
e <- merge(g1, g2)
summary(e)
```
As we have access to the joint asymptotic distribution we can for example test
for whether the odds-ratio is the same for the two responses:

```{r hypo1}
estimate(e, cbind(0,1,0,-1), null=0)
```
More details an be found in the Section on [hypothesis testing](#linear-contrasts-and-hypothesis-testing).

## Imbalanced data

Let $O_{1} = (Z_{1}R_{1}, R_{1}), \ldots, O_{N}=(Z_{N}R_{N}, R_{N})$ be iid with
$R_{i}\indep Z_i$ and let the full-data IF for some estimator of a parameter
$\theta\in\mathbb{R}^p$ be $IC(\cdot; \Pz)$. For convenience let the data be ordered
$R_{i}=\one(i\leq n)$ where $n$ is the number of observed data points, then the complete-case estimator is consistent and based
on same IF

\begin{align*}
\sqrt{n}(\widehat{\theta}-\theta) = \frac{1}{\sqrt{n}}\sum_{i=1}^n IC(Z_i; \Pz) + \op(1).
\end{align*}

This estimator can also be decomposed in terms of the observed data
$O_1,\ldots,O_N$ noting that

\begin{align*}
\sqrt{N}(\widehat{\theta}-\theta) = \frac{1}{\sqrt{N}}\sum_{i=1}^N IC(Z_i;
P)\frac{R_i N}{n} + \op(1).
\end{align*}

where the term $\frac{R_i N}{n}$ corresponds to an inverse probability
weighting with the empirical plugin estimate of the proportion of observed data $R=1$.
Under a missing completely at random assumption we can therefore combine estimators that are estimated
on different datasets. Let the observed data be
$(Z_{11}R_{11}, R_{11}, Z_{21}R_{21}, R_{21}), \ldots, (Z_{1N}R_{1N}, R_{1N},
Z_{2N}R_{2N}, R_{2N}))$ with complete-case estimators $\widehat{\theta}_1$ and
$\widehat{\theta}_2$ for parameters $\theta_1$ and
$\theta_2$ based on $(Z_{11}R_{11}, \ldots, Z_{1N}R_{1N})$ and
$(Z_{21}R_{21}, \ldots, Z_{2N}R_{2N})$, respectively, and let the
corresponding IFs be
$IC_{1}(\cdot; \Pz)$ and $IC_{2}(\cdot;\ P)$. It then follows that

\begin{align*}
\sqrt{N}\left\{
\begin{pmatrix}
\widehat{\theta}_1 \\
\widehat{\theta}_2
\end{pmatrix}
-
\begin{pmatrix}
\vphantom{\widehat{\theta}_1}\theta_1 \\
\vphantom{\widehat{\theta}_1}\theta_2
\end{pmatrix}
\right\}
=
\frac{1}{\sqrt{N}}\sum_{i=1}^N
\begin{pmatrix}
IC_1(Z_{1i}; \Pz)\frac{R_{1i}N}{R_{1\bullet}} \\
IC_2(Z_{2i}; \Pz)\frac{R_{2i}N}{R_{2\bullet}}
\end{pmatrix}
+ \op(1)
\end{align*}

with $R_{k\bullet} = \sum_{i=1}^{N}R_{ki}.$ Returning to the example, we can
combine the marginal estimates of two model objects that have been estimated
from different datasets (as the outcome $Y_3$ is only available in half of the
data) with the `merge` function

```{r glmmargmis}
g2 <- glm(y2 ~ 1, family = binomial, data = dw)
summary(g2)
dwc <- na.omit(dw)
g3 <- glm(y3 ~ 1, family = binomial, data = dwc)
summary(g3)

e2 <- estimate(g2, id = dw$id)
e3 <- estimate(g3, id = "id", data=dwc)

ecomb <- merge(e2, e3)
IC(ecomb) |> Print()
vcov(ecomb)
## Same marginals as
list(vcov(e2), vcov(e3))
```

Note, it is also possible to directly specify the id-variables in the `merge` call:
```{r merge}
merge(e2, e3, id = list(dw$id, dwc$id))
```

In the above example the `id` argument defines the identifier that makes it
possible to link the rows in the different IFs that should be glued together.
If omitted then the `id` will automatically be extracted from the model-specific
`IC` method (deriving it from the original data.frame used for estimating the
model). This automatically works with all models and `IC` methods described in this document.
```{r estimatenoid}
estimate(g2) |>
  IC() |> head()
vcov(merge(estimate(g2), estimate(g3)))
merge(estimate(g2), estimate(g3)) |>
  (rownames %++% head %++% IC)()
```

To force that the id variables are not overlapping between the merged model
objects, i.e., assuming that there is complete independence between the
estimates, the argument `id=NULL` can be used
```{r merge_idnull}
merge(g1, g2, id = NULL) |> (Print %++% IC)()
merge(g1, g2, id = NULL) |> vcov()
```

## Renaming and subsetting parameters

To only keep a subset of the parameters the `keep` argument can be used.
```{r mergekeep}
merge(g1, g2, keep = c("(Intercept)", "(Intercept).1"))
```
The argument can be given either as character vector or a vector of indices:
```{r merge2}
merge(g1, g2, keep=c(1, 3))
```
or as a vector of perl-style regular expressions
```{r merge3}
merge(g1, g2, keep = "cept", regex = TRUE)
merge(g1, g2, keep = c("\\)$", "^a$"), regex = TRUE, ignore.case = TRUE)
```

When merging estimates unique parameter names are created. It is also possible
to rename the parameters with the `labels` argument
```{r merge4}
merge(g1, g2, labels = c("a", "b", "c")) |> estimate(keep = c("a", "c"))
merge(g1, g2,
      labels = c("a", "b", "c"),
      keep = c("a", "c")
)
estimate(g1, labels=c("a", "b"))
```

Finally, the `subset` argument can be used to subset the parameters and IFs
before the actual merging is being done
```{r merge5}
merge(g1, g2, subset="(Intercept)")
```

## Clustered data (non-iid case)

Let $Z_i = (Z_{i1},\ldots,Z_{iN_{i}})$ and assume that
$(Z_{i}, N_{i}) \sim P$, $i=1,\ldots,n$ are iid and $N_i\indep
Z_{ij}$. The variables $Z_{i1},\ldots,Z_{iN_{i}}$ we assume are exchangeable
but not necessarily independent. Define $N = \sum_{i=1}^{n} N_i$, and
assume that a parameter estimate, $\widehat{\theta}\in\mathbb{R}^p$ has the decomposition
$$
\sqrt{N}(\widehat{\theta}-\theta) =
\frac{1}{\sqrt{N}}
\sum_{i=1}^{n} \sum_{k=1}^{N_{i}} IC(Z_{ik}; \Pz) + \op(1).
$$
It then follows that
$$
\sqrt{n}(\widehat{\theta}-\theta) =
\frac{1}{\sqrt{n}}
\sum_{i=1}^{n} \widetilde{\IC}(Z_{i}; \Pz) + \op(1)
$$
with $\widetilde{\IC}(Z_{i}; \Pz)
= \sum_{k=1}^{N_{i}} \frac{n}{N}IC(Z_{ik}; \Pz)$, $i=1,\ldots,n$ which are iid
an therefore admits the usual CLT to derive the asymptotic variance of $\widehat{\theta}$.
Turning back to the example data, we can estimate the marginal model
```{r cluster1}
g0 <- glm(y ~ a + w + x, data = dl, family = binomial())
```
The asymptotic variance estimate ignoring that the observations are not
independent is not consistent. Instead we can calculate the cluster robust
standard errors from the above iid decomposition
```{r cluster2}
estimate(g0, id=dl$id)
```
We can confirm that this situation is equivalent to the variance estimates we
obtain from a GEE marginal model with working independence correlation structure [@r_geepack]
```{r geepack, eval=has_geepack }
gee0 <- geepack::geeglm(y ~ a + w + x, data = dl, id = dl$id, family=binomial)
summary(gee0)
```

## Computational aspects

Working with large and potentially multiple different IFs can be memory-intensive. A
remedy is to use the idea of aggregating the IFs by introducing a random
coarser grouping variable. Following the same arguments as in the previous
section, the aggregated IF will still be iid and allows us to estimate the
asymptotic variance. Obviously, the same grouping must be used across estimates
when combining IFs.
```{r aggregate}
set.seed(1)
y <- cbind(rnorm(1e5))
N <- 2e2 ## Number of aggregated groups, the number of observations in the new IF
id <- foldr(nrow(y), N, list=FALSE)
Print(cbind(table(id)))

## Aggregated IF
e <- estimate(cbind(y), id = id)
object.size(e)
e
```

# IF building blocks: transformations and the delta theorem

Let $\phi\colon \mathbb{R}^p\to\mathbb{R}^m$ be differentiable at $\theta$
and assume that $\widehat{\theta}_n$ is RAL estimator with IF given by
$\IC(\cdot; \Pz)$ such that
\begin{align*}
\sqrt{n}(\widehat{\theta}_n - \theta) =
\frac{1}{\sqrt{n}}\sum_{i=1}^n \IC(Z_i; \Pz) + \op(1),
\end{align*}
then by the delta method [@vaart_1998_asymp Theorem 3.1]
\begin{align*}
\sqrt{n}\{\phi(\widehat{\theta}_n) - \phi(\theta)\}  =
\frac{1}{\sqrt{n}}\sum_{i=1}^n \nabla\phi(\theta)\IC(Z_i; \Pz) + \op(1),
\end{align*}

where $\phi\colon \theta\mapsto
(\phi_{1}(\theta),\ldots,\phi_{m}(\theta))^\top$ and $\nabla$ is the partial
derivative operator
\begin{align*}
\nabla\phi(\theta) =
\begin{pmatrix}
\tfrac{\partial}{\partial\theta_1}\phi_1(\theta) & \cdots &
\tfrac{\partial}{\partial\theta_p}\phi_1(\theta) \\
\vdots & \ddots & \vdots \\
\tfrac{\partial}{\partial\theta_1}\phi_m(\theta) & \cdots &
\tfrac{\partial}{\partial\theta_p}\phi_m(\theta) \\
\end{pmatrix}.
\end{align*}

Together with the ability to derive the joint IF from marginal IFs, this
provides us with a powerful tool for constructing new estimates using the IFs as
the fundamental building blocks.

To apply the delta method the transformation of the parameters function must be
supplied to the `estimate` method (argument `f`)
```{r delta1}
estimate(g1, sum)
estimate(g1, function(p) list(a = sum(p))) # named list
## Multiple parameters
estimate(g1, function(x) c(x, x[1] + exp(x[2]), inv = 1 / x[2]))
estimate(g1, exp)
```
The gradient can be provided as the attribute `grad` and otherwise numerical differentiation is applied.


An often more convenient syntax is to directly transform the `estimate` objects
via standard mathematical R operations. An advantage of this is also that the
necessary derivatives are automatically calculated exactly.
To illustrate this consider the following simple example where we consider two
`estimate` objects each with a single parameter estimate
```{r estimate_calculus_models}
a <- estimate(coef=c("a"=0.5), IC=scale(rnorm(10)), id=1:10)
b <- estimate(coef=c("b"=0.8), IC=scale(rnorm(10)), id=1:10)
```

Parameter transformation can now be calculated directly as in the following examples
```{r estimate_calc_ex}
a * b
(3 * cos(a) / sqrt(b) + 1) / a
c(sum=sum(e), sum2=a+b,
  prod=prod(e), prod2=a*b) # sum and prod function
e <- c(a,b) # merge
e %*% e # inner prod.
c(1, 2) %*% e
c(pow = a^b) # power-function, rename parameter
a^c(0.5, 2)
c(e["a"] * e["b"] / a, e["b"])
```

For the `%*%` operator we can also use a general contrast matrix
(see also Section on [Linear contrasts](#linear-contrasts-and-hypothesis-testing)
```{r estimate-contrast}
B <- rbind(c(1,-1), c(1,0), c(0,1))
B %*% e
B %*% e == c(1,1,0)
```


The following transformations are implemented

- trigonometric functions: `cos`, `sin`, `tan`
- inverse trigonometric functions: `acos`, `asin`, `atan`
- hyperbolic functions: `cosh`, `sinh`, `tanh`
- inverse hyperbolic functions: `acosh`, `asinh`, `atanh`
- other mathematical functions: `log`, `log1p`, `exp`, `expm1`, `sqrt`
- mathematical operators: `+`, `-`, `*`, `/`, `^`,
 `prod`, `sum`, `%*%`

and any compositions of these functions, which also means we can immediately
 apply most existing user-defined functions. To give an example consider the `logit` function
```{r}
lava::logit
logit(b)
expit(c(a,b))
```

In the above example we also demonstrated the user of the subset operator `[]`
and the merge method `c()`. These are basically equivalent to the `subset` and
`merge` methods. There is also a `transform` method for constructing parameter
transformations above, and the `labels` function for renaming parameters.
Together we can combine these methods via the pipe operator to create readable
code:
```{r estimate-pipe}
merge(a, b) |>  # merges the two `estimate` objects
  transform(prod) |> # calculates product of parameter estimates
  subset(1) |> # nothing happens here as the result was already 1-dim.
  labels("prod") # rename parameter
```

Finally, the `with` function can be used to reference parameter names
```{r estimate-with}
e <- c("e1"=a, "e2"=b)
with(e, c(est = e1*e2))
```


## Example: Pearson correlation
As a simple toy example consider the problem of estimating the covariance of two
variables $X_1$ and $X_2$
\begin{align*}
\widehat{\cov}(X_1,X_2) = \mathbb{P}_n(X_1-\mathbb{P}_n X_1)(Y_1-\mathbb{P}_n Y_1).
\end{align*}
It is easily verified that the IF of the sample estimate of $(\E X_{1}, \E
X_{2}, \E\{X_{1}X_{2}\})^\top$ given by
is $\IC(X1,X2; \Pz) = (X_{1}-\E X_{1}, X_{2}-\E X_{2},
X_{1}X_{2}-\E\{X_{1}X_{2}\})^\top$. By the delta theorem with $\phi(x,y,z) =
z-xy$ we have  $\nabla\phi(x,y,z) = (-y, -x, 1)^\top$ and thus the IF for the
sample covariance estimate becomes
\begin{align*}
\IC_{x_1, x_2}(X_1, X_2; \Pz) = (X_1 - \E X_1)(X_2 - \E X_2) - \cov(X_1,X_2)
\end{align*}

We can implement this directly using the `estimate` function via the `IC`
argument which allows us to provide a user-specificed IF and with the point
estimate given by the `coef` argument
```{r cov}
Cov <- function(x, y, ...) {
  est <- mean(x * y)-mean(x)*mean(y)
    estimate(
      coef = est,
      IC = (x - mean(x)) * (y - mean(y)) - est,
      ...
    )
}
with(dw, Cov(x1, x2))
```

As an illustration we could also derive this estimate from simpler building
blocks of $\E X_{1}$, $\E X_{2}$, and $\E(X_{1}X_{2})$.
```{r cov2}
est <- lm(cbind(x1, x2, x1 * x2) ~ 1, data = dw) |>
  estimate(labels = c("E1", "E2", "E12"))
est

est["E12"] - est["E2"]*est["E1"]
# transform(e1, function(x) c(x, cov=with(as.list(x), E12 - E2* E1))) # Same result
```
The variance estimates can be estimated in the same way and combined to estimate the correlation
```{r rho}
v12 <- with(dw, Cov(x1, x2, id = id))
v1  <- with(dw, Cov(x1, x1, id = id))
v2  <- with(dw, Cov(x2, x2, id = id))

rho <- c(rho = v12 / sqrt(v1 * v2))
rho
```

by using a variance stabilizing transformation, Fishers
$z$-transform [@lehmann2023_testing],
$z = \arctanh(\widehat{\rho}) =
\frac{1}{2}\log\left(\frac{1+\widehat{\rho}}{1-\widehat{\rho}}\right)$,
confidence limits with general better coverage can be obtained
```{r tanh}
estimate(atanh(rho), back.transform = tanh)
```
The confidence limits are calculated on the $\arctanh$-scale and transformed
back to the original correlation scale via the `back.transform` argument.
In this case, where the estimates are far away from the boundary of the parameter space, the variance
stabilizing transform does almost not have any impact, and the confidence limits
agrees with the original symmetric confidence limits.

## Linear contrasts and hypothesis testing

An important special case of parameter transformations are linear
transformations. A particular interest may be formulated around testing
null-hypotheses of the form
\begin{align*}
H_0\colon\quad \bm{B}\theta = \bm{b}_0
\end{align*}

where $\bm{B}\in\mathbb{R}^{m\times p}$ is a matrix of estimable contrasts and
$\bm{b}_0\in\mathbb{R}^{m}$.

As an example consider marginal models for the binary response variables
  $Y_1, Y_2, Y_3, Y_4$
```{r estglmmerge}
g <- lapply(
  list(y1 ~ a, y2 ~ a, y3 ~ a), #, y4 ~ a+x4),
  function(f) glm(f, family = binomial, data = dw)
)
gg <- Reduce(merge, g)
gg
```

A linear transformation can be specified via the `f` as a matrix argument
instead of function object
```{r contrast1}
B <- cbind(0,1, 0,-1, 0,0)
estimate(gg, B)
```

The $\bm{b}_0$ vector (default assumed to be zero) can be specified via the `null` argument
```{r estcontrast1}
estimate(gg, B, null=1)
```

For testing multiple hypotheses we use that
$(\bm{B}\widehat{\theta}-\bm{b}_0)^{\top}
(\bm{B}\widehat{\Sigma}\bm{B}^{\top})^{-1}
(\bm{B}\widehat{\theta}-\bm{b}_0) \sim
 \chi^2_{\operatorname{rank}(B)}$ under the null hypothesis where
 $\widehat{\Sigma}$ is the estimated variance of $\theta$ (i.e.,
`vcov(gg)`)
```{r estcontrast2}
B <- rbind(cbind(0,1, 0,-1, 0,0),
           cbind(0,1, 0,0, 0,-1))
estimate(gg, B)
```

Such linear statistics can also be specified directly as expressions of the parameter names
```{r}
estimate(gg, a + a.1, 2*a - a.2, a, null=c(2,1,1))
```

We refer to the function `lava::contr`  and `lava::parsedesigns` for defining contrast matrices.
```{r contr}
contr(list(1, c(1, 2), c(1, 4)), n = 5)
```
A particular useful contrast is the following for considering all pairwise
comparisons of different exposure estimates:

```{r pairwise.diff}
pairwise.diff(3)
estimate(gg, pairwise.diff(3), null=c(1,1,1), use=c(2,4,6))
```

When conducting multiple tests each at a nominal-level of $\alpha$ the overall
type I error is not controlled at $\alpha$-level.
The influence function also allows for adjusting for multiple comparisons.
Let $Z_{1},\ldots,Z_{p}$ denote $Z$-statistics from $p$ distinct two-sided
hypothesis tests which we will assume is asymptotically distributed under the
null hypothesis as a zero-mean Gaussian distribution with correlation matrix
$R.$ Let $ Z_{max} = \max_{i=1,\ldots,p} |Z_i|$ then the family-wise error
rate (FWER) under the null can be approximated by
$$
P(Z_{max} > z) =
1-\int_{-z}^{z} \cdots \int_{-z}^{z} \phi_{R}(x_{1},\ldots,x_{p})
\,dx_{1}\cdots\,dx_{p}
$$
where $\phi_{R}$ is the multivariate normal density
function with mean 0 and variance given by the correlation matrix $R$. The
adjusted $p$-values can then be calculated as
$$
P(Z_{max} >
\Phi^{-1}(1-p/2))
$$
where $\Phi$ is the standard Gaussian CDF. As described in
in [@RitzPipper_multcomp] the joint distribution of $Z_{1},\ldots,Z_{p}$
can be estimated from the IFs. This is implemented in the `alpha_zmax` method

```{r pcorrect, eval=has_mets }
gg0 <- estimate(gg, use="^a", regex=TRUE, null=rep(.8, 3))
alpha_zmax(gg0)
```

While this always yields a more powerful test compared to Bonferroni
adjustments, a more powerful closed-testing procedure [@marcus1976], can be generally
obtained by considering all intersection hypotheses.

![*Figure: closed testing via Wald tests of all intersections hypotheses.*](../man/figures/closedtesting.svg){width=60%}

To reject the $H_{1}$ at
an correct $\alpha$-level we should test all intersection hypotheses involving
$H_{1}$ and check if there all are rejected at the $\alpha$-level. The
adjusted $p$-values can here be obtained as the maximum $p$-value across all
the composite hypothesis tests. Unfortunately, this only works for relatively
few comparisons as the number of tests grows exponentially.

```{r closedtesting, eval=has_mets}
closed_testing(gg0, test = test_wald)
```


## Averaging

Some parameters of interest are expressed as averages over functions of the
observed data and estimated parameters of a model. The asymptotic distribution
can in some of these cases also be derived from the influence function.
Let $Z_{1},\ldots,Z_{n}$ be iid observations, $Z_{1}\sim \Pz$ and let  $X_{i}\subset Z_{i}$.

Assume that $\widehat{\theta}$ is RAL estimator of $\theta\in\Omega\subset \mathbb{R}^{p}$
$$\sqrt{n}(\widehat{\theta}-\theta) = \frac{1}{\sqrt{n}}\sum_{i=1}^{n}\phi(Z_{i}; \Pz) + \op(1).
$$
Let $f:\mathcal{X}\times\Omega\to\mathbb{R}$ be continuous differentiable in $\theta$
$$\sqrt{n}\{f(X; \widehat{\theta})-f(X; \theta)\} =
\frac{1}{\sqrt{n}}\nabla_{\theta}f(X;\theta)\sum_{i=1}^{n}\phi(Z_{i}; P) + \op(1).
$$

Let $\Psi = \Pz f(X;\theta)$ and $\widehat{\Psi} = P_{n} f(X;\widehat{\theta})$. $\Pz$ and $P_{n}$ are here everywhere the integrals wrt. $X$. It is easily verified that
\begin{align*}
\widehat{\Psi}-\Psi &= (P_{n}-\Pz)(f(X; \theta)-\Psi) + P[f(X;\widehat{\theta})-f(X;\theta)] \\
&\quad + (P_{n}-\Pz)[f(X;\widehat{\theta})-f(X;\theta)]
\end{align*}

From Lemma 19.24 [@vaart_1998_asymp] it follows that for the last term
\begin{align*}
\sqrt{n}(P_{n}-\Pz)[f(X;\widehat{\theta})-f(X;\theta)] = \op(1)
\end{align*}
when $f$ for example is Lipschitz and more generally when $f(X;\theta)$ forms a $\Pz$-Donsker class.

It therefore follows that
\begin{align*}
\sqrt{n}(\widehat{\Psi}-\Psi) &= \sqrt{n}P_{n}(f(X; \theta)-\Psi) +
\frac{1}{\sqrt{n}}P\nabla_{\theta}f(X;\theta)\sum_{i=1}^{n}\phi(Z_{i}; \Pz) + \op(1) \\
&=
\frac{1}{\sqrt{n}}\sum_{i=1}^{n}\{f(X;\theta)-\Psi\} +
\frac{1}{\sqrt{n}}P\nabla_{\theta}f(X;\theta)\sum_{i=1}^{n}\phi(Z_{i}, \Pz) + \op(1)
\end{align*}
Hence the IF for $\widehat{\Psi}$ becomes
$$
IC(Z; \Pz) = f(X;\theta)-\Psi +
[\Pz\nabla_{\theta}f(X;\theta)]\phi(Z).
$$


Turning back to the example we can estimate the logistic regression model
$\operatorname{logit}(E\{Y_1 | A,X_1,W\}) = \beta_0 + \beta_a A + \beta_{x_1}
X_1 + \beta_w W$, and from this we want to estimate the target parameter
$$
\theta(P) = \E_{P}[E(Y\mid A=1, X_{1}, W)].
$$
To do this we need first to estimate the model and then define a function that
gives the predicted probability $\pr(Y=1\mid,A=a,X_{1},W)$ for any observed
values of $X_1,W$ but with the treatment variable $A$ kept fixed at the
value $1$
```{r estpred}
g <- glm(y1 ~ a + x1 + w, data=dw, family=binomial)
pr <- function(p, data, ...)
  with(data, expit(p[1] + p["a"] + p["x1"]*x1 + p["w"]*w))
pr(coef(g), dw) |> head()
```
The target parameter can now be estimated with the syntax
```{r average}
id <- foldr(NROW(dw), 100, list=FALSE)
ea <- estimate(g, pr, average=TRUE, id=id)
ea
IC(ea) |> head()
```

## Average Treatment Effects

Let the observed data be $Z = (W,A,Y)$, where $W$ are covariates, $A$ a binary treatment
variable, and $Y$ the outcome.
In the following we are interested in estimating the target parameter
$\psi_a(P)
= E_P[Y(a)]$, where $Y(a)$ is the *potential outcome* we would have
observed if treatment $a$ had been administered, possibly contrary to
the actual treatment that was observed, i.e., $Y = Y(A)$. To assess the
treatment effect we can then the consider the *average treatment effect*
(ATE) $$E_P[Y(1)]-E_P[Y(0)],$$ or some other contrast of interest
$g(\psi_1(P), \psi_0(P))$. Under the following assumptions

1)  Stable Unit Treatment Values Assumption (the treatment of a specific
    subject is not affecting the potential outcome of other subjects)
2)  Positivity, $P(A\mid W)>\epsilon$ for some $\epsilon>0$ and baseline
    covariates $W$
3)  No unmeasured confounders, $Y(a)\perp \!\!\! \perp A|W$


then the target parameter can be identified from the observed data
distribution as $$E(E[Y|W,A=a]) = E(E[Y(a)|W]) = E[Y(a)]$$ or
$$E[Y I(A=a)/P(A=a|W)] = E[Y(a)].$$

This suggests estimators based on outcome regression ($g$-computation)
or inverse probability weighting. More generally, under the above
assumption we can constructor a *one-step* estimator from the *Efficient
Influence Function* combining these two
$$ E\left[\frac{I(A=a)}{\Pi_a(W)}(Y-Q(W,A)) + Q(W,a)\right].$$
In practice, this requires plugin estimates of both the outcome model,
$Q(W,A) := E(Y\mid A, W)$, and of the treatment propensity model
$\Pi_a(W) := P(A=a\mid W)$. The corresponding estimator is consistent
even if just one of the two nuisance models is correctly specified.

To illustrate this we manually construct the above estimating equations based on
logistic regressions models for the two nuisance models
```{r ate-nuisancemodels}
qmod <- glm(y1 ~ a * w, family = binomial, data = dw) # E(Y|W,A) := Q(W,A)
amod <- glm(a ~ w, family = binomial, data = dw) # P(A=1|W) = Pi_1(W)
q0 <- predict(qmod, transform(dw, a=0), type="response") # Q(W,0)
q1 <- predict(qmod, transform(dw, a=1), type="response") # Q(W,1)
p1 <- predict(amod, dw, type="response") # P(A=1|W)
e0 <- with(dw, (1-a) / (1-p1) * (y1 - q0) + q0)
e1 <- with(dw, a / p1 * (y1 - q1) + q1)
head(cbind(e0, e1))
```

We can then easily construct estimates of the two expected potential outcomes
and estimate their joint asymptotic distribution
```{r est-potentialoutcomes}
est0 <- estimate(coef=mean(e0), IC=e0-mean(e0)) # E[Y(0)]
est1 <- estimate(coef=mean(e1), IC=e1-mean(e1)) # E[Y(1)]

potential_outcomes <- merge(est0, est1, paired=TRUE, labels=c("y(0)", "y(1)"))
potential_outcomes
head(IC(potential_outcomes))
vcov(potential_outcomes)
```

Finally, we can obtain the Average Treatment Effect (ATE) (risk-difference)
\begin{align*}
E[Y(1)]  - E[Y(0)]
\end{align*}

```{r est-ate}
estimate(potential_outcomes, cbind(-1, 1), labels="ate")
```

Alternatively, we could get the estimate of the treatment effect on the log-odds
scale:
```{r est-ate-or}
est <- with(potential_outcomes, logit(`y(1)`)-logit(`y(0)`))
est
transform(est, labels="OR", back.transform=exp)
#logor <- function(p) logit(p[2]) - logit(p[1])
#transform(potential_outcomes, logor, labels="logOR")
#transform(potential_outcomes, logor, labels="OR", back.transform=exp)
```

We refer to the `targeted` package [@r-targeted] for a general implementation
where the nuisance models can be based on machine learning models using
cross-fitting. The above estimation can be obtained with the  `targeted::cate` function in the following way
```{r est-cate}
#| eval: false
amod <- targeted::learner_glm(a ~ w, family=binomial)
qmod <- targeted::learner_glm(y1 ~ a * w, family=binomial)
est <- targeted::cate(qmod, amod, data=dw, second.order = FALSE)
#>             Estimate Std.Err    2.5%  97.5%   P-value
#> E[y1(1)]      0.7050 0.03721 0.63207 0.7779 4.934e-80
#> E[y1(0)]      0.5359 0.03535 0.46666 0.6052 6.290e-52
#> ───────────
#> (Intercept)   0.1691 0.04996 0.07115 0.2670 7.138e-04
```

# SessionInfo

```{r sessionInfo}
sessionInfo()
```

# Bibliography

---
# YAML header created by ox-ravel
title: Estimating partial correlations with lava
author: Klaus Kähler Holst
date: "`r Sys.Date()`"
output:
  rmarkdown::html_vignette:
    fig_caption: yes
    fig_width: 6
    fig_height: 4
vignette: >
  %\VignetteIndexEntry{Estimating partial correlations with lava}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
bibliography: ref.bib
---
<!-- correlation.Rmd is generated from correlation.org. Please edit that file -->

```{r  include=FALSE }
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  dev = "svg",
  fig.ext = "svg"
)
mets <- lava:::versioncheck('mets', 1)
```

\newcommand{\arctanh}{\operatorname{arctanh}}

This document illustrates how to estimate partial correlation
coefficients using `lava`.

Assume that $Y_{1}$ and $Y_{2}$ are conditionally normal
distributed given $\mathbf{X}$ with the following linear structure
$$Y_1 = \mathbf{\beta}_1^{t}\mathbf{X} + \epsilon_1$$
$$Y_2 = \mathbf{\beta}_2^{t}\mathbf{X} + \epsilon_2$$
with covariates $\mathbf{X} = (X_1,\ldots,X_k)^{t}$ and measurement errors
$$\begin{pmatrix}
    \epsilon_{1} \\
    \epsilon_{2}
    \end{pmatrix} \sim \mathcal{N}\left(0, \mathbf{\Sigma} \right), \quad \mathbf{\Sigma}
    =
    \begin{pmatrix}
    \sigma_1^2 & \rho\sigma_{1}\sigma_{2} \\
    \rho\sigma_{1}\sigma_{2} & \sigma_2^2
    \end{pmatrix}.$$

```{r  plotload, results="hide",message=FALSE,warning=FALSE }
library("lava")
m0 <- lvm(y1+y2 ~ x, y1 ~~ y2)
edgelabels(m0, y1 + y2 ~ x) <- c(expression(b[1]), expression(b[2]))
edgelabels(m0, y1 ~ y2) <- "r"
plot(m0, layoutType="circo")
```

Here we focus on inference with respect to the correlation parameter $\rho$.


# Simulation

As an example, we will simulate data from this model with a single covariate.

The model can be specified (here using the pipe notation)
with the following syntax where the correlation parameter here is
given the label '`r`':

```{r  m0 }
m0 <- lvm() |>
  covariance(y1 ~ y2, value='r') |>
  regression(y1 + y2 ~ x)
```

To simulate from the model we can now simply use the `sim` method. The
parameters of the models are set through the argument `p` which must be a
named numeric vector of parameters of the model. The parameter names
can be inspected with the `coef` method

```{r  coef }
coef(m0, labels=TRUE)
```

The default simulation parameters are zero for all intercepts (`y1`, `y2`)
and one for all regression coefficients (`y1~x`, `y2~x`) and residual
variance parameters (`y1~~y1`, `y2~~y2`).

```{r  sim }
d <- sim(m0, 500, p=c(r=0.9), seed=1)
head(d)
```

Under Gaussian and coarsening at random assumptions we can also
consistently estimate the correlation in the presence of censoring or
missing data. To illustrate this, we add left and right censored data
types to the model output using the `transform` method.

```{r  defcens }
cens1 <- function(threshold,type='right') {
  function(x) {
    x <- unlist(x)
    if (type=='left')
      return( survival::Surv(pmax(x,threshold), x>=threshold, type='left') )
      return ( survival::Surv(pmin(x,threshold), x<=threshold) )
  }
}

m0 <-
  transform(m0, s1 ~ y1, cens1(-2, 'left')) |>
  transform(s2 ~ y2, cens1(2,  'right'))
```

```{r  sim2 }
d <- sim(m0, 500, p=c(r=0.9), seed=1)
head(d)
```

# Estimation and inference

The Maximum Likelihood Estimate can be obtainted using the `estimate` method:

```{r  est1 }
m <- lvm() |>
     regression(y1 + y2 ~ x) |>
     covariance(y1 ~ y2)

e <- estimate(m, data=d)
e
```

The estimate `y1~~y2` gives us the estimated covariance between the
residual terms in the model. To estimate the correlation we can apply
the delta method using the `estimate` method again

```{r  delta }
estimate(e, function(p) p['y1~~y2']/(p['y1~~y1']*p['y2~~y2'])^.5)
```

Alternatively, the correlations can be extracted using the `correlation` method

```{r  correlation }
correlation(e)
```

Note, that in this case the confidence intervals are constructed
by using a variance stabilizing transformation, Fishers
$z$-transform [@lehmann2023_testing],

$$
z = \arctanh(\widehat{\rho}) =
\frac{1}{2}\log\left(\frac{1+\widehat{\rho}}{1-\widehat{\rho}}\right)
$$

where $\widehat{\rho}$ is the MLE.  This estimate has an approximate
asymptotic normal distribution
$\mathcal{N}(\arctanh(\rho),\frac{1}{n-3})$. Hence a asymptotic 95%
confidence interval is given by

$$\widehat{z} \pm \frac{1.96}{\sqrt{n-3}}$$

and the confidence interval for $\widehat{\rho}$ can directly be calculated by
the inverse transformation:

$$\widehat{\rho} = \tanh(z) = \frac{e^{2z}-1}{e^{2z}+1}.$$

This is equivalent to the direct calculations using the delta method
(except for the small sample bias correction $3$) where the
estimate and confidence interval are transformed back to the original
scale using the `back.transform` argument.

```{r   }
estimate(e, function(p) atanh(p['y1~~y2']/(p['y1~~y1']*p['y2~~y2'])^.5), back.transform=tanh)
```

The transformed confidence interval will generally have improved
coverage especially near the boundary $\rho \approx \pm 1$.

While the estimates of this particular model can be obtained in closed
form, this is generally not the case when for example considering
parameter constraints, latent variables, or missing and censored
observations. The MLE is therefore obtained using iterative
optimization procedures (typically Fisher scoring or Newton-Raphson
methods). To ensure that the estimated variance parameters leads to a
meaningful positive definite structure and to avoid potential problems
with convergence it can often be a good idea to parametrize the model
in a way that such parameter constraints are naturally fulfilled.
This can be achieved with the `constrain` method.

```{r  constraints }
m2 <- m |>
    parameter(~ l1 + l2 + z) |>
    variance(~ y1 + y2, value=c('v1','v2')) |>
    covariance(y1 ~ y2, value='c') |>
    constrain(v1 ~ l1, fun=exp) |>
    constrain(v2 ~ l2, fun=exp) |>
    constrain(c ~ z+l1+l2, fun=function(x) tanh(x[1])*sqrt(exp(x[2])*exp(x[3])))
```

In the above code,  we first add new parameters `l1` and `l2` to hold the log-variance
parameters, and `z` which will be the z-transform of the correlation
parameter.
Next we label the variances and covariances: The variance of `y1` is called `v1`;
the variance of `y2` is called `v2`; the covariance of `y1` and `y2` is called `c`.
Finally, these parameters are tied to the previously defined
parameters using the `constrain` method such that `v1` := $\exp(\mathtt{l1})$
`v2` := $\exp(\mathtt{l1})$ and `z` := $\tanh(\mathtt{z})\sqrt{\mathtt{v1}\mathtt{v2}}$.
In this way there is no constraints on the actual estimated parameters
`l1`, `l2`, and `z` which can take any values in $\mathbb{R}^{3}$, while we at the
same time are guaranteed a proper covariance matrix which is positive
definite.

```{r  estconstraints }
e2 <- estimate(m2, d)
e2
```

The correlation coefficient can then be obtained as

```{r  deltaconstraints }
estimate(e2, 'z', back.transform=tanh)
```

In practice, a much shorter syntax can be used to obtain the above
parametrization. We can simply use the argument `constrain`
when specifying the covariances (the argument `rname` specifies the
parameter name of the $\arctanh$ transformed correlation
coefficient, and `lname`, `lname2` can be used to specify the parameter
names for the log variance parameters):

```{r  constraints2 }
m2 <- lvm() |>
  regression(y1 + y2 ~ x) |>
  covariance(y1 ~ y2, constrain=TRUE, rname='z')

e2 <- estimate(m2, data=d)
e2
```

```{r  e2backtransform }
estimate(e2, 'z', back.transform=tanh)
```

As an alternative to the Wald confidence intervals (with or without
transformation) is to profile the likelihood. The profile likelihood
confidence intervals can be obtained with the `confint` method:

```{r  profileci, cache=TRUE }
tanh(confint(e2, 'z', profile=TRUE))
```

Finally, a non-parametric bootstrap (in practice a larger number of
replications would be needed) can be calculated in the following way

```{r  bootstrap, cache=TRUE }
set.seed(1)
b <- bootstrap(e2, data=d, R=50, mc.cores=1)
b
```

```{r  cache=TRUE }
quantile(tanh(b$coef[,'z']), c(.025,.975))
```


## Censored observations

Letting one of the variables be right-censored (Tobit-type model) we
can proceed in exactly the same way (note, this functionality is only
available with the `mets` package installed - available from CRAN). The
only difference is that the variables that are censored must all be
defined as `Surv` objects (from the `survival` package which is
automatically loaded when using the `mets` package) in the data frame.

```{r  cache=TRUE, eval=mets }
m3 <- lvm() |>
  regression(y1 + s2 ~ x) |>
  covariance(y1 ~ s2, constrain=TRUE, rname='z')

e3 <- estimate(m3, d)
```

```{r  eval=mets }
e3
```

```{r  cache=TRUE, eval=mets }
estimate(e3, 'z', back.transform=tanh)
```

And here the same analysis with `s1` being left-censored and `s2` right-censored:

```{r  cache=TRUE, eval=mets }
m3b <- lvm() |>
  regression(s1 + s2 ~ x) |>
  covariance(s1 ~ s2, constrain=TRUE, rname='z')

e3b <- estimate(m3b, d)
e3b
```

```{r  eval=mets }
e3b
```

```{r  cache=TRUE, eval=mets }
estimate(e3b, 'z', back.transform=tanh)
```

```{r  profilecens, cache=TRUE, eval=mets }
tanh(confint(e3b, 'z', profile=TRUE))
```


# SessionInfo

```{r   }
sessionInfo()
```
# Bibliography

---
title: Non-linear latent variable models and error-in-variable models
author: Klaus Kähler Holst
date: "`r Sys.Date()`"
output:
  rmarkdown::html_vignette:
    fig_caption: yes
    fig_width: 6 
    fig_height: 4 
vignette: >
  %\VignetteIndexEntry{Non-linear latent variable models and error-in-variable models}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
bibliography: ref.bib
---

```{r  include=FALSE }
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  dev = "svg",
  fig.ext = "svg"
)
## library("svglite")
mets <- lava:::versioncheck('mets', 1)
fullVignette <- Sys.getenv("_R_FULL_VIGNETTE_") %in% c("1","TRUE")
```

```{r load, results="hide",message=FALSE,warning=FALSE }
library("lava")
```

We consider the measurement models given by

$$X_{j} = u_{1} + \epsilon_{j}^{x}, \quad j=1,2,3$$
$$Y_{j} = u_{2} + \epsilon_{j}^{y}, \quad j=1,2,3$$
and with a structural model given by
$$u_{2} = f(u_{1}) + Z + \zeta_{2}$$
$$u_{1} = Z + \zeta_{1}$$
with iid measurement errors
$\epsilon_{j}^{x},\epsilon_{j}^{y},\zeta_{1},\zeta_{2}\sim\mathcal{N}(0,1),
j=1,2,3.$ and standard normal distributed covariate $Z$.  To
simulate from this model we use the following syntax:

```{r  sim }
f <- function(x) cos(1.25*x) + x - 0.25*x^2
m <- lvm(x1+x2+x3 ~ u1, y1+y2+y3 ~ u2, latent=~u1+u2)
regression(m) <- u1+u2 ~ z
functional(m, u2~u1) <- f

d <- sim(m, n=200, seed=42) # Default is all parameters are 1
```

```{r   }
## plot(m, plot.engine="visNetwork")
plot(m)
```

We refer to [@holst_budtzjorgensen_2013] for details on the syntax for
model specification.


# Estimation

To estimate the parameters using the two-stage estimator described in [@holst_budtzjorgensen_2020], the first step is now to specify the measurement models

```{r  specifymodels }
m1 <- lvm(x1+x2+x3 ~ u1, u1 ~ z, latent=~u1)
m2 <- lvm(y1+y2+y3 ~ u2, u2 ~ z, latent=~u2)
```

Next, we specify a quadratic relationship between the two latent variables

```{r   }
nonlinear(m2, type="quadratic") <- u2 ~ u1
```

and the model can then be estimated using the two-stage estimator

```{r  twostage1 }
e1 <- twostage(m1, m2, data=d)
e1
```

We see a clear statistically significant effect of the second order
term (`u2~u1_2`). For comparison we can also estimate the full MLE
of the linear model:

```{r  linear_mle }
e0 <- estimate(regression(m1%++%m2, u2~u1), d)
estimate(e0,keep="^u2~[a-z]",regex=TRUE) ## Extract coef. matching reg.ex.
```

Next, we calculate predictions from the quadratic model using the estimated parameter coefficients
$$
\mathbb{E}_{\widehat{\theta}_{2}}(u_{2} \mid u_{1}, Z=0),
$$

```{r  pred1 }
newd <- expand.grid(u1=seq(-4, 4, by=0.1), z=0)
pred1 <- predict(e1, newdata=newd, x=TRUE)
head(pred1)
```

To obtain a potential better fit we next proceed with a natural cubic spline

```{r  spline_twostage }
kn <- seq(-3,3,length.out=5)
nonlinear(m2, type="spline", knots=kn) <- u2 ~ u1
e2 <- twostage(m1, m2, data=d)
e2
```

Confidence limits can be obtained via the Delta method using the `estimate` method:

```{r  spline_ci }
p <- cbind(u1=newd$u1,
	  estimate(e2,f=function(p) predict(e2,p=p,newdata=newd))$coefmat)
head(p)
```

The fitted function can be obtained with the following code:

```{r  figpred2 }
plot(I(u2-z) ~ u1, data=d, col=Col("black",0.5), pch=16,
     xlab=expression(u[1]), ylab=expression(u[2]), xlim=c(-4,4))
lines(Estimate ~ u1, data=as.data.frame(p), col="darkblue", lwd=5)
confband(p[,1], lower=p[,4], upper=p[,5], polygon=TRUE,
	 border=NA, col=Col("darkblue",0.2))
```


# Cross-validation

A more formal comparison of the different models can be obtained by
cross-validation. Here we specify linear, quadratic and cubic spline
models with 4 and 9 degrees of freedom.

```{r  spline_several }
m2a <- nonlinear(m2, type="linear", u2~u1)
m2b <- nonlinear(m2, type="quadratic", u2~u1)
kn1 <- seq(-3,3,length.out=5)
kn2 <- seq(-3,3,length.out=8)
m2c <- nonlinear(m2, type="spline", knots=kn1, u2~u1)
m2d <- nonlinear(m2, type="spline", knots=kn2, u2~u1)
```

To assess the model fit average RMSE is estimated with 5-fold
cross-validation repeated two times

```{r  cv_fit, cache=TRUE, eval=fullVignette }
## Scale models in stage 2 to allow for a fair RMSE comparison
d0 <- d
for (i in endogenous(m2))
    d0[,i] <- scale(d0[,i],center=TRUE,scale=TRUE)
## Repeated 5-fold cross-validation:
ff <- lapply(list(linear=m2a,quadratic=m2b,spline4=m2c,spline6=m2d),
	    function(m) function(data,...) twostage(m1,m,data=data,stderr=FALSE,control=list(start=coef(e0),contrain=TRUE)))
fit.cv <- lava:::cv(ff,data=d,K=5,rep=2,mc.cores=parallel::detectCores(),seed=1)
```

```{r  results="hide", echo=FALSE }
## To save time building the vignettes on CRAN, we cache time consuming computations
if (fullVignette) {
  fit.cv$fit <- NULL
  saveRDS(fit.cv, "data/nonlinear_fitcv.rds")
} else {
  fit.cv <- readRDS("data/nonlinear_fitcv.rds")
}
```

```{r   }
fit.cv$coef
```

Here the RMSE is in favour of the splines model with 4 degrees of freedom:

```{r  multifit }
fit <- lapply(list(m2a,m2b,m2c,m2d),
	     function(x) {
		 e <- twostage(m1,x,data=d)
		 pr <- cbind(u1=newd$u1,predict(e,newdata=newd$u1,x=TRUE))
		 return(list(estimate=e,predict=as.data.frame(pr)))
	     })

plot(I(u2-z) ~ u1, data=d, col=Col("black",0.5), pch=16,
     xlab=expression(eta[1]), ylab=expression(eta[2]), xlim=c(-4,4))
col <- c("orange","darkred","darkgreen","darkblue")
lty <- c(3,4,1,5)
for (i in seq_along(fit)) {
    with(fit[[i]]$pr, lines(u2 ~ u1, col=col[i], lwd=4, lty=lty[i]))
}
legend("bottomright",
      c("linear","quadratic","spline(df=4)","spline(df=6)"),
      col=col, lty=lty, lwd=3)
```

For convenience, the function `twostageCV` can be used to do the
cross-validation (also for choosing the mixture distribution via the `nmix`
argument, see the section below). For example,

```{r  twostageCV, cache=TRUE, eval=fullVignette }
set.seed(1)
selmod <- twostageCV(m1, m2, data=d, df=2:4, nmix=1:2,
	    nfolds=5, rep=2, mc.cores=parallel::detectCores())
```

```{r  results="hide", echo=FALSE }
## To save time building the vignettes on CRAN, we cache time consuming computations
if (fullVignette) {
  saveRDS(summary(selmod), "data/nonlinear_selmod.rds")
} else {
  selmod <- readRDS("data/nonlinear_selmod.rds")
}
```

applies cross-validation (here just 2 folds for simplicity) to select the best splines with
degrees of freedom varying from 1-3 (the linear model is
automatically included)

```{r   }
selmod
```


# Specification of general functional forms

Next, we show how to specify a general functional relation of
multiple different latent or exogenous variables. This is achieved via
the `predict.fun` argument. To illustrate this we include interactions
between the latent variable $u_{1}$ and a dichotomized version of
the covariate $z$

```{r   }
d$g <- (d$z<0)*1 ## Group variable
mm1 <- regression(m1, ~g)  # Add grouping variable as exogenous variable (effect specified via 'predict.fun')
mm2 <- regression(m2, u2~ u1+u2+u1:g+u2:g+z)
pred <- function(mu,var,data,...) {
    cbind("u1"=mu[,1],"u2"=mu[,1]^2+var[1],
	  "u1:g"=mu[,1]*data[,"g"],"u2:g"=(mu[,1]^2+var[1])*data[,"g"])
}
ee1 <- twostage(mm1, model2=mm2, data=d, predict.fun=pred)
estimate(ee1,keep="u2~u",regex=TRUE)
```

A formal test show no statistically significant effect of this interaction

```{r   }
summary(estimate(ee1,keep="(:g)", regex=TRUE))
```


# Mixture models

Lastly, we demonstrate how the distributional assumptions of stage 1
model can be relaxed by letting the conditional distribution of the
latent variable given covariates follow a Gaussian mixture
distribution. The following code explictly defines the parameter
constraints of the model by setting the intercept of the first
indicator variable, $x_{1}$, to zero and the factor loading
parameter of the same variable to one.

```{r   }
m1 <- baptize(m1)  ## Label all parameters
intercept(m1, ~x1+u1) <- list(0,NA) ## Set intercept of x1 to zero. Remove the label of u1
regression(m1,x1~u1) <- 1 ## Factor loading fixed to 1
```

The mixture model may then be estimated using the `mixture` method
(note, this requires the `mets` package to be installed), where the
Parameter names shared across the different mixture components given
in the `list` will be constrained to be identical in the mixture
model. Thus, only the intercept of $u_{1}$ is allowed to vary
between the mixtures.

```{r  mixture1, cache=TRUE, eval=fullVignette }
set.seed(1)
em0 <- mixture(m1, k=2, data=d)
```

To decrease the risk of using a local maximizer of the likelihood we
can rerun the estimation with different random starting values

```{r  estmixture, cache=TRUE,warnings=FALSE,messages=FALSE,eval=FALSE }
em0 <- NULL
ll <- c()
for (i in 1:5) {
    set.seed(i)
    em <- mixture(m1, k=2, data=d, control=list(trace=0))
    ll <- c(ll,logLik(em))
    if (is.null(em0) || logLik(em0)<tail(ll,1))
	em0 <- em
}
```

```{r  results="hide", echo=FALSE }
## To save time building the vignettes on CRAN, we cache time consuming computations
if (fullVignette) {
  saveRDS(em0, "data/nonlinear_em0.rds")
} else {
  em0 <- readRDS("data/nonlinear_em0.rds")
}
```

```{r   }
summary(em0)
```

Measured by AIC there is a slight improvement in the model fit using the mixture model

```{r  eval=mets }
e0 <- estimate(m1,data=d)
AIC(e0,em0)
```

The spline model may then be estimated as before with the `two-stage` method

```{r  eval=mets }
em2 <- twostage(em0,m2,data=d)
em2
```

In this example the results are very similar to the Gaussian model:

```{r  mixturefit, eval=mets }
plot(I(u2-z) ~ u1, data=d, col=Col("black",0.5), pch=16,
     xlab=expression(eta[1]), ylab=expression(eta[2]))

lines(Estimate ~ u1, data=as.data.frame(p), col="darkblue", lwd=5)
confband(p[,1], lower=p[,4], upper=p[,5], polygon=TRUE,
	 border=NA, col=Col("darkblue",0.2))

pm <- cbind(u1=newd$u1,
	    estimate(em2, f=function(p) predict(e2,p=p,newdata=newd))$coefmat)
lines(Estimate ~ u1, data=as.data.frame(pm), col="darkred", lwd=5)
confband(pm[,1], lower=pm[,4], upper=pm[,5], polygon=TRUE,
	 border=NA, col=Col("darkred",0.2))
legend("bottomright", c("Gaussian","Mixture"),
       col=c("darkblue","darkred"), lwd=2, bty="n")
```

# SessionInfo

```{r   }
sessionInfo()
```


# Bibliography

# First steps

To contribute to LAVA, start by reviewing the following guidelines:

- [Code of Conduct](./code-of-conduct.md)
- [Contribution guide](./methodology/contribute.md)
- [Contribution process](./tutorials/contributing.md)

Once you're familiar with the foundational basics, explore the development
guides to [setup](./development/setup.md) your environment,
[build](./development/build.md) and [test](./development/test.md) your changes.

# Getting Started

New to LAVA? You're in the right place. Follow the steps below to get up and
running in minutes.

## Step 1. Install a LAVA instance

A full LAVA instance consists of a LAVA server and one or more LAVA workers.
The server handles job submission, scheduling, live log and result collection.
Workers connect to the server and execute test jobs on physical or virtual
devices.

If you don't have access to an existing LAVA instance, you can set one up
locally using Docker in a few minutes:

[Install LAVA with Docker](../admin/basic-tutorials/instance/install.md#docker)

A worker and two virtual devices are pre-configured for your convenience. Once
the instance is running, verify the following:

- You can [login](http://localhost/accounts/login/)
- Worker [worker0](http://localhost/scheduler/allworkers) is online
- [Devices](http://localhost/scheduler/alldevices/active) `qemu0` and `docker0`
  appear.

Once verified, you're ready to submit your first job.

## Step 2. Submit a LAVA job

A LAVA job is defined in YAML and describes what images to deploy, how to boot
the DUT, and which tests to run, etc.

For your first job, try one of these:

- [Docker job](../admin/basic-tutorials/device-setup/docker.md#submit-a-job) -
  the job creates a Docker container and attaches to the LAVA docker test shell
  and runs inline test definitions that can be easily customized.

- [QEMU job](../user/basic-tutorials/submit.md) — the job boots an image and runs
  test definitions fetched from a Git repository.

For simplicity, submit via the [web interface](../user/basic-tutorials/submit.md#web-interface).

## Step 3. Inspect job output

After your job completes, follow the guide below to review the results and logs
on the job page:

[Understanding job output](../user/basic-tutorials/job-output.md)

## What's next

- [Learn job definition schema](../user/basic-tutorials/job-definition.md)
- [Learn test definition format](../user/basic-tutorials/test-definition.md)

--8<-- "refs.txt"

# TFTP

The `tftp` deploy action is used to deploy files via
[TFTP](../../../../introduction/glossary.md#tftp) for network booting. This
is commonly used with bootloaders like U-Boot that can fetch kernel, device
tree, ramdisk, and other files over the network.

Files are downloaded to a temporary directory in the TFTP tree and the
file names are substituted into the bootloader commands specified in the device
configuration or overridden in the job definition.

```yaml
actions:
- deploy:
    to: tftp
    kernel:
      url: http://example.com/vmlinuz
      type: zimage
    dtb:
      url: http://example.com/device.dtb
    ramdisk:
      url: http://example.com/initrd.gz
      compression: gz
```

## url

See [url](./index.md#url)

## compression

See [compression](./index.md#compression)

## kernel

The kernel is a required artifact for TFTP deployments.

### type

Specifies the type of kernel being downloaded.

Supported types:

- `image` - used with `booti` command
- `zimage` - used with `bootz` command
- `uimage` - used with `bootm` command

Boot methods like `u-boot` use this information to determine the appropriate
load addresses and boot commands.

## dtb

See [DTB](../../../../introduction/glossary.md#dtb)

## dtbo

Device Tree Blob for Overlay (DTBO) - additional device tree files that can be
applied on top of the base DTB. Multiple overlays can be specified as a list.

See also: [U-Boot Device Tree Overlays](https://docs.u-boot.org/en/latest/usage/fdt_overlays.html)

```yaml
dtbo:
- url: https://example.com/overlay1.dtbo
- url: https://example.com/overlay2.dtbo
```

### Enabling DTBO

By default, dtb overlay is not enabled for uboot-based devices, you need to
enable it in your device type dictionary:

```jinja
{% set enable_dtbo_support = 'true' %}
```

Additionally, configure custom DTBO load addresses and `dtb_base_resize` if
your device configuration differs from the defaults:

```jinja
{% set bootm_dtbo_addr = '0x93c00000' %}
{% set bootz_dtbo_addr = bootm_dtbo_addr %}
{% set dtb_base_resize = dtb_base_resize | default(1048576) %}
```

The same DTBO load address can be used for every overlay, as they are loaded
and applied by LAVA sequentially in the following order.

1. Load the base DTB
2. Resize it to accommodate overlays
3. Load each overlay to the same address
4. Apply each overlay in sequence using `fdt apply`

```shell title="Example U-Boot commands:"
- fdt addr ${fdt_addr}
- fdt resize 1048576
- tftp 0x93c00000 577262/tftp-deploy-o40xvdqb/dtbo0/overlay1.dtbo
- fdt apply 0x93c00000
- tftp 0x93c00000 577262/tftp-deploy-o40xvdqb/dtbo1/overlay2.dtbo
- fdt apply 0x93c00000
```

## modules

```yaml
modules:
  url: http://example.com/modules.tar.gz
  compression: gz
```

A tarball of kernel modules for the supplied kernel. The file **must** be a tar
file and the compression method **must** be specified.

Modules are required when:

- The kernel needs modules to locate the rootfs (e.g., NFS or specific
  filesystem drivers)
- Tests within the ramdisk require certain kernel modules

When modules are provided, LAVA can unpack the ramdisk and add the modules
before repacking.

## ramdisk

An initial ramdisk (initrd/initramfs) that can be loaded alongside the kernel.
The ramdisk needs to be unpacked and modified in either of the following two
use cases:

- The LAVA test shell is expected to run inside the ramdisk
- Kernel modules need to be added (e.g., to load network drivers for NFS boot)

If the ramdisk is compressed, the compression method **must** be specified.

### header

If a header is already applied to the ramdisk, the `header` value **must**
specify the type of header. Currently only `u-boot` is supported.

```yaml
ramdisk:
  url: http://example.com/initrd.img.gz
  compression: gz
  header: u-boot
```

This header will be removed before unpacking so that LAVA can overlay files.
LAVA will add the header back if the device requires it. This is controlled by
the `add_header` parameter in the device type template.

### install_modules

The default is `true`. Set to `false` to skip installing kernel modules into the
ramdisk.

```yaml
ramdisk:
  url: http://example.com/initrd.gz
  compression: gz
  install_modules: false
```

### install_overlay

The default is `true`. The overlay is only applied when a test action is present
in the job. Set to `false` to skip installing the LAVA overlay into the ramdisk.

```yaml
ramdisk:
  url: http://example.com/initrd.gz
  compression: gz
  install_overlay: false
```

## nfsrootfs

A tarball containing the root filesystem to be exported via NFS. The device
boots with this filesystem mounted as the root.

```yaml
nfsrootfs:
  url: http://example.com/rootfs.tar.xz
  compression: xz
  format: tar
  overlays:
    kselftest:
      url: http://example.com/kselftest.tar.xz
      compression: xz
      format: tar
      path: /opt/kselftest
```

!!! note "Context options"
    Additional options can be specified via the job definition context:

    - **`extra_nfsroot_args`**: Append NFS mount options (e.g., protocol version)
    - **`extra_kernel_args`**: Append kernel command line arguments

    ```yaml
    context:
      extra_nfsroot_args: ",vers=3"
      extra_kernel_args: "loglevel=7 earlycon"
    ```

### prefix

Optional prefix path within the tarball to use as the root filesystem.

### install_modules

See [install_modules](#install_modules)

### install_overlay

See [install_overlay](#install_overlay)

### overlays

See [overlays](./index.md#overlays)

## persistent_nfs

A persistent NFS URL can be used instead of a compressed tarball. This allows
multiple jobs to share the same NFS export.

!!! warning "Known caveats"
    - Modules are not extracted into the persistent NFS mount
    - LAVA does not shut down the device or attempt to unmount the NFS
      filesystem when the job finishes. The device is simply powered off
    - The test writer needs to ensure that any background processes started
      by the test have been stopped before the test finishes

!!! note
    Only one of `nfsrootfs` or `persistent_nfs` can be specified in a single
    deploy action.

### address

Specifies the address to use for the persistent filesystem.

The `address` **must** include the IP address of the NFS server and the full
path to the directory which contains the root filesystem, separated by a single
colon. In YAML, all values containing a colon **must** be quoted:

```yaml
- deploy:
    to: tftp
    kernel:
      url: http://example.com/vmlinuz
      type: zimage
    persistent_nfs:
      address: "192.168.1.1:/var/lib/lava/dispatcher/tmp/nfsroot"
```

The address supports placeholders (see [URL placeholders](./index.md#url)):

### install_overlay

See [install_overlay](#install_overlay)

## tee

Trusted Execution Environment (TEE) image for devices that support secure
boot with TEE.

When booting with `u-boot`, LAVA can load tee over TFTP. User should provide a
resource called `tee` in the deploy action.

```yaml
- deploy:
    to: tftp
    tee:
      url: file:/local/lava-ref-binaries/fsl-imx6q-sabresd-linux/uTee-6qsdb
```

## preseed

Configuration file for automated OS installations.

When `os` is set to `debian_installer` or `centos_installer`, LAVA
automatically modifies the file to inject the LAVA overlay into the installed
system.

```yaml
- deploy:
    to: tftp
    os: debian_installer
    kernel:
      url: http://example.com/debian-installer/linux
    ramdisk:
      url: http://example.com/debian-installer/initrd.gz
      compression: gz
    preseed:
      url: http://example.com/preseed.cfg
```

Device boot commands can reference:

- `{PRESEED_CONFIG}` — TFTP path to the preseed file (e.g., `url=tftp://{SERVER_IP}/{PRESEED_CONFIG}`)
- `{PRESEED_LOCAL}` — file name embedded in the ramdisk root (e.g, `preseed.cfg`)


## Example jobs

* [Booting from ramdisk](../../../../admin/basic-tutorials/device-setup/u-boot.md#booting-from-ramdisk)
* [Booting from NFS](../../../../admin/basic-tutorials/device-setup/u-boot.md#booting-from-nfs)

# Backup LAVA

In order to backup LAVA you will have to backup both the database and specific
directories.

## Database

The database is managed by
[PostgreSQL](../../../technical-references/services/postgresql.md). See the
[PostgreSQL documentation][pgsql-backup] for
details.

To backup the database run:

```shell
sudo -u lavaserver pg_dump lavaserver > lava-server.sql
```

## Filesystem

The configuration files are stored in:

* `/etc/lava-coordinator/`
* `/etc/lava-dispatcher/`
* `/etc/lava-server/`

The data (job outputs, ...) are stored in `/var/lib/lava-server/default/media`.

To backup the filesystem run:

```shell
tar czf lava-server-etc.tar.gz /etc/lava-coordinator/ \
                               /etc/lava-dispatcher/ \
                               /etc/lava-server/
tar czf lava-server-data.tar.gz /var/lib/lava-server/default/media/
```

--8<-- "refs.txt"

# Device dictionary

The device dictionary is a [jinja2][jinja2] template that should extend one
[device-type template](./device-type-template.md).

While device-type templates are usually provided and maintained by LAVA, device
dictionaries are provided and maintained by the admin.

A device dictionary will contain device specific information like:

* connection command
* power on/off commands
* bootloader prompt
* character delays
* ...

See [Parameters](#parameters)

## Configuration file

Device dictionaries are stored on the server in
`/etc/lava-server/dispatcher-config/devices/<hostname>.jinja2`.

Admin could update device dictionaries using lavacli:

```shell
lavacli devices dict set qemu01 qemu01.jinja2
```

## Parameters

The variables you may define in a device dictionary depend on the
[device-type template](./device-type-template.md) it extends. Common
parameters include connection and power commands. See the device type template
and its base template for the full set of supported parameters.

### Commands

All the command parameters in this section accept either a single string or a
list of strings. When a list is used, LAVA executes each command in order.

#### power_on_command

Command to supply power to the device remotely. The device **must** start the
boot sequence when the `power_on_command` is called. If the DUT needs a button
to be pressed to boot, then the command to press the button should be part of
this one.

```jinja
{% set power_on_command =  [
    'laacli power 5v on',
    'laacli power 3v3 on'
] %}
```

#### power_off_command

Command to remove power from the device remotely.

```jinja
{% set power_off_command = [
    'laacli power 3v3 off',
    'laacli power 5v off'
] %}
```

#### soft_reboot_command

Command to run at a shell prompt on a running system to request a reboot (e.g.
`reboot` or `systemctl reboot`).

```jinja
{% set soft_reboot_command = 'reboot' %}
```

#### hard_reset_command

Command to cut power and then restore it remotely (e.g. via PDU or hub). You can
insert a delay using `sleep` between cutting and restoring power if the device
requires it.

```jinja
{% set hard_reset_command = ['usbrelay 1_2=0', 'sleep 1', 'usbrelay 1_2=1'] %}
```

#### pre_power_command

Ancillary command for special cases. Behavior and support depend on the
deployment method and device-type template.

```jinja
{% set pre_power_command = 'laacli usb 1 on' %}
```

#### pre_os_command

Ancillary command for special cases. Behavior and support depend on the
deployment method and device-type template.

```jinja
{% set pre_os_command = 'laacli usb 1 off' %}
```

### Connections

#### connection_command

Command to access the serial port of the device.

```jinja
{% set connection_command = 'telnet dispatcher03 7004' %}
```

#### connection_list

The list of hardware ports which are configured for serial connections to the device.

```jinja
{% set connection_list = ['uart0'] %}
```

#### connection_commands

A dictionary of the commands to start each connection.

```jinja
{% set connection_commands = {'uart0': 'telnet w1 7005'} %}
```

#### connection_tags

Each connection can include `tags` - extra pieces of metadata to describe the connection.

```jinja
{% set connection_tags = {'uart0': ['primary', 'telnet']} %}
```

There must always be one (and only one) connection with the `primary` tag, denoting
the connection that will be used for firmware, bootloader and kernel interaction.

Other tags may describe the *type* of connection, as extra information that LAVA
can use to determine how to close the connection cleanly when a job finishes
(e.g `telnet` and `ssh`).

### Device

Some devices or the elements of the device configuration can be exposed to the
LAVA test shell, where it is safe to do so. Each parameter must be explicitly
set in each device dictionary.

For ease of use, LAVA will directly export the content of the `device_info`,
`environment`, `static_info` and `storage_info` dictionaries into the test
shell environment. The dictionaries and lists will be unrolled, for example:

```jinja
{% set environment = {"RELAY_ADDRESS": "10.66.16.103", "REMOTE_SERIAL_PORT": "/dev/ttyUSB2"} %}
{% set device_info = [{'board_id': 'a2c22e48'}] %}
{% set static_info = [{"board_id": "S_NO81730000"}, {"board_id": "S_NO81730001"}] %}
{% set storage_info = [{'SATA': '/dev/disk/by-id/ata-ST500DM002-1BD142_W3T79GCW'}] %}
```

will become:

```shell
export RELAY_ADDRESS='10.66.16.103'
export REMOTE_SERIAL_PORT='/dev/ttyUSB2'
export LAVA_DEVICE_INFO_0_board_id='a2c22e48'
export LAVA_STATIC_INFO_0_board_id='S_NO81730000'
export LAVA_STATIC_INFO_1_board_id='S_NO81730001'
export LAVA_STORAGE_INFO_0_SATA='/dev/disk/by-id/ata-ST500DM002-1BD142_W3T79GCW'
```

The environment can be overridden in the job definition, see
[job environment](../job-definition/environment.md)

#### device_info

`device_info` is a list of dictionaries, where each dictionary value can contain
keys such as `board_id`, `usb_vendor_id`, `usb_product_id`, `wait_device_board_id`,
which can be made available to the Docker container for device specific tasks
dynamically, whenever the device is reset, using a `udev` rule.

```jinja
{% set device_info = [{'board_id': '8c5f2290'}] %}
```

#### environment

A dictionary containing device-specific shell variables, which will be available
in the LAVA test shell. These can be used, for example, to describe physical
hardware connections between the DUT and interfaces on the worker or other
addressable hardware.

```jinja
{% set environment = {
    'RELAY_ADDRESS': '10.66.16.103',
    'REMOTE_SERIAL_PORT': '/dev/ttyUSB2',
} %}
```

#### static_info

A list of dictionaries describing static information. A common use case is
defining a file server address for URL substitution. For example:

```jinja
{% set static_info = [{'FILE_SERVER_IP': "10.192.244.104"}] %}
```

See also [URL placeholders](../job-definition/actions/deploy/index.md#url).

#### storage_info

A list of dictionaries, where each dictionary value can contain keys describing
the storage name (e.g. USB or SATA) and a value stating the device node of the
top level block device which is available to the test writer.

```jinja
{% set storage_info = [{'SATA': '/dev/disk/by-id/ata-ST500DM002-1BD142_W3T79GCW'}] %}
```

#### device_ip

A single fixed IPv4 address of this device. The value will be exported into the
test shell using
[`lava-target-ip`](../../user/basic-tutorials/test-definition.md#lava-target-ip).

```jinja
{% set device_ip = "10.66.16.24" %}
```

#### device_mac

Similar to `device_ip` but for a single MAC address. The value will be exported
into the test shell using
[`lava-target-mac`](../../user/basic-tutorials/test-definition.md#lava-target-mac).

```jinja
{% set device_mac = '00:02:F7:00:58:53' %}
```

### Character delays

LAVA supports specifying character delays in the deploy, boot, and test
actions to help with serial reliability (e.g. when the DUT or connection
drops or corrupts characters when input is sent too quickly). These are
device-specific, so they are best set in the device dictionary (or
device-type template).

The `deploy` and `boot` actions are affected more often than the `test`
action, because they interacts with firmware or bootloader processes where
input handling can be more limited than in a POSIX test environment.

#### deploy_character_delay

Set the number of milliseconds to add between each character of every
string sent to the DUT during the `deploy` action. This is useful for
deployment methods that use a connection (e.g. `vemsd`):

```jinja
{% set deploy_character_delay = 30 %}
```

#### boot_character_delay

Set the number of milliseconds to add between each character of every
string sent to the DUT during the `boot` action:

```jinja
{% set boot_character_delay = 20 %}
```

Some devices need more (e.g. 100 or 500 ms). For long delays, also consider
the overall boot timeout and set a minimum for the relevant boot action in
the device-type template.

#### test_character_delay

Set the number of milliseconds to add between each character of every
string sent to the DUT during the `test` action:

```jinja
{% set test_character_delay = 10 %}
```

!!! note
    LAVA also waits `test_character_delay` milliseconds before sending each
    test signal. The delay can be helpful on slow serial connections to avoid
    character interleaving with other inputs, such as kernel dmesg.

### Depthcharge

The following variables can be used in a Depthcharge device dictionary to
configure the [`depthcharge`](../job-definition/actions/boot/method-depthcharge.md)
boot method.

#### start_message

The string that Depthcharge prints when it is ready to accept netboot
commands. LAVA waits for this message after resetting the device.

```jinja
{% set start_message = start_message | default('Starting depthcharge on Volmar...') %}
```

Defaults to `Starting netboot`.

#### bootloader_prompt

The Depthcharge command line interface prompt string. LAVA uses this to detect
when the bootloader is ready for sending the next command.

```jinja
{% set bootloader_prompt = bootloader_prompt | default('volmar:') %}
```

Defaults to `dpch:`.

#### depthcharge_final_message

The string that depthcharge prints when it has finished. LAVA waits for this
message before continuing.

```jinja
{% set depthcharge_final_message = depthcharge_final_message | default('jumping to kernel') %}
```

Defaults to `Starting kernel`.

See also [bootloader-final-message](../job-definition/actions/boot/common.md#bootloader-final-message).

### DFU

`board_id` must be provided for adding a new DFU device.

```jinja
{% set board_id = "LCES2-0xf775916400000000" %}
```

### Fastboot

The following variables can be used in fastboot device dictionary to configure
fastboot device.

#### adb_serial_number

A string to specify the serial number of the device in ADB mode.

```jinja
{% set adb_serial_number = '25564f71' %}
```

#### fastboot_serial_number

A string to specify the serial number of the device in fastboot mode.

```jinja
{% set fastboot_serial_number = '25564f71' %}
```

#### fastboot_auto_detection

A boolean. When `true`, LAVA tries to detect the device serial number
automatically.

```jinja
{% set fastboot_auto_detection = true %}
```

!!! note
    `fastboot_auto_detection` assumes the ADB serial number and fastboot serial
    number are the same for the same device.

When `false` or unset, you must set `adb_serial_number` and
`fastboot_serial_number` explicitly.

#### fastboot_options

A list of strings, used for specifying additional options to the `fastboot` command.

```jinja
{% set fastboot_options = ['-S', '256M'] %}
```

#### flash_cmds_order

A list of strings, used for specifying the order in which the images should be
flashed to the DUT using the `fastboot` command.

```jinja
{% set flash_cmds_order = ['update', 'ptable', 'partition', 'hyp', 'modem',
                           'rpm', 'sbl1', 'sbl2', 'sec', 'tz', 'aboot',
                           'boot', 'rootfs', 'vendor', 'system', 'cache',
                           'userdata', ] %}
```

#### fastboot_sequence

A list of strings. In practice, you only need to set the list to use one of the
following method for most of the device types, then LAVA will populate and
execute the corresponding action.

##### boot

Runs `fastboot boot <boot-image>`, loading the `boot` image from the preceding
deploy action directly into RAM and executing it.

```jinja
{% set fastboot_sequence = ['boot'] %}
```

##### reboot

Runs `fastboot reboot`, rebooting the device into the OS from its previously
flashed partitions. This is the typical path when all images (including `boot`)
were flashed during the deploy action.

```jinja
{% set fastboot_sequence = ['reboot'] %}
```

##### no-flash-boot

Same behaviour as `boot` at boot time — the image is loaded into RAM via
`fastboot boot`. Additionally, during the deploy phase, LAVA skips flashing the
`boot` partition even if a `boot` image was listed under `images` in the deploy
action.

```jinja
{% set fastboot_sequence = ['no-flash-boot'] %}
```

### Flasher

The following `flasher_deploy_commands` variable must be configured for using
the [flasher](../job-definition/actions/deploy/to-flasher.md) deployment method.

#### flasher_deploy_commands

A list of commands that LAVA executes on the worker for flashing images onto the
DUT. For example, using a USB-SD-Mux to write an image to the DUT's SD card:

```jinja
{% set flasher_deploy_commands = [
        '/root/.local/bin/usbsdmux /dev/sg1 host',
        'sleep 3',
        '/root/.local/bin/usbsdmux /dev/sg1 info',
        'dd if={RECOVERY_IMAGE} of=/dev/disk/by-id/usb-LinuxAut_sdmux_HS-SD_MMC_000000001156-0:0 bs=4M oflag=sync conv=nocreat',
        'sleep 3',
        '/root/.local/bin/usbsdmux /dev/sg1 client',
    ]
%}
```

The following placeholders can be used in the commands. LAVA substitutes them
with the actual values at runtime.

| Variable | Value |
| --- | --- |
| `{UPPER_CASE_IMAGE_KEY}` | Downloaded image path (e.g., `{RECOVERY_IMAGE}` for `images.recovery_image`) |
| `{POWER_ON_COMMAND}` | [`power_on_command`](#power_on_command) |
| `{POWER_OFF_COMMAND}` | [`power_off_command`](#power_off_command) |
| `{SOFT_REBOOT_COMMAND}` | [soft_reboot_command](#soft_reboot_command) |
| `{HARD_RESET_COMMAND}` | [`hard_reset_command`](#hard_reset_command) |
| `{PRE_OS_COMMAND}` | [`pre_os_command`](#pre_os_command) |
| `{PRE_POWER_COMMAND}` | [`pre_power_command`](#pre_power_command) |
| `{DEVICE_INFO}` | YAML dump of [`device_info`](#device_info) |
| `{STATIC_INFO}` | YAML dump of [`static_info`](#static_info) |

### JLink

#### board_id

The serial number of the JLink probe or target board.

```jinja
{% set board_id = '000380000008' %}
```

### QDL

#### board_qdl_id

When a Qualcomm device goes into EDL (Emergency Download) mode it can be identified by its ID.
This ID must be set in the device dictionary as `board_qdl_id` like so:

```jinja
{% set board_qdl_id = 'D902AA38' %}
```

To find the EDL ID one should boot the device into EDL mode.
The device must be connected to a Linux host over USB.
To obtain the device's EDL ID, run the following command:

```bash
# lsusb -v -d 05c6:9008 | grep iProduct
  iProduct                2 QUSB_BULK_CID:040E_SN:BA9B2FEB
```

`board_qdl_id` should be set to the value that comes after `:`
In the example above it is `BA9B2FEB`

#### qdl_enter_commands

Normally to force a device to boot into EDL mode,
a physical button must be pressed or a DIP switch must be set on the device.
In automated setups, like LAVA, it's possible to force a device into EDL mode using the automation interface.
Some Qualcomm devices are shipped with an FTDI or UART based GPIO interface that allows remote control.
More details on that can be found in the [Qualcomm Test Automation Controller] (https://github.com/qualcomm/qcom-test-automation-controller/) documentation.

Note: `qdl_enter_commands` must be a list of strings.

```jinja
{% set qdl_enter_commands = ['/usr/local/bin/tac-api.py --serial DP05DD00 --command bootToEDL'] %}
```

### Secondary media

Use `<media>_label` and `<media>_uuid` to configure a secondary media in your
device dictionary. The `/dev/disk/by-id/<media>_uuid` must exist on the LAVA
worker.

#### USB

```jinja2
{% set usb_label = 'SanDiskCruzerBlade' %}
{% set usb_uuid = 'usb-SanDisk_Cruzer_Blade_20060266531DA442AD42-0:0' %}
```

#### SATA

```jinja2
{% set sata_label = 'ST160LM003' %}
{% set sata_uuid = "ata-ST160LM003_HN-M160MBB_S2SYJ9KC102184" %}
```

#### SD

```jinja2
{% set sd_label = 'sdcard' %}
{% set sd_uuid = 'mmc-SD16G_0xda85ac89' %}
```

### SSH

SSH device can be configured using the following parameters.

```jinja
{% extends 'ssh.jinja2' %}

{% set ssh_host = '192.168.1.100' %}
{% set ssh_port = 22 %}
{% set ssh_user = 'root' %}
{% set ssh_identity_file = '/<path>/private_key' %}
```

#### ssh_host

Required. The hostname or IP address of the device.

```jinja
{% set ssh_host = '192.168.1.100' %}
```

#### ssh_port

The SSH port on the device. Defaults to `22`.

```jinja
{% set ssh_port = 2222 %}
```

#### ssh_user

The login user for SSH and SCP connections. Defaults to `root`.

```jinja
{% set ssh_user = 'testuser' %}
```

#### ssh_identity_file

Path to the SSH private key on the LAVA worker. The matching public key must be
added to the `~/.ssh/authorized_keys` file on the device for the SSH login
authentication.

```jinja
{% set ssh_identity_file = '/root/.ssh/id_rsa' %}
```

If this parameter is not defined, the insecure `lava` private key distributed
with LAVA is used by default. The corresponding public key is provided below.

```plain
ssh-rsa AAAAB3NzaC1yc2EAAAADAQABAAABAQDLsQXteu02Mvs8Srs8TI/XJqTnfoNDjT5zJVQNI6BqUvvaHSna0iZpcPln9lbmRwAkf84rZAP3eAn051l+GYcRAVAH3bu9HDA8XXIbA4EkCZJ9aCPX7jqtSTBLaIUH28JRPhvP6iZWvqSQck4OmoyrBaMJByBm5CaPR4IhpcAyORF88AGmRW/qFIZTxNm/z1JN/WO+4+C/uM07T+KuInAPBQCTY9pYk4Vd2tZ4msWMYuWs3uVKRdN8aTgyqeyOE3zmXN8Tr2r5uFU0SAe1ZmVnex3s9ZF4YhgmX9SUBuxQw/FjNUajx2D18x/+RQuWXxZOpPQ5ecAysDKROTFWl6QB lava insecure public key
```

### USBG MS

For using the [`usbg-ms`](../job-definition/actions/deploy/to-usbg-ms.md)
deployment method, the device dictionary must define the `usbg_ms_commands`
variable with `enable` and `disable` entries that accept a single string or a
list of strings.

```jinja
{% set usbg_ms_commands = {
    "enable": ["laacli usbg-ms on --image {IMAGE}"],
    "disable": ["laacli usbg-ms off"],
} %}
```

#### enable

The commands for setting up the USB gadget.

The `enable` commands should contain the `{IMAGE}` placeholder which LAVA
replaces with the path to the downloaded image file.

#### disable

The commands for removing the USB gadget.

### UUU

The following variables should be used to configure UUU devices for using
[`uuu`](../job-definition/actions/boot/method-uuu.md) boot method to flash
these devices.

#### uuu_usb_otg_path

The USB OTG path of the device.

```jinja
{% set uuu_usb_otg_path = '2:143' %}
```

The path can be obtained using the command below:

```shell
$ uuu -lsusb
uuu (Universal Update Utility) for nxp imx chips -- libuuu_1.3.102-1-gddf2649
Connected Known USB Devices
    Path	 Chip	 Pro	 Vid	 Pid	 BcdVersion
    ==================================================
    2:143	 MX8MQ	 SDP:	 0x1FC9	0x012B	 0x0001
```

Multiple paths can be provided as a list:

```jinja
{% set uuu_usb_otg_path = ['1:143', '2:143'] %}
```

Multiple paths are needed when using USB 3.0 because UUU can use different USB
protocols at different stages. During the initial `SDPS` stage, the device
typically appears as a USB 2.0 device. After that, UUU may switch to the `FB`
stage, where the device can re-enumerate on a different USB path, often as a
USB 3.0 device for faster flashing.

USB 3.0 is only used when both of the following are true:

1. The board supports USB 3.0 in the relevant flashing stage.
2. The board is connected to a USB 3.0-capable port on the worker, directly or
   through a USB 3.0 hub.

!!! note
    If the flashing process hangs at `SDPS` completion or when entering `FB`, a
    common cause is that the device has switched to a different USB path and the
    device dictionary only contains the first one. You can run `uuu -lsusb` on the
    LAVA worker while the device is transitioning to `FB` to identify the second
    path and then add it to `uuu_usb_otg_path`.

#### uuu_usb_otg_path_command

The command allows running a custom command on the worker to find the OTG paths
on the LAVA worker at runtime so you don't have to modify the device template
for changing device or USB port. In these cases, you will need to update the
command to send the new paths.

```jinja
{% set uuu_usb_otg_path_command = ['echo', '2:143'] %}
```

For multiple OTG paths, your command must print a well formatted usb path
accepted by `uuu` on each line with no new-line at the end of the output.

```jinja
{% set uuu_usb_otg_path_command = ['echo', '-ne', '1:143\n2:143'] %}
```

!!! note
    In practice, the command would be a script that runs on the LAVA worker to
    query and output the OTG paths in the required format, rather than a hardcoded
    `echo` in the examples.

!!! warning
    If `uuu_usb_otg_path` is not set or does not match the expected format,
    LAVA falls back to `uuu_usb_otg_path_command`. You must provide one of them
    in your device dictionary.

#### uuu_corrupt_boot_media_command

Required. A list of commands to execute on the platform within U-Boot to corrupt
the primary boot media. On the next reboot, serial download protocol must be
available on the platform to flash future images using `uuu`.

```jinja
{% set uuu_corrupt_boot_media_command = ['mmc dev 1', 'mmc erase 0 0x400'] %}
```

#### uuu_power_off_before_corrupt_boot_media

In case U-Boot continuously restarts, it will interfere with interrupting U-Boot
while attempting to corrupt the boot media. Set this option to `true` to power
off the device before the boot media corrupting action.

```jinja
{% set uuu_power_off_before_corrupt_boot_media = true %}
```

#### uuu_docker_image

The Docker image used to run UUU commands. The image must contain the `uuu`
binary.

```jinja
{% set uuu_docker_image = 'atline/uuu:1.5.239' %}
```

!!! note
    A `docker` block in the job definition overrides this value. See
    [docker](../job-definition/actions/boot/method-uuu.md#docker).

#### uuu_remote_options

When the UUU device is not directly connected to the LAVA worker, you can use
this parameter to provide the Docker client options for running `uuu` inside a
container on a remote machine.

```jinja
{% set uuu_remote_options = '--tlsverify --tlscacert=/certs/ca.pem --tlscert=/certs/cert.pem --tlskey=/certs/key.pem -H 10.192.244.5:2376' %}
```

See the Docker documentation for [remote access](https://docs.docker.com/engine/daemon/remote-access/)
and [TLS protection](https://docs.docker.com/engine/security/https/).

!!! note
    The minimal Docker version to run uuu is 19.03. This is due to a bug in
    earlier docker versions. See [moby/moby#37665](https://github.com/moby/moby/pull/37665).

### BCU

Most recent i.MX boards (imx8dxl, imx8mp, imx8ulp, imx93 as of july-2022)
support BCU, a remote control utility. BCU allows changing the board's boot
configuration (mainly SD card, eMMC or USB Serial Download Protocol) through a
serial interface. You can use the following parameters to configure BCU.

#### bcu_board_name

Required. The BCU board model name.

```jinja
{% set bcu_board_name = 'imx93evk11b1' %}
```

The list of supported board model can be obtained with `bcu lsboard`:

```shell
$ bcu lsboard
version bcu_1.1.128-0-ge7027dc

list of supported board model:

	imx8dxlevk
	imx8dxlevkc1
	imx8dxl_ddr3_evk
	imx8dxl_obx
	imx8mpevkpwra0
	imx8mpevkpwra1
	imx8mpevk
	imx8mpddr4
	imx8ulpevkb2
	imx8ulpevk9
	imx8ulpwatchval
	imx8ulpwatchuwb
	val_board_1
	val_board_2
	imx91qsb
	imx91evk11
	imx93evk11
	imx93evk11b1
	imx93wevk
	val_board_3
	imx93qsb
	imx93evk14
	imx95evk19
	imx95evk15
	imx952evk15
	imx952evk19
	nxp_custom
	nxp_custom_revB
	val_board_4
	val_board_5
	val_board_6
	val_board_8
	bench_imx8qm
	bench_imx8qm_revB
	bench_imx8qxp
	bench_imx8qxp_revB
	bench_imx8mq
	bench_imx6ull
	bench_mcu
	imx943evk19a0
	imx943evk19b1
	imx943obx
	val_board_7
	val_board_9
	val_board_10
	val_board_11
	val_board_12
done
```

#### bcu_board_id

The BCU board ID. It can be obtained with `bcu lsftdi`.

```jinja
{% set bcu_board_id = "1-3.2.1" %}
```

#### bcu_board_id_command

A command that prints the BCU board identifier at runtime, allowing per-worker
customisation. The command must print a single line with a valid board ID
accepted by `bcu`.

```jinja
{% set bcu_board_id_command = ['echo', '2-1.3'] %}
```

!!! warning
    If `bcu_board_id` is not set or does not match the expected format,
    LAVA falls back to `bcu_board_id_command`. You must provide one of them
    in your device dictionary.

#### additional_bcu_cleanup_commands

A list of additional BCU commands to run during `uuu` boot action cleanup.

```jinja
{% set additional_bcu_cleanup_commands = ["set_gpio ft_fta_sel 0"] %}
```

## Examples

### Beaglebone-black

```jinja
{% extends 'beaglebone-black.jinja2' %}

{% set connection_command = 'telnet dispatcher02 7001' %}

{% set hard_reset_command = 'pdu_control -host pdu20 --cmd reboot --port 10' %}
{% set power_off_command = 'pdu_control --host pdu20 --cmd off --port 10' %}
{% set power_on_command = 'pdu_control --host pdu20 --cmd on --port 10' %}

```

### dragonboard-410c

```jinja
{% extends 'dragonboard-410c.jinja2' %}

{% set connection_command = 'telnet dispatcher03 7004' %}

{% set adb_serial_number = '8c5f2290' %}
{% set fastboot_serial_number = '8c5f2290' %}

{# OTG power control #}
{% set pre_power_command = 'hub_control -n dispatcher03-hub01 -m sync -u 02' %}
{% set pre_os_command = 'hub_control -n dispatcher03-hub01 -m off -u 02' %}
{% set hard_reset_command = [
        'hub_control -n dispatcher03-hub01 -m off -u 02',
        'pdu_control --host pdu04 --command reboot --port 19 --delay 50',
        'hub_control -n dispatcher03-hub01 -m sync -u 02'] %}
{% set power_on_command = 'pdu_control --host pdu04 --command on --port 19' %}
{% set power_off_command = ['pdu_control --host pdu04 --command off --port 19',
                            'hub_control -n dispatcher03-hub01 -m off -u 02'] %}

{% set device_info = [{'board_id': '8c5f2290'}] %}
{% set flash_cmds_order = ['update', 'ptable', 'partition', 'hyp', 'modem',
                           'rpm', 'sbl1', 'sbl2', 'sec', 'tz', 'aboot',
                           'boot', 'rootfs', 'vendor', 'system', 'cache',
                           'userdata', ] %}
{% set device_ip = "10.7.0.74" %}
```

--8<-- "refs.txt"

# Device dictionary

The device dictionary is a [jinja2][jinja2] template that should extend one
[device-type template](./device-type-template.md).

While device-type templates are usually provided and maintained by LAVA, device
dictionaries are provided and maintained by the admin.

A device dictionary will contain device specific information like:

* connection command
* power on/off commands
* bootloader prompt
* character delays
* ...

See [Parameters](#parameters)

## Configuration file

Device dictionaries are stored on the server in
`/etc/lava-server/dispatcher-config/devices/<hostname>.jinja2`.

Admin could update device dictionaries using lavacli:

```shell
lavacli devices dict set qemu01 qemu01.jinja2
```

## Parameters

The variables you may define in a device dictionary depend on the
[device-type template](./device-type-template.md) it extends. Common
parameters include connection and power commands. See the device type template
and its base template for the full set of supported parameters.

### Commands

All the command parameters in this section accept either a single string or a
list of strings. When a list is used, LAVA executes each command in order.

#### power_on_command

Command to supply power to the device remotely. The device **must** start the
boot sequence when the `power_on_command` is called. If the DUT needs a button
to be pressed to boot, then the command to press the button should be part of
this one.

```jinja
{% set power_on_command =  [
    'laacli power 5v on',
    'laacli power 3v3 on'
] %}
```

#### power_off_command

Command to remove power from the device remotely.

```jinja
{% set power_off_command = [
    'laacli power 3v3 off',
    'laacli power 5v off'
] %}
```

#### soft_reboot_command

Command to run at a shell prompt on a running system to request a reboot (e.g.
`reboot` or `systemctl reboot`).

```jinja
{% set soft_reboot_command = 'reboot' %}
```

#### hard_reset_command

Command to cut power and then restore it remotely (e.g. via PDU or hub). You can
insert a delay using `sleep` between cutting and restoring power if the device
requires it.

```jinja
{% set hard_reset_command = ['usbrelay 1_2=0', 'sleep 1', 'usbrelay 1_2=1'] %}
```

#### pre_power_command

Ancillary command for special cases. Behavior and support depend on the
deployment method and device-type template.

```jinja
{% set pre_power_command = 'laacli usb 1 on' %}
```

#### pre_os_command

Ancillary command for special cases. Behavior and support depend on the
deployment method and device-type template.

```jinja
{% set pre_os_command = 'laacli usb 1 off' %}
```

### Connections

#### connection_command

Command to access the serial port of the device.

```jinja
{% set connection_command = 'telnet dispatcher03 7004' %}
```

#### connection_list

The list of hardware ports which are configured for serial connections to the device.

```jinja
{% set connection_list = ['uart0'] %}
```

#### connection_commands

A dictionary of the commands to start each connection.

```jinja
{% set connection_commands = {'uart0': 'telnet w1 7005'} %}
```

#### connection_tags

Each connection can include `tags` - extra pieces of metadata to describe the connection.

```jinja
{% set connection_tags = {'uart0': ['primary', 'telnet']} %}
```

There must always be one (and only one) connection with the `primary` tag, denoting
the connection that will be used for firmware, bootloader and kernel interaction.

Other tags may describe the *type* of connection, as extra information that LAVA
can use to determine how to close the connection cleanly when a job finishes
(e.g `telnet` and `ssh`).

### Device

Some devices or the elements of the device configuration can be exposed to the
LAVA test shell, where it is safe to do so. Each parameter must be explicitly
set in each device dictionary.

For ease of use, LAVA will directly export the content of the `device_info`,
`environment`, `static_info` and `storage_info` dictionaries into the test
shell environment. The dictionaries and lists will be unrolled, for example:

```jinja
{% set environment = {"RELAY_ADDRESS": "10.66.16.103", "REMOTE_SERIAL_PORT": "/dev/ttyUSB2"} %}
{% set device_info = [{'board_id': 'a2c22e48'}] %}
{% set static_info = [{"board_id": "S_NO81730000"}, {"board_id": "S_NO81730001"}] %}
{% set storage_info = [{'SATA': '/dev/disk/by-id/ata-ST500DM002-1BD142_W3T79GCW'}] %}
```

will become:

```shell
export RELAY_ADDRESS='10.66.16.103'
export REMOTE_SERIAL_PORT='/dev/ttyUSB2'
export LAVA_DEVICE_INFO_0_board_id='a2c22e48'
export LAVA_STATIC_INFO_0_board_id='S_NO81730000'
export LAVA_STATIC_INFO_1_board_id='S_NO81730001'
export LAVA_STORAGE_INFO_0_SATA='/dev/disk/by-id/ata-ST500DM002-1BD142_W3T79GCW'
```

The environment can be overridden in the job definition, see
[job environment](../job-definition/environment.md)

#### device_info

`device_info` is a list of dictionaries, where each dictionary value can contain
keys such as `board_id`, `usb_vendor_id`, `usb_product_id`, `wait_device_board_id`,
which can be made available to the Docker container for device specific tasks
dynamically, whenever the device is reset, using a `udev` rule.

```jinja
{% set device_info = [{'board_id': '8c5f2290'}] %}
```

#### environment

A dictionary containing device-specific shell variables, which will be available
in the LAVA test shell. These can be used, for example, to describe physical
hardware connections between the DUT and interfaces on the worker or other
addressable hardware.

```jinja
{% set environment = {
    'RELAY_ADDRESS': '10.66.16.103',
    'REMOTE_SERIAL_PORT': '/dev/ttyUSB2',
} %}
```

#### static_info

A list of dictionaries describing static information. A common use case is
defining a file server address for URL substitution. For example:

```jinja
{% set static_info = [{'FILE_SERVER_IP': "10.192.244.104"}] %}
```

See also [URL placeholders](../job-definition/actions/deploy/index.md#url).

#### storage_info

A list of dictionaries, where each dictionary value can contain keys describing
the storage name (e.g. USB or SATA) and a value stating the device node of the
top level block device which is available to the test writer.

```jinja
{% set storage_info = [{'SATA': '/dev/disk/by-id/ata-ST500DM002-1BD142_W3T79GCW'}] %}
```

#### device_ip

A single fixed IPv4 address of this device. The value will be exported into the
test shell using
[`lava-target-ip`](../../user/basic-tutorials/test-definition.md#lava-target-ip).

```jinja
{% set device_ip = "10.66.16.24" %}
```

#### device_mac

Similar to `device_ip` but for a single MAC address. The value will be exported
into the test shell using
[`lava-target-mac`](../../user/basic-tutorials/test-definition.md#lava-target-mac).

```jinja
{% set device_mac = '00:02:F7:00:58:53' %}
```

### Character delays

LAVA supports specifying character delays in the deploy, boot, and test
actions to help with serial reliability (e.g. when the DUT or connection
drops or corrupts characters when input is sent too quickly). These are
device-specific, so they are best set in the device dictionary (or
device-type template).

The `deploy` and `boot` actions are affected more often than the `test`
action, because they interacts with firmware or bootloader processes where
input handling can be more limited than in a POSIX test environment.

#### deploy_character_delay

Set the number of milliseconds to add between each character of every
string sent to the DUT during the `deploy` action. This is useful for
deployment methods that use a connection (e.g. `vemsd`):

```jinja
{% set deploy_character_delay = 30 %}
```

#### boot_character_delay

Set the number of milliseconds to add between each character of every
string sent to the DUT during the `boot` action:

```jinja
{% set boot_character_delay = 20 %}
```

Some devices need more (e.g. 100 or 500 ms). For long delays, also consider
the overall boot timeout and set a minimum for the relevant boot action in
the device-type template.

#### test_character_delay

Set the number of milliseconds to add between each character of every
string sent to the DUT during the `test` action:

```jinja
{% set test_character_delay = 10 %}
```

!!! note
    LAVA also waits `test_character_delay` milliseconds before sending each
    test signal. The delay can be helpful on slow serial connections to avoid
    character interleaving with other inputs, such as kernel dmesg.

### Depthcharge

The following variables can be used in a Depthcharge device dictionary to
configure the [`depthcharge`](../job-definition/actions/boot/method-depthcharge.md)
boot method.

#### start_message

The string that Depthcharge prints when it is ready to accept netboot
commands. LAVA waits for this message after resetting the device.

```jinja
{% set start_message = start_message | default('Starting depthcharge on Volmar...') %}
```

Defaults to `Starting netboot`.

#### bootloader_prompt

The Depthcharge command line interface prompt string. LAVA uses this to detect
when the bootloader is ready for sending the next command.

```jinja
{% set bootloader_prompt = bootloader_prompt | default('volmar:') %}
```

Defaults to `dpch:`.

#### depthcharge_final_message

The string that depthcharge prints when it has finished. LAVA waits for this
message before continuing.

```jinja
{% set depthcharge_final_message = depthcharge_final_message | default('jumping to kernel') %}
```

Defaults to `Starting kernel`.

See also [bootloader-final-message](../job-definition/actions/boot/common.md#bootloader-final-message).

### DFU

`board_id` must be provided for adding a new DFU device.

```jinja
{% set board_id = "LCES2-0xf775916400000000" %}
```

### Fastboot

The following variables can be used in fastboot device dictionary to configure
fastboot device.

#### adb_serial_number

A string to specify the serial number of the device in ADB mode.

```jinja
{% set adb_serial_number = '25564f71' %}
```

#### fastboot_serial_number

A string to specify the serial number of the device in fastboot mode.

```jinja
{% set fastboot_serial_number = '25564f71' %}
```

#### fastboot_auto_detection

A boolean. When `true`, LAVA tries to detect the device serial number
automatically.

```jinja
{% set fastboot_auto_detection = true %}
```

!!! note
    `fastboot_auto_detection` assumes the ADB serial number and fastboot serial
    number are the same for the same device.

When `false` or unset, you must set `adb_serial_number` and
`fastboot_serial_number` explicitly.

#### fastboot_options

A list of strings, used for specifying additional options to the `fastboot` command.

```jinja
{% set fastboot_options = ['-S', '256M'] %}
```

#### flash_cmds_order

A list of strings, used for specifying the order in which the images should be
flashed to the DUT using the `fastboot` command.

```jinja
{% set flash_cmds_order = ['update', 'ptable', 'partition', 'hyp', 'modem',
                           'rpm', 'sbl1', 'sbl2', 'sec', 'tz', 'aboot',
                           'boot', 'rootfs', 'vendor', 'system', 'cache',
                           'userdata', ] %}
```

#### fastboot_sequence

A list of strings. In practice, you only need to set the list to use one of the
following method for most of the device types, then LAVA will populate and
execute the corresponding action.

##### boot

Runs `fastboot boot <boot-image>`, loading the `boot` image from the preceding
deploy action directly into RAM and executing it.

```jinja
{% set fastboot_sequence = ['boot'] %}
```

##### reboot

Runs `fastboot reboot`, rebooting the device into the OS from its previously
flashed partitions. This is the typical path when all images (including `boot`)
were flashed during the deploy action.

```jinja
{% set fastboot_sequence = ['reboot'] %}
```

##### no-flash-boot

Same behaviour as `boot` at boot time — the image is loaded into RAM via
`fastboot boot`. Additionally, during the deploy phase, LAVA skips flashing the
`boot` partition even if a `boot` image was listed under `images` in the deploy
action.

```jinja
{% set fastboot_sequence = ['no-flash-boot'] %}
```

### Flasher

The following `flasher_deploy_commands` variable must be configured for using
the [flasher](../job-definition/actions/deploy/to-flasher.md) deployment method.

#### flasher_deploy_commands

A list of commands that LAVA executes on the worker for flashing images onto the
DUT. For example, using a USB-SD-Mux to write an image to the DUT's SD card:

```jinja
{% set flasher_deploy_commands = [
        '/root/.local/bin/usbsdmux /dev/sg1 host',
        'sleep 3',
        '/root/.local/bin/usbsdmux /dev/sg1 info',
        'dd if={RECOVERY_IMAGE} of=/dev/disk/by-id/usb-LinuxAut_sdmux_HS-SD_MMC_000000001156-0:0 bs=4M oflag=sync conv=nocreat',
        'sleep 3',
        '/root/.local/bin/usbsdmux /dev/sg1 client',
    ]
%}
```

The following placeholders can be used in the commands. LAVA substitutes them
with the actual values at runtime.

| Variable | Value |
| --- | --- |
| `{UPPER_CASE_IMAGE_KEY}` | Downloaded image path (e.g., `{RECOVERY_IMAGE}` for `images.recovery_image`) |
| `{POWER_ON_COMMAND}` | [`power_on_command`](#power_on_command) |
| `{POWER_OFF_COMMAND}` | [`power_off_command`](#power_off_command) |
| `{SOFT_REBOOT_COMMAND}` | [soft_reboot_command](#soft_reboot_command) |
| `{HARD_RESET_COMMAND}` | [`hard_reset_command`](#hard_reset_command) |
| `{PRE_OS_COMMAND}` | [`pre_os_command`](#pre_os_command) |
| `{PRE_POWER_COMMAND}` | [`pre_power_command`](#pre_power_command) |
| `{DEVICE_INFO}` | YAML dump of [`device_info`](#device_info) |
| `{STATIC_INFO}` | YAML dump of [`static_info`](#static_info) |

### JLink

#### board_id

The serial number of the JLink probe or target board.

```jinja
{% set board_id = '000380000008' %}
```

### Secondary media

Use `<media>_label` and `<media>_uuid` to configure a secondary media in your
device dictionary. The `/dev/disk/by-id/<media>_uuid` must exist on the LAVA
worker.

#### USB

```jinja2
{% set usb_label = 'SanDiskCruzerBlade' %}
{% set usb_uuid = 'usb-SanDisk_Cruzer_Blade_20060266531DA442AD42-0:0' %}
```

#### SATA

```jinja2
{% set sata_label = 'ST160LM003' %}
{% set sata_uuid = "ata-ST160LM003_HN-M160MBB_S2SYJ9KC102184" %}
```

#### SD

```jinja2
{% set sd_label = 'sdcard' %}
{% set sd_uuid = 'mmc-SD16G_0xda85ac89' %}
```

### SSH

SSH device can be configured using the following parameters.

```jinja
{% extends 'ssh.jinja2' %}

{% set ssh_host = '192.168.1.100' %}
{% set ssh_port = 22 %}
{% set ssh_user = 'root' %}
{% set ssh_identity_file = '/<path>/private_key' %}
```

#### ssh_host

Required. The hostname or IP address of the device.

```jinja
{% set ssh_host = '192.168.1.100' %}
```

#### ssh_port

The SSH port on the device. Defaults to `22`.

```jinja
{% set ssh_port = 2222 %}
```

#### ssh_user

The login user for SSH and SCP connections. Defaults to `root`.

```jinja
{% set ssh_user = 'testuser' %}
```

#### ssh_identity_file

Path to the SSH private key on the LAVA worker. The matching public key must be
added to the `~/.ssh/authorized_keys` file on the device for the SSH login
authentication.

```jinja
{% set ssh_identity_file = '/root/.ssh/id_rsa' %}
```

If this parameter is not defined, the insecure `lava` private key distributed
with LAVA is used by default. The corresponding public key is provided below.

```plain
ssh-rsa AAAAB3NzaC1yc2EAAAADAQABAAABAQDLsQXteu02Mvs8Srs8TI/XJqTnfoNDjT5zJVQNI6BqUvvaHSna0iZpcPln9lbmRwAkf84rZAP3eAn051l+GYcRAVAH3bu9HDA8XXIbA4EkCZJ9aCPX7jqtSTBLaIUH28JRPhvP6iZWvqSQck4OmoyrBaMJByBm5CaPR4IhpcAyORF88AGmRW/qFIZTxNm/z1JN/WO+4+C/uM07T+KuInAPBQCTY9pYk4Vd2tZ4msWMYuWs3uVKRdN8aTgyqeyOE3zmXN8Tr2r5uFU0SAe1ZmVnex3s9ZF4YhgmX9SUBuxQw/FjNUajx2D18x/+RQuWXxZOpPQ5ecAysDKROTFWl6QB lava insecure public key
```

### USBG MS

For using the [`usbg-ms`](../job-definition/actions/deploy/to-usbg-ms.md)
deployment method, the device dictionary must define the `usbg_ms_commands`
variable with `enable` and `disable` entries that accept a single string or a
list of strings.

```jinja
{% set usbg_ms_commands = {
    "enable": ["laacli usbg-ms on --image {IMAGE}"],
    "disable": ["laacli usbg-ms off"],
} %}
```

#### enable

The commands for setting up the USB gadget.

The `enable` commands should contain the `{IMAGE}` placeholder which LAVA
replaces with the path to the downloaded image file.

#### disable

The commands for removing the USB gadget.

### UUU

The following variables should be used to configure UUU devices for using
[`uuu`](../job-definition/actions/boot/method-uuu.md) boot method to flash
these devices.

#### uuu_usb_otg_path

The USB OTG path of the device.

```jinja
{% set uuu_usb_otg_path = '2:143' %}
```

The path can be obtained using the command below:

```shell
$ uuu -lsusb
uuu (Universal Update Utility) for nxp imx chips -- libuuu_1.3.102-1-gddf2649
Connected Known USB Devices
    Path	 Chip	 Pro	 Vid	 Pid	 BcdVersion
    ==================================================
    2:143	 MX8MQ	 SDP:	 0x1FC9	0x012B	 0x0001
```

Multiple paths can be provided as a list:

```jinja
{% set uuu_usb_otg_path = ['1:143', '2:143'] %}
```

Multiple paths are needed when using USB 3.0 because UUU can use different USB
protocols at different stages. During the initial `SDPS` stage, the device
typically appears as a USB 2.0 device. After that, UUU may switch to the `FB`
stage, where the device can re-enumerate on a different USB path, often as a
USB 3.0 device for faster flashing.

USB 3.0 is only used when both of the following are true:

1. The board supports USB 3.0 in the relevant flashing stage.
2. The board is connected to a USB 3.0-capable port on the worker, directly or
   through a USB 3.0 hub.

!!! note
    If the flashing process hangs at `SDPS` completion or when entering `FB`, a
    common cause is that the device has switched to a different USB path and the
    device dictionary only contains the first one. You can run `uuu -lsusb` on the
    LAVA worker while the device is transitioning to `FB` to identify the second
    path and then add it to `uuu_usb_otg_path`.

#### uuu_usb_otg_path_command

The command allows running a custom command on the worker to find the OTG paths
on the LAVA worker at runtime so you don't have to modify the device template
for changing device or USB port. In these cases, you will need to update the
command to send the new paths.

```jinja
{% set uuu_usb_otg_path_command = ['echo', '2:143'] %}
```

For multiple OTG paths, your command must print a well formatted usb path
accepted by `uuu` on each line with no new-line at the end of the output.

```jinja
{% set uuu_usb_otg_path_command = ['echo', '-ne', '1:143\n2:143'] %}
```

!!! note
    In practice, the command would be a script that runs on the LAVA worker to
    query and output the OTG paths in the required format, rather than a hardcoded
    `echo` in the examples.

!!! warning
    If `uuu_usb_otg_path` is not set or does not match the expected format,
    LAVA falls back to `uuu_usb_otg_path_command`. You must provide one of them
    in your device dictionary.

#### uuu_corrupt_boot_media_command

Required. A list of commands to execute on the platform within U-Boot to corrupt
the primary boot media. On the next reboot, serial download protocol must be
available on the platform to flash future images using `uuu`.

```jinja
{% set uuu_corrupt_boot_media_command = ['mmc dev 1', 'mmc erase 0 0x400'] %}
```

#### uuu_power_off_before_corrupt_boot_media

In case U-Boot continuously restarts, it will interfere with interrupting U-Boot
while attempting to corrupt the boot media. Set this option to `true` to power
off the device before the boot media corrupting action.

```jinja
{% set uuu_power_off_before_corrupt_boot_media = true %}
```

#### uuu_docker_image

The Docker image used to run UUU commands. The image must contain the `uuu`
binary.

```jinja
{% set uuu_docker_image = 'atline/uuu:1.5.239' %}
```

!!! note
    A `docker` block in the job definition overrides this value. See
    [docker](../job-definition/actions/boot/method-uuu.md#docker).

#### uuu_remote_options

When the UUU device is not directly connected to the LAVA worker, you can use
this parameter to provide the Docker client options for running `uuu` inside a
container on a remote machine.

```jinja
{% set uuu_remote_options = '--tlsverify --tlscacert=/certs/ca.pem --tlscert=/certs/cert.pem --tlskey=/certs/key.pem -H 10.192.244.5:2376' %}
```

See the Docker documentation for [remote access](https://docs.docker.com/engine/daemon/remote-access/)
and [TLS protection](https://docs.docker.com/engine/security/https/).

!!! note
    The minimal Docker version to run uuu is 19.03. This is due to a bug in
    earlier docker versions. See https://github.com/moby/moby/pull/37665.

### BCU

Most recent i.MX boards (imx8dxl, imx8mp, imx8ulp, imx93 as of july-2022)
support BCU, a remote control utility. BCU allows changing the board's boot
configuration (mainly SD card, eMMC or USB Serial Download Protocol) through a
serial interface. You can use the following parameters to configure BCU.

#### bcu_board_name

Required. The BCU board model name.

```jinja
{% set bcu_board_name = 'imx93evk11b1' %}
```

The list of supported board model can be obtained with `bcu lsboard`:

```shell
$ bcu lsboard
version bcu_1.1.128-0-ge7027dc

list of supported board model:

	imx8dxlevk
	imx8dxlevkc1
	imx8dxl_ddr3_evk
	imx8dxl_obx
	imx8mpevkpwra0
	imx8mpevkpwra1
	imx8mpevk
	imx8mpddr4
	imx8ulpevkb2
	imx8ulpevk9
	imx8ulpwatchval
	imx8ulpwatchuwb
	val_board_1
	val_board_2
	imx91qsb
	imx91evk11
	imx93evk11
	imx93evk11b1
	imx93wevk
	val_board_3
	imx93qsb
	imx93evk14
	imx95evk19
	imx95evk15
	imx952evk15
	imx952evk19
	nxp_custom
	nxp_custom_revB
	val_board_4
	val_board_5
	val_board_6
	val_board_8
	bench_imx8qm
	bench_imx8qm_revB
	bench_imx8qxp
	bench_imx8qxp_revB
	bench_imx8mq
	bench_imx6ull
	bench_mcu
	imx943evk19a0
	imx943evk19b1
	imx943obx
	val_board_7
	val_board_9
	val_board_10
	val_board_11
	val_board_12
done
```

#### bcu_board_id

The BCU board ID. It can be obtained with `bcu lsftdi`.

```jinja
{% set bcu_board_id = "1-3.2.1" %}
```

#### bcu_board_id_command

A command that prints the BCU board identifier at runtime, allowing per-worker
customisation. The command must print a single line with a valid board ID
accepted by `bcu`.

```jinja
{% set bcu_board_id_command = ['echo', '2-1.3'] %}
```

!!! warning
    If `bcu_board_id` is not set or does not match the expected format,
    LAVA falls back to `bcu_board_id_command`. You must provide one of them
    in your device dictionary.

#### additional_bcu_cleanup_commands

A list of additional BCU commands to run during `uuu` boot action cleanup.

```jinja
{% set additional_bcu_cleanup_commands = ["set_gpio ft_fta_sel 0"] %}
```

## Examples

### Beaglebone-black

```jinja
{% extends 'beaglebone-black.jinja2' %}

{% set connection_command = 'telnet dispatcher02 7001' %}

{% set hard_reset_command = 'pdu_control -host pdu20 --cmd reboot --port 10' %}
{% set power_off_command = 'pdu_control --host pdu20 --cmd off --port 10' %}
{% set power_on_command = 'pdu_control --host pdu20 --cmd on --port 10' %}

```

### dragonboard-410c

```jinja
{% extends 'dragonboard-410c.jinja2' %}

{% set connection_command = 'telnet dispatcher03 7004' %}

{% set adb_serial_number = '8c5f2290' %}
{% set fastboot_serial_number = '8c5f2290' %}

{# OTG power control #}
{% set pre_power_command = 'hub_control -n dispatcher03-hub01 -m sync -u 02' %}
{% set pre_os_command = 'hub_control -n dispatcher03-hub01 -m off -u 02' %}
{% set hard_reset_command = [
        'hub_control -n dispatcher03-hub01 -m off -u 02',
        'pdu_control --host pdu04 --command reboot --port 19 --delay 50',
        'hub_control -n dispatcher03-hub01 -m sync -u 02'] %}
{% set power_on_command = 'pdu_control --host pdu04 --command on --port 19' %}
{% set power_off_command = ['pdu_control --host pdu04 --command off --port 19',
                            'hub_control -n dispatcher03-hub01 -m off -u 02'] %}

{% set device_info = [{'board_id': '8c5f2290'}] %}
{% set flash_cmds_order = ['update', 'ptable', 'partition', 'hyp', 'modem',
                           'rpm', 'sbl1', 'sbl2', 'sec', 'tz', 'aboot',
                           'boot', 'rootfs', 'vendor', 'system', 'cache',
                           'userdata', ] %}
{% set device_ip = "10.7.0.74" %}
```

--8<-- "refs.txt"

# Backup LAVA

In order to backup LAVA you will have to backup both the database and specific
directories.

## Database

### For a native Debian install
The database is managed by
[PostgreSQL](../../../technical-references/services/postgresql.md). See the
[PostgreSQL documentation][pgsql-backup] for
details.

To backup the database run:

```shell
sudo -u lavaserver pg_dump lavaserver > lava-server.sql
```

### For the docker-compose solution
If you run your LAVA instance with [docker-compose](https://gitlab.com/lava/pkg/docker-compose),
you can backup the database like this:
```shell
docker exec --user postgres docker-compose_db_1 bash -c \
            "pg_dump --username=lavaserver lavaserver > /tmp/lavaserver.sql"
```

On the host, you can then retrieve the backup file.
Example:
```shell
docker cp docker-compose_db_1:/tmp/lavaserver.sql .
```

## Filesystem

The configuration files are stored in:

* `/etc/lava-coordinator/`
* `/etc/lava-dispatcher/`
* `/etc/lava-server/`

The data (job outputs, ...) are stored in `/var/lib/lava-server/default/media`.

To backup the filesystem run:

```shell
tar czf lava-server-etc.tar.gz /etc/lava-coordinator/ \
                               /etc/lava-dispatcher/ \
                               /etc/lava-server/
tar czf lava-server-data.tar.gz /var/lib/lava-server/default/media/
```

--8<-- "refs.txt"

# lavacli

[lavacli] is a command line tool to interact with one or many LAVA instances
using XML-RPC.

## Install

### Debian

A package is provided for both Debian and Ubuntu:

```shell
apt-get install lavacli
```

### Pypi

lavacli is also available on PyPi:

```shell
python3 -m pip install lavacli
```

## Configure

### Create a token

In order to access to restricted API methods, lavacli will need a `token` to authenticate.

In the web interface, go to [/api/tokens/](/api/tokens/).

![submit](token-menu.png)

Click on `new` to create a token.

![submit](token-new.png)

In the modal, you can add a description to the new token. This description
can be left empty for the moment.

![submit](token-new-dialog.png)

The token has been created, click on the green eye icon and copy the value
shown in the modal dialog.

![submit](token-show.png)

### Create an identity

You can now add this token to lavacli identities with:

```shell
lavacli identities add --uri https://<lava-server>/RPC2/ \
                       --username <username> \
                       --token <token> \
                       myserver
```

### Use the identity

In order to use the identity, call `lavacli -i myserver <command>`.

!!! tip "default identity"
    If the identity is called `default` lavacli will use it for every command.

    `lavacli -i default <command>` and `lavacli <command>` will use the same identity.

## Help

For more information, refer to the [lavacli documentation][lavacli].

--8<-- "refs.txt"

# QEMU device setup

LAVA can use qemu as a DUT and run test inside QEMU.

## Create device-type

Create a device-type in the [admin interface](/admin/lava_scheduler_app/devicetype/add/).

The only relevant information is the device-type name that should be **qemu**.

!!! tip "Command line"

    ```shell tab="lavacli"
    lavacli device-types add qemu
    ```

    ```shell tab="lava-server"
    lava-server manage device-types add qemu
    ```

## Create device

Create a qemu device in the [admin interface](/admin/lava_scheduler_app/device/add/):

* hostname: name of the device
* device-type: **qemu**
* worker host: worker that will run the job

!!! tip "Command line"

    ```shell tab="lavacli"
    lavacli devices add --type qemu --worker <worker> <hostname>
    ```

    ```shell tab="lava-server"
    lava-server manage devices add \
        --device-type qemu \
        --worker <worker> \
        <hostname>
    ```

## Device configuration

In order to submit jobs to the newly created device, LAVA requires a device
dictionary. For a simple qemu job, this device dictionary would work:

```jinja
{% extends "qemu.jinja2" %}

{% set netdevice = "user" %}
{% set memory = 1024 %}
```

This file should be pushed to the LAVA server under
`/etc/lava-server/dispatcher-config/devices/<hostname>.jinja2`.

!!! tip "Command line"

    ```shell tab="lavacli"
    lavacli devices dict set <hostname> <filename>
    ```

    ```bash tab="lava-server"
    cp <filename> /etc/lava-server/dispatcher-config/devices/<hostname>.jinja2
    chown lavaserver:lavaserver /etc/lava-server/dispatcher-config/devices/<hostname>.jinja2
    ```

## Activate the device

By default, a new device is put in maintenance.

As the device is now configure, admins can put it online in the [device page](/scheduler/device/<hostname>).

!!! tip "Command line"

    ```shell tab="lavacli"
    lavacli devices update --health UNKNOWN <hostname>
    ```

    ```bash tab="lava-server"
    lava-server manage devices update --health UNKNOWN <hostname>
    ```

## Submit a job

Submit this simple test job:

```yaml
--8<-- "jobs/qemu.yaml"
```

The job page will look like [this](https://validation.linaro.org/scheduler/job/2009038).

--8<-- "refs.txt"

# Adding a new device type

The integration process is different for every new device type. Therefore,
this documentation can only provide hints about such devices, based on
experience within the LAVA software and lab teams. **Please** talk to us
**before** starting on the integration of a new device type using the
mailing lists.

Integrating a new device type will involve some level of development
work. Testing new device type templates requires setting up a developer
workflow and running unit tests as well as running test jobs on a LAVA
instance. If the new device type involves a new boot or deployment
method, there will also need to be changes in the `lava-dispatcher`
codebase.

## Device requirements

It is not possible to automate every piece of hardware, there are a number
of critical limitations.

### Power

Devices **MUST** support automated resets either by the removal of all
power supplied to the DUT or a full reboot or other reset which clears
all previous state.

**Every** boot **must** reliably start, without interaction, directly
from the first application of power.

!!! warning "Power leaks"
    Some devices are capable of drawing power over the serial line used
    to control the device, despite the actual power supply being
    disconnected. Sometimes this requires a period of time to discharge
    capacitors on the board (fixable by adding a sleep in the power
    control commands). Sometimes this power leak can cause the device to
    latch into a particular bootloader mode or other state which prevents
    the automation from proceeding.

### Serial console

LAVA expects to automate devices by interacting with the serial port
immediately after power is applied to the device. The bootloader
**must** interact with the serial port. If a serial port is not
available on the device, suitable additional hardware **must** be
provided before integration can begin.

### Networking

**Ethernet** — all devices using Ethernet interfaces in LAVA **must**
have a unique MAC address on each interface. The MAC address **must** be
persistent across reboots.

### Reproducibility

Reproducibility is the ability to deploy exactly the same software to
the same board(s) and running exactly the same tests many times in a
row, getting exactly the same results each time.

For automation to work, all device functions which need to be used in
automation **must** always produce the same results on each device of a
specific device type, irrespective of any previous operations on that
device, given the same starting hardware configuration.

There is no way to automate a device which behaves unpredictably.

### Reliability

Reliability is the ability to run a wide range of test jobs, stressing
different parts of the overall deployment, with a variety of tests and
**always** getting a `Complete` test job. There must be no `JobError` or
`InfrastructureError` failures and there should be limited variability
in the time taken to run the test jobs.

### Scriptability

The device **must** support deployment of files and booting without
**any** need for a human to monitor or interact with the process. The
need to press buttons is undesirable but can be managed in some cases by
using relays.

### Scalability

All methods used to automate a device **must** have minimal footprint
in terms of load on the workers, complexity of scripting support and
infrastructure requirements.

This is a complex area and can trivially impact on both reliability and
reproducibility as well as making it much more difficult to debug problems
which do arise. Admins must also consider the complexity of combining
multiple different devices which each require multiple layers of support.

Some devices may need:

- Relays to work around buttons
- Specialized hardware to work around deployment limitations
- Complex scripting around power control
- A need to use Docker for automation

Any one of these burdens will make debugging issues on the worker and on
the devices difficult. Any combination of these burdens make debugging
many times more difficult than any one burden alone.

## Integration process

To add support for a new device type, a certain amount of development
and testing **will** be required.

For most device types, extending an existing base template should be
enough. For other device types, new dispatcher action classes and new or
modified strategy classes will be needed. This typically involves a lot
of development time — make sure that you contribute upstream so that your
local changes do not break when you next upgrade your LAVA instance(s).

### Find a similar existing device type

Check the already
[supported device types](https://gitlab.com/lava/lava/-/tree/master/etc/dispatcher-config/device-types)
for:

- Similar bootloader
- Similar deployment type
- Similar deployment or boot process
- Similar sequence of boot steps

If you do not find something similar, we strongly recommend that you
**stop** and [talk to us](../../introduction/contact.md) before doing
anything else.

### Extend from an existing device type template

All new device type templates need to `extend 'base.jinja2'`, but there
are other base templates which simplify the process for certain bootloaders.
See [base templates](../../technical-references/configuration/device-type-template.md#base-templates)
for the list.

Avoid directly extending any of the templates which do not have the
`base` prefix — instead copy the existing template for your new device
type.

### Extend the template unit tests

All device type template files in `tests/lava_scheduler_app/devices`
will be checked for simple YAML validity by the `test_all_templates`
unit test. However, a dedicated unit test is recommended for all but the
simplest of new device type templates. You can edit
`tests/lava_scheduler_app/test_templates.py` and add a new unit test
for your device-type based on one of the existing test functions.

Every time you make a change to the new template, re-run the template tests:

```shell
pytest tests/lava_scheduler_app/test_templates.py
```

## See also

- [Adding new actions](./new-actions.md)
- [Device type templates](../../technical-references/configuration/device-type-template.md)
- [Device dictionary](../../technical-references/configuration/device-dictionary.md)

# lavacli

[lavacli] is a command line tool to interact with one or many LAVA instances
using XML-RPC.

## Install

### Debian

A package is provided for both Debian and Ubuntu:

```shell
apt-get install lavacli
```

### Pypi

lavacli is also available on PyPi:

```shell
python3 -m pip install lavacli
```

## Configure

### Create a token

In order to access to restricted API methods, lavacli will need a `token` to authenticate.

In the web interface, go to `/api/tokens/` on your LAVA server.

![submit](token-menu.png)

Click on `new` to create a token.

![submit](token-new.png)

In the modal, you can add a description to the new token. This description
can be left empty for the moment.

![submit](token-new-dialog.png)

The token has been created, click on the green eye icon and copy the value
shown in the modal dialog.

![submit](token-show.png)

### Create an identity

You can now add this token to lavacli identities with:

```shell
lavacli identities add --uri https://<lava-server>/RPC2/ \
                       --username <username> \
                       --token <token> \
                       myserver
```

### Use the identity

In order to use the identity, call `lavacli -i myserver <command>`.

!!! tip "default identity"
    If the identity is called `default` lavacli will use it for every command.

    `lavacli -i default <command>` and `lavacli <command>` will use the same identity.

## Help

For more information, refer to the [lavacli documentation][lavacli].

--8<-- "refs.txt"

# QEMU device setup

LAVA can use qemu as a DUT and run test inside QEMU.

## Create device-type

[Create the device type](common.md#create-device-type) using the name **`qemu`**.

## Create device

1. [Add the device](common.md#add-device) using the following settings:
    * **Device Type:** `qemu`
    * **Hostname:** A unique name (e.g., `qemu-01`)
2. [Add the device configuration](common.md#add-device-configuration).

    For a simple qemu job, this device dictionary would work:

    ```jinja
    {% extends "qemu.jinja2" %}

    {% set netdevice = "user" %}
    {% set memory = 1024 %}
    ```

    !!! tip
        If `/dev/kvm` is unavailable on the worker, add `{% set no_kvm = True %}` to
        the dictionary.

## Submit a job

Submit this simple test job:

```yaml
--8<-- "jobs/qemu.yaml"
```

The job page will look like [this](https://validation.linaro.org/scheduler/job/2009038).

## Configure bridged network

For `qemu-nfs` and `qemu-iso` boot methods, `netdevice` must be set to `tap`
in either the job context or the device dictionary. The tap interface must be
linked to a bridged interface that provides access to the worker and maybe also
the Internet.

If the network bridge is not configured yet, follow the steps below to create
one.

!!! note
    These instructions are only validated on Debian. You may need to adjust
    them for other distributions.

1. Install required packages:

    ```bash
    sudo apt install iproute2 dnsmasq
    ```

2. Create the bridge:

    ```bash
    sudo ip link add name br-lava type bridge
    sudo ip addr add 192.168.66.1/24 dev br-lava
    sudo ip link set br-lava up
    ```

3. Make the bridge persistent:

    ```bash
    sudo mkdir -p /etc/network/interfaces.d
    sudo tee /etc/network/interfaces.d/br-lava > /dev/null <<'EOF'
    auto br-lava
    iface br-lava inet static
        address 192.168.66.1
        netmask 255.255.255.0
        bridge_ports none
        bridge_stp off
    EOF
    ```

4. Enable DHCP for the bridge:

    ```bash
    sudo tee /etc/dnsmasq.d/br-lava.conf > /dev/null << 'EOF'
    interface=br-lava
    bind-interfaces
    dhcp-range=192.168.66.2,192.168.66.100,12h
    dhcp-option=option:router,192.168.66.1
    dhcp-option=option:dns-server,8.8.8.8
    except-interface=lo
    port=0
    EOF
    sudo systemctl restart dnsmasq
    ```

5. Configure QEMU to use the bridge:

    ```bash
    sudo cp /etc/qemu-ifup /etc/qemu-ifup.original
    sudo tee /etc/qemu-ifup > /dev/null << 'EOF'
    #!/bin/sh -ex

    TAP="$1"
    BRIDGE=br-lava

    ip link set "$TAP" up
    ip link set "$TAP" master "$BRIDGE"
    EOF
    ```

6. Optionally, allow the `br-lava` interface to access the Internet.

    Add NAT masquerading rules:

    ```shell
    sudo iptables -t nat -A POSTROUTING -o <eth0> -j MASQUERADE
    sudo iptables -A FORWARD -i br-lava -o <eth0> -j ACCEPT
    sudo iptables -A FORWARD -i <eth0> -o br-lava -m state --state RELATED,ESTABLISHED -j ACCEPT
    ```

    !!! note
        Replace `<eth0>` with the name of the interface that provides Internet access.

    Make rules persistent:

    ```shell
    sudo apt-get install -y iptables-persistent
    sudo netfilter-persistent save
    ```

--8<-- "refs.txt"

# tmate-based Hacking session

Remote access to a DUT(Device Under Test) in LAVA is supported via a hacking
session. A LAVA hacking session is a special LAVA test job that provides
interactive SSH access to the DUT inside a pre-defined test environment.

Setting up an SSH server on DUT OS could be a pain when package management isn't
supported. Connecting to the DUT SSH server could be another pain when a VPN or
SSH tunnel is needed. This is where tmate comes to the rescue.

[tmate](https://tmate.io/) works by using SSH connections to backend servers.
This means users always connect to the DUT via the backend servers. It also
provides static build for most Linux machine types. Because of its simplicity,
tmate-based LAVA hacking session is recommended. The doc will explain how to use
it in LAVA.

## Prerequisites

To use a tmate-based LAVA hacking session, the following prerequisites must be
in place:

* Both the user and the DUT should have access to `ssh.tmate.io`. It is the CDN
domain name for connecting to the tmate backend servers.
* In the hacking session test definition repository, we included i386, x86_64,
arm32v6, arm32v7, and arm64v8 Linux static tmate builds for convenience. The DUT
you want to use should be one of the supported types.

## Starting a hacking session

A LAVA hacking session is a special LAVA test job with pre-defined deploy, boot,
and test actions. Deploy and boot actions are image and device type relevant. The
test action for starting hacking is always the same.

Refer to the below job example to define your hacking job. The following changes
are required.

* ***Always set job `visibility` to `personal` to avoid leaking SSH access information.***
* ***Provide your SSH public key using the `parameters.PUB_KEY` test parameter. It
is mandatory. tmate session wouldn't be launched if the key is empty.***
* Define a meaningful job name.
* Change the device type to the one that you want to use.
* Define the deploy and boot actions for the device type.
* If needed, modify the job level timeout `timeouts.job.minutes`.
* When needed, modify the test action timeout `test.timeout.minutes`.

```yaml
job_name: BBB NFS boot hacking session
device_type: beaglebone-black

timeouts:
  job:
    minutes: 120
  action:
    minutes: 5
  connection:
    minutes: 2

priority: medium
visibility: personal

actions:
- deploy:
    ...

- boot:
    ...

- test:
    timeout:
      minutes: 90
    definitions:
    - repository: https://gitlab.com/lava/hacking-session.git
      from: git
      path: hacking-session-tmate.yaml
      parameters:
          PUB_KEY: "YOUR_PUB_KEY"
      name: hacking-session
```

## Stopping a hacking session

During a hacking session, your test device can't be used for other jobs. This
will block other users who may want to run tests using the device. So please
stop it just after you finish your work.

* **Cancel** the job using the button on the LAVA job pages. It ends the entire
job immediately.
* **Kill** the tmate process using the command `pkill tmate`. Once the process
is killed, the test will be finished immediately. If no other tests are defined
after the test, the entire job will be finished very quickly.

## Using a custom tmate server

### Host a custom tmate server

#### Create SSH Keys

You need to generate SSH keys for the tmate server. You can use the
`create_keys.sh` script provided by tmate to do this. Run the following commands
on your server:

```bash
mkdir tmate && cd tmate
curl -s -q https://raw.githubusercontent.com/tmate-io/tmate-ssh-server/master/create_keys.sh | bash
```

This will create a keys directory containing the necessary SSH keys. Write down
the RSA and ED25519 public key fingerprints. You can see them in the output.
They are needed in the following LAVA job definition.

#### Start the tmate server Docker container

Use the following command, adjust the `SSH_HOSTNAME` to your server's domain
or IP, and `SSH_PORT_LISTEN` to the port that your server is using.

```bash
docker run -d --name="tmate-server" \
  --cap-add SYS_ADMIN \
  -v $(pwd)/keys:/keys \
  -e SSH_KEYS_PATH=/keys \
  -p 22:22 \
  -e SSH_PORT_LISTEN=22 \
  -e SSH_HOSTNAME=<domain/IP> \
  -e USE_PROXY_PROTOCOL=0 \
  tmate/tmate-ssh-server:prod
```

### Use the custom tmate server

The following parameters are provided to specify custom tmate server.

* `SERVER_HOST`: tmate server domain name or IP.
* `SERVER_PORT`: tmate sever port. Defaults to 22.
* `RAS_FINGERPRINT`: ras public key fingerprint. e.g. SHA256:j9OzsEpYzOOnUAMLdGvMyj2KiiVWCFGLQB+vfIA2rE4
* `ED25519_FINGERPRINT`: ed25519 public key fingerprint. e.g. SHA256:amJtQzWuZ1rSpEIMnE6qeHf7rcTh8eRazOTLVNsuJN4

Refer to the below test action example to use a custom tmate server. You need to
replace the `YOUR_*` strings with the real values.

```yaml
- test:
    timeout:
      minutes: 90
    definitions:
    - repository: https://gitlab.com/lava/hacking-session.git
      from: git
      path: hacking-session-tmate.yaml
      parameters:
          PUB_KEY: "YOUR_PUB_KEY"
          SERVER_HOST: "YOUR_SERVER_HOST"
          SERVER_PORT: "YOUR_SERVER_PORT"
          RAS_FINGERPRINT: "YOUR_RAS_FINGERPRINT"
          ED25519_FINGERPRINT: "YOUR_ED25519_FINGERPRINT"
      name: hacking-session
```

## SSH-based hacking session

In case you still need the pure SSH-based hacking sessions, see
<https://docs.lavasoftware.org/lava/hacking-session.html>.

# Debugging Unexpected Failures

There are many potential reasons why tests running in LAVA might fail, or
produce unexpected behavior. Some of them can be easy to track down, but
others may be more difficult.

## Debugging job failures

The devices, software and test suites can vary massively from one test job to
the next, but nonetheless a few common ideas may help you to work out what's
going wrong.

### Read the logs

This may seem obvious, but it is all too easy to miss real problems in the test
logs! For people not used to diagnosing failures, it is worth reading all the
way from deployment through test device boot to the end of the log file. If a
test job fails to complete successfully, it can often be caused by a problem
much earlier in the test - don't assume that the final few lines of the log file
will tell the whole story:

* A kernel message about failing to load a module or a device failing to
  initialize may be very easily missed in a long stream of kernel boot
  messages. Equally, check that the expected device and module messages are
  present.

* A 'syntax error' from a shell script early in the test run could easily
  propagate errors to later test results.

* A test shell may time out due to an earlier 'Command not found' or 'No such
  file or directory'.

When writing tests, [make things verbose](#be-verbose) to
give yourself more useful logs in case they fail.

### Read the failure comment

Certain operations will cause a failure comment to be automatically added to
the test job.

### Boot failure

If the test system does not (seem to) boot at all, there are a few things worth
checking:

* Did the LAVA dispatcher manage to download and deploy the correct files that
  were specified in the test job? Check that all the files downloaded OK, and
  that any additional work needed on them worked OK (e.g. deploying any
  overlays). You can also specify the expected checksum for each file your test
  job deploys, to guard against download corruption
  (see [deploy checksums](../../technical-references/job-definition/actions/deploy/index.md#checksums)).

* Did you specify the correct files in your test job? It's quite easy to make a
  mistake and use the wrong kernel build, or cut and paste the wrong URL from a
  directory listing.

* (Where needed) Did you use the correct DTB for the test device?
  Symptoms here could include apparent boot failure, as the kernel will either
  not boot or boot but not provide any useful boot messages.

### Fail to find/mount the rootfs

Did the kernel boot OK but then fail to find the root filesystem? This is a
common failure mode, and there are quite a few possible causes. Here are some
of the more common failure cases.

* Again, check that the LAVA dispatcher could download and deploy the correct
  rootfs as specified in the test job.

* Make sure that your kernel has the right drivers available that it needs to
  support the rootfs. Depending on your particular setup, they may need to be
  built-in, or your kernel may need modules to be supplied by an initramfs.
  Typical modules that may be needed will be for disk devices (e.g.
  `sd_mod`), filesystems (e.g. `ext4`) or network interfaces (e.g.
  `e1000e`) if you're using NFS for the rootfs. You should be able to see
  what devices are found by the kernel by reading the boot messages; check that
  the device you are expecting to use does show up there.

* Make sure that you have specified the correct root device on the kernel
  command line, using the `root=` parameter.

* Make sure that the rootfs includes a working `init` program, in the correct
  location. In an initramfs, the default location is `/init`; this can be
  over-ridden on the kernel command line using the `init=` parameter.

### Start simple

This is a common theme throughout the suggested workflow for developing tests
in LAVA. Start with simple test jobs and verify they work as expected. Add
complexity one step at a time, ensuring that each new option or test suite
added behaves as expected. It's much easier to work out what has broken in a
test job if you've made just one small change to a previous test job that
worked fine.

Similarly, if you have a complex test job that's not working correctly then
often the easiest way to find the problem is to simplify the job - remove some
of the complexity and re-test. By removing the complex setup in the test, it
should be possible to identify the cause of the failure.

If there are standard test jobs available for the device type in question, it
might be useful to compare your test job to one of those standard jobs, or even
start with an example jobs and append your test definitions.

### Change one thing at a time

When developing a test, resist the urge to make too many changes at once - test
one element at a time. Avoid changing the deployed files and the test
definition in the same job. When the deployed files change, use an older test
definition and an inline definition to explicitly check for any new support
your test will want to use from those new files. If you change too many
variables at once, it may become impossible to work out what change caused
things to break.

### Be verbose

Especially when developing a new test, add plenty of output to explain what is
going on. If you are starting with a new test device or new boot files, make it
easy to diagnose problems later by adding diagnostics early in the process. In
general, it is much easier to debug a failed test when it is clear about what
it expects to be happening than one which just stops or says "error" in the
middle of a test. The presence of debug information in a known working test job
can be invaluable when checking why a different test job or test case failed.

* If your test configures one or more network interfaces, add the output of
  `ifconfig` or `ip a show` after wards to show that it worked. Consider adding
  calls to `route` or running `cat /etc/resolv.conf` as well.

* If your test uses a specific block device or filesystem, add the output of
  `df` or `mount` to show what devices and filesystems are available.

* Check the kernel support available inside the test image by running commands
  to output details into the test job log file. Once you know which parts of
  `/dev/`, `/proc/` and `/sys` are relevant to the commands used in your test
  definition, use `grep` and `cat` to ensure that details about the available
  support are available when you come to debug the test job.

* Check the available kernel modules using `lsmod` or by outputting the contents
  of `modules.dep`, depending on the configuration of the kernel used in the
  test job.

* Use the metadata to reference the build log and configuration of files used in
  the test job, especially the kernel, initramfs and / or NFS.

If you are writing shell scripts to wrap tests, try using `set -x` - this
will tell the shell to log all lines of your script as it runs them.

For example:

```shell
#!/bin/sh
set -e
set -x
echo "foo"
a=1
if [ $a -eq 1 ]; then
  echo "yes"
fi
```

will give the following output:

```shell
+ echo foo
foo
+ a=1
+ [ 1 -eq 1 ]
+ echo yes
yes
```

### Provide debug data by default

The debug statements used when the test definitions are being developed can be
retained in the final test definitions for later reference. It is much better
to have the debug information available in every test than to have to resubmit
the test job only to find that the problem is intermittent or can only be
reproduced in particular operations.

Debug checks which become common across a range of test job definitions or
which are particularly important for quick triage can also be run as test cases
so that the presence or absence of a critical element of the test shows up as a
pass or fail. Many such checks will need to use scripts to isolate the relevant
information from the available data in `proc` or `dmesg` etc.

## Common pitfalls

There are some common mistakes using LAVA which can cause issues. If you are
experiencing weird problems with your test job, maybe considering these will
help.

### Handling locally built files

Triage will be a lot easier if you follow these guidelines when using files
you have built or modified yourself in LAVA test jobs:

* Use a checksums on all downloaded copies of locally rebuilt files. Frequent
  rebuilds lead to confusion about whether the file you have just built is the
  same file as the test job uses. Even when you are sure you have updated the
  file correctly, there may be caches between the upload location and the worker.

* Always update the metadata every time a local file is rebuilt for use in a
  test job. Include details of what was changed to require the file to be
  rebuilt and when that change was made.

* Always include and update files describing the configuration of the locally
  built file. If building a kernel, enabling `/proc/config.gz` can save large
  amounts of time in triage. Upload the full configuration and build log of all
  files and include the URL to those files in the metadata. It can be very
  difficult for anyone to help you debug your test jobs if the details of how
  the test job files were built is not available. Consider using version control
  software for the test job definitions, configuration files, build logs or
  changelogs to make it easier to track what has changed. When rebuilding local
  files for your test jobs, please remember:
  [Change one thing at a time](#change-one-thing-at-a-time).

* Retain old copies of locally built files, especially if test jobs using those
  files ran successfully.

* Compare your configuration with known working test jobs.

### Avoid using shell operators in YAML lines

Pipes, redirects and nested sub-shells will not work reliably when put directly
into the YAML. While quoting them may work, using a wrapper script is safer:

```shell
#!/bin/sh

set -e
set -x
ifconfig|grep "inet addr"|grep -v "127.0.0.1"|cut -d: -f2|cut -d' ' -f1
```

### Be obsessive about paths and scripts

* If you use `cd` in your YAML, always store where you were and where you end
  up using `pwd`.
* Output your location prior to calling local wrapper scripts.
* Ensure that all wrapper scripts are executable in your VCS.
* Ensure that the relevant interpreter is installed. For example, python is not
  necessarily part of the test image.
* Consider installing `realpath` and use that to debug your directory structure.

## Debugging automation failures

A first step in triage of a test job failure can be to replicate the
steps manually. If this works, then consider the differences between
running a test manually and through automation:

### Infrastructure effects

Some devices have substantial requirements for infrastructure to support the
automation: switchable USB hubs, relays, remote power control, multiple serial
connections, etc.

Triaging of test job failures in one automated system typically needs to be done
on the same instance or, if using another instance, using infrastructure which
is as close as possible to the original instance. It will still be difficult to
identify the problem, especially with intermittent failures, unless key elements
of the test instance can be disabled, replaced or otherwise eliminated from the
test process without generating new failures.

LAVA tries to identify the likely cause of the error and raise the correct
exception. This can be tracked in the `job` test case created by every test job
in the `lava` test suite of the results.

It can be particularly hard to identify the cause of timeouts. Pay close
attention to all devices across the instance to see if a third party element
(like a distribution mirror) is the cause. Look for common factors - both those
which trigger a failure and those which do not.

When investigating intermittent errors, see if the error can be provoked in a
health check and then use looping mode to generate data on how often the error
occurs whilst keeping the test job identical.

If the health check does generate the error, the device will go offline.
Infrastructure problems can be debugged whilst keeping the device(s) offline by
running `lava-run` directly.

### Hidden assumptions in the manual operations

It is common to find that a manual user will know that something is meant to
happen or that an error can simply be ignored and re-tried or simply add an
extra command "just in case". Automation will not do those steps and if the
underlying problem is intermittent, a lot of engineering time will be wasted
trying to work out why. Be meticulous in logging every operation done on the
device to run a test job manually. Pay particular attention to:

* **Changes in prompts** - exactly when and under what circumstances?
* **Hidden time limits** - interrupting a process or waiting for an operation to
  take place. These will need to be carefully written into the device integration.
* **Extra commands** - often not needed every single time but just *sometimes*.
  Define exactly when and prove whether the commands can be safely issued anyway
  or whether there are specific circumstances.

### Differences in input speeds

Manual keyboard entry has noticeable gaps between every key press. Even
the fastest typist will not approach the speed at which a computer can
transmit the same string over a serial connection.

In automation, strings will be sent as quickly as the connection
allows. Some devices may then fail to process the characters correctly.
This might manifest in several ways, including:

* **Missing characters** - `rot` instead of `root` or `erverip`
  instead of `set serverip`.  Often at the start of a line, although
  also includes loss of the newline itself, causing lines to join
  together. `set foo ; set bar` can be changed to `set foo set
  bar`, causing a failure to process `bar`.

* **Reordered characters** - `orot` instead of `root`. This is less
  common than missing characters and can sometimes indicate a hardware
  problem on the device. However, replicating an input speed which is
  closer to human typing can still alleviate the problem.

!!! note
    This is **not** the same as the replacement of characters
    by invalid characters which is a different type of serial
    corruption. If you see ASCII strings being output to the device but
    unprintable or otherwise incorrect characters being received, then
    this could be a hardware problem with the DUT or the connections to
    it. Intermittent single bit flips in the serial data stream are all
    too common.

LAVA supports specifying
[**character delays**](../../technical-references/configuration/device-dictionary.md#character-delays)
in the `deploy`, `boot` and `test` actions to help alleviate these problems.
These are device-specific features, so best controlled in the device configuration.

The `boot` action suffers from this problem more frequently than the `test`
action, typically because `boot` has to interact with processes executed by
firmware or a bootloader where processing can be more limited than in a
POSIX-type test environment.

# Debugging Unexpected Failures

There are many potential reasons why tests running in LAVA might fail, or
produce unexpected behavior. Some of them can be easy to track down, but
others may be more difficult.

## Debugging job failures

The devices, software and test suites can vary massively from one test job to
the next, but nonetheless a few common ideas may help you to work out what's
going wrong.

### Read the logs

This may seem obvious, but it is all too easy to miss real problems in the test
logs! For people not used to diagnosing failures, it is worth reading all the
way from deployment through test device boot to the end of the log file. If a
test job fails to complete successfully, it can often be caused by a problem
much earlier in the test - don't assume that the final few lines of the log file
will tell the whole story:

* A kernel message about failing to load a module or a device failing to
  initialize may be very easily missed in a long stream of kernel boot
  messages. Equally, check that the expected device and module messages are
  present.

* A 'syntax error' from a shell script early in the test run could easily
  propagate errors to later test results.

* A test shell may time out due to an earlier 'Command not found' or 'No such
  file or directory'.

When writing tests, [make things verbose](#make-your-tests-and-setup-verbose) to
give yourself more useful logs in case they fail.

### Read the failure comment

Certain operations will cause a failure comment to be automatically added to
the test job.

### Boot failure

If the test system does not (seem to) boot at all, there are a few things worth
checking:

* Did the LAVA dispatcher manage to download and deploy the correct files that
  were specified in the test job? Check that all the files downloaded OK, and
  that any additional work needed on them worked OK (e.g. deploying any
  overlays). You can also specify the expected checksum for each file your test
  job deploys, to guard against download corruption
  (see [deploy checksums](../../technical-references/job-definition/actions/deploy/index.md#checksums)).

* Did you specify the correct files in your test job? It's quite easy to make a
  mistake and use the wrong kernel build, or cut and paste the wrong URL from a
  directory listing.

* (Where needed) Did you use the correct DTB for the test device?
  Symptoms here could include apparent boot failure, as the kernel will either
  not boot or boot but not provide any useful boot messages.

### Fail to find/mount the rootfs

Did the kernel boot OK but then fail to find the root filesystem? This is a
common failure mode, and there are quite a few possible causes. Here are some
of the more common failure cases.

* Again, check that the LAVA dispatcher could download and deploy the correct
  rootfs as specified in the test job.

* Make sure that your kernel has the right drivers available that it needs to
  support the rootfs. Depending on your particular setup, they may need to be
  built-in, or your kernel may need modules to be supplied by an initramfs.
  Typical modules that may be needed will be for disk devices (e.g.
  `sd_mod`), filesystems (e.g. `ext4`) or network interfaces (e.g.
  `e1000e`) if you're using NFS for the rootfs. You should be able to see
  what devices are found by the kernel by reading the boot messages; check that
  the device you are expecting to use does show up there.

* Make sure that you have specified the correct root device on the kernel
  command line, using the `root=` parameter.

* Make sure that the rootfs includes a working `init` program, in the correct
  location. In an initramfs, the default location is `/init`; this can be
  over-ridden on the kernel command line using the `init=` parameter.

### Start simple

This is a common theme throughout the suggested workflow for developing tests
in LAVA. Start with simple test jobs and verify they work as expected. Add
complexity one step at a time, ensuring that each new option or test suite
added behaves as expected. It's much easier to work out what has broken in a
test job if you've made just one small change to a previous test job that
worked fine.

Similarly, if you have a complex test job that's not working correctly then
often the easiest way to find the problem is to simplify the job - remove some
of the complexity and re-test. By removing the complex setup in the test, it
should be possible to identify the cause of the failure.

If there are standard test jobs available for the device type in question, it
might be useful to compare your test job to one of those standard jobs, or even
start with an example jobs and append your test definitions.

### Change one thing at a time

When developing a test, resist the urge to make too many changes at once - test
one element at a time. Avoid changing the deployed files and the test
definition in the same job. When the deployed files change, use an older test
definition and an inline definition to explicitly check for any new support
your test will want to use from those new files. If you change too many
variables at once, it may become impossible to work out what change caused
things to break.

### Be verbose

Especially when developing a new test, add plenty of output to explain what is
going on. If you are starting with a new test device or new boot files, make it
easy to diagnose problems later by adding diagnostics early in the process. In
general, it is much easier to debug a failed test when it is clear about what
it expects to be happening than one which just stops or says "error" in the
middle of a test. The presence of debug information in a known working test job
can be invaluable when checking why a different test job or test case failed.

* If your test configures one or more network interfaces, add the output of
  `ifconfig` or `ip a show` after wards to show that it worked. Consider adding
  calls to `route` or running `cat /etc/resolv.conf` as well.

* If your test uses a specific block device or filesystem, add the output of
  `df` or `mount` to show what devices and filesystems are available.

* Check the kernel support available inside the test image by running commands
  to output details into the test job log file. Once you know which parts of
  `/dev/`, `/proc/` and `/sys` are relevant to the commands used in your test
  definition, use `grep` and `cat` to ensure that details about the available
  support are available when you come to debug the test job.

* Check the available kernel modules using `lsmod` or by outputting the contents
  of `modules.dep`, depending on the configuration of the kernel used in the
  test job.

* Use the metadata to reference the build log and configuration of files used in
  the test job, especially the kernel, initramfs and / or NFS.

If you are writing shell scripts to wrap tests, try using `set -x` - this
will tell the shell to log all lines of your script as it runs them.

For example:

```shell
#!/bin/sh
set -e
set -x
echo "foo"
a=1
if [ $a -eq 1 ]; then
  echo "yes"
fi
```

will give the following output:

```shell
+ echo foo
foo
+ a=1
+ [ 1 -eq 1 ]
+ echo yes
yes
```

### Provide debug data by default

The debug statements used when the test definitions are being developed can be
retained in the final test definitions for later reference. It is much better
to have the debug information available in every test than to have to resubmit
the test job only to find that the problem is intermittent or can only be
reproduced in particular operations.

Debug checks which become common across a range of test job definitions or
which are particularly important for quick triage can also be run as test cases
so that the presence or absence of a critical element of the test shows up as a
pass or fail. Many such checks will need to use scripts to isolate the relevant
information from the available data in `proc` or `dmesg` etc.

## Common pitfalls

There are some common mistakes using LAVA which can cause issues. If you are
experiencing weird problems with your test job, maybe considering these will
help.

### Handling locally built files

Triage will be a lot easier if you follow these guidelines when using files
you have built or modified yourself in LAVA test jobs:

* Use a checksums on all downloaded copies of locally rebuilt files. Frequent
  rebuilds lead to confusion about whether the file you have just built is the
  same file as the test job uses. Even when you are sure you have updated the
  file correctly, there may be caches between the upload location and the worker.

* Always update the metadata every time a local file is rebuilt for use in a
  test job. Include details of what was changed to require the file to be
  rebuilt and when that change was made.

* Always include and update files describing the configuration of the locally
  built file. If building a kernel, enabling `/proc/config.gz` can save large
  amounts of time in triage. Upload the full configuration and build log of all
  files and include the URL to those files in the metadata. It can be very
  difficult for anyone to help you debug your test jobs if the details of how
  the test job files were built is not available. Consider using version control
  software for the test job definitions, configuration files, build logs or
  changelogs to make it easier to track what has changed. When rebuilding local
  files for your test jobs, please remember:
  [Change one thing at a time](#change-one-thing-at-a-time).

* Retain old copies of locally built files, especially if test jobs using those
  files ran successfully.

* Compare your configuration with known working test jobs.

### Avoid using shell operators in YAML lines

Pipes, redirects and nested sub-shells will not work reliably when put directly
into the YAML. While quoting them may work, using a wrapper script is safer:

```shell
#!/bin/sh

set -e
set -x
ifconfig|grep "inet addr"|grep -v "127.0.0.1"|cut -d: -f2|cut -d' ' -f1
```

### Be obsessive about paths and scripts

* If you use `cd` in your YAML, always store where you were and where you end
  up using `pwd`.
* Output your location prior to calling local wrapper scripts.
* Ensure that all wrapper scripts are executable in your VCS.
* Ensure that the relevant interpreter is installed. For example, python is not
  necessarily part of the test image.
* Consider installing `realpath` and use that to debug your directory structure.

## Debugging automation failures

A first step in triage of a test job failure can be to replicate the
steps manually. If this works, then consider the differences between
running a test manually and through automation:

### Infrastructure effects

Some devices have substantial requirements for infrastructure to support the
automation: switchable USB hubs, relays, remote power control, multiple serial
connections, etc.

Triaging of test job failures in one automated system typically needs to be done
on the same instance or, if using another instance, using infrastructure which
is as close as possible to the original instance. It will still be difficult to
identify the problem, especially with intermittent failures, unless key elements
of the test instance can be disabled, replaced or otherwise eliminated from the
test process without generating new failures.

LAVA tries to identify the likely cause of the error and raise the correct
exception. This can be tracked in the `job` test case created by every test job
in the `lava` test suite of the results.

It can be particularly hard to identify the cause of timeouts. Pay close
attention to all devices across the instance to see if a third party element
(like a distribution mirror) is the cause. Look for common factors - both those
which trigger a failure and those which do not.

When investigating intermittent errors, see if the error can be provoked in a
health check and then use looping mode to generate data on how often the error
occurs whilst keeping the test job identical.

If the health check does generate the error, the device will go offline.
Infrastructure problems can be debugged whilst keeping the device(s) offline by
running `lava-run` directly.

### Hidden assumptions in the manual operations

It is common to find that a manual user will know that something is meant to
happen or that an error can simply be ignored and re-tried or simply add an
extra command "just in case". Automation will not do those steps and if the
underlying problem is intermittent, a lot of engineering time will be wasted
trying to work out why. Be meticulous in logging every operation done on the
device to run a test job manually. Pay particular attention to:

* **Changes in prompts** - exactly when and under what circumstances?
* **Hidden time limits** - interrupting a process or waiting for an operation to
  take place. These will need to be carefully written into the device integration.
* **Extra commands** - often not needed every single time but just *sometimes*.
  Define exactly when and prove whether the commands can be safely issued anyway
  or whether there are specific circumstances.

### Differences in input speeds

Manual keyboard entry has noticeable gaps between every key press. Even
the fastest typist will not approach the speed at which a computer can
transmit the same string over a serial connection.

In automation, strings will be sent as quickly as the connection
allows. Some devices may then fail to process the characters correctly.
This might manifest in several ways, including:

* **Missing characters** - `rot` instead of `root` or `erverip`
  instead of `set serverip`.  Often at the start of a line, although
  also includes loss of the newline itself, causing lines to join
  together. `set foo ; set bar` can be changed to `set foo set
  bar`, causing a failure to process `bar`.

* **Reordered characters** - `orot` instead of `root`. This is less
  common than missing characters and can sometimes indicate a hardware
  problem on the device. However, replicating an input speed which is
  closer to human typing can still alleviate the problem.

!!! note
    This is **not** the same as the replacement of characters
    by invalid characters which is a different type of serial
    corruption. If you see ASCII strings being output to the device but
    unprintable or otherwise incorrect characters being received, then
    this could be a hardware problem with the DUT or the connections to
    it. Intermittent single bit flips in the serial data stream are all
    too common.

LAVA supports specifying
[**character delays**](../../technical-references/configuration/device-dictionary.md#character-delays)
in the `deploy`, `boot` and `test` actions to help alleviate these problems.
These are device-specific features, so best controlled in the device configuration.

The `boot` action suffers from this problem more frequently than the `test`
action, typically because `boot` has to interact with processes executed by
firmware or a bootloader where processing can be more limited than in a
POSIX-type test environment.

# Environment

The job can define environment variables for the device. The variables are made
available in the test shell environment and can be referenced in test scripts then.

Environment variables defined in the job will override environment variables of
the same name defined in the
[device dictionary](../configuration/device-dictionary.md#device).

## Signal node

```yaml
environment:
  FOO: bar
  BAR: baz
```

## Multinode

For multinode jobs, the environment has to be defined for each multinode role
separately:

```yaml
protocols:
  lava-multinode:
    roles:
      node_a:
        environment:
          FOO: bar
      node_b:
        environment:
          BAR: baz
```

# Testing LAVA

## Code style

We use [black] to format the LAVA source code.

The GitLab CI will run black on every merge request.

You can run it locally:

```shell
apt-get install black
black lava/
```

## Unittests

We use [pytest] for the python test harness.

To run the full test suite:

```shell
pytest-3 -v tests/
```

You can execute every tests that are defined in a given file:

```shell
pytest-3 -v tests/lava_dispatcher/test_utils.py
```

You can also execute a specific test:
```shell
pytest-3 -v tests/lava_dispatcher/test_utils.py::test_simple_clone
```

## Static analysis

We use [pylint] and [bandit] for static analysis.

You can run them locally:

```shell
.gitlab-ci/analyze/pylint.sh
bandit -r .
```

## Job schema

If you change the device or job schemas, you will have to ensure that the device and job schemas are valid.

This is checked by:

```shell
```

--8<-- "refs.txt"

# Environment

The job can define environment variables for the device. The variables are made
available in the test shell environment and can be referenced in test scripts then.

Environment variables defined in the job will override environment variables of
the same name defined in the
[device dictionary](../configuration/device-dictionary.md#device).

LAVA also writes a small set of metadata into the same overlay `environment`
file. These `LAVA_*` variables are sourced by the test shell after the job and
device environment, so they take precedence if a name clashes.

Do not confuse these with the bootloader placeholders such as `{LAVA_JOB_ID}`
documented in [boot common](./actions/boot/common.md). Those are substituted
into boot commands on the dispatcher, not exported into the DUT test shell.

## LAVA-injected variables

| Variable | Description |
| --- | --- |
| `LAVA_JOB_ID` | Numeric job ID assigned by the scheduler |
| `LAVA_DISPATCHER_IP` | IP address of the LAVA worker running this job |
| `LAVA_DISPATCHER_PREFIX` | Optional dispatcher `prefix` from the worker config |
| `LAVA_JOB_TAGS` | Comma-separated [job tags](./job.md#tags). Empty when the job requested none |
| `LAVA_DEVICE_HOSTNAME` | Hostname of the assigned device, if present in the device config |
| `LAVA_DEVICE_TYPE` | Device type of the assigned device, if present in the device config |
| `HTTP_CACHE` | Dispatcher `http_url_format_string`, when configured |

## Signal node

```yaml
environment:
  FOO: bar
  BAR: baz
```

## Multinode

For multinode jobs, the environment has to be defined for each multinode role
separately:

```yaml
protocols:
  lava-multinode:
    roles:
      node_a:
        environment:
          FOO: bar
      node_b:
        environment:
          BAR: baz
```

# LAVA Software Community Project Code of Conduct

*Version 1.1 - September 2018*

The LAVA Software Community Project (LSCP), the producer of the LAVA
Software, has adopted a code of conduct for participants. This is
intended to cover interactions and behavior everywhere within the
project, including but not limited to its mailing lists, IRC channels
and other modes of communication.

## Be respectful

In a project of any size, inevitably there will be people with whom
you may disagree, or find it difficult to cooperate. Accept that, but
even so remain respectful. Disagreement is no excuse for poor
behavior or personal attacks, and a community in which people feel
threatened is not a healthy community.

## Assume good faith

LAVA Software Contributors have many ways of reaching our common goal
of reliable and scalable automated testing, which may differ from your
ways. Assume that other people are working towards this goal.

Also note that our Contributors and Maintainers may not be native
English speakers, or may have different cultural backgrounds.

## Be collaborative

LAVA Software is a large and complex project; there is always more to
learn within LAVA and automation. It's good to ask for help when you
need it. Similarly, offers for help should be seen in the context of
our shared goal of improving LAVA.

When you make something for the benefit of the project, be willing to
explain to others how it works, so that they can build on your work to
make it even better.

## Try to be concise

Keep in mind that what you write once will be read by hundreds of
persons. Writing a short email means people can understand the
conversation as efficiently as possible. When a long explanation is
necessary, consider adding a summary.

Try to bring new arguments to a conversation so that each mail adds
something unique to the thread, keeping in mind that the rest of the
thread still contains the other messages with arguments that have
already been made.

Try to stay on topic, especially in discussions that are already fairly
large.

## Be open

Most ways of communication used within LAVA allow for public and
private communication. You should preferably use public methods of
communication for LAVA-related messages, unless posting something
sensitive.

This applies to messages for help or LAVA-related support, too; not
only is a public support request much more likely to result in an
answer to your question, it also makes sure that any inadvertent
mistakes made by people answering your question will be more easily
detected and corrected.

## In case of problems

While this code of conduct should be adhered to by participants, we
recognize that sometimes people may have a bad day, or be unaware of
some of the guidelines in this code of conduct. When that happens, you
may reply to them and point out this code of conduct. Such messages may
be in public or in private, whatever is most appropriate. However,
regardless of whether the message is public or not, it should still
adhere to the relevant parts of this code of conduct; in particular, it
should not be abusive or disrespectful. Assume good faith; it is more
likely that participants are unaware of their bad behavior than that
they are intentionally trying to degrade the quality of the discussion.

Serious or persistent offenders will be temporarily or permanently
banned from communicating through LSCP's systems. Complaints should be
made (in private) to the administrators of the forum in question. To
find contact information for these administrators, please see the
documentation on [getting support within LAVA](../introduction/contact.md).

# AVH

The AVH device type in LAVA refers to
[Arm Virtual Hardware](https://www.arm.com/products/development-tools/simulation/virtual-hardware).

AVH for third-party boards provides cloud-based functionally accurate virtual
models of popular IoT development kits, simulating CPU, peripherals, sensors
and board components. It allows developers to execute the same binaries as on
the real hardware and so leverage the board SDKs and software code examples.

AVH integration in LAVA enables efficient CI workflows for modern agile software
development.

## LAVA worker setup

LAVA uses the [avh-api](https://pypi.org/project/avh-api/) Python library to
communicate with AVH's REST API for managing virtual devices. If you are a LAVA
Docker worker user, the library is already pre-installed in the LAVA dispatcher
Docker image. If your LAVA worker is installed via APT, you will need to
install the `avh-api` library manually using the following commands. This is
mainly because the library isn't available via APT installation yet.

```bash
apt-get install --no-install-recommends --yes python3-pip
python3 -m pip install avh-api==1.0.5 --break-system-packages
```

## AVH API authentication

The following steps are required to authorize LAVA for API access to AVH.

1. Generate an AVH API token.

    Log in to AVH with your account at
    [https://app.avh.corellium.com](https://app.avh.corellium.com). Navigate to
    your profile by clicking on your name in the top right corner. Change to the
    API tab, then click the `GENERATE` button to generate your AVH API token.

2. Add a LAVA remote artifact token.

    Log in to the LAVA server. Click on your name in the top right corner, and
    then click Profile. On the profile page, click the `Create` button in the
    `Remote artifact tokens` section and input `avh_api_token` as the
    `Token name` and the above AVH API token as the `Token string`. Save the
    token.

3. Define secrets block in LAVA job definition.

    ```yaml
    secrets:
        avh_api_token: avh_api_token
    ```

    The secrets block should include the `avh_api_token` key. LAVA dispatcher
    needs it for AVH API authentication. The key value should be the **name** of
    the above LAVA remote artifact token. At run time, the token name will
    be replaced with the token string by the LAVA server. This is mainly for
    hiding the real token in a public LAVA job.

## AVH platform limitations

Concurrent uploads of AVH firmware images can only be 5 at a time. The
limitation is per AVH project. After 5 images are uploaded to the same project,
when another one is uploaded it overrides the first. Trying to run more than 5
AVH LAVA jobs using the same AVH project may encounter an instance creation
error as the image uploaded could be overridden by another one at any time.

Multiple projects are possible but only in enterprise/domain accounts so far.
For these accounts, a lab administrator can extend the base AVH device type by
setting the `avh_project_name` variable to a different AVH project name to
create another device type. The variable defaults to `Default Project`. The
number of devices for these device types should always be equal to or less than
5.

LAVA users can also overwrite the AVH project name in the device dictionary
or in the job definition using the `deploy.options.project_name` key.

## Creating device type

[Create the device type](common.md#create-device-type) using the name **`avh`**.

## Creating device

1. [Add the device](common.md#add-device) using the following settings:
    * **Device Type:** `avh`
    * **Hostname:** A unique name (e.g., `avh-01`)
2. [Add the device configuration](common.md#add-device-configuration).

    For a minimal configuration, simply extend the base template:

    ```jinja
    {% extends "avh.jinja2" %}
    ```

    To set or override the
    [default settings](../../../technical-references/job-definition/actions/deploy/to-avh.md#options),
    define the additional variables below:

    ```jinja
    {% set avh_model = <value> %}
    {% set avh_api_end_point = <value> %}
    {% set avh_project_name = <value> %}
    ```

## Sample job definition

```yaml
device_type: avh
job_name: avh-rpi4b-health-check

timeouts:
  job:
    minutes: 60

priority: medium
visibility: public

secrets:
  avh_api_token: avh_api_token

actions:
- deploy:
    to: avh
    options:
      model: rpi4b
    timeout:
      minutes: 30
    images:
      rootfs:
        url: https://example.com/rpi4b/nand
        format: ext4
        root_partition: 1
      kernel:
        url: https://example.com/rpi4b/kernel
      dtb:
        url: https://example.com/rpi4b/devicetree

- boot:
    method: avh
    timeout:
      minutes: 20
    prompts:
    - "pi@raspberrypi:"
    - "root@raspberrypi:"
    auto_login:
      login_prompt: "login:"
      username: pi
      password_prompt: 'Password:'
      password: raspberry
      login_commands:
      - sudo su

- test:
    timeout:
      minutes: 10
    definitions:
    - from: inline
      repository:
        metadata:
          format: Lava-Test Test Definition 1.0
          name: health checks
        run:
          steps:
          - lava-test-case kernel-info --shell uname -a
          - lava-test-case network-info --shell ip a
      name: health-checks
      path: inline/health-checks.yaml
```

# Testing LAVA

## Code style

We use [black] to format the LAVA source code.

The GitLab CI will run black on every merge request.

You can run it locally:

```shell
apt-get install black
black lava/
```

## Unittests

We use [pytest] for the python test harness.

To run the full test suite:

```shell
python3 -m pytest -v tests/
```

You can execute every tests that are defined in a given file:

```shell
python3 -m pytest -v tests/lava_dispatcher/test_utils.py
```

You can also execute a specific test:
```shell
python3 -m pytest -v tests/lava_dispatcher/test_utils.py::test_simple_clone
```

## Static analysis

We use [pylint] and [bandit] for static analysis.

You can run them locally:

```shell
.gitlab-ci/analyze/pylint.sh
bandit -r .
```

## Job schema

If you change the device or job schemas, you will have to ensure that the device
and job schemas are valid.

This is checked by:

```shell
.gitlab-ci/analyze/schemas.sh
```

--8<-- "refs.txt"

# Backfilling job metadata

The [metadata](../../technical-references/job-metadata.md) of a job definition
is stored with the test job when the job is submitted, so that jobs can be
filtered by it. Jobs that were submitted before the instance was upgraded to
the LAVA version that introduced this feature have no metadata and are not
returned by any metadata filter.

The `backfill-metadata` sub command fills the metadata in from the definition
that is already stored with each job:

!!! example "Backfill job metadata"

    === "docker-compose"
        ```shell
        docker-compose exec lava-server lava-server manage jobs backfill-metadata
        ```

    === "debian"
        ```shell
        lava-server manage jobs backfill-metadata
        ```

Only the jobs with no metadata are considered, so the command can be
interrupted and started again. Jobs whose definition has no metadata are simply
skipped, at the cost of parsing their definition on every run.

The command parses one definition per job, which takes a while on instances
with a large number of jobs. Use `--start-id` and `--end-id` to split the work
into several runs, starting with the most recent jobs:

```shell
lava-server manage jobs backfill-metadata --start-id 900000
lava-server manage jobs backfill-metadata --start-id 800000 --end-id 899999
```

| Option         | Description                                          |
| -------------- | ---------------------------------------------------- |
| `--batch-size` | Number of jobs to update per query, 1000 by default   |
| `--start-id`   | First job id to consider                             |
| `--end-id`     | Last job id to consider                              |
| `--dry-run`    | Do not update the database, only report what would be |
| `--slow`       | Sleep between batches to reduce the load on the database |

# Adding new Actions

The expectation is that new tasks for the LAVA dispatcher will be created
by adding more specialist Actions and organizing the existing Action
classes into a new pipeline for the new task.

## Two-step process

Adding new behavior is a two-step process:

1. **Always** add a new Action, usually with an internal pipeline, to
   implement the new behavior.
2. Add a new Strategy class which creates a suitable pipeline to use
   that Action.

A test Job will consist of multiple strategies, one for each of the
listed *actions* in the YAML file. Typically, this may include a
Deployment strategy, a Boot strategy and a Test strategy. Jobs can have
multiple deployment, boot, or test actions.

## Guidelines for adding new actions

1. A Strategy class is simply a way to `select()` which top level Action
   class is instantiated.

2. Ensure that the `accepts()` routine can uniquely identify this strategy
   without interfering with other strategies.

3. A top level Action class creates an internal pipeline in `populate()`.
   Actions are added to the internal pipeline to do the rest of the work.

4. A top level Action will generally have a basic `run()` function which
   calls `run_actions()` on the internal pipeline.

5. Respect the existing classes — reuse wherever possible and keep all
   classes as pure as possible. There should be one class for each type
   of operation and no more.

6. Expose all configuration in the device or job definitions, not python.
   Extend the device or job schemas if new values are needed.

7. Take care with YAML structure. Always check your YAML changes in the
   [Online YAML Parser](http://yaml-online-parser.appspot.com/?yaml=&type=json).

8. Cherry-pick existing classes alongside new classes to create new
   pipelines and keep all Action classes to a single operation.

9. Code defensively:
    * Check that parameters exist in validation steps.
    * Call `super()` on the base class `validate()` in each `Action.validate()`.
    * Handle missing data in the dynamic context. Don’t assume namespace
      data from earlier actions is always present.
    * Use `cleanup()` and keep actions idempotent.

## Adding retry actions

For a `RetryAction` to validate, the `RetryAction` subclass must be a
wrapper class around a new pipeline to allow the `RetryAction.run()`
function to handle all the retry functionality in one place.

An Action which needs to support `failure_retry` or which wants to use
`RetryAction` support internally, needs a new class added which derives
from `RetryAction`, sets a useful name, summary and description and
defines a `populate()` function which creates the pipeline.

## Always add unit tests for new actions

Wherever a new class is added, that new class should be tested. Always
create a new file in the tests directory for new functionality.

All unit tests need to be in a file with the `test_` prefix and add a
new YAML file to the `sample_jobs` so that the strategies to select the
new code can be tested.

To run a single unit-test, for example `test_pipeline` in a class called
`TestFlasher` in a file called `test_flasher.py`, use:

```shell
pytest tests/lava_dispatcher/test_flasher.py::TestFlasher::test_pipeline
```

## Coding conventions

When adding or modifying `run`, `validate`, `populate` or `cleanup`
functions, always ensure that `super` is called appropriately:

```python
super().validate()
```

When adding or modifying `run` functions in subclasses of `Action`,
always ensure that each return point returns the `connection` object:

```python
connection = super().run(connection, max_end_time)
# Use the connection or initiate a new one.
return connection
```

When adding new classes, use hyphens (`-`) as separators in `self.name`,
not underscores (`_`). Action names need to all be lowercase:

```python
self.name = 'do-something-at-runtime'
```

Use namespaces for all dynamic data. Parameters of actions are
**immutable**. Use the namespace functions when an action needs to store
dynamic data:

```python
self.set_namespace_data(
    action="run-fvp",
    label="fvp",
    key="serial_port",
    value=serial_port,
)
```

The data can be retrieved in the following actions in the pipeline:

```python
serial_port = self.get_namespace_data(
            action="run-fvp", label="fvp", key="serial_port"
        )
```

# SSH

The `ssh` boot method connects the LAVA worker to an already running device over
SSH. The device must be booted up and reachable at the address configured in
the [SSH device dictionary](../../../configuration/device-dictionary.md#ssh).

```yaml
- boot:
    method: ssh
    prompts:
    - 'root@device:~#'
```

The boot action:

1. Opens an SSH session from the worker to the device.
2. Copies the LAVA test overlay to the device using SCP.
3. Waits for the expected shell prompt.
4. Unpacks the overlay on the device for test execution.

## prompts

See [prompts](./common.md#prompts)

## Job example

```yaml
job_name: SSH job example
device_type: ssh
visibility: public

timeouts:
  job:
    minutes: 5
  action:
    minutes: 2

actions:
- deploy:
    to: ssh
    os: debian

- boot:
    method: ssh
    prompts:
    - 'root@device:~#'

- test:
    definitions:
    - from: inline
      repository:
        metadata:
          format: Lava-Test Test Definition 1.0
          name: test-definition-example
        run:
          steps:
          - lava-test-case run-uname-a --shell uname -a
          - lava-test-case check-os-id --shell 'cat /etc/os-release | grep "ID=debian"'
      path: inline/test-definition-example.yaml
      name: test-suite-example
```

# PostgreSQL

Store the different objects like Device, DeviceType, TestJob, ...

## Command line

This service is usually only manage via systemd.

## Service

The systemd service is called `postgresql`.

## Logs

The logs are stored under `/var/log/posqtgresql/`

The log rotation is configured in `/etc/logrotate.d/postgresql-common`.

# Authorization

## Private instance

By default, device types, devices and test jobs are publicly visible meaning
that a non-authenticated users can view everything. This does not apply to
submit and change permissions.

It is possible to require all users to login before accessing any page outside
the home page, documentation pages and the login page itself by setting the
`REQUIRE_LOGIN: true` in any YAML configuration file under
`/etc/lava-server/settings.d/*.yaml`.

For a public instance, the access can be controlled globally or per-object.

## Global authorization

LAVA's global authorization is built on top of
[Django's permission system](https://docs.djangoproject.com/en/stable/topics/auth/default/#topic-authorization).
A user can hold a global permission for any system entity. Global permissions
are checked first and take precedence over [per-object permissions](#per-object-authorization).
They are managed through the Django admin user settings.

## Per-object authorization

LAVA per-object authorization is used to apply fine-grained access to a specific
objects in the system. For example, it can be used to hide a specific device
(and all jobs on that device) from the public.

The per-object system is **group-based** and applies to:

- **Device Type**
- **Device**

!!! note "Test Job"
    Test jobs are **not** directly controlled through per-object group
    permissions. Test job visibility is determined by inheritance from the
    device or device type it belongs to, along with the job-level
    [`viewing_groups`](#viewing_groups) and [`is_public`](#is_public) fields.

The supported per-object permissions are:

- **view**
- **submit**
- **change**

!!! important
    LAVA per-object authorization works in an **inverse** manner. This means
    that if no permissions are assigned to an object, that object is considered
    **unrestricted** and is visible to everyone. Once a permission is assigned
    to an object for a specific group, only members of that group (and
    superusers) are granted the corresponding access.

## Permission inheritance

As we already covered that the object is permission restricted if it has any
permission assigned for a specific group and permission unrestricted otherwise.

Since every device belongs to a device type, and every test job is associated
with either a device (actual) or a device type (requested), permissions cascade
from parent to child when the child object is unrestricted.

For example, the device `qemu01` is permission unrestricted but the `qemu`
device type has a permission to allow group `lkft` to view it. Only users from
this particular group will be able to see device `qemu01` even if it’s not
permission restricted because the underlying device type has the restriction.
The same goes for test jobs. If the test job is not permission restricted it
will be visible only if device and device type it belongs to are unrestricted,
or if the user belongs to one of the groups the view permission applies to.

This means that setting a restriction on a device type automatically restricts
all devices of that type (and all jobs on those devices) without requiring
individual configuration.

This particular behavior allows admin to set the permissions on higher level,
and it will be applied to all the lower level objects. For example, if you set
the view permission for aforementioned `lkft` group for the device type, all the
devices and test job will be automatically hidden for non-lkft users.

Conversely, a more specific permission at the lower level takes precedence.
For example, if a group has direct access to a device, that access applies
regardless of whether the same group has any permission on the parent device
type.

## Test job visibility

Test job visibility has two additional fields that override the inherited
device/device-type permissions.

### `viewing_groups`

If `viewing_groups` is set on a test job, the user must belong to **all** the
specified groups to view the job. This field has higher priority than per-object
device/device-type authorization and can be set in the job definition by setting
job [visibility](./job-definition/job.md#visibility) to a list of groups.

### `is_public`

Test job field `is_public` can be used to completely hide job from public. If
set to `False`, only submitter, superusers and users belonging to
`viewing_groups` field (if set) will be able to view the test job. If set to
`True`, regular per-object authorization will be applied. This field can also be
set in the job definition by setting job
[visibility](./job-definition/job.md#visibility) to `personal`.

## Visibility decision trees

### Device type

![Device type visibility](./images/device-type-decision-tree.png)

### Device

![Device visibility](./images/device-decision-tree.png)

### Test job

![Test job visibility](./images/test-job-decision-tree.png)

## Examples

For the sake of simplicity, let's say we have `device-type1`, `device1` (of
`device-type1`), and groups `group1` and `group2`.

1. No per-object permissions set on `device-type1` nor `device1`: All authenticated
   as well as anonymous users are able to access `device-type1`, `device1` page
   and all test jobs view pages (e.g., `/scheduler/device_type/qemu`) as well as
   plain logs etc. All authenticated users are able to submit jobs to `device1`.
   No anonymous users can submit jobs (ever).
2. `device1` has a per-object permission `Can submit to device` set to `group1`:
   Now only users from `group1` are able to submit jobs to `device1`. All
   authenticated and anonymous users are still able to **view** `device1` and
   all jobs running on it since there's no restriction for **view** permission.
3. `group1` has a per-object permission **view** for `device-type1` and `device1`
   has no per-object permissions set: only users from `group1` can **view** the
   `device-type1`, `device1` and all associated test jobs. No-one else.
4. `group1` has a per-object permission **view** for `device-type1` and `group2`
    has a per-object permission **view** for `device1`: users from `group1` are
    not able to **view** `device1` (nor any test jobs running on it) but they
    will be able to access `device-type1` details page. Users from `group1` will
    be able to **view** other devices (if any) of the `device-type1`, but only
    if those devices have no **view** permissions set for them (or they belong
    to some other groups that are included in those permissions). Users from
    `group2` will be able to **view** the `device1` and all the test jobs running
    on it but will not be able to access the `device-type1` detail page.

## See also

[Configure permissions](../admin/basic-tutorials/instance/permissions.md)

# ISO Installer

Provides QEMU operations using an operating system installer into a new image
and then boot into the installed image to run the specified test definitions.

!!! note
    Currently only tested with Debian installer and the `qemu-iso` boot method.

```yaml
- deploy:
    to: iso-installer
    os: debian_installer
    images:
      iso:
        url: https://cdimage.debian.org/debian-cd/13.3.0/amd64/iso-cd/debian-13.3.0-amd64-netinst.iso
        image_arg: -drive file={iso},index=2,media=cdrom,readonly=on
      preseed:
        url: https://storage.lavacloud.io/health-checks/qemu/iso/preseed.cfg
    iso:
      kernel: /install.amd/vmlinuz
      initrd: /install.amd/initrd.gz
      installation_size: 3G
    timeout:
      minutes: 30
```

## images

The ISO and the preseed file can be tightly coupled but point releases of a
stable release will typically continue to work with existing preseed files.
LAVA tests of the installer tend to only install the base system in order to
test kernel functionality rather than the operating system itself.

### iso

The operating system installer ISO image. This ISO should contain the installer
that will be used to install the operating system onto a new disk image.

#### image_arg

QEMU requires `,media=cdrom,readonly=on` to handle the ISO correctly.

### preseed

Debian Installer can retrieve settings from a `preseed` file to allow the
installation to proceed without prompting for information.

## iso

### kernel

Absolute path to the kernel within the installer ISO. LAVA extracts the kernel
from the ISO so it can be booted directly by QEMU with custom kernel
command-line options required for automated installation.

### initrd

Absolute path to the initial ramdisk within the installer ISO. LAVA extracts
the initrd from the ISO so it can be passed directly to QEMU, enabling
custom boot configuration for the installer.

!!! note
    Both `kernel` and `initrd` paths must:

    - Start with `/` (absolute path from the ISO root)
    - Be unique within the ISO

### installation_size

Size of the empty image to be provided to the installer. Typically, a maximum of
6G. Use megabytes for a smaller image, although ~3G is likely to be the
smallest practical size for a recent Debian installer.

# pyOCD

The `pyocd` boot method is used to flash and boot devices using the
[pyOCD](https://pyocd.io/) tool. It takes no parameters.

```yaml
- boot:
    method: pyocd
    timeout:
      minutes: 10
```

The boot method requires configuration on the device type level. The supported
configurations are:

* `command` - executable command to invoke
* `options` - list of options to pass to the executable
* `connect_before_flash` - if `true`, connect to device before running pyocd
flash command, otherwise after the command (default is `false`)

# Welcome to LAVA

**LAVA** stand for **L**inaro **A**utomated **V**alidation **A**rchitecture.

LAVA is a *continuous integration* system for deploying operating systems onto
physical and virtual hardware for running tests.
Tests can be simple boot testing, bootloader testing and system level testing,
although extra hardware may be required for some system tests.
Results are tracked over time and data can be exported for further analysis.

# Documentation layout

The documentation is made of five sections:

* [Introduction](introduction/concepts.md)
* [User guide](user/first-steps.md)
* [Admin guide](admin/first-steps.md)
* [Developer guide](developer/first-steps.md)
* [Technical references](technical-references/architecture.md)

The guides are targeting specific audiences: Users, Admins or Developers.

# Concepts

## What is LAVA?

LAVA is the **L**inaro **A**utomated **V**alidation **A**rchitecture.

LAVA is a continuous integration system for deploying operating systems onto
physical and virtual hardware for running tests. Tests can be simple boot
testing, bootloader testing and system level testing, although extra hardware
may be required for some system tests. Results are tracked over time and data
can be exported for further analysis.

LAVA is a collection of participating components in an evolving architecture.
LAVA aims to make systematic, automatic and manual quality control more
approachable for projects of all sizes.

LAVA is designed for validation during development - testing whether the code
that engineers are producing "works", in whatever sense that means. Depending
on context, this could be many things, for example:

* Testing whether changes in a bootloader initialize and boot the board
* Testing whether changes in the Linux kernel compile and boot
* Testing whether a kernel scheduler change reduces power consumption
* Testing whether the code produced by GCC is smaller or faster

LAVA is good for automated validation. LAVA tests the Linux kernel on a range
of supported boards every day. LAVA tests proposed android changes in Gerrit
before they are landed, and does the same for other projects like GCC.

LAVA is good for providing developers with the ability to run customized tests
on a variety of different types of hardware, some of which may be difficult to
obtain or integrate. LAVA has great support for emulation platforms like AVH,
FVP and QEMU, and is best at providing test support for real hardware devices.

LAVA is principally aimed at testing changes made by developers across
multiple hardware platforms to aid portability and encourage multi-platform
development.

## What is LAVA **not**?

LAVA is not a set of tests. It is the infrastructure to enable users to run
their own tests. LAVA concentrates on providing a range of deployment methods
and boot methods. Once the login is complete, the test consists of whatever
scripts the test writer chooses to execute in that environment.

LAVA is not a test lab. It is the software that can be used in a test lab to
manage and control test devices.

LAVA is not a complete CI system. It is the software that can form part of a
CI loop. LAVA supports data extraction to make it easier to produce a frontend
which is directly relevant to particular groups of developers.

LAVA is not a build farm. Other tools need to be used to prepare binaries
which can be deployed to the devices using LAVA.

## LAVA in CI

Continuous Integration (CI) is a development model where developers make changes
which are continuously integrated into the main development tree. Testing of
that tree allows developers to know if their changes have introduced new bugs,
regressions, or exposed side effects in other systems.

LAVA is focused on bootloader, kernel and system validation as part of a CI loop
and provides a way to test on a wide variety of different hardware to aid
portability and develop multi-platform systems.

LAVA fits into a CI loop where code changes trigger builds that are tested on
virtual and real hardware. LAVA handles the testing and validation phase,
providing results that feed back to developers to close the loop.

```mermaid
graph LR
    A[[Developer<br>Workflow]]
    B[[Build<br>Farm]]
    C([In-build Tests])
    D[(Build<br>Artifacts)]
    E((LAVA Jobs))
    F[(Results &<br>Logs)]
    G[[Reporting]]

    A --> B
    B --> C
    B --> D
    C --> F
    D --> E
    E --> F
    F --> G
    G --> A

    classDef greyNode fill:#F5F5F5,stroke:#333,stroke-width:1px;
    classDef whiteNode fill:#ffffff,stroke:#333,stroke-width:1px;
    classDef greenNode fill:#9BCB63,stroke:#5E8F2A,stroke-width:1.1px;

    class A,B,C,G whiteNode;
    class D,F greyNode;
    class E greenNode;
```

LAVA's role in this loop is to:

1. **accept** job submission
2. **deploy** build artifacts onto physical and virtual DUTs
3. **boot** DUT to expected environment
4. **execute** predefined tests
5. **notify** reporting system to pull job and test results

For CI to work effectively, tests need to be relevant, reproducible, and timely.
Test jobs submitted to LAVA should follow a basic rule:
**test one element at a time**. If a test changes the kernel, deployment method,
and OS simultaneously, failures cannot be attributed to a specific cause.

# Developer setup

In this document, we assume that you are developing on a Debian system. We also
assume that the version is Buster (or any more recent versions).

??? note "Other distributions"
    Even if we assume the use of Debian, you can also use any non-Debian
    distributions. Just adapt the steps above to match your distribution.

We currently recommend two methods for developing LAVA.

* [from the Debian packages](#from-debian-packages)
* [from sources](#from-sources)

Both methods are currently used by LAVA core developers. Choosing one or the other is a matter of taste.

## From Debian packages

Install the Debian packages:

```shell
apt-get install git postgresql
apt-get install lava-coordinator lava-dispatcher lava-server
```

Fetch the sources somewhere in your home directory.

```shell
cd ~/
git clone https://git.lavasoftware.org/lava/lava
cd ~/lava/
```

Remove the lava binaries and packages from the system directories:

```shell
# python libraries
rm -rf /usr/lib/python3/dist-packages/lava/
rm -rf /usr/lib/python3/dist-packages/lava_common/
rm -rf /usr/lib/python3/dist-packages/lava_dispatcher/
rm -rf /usr/lib/python3/dist-packages/lava_dispatcher_host/
rm -rf /usr/lib/python3/dist-packages/lava_rest_app/
rm -rf /usr/lib/python3/dist-packages/lava_results_app/
rm -rf /usr/lib/python3/dist-packages/lava_scheduler_app/
rm -rf /usr/lib/python3/dist-packages/lava_server/
rm -rf /usr/lib/python3/dist-packages/linaro_django_xmlrpc/

# binaries
rm -f /usr/bin/lava-coordinator
rm -f /usr/bin/lava-dispatcher-host
rm -f /usr/bin/lava-run
rm -f /usr/bin/lava-server
rm -f /usr/bin/lava-worker
```

Add symbolic links to your local clone:

```shell
# python libraries
cd /usr/lib/python3/dist-packages/
ln -s ~/lava/lava/ .
ln -s ~/lava/lava_common/ .
ln -s ~/lava/lava_dispatcher/ .
ln -s ~/lava/lava_dispatcher_host/ .
ln -s ~/lava/lava_rest_app/ .
ln -s ~/lava/lava_results_app/ .
ln -s ~/lava/lava_scheduler_app/ .
ln -s ~/lava/lava_server/ .
ln -s ~/lava/linaro_django_xmlrpc/ .

# binaries
cd /usr/bin/
ln -s ~/lava/lava/coordinator/lava-coordinator .
ln -s ~/lava/lava/dispatcher/lava-run .
ln -s ~/lava/lava/dispatcher/lava-worker .
ln -s ~/lava/lava_dispatcher_host/lava-dispatcher-host .
ln -s ~/lava/manage.py lava-server
```

Install the lava-dispatcher-host udev rules:

```shell
lava-dispatcher-host rules install
```

Restart the services:

```shell
service lava-coordinator restart
service lava-publisher restart
service lava-scheduler restart
service lava-server-gunicorn restart
service lava-worker restart
```

LAVA is now accessible at [http://localhost/](http://localhost/)

## From sources

Install basic tools:

```shell
apt-get install black git nfs-kernel-server \
    postgresql postgresql-client \
    python3 python3-yaml python3-django-extensions python3-werkzeug \
    ruby-foreman
```

Fetch the sources somewhere in your home directory.

```shell
cd ~/
git clone https://git.lavasoftware.org/lava/lava
cd ~/lava/
```

Install `lava-dispatcher` and  `lava-server` dependencies:

```shell
apt-get install \
    $(python3 share/requires.py  -p lava-dispatcher -d debian -s buster -n) \
    $(python3 share/requires.py  -p lava-dispatcher -d debian -s buster -n -u) \
    $(python3 share/requires.py  -p lava-server -d debian -s buster -n) \
    $(python3 share/requires.py  -p lava-server -d debian -s buster -n -u)
```

Create the test database:

```shell
sudo -u postgres psql -c "CREATE ROLE devel NOSUPERUSER CREATEDB NOCREATEROLE INHERIT LOGIN ENCRYPTED PASSWORD 'devel'"
sudo -u postgres psql -c "CREATE DATABASE devel OWNER devel"
```

Apply the django migrations

```shell
python3 lava_server/manage.py migrate
```

Install the lava-dispatcher-host udev rules

```shell
$(pwd)/lava_dispatcher_host/lava-dispatcher-host rules install
```

The above will make the installed udev rules call the lava-dispatcher-host
program inside your source directory. If you move your source directory to a
different location, you will have to install the rules again.

Start the services

```shell
foreman start
```

LAVA is now accessible at [http://localhost:8000/](http://localhost:8000/)

# Docker

## Docker deploy

Download the given docker image:

```yaml
actions:
- deploy:
    to: docker
    os: debian
    image: debian:sid
```

## Local images

The deploy action can use local docker images, just add `local: true`:

```yaml
actions:
- deploy:
    to: docker
    os: debian
    image:
      name: debian:sid
      local: true
```

# Docker

## Docker deploy

Download the given docker image:

```yaml
actions:
- deploy:
    to: docker
    os: debian
    image: debian:sid
```

## Local images

The deploy action can use local docker images, just add `local: true`:

```yaml
actions:
- deploy:
    to: docker
    os: debian
    image:
      name: debian:sid
      local: true
```

## Docker login

To pull images from a password protected registry add a login section
with the username and password.

```yaml
actions:
- deploy:
    to: docker
    os: debian
    image:
      name: debian:sid
      login:
        user: foobar
        password: my_password
```

**Warning**: The pulled image will be cached on the local daemon and will
be available to other jobs running on same worker.

# Minimal

The `minimal` method is used to power-on the DUT and to let the DUT boot without
any interaction.

```yaml
- boot
    method: minimal
    prompts:
    - 'root@debian:~#'
```

## reset

By default, LAVA will reset the board power when executing this action. You can
skip this step by setting `reset: false`.

```yaml
- boot
    method: minimal
    reset: false
```

This can be useful when testing bootloader in interactive tests and then
booting to the OS without resetting the device.

## pre_power_command

`pre_power_command` defined in the device dictionary can be used to activate or
deactivate external hardware before applying power to the device. Set
`pre_power_command: true` to execute the command.

```yaml
- boot
    method: minimal
    pre_power_command: true
```

## pre_os_command

`pre_os_command` defined in the device dictionary can be used to activate or
deactivate external hardware before booting the OS. Set `pre_os_command: true`
to execute the command.

```yaml
- boot
    method: minimal
    pre_os_command: true
```

# IoT

LAVA supports testing IoT and embedded devices using a set of deployment and
boot methods designed for resource-constrained hardware.

## Supported boot methods

- [DFU](../../technical-references/job-definition/actions/boot/method-dfu.md)
- [Musca](../../technical-references/job-definition/actions/boot/method-musca.md)
- [CMSIS-DAP](../../technical-references/job-definition/actions/boot/method-cmsis-dap.md)
- [JLink](../../technical-references/job-definition/actions/boot/method-jlink.md)
- [OpenOCD](../../technical-references/job-definition/actions/boot/method-openocd.md)
- [pyOCD](../../technical-references/job-definition/actions/boot/method-pyocd.md)

## Test monitors

Many IoT devices do not have a login shell. Instead, the firmware produces
structured output on the serial console. LAVA test
[monitors](../../technical-references/job-definition/actions/test.md#monitors)
action can be used to parse these output against expected patterns to determine
`pass/fail/skip/unknown` results without requiring an interactive shell.

## Example jobs

- [TF-M test on musca-b](../../technical-references/job-definition/actions/boot/method-musca.md#job-example)
- [Zephyr test on arduino101](../../technical-references/job-definition/actions/boot/method-dfu.md#hardware-dfu)

# Fastboot

The `fastboot` boot method boots a device that has been deployed using the
[Fastboot](https://en.wikipedia.org/wiki/Fastboot) protocol.

```yaml
- boot:
    method: fastboot
    docker:
      image: linaro/noble-adb-fastboot
      local: true
    prompts:
    - 'root@linaro-developer:~#'
```

For boot sequence,see [fastboot_sequence](../../../../technical-references/configuration/device-dictionary.md#fastboot_sequence)

## docker

See [docker](../deploy/to-fastboot.md#docker)

## commands

Some test jobs need to send additional `fastboot` commands before rebooting the
device. For example, when images are deployed to A/B partition slots, a command
is needed to activate the correct slot before boot.

If the deployment action used `boot_a` as the partition label instead of `boot`,
then the following command ensures the device boots from `boot_a` instead of
`boot_b` (which may contain a stale deployment or be empty):

```yaml
- boot:
    method: fastboot
    commands:
    - --set-active=a
```

Each entry in `commands` is passed as an argument to the `fastboot` binary and
issued by LAVA sequentially.

## prompts

See [prompts](./common.md#prompts)

# GRUB EFI

The `grub-efi` boot method is used to boot a device via GRUB loaded from UEFI.

```yaml
- boot:
    method: grub-efi
    commands: nfs
    auto_login:
      login_prompt: "login:"
      username: root
    prompts:
    - 'root@stretch:'
```

!!! note
    In most cases, starting GRUB from UEFI requires using the
    [uefi-menu](./method-uefi-menu.md) method as well. Admins can refer to the
    [`mustang-grub-efi.jinja2`](https://gitlab.com/lava/lava/-/blob/master/etc/dispatcher-config/device-types/mustang-grub-efi.jinja2)
    device type template for an example of how to make selections from a UEFI
    menu to load GRUB.

## Parameters

This boot method shares the same job parameter supported by the
[grub](./method-grub.md) boot method.

## Example job

```yaml
job_name: mustang grub-efi example job
device_type: mustang

priority: medium
visibility: public

timeouts:
  job:
    minutes: 15
  connection:
    minutes: 2

actions:
- deploy:
    to: tftp
    kernel:
      url: https://example.com/vmlinuz-4.9.0-2-arm64
      type: zimage
    ramdisk:
      url: https://example.com/initrd.img-4.9.0-2-arm64
      compression: gz
    modules:
      url: https://example.com/modules.tar.gz
      compression: gz
    nfsrootfs:
      url: https://example.com/stretch-arm64-nfs.tar.gz
      compression: gz
    timeout:
      minutes: 5

- boot:
    method: grub-efi
    commands: nfs
    auto_login:
      login_prompt: "login:"
      username: root
    prompts:
    - 'root@stretch:'
    timeout:
      minutes: 5

- test:
    definitions:
    - repository: https://github.com/Linaro/test-definitions
      from: git
      path: automated/linux/smoke/smoke.yaml
      name: smoke-tests
    timeout:
      minutes: 5
```

--8<-- "refs.txt"

---
hide:
  - navigation
---

# Welcome to LAVA

**LAVA** stands for **L**inaro **A**utomated **V**alidation **A**rchitecture.

LAVA is a *continuous integration* system for deploying operating systems onto
physical and virtual hardware for running tests.
Tests can be simple boot testing, bootloader testing and system level testing,
although extra hardware may be required for some system tests.
Results are tracked over time and data can be exported for further analysis.

## Documentation layout

The documentation is made of five sections:

* [Introduction](introduction/concepts.md)
* [User guide](user/first-steps.md)
* [Admin guide](admin/first-steps.md)
* [Developer guide](developer/first-steps.md)
* [Technical references](technical-references/architecture.md)

The guides are targeting specific audiences: Users, Admins or Developers.
Depending on how you expect to use LAVA, the **First steps** of each section
describe several useful starting points.

## Getting started

New to LAVA? Start with the
[Getting started](introduction/getting-started.md) guide.

# LAVA server gunicorn

`lava-server-gunicorn` is the web server that will handle http requests coming
from users.

This process will not serve static files that are handled directly by
[apache2](./apache2.md).

## Command line

Run `gunicorn lava-server.wsgi`

## Service

The systemd service is called `lava-server-gunicorn`.

## Dependencies

lava-server-gunicorn should be able to:

* connect to the [postgresql](./postgresql.md) database
* open a socket on port `8000`

## Configuration

Daemon start options:

* `/etc/default/lava-server-gunicorn`
* `/etc/lava-server/lava-server-gunicorn`

Django configuration:

* `/etc/lava-server/instance.conf`
* `/etc/lava-server/settings.conf`

## Logs

The logs are stored in `/var/log/lava-server/gunicorn.log` and
`/var/log/lava-server/django.log`.

The log rotation is configured in `/etc/logrotate.d/lava-server-gunicorn-log`
and `/etc/logrotate.d/django-log`.

## Security

This process should be always behind a reverse proxy like
[apache2](./apache2.md).

# Developer setup

In this document, we assume that you are developing on a Debian system. We also
assume that the version is Trixie (or any more recent versions).

??? note "Other distributions"
    Even if we assume the use of Debian, you can also use any non-Debian
    distributions. Just adapt the steps above to match your distribution.

We currently recommend three methods for developing LAVA.

* [using uv and pyproject.toml](#using-uv) (recommended)
* [from the Debian packages](#from-debian-packages)
* [from sources](#from-sources)

## Using uv

[uv](https://docs.astral.sh/uv/) is the recommended way to set up a
development environment. LAVA provides a `pyproject.toml` with optional
dependency groups for each component.

Available dependency groups:

* `dispatcher` -- dependencies for `lava-dispatcher`
* `server` -- dependencies for `lava-server`
* `dispatcher-host` -- dependencies for `lava-dispatcher-host`
* `coordinator` -- dependencies for `lava-coordinator`
* `full` -- all of the above
* `test` -- test dependencies (pytest, coverage, etc.)
* `dev` -- development tools (ruff, pylint, etc.)
* `docs` -- documentation build dependencies (sphinx)

Fetch the sources and install dependencies:

```shell
git clone https://gitlab.com/lava/lava
cd lava/
uv sync --all-extras
```

### Running the development server

LAVA uses PostgreSQL in production, but for local development you can
use SQLite:

```shell
export DATABASE_URL="sqlite:////tmp/lava.sqlite3"
uv run --extra server manage.py migrate
uv run --extra server manage.py collectstatic --noinput
uv run --extra server manage.py createsuperuser
uv run --extra server manage.py runserver
```

LAVA is now accessible at [http://localhost:8000/](http://localhost:8000/)

### Running tests and linters

```shell
uv run --frozen --all-extras -- pre-commit run --show-diff-on-failure --color=always --all-files
uv run python -m pytest tests/
```

!!! note
    Production deployments use Debian packages, not pip or uv. The
    `pyproject.toml` dependency groups are provided for development
    convenience only. System-level files (systemd units, configuration
    in `/etc`, etc.) are handled by the Debian packaging and `setup.py`,
    not by `pyproject.toml`.

### Using pip

Alternatively, you can use pip in a virtual environment:

```shell
python3 -m venv .venv
source .venv/bin/activate
pip install -e ".[dev]"
```

## From Debian packages

Install the Debian packages:

```shell
apt-get install git postgresql
apt-get install lava-coordinator lava-dispatcher lava-server
```

Fetch the sources somewhere in your home directory.

```shell
cd ~/
git clone https://gitlab.com/lava/lava
cd ~/lava/
```

Remove the lava binaries and packages from the system directories:

```shell
# python libraries
rm -rf /usr/lib/python3/dist-packages/lava/
rm -rf /usr/lib/python3/dist-packages/lava_common/
rm -rf /usr/lib/python3/dist-packages/lava_dispatcher/
rm -rf /usr/lib/python3/dist-packages/lava_dispatcher_host/
rm -rf /usr/lib/python3/dist-packages/lava_rest_app/
rm -rf /usr/lib/python3/dist-packages/lava_results_app/
rm -rf /usr/lib/python3/dist-packages/lava_scheduler_app/
rm -rf /usr/lib/python3/dist-packages/lava_server/
rm -rf /usr/lib/python3/dist-packages/linaro_django_xmlrpc/

# binaries
rm -f /usr/bin/lava-coordinator
rm -f /usr/bin/lava-dispatcher-host
rm -f /usr/bin/lava-run
rm -f /usr/bin/lava-server
rm -f /usr/bin/lava-worker
```

Add symbolic links to your local clone:

```shell
# python libraries
cd /usr/lib/python3/dist-packages/
ln -s ~/lava/lava/ .
ln -s ~/lava/lava_common/ .
ln -s ~/lava/lava_dispatcher/ .
ln -s ~/lava/lava_dispatcher_host/ .
ln -s ~/lava/lava_rest_app/ .
ln -s ~/lava/lava_results_app/ .
ln -s ~/lava/lava_scheduler_app/ .
ln -s ~/lava/lava_server/ .
ln -s ~/lava/linaro_django_xmlrpc/ .

# binaries
cd /usr/bin/
ln -s ~/lava/lava/coordinator/lava-coordinator .
ln -s ~/lava/lava/dispatcher/lava-run .
ln -s ~/lava/lava/dispatcher/lava-worker .
ln -s ~/lava/lava_dispatcher_host/lava-dispatcher-host .
ln -s ~/lava/manage.py lava-server
```

Install the lava-dispatcher-host udev rules:

```shell
lava-dispatcher-host rules install
```

Restart the services:

```shell
systemctl restart lava-coordinator
systemctl restart lava-publisher
systemctl restart lava-scheduler
systemctl restart lava-server-gunicorn
systemctl restart lava-worker
```

LAVA is now accessible at [http://localhost/](http://localhost/)

## From sources

Install basic tools:

```shell
apt-get install git nfs-kernel-server \
    postgresql postgresql-client \
    python3 python3-yaml python3-django-extensions python3-werkzeug \
    ruby-foreman
```

Fetch the sources somewhere in your home directory.

```shell
cd ~/
git clone https://gitlab.com/lava/lava
cd ~/lava/
```

Install `lava-dispatcher` and  `lava-server` dependencies:

```shell
apt-get install \
    $(python3 share/requires.py  -p lava-common -d debian -s trixie -n) \
    $(python3 share/requires.py  -p lava-dispatcher -d debian -s trixie -n) \
    $(python3 share/requires.py  -p lava-dispatcher -d debian -s trixie -n -u) \
    $(python3 share/requires.py  -p lava-dispatcher-host -d debian -s trixie -n) \
    $(python3 share/requires.py  -p lava-server -d debian -s trixie -n) \
    $(python3 share/requires.py  -p lava-server -d debian -s trixie -n -u)
```

Create the test database:

```shell
sudo -u postgres psql -c "CREATE ROLE devel NOSUPERUSER CREATEDB NOCREATEROLE INHERIT LOGIN ENCRYPTED PASSWORD 'devel'"
sudo -u postgres psql -c "CREATE DATABASE devel OWNER devel"
```

Apply the django migrations

```shell
python3 manage.py migrate
```

Install the lava-dispatcher-host udev rules

```shell
$(pwd)/lava_dispatcher_host/lava-dispatcher-host rules install
```

The above will make the installed udev rules call the lava-dispatcher-host
program inside your source directory. If you move your source directory to a
different location, you will have to install the rules again.

Start the services

```shell
foreman start
```

LAVA is now accessible at [http://localhost:8000/](http://localhost:8000/)

# Developer setup

In this document, we assume that you are developing on a Debian system. We also
assume that the version is Bullseye (or any more recent versions).

??? note "Other distributions"
    Even if we assume the use of Debian, you can also use any non-Debian
    distributions. Just adapt the steps above to match your distribution.

We currently recommend two methods for developing LAVA.

* [from the Debian packages](#from-debian-packages)
* [from sources](#from-sources)

Both methods are currently used by LAVA core developers. Choosing one or the other is a matter of taste.

## From Debian packages

Install the Debian packages:

```shell
apt-get install git postgresql
apt-get install lava-coordinator lava-dispatcher lava-server
```

Fetch the sources somewhere in your home directory.

```shell
cd ~/
git clone https://gitlab.com/lava/lava
cd ~/lava/
```

Remove the lava binaries and packages from the system directories:

```shell
# python libraries
rm -rf /usr/lib/python3/dist-packages/lava/
rm -rf /usr/lib/python3/dist-packages/lava_common/
rm -rf /usr/lib/python3/dist-packages/lava_dispatcher/
rm -rf /usr/lib/python3/dist-packages/lava_dispatcher_host/
rm -rf /usr/lib/python3/dist-packages/lava_rest_app/
rm -rf /usr/lib/python3/dist-packages/lava_results_app/
rm -rf /usr/lib/python3/dist-packages/lava_scheduler_app/
rm -rf /usr/lib/python3/dist-packages/lava_server/
rm -rf /usr/lib/python3/dist-packages/linaro_django_xmlrpc/

# binaries
rm -f /usr/bin/lava-coordinator
rm -f /usr/bin/lava-dispatcher-host
rm -f /usr/bin/lava-run
rm -f /usr/bin/lava-server
rm -f /usr/bin/lava-worker
```

Add symbolic links to your local clone:

```shell
# python libraries
cd /usr/lib/python3/dist-packages/
ln -s ~/lava/lava/ .
ln -s ~/lava/lava_common/ .
ln -s ~/lava/lava_dispatcher/ .
ln -s ~/lava/lava_dispatcher_host/ .
ln -s ~/lava/lava_rest_app/ .
ln -s ~/lava/lava_results_app/ .
ln -s ~/lava/lava_scheduler_app/ .
ln -s ~/lava/lava_server/ .
ln -s ~/lava/linaro_django_xmlrpc/ .

# binaries
cd /usr/bin/
ln -s ~/lava/lava/coordinator/lava-coordinator .
ln -s ~/lava/lava/dispatcher/lava-run .
ln -s ~/lava/lava/dispatcher/lava-worker .
ln -s ~/lava/lava_dispatcher_host/lava-dispatcher-host .
ln -s ~/lava/manage.py lava-server
```

Install the lava-dispatcher-host udev rules:

```shell
lava-dispatcher-host rules install
```

Restart the services:

```shell
systemctl restart lava-coordinator
systemctl restart lava-publisher
systemctl restart lava-scheduler
systemctl restart lava-server-gunicorn
systemctl restart lava-worker
```

LAVA is now accessible at [http://localhost/](http://localhost/)

## From sources

Install basic tools:

```shell
apt-get install black git nfs-kernel-server \
    postgresql postgresql-client \
    python3 python3-yaml python3-django-extensions python3-werkzeug \
    ruby-foreman
```

Fetch the sources somewhere in your home directory.

```shell
cd ~/
git clone https://gitlab.com/lava/lava
cd ~/lava/
```

Install `lava-dispatcher` and  `lava-server` dependencies:

```shell
apt-get install \
    $(python3 share/requires.py  -p lava-common -d debian -s bullseye -n) \
    $(python3 share/requires.py  -p lava-dispatcher -d debian -s bullseye -n) \
    $(python3 share/requires.py  -p lava-dispatcher -d debian -s bullseye -n -u) \
    $(python3 share/requires.py  -p lava-dispatcher-host -d debian -s bullseye -n) \
    $(python3 share/requires.py  -p lava-server -d debian -s bullseye -n) \
    $(python3 share/requires.py  -p lava-server -d debian -s bullseye -n -u)
```

Create the test database:

```shell
sudo -u postgres psql -c "CREATE ROLE devel NOSUPERUSER CREATEDB NOCREATEROLE INHERIT LOGIN ENCRYPTED PASSWORD 'devel'"
sudo -u postgres psql -c "CREATE DATABASE devel OWNER devel"
```

Apply the django migrations

```shell
python3 manage.py migrate
```

Install the lava-dispatcher-host udev rules

```shell
$(pwd)/lava_dispatcher_host/lava-dispatcher-host rules install
```

The above will make the installed udev rules call the lava-dispatcher-host
program inside your source directory. If you move your source directory to a
different location, you will have to install the rules again.

Start the services

```shell
foreman start
```

LAVA is now accessible at [http://localhost:8000/](http://localhost:8000/)

# LAVA worker

Manage local lava jobs for attached DUTs.

## Command line

Run `/usr/bin/lava-worker`.

## Service

The systemd service is called `lava-worker`.

## Dependencies

lava-worker should be able to:

* connect to [lava-server-gunicorn](./lava-server-gunicorn.md)

## Configuration

Daemon start options:

* `/etc/default/lava-worker`
* `/etc/lava-server/lava-worker`

## Logs

The logs are stored in `/var/log/lava-dispatcher/lava-worker.log`

The log rotation is configured in `/etc/logrotate.d/lava-worker-log`.

## Security

TODO: should activate encryption

# Proxy

When running jobs, LAVA is fetching many resources from remote servers over
http or https. In some situation or when many jobs are running in parallel, the
network performances could become a bottleneck.

To improve network performances, you could setup a caching service that will
keep a local version of the resources used by LAVA.

You can choose among two kind of caching service:

* generic http proxy like [squid][squid]
* http caching service like [KissCache][KissCache]

## SQUID

SQUID is a well know caching proxy that will cache resources available over
http only.

!!! warning "SQUID and https"
    Configuring SQUID to cache https resources is an advance topic.
    If you need to cache https resources, you should have a look at
    [KissCache](#kisscache).

LAVA dispatcher uses the proxy configured in the `HTTP_PROXY` environment
variable.

!!! info "dispatcher environment"
    Environment variables are set in:

    * `/etc/lava-server/env.yaml` for every dispatchers
    * `/etc/lava-server/dispatcher.d/<name>/env.yaml` for a specific dispatcher

    Read [the documentation](../../technical-references/configuration/dispatcher.md)
    for more information.

??? tip "DUT environment"
    LAVA can also set the environment variable on the DUT.

    Read [the documentation](../../technical-references/configuration/dispatcher.md)
    for more information.

## KissCache

KissCache is a simple and stupid caching server.

KissCache is able to fetch and cache remote resources over `http` and `https`.
It can also download and stream the resource to many clients in parallel.

In order to use KissCache, users should prefix the resources's URLs by the
KissCache instance URL.

LAVA will use the KissCache instance specified in the dispatcher configuration:
```yaml
http_url_format_string: "https://kisscache-instance/api/v1/fetch?url=%s"
```

--8<-- "refs.txt"

# Raspberry Pi (U-Boot)

This guide covers setting up Raspberry Pi devices (RPi3, RPi4, RPi5) that use
U-Boot as their bootloader in LAVA.

LAVA has extensive support for devices that use U-Boot as their bootloader,
allowing you to deploy kernel, ramdisk, device tree blob (DTB), and other
files via `tftp` deploy action, and boot them using `u-boot` boot method.

## Hardware setup

Before adding a Raspberry Pi to LAVA, you need to set up the required hardware.
This includes configuring the serial console, network connectivity, power
control, and preparing an SD card with U-Boot installed.

### Serial console

[Setup serial console](common.md#serial-console).

For Raspberry Pi, connect the USB to serial adapter to the GPIO pins:

| RPi GPIO Pin | Signal | USB-Serial Adapter |
|--------------|--------|--------------------|
| Pin 6        | GND    | GND                |
| Pin 8        | TXD    | RXD                |
| Pin 10       | RXD    | TXD                |

### Network connectivity

The Raspberry Pi must be able to access the LAVA worker’s TFTP and NFS services
in order to download boot files and mount the root filesystem over NFS.
Therefore, the Raspberry Pi should be connected to the same network as the
LAVA worker.

### Power control

[Setup power control](common.md#power-control).

### SD Card

You need to prepare an SD Card with U-Boot configured to:

1. Display an interrupt prompt (e.g., `Hit any key to stop autoboot`)
2. Wait at least 5 seconds, so LAVA can match and interrupt the autoboot
3. Support TFTP boot commands

You can follow [this guide](https://docs.lavacloud.io/devices/rpi-3-and-4.html#sdcard)
to flash your SD Card.

### See also

- [How LAA simplifies device setup for LAVA](https://docs.lavacloud.io/generality/laa.html)
- [RPi device setup with LAA](https://docs.lavacloud.io/devices/rpi-3-and-4.html)

## Creating device type

[Create the device type](common.md#create-device-type) using one of the
following names that match the existing device type templates:

| Raspberry Pi Model | Device Type Name | Architecture |
| ------------------ | ---------------- | ------------ |
| RPi 3 Model B (32-bit) | `bcm2837-rpi-3-b-32` | ARM32 |
| RPi 3 Model B (64-bit) | `bcm2837-rpi-3-b` | ARM64 |
| RPi 4 Model B | `bcm2711-rpi-4-b` | ARM64 |
| RPi 5 Model B | `bcm2712-rpi-5-b` | ARM64 |

## Creating device

Each Raspberry Pi device requires a device dictionary that specifies
device-specific settings such as serial connection and power control commands.
The device dictionary should extend the device type template and add the
device-specific configurations.

1. [Add the device](common.md#add-device) using the following settings:
    - **Device Type:** See device type name in the
    [above table](#creating-device-type) (e.g., `bcm2711-rpi-4-b`)
    - **Hostname:** A unique name (e.g., `rpi4-01`)
2. [Add the device configuration](common.md#add-device-configuration):

    ```jinja
    {% extends "<device_type>.jinja2" %}

    {% set connection_command = "telnet localhost <port>" %}

    {% set hard_reset_command = "<power_reset_command>" %}
    {% set power_off_command = "<power_off_command>" %}
    {% set power_on_command = "<power_on_command>" %}
    ```

    !!! note
        Replace all the placeholders marked with `<>` with their corresponding
        actual values.

## Sample job definitions

### Booting from NFS

```yaml
job_name: u-boot nfs
device_type: bcm2837-rpi-3-b-32

priority: medium
visibility: public

timeouts:
  job:
    minutes: 30

actions:
- deploy:
    to: tftp
    timeout:
      minutes: 15
    dtb:
      url: http://example.com/nfs/bcm2837-rpi-3-b-plus.dtb
    kernel:
      url: http://example.com/nfs/zImage
      type: zimage
    modules:
      url: http://example.com/nfs/modules.tar.xz
      compression: xz
    ramdisk:
      url: http://example.com/nfs/initrd.cpio.gz
      compression: gz
    nfsrootfs:
      url: http://example.com/nfs/full.rootfs.tar.xz
      format: tar
      overlays:
        kselftest:
          url: http://example.com/nfs/kselftest.tar.xz
          compression: xz
          format: tar
          path: /opt/kselftest

- boot:
    method: u-boot
    commands: nfs
    prompts:
    - '/ #'
    timeout:
      minutes: 5

- test:
    timeout:
      minutes: 5
    definitions:
    - from: inline
      repository:
        metadata:
          format: Lava-Test Test Definition 1.0
          name: health checks
        run:
          steps:
          - lava-test-case kernel-info --shell uname -a
          - lava-test-case network-info --shell ip a
      name: health-checks
      path: inline/health-checks.yaml
```

### Booting from ramdisk

```yaml
device_type: bcm2837-rpi-3-b-32
job_name: u-boot boot ramdisk

priority: medium
visibility: public

timeouts:
  job:
    minutes: 15

actions:
- deploy:
    dtb:
      url: http://example.com/ramdisk/bcm2837-rpi-3-b.dtb
    kernel:
      type: zimage
      url: http://example.com/ramdisk/zImage
    ramdisk:
      compression: gz
      url: http://example.com/ramdisk/rootfs.cpio.gz
    timeout:
      minutes: 3
    to: tftp

- boot:
    method: u-boot
    commands: ramdisk
    prompts:
    - '/ #'
    timeout:
      minutes: 5

- test:
    timeout:
      minutes: 5
    definitions:
    - from: inline
      repository:
        metadata:
          format: Lava-Test Test Definition 1.0
          name: health checks
        run:
          steps:
          - lava-test-case kernel-info --shell uname -a
          - lava-test-case network-info --shell ip a
      name: health-checks
      path: inline/health-checks.yaml
```

--8<-- "refs.txt"

# Timeouts

## Timeout syntax

A timeout is a dictionary with unit and value:

```yaml
timeout:
  <unit>: <value>
```

Valid units are `seconds`, `minutes`, `hours` and `days` while the value should be an integer.

## Job timeout

The entire job will have an overall timeout. The job will fail if this timeout
is exceeded, whether or not any other timeout is longer.

```yaml
timeouts:
  job:
    minutes: 15
```

## Action timeout

Each action of a job will have a timeout. The action will not run longer than the defined timeout.

To define the timeout, you should add a `timeout` dictionary to the action definition:

```yaml
actions:
- deploy:
    timeout:
      minutes: 5
    [...]
```

!!! warning "action timeout"
    No action timeout can be larger than the job timeout.
    If the action timeout is larger than the job timeout, LAVA will set the
    action timeout to the job timeout and warn during validation.


### Default timeout

You can define the default duration of every action:

```yaml
timeouts:
  action:
    seconds: 60
```

You can also define individual action timeout in the root timeout dictionary:

```yaml
timeouts:
  actions:
    extract-nfsrootfs:
      seconds: 60
```

## Connection timeout

When an action interact with the DUT, this action will use the
`connection_timeout`. Each interaction with the DUT should take less than the
`connection_timeout` duration.

You can define the default duration of every connection timeout:

```yaml
timeouts:
  connection:
    seconds: 60
```

You can also define individual connection timeout in the root timeout dictionary:

```yaml
timeouts:
  connections:
    lava-test-shell:
      seconds: 60
```


## Skipping timeout

!!! warning "TODO"

# Musca

The `musca` boot method is used to boot Musca devices deployed by the
[`musca`](../deploy/to-musca.md) deployment method.

```yaml
- deploy:
    to: musca
    images:
      test_binary:
        url: https://storage.lavacloud.io/health-checks/musca-b/tfm.hex

- boot:
    method: musca
```

Unlike the [`minimal`](./method-minimal.md) boot method, the board has to be
powered on before the serial will be available as the board is powered by the
USB that provides the serial connection also. Therefore, the board is powered on
then connection to the serial is made.

## prompts

Optional. If the test binary has output, `prompts` can be used to wait for a
specific string from the serial output before continuing.

```yaml
- boot:
    method: musca
    prompts: "string"
```

## monitors

No shell is expected, and no boot string is checked. All checking should be
done with [test monitors](../test.md#monitors).

## Job example

```yaml
job_name: musca-b-health-check
device_type: musca-b
priority: medium
visibility: public

timeouts:
  job:
    minutes: 20
  action:
    minutes: 10

actions:
    - deploy:
        to: musca
        images:
          test_binary:
            url: https://storage.lavacloud.io/health-checks/musca-b/tfm.hex

    - boot:
        method: musca

    - test:
        monitors:
        - name: "Secure_Test_Suites_Summary"
          start: "Non-Secure system"
          end: "starting\\.{3}"
          pattern: "Non-Secure system starting\\.{3}"
          fixupdict:
              '!': pass
              '': fail
```

# Glossary of terms

<!--
Please add new terms in alphabetical order and feel free to relocate
existing terms to match. Ensure that new terms added to the glossary are
also linked from the body of the documentation. The glossary is a reference
only, users are not expected to need to read the entire glossary to find
the information.
-->

**A** [action level](#action-level) | [alias](#alias)

**B** [BMC](#bmc)

**C** [chart](#chart) | [CI loop](#ci-loop)

**D** [device](#device) | [device dictionary](#device-dictionary) |
[device status transition](#device-status-transition) |
[device tag](#device-tag) | [device type](#device-type) |
[developer image](#developer-image) | [dispatcher](#dispatcher) |
[distributed deployment](#distributed-deployment) | [DTB](#dtb) | [DUT](#dut)

**F** [frontend](#frontend)

**G** [group](#group)

**H** [hacking session](#hacking-session) | [health check](#health-check) |
[hidden device type](#hidden-device-type) | [hostname](#hostname)

**I** [inline](#inline) | [interface tag](#interface-tag)

**J** [jinja2](#jinja2) | [job context](#job-context) |
[job definition](#job-definition)

**M** [messageID](#messageid) | [metadata](#metadata) | [MultiNode](#multinode)

**N** [namespace](#namespace)

**O** [offline](#offline)

**P** [parameters](#parameters) | [PDU](#pdu) |
[physical access](#physical-access) | [pipeline](#pipeline) |
[priority](#priority) | [production image](#production-image) |
[prompts](#prompts) | [protocol](#protocol)

**Q** [query](#query)

**R** [remote worker](#remote-worker) |
[restricted device](#restricted-device) | [results](#results) |
[retired](#retired) | [role](#role) | [rootfs](#rootfs) |
[rootfstype](#rootfstype)

**S** [scheduler](#scheduler)

**T** [target_group](#target_group) | [test case](#test-case) |
[test run](#test-run) | [test shell](#test-shell) | [test set](#test-set) |
[test suite](#test-suite) | [tftp](#tftp) | [token](#token)

**U** [UART](#uart)

**V** [visibility](#visibility) | [VLANd](#vland)

**W** [worker](#worker)

---

## action level

The pipeline is organized into sections and levels. The first section of the
pipeline is given level 1. Sub-tasks of that section start with level 1.1 and
so on.

## alias

A string which can be used to relate the descriptive device-type name to a
particular list of aliases which could be used to look up the matching
device-type. This can be useful to list the device tree blobs (DTB) which can
be used with this device-type. Aliases can be used in job submissions directly.

## BMC

A Baseboard Management Controller (BMC) is an embedded controller on a computer
mainboard which allows external monitoring and management of the computer
system.

## chart

A chart allows users to track results over time using queries.

## CI loop

Continuous Integration (CI) typically involves repeated automated submissions
using automated builds of the artifacts prompted by modifications made by
developers. Providing feedback to the developers on whether the automated build
passed or failed creates a loop. LAVA is designed as one component of a CI
loop.

## device

A device in LAVA is an instance of a device type.

* Test writers: see [device tag](#device-tag)
* Admins: see [device dictionary](#device-dictionary)

## device dictionary

The device dictionary holds data which is specific to one device within a group
of devices of the same device type. For example, the power control commands
which reference a single port number. The dictionary itself is a `key:value`
store within the LAVA server database which admins can modify to set
configuration values according to the pipeline design.

## device status transition

A record of when a device changed status, who caused the transition, when the
transition took place as well as any message assigned to the transition.

## device tag

A tag is a device specific label which describes specific hardware capabilities
of this specific device. Test jobs using tags will fail if no suitable devices
exist matching the requested device tag or tags. Tags are typically used when
only a proportion of the devices of the specified type have hardware support
for a particular feature, possibly because those devices have peripheral
hardware connected or enabled. A device tag can only be created or assigned to
a particular device by a lab admin. When requesting tags, remember to include a
description of what the tagged device can provide to a Test Job.

## device type

The common type of a number of devices in LAVA. The device type may have a
health check defined. Devices with the same device type will run the same
health check at regular intervals.

## developer image

A build of Android which, when deployed to a device, means that the device **is
visible** to `adb`. Devices configured this way will be able to have the image
replaced using any machine, just be connecting a suitable cable, so these
images are not typically deployed onto hardware which will be sold to the
customer without having this image replaced with a production image.

## dispatcher

A machine to which multiple devices are connected. The dispatcher has
`lava-dispatcher` installed and passes the commands to the device and other
processes involved in running the LAVA test. A dispatcher does not need to be
at the same location as the server which runs the scheduler. The term
`dispatcher` relates to how the machine operates the `lava-dispatch` process
using `lava-worker`. The related term [worker](#worker) relates to how the
machine appears from the server.

## distributed deployment

A method of installing LAVA involving a single server and one or more remote
workers which communicate with the master using HTTP. This method spreads the
load of running tests on devices across multiple dispatchers.

## DTB

Device Tree Blob - file describing hardware configuration, commonly used on ARM
devices with the Linux kernel. See
<https://en.wikipedia.org/wiki/Device_tree> for more information.

## DUT

Device Under Test - a quick way to refer to the device in LAVA.

## frontend

`lava-server` provides a generic frontend consisting of the Results, Queries,
Job tables, Device tables and Charts. Many projects will need to customize this
data to make it directly relevant to the developers. This is supported using
the XML-RPC and REST API support.

## group

LAVA uses the Django local group configuration (synchronizing Django groups
with external groups like LDAP is **not** supported). Users can be added to
groups after the specified group has been created by admins using the Django
admin interface or the `lava-server manage groups` and `lava-server manage
users` command line support.

## hacking session

A test job which uses a particular type of test definition to allow users to
connect to a test device and interact with the test environment directly.
Normally implemented by installing and enabling an SSH daemon inside the test
image. Not all devices can support hacking sessions.

## health check

A test job for one specific device type which is automatically run at regular
intervals to ensure that the physical device is capable of performing the
minimum range of tasks. If the health check fails on a particular device, LAVA
will automatically put that device offline. Health checks have higher priority
than any other jobs.

## hidden device type

A device type can be hidden by the LAVA administrators. Devices of a hidden
device type will only be visible to owners of at least one device of this type.
Other users will not be able to access the job output, device status transition
pages or bundle streams of devices of a hidden type. Devices of a hidden type
will be shown as `Unavailable` in tables of test jobs and omitted from tables
of devices and device types if the user viewing the table does not own any
devices of the hidden type.

## hostname

The unique name of this device in this LAVA instance, used to link all jobs,
results and device information to a specific device configuration.

## inline

A type of test definition which is contained within the job submission instead
of being fetched from a URL. These are useful for debugging tests and are
recommended for the synchronization support within MultiNode test jobs.

## interface tag

An interface tag is similar to [device tag](#device-tag) but operates **solely**
within the VLANd support. An interface tag may be related to the link speed
which is achievable on a particular switch and port - it may also embed
information about that link.

## jinja2

Jinja2 is a templating language for Python, modelled after Django's templates.
It is used in LAVA for device-type configuration, as it allows conditional
logic and variable substitution when generating device configuration for the
dispatcher.

See: <http://jinja.pocoo.org/docs/dev/>

## job context

Test job definitions can include the `context:` dictionary at the top level.
This is used to set values for selected variables in the device configuration,
subject to the administrator settings for the device templates and device
dictionary. A common example is to instruct the template to use the
`qemu-system-x86_64` executable when starting a QEMU test job using the value
`arch: amd64`. All device types support variables in the job context.

## job definition

The original YAML submitted to create a job in LAVA is retained in the database
and can be viewed directly from the job page. When resubmitting old job
definitions, be aware that device templates and other system configuration may
have changed since the original submission, so the job may behave differently or
fail validation. If the job was a MultiNode job, the MultiNode definition will
be the unchanged YAML from the original submission; the job definition will be
the parsed YAML for this particular device within the MultiNode job.

## messageID

Each message sent using the MultiNode API uses a `messageID` which is a string,
unique within the group. It is recommended to make these strings descriptive
(use underscores instead of spaces). The messageID will be included in the log
files of the test.

## metadata

Test jobs should include metadata relating to the files used within the job.
Metadata consists of a key and a value, there is no limit to the number of key
value pairs as long as each key is unique within the metadata for that test
job.

## MultiNode

A single test job which runs across multiple devices, or using multiple
independent connections to the same device.

## namespace

A simple text label which is used to tie related actions together within a test
job submission where multiple deploy, boot or test actions are defined. A
common use case for namespaces is the use of container in a test job where some
actions are to be executed inside the container and some on the DUT. The
namespace is used to store the temporary locations of files and other dynamic
data during the running of the test job so that, for example, the test runner is
able to execute the correct test definition YAML. Namespaces are set in the test
job submission.

## offline

A status of a device which allows jobs to be submitted and reserved for the
device but where the jobs will not start to run until the device is online.
Devices enter the offline state when a health check fails on that device or the
administrator puts the device offline.

## parameters

Parameters are used in a number of contexts in LAVA.

* For the use of parameters to control test jobs, see the test action
  parameters documentation.
* For the use of parameters within the codebase of the pipeline, see the
  developer guide.

## PDU

PDU is an abbreviation for Power Distribution Unit - a network-controlled set
of relays which allow the power to the devices to be turned off and on
remotely. Many PDUs are supported by `pdudaemon` to be able to hard reset
devices in LAVA.

## physical access

The user or group with physical access to the device, for example to fix a
broken SD card or check for possible problems with physical connections. The
user or group with physical access is recommended to be one of the superusers.

## pipeline

Within LAVA, the pipeline is the model for the dispatcher code where submitted
jobs are converted to a pipeline of discrete actions - each pipeline is
specific to the structure of that submission and the entire pipeline is
validated before the job starts. The model integrates concepts like fail-early,
error identification, avoid defaults, fail and diagnose later, as well as
giving test writers more rope to make LAVA more transparent.

## priority

A job has a default priority of `Medium`. This means that the job will be
scheduled according to the submit time of the job, in a list of jobs of the
same priority. Every health check has a higher priority than any submitted job
and if a health check is required, it will **always** run before any other
jobs. Priority only has any effect while the job is queued as `Submitted`.

## production image

A build of Android which, when deployed to a device, means that the device is
**not** visible to `adb`. This is typically how a device is configured when
first sold to the consumer.

## prompts

A list of prompt strings which the test writer needs to specify in advance and
which LAVA will use to determine whether the boot was successful. One of the
specified prompts **must** match before the test can be started.

## protocol

A protocol in LAVA is a method of interacting with external services using an
API (Application Programming Interface) instead of with direct shell commands
or via a test shell. Examples of services in LAVA which use protocols include
MultiNode and VLANd. The protocol defines which API calls are available through
the LAVA interface and the Pipeline determines when the API call is made.

## query

Queries are used to identify test jobs and associated results which match
specified criteria based on the results or metadata.

## remote worker

A dispatcher with devices attached which does not have a web frontend but which
uses an HTTP connection to a remote lava-server to control the operation of
test jobs on the attached devices.

## restricted device

A restricted device can only accept job submissions from the device owner. If
the device owner is a group, all users in that group can submit jobs to the
device.

## results

LAVA results provide a generic view of how the tests performed within a test
job. Results from test jobs provide support for queries, charts and downloading
results to support later analysis and frontends. Results can be viewed whilst
the test job is running. Results are also generated during the operation of the
test job outside the test action itself. All results are referenced solely
using the test job ID.

## retired

A device is retired when it can no longer be used by LAVA. A retired device
allows historical data to be retained in the database, including log files,
result bundles and state transitions. Devices can also be retired when the
device is moved from one instance to another.

## role

An arbitrary label used in MultiNode tests to determine which tests are run on
the devices and inside the YAML to determine how the devices communicate.

## rootfs

A tarball for the root file system.

## rootfstype

Filesystem type for the root filesystem, e.g. ext2, ext3, ext4.

## scheduler

There is a single scheduler in LAVA, running on the server. The scheduler is
responsible for assigning devices to submitted test jobs.

## target_group

In MultiNode, the single submission is split into multiple test jobs which all
share a single `target_group` which uses a string as a unique ID. The
`target_group` is usually transparent to test writers but underpins how the
rest of the MultiNode API operates.

## test case

An individual test case records a single test event as a pass or fail along
with measurements, units or a reference.

## test run

The result from a single test definition execution. The individual id and
result of a single test within a test run is called the test case.

## test shell

Most test jobs will boot into a POSIX type shell, much like if the user had
used `ssh`. LAVA uses the test shell to execute the tests defined in the Lava
Test Shell Definition(s) specified in the job definition.

## test set

Test writers can choose to subdivide a single test suite into multiple sets,
for example to handle repetition or changes to the parameters used to run the
tests.

## test suite

Individual test cases are aggregated into a test suite and given the name
specified in the test job definition. The Test Suite is created when results
are generated in the running test job. LAVA uses a reserved test suite called
`lava` for results generated by the actions running the test job itself.
Results in the `lava` suite contain details like the commit hash of the test
definitions, messages from exceptions raised if the job ends Incomplete and
other data about how the test behaved.

## tftp

Trivial File Transfer Protocol (TFTP) is a file transfer protocol, mainly to
serve boot images over the network to other machines (e.g. for PXE booting).
The protocol is managed by the
[tftpd-hpa package](https://tracker.debian.org/pkg/tftp-hpa) and **not** by
LAVA directly.

## token

LAVA uses tokens to authenticate users via the XML-RPC and REST APIs.

## UART

A UART (Universal asynchronous receiver-transmitter) is the most common way to
make a serial connection to a DUT. Some devices can support multiple UARTs.
This can be useful as a way to isolate the test shell processing from kernel
messages.

See: [UART article on Wikipedia](https://en.wikipedia.org/wiki/Universal_asynchronous_receiver-transmitter)

## visibility

Supports values of `public`, `personal` and `group` and controls who is allowed
to view the job and the results generated by the job. This includes whether the
results are available to queries and to charts:

```yaml
visibility: personal
```

or:

```yaml
visibility: public
```

`group` visibility setting should list the groups users must be in to be
allowed to see the job. If more than one group is listed, users must be in all
the listed groups to be able to view the job or the results:

```yaml
visibility:
  group:
    - developers
    - project
```

In this example, users must be members of both `developers` group and `project`
group. Groups must already exist in the Django configuration for the instance.

## VLANd

VLANd is a daemon to support virtual local area networks in LAVA. This support
is specialized and requires careful configuration of the entire LAVA instance,
including the physical layout of the switches and the devices of that instance.

## worker

The worker is responsible for running the `lava-worker` daemon to start and
monitor test jobs running on the dispatcher. Each server has a worker installed
by default. When a dispatcher is added to the master as a separate machine,
this worker is a remote worker. The admin decides how many devices to assign to
which worker. In large instances, it is common for all devices to be assigned
to remote workers to manage the load on the master.

# First steps

Running a test in LAVA is done by submitting a test job that describes what
software to deploy, how to boot the DUT, and what tests to run.

The following diagram shows the typical workflow for running your first test.
Each step is explained in detail in the basic tutorials below.

```mermaid
flowchart LR
    1([Sign in]) --> 2([Write job definition]) --> 3([Submit job]) --> 4([Inspect job output])
```

## Basic tutorials

| Tutorial | Description |
| -------- | ----------- |
| [Signing in](basic-tutorials/signing-in.md) | Sign in to a LAVA instance |
| [Job definition](basic-tutorials/job-definition.md) | Understand the structure of a job definition |
| [Job submission](basic-tutorials/submit.md) | Submit your first job via web UI or lavacli |
| [Test definition](basic-tutorials/test-definition.md) | Write test definitions for your jobs |
| [Job output](basic-tutorials/job-output.md) | Navigate and understand job logs and results |
| [lavacli](basic-tutorials/lavacli.md) | Install and configure the command-line tool |

## What's next

Explore the advanced tutorials to get the most out of LAVA.

--8<-- "refs.txt"

# SSH

The `ssh` deployment method is used to prepare the LAVA overlay needed by the
[ssh boot method](../boot/method-ssh.md).

```yaml
- deploy:
    to: ssh
    os: debian
```

## os

The operating system running on the device. LAVA uses this to select the
right options for applying the LAVA overlay.

# GRUB

The `grub` boot method is used to boot a device using the configured grub boot
commands in device type or job definition.

```yaml
- boot:
    method: grub
    commands: ramdisk
    prompts:
    - 'root@debian:~#'
    timeout:
      minutes: 5
```

## commands

See [commands](./common.md#commands).

## use_bootscript

See [use_bootscript](./method-bootloader.md#use_bootscript).

## reset

Defaults to `true`. When `false`, LAVA will not reset the device at the start
of the boot action.

This can be used when the device has already been booted into the bootloader by
a previous action.

```yaml
- boot:
    method: grub
    commands: ramdisk
    reset: false
```

## boot_finished

A string or list of strings that LAVA will wait for when no `prompts` are set.
This can be used with an OS installer that does not present a login prompt.
After the string is matched the device is powered off.

```yaml
- boot:
    method: grub
    commands: ramdisk
    boot_finished:
    - 'Installation complete'
    - 'reboot: Restarting system'
```

## Example job

```yaml
job_name: 'ampereone grub boot example job'
device_type: ampereone

priority: medium
visibility: public

timeouts:
  actions:
    finalize:
      seconds: 60
    power-off:
      seconds: 60
  connection:
    minutes: 2
  job:
    minutes: 65
  queue:
    hours: 72

context:
  extra_kernel_args: ' rw'

actions:
- deploy:
    to: tftp
    kernel:
      compression: gz
      type: image
      url: https://example.com/Image.gz
    nfsrootfs:
      compression: xz
      format: tar
      overlays:
        modules:
          compression: xz
          format: tar
          path: /usr/
          url: https://example.com/modules.tar.xz
        overlay-00:
          compression: xz
          format: tar
          path: /
          url: https://storage.tuxboot.com/overlays/debian/trixie/arm64/ltp/master/ltp.tar.xz
      url: https://storage.tuxboot.com/debian/20250722/trixie/arm64/rootfs.tar.xz
    os: debian
    timeout:
      minutes: 30

- boot:
    method: grub
    commands: nfs
    auto_login:
      login_prompt: 'login:'
      username: root
    prompts:
    - root@(.*):[/~]#
    - '/ #'
    timeout:
      minutes: 30

- test:
    definitions:
    - from: inline
      name: prep-inline
      path: inline/prep.yaml
      repository:
        metadata:
          description: Device preparation
          format: Lava-Test Test Definition 1.0
          name: prep-tests
        run:
          steps:
          - export STORAGE_DEV=/dev/nvme0n1p2
          - mkfs.ext4 -F "$STORAGE_DEV" || lava-test-raise "mkfs.ext4 $STORAGE_DEV
            failed; job exit"
          - mkdir -p /scratch && mount "$STORAGE_DEV" /scratch || lava-test-raise
            "mount $STORAGE_DEV failed; job exit"
          - df -h
          - mount
    timeout:
      minutes: 5

- test:
    definitions:
    - compression: zstd
      from: url
      lava-signal: kmsg
      name: ltp-smoke
      parameters:
        ENVIRONMENT: production
        KIRK_WORKERS: 1
        LTP_INSTALL_PATH: /opt/ltp/
        LTP_TMPDIR: /scratch
        RUNNER: kirk
        SHARD_INDEX: 1
        SHARD_NUMBER: 1
        SKIPFILE: skipfile-lkft.yaml
        SKIP_INSTALL: 'true'
        TIMEOUT_MULTIPLIER: 5
        TST_CMDFILES: smoketest
      path: automated/linux/ltp/ltp.yaml
      repository: https://github.com/Linaro/test-definitions/releases/download/2025.10.01/2025.10.tar.zst
    timeout:
      minutes: 5
```

--8<-- "refs.txt"

# JLink

The `jlink` boot method is used to flash and boot devices using the
[J-Link](https://www.segger.com/products/debug-probes/j-link/) debug probe from
SEGGER. It supports flashing single or multiple images and can interact with
devices that have multiple cores.

```yaml
- boot:
    method: jlink
    commands:
    - loadfile {shell} 0x0
    prompts:
    - 'SHELL>>'
    timeout:
      minutes: 2
```

!!! note
    Images passed to JLink commands must be first deployed using the
    [`tmpfs`](../deploy/to-tmpfs.md) deploy action.

## Installation

`jlink` must be installed manually on the LAVA worker since it is not provided
as a dependency by LAVA. See
[JLink Software and Documentation Pack](https://www.segger.com/downloads/jlink/)
for installing the tool.

## Configuration

For adding a new device type, see
[JLink device type](../../../configuration/device-type-template.md#jlink).

For adding a new device, refer to
[JLink device configuration](../../../configuration/device-dictionary.md#jlink).

## commands

Optional. A list of JLink commands to execute. These can include `loadfile` or
any other valid JLink command. Variables in the commands reference images
deployed during the prior `tmpfs` deploy action.

### Flashing signal image

When only a single image deployed with `tmpfs` deployment method and the
`commands` is not specified in this boot method, LAVA flashes the single image
to the default load address `0x0`.

```yaml
- boot:
    method: jlink
    timeout:
      minutes: 2
```

### Flashing multiple images

```yaml
- boot:
    method: jlink
    commands:
    - loadfile {shell} 0x0
    - loadfile {hello_world} 0x0
```

In this example, `shell` and `hello_world` are the image keys used in the
`tmpfs` deploy action.

## coretype

For devices with multiple cores, specifies which core to connect to. The
available core types are defined by `supported_core_types` in the device type
template. If not specified, the first core in the list is used.

```yaml
- boot:
    method: jlink
    coretype: M7
    commands:
    - loadfile {test}
    prompts:
    - 'SHELL>>'
    timeout:
      minutes: 2
```

## prompts

See [prompts](./common.md#prompts).

## timeout

See [timeouts](../../timeouts.md).

## Example job

```yaml
device_type: frdm-kw36zj
job_name: health-check

timeouts:
  job:
    minutes: 10
  action:
    minutes: 3

priority: medium
visibility: public

actions:
- deploy:
    to : tmpfs
    images :
      boot :
          url: https://example.com/frdm-kw36-shell.bin

- boot:
    method: jlink

- test:
    monitors:
    - name: tests
      start: Running test suite common_test
      end: PROJECT EXECUTION SUCCESSFUL
      pattern: '(?P<result>(PASS|FAIL)) - (?P<test_case_id>.*)\.'
      fixupdict:
        PASS: pass
        FAIL: fail
```

## Using multiple MCUs

To set up a test requiring multiple MCUs, such as testing Bluetooth
connectivity, you can use a multinode job. In this setup, MCU 1 and MCU 2 are
placed on the same worker, and a Docker device is used to execute tests once
both MCUs have completed their deployment and boot processes.

![Multiple MCU JLink Setup](./multi_mcu_jlink_setup.png)

The test setup consists of:

- **MCU 1 and MCU 2:** These are the target devices to be tested.
- **Docker Device:** A Docker container is deployed to execute the tests. This
Docker device is responsible for coordinating the test execution once both MCUs
are ready.

Here’s how the process works:

1. **Deployment and Boot:** The necessary images are deployed to both MCUs using
tmpfs. Once deployed, the MCUs are booted to initiate the test process.
2. **Test Execution:** Once both MCUs have completed their boot processes, the
Docker device executes the test suite.
3. **Results Analysis:** The test results are collected and analyzed by a LAVA
test to verify the functionality of the MCUs and ensure successful connectivity.

Example job:

```yaml
job_name: Multiple MCUs

timeouts:
  job:
    minutes: 5
  action:
    minutes: 5
  connection:
    minutes: 2

priority: medium
visibility: public

protocols:
  lava-multinode:
    roles:
      host:
        device_type: rw610bga
        count: 1
        tags:
          - rw610bga-fr01
        timeout:
          minutes: 30
      guest:
        device_type: rw610bga
        tags:
          - rw610bga-fr02
        count: 1
        timeout:
          minutes: 30
      docker:
        device_type: docker
        count: 1

actions:
- deploy:
    role:
      - host
      - guest
    to : tmpfs
    images :
      boot :
          url: https://example.com/bt_ble_cli.out
      cpu1 :
          url: https://example.com/rw61xw_raw_cpu1_a1.bin
      cpu2 :
          url: https://example.com/rw61xn_raw_cpu2_ble_a1.bin
      combo :
          url: https://example.com/rw61xn_combo_raw_cpu2_ble_15_4_combo_a1.bin

- deploy:
    role:
      - docker
    to: docker
    os: ubuntu
    image:
      name: connectivity_test
      local: true

- boot:
    role:
      - host
      - guest
    method: jlink
    commands :
      - loadfile {boot}
      - loadfile {cpu1} 0x8400000
      - loadfile {cpu2} 0x8540000
      - loadfile {combo} 0x85E0000
    timeout:
      minutes: 2

- boot:
    role:
      - docker
    method: docker
    command: bash
    prompts:
    - 'root@\w*:(.*)#'
    timeout:
      minutes: 2
```

# QDL

The qdl deployment action downloads a flat build tarball to Qualcomm devices
using qdl [qdl](https://github.com/linux-msm/qdl).
It is possible to add a LAVA overlay to one of the partition images
within the tarball.
QDL deployment is required to boot into QDL mode and
flash the contents of the tarball onto the board.

```yaml
- deploy:
    rootfs_image: rootfs.img
    overlay_path: /home
    qcomflash:
      url: ...
      apply-overlay: true
    to: qdl
```

## qcomflash

The `qcomflash` block specifies the location of the tarball to be downloaded.
It uses the usual [download syntax](./index.md#artifacts).
The tarball should not be decompressed by the download action.
It is assumed that the archive is compressed.

## rootfs_image

This parameter points to a partition image where the LAVA overlay should be added.
The value should be a path relative to the main directory in the tarball.

## overlay_path

This parameter names the path inside the [rootfs_image](#rootfs_image)
where the LAVA overlay should be added.
The value should be in line with the variable `lava_test_results_dir` defined in the job context.

## qcomflash

This parameter represents the tarball containing the build to be flashed to the device.

### apply-overlay

LAVA can apply the [overlay](../boot/common.md#transfer_overlay) to the image before flashing.
The overlay is applied to the [overlay_path](#overlay_path).

# CMSIS-DAP

The `cmsis-dap` boot method takes no parameters. It is used to flash and boot
a device by copying the images to the USB Mass Storage device exposed by the
[CMSIS-DAP](https://arm-software.github.io/CMSIS-DAP/latest/) interface.

```yaml
- boot
    method: cmsis-dap
    timeout:
      minutes: 10
```

# New Connection

The `new_connection` boot method can be used to switch to a new connection.
If the kernel and the device are both appropriately configured, a test can use
a new connection to isolate test and kernel messages.

If the second [connection](../../../configuration/device-dictionary.md#connections)
`uart0` is configured in the device dictionary, the connection can be created in
a separate namespace.

```yaml
- boot:
    namespace: isolation
    connection: uart0
    method: new_connection
```

The `new_connection` boot method **must** use a different namespace to all other
actions in the test job. The test actions **must** pass this namespace label as
the `connection-namespace`.

```yaml
- test:
    namespace: hikey-oe
    connection-namespace: isolation
    definitions:
    - from: git
      repository: https://gitlab.com/lava/functional-tests.git
      path: posix/smoke-tests-basic.yaml
      name: smoke-tests
    timeout:
      minutes: 5
```

## Example device

```jinja title="Connection snippet"
{% extends 'n1sdp.jinja2' %}

{% set connection_list = ['uart0', 'uart1', 'uart2', 'uart3'] %}
{% set connection_commands = {
  'uart0': 'telnet w1 7007',
  'uart1': 'telnet w1 7008',
  'uart2': 'telnet w1 7009',
  'uart3': 'telnet w1 7010'
} %}
{% set connection_tags = {
  'uart0': ['primary', 'telnet'],
  'uart1': ['telnet'],
  'uart2': ['telnet'],
  'uart3': ['telnet']
} %}
```

## Example job

```yaml
job_name: new_connection job
device_type: n1sdp

priority: medium
visibility: public

timeouts:
  job:
    minutes: 15
  connection:
    minutes: 5

actions:
- deploy:
    namespace: default
    to: flasher
    images:
      recovery_image:
        url: https://storage.lavacloud.io/health-checks/n1sdp/board-firmware.zip
        compression: zip
    timeout:
      minutes: 5

- boot:
    namespace: default
    method: minimal
    timeout:
      minutes: 5

- boot:
    namespace: uart_one
    method: new_connection
    connection: uart1
    timeout:
      minutes: 5

- test:
    namespace: uart_one
    connection-namespace: uart_one
    interactive:
    - name: int_1
      prompts: ["Press ESCAPE for boot options"]
      script:
      - command:
    timeout:
      minutes: 10
```

# LAVA logs in database

LAVA logs are received and stored by lava-server-gunicorn. The default
storage backend uses filesystem for storing logs and the default location of
`/var/lib/lava-server/default/media/job-output/`.

LAVA also supports storing the logs in a few NoSQL based database engines.

## Configuration

This can be set up in the [LAVA configuration](../basic-tutorials/instance/configure.md), like so:

```yaml
LAVA_LOG_BACKEND: "lava_scheduler_app.logutils.LogsFilesystem"
```

Available backends are:

* `lava_scheduler_app.logutils.LogsMongo`
* `lava_scheduler_app.logutils.LogsElasticsearch`
* `lava_scheduler_app.logutils.LogsFirestore`

The list can be also found in [the source code](https://git.lavasoftware.org/lava/lava/-/blob/master/lava_server/settings/common.py)

### MongoDB

Integration with MongoDB requires two variables to be set in the [LAVA settings](../basic-tutorials/instance/configure.md):

* `MONGO_DB_URI` - connection string URI format for mongodb (i.e. mongodb://example.com:27017)
* `MONGO_DB_DATABASE` - database name for LAVA logs (i.e. 'lava-logs')

Every log line in LAVA is one document in MongoDB database.

### ElasticSearch

Integration with Elasticsearch db requires three variables to be set in the [LAVA settings](../basic-tutorials/instance/configure.md):

* `ELASTICSEARCH_URI` - connection string URI format for Elasticsearch (i.e. http://localhost:9200/)
* `ELASTICSEARCH_INDEX` - database(index) name for LAVA logs (i.e. 'lava-logs')
* `ELASTICSEARCH_APIKEY` - API key to be used in Authorization header when
  sending requests to the Elasticsearch API. Can be obtained via [this request](https://www.elastic.co/guide/en/elasticsearch/reference/current/security-api-create-api-key.html).

Every log line in LAVA is one document in Elasticsearch database.

### Firestore

Still proof of concept, does not cover full integration with LAVA logs.
Patches are welcome.

In order to user Google firestore, you first have to create a project and
generate a private key file for the service account [here](https://console.firebase.google.com/project/_/settings/serviceaccounts/adminsdk).

The root collection is currently hardcoded to `logs`.

Firestore integration does not currently support limited read for each job,
meaning that incremental read on the job UI will not work properly.

## Log database migration

If the logging database is used from the very first installation of one LAVA
instance then all the logs will be stored in the said database, but if the
switch is made at one later point in the life of a LAVA instance, then you
can use the management command [copy-logs](https://git.lavasoftware.org/lava/lava/-/blob/master/lava_server/management/commands/copy-logs.py) to copy the logs
from the filesystem to the database of your choosing.
It is mandatory to have appropriate configuration variables set in LAVA settings in order to properly connect to the database.

!!! example "From filesystem to MongoDB"

    ```shell
    lava-server manage copy-logs LogsMongo
    ```

!!! example "From filesystel to ElasticSearch"

    ```shell
    lava-server manage copy-logs --dry-run LogsElasticsearch
    ```

!!! warning "copy-logs compatibility"
    Management command `copy-logs` is only applicable to the MongoDB and
    Elasticsearch database engines. Firestore is not yet supported.

# Hardware

LAVA supports both `amd64` and `arm64` architectures. It uses a
[server–worker](../../technical-references/architecture.md) model. The server
and worker hardware requirements described here are guidelines, not hard
limits. You can start with the recommendations, measure real load, and add
capacity when needed.

## LAVA server

A small LAVA instance can be deployed on fairly modest hardware.

We recommend a minimum of 4GB of RAM to cover the runtime needs of the database
server, the application server, and the web server. 8 GB RAM is preferred, as
PostgreSQL alone may consume 1–2 GB on a busy instance.

For storage, reserve about 80GB for job data to avoid frequent cleanup of
job logs. Keep in mind that storage requirements will grow over time. Unless
storage is unlimited, a job log retention or cleanup policy is necessary. SSDs
are strongly preferred, as spinning disk is a common performance bottleneck.

If you are deploying many devices and expect to be running large numbers of
concurrent jobs, you will obviously need more RAM and disk space.

## LAVA worker

### Classic LAVA worker

A LAVA worker can run on fairly modest hardware. Exact requirements depend on
the number of attached tests devices and concurrent jobs the worker is expected
to handle.

When a standard PC is used as a worker, several additional pieces of equipment
are needed to support test automation:

- PDU (Power Distribution Unit) for power-cycling DUTs remotely (e.g., APC PDU)
- USB serial adapters for serial console access to each DUT (e.g., FTDI-based adapters)
- Powered USB hub for switching DUT USB ports
- Ethernet switch for providing isolated network environment

Sourcing reliable, compatible equipment and debugging hardware-level issues
(e.g., flaky USB, power glitches, serial dropouts) can be surprisingly
time-consuming. An integrated solution like the [**LAA**](#linaro-automation-appliance)
is designed and well tested to eliminate most of that effort.

### Linaro Automation Appliance

The [**Linaro Automation Appliance (LAA)**](https://docs.lavacloud.io/) is a
fully integrated device testing solution. It provides everything needed by most
hardware automation tasks while also allowing to share DUTs with remote users
and customers.

The LAA integrates power control, USB switching, USB OTG (mass storage gadget),
serial console, and secure network connectivity into a single managed unit. This
solution significantly reduces per-device wiring complexity and avoids the
reliability issues commonly associated with assembling untested generic hubs
and PDUs.

The LAA is the recommended worker for efficient LAVA deployments. It significantly
accelerates test device enablement in LAVA while ensuring more repeatable and
reliable test results.

For more details, please refer to [LAA documentation][LAA].

## See also

- [LAVA architecture](../../technical-references/architecture.md)
- [LAVA deployment topology](./topology.md)
- [LAVA installation](../basic-tutorials/instance/install.md)
- [LAVA Docker worker](../advanced-tutorials/docker-worker.md)
- [Raspberry Pi4B as a LAVA worker](../advanced-tutorials/deploying-rpi4b-as-worker.md)

--8<-- "refs.txt"

# LTP

The [Linux Test Project (LTP)](https://github.com/linux-test-project/ltp)
provides tests to the open source community for validating reliability,
robustness, and stability of the Linux Kernel.

This example job demonstrates running the LTP `smoketest` suite on an `ampereone`
server. The `ltp.tar.xz` pre-built binaries are deployed to the root filesystem
as an overlay. A local NVMe partition mounted to `/scratch` is used as the LTP
TMPDIR.

```yaml
device_type: ampereone
job_name: arm64-ltp-smoke

visibility: public
priority: 80

timeouts:
  job:
    minutes: 60
  connection:
    minutes: 2
  actions:
    finalize:
      seconds: 60
    power-off:
      seconds: 60

context:
  arch: arm64
  extra_kernel_args: ' rw'

actions:
- deploy:
    to: tftp
    kernel:
      url: https://storage.tuxsuite.com/public/ampere/ci/builds/3BgK8gsSp9l33hXcgInsWQXEYEy/Image.gz
      compression: gz
      type: image
    nfsrootfs:
      url: https://storage.tuxboot.com/debian/20250722/trixie/arm64/rootfs.tar.xz
      compression: xz
      format: tar
      overlays:
        modules:
          url: https://storage.tuxsuite.com/public/ampere/ci/builds/3BgK8gsSp9l33hXcgInsWQXEYEy/modules.tar.xz
          compression: xz
          format: tar
          path: /usr/
        overlay-00:
          url: https://storage.tuxboot.com/overlays/debian/trixie/arm64/ltp/master/ltp.tar.xz
          compression: xz
          format: tar
          path: /
    os: debian
    timeout:
      minutes: 30

- boot:
    method: grub
    commands: nfs
    auto_login:
      login_prompt: 'login:'
      username: root
    prompts:
    - root@(.*):[/~]#
    - '/ #'
    timeout:
      minutes: 30

- test:
    definitions:
    - from: inline
      name: prep-inline
      path: inline/prep.yaml
      repository:
        metadata:
          format: Lava-Test Test Definition 1.0
          name: prep-tests
          description: Device preparation
        run:
          steps:
          - export STORAGE_DEV=/dev/nvme0n1p2
          - mkfs.ext4 -F "$STORAGE_DEV" || lava-test-raise "mkfs.ext4 $STORAGE_DEV
            failed; job exit"
          - mkdir -p /scratch && mount "$STORAGE_DEV" /scratch || lava-test-raise
            "mount $STORAGE_DEV failed; job exit"
          - df -h
          - mount
    timeout:
      minutes: 5

- test:
    definitions:
    - from: url
      name: ltp-smoke
      path: automated/linux/ltp/ltp.yaml
      repository: https://github.com/Linaro/test-definitions/releases/download/2025.10.01/2025.10.tar.zst
      compression: zstd
      lava-signal: kmsg
      parameters:
        ENVIRONMENT: production
        KIRK_WORKERS: 1
        LTP_INSTALL_PATH: /opt/ltp/
        LTP_TMPDIR: /scratch
        RUNNER: kirk
        SHARD_INDEX: 1
        SHARD_NUMBER: 1
        SKIPFILE: skipfile-lkft.yaml
        SKIP_INSTALL: 'true'
        TIMEOUT_MULTIPLIER: 5
        TST_CMDFILES: smoketest
    timeout:
      minutes: 5
```

See [ltp test definition](https://github.com/Linaro/test-definitions/blob/master/automated/linux/ltp/ltp.yaml)
for what each test parameter does.

# Upgrade LAVA

## When to upgrade

We advice to always upgrade to the latest released version. Only the latest
release will include new features and bug fixes (including security issues).

## How to upgrade

Currently, LAVA does not allow for zero-downtime upgrade.

We advice to put the instance in maintenance before any upgrade.

!!! example "Maintenance mode"

    ```shell tab="docker-compose"
    docker-compose exec lava-server lava-server manage maintenance
    ```

    ```shell tab="debian"
    lava-server manage maintenance
    ```

    ```shell tab="lavacli"
    lavacli system maintenance
    ```


### Docker

When a new version of LAVA is released, a Docker image is published on [docker hub][lava-docker-hub].

In order to upgrade, admins should just pull the latest **docker-compose
configuration**:

```shell
git pull --rebase
```

The restarting docker-compose will enough:

```shell
docker-compose pull
docker-compose up
```

??? tip "Downtime during upgrade"
    The current setup does not allow for zero-downtime upgrade without a
    management layer like [docker swarm](https://docs.docker.com/engine/swarm/)
    or [kubernetes](https://kubernetes.io/).

### Debian

When a new version of LAVA is released, a new Debian package is published on
the official [Debian repository][lava-debian].

You can upgrade using **apt**:

```shell
apt-get update
apt-get upgrade
```

??? tip "Downtime during upgrade"
    Zero-downtime upgrade is not possible with the current Debian packages.
    During the upgrade, every services will be stopped while upgrading.

## Upgrade notifications

Different version compatibility in LAVA between server and worker
instances is not guaranteed at the moment.

When a LAVA server instance is upgraded, if the specific flag in settings is
provided, the system will send out email notifications reminding admins that
the worker instances need to be upgraded as well.

```yaml
"MASTER_UPGRADE_NOTIFY": false
```

The default value for this flag is False so admin action is required in order
to use this feature.

Each user that has a 'change_worker' permission over the at least one worker
(which is not of the same version as the LAVA server) will get an email
with the list of workers which require an upgrade.
If the user is superuser, he will be regarded as having permission over all the
workers in the system, so that kind of user will get the full list of the
workers which require an upgrade.


## Rollback

!!! warning "TODO"

--8<-- "refs.txt"

# Timeouts

## Timeout syntax

A timeout is a dictionary with unit and value:

```yaml
timeout:
  <unit>: <value>
```

Valid units are `seconds`, `minutes`, `hours` and `days` while the value should be an integer.

## Job timeout

The entire job will have an overall timeout. The job will fail if this timeout
is exceeded, whether or not any other timeout is longer.

```yaml
timeouts:
  job:
    minutes: 15
```

## Action timeouts

Each action of a job will have a timeout. The action will not run longer than the
defined timeout.

### Job generic

In the job’s top-level `timeouts` block, you can set default action and individual
action timeouts.

#### Action timeout

Set the default duration of every action:

```yaml
timeouts:
  action:
    minutes: 5
```

#### Individual action timeout

Set the default duration of individual named action timeout:

```yaml
timeouts:
  actions:
    extract-nfsrootfs:
      minutes: 3
```

### Action block

You can define or override the job generic action timeouts in the action
definition block.

#### Action timeout

To define or override action timeout, you should add a `timeout` dictionary to
the action definition.

```yaml
actions:
- deploy:
    timeout:
      minutes: 5
    [...]
```

!!! warning "Action timeout"
    No action timeout can be larger than the job timeout. If the action timeout
    is larger than the job timeout, LAVA logs a warning and terminates the
    action when the job timeout is reached.

#### Individual action timeout

To define or override the individual named action timeout, you should
add a `timeouts` dictionary to the action definition.

```yaml
actions:
- deploy:
    timeout:
      minutes: 5
    timeouts:
      http-download:
        minutes: 1
```

### Repeatable action timeout

In order to have enough time to repeat or retry actions that have `repeat` or
`failure_retry` set (see [actions](./job.md#actions)), LAVA divides the timeout
by number of retries for each attempt.

For example:

```yaml
- boot:
   failure_retry: 4
   timeout:
     minutes: 20
```

This boot action will have a 5 minutes timeout for each of the 4 attempts. The
timeout can be further divided when a repeatable action is nested inside another
repeatable action.

### Timeout priority

Action timeout priority from lowest to highest:

1. [Job generic action timeout](#action-timeout)
2. Device individual action timeout
3. [Action block action timeout](#action-timeout_1)
4. [Repeatable action timeout](#repeatable-action-timeout)
5. [Job generic individual action timeout](#individual-action-timeout)
6. [Action block individual action timeout](#individual-action-timeout_1)

## Connection timeout

When an action interact with the DUT, this action will use the
`connection_timeout`. Each interaction with the DUT should take less than the
`connection_timeout` duration.

You can define the default duration of every connection timeout:

```yaml
timeouts:
  connection:
    seconds: 60
```

You can also define individual connection timeout in the root timeout dictionary:

```yaml
timeouts:
  connections:
    lava-test-shell:
      seconds: 60
```

## Queue timeout

You can specify how long the job can wait in the queue before it is
automatically canceled.

```yaml
timeouts:
  job:
    minutes: 60
  queue:
    hours: 72
```

## Skipping timeout

In some cases, a test action is known to hang or otherwise cause a timeout. If
the device is capable of booting again, then the first test action timeout can
set `skip: true` which stops the job finishing as `Incomplete` when the timeout
occurs. The job will then continue to the following `boot` actions, allowing the
device to reset to a known state and start the second test action. A second
`deploy` action might be also needed to get to a point where the device could
`boot` again.

```yaml hl_lines="4 8"
- test:
    timeout:
      minutes: 5
      skip: true
    definitions:
      [...]

- boot:
    timeout:
      minutes: 2
    [...]

- test:
    timeout:
      minutes: 5
    definitions:
      [...]
```

!!! note
    Skipping a test action timeout is intended for tests which may cause a
    kernel panic or other deadlock of the currently executing test shell. When a
    potential test timeout is expected, test writers should consider using
    utilities like the `timeout` command inside their own test shell scripts to
    retain control within the currently executing test shell scripts.

# State machine

The state machine describes and controls the state and health of Workers,
Devices and Test Jobs.

## Workers

For each worker, two variables describe the current status: **state** and **health**.

### State

The **state** is an internal variable, set by [lava-server-gunicorn](./services/lava-server-gunicorn.md) when receiving (or not) pings from each worker.

* *Online*: the worker is sending PING to the server
* *Offline*: the worker hasn't sent any messages for a while

### Health

The **health** can be used by admins to control the **health** of all attached devices.

For instance, when set to *Maintenance*, all attached devices will be automatically put into maintenance mode so that no jobs will be scheduled on those devices.

The worker health can be:

* *Active*
* *Maintenance*
* *Retired*


!!! warning
    When a worker is *Offline*, none of the attached devices will be used to schedule new jobs.

## Devices


For each device, two variables describe the current status: **state** and **health**.

### State

The **state** is an internal variable, set by [lava-scheduler](./services/lava-scheduler.md) and [lava-server-gunicorn](./services/lava-server-gunicorn.md) when scheduling, starting, canceling and ending test jobs.

* *Idle*: not in use by any test job
* *Reserved*: has been reserved for a test job but the test job is not running yet
* *Running*: currently running a test job

### Health

The **health** can be used by admins to indicate if a device should be used by the scheduler or not.

Moreover, when ending an health-check, the device health will be set according to the test job health.

* *Good*: the device passed the health-check
* *Unknown*
* *Looping*: should run health-checks in a loop
* *Bad*: the device failed the health-check
* *Maintenance*
* *Retired*


## TestJobs

For each test job, two variables are describing the current status: **state** and **health**.

### State

* *Submitted*: waiting in the queue
* *Scheduling*: part of a multinode test job where some sub-jobs are still in *Submitted*
* *Scheduled*: has been scheduled. For multinode, it means that all sub-jobs are also scheduled
* *Running*: currently running on a device
* *Canceling*: has been canceled but not ended yet
* *Finished*

!!! note Multinode scheduling
    Only multinode test jobs use *Scheduling*. When all
    sub-jobs are in *Scheduling*, [lava-scheduler](./services/lava-scheduler.md) will transition all test
    jobs to *Scheduled*.

### Health

* *Unknown*: default value that will be overridden when the job is finished.
* *Complete*: the job was able to finish
* *Incomplete*: the job was not able to dinish
* *Canceled*: the test job was canceled.


# Scheduler

The scheduler is called by [lava-scheduler](./services/lava-scheduler.md)
approximately every 20 seconds or when receiving specific events.
The scheduler starts by scheduling health-checks. The remaining devices are
then considered for test jobs.

## Health-checks

To ensure that health-checks are always scheduled when needed, they will be
considered first by the scheduler before regular test jobs.

The scheduler will only consider devices where:

* `device state` is *Idle*
* `device health` is *Good*, *Unknown* or *Looping*
* `worker state` is *Online*

## Test jobs

The scheduler will only consider devices where:

* ``device state`` is *Idle*
* ``device health`` is *Good* or *Unknown*
* ``worker state`` is *Online*

# Synchronizing records from device dictionary

LAVA has the ability to synchronize database records with the device
dictionary. It currently supports these models:

* Device
* Device type
* Tag
* Alias

This can make administration less cumbersome and can help with Ansible and
similar setups. By using this feature administrators are able to keep the list
of these records inside the version control.

For device records, a flag named **is_synced** is used to recognize devices
which are synced to/from the device dictionary. This option can be updated via
usual channels (web UI admin, APIs).

Once the new device dictionary is added to the filesystem it needs to include
the following dictionary variable which will control which records should be
created:

```jinja
{% set sync_to_lava = {
   "device_type": "qemu",
   "worker": "worker-01"
   "tags": ["tag1", "tag2"],
   "aliases": ["alias1", "alias2"],
   }
%}
```

A management command is used to perform a sync to the database:

```shell tab="debian"
lava-server manage sync
```

```shell tab="docker-compose"
docker-compose exec lava-server lava-server manage maintenance
```

You can for example set this to run when the contents of the device dictionary
folder change using a watch utility.

In the above example, a system will create:

* device named after the device dictionary file
* device type named qemu (if it doesn't exist already) and relate this device to it
* worker named worker-01 (if it doesn't exist already) and relate this device to it
* aliases in the list; and relate them to the specified device type
* tags in the list; and relate them to the device

If at any point the device dictionary is removed (and once the sync command is
executed) the system will move the device that was created to it to **Retired**
health. If all the devices of the synced device type are **Retired** that
particular device type will be marked as invisible in the system.

# MPS

The `mps` deployment method is used to write a recovery image and test binaries
to the MPS devices. The support is similar to the [VEMSD](./to-vemsd.md).

```yaml
  - deploy:
      to: mps
      images:
        recovery_image:
          url: mps2_sse200_an512.tar.gz
          compression: gz
        test_binary_1:
          url: tfm_sign.bin
        test_binary_2:
          url: mcuboot.bin
```

## recovery_image

The image ready to be unpacked onto the USB filesystem of the MPS device.

## test_binary

Download test binary to MPS device and rename if required.

Multiple test binaries can be flashed in the same deploy action by listing all
of them. The keys should start with `test_binary_`.

### rename

Renames the test_binary if required.

If the `recovery_image` expects to flash a specific image and the file
downloaded is not named this way, this option will save it with a different
name on the board.

If not specified, the test_binary is copied as-is, no renaming takes place.

```yaml
  - deploy:
      to: mps
      images:
        recovery_image:
          url: mps2_sse200_an512.tar.gz
          compression: gz
        test_binary_1:
          url: tfm_sign_20191121.bin
          rename: tfm_sign.bin
```

## url

See [url](./index.md#url)

## compression

If the image is compressed, the compression method **must** be specified.

See [compression](./index.md#compression)

# Core Features

## Automated validation

LAVA is designed for automated workflows to create, submit, and process test
jobs to validate the development process.

## Distributed architecture

LAVA uses a distributed architecture built around a central LAVA server and
multiple LAVA workers.

LAVA workers can be deployed behind firewalls, allowing organizations to
integrate internal test infrastructure while preserving network isolation and
security. Labs that host LAVA workers and DUTs can be deployed remotely as
remote labs.

This design enables horizontal scaling and flexible deployment across
geographically distributed and network-restricted environments.

## Privacy support

LAVA test jobs or device types can be kept private to selected groups,
individuals or teams. Entire LAVA instance can also be configured as private.
A private instance requires authentication for all access.

## Parallel scheduling

Multiple test jobs run at the same time across multiple devices.

## MultiNode test jobs

LAVA test jobs can be run as a single group of tests involving multiple devices.

## Containerized testing control

Test jobs can run arbitrary code in isolated containers on the worker to control
test devices. The feature enables single node jobs for scenarios that used to
require complex MultiNode test jobs.

## Hardware sharing

Uncommon hardware is shared between disparate groups to maximize usage. Remote
access via hacking sessions allows developers to interactively debug on physical
devices.

## Wide device coverage

A large number of device types can be supported, with instances ranging from one
to more than a hundred devices available for test jobs.

## Multi-platform support

Test jobs can be run on systems running various UNIX flavors, Android, or IoT
platforms like Zephyr. LAVA supports a wide range of deploy, boot, and test
methods to accommodate testing with different platforms.

## Complex network testing

LAVA supports testing scenarios that require reconfigurable networking across
multiple devices, including VLAN configuration and multiple network interfaces.

## Live result reporting

If a test job fails, all results up to the point of failure are retained.

## Result Queries and charts

LAVA result queries allow user to track test results over time. LAVA result
charts can visualize trends in test pass rates, performance measurements, and
other metrics.

## Flexible notifications

LAVA job completion can trigger notifications via email, IRC, or webhook
callbacks.

## Data export for customization

LAVA provides APIs for exporting test job data to support custom frontends and
dashboards tailored to specific needs.

## REST API

LAVA provides REST APIs alongside the `lavacli` client, enabling seamless
integration with CI systems and supporting LAVA instance management through
Infrastructure as Code (IaC).

--8<-- "refs.txt"

---
title: The Art of Influence
subtitle: A Practical Guide to Working with Influence Functions 
author: Klaus Kähler Holst
Rdate: "`r Sys.Date()`"
output:
  # html_document:  
  rmarkdown::html_vignette:
    # toc_float: true
    fig_width: 6
    fig_height: 4
    fig_caption: true
    toc: true
    toc_depth: 2
    mathjax: "https://cdnjs.cloudflare.com/ajax/libs/mathjax/2.7.5/MathJax.js?config=TeX-AMS_CHTML.js"
vignette: >
  %\VignetteIndexEntry{The Art of Influence}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
bibliography: ref.bib
---

```{r  setup,include=FALSE }
library("lava")
knitr::opts_chunk$set(
  collapse = TRUE,
  cache = TRUE,
  comment = "#>",
  dev = "svg",
  fig.ext = "svg"
  )
has_mets <- lava:::versioncheck('mets', 1)
has_geepack <- lava:::versioncheck('geepack', 1)
```

\newcommand{\arctanh}{\operatorname{arctanh}}
\newcommand{\indep}{\!\perp\!\!\!\!\perp\!} 
\newcommand{\pr}{\mathbb{P}}
\newcommand{\E}{\mathbb{E}}
\newcommand{\var}{\mathbb{V}\!\text{ar}}
\newcommand{\cov}{\mathbb{C}\!\text{ov}}
\newcommand{\cor}{\mathbb{C}\!\text{or}}
\newcommand{\IC}{\operatorname{IC}}
\newcommand{\one}{\mathbf{1}}
<!-- \newcommand{\Dto}{\rightsquigarrow} -->
\newcommand{\Dto}{\overset{\mathcal{D}}{\longrightarrow}}
\newcommand{\bm}[1]{\mathbf{#1}}
\newcommand{\Pz}{P_{0}}
\newcommand{\op}{o_{P}}

# Influence functions

Estimators that have parametric convergence rates can often be fully
characterized by their *influence function* (IF), also referred to as an influence
curve or canonical gradient [@bickel_effic_adapt_estim_semip_model; @vaart_1998_asymp]. The IF allows for the direct estimation of properties of the estimator, including its asymptotic variance. Moreover,
estimates of the IF enable the simple combination and transformation of
estimators into new ones. 
This vignette describes how to estimate and manipulate IFs using the R-package
`lava` [@holst_budtzjorgensen_2013].


Formally, let $Z_1,\ldots,Z_n$ be iid $k$-dimensional stochastic variables, $Z_i=(Y_{i},A_{i},W_{i})\sim \Pz$, and $\widehat{\theta}$ a
consistent estimator for the parameter $\theta\in\mathbb{R}^p$. When
$\widehat{\theta}$ is a *regular and asymptotic linear* (RAL) estimator, it has
a unique iid decomposition 
\begin{align*} \sqrt{n}(\widehat{\theta}-\theta) =
\frac{1}{\sqrt{n}}\sum_{i=1}^n \IC(Z_i; \Pz) + \op(1), 
\end{align*} 
where the
function $\IC$ is the unique *Influence Function* s.t. $\mathbb{E}\{\IC(Z_{i};
\Pz)\}=0$ and $\var\{\IC(Z_{i}; \Pz)^{2}\}<\infty$ [@tsiatis2006semiparametric;
@vaart_1998_asymp]. The influence
function thus fully characterizes the asymptotic behaviour of the estimator and by the central limit
theorem it follows that the estimator converges weakly to a Gaussian distribution
$$
\sqrt{n}(\widehat{\theta}-\theta) \Dto
\mathcal{N}(0, \var\{\IC(Z; \Pz)\}),
$$
where the empirical variance of the plugin estimator, $\pr_{n}\IC(Z;
\widehat{P})^{\otimes 2} = \frac{1}{n}\sum_{i=1}^n \IC(Z_{i};
\widehat{P})\IC(Z_{i};
\widehat{P})^{\top}$ can be used to obtain a consistent estimate of the
asymptotic variance. Note, in practice the estimate $\widehat{P}$ used in
the plugin-estimate, needs only to capture the parts of the distribution of
$Z$ that is necessary to evaluate the IF. In some cases this nuisance
parameter can be estimated using flexible machine learning components and in
other (parametric) cases be derived directly from $\widehat{\theta}$.

The IFs are easily derived for the parameters of many parametric statistical
models as illustrated in the [next example sections](#example-generalized-linear-model). More generally, the IF can
also be derived for a smooth target parameter $\Psi:
\mathcal{P}\to\mathbb{R}$ where $\mathcal{P}$ is a family of probability
distributions forming the statistical model, which often can be left completely
non-parametric. 
Formally, the parameter must be *pathwise differentiable* see [@vaart_1998_asymp]
in the sense that there exists linear bounded function  $\dot\Psi\colon
L_{2}(P_{0})\to\mathbb{R}$ such that
$[\Psi(P_{t}) - \Psi(P_{0}))]t^{-1} \to \dot\Psi(P_{0})(g)$
as $t\to 0$ for any parametric submodel $P_t$
with score model $g(z)= \partial/(\partial t) \log (p_t)(z)|_{t=0}$.  Riesz's
representation theorem then tells us that the directional derivative has a 
unique representer, $\phi_{P_{0}}$ lying in the closure of the submodel score
space (the *tangent space*), s.t.
\begin{align*}
    \dot\Psi(P_0)(g) = \langle\phi_{P_0}, g\rangle =
    \int \phi_{P_0}(Z)g(X)\,dP_0
\end{align*}
The unique representer is exactly the IF which can be found by solving the above
integral equation.
For more details on how to derive influence
functions, we refer to [@targetedlearning_2011; @hines2022].

As an example we might be interested in the target parameter $\Psi(P) =
\E_P(Z)$ which can be shown to have the unique (and thereby efficient)
influence function $Z\mapsto Z-\E_P(Z)$ under the non-parametric model.
Another target parameter could be $\Psi_{a}(P) = \E_{P}[\E_{P}(Y\mid A=a, W)]$ which is often
a key interest in causal inference and which has the IF 
\begin{align*}
\IC(Y,A,W; P) = \frac{\one(A=a)}{\pr(A=a\mid W)}(Y-\E_{P}[Y\mid A=a,W]) + \E_{P}[Y\mid
A=a,W] - \Psi_{a}(P)
\end{align*}
See section on [average treatment effects](#average-treatment-effects).

# Working with influence functions with `lava::estimate`

The main functions for working with influence functions are

- `estimate` which prepares a model object and estimates the IF and
  corresponding robust standard errors. Can also be used to transform model
  parameters by application of the Delta Theorem.
- `merge`  method for combining estimates via their estimated IFs
- `IC`, `coef`, `vcov` methods to extract the estimated IF, coefficients, and
  asymptotic covariance matrix
- `subset` for extracting subset of parameters and IFs
- `transform` method for transforming parameter estimates
- `labels` for renaming parameters

The `estimate` function is the primary tool for obtaining parameter estimates
and related information. It returns an object of the class type `estimate`,
which is a general container for holding information about estimated parameters.
The estimate function takes as input either a model object (the first argument
`x`), or a parameter vector and corresponding influence function (IF) matrix
specified using the `coef` and `IC` arguments. If the primary goal is to apply the
delta method or test linear hypotheses, it is also possible to provide the
asymptotic variance estimate via the `vcov` argument, without specifying the IC
matrix.

```{r estimate.syntax, eval=FALSE}
estimate(x=, ...) 
estimate(coef=, IC=, ...)
estimate(coef=, vcov=, ...)
```

A typical call could look like
```{r estimate.examples, eval=FALSE}
merge(subset(estimate(x), 1), estimate(coef=p, IC=ic, id=id)) |> 
  transform(function(x) c(exp(x), exp(x[1]))) |> # parameter transformation
  labels(c("a", "b")) # rename parameters
```

Direct calculations is also possible, if `a` and `b` are estimate objects we can
obtain new transformed parameter estimates with the syntax
```{r estimate.direct, eval=FALSE}
c(exp(b)^0.5, exp(b * a) / (1 + exp(-b)))
```
with the necessary derivatives exactly calculated automatically.
We refer to section
[IF building blocks](#if-building-blocks-transformations-and-the-delta-theorem) for more details.


# Examples
    
To illustrate the methods we consider data arising from the model $Y_{ij} \sim
Bernoulli\{\operatorname{expit}(X_{ij} + A_{i} + W_{i})\}, A_{i} \sim
Bernoulli\{\operatorname{expit}(W_{i})\}$
with independent covariates
$X_{ij}\sim\mathcal{N}(0,1), W_{i}\sim\mathcal{N}(0,1)$.
```{r sim_model}
m <- lvm() |>
  regression(y1 ~ x1 + a + w) |>
  regression(y2 ~ x2 + a + w) |>
  regression(y3 ~ x3 + a + w) |>
  regression(y4 ~ x4 + a + w) |>
  regression(a ~ w) |>
  distribution(~ y1 + y2 + y3 + y4 + a, value = binomial.lvm()) |>
  distribution(~id, value = Sequence.lvm(integer = TRUE))
```

We simulate from the model where $Y_3$ is only observed for half of the subjects
```{r simulate}
n <- 4e2
dw <- sim(m, n, seed = 1) |>
  transform(y3 = y3 * ifelse(id > n / 2, NA, 1))
Print(dw)
## Data in long format
dl <- reshape(dw,
        varying = list(paste0("y",1:4),
                       paste0("x",1:4)),
        v.names=c("y", "x"), direction="long") |>
  na.omit()
dl <- dl[order(dl$id), ]
## dl <- mets::fast.reshape(dw, varying = c("y", "x")) |> na.omit()
Print(dl)
```

## Example: population mean

Here we first consider the problem of estimating the IF of the mean. For a
general transformation $f: \mathbb{R}^k\to\mathbb{R}^p$ we have that
$$
\sqrt{n}\{\mathbb{P}_{n}f(X) - \E[f(X)]\} = \frac{1}{\sqrt{n}}\sum_{i=1}^{n} 
f(X_{i}) - \E[f(X)]
$$
<!-- $$
<!-- \sqrt{n}\{\mathbb{P}_{n}f(X) - \pr(Y_{1}=1)\} = \frac{1}{\sqrt{n}}\sum_{i=1}^{n} -->
<!-- \one(Y_{1i}=1) - \pr(Y_{1}=1) -->
<!-- $$  -->
and hence for the problem of estimating the proportion of the binary outcome
$Y_1$, the IF is given by $\one(Y_{1}=1) - \pr(Y_{1}=1)$.

To estimate this parameter and its IF we will use the `estimate` function
```{r inp1}
inp <- as.matrix(dw[, c("y1", "y2")])
e <- estimate(inp[, 1, drop = FALSE], type="mean") 
class(e)
e
```
The reported standard errors from the `estimate` method
are the robust standard errors obtained from the IF.
The variance estimate and the parameters can be extracted with the `vcov`
and `coef` methods.
The IF itself can be extracted with the `IC` (or  `influence`) method:
```{r ic1}
IC(e) |> Print()
```

It is also possible to simultaneously estimate the proportions of each of the two
binary outcomes 
```{r inp2}
estimate(inp)
```
or alternatively the input can be a model object, here a `mlm` object:
```{r mlm}
e <- lm(cbind(y1, y2) ~ 1, data = dw) |>
  estimate()
IC(e) |> head()
```

Different methods are available for inspecting an `estimate`  object
```{r estimatemethods}
summary(e)
## extract parameter coefficients
coef(e)
## ## Asymptotic (robust) variance estimate
vcov(e)
## Matrix with estimates and confidence limits
estimate(e, null=0, level = 0.99) |> parameter()
## Influence curve
IC(e) |> head()
## Join estimates
ee <- merge(e, e)
ee
## Forest plots
plot(ee, null=0.5, digits=2)
```

<!-- The IF for both the empirical mean and variance can also be estimated directly with the `IC` method: -->
<!-- ```{r} -->
<!-- with(dw, IC(y1)) |> head() -->
<!-- ## point estimates stores in the attributes -->
<!-- with(dw, IC(y1)) |> attributes() -->
<!-- ``` -->


## Example: generalized linear model

For a $Z$-estimator defined by the score equation $E[U(Z; \theta)] = 0$, the
IF is given by \begin{align*} IC(Z; \theta) =
-\E\Big\{\frac{\partial}{\partial\theta^\top}U(\theta; Z)\Big\}^{-1}U(Z; \theta)
\end{align*}
In particular, for a maximum likelihood estimator the score, $U$, is given by the partial derivative of
the log-likelihood function.

As an example, we can obtain the estimates with robust standard errors for a
logistic regression model:
```{r glm}
g <- glm(y1 ~ a + x1, data = dw, family = binomial)
estimate(g)
```
We can compare that to the usual (non-robust) standard errors:
```{r glm.std}
estimate(g, robust = FALSE)
```

The IF can be extracted from the `estimate` object or directly from the 
model object
```{r ifglm}
IC(g) |> head()
```


The same estimates can be obtained with a 
*cumulative link regression* model which also generalizes to 
ordinal outcomes. Here we consider the proportional odds model given by 
\begin{align*}
\log\left(\frac{\pr(Y\leq j\mid x)}{1-\pr(Y\leq j\mid x)}\right) = \alpha_{j} + \beta^{t}x, \quad j=1,\ldots,J
\end{align*}
```{r ordreg}
ordreg(y1 ~ a + x1, dw, family=binomial(logit)) |> estimate()
```

Note that the `sandwich::estfun` function from the `sandwich` library [@r_sandwich] can also estimate the IF
for different parametric models, but does not provide the tools for combining
and transforming these.

## Example: right-censored outcomess

To illustrate the methods on survival data we will use the Mayo Clinic Primary
Biliary Cholangitis Data [@therneau00surv]

```{r mets}
library("survival")
data(pbc, package="survival")
```

The Cox proportional hazards model can be fitted with the `mets::phreg` method
which can estimate the IF for both the partial likelihood parameters and the
baseline hazard. Here we fit a survival model with right-censored event times
```{r phreg, eval=has_mets }
fit.phreg <- mets::phreg(Surv(time, status > 0) ~ age + sex, data = pbc)
fit.phreg
IC(fit.phreg) |> head()
```

The IF for the baseline cumulative hazard at a specific time point
\begin{align*}
\Lambda_0(t) = \int_0^t \lambda_0(u)\,du, 
\end{align*}
where $\lambda_0(t)$ is the baseline hazard, can be estimated in similar way:
```{r phreg-baseline, eval=has_mets }
baseline <- function(object, time, ...) {
  ic <- mets::IC(object, baseline = TRUE, time = time, ...)
  est <- mets::predictCumhaz(object$cumhaz, new.time = time)[1, 2]
  estimate(NULL, coef = est, IC = ic, labels = paste0("chaz:", time))
}
tt <- 2000
baseline(fit.phreg, tt)
```

The `estimate` and `IF` methods are also available for parametric survival
models via `survival::survreg`, here a Weibull model:
```{r survreg, eval=has_mets }
survival::survreg(Surv(time, status > 0) ~ age + sex, data = pbc, dist="weibull") |>
  estimate()
```

<!-- glm poisson -->
<!-- ```{r lifetable} -->
<!-- pbc2 <- mets::lifetable(Surv(time, status > 0) ~ sex, data = pbc, breaks = seq(0, 5000, length.out = 10)) -->
<!-- head(pbc2) -->
<!-- glm(events ~ offset(log(atrisk)) + sex + factor(int.end), data = pbc2, family=poisson(log)) -->
<!-- ``` -->

  
## Example: random effects model / structural equation model

General structural equation models (SEMs) can be estimated with `lava::lvm`. 
Here we fit a random effects probit model
$$
\pr(Y_{ij} = 1 \mid U_{i}, W_{ij})=\Phi(\mu_{j} + \beta_{j} W_{ij} + U_{i}), \quad U_{i}\sim\mathcal{N}(0,\sigma_{u}^{2}),\quad j=1,2
$$
to the simulated dataset
```{r semfit, eval=has_mets }
sem <- lvm(y1 + y2 ~ 1 * u + w) |>
  latent(~ u) |>
  ordinal(K=2, ~ y1 + y2)
semfit <- estimate(sem, data = dw)

## Robust standard errors
estimate(semfit)
```

## Example: quantile

Let $\beta$ denote the $\tau$th quantile of $X$, with IF
\begin{align*}
\IC(x; \Pz) = \tau - \one(x\leq \beta)f_{0}(\beta)^{-1} 
\end{align*}

where $f_{0}$ is the density function of $X$.

To calculate the variance estimate, an estimate of the density is thus needed which can be
obtained by a kernel estimate. Alternatively, 
the resampling method of [@zenglin2008] can be applied.
Here we use a kernel smoother (additional arguments to the `estimate` function
are parsed on to `stats::density.default`) to estimate the quantiles and IF for 
the 25%, 50%, and 75% quantiles of $W$ and $X_1$
```{r quantiles}
eq <- estimate(dw[, c("w", "x1")], type = "quantile", probs = c(0.25, 0.5, 0.75))
eq
IC(eq) |> head()
```


# Combining influence functions

A key benefit of working with the IFs of estimators is that this allows
for transforming or combining different estimates while easily deriving
the resulting IF and thereby asymptotic distribution of the new estimator. 

Let $\widehat{\theta}_{1}, \ldots, \widehat{\theta}_{M}$ be $M$
different estimators with decompositions
\begin{align*}
\sqrt{n}(\widehat{\theta}_{m}-\theta_{m}) = \frac{1}{\sqrt{n}}\sum_{i=1}^{n}
\IC_m(Z_i; \Pz) + \op(1)
\end{align*}
based on iid data $Z_1,\ldots,Z_n$. It then follows immediately
[@vaart_1998_asymp Theorem 18.10[vi]] that the
joint distribution of $\widehat{\theta} - {\theta}=
(\widehat{\theta}_{1}^{\top},\ldots,\widehat{\theta}_{M}^{\top})^\top-
({\theta}_{1}^{\top},\ldots,{\theta}_{M}^{\top})^\top
$ is given by
\begin{align*}
\sqrt{n}(\widehat{\theta}-\theta) &= \frac{1}{\sqrt{n}}\sum_{i=1}^{n}
\underbrace{[\IC_{1}(Z_i; \Pz)^\top,\ldots,\IC_{M}(Z_i; \Pz)^\top]^{\top}}_{\overline{\IC}(Z_i; \Pz)} + \op(1) \\
&\Dto \mathcal{N}(0,\Sigma)
\end{align*}
by the CLT, and under regulatory conditions
$\mathbb{P}_{n}\overline{\IC}(Z_i; \widehat{P})^{\otimes 2} \overset{P}{\longrightarrow}\Sigma$
as $n\to\infty$.

To illustrate this we consider two marginal logistic regression models fitted
separately for $Y_1$ and $Y_2$ and combine the estimates and IFs using the
`merge` method
```{r glmmarg}
g1 <- glm(y1 ~ a, family=binomial, data=dw)
g2 <- glm(y2 ~ a, family=binomial, data=dw)
e <- merge(g1, g2)
summary(e)
```
As we have access to the joint asymptotic distribution we can for example test
for whether the odds-ratio is the same for the two responses: 

```{r hypo1}
estimate(e, cbind(0,1,0,-1), null=0)
``` 
More details an be found in the Section on [hypothesis testing](#linear-contrasts-and-hypothesis-testing).

## Imbalanced data

Let $O_{1} = (Z_{1}R_{1}, R_{1}), \ldots, O_{N}=(Z_{N}R_{N}, R_{N})$ be iid with
$R_{i}\indep Z_i$ and let the full-data IF for some estimator of a parameter
$\theta\in\mathbb{R}^p$ be $IC(\cdot; \Pz)$. For convenience let the data be ordered
$R_{i}=\one(i\leq n)$ where $n$ is the number of observed data points, then the complete-case estimator is consistent and based
on same IF

\begin{align*}
\sqrt{n}(\widehat{\theta}-\theta) = \frac{1}{\sqrt{n}}\sum_{i=1}^n IC(Z_i; \Pz) + \op(1).
\end{align*}

This estimator can also be decomposed in terms of the observed data
$O_1,\ldots,O_N$ noting that

\begin{align*}
\sqrt{N}(\widehat{\theta}-\theta) = \frac{1}{\sqrt{N}}\sum_{i=1}^N IC(Z_i;
P)\frac{R_i N}{n} + \op(1).
\end{align*}

where the term $\frac{R_i N}{n}$ corresponds to an inverse probability
weighting with the empirical plugin estimate of the proportion of observed data $R=1$.
Under a missing completely at random assumption we can therefore combine estimators that are estimated
on different datasets. Let the observed data be
$(Z_{11}R_{11}, R_{11}, Z_{21}R_{21}, R_{21}), \ldots, (Z_{1N}R_{1N}, R_{1N},
Z_{2N}R_{2N}, R_{2N}))$ with complete-case estimators $\widehat{\theta}_1$ and
$\widehat{\theta}_2$ for parameters $\theta_1$ and
$\theta_2$ based on $(Z_{11}R_{11}, \ldots, Z_{1N}R_{1N})$ and
$(Z_{21}R_{21}, \ldots, Z_{2N}R_{2N})$, respectively, and let the
corresponding IFs be
$IC_{1}(\cdot; \Pz)$ and $IC_{2}(\cdot;\ P)$. It then follows that

\begin{align*}
\sqrt{N}\left\{
\begin{pmatrix}
\widehat{\theta}_1 \\
\widehat{\theta}_2 
\end{pmatrix}
- 
\begin{pmatrix}
\vphantom{\widehat{\theta}_1}\theta_1 \\
\vphantom{\widehat{\theta}_1}\theta_2 
\end{pmatrix}
\right\}
= 
\frac{1}{\sqrt{N}}\sum_{i=1}^N 
\begin{pmatrix}
IC_1(Z_{1i}; \Pz)\frac{R_{1i}N}{R_{1\bullet}} \\
IC_2(Z_{2i}; \Pz)\frac{R_{2i}N}{R_{2\bullet}}
\end{pmatrix} 
+ \op(1)
\end{align*}

with $R_{k\bullet} = \sum_{i=1}^{N}R_{ki}.$ Returning to the example, we can
combine the marginal estimates of two model objects that have been estimated
from different datasets (as the outcome $Y_3$ is only available in half of the
data) with the `merge` function

```{r glmmargmis}
g2 <- glm(y2 ~ 1, family = binomial, data = dw)
summary(g2)
dwc <- na.omit(dw) 
g3 <- glm(y3 ~ 1, family = binomial, data = dwc)
summary(g3)

e2 <- estimate(g2, id = dw$id)
e3 <- estimate(g3, id = "id", data=dwc)

ecomb <- merge(e2, e3)
IC(ecomb) |> Print()
vcov(ecomb)
## Same marginals as
list(vcov(e2), vcov(e3))
```

Note, it is also possible to directly specify the id-variables in the `merge` call:
```{r merge}
merge(e2, e3, id = list(dw$id, dwc$id))
```

In the above example the `id` argument defines the identifier that makes it
possible to link the rows in the different IFs that should be glued together. 
If omitted then the `id` will automatically be extracted from the model-specific
`IC` method (deriving it from the original data.frame used for estimating the
model). This automatically works with all models and `IC` methods described in this document.
```{r estimatenoid}
estimate(g2) |>
  IC() |> head()
vcov(merge(estimate(g2), estimate(g3)))
merge(estimate(g2), estimate(g3)) |>
  (rownames %++% head %++% IC)()
```

To force that the id variables are not overlapping between the merged model
objects, i.e., assuming that there is complete independence between the
estimates, the argument `id=NULL` can be used
```{r merge_idnull}
merge(g1, g2, id = NULL) |> (Print %++% IC)()
merge(g1, g2, id = NULL) |> vcov()
```

## Renaming and subsetting parameters

To only keep a subset of the parameters the `keep` argument can be used.  
```{r mergekeep}
merge(g1, g2, keep = c("(Intercept)", "(Intercept).1"))
```
The argument can be given either as character vector or a vector of indices:
```{r merge2}
merge(g1, g2, keep=c(1, 3))
```
or as a vector of perl-style regular expressions 
```{r merge3}
merge(g1, g2, keep = "cept", regex = TRUE)
merge(g1, g2, keep = c("\\)$", "^a$"), regex = TRUE, ignore.case = TRUE)
```

When merging estimates unique parameter names are created. It is also possible
to rename the parameters with the `labels` argument 
```{r merge4}
merge(g1, g2, labels = c("a", "b", "c")) |> estimate(keep = c("a", "c"))
merge(g1, g2,
      labels = c("a", "b", "c"),
      keep = c("a", "c")
)
estimate(g1, labels=c("a", "b"))
```

Finally, the `subset` argument can be used to subset the parameters and IFs
before the actual merging is being done 
```{r merge5}
merge(g1, g2, subset="(Intercept)")
```

## Clustered data (non-iid case)

Let $Z_i = (Z_{i1},\ldots,Z_{iN_{i}})$ and assume that
$(Z_{i}, N_{i}) \sim P$, $i=1,\ldots,n$ are iid and $N_i\indep
Z_{ij}$. The variables $Z_{i1},\ldots,Z_{iN_{i}}$ we assume are exchangeable
but not necessarily independent. Define $N = \sum_{i=1}^{n} N_i$, and
assume that a parameter estimate, $\widehat{\theta}\in\mathbb{R}^p$ has the decomposition
$$
\sqrt{N}(\widehat{\theta}-\theta) =
\frac{1}{\sqrt{N}}
\sum_{i=1}^{n} \sum_{k=1}^{N_{i}} IC(Z_{ik}; \Pz) + \op(1).
$$
It then follows that 
$$
\sqrt{n}(\widehat{\theta}-\theta) = 
\frac{1}{\sqrt{n}}
\sum_{i=1}^{n} \widetilde{\IC}(Z_{i}; \Pz) + \op(1)
$$
with $\widetilde{\IC}(Z_{i}; \Pz)
= \sum_{k=1}^{N_{i}} \frac{n}{N}IC(Z_{ik}; \Pz)$, $i=1,\ldots,n$ which are iid
an therefore admits the usual CLT to derive the asymptotic variance of $\widehat{\theta}$.
Turning back to the example data, we can estimate the marginal model
```{r cluster1}
g0 <- glm(y ~ a + w + x, data = dl, family = binomial())
```
The asymptotic variance estimate ignoring that the observations are not
independent is not consistent. Instead we can calculate the cluster robust
standard errors from the above iid decomposition
```{r cluster2}
estimate(g0, id=dl$id)
```
We can confirm that this situation is equivalent to the variance estimates we 
obtain from a GEE marginal model with working independence correlation structure [@r_geepack]
```{r geepack, eval=has_geepack }
gee0 <- geepack::geeglm(y ~ a + w + x, data = dl, id = dl$id, family=binomial)
summary(gee0)
```

## Computational aspects 

Working with large and potentially multiple different IFs can be memory-intensive. A
remedy is to use the idea of aggregating the IFs by introducing a random
coarser grouping variable. Following the same arguments as in the previous
section, the aggregated IF will still be iid and allows us to estimate the
asymptotic variance. Obviously, the same grouping must be used across estimates
when combining IFs.   
```{r aggregate}
set.seed(1)
y <- cbind(rnorm(1e5))
N <- 2e2 ## Number of aggregated groups, the number of observations in the new IF
id <- foldr(nrow(y), N, list=FALSE)
Print(cbind(table(id)))

## Aggregated IF
e <- estimate(cbind(y), id = id) 
object.size(e)
e
```

# IF building blocks: transformations and the delta theorem

Let $\phi\colon \mathbb{R}^p\to\mathbb{R}^m$ be differentiable at $\theta$
and assume that $\widehat{\theta}_n$ is RAL estimator with IF given by
$\IC(\cdot; \Pz)$ such that
\begin{align*} 
\sqrt{n}(\widehat{\theta}_n - \theta) =
\frac{1}{\sqrt{n}}\sum_{i=1}^n \IC(Z_i; \Pz) + \op(1), 
\end{align*} 
then by the delta method [@vaart_1998_asymp Theorem 3.1]
\begin{align*} 
\sqrt{n}\{\phi(\widehat{\theta}_n) - \phi(\theta)\}  =
\frac{1}{\sqrt{n}}\sum_{i=1}^n \nabla\phi(\theta)\IC(Z_i; \Pz) + \op(1), 
\end{align*} 

where $\phi\colon \theta\mapsto
(\phi_{1}(\theta),\ldots,\phi_{m}(\theta))^\top$ and $\nabla$ is the partial
derivative operator
\begin{align*}
\nabla\phi(\theta) = 
\begin{pmatrix}
\tfrac{\partial}{\partial\theta_1}\phi_1(\theta) & \cdots &
\tfrac{\partial}{\partial\theta_p}\phi_1(\theta) \\
\vdots & \ddots & \vdots \\
\tfrac{\partial}{\partial\theta_1}\phi_m(\theta) & \cdots &
\tfrac{\partial}{\partial\theta_p}\phi_m(\theta) \\
\end{pmatrix}.
\end{align*}

Together with the ability to derive the joint IF from marginal IFs, this
provides us with a powerful tool for constructing new estimates using the IFs as
the fundamental building blocks.

To apply the delta method the transformation of the parameters function must be
supplied to the `estimate` method (argument `f`)
```{r delta1}
estimate(g1, sum)
estimate(g1, function(p) list(a = sum(p))) # named list
## Multiple parameters
estimate(g1, function(x) c(x, x[1] + exp(x[2]), inv = 1 / x[2]))
estimate(g1, exp)         
```
The gradient can be provided as the attribute `grad` and otherwise numerical differentiation is applied. 


An often more convenient syntax is to directly transform the `estimate` objects
via standard mathematical R operations. An advantage of this is also that the
necessary derivatives are automatically calculated exactly.
To illustrate this consider the following simple example where we consider two
`estimate` objects each with a single parameter estimate
```{r estimate_calculus_models}
a <- estimate(coef=c("a"=0.5), IC=rnorm(10), id=1:10)
b <- estimate(coef=c("b"=0.8), IC=rnorm(10), id=1:10)
```

Parameter transformation can now be calculated directly as in the following examples
```{r estimate_calc_ex}
a * b
(3 * cos(a) / sqrt(b) + 1) / a
c(sum=sum(e), sum2=a+b,
  prod=prod(e), prod2=a*b) # sum and prod function
e <- c(a,b) # merge
e %*% e # inner prod.
c(1, 2) %*% e
c(pow = a^b) # power-function, rename parameter
a^c(0.5, 2)
c(e["a"] * e["b"] / a, e["b"])
```

For the `%*%*` operator we can also use a general contrast matrix
(see also Section on [Linear contrasts](#linear-contrasts-and-hypothesis-testing)
```{r estimate-contrast}
B <- rbind(c(1,-1), c(1,0), c(0,1))
B %*% e
B %*% e == c(1,1,0)
```


The following transformations are implemented

- trigonometric functions: `cos`, `sin`, `tan`
- inverse trigonometric functions: `acos`, `asin`, `atan`
- hyperbolic functions: `cosh`, `sinh`, `tanh`
- inverse hyperbolic functions: `acosh`, `asinh`, `atanh`
- other mathematical functions: `log`, `log1p`, `exp`, `expm1`, `sqrt`
- mathematical operators: `+`, `-`, `*`, `/`, `^`,
 `prod`, `sum`, `%*%`

and any compositions of these functions, which also means we can immediately
 apply most existing user-defined functions. To give an example consider the `logit` function
```{r}
lava::logit
logit(b)
expit(c(a,b))
```

In the above example we also demonstrated the user of the subset operator `[]`
and the merge method `c()`. These are basically equivalent to the `subset` and
`merge` methods. There is also a `transform` method for constructing parameter
transformations above, and the `labels` function for renaming parameters.
Together we can combine these methods via the pipe operator to create readable
code:
```{r estimate-pipe}
merge(a, b) |>  # merges the two `estimate` objects
  transform(prod) |> # calculates product of parameter estimates
  subset(1) |> # nothing happens here as the result was already 1-dim.
  labels("prod") # rename parameter
```

Finally, the `with` function can be used to reference parameter names
```{r estimate-with}
e <- c("e1"=a, "e2"=b)
with(e, c(est = e1*e2))
```


## Example: Pearson correlation
As a simple toy example consider the problem of estimating the covariance of two
variables $X_1$ and $X_2$
\begin{align*}
\widehat{\cov}(X_1,X_2) = \mathbb{P}_n(X_1-\mathbb{P}_n X_1)(Y_1-\mathbb{P}_n Y_1).
\end{align*}
It is easily verified that the IF of the sample estimate of $(\E X_{1}, \E
X_{2}, \E\{X_{1}X_{2}\})^\top$ given by
is $\IC(X1,X2; \Pz) = (X_{1}-\E X_{1}, X_{2}-\E X_{2},
X_{1}X_{2}-\E\{X_{1}X_{2}\})^\top$. By the delta theorem with $\phi(x,y,z) =
z-xy$ we have  $\nabla\phi(x,y,z) = (-y, -x, 1)^\top$ and thus the IF for the
sample covariance estimate becomes
\begin{align*}
\IC_{x_1, x_2}(X_1, X_2; \Pz) = (X_1 - \E X_1)(X_2 - \E X_2) - \cov(X_1,X_2)
\end{align*}

We can implement this directly using the `estimate` function via the `IC`
argument which allows us to provide a user-specificed IF and with the point
estimate given by the `coef` argument
```{r cov}
Cov <- function(x, y, ...) {
  est <- mean(x * y)-mean(x)*mean(y)
    estimate(
      coef = est,
      IC = (x - mean(x)) * (y - mean(y)) - est,
      ...
    )
}
with(dw, Cov(x1, x2))
```

As an illustration we could also derive this estimate from simpler building
blocks of $\E X_{1}$, $\E X_{2}$, and $\E(X_{1}X_{2})$.
```{r cov2}
est <- lm(cbind(x1, x2, x1 * x2) ~ 1, data = dw) |>
  estimate(labels = c("E1", "E2", "E12"))
est

est["E12"] - est["E2"]*est["E1"] 
# transform(e1, function(x) c(x, cov=with(as.list(x), E12 - E2* E1))) # Same result
```
The variance estimates can be estimated in the same way and combined to estimate the correlation
```{r rho}
v12 <- with(dw, Cov(x1, x2, id = id))
v1  <- with(dw, Cov(x1, x1, id = id))
v2  <- with(dw, Cov(x2, x2, id = id))

rho <- c(rho = v12 / sqrt(v1 * v2)) 
rho
```

by using a variance stabilizing transformation, Fishers
$z$-transform [@lehmann2023_testing],
$z = \arctanh(\widehat{\rho}) =
\frac{1}{2}\log\left(\frac{1+\widehat{\rho}}{1-\widehat{\rho}}\right)$,
confidence limits with general better coverage can be obtained
```{r tanh}
estimate(atanh(rho), back.transform = tanh)
```
The confidence limits are calculated on the $\arctanh$-scale and transformed
back to the original correlation scale via the `back.transform` argument.
In this case, where the estimates are far away from the boundary of the parameter space, the variance
stabilizing transform does almost not have any impact, and the confidence limits
agrees with the original symmetric confidence limits.

## Linear contrasts and hypothesis testing

An important special case of parameter transformations are linear
transformations. A particular interest may be formulated around testing
null-hypotheses of the form
\begin{align*}
H_0\colon\quad \bm{B}\theta = \bm{b}_0
\end{align*}

where $\bm{B}\in\mathbb{R}^{m\times p}$ is a matrix of estimable contrasts and
$\bm{b}_0\in\mathbb{R}^{m}$.

As an example consider marginal models for the binary response variables
  $Y_1, Y_2, Y_3, Y_4$
```{r estglmmerge}
g <- lapply(
  list(y1 ~ a, y2 ~ a, y3 ~ a), #, y4 ~ a+x4),
  function(f) glm(f, family = binomial, data = dw)
)
gg <- Reduce(merge, g)
gg
```

A linear transformation can be specified via the `f` as a matrix argument
instead of function object
```{r contrast1}
B <- cbind(0,1, 0,-1, 0,0)
estimate(gg, B)
```

The $\bm{b}_0$ vector (default assumed to be zero) can be specified via the `null` argument
```{r estcontrast1}
estimate(gg, B, null=1)
```

For testing multiple hypotheses we use that 
$(\bm{B}\widehat{\theta}-\bm{b}_0)^{\top}
(\bm{B}\widehat{\Sigma}\bm{B}^{\top})^{-1}
(\bm{B}\widehat{\theta}-\bm{b}_0) \sim
 \chi^2_{\operatorname{rank}(B)}$ under the null hypothesis where
 $\widehat{\Sigma}$ is the estimated variance of $\theta$ (i.e.,
`vcov(gg)`) 
```{r estcontrast2}
B <- rbind(cbind(0,1, 0,-1, 0,0),
           cbind(0,1, 0,0, 0,-1))
estimate(gg, B)
```

Such linear statistics can also be specified directly as expressions of the parameter names
```{r}
estimate(gg, a + a.1, 2*a - a.2, a, null=c(2,1,1))
```

We refer to the function `lava::contr`  and `lava::parsedesigns` for defining contrast matrices.
```{r contr}
contr(list(1, c(1, 2), c(1, 4)), n = 5)
```
A particular useful contrast is the following for considering all pairwise
comparisons of different exposure estimates:

```{r pairwise.diff}
pairwise.diff(3)
estimate(gg, pairwise.diff(3), null=c(1,1,1), use=c(2,4,6))
```

When conducting multiple tests each at a nominal-level of $\alpha$ the overall
type I error is not controlled at $\alpha$-level.
The influence function also allows for adjusting for multiple comparisons.
Let $Z_{1},\ldots,Z_{p}$ denote $Z$-statistics from $p$ distinct two-sided
hypothesis tests which we will assume is asymptotically distributed under the
null hypothesis as a zero-mean Gaussian distribution with correlation matrix
$R.$ Let $ Z_{max} = \max_{i=1,\ldots,p} |Z_i|$ then the family-wise error
rate (FWER) under the null can be approximated by
$$
P(Z_{max} > z) =
1-\int_{-z}^{z} \cdots \int_{-z}^{z} \phi_{R}(x_{1},\ldots,x_{p})
\,dx_{1}\cdots\,dx_{p}
$$
where $\phi_{R}$ is the multivariate normal density
function with mean 0 and variance given by the correlation matrix $R$. The
adjusted $p$-values can then be calculated as
$$
P(Z_{max} >
\Phi^{-1}(1-p/2))
$$
where $\Phi$ is the standard Gaussian CDF. As described in
in [@RitzPipper_multcomp] the joint distribution of $Z_{1},\ldots,Z_{p}$
can be estimated from the IFs. This is implemented in the `alpha_zmax` method

```{r pcorrect, eval=has_mets }
gg0 <- estimate(gg, use="^a", regex=TRUE, null=rep(.8, 3))
alpha_zmax(gg0)
```

While this always yields a more powerful test compared to Bonferroni
adjustments, a more powerful closed-testing procedure [@marcus1976], can be generally
obtained by considering all intersection hypotheses. 

![*Figure: closed testing via Wald tests of all intersections hypotheses.*](../man/figures/closedtesting.svg){width=60%}

To reject the $H_{1}$ at
an correct $\alpha$-level we should test all intersection hypotheses involving
$H_{1}$ and check if there all are rejected at the $\alpha$-level. The
adjusted $p$-values can here be obtained as the maximum $p$-value across all
the composite hypothesis tests. Unfortunately, this only works for relatively
few comparisons as the number of tests grows exponentially.

```{r closedtesting, eval=has_mets}
closed_testing(gg0, test = test_wald)
```


## Averaging

Some parameters of interest are expressed as averages over functions of the
observed data and estimated parameters of a model. The asymptotic distribution
can in some of these cases also be derived from the influence function.
Let $Z_{1},\ldots,Z_{n}$ be iid observations, $Z_{1}\sim \Pz$ and let  $X_{i}\subset Z_{i}$.

Assume that $\widehat{\theta}$ is RAL estimator of $\theta\in\Omega\subset \mathbb{R}^{p}$
$$\sqrt{n}(\widehat{\theta}-\theta) = \frac{1}{\sqrt{n}}\sum_{i=1}^{n}\phi(Z_{i}; \Pz) + \op(1).
$$
Let $f:\mathcal{X}\times\Omega\to\mathbb{R}$ be continuous differentiable in $\theta$
$$\sqrt{n}\{f(X; \widehat{\theta})-f(X; \theta)\} =
\frac{1}{\sqrt{n}}\nabla_{\theta}f(X;\theta)\sum_{i=1}^{n}\phi(Z_{i}; P) + \op(1).
$$

Let $\Psi = \Pz f(X;\theta)$ and $\widehat{\Psi} = P_{n} f(X;\widehat{\theta})$. $\Pz$ and $P_{n}$ are here everywhere the integrals wrt. $X$. It is easily verified that
\begin{align*}
\widehat{\Psi}-\Psi &= (P_{n}-\Pz)(f(X; \theta)-\Psi) + P[f(X;\widehat{\theta})-f(X;\theta)] \\
&\quad + (P_{n}-\Pz)[f(X;\widehat{\theta})-f(X;\theta)]
\end{align*}

From Lemma 19.24 [@vaart_1998_asymp] it follows that for the last term
\begin{align*}
\sqrt{n}(P_{n}-\Pz)[f(X;\widehat{\theta})-f(X;\theta)] = \op(1)
\end{align*}
when $f$ for example is Lipschitz and more generally when $f(X;\theta)$ forms a $\Pz$-Donsker class.

It therefore follows that
\begin{align*}
\sqrt{n}(\widehat{\Psi}-\Psi) &= \sqrt{n}P_{n}(f(X; \theta)-\Psi) +
\frac{1}{\sqrt{n}}P\nabla_{\theta}f(X;\theta)\sum_{i=1}^{n}\phi(Z_{i}; \Pz) + \op(1) \\
&=
\frac{1}{\sqrt{n}}\sum_{i=1}^{n}\{f(X;\theta)-\Psi\} +
\frac{1}{\sqrt{n}}P\nabla_{\theta}f(X;\theta)\sum_{i=1}^{n}\phi(Z_{i}, \Pz) + \op(1)
\end{align*}
Hence the IF for $\widehat{\Psi}$ becomes
$$
IC(Z; \Pz) = f(X;\theta)-\Psi +
[\Pz\nabla_{\theta}f(X;\theta)]\phi(Z).
$$


Turning back to the example we can estimate the logistic regression model
$\operatorname{logit}(E\{Y_1 | A,X_1,W\}) = \beta_0 + \beta_a A + \beta_{x_1}
X_1 + \beta_w W$, and from this we want to estimate the target parameter
$$
\theta(P) = \E_{P}[E(Y\mid A=1, X_{1}, W)].
$$
To do this we need first to estimate the model and then define a function that
gives the predicted probability $\pr(Y=1\mid,A=a,X_{1},W)$ for any observed
values of $X_1,W$ but with the treatment variable $A$ kept fixed at the
value $1$
```{r estpred}
g <- glm(y1 ~ a + x1 + w, data=dw, family=binomial)
pr <- function(p, data, ...)
  with(data, expit(p[1] + p["a"] + p["x1"]*x1 + p["w"]*w))
pr(coef(g), dw) |> head()
```
The target parameter can now be estimated with the syntax
```{r average}
id <- foldr(NROW(dw), 100, list=FALSE)
ea <- estimate(g, pr, average=TRUE, id=id)
ea
IC(ea) |> head()
```

## Average Treatment Effects

Let the observed data be $Z = (W,A,Y)$, where $W$ are covariates, $A$ a binary treatment
variable, and $Y$ the outcome.
In the following we are interested in estimating the target parameter
$\psi_a(P)
= E_P[Y(a)]$, where $Y(a)$ is the *potential outcome* we would have
observed if treatment $a$ had been administered, possibly contrary to
the actual treatment that was observed, i.e., $Y = Y(A)$. To assess the
treatment effect we can then the consider the *average treatment effect*
(ATE) $$E_P[Y(1)]-E_P[Y(0)],$$ or some other contrast of interest
$g(\psi_1(P), \psi_0(P))$. Under the following assumptions

1)  Stable Unit Treatment Values Assumption (the treatment of a specific
    subject is not affecting the potential outcome of other subjects)
2)  Positivity, $P(A\mid W)>\epsilon$ for some $\epsilon>0$ and baseline
    covariates $W$
3)  No unmeasured confounders, $Y(a)\perp \!\!\! \perp A|W$


then the target parameter can be identified from the observed data
distribution as $$E(E[Y|W,A=a]) = E(E[Y(a)|W]) = E[Y(a)]$$ or
$$E[Y I(A=a)/P(A=a|W)] = E[Y(a)].$$

This suggests estimators based on outcome regression ($g$-computation)
or inverse probability weighting. More generally, under the above
assumption we can constructor a *one-step* estimator from the *Efficient
Influence Function* combining these two
$$ E\left[\frac{I(A=a)}{\Pi_a(W)}(Y-Q(W,A)) + Q(W,a)\right].$$
In practice, this requires plugin estimates of both the outcome model,
$Q(W,A) := E(Y\mid A, W)$, and of the treatment propensity model
$\Pi_a(W) := P(A=a\mid W)$. The corresponding estimator is consistent
even if just one of the two nuisance models is correctly specified.

To illustrate this we manually construct the above estimating equations based on
logistic regressions models for the two nuisance models
```{r ate-nuisancemodels}
qmod <- glm(y1 ~ a * w, family = binomial, data = dw) # E(Y|W,A) := Q(W,A)
amod <- glm(a ~ w, family = binomial, data = dw) # P(A=1|W) = Pi_1(W)
q0 <- predict(qmod, transform(dw, a=0), type="response") # Q(W,0)
q1 <- predict(qmod, transform(dw, a=1), type="response") # Q(W,1)
p1 <- predict(amod, dw, type="response") # P(A=1|W)
e0 <- with(dw, (1-a) / (1-p1) * (y1 - q0) + q0)
e1 <- with(dw, a / p1 * (y1 - q1) + q1)
head(cbind(e0, e1))
```

We can then easily construct estimates of the two expected potential outcomes
and estimate their joint asymptotic distribution
```{r est-potentialoutcomes}
est0 <- estimate(coef=mean(e0), IC=e0-mean(e0)) # E[Y(0)]
est1 <- estimate(coef=mean(e1), IC=e1-mean(e1)) # E[Y(1)]

potential_outcomes <- merge(est0, est1, paired=TRUE, labels=c("y(0)", "y(1)"))
potential_outcomes
head(IC(potential_outcomes))
vcov(potential_outcomes)
```

Finally, we can obtain the Average Treatment Effect (ATE) (risk-difference)
\begin{align*}
E[Y(1)]  - E[Y(0)]
\end{align*}

```{r est-ate}
estimate(potential_outcomes, cbind(-1, 1), labels="ate")
```

Alternatively, we could get the estimate of the treatment effect on the log-odds
scale:
```{r est-ate-or}
est <- with(potential_outcomes, logit(`y(1)`)-logit(`y(0)`))
est
transform(est, labels="OR", back.transform=exp)
#logor <- function(p) logit(p[2]) - logit(p[1])
#transform(potential_outcomes, logor, labels="logOR")
#transform(potential_outcomes, logor, labels="OR", back.transform=exp)
```

We refer to the `targeted` package [@r-targeted] for a general implementation
where the nuisance models can be based on machine learning models using
cross-fitting. The above estimation can be obtained with the  `targeted::cate` function in the following way
```{r est-cate}
#| eval: false
amod <- targeted::learner_glm(a ~ w, family=binomial)
qmod <- targeted::learner_glm(y1 ~ a * w, family=binomial)
est <- targeted::cate(qmod, amod, data=dw, second.order = FALSE)
#>             Estimate Std.Err    2.5%  97.5%   P-value
#> E[y1(1)]      0.7050 0.03721 0.63207 0.7779 4.934e-80
#> E[y1(0)]      0.5359 0.03535 0.46666 0.6052 6.290e-52
#> ───────────
#> (Intercept)   0.1691 0.04996 0.07115 0.2670 7.138e-04
```

# SessionInfo

```{r sessionInfo}
sessionInfo()
```

# Bibliography

# LAVA celery worker

Run the celery worker for lava.

## Command line

This daemon is started by celery using `python3 -m celery`.

## Service

The systemd service is called `lava-celery-worker`.

## Dependencies

lava-celery-worker should be able to:

* connect to the celery broker (rabbitmq, redis, ...)
* connect to the [postgresql](./postgresql.md) database

## Configuration

Daemon start options:

* `/etc/default/lava-celery-worker`
* `/etc/lava-server/lava-celery-worker`

Django configuration:

* `/etc/lava-server/settings.conf`
* `/etc/lava-server/settings.yaml`
* `/etc/lava-server/settings.d/*.yaml`

## Logs

The logs are stored in `/var/log/lava-server/lava-celery-worker.log`

The log rotation is configured in `/etc/logrotate.d/lava-celery-worker-log`.

# Setting up remote worker

Test execution in LAVA is performed by 'lava-worker'. It can run on the same
physical hardware as 'lava-server' but also can run separately on different
physical host. The latter case is called 'remote worker'. Remote workers can
connect to server on local network or using Internet. Connection is established
over http/https protocol.

## lava-dispatcher settings

In order to point lava-worker to the correct server, it needs the following
settings:
```shell
URL="http://<lava-server-dns>/"
```

Location of the settings depends on the way lava-dispatcher is started.

When using standalone installation either from sources or from debian package,
settings can be found in `/etc/lava-dispatcher/lava-worker` file. In case
[docker-compose](https://git.lavasoftware.org/lava/pkg/docker-compose) is used
settings should be updated in .env file:

```shell
DC_LAVA_SERVER_HOSTNAME="<lava-server-dns>"
```

## connection encryption

It is advised to always encrypt the connection between server and workers. We
advice to use https instead of http for worker connection.

## http_proxy settings

Worker specific http_proxy settings should be defined on the server in
```shell
/etc/lava-server/dispatcher.d/<name>/env.yaml
```
It can be directly edited on the server host or created/updated using API calls:
 * ```scheduler.workers.set_env``` XML-RPC
 * POST on ```/api/v0.2/workers/``` REST API
 * PUT on ```/api/v0.2/workers/<hostname>/env``` REST API

## local devices connected to remote worker

All device dictionaries are stored on the server node and passed to workers
before starting test jobs. In case both workers and server are administered by
the same admin there should be no issues.

However in case remote worker admin is different than server admin, there needs
to be coordination of device dictionaries. It is advised that device dictionary
files are maintained in version control system. This way bad changes can be
reverted quickly and full instance recovery is possible without major problems.

Current APIs allow to upload device dictionaries directly to the server. These
are ```scheduler.devices.set_dictionary``` XML-RPC method and POST request on
```/api/v0.2/devices/<hostname>/dictionary``` REST API.

Alternatively server admins might decide to use a configuration tool (salt,
ansible, etc.) to copy files from version control to the lava-server host.

### database entries for device types and devices

Although all device specific settings are stored in device dictionary, there is
still a need to create a database entry. This can be done using django admin UI,
XML-RPC API ```scheduler.devices.add``` or issuing POST request on
```/api/v0.2/devices/```.

## local device types (not yet available in LAVA)

Devices on which test jobs are executed should be supported by LAVA. This means
that "device type" should be present in the server node. Device type is used to
render full device dictionary. There may be cases that remote workers include
devices that are not yet supported by LAVA. In such case the following actions
should be performed:

 * new device type should be added to LAVA code base
 * temporarily server node admins might add new device type on their host. This
   operation is depending on how server configuration is maintained. As with
   device dictionaries it is advised to use version control.

## dispatcher version

Currently lava-dispatcher version should be in sync with lava-server version.
In case of version mismatch, lava-worker connection will be rejected by
lava-server.

### upgrades

Currently LAVA doesn't support any mechanism for remote dispatcher updates. It
is responsibility of admins to perform the updates.

# remote dispatcher - host installation

The instructions assume installation from debian packages but should also be
valid for installation from sources. These steps should result in remote worker
registering with LAVA server.

## install required packages

```shell
apt-key adv --keyserver keyserver.ubuntu.com --recv-keys A791358F2E49B100
apt-get update
apt-get install lava-dispatcher lava-dispatcher-host
```

### installing from source

lava-dispatcher-host package installs udev rules file. In order to perform this
step when installing from source one needs to run:
```shell
lava-dispatcher-host rules install
```

## edit /etc/lava-dispatcher/lava-worker and update settings

```shell
URL="http://<lava-server-dns>/"
```

## start lava-dispatcher

```shell
systemctl start lava-worker
```

# remote dispatcher - docker installation

docker-compose file to setup an instance **lava-dispatcher** from scratch. In this
setup, every service will be running in a separate container.

## Usage

###  Requirements

In order to use this docker-compose file, you need:

* docker.io
* docker-compose

You can install the dependencies using:

```shell
apt install docker.io docker-compose
```

### Installing

You just need to fetch the sources:

```shell
git clone https://git.lavasoftware.org/lava/pkg/docker-compose
cd docker-compose
```

### Using it

#### Configuration (simple, for QEMU purposes)

All configuration is stored in `.env` file. Some of the steps are required
whilst others are optional.

* Change DC_LAVA_SERVER_HOSTNAME to <server_name>
  which points to the running LAVA server instance.
* Execute `make lava-dispatcher`; at this point multiple containers should be
  up and running and the worker should connect to the LAVA server instance of
  your choosing.
* Add a new device and set its' device template (alternatively you can update
  existing device to use this new worker)
  Example QEMU device template:
  ```jinja
  {% extends 'qemu.jinja2' %}
  {% set mac_addr = 'DF:AD:BE:EF:33:02' %}
  {% set memory = 1024 %}
  ```
  You can do this via [XMLRPC](https://validation.linaro.org/api/help/#scheduler.devices.set_dictionary), [lavacli](https://docs.lavasoftware.org/lavacli/) or [REST API](https://staging.validation.linaro.org/api/v0.2/devices/staging-qemu01/dictionary/) (if using version 2020.01 and higher).
* (optional) If the lab where this container runs is behind a proxy or you
  require any specific worker environment settings, you will need to update the
  proxy settings by setting the [worker environment](https://docs.lavasoftware.org/lava/proxy.html#using-the-http-proxy)
  You can do this via this [XMLRPC API call](https://validation.linaro.org/api/help/#scheduler.workers.set_env).

!!! note "Firewall"

    If the server instance is behind a firewall, you will need to create a
    port forwarding so that ports 80 and maybe 443 are open to the public.


#### Configuration (advanced, for physical DUT purposes)

If you're setting up a standalone dispatcher container, make sure you go
through the above configuration first, it is mandatory for this step.
In order to run test jobs on physical devices we will need a couple of
additional setup steps:

* PDU control:
  * The dispatcher docker container will already download pdu scripts from
    [lava-lab repo](https://git.linaro.org/lava/lava-lab.git/) which you can use
    in device configuration but if you use custom PDU scripts you need to
    provide them and copy them into `dispatcher/power-control` directory; they
    will be copied into `/root/power-control` path in the container.
  * If you need SSH keys for PDU control, copy the private key to the
    `dispatcher/ssh` directory and the public key on to the PDU
  * SSH config - if there's a need for a specific SSH configuration (like
    tunnel passthrough, proxy, strict host checking, kexalgorithm etc), create
    the config file with relevant settings and copy it into `dispatcher/ssh`
    dir; it will be copied to `/root/.ssh` directory on the dispatcher
    container.
* ser2net config - update `ser2net/ser2net.config` with the corresponding
  serial port and device settings
* Update/add [device dictionary](https://docs.lavasoftware.org/lava/glossary.html#term-device-dictionary) with power commands and connection command
* Add dispatcher_ip setting to the [dispatcher configuration](https://validation.linaro.org/api/help/#scheduler.workers.set_config). Alternatively you can use
[REST API](https://lava_server/api/v0.2/workers/docker_dispatcher_hostname/config/) if you are using version 2020.01 or higher:
  * `dispatcher_ip: <docker host ip address>`
* Disable/stop rpcbind service on host machine if it's running - docker service
  nfs will need port 111 available on the host.


### Running

In order to start the containers, run:

    docker-compose build lava-dispatcher
    docker-compose up lava-dispatcher

or, alternatively:

    make lava-dispatcher

# Fastboot

The `fastboot` deploy action downloads images and flashes them to the device
using the [Fastboot](https://en.wikipedia.org/wiki/Fastboot) protocol.

```yaml
- deploy:
    to: fastboot
    docker:
      image: linaro/noble-adb-fastboot
      local: true
    images:
      boot:
        url: https://example.com/boot.img
      rootfs:
        url: https://example.com/rootfs.img.xz
        compression: xz
        apply-overlay: true
```

## docker

LAVA supports running fastboot directly on the worker or within a Docker
container. Using Docker is recommended as it allows using custom fastboot
binaries and provides better isolation between jobs.

The `docker` block specifies the container image used to run fastboot commands.
This image must contain the `fastboot` binary.

```yaml
- deploy:
    to: fastboot
    docker:
      image: linaro/noble-adb-fastboot
      local: true
```

### image

The Docker image name (e.g., `linaro/noble-adb-fastboot`).

### local

Optional. If `true`, LAVA will use the image if it already exists locally on
the worker without pulling from a registry.

## images

The `images` block specifies a set of images to be downloaded and flashed to
the device. Each key in the `images` dictionary is the partition name to flash.

```yaml
- deploy:
    to: fastboot
    images:
      boot:
        url: https://example.com/boot.img
      system:
        url: https://example.com/system.img.xz
        compression: xz
```

### partition

The key name under `images` is the partition to which the image will be
flashed. It is passed directly to the `fastboot flash` command:

```shell
fastboot flash <image_key> <image>
```

!!! note
    If [`fastboot_sequence`](../../../configuration/device-dictionary.md#fastboot_sequence)
    is set to `no-flash-boot`, LAVA skips flashing the `boot` partition even if
    a `boot` image was listed under `images`.

### apply-overlay

Set to `true` to apply the LAVA test overlay to this image. The overlay
contains the test definitions and LAVA helper scripts.

```yaml
- deploy:
    to: fastboot
    images:
      rootfs:
        url: https://example.com/rootfs.img.xz
        compression: xz
        apply-overlay: true
```

### sparse

System images shipped as sparse images requires special handling with tools such
as `simg2img` and `img2simg` in order to apply LAVA overlays.

By default, LAVA assumes that any image with `apply-overlay: true` is a sparse
image. If the image is not a sparse image, set `sparse: false` so that LAVA
treats it as a plain ext4 image:

```yaml
- deploy:
    to: fastboot
    images:
      system:
        url: https://example.com/system.ext4.xz
        compression: xz
        sparse: false
        apply-overlay: true
```

### reboot

If the device needs to be restarted after flashing a image, specify the reboot
method. This is optional and only needed when an intermediate reboot is required.

```yaml
- deploy:
    to: fastboot
    images:
      partition:
        url: https://example.com/gpt_both0.bin
        reboot: hard-reset
      boot:
        url: https://example.com/boot.img
```

Allowed values:

- `hard-reset`
- `fastboot-reboot`
- `fastboot-reboot-bootloader`
- `fastboot-reboot-fastboot`

### url

See [url](./index.md#url)

### compression

See [compression](./index.md#compression)

# LAVA logs in database

LAVA logs are received and stored by lava-server-gunicorn. The default
storage backend uses filesystem for storing logs and the default location of
`/var/lib/lava-server/default/media/job-output/`.

LAVA also supports storing the logs in a few NoSQL based database engines.

## Configuration

This can be set up in the [LAVA configuration](../basic-tutorials/instance/configure.md), like so:

```yaml
LAVA_LOG_BACKEND: "lava_scheduler_app.logutils.LogsFilesystem"
```

Available backends are:

* `lava_scheduler_app.logutils.LogsMongo`
* `lava_scheduler_app.logutils.LogsElasticsearch`
* `lava_scheduler_app.logutils.LogsFirestore`

The list can be also found in [the source code](https://gitlab.com/lava/lava/-/blob/master/lava_server/settings/common.py)

### MongoDB

Integration with MongoDB requires two variables to be set in the [LAVA settings](../basic-tutorials/instance/configure.md):

* `MONGO_DB_URI` - connection string URI format for mongodb (i.e. mongodb://example.com:27017)
* `MONGO_DB_DATABASE` - database name for LAVA logs (i.e. 'lava-logs')

Every log line in LAVA is one document in MongoDB database.

### ElasticSearch

Integration with Elasticsearch db requires three variables to be set in the [LAVA settings](../basic-tutorials/instance/configure.md):

* `ELASTICSEARCH_URI` - connection string URI format for Elasticsearch (i.e. http://localhost:9200/)
* `ELASTICSEARCH_INDEX` - database(index) name for LAVA logs (i.e. 'lava-logs')
* `ELASTICSEARCH_APIKEY` - API key to be used in Authorization header when
  sending requests to the Elasticsearch API. Can be obtained via [this request](https://www.elastic.co/guide/en/elasticsearch/reference/current/security-api-create-api-key.html).

Every log line in LAVA is one document in Elasticsearch database.

### Firestore

Still proof of concept, does not cover full integration with LAVA logs.
Patches are welcome.

In order to user Google firestore, you first have to create a project and
generate a private key file for the service account [here](https://console.firebase.google.com/project/_/settings/serviceaccounts/adminsdk).

The root collection is currently hardcoded to `logs`.

Firestore integration does not currently support limited read for each job,
meaning that incremental read on the job UI will not work properly.

## Log database migration

If the logging database is used from the very first installation of one LAVA
instance then all the logs will be stored in the said database, but if the
switch is made at one later point in the life of a LAVA instance, then you
can use the management command [copy-logs](https://gitlab.com/lava/lava/-/blob/master/lava_server/management/commands/copy-logs.py) to copy the logs
from the filesystem to the database of your choosing.
It is mandatory to have appropriate configuration variables set in LAVA settings in order to properly connect to the database.

!!! example "From filesystem to MongoDB"

    ```shell
    lava-server manage copy-logs LogsMongo
    ```

!!! example "From filesystel to ElasticSearch"

    ```shell
    lava-server manage copy-logs --dry-run LogsElasticsearch
    ```

!!! warning "copy-logs compatibility"
    Management command `copy-logs` is only applicable to the MongoDB and
    Elasticsearch database engines. Firestore is not yet supported.

# Command action

The `command` action is used to execute a pre-defined command on the dispatcher
itself. A typical use case would be to set the bootmode for a board by
controlling a relay.

## Job definition

In order to use the action, the action block should only define the name of the
command to execute:

```yaml
actions:
- command:
    name: set_boot_to_nand
```

## Device dictionary

Admins should add the commands to the device dictionary:

```jinja
{% set user_commands = {'set_boot_to_qspi': {'do': 'webrelay --relay 1 on',
                                             'undo': 'webrelay --relay 1 off'},
                        'set_boot_to_nand': {'do': 'webrelay --relay 1 off',
                                             'undo': 'webrelay --relay 1 off'}} %}
```

Admins can define two commands:

* `do`: called when running the action
* `undo`: called when finalizing the job

!!! note "Failing jobs"
    When an action of a job is failing, the full job will be terminated.

    In such case, the `do` command of every subsequent actions will not be
    called. However, the job will be finalized and every `undo` commands will be
    called.

## Power commands

When defined, the power commands are automatically added to the list of available commands:

* `hard_reset`
* `pre_power_command`
* `pre_os_command`
* `power_on`
* `power_off`
* `recovery_mode`
* `recovery_exit`

!!! warn "`undo` is undefined"
    When using such commands, LAVA will not run any `undo` command.

# Overlay

The `overlay` deployment method creates the LAVA overlay without downloading
or flashing any images. The overlay is a tarball that contains LAVA helper
scripts and test definitions needed for running tests in LAVA test shell on the
DUT.

```yaml
- deploy:
    to: overlay
```

## Use case

The method is useful when the DUT is booted with a pre-installed OS. The overlay
can be transferred to the running OS using
[`transfer_overlay`](../boot/common.md#transfer_overlay) in the subsequent boot
method.

## Example job

[Flashing and testing rzn1 FW using a pre-installed OS](../boot/method-dfu.md#u-boot-dfu)

--8<-- "refs.txt"

# Configure LAVA

## Creating a superuser

Admins can create superuser accounts from the command line:

!!! example ""

    ```shell tab="Docker"
    docker-compose exec lava-server lava-server manage createsuperuser
    ```

    ```shell tab="Debian"
    lava-server manage createsuperuser
    ```

## Settings

LAVA settings are stored in **yaml** in:

* `/etc/lava-server/settings.conf`
* `/etc/lava-server/settings.yaml`
* `/etc/lava-server/settings.d/*.yaml`

LAVA will load the files in this exact order. Files in the `settings.d`
directory will be alphabetically ordered.

!!! tip "Merging files"
    If a variable is defined in two files, the value from the last file will
    override the value from the first one.

In a fresh installation, LAVA installer will automatically create some setting files
for you:

* `/etc/lava-server/settings.d/00-secret-key.yaml`: django [SECRET_KEY]
* `/etc/lava-server/settings.d/00-database.yaml`: django [DATABASES]

You can then create some file to customize your instance.
The list of available values is listed in:

* [Django settings](https://docs.djangoproject.com/en/2.2/ref/settings/)
* [LAVA settings](https://git.lavasoftware.org/lava/lava/-/blob/master/lava_server/settings/common.py)

!!! warning "Legacy configuration"
    In previous LAVA versions, the settings where saved in three files:

    * `/etc/lava-server/settings.conf` (json)
    * `/etc/lava-server/secret_key.conf` (json)
    * `/etc/lava-server/instance.conf` (ini)

    LAVA will still load this files prior to load the new setting files.

!!! tip "Applying changes"
    When updating the settings, you should restart **every** LAVA services.

## Worker configuration

### Devices

The device dictionaries are stored into `/etc/lava-server/dispatcher-config/devices/`.
The files should belong to `lavaserver` user and group.

More information in the [technical reference](../../../technical-references/configuration/device-dictionary.md).

### Device-types

The device-type templates are stored into `/etc/lava-server/dispatcher-config/device-types/`
The files should belong to `lavaserver` user and group.

For supported device-types, the templates provided by LAVA are stored in
`/usr/share/lava-server/device-types`. LAVA looks for the template first in
`/etc/lava-server/dispatcher-config/device-types/` and then fallback to
`/usr/share/lava-server/device-types`.

More information in the [technical reference](../../../technical-references/configuration/device-type-template.md).

### Health-checks

The health-checks are stored into `/etc/lava-server/dispatcher-config/health-checks/`
The files should belong to `lavaserver` user and group.

More information in the [technical reference](../../../technical-references/configuration/health-check.md).

### Dispatcher

!!! warning "TODO"

--8<-- "refs.txt"

# Upgrade LAVA

## When to upgrade

We advice to always upgrade to the latest released version. Only the latest
release will include new features and bug fixes (including security issues).

## How to upgrade

Currently, LAVA does not allow for zero-downtime upgrade.

We advice to put the instance in maintenance before any upgrade.

!!! example "Maintenance mode"

    === "docker-compose"
        ```shell
        docker-compose exec lava-server lava-server manage maintenance
        ```

    === "debian"
        ```shell
        lava-server manage maintenance
        ```

    === "lavacli"
        ```shell
        lavacli system maintenance
        ```


### Docker

When a new version of LAVA is released, a Docker image is published on [docker hub][lava-docker-hub].

In order to upgrade, admins should just pull the latest **docker-compose
configuration**:

```shell
git pull --rebase
```

The restarting docker-compose will enough:

```shell
docker-compose pull
docker-compose up
```

??? tip "Downtime during upgrade"
    The current setup does not allow for zero-downtime upgrade without a
    management layer like [docker swarm](https://docs.docker.com/engine/swarm/)
    or [kubernetes](https://kubernetes.io/).

### Debian

When a new version of LAVA is released, a new Debian package is published on
the official [Debian repository][lava-debian].

You can upgrade using **apt**:

```shell
apt-get update
apt-get upgrade
```

??? tip "Downtime during upgrade"
    Zero-downtime upgrade is not possible with the current Debian packages.
    During the upgrade, every services will be stopped while upgrading.

## Upgrade notifications

Different version compatibility in LAVA between server and worker
instances is not guaranteed at the moment.

When a LAVA server instance is upgraded, if the specific flag in settings is
provided, the system will send out email notifications reminding admins that
the worker instances need to be upgraded as well.

```yaml
"MASTER_UPGRADE_NOTIFY": false
```

The default value for this flag is False so admin action is required in order
to use this feature.

Each user that has a 'change_worker' permission over the at least one worker
(which is not of the same version as the LAVA server) will get an email
with the list of workers which require an upgrade.
If the user is superuser, he will be regarded as having permission over all the
workers in the system, so that kind of user will get the full list of the
workers which require an upgrade.


## Rollback

!!! warning "TODO"

--8<-- "refs.txt"

# Job metadata

The [`metadata`](./job-definition/job.md#metadata) dictionary of a job
definition is stored with the test job. It is meant for the information that
identifies what a job was testing, for example a build id, a git branch or a
toolchain version:

```yaml
metadata:
  build_id: 1234
  branch: release/1.0
  build-url: https://ci.example.com/builds/42
```

Metadata is read when the job is submitted and is never updated afterwards.
Multinode jobs copy the metadata into every sub job.

## Value types

Scalar values are stored as strings, so a filter does not have to know which
type YAML gave them: `build_id: 1234` and `build_id: "1234"` are both matched
by `?metadata__build_id=1234`. Booleans are stored as `true` and `false`.
Dictionaries and lists are kept as they are.

## Retrieving metadata

The metadata of a job is part of the job object in the REST API and is also
available on its own:

```shell
curl https://validation.linaro.org/api/v0.2/jobs/1234/metadata/
```

The results page of a job has a **Job metadata** button that downloads the same
data as YAML.

## Filtering jobs by metadata

Any metadata key can be used as a `metadata__<key>` filter on the jobs
endpoint:

```shell
curl "https://validation.linaro.org/api/v0.2/jobs/?metadata__build_id=1234"
```

Nested dictionaries are reached by adding the keys in order, so
`metadata__build__id=1234` matches:

```yaml
metadata:
  build:
    id: 1234
```

The key can be followed by a lookup. Without one, `exact` is used:

| Lookup        | Description                                     |
| ------------- | ----------------------------------------------- |
| `exact`       | Exact match, the default                        |
| `iexact`      | Exact match, ignoring case                      |
| `icontains`   | Contains the value, ignoring case               |
| `startswith`  | Starts with the value                           |
| `istartswith` | Starts with the value, ignoring case            |
| `endswith`    | Ends with the value                             |
| `iendswith`   | Ends with the value, ignoring case              |
| `in`          | Matches one of the values, separated by commas  |
| `isnull`      | The key is (not) set                            |
| `regex`       | Matches the regular expression                  |
| `iregex`      | Matches the regular expression, ignoring case   |

```shell
curl "https://validation.linaro.org/api/v0.2/jobs/?metadata__branch__startswith=release/"
curl "https://validation.linaro.org/api/v0.2/jobs/?metadata__build_id__in=1234,5678"
```

Several metadata parameters can be combined with any other job filter. Only the
jobs matching all of them are returned:

```shell
curl "https://validation.linaro.org/api/v0.2/jobs/?metadata__branch=main&metadata__build_id=1234&health=Complete"
```

These lookups work on the metadata dictionary as a whole:

| Filter                    | Description                                    |
| ------------------------- | ---------------------------------------------- |
| `metadata__has_key`       | Has this key                                   |
| `metadata__has_keys`      | Has all of these keys, separated by commas     |
| `metadata__has_any_keys`  | Has any of these keys, separated by commas     |
| `metadata__contains`      | Metadata contains this JSON object             |

```shell
curl "https://validation.linaro.org/api/v0.2/jobs/?metadata__has_key=build_id"
curl -G https://validation.linaro.org/api/v0.2/jobs/ \
     --data-urlencode 'metadata__contains={"branch": "main", "build_id": "1234"}'
```

!!! note "Performance"
    On PostgreSQL, exact matches and the dictionary lookups use the index on
    the metadata column. The other lookups cannot, so they are much slower on
    instances with a large number of test jobs. Combine them with an exact
    match or with any other job filter whenever possible.

!!! warning "Jobs submitted before the upgrade"
    Metadata is stored at submission time. Jobs submitted before the instance
    was upgraded to the LAVA version that introduced this feature have no
    metadata until the administrator runs
    [`lava-server manage jobs backfill-metadata`](../admin/advanced-tutorials/backfill-job-metadata.md).

# Device setup

## Available methods

The following methods can be used to add, update and delete device types and
devices in LAVA.

!!!note
    Only one method is required - choose the option that works best for you.

### Admin interface

The LAVA admin interface is a web-based graphical interface accessible through
your browser. It provides easy to use forms for managing device types and
devices. A user with admin permission is required to access it. To access it,
navigate to `/admin/lava_scheduler_app/`.

### Command line

One of the following tools can be used for managing device types and devices
from the command line:

#### lavacli

The `lavacli` utility is the recommended command line tool for interacting with
the LAVA server. Refer to [lavacli](../../../user/basic-tutorials/lavacli.md)
for usage details and setup steps.

#### lava-server

The `lava-server manage` command offers a suite of sub-commands for managing
LAVA server. Because it interacts directly with the database and system
configuration files, it must be executed as root on the server itself.

## Create device type

A device type can be created using one of the following methods. For detailed
usage of the methods, refer to [Available methods](#available-methods).

### From admin interface

1. Navigate to `/admin/lava_scheduler_app/devicetype/add/` in your browser.
2. Input device type `Name` which is the only required field. It must match
exactly the device type name provided in device specific setup guide
(e.g., `docker`, `qemu`, `avh`).

### Via command line

=== "lavacli"
    ```shell
    lavacli device-types add <device-type>
    ```
=== "lava-server"
    ```shell
    lava-server manage device-types add <device-type>
    ```

## Create device

### Add device

A device can be created using one of the following methods. For detailed usage
of the methods, refer to [Available methods](#available-methods).

#### From admin interface

1. Navigate to `/admin/lava_scheduler_app/device/add/` in your browser.
2. Provide the following information:

    * **hostname**: unique identifier for the device
    * **device-type**: the device type name provided in device setup guide
    * **worker host**: the worker that this device is attached to

#### Via command line

=== "lavacli"
    ```shell
    lavacli devices add --type <device-type> --worker <worker> <hostname>
    ```
=== "lava-server"
    ```shell
    lava-server manage devices add \
        --device-type <device-type> \
        --worker <worker> \
        <hostname>
    ```

### Add device configuration

Before a device can accept jobs, it requires a device configuration that defines
its specific characteristics. The device configuration is a [Jinja2][jinja2]
template file that extends a base device type template and can override or add
device specific configurations. For example, serial connection command and
power reset, off, and on commands.

The device configuration file must be placed at
`/etc/lava-server/dispatcher-config/devices/<hostname>.jinja2` on the LAVA server.
It can be copied or uploaded to this location using one of the following methods.
For detailed usage of the methods, refer to [Available methods](#available-methods).

Before running the command below:

* Replace the `<hostname>` with the actual device hostname.
* Create the device configuration file `<hostname>.jinja2`. Refer to device
specific setup guides for examples.

=== "lavacli"
    ```shell
    lavacli devices dict set <hostname> <hostname>.jinja2
    ```
=== "lava-server"
    ```shell
    cp <filename> /etc/lava-server/dispatcher-config/devices/<hostname>.jinja2
    chown lavaserver:lavaserver /etc/lava-server/dispatcher-config/devices/<hostname>.jinja2
    ```

### Change device health

After a device configuration file added, now it is time to put the device online
by changing its health using one of the following methods. For detailed usage of
the methods, refer to [Available methods](#available-methods).

#### From admin interface

1. Navigate to `/admin/scheduler/device/<hostname>` in your browser.
2. Find the `Status` section, change the `Health` field to either `Unknown` or
`Good`.

#### Via command line

Replace `<health>` with either `UNKNOWN` or `GOOD`.

=== "lavacli"
    ```shell
    lavacli devices update --health <health> <hostname>
    ```
=== "lava-server"
    ```shell
    lava-server manage devices update --health <health> <hostname>
    ```

A device in unknown health is already ready to accept and run test jobs. If a
[health check](../../../technical-references/configuration/health-check.md)
job is configured, the device will be tested automatically. A successful health
check run sets the device health to `Good`, while a failed run sets it to `Bad`.

For more information about device state and health, refer to
[device state and health](../../../technical-references/state-machine.md#devices).

## Hardware setup

### Power control

LAVA needs to control power to the device to automate power on, off, and reset
operations. Common options include:

* USB or GPIO controlled relay
* PDU (Power Distribution Unit)

Configure your power control scripts and note the commands for power on, power
off, and reset operations.

### Serial console

LAVA communicates with the device through a serial connection. You will need to
connect the device's serial console to the LAVA worker (e.g., via USB-to-TTL adapter).

Configure `ser2net` on the worker to expose the serial port over telnet by adding
an entry to `/etc/ser2net.yaml`:

```yaml
connection: &device-01
  accepter: telnet(rfc2217),tcp,2001
  enable: on
  connector: serialdev,/dev/serial/by-id/<your-serial-device>,115200n81,local
  options:
    banner: \r\nser2net port \p device \d [\B] \r\n\r\n
    telnet-brk-on-sync: true
    kickolduser: true
    max-connections: 2
```

!!! tip
    * Always use paths in `/dev/serial/by-id/` to ensure stable device naming
    across reboots.
    * Setting `max-connections: 2` allows both LAVA and you to connect to the
    serial console simultaneously for debugging.

Restart or reload ser2net service:

```shell
sudo systemctl restart ser2net.service
```

--8<-- "refs.txt"

# Contribution Guide

We want to make it as easy as possible for LAVA Software users to become
LAVA Software Community Project contributors, so we have created this
guide to help you get started.

The LAVA Software Community Project has published this Contribution
Guide, and all contributors will be expected to adhere to these
guidelines when submitting issues or merge requests. They are designed
to clarify the requirements for contributions, to make contributing more
efficient for all involved.

Following the guidelines is a great way to prevent your contributions
from being rejected or delayed. Most maintainers won't intend to
discredit your work or be tough on contributors. However, many are busy
and some may be working on LAVA in their free time. Well-formed
contributions are much easier to review and work with.

!!! info "See also"
    - [Code of Conduct](../code-of-conduct.md)
    - [Contributing Process](../tutorials/contributing.md)

### Conflicting priorities

Sometimes a request will be turned down because of conflicting
priorities. It is important to talk about the reasons on the mailing
list. Whether you're requesting a new feature, or providing a fix,
remember that the maintainer has to weigh up your contribution. They are
the people who may have to support the new code in the future, and
resources are often scarce. Try not to be discouraged if your feature
request or merge request is turned down. Be open-minded and, if
necessary, propose an alternative idea on the mailing list after hearing
their concerns.

## Pre-requisites to start

* LAVA is written in [Python](http://www.python.org/), so you will need
  to know (or be willing to learn) the language.
* The web interface is a [Django](https://www.djangoproject.com/)
  application so you will need to use and debug Django if you need to
  modify the web interface.
* LAVA uses [YAML][yaml] heavily internally, so you'll likely need to
  understand the syntax.
* LAVA also uses [Jinja2][jinja2].
* All LAVA software is maintained in [git](https://www.git-scm.org/).
* Some familiarity with [Debian](https://www.debian.org/) is going to be
  useful; helper scripts are available when preparing updated `.deb`
  packages based on your modifications.

LAVA is complex and designed to solve complex problems. This has
implications for how LAVA is developed, tested, deployed and used.

## Other elements involved

* The Django backend used with LAVA is
  [PostgreSQL](https://www.postgresql.org/).
* The LAVA UI includes JavaScript and CSS.
* LAVA also uses [ZMQ](http://zeromq.org/) and XML-RPC.

In addition, test jobs and device support can involve use of U-Boot,
GuestFS, fastboot, ADB, QEMU, Grub, SSH, Docker, and a wide variety
of other systems and tools.

## Updating documentation

We welcome contributions to improve our documentation. If you are
considering adding new features to LAVA or changing current behavior,
also please ensure that the changes include matching updates for the
documentation.

Wherever possible, all new sections of documentation should come
**with worked examples**.

* If the change relates to or includes particular test definitions to
  demonstrate the new support, add a test definition YAML file as an
  example.
* Use comments in the examples and link to existing terms and sections.

--8<-- "refs.txt"

# Test Actions

The test action executes tests on the DUT after booting. It collects output,
parses results, and reports test outcomes.

**Key rules for using test actions:**

* A test action **never** boots the device - a [boot action](./boot/common.md)
  **must** be defined before it.

* LAVA support scripts are prepared by the [deploy action](deploy/index.md) and
  remain available for all test definitions until the next deploy block.

* Multiple test definitions can be listed within the same test block to run
  sequentially without rebooting the device.

There are 4 types of test actions:

* [**Lava-test-shell definitions**](#definitions) (YAML directive: `definitions`)
  are used for POSIX compliant operating systems on the DUT. The deployed system
  is expected to support a POSIX shell environment (`/bin/ash`, `/bin/dash` or
  `/bin/bash` are the most common) so that LAVA can execute the LAVA Test Shell
  Helper scripts.

* [**Output monitors**](#monitors) (YAML directive: `monitors`) are used for
  devices which have no POSIX shell and start the test (and corresponding output)
  immediately after booting, for example microcontroller/IoT boards.

* [**Interactive tests**](#interactive) (YAML directive: `interactive`) are
  further extension of `monitor` tests idea, allowing not just matching some
  output from a device, but also feeding some input. They are useful for
  non-POSIX shells like bootloaders (u-boot for instance) and other interactive
  command-line applications.

* [**Test services**](#services) (YAML directive: services) are used for
  starting docker-compose based containers on the LAVA worker. It enables
  running custom services for the needs in test definitions.

## Definitions

```yaml title="Lava-test-shell definitions"
- test:
    timeout:
      minutes: 5
    definitions:
    - name: smoke-tests
      from: git
      repository: https://github.com/Linaro/test-definitions
      path: automated/linux/smoke/smoke.yaml
    - name: smoke-tests-net
      from: git
      repository: https://gitlab.com/lava/functional-tests.git
      path: posix/smoke-tests-net.yaml
```

### name

The unique test suite name to use. It replaces the name from the YAML.

### repository

A publicly readable repository location.

### from

The type of the repository is **not** guessed, it **must** be specified
explicitly. Support is available for [git](#git), [url](#url) and [inline](#inline).

#### git

A remote git repository which needs to be cloned by the LAVA worker.

Optional parameters:

* revision (`str`, default: `null`): specific commit SHA to check out
* branch (`str`, default: `null`): git branch to clone
* shallow (`bool`, default: `true`): perform a shallow git clone
* history (`bool`, default: `false`): include full git history
* recursive (`bool`, default: `false`): clone submodules recursively

#### url

```yaml title="Test definition from url" hl_lines="4-5"
- test:
    definitions:
    - name: smoke-tests
      from: url
      repository: https://github.com/Linaro/test-definitions/releases/download/2025.10.01/2025.10.tar.zst
      compression: zstd
      path: automated/linux/smoke/smoke.yaml
```

The repository must be a tar archive containing the test definitions.

`compression` is an optional string for providing decompression method to apply
before extracting. Supported: `bz2`, `gz`, `xz` and `zstd`.

#### inline

A simple test definition present in the same file as the job submission,
instead of from a separate file or VCS repository. This allows tests to be run
based on a single file. When combined with `file://` URLs to the `deploy`
parameters, this allows tests to run without needing external access.

```yaml title="Test definition from inline" hl_lines="4-14"
- test:
    definitions:
    - name: health-checks
      from: inline
      path: inline/health-checks.yaml
      repository:
        metadata:
          format: Lava-Test Test Definition 1.0
          name: health checks
        run:
          steps:
            - lava-test-case check-version --shell cat /proc/version
            - lava-test-case lscpu --shell lscpu
            - lava-test-case lsblk --shell lsblk
```

### path

The path within that repository to the YAML file containing the test definition.

### parameters

(optional) Pass parameters to the Lava Test Shell Definition.

The key name can be `parameters` or `params`, both are supported.

The format is a YAML dictionary - the dictionary key is the name of the variable
to be made available to the test shell, the value is the value of that variable.

```yaml title="Test definition parameters" hl_lines="7-9"
- test:
    definitions:
    - name: smoke-test-lsblk
      from: git
      repository: https://github.com/Linaro/test-definitions
      path: automated/linux/smoke/smoke.yaml
      parameters:
        "SKIP_INSTALL": "true"
        "TESTS": "lsblk"
```

### expected

(optional) Provide an expected test case list. Both unexpected and missing test
cases are reported as fail.

```yaml hl_lines="10-11"
- test:
  definitions:
    - name: smoke-test-lsblk
      from: git
      repository: https://github.com/Linaro/test-definitions
      path: automated/linux/smoke/smoke.yaml
      parameters:
        "SKIP_INSTALL": "true"
        "TESTS": "lsblk"
      expected:
        - lsblk
```

For test definitions from `inline`, the expected list must be defined at the job
level using the `expected` key, **not** inside the inline definition.

```yaml hl_lines="14-16"
  - test:
      definitions:
        - from: inline
          repository:
            metadata:
                format: Lava-Test Test Definition 1.0
                name: expected-tests
            run:
                steps:
                    - lava-test-case tc1 --result pass
                    - lava-test-case tc2 --result fail
          name: expected-tests
          path: inline/expected-tests.yaml
          expected:
            - tc1
            - tc2
```

!!!note "Always reported"
    Expected test cases defined in the job definition are always reported, even
    if the tests are not executed due to earlier job errors.

See also
[expected test cases in the test definition](../../../user/basic-tutorials/test-definition.md#expected).

### Additional support

#### Result checks

LAVA collects results from internal operations, these form the `lava` test
suite results as well as from the submitted test definitions. The full set of
results for a job are available at:

```sh
results/1234
```

LAVA records when a submitted test definition starts execution on the test
device. If the number of test definitions which started is not the same as the
number of test definitions submitted (allowing for the `lava` test suite
results), a warning will be displayed on this page.

#### TestSets

A TestSet is a group of lava test cases which will be collated within the LAVA
Results. This allows queries to look at a set of related test cases within a
single definition.

```yaml
- test:
   definitions:
   - repository:
       run:
         steps:
         - lava-test-set start first_set
         - lava-test-case date --shell ntpdate-debian
         - ls /
         - lava-test-case mount --shell mount
         - lava-test-set stop
         - lava-test-case uname --shell uname -a
```

This results in the `date` and `mount` test cases being included into a
`first_set` TestSet, independent of other test cases. The TestSet is
concluded with the `lava-test-set stop` command, meaning that the `uname`
test case has no test set, providing a structure like:

```yaml
results:
  first_set:
    date: pass
    mount: pass
  uname: pass
```

```python
{'results': {'first_set': {'date': 'pass', 'mount': 'pass'}, 'uname': 'pass'}}
```

Each TestSet name must be valid as a URL.

For test job `1234`, the `uname` test case would appear as:

```sh
results/1234/testset-def/uname
```

The `date` and `mount` test cases are referenced via the TestSet:

```sh
results/1234/testset-def/first_set/date
results/1234/testset-def/first_set/mount
```

A single test definition can start and stop different TestSets in sequence, as
long as the name of each TestSet is unique for that test definition.

See also [test definitions](../../../user/basic-tutorials/test-definition.md)

## Monitors

Test jobs using Monitors **must**:

1. Execute automatically after boot.

2. Emit a unique `start` string before test runs.

3. Emit a unique `end` string after tests complete.

4. Provide a regex that captures test output and maps it to results without
   producing excessively long test case names.

### name

The name of the test suite.

### start

String or regex pattern that signals the beginning of test suite output.

If `start` does not match, the job will time out with no results.

### end

String or regex pattern that signals the end of test suite output.

If `end` does not match, the job will time out but the results of the current
boot will already have been reported.

!!! note
    `start` and `end` strings will match part of a line but make sure that each
    string is long enough that it can only match once per boot.

### pattern

Regex with named capture groups to extract test results from the test output.

For the following Zephyr test suite monitor:

```yaml
- test:
    monitors:
    - name: tests
      start: BOOTING ZEPHYR
      end: PROJECT EXECUTION SUCCESSFUL
      pattern: '(?P<test_case_id>\d+ *- [^-]+) (?P<measurement>\d+) tcs = [0-9]+ nsec'
      fixupdict:
        PASS: pass
        FAIL: fail
```

If the device output is of the form:

```text
***** BOOTING ZEPHYR OS v1.7.99 - BUILD: Apr 18 2018 10:00:55 *****
|-----------------------------------------------------------------------------|
|                            Latency Benchmark                                |
|-----------------------------------------------------------------------------|
|  tcs = timer clock cycles: 1 tcs is 12 nsec                                 |
|-----------------------------------------------------------------------------|
| 1 - Measure time to switch from ISR back to interrupted thread              |
| switching time is 107 tcs = 1337 nsec                                       |
|-----------------------------------------------------------------------------|

...

PROJECT EXECUTION SUCCESSFUL
```

The above regular expression can result in test case names like:

```text
1_measure_time_to_switch_from_isr_back_to_interrupted_thread_switching_time_is
```

The raw data will be logged as:

```text
test_case_id: 1 - Measure time to switch from ISR back to interrupted thread              |
| switching time is
```

!!! warning
    Notice how the regular expression has not closed the match at the end of
    the "line" but has continued on to the first non-matching character. The
    test case name then concatenates all whitespace and invalid characters to a
    single underscore. LAVA uses pexpect to perform output parsing. pexpect
    docs explain how to find line ending strings:
    <https://pexpect.readthedocs.io/en/stable/overview.html#find-the-end-of-line-cr-lf-conventions>

```python
r'(?P<test_case_id>\d+ *- [^-]+) (?P<measurement>\d+) tcs = [0-9]+ nsec'
```

The test_case_id will be formed from the match of the expression `\d+ *- [^-]+`
followed by a single space - but **only** if the rest of the expression matches
as well.

The measurement will be taken from the match of the expression `\d+` preceded
by a single space and followed by the **exact** string `tcs = ` which itself
must be followed by a number of digits, then a single space and finally the
**exact** string `nsec` - but only if the rest of the expression also matches.

See also: [Regular Expression HOWTO for Python3][regex]

### fixupdict

(optional) Maps non-standard result strings in test output to LAVA's expected
values. It is a translation map for raw result values.

```yaml title="fixupdict example"
fixupdict:
  PASSED: pass
  FAILED: fail
  SKIPPED: skip
```

### expected

(optional) Provide an expected test case list. Both unexpected and missing test
cases are reported as fail.

```yaml title="Expected test cases"
- test:
    monitors:
    - name: "mcuboot_suite"
      start: "Execute test suites for the MCUBOOT area"
      end: "End of MCUBOOT test suites"
      pattern: "TEST: (?P<test_case_id>.+?) - (?P<result>(PASSED|FAILED|SKIPPED))"
      fixupdict:
         'PASSED': pass
         'FAILED': fail
         'SKIPPED': skip
      expected:
      - "tfm_mcuboot_integration_test_0001"
```

!!!note "Always reported"
    Expected test cases provided for the monitors are always reported, even if
    the tests are not executed due to earlier job errors.

### monitors

Multiple monitors can be defined for a single boot. LAVA processes each monitor
sequentially — waiting for the first monitor's `start` pattern, parsing until
its `end`, then moving to the next monitor.

```yaml title="Monitors for TF-M test suite"
- test:
    monitors:
    - name: "secure_regression_suite"
      start: "Execute test suites for the Secure area"
      end: "End of Secure test suites"
      pattern: "'(?P<test_case_id>TFM_S_[A-Z]+_TEST_[0-9]+)'.*?(?P<result>PASSED|FAILED|SKIPPED|Assertion failed)"
      fixupdict:
         'PASSED': pass
         'FAILED': fail
         'SKIPPED': skip
    - name: "non_secure_regression_suite"
      start: "Execute test suites for the Non-secure area"
      end: "End of Non-secure test suites"
      pattern: "TEST: (?P<test_case_id>.+?) - (?P<result>(PASSED|FAILED|SKIPPED))"
      fixupdict:
         'PASSED': pass
         'FAILED': fail
         'SKIPPED': skip

```

## Interactive

An interactive test action allows to interact with a non-POSIX shell or just
arbitrary interactive application. For instance, the shell of u-boot bootloader.

```yaml title="U-boot interactive tests"
actions:
- boot:
    method: bootloader
    bootloader: u-boot
    commands: []
    prompts:
    - "U-Boot>"

- test:
    interactive:
    - name: network
      prompts:
      - "U-Boot>"
      - "/ # "
      echo: discard
      script:
      - name: dhcp
        command: dhcp
        successes:
        - message: "DHCP client bound to address"
        failures:
        - message: "TIMEOUT"
          exception: InfrastructureError
          error: "dhcp failed"
      - name: setenv
        command: "setenv serverip {SERVER_IP}"
```

The workflow of the interactive test action is:

1. Send the `command` to the DUT, unless empty.
2. If `echo: discard` is specified, discard next output line (assumed to be an
   echo of the command).
3. Wait for a match from the `prompts`, `successes` or `failures`.
4. If a `name` is defined, log the result for this command (as soon as a prompt
   or a message is matched).
5. If a `successes` or `failures` was matched, wait for the `prompts` before
   proceeding.

!!! note
    The interactive test action expects the prompt to be already matched
    before it starts. If this is not the case, then you will to wait the prompt
    by adding an empty `command` directive as described below. Note that empty
    `command:` is different from empty string `command: ""`. In the latter
    case, a newline will be sent to device.

```yaml title="Wait for prompt first" hl_lines="9-10"
- test:
    interactive:
    - name: network
      prompts:
      - "U-Boot>"
      - "/ # "
      echo: discard
      script:
      - name: 'match-prompt'
        command:
      - name: setenv
        command: "setenv serverip {SERVER_IP}"
```

### name

A unique test suite name across all interactive test actions in the same
namespace.

### prompts

The list of possible prompts for the interactive session. In many cases, there
is just one prompt, but if shell has different prompts for different states, it
can be accommodated. Prompts can also include regular expressions, as any other
match strings.

### echo

If set to `discard`, after each sent `command` of a `script`, discard the next
output line (assumed to be an echo of the command). This option should be set
when interacting with shell (like u-boot shell) that will echo the command, to
avoid false positive matches. Note that this options applies to every `command`
in the script. If you need different value of this option for different
commands, you would need to group them in different `script`s.

### script

A list of commands to send and what kind of output to expect.

#### name

If present, log the result (pass/fail) of this command under the given name as
a test case). If not present, and the command fails, the entire test will
fail with `TestError`.

#### command

The command (string) to send to device, followed by newline. The command can use
variables that will be substituted with live data, like `{SERVER_IP}`. If
command value is empty (`command:` in YAML), nothing is sent, but output
matching (prompts/successes/failures) will be performed as usual.

#### successes

A list of dictionaries with a single `message` key. The message should be a list
of string or regex to match. It uses substring matching, so `message: 4` matches
"14", "41", etc.

#### failures

A list of dictionaries with:

* message (`str`): the string or regex to match (substring matching).
* exception (`str`, optional): if the message indicates a fatal problem,
  an exception can be raised. The exception can be one of `InfrastructureError`,
  `JobError`, or `TestError`.  If omitted, the failure is recorded as a failed
  test case without stopping the job.
* error (`str`, optional): custom exception message for the job log.

If `successes` is defined, but LAVA matches one of the prompts instead, an
error will be recorded (following the logic that the lack of expected success
output is an error). This means that in many cases you don't need to specify
`failures` - any output but the successes will be recorded as an error.

However, if `successes` is not defined, then matching a prompt will generate a
passing result. This is useful for interactive commands which don't generate
any output on success; of course, in this case you would need to specify
`failures` to catch them.

### expected

(optional) Provide an expected test case list. Both unexpected and missing test
cases are reported as fail.

```yaml title="Expected test cases" hl_lines="9-11"
- test:
    interactive:
    - name: network
      prompts:
      - "U-Boot>"
      script:
      - name: setenv-serverip
        command: "setenv serverip {SERVER_IP}"
      expected:
      - setenv-serverip
```

!!!note "Always reported"
    The expected test cases are always reported, even if the tests are not
    executed due to earlier job errors.

## Services

The feature is disabled by default. It must be explicitly enabled in the device
dictionary by lab admin or device owner. And the worker that the device attached
to should have access to the Docker images needed by the services.

!!! danger
    Enabling the feature allows the users of the device to run any Docker
    containers with any permissions on the LAVA worker. Access to the device
    should be strictly restricted to trusted users using LAVA user groups.

```jinja2 title="Enable test services"
{% set allow_test_services = true %}
```

The services are defined in a `test` action using the `services` key.

```yaml title="Custom test services"
- test:
    services:
    - name: srv1
      from: git
      repository: https://example.com/org/srv1.git
      path: docker-compose.yml
    - name: srv2
      from: url
      repository: https://example.com/srv2.tar.gz
      path: docker-compose.yaml
      service: srv-name
      compression: gz
```

### name

Unique test service name.

### from

The type of the repository. Supported: `git` or `url`.

#### git

See [git](#git)

#### url

See [url](#url)

### repository

A publicly readable repository location.

### path

Path to the Docker Compose file within the repository.

### service

A specific service to start. If not set, all services in the compose file are
started.

### Environment

LAVA dispatcher writes environment variables from both the
[device dictionary](../../configuration/device-dictionary.md#environment)
and the job-level [`environment`](../environment.md) to the repo's `.env`
file. Values defined at the job level take precedence over those from the
device dictionary.

Docker Compose reads the `.evn` automatically. This means you can define
an environment variable in the device or job:

```jinja title="Device dictionary environment"
{% set environment = { 'BOARD_QDL_ID': 'D902AA38'} %}
```

And then expose the variable to your Docker container:

```yaml title="Docker Compose environment"
services:
  srv1:
    environment:
      - BOARD_QDL_ID
```

!!! note
    This is defined in your Docker Compose file, not the LAVA job
    definition. For more about exposing variables inside your container,
    see [set environment variables](https://docs.docker.com/compose/how-tos/environment-variables/set-environment-variables/).

### Lifecycle

Services started remain active until the end of the job. You can use command
`stop_test_services` to stop services earlier.

```yaml title="Stop test services"
- command:
    name: stop_test_services
```

--8<-- "refs.txt"

# Job output

A simple [QEMU job page](https://validation.linaro.org/scheduler/job/2009038) will look like this:

![job output](job-output.png)

## Header

At the top of the page, job meta-data are listed in two columns:

* **static data** on the left (description, device-type, submitter, ...)
* **dynamic data** on the right (job id, actual device, duration, ...)

## Summary

This is the main tabular as it includes the logs of the job.

You can select the levels that should be displayed or hidden by clicking on the corresponding buttons:

* `debug`, `info`, `warning` and `error`: LAVA logs
* `input`: DUT input
* `target` and `feedback`: DUT output
* `results`: LAVA results

!!! tip "`target` vs `feedback`"
    Both `target` and `feedback` are DUT output.

    When using [namespaces](../../technical-references/job-definition/job.md#namespace), DUT
    output for the current namespace will be in `target` while outputs coming
    from over namespace will be in `feedback`.

### Pipeline

On the right side of the page, the `pipeline` element allows you to jump
directly at the log of a specific action of the pipeline.


## Details

In this tabular, you have access to the internal details of the LAVA pipeline
that was running for this specific job definition.

## Timing

This tabular list each individual action with it's duration along with its timeout.

This helps test writers to check for actions with a duration which is
much shorter than the requested timeout. Reducing these timeouts will allow
failures to be identified more quickly.

The graph only shows actions that are longer than 1 second. The full list is
available in the table.

--8<-- "refs.txt"

# Test Actions

The test action executes tests on the DUT after booting. It collects output,
parses results, and reports test outcomes.

**Key rules for using test actions:**

* A test action **never** boots the device - a [boot action](./boot/common.md)
  **must** be defined before it.

* LAVA support scripts are prepared by the [deploy action](deploy/index.md) and
  remain available for all test definitions until the next deploy block.

* Multiple test definitions can be listed within the same test block to run
  sequentially without rebooting the device.

There are 4 types of test actions:

* [**Lava-test-shell definitions**](#definitions) (YAML directive: `definitions`)
  are used for POSIX compliant operating systems on the DUT. The deployed system
  is expected to support a POSIX shell environment (`/bin/ash`, `/bin/dash` or
  `/bin/bash` are the most common) so that LAVA can execute the LAVA Test Shell
  Helper scripts.

* [**Output monitors**](#monitors) (YAML directive: `monitors`) are used for
  devices which have no POSIX shell and start the test (and corresponding output)
  immediately after booting, for example microcontroller/IoT boards.

* [**Interactive tests**](#interactive) (YAML directive: `interactive`) are
  further extension of `monitor` tests idea, allowing not just matching some
  output from a device, but also feeding some input. They are useful for
  non-POSIX shells like bootloaders (u-boot for instance) and other interactive
  command-line applications.

* [**Test services**](#services) (YAML directive: services) are used for
  starting docker-compose based containers on the LAVA worker. It enables
  running custom services for the needs in test definitions.

## Definitions

```yaml title="Lava-test-shell definitions"
- test:
    timeout:
      minutes: 5
    definitions:
    - name: smoke-tests
      from: git
      repository: https://github.com/Linaro/test-definitions
      path: automated/linux/smoke/smoke.yaml
    - name: smoke-tests-net
      from: git
      repository: https://gitlab.com/lava/functional-tests.git
      path: posix/smoke-tests-net.yaml
```

### name

The unique test suite name to use. It replaces the name from the YAML.

### repository

A publicly readable repository location.

### from

The type of the repository is **not** guessed, it **must** be specified
explicitly. Support is available for [git](#git), [url](#url) and [inline](#inline).

#### git

A remote git repository which needs to be cloned by the LAVA worker.

Optional parameters:

* revision (`str`, default: `null`): specific commit SHA to check out
* branch (`str`, default: `null`): git branch to clone
* shallow (`bool`, default: `true`): perform a shallow git clone
* history (`bool`, default: `false`): include full git history
* recursive (`bool`, default: `false`): clone submodules recursively

#### url

```yaml title="Test definition from url" hl_lines="4-5"
- test:
    definitions:
    - name: smoke-tests
      from: url
      repository: https://github.com/Linaro/test-definitions/releases/download/2025.10.01/2025.10.tar.zst
      compression: zstd
      path: automated/linux/smoke/smoke.yaml
```

The repository must be a tar archive containing the test definitions.

`compression` is an optional string for providing decompression method to apply
before extracting. Supported: `bz2`, `gz`, `xz` and `zstd`.

#### inline

A simple test definition present in the same file as the job submission,
instead of from a separate file or VCS repository. This allows tests to be run
based on a single file. When combined with `file://` URLs to the `deploy`
parameters, this allows tests to run without needing external access.

```yaml title="Test definition from inline" hl_lines="4-14"
- test:
    definitions:
    - name: health-checks
      from: inline
      path: inline/health-checks.yaml
      repository:
        metadata:
          format: Lava-Test Test Definition 1.0
          name: health checks
        run:
          steps:
            - lava-test-case check-version --shell cat /proc/version
            - lava-test-case lscpu --shell lscpu
            - lava-test-case lsblk --shell lsblk
```

### path

The path within that repository to the YAML file containing the test definition.

### parameters

(optional) Pass parameters to the Lava Test Shell Definition.

The key name can be `parameters` or `params`, both are supported.

The format is a YAML dictionary - the dictionary key is the name of the variable
to be made available to the test shell, the value is the value of that variable.

```yaml title="Test definition parameters" hl_lines="7-9"
- test:
    definitions:
    - name: smoke-test-lsblk
      from: git
      repository: https://github.com/Linaro/test-definitions
      path: automated/linux/smoke/smoke.yaml
      parameters:
        "SKIP_INSTALL": "true"
        "TESTS": "lsblk"
```

### expected

(optional) Provide an expected test case list. Both unexpected and missing test
cases are reported as fail.

```yaml hl_lines="10-11"
- test:
  definitions:
    - name: smoke-test-lsblk
      from: git
      repository: https://github.com/Linaro/test-definitions
      path: automated/linux/smoke/smoke.yaml
      parameters:
        "SKIP_INSTALL": "true"
        "TESTS": "lsblk"
      expected:
        - lsblk
```

For test definitions from `inline`, the expected list must be defined at the job
level using the `expected` key, **not** inside the inline definition.

```yaml hl_lines="14-16"
  - test:
      definitions:
        - from: inline
          repository:
            metadata:
                format: Lava-Test Test Definition 1.0
                name: expected-tests
            run:
                steps:
                    - lava-test-case tc1 --result pass
                    - lava-test-case tc2 --result fail
          name: expected-tests
          path: inline/expected-tests.yaml
          expected:
            - tc1
            - tc2
```

!!!note "Always reported"
    Expected test cases defined in the job definition are always reported, even
    if the tests are not executed due to earlier job errors.

See also
[expected test cases in the test definition](../../../user/basic-tutorials/test-definition.md#expected).

### Additional support

#### Result checks

LAVA collects results from internal operations, these form the `lava` test
suite results as well as from the submitted test definitions. The full set of
results for a job are available at:

```sh
results/1234
```

LAVA records when a submitted test definition starts execution on the test
device. If the number of test definitions which started is not the same as the
number of test definitions submitted (allowing for the `lava` test suite
results), a warning will be displayed on this page.

#### TestSets

A TestSet is a group of lava test cases which will be collated within the LAVA
Results. This allows queries to look at a set of related test cases within a
single definition.

```yaml
- test:
   definitions:
   - repository:
       run:
         steps:
         - lava-test-set start first_set
         - lava-test-case date --shell ntpdate-debian
         - ls /
         - lava-test-case mount --shell mount
         - lava-test-set stop
         - lava-test-case uname --shell uname -a
```

This results in the `date` and `mount` test cases being included into a
`first_set` TestSet, independent of other test cases. The TestSet is
concluded with the `lava-test-set stop` command, meaning that the `uname`
test case has no test set, providing a structure like:

```yaml
results:
  first_set:
    date: pass
    mount: pass
  uname: pass
```

```python
{'results': {'first_set': {'date': 'pass', 'mount': 'pass'}, 'uname': 'pass'}}
```

Each TestSet name must be valid as a URL.

For test job `1234`, the `uname` test case would appear as:

```sh
results/1234/testset-def/uname
```

The `date` and `mount` test cases are referenced via the TestSet:

```sh
results/1234/testset-def/first_set/date
results/1234/testset-def/first_set/mount
```

A single test definition can start and stop different TestSets in sequence, as
long as the name of each TestSet is unique for that test definition.

See also [test definitions](../../../user/basic-tutorials/test-definition.md)

## Monitors

Test jobs using Monitors **must**:

1. Execute automatically after boot.

2. Emit a unique `start` string before test runs.

3. Emit a unique `end` string after tests complete.

4. Provide a regex that captures test output and maps it to results without
   producing excessively long test case names.

### name

The name of the test suite.

### start

String or regex pattern that signals the beginning of test suite output.

If `start` does not match, the job will time out with no results.

### end

String or regex pattern that signals the end of test suite output.

If `end` does not match, the job will time out but the results of the current
boot will already have been reported.

!!! note
    `start` and `end` strings will match part of a line but make sure that each
    string is long enough that it can only match once per boot.

### pattern

Regex with named capture groups to extract test results from the test output.

For the following Zephyr test suite monitor:

```yaml
- test:
    monitors:
    - name: tests
      start: BOOTING ZEPHYR
      end: PROJECT EXECUTION SUCCESSFUL
      pattern: '(?P<test_case_id>\d+ *- [^-]+) (?P<measurement>\d+) tcs = [0-9]+ nsec'
      fixupdict:
        PASS: pass
        FAIL: fail
```

If the device output is of the form:

```text
***** BOOTING ZEPHYR OS v1.7.99 - BUILD: Apr 18 2018 10:00:55 *****
|-----------------------------------------------------------------------------|
|                            Latency Benchmark                                |
|-----------------------------------------------------------------------------|
|  tcs = timer clock cycles: 1 tcs is 12 nsec                                 |
|-----------------------------------------------------------------------------|
| 1 - Measure time to switch from ISR back to interrupted thread              |
| switching time is 107 tcs = 1337 nsec                                       |
|-----------------------------------------------------------------------------|

...

PROJECT EXECUTION SUCCESSFUL
```

The above regular expression can result in test case names like:

```text
1_measure_time_to_switch_from_isr_back_to_interrupted_thread_switching_time_is
```

The raw data will be logged as:

```text
test_case_id: 1 - Measure time to switch from ISR back to interrupted thread              |
| switching time is
```

!!! warning
    Notice how the regular expression has not closed the match at the end of
    the "line" but has continued on to the first non-matching character. The
    test case name then concatenates all whitespace and invalid characters to a
    single underscore. LAVA uses pexpect to perform output parsing. pexpect
    docs explain how to find line ending strings:
    <https://pexpect.readthedocs.io/en/stable/overview.html#find-the-end-of-line-cr-lf-conventions>

```python
r'(?P<test_case_id>\d+ *- [^-]+) (?P<measurement>\d+) tcs = [0-9]+ nsec'
```

The test_case_id will be formed from the match of the expression `\d+ *- [^-]+`
followed by a single space - but **only** if the rest of the expression matches
as well.

The measurement will be taken from the match of the expression `\d+` preceded
by a single space and followed by the **exact** string `tcs = ` which itself
must be followed by a number of digits, then a single space and finally the
**exact** string `nsec` - but only if the rest of the expression also matches.

See also: [Regular Expression HOWTO for Python3][regex]

### fixupdict

(optional) Maps non-standard result strings in test output to LAVA's expected
values. It is a translation map for raw result values.

```yaml title="fixupdict example"
fixupdict:
  PASSED: pass
  FAILED: fail
  SKIPPED: skip
```

### expected

(optional) Provide an expected test case list. Both unexpected and missing test
cases are reported as fail.

```yaml title="Expected test cases"
- test:
    monitors:
    - name: "mcuboot_suite"
      start: "Execute test suites for the MCUBOOT area"
      end: "End of MCUBOOT test suites"
      pattern: "TEST: (?P<test_case_id>.+?) - (?P<result>(PASSED|FAILED|SKIPPED))"
      fixupdict:
         'PASSED': pass
         'FAILED': fail
         'SKIPPED': skip
      expected:
      - "tfm_mcuboot_integration_test_0001"
```

!!!note "Always reported"
    Expected test cases provided for the monitors are always reported, even if
    the tests are not executed due to earlier job errors.

### monitors

Multiple monitors can be defined for a single boot. LAVA processes each monitor
sequentially — waiting for the first monitor's `start` pattern, parsing until
its `end`, then moving to the next monitor.

```yaml title="Monitors for TF-M test suite"
- test:
    monitors:
    - name: "secure_regression_suite"
      start: "Execute test suites for the Secure area"
      end: "End of Secure test suites"
      pattern: "'(?P<test_case_id>TFM_S_[A-Z]+_TEST_[0-9]+)'.*?(?P<result>PASSED|FAILED|SKIPPED|Assertion failed)"
      fixupdict:
         'PASSED': pass
         'FAILED': fail
         'SKIPPED': skip
    - name: "non_secure_regression_suite"
      start: "Execute test suites for the Non-secure area"
      end: "End of Non-secure test suites"
      pattern: "TEST: (?P<test_case_id>.+?) - (?P<result>(PASSED|FAILED|SKIPPED))"
      fixupdict:
         'PASSED': pass
         'FAILED': fail
         'SKIPPED': skip

```

## Interactive

An interactive test action allows to interact with a non-POSIX shell or just
arbitrary interactive application. For instance, the shell of u-boot bootloader.

```yaml title="U-boot interactive tests"
actions:
- boot:
    method: bootloader
    bootloader: u-boot
    commands: []
    prompts:
    - "U-Boot>"

- test:
    interactive:
    - name: network
      prompts:
      - "U-Boot>"
      - "/ # "
      echo: discard
      script:
      - name: dhcp
        command: dhcp
        successes:
        - message: "DHCP client bound to address"
        failures:
        - message: "TIMEOUT"
          exception: InfrastructureError
          error: "dhcp failed"
      - name: setenv
        command: "setenv serverip {SERVER_IP}"
```

The workflow of the interactive test action is:

1. Send the `command` to the DUT, unless empty.
2. If `echo: discard` is specified, discard next output line (assumed to be an
   echo of the command).
3. Wait for a match from the `prompts`, `successes` or `failures`.
4. If a `name` is defined, log the result for this command (as soon as a prompt
   or a message is matched).
5. If a `successes` or `failures` was matched, wait for the `prompts` before
   proceeding.

!!! note
    The interactive test action expects the prompt to be already matched
    before it starts. If this is not the case, then you will to wait the prompt
    by adding an empty `command` directive as described below. Note that empty
    `command:` is different from empty string `command: ""`. In the latter
    case, a newline will be sent to device.

```yaml title="Wait for prompt first" hl_lines="9-10"
- test:
    interactive:
    - name: network
      prompts:
      - "U-Boot>"
      - "/ # "
      echo: discard
      script:
      - name: 'match-prompt'
        command:
      - name: setenv
        command: "setenv serverip {SERVER_IP}"
```

### name

A unique test suite name across all interactive test actions in the same
namespace.

### prompts

The list of possible prompts for the interactive session. In many cases, there
is just one prompt, but if shell has different prompts for different states, it
can be accommodated. Prompts can also include regular expressions, as any other
match strings.

### echo

If set to `discard`, after each sent `command` of a `script`, discard the next
output line (assumed to be an echo of the command). This option should be set
when interacting with shell (like u-boot shell) that will echo the command, to
avoid false positive matches. Note that this options applies to every `command`
in the script. If you need different value of this option for different
commands, you would need to group them in different `script`s.

### script

A list of commands to send and what kind of output to expect.

#### name

If present, log the result (pass/fail) of this command under the given name as
a test case). If not present, and the command fails, the entire test will
fail with `TestError`.

#### command

The command (string) to send to device, followed by newline. The command can use
variables that will be substituted with live data, like `{SERVER_IP}`. If
command value is empty (`command:` in YAML), nothing is sent, but output
matching (prompts/successes/failures) will be performed as usual.

#### successes

A list of dictionaries with a single `message` key. The message should be a list
of string or regex to match. It uses substring matching, so `message: 4` matches
"14", "41", etc.

#### failures

A list of dictionaries with:

* message (`str`): the string or regex to match (substring matching).
* exception (`str`, optional): if the message indicates a fatal problem,
  an exception can be raised. The exception can be one of `InfrastructureError`,
  `JobError`, or `TestError`.  If omitted, the failure is recorded as a failed
  test case without stopping the job.
* error (`str`, optional): custom exception message for the job log.

If `successes` is defined, but LAVA matches one of the prompts instead, an
error will be recorded (following the logic that the lack of expected success
output is an error). This means that in many cases you don't need to specify
`failures` - any output but the successes will be recorded as an error.

However, if `successes` is not defined, then matching a prompt will generate a
passing result. This is useful for interactive commands which don't generate
any output on success; of course, in this case you would need to specify
`failures` to catch them.

### expected

(optional) Provide an expected test case list. Both unexpected and missing test
cases are reported as fail.

```yaml title="Expected test cases" hl_lines="9-11"
- test:
    interactive:
    - name: network
      prompts:
      - "U-Boot>"
      script:
      - name: setenv-serverip
        command: "setenv serverip {SERVER_IP}"
      expected:
      - setenv-serverip
```

!!!note "Always reported"
    The expected test cases are always reported, even if the tests are not
    executed due to earlier job errors.

## Services

The feature is disabled by default. It must be explicitly enabled in the device
dictionary by lab admin or device owner. And the worker that the device attached
to should have access to the Docker images needed by the services.

!!! danger
    Enabling the feature allows the users of the device to run any Docker
    containers with any permissions on the LAVA worker. Access to the device
    should be strictly restricted to trusted users using LAVA user groups.

```jinja2 title="Enable test services"
{% set allow_test_services = true %}
```

The services are defined in a `test` action using the `services` key.

```yaml title="Custom test services"
- test:
    services:
    - name: srv1
      from: git
      repository: https://example.com/org/srv1.git
      path: docker-compose.yml
    - name: srv2
      from: url
      repository: https://example.com/srv2.tar.gz
      path: docker-compose.yaml
      service: srv-name
      compression: gz
```

### name

Unique test service name.

### from

The type of the repository. Supported: `git` or `url`.

#### git

See [git](#git)

#### url

See [url](#url)

### repository

A publicly readable repository location.

### path

Path to the Docker Compose file within the repository.

### service

A specific service to start. If not set, all services in the compose file are
started.

### Lifecycle

Services started remain active until the end of the job. You can use command
`stop_test_services` to stop services earlier.

```yaml title="Stop test services"
- command:
    name: stop_test_services
```

--8<-- "refs.txt"

# State machine

The state machine describes and controls the state and health of Workers,
Devices and Test Jobs.

## Workers

For each worker, two variables describe the current status: **state** and **health**.

### State

The **state** is an internal variable, set by [lava-server-gunicorn](./services/lava-server-gunicorn.md) when receiving (or not) pings from each worker.

* *Online*: the worker is sending PING to the server
* *Offline*: the worker hasn't sent any messages for a while

### Health

When worker **health** is set to *Maintenance*, no jobs will be ran on the
attached devices.

The worker health can be:

* *Active*
* *Maintenance*
* *Retired*


!!! warning
    When a worker is *Offline*, none of the attached devices will be used to schedule new jobs.

## Devices


For each device, two variables describe the current status: **state** and **health**.

### State

The **state** is an internal variable, set by [lava-scheduler](./services/lava-scheduler.md) and [lava-server-gunicorn](./services/lava-server-gunicorn.md) when scheduling, starting, canceling and ending test jobs.

* *Idle*: not in use by any test job
* *Reserved*: has been reserved for a test job but the test job is not running yet
* *Running*: currently running a test job

### Health

The **health** can be used by admins to indicate if a device should be used by the scheduler or not.

Moreover, when ending an health-check, the device health will be set according to the test job health.

* *Good*: the device passed the health-check
* *Unknown*
* *Looping*: should run health-checks in a loop
* *Bad*: the device failed the health-check
* *Maintenance*
* *Retired*


## TestJobs

For each test job, two variables are describing the current status: **state** and **health**.

### State

* *Submitted*: waiting in the queue
* *Scheduling*: part of a multinode test job where some sub-jobs are still in *Submitted*
* *Scheduled*: has been scheduled. For multinode, it means that all sub-jobs are also scheduled
* *Running*: currently running on a device
* *Canceling*: has been canceled but not ended yet
* *Finished*

!!! note Multinode scheduling
    Only multinode test jobs use *Scheduling*. When all
    sub-jobs are in *Scheduling*, [lava-scheduler](./services/lava-scheduler.md) will transition all test
    jobs to *Scheduled*.

### Health

* *Unknown*: default value that will be overridden when the job is finished.
* *Complete*: the job was able to finish
* *Incomplete*: the job was not able to dinish
* *Canceled*: the test job was canceled.


# Scheduler

The scheduler is called by [lava-scheduler](./services/lava-scheduler.md)
approximately every 20 seconds or when receiving specific events.
The scheduler starts by scheduling health-checks. The remaining devices are
then considered for test jobs.

## Health-checks

To ensure that health-checks are always scheduled when needed, they will be
considered first by the scheduler before regular test jobs.

The scheduler will only consider devices where:

* `device state` is *Idle*
* `device health` is *Good*, *Unknown* or *Looping*
* `worker state` is *Online*

## Test jobs

The scheduler will only consider devices where:

* ``device state`` is *Idle*
* ``device health`` is *Good* or *Unknown*
* ``worker state`` is *Online*

# VEMSD

The `vemsd` (Versatile Express MicroSD) deployment method is used to write a
new recovery image to the ARM Versatile Express Hardware Platforms like Juno.

```yaml
- deploy:
    to: vemsd
    recovery_image:
      url: https://storage.lavacloud.io/health-checks/juno/20.01-oe-uboot.zip
      compression: zip
```

Note that the image is unpacked onto the filesystem of the VEMSD by the
deployment method, however, the actual flashing is deferred and executed
automatically during the next boot.

## recovery_image

The image ready to be unpacked onto the MicroSD.

### url

See [url](./index.md#url)

### compression

If the image is compressed, the compression method **must** be specified.

See [compression](./index.md#compression)

# VEMSD

The `vemsd` (Versatile Express MicroSD) deployment method is used to write a
new recovery image to the ARM Versatile Express Hardware Platforms like Juno.

```yaml
- deploy:
    to: vemsd
    recovery_image:
      url: https://storage.lavacloud.io/health-checks/juno-r2/20.01-oe-uboot.zip
      compression: zip
```

Note that the image is unpacked onto the filesystem of the VEMSD by the
deployment method, however, the actual flashing is deferred and executed
automatically during the next boot.

## recovery_image

The image ready to be unpacked onto the MicroSD.

### url

See [url](./index.md#url)

### compression

If the image is compressed, the compression method **must** be specified.

See [compression](./index.md#compression)

# Synchronizing records from device dictionary

LAVA has the ability to synchronize database records with the device
dictionary. It currently supports these models:

* Device
* Device type
* Tag
* Alias
* Physical Owner
* Physical Group
* Group Device Permissions

This can make administration less cumbersome and can help with Ansible and
similar setups. By using this feature administrators are able to keep the list
of these records inside the version control.

For device records, a flag named **is_synced** is used to recognize devices
which are synced to/from the device dictionary. This option can be updated via
usual channels (web UI admin, APIs).

Once the new device dictionary is added to the filesystem it needs to include
the following dictionary variable which will control which records should be
created:

```jinja
{% set sync_to_lava = {
   "device_type": "qemu",
   "worker": "worker-01"
   "tags": ["tag1", "tag2"],
   "aliases": ["alias1", "alias2"],
   "physical_owner": "owner-01",
   "physical_group": "group-01",
   "group_device_permissions": [
       ["change_device", "group-02"],
       ["view_device", "group-02"],
       ["submit_to_device", "group-02"],
   ]
   }
%}
```

A management command is used to perform a sync to the database:

=== "Debian"
    ```shell
    lava-server manage sync
    ```

=== "docker-compose"
    ```shell
    docker-compose exec lava-server lava-server manage maintenance
    ```

You can for example set this to run when the contents of the device dictionary
folder change using a watch utility.

In the above example, a system will create:

* device named after the device dictionary file
* device type named qemu (if it doesn't exist already) and relate this device to it
* worker named worker-01 (if it doesn't exist already) and relate this device to it
* aliases in the list; and relate them to the specified device type
* tags in the list; and relate them to the device
* relate specified physical owner (existed) to the device
* relate specified physical group (existed) to the device
* assign/delete group device permissions to the device

If at any point the device dictionary is removed (and once the sync command is
executed) the system will move the device that was created to it to **Retired**
health. If all the devices of the synced device type are **Retired** that
particular device type will be marked as invisible in the system.

# LAVA publisher

Receive and forward events coming from lava services.

## Command line

This daemon is part of lava-server and is started by: `lava-server manage lava-publisher`.

## Service

The systemd service is called `lava-publisher`.

## Dependencies

lava-publisher should be able to:

* open a socket on `/tmp/lava.events`
* open a socket on port `5500`
* open a socket on port `8001`

The services that generate events
([lava-server-gunicorn](./lava-server-gunicorn.md)) should be able to write to
the local socket.

## Configuration

Daemon start options:

* `/etc/default/lava-publisher`
* `/etc/lava-server/lava-publisher`

Django configuration:

* `/etc/lava-server/settings.conf`
* `/etc/lava-server/settings.yaml`
* `/etc/lava-server/settings.d/*.yaml`

## Logs

The logs are stored in `/var/log/lava-server/lava-publisher.log`

The log rotation is configured in `/etc/logrotate.d/lava-publisher-log`.

# FVP

The FVP device type in LAVA refers to running
[Fixed Virtual Platforms](https://developer.arm.com/tools-and-software/simulation-models/fixed-virtual-platforms).

## LAVA FVP worker setup

LAVA executes FVP devices inside Docker containers. Therefore, like any other
LAVA worker that can run `docker` device types, Docker needs to be
installed.

## Creating device type

[Create the device type](common.md#create-device-type) using the name **`fvp`**.

## Creating device

1. [Add the device](common.md#add-device) using the following settings:
    * **Device Type:** `fvp`
    * **Hostname:** A unique name (e.g., `fvp-01`)
2. [Add the device configuration](common.md#add-device-configuration).

    For a minimal configuration, simply extend the base template:

    ```jinja
    {% extends "fvp.jinja2" %}
    ```

## FVP Docker images

LAVA does not handle the download of FVP binaries. These are assumed to be in
the Docker image defined in the LAVA job.

You can use Docker images published by the
[shrinkwraptool](https://gitlab.arm.com/tooling/shrinkwrap) project. The image
`docker.io/shrinkwraptool/base-slim:2025.12.0` is known to work with LAVA.

## Building FVP Docker images

Here is a sample Dockerfile for building your own Docker images for running
FVPs in LAVA.

```shell
FROM ubuntu:22.04

# Install telnet package
RUN apt-get update && \
    apt-get install --no-install-recommends --yes bc libatomic1 telnet libdbus-1-3 && \
    rm -rf /var/cache/apt

# Add FVP Binaries
RUN mkdir /opt/fvp
ADD Foundation_Platform_11.30_27_Linux64.tgz /opt/fvp
```

This example is for the free Foundation model. Download this by going to
[FVP Downloads](https://developer.arm.com/Tools%20and%20Software/Fixed%20Virtual%20Platforms/Arm%20Architecture%20FVPs)
and downloading the `Foundation_Platform_*_Linux64.tgz` archive. Put the
Dockerfile and the archive in the same directory, then build the image from
there:

```shell
docker build -t fvp_foundation_platform:11.30 .
```

Run the following command to test the image. Adjust the versions and binary
paths as needed. You should see `Hello, 64-bit world!` printed.

```shell
docker run --rm \
  fvp_foundation_platform:11.30 \
  /opt/fvp/Foundation_Platformpkg/models/Linux64_GCC-9.3/Foundation_Platform \
  --image /opt/fvp/Foundation_Platformpkg/examples/hello.axf
```

## Networking inside models

Optionally, if you require networking in the model, here is a way to enable
this.

* Create the `network` file with the following contents:

    ```bash
    #!/bin/bash

    # If the change occurs to the "default" libvirt managed network
    if [ "${1}" = "default" ] ; then
      # If the network is started
      if [ "${2}" = "started" ] ; then
        ip tuntap add mode tap tap01
        ip link set tap01 promisc on
        ip link set tap01 up
        ip link set tap01 master virbr0
      fi
    fi
    ```

* Create the `entrypoint.sh` file with the following contents:

    ```bash
    #!/bin/bash

    set -ex

    /usr/sbin/libvirtd &
    sleep 3

    exec "$@"
    ```

    Use this `Dockerfile` instead for installing the additional packages and
    adding the scripts:

    ```dockerfile
    FROM ubuntu:22.04

    # Install packages
    RUN apt-get update && \
        apt-get install --no-install-recommends --yes bc libatomic1 telnet libdbus-1-3 \
        libvirt-daemon-system iproute2 dnsmasq&& \
        rm -rf /var/cache/apt

    # Add FVP Binaries
    RUN mkdir /opt/fvp
    ADD Foundation_Platform_11.30_27_Linux64.tgz /opt/fvp

    COPY network /etc/libvirt/hooks/network
    RUN chmod +x /etc/libvirt/hooks/network

    COPY entrypoint.sh /usr/local/bin/entrypoint.sh
    RUN chmod +x /usr/local/bin/entrypoint.sh

    ENTRYPOINT ["/usr/local/bin/entrypoint.sh"]
    ```

* Configure your FVP device type to run with privileged mode for creating the
  interfaces.

    ```jinja
    {% extends "fvp.jinja2" %}

    {% set fvp_docker_privileged = True %}
    ```

    !!! warning
        Access to devices running with privileged mode should be strictly
        controlled via user groups to mitigate security concerns.

* In your LAVA job, add the following arguments to your foundation model
(other models will differ):

    ```yaml
    arguments:
    - ...
    - "--network=bridged"
    - "--network-bridge=tap01"
    ```

This will be required if you require the use of `transfer_overlay`. This could
be useful in the event you want to pass binaries to the model that contains the
filesystem but stored in a way LAVA cannot currently put the overlay into.

```yaml
transfer_overlay:
  # It may be required to suppress some kernel messages
  download_command: echo 3 > /proc/sys/kernel/printk ; wget
  unpack_command: tar -C / -xzf
```

## Sample job definition

```yaml
device_type: fvp
job_name: sample fvp job

timeouts:
  job:
    minutes: 60

priority: medium
visibility: public

actions:
- deploy:
    to: fvp
    timeout:
      minutes: 15
    uniquify: false
    images:
      startup:
        # Create the 'startup.nsh' file with the below content, host it locally
        # or on a http file server, and then update the url below.
        # Image dtb=fvp-base-revc.dtb systemd.log_level=warning console=ttyAMA0 earlycon=pl011,0x1c090000 root=/dev/vda ip=dhcp
        url: file:///<path>/startup.nsh
      uefi:
        url: https://storage.tuxboot.com/buildroot/fvp-aemva/FVP_AARCH64_EFI.fd
      bl1:
        url: https://storage.tuxboot.com/buildroot/fvp-aemva/bl1.bin
      fip:
        url: https://storage.tuxboot.com/buildroot/fvp-aemva/fip.bin
      dtb:
        url: https://storage.tuxboot.com/buildroot/fvp-aemva/fvp-base-revc.dtb
      kernel:
        url: https://storage.tuxboot.com/buildroot/fvp-aemva/Image
      rootfs:
        url:
          https://storage.tuxboot.com/buildroot/fvp-aemva/rootfs.ext4.zst
        compression: zstd
        format: ext4
        overlays:
          lava: true

- boot:
    method: fvp
    docker:
      name: docker.io/shrinkwraptool/base-slim:2025.12.0
      local: true
    image: /tools/Base_RevC_AEMvA_pkg/models/Linux64_GCC-9.3/FVP_Base_RevC-2xAEMvA
    version_string: Fast Models [^\n]+
    timeout:
      minutes: 10
    console_string: 'terminal_0: Listening for serial connection on port (?P<PORT>\d+)'
    feedbacks:
    - '(?P<NAME>terminal_1): Listening for serial connection on port (?P<PORT>\d+)'
    - '(?P<NAME>terminal_2): Listening for serial connection on port (?P<PORT>\d+)'
    - '(?P<NAME>terminal_3): Listening for serial connection on port (?P<PORT>\d+)'
    arguments:
    - --stat
    - -C bp.dram_size=4
    - -C bp.flashloader0.fname='{FIP}'
    - -C bp.flashloader1.fname='{UEFI}'
    - -C bp.hostbridge.userNetPorts=8022=22
    - -C bp.hostbridge.userNetworking=1
    - -C bp.refcounter.non_arch_start_at_default=1
    - -C bp.secure_memory=1
    - -C bp.secureflashloader.fname='{BL1}'
    - -C bp.smsc_91c111.enabled=1
    - -C bp.terminal_0.mode=telnet
    - -C bp.terminal_0.start_telnet=0
    - -C bp.terminal_1.mode=raw
    - -C bp.terminal_1.start_telnet=0
    - -C bp.terminal_2.mode=raw
    - -C bp.terminal_2.start_telnet=0
    - -C bp.terminal_3.mode=raw
    - -C bp.terminal_3.start_telnet=0
    - -C bp.ve_sysregs.exit_on_shutdown=1
    - -C bp.virtio_rng.enabled=1
    - -C bp.virtioblockdevice.image_path='{ROOTFS}'
    - -C bp.virtiop9device.root_path=
    - -C bp.vis.disable_visualisation=1
    - -C cache_state_modelled=0
    - -C cluster0.NUM_CORES=4
    - -C cluster0.PA_SIZE=48
    - -C cluster0.check_memory_attributes=0
    - -C cluster0.clear_reg_top_eret=2
    - -C cluster0.cpu0.semihosting-cwd={ARTIFACT_DIR}
    - -C cluster0.ecv_support_level=2
    - -C cluster0.enhanced_pac2_level=3
    - -C cluster0.gicv3.cpuintf-mmap-access-level=2
    - -C cluster0.gicv3.without-DS-support=1
    - -C cluster0.gicv4.mask-virtual-interrupt=1
    - -C cluster0.has_16k_granule=1
    - -C cluster0.has_amu=1
    - -C cluster0.has_arm_v8-1=1
    - -C cluster0.has_arm_v8-2=1
    - -C cluster0.has_arm_v8-3=1
    - -C cluster0.has_arm_v8-4=1
    - -C cluster0.has_arm_v8-5=1
    - -C cluster0.has_arm_v8-6=1
    - -C cluster0.has_arm_v8-7=1
    - -C cluster0.has_arm_v8-8=1
    - -C cluster0.has_arm_v8-9=1
    - -C cluster0.has_arm_v9-0=1
    - -C cluster0.has_arm_v9-1=1
    - -C cluster0.has_arm_v9-2=1
    - -C cluster0.has_arm_v9-3=1
    - -C cluster0.has_arm_v9-4=1
    - -C cluster0.has_arm_v9-5=1
    - -C cluster0.has_branch_target_exception=1
    - -C cluster0.has_brbe=1
    - -C cluster0.has_brbe_v1p1=1
    - -C cluster0.has_const_pac=1
    - -C cluster0.has_gcs=1
    - -C cluster0.has_hpmn0=1
    - -C cluster0.has_large_system_ext=1
    - -C cluster0.has_large_va=1
    - -C cluster0.has_permission_indirection_s1=1
    - -C cluster0.has_permission_indirection_s2=1
    - -C cluster0.has_permission_overlay_s1=1
    - -C cluster0.has_permission_overlay_s2=1
    - -C cluster0.has_rndr=1
    - -C cluster0.has_sve=1
    - -C cluster0.max_32bit_el=0
    - -C cluster0.pmb_idr_external_abort=1
    - -C cluster0.stage12_tlb_size=1024
    - -C cluster0.sve.has_sme2=1
    - -C cluster0.sve.has_sme=1
    - -C cluster0.sve.has_sve2=1
    - -C cluster1.NUM_CORES=4
    - -C cluster1.PA_SIZE=48
    - -C cluster1.check_memory_attributes=0
    - -C cluster1.clear_reg_top_eret=2
    - -C cluster1.ecv_support_level=2
    - -C cluster1.enhanced_pac2_level=3
    - -C cluster1.gicv3.cpuintf-mmap-access-level=2
    - -C cluster1.gicv3.without-DS-support=1
    - -C cluster1.gicv4.mask-virtual-interrupt=1
    - -C cluster1.has_16k_granule=1
    - -C cluster1.has_amu=1
    - -C cluster1.has_arm_v8-1=1
    - -C cluster1.has_arm_v8-2=1
    - -C cluster1.has_arm_v8-3=1
    - -C cluster1.has_arm_v8-4=1
    - -C cluster1.has_arm_v8-5=1
    - -C cluster1.has_arm_v8-6=1
    - -C cluster1.has_arm_v8-7=1
    - -C cluster1.has_arm_v8-8=1
    - -C cluster1.has_arm_v8-9=1
    - -C cluster1.has_arm_v9-0=1
    - -C cluster1.has_arm_v9-1=1
    - -C cluster1.has_arm_v9-2=1
    - -C cluster1.has_arm_v9-3=1
    - -C cluster1.has_arm_v9-4=1
    - -C cluster1.has_arm_v9-5=1
    - -C cluster1.has_branch_target_exception=1
    - -C cluster1.has_brbe=1
    - -C cluster1.has_brbe_v1p1=1
    - -C cluster1.has_const_pac=1
    - -C cluster1.has_gcs=1
    - -C cluster1.has_hpmn0=1
    - -C cluster1.has_large_system_ext=1
    - -C cluster1.has_large_va=1
    - -C cluster1.has_permission_indirection_s1=1
    - -C cluster1.has_permission_indirection_s2=1
    - -C cluster1.has_permission_overlay_s1=1
    - -C cluster1.has_permission_overlay_s2=1
    - -C cluster1.has_rndr=1
    - -C cluster1.has_sve=1
    - -C cluster1.max_32bit_el=0
    - -C cluster1.pmb_idr_external_abort=1
    - -C cluster1.stage12_tlb_size=1024
    - -C cluster1.sve.has_sme2=1
    - -C cluster1.sve.has_sme=1
    - -C cluster1.sve.has_sve2=1
    - -C gic_distributor.has_nmi=1
    - -C pci.pci_smmuv3.mmu.SMMU_AIDR=2
    - -C pci.pci_smmuv3.mmu.SMMU_IDR0=135263935
    - -C pci.pci_smmuv3.mmu.SMMU_IDR1=216481056
    - -C pci.pci_smmuv3.mmu.SMMU_IDR3=5908
    - -C pci.pci_smmuv3.mmu.SMMU_IDR5=4294902901
    - -C pci.pci_smmuv3.mmu.SMMU_S_IDR1=2684354562
    - -C pci.pci_smmuv3.mmu.SMMU_S_IDR2=0
    - -C pci.pci_smmuv3.mmu.SMMU_S_IDR3=0
    - -C pctl.startup=0.0.0.0
    auto_login:
      login_prompt: 'login:'
      username: root
    prompts:
    - root@(.*):[/~]#

- test:
    timeout:
      minutes: 10
    definitions:
    - from: inline
      repository:
        metadata:
          format: Lava-Test Test Definition 1.0
          name: health checks
        run:
          steps:
          - lava-test-case kernel-info --shell uname -a
          - lava-test-case network-info --shell ip a
      name: health-checks
      path: inline/health-checks.yaml
```

# Test definitions

The [test definition](../../technical-references/test-definition.md) is a `yaml` file that describe the tests that you want LAVA
to run on the DUT.

## Smoke example

Let's look at this example:

```yaml
--8<-- "tests/smoke.yaml"
```

## Metadata

The `metadata` dictionary describes the test. `format` and `name` are mandatory
while `description` is optional.

## Run

The dictionary list the actions that LAVA should execute on the DUT. The only
available key is `steps`.

`steps` is an array that will be transformed into a shell script that LAVA will
run on the DUT.

In this example, `run.steps` will be rendered into the following shell script:

```shell
--8<-- "tests/smoke.sh"
```

## Install
!!! tip "install"
    LAVA can also install some packages or sources before running the tests.
    See the [technical documentation](../../technical-references/test-definition.md)
    for more information.

--8<-- "refs.txt"

# AVH

Creates and boots Arm Virtual Hardware models.

```yaml
- boot:
    method: avh
    bootargs:
      normal: earlycon=uart8250,mmio32,0xfe215040 console=ttyS0,115200n8 rw rootwait root=/dev/mmcblk0p2 coherent_pool=1M 8250.nr_uarts=1 cma=64M
      restore: earlycon=uart8250,mmio32,0xfe215040 console=ttyS0,115200n8 console=tty0 rw rootwait root=/dev/mmcblk0p2 coherent_pool=1M 8250.nr_uarts=1 cma=64M init=/usr/lib/raspi-config/init_resize.sh
    docker:
      image: ghcr.io/vi/websocat:v1.14.1
      local: true
```

## bootargs

(optional) The `bootargs` dictionary allows you to override the default Kernel
bootargs provided by the AVH model.

The `normal` key is used for every regular boot.

The `restore` key, if present, is used for a first boot prior to the device
being declared ready. It is expected that the device will reboot itself to
indicate that this phase is complete. It is used, for example, on the Raspberry
Pi 4 to expand the root FS.

## docker

(optional) `avh` boot method uses [websocat](https://github.com/vi/websocat) for
connecting to AVH device serial console.

[ghcr.io/vi/websocat:v1.14.1](https://github.com/vi/websocat/pkgs/container/websocat/623664453?tag=v1.14.1)
docker image is used by default. The docker image specified in the job definition
overrides the default one.

!!! note
    An `avh` deploy action before this boot action is required. The boot method
    uses the image path from the deploy action to create the AVH instance.

# iPXE

The `ipxe` boot method is used to boot DUT using iPXE commands.

```yaml
- boot:
    method: ipxe
    commands: nfs
    prompts:
    - 'root@debian:~#'
```

!!! note
    A prior `tftp` deploy action is required.

## commands

See [commands](./common.md#commands).

## use_bootscript

See [use_bootscript](./method-bootloader.md#use_bootscript).

## Sample job

```yaml
device_type: x86

job_name: x86_64 iPXE sample job

timeouts:
  job:
    minutes: 30
  connection:
    minutes: 2

priority: medium
visibility: public

context:
  extra_nfsroot_args: ',vers=3'

actions:
- deploy:
    to: tftp
    kernel:
      url: https://storage.lavacloud.io/health-checks/x86/bzImage.bin
    nfsrootfs:
      url: https://storage.lavacloud.io/health-checks/x86/rootfs.tar.xz
      compression: xz
    timeout:
      minutes: 20

- boot:
    method: ipxe
    commands: nfs
    parameters:
      shutdown-message: "reboot: Restarting system"
    auto_login:
      login_prompt: 'login:'
      username: root
    prompts:
    - 'root@intel-core2-32:'
    timeout:
      minutes: 10

- test:
    definitions:
    - from: git
      repository: https://gitlab.com/lava/functional-tests.git
      path: posix/smoke-tests-basic.yaml
      name: smoke-tests
    timeout:
      minutes: 5
```

--8<-- "refs.txt"

# FVP

The `fvp` boot method allows you to run Fixed Virtual Platforms.

```yaml
- boot:
    method: fvp
    docker:
      name: "fvp_foundation:11.8"
      local: true
    image: /path/to/FVP_Binary
    ubl_license: "/root/ubl.lic"
    arguments:
      - "-C board.virtioblockdevice.image_path={DISK}"
    version_string: Fast Models [^\n]+
    console_string: 'terminal_0: Listening for serial connection on port (?P<PORT>\d+)'
    feedbacks:
    - '(?P<NAME>terminal_1): Listening for serial connection on port (?P<PORT>\d+)'
    - '(?P<NAME>terminal_2): Listening for serial connection on port (?P<PORT>\d+)'
    - '(?P<NAME>terminal_3): Listening for serial connection on port (?P<PORT>\d+)'
```

This boot method will launch the FVP binary
(already present in the docker image) specified with the `image` key with the
`arguments` as parameters.

!!! note
    The docker image must have the fastmodel in it and must have the required
    tools, such as `telnet`.

## ubl_license

(optional) Specifies a Universal Base License (UBL) for FVP models. The value
can be either:

- A file path to a license file
- A license activation code matches the pattern `XXXXX-XXXXX-XXXXX-XXXXX-XXXXX`

!!! note
    Using a license file is recommended so you don't hit the activation limit
    set on the license server by running many jobs in parallel that try to
    connect to the server for activation.

## arguments

A list of arguments passed to the FVP binary.

You can use `{IMAGE_NAME}` which will be replaced with the path to the image
with the same key under `images` in the previous `fvp` deploy stage.
`{ARTIFACT_DIR}` can also be used for the directory where all images are
deployed.

!!! note
    Previous to running an `fvp` boot, you should run an `fvp` deploy.

## version_string

A regular expression pattern defaults to `Fast Models[^\n]+` for matching the
FVP version output. The version output is saved as extra information of test
result `lava/fvp-version`.

## console_string

A required regular expression pattern that matches console output from the FVP
model. This pattern must contain a named group `(?P<PORT>\d+)` to extract the
serial port number that the FVP terminal is listening on. LAVA uses this port
number to establish a connection to the FVP console.

## feedbacks

Sometimes models offer more than one console that produces useful output. LAVA
can only write to one console at a time. Reading can be done from multiple
consoles. In some cases it's essential to read from all consoles to prevent
model from hanging. This happens when internal model buffer is not able to
accept more output because previously generated output is not consumed. FVP
boot method allows to define additional regexes to match more than one console.
This is done with the `feedbacks` keyword.

Feedbacks will be read twice during boot process (before matching login prompt)
and periodically during test-shell.

## erpc_app

You can provide `erpc_app: image_key` field to run a eRPC app. Both the server
and test apps must be deployed in advance. The `image_key` should be the name
of the test app. eRPC-related parameters for communication between the apps
should be configured at build time. Using a separate UART device is recommended
to avoid collision with other data being sent on UART. A `monitors` test action
can be defined right after the boot action to enable log parsing and wait for
run completion.

Example definition:

```yaml
actions:
- deploy:
    timeout:
      minutes: 5
    to: fvp
    images:
      APP:
        url: https://example.com/bl2.axf
      SERVER_APP:
        url: https://example.com/tfm_s_ns_signed.bin
      TEST_APP:
        url: https://example.com/erpc_main

- boot:
    timeout:
      minutes: 5
    method: fvp
    docker:
      name: localhost/fvp-std-lib:erpc
      local: true
    version_string: "Fast Models [^\\n]+"
    image: "/opt/model/FVP_ARM_Std_Library/FVP_MPS2/FVP_MPS2_AEMv8M"
    ubl_license: "/root/swskt-linaro-root-20241031-365d.lic"
    arguments:
    - "--parameter fvp_mps2.platform_type=2"
    - "--parameter cpu0.baseline=0"
    - "--parameter cpu0.INITVTOR_S=0x10000000"
    - "--parameter cpu0.semihosting-enable=0"
    - "--parameter fvp_mps2.DISABLE_GATING=0"
    - "--parameter fvp_mps2.telnetterminal0.start_telnet=1"
    - "--parameter fvp_mps2.telnetterminal1.start_telnet=0"
    - "--parameter fvp_mps2.telnetterminal2.start_telnet=0"
    - "--parameter fvp_mps2.telnetterminal0.quiet=0"
    - "--parameter fvp_mps2.telnetterminal1.quiet=1"
    - "--parameter fvp_mps2.telnetterminal2.quiet=1"
    - "--parameter fvp_mps2.telnetterminal0.start_port=5000"
    - "--parameter fvp_mps2.telnetterminal1.start_port=5001"
    - "--parameter fvp_mps2.telnetterminal1.mode=raw"
    - "--parameter fvp_mps2.UART1.unbuffered_output=1"
    - "--application cpu0={APP}"
    - "--data cpu0={SERVER_APP}@0x10080000"
    - "-M 1"
    console_string: 'telnetterminal0: Listening for serial connection on port (?P<PORT>\d+)'
    use_telnet: true
    prompts:
    - "Non-Secure system starting..."
    erpc_app: "TEST_APP"
```

## commands

(optional) A list of shell commands to execute in the FVP Docker container after
the FVP model has started. These commands are joined with `&&` and executed
sequentially using `docker exec --tty fvp_container sh -c 'joined_commands'`.

Example definition:

```yaml
- boot:
    method: fvp
    docker:
      name: fvp_rd_v3_r1:xvfb
      local: true
    image: /opt/model/FVP_RD_V3_R1/models/Linux64_GCC-9.3/FVP_RD_V3_R1
    version_string: Fast Models [^\n]+
    timeout:
      minutes: 15
    console_string: 'terminal0: Listening for serial connection on port (?P<PORT>\d+)'
    arguments:
    - -C socket0.css0.sysctrl.rse.rom.raw_image='{ROM}'
    - -C socket0.css1.sysctrl.rse.rom.raw_image='{ROM}'
    - -C socket0.css0.sysctrl.rse.intchecker.ICBC_RESET_VALUE=0x0000011B
    - -C socket0.css1.sysctrl.rse.intchecker.ICBC_RESET_VALUE=0x0000011B
    - -C socket0.board0.rse2rse_mhu0.NUM_CH=16
    - -C socket0.board1.rse2rse_mhu0.NUM_CH=16
    - -C socket0.css0.sysctrl.rse.DISABLE_GATING=true
    - -C socket0.css1.sysctrl.rse.DISABLE_GATING=true
    - -C socket0.css0.sysctrl.rse_uart.out_file=rse_uart_css0.log
    - -C socket0.css0.sysctrl.rse_uart.unbuffered_output=1
    - -C socket0.css0.sysctrl.rse_uart.uart_enable=true
    - -C socket0.css1.sysctrl.rse_uart.out_file=rse_uart_css1.log
    - -C socket0.css1.sysctrl.rse_uart.unbuffered_output=1
    - -C socket0.css1.sysctrl.rse_uart.uart_enable=true
    - -C socket0.css0.sysctrl.otpw.GP_AON_0_INIT_VAL=0x80000000
    - -C socket0.css1.sysctrl.otpw.GP_AON_0_INIT_VAL=0x80000000
    - --data socket0.css0.sysctrl.rse.cpu='{TESTS}'@0x31000400
    - -IRp
    commands:
    - export PYTHONPATH=$PYTHONPATH:/opt/model/FVP_RD_V3_R1/Iris/Python/
    - python3 <path>/test_dcsu.py --backend iris --iris_dcsu_component_string socket0.css0.sysctrl.dcsu
```

A `monitors` test action can be defined after the boot action to enable
log parsing and wait for the commands to finish.

```yaml
- test:
    timeout:
      minutes: 5
    monitors:
    - name: "rse_bl1_tests"
      start: "INFO:DCSU:Using iris backend"
      end: "TEST: Send completion PASSED"
      pattern: 'TEST: (?P<test_case_id>.+?) (?P<result>PASSED|FAILED|SKIPPED)'
      fixupdict:
         PASSED: pass
         FAILED: fail
         SKIPPED: skip
```

# Security

## ALLOWED_HOSTS

Django requires the allowed hosts to be explicitly set in the LAVA settings, as
a list of host names or IP addresses which LAVA is allowed to use.

By default, only access from the local machine is allowed.

```python
ALLOWED_HOSTS = ["[::1]", "127.0.0.1", "localhost"]
```

To enable remote access, set it to the domain or IP used to access LAVA server,
otherwise, the UI will return HTTP 500.

```yaml
ALLOWED_HOSTS: ["127.0.0.1", "localhost", "lava.example.com"]
```

## SECURE_PROXY_SSL_HEADER

LAVA server is typically deployed behind a reverse proxy (Apache by default) that
performs SSL termination and forwards plain HTTP to a local
[lava-server-gunicorn](../../technical-references/services/lava-server-gunicorn.md)
process.

When SSL is terminated at the reverse proxy, Django may receive requests over
plain HTTP internally. This creates a protocol mismatch that triggers Django’s
CSRF protection, leading to CSRF failures like:

```plain
Forbidden (403)
CSRF verification failed. Request aborted.
```

### LAVA configuration

The recommended fix is to configure LAVA to trust the header provided by your
proxy.

Add this in `/etc/lava-server/settings.d/01-secure.yaml` on your LAVA server:

```yaml
SECURE_PROXY_SSL_HEADER: ["HTTP_X_FORWARDED_PROTO", "https"]
```

!!! note
    Do not work around CSRF errors by adding your site to
    [`CSRF_TRUSTED_ORIGINS`](https://docs.djangoproject.com/en/4.2/ref/settings/#csrf-trusted-origins)
    unless you have a genuine cross-origin or subdomain setup.

!!! danger
    Make sure **ALL** the following are true before setting this:

    - Your LAVA server is strictly served via HTTPS.
    - `lava-server-gunicorn` is bound to `localhost` and is inaccessible from
    the public internet.
    - Your proxy strips the `X-Forwarded-Proto` header from all incoming requests
    and sets its own. This prevents header spoofing.

    For more information, refer to the Django
    [SECURE_PROXY_SSL_HEADER](https://docs.djangoproject.com/en/4.2/ref/settings/#secure-proxy-ssl-header)
    reference.

### Proxy configuration

Ensure your reverse proxy is configured to set the header on every forwarded
request.

#### Apache2

Enable `mod_headers`:

```shell
a2enmod headers
```

Add the directive to your virtual host:

```shell
RequestHeader set X-Forwarded-Proto "https"
```

#### Nginx

Add or verify the following directive within your `location` block. This is
often pre-configured in `/etc/nginx/proxy_params`.

```shell
proxy_set_header X-Forwarded-Proto $scheme;
```

## Cookie

LAVA defaults to secure cookie settings that suitable for HTTPS deployments:

```python
SESSION_COOKIE_SECURE = True
CSRF_COOKIE_SECURE = True
```

With these settings, browsers will only send session and CSRF cookies over
HTTPS. Accessing LAVA over plain HTTP from a non-localhost address will cause
login to fail.

If LAVA is only accessed over HTTP (e.g., a local lab not exposed to the
internet), you can disable the secure cookie requirements:

```yaml
SESSION_COOKIE_SECURE: false
CSRF_COOKIE_SECURE: false
```

!!! note "localhost exemption"
    Modern browsers treat `localhost`, `127.0.0.1`, and `::1` as secure contexts
    even over plain HTTP, so logging in via `http://localhost` may work regardless
    of these settings.

!!! danger
    Do not disable these settings on a server accessible over the internet, as
    it allows session and CSRF tokens to be transmitted in plain text.

# Install LAVA

Two installation methods are currently supported:

* [docker](#docker)
* [debian](#debian)

We advice to use docker image for a first test of LAVA.

For a production instance, both methods would be suitable.

## Docker

In order to install LAVA using the official docker images, we advice to use the
provided **docker-compose** file.

### Dependencies

Install the dependencies: `docker` and `docker-compose`:

```shell
apt-get update
apt-get install docker.io docker-compose git
```

??? tip "Old Debian version"
    On old Debian versions, docker should be installed from the official docker repository.
    ```shell
    apt-get install ca-certificates gnupg2 wget
    wget https://download.docker.com/linux/debian/gpg
    apt-key add gpg
    echo "deb [arch=amd64] https://download.docker.com/linux/debian buster stable" > /etc/apt/sources.list.d/docker.list
    apt-get update
    apt-get install docker-ce
    ```

### Install

Get the **docker-compose** files from [gitlab][lava-docker-compose] and use it.

```shell
git clone https://git.lavasoftware.org/lava/pkg/docker-compose
cd docker-compose/
docker-compose pull
docker-compose up
```

The newly created instance is now available at [localhost].

## Debian

In order to install LAVA using the Debian packages, we advice to use the
repositories that we manage to get the latest version.

!!! note "Supported Debian versions"
    LAVA is only supported on Debian *Buster* and *Bullseye*.

### Dependencies

In order to follow this tutorial, you would have to install some dependencies:

```shell
apt-get install ca-certificates gnupg2 wget
```

### Repository

Add the lavasoftware debian repository:

```shell
wget https://apt.lavasoftware.org/lavasoftware.key.asc
apt-key add lavasoftware.key.asc
echo "deb https://apt.lavasoftware.org/release buster main" > /etc/apt/sources.list.d/lava.list
```

### Install

Install **postgresql** and **lava** debian packages:

```shell
apt-get update
apt-get install postgresql
pg_ctlcluster 11 main start
apt-get install lava
```

### Starting

You can start the different services:

```shell
service apache2 start
service postgresql start
service lava-server-gunicorn start
service lava-publisher start
service lava-scheduler start
service lava-worker start
```

The newly created instance is now available at [localhost].

--8<-- "refs.txt"

# Setting up remote worker

Test execution in LAVA is performed by 'lava-worker'. It can run on the same
physical hardware as 'lava-server' but also can run separately on different
physical host. The latter case is called 'remote worker'. Remote workers can
connect to server on local network or using Internet. Connection is established
over http/https protocol.

## lava-dispatcher settings

In order to point lava-worker to the correct server, it needs the following
settings:
```shell
URL="http://<lava-server-dns>/"
```

Location of the settings depends on the way lava-dispatcher is started.

When using standalone installation either from sources or from debian package,
settings can be found in `/etc/lava-dispatcher/lava-worker` file. In case
[docker-compose](https://gitlab.com/lava/pkg/docker-compose) is used
settings should be updated in .env file:

```shell
DC_LAVA_SERVER_HOSTNAME="<lava-server-dns>"
```

## connection encryption

It is advised to always encrypt the connection between server and workers. We
advice to use https instead of http for worker connection.

## http_proxy settings

Worker specific http_proxy settings should be defined on the server in
```shell
/etc/lava-server/dispatcher.d/<name>/env.yaml
```
It can be directly edited on the server host or created/updated using API calls:
 * ```scheduler.workers.set_env``` XML-RPC
 * POST on ```/api/v0.2/workers/``` REST API
 * PUT on ```/api/v0.2/workers/<hostname>/env``` REST API

## local devices connected to remote worker

All device dictionaries are stored on the server node and passed to workers
before starting test jobs. In case both workers and server are administered by
the same admin there should be no issues.

However in case remote worker admin is different than server admin, there needs
to be coordination of device dictionaries. It is advised that device dictionary
files are maintained in version control system. This way bad changes can be
reverted quickly and full instance recovery is possible without major problems.

Current APIs allow to upload device dictionaries directly to the server. These
are ```scheduler.devices.set_dictionary``` XML-RPC method and POST request on
```/api/v0.2/devices/<hostname>/dictionary``` REST API.

Alternatively server admins might decide to use a configuration tool (salt,
ansible, etc.) to copy files from version control to the lava-server host.

### database entries for device types and devices

Although all device specific settings are stored in device dictionary, there is
still a need to create a database entry. This can be done using django admin UI,
XML-RPC API ```scheduler.devices.add``` or issuing POST request on
```/api/v0.2/devices/```.

## local device types (not yet available in LAVA)

Devices on which test jobs are executed should be supported by LAVA. This means
that "device type" should be present in the server node. Device type is used to
render full device dictionary. There may be cases that remote workers include
devices that are not yet supported by LAVA. In such case the following actions
should be performed:

 * new device type should be added to LAVA code base
 * temporarily server node admins might add new device type on their host. This
   operation is depending on how server configuration is maintained. As with
   device dictionaries it is advised to use version control.

## dispatcher version

Currently lava-dispatcher version should be in sync with lava-server version.
In case of version mismatch, lava-worker connection will be rejected by
lava-server.

### upgrades

Currently LAVA doesn't support any mechanism for remote dispatcher updates. It
is responsibility of admins to perform the updates.

# remote dispatcher - host installation

The instructions assume installation from debian packages but should also be
valid for installation from sources. These steps should result in remote worker
registering with LAVA server.

## install required packages

```shell
apt-key adv --keyserver keyserver.ubuntu.com --recv-keys A791358F2E49B100
apt-get update
apt-get install lava-dispatcher lava-dispatcher-host
```

### installing from source

lava-dispatcher-host package installs udev rules file. In order to perform this
step when installing from source one needs to run:
```shell
lava-dispatcher-host rules install
```

## edit /etc/lava-dispatcher/lava-worker and update settings

```shell
URL="http://<lava-server-dns>/"
```

## start lava-dispatcher

```shell
systemctl start lava-worker
```

# remote dispatcher - docker installation

docker-compose file to setup an instance **lava-dispatcher** from scratch. In this
setup, every service will be running in a separate container.

## Usage

###  Requirements

In order to use this docker-compose file, you need:

* docker.io
* docker-compose

You can install the dependencies using:

```shell
apt install docker.io docker-compose
```

### Installing

You just need to fetch the sources:

```shell
git clone https://gitlab.com/lava/pkg/docker-compose
cd docker-compose
```

### Using it

#### Configuration (simple, for QEMU purposes)

All configuration is stored in `.env` file. Some of the steps are required
whilst others are optional.

* Change DC_LAVA_SERVER_HOSTNAME to <server_name>
  which points to the running LAVA server instance.
* Execute `make lava-dispatcher`; at this point multiple containers should be
  up and running and the worker should connect to the LAVA server instance of
  your choosing.
* Add a new device and set its' device template (alternatively you can update
  existing device to use this new worker)
  Example QEMU device template:
  ```jinja
  {% extends 'qemu.jinja2' %}
  {% set mac_addr = 'DF:AD:BE:EF:33:02' %}
  {% set memory = 1024 %}
  ```
  You can do this via [XMLRPC](https://validation.linaro.org/api/help/#scheduler.devices.set_dictionary), [lavacli](https://docs.lavasoftware.org/lavacli/) or [REST API](https://staging.validation.linaro.org/api/v0.2/devices/staging-qemu01/dictionary/) (if using version 2020.01 and higher).
* (optional) If the lab where this container runs is behind a proxy or you
  require any specific worker environment settings, you will need to update the
  proxy settings by setting the [worker environment](https://docs.lavasoftware.org/lava/proxy.html#using-the-http-proxy)
  You can do this via this [XMLRPC API call](https://validation.linaro.org/api/help/#scheduler.workers.set_env).

!!! note "Firewall"

    If the server instance is behind a firewall, you will need to create a
    port forwarding so that ports 80 and maybe 443 are open to the public.


#### Configuration (advanced, for physical DUT purposes)

If you're setting up a standalone dispatcher container, make sure you go
through the above configuration first, it is mandatory for this step.
In order to run test jobs on physical devices we will need a couple of
additional setup steps:

* PDU control:
  * The dispatcher docker container will already download pdu scripts from
    [lava-lab repo](https://git.linaro.org/lava/lava-lab.git/) which you can use
    in device configuration but if you use custom PDU scripts you need to
    provide them and copy them into `dispatcher/power-control` directory; they
    will be copied into `/root/power-control` path in the container.
  * If you need SSH keys for PDU control, copy the private key to the
    `dispatcher/ssh` directory and the public key on to the PDU
  * SSH config - if there's a need for a specific SSH configuration (like
    tunnel passthrough, proxy, strict host checking, kexalgorithm etc), create
    the config file with relevant settings and copy it into `dispatcher/ssh`
    dir; it will be copied to `/root/.ssh` directory on the dispatcher
    container.
* ser2net config - update `ser2net/ser2net.config` with the corresponding
  serial port and device settings
* Update/add [device dictionary](https://docs.lavasoftware.org/lava/glossary.html#term-device-dictionary) with power commands and connection command
* Add dispatcher_ip setting to the [dispatcher configuration](https://validation.linaro.org/api/help/#scheduler.workers.set_config). Alternatively you can use
[REST API](https://lava_server/api/v0.2/workers/docker_dispatcher_hostname/config/) if you are using version 2020.01 or higher:
  * `dispatcher_ip: <docker host ip address>`
* Disable/stop rpcbind service on host machine if it's running - docker service
  nfs will need port 111 available on the host.


### Running

In order to start the containers, run:

    docker-compose build lava-dispatcher
    docker-compose up lava-dispatcher

or, alternatively:

    make lava-dispatcher

# Mailing-lists

The best way to get in touch with the community and the LAVA developers is via [mailing-lists](https://lists.lavasoftware.org).

## [lava-announce](mailto:lava-announce@lists.lavasoftware.org)

Subscribing to the **lava-announce** list is recommended for everyone using LAVA, whether writing tests or viewing reports or administering a LAVA instance.

The release notes for each production release are sent to the **lava-announce** mailing list.

Replies to this list are sent to the **lava-users** list.


## [lava-devel](mailto:lava-devel@lists.lavasoftware.org)

Subscribing to the **lava-devel** list is recommended for developers of LAVA.

**lava-devel** is aimed at supporting code contributors, device integration engineers and instance admins who are working with the LAVA codebase.
Discussions about planning and new LAVA features also take place here.

## [lava-users](mailto:lava-users@lists.lavasoftware.org)

The **lava-users** mailing list concentrates on support for setting up and using LAVA.

Subscribers include test writers, and individual admins.
Users are encouraged to contribute to answer queries from other users.
Replies to the **lava-announce** list are directed here.


# IRC

IRC is a common support method for developers. Our team is spread geographically around the world, with members in Europe, America and Asia.

The LAVA Software Community Project has an IRC channel, `#lavasoftware` on [irc.freenode.net](https://webchat.freenode.net/).

We can also be found on a IRC channel used for topics relating to the Linaro Lab in Cambridge, UK: `#linaro-lava` on [irc.freenode.net](https://webchat.freenode.net/).

# QEMU

LAVA provides three QEMU-based boot methods.

| Boot Method | Media   | Description |
| ----------- | ------- | ----------- |
| `qemu`      | `tmpfs` | Boots downloaded image directly using QEMU |
| `qemu-nfs`  | `nfs`   | Boots kernel with NFS-deployed root filesystem |
| `qemu-iso`  | `img`   | Boots downloaded ISO installer using QEMU |

These methods run the QEMU command line on the worker. Certain elements of the
command line are available for modification using the [job context](#job-context).

The version of QEMU installed on the worker is a choice made by the admin.
Generally, this will be the same as the version of QEMU available from Debian in
the same suite as the rest of the packages installed on the worker.
Information on the available versions of QEMU in Debian is available at
[tracker.debian.org/pkg/qemu](https://tracker.debian.org/pkg/qemu).

## Job context

The `context` section at the top level of the job definition allows
customization of the QEMU command line. The available options depend on the
device type template configuration. For example, many admins restrict the
available memory of each QEMU device, so the `memory` option in the job context
may be ignored.

```yaml
context:
  arch: aarch64
  cpu: cortex-a57
  machine: virt
  memory: 2048
  netdevice: user
  extra_options:
  - -smp
  - 1
  - -global
  - virtio-blk-device.scsi=off
  - -device virtio-scsi-device,id=scsi
  - --append "console=ttyAMA0 root=/dev/vda rw"
```

Common context options:

* `arch` (Required) - Target architecture (e.g., `amd64`, `aarch64`)
* `cpu` - CPU model (e.g., `cortex-a57`)
* `machine` - Machine type (e.g., `virt` for `aarch64`)
* `memory` - Memory allocation in MB
* `netdevice` - Network device type (e.g., `tap`, `user`)
* `extra_options` - List of additional QEMU command line options

!!! note
    The `arch` parameter in the context section is required  for all QEMU boot
    methods. It is used to determine which `qemu-system-<arch>` binary to
    execute.

## qemu

The `qemu` method is used to boot the downloaded images from the `tmpfs`
deployment action using QEMU.

```yaml
- boot:
    method: qemu
    media: tmpfs
    prompts:
    - 'root@debian:'
    auto_login:
      login_prompt: 'login:'
      username: root
```

### media

When booting QEMU image files directly, the `media` needs to be specified as
`tmpfs`.

### docker

QEMU can be run inside a Docker container. This is useful when the required
QEMU version or configuration is not available on the worker.

```yaml
- boot:
    method: qemu
    media: tmpfs
    docker:
      image: qemu-image:latest
      binary: qemu-system-aarch64
    prompts:
    - 'root@debian:'
```

The `docker` option accepts the following parameters:

* `image` (required) - the docker image to use
* `local` (optional) - whether the image is local (`docker pull` is skipped if
  `local: true` and the image exists on the local)
* `binary` (optional) - QEMU binary path inside the container

### Job example

```yaml
--8<-- "jobs/qemu.yaml"
```


## qemu-nfs

The `qemu-nfs` method is used to boot a downloaded `kernel` with a root
filesystem deployed on the worker via NFS.

```yaml
- boot:
    method: qemu-nfs
    auto_login:
      login_prompt: 'login:'
      username: 'root'
    prompts:
      - 'root@jessie:~#'
```

### media

When booting a QEMU image using NFS, the `media` is implicitly `nfs`. The
deploy action should use `to: nfs`.

!!! note
    When using `qemu-nfs`, the hostname element of the prompt may vary
    according to the worker running QEMU. Use a regex pattern like
    `'root@(.*):'` to match the prompt.

### netdevice

This boot method requires access to the NFS service running on the worker to
mount the root filesystem. Therefore, `netdevice` must be set to `tap` in
either the job context or the device dictionary.

If no network bridge configured for the tap interface yet, see
[Configure bridged network](../../../../admin/basic-tutorials/device-setup/qemu.md#configure-bridged-network).

### Job example

```yaml
--8<-- "jobs/qemu-nfs.yaml"
```

## qemu-iso

The `qemu-iso` method is used to boot a downloaded installer from the deployment
action using QEMU. This is typically used for automated OS installations.

```yaml
- boot:
    method: qemu-iso
    media: img
    auto_login:
      login_prompt: 'login:'
      username: root
      password_prompt: 'Password:'
      password: root
    prompts:
    - 'root@debian:'
```

### media

When booting an installer using QEMU, the `media` needs to be specified as
`img`. The deploy action should use `to: iso-installer`.

### netdevice

This boot method requires access to the TFTP service running on the worker for
loading the preseed file and the Internet for OS installation. Therefore,
`netdevice` must be set to `tap` in either the job context or the device
dictionary.

If no network bridge configured for the tap interface yet, see
[Configure bridged network](../../../../admin/basic-tutorials/device-setup/qemu.md#configure-bridged-network).

### Job example

```yaml
--8<-- "jobs/qemu-iso.yaml"
```

# VTS / CTS

The Android [Vendor Test Suite (VTS)](https://source.android.com/docs/core/tests/vts)
and
[Compatibility Test Suite (CTS)](https://source.android.com/docs/compatibility/cts)
are test frameworks used to validate Android device implementations.

LAVA provides sophisticated methods for deploying and booting Android images, as
well as for executing VTS/CTS test suites.

```yaml
device_type: dragonboard-845c
job_name: android-mainline-cts-test

timeouts:
  job:
    minutes: 600
  connection:
    minutes: 2
  actions:
    finalize:
      seconds: 60

priority: 50
visibility: public

context:
  test_character_delay: 10

secrets:
  SQUAD_ARCHIVE_SUBMIT_TOKEN: SQUAD_ARCHIVE_SUBMIT_TOKEN

actions:
- deploy:
    to: fastboot
    docker:
      image: linaro/lava-android-test:focal-2024.02.20-01
      local: true
    images:
      partition:0:
        url: https://example.com/gpt_both0.bin
      boot:
        url: https://example.com/boot.img
      super:
        url: https://example.com/super.img
      vendor_boot:
        url: https://example.com/vendor_boot.img
      userdata:
        url: https://example.com/userdata.img
    timeout:
      minutes: 20

- test:
    docker:
      image: linaro/lava-android-test:focal-2024.02.20-01
      local: true
    definitions:
    - from: inline
      path: format-metatdata.yaml
      name: format-metatdata
      repository:
        metadata:
          format: Lava-Test Test Definition 1.0
          name: format-metatdata
          description: format-metatdata
        run:
          steps:
          - lava-test-case "format-metadata" --shell fastboot format:ext4 metadata
    timeout:
      minutes: 5

- test:
    docker:
      image: linaro/lava-android-test:focal-2024.02.20-01
      local: true
    definitions:
    - from: inline
      path: select-display-panel.yaml
      name: select-display-panel
      repository:
        metadata:
          format: Lava-Test Test Definition 1.0
          name: select-display-panel
          description: select-display-panel
        run:
          steps:
          - lava-test-case "select-display-panel-1" --shell fastboot oem select-display-panel
            hdmi
          - lava-test-case "reboot-bootloader-1" --shell fastboot reboot bootloader
          - lava-test-case "select-display-panel-2" --shell fastboot oem select-display-panel
            hdmi
          - lava-test-case "reboot-bootloader-2" --shell fastboot reboot bootloader
          - lava-test-case "select-display-panel-3" --shell fastboot oem select-display-panel
            hdmi
          - lava-test-case "reboot" --shell fastboot reboot
    timeout:
      minutes: 5

- boot:
    method: fastboot
    docker:
      image: linaro/lava-android-test:focal-2024.02.20-01
      local: true
    prompts:
    - console:/
    - root@(.*):[/~]#
    timeout:
      minutes: 15

- test:
    timeout:
      minutes: 10
    interactive:
    - name: sleep-before-adb-available
      prompts:
      - console:/
      - root@(.*):[/~]#
      script:
      - command: echo ===========================
      - command: i=0 && while ! getprop sys.boot_completed|grep 1 && [ $i -le 30 ];
          do let i=i+1; echo sleep $i*10s for sys.boot_completed; sleep 10; done;
          echo "for prompt";
      - command: if ! getprop sys.boot_completed|grep 1 ; then logcat -d; echo "Failed
          to boot successfully"; exit 1; fi
      - command: echo ===========================
      - command: i=0 && while ! getprop init.svc.adbd|grep running && [ $i -le 15
          ]; do let i=i+1; echo sleep $i*10s for init.svc.adbd; sleep 10; done; echo
          "for prompt";
      - command: if ! getprop init.svc.adbd|grep running; then logcat -d; echo "Failed
          to have adbd running"; exit 1; fi
      - command: echo ===========================
      - command: getprop | grep adb
      - command: echo ===========================

- test:
    docker:
      image: linaro/lava-android-test:focal-2024.02.20-01
      local: true
    definitions:
    - from: inline
      path: android-boot.yaml
      name: android-boot
      repository:
        metadata:
          format: Lava-Test Test Definition 1.0
          name: android-boot
          description: android-boot
        run:
          steps:
          - lava-test-case "android-boot-wait-for-device" --shell adb wait-for-device
          - lava-test-case "android-boot-boot-completed" --shell "while ! adb shell
            getprop sys.boot_completed|grep 1; do sleep 2; done"
          - lava-test-case "android-boot-set-power-stayon" --shell adb shell su 0
            svc power stayon true
          - lava-test-case "android-boot-screencap" --shell adb shell screencap -p
            /data/local/tmp/screencap.png
          - lava-test-case "android-boot-fstab" --shell adb shell "su 0 cat /vendor/etc/fstab.*"
          - lava-test-case "android-boot-kernel-version" --shell adb shell su 0 cat
            /proc/version
          - lava-test-case "android-boot-kernel-cmdline" --shell adb shell su 0 cat
            /proc/cmdline
    timeout:
      minutes: 10

- test:
    docker:
      image: linaro/lava-android-test:focal-2024.02.20-01
      local: true
    definitions:
    - repository: https://github.com/Linaro/test-definitions.git
      from: git
      path: automated/android/noninteractive-tradefed/tradefed.yaml
      params:
        TEST_PARAMS: cts --include-filter CtsAslrMallocTestCases --include-filter
          CtsBionicTestCases --include-filter CtsBluetoothTestCases --include-filter
          CtsCameraTestCases --include-filter CtsDisplayTestCases --include-filter
          CtsDramTestCases --include-filter CtsDrmTestCases --include-filter CtsGraphicsTestCases
          --include-filter CtsHardwareTestCases --include-filter CtsJankDeviceTestCases
          --include-filter CtsJniTestCases --include-filter CtsLibcoreLegacy22TestCases
          --include-filter CtsLibcoreTestCases --include-filter CtsMonkeyTestCases
          --include-filter CtsOsTestCases --include-filter CtsSystemUiTestCases --include-filter
          CtsSystemUiRenderingTestCases --include-filter CtsUsbTests --exclude-filter
          "CtsOsTestCases android.os.cts.BuildVersionTest#testBuildFingerprint" --exclude-filter
          "CtsOsTestCases android.os.cts.SecurityFeaturesTest#testPrctlDumpable" --disable-reboot
        TEST_URL: https://example.com/android-cts.zip
        TEST_PATH: android-cts
        RESULTS_FORMAT: aggregated
        ANDROID_VERSION: aosp-main
        SQUAD_UPLOAD_URL: https://qa-reports.linaro.org/api/submit/android-lkft/mainline-gki-16k-aosp-master-db845c/6.12.0-b405a1c95660/dragonboard-845c
        INTERNET_ACCESS: 'true'
      name: cts-lkft
    timeout:
      minutes: 480
```

See [tradefed test definition](https://github.com/Linaro/test-definitions/blob/master/automated/android/noninteractive-tradefed/tradefed.yaml)
for what each test parameter does.

# UUU

The `uuu` boot method flashes and boots NXP i.MX devices using the
[UUU (Universal Update Utility)](https://github.com/NXPmicro/mfgtools). See the
[wiki](https://github.com/NXPmicro/mfgtools/wiki) for the complete documentation
of the tool.

```yaml
- boot:
    method: uuu
    commands:
    - uuu: -b sd {boot}
    timeout:
      minutes: 5
```

## Installation

LAVA supports running `uuu` directly on the worker or inside a Docker container.
See [docker](#docker) for running with docker. When running directly, `uuu`
must be installed manually on the LAVA worker since it is not provided as a
dependency by LAVA.

The latest release is available at
<https://github.com/NXPmicro/mfgtools/releases/latest>. A specific version can
be installed using the following commands:

```shell
wget https://github.com/NXPmicro/mfgtools/releases/download/<UUU_VERSION>/uuu
chmod a+x uuu
mv uuu /usr/bin/uuu
```

The `uuu` boot method supports to run [`bcu`](https://github.com/NXPmicro/bcu)
commands, see [using BCU commands](#using-bcu-commands). When needed, the tool
can be installed in the same way.

## Device configuration

Follow the instructions in [UUU](../../../configuration/device-dictionary.md#uuu)
and [BCU](../../../configuration/device-dictionary.md#bcu) to configure your
UUU device.

## commands

Required. The `commands` field is a list of commands to execute. Each entry is
a dictionary with a single `key:value` pair where the key is the protocol and the
value is the command string. Image placeholders like `{boot}` or `{system}` are
replaced with paths to images downloaded during the
[`uuu`](../deploy/to-uuu.md) deployment action.

### Using built-in scripts

UUU provides built-in scripts invoked via the `uuu` protocol with the `-b`
flag:

```yaml
- boot:
    method: uuu
    commands:
    - uuu: -b sd {boot}
```

Non-exhaustive list of available built-in scripts:

```yaml
- uuu: -b emmc {boot}              # Write bootloader to eMMC
- uuu: -b emmc_all {boot} {system} # Write bootloader & rootfs to eMMC
- uuu: -b sd {boot}                # Write bootloader to SD card
- uuu: -b sd_all {boot} {system}   # Write bootloader & rootfs to SD card
```

### Using commands

```yaml
- boot:
    method: uuu
    commands:
    - SDPS: boot -f {boot}
    - FB: continue
    - FB: done
```

Each entry is passed to the `uuu` binary as a `<Protocol>: <Command>` pair.

### Using BCU commands

Devices that support [BCU (Board Control Utility)](https://github.com/NXPmicro/bcu)
can use `bcu` commands using the commands list:

```yaml
- boot:
    method: uuu
    commands:
    - bcu: reset usb
    - uuu: -b emmc {boot}
    - bcu: set_boot_mode emmc
```

Non-exhaustive list of available bcu commands :

```yaml
- reset BOOTMODE_NAME # Reset the board and then boots from mentioned BOOTMODE_NAME.
                      # Replace BOOTMODE_NAME with different options like emmc,sd,
                      # usb which can be obtained from command bcu lsbootmode.
                      # Replace the BOOTMODE_NAME with anyone of the mentioned.
- lsftdi              # List all the boards connected by ftdi device
- lsboard             # List all supported board models
- get_boot_mode       # Displays the boot mode set by BCU
```

!!! note
    The serial availability check and bootloader corruption actions are skipped
    when the first command is `bcu: reset usb`, or when the `commands` block
    contains `bcu` commands only. This behavior is useful to recover bricked
    devices or to use `bcu` as a standalone action.

## docker

UUU can run inside a Docker container. The `docker` block specifies the
container image, which must contain the `uuu` binary.

```yaml
- boot:
    method: uuu
    docker:
      image: atline/uuu:1.5.239
    commands:
    - uuu: -b sd {boot}
```

!!! note
    A docker image specified in the job definition overrides the
    `uuu_docker_image` value from the device configuration.

### image

The Docker image name.

### local

Optional. If `true`, LAVA will use the image if it already exists locally on
the worker without pulling from a registry.

## skip_uuu_if_bootloader_valid

When set to `true`, LAVA checks whether the device already has a valid
bootloader by attempting to boot to U-Boot. If the bootloader is valid, the
entire UUU flash sequence is skipped. This is useful for jobs that only need to
flash when the bootloader is missing or corrupted.

```yaml
- boot:
    method: uuu
    skip_uuu_if_bootloader_valid: true
    commands:
    - bcu: set_boot_mode emmc_s
    timeout:
      minutes: 2
```

# Command action

The `command` action is used to execute a pre-defined command on the dispatcher
itself. A typical use case would be to set the boot mode for a board by
controlling a relay.

## Job definition

In order to use the action, the action block should only define the name of the
command to execute:

```yaml
actions:
- command:
    name: set_boot_to_nand
```

## Device dictionary

Admins should add the commands to the device dictionary:

```jinja
{% set user_commands = {'set_boot_to_qspi': {'do': 'webrelay --relay 1 on',
                                             'undo': 'webrelay --relay 1 off'},
                        'set_boot_to_nand': {'do': 'webrelay --relay 1 off',
                                             'undo': 'webrelay --relay 1 off'}} %}
```

Admins can define two commands:

* `do`: called when running the action
* `undo`: called when finalizing the job

!!! note "Failing jobs"
    When an action of a job is failing, the full job will be terminated.

    In such case, the `do` command of every subsequent actions will not be
    called. However, the job will be finalized and every `undo` commands will be
    called.

## Power commands

When defined, the power commands are automatically added to the list of available commands:

* `hard_reset`
* `pre_power_command`
* `pre_os_command`
* `power_on`
* `power_off`
* `recovery_mode`
* `recovery_exit`

!!! warning "`undo` is undefined"
    When using such commands, LAVA will not run any `undo` command.

# LDAP Authentication

LAVA server can be configured to authenticate users via Lightweight Directory
Access Protocol (LDAP). LAVA uses the
[django-auth-ldap](https://django-auth-ldap.readthedocs.io/en/latest/) backend
for LDAP authentication.

## Configuration

LDAP authentication is enabled by setting `AUTH_LDAP_SERVER_URI`. All the
following settings can be placed in `/etc/lava-server/settings.d/01-ldap.yaml`.

```yaml
AUTH_LDAP_SERVER_URI: "ldap://ldap.example.com"
AUTH_LDAP_BIND_DN: ""
AUTH_LDAP_BIND_PASSWORD: ""
AUTH_LDAP_USER_SEARCH: "LDAPSearch('ou=users,dc=example,dc=com', ldap.SCOPE_SUBTREE, '(uid=%(users)')"
AUTH_LDAP_USER_ATTR_MAP:
  first_name: "givenName"
  last_name : "sn"
  email: "mail"
```

Use the following parameter to configure a custom LDAP login page message:

```yaml
LOGIN_MESSAGE_LDAP: "If your email is first.second@example.org then use first.second as your username"
```

Restart [lava-server-gunicorn](../../technical-references/services/lava-server-gunicorn.md)
and [apache2](../../technical-references/services/apache2.md) after any changes.

## User management

To create a superuser, first log in once through the web UI with the target
LDAP `username`, then promote it to superuser:

```shell
sudo lava-server manage authorize_superuser --username <username>
```

Alternatively, create and promote an LDAP user in a single step using `addldapuser`:

```shell
sudo lava-server manage addldapuser --username <username> --superuser
```

An existing local Django user account can be converted to an LDAP user account
without losing data, provided the LDAP username does not already exist in the
LAVA instance:

```shell
sudo lava-server manage mergeldapuser --lava-user <lava_user> --ldap-user <ldap_user>
```

# Health-Check

A health-check is specific LAVA jobs that is automatically and regularly
scheduled to check the health of DUT.

If for any reason the job fails, the DUT `health` will be set to `Bad` (see
[device health](../state-machine.md#health_1)). The device will not be used
anymore by the scheduler until an admin set the health to either `unknown` or
`good`.

## Reports

Health-check reports are available at
[http://localhost/scheduler/reports](http://localhost/scheduler/reports). The
report shows health-check failures and track the general health of the devices.

There is also a table providing information on the fails checks at
[http://localhost/scheduler/reports/failures?health-checks=1](http://localhost/scheduler/reports/failures?health-checks=1).

## Recommendations

### Golden image

In order to provide constant results, we advice to only use Golden image for
health-checks.

### The infrastructure

As health-checks are normal LAVA jobs, an health-check can test any part of the
infrastructure that is normally used by a LAVA job.

For instance, we recommend to test:

* bootloader methods
* attached devices (hard-drive, probes, ...)
* network access (local or remote servers)

!!! info "lava-test-raise"
    In order to fail a job during the test, call `lava-test-raise <message>`.
    This will return immediately and set the job health to `Incomplete`.

## Configuration file

The health-checks are stored on the server in
`/etc/lava-server/dispatcher-config/health-checks/<name>.yaml`

By default, LAVA creates device-types with a routine health-check frequency of
24 hours. Admins can change this default for an instance by setting
`HEALTH_FREQUENCY_HOURS` in one of the [LAVA settings files](../../admin/basic-tutorials/instance/configure.md#configuration-files), for example:

```yaml
HEALTH_FREQUENCY_HOURS: 168
```

Admin could update device-type health-check using [lavacli]:

```shell
lavacli device-types health-check set qemu qemu.jinja2
```

!!! info "filename"
    In order to compute the health-check of a DUT, LAVA will look in the device
    dictionary for the `{% extends device-type.jinja2 %}` line. The
    health-check filename is `device-type.yaml`.

--8<-- "refs.txt"

# Health-Check

A health-check is specific LAVA jobs that is automatically and regularly
scheduled to check the health of DUT.

If for any reason the job fails, the DUT `health` will be set to `Bad` (see
[device health](../state-machine.md#health_1)). The device will not be used
anymore by the scheduler until an admin set the health to either `unknown` or
`good`.

## Reports

Health-check reports are available at
[http://localhost/scheduler/reports](http://localhost/scheduler/reports). The
report shows health-check failures and track the general health of the devices.

There is also a table providing information on the fails checks at
[http://localhost/scheduler/reports/failures?health-checks=1](http://localhost/scheduler/reports/failures?health-checks=1).

## Recommendations

### Golden image

In order to provide constant results, we advice to only use Golden image for
health-checks.

### The infrastructure

As health-checks are normal LAVA jobs, an health-check can test any part of the
infrastructure that is normally used by a LAVA job.

For instance, we recommend to test:

* bootloader methods
* attached devices (hard-drive, probes, ...)
* network access (local or remote servers)

!!! info "lava-test-raise"
    In order to fail a job during the test, call `lava-test-raise <message>`.
    This will return immediately and set the job health to `Incomplete`.

## Configuration file

The health-checks are stored on the server in
`/etc/lava-server/dispatcher-config/health-checks/<name>.yaml`

Admin could update device-type health-check using [lavacli]:

```shell
lavacli device-types health-check set qemu qemu.jinja2
```

!!! info "filename"
    In order to compute the health-check of a DUT, LAVA will look in the device
    dictionary for the `{% extends device-type.jinja2 %}` line. The
    health-check filename is `device-type.yaml`.

--8<-- "refs.txt"

# Building LAVA

LAVA can be distributed either as [Debian packages](#debian-packages) or [Docker images](#docker-images).

Both formats can be built locally from sources.

!!! tip "Version"
    The version will be computed based on `git describe`.
    Run `./lava_common/version.py` to print this version.

## Debian packages

Run the GitLab CI script to build the Debian packages:

```shell
.gitlab-ci/build/debian/10.sh
```

The packages will be available under `_build/`.

!!! warning "Uncommitted changes"
    The script will refuse to build the packages is you have any uncommitted changes.

## Docker images

Run the GitLab CI script to build either the `lava-dispatcher` or `lava-server` image:

```shell
.gitlab-ci/build/docker.sh dispatcher
.gitlab-ci/build/docker.sh server
```

Images will be tagged with:

```
hub.lavasoftware.org/lava/lava/amd64/lava-dispatcher:<VERSION>
hub.lavasoftware.org/lava/lava/amd64/lava-server:<VERSION>
```

!!! tip "Base images"
    The script will create base images for both `lava-dispatcher` and `lava-server`.
    Bases images will be tagged with
    ```
    hub.lavasoftware.org/lava/lava/amd64/lava-dispatcher-base:<VERSION>
    hub.lavasoftware.org/lava/lava/amd64/lava-server-base:<VERSION>
    ```

# Architecture

```mermaid
graph TD
    subgraph server ["Server"]
        web([Web UI & API])
        scheduler(["Scheduler"])
        publisher(["Publisher"])
        coordinator([Coordinator])
        db[("Database")]

        web <--> db
        %% db --> publisher breaks the layout; a labeled link is used instead.
        publisher-- Query ---db
        scheduler <--> db
        publisher <--ZMQ--> scheduler
        web --ZMQ--> publisher
    end

    subgraph lab [Lab]
        worker([Worker])
        dut1([DUT<br>Linux])
        dut2([DUT<br>IoT])
        dut3([DUT<br>AOSP])

        worker <--> dut1
        worker <--> dut2
        worker <--> dut3
    end

    publisher --WS--> worker
    web <--HTTP--> worker
    coordinator <--TCP--> worker
```

LAVA consists of several components with different roles, organized in a
server-worker model.

The simplest possible configuration is to run the server and worker components
on a single machine. Larger instances can be configured with multiple workers
controlling a greater number of attached test devices.

## Server

### Web interface

The LAVA web interface is built using the [apache2](./services/apache2.md) web
server, the [gunicorn](./services/lava-server-gunicorn.md) application server
and the Django web framework. It also provides XML-RPC access and the REST API.

!!! danger
    In production, do not serve the web application or the [publisher](#publisher)
    over plain HTTP. Put them behind a reverse proxy that terminates TLS.

### Publisher

The [lava-publisher](./services/lava-publisher.md) receives events from LAVA
services and forwards them to subscribers.

### Scheduler

The [lava-scheduler](./services/lava-scheduler.md) periodically schedules jobs
in the queue to idle devices.

### Coordinator

The [lava-coordinator](./services/lava-coordinator.md) daemon receives, stores,
and dispatches messages for
[MultiNode](../user/advanced-tutorials/multinode.md) jobs.


## Worker

The [lava-worker](./services/lava-worker.md) daemon pulls scheduled jobs for
this worker from the server over HTTP. For each job it starts a `lava-run`
process to manage all the operations on the DUT. It also keeps a WebSocket
connection to the [publisher](#publisher) for real-time events, reducing
latency when jobs are assigned or updated.

## DUT

Physical or virtual devices under test. Multiple DUTs can be attached to a
single [worker](#worker).

# Bootloader

The `bootloader` boot method is used to power-on the DUT, interrupt the
bootloader, and wait for the bootloader prompt.

```yaml
- boot:
    method: bootloader
    bootloader: u-boot
    commands: []
```

## bootloader

In order to interrupt the bootloader, the bootloader type should be specified
in the `bootloader` parameter.

!!! note
    The bootloader method type should match a boot method supported by the
    given device type. For example `fastboot`, `minimal`, `pyocd`, `u-boot`, etc.

## commands

The `commands` parameter is required but can be kept empty. If some commands
should be sent to the bootloader before the end of the action, provide them as
a list in the `commands` parameter.

See also [commands](./common.md#commands)

## use_bootscript

When set to `true`, the boot commands are written to an iPXE script (`script.ipxe`)
stored in the TFTP directory. The bootloader then executes a chainload command
to run this script (e.g., `dhcp net0; chain tftp://<worker_ip>/<path>/script.ipxe`).

```yaml
- boot:
    method: ipxe
    commands: ramdisk
    use_bootscript: true
```

!!! note
    A prior `tftp` deploy action is required.

## reset_connection

By default, LAVA will reset the previous connection when executing this action.
You can skip this step by setting `reset_connection: false`.

```yaml
- boot:
    method: bootloader
    bootloader: u-boot
    reset_connection: false
```

## reset_device

By default, LAVA will reset the board power when executing this action. You can
skip this step by setting `reset_device: false`.

```yaml
- boot:
    method: bootloader
    bootloader: u-boot
    reset_device: false
```

# Health-Check

A health-check is a specific LAVA job that is automatically and regularly
scheduled to check the health of DUT.

If for any reason the job fails, the DUT `health` will be set to `Bad` (see
[device health](../state-machine.md#health_1)). The device will not be used
anymore by the scheduler until an admin set the health to either `unknown` or
`good`.

## Reports

Health-check reports are available at
[http://localhost/scheduler/reports](http://localhost/scheduler/reports). The
report shows health-check failures and track the general health of the devices.

There is also a table providing information on the fails checks at
[http://localhost/scheduler/reports/failures?health-checks=1](http://localhost/scheduler/reports/failures?health-checks=1).

## Recommendations

### Golden image

In order to provide constant results, we advice to only use Golden image for
health-checks.

### The infrastructure

As health-checks are normal LAVA jobs, an health-check can test any part of the
infrastructure that is normally used by a LAVA job.

For instance, we recommend to test:

* bootloader methods
* attached devices (hard-drive, probes, ...)
* network access (local or remote servers)

!!! info "lava-test-raise"
    In order to fail a job during the test, call [`lava-test-raise <message>`](../../user/basic-tutorials/test-definition.md#lava-test-raise).
    This will return immediately and set the job health to `Incomplete`.

## Configuration file

The health-checks are stored on the server in
`/etc/lava-server/dispatcher-config/health-checks/<name>.yaml`

By default, LAVA creates device-types with a routine health-check frequency of
24 hours. Admins can change this default for an instance by setting
`HEALTH_FREQUENCY_HOURS` in one of the [LAVA settings files](../../admin/basic-tutorials/instance/configure.md#configuration-files), for example:

```yaml
HEALTH_FREQUENCY_HOURS: 168
```

Admin could update device-type health-check using [lavacli]:

```shell
lavacli device-types health-check set qemu qemu.jinja2
```

!!! info "filename"
    In order to compute the health-check of a DUT, LAVA will look in the device
    dictionary for the `{% extends device-type.jinja2 %}` line. The
    health-check filename is `device-type.yaml`.

## Secrets

A health-check may need credentials, for instance to authenticate against a
service or to download a private artifact.

As for any other job, a health-check can use [secrets](../job-definition/job.md#secrets)
and refer to a remote artifact token by name, so the token value never appears
in the health-check definition. LAVA resolves the token names against the
remote artifact tokens of the **job submitter**. Health-checks are submitted
automatically by LAVA using the `lava-health` service account, so the remote
artifact tokens are resolved from this user.

The `lava-health` account cannot log in, so its tokens cannot be modified
through the usual profile page. Admins should add them from the Django admin
interface: edit the `lava-health` user and add the tokens in the
**Remote artifacts auths** form, filling the `Name` and `Token` fields.

!!! warning "REST API"
    The `/api/v0.2/remote-artifact-tokens/` endpoint always operates on the
    authenticated user. Since the `lava-health` user cannot log in, this
    endpoint cannot be used to manage its tokens; use the Django admin
    interface instead.

For example, add a token named `example-token-reference` to the `lava-health`
user, then refer to it from the health-check definition:

```yaml
job_name: example health check

secrets:
  API_TOKEN: example-token-reference

actions:
- deploy:
    # ...

- boot:
    # ...

- test:
    definitions:
    - from: inline
      name: example
      path: inline/example.yaml
      repository:
        metadata:
          format: Lava-Test Test Definition 1.0
          name: example
        run:
          steps:
          - . /lava-*/secrets
          - curl -H "Authorization: Bearer ${API_TOKEN}" https://example.com/check
```

LAVA replaces `example-token-reference` with the real token value when the job
runs; the definition shown to users keeps the token name.

### Downloading a private artifact

The `secrets` block is only exported to the test shell, so it cannot be used to
authenticate a download. Use an
[authenticated download](../job-definition/actions/deploy/index.md#authenticated-downloads)
instead: token names used in the artifact `headers` are resolved from the
`lava-health` user, just like the ones in `secrets`.

--8<-- "refs.txt"

# User notifications in LAVA

Users can receive notifications containing information about submitted test
jobs and the accompanying results by using settings in job definition yaml.
Users can also compare the job results within notifications against previous
results which satisfy certain conditions.

The basic setup of the notify block in job definitions will have
**criteria**, **verbosity**, **recipients** and **compare** blocks.

+ **Criteria** tells the system when the notifications should be sent.

+ **verbosity** will tell the system how detailed the email notification should
  be.

+ **Criteria** can be set by job state (running, complete, incomplete,
  canceled) and the type (regression, progression). There is a special status of
  'finished', which will match any of (complete, incomplete, canceled). You
  could also use "all" to receive a notification for both "running" and
  "finished" states.

The basic **email** template includes job details, device details and results,
based on the level of **verbosity** (verbose, quiet, status-only).
verbose level job information will only be included if the job finished as
complete or incomplete, not when the job was canceled.

## Notification recipients

The method and destination of each notification can be set for each recipient.
Currently, two notification methods are supported, **email** and **IRC**.

Recipients can be specified using the LAVA username or in full.

If the **recipients** section is omitted in the notify block, the system will
send an email to the job submitter only, provided the **criteria** is satisfied,
and there is no **callbacks** section.

## Notification callbacks

In addition to sending email and IRC messages to **recipients**, the system can
send multiple URL callback actions. This will do a GET or POST request to the
specified URL in the **callbacks** subsection. This can be used to trigger some
action remotely. If a callback uses the POST request, the system will attach job data as described below.
The **callbacks** section supports list of the following options:

+ **url** The URL for the request. This also supports field value
  substitution, i.e. in http://example.com/{ID}/{STATUS} **id** and
  **status** will be replaced with corresponding values from the job.

+ **method** GET or POST

+ **token** This option is used to supply the API token of the
  authenticated user, appended as the POST request parameter. If the submitting
  user has an XMLRPC auth token with a description that matches this field, that
  token is returned instead. The token is included in the POST data, and also in
  an Authorization header.

+ **dataset** This option specifies style of data that the system
  will provide in the callback. It applies only for the POST request. The format
  of the data and possible options are as following:

    + **minimal** This will provide basic job info such as job id, status,
    submit_time, start_time, end_time, submitter_username, submitter_email,
    failure_comment, priority, description, actual_device_id, definition,
    health_check and metadata.
    + **logs** In addition to minimal data this will also attach the job log
    output in the url encoded format
    + **results** In addition to minimal data this will also attach the job
    results as a list of test suites exported in yaml format.
    + **all** In addition to minimal data this will include both the **logs**
    and **results** datasets as described above.

+ **content-type** This option is used to determine how the POST data is submitted:

    + **urlencoded** (Default) Will return a standard HTTP POST request, with an
    application/x-www-form-urlencoded Content-Type header and data sent as an
    urlencoded query string.
    + **json** The data is dumped into JSON and returned with a ``Content-Type:
    application/json`` header.

Example callback usage:

```yaml
notify:
  criteria:
    status: finished
  callbacks:
    - url: https://my-url/callback
      method: GET
    - url: https://remote-site/callback/lava?status={STATUS}&status_string={STATUS_STRING}
      method: POST
      dataset: all
      token: ci-callback
      content-type: json

```

### Restricting callback URLs

By default, notification callbacks can target any URL. Administrators can
restrict which hosts are permitted by setting `CALLBACK_ALLOWED_HOSTS` in
`/etc/lava-server/settings.yaml` or a file in `/etc/lava-server/settings.d/`:

```yaml
CALLBACK_ALLOWED_HOSTS:
  - "example.com"
  - "*.example.com"
  - "*.ci.example.com"
```

When this setting is defined, only callback URLs whose hostname matches one of
the patterns will be accepted. All other callback URLs will be rejected and a
warning logged. Shell-style wildcards are supported:

- `example.com` — matches `example.com` only
- `*.example.com` — matches `ci.example.com`, `api.example.com` but not
  the bare `example.com`
- `*.ci.example.com` — matches `build.ci.example.com`,
  `test.ci.example.com`

When `CALLBACK_ALLOWED_HOSTS` is not set, all callback URLs are permitted.

!!! warning "Security recommendation"

    It is recommended to configure `CALLBACK_ALLOWED_HOSTS` in production
    deployments to prevent the LAVA server from being used to reach internal
    network services via crafted callback URLs.

### Debugging notification callbacks

The job data can also be retrieved using the REST API (which supports
authentication). For example::

```
wget -O job_data.gz http://localhost/scheduler/job/2126/job_data
```

```
wget -O job_data.gz "http://localhost/scheduler/job/2127/job_data?user=joe&token=22yj3ls...."
```

Returns a gzip file containing the job data as JSON.

!!! info "callback data retrieval"

    Only test jobs which are configured to use the notification callback
    will create notification callback data for later retrieval. Other jobs will
    generate a 404 error.

### Using profile settings

LAVA users can configure the ``IRC settings`` or email address in their own
profile data in the instance. This allows the recipient to be specified using
only the LAVA username.

![profile menu](profile-menu.png)

### Direct listing of recipients

If the user has not configured their own profile data, the recipient details
must be specified in full.

Examples for user vs manual addressing:

```yaml
notify:
  recipients:
  - to:
     method: irc
     user: neil.williams
  - to:
     method: irc
     handle: codehelp
     server: irc.freenode.net
  - to:
     method: email
     user: neil.williams
  - to:
     method: email
     email: neil.williams@linaro.org
```

IRC and email notifications use different templates since emails allow for more
verbosity, so some options which are present in the ``notify`` block for email
recipients do not apply for IRC recipients, like **comparing** options and
**verbosity**.


## Result comparison in notifications

Users can use notifications to compare the job status against the last few jobs
and also compare job results, like test case/suites difference. The
notification will report if there are new test cases compared to previous test
jobs, if there are some missing and how the said results differ.

Which test jobs the current one is compared to is determined by the ``queries``
system, where those previous jobs will be taken from the query specified in the
notification block. ``Custom queries`` can also be specified.

**Blacklisting** test cases is also an option. Ignoring some of the test cases
in the comparing section of the notification is possible through an option in
the notification block.

Here are some comparing setup examples from test definition excerpts:

```yaml
notify:
  recipients:
  - to:
     method: email
     user: neil.williams
  criteria:
    status: finished
  verbosity: verbose
  compare:
    query:
      name: query_name
      username: query_owner
    blacklist: [lava, singlenode-advanced]
```
```yaml
notify:
  recipients:
  - to:
     method: email
     user: neil.williams
  criteria:
    status: finished
  verbosity: verbose
  compare:
    query:
      entity: testjob
      conditions:
        submitter: stevan.radakovic
        requested_device_type: qemu
    blacklist: [singlenode-advanced]
```

# Configure LAVA

## Creating a superuser

Admins can create superuser accounts from the command line:

=== "Docker"

    ```shell
    docker-compose exec lava-server lava-server manage createsuperuser
    ```

=== "Debian"
    ```shell
    lava-server manage createsuperuser
    ```

## Settings

### Configuration files

LAVA settings are stored in **yaml** in:

* `/etc/lava-server/settings.conf`
* `/etc/lava-server/settings.yaml`
* `/etc/lava-server/settings.d/*.yaml`

LAVA will load the files in this exact order. Files in the `settings.d`
directory will be alphabetically ordered.

!!! tip "Merging files"
    If a variable is defined in two files, the value from the last file will
    override the value from the first one.

In a fresh installation, LAVA installer will automatically create some setting files
for you:

* `/etc/lava-server/settings.d/00-secret-key.yaml`: django [SECRET_KEY]
* `/etc/lava-server/settings.d/00-database.yaml`: django [DATABASES]

You can then create some file to customize your instance.
The list of available values is listed in:

* [Django settings](https://docs.djangoproject.com/en/3.2/ref/settings/)
* [LAVA settings](https://gitlab.com/lava/lava/-/blob/master/lava_server/settings/common.py)

!!! warning "Legacy configuration"
    In previous LAVA versions, the settings where saved in three files:

    * `/etc/lava-server/settings.conf` (json)
    * `/etc/lava-server/secret_key.conf` (json)
    * `/etc/lava-server/instance.conf` (ini)

    LAVA will still load this files prior to load the new setting files.

!!! tip "Applying changes"
    When updating the settings, you should restart **every** LAVA services.


### Environment variables

Admins can also pass settings via environment variables. This is mainly used in
cloud setup like docker-compose or k8s.

The following variables are supported:

* `ALLOWED_HOSTS`
* `DATABASE_URL` (see django-url documentation)
* `INSTANCE_NAME`
* `SECRET_KEY`
* `LAVA_SETTINGS_*` (without the `LAVA_SETTINGS_` prefix).

In order to add complex types, admins can use `LAVA_JSON_SETTINGS`. This
variable should contain a dictionary of settings as a json string base64
encoded.

```python
import base64
import json

data = {
  "WORKER_AUTO_REGISTER_NETMASK": ["::1"]
}
print(base64.b64encode(json.dumps(data).encode("utf-8")).decode("utf-8"))
```

In order to use it, add to the environment: `LAVA_JSON_SETTINGS="eyJXT1JLRVJfQVVUT19SRUdJU1RFUl9ORVRNQVNLIjogWyI6OjEiXX0="`.

## Worker configuration

### Devices

The device dictionaries are stored into `/etc/lava-server/dispatcher-config/devices/`.
The files should belong to `lavaserver` user and group.

More information in the [technical reference](../../../technical-references/configuration/device-dictionary.md).

### Device-types

The device-type templates are stored into `/etc/lava-server/dispatcher-config/device-types/`
The files should belong to `lavaserver` user and group.

For supported device-types, the templates provided by LAVA are stored in
`/usr/share/lava-server/device-types`. LAVA looks for the template first in
`/etc/lava-server/dispatcher-config/device-types/` and then fallback to
`/usr/share/lava-server/device-types`.

More information in the [technical reference](../../../technical-references/configuration/device-type-template.md).

### Health-checks

The health-checks are stored into `/etc/lava-server/dispatcher-config/health-checks/`
The files should belong to `lavaserver` user and group.

More information in the [technical reference](../../../technical-references/configuration/health-check.md).

### Dispatcher

See [dispatcher configuration](../../../technical-references/configuration/dispatcher.md).

--8<-- "refs.txt"

# Boot Testing

LAVA supports a large number of
[deploy methods](../../technical-references/job-definition/actions/deploy/index.md)
and
[boot methods](../../technical-references/job-definition/actions/boot/common.md),
making it easy to validate that a firmware, kernel, or root filesystem boots
correctly across many different hardware platforms.

## Boot test

A typical boot test deploys an image, boots the device, waits for a login prompt:

```yaml
job_name: Boot test
device_type: qemu

timeouts:
  job:
    minutes: 20

priority: medium
visibility: public

context:
  arch: amd64

actions:
- deploy:
    to: tmpfs
    images:
      rootfs:
        image_arg: -drive format=qcow2,file={rootfs}
        url: https://storage.lavacloud.io/health-checks/qemu/tmpfs/debian-buster.qcow2.zst
        compression: zstd
    timeout:
      minutes: 20

- boot:
    method: qemu
    media: tmpfs
    prompts:
    - 'root@debian:~#'
    auto_login:
      login_prompt: "login:"
      username: root
    timeout:
      minutes: 5
```

## Boot duration

LAVA records how long the `auto_login` action takes using the `lava/login-action`
test case in job results, allowing teams to track boot duration over time. This
makes it easy to detect regressions in boot performance across kernel or
firmware changes.

For more precise measurement, if supported by the booted OS, `systemd-analyze`
can be used:

```yaml
- test:
    interactive:
    - name: systemd-analyze
      prompts:
      - 'root@(.*):'
      script:
      - command: 'systemd-analyze'
      - command: 'systemd-analyze blame --no-pager | head -20'
    timeout:
      minutes: 5
```

# Trusted by

LAVA is the backbone of hardware testing for some of the world's most important
open-source projects. The organizations below rely on LAVA to validate their
code on real physical hardware or emulated platforms (QEMU, FVP).

<style>
.grid > div:hover img { opacity: 0.3; }
.grid > div:hover > div[style*="position: absolute"] { opacity: 1 !important; }
.grid > div > div[style*="position: absolute"] {
  top: auto !important;
  bottom: 0.5rem !important;
  transform: translate(-50%, 0) !important;
}
</style>

## KernelCI — Linux kernel validation

KernelCI uses LAVA as its **primary backend** for automated Linux kernel
hardware testing. Every kernel patch that lands in mainline has likely been
validated by LAVA running on dozens of boards across the KernelCI labs network.

> As KernelCI continues to grow and mature, the complexity of managing
> hardware testing laboratories has become increasingly apparent. While we
> currently support LAVA as our primary backend, the community has expressed
> strong interest in expanding our capabilities.
>
> — [KernelCI Labs Working Group announcement](https://kernelci.org)

KernelCI's Labs Working Group was created specifically to strengthen the
LAVA integration and scale hardware testing across the Linux community.

## 10x Engineers — RISC-V CI

10x Engineers uses LAVA through their Cloud-V platform for automated firmware
and kernel testing on RISC-V hardware. Their LAVA lab supports testing on
boards like StarFive VisionFive 2, SiFive HiFive Unleashed, Banana Pi F3,
Milk-V Jupiter, and Milk-V Pioneer Box.

## Android — pre-merge validation

LAVA tests proposed Android changes in Gerrit before they are landed, running
boot tests, VTS/CTS suites, and hardware compatibility checks on physical
Android development boards.

## Apertis — Debian-based IoT platform testing

Apertis uses LAVA for integration testing and package testing of their
Debian-based IoT platform. They have a dedicated LAVA test shell distro
profile, use Robot Framework integrated on LAVA, and run complete test
automation on LAVA infrastructure for all their reference hardware.

Apertis developers can also submit personal LAVA jobs during development
to debug tests or verify changes before final integration.

## GCC — compiler validation

LAVA tests GCC compiler output on real hardware, validating that code
produced by GCC is correct and performs as expected across a wide range of
embedded and server processors.

## Linux kernel — daily validation

LAVA tests the Linux kernel on a range of supported boards every day. It
validates bootloader changes, kernel boot sequences, and system-level
functionality across multiple architectures.

## Qualcomm — Debian and Yocto image validation

Qualcomm uses LAVA to automatically test their Linux images on physical
Qualcomm development boards. Their CI pipeline integrates LAVA via GitHub
Actions, submitting test jobs, monitoring execution, and publishing results
back to their repositories.

---

## Supported by

LAVA is built and maintained by a global community of engineers from leading
technology companies. The organizations below are the testimony of LAVA's
success.

## Core contributors

These organizations account for the vast majority of LAVA's codebase:

<div class="grid" style="grid-template-columns: repeat(2, 1fr); gap: 1.5rem; margin: 2rem 0;">

<div style="text-align: center; padding: 1.5rem; border: 1px solid var(--md-default-fg-color--lightest); border-radius: 8px; position: relative;">
<img src="../assets/images/logos/linaro.svg" alt="Linaro" width="160" style="cursor: pointer;">
<div style="position: absolute; top: 50%; left: 50%; transform: translate(-50%, -50%); opacity: 0; transition: opacity 0.2s; font-weight: 500; font-size: 0.9rem; pointer-events: none;">Linaro</div>
</div>

<div style="text-align: center; padding: 1.5rem; border: 1px solid var(--md-default-fg-color--lightest); border-radius: 8px; position: relative;">
<img src="../assets/images/logos/collabora.svg" alt="Collabora" width="128" style="cursor: pointer;">
<div style="position: absolute; top: 50%; left: 50%; transform: translate(-50%, -50%); opacity: 0; transition: opacity 0.2s; font-weight: 500; font-size: 0.9rem; pointer-events: none;">Collabora</div>
</div>

<div style="text-align: center; padding: 1.5rem; border: 1px solid var(--md-default-fg-color--lightest); border-radius: 8px; position: relative;">
<img src="../assets/images/logos/debian.svg" alt="Debian" width="64" style="cursor: pointer;">
<div style="position: absolute; top: 50%; left: 50%; transform: translate(-50%, -50%); opacity: 0; transition: opacity 0.2s; font-weight: 500; font-size: 0.9rem; pointer-events: none;">Debian</div>
</div>

<div style="text-align: center; padding: 1.5rem; border: 1px solid var(--md-default-fg-color--lightest); border-radius: 8px; position: relative;">
<img src="../assets/images/logos/canonical.svg" alt="Canonical" width="128" style="cursor: pointer;">
<div style="position: absolute; top: 50%; left: 50%; transform: translate(-50%, -50%); opacity: 0; transition: opacity 0.2s; font-weight: 500; font-size: 0.9rem; pointer-events: none;">Canonical</div>
</div>

</div>

## Hardware partners

Major semiconductor and hardware companies:

<div class="grid" style="grid-template-columns: repeat(3, 1fr); gap: 1.5rem; margin: 2rem 0;">

<div style="text-align: center; padding: 1.5rem; border: 1px solid var(--md-default-fg-color--lightest); border-radius: 8px; position: relative;">
<img src="../assets/images/logos/nxp.svg" alt="NXP" width="64" style="cursor: pointer;">
<div style="position: absolute; top: 50%; left: 50%; transform: translate(-50%, -50%); opacity: 0; transition: opacity 0.2s; font-weight: 500; font-size: 0.9rem; pointer-events: none;">NXP</div>
</div>

<div style="text-align: center; padding: 1.5rem; border: 1px solid var(--md-default-fg-color--lightest); border-radius: 8px; position: relative;">
<img src="../assets/images/logos/baylibre.svg" alt="BayLibre" width="128" style="cursor: pointer;">
<div style="position: absolute; top: 50%; left: 50%; transform: translate(-50%, -50%); opacity: 0; transition: opacity 0.2s; font-weight: 500; font-size: 0.9rem; pointer-events: none;">BayLibre</div>
</div>

<div style="text-align: center; padding: 1.5rem; border: 1px solid var(--md-default-fg-color--lightest); border-radius: 8px; position: relative;">
<img src="../assets/images/logos/arm.svg" alt="ARM" width="64" style="cursor: pointer;">
<div style="position: absolute; top: 50%; left: 50%; transform: translate(-50%, -50%); opacity: 0; transition: opacity 0.2s; font-weight: 500; font-size: 0.9rem; pointer-events: none;">ARM</div>
</div>

<div style="text-align: center; padding: 1.5rem; border: 1px solid var(--md-default-fg-color--lightest); border-radius: 8px; position: relative;">
<img src="../assets/images/logos/qualcomm.svg" alt="Qualcomm" width="128" style="cursor: pointer;">
<div style="position: absolute; top: 50%; left: 50%; transform: translate(-50%, -50%); opacity: 0; transition: opacity 0.2s; font-weight: 500; font-size: 0.9rem; pointer-events: none;">Qualcomm</div>
</div>

<div style="text-align: center; padding: 1.5rem; border: 1px solid var(--md-default-fg-color--lightest); border-radius: 8px; position: relative;">
<img src="../assets/images/logos/pengutronix.svg" alt="Pengutronix" width="128" style="cursor: pointer;">
<div style="position: absolute; top: 50%; left: 50%; transform: translate(-50%, -50%); opacity: 0; transition: opacity 0.2s; font-weight: 500; font-size: 0.9rem; pointer-events: none;">Pengutronix</div>
</div>

<div style="text-align: center; padding: 1.5rem; border: 1px solid var(--md-default-fg-color--lightest); border-radius: 8px; position: relative;">
<img src="../assets/images/logos/renesas.svg" alt="Renesas" width="128" style="cursor: pointer;">
<div style="position: absolute; top: 50%; left: 50%; transform: translate(-50%, -50%); opacity: 0; transition: opacity 0.2s; font-weight: 500; font-size: 0.9rem; pointer-events: none;">Renesas</div>
</div>

<div style="text-align: center; padding: 1.5rem; border: 1px solid var(--md-default-fg-color--lightest); border-radius: 8px; position: relative;">
<img src="../assets/images/logos/st.svg" alt="STMicroelectronics" width="128" style="cursor: pointer;">
<div style="position: absolute; top: 50%; left: 50%; transform: translate(-50%, -50%); opacity: 0; transition: opacity 0.2s; font-weight: 500; font-size: 0.9rem; pointer-events: none;">STMicroelectronics</div>
</div>

<div style="text-align: center; padding: 1.5rem; border: 1px solid var(--md-default-fg-color--lightest); border-radius: 8px; position: relative;">
<img src="../assets/images/logos/ti.svg" alt="Texas Instruments" width="160" style="cursor: pointer;">
<div style="position: absolute; top: 50%; left: 50%; transform: translate(-50%, -50%); opacity: 0; transition: opacity 0.2s; font-weight: 500; font-size: 0.9rem; pointer-events: none;">Texas Instruments</div>
</div>

</div>

## Additional contributors and users

<div class="grid" style="grid-template-columns: repeat(3, 1fr); gap: 1rem; margin: 2rem 0;">

<div style="text-align: center; padding: 1rem; border: 1px solid var(--md-default-fg-color--lightest); border-radius: 8px; position: relative;">
<img src="../assets/images/logos/ubuntu.svg" alt="Ubuntu" width="48" style="cursor: pointer;">
<div style="position: absolute; top: 50%; left: 50%; transform: translate(-50%, -50%); opacity: 0; transition: opacity 0.2s; font-weight: 500; font-size: 0.9rem; pointer-events: none;">Ubuntu</div>
</div>

<div style="text-align: center; padding: 1rem; border: 1px solid var(--md-default-fg-color--lightest); border-radius: 8px; position: relative;">
<img src="../assets/images/logos/foundriesio.svg" alt="Foundries.io" width="160" style="cursor: pointer;">
<div style="position: absolute; top: 50%; left: 50%; transform: translate(-50%, -50%); opacity: 0; transition: opacity 0.2s; font-weight: 500; font-size: 0.9rem; pointer-events: none;">Foundries.io</div>
</div>

<div style="text-align: center; padding: 1rem; border: 1px solid var(--md-default-fg-color--lightest); border-radius: 8px; position: relative;">
<img src="../assets/images/logos/amd.svg" alt="AMD" width="64" style="cursor: pointer;">
<div style="position: absolute; top: 50%; left: 50%; transform: translate(-50%, -50%); opacity: 0; transition: opacity 0.2s; font-weight: 500; font-size: 0.9rem; pointer-events: none;">AMD</div>
</div>

<div style="text-align: center; padding: 1rem; border: 1px solid var(--md-default-fg-color--lightest); border-radius: 8px; position: relative;">
<img src="../assets/images/logos/amlogic.svg" alt="Amlogic" width="96" style="cursor: pointer;">
<div style="position: absolute; top: 50%; left: 50%; transform: translate(-50%, -50%); opacity: 0; transition: opacity 0.2s; font-weight: 500; font-size: 0.9rem; pointer-events: none;">Amlogic</div>
</div>

<div style="text-align: center; padding: 1rem; border: 1px solid var(--md-default-fg-color--lightest); border-radius: 8px; position: relative;">
<img src="../assets/images/logos/cypress.svg" alt="Cypress" width="96" style="cursor: pointer;">
<div style="position: absolute; top: 50%; left: 50%; transform: translate(-50%, -50%); opacity: 0; transition: opacity 0.2s; font-weight: 500; font-size: 0.9rem; pointer-events: none;">Cypress</div>
</div>

<div style="text-align: center; padding: 1rem; border: 1px solid var(--md-default-fg-color--lightest); border-radius: 8px; position: relative;">
<img src="../assets/images/logos/fairphone.svg" alt="Fairphone" width="96" style="cursor: pointer;">
<div style="position: absolute; top: 50%; left: 50%; transform: translate(-50%, -50%); opacity: 0; transition: opacity 0.2s; font-weight: 500; font-size: 0.9rem; pointer-events: none;">Fairphone</div>
</div>

<div style="text-align: center; padding: 1rem; border: 1px solid var(--md-default-fg-color--lightest); border-radius: 8px; position: relative;">
<img src="../assets/images/logos/siemens.svg" alt="Siemens" width="96" style="cursor: pointer;">
<div style="position: absolute; top: 50%; left: 50%; transform: translate(-50%, -50%); opacity: 0; transition: opacity 0.2s; font-weight: 500; font-size: 0.9rem; pointer-events: none;">Siemens</div>
</div>

<div style="text-align: center; padding: 1rem; border: 1px solid var(--md-default-fg-color--lightest); border-radius: 8px; position: relative;">
<img src="../assets/images/logos/linuxfoundation.svg" alt="Linux Foundation" width="96" style="cursor: pointer;">
<div style="position: absolute; top: 50%; left: 50%; transform: translate(-50%, -50%); opacity: 0; transition: opacity 0.2s; font-weight: 500; font-size: 0.9rem; pointer-events: none;">Linux Foundation</div>
</div>

<div style="text-align: center; padding: 1rem; border: 1px solid var(--md-default-fg-color--lightest); border-radius: 8px; position: relative;">
<img src="../assets/images/logos/realtek.svg" alt="Realtek" width="128" style="cursor: pointer;">
<div style="position: absolute; top: 50%; left: 50%; transform: translate(-50%, -50%); opacity: 0; transition: opacity 0.2s; font-weight: 500; font-size: 0.9rem; pointer-events: none;">Realtek</div>
</div>

<div style="text-align: center; padding: 1rem; border: 1px solid var(--md-default-fg-color--lightest); border-radius: 8px; position: relative;">
<img src="../assets/images/logos/hitachi.svg" alt="Hitachi Energy" width="96" style="cursor: pointer;">
<div style="position: absolute; top: 50%; left: 50%; transform: translate(-50%, -50%); opacity: 0; transition: opacity 0.2s; font-weight: 500; font-size: 0.9rem; pointer-events: none;">Hitachi Energy</div>
</div>

<div style="text-align: center; padding: 1rem; border: 1px solid var(--md-default-fg-color--lightest); border-radius: 8px; position: relative;">
<img src="../assets/images/logos/se.svg" alt="Schneider Electric" width="96" style="cursor: pointer;">
<div style="position: absolute; top: 50%; left: 50%; transform: translate(-50%, -50%); opacity: 0; transition: opacity 0.2s; font-weight: 500; font-size: 0.9rem; pointer-events: none;">Schneider Electric</div>
</div>

</div>

# Kselftest

The Linux kernel contains a set of "self tests" under the
`tools/testing/selftests/` directory. These are intended to be small tests to
exercise individual code paths in the kernel.

You can use LAVA to deploy a kernel with matching kselftest binaries and run
these tests on both virtual and physical hardware.

This example job runs the `arm64` kselftest on an `ampereone` server. The matching
kselftest binaries are deployed as an overlay into the NFS root filesystem.

```yaml
device_type: ampereone
job_name: arm64-kselftest
visibility: public
priority: 27

timeouts:
  job:
    minutes: 120
  connection:
    minutes: 2
  actions:
    finalize:
      seconds: 60
    power-off:
      seconds: 60

context:
  arch: arm64
  extra_kernel_args: ' rw kvm-arm.mode=nvhe'

actions:
- deploy:
    to: tftp
    kernel:
      url: https://storage.tuxsuite.com/public/ampere/ci/builds/3BgK8l8H3lwyccIe08ogX16JPj2/Image.gz
      compression: gz
      type: image
    nfsrootfs:
      url: https://storage.tuxboot.com/debian/20250722/trixie/arm64/rootfs.tar.xz
      compression: xz
      format: tar
      overlays:
        kselftest:
          url: https://storage.tuxsuite.com/public/ampere/ci/builds/3BgK8l8H3lwyccIe08ogX16JPj2/kselftest.tar.xz
          compression: xz
          format: tar
          path: /opt/kselftests/default-in-kernel/
        modules:
          url: https://storage.tuxsuite.com/public/ampere/ci/builds/3BgK8l8H3lwyccIe08ogX16JPj2/modules.tar.xz
          compression: xz
          format: tar
          path: /usr/
    os: debian
    timeout:
      minutes: 30

- boot:
    method: grub
    commands: nfs
    auto_login:
      login_prompt: 'login:'
      username: root
    prompts:
    - root@(.*):[/~]#
    - '/ #'
    timeout:
      minutes: 30

- test:
    definitions:
    - from: url
      name: kselftest-arm64
      path: automated/linux/kselftest/kselftest.yaml
      repository: https://github.com/Linaro/test-definitions/releases/download/2025.10.01/2025.10.tar.zst
      compression: zstd
      lava-signal: kmsg
      parameters:
        ENVIRONMENT: production
        KSELFTEST_PATH: /opt/kselftests/default-in-kernel
        SHARD_INDEX: 1
        SHARD_NUMBER: 1
        SKIPFILE: skipfile-lkft.yaml
        SKIP_INSTALL: 'true'
        TST_CMDFILES: arm64
    timeout:
      minutes: 60
```

See [kselftest test definition](https://github.com/Linaro/test-definitions/blob/master/automated/linux/kselftest/kselftest.yaml)
for what each test parameter does.

# Security Policy

## Reporting a vulnerability

The LAVA project is stewarded by Linaro. Security vulnerabilities in LAVA are
handled through the [Linaro Security Incident Handling Process (VDP)](https://www.linaro.org/vdp).

To report a vulnerability, email the Linaro Product Security Incident Response
Team (PSIRT) at **psirt@linaro.org**.

When reporting, please include:

- the affected component (server, dispatcher, worker, lava-dispatcher-host)
- the LAVA version or git commit
- steps to reproduce the issue
- exploit code, if available

Machine-readable details are published at <https://www.linaro.org/.well-known/security.txt>.

## Scope

In scope: the LAVA components in this repository: the LAVA server, dispatcher,
worker, and `lava-dispatcher-host`.

Out of scope:

- third-party dependencies; report those to their maintainers
- operator-configured deployments; misconfiguration of the host system
  (systemd units, Docker, network) is the operator's responsibility

## Disclosure and response

Please do not disclose the vulnerability publicly until a fix has been
released. The Linaro PSIRT triages reports, performs a risk assessment, and
works on a permanent solution (hotfix, support release, or major release
depending on the impact), as described in the VDP.

## Bug bounty

The LAVA project or Linaro does not offer a bug bounty program.

# Job Schema

A LAVA job definition is a YAML file that describes the test job to execute. It
specifies what device to use, how to deploy software, how to boot, and what
tests to run.

In general, the schema used for the job definition is constrained. The schema
requires that specific elements must exist in specific formats. The LAVA server
performs basic schema validation when the job is submitted. Full validation is
performed on the LAVA worker at runtime during the job validation phase.

This page documents the top-level elements supported in LAVA job definitions,
with links to more detailed schemas.

## job_name

**Required** `string` (1–200 characters)

Name of the job, sometime called `description`.

```yaml
job_name: network-smoke-test
```

## device_type

`string` (1–50 characters)

**Required** by single node jobs. These jobs will fail to validate if no device
type is given.

```yaml
device_type: qemu
```

For [multinode](../../user/advanced-tutorials/multinode.md) jobs, the device type
is defined per-role inside the `lava-multinode` protocol block.

!!! note
    The device type **must** exist on the LAVA instance for the submission to
    be accepted by the scheduler.

## tags

`list[str]`

A list of device tags that the assigned device must have. This is used to
select a specific subset of devices of the given device type.

```yaml
tags:
- usb-eth
```

## timeouts

**Required** `dict`

See [Timeouts](./timeouts.md)

## visibility

**Required** `string` or `dict`

Controls who can view the job and its results.

```yaml
visibility: "public"
```

```yaml
visibility:
  group:
  - developers
  - project
```

Supported values:

| Value        | Description                                    |
| ------------ | ---------------------------------------------- |
| `"public"`   | Visible to everyone, including anonymous users |
| `"personal"` | Visible only to the submitter                  |
| `group`      | Visible to members of the specified group(s)   |

## priority

`string` or `int`

Controls the scheduling priority of the job.

```yaml
priority: high
```

Supported values:

| Value              | Description                                |
| ------------------ | ------------------------------------------ |
| `"high"`           | High priority                              |
| `"medium"`         | Medium priority (default)                  |
| `"low"`            | Low priority                               |
| `0`–`100`          | Higher value means higher priority         |

## context

`dict`

Allows individual jobs to override selected device configuration values. The
keys must be from the set of allowed context variables.

??? "Allowed context variables"
    - `arch`
    - `boot_character_delay`
    - `boot_console`
    - `boot_retry`
    - `boot_root`
    - `booti_dtb_addr`
    - `booti_kernel_addr`
    - `booti_ramdisk_addr`
    - `bootloader_prompt`
    - `bootm_dtb_addr`
    - `bootm_kernel_addr`
    - `bootm_ramdisk_addr`
    - `bootz_dtb_addr`
    - `bootz_kernel_addr`
    - `bootz_ramdisk_addr`
    - `console_device`
    - `cpu`
    - `custom_kernel_args`
    - `deploy_character_delay`
    - `extra_kernel_args`
    - `extra_nfsroot_args`
    - `extra_options`
    - `failure_retry`
    - `guestfs_driveid`
    - `guestfs_interface`
    - `guestfs_size`
    - `kernel_loglevel`
    - `kernel_start_message`
    - `lava_test_results_dir`
    - `machine`
    - `memory`
    - `menu_interrupt_prompt`
    - `model`
    - `monitor`
    - `mustang_menu_list`
    - `netdevice`
    - `no_kvm`
    - `no_network`
    - `serial`
    - `test_character_delay`
    - `tftp_mac_address`
    - `uboot_altbank`
    - `uboot_extra_error_message`
    - `uboot_needs_interrupt`
    - `vga`

Check device type and action documentation for the detailed usages.

## environment

`dict`

See [environment](./environment.md)

## secrets

`dict[str: str]`

A dictionary of secret values.

```yaml
secrets:
  API_USER: kernel-ci
  API_KEY: b43614a9583f9c74b989914a91d1cfd9
```

Remote artifact tokens with `Token name` and `Token string` paires can be added
on your profile page. The value of a secret can be the `Token name` so the real
token does not need to appear in the job definition. LAVA replaces that name
with the actual `Token string` for the job runs.

In the following example,  LAVA uses `token-name` to show the job definition and
replaces it with the actual `Token string` for the job run.

```yaml
secrets:
  API_KEY: token-name
```

This dictionary will be written to the LAVA overlay using the `secrets` file
name. The file can be discovered and sourced from a test shell:

```shell
lava_dir="$(find /lava-* -maxdepth 0 -type d | grep -E '^/lava-[0-9]+' 2>/dev/null | sort | tail -1)"
source "$lava_dir/secrets"
```

!!! note
    `/lava-*` is the default LAVA overlay path; adjust it when needed.

!!! warning "Avoid secret leaks"
    - Use per-user remote artifact tokens whenever possible.
    - Always set job visibility to `personal` so only you can see the secrets.
    - Do not print the secret values in test log.
    - A compromise of the DUT or its filesystem can leak your secrets. Do not
    use this feature with highly sensitive credentials — use a proper secrets
    management system instead.

## metadata

`dict`

An arbitrary set of key-value pairs attached to the job. The data can be
retrieved via LAVA APIs.

```yaml
metadata:
  build-url: https://ci.example.com/builds/42
  toolchain: clang-r377782b
```

## actions

**Required** `list`

A list of actions blocks to execute. Every action block supports these optional
parameters:

| Key                      | Type       | Description                         |
| ------------------------ | ---------- | ----------------------------------- |
| `timeout`                | `dict`     | Timeout for this action             |
| `timeouts`               | `dict`     | Per-child-action timeout            |
| `namespace`              | `string`   | Namespace for this action           |
| `connection-namespace`   | `string`   | Namespace to use for the connection |
| `failure_retry`          | `int`      | Number of times to retry on failure |
| `failure_retry_interval` | `int`      | Seconds to wait between retries     |
| `repeat`                 | `int`      | Number of times to repeat           |

The supported action types are:

### deploy

`dict`

See [Deploy action](./actions/deploy/index.md) for image deployment.

### boot

`dict`

See [Boot action](./actions/boot/common.md) for booting a device.

### test

`dict`

See [Test action](./actions/test.md) for running tests.

### command

`dict`

See [Command action](./actions/command.md) for executing a pre-defined command
on the LAVA worker.

## notify

`dict`

See [Notifications](./notifications.md)

## protocols

`dict`

See [Protocols](./protocols.md)

# Flasher

The `flasher` deployment method downloads the specific images and runs a list of
commands defined in the device configuration to flash the images onto the DUT.
It can be used for devices that require custom flashing procedures.

```yaml
actions:
- deploy:
    to: flasher
    images:
      recovery_image:
        url: https://example.com/image.img.xz
        compression: xz
    timeout:
      minutes: 15
```

The device dictionary must define the
[flasher_deploy_commands](../../../configuration/device-dictionary.md#flasher_deploy_commands).

## images

A dictionary of named images to download. Each image accepts the standard
[artifact parameters](./index.md#artifacts).

### overlays

See [overlays](./index.md#overlays). The [LAVA overlay](./index.md#lava-overlay)
must be applied if the job definition contains tests.

## uniquify

Optional boolean. By default, LAVA saves each downloaded file in a separate
subdirectory named after image key to avoid filename collisions.

Set to `false` if the flash commands expect files in a flat directory:

```yaml
- deploy:
    to: flasher
    uniquify: false
    images:
      firmware0:
        url: https://example.com/firmware0.bin
      firmware1:
        url: https://example.com/firmware1.bin
```

## Sample job

This sample job demonstrates using a USB-SD-Mux with custom flashing commands
defined in the device dictionary to flash the Debian image onto the DUT's SD card.

```yaml
job_name: RPi3B flasher sample job
device_type: bcm2837-rpi-3-b-plus

priority: medium
visibility: public

timeouts:
  job:
    minutes: 30

actions:
- deploy:
    to: flasher
    images:
      recovery_image:
        url: https://raspi.debian.net/tested/20231109_raspi_3_bookworm.img.xz
        compression: xz
        format: ext4
        partition: 1
        overlays:
          lava: true
    timeout:
      minutes: 15

- boot:
    method: minimal
    auto_login:
      login_prompt: 'login:'
      username: root
    prompts:
    - 'root@rpi3-20231109:'
    timeout:
      minutes: 5

- test:
    definitions:
    - from: inline
      repository:
        metadata:
          format: Lava-Test Test Definition 1.0
          name: test-definition-example
        run:
          steps:
          - lava-test-case run-uname-a --shell uname -a
          - lava-test-case check-os-id --shell 'cat /etc/os-release | grep "ID=debian"'
      path: inline/test-definition-example.yaml
      name: test-suite-example
    timeout:
      minutes: 5
```

--8<-- "refs.txt"

# UEFI Menu

The `uefi-menu` method selects a pre-defined menu entry in the UEFI
configuration for the device. In most cases this is used to execute a different
bootloader. For example, a `fastboot` device can execute `fastboot` from a menu
item, or a device could execute a PXE menu item to download and execute GRUB.

```yaml
- boot:
    method: uefi-menu
    commands: fastboot
    prompts:
    - 'root@debian:~#'
```

!!! warning
    Although it *is* possible to create new menu entries in UEFI each time a
    test job starts, this has proven to be unreliable on the device types tested
    so far. If the build of UEFI is not able to download a bootloader using PXE,
    then experiment with creating a UEFI menu item that executes a local file,
    and place a build of GRUB on local storage. Build the GRUB binary using
    `grub-mkstandalone` to ensure that all modules are available.

UEFI menus renumber themselves each time an item is added or removed, so LAVA
must match the *description* of the menu item and then identify the correct
selector to send the right character to execute that menu option. This means
admins must create the same menu structures on each device of the same device
type and correlate the text content of the menu with the
[device-type templates](../../../configuration/device-type-template.md).

To use `uefi-menu`, the device must offer a menu from which items can be
selected using a standard regular expression (`a-zA-Z0-9\s\:`). For example:

```text
[1] bootloader
[2] boot from eMMC
[3] Boot Manager
[4] Reboot
```

The device-type template needs to specify the `separator` (whitespace in the
example above) and how to identify the **item** matching the requested selector:

```yaml
- boot:
    method: uefi-menu
    parameters:
      item_markup:
      - "["
      - "]"
      item_class: '0-9'
      separator: ' '
```

This allows LAVA to match a menu item matching `\[[0-9]\] ['a-zA-Z0-9\s\:']`
and select the correct selector when the menu item string matches one (and only
one) line output by the UEFI menu. In this example the selector must be a digit.

## commands

The template must also specify which menu item to select, according to the
`commands` set in the test job:

```yaml
- boot:
    method: uefi-menu
    commands: fastboot
```

The device type template would then need:

```yaml
- boot:
    method: uefi-menu
    parameters:
      fastboot:
      - select:
          items:
          - 'boot from eMMC'
```

See also [commands](./common.md#commands).

## line_separator

Specifies the line ending sent to the UEFI console. Supported values are `dos`
(the default, `\r\n`) and `unix` (`\n`). Overrides the value set in device type.

```yaml
- boot:
    method: uefi-menu
    line_separator: unix
```

!!! warning
    This documentation may be outdated. If you use the boot method and see any
    issues, please submit an update. [Contact](../../../../introduction/contact.md)
    the LAVA dev team if help needed.

--8<-- "refs.txt"

# Running arbitrary code with docker

## Introduction

Testing in LAVA will often require running arbitrary code on the LAVA
dispatcher. Of course, no lab admin would ever allow users running arbitrary
code on their systems, so we need a solution to have users run their arbitrary
code in isolated containers.

This document describes how to use docker to cover the use cases were users
need to run arbitrary code on the lava dispatcher.

## Use case 1: Fastboot deploy from a docker container

Deploying and booting fastboot devices using docker allows you to provide your
own image with a pre-installed fastboot binary, making test jobs faster. To do
this, you just need to add the `docker` section to the fastboot deploy and boot
actions:

```yaml
actions:
# ...
    - deploy:
        to: fastboot
        docker:
            image: my-fastboot-image
        timeout:
            minutes: 15
        images:
            boot:
                url: http://example.com/images/aosp/hikey/boot.img
                reboot: hard-reset
# ...
    - boot:
        method: fastboot
        docker:
            image: my-fastboot-image
        prompts:
            - 'healthd: No battery devices found'
            - 'hikey: '
            - 'console:'
        timeout:
            minutes: 15
```

## Use case 2: Manipulating downloaded images

Some use cases involve downloading different build images and combining them
somehow. Examples include but are not limited to:

* Injecting kernel modules into a rootfs
* Downloading separate kernel/modules/rootfs and combining them in a single
  image for flashing.

This can be achieved using the "**downloads**" deploy method (note
"**downloads**", plural; "**download**"), plus postprocessing instructions:

```yaml
actions:
# ...
    - deploy:
        to: downloads
        images:
            # [...]
            kernel:
                url: http://images.com/.../Image
            modules:
                url: http://images.com/.../modules.tar.xz
            rootfs:
                url: http://images.com/.../rootfs.ext4.gz
                apply-overlay: true
        postprocess:
            docker:
                image: my-kir-image
                steps:
                    - /kir/lava/board_setup.sh hi6220-hikey-r2
```

This will cause all the specified images to be downloaded, and then a docker container
running the specified will be executed.

* The container will have the download directory as the current directory.
    * i.e. the downloaded images will be present in the current directory.
* The steps listed in `steps:` will be executed in order
* Any file modified or created by the steps is left around for later usage.

After the postprocessing fininshes, the resulting images can be used by
specifying their location using the `downloads://` pseudo-URL in a subsequent
deploy action:

```yaml
# ...
    - deploy:
        to: fastboot
        images:
            system:
                rootfs: downloads://rootfs.img
            boot:
                url: downloads://boot.img
```

Those pseudo-URLs are relative to the download directory, from where the
container was executed.

## Use case 3: Running tests from the docker container

To run tests from a docker container, you just need to add a `docker` section
to the well-known LAVA test shell action:

```yaml
# ...
    - test:
        docker:
            image: my-adb-image
        timeout:
        minutes: 5
        definitions:
            - repository:
                # [...]
                from: inline
                path: inline-smoke-test
                name: docker-test
# ...
```

The specified test definitions will be executed inside a container running the
specified image, and the following applies:

* The USB connection to the device is shared with the container, so that you
  can run `adb` and have it connect to the device.
    * For example this can be used in AOSP jobs to run CTS/VTS against the
      device.
* The device connection settings are exposed to the tests running in the
  container via environment variables. For example, assume the given connection
  commands in the device configuration:
    ```jinja
    {% set connection_list = ['uart0', 'uart1'] %}
    {% set connection_commands = {
        'uart0': 'telnet localhost 4002',
        'uart1': 'telnet 192.168.1.200 8001',
        }
    %}
    {% set connection_tags = {'uart1': ['primary', 'telnet']} %}
    ```

    These connection settings will be exported to the container environment as:

    ```shell
    LAVA_CONNECTION_COMMAND='telnet 192.168.1.200 8001'
    LAVA_CONNECTION_COMMAND_UART0='telnet localhost 4002'
    LAVA_CONNECTION_COMMAND_UART1='telnet 192.168.1.200 8001'
    ```

    Of course, for this to work the network addresses used in the configuration
    need to be resolvable from inside the docker container. This requires
    coordination with the lab administration.
* The device power control commands are also exposed in the following
  environment variables: `LAVA_HARD_RESET_COMMAND`, `LAVA_POWER_ON_COMMAND`,
  and `LAVA_POWER_OFF_COMMAND`.

  The same caveat as with the connection commands: any network addresses used
  in such commands need to be accessible from inside the container.

  Note that each of these operations can actually require more than one
  command, in which case the corresponding environment variable will have the
  multiple commands with `&&` between them. Because of this, the safest way to
  run the commands is passing the entire contents of the variable as a single
  argument to `sh -c`, like this:

  ```bash
  sh -c "${LAVA_HARD_RESET_COMMAND}"
  ```

## See also

* LAVA release notes:
    * [2020.01](https://gitlab.com/lava/lava/-/wikis/releases/2020.01)
    * [2020.02](https://gitlab.com/lava/lava/-/wikis/releases/2020.02)
    * [2020.04](https://gitlab.com/lava/lava/-/wikis/releases/2020.04)
* [Improved Android Testing in LAVA with Docker](https://connect.linaro.org/resources/ltd20/ltd20-304/). Talk at Linaro Tech Days 2020.

# Running arbitrary code with docker

## Introduction

Testing in LAVA will often require running arbitrary code on the LAVA
dispatcher. Of course, no lab admin would ever allow users running arbitrary
code on their systems, so we need a solution to have users run their arbitrary
code in isolated containers.

This document describes how to use docker to cover the use cases were users
need to run arbitrary code on the lava dispatcher.

## Use case 1: Fastboot deploy from a docker container

Deploying and booting fastboot devices using docker allows you to provide your
own image with a pre-installed fastboot binary, making test jobs faster. To do
this, you just need to add the `docker` section to the fastboot deploy and boot
actions:

```yaml
actions:
# ...
    - deploy:
        to: fastboot
        docker:
            image: my-fastboot-image
        timeout:
            minutes: 15
        images:
            boot:
                url: http://example.com/images/aosp/hikey/boot.img
                reboot: hard-reset
# ...
    - boot:
        method: fastboot
        docker:
            image: my-fastboot-image
        prompts:
            - 'healthd: No battery devices found'
            - 'hikey: '
            - 'console:'
        timeout:
            minutes: 15
```

## Use case 2: Manipulating downloaded images

Some use cases involve downloading different build images and combining them
somehow. Examples include but are not limited to:

* Injecting kernel modules into a rootfs
* Downloading separate kernel/modules/rootfs and combining them in a single
  image for flashing.

This can be achieved using the "**downloads**" deploy method (note
"**downloads**", plural; "**download**"), plus postprocessing instructions:

```yaml
actions:
# ...
    - deploy:
        to: downloads
        images:
            # [...]
            kernel:
                url: http://images.com/.../Image
            modules:
                url: http://images.com/.../modules.tar.xz
            rootfs:
                url: http://images.com/.../rootfs.ext4.gz
                apply-overlay: true
        postprocess:
            docker:
                image: my-kir-image
                steps:
                    - /kir/lava/board_setup.sh hi6220-hikey-r2
```

This will cause all the specified images to be downloaded, and then a docker container
running the specified will be executed.

* The container will have the download directory as the current directory.
    * i.e. the downloaded images will be present in the current directory.
* The steps listed in `steps:` will be executed in order
* Any file modified or created by the steps is left around for later usage.

After the postprocessing fininshes, the resulting images can be used by
specifying their location using the `downloads://` pseudo-URL in a subsequent
deploy action:

```yaml
# ...
    - deploy:
        to: fastboot
        images:
            system:
                rootfs: downloads://rootfs.img
            boot:
                url: downloads://boot.img
```

Those pseudo-URLs are relative to the download directory, from where the
container was executed.

## Use case 3: Running tests from the docker container

To run tests from a docker container, you just need to add a `docker` section
to the well-known LAVA test shell action:

```yaml
# ...
    - test:
        docker:
            image: my-adb-image
        timeout:
        minutes: 5
        definitions:
            - repository:
                # [...]
                from: inline
                path: inline-smoke-test
                name: docker-test
# ...
```

The specified test definitions will be executed inside a container running the
specified image, and the following applies:

* The USB connection to the device is shared with the container, so that you
  can run `adb` and have it connect to the device.
    * For example this can be used in AOSP jobs to run CTS/VTS against the
      device.
* The device connection settings are exposed to the tests running in the
  container via environment variables. For example, assume the given connection
  commands in the device configuration:
    ```jinja
    {% set connection_list = ['uart0', 'uart1'] %}
    {% set connection_commands = {
        'uart0': 'telnet localhost 4002',
        'uart1': 'telnet 192.168.1.200 8001',
        }
    %}
    {% set connection_tags = {'uart1': ['primary', 'telnet']} %}
    ```

    These connection settings will be exported to the container environment as:

    ```shell
    LAVA_CONNECTION_COMMAND='telnet 192.168.1.200 8001'
    LAVA_CONNECTION_COMMAND_UART0='telnet localhost 4002'
    LAVA_CONNECTION_COMMAND_UART1='telnet 192.168.1.200 8001'
    ```

    Of course, for this to work the network addresses used in the configuration
    need to be resolvable from inside the docker container. This requires
    coordination with the lab administration.
* The device power control commands are also exposed in the following
  environment variables: `LAVA_HARD_RESET_COMMAND`, `LAVA_POWER_ON_COMMAND`,
  and `LAVA_POWER_OFF_COMMAND`.

  The same caveat as with the connection commands: any network addresses used
  in such commands need to be accessible from inside the container.

  Note that each of these operations can actually require more than one
  command, in which case the corresponding environment variable will have the
  multiple commands with `&&` between them. Because of this, the safest way to
  run the commands is passing the entire contents of the variable as a single
  argument to `sh -c`, like this:

  ```bash
  sh -c "${LAVA_HARD_RESET_COMMAND}"
  ```
#### Using password protected docker images

To pull images from a password protected registry add a login section
with the registry domain name, username and password.

```yaml
# ...
    - test:
        docker:
            image: example.com/my-adb-image
                login:
                    registry: example.com
                    user: foobar
                    password: my_password
# ...
```

**Warning**: The pulled image will be cached on the local daemon and will
be available to other jobs running on same worker.

## See also

* LAVA release notes:
    * [2020.01](https://gitlab.com/lava/lava/-/wikis/releases/2020.01)
    * [2020.02](https://gitlab.com/lava/lava/-/wikis/releases/2020.02)
    * [2020.04](https://gitlab.com/lava/lava/-/wikis/releases/2020.04)
* [Improved Android Testing in LAVA with Docker](https://connect.linaro.org/resources/ltd20/ltd20-304/). Talk at Linaro Tech Days 2020.

# Job output

A simple [QEMU job page](https://validation.linaro.org/scheduler/job/2009038) will look like this:

![job output](job-output.png)

## Header

At the top of the page, job meta-data are listed in two columns:

* **static data** on the left (description, device-type, submitter, ...)
* **dynamic data** on the right (job id, actual device, duration, ...)

## Summary

This is the main tabular as it includes the logs of the job.

You can select the levels that should be displayed or hidden by clicking on the corresponding buttons:

* `debug`, `info`, `warning` and `error`: LAVA logs
* `input`: DUT input
* `target` and `feedback`: DUT output
* `results`: LAVA results

!!! tip "`target` vs `feedback`"
    Both `target` and `feedback` are DUT output.

    When using [namespaces](../../introduction/glossary.md#namespace), DUT
    output for the current namespace will be in `target` while outputs coming
    from over namespace will be in `feedback`.

### Pipeline

On the right side of the page, the `pipeline` element allows you to jump
directly at the log of a specific action of the pipeline.


## Details

In this tabular, you have access to the internal details of the LAVA pipeline
that was running for this specific job definition.

## Timing

This tabular list each individual action with it's duration along with its timeout.

This helps test writers to check for actions with a duration which is
much shorter than the requested timeout. Reducing these timeouts will allow
failures to be identified more quickly.

The graph only shows actions that are longer than 1 second. The full list is
available in the table.

--8<-- "refs.txt"

# Job submit

Jobs can be submitted from either the [web interface](#web-interface), [lavacli](#lavacli) or the REST API.

## QEMU job

You will start with this simple QEMU job:

```yaml
--8<-- "jobs/qemu.yaml"
```

## Web interface

In order to submit from the web interface, you should first login and then go to
`Scheduler > Submit`.

![submit menu](submit-job-menu.png)

In the submit page, copy and paste the job definition and click on `Validate`.

![submit validate](submit-job-validate.png)

The server will validate the job definition and return any errors or warning found.

If the job is valid, the `Submit` button will become active. Clicking on it
will actually submit the job.

![submit](submit-job-submit.png)


## lavacli

In order to submit using [lavacli][lavacli], start by installing and configure
lavacli with the [lavacli tutorial](./lavacli.md).

Submit the job with:
```shell
lavacli -i <identity> jobs submit qemu.yaml
```

lavacli will submit and print the job identifier that you can use to inspect
the current job.

```shell
lavacli -i <identity> jobs show <id>
lavacli -i <identity> jobs logs <id>
lavacli -i <identity> jobs wait <id>
```

--8<-- "refs.txt"

# TMPFS

Downloads files to a temporary directory on the LAVA worker, making them
available locally for boot methods that flash or run on the worker.

To date, this deployment method is used by the following boot methods:
`cmsis-dap`, `dfu`, `jlink`, `openocd`, `pyocd`, and `qemu`.

```yaml title="QEMU example" hl_lines="1-7"
- deploy:
    to: tmpfs
    images:
        rootfs:
          image_arg: -drive format=qcow2,file={rootfs}
          url: http://example.com/debian-buster.qcow2.zst
          compression: zstd
    timeout:
      minutes: 5

- boot:
    method: qemu
    media: tmpfs
    prompts:
    - "root@debian:"
    auto_login:
      login_prompt: "login:"
      username: root
    timeout:
      minutes: 5
```


```yaml title="pyOCD example" hl_lines="1-5"
- deploy:
    to: tmpfs
    images:
        zephyr:
          url: http://example.com/zephyr.bin
    timeout:
      minutes: 5

- boot:
    method: pyocd
    timeout:
      minutes: 5
```

## images

To deploy images using tmpfs, the job definition writer needs to specify a
unique **label** for each image. Depending on the boot method, an `image_arg`
may also be required to specify how the tool handles the image.

### label

The label is the key under `images:` that identifies each image. It is used to
identify the image by the following boot actions. It also can be used as a
placeholder `{label}` in the `image_arg` to reference the downloaded file path.

### image_arg

The `image_arg` determines how the boot method handles the image. When specified,
it should include a placeholder (referencing the label) which exactly matches
the key of the same block in the list of images. The actual location of the
downloaded file will then replace the placeholder.

In the QEMU example, the label is `rootfs` and the `image_arg` includes the
matching placeholder `{rootfs}`. If the final location of the downloaded image
is `/<DISPATCHER_DOWNLOAD_DIR>/tmp/442/deployimages-td1ofegd/rootfs/debian-buster.qcow2`,
then the final argument passed to QEMU would include
`-drive format=raw,file=/<DISPATCHER_DOWNLOAD_DIR>/tmp/442/deployimages-td1ofegd/rootfs/debian-buster.qcow2`.

!!! note
    Single brace before and after the label and **no whitespace**. This is test
    job definition syntax, not Jinja.

Multiple images can be supplied, but the test job definition writer is
responsible for ensuring that the `image_arg` values make sense to the target
tool.

!!! note
    For some boot methods like `pyocd`, `image_arg` is optional. The flashing
    tool will automatically use the downloaded file path with flashing arguments
    defined in the device type. For `qemu`, `image_arg` is required because QEMU
    needs to know how to use each image (e.g., as a disk drive via `-drive`, as
    a kernel via `-kernel`, or as an initrd via `-initrd`).

### url

See [url](./index.md#url)

### compression

See [compression](./index.md#compression)

# Deploying a Raspberry Pi4B as a LAVA worker

Due to the distributed nature of LAVA architecture, worker part of the software
does not require latest and most expensive hardware. LAVA lab is already using
it in it's own setup for particular instances.

Here we will provide a reference and how-to for deploying one such RPi4 as a
LAVA worker.

## Get the hardware

What you will need for a worker is:

* RPi4 with power supply
* SD card with min 8GB (only temporarily)
* SSD / Flash drive
* SATA(M.2, etc) to USB adapter
* FTDI USB to serial UART cable

There is a list of know working and also problematic [USB adapters](https://jamesachambers.com/raspberry-pi-4-usb-boot-config-guide-for-ssd-flash-drives/)
along with the more extensive setup instructions which we will cover here.

## Prepare the Pi

Download the [Raspberry Pi OS](https://www.raspberrypi.org/software/operating-systems/) and flash it onto an SD card.
Once you get to login prompt the user is `pi` and password is `raspberry`
then perform the following steps:

```shell
apt update
apt full-upgrade
rpi-update
reboot
sudo rpi-eeprom-update -d -a
reboot
```

After that launch the raspi-config:

```shell
raspi-config
```

In the menu, find the `Boot ROM Version` (it defers where it is depending on
the version of raspi-config, either under `Boot options` or
`Advanced settings`).
Select **Latest** version of boot ROM software. In the next menu, for a question "Reset boot ROM to defaults?", select **NO**.

Then in the same submenu select "Boot Order" and choose "USB Boot". That's
it.

## Prepare the SSD

Currently we only support a Debian images for LAVA installation (unless you'd
like to use docker).

Get the latest Debian image from [Tested images](https://raspi.debian.net/tested-images/) - download the one for the RPi 4 family. This is a good
stock image for a number of reasons, not least of which is that it already has
enable-uart set to 1 in the boot config.txt, which enables serial output.
The user for this image is root and there is no password.

Unpack it and “dd” it onto the SSD.
If everything is OK, you should be able to remove the SD card from th Pi and
boot the system from the SSD over USB.

We're now ready to install the LAVA worker.

## Installing LAVA worker

* Start with upgrading the system and installing the packages:

```shell
apt update && apt dist-upgrade
apt install net-tools wget gnugpg curl
```

* Edit the /etc/hostname and /etc/hosts files with appropriate hostname / IP
address.
* Add the LAVA repo to apt:

```shell
echo "deb https://apt.lavasoftware.org/release buster main" > /etc/apt/sources.list.d/lava.list
```

* Add LAVA archive signing keys to apt:

```shell
curl -fsSL https://apt.lavasoftware.org/lavasoftware.key.asc | apt-key add -
```

* Install the lava-dispatcher package

```shell
apt update
apt install lava-dispatcher
apt dist-upgrade -t buster
```

## Auto-registration of worker (or token)

We need to tell LAVA master to `trust` this worker and there's two ways to
do this. You can either add the following line to `/etc/lava-server/settings.d/01-autoregister.yaml` on the LAVA master (remember to change the netmask for
the subnet where the Pi is):

```yaml
WORKER_AUTO_REGISTER_NETMASK: ["172.0.0.0/8"]
```

Or, alternatively you can add a token of your choice in the LAVA worker (Pi 4)
in `/etc/lava-dispatcher/lava-worker` settings and paste the same token in the
LAVA admin UI on the master in the worker section for this particular worker.

## Setting up the server url

Once you have lava-worker running you need to point it to the LAVA server it
will connect to. Do this in `/etc/lava-dispatcher/lava-worker` on the Pi.

After that RPi4 worker should be visible in the LAVA server UI and marked as
**Online**.

# LAVA scheduler

This process is is responsible for scheduling jobs on devices.

## Command line

This daemon is part of lava-server and is started by: `lava-server manage lava-scheduler`

## Service

The systemd service is called `lava-scheduler`.

## Dependencies

lava-scheduler should be able to:

* connect to the [postgresql](./postgresql.md) database
* connect to [lava-publisher](./lava-publisher.md) PUB socket.

## Configuration

Daemon start options:

* `/etc/default/lava-scheduler`
* `/etc/lava-server/lava-scheduler`

Django configuration:

* `/etc/lava-server/settings.conf`
* `/etc/lava-server/settings.yaml`
* `/etc/lava-server/settings.d/*.yaml`

## Logs

The logs are stored in `/var/log/lava-server/lava-scheduler.log`

The log rotation is configured in `/etc/logrotate.d/lava-scheduler-log`.

# First steps

These tutorials cover the essentials of managing your LAVA instance and lab.

| Tutorial | Description |
| -------- | ----------- |
| [Install](./basic-tutorials/instance/install.md) | Install a LAVA instance |
| [Configure](./basic-tutorials/instance/configure.md) | Configure a LAVA instance |
| [Upgrade](./basic-tutorials/instance/upgrade.md) | Upgrade a LAVA instance |
| [Backup](./basic-tutorials/instance/backup.md) | Back up a LAVA instance |
| [Permissions](./basic-tutorials/instance/permissions.md) | Manage user permissions |
| [Docker worker](./advanced-tutorials/docker-worker.md) | Install a LAVA Docker worker |
| [Device setup](basic-tutorials/device-setup/common.md) | Add devices to LAVA |

For the next steps, explore the advanced tutorials to learn more about LAVA
administration.

# Device-type template

The device-type template is a [jinja2][jinja2] template that will be extended
by [device dictionaries](./device-dictionary.md). The resulting file is the
device configuration file.

This [yaml][yaml] file is used by `lava-dispatcher` to know how to flash, boot
and communicate with a specific device.

# Configuration file

The device-type templates that are supported and provided by LAVA are stored
in `/usr/share/lava-server/device-types/`.

Admins could also provide their own device-type template using lavacli:

```shell
lavacli device-types template set qemu qemu.jinja2
```

Device-types provided by admins will be stored in
`/etc/lava-server/dispatcher-config/device-types/<name>.jinja2`.

LAVA will always look first in `/etc/lava-server/dispatcher-config/device-types/` and fallback to `/usr/share/lava-server/device-types/`.

This mean that admins can override every device-type templates, including the
ones provided by LAVA.

--8<-- "refs.txt"

# Bootloader Testing

## Interactive bootloader tests

LAVA supports testing bootloaders such as U-Boot by interacting directly with
the bootloader prompt over the device serial console.

This example job boots a Raspberry Pi 4B to the U-Boot prompt and runs
[interactive](../../technical-references/job-definition/actions/test.md#interactive)
U-Boot tests directly.

```yaml
job_name: Interactive U-Boot test
device_type: bcm2711-rpi-4-b

priority: medium
visibility: public

timeouts:
  job:
    minutes: 20
  action:
    minutes: 5
  connection:
    minutes: 2

context:
  extra_kernel_args: console_msg_format=syslog earlycon
  console_device: ttyS1

secrets:
  GITLAB_TOKEN: gitlab-private-code

actions:
- deploy:
    to: downloads
    failure_retry: 3
    images:
      gitlab_files:
        url: https://gitlab.com/api/v4/projects/58465263/jobs/9656175138/artifacts/files_from_gl.tar.xz
        compression: xz
        archive: tar
        headers:
          PRIVATE-TOKEN: gitlab-private
    timeout:
      minutes: 5

- deploy:
    to: usbg-ms
    image:
      url: https://gitlab.com/api/v4/projects/58465263/jobs/9656175138/artifacts/disk.img.gz
      compression: gz
      headers:
        PRIVATE-TOKEN: gitlab-private
    timeout:
      minutes: 5

- boot:
    method: bootloader
    bootloader: u-boot
    commands: []
    prompts:
    - 'U-Boot> '
    timeout:
      minutes: 5

- test:
    timeout:
      minutes: 5
    interactive:
    - name: basic-cmds
      prompts: ["U-Boot> ", "=> ", "/ # "]
      script:
      - command: echo "u-boot echo test"
        name: echo
        successes:
        - message: "u-boot echo test"
      - command: version
        name: version
        successes:
        - message: "U-Boot "
      - command: help test
        name: help
        successes:
        - message: "test - minimal test like /bin/sh"
      - command: setenv test_var test123printenv
      - command: printenv
        name: setenv-and-printenv
        successes:
        - message: "test_var=test123"
    - name: memory-test
      prompts: ["U-Boot> ", "=> ", "/ # "]
      script:
      - command: base
        name: print-default-base-address-offset
        successes:
        - message: "Base Address: 0x"
      - command: base 40000000
        name: set-address-offset-0x40000000
        successes:
        - message: "Base Address: 0x40000000"
      - command: base
        name: check-address-offset-0x40000000
        successes:
        - message: "Base Address: 0x40000000"
      - command: mw.b 40000000 aa 400
      - command: crc 40000000 400
        name: compute-CRC32-checksum
        successes:
        - message: "CRC32 for 40000000 ... 400003ff ==> efb5af2e"
      - command: mw 100000 aabbccdd 10
      - command: md 100000 10
        name: mw-md-100000
        successes:
        - message: "aabbccdd"
      - command: cp 100000 200000 10
      - command: md 200000 10
        name: cp-md-200000
        successes:
        - message: "aabbccdd"
      - command: cmp 100000 200000 10
        name: cmp-100000-200000-10
        successes:
        - message: 'Total of 16 word\(s\) were the same'
    - name: network
      prompts: ["U-Boot> ", "=> ", "/ # "]
      script:
      - command: dhcp
        name: dhcp
        successes:
        - message: "DHCP client bound to address"
        failures:
        - message: "TIMEOUT"
          exception: InfrastructureError
          error: "dhcp failed"
    - name: tftp-cmds
      prompts: ["U-Boot> ", "=> ", "/ # "]
      script:
      - command: setenv serverip {SERVER_IP} ; tftp 0x1000000 {JOB_ID}/downloads/common/gitlab_files/helloworld.efi
        name: tftp
        successes:
        - message: "Bytes transferred ="
```

## Running bootloader test suite

LAVA allows test scripts running from a
[docker test shell](../../user/advanced-tutorials/running-arbitrary-code-with-docker.md)
to connect to and control the DUT. This enables running existing bootloader test
suites unmodified inside an isolated container.

The following example job deploys firmware images and then runs the U-Boot test
suite inside a Docker container. LAVA exposes the DUT’s power control and serial
connection commands to the container, enabling the test framework to power-cycle
the board and interact directly with the bootloader.

```yaml
job_name: Running U-Boot test suite
device_type: kv260

visibility: public
priority: high

timeouts:
  job:
    minutes: 60
  connection:
    minutes: 2
  actions:
    finalize:
      seconds: 60

context:
  lava_test_results_dir: /var/lib/lava-%s

actions:
- deploy:
    to: downloads
    images:
      firmware:
        url: https://linaro.gitlab.io/trustedsubstrate/meta-ts//ImageA.bin.xz
        compression: xz
      os:
        url: https://gitlab.com/Linaro/trustedsubstrate/ts-testing/-/jobs/artifacts/main/raw/image/efiboot.wic.xz?job=build
        compression: xz
      tested_os:
        url: https://linaro.gitlab.io/trustedsubstrate/gpit/gpit-genericarm64.img.xz
        compression: xz
      capsule:
        url: https://linaro.gitlab.io/trustedsubstrate/meta-ts//zynqmp-kria-starter_fw.capsule
      invalid_sig:
        url: https://linaro.gitlab.io/trustedsubstrate/meta-ts//zynqmp-kria-starter_fw_invalid_sig.capsule
    postprocess:
      docker:
        image: registry.gitlab.com/linaro/trustedsubstrate/dockerfiles/lava-postprocess:arm64-38fee7c
        steps:
        - mkdir -p EFI/BOOT/
        - mcopy -i efiboot.wic@@1048576 ::/EFI/BOOT/startup.nsh EFI/BOOT/
        - mcopy -o -i efiboot.wic@@1048576 zynqmp-kria-starter_fw.capsule ::/EFI/BOOT/valid.capsule
        - mcopy -o -i efiboot.wic@@1048576 zynqmp-kria-starter_fw_invalid_sig.capsule ::/EFI/BOOT/invalid.capsule
        - sed -i "s#ledge.efi -f.*#ledge.efi -f -u ${LAVA_DISPATCHER_IP}/tmp/${LAVA_JOB_ID}/downloads/common/gpit-genericarm64.img#g" EFI/BOOT/startup.nsh
        - cat EFI/BOOT/startup.nsh
        - mcopy -o -i efiboot.wic@@1048576 EFI/BOOT/startup.nsh ::/EFI/BOOT/
        - fdisk -l efiboot.wic
        - fdisk -l ImageA.bin
    timeout:
      minutes: 20

- deploy:
    to: flasher
    images:
      ImageA:
        url: downloads://ImageA.bin
      ImageB:
        url: downloads://ImageA.bin
      sd:
        url: downloads://efiboot.wic
    timeout:
      minutes: 60

- boot:
    method: minimal
    prompts:
    - (.*)login:(.*)
    timeout:
      minutes: 5

- test:
    docker:
      image: registry.gitlab.com/linaro/trustedsubstrate/dockerfiles/u-boot-ci:arm64-77bfe23
    timeout:
      minutes: 30
    definitions:
    - from: inline
      path: inline-docker-test
      name: u-boot-ci
      repository:
        metadata:
          format: Lava-Test Test Definition 1.0
          name: inline-repo
          description: "Inline repository test for U-Boot"
        run:
          steps:
          - |
            mkdir src && cd src
            retry --delay=30 --times=3 -- git clone https://source.denx.de/u-boot/u-boot.git
            retry --delay=30 --times=3 -- git clone --depth 1 https://source.denx.de/u-boot/u-boot-test-hooks.git
            cd u-boot
            git checkout "127a42c7257a6ffbbd1575ed1cbaa8f5408a44b3"
            git describe
          - |
            python -m virtualenv /lava-${LAVA_JOB_ID}/0/uboot-venv
            . /lava-${LAVA_JOB_ID}/0/uboot-venv/local/bin/activate
            python -m pip install --root-user-action=ignore --upgrade pip
            python -m pip install --root-user-action=ignore -r ./test/py/requirements.txt
            python -m pip install --root-user-action=ignore -r ./tools/buildman/requirements.txt
          - |
            export KBUILD_OUTPUT=/lava-${LAVA_JOB_ID}/0/build/xilinx_zynqmp_kria
            mkdir -p ${KBUILD_OUTPUT}
            tools/buildman/buildman \
              -o ${KBUILD_OUTPUT} \
              -w -E -W -e -V \
              --boards xilinx_zynqmp_kria || { echo "==== Buildman failed ===="; exit 1; }
          - |
            mkdir ../u-boot-test-hooks/bin/${HOSTNAME}
            mkdir ../u-boot-test-hooks/py/${HOSTNAME}

            echo 'bash -c "${LAVA_POWER_ON_COMMAND}"' > ../u-boot-test-hooks/bin/poweron.laa
            echo 'bash -c "${LAVA_POWER_OFF_COMMAND}"' > ../u-boot-test-hooks/bin/poweroff.laa
            echo 'bash -c "${LAVA_HARD_RESET_COMMAND}"' > ../u-boot-test-hooks/bin/reset.laa

            echo 'exec telnet 172.17.0.1 2001' > ../u-boot-test-hooks/bin/console.laa

            cat > ../u-boot-test-hooks/bin/${HOSTNAME}/conf.xilinx_zynqmp_kria_laa-xilinx_zynqmp_kria << _EOF
            flash_impl=none
            power_impl=laa
            console_impl=laa
            reset_impl=laa
            _EOF

            cat > ../u-boot-test-hooks/py/${HOSTNAME}/u_boot_boardenv_xilinx_zynqmp_kria.py << __EOF
            env__net_dhcp_server = True
            env__spl_skipped = True
            env__dhcp_abort_test_skip = False
            env_spl_banner_times = 0
            env__net_static_env_vars = [
              ("ipaddr", "198.18.0.2"),
              ("netmask", "255.255.255.0"),
              ("serverip", "198.18.0.1"),
              ("tftpserverip", "198.18.0.1"),
            ]
            __EOF
          - |
            U_BOOT_TEST_HOOKS_PATH=/var/lib/lava-${LAVA_JOB_ID}/0/tests/0_u-boot-ci/src/u-boot-test-hooks
            export PATH=${U_BOOT_TEST_HOOKS_PATH}/bin:./tools/buildman:${PATH}
            export PYTHONPATH=${U_BOOT_TEST_HOOKS_PATH}/py/${HOSTNAME}/:${PYTHONPATH}

            ./test/py/test.py --tb=long \
              -s -v -ra \
              --bd xilinx_zynqmp_kria \
              --id laa-xilinx_zynqmp_kria \
              --build-dir ${KBUILD_OUTPUT} \
              --junitxml=results.xml --color=no || { echo "==== U-Boot CI failed ==="; exit 1; }

            xmllint --format results.xml --output formatted_results.xml
            cat formatted_results.xml
```

!!! note
    The inline test definition demonstrates how the `LAVA_*` environment variables
    are used to control and connect to the DUT from LAVA Docker test shell. When
    needed, you can convert these steps into a custom shell script with a result
    parser to store test results in LAVA.

# Protocols

A protocol in LAVA is a method of interacting with external services using the
supported APIs instead of with direct shell commands. The protocol defines which
APIs are available, and the job pipeline determines when the API call is made.

Note that not all protocols can be called from all actions and not all protocols
are able to share data between actions.

## MultiNode

The MultiNode protocol allows multiple devices to work together as a group,
synchronizing operations and sharing data between them.

### API

The MultiNode API provides helper scripts that are available inside the test
shell on the DUT. The APIs can be used from the LAVA test shell definitions to
pass messages between devices in a group.

See [MultiNode job definition](../../user/advanced-tutorials/multinode.md#coordinating-via-apis)
for job example.

#### lava-group

Print one line per device in the group:

```shell
lava-group
```

Output format: `<job_id>\t<role>`:

```plain
12345     client
12346     loadbalancer
12347     backend
12348     backend
```

Print job IDs for devices with the specified role, one per line:

```shell
lava-group <role>
```

Example output:

```plain
$ lava-group client
12345
$ lava-group backend
12347
12348
```

Exit with non-zero if the role doesn't exist:

```shell
$ lava-group server ; echo $?
1
```

#### lava-role

Print the role of the current device:

```shell
lava-role
```

Example output:

```plain
server
```

Print all roles in the MultiNode group, one role per line:

```shell
lava-role list
```

Example output:

```plain
server
client
```

#### lava-self

Report the job ID of the current device:

```shell
lava-self
```

#### lava-send

Sends a message to the group with optional key-value data. This is
non-blocking — the message is guaranteed to be available to all members, but
some may never retrieve it.

The message ID is persistent for the lifetime of the group. Re-sending a
different message with the same ID is not supported.

```shell
lava-send <message-id> [key1=val1 [key2=val2] ...]
```

!!! warning
    In the whitespace-separated `key=value` pairs, the key name must match the
    `\w+` pattern. This means only letters (`a–z, A–Z`), digits (`0–9`), and
    underscore (`_`) are allowed in the key name. In the values, quoted
    white spaces are allowed.

#### lava-sync

Global synchronization primitive. Sends a message, and waits for the same
message from all the other devices.

```shell
lava-sync <message>
```

!!! note
    All devices in the group must call `lava-sync` with the same message,
    otherwise it will block until timeout. `lava-sync foo` is effectively the
    same as `lava-send foo` followed by `lava-wait-all foo`.

#### lava-wait

Blocks until any device in the group sends a message with the given ID.

```shell
lava-wait <message-id>
```

Data from the message is written to `/tmp/lava_multi_node_cache.txt` as
`key=value` pairs.

!!! note
    The message ID data is persistent for the life of the MultiNode group. The
    data can be retrieved at any later stage using `lava-wait` and as the data
    is already available, there will be no waiting time for repeat calls. If
    devices continue to send data with the associated message ID, that new data
    will continue to be added to the stored data for that message ID and will be
    returned by subsequent calls to `lava-wait` for that message ID. Use
    different message ID(s) if you don’t want this effect.

#### lava-wait-all

`lava-wait-all` operates in different ways, depending on the presence of the
role parameter.

Waits until **all** devices in the group send a message with the given ID.
Every device must use `lava-send` with the same ID, or this call will block
until the timeout.

```shell
lava-wait-all <message-id>
```

Wait until all devices with the specified role send the message:

```shell
lava-wait-all <message-id> <role>
```

As with `lava-wait`, the message ID is persistent for the duration of the
MultiNode group, and the key-value pairs with the message will be stored in the
cache file.

### Request

This protocol allows actions within the pipeline to make requests using the
MultiNode API outside a test definition by wrapping the call inside the protocol.
This allows synchronization to happen during deploy or boot, not just within test
shells.

Actions check for protocol calls at the start of the run step. Only the named
action instance inside the pipeline will make the call. The use of the protocol
requests is an advanced use of LAVA and relies on the test job writer carefully
planning how the job will work.

See [MultiNode job definition](../../user/advanced-tutorials/multinode.md#coordinating-via-requests)
for job example.

#### lava-send

Send a message with optional key-value message to the group:

```yaml
protocols:
  lava-multinode:
  - action: prepare-scp-overlay
    request: lava-send
    messageID: ipv4
    message:
      ipaddr: '<value>'
    timeout:
      minutes: 5
```

#### lava-sync

Synchronize all devices in the group at a specific point in the pipeline:

```yaml
protocols:
  lava-multinode:
  - action: prepare-scp-overlay
    request: lava-sync
    messageID: start
    timeout:
      minutes: 5
```

#### lava-wait

Waits for any device in the group to send a message with the given ID. To
update the value available to the action, ensure the key exists in the matching
`lava-send` and prefix the value with `$` in the job submission:

```yaml
protocols:
  lava-multinode:
  - action: prepare-scp-overlay
    request: lava-wait
    messageID: ipv4
    message:
      ipaddr: $ipaddr
    timeout:
      minutes: 5
```

This makes the received data available to the action:

```yaml
{'message': {'ipaddr': '192.168.0.3'}, 'messageID': 'ipv4'}
```

#### lava-wait-all

Wait for all devices to send a message:

```yaml
protocols:
  lava-multinode:
  - action: prepare-scp-overlay
    request: lava-wait-all
    messageID: ipv4
    message:
      ipaddr: $ipaddr
    timeout:
      minutes: 5
```

Wait for all devices with a specific role to send a message:

```yaml
protocols:
  lava-multinode:
  - action: prepare-scp-overlay
    request: lava-wait-all
    role: server
    messageID: ipv4
    message:
        ipaddr: $ipaddr
    timeout:
      minutes: 5
```

# Docker device setup

LAVA can use docker as a DUT and run test under Docker.

## Create device-type

[Create the device type](common.md#create-device-type) using the name **`docker`**.

## Create device

1. [Add the device](common.md#add-device) using the following settings:
    * **Device Type:** `docker`
    * **Hostname:** A unique name (e.g., `docker-01`)
2. [Add the device configuration](common.md#add-device-configuration).

    For a standard docker device and a simple docker job, the following device
    dictionary should be sufficient:

    ```jinja
    {% extends "docker.jinja2" %}
    ```

## Submit a job

Submit this simple test job:

```yaml
--8<-- "jobs/docker.yaml"
```

--8<-- "refs.txt"

# Running the LAVA dispatcher in a Docker container

Running a LAVA dispatcher in a Docker container requires quite a few special
options, to ensure that the dispatcher has access to system resources that
are, in most Docker use cases, invisible to containers. This is automated by
the `lava-docker-worker` program, provided by the `lava-dispatcher-host`
package.

`lava-docker-worker` should support all use cases that the regular LAVA worker
supports, except running LXC containers. In particular, `lava-docker-work`
**does support** running other docker containers, QEMU/KVM, and other regular
device types, as long as the required user space tools are included in the LAVA
worker docker image.

## 1: install and configure lava-dispatcher-host

**lava-dispatcher-host needs to be installed in the host machine** (bare metal)
for the LAVA dispatcher to work inside a container, as well as Docker itself.

### Debian package

If you are using the Debian package, just installing it will do all necessary
setup.

### Other means

If you are installing lava-dispatcher-host in some other way, you need to run
the following command, as root:

```
# lava-dispatcher-host rules install
```

This step needs to be repeated when lava-dispatcher-host is upgraded, to make
sure you have the most up to date udev rules file for it.

## 2: start lava-docker-worker

### Debian package

To run the Docker Dispatcher, first edit
`/etc/lava-dispatcher-host/lava-docker-worker` and set the connection
parameters. At least URL and TOKEN are probably needed:

```
WORKER_NAME="--name myworker"

# ...

# Server connection
URL="https://my.lava.server/"
TOKEN="--token 0123456789012345678901234567890123456789"
```

Then enable and start the `lava-docker-worker` service:

```
sudo systemctl enable lava-docker-worker.service
sudo systemctl start lava-docker-worker.service
```

### Other means

If you installed LAVA in a way that doesn't provide a lava-docker-worker
systemd service, you can just run `lava-docker-worker` as any user that can run
docker. You need to pass the relevant arguments to have it connect to your LAVA
server:

```
$ lava-docker-worker --url https://my.lavaserver.com/ --token 0123456789012345678901234567890123456789
```

# Test definitions

The test definition is a `yaml` file that describe the tests that you want LAVA
to run on the DUT.

## Smoke example

Let's look at this example:

```yaml
--8<-- "tests/smoke.yaml"
```

## Metadata

The `metadata` dictionary describes the test. `format` and `name` are mandatory
while `description` is optional.

## Run

The dictionary list the actions that LAVA should execute on the DUT. The only
available key is `steps`.

`steps` is an array that will be transformed into a shell script that LAVA will
run on the DUT.

In this example, `run.steps` will be rendered into the following shell script:

```shell
--8<-- "tests/smoke.sh"
```

### LAVA test helpers

The LAVA Test Helpers are scripts maintained in the LAVA codebase, like
`lava-test-case`. These are designed to work using only the barest
minimum of operating system support, to make them portable to all deployments.
The helpers are mainly used to:

* embed information from LAVA into the test shell
* support communication with LAVA during test runs

#### lava-test-case

##### Result

Record a test result using the `--result` argument:

```shell
lava-test-case <test-case-id> --result <pass|fail|skip|unknown> [--measurement <val>] [--units <unit>] [--output <file>]
```

##### Shell

Record a test result using the `--shell` command exit code:

```shell
lava-test-case <test-case-id> --shell <command>
```

If `<command>` exits with 0, the result is `pass`. Otherwise, it is `fail`.

!!! warning
    When using `--shell`, multiple commands or combined on-liner command with pipes
    and redirects should be quoted so it is passed as a single argument and the final
    exit code is checked.
    ```yaml title="Correct"
    - lava-test-case check-os --shell "cat /etc/os-release | grep 'ID=debian'"
    ```
    If not quoted, only the exit code of the first command is checked and a
    `| grep <pattern>` check could potentially prevent the helper from sending
    the LAVA test result signal and lead to missing test result.
    ```yaml title="Incorrect"
    - lava-test-case check-os --shell cat /etc/os-release | grep "ID=debian"
    ```

##### Output

Attach output from a file to a test result using `--output`:

```shell
lava-test-case <test-case-id> --result <result> --output <file>
```

This streams the contents of `<file>` between start/end test case signals,
similar to how `--shell` captures command output. Useful when the test output
has already been saved to a file by a previous step.

#### lava-test-set

Group test cases into named sets. This allows test writers to subdivide
results within a single test definition using an arbitrary label:

```shell
lava-test-set start <name>
lava-test-set stop
```

```yaml
steps:
  - lava-test-set start network-tests
  - lava-test-case ping --shell ping -c4 localhost
  - lava-test-set stop
```

#### lava-test-reference

Some test cases may relate to specific bug reports or have specific URLs
associated with the result. This help can be used to associate a URL with a
test result:

```shell
lava-test-reference <test-case-id> --result <result> --reference <url>
```

!!! note
    The URL should be a simple file reference, complex query strings could
    fail to be parsed.

#### lava-test-raise

Raise a `TestError` to abort the current test job immediately:

```shell
lava-test-raise <message>
```

This is useful when a setup step fails and running subsequent tests would be
pointless. The message is included in the error reported by LAVA.

#### lava-background-process-

Manage background processes during the test:

```yaml
steps:
  - lava-background-process-start MON --cmd "top -b"
  - ./run-my-tests.sh
  - lava-background-process-stop MON
```

### Device Information Helpers

Some elements of the static device configuration are exposed to the test shell,
where it is safe to do so and where the admin has explicitly configured the
information.

#### lava-target-ip

Prints the target's IP address if configured by the admin in device dictionary
using [device_ip](../../technical-references/configuration/device-dictionary.md#device_ip).
Devices with a fixed IPv4 address configured in the device dictionary will populate
this field. Test writers can use this in an Docker container to connect to the device:

```shell
ping -c4 $(lava-target-ip)
```

#### lava-target-mac

Prints the target's MAC address if configured by the admin in device dictionary
using [device_mac](../../technical-references/configuration/device-dictionary.md#device_mac).
This can be useful to look up the IP address of the device:

```shell
echo $(lava-target-mac)
```

#### lava-echo-ipv4

Prints the IPv4 address for a given network interface using `ifconfig` or `ip`:

```shell
lava-echo-ipv4 <interface>
```

#### lava-target-storage

Prints available storage devices if configured by the admin in device dictionary
using [storage_info](../../technical-references/configuration/device-dictionary.md#storage_info).

Without arguments, outputs one line per device with the name and value separated
by a tab:

```shell
$ lava-target-storage
UMS	/dev/disk/by-id/usb-Linux_UMS_disk_0_WaRP7-0xac2400d300000054-0:0
SATA	/dev/disk/by-id/ata-ST500DM002-1BD142_W3T79GCW
```

With a name filter, outputs only the matching device value:

```shell
$ lava-target-storage UMS
/dev/disk/by-id/usb-Linux_UMS_disk_0_WaRP7-0xac2400d300000054-0:0
```

If there is no matching name, exits non-zero and outputs nothing.

### Using custom scripts

When multiple steps are necessary to run a test and get usable output, write a
custom script to go alongside the YAML and execute it as a run step:

```yaml
run:
  steps:
    - ./my-script.sh arguments
```

You can choose whatever scripting language you prefer, as long as it is
available in the test image. The best practices are:

* Be verbose - Add progress messages, error handling, and debug output so
  that test logs are useful during triage. Control the total amount of output
  to keep logs readable.
* Watch out for `cd` - If you change directories inside a script, save the
original working directory first and return to it at the end.
* Wait for subprocesses - In Python, use `subprocess.check_call()` or
`subprocess.run()` instead of `subprocess.Popen()` when calling LAVA helpers.
`Popen` returns immediately, which can cause output to arrive after the test
definition finishes and lead to missing results.
* Make it portable - Scripts should check if the LAVA helper in `$PATH` to
determine whether they are running inside LAVA. When the helper is not available,
report results with `echo` or `print()` instead. This allows the same script to
run both inside and outside of LAVA, helping developers reproduce issues without
the full CI system.

    ```shell
    #!/bin/sh
    if command -v lava-test-case >/dev/null 2>&1; then
        lava-test-case my-test --result pass
    else
        echo "my-test: pass"
    fi
    ```

See also [test writing guidelines](https://github.com/Linaro/test-definitions/blob/master/docs/test-writing-guidelines.md#test-writing-guidelines), typically the **Running in
LAVA** section. For test definition examples that follow the best practices,
see [automated](https://github.com/Linaro/test-definitions/tree/master/automated).

## Install

Before running the run steps, LAVA can also `install` some Git repositories and
run arbitrary shell commands to prepare for the following test runs.

### git-repos

Specifies the Git repositories to be cloned into the test working directory.
The repositories are cloned on the LAVA worker and applied to the test image as
part of the LAVA overlay.

```yaml
install:
  git-repos:
    - url: https://gitlab.com/lava/lava.git
      branch: 2026.01
      destination: lava202601
```

### steps

Specifies arbitrary shell commands to run during the install phase. The install
steps are run directly on the DUT.

```yaml
install:
  git-repos:
    - url: https://example.com/repo.git
  steps:
    - cd repo
    - make install
```

## Expected

The `expected` dictionary allows users to define a list of expected test cases.
Test case present in the results but not in the expected list is forcibly reported
as `fail`. At the end of each test run, missing expected test cases from the test
results are reported as `fail` as well.

With the following test definition example, tc5 and tc6 will be reported as
`fail` because they are unexpected. Likewise, tc3 and tc4 will be reported as
`fail` because they are missing.

```yaml title="Test definition"
metadata:
  format: Lava-Test Test Definition 1.0
  name: expected-testdef-example
run:
  steps:
  - lava-test-case tc1 --result pass
  - lava-test-case tc2 --result fail
  - lava-test-case tc5 --result pass
  - lava-test-case tc6 --result fail
expected:
  - tc1
  - tc2
  - tc3
  - tc4
```

The list can be defined in either the test definition or the job definition. If
both are provided, the value in the job definition takes precedence.

!!!warning "Limitation"
    Expected test cases defined in test definition from `git` are checked and
    reported after test execution. If the test definition not get executed at
    all, they are not reported. The limitation exists because these test
    definitions and the expected test case lists defined inside may be
    unreachable or not deployed yet when job exists on error. If you need
    consistent and predictable job results for the test suites and cases, you
    should provide the expected test cases in the
    [job definition](../../technical-references/job-definition/actions/test.md#expected).

--8<-- "refs.txt"

# Job definition

The [job definition](../../technical-references/job-definition/job.md) is a `yaml` file that you submit to the LAVA server.
This file is describing everything that LAVA needs to know to run your tests on
a DUT.

## QEMU example

Let's look at this example:

```yaml
--8<-- "jobs/qemu.yaml"
```

## Structure

The job definition is made of:

* `device_type`: requested device-type
* `job_name`: name of the job, sometime called `description`
* `timeouts`: [timeout definition](../../technical-references/job-definition/timeouts.md)
* `priority`: [job priority](../../technical-references/job-definition/job.md#priority)
* `visibility`: [job visibility](../../technical-references/job-definition/job.md#visibility)
* `actions`: list of actions to run

In this definition, we request to run a job called `simple qemu tjob` on a `qemu` DUT.
The job is visible to everyone (including anonymous users).
The job will not run for more than 10 minutes at `medium` priority.

## Actions

`actions` is a list of actions that LAVA will have to execute for the given
job.

Currently, LAVA support four type of actions:

* [command](../../technical-references/job-definition/actions/command.md): run commands on the dispatcher
* [deploy](../../technical-references/job-definition/actions/deploy.md): deploy software on the DUT
* [boot](../../technical-references/job-definition/actions/boot.md): boot the DUT
* [test](../../technical-references/job-definition/actions/test.md): run some tests

### Deploy

The deploy action will deploy the software on the DUT, using the method
specified by the `to`.

In this example, we are requesting to deploy to `tmpfs`. The `rootfs` will be
downloaded from the given url.

!!! info "deployment methods"
    LAVA supports many deployment methods. However, only some methods are
    supported by each device-type.

### Boot

The boot action will boot the DUT, using the method specified by the `method`.

In this example, we are requesting to boot `qemu` using the software from
`tmpfs`.

!!! info "boot methods"
    LAVA supports many boot methods. However, only some methods are supported
    by each device-type.

### Test

The test action will run the given tests on the DUT.

In this example, LAVA will download the [test definitions](./test-definition.md) from two git repositories.

!!! info "test methods"
    LAVA supports three test methods that do not depend on the device-type.

--8<-- "refs.txt"

# Topology

LAVA's [distributed architecture](../../technical-references/architecture.md)
enables horizontal scaling and flexible deployment across geographically
distributed and network-restricted environments.

A LAVA server can be deployed in an isolated lab or in the cloud. LAVA workers
can be deployed from any location with network connectivity to the LAVA server.

## Local LAVA deployment

LAVA server and workers can be deployed on local machines. Several workers can be
connected to a LAVA server. Multiple DUTs can be connected to a worker.

```mermaid
graph TD
    server([LAVA Server])
    worker1([Worker1])
    dut1([DUT1])
    dut2([DUT2])
    worker2([Worker2])
    dut3([DUT3])
    dut4([DUT4])

    server <--> worker1
    server <--> worker2

    worker1 <--> dut1
    worker1 <--> dut2

    worker2 <--> dut3
    worker2 <--> dut4
```

## Cloud-hosted Services

LAVA server can be hosted in the cloud using
[docker-composer](https://gitlab.com/lava/pkg/docker-compose) or K8s. Some
companies provide managed LAVA instances.

For example, Linaro offers both cloud-hosted LAVA instances and
[Linaro Automation Appliance](https://docs.lavacloud.io/) (LAA) that managed by
the LAVA Managed Service (LMS) at [lavacloud.io](https://lavacloud.io/). This
integrated solution significantly simplifies and accelerates LAVA deployment and
device enablement in LAVA.

While the LAVA server and LMS are cloud-hosted, the LAA can be deployed from any
location. LMS provides seamless LAVA integration, secure remote access and
Over-The-Air (OTA) upgrades for your LAAs.

```mermaid
graph TD
    server([cloud-hosted LAVA Server])
    lms([cloud-hosted LMS])
    laa1([LAA1 + DUT1])
    laa2([LAA2 + DUT2])

    server <--> laa1
    server <--> laa2
    laa1 <-.-> lms
    server <-.-> lms
    laa2 <-.-> lms
```

Each LAA serves as a pre-configured LAVA worker. It provides everything needed
by most hardware automation tasks. The DUT is attached to the LAA via the
[MIB](https://docs.lavacloud.io/hardware/mibs.html) directly. Simply plug
in the LAA power and Ethernet cables, then it will automatically connect to the
LMS. You can connect it to your LAVA server and add device configurations with
just a few clicks on the LMS web UI.

See also [LAVA worker](./hardware.md#lava-worker).

--8<-- "refs.txt"

# Install LAVA

Two installation methods are currently supported:

* [docker](#docker)
* [debian](#debian)

We advise to use docker image for a first test of LAVA.

For a production instance, both methods would be suitable.

## Docker

In order to install LAVA using the official docker images, we advise to use the
provided **docker-compose** file.

### Dependencies

Install the dependencies: `docker` and `docker-compose`:

```shell
sudo apt update
sudo apt install docker-compose
```

The ports needed by the lava services must be available on the host. You can
check if they are occupied with the below command. Stop the programs that using
the port if any.

```shell
sudo ss -tulpn | grep -E ':(69|80|111|2049|3079|5500|7101|8001|35543)\s'
```

### Install

1. Get the **docker-compose** files from [gitlab][lava-docker-compose].

    ```shell
    git clone https://gitlab.com/lava/pkg/docker-compose
    cd docker-compose/
    ```

2. Configure the `.env` file to set LAVA admin username and password. You will
need them for using and managing your LAVA instance.

    ```shell
    DC_LAVA_ADMIN_USERNAME=<your_user_name>
    DC_LAVA_ADMIN_PASSWORD=<your_user_password>
    ```

3. Start the services

    ```shell
    docker-compose pull
    docker-compose build
    docker-compose up
    ```

    !!! tip
        If `/dev/kvm` is unavailable on your host, comment out the `- /dev/kvm`
        line in the `docker-compose.yaml`.

The newly created instance is now available at [localhost].

## Debian

In order to install LAVA using the Debian packages, we advice to use the
repositories that we manage to get the latest version.

!!! note "Supported Debian versions"
    LAVA is only supported on Debian *Bullseye* and *Bookworm*.

### Dependencies

In order to follow this tutorial, you would have to install some dependencies:

```shell
apt install ca-certificates gnupg2 wget
```

### Repository

Add the lavasoftware debian repository:

=== "Bullseye"
    ```shell
    wget https://apt.lavasoftware.org/lavasoftware.key.asc
    apt-key add lavasoftware.key.asc
    echo "deb https://apt.lavasoftware.org/release bullseye main" > /etc/apt/sources.list.d/lava.list
    echo "deb http://deb.debian.org/debian bullseye-backports main" > /etc/apt/sources.list.d/backports.list
    ```

=== "Bookworm"
    ```shell
    wget https://apt.lavasoftware.org/lavasoftware.key.asc
    apt-key add lavasoftware.key.asc
    echo "deb https://apt.lavasoftware.org/release bookworm main" > /etc/apt/sources.list.d/lava.list
    ```

### Install

Install **postgresql** and **lava** debian packages:

=== "Bullseye"
    ```shell
    apt update
    apt install postgresql
    pg_ctlcluster 13 main start
    apt install lava-server
    apt install -t bullseye-backports lava-dispatcher
    ```

=== "Bookworm"
    ```shell
    apt update
    apt install postgresql
    pg_ctlcluster 15 main start
    apt install lava-server lava-dispatcher
    ```

### Starting

Enable default LAVA server Apache configuration:

```shell
a2dissite 000-default
a2enmod proxy
a2enmod proxy_http
a2ensite lava-server.conf
systemctl restart apache2
```

You can start the different services:

```shell
systemctl start apache2
systemctl start postgresql
systemctl start lava-server-gunicorn
systemctl start lava-publisher
systemctl start lava-scheduler
systemctl start lava-worker
```

The newly created instance is now available at [localhost].

--8<-- "refs.txt"

# Mailing-lists

The best way to get in touch with the community and the LAVA developers is via
[mailing-lists](https://lists.lavasoftware.org).

## [lava-announce](mailto:lava-announce@lists.lavasoftware.org)

Subscribing to the **lava-announce** list is recommended for everyone using
LAVA, whether writing tests or viewing reports or administering a LAVA instance.

The release notes for each production release are sent to the **lava-announce**
mailing list.

Replies to this list are sent to the **lava-users** list.

## [lava-devel](mailto:lava-devel@lists.lavasoftware.org)

Subscribing to the **lava-devel** list is recommended for developers of LAVA.

**lava-devel** is aimed at supporting code contributors, device integration
engineers and instance admins who are working with the LAVA codebase.

Discussions about planning and new LAVA features also take place here.

## [lava-users](mailto:lava-users@lists.lavasoftware.org)

The **lava-users** mailing list concentrates on support for setting up and using
LAVA.

Subscribers include test writers, and individual admins. Users are encouraged to
contribute to answer queries from other users. Replies to the **lava-announce**
list are directed here.

## Discord

Our team is distributed globally, with members in Europe, North America, and
Asia. Join the LAVA Software Community Project on Discord for real-time
discussions and support.

Use this [invitation link](https://discord.com/invite/aUjpUFBvD2) to join our
server.

# Dispatcher configuration

The dispatcher configuration files allows you to configure some of the beavior
of the dispatchers.

Available files are:

* dispatcher configuration
* environment
* DUT environment

## Dispatcher configuration

Environment configuration file is located on the server in:

* `/etc/lava-server/<hostname>.yaml` old configuration file
* `/etc/lava-server/dispatcher.d/<hostname>/dispatcher.yaml` new configuration file

When loading the configuration, LAVA will look at the new configuration file
and fallback to the old one

```yaml
--8<-- "config/dispatcher.yaml"
```

## Environment

The dispatcher environment is used to set the process environment when spawning
`lava-run`.

Environment configuration file is located on the server in:

* `/etc/lava-server/env.yaml` default, for every dispatchers
* `/etc/lava-server/dispatcher.d/<hostname>/env.yaml` for a specific dispatcher

When loading the configuration, LAVA will look at the dispatcher specific
configuration and fallback to the default configuration.

```yaml
--8<-- "config/env.yaml"
```

## DUT environment

The DUT environment is used to set some environment variable on the DUT when
running tests.

Environment configuration file is located on the server in:

* `/etc/lava-server/env-dut.yaml` default, for every dispatchers
* `/etc/lava-server/dispatcher.d/<hostname>/env-dut.yaml` for a specific dispatcher

When loading the configuration, LAVA will look at the dispatcher specific
configuration and fallback to the default configuration.

The format is the same as for the environment file.

--8<-- "refs.txt"

# FVP

Allows running of FVP (or Fixed Virtual Platforms) from a Docker container.
Generally speaking, FVP are launched in Docker and UART output is served over a
telnet connection locally. A pattern is given in the job definition to find the
port of the UART from the FVP output. LAVA will then connect via `telnet` to
view UART output.

```yaml
- deploy:
    to: fvp
    images:
      disk:
        url: http://fileserver/path/to/fvp/grub-busybox.img
        format: ext4
        overlays:
          lava: true
```

## images

See [images](./index.md#artifacts)

### overlays

See [overlays](./index.md#overlays)

# Deploy action

The deploy actions are deploying the provided software on the DUT using the
method specified in the job definition.

## Artifacts

In the deploy action, the following parameters are available for artifacts
downloads:

* archive
* checksums
* compression
* headers
* url

### Archive

When specifying `archive: tar`, LAVA will extract the tar archive prior to
using it.

```yaml
- deploy:
    images:
      boot:
        url: http://example.com/boot.tar.xz
        compression: xz
        archive: tar
```

### Checksums

LAVA is able to compute the checksums of the download
artifacts using `md5sum`, `sha256sum` and `sha512sum`:

```yaml
- deploy:
    tmpfs:
      rootfs:
        url: http://example.com/rootfs.img.xz
        compression: xz
        md5sum: d8784b27867b3dcad90cbea66eacc264
```

!!! info "Multiple checksums"
    If needed, you can provide multiple checksum algorithms for the same
    artifact.

### Compression

If needed, LAVA can uncompress a compressed artifact by specifying
`compression`.

```yaml
- deploy:
    tmpfs:
      rootfs:
        url: http://example.com/rootfs.img.xz
        compression: xz
```

The supported formats are: `bz2`, `gz`, `xz`, `zip` or `zstd`.

### Headers

For http(s) artifacts, you can provide additional headers:

```yaml
- deploy:
    tmpfs:
      rootfs:
        url: http://example.com/rootfs.img.xz
        compression: xz
        headers:
          my-header1: value
```

### URL

Specifies the URL to download.

URLs **must** use one of the supported protocols:

* `http://`
* `https://`
* `file://`
* `scp://`
* `downloads://`

URLs are checked during the test job validation to ensure that the file can be
downloaded. Missing files will cause the test job to end as `Incomplete`.

URLs allow placeholders for all supported protocols.

```yaml
- deploy:
    to: tftp
    kernel:
      url: http://{FILE_SERVER_IP}/linux/Image-imx8mmevk.bin
      type: image
    persistent_nfs:
      address: "{FILE_SERVER_IP}:/var/lib/lava/dispatcher/tmp/linux/imx8mm_rootfs"
    dtb:
      url: http://{FILE_SERVER_IP}/linux/imx8mm-evk.dtb
    os: debian
```

!!! note
    Admin can define any placeholder and assign an address to it in device
    dictionary. LAVA then substitutes the placeholders in job with the
    `static_info` to generate a new `url`.

```jinja
{% set static_info = [{'FILE_SERVER_IP': "10.192.244.104"}] %}
```

## Overlays

LAVA can apply a set of overlays to every artifact. The configuration should
look like:

```yaml
- deploy:
    tmpfs:
      rootfs:
        url: http://example.com/rootfs.img.xz
        compression: xz
        format: ext4
        overlays:
          lava: true
          kselftest:
            url: https://exampl.com/kselftes.tar.xz
            compression: xz
            format: tar
            path: /
```

You should provide:

* `format`: the format of the artifact to update (`cpio.newc`, `ext4` or `tar`)
* `overlays`: a dictionary of overlays to insert

You can also provide:

* `partition`: to update a given partition (for `ext4` with multiple partitions)
* `sparse`: set to `true` if the artifact is a sparse image

### LAVA overlay

In order to insert the LAVA overlay (that include the test definitions and
helpers), use `lava: true` as overlay.

```yaml
- deploy:
    to: usbg-ms
    image:
      url: https://raspi.debian.net/tested/20231109_raspi_4_bookworm.img.xz
      compression: xz
      format: ext4
      partition: 1
      overlays:
        lava: true
```

### Overlays

You can insert a tar archive or a file in the artifact. You should provide:

* `format`: the format of the overlay (`file` or `tar`).
* `path`: the path in the artifact

# Deploy action

The deploy actions are deploying the provided software on the DUT using the
method specified in the job definition.

## Artifacts

In the deploy action, the following parameters are available for artifacts
downloads:

* archive
* checksums
* compression
* headers
* url

### Archive

When specifying `archive: tar`, LAVA will extract the tar archive prior to
using it.

```yaml
- deploy:
    images:
      boot:
        url: http://example.com/boot.tar.xz
        compression: xz
        archive: tar
```

### Checksums

LAVA is able to compute the checksums of the download
artifacts using `md5sum`, `sha256sum` and `sha512sum`:

```yaml
- deploy:
    tmpfs:
      rootfs:
        url: http://example.com/rootfs.img.xz
        compression: xz
        md5sum: d8784b27867b3dcad90cbea66eacc264
```

!!! info "Multiple checksums"
    If needed, you can provide multiple checksum algorithms for the same
    artifact.

### Compression

If needed, LAVA can uncompress a compressed artifact by specifying
`compression`.

```yaml
- deploy:
    tmpfs:
      rootfs:
        url: http://example.com/rootfs.img.xz
        compression: xz
```

The supported formats are: `bz2`, `gz`, `xz`, `zip` or `zstd`.

### Headers

For http(s) artifacts, you can provide additional headers:

```yaml
- deploy:
    tmpfs:
      rootfs:
        url: http://example.com/rootfs.img.xz
        compression: xz
        headers:
          my-header1: value
```

### URL

Specifies the URL to download.

URLs **must** use one of the supported protocols:

* `http://`
* `https://`
* `file://`
* `scp://`
* `downloads://`

URLs are checked during the test job validation to ensure that the file can be
downloaded. Missing files will cause the test job to end as `Incomplete`.

URLs allow placeholders for all supported protocols.

```yaml
- deploy:
    to: tftp
    kernel:
      url: http://{FILE_SERVER_IP}/linux/Image-imx8mmevk.bin
      type: image
    persistent_nfs:
      address: "{FILE_SERVER_IP}:/var/lib/lava/dispatcher/tmp/linux/imx8mm_rootfs"
    dtb:
      url: http://{FILE_SERVER_IP}/linux/imx8mm-evk.dtb
    os: debian
```

!!! note
    Admin can define any placeholder and assign an address to it in device
    dictionary. LAVA then substitutes the placeholders in job with the
    `static_info` to generate a new `url`.

```jinja
{% set static_info = [{'FILE_SERVER_IP': "10.192.244.104"}] %}
```

## Overlays

LAVA can apply a set of overlays to every artifact. The configuration should
look like:

```yaml
- deploy:
    tmpfs:
      rootfs:
        url: http://example.com/rootfs.img.xz
        compression: xz
        format: ext4
        overlays:
          lava: true
          kselftest:
            url: https://exampl.com/kselftes.tar.xz
            compression: xz
            format: tar
            path: /
```

You should provide:

* `format`: the format of the artifact to update (`cpio.newc`, `ext4` or `tar`)
* `overlays`: a dictionary of overlays to insert

You can also provide:

* `partition`: to update a given partition (for `ext4` with multiple partitions)
* `sparse`: set to `true` if the artifact is a sparse image
* `overlay_backend`: which tool to use to modify `ext4` images: `auto`
  (the default), `e2fsprogs` or `guestfs`. With `auto`, LAVA uses `e2fsprogs`
  (`debugfs`) when it is available on the worker and falls back to
  `libguestfs`. Set this to override that choice, for example to force
  `guestfs` if an image does not work with the `e2fsprogs` backend.

`overlay_backend` can also be set directly under `deploy:` to apply to every
image in the deployment; a per-image value takes precedence over the
deploy-level one.

### LAVA overlay

In order to insert the LAVA overlay (that include the test definitions and
helpers), use `lava: true` as overlay.

```yaml
- deploy:
    to: usbg-ms
    image:
      url: https://raspi.debian.net/tested/20231109_raspi_4_bookworm.img.xz
      compression: xz
      format: ext4
      partition: 1
      overlays:
        lava: true
```

### Overlays

You can insert a tar archive or a file in the artifact. You should provide:

* `format`: the format of the overlay (`file` or `tar`).
* `path`: the path in the artifact

# LAVA coordinator

This service is receiving, storing and dispatching multinode messages.

Only one instance of this process should be running for a given lava instance. Usually, this instance is running on the server but this is not required.

## Command line

Run `/usr/bin/lava-coordinator`

## Service

The systemd service is called `lava-coordinator`.

## Dependencies

The process should be able to:

* open a socket on port `3079`

The process should be visible to every dispatchers.

## Configuration

Daemon start options:

* `/etc/default/lava-coordinator`

In order to connect to the coordinator, configuration file
`/etc/lava-coordinator/lava-coordinator.conf` should be copied on each
dispatcher.

## Logs

The logs are stored in `/var/log/lava-coordinator.log`

The log rotation is configured in `/etc/logrotate.d/lava-coordinator-log`.

## Security

It's currently advised to restrict connections to the private network.

# Job results

A test result in a LAVA job is called a test case that include a result. Each
test case can optionally include a measurement and a units string. Test cases
are aggregated into a test suite. Multiple test suites can be generated for each
test job.

## Test suite

The name of a test suite is defined in the test job definition. LAVA creates
one test suite for each [test](./job-definition/actions/test.md) definition,
using its `name` field. For the
[`definitions`](./job-definition/actions/test.md#definitions) action, LAVA
prepends a zero-based sequence number to the `name` to ensure test suite is
unique within the job.

Every test job also generates a reserved `lava` test suite that contains results
produced by the job actions.

## Test set

Test Set is optional and allows test writers to subdivide individual results
within a single Lava test definition using an arbitrary label. This is useful
when the same test is run multiple times with different parameters.

See [lava-test-set](../user/basic-tutorials/test-definition.md#lava-test-set) for
recording a test set.

## Test case

Each test case has a `name` and a `result`. Optionally it can have a
`measurement` and `units`.

See [lava-test-case](../user/basic-tutorials/test-definition.md#lava-test-case)
for recording a test result.

!!! note
    - Whitespace is not allowed in test case name.
    - Supported result types are: `pass`, `fail`, `skip` or `unknown`.

## Retrieving results

### API

See `/api/v0.2/jobs/` for the endpoints that can be used to retrieve LAVA job
results through the REST API.

### URL

LAVA job results also can be retrieved via the following simple URLs:

| URL | Description |
| --- | ----------- |
| `/results/<job_id>/csv` | All job results as CSV |
| `/results/<job_id>/yaml` | All job results as YAML |
| `/results/<job_id>/yaml_summary` | All test suite names as YAML |
| `/results/<job_id>/<suite_name>/csv` | Results for a specific test suite as CSV |
| `/results/<job_id>/<suite_name>/yaml` | Results for a specific test suite as YAML |
| `/results/testcase/<testcase_id>/yaml` | Result for a specific test case as YAML |

# Depthcharge

The `depthcharge` boot method is used to boot a device using
[Depthcharge](https://chromium.googlesource.com/chromiumos/platform/depthcharge/),
the bootloader used by ChromeOS devices. LAVA waits for the Depthcharge command
line interface, and then sends the configured boot commands.

```yaml
- boot:
    method: depthcharge
    commands: nfs
    prompts:
    - 'root@debian:~#'
    timeout:
      minutes: 5
```

See [Depthcharge](../../../configuration/device-dictionary.md#depthcharge) for
supported device configurations.

!!! note
    A prior `tftp` deploy action is required.

## commands

See [commands](./common.md#commands).

The boot method provides the following additional placeholders that can be used
in the boot commands defined in the device type or job definition.

| Placeholder | Description |
| --- | --- |
| `{DEPTHCHARGE_KERNEL}` | TFTP path to the FIT image (if available) or kernel image |
| `{CMDLINE}` | TFTP path to the generated `cmdline` file |
| `{DEPTHCHARGE_RAMDISK}` | TFTP path to the ramdisk (empty when using a FIT image) |

## extra_kernel_args

Extra kernel command line arguments to append to the kernel `cmdline` defined
in the device type.

```yaml
- boot:
    method: depthcharge
    commands: nfs
    extra_kernel_args: "debug loglevel=7"
```

## Example job

```yaml
device_type: acer-cbv514-1h-34uz-brya
job_name: depthcharge NFS boot example

timeouts:
  job:
    minutes: 30
  connection:
    minutes: 2

priority: medium
visibility: public

actions:
- deploy:
    to: tftp
    kernel:
      url: https://example.com/bzImage
    modules:
      url: https://example.com/modules.tar.xz
      compression: xz
    ramdisk:
      url: https://example.com/bullseye-rootfs-amd64-initramfs.gz
      compression: gz
    nfsrootfs:
      url: https://example.com/bullseye-rootfs-amd64.tar.gz
      compression: gz
    timeout:
      minutes: 10

- boot:
    method: depthcharge
    commands: nfs
    auto_login:
        login_prompt: 'login:'
        username: user
        password_prompt: 'Password:'
        password: user
        login_commands:
          - sudo su
    prompts:
      - 'root@health'
      - 'user@health'
    timeout:
      minutes: 5

- test:
    definitions:
    - from: git
      repository: https://gitlab.com/lava/functional-tests.git
      path: posix/smoke-tests-basic.yaml
      name: smoke-tests
    timeout:
      minutes: 5
```

# Building LAVA

LAVA can be distributed either as [Debian packages](#debian-packages) or [Docker images](#docker-images).

Both formats can be built locally from sources.

!!! tip "Version"
    The version will be computed based on `git describe`.
    Run `./lava_common/version.py` to print this version.

## Debian packages

Run the GitLab CI script to build the Debian packages:

```shell
.gitlab-ci/build/debian/10.sh
```

The packages will be available under `_build/`.

!!! warning "Uncommitted changes"
    The script will refuse to build the packages is you have any uncommitted changes.

## Docker images

Run the GitLab CI script to build either the `lava-dispatcher` or `lava-server` image:

```shell
.gitlab-ci/build/docker.sh dispatcher
.gitlab-ci/build/docker.sh server
```

Images will be tagged with:

```
registry.gitlab.com/lava/lava/amd64/lava-dispatcher:<VERSION>
registry.gitlab.com/lava/lava/amd64/lava-server:<VERSION>
```

!!! tip "Base images"
    The script will create base images for both `lava-dispatcher` and `lava-server`.
    Bases images will be tagged with
    ```
    registry.gitlab.com/lava/lava/amd64/lava-dispatcher-base:<VERSION>
    registry.gitlab.com/lava/lava/amd64/lava-server-base:<VERSION>
    ```

# Boot Action

The `boot` action is used to boot the device using the deployed files.
Depending on the parameters in the job definition, this could be by executing a
command on the dispatcher (e.g., `qemu-system-x86_64`) or by connecting to the
device over serial or SSH. Depending on the power state of the device and the
device configuration, the device may be powered up or reset to provoke the
boot.

Every `boot` action **must** specify a [method](#method) which determines how
to boot the deployed files on the device. Depending on the method, other
parameters may be required.

Boot actions which result in a POSIX-type login or shell must specify a list of
expected [prompts](#prompts) which will be matched against the output to
determine the endpoint of the boot process. There are no default prompts. The
job definition writer is responsible for providing a list of all possible
prompts.

Boot is a top level action that is part of the `actions` list. Here is a common
example of boot action from the test job definition:

```yaml
actions:
- boot:
    timeout:
      minutes: 15
    method: u-boot
    commands: nfs
    auto_login:
      login_prompt: 'login:'
      username: root
      password_prompt: "Password:"
      password: "secret"
    prompts:
    - 'root@device:~#'
```

## timeout

See [Action timeout](../../timeouts.md#action-timeout)

## method

The boot `method` determines how the device is booted and which commands and
prompts are used to determine a successful boot. Each method is documented
separately.

## commands

For the following bootloader-based boot methods, one of the key definitions in
the device type template is the list of boot `commands` that the device needs.
These are sets of specific commands that will be run to boot the device.

- `bootloader`
- `depthcharge`
- `ipxe`
- `grub`
- `grub-efi`
- `u-boot`

The `commands` parameter can be specified in two ways:

1. **A predefined command set name** — predefined in the device type template,
such as `nfs`, `ramdisk` and `usb`. See your device dictionary for the full
list.
2. **[A list of custom boot commands](#custom-boot-commands)** — specified
directly in the job definition

### Placeholder substitution

The commands typically include placeholders. At runtime, LAVA substitutes them
with dynamic data. For example, this is the raw predefined U-Boot commands for
booting from `nfs`:

```yaml
commands:
- "usb start"
- "setenv autoload no"
- "setenv initrd_high 0xffffffff"
- "setenv fdt_high 0xffffffff"
- "dhcp"
- "setenv serverip {SERVER_IP}"
- "tftp {KERNEL_ADDR} {KERNEL}"
- "tftp {RAMDISK_ADDR} {RAMDISK}"
- "tftp {TEE_ADDR} {TEE}"
- "setenv initrd_size ${filesize}"
- "tftp {DTB_ADDR} {DTB}"
- "setenv bootargs ' root=/dev/nfs rw nfsroot={NFS_SERVER_IP}:{NFSROOTFS},tcp,hard earlycon=uart8250,mmio32,0x3f215040 ip=dhcp'"
- "{BOOTX}"
```

!!! note
    In some cases, the boot commands list in the template may not provide
    **all** of the commands used; lines will also be generated from other data
    in the template. For example, the command `setenv kernel_addr_r '0x82000000'`
    might be generated from load addresses which match the type of kernel being
    deployed.

The final parsed and expanded boot commands are reported in the job logs:

```text
- usb start
- setenv autoload no
- setenv initrd_high 0xffffffff
- setenv fdt_high 0xffffffff
- dhcp
- setenv serverip 192.168.18.116
- tftp 0x01000000 347/tftp-deploy-lf41516b/kernel/zImage
- tftp 0x04000000 347/tftp-deploy-lf41516b/ramdisk/ramdisk.cpio.gz.uboot
- setenv initrd_size ${filesize}
- tftp 0x03f00000 347/tftp-deploy-lf41516b/dtb/bcm2837-rpi-3-b-plus.dtb
- setenv bootargs ' root=/dev/nfs rw nfsroot=192.168.18.116:/var/lib/lava/dispatcher/tmp/347/extract-nfsrootfs-bzt0mf_r,tcp,hard earlycon=uart8250,mmio32,0x3f215040 ip=dhcp'
- bootz 0x01000000 0x04000000 0x03f00000
```

Common placeholders used in multiple boot commands:

| Placeholder | Description |
| --- | --- |
| `{LAVA_DISPATCHER_IP}` | IP address of the LAVA worker |
| `{LAVA_JOB_ID}` | The numeric LAVA job ID |
| `{LAVA_MAC}` | DUT MAC address configured in device dictionary |
| `{SERVER_IP}` | Alias for `{LAVA_DISPATCHER_IP}` |
| `{KERNEL}` | TFTP path to the `kernel` image |
| `{RAMDISK}` | TFTP path to the `ramdisk` image |
| `{DTB}` | TFTP path to the `dtb` file |
| `{INITRD}` | Alias for `{RAMDISK}` |
| `{TEE}` | TFTP path to the `tee` binary |
| `{NFS_SERVER_IP}` | Alias for `{LAVA_DISPATCHER_IP}` |
| `{NFSROOTFS}` | Path to the extracted `nfsroot` filesystem |
| `{KERNEL_ADDR}` | Load address for the kernel |
| `{RAMDISK_ADDR}` | Load address for the ramdisk |
| `{DTB_ADDR}` | Load address for the device tree blob |
| `{DTB_BASE_ADDR}` | Base address for the device tree (used with overlays) |
| `{DTB_BASE_RESIZE}` | Resize value for the base device tree (used with overlays) |
| `{DTBO_ADDR}` | Load address for device tree overlays |
| `{TEE_ADDR}` | Load address for the TEE binary |
| `{NBDSERVERIP}` | Alias for `{LAVA_DISPATCHER_IP}` |
| `{NBDSERVERPORT}` | Port of the NBD server |
| `{PRESEED_CONFIG}` | TFTP path to the preseed configuration file |
| `{PRESEED_LOCAL}` | File name embedded in the ramdisk root |
| `{BOOTX}` | The final boot command with appropriate addresses |

### Custom boot commands

During testing and development, it can be useful to override the boot commands
from the device-type template. These custom commands must still **include the
placeholders** so that LAVA can substitute paths and values:

```yaml
- boot:
    method: ipxe
    commands:
    - dhcp net0
    - set console console=ttyS0,115200n8 lava_mac={LAVA_MAC}
    - set extraargs ip=dhcp root=/dev/sda1 rw
    - kernel tftp://{SERVER_IP}/{KERNEL} ${extraargs} ${console}
    - initrd tftp://{SERVER_IP}/{RAMDISK}
    - boot
```

!!! warning
    This support is recommended for corner cases that can't be fixed at the
    device-type level. LAVA will raise a warning each time custom commands are
    used. Abuse of this feature can potentially stop devices from working in
    subsequent tests, or even damage them permanently.

    If you need to use these commands regularly, request that a label be created
    in the device type for this command set, or propose a patch.

### Extra kernel arguments

A test job may require extra kernel command line options. The
[job context](../../../../introduction/glossary.md#job-context) can be used to
append strings to the kernel command line:

!!! note
    `context` is a top-level element of the LAVA job definition, not part of
    the `boot` section.

```yaml
context:
  extra_kernel_args: vsyscall=native
```

The parameter values need to be separated by whitespace and will be added to
the command line with a prefix and suffix of a single space.

The possible values which can be used are determined solely by the support
available within the kernel provided to the DUT.

### nfsroot arguments

To append values to the NFS options, use `extra_nfsroot_args`:

```yaml
context:
  extra_nfsroot_args: ",rsize=4096 nfsrootdebug"
```

!!! note
    `extra_nfsroot_args` are appended directly to the existing NFS flags
    `nfsroot={NFS_SERVER_IP}:{NFSROOTFS},tcp,hard`. If the appended string
    contains an extra flag, it must come first and the string must start with
    a comma. Other options can be separated by a space.

    Example result: `nfsroot=10.0.0.1:/nfs,tcp,hard,rsize=4096 nfsrootdebug`

See also: [Kernel documentation for NFSROOT](https://www.kernel.org/doc/Documentation/filesystems/nfs/nfsroot.txt)

## auto_login

Some systems require the test job to specify a username and optionally a
password to login. These values must be specified in the test job definition.
If the system boots directly to a prompt without needing a login, the
`auto_login` section can be omitted.

### login_prompt

The prompt to match when the system requires a login. This prompt needs to be
unique across the entire boot sequence, so typically includes `:` and should
be quoted:

```yaml
- boot:
    auto_login:
      login_prompt: 'login:'
      username: root
```

!!! note
    If `login_prompt` is not matched during boot, LAVA will send control
    characters to the shell assuming a kernel alert occurred. This may result
    in incorrect login attempts, but LAVA will automatically retry after
    recognizing the `Login incorrect` message.

### username

Whenever a `login_prompt` is specified, a `username` is also required. The
username should either be `root` or a user with passwordless `sudo` access.

### password_prompt

If the login requires a password as well as a username, the `password_prompt`
must be specified:

```yaml
- boot:
    auto_login:
      login_prompt: 'login:'
      username: root
      password_prompt: 'Password:'
      password: rootme
```

!!! note
    If `password_prompt` is not matched during login, or a password is required
    but not provided, LAVA will recognize the `Login timed out` message, stop
    the job execution, and log the error.

### password

Whenever a `password_prompt` is specified, a `password` is also required.

### login_commands

A list of commands to run after the initial login and before setting the shell
prompt. This is typically used to switch from a regular user to root:

```yaml
- boot:
    auto_login:
      login_prompt: 'login:'
      username: user
      password_prompt: 'Password:'
      password: pass
      login_commands:
      - sudo su
```

!!! note
    No interactive input (such as a password) can be provided with
    `login_commands`.

## prompts

After login (or directly from boot if no login is required), LAVA needs to
match the first prompt offered by the booted system. The full list of possible
prompts **must** be specified in the test job definition.

Each prompt needs to be unique across the entire boot sequence, so typically
includes `:` and needs to be quoted. If the hostname of the device is included
in the prompt, this can be included in the prompt string.

```yaml
- boot:
    prompts:
    - 'root@debian:'
```

When the hostname varies, use a regex pattern:

```yaml
- boot:
    prompts:
    - 'root@(.*):'
```

When using a ramdisk, the prompt may contain brackets which need to be escaped:

```yaml
- boot:
    prompts:
    # escape the brackets to ensure that the prompt does not match
    # kernel debug lines which may mention initramfs
    - '\(initramfs\)'
```

When using `login_commands` that change the user (such as `sudo su`), include
prompts for both the initial user and the final user in the prompts list:

```yaml
- boot:
    auto_login:
      login_prompt: "login:"
      username: pi
      password_prompt: 'Password:'
      password: raspberry
      login_commands:
      - sudo su
    prompts:
    - "pi@raspberrypi:"
    - "root@raspberrypi:"
```

!!! warning "Take care with the specified prompts:"

    - Prompt strings which do not include enough characters can match early,
      resulting in a failed login
    - Prompt strings which include extraneous characters may fail to match
    - Avoid user-specific prompt elements like `$` (unprivileged user) or `#`
      (superuser) or `~` (home directory)
    - The prompt string should **include and usually end with** a colon `:` or
      a colon and space

## transfer_overlay

A LAVA test overlay is a tarball of scripts which run the LAVA Test Shell for
the test job. It also includes the git clones or the downloads of repositories
specified in the test job submission and the LAVA helper scripts. Normally,
applying the overlay is integrated into the test job automatically. When the
rootfs is not deployed directly by LAVA, for example booting from persistent
storage or a pre-installed system, it is necessary to transfer the overlay from
the LAVA worker to the DUT using commands within the booted system prior to
starting to run the test shell.

The `transfer_overlay` option allows LAVA to transfer the LAVA test overlay to
the DUT after boot.

```yaml
- boot:
    method: minimal
    transfer_overlay:
      download_command: wget
      unpack_command: tar -C / -xzf
    prompts:
    - 'root@debian:~#'
```

!!!note
    The overlay is transferred before any test shell operations run. Therefore,
    the commands required for the selected `transfer_method` must be available
    and functional after boot and before running the test shell. For `http` and
    `nfs` transfers this means the network, the download command and the
    unpack command **must** work. For the `zmodem` transfer method only the
    unpack command is required locally.

The `download_command` and `unpack_command` can include one or more shell
commands. Avoid using redirects (`>` or `>>`) or other complex shell syntax.
This example below changes to `/tmp` to ensure there is enough writeable space
for the download. This feature can also be used to install the utilities or
wait for the network needed for the transfer.


```yaml
- boot:
    transfer_overlay:
      download_command: cd /tmp ; wget
      unpack_command: tar -C / -xzf
```

### transfer_method

Optional. Defaults to `http`. Can be set to `nfs` for NFS-based transfer or to
`zmodem` for ZMODEM-based transfer.

#### http

This will transfer overlay through the worker's Apache service.

```yaml
- boot:
    transfer_overlay:
      transfer_method: http
      download_command: wget --progress=dot:giga
      unpack_command: tar -C / -xzf
```

The `--progress=dot:giga` options to wget in the example above optimize the
output for serial console logging to avoid wasting line upon line of progress
percentage dots. If the system uses `busybox`, these options may not be
supported by the version of `wget` on the device.

#### nfs

This will transfer overlay through the worker's NFS server service.

```yaml
- boot:
    transfer_overlay:
      transfer_method: nfs
      download_command: mount -t nfs -o nolock
      unpack_command: cp -rf
```

Since the `mount` command requires the NFS mount helper program to mount NFS
shares, the target device **must** have NFS client utilities installed. This is
typically provided by:

- `nfs-common` package on Debian-based systems
- `nfs-utils` package on Fedora-based systems

Example for installing the package for Debian at job runtime:

```yaml
- boot:
    transfer_overlay:
      transfer_method: nfs
      download_command: apt update && apt-get install -y nfs-common && mount -t nfs -o nolock
      unpack_command: cp -rf
```

When needed, you can append `-o Acquire::Check-Valid-Until=false` to the
`apt update` command for skipping APT's repository time validity check.

#### zmodem

This will transfer overlay through the ZMODEM inline file transfer protocol
from the worker to the target device.

```yaml
- boot:
    transfer_overlay:
      transfer_method: zmodem
      unpack_command: tar -C / -xzf
```

The worker and target device **must** have respectively the `sz` and `rz`
utilities installed. This is typically provided by:

- `lrzsz` package on Debian-based or Fedora-based systems

The `zmodem` transfer method is interesting for target device that does not
have an Ethernet connection. As a consequence, the `rz` utility should be
available by default in the flashed image.

The worker and the target device communicate over a `UART`. This `UART` usually
appears on the worker as `/dev/ttyUSB<N>` or `/dev/ttyACM<N>`. This value
**must** be defined in the device dictionary using the `device_info` variable
and the `uart` field. This is an example taken from a device dictionary.

```jinja
{% set device_info = [{'uart': '/dev/ttyUSB0'}] %}
```


### download_command

The command used to download the overlay file from the LAVA worker. This will
be called with the URL to the overlay file appended. For example:

```shell
wget http://192.168.18.116/tmp/427/compress-overlay-5jg6nqov/overlay-1.1.5.tar.gz
```

### unpack_command

The command used to extract the overlay archive. The overlay filename will be
appended to this command. For example: `tar -C / -xzf overlay-1.1.5.tar.gz`

!!! note
    The `-C /` command to tar is **essential** or the test shell will not be
    able to start. The overlay will use `gzip` compression, so pass the `z`
    option to `tar`.

Some systems do not store the current time between boots. The `--warning
no-timestamp` option is a useful addition for `tar` for those systems but note
that `busybox tar` does not support this option.

## parameters

The optional `parameters` block in the job definition boot action allows you to
override the following device constants.

### bootloader-final-message

The string that the bootloader prints when it has finished. After sending the
last bootloader command, LAVA waits for this message before continuing.

See the boot method in your device type for the default value. Adjust the value
in the device type or in the boot action parameters block:

```yaml
- boot:
    method: u-boot
    commands: nfs
    parameters:
      bootloader-final-message: "Booting kernel"
    prompts:
    - 'root@device:~#'
```

### kernel-start-message

The string that indicates the kernel has started booting. LAVA waits for this
message and then starts parsing the kernel boot logs to detect common kernel
boot errors.

The default is `Linux version [0-9]`. Adjust it in your device type or in the
boot action parameters block:

```yaml
- boot:
    method: minimal
    parameters:
      kernel-start-message: 'Starting kernel'
    prompts:
    - 'root@device:~#'
```

!!!note
    Only after the start message is matched, LAVA starts waiting for the boot
    prompts or login prompt to appear. When needed, set the string to `""` to
    skip the matching.

### shutdown-message

The message indicates the system is shutting down or rebooting. It is used when
performing a [soft reboot](#soft_reboot). LAVA waits for this string to confirm
that the device has started shutting down or rebooting.

The default is `The system is going down for reboot NOW`. You can override it
in your device configuration or in the boot action parameters.

```yaml
- boot:
    method: u-boot
    commands: ramdisk
    parameters:
      shutdown-message: "reboot: Restarting system"
    prompts:
    - 'root@device:~#'
```

## soft_reboot

When LAVA needs to reset the device and the `hard_reset_command` is not defined
in the device configuration, it sends a soft reboot command.

By default, LAVA sends the command in the default list
`["reboot", "reboot -n", "reboot -nf"]` one by one until the
[shutdown-message](#shutdown-message) is matched.

You can override the default in your device configure using the
`soft_reboot_command` variable or in the boot action parameters using the
`soft_reboot` option.

```yaml
- boot:
    method: u-boot
    commands: ramdisk
    soft_reboot: reboot -nf
    prompts:
    - 'root@device:~#'
```

## ignore_kernel_messages

Some test scenarios deliberately force a kernel panic. To prevent LAVA from
stopping the job when this happens, set `ignore_kernel_messages` to `true`:

```yaml
- boot:
    method: u-boot
    ignore_kernel_messages: true
    prompts:
    - 'root@device:~#'
```

!!! warning
    When `ignore_kernel_messages` is `true`, LAVA won't be able to detect any
    "legitimate" kernel crashes either.

This option also disables detection of unexpected board resets. The pattern
`U-Boot SPL 20[0-9][0-9]` is used to detect board reset. Disabling kernel
message parsing can be useful when a board reset is expected.

Default value: `false`

# How to find issues

 When something is not working as expected, check the following areas:

* Job logs
* Job not terminating
* [Crashes in LAVA logs](../../technical-references/services/lava-worker.md)
* [State machine](../../technical-references/state-machine.md)
* Job queue not empty
* Performances

# Health Checks

See [health checks](../../technical-references/configuration/health-check.md) for
adding device health checks.

[Notifications](../../technical-references/job-definition/notifications.md) should
be enabled to ensure administrators are alerted when health checks fail.

# Contribution Process

To contribute changes to LAVA, follow this process:

1. [Create a GitLab account](#create-a-gitlab-account)
2. [Fork the code](#fork-the-code)
3. [Create a development branch](#create-a-development-branch)
4. [Make and commit your changes](#make-and-commit-your-changes)
5. [Push your changes](#push-your-changes)
6. [Submit a Merge Request](#submit-a-merge-request)
7. [Review and fix](#review-and-fix)
8. [Merging](#merging)

!!! note
    It is worth checking if someone already has a merge request which
    relates to your proposed changes. Check for open merge requests at
    <https://gitlab.com/lava/lava/merge_requests>

## Create a GitLab account

To be able to work with the LAVA Software Community Project, start by
creating an account on <https://gitlab.com/lava/>. Fill in details in
your profile, and make sure you add a public SSH key to your account.
You will need that to be able to push code changes.

## Fork the code

Fork the lava project in the GitLab web interface. This will set up a
copy of the lava project in your own personal namespace. From here, you
can create new branches as you like, ready for making changes.

## Create a development branch

Clone your fork of the lava software repository:

```shell
git clone git@gitlab.com:yourname/lava.git
```

We recommend always making a new local branch for your changes:

```shell
cd lava
git checkout -b <name-of-branch>
```

See also
[Create a branch](https://docs.gitlab.com/topics/git/branch/#create-a-branch).

## Make and commit your changes

Make and test the changes you need. The details here are down to you!

### Coding style

* Follow PEP8 style for Python code.
* Use one topic branch for each logical change.
* Include new unit tests in the proposed merge request.
* Write good commit messages — describe *what* you've changed, and *why*

See also
[How to Write a Git Commit Message](https://chris.beams.io/git-commit).

### Source code formatting

`ruff check` and `ruff format` should be applied to **all** LAVA source code
files. Merge requests will **fail** CI if a change breaks the formatting.
`ruff check` should be run with `--select I` option before `ruff format` to
ensure compatibility.

When changing files formatted by `ruff format`, make your changes and then run
`ruff format` on all modified Python files before pushing the branch to GitLab.

### Signing off your commits

Use the `--signoff` or `-s` option to `git commit` to acknowledge that
you have the rights to submit this change under the terms of the
licenses applicable to the LAVA Software. This is commonly known as the
"Developer's Certificate of Origin" ([DCO](https://developercertificate.org/)),
and is used in a wide variety of other Open Source projects like the
Linux kernel.

```shell
git commit -s
```

GitLab supports including multiple commits in a single merge request, so
feel free to collect your changes in as many logical changesets as you
like. Don't include unrelated changes — use a separate branch (and
therefore a separate merge request) instead.

## Push your changes

Use `git push` to publish the changes on your branch back to your own
fork:

```shell
git push --set-upstream my_username my_branch
```

You can push here as many times as you like, as you make more changes.

Pushing to your fork will trigger the CI process — your changes will now
be automatically tested, and the results will be displayed for the MR.

## Submit a Merge Request

When your code is clean and ready to be reviewed, create a merge request
against the *master* branch of the original lava project. GitLab will
track all the changes that you have pushed to your development branch,
and present them together for review in one patchset.

There are some headlines that we expect in each merge request:

In the git commit message:

1. What does this change do?
2. Why was this change needed?
3. What are the relevant issue numbers?

In the merge request, as comments:

1. Are there points in the code the reviewer needs to double-check?
2. Screenshots or test job log files as links or attachments (if
   relevant)
3. Links to external resources like gold standard images to demonstrate
   how to use and/or test a new feature.

## Review and fix

If your MR failed its tests, you will receive a detailed email
explaining where the failures occurred. It is up to you to make any
fixes required.

If you are not sure how to fix things, please ask for help!

Fixes for test failures should be pushed to the same GitLab branch. Each
time you push, GitLab will automatically update your related merge
request and re-run the CI loop. As and when the code is functional,
maintainers will comment on your changes and if all is well they will
approve the merge. They may also ask you to make more changes — this is
an iterative process.

## Merging

As the final step in merging a change, we will want the list of commits
in the merge request to be squashed. The objective is to ensure that
each commit on the master branch is clean and intact, while also keeping
logical changes in separate commits.

* Ensure that commits to fix unit test failures, CI failures or other
  breakage are squashed into the parent commit.
* Ensure that separate logical changes remain as separate commits. It is
  often easier to use separate branches for this reason.
* Ensure that your commits are all rebased onto the current master
  branch.

Pushing the squashed branch will need you to use `git push --force` to
replace the existing commits in your merge request. The merge request
will get one final code review and if a Maintainer approves of the final
state, the change will be merged when the CI completes successfully.

!!! warning
    This is the **only** time that `git push --force` is ever
    recommended. Forcing a push makes it hard for other contributors to
    work on the changes by triggering lots of merge conflicts.

!!! info "See also"
    - [Contribution Guide](../methodology/contribute.md)
    - [Code of Conduct](../code-of-conduct.md)

# Contribution Process

To contribute changes to LAVA, follow this process:

1. [Create a GitLab account](#create-a-gitlab-account)
2. [Fork the code](#fork-the-code)
3. [Create a development branch](#create-a-development-branch)
4. [Make and commit your changes](#make-and-commit-your-changes)
5. [Push your changes](#push-your-changes)
6. [Submit a Merge Request](#submit-a-merge-request)
7. [Review and fix](#review-and-fix)
8. [Merging](#merging)

!!! note
    It is worth checking if someone already has a merge request which
    relates to your proposed changes. Check for open merge requests at
    <https://gitlab.com/lava/lava/merge_requests>

## Create a GitLab account

To be able to work with the LAVA Software Community Project, start by
creating an account on <https://gitlab.com/lava/>. Fill in details in
your profile, and make sure you add a public SSH key to your account.
You will need that to be able to push code changes.

## Fork the code

Fork the lava project in the GitLab web interface. This will set up a
copy of the lava project in your own personal namespace. From here, you
can create new branches as you like, ready for making changes.

## Create a development branch

Clone your fork of the lava software repository:

```shell
git clone git@gitlab.com:yourname/lava.git
```

We recommend always making a new local branch for your changes:

```shell
cd lava
git checkout -b <name-of-branch>
```

See also
[Create a branch](https://docs.gitlab.com/topics/git/branch/#create-a-branch).

## Make and commit your changes

Make and test the changes you need. The details here are down to you!

### Coding style

* Follow PEP8 style for Python code.
* Use one topic branch for each logical change.
* Include new unit tests in the proposed merge request.
* Write good commit messages — describe *what* you've changed, and *why*

See also
[How to Write a Git Commit Message](https://chris.beams.io/git-commit).

### Source code formatting

`black` and `isort` should be applied to **all** LAVA source code files.
Merge requests will **fail** CI if a change breaks the formatting.
`isort` should be run with `--profile black` option to ensure compatibility
with `black`.

When changing files formatted by `black`, make your changes and then run
`black` on all modified Python files before pushing the branch to GitLab.

### Signing off your commits

Use the `--signoff` or `-s` option to `git commit` to acknowledge that
you have the rights to submit this change under the terms of the
licenses applicable to the LAVA Software. This is commonly known as the
"Developer's Certificate of Origin" ([DCO](https://developercertificate.org/)),
and is used in a wide variety of other Open Source projects like the
Linux kernel.

```shell
git commit -s
```

GitLab supports including multiple commits in a single merge request, so
feel free to collect your changes in as many logical changesets as you
like. Don't include unrelated changes — use a separate branch (and
therefore a separate merge request) instead.

## Push your changes

Use `git push` to publish the changes on your branch back to your own
fork:

```shell
git push --set-upstream my_username my_branch
```

You can push here as many times as you like, as you make more changes.

Pushing to your fork will trigger the CI process — your changes will now
be automatically tested, and the results will be displayed for the MR.

## Submit a Merge Request

When your code is clean and ready to be reviewed, create a merge request
against the *master* branch of the original lava project. GitLab will
track all the changes that you have pushed to your development branch,
and present them together for review in one patchset.

There are some headlines that we expect in each merge request:

In the git commit message:

1. What does this change do?
2. Why was this change needed?
3. What are the relevant issue numbers?

In the merge request, as comments:

1. Are there points in the code the reviewer needs to double-check?
2. Screenshots or test job log files as links or attachments (if
   relevant)
3. Links to external resources like gold standard images to demonstrate
   how to use and/or test a new feature.

## Review and fix

If your MR failed its tests, you will receive a detailed email
explaining where the failures occurred. It is up to you to make any
fixes required.

If you are not sure how to fix things, please ask for help!

Fixes for test failures should be pushed to the same GitLab branch. Each
time you push, GitLab will automatically update your related merge
request and re-run the CI loop. As and when the code is functional,
maintainers will comment on your changes and if all is well they will
approve the merge. They may also ask you to make more changes — this is
an iterative process.

## Merging

As the final step in merging a change, we will want the list of commits
in the merge request to be squashed. The objective is to ensure that
each commit on the master branch is clean and intact, while also keeping
logical changes in separate commits.

* Ensure that commits to fix unit test failures, CI failures or other
  breakage are squashed into the parent commit.
* Ensure that separate logical changes remain as separate commits. It is
  often easier to use separate branches for this reason.
* Ensure that your commits are all rebased onto the current master
  branch.

Pushing the squashed branch will need you to use `git push --force` to
replace the existing commits in your merge request. The merge request
will get one final code review and if a Maintainer approves of the final
state, the change will be merged when the CI completes successfully.

!!! warning
    This is the **only** time that `git push --force` is ever
    recommended. Forcing a push makes it hard for other contributors to
    work on the changes by triggering lots of merge conflicts.

!!! info "See also"
    - [Contribution Guide](../methodology/contribute.md)
    - [Code of Conduct](../code-of-conduct.md)

# Interactive Tests

[Interactive](../../technical-references/job-definition/actions/test.md#interactive)
test action can be used for adding tests with a non-POSIX shell or just arbitrary
interactive applications. Examples are provided in the link.

# AVH

Creates a custom AVH package from a set of `images`, or deploys a pre-built AVH
`fw_package`.

## Custom package

Creates a custom AVH Linux firmware package from individual images.

```yaml
- deploy:
    to: avh
    options:
      model: rpi4b
      api_endpoint: https://app.avh.corellium.com/api
      project_name: "Default Project"
    images:
      rootfs:
        url: https://example.com/rpi4b/nand
        format: ext4
        # root partition for LAVA test overlay
        root_partition: 1
        # partition for custom overlays
        partition: 1
        overlays:
          modules:
            compression: xz
            format: tar
            path: /
            url: https://example.com/rpi4b/modules.tar.xz
      kernel:
        url: https://example.com/rpi4b/kernel
      dtb:
        url: https://example.com/rpi4b/devicetree
```

The deployment action runs the following tasks in serial:

* Download the `kernel`, `dtb` and `rootfs` images required by AVH custom firmware.
* Apply lava overlay to `rootfs` image when needed.
* Generate an `Info.plist` file for the meta information about firmware version,
  type, build, unique identifier, and device identifier.
* Create a zip package that contains the images and the `Info.plist` file.

### rootfs

OS disk image file.

#### url

See [url](./index.md#url)

#### format

Specify disk image format for LAVA to apply overlay. `ext4` is the only
supported format now.

#### root_partition

Specify disk image root partition index for LAVA to apply overlay. Disk image
partition index 0 usually is the boot partition, while index 1 usually is the
root partition.

#### overlays

AVH doesn't support adding custom Kernel modules to custom Linux firmware
package yet. Fortunately, this is supported in LAVA by applying custom
`overlays.modules` into the `rootfs` image.

See [overlays](./index.md#overlays)

### kernel

The Linux kernel in the `Image` format.

### dtb

The device tree for Linux in binary `.dtb` format.

See also [DTB](../../../../introduction/glossary.md#dtb)

## Pre-built firmware package

Deploys a pre-built AVH firmware package.

```yaml
- deploy:
    to: avh
    options:
      model: kronos
      api_endpoint: https://app.avh.corellium.com/api
      project_name: "Default Project"
    fw_package:
      url: https://example.com/kronos-baremetal-psa-secure-storage-1.0.zip
      storage_file: virtio_0
      root_partition: 1
```

### storage_file

The name of the storage file inside the firmware package. It should be a disk
image in ext4 format.

See [AVH Storage](https://support.avh.corellium.com/devices/firmware/linux-firmware#storage)
for details.

### root_partition

See [root_partition](#root_partition)

## Options

A dictionary of AVH configuration.

### model

This `model` key is crucial and mandatory. `avh` deploy action needs it to
generate `Info.plist` file. `avh` boot action needs it to create instance.

Here are the tested models in LAVA.

* rpi4b
* imx93
* kronos

Other models that capable to run Linux should work too.

### api_endpoint

(optional) AVH API endpoint defaults to https://app.avh.corellium.com/api.

### project_name

(optional) AVH project name defaults to `Default Project`.

# U-Boot

The `u-boot` boot method boots the downloaded files using U-Boot commands.

```yaml
- boot:
    method: u-boot
    commands: nfs
    prompts:
    - '/ #'
```

## commands

Specifies a predefined set of U-Boot commands that LAVA will execute to boot
the device. The commands are templates where LAVA substitutes placeholders with
actual values, such as:

- File locations of downloaded kernel, ramdisk, and dtb files
- Network configuration details (e.g., `SERVERIP` for TFTP server address)
- NFS server location and mount paths (when using NFS boot)

See your device configuration for the complete list of supported commands.
Common command sets include:

| Command | Description |
| ------- | ----------- |
| `ramdisk` | Boot from kernel and ramdisk loaded via TFTP |
| `nfs` | Boot kernel via TFTP, mount root filesystem via NFS |
| `nbd` | Boot using Network Block Device |
| `usb` | Boot from USB storage |
| `sata` | Boot from SATA storage |

Certain elements of the command line are available for modification using the
[job context](../../../../introduction/glossary.md#job-context). For example,
NXP Layerscape platforms support booting from Alternate Bank, keeping Main Bank
safe. If you want to boot the board from Alternate Bank, you can do it by
adding context variable `uboot_altbank: true`. By default, its value is set to
`false`.

```yaml
context:
  uboot_altbank: true
```

## reset

By default, LAVA will reset the board power when executing this action. You can
skip this step by setting `reset: false`.

This is useful when the device is already booted into U-Boot from a previous
boot action (such as `fastboot`).

The following example demonstrates a workflow where U-Boot is first deployed
and booted via `fastboot`, then the `u-boot` boot method is used without
resetting.

```yaml
- deploy:
    to: fastboot
    docker:
      image: lavalabteam/adb-fastboot
      local: true
    images:
      boot:
        url: https://example.com/uboot.img
    timeout:
      seconds: 90

- boot:
    method: fastboot
    docker:
      image: lavalabteam/adb-fastboot
      local: true
    timeout:
      seconds: 30

# Boot with u-boot commands without resetting as device is already in U-Boot
- boot:
    method: u-boot
    reset: false
    commands: nfs
    prompts:
    - 'root@board:~#'
```

## See also

[Raspberry Pi (U-Boot) device setup](../../../../admin/basic-tutorials/device-setup/u-boot.md)

# QDL

The `qdl` boot method allows to flash Qualcomm devices using the [qdl](https://github.com/linux-msm/qdl) tool.

```
- boot:
    method: qdl
    firehose_program: "prog_firehose_ddr.elf"
    rawprogram: "rawprogram*.xml"
    patch: "patch*.xml"
    path: "path-to-dir-inside-tarball"
    storage: "emmc"
    timeout:
      minutes: 5
```

## Installation

LAVA supports running `qdl` directly on the worker host or from a Docker container.
In both cases, LAVA administrators have to make sure `qdl` is installed on the worker.

The latest release is available at [https://github.com/linux-msm/qdl/releases](https://github.com/linux-msm/qdl/releases).

## Device configuration

## qdl parameters

Some of the `qdl` parameters must be provided in the job definition.

### firehose_program

Since each Qualcomm devices uses a different `firehose` protocol implementation,
the user must specify the filename of the `firehose` program to be used by `qdl`.
This filename is relative to the top tarball directory.
See [deploy-to-qdl](../deploy/to-qdl.md) for more details.

### rawprogram

List of `rawprogram` files to be used by `qdl`. The filenames should be delimited by whitespace
and should be specified relative to the root of the tarball defined in `qcomflash`.
See [deploy-to-qdl](../deploy/to-qdl.md) for more details.

### patch

List of `patch` files used by `qdl`. The filenames should be delimited by whitespace
and should be specified relative to the root of the tarball defined in `qcomflash`.
See [deploy-to-qdl](../deploy/to-qdl.md) for more details.

### storage

Storage device for `qdl` to write data to. Supported values include `emmc`, `ufs`, `spinor`, etc.
See [qdl documentation](https://github.com/linux-msm/qdl/blob/master/README.md) for more details.

### path

Path inside the downloaded tarball containing the `rawprogram` and `patch` files.
The paths referenced by `rawprogram` and `patch` files are relative, so `qdl` must be ran from this directory.

# Downloads

## Downloads deploy

Download the given images and allow to post-process the images.

```yaml
actions:
- deploy:
    to: downloads
    images:
      rootfs:
        url: https://example.com/rootfs.img.xz
        compression: xz
```

## Post-processing

The `downloads` deploy action allows to run a custom post-processing script. The configuration would look like the following:

```yaml
actions:
- deploy:
    to: downloads
    images:
      rootfs:
        url: https://example.com/sid.img.xz
        compression: xz
    postprocess:
      docker:
        image: debian
        steps:
        - cp sid.img sid-modified.img
        - ls -lhR
```

The post-processing `steps` are ran inside the provided docker container. The
artifacts will be in sub-directories of the working directory. The name of
the sub-directories are the keys of the `images` dictionary (`rootfs` in the
example).

!!! info "Using the artifacts"
    In subsequent deploy actions, you can use the artifacts created by the
    postprocessing step with `downloads://sid-modified.img`.

### LAVA overlay

LAVA overlay is prepared when there is a test action available in the job.
In some cases LAVA is able to add overlay to image as part of the download
process. As this is not always the case, it is useful to have access to
overlay tarball in the postprocessing step. The file is available in the
`downloads` directory. The file name isn't constant as it depends on the
test job definition structure. Example file name can look like:

    overlay-1.1.5.tar.gz

It always starts with `overlay` and ends with `.tar.gz`. The middle part
depends is the ID of the action in the LAVA job definition and will be different
for different job definitions.

Example use of overlay in the postprocessing step can be adding it to the ramdisk

```yaml
- deploy:
    images:
      image:
        url: 'https://example.com/Image'
      dtb:
        url: 'https://example.com/device.dtb'
      ramdisk:
        url: 'https://example.com/ramdisk.gz'
    timeout:
      minutes: 5
    to: downloads
    postprocess:
      docker:
        image: mkbootimage:master
        steps:
        - mkdir lava_overlay
        - tar xvf overlay*.tar.gz -C lava_overlay
        - "(cd lava_overlay ; find * | cpio -o -H newc -R +0:+0 | gzip -9 >> ../ramdisk.gz)"
        - mkbootimg --header_version 2 --kernel Image --dtb device.dtb --cmdline "earlycon clk_ignore_unused pd_ignore_unused audit=0" --ramdisk ramdisk.gz --output boot.img
```

# LAVA docker worker

Start lava-worker inside a docker container, matching the server version.

## Command line

Run `/usr/bin/lava-docker-worker`.

## Service

The systemd service is called `lava-docker-worker`.

## Dependencies

lava-docker-worker should be able to:

* use docker

## Configuration

Daemon start options:

* `/etc/default/lava-docker-worker`
* `/etc/lava-server/lava-docker-worker`

## Host directories

If /usr/share/lava-docker-worker exists, it is bind mounted into the container,
and if /usr/share/lava-docker-worker/bin exists it is added to PATH.

If /usr/local/share/lava-docker-worker exists, it is bind mounted into the
container, and if /usr/local/share/lava-docker-worker/bin exists it is added to
PATH.

## Logs

The logs are stored in `/var/log/lava-dispatcher-host/lava-docker-worker.log`

The log rotation is configured in `/etc/logrotate.d/lava-docker-worker-log`.

## Security

TODO: should activate encryption

# User permissions

## Authorization

For how LAVA manage user permissions, see [authorization](../../../technical-references/authorization.md).

## Private instance

See [private instance](../../../technical-references/authorization.md#private-instance).

## Private devices/device-types

User access to devices and device types in LAVA can be configured from the Django
admin UI or via the XML-RPC API.

### Django admin UI

Navigate to the individual **device type** or **device** object in the Django
admin UI, find the **Group device (type) permissions** table, and add or remove
group permissions directly on the object.

### XML-RPC API

The XML-RPC methods are provided for managing device and device type permission
 (see `/api/help/`). You can use
 [lavacli](../../../user/basic-tutorials/lavacli.md) that based on the APIs to
 manage permissions.

#### Permission codenames:

| Object      | View              | Submit                 | Change              |
|-------------|-------------------|------------------------|---------------------|
| Device type | `view_devicetype` | `submit_to_devicetype` | `change_devicetype` |
| Device      | `view_device`     | `submit_to_device`     | `change_device`     |

#### Examples

Allow group `lkft` to view and submit to the `qemu` device type:

```shell
lavacli -i <instance> device-types perms add qemu lkft view_devicetype
lavacli -i <instance> device-types perms add qemu lkft submit_to_devicetype
```

Remove the submit permission:

```shell
lavacli -i <instance> device-types perms delete qemu lkft submit_to_devicetype
```

List permissions on the `qemu` device type:

```shell
lavacli -i <instance> device-types perms list qemu
```

# User permissions

## Authorization

For how LAVA manage user permissions, see [authorization](../../../technical-references/authorization.md).

## Private instance

See [private instance](../../../technical-references/authorization.md#prviate-instance).

## Private devices/device-types

User access to devices and device types in LAVA can be configured from the Django
admin UI or via the XML-RPC API.

### Django admin UI

Navigate to the individual **device type** or **device** object in the Django
admin UI, find the **Group device (type) permissions** table, and add or remove
group permissions directly on the object.

### XML-RPC API

The XML-RPC methods are provided for managing device and device type permission
 (see `/api/help/`). You can use
 [lavacli](../../../user/basic-tutorials/lavacli.md) that based on the APIs to
 manage permissions.

#### Permission codenames:

| Object      | View              | Submit                 | Change              |
|-------------|-------------------|------------------------|---------------------|
| Device type | `view_devicetype` | `submit_to_devicetype` | `change_devicetype` |
| Device      | `view_device`     | `submit_to_device`     | `change_device`     |

#### Examples

Allow group `lkft` to view and submit to the `qemu` device type:

```shell
lavacli -i <instance> device-types perms add qemu lkft view_devicetype
lavacli -i <instance> device-types perms add qemu lkft submit_to_devicetype
```

Remove the submit permission:

```shell
lavacli -i <instance> device-types perms delete qemu lkft submit_to_devicetype
```

List permissions on the `qemu` device type:

```shell
lavacli -i <instance> device-types perms list qemu
```

# Job submit

Jobs can be submitted from either the [web interface](#web-interface), [lavacli](#lavacli) or the REST API.

## QEMU job

You will start with this simple QEMU job:

```yaml
--8<-- "jobs/qemu.yaml"
```

!!! tip
    If `/dev/kvm` is unavailable on the worker, it can be disabled by adding
    `no_kvm: true` to the job context.

## Web interface

In order to submit from the web interface, you should first login and then go to
`Scheduler > Submit`.

![submit menu](submit-job-menu.png)

In the submit page, copy and paste the job definition and click on `Validate`.

![submit validate](submit-job-validate.png)

The server will validate the job definition and return any errors or warning found.

If the job is valid, the `Submit` button will become active. Clicking on it
will actually submit the job.

![submit](submit-job-submit.png)


## lavacli

In order to submit using [lavacli][lavacli], start by installing and configure
lavacli with the [lavacli tutorial](./lavacli.md).

Submit the job with:
```shell
lavacli -i <identity> jobs submit qemu.yaml
```

lavacli will submit and print the job identifier that you can use to inspect
the current job.

```shell
lavacli -i <identity> jobs show <id>
lavacli -i <identity> jobs logs <id>
lavacli -i <identity> jobs wait <id>
```

--8<-- "refs.txt"

# Deploying a Raspberry Pi4B as a LAVA worker

Due to the distributed nature of LAVA architecture, worker part of the software
does not require latest and most expensive hardware. LAVA lab is already using
it in it's own setup for particular instances.

Here we will provide a reference and how-to for deploying one such RPi4 as a
LAVA worker.

## Get the hardware

What you will need for a worker is:

* RPi4 with power supply
* SD card with min 8GB (only temporarily)
* SSD / Flash drive
* SATA(M.2, etc) to USB adapter
* FTDI USB to serial UART cable

There is a list of know working and also problematic [USB adapters](https://jamesachambers.com/raspberry-pi-4-usb-boot-config-guide-for-ssd-flash-drives/)
along with the more extensive setup instructions which we will cover here.

## Prepare the Pi

Download the [Raspberry Pi OS](https://www.raspberrypi.org/software/operating-systems/) and flash it onto an SD card.
Once you get to login prompt the user is `pi` and password is `raspberry`
then perform the following steps:

```shell
apt update
apt full-upgrade
rpi-update
reboot
sudo rpi-eeprom-update -d -a
reboot
```

After that launch the raspi-config:

```shell
raspi-config
```

In the menu, find the `Boot ROM Version` (it defers where it is depending on
the version of raspi-config, either under `Boot options` or
`Advanced settings`).
Select **Latest** version of boot ROM software. In the next menu, for a question "Reset boot ROM to defaults?", select **NO**.

Then in the same submenu select "Boot Order" and choose "USB Boot". That's
it.

## Prepare the SSD

Currently we only support a Debian images for LAVA installation (unless you'd
like to use docker).

Get the latest Debian image from [Tested images](https://raspi.debian.net/tested-images/) - download the one for the RPi 4 family. This is a good
stock image for a number of reasons, not least of which is that it already has
enable-uart set to 1 in the boot config.txt, which enables serial output.
The user for this image is root and there is no password.

Unpack it and “dd” it onto the SSD.
If everything is OK, you should be able to remove the SD card from th Pi and
boot the system from the SSD over USB.

We're now ready to install the LAVA worker.

## Installing LAVA worker

* Start with upgrading the system and installing the packages:

```shell
apt update && apt dist-upgrade
apt install net-tools wget gnupg curl ca-certificates
```

* Edit the /etc/hostname and /etc/hosts files with appropriate hostname / IP
address.
* Add the LAVA repo to apt:

```shell
echo "deb https://apt.lavasoftware.org/release bullseye main" > /etc/apt/sources.list.d/lava.list
```

* Add LAVA archive signing keys to apt:

```shell
curl -fsSL https://apt.lavasoftware.org/lavasoftware.key.asc | apt-key add -
```

* Install the lava-dispatcher package

```shell
apt update
apt install lava-dispatcher
apt dist-upgrade -t bullseye
```

## Auto-registration of worker (or token)

We need to tell LAVA master to `trust` this worker and there's two ways to
do this. You can either add the following line to `/etc/lava-server/settings.d/01-autoregister.yaml` on the LAVA master (remember to change the netmask for
the subnet where the Pi is):

```yaml
WORKER_AUTO_REGISTER_NETMASK: ["172.0.0.0/8"]
```

Or, alternatively you can add a token of your choice in the LAVA worker (Pi 4)
in `/etc/lava-dispatcher/lava-worker` settings and paste the same token in the
LAVA admin UI on the master in the worker section for this particular worker.

## Disable master/worker version mismatch checking

LAVA master rejects workers which do not run strictly the same version.
You can disable this in eg. `/etc/lava-server/settings.d/01-autoregister.yaml`

```yaml
ALLOW_VERSION_MISMATCH: true
```
Note: Keep in mind that the version check is here to prevent against strange
issues when version are not compatible. Like when the device-dictionary are
changing and the dispatcher depends on such changes.

## Setting up the server url

Once you have lava-worker running you need to point it to the LAVA server it
will connect to. Do this in `/etc/lava-dispatcher/lava-worker` on the Pi.

After that RPi4 worker should be visible in the LAVA server UI and marked as
**Online**.

# UUU

The `uuu` deployment action downloads images for flashing to NXP i.MX devices
using the [UUU (Universal Update Utility)](https://github.com/NXPmicro/mfgtools).
The job definition schema is very similar to the [`fastboot`](./to-fastboot.md)
deployment method, with a required boot partition.

```yaml
- deploy:
    to: uuu
    images:
      boot:
        url: https://example.com/imx-boot-sd.bin-flash
      system:
        url: https://example.com/imx-image-multimedia.rootfs.wic
        apply-overlay: true
        root_partition: 1
```

## images

The `images` block specifies a set of images to be downloaded and deployed to
the device.

### partition

Each key in the `images` dictionary is a partition name that can be referenced
in the [`uuu`](../boot/method-uuu.md) boot method `commands` using
`{partition_name}` placeholders.

```yaml
- deploy:
    to: uuu
    images:
      boot:
        url: https://example.com/imx-boot-sd.bin-flash
- boot:
    method: uuu
    commands:
    - uuu: -b sd {boot}
```

!!! warning
    Partition `boot` is required by `uuu` boot method to perform USB serial
    download availability check. The check attempts to write the boot image into
    memory using the command `uuu {boot}`. If the command does not complete
    within 10 seconds, primary boot-media will be erased using
    `uuu_corrupt_boot_media_command`, so the device can enter USB serial
    download mode after the following reset.

### apply-overlay

The default is `false`. Set to `true` to apply the LAVA test overlay to this
image. The overlay contains the test definitions and LAVA helper scripts.

```yaml
- deploy:
    to: uuu
    images:
      system:
        url: https://example.com/system.img.xz
        compression: xz
        apply-overlay: true
```

### root_partition

Specify the root partition index within a disk image for LAVA to apply the
overlay. Partition index 0 is typically the boot partition, while index 1 is
the root partition.

```yaml
- deploy:
    to: uuu
    images:
      system:
        url: https://example.com/imx-image-multimedia.rootfs.wic
        apply-overlay: true
        root_partition: 1
```

### sparse

System images shipped as sparse images require special handling with tools such
as `simg2img` and `img2simg` in order to apply LAVA overlays.

The default is `false`. Set `sparse: true` if the image is a sparse image:

```yaml
- deploy:
    to: uuu
    images:
      system:
        url: https://example.com/system.img.xz
        compression: xz
        sparse: true
        apply-overlay: true
```

### url

See [url](./index.md#url)

### compression

See [compression](./index.md#compression)

## uniquify

By default, LAVA stores each downloaded image in a separate subdirectory named
after the image key to avoid filename collisions. When needed, set
`uniquify: false` to store all images in the same directory. E.g., the
`uuu.lst` you are using may assume all artifacts in the same folder.

```yaml
- deploy:
    to: uuu
    uniquify: false
    images:
      boot:
        url: https://example.com/imx-boot-sd.bin-flash
      system:
        url: https://example.com/imx-image-multimedia.rootfs.wic
```

# Device-type template

The device-type template is a [jinja2][jinja2] template that will be extended
by [device dictionaries](./device-dictionary.md). The resulting file is the
device configuration file.

This [yaml][yaml] file is used by `lava-dispatcher` to know how to flash, boot
and communicate with a specific device.

## Configuration file

The device-type templates that are supported and provided by LAVA are stored
in `/usr/share/lava-server/device-types/`.

Admins could also provide their own device-type template using lavacli:

```shell
lavacli device-types template set qemu qemu.jinja2
```

Device-types provided by admins will be stored in
`/etc/lava-server/dispatcher-config/device-types/<name>.jinja2`.

LAVA will always look first in `/etc/lava-server/dispatcher-config/device-types/` and fallback to `/usr/share/lava-server/device-types/`.

This mean that admins can override every device-type templates, including the
ones provided by LAVA.

## Base templates

Every device-type template should inherit from `base.jinja2` which contains
many default structures.

For simplicity multiple base templates are provided for the most common bootloaders:

* `base-barebox.jinja2`
* `base-depthcharge.jinja2`
* `base-edk2.jinja2`
* `base-fastboot.jinja2`
* `base-grub.jinja2`
* `base-nxp-mcu.jinja2`
* `base-uboot.jinja2`

## Device types

For most device-types, extending a base template should be enough. Additional
requirements are documented below.

### DFU

`usb_vendor_id` and `usb_product_id` must be provided.

```jinja
{% extends "base-uboot.jinja2" %}

{% block body %}

board_id: '{{ board_id|default('LCES2-0x0000000000000000') }}'
usb_vendor_id: '045b'
usb_product_id: '0239'

{{ super() }}
{% endblock body %}
```

`deploy_dfu_commands` is required for entering DFU from u-Boot.

```jinja
{% set deploy_dfu_commands = deploy_dfu_commands|default(["nand erase.part fs1", "dfu"]) %}
```

### JLink

```jinja
  {% extends 'base-nxp-mcu.jinja2' %}

  {% set usb_vendor_id = '1366' %}
  {% set usb_product_id = '1024' %}

  {% set processor = 'MIMXRT1189XXX8' %}
  {% set supported_core_types = ['M33', 'M7'] %}

  {% set erase_command = ['erase', 'r', 'connect'] %}
  {% set reset_command = ['r'] %}

  {% set device_info = device_info|default([{'board_id': board_id, 'usb_vendor_id': usb_vendor_id, 'usb_product_id': usb_product_id}]) %}

  {% block jlink_options %}
  - '-if {{target_interface|default("SWD")}}'
  - '-speed {{ speed_ti|default(4000)}}'
  {% endblock jlink_options %}
```

You can configure the `erase_command` (default: `["erase"]`) and
`reset_command` (default: `["r"]`) parameters for each device type. These
parameters are mandatory for the jlink boot action, refer to the base
`base-nxp-mcu.jinja2` template for reference.

The `supported_core_types` parameter in the device type definition is
optional (default: `None`). It is used by JLink to connect to the board in cases
where the board has multiple cores, such as with the M33 and M7 cores. By
default, the connection is made using the first core in the list, which in this
case is the M33. To connect to the second core (e.g., M7), you need to use the
[`coretype`](../job-definition/actions/boot/method-jlink.md#coretype) parameter
in the JLink boot method.

!!! note
    The variables needed in `jlink_options` block should be provided in device
    dictionary.

### Raspberry Pi 4b

For the Raspberry Pi 4b, the device-type could look like the following. This
will only define some memory address and prompts that are specific to the board.

```jinja
{% extends 'base-uboot.jinja2' %}

{% set interrupt_ctrl_list = ['c'] %}
{% set action_timeout_bootloader_interrupt = '60' %}

{% set booti_kernel_addr = booti_kernel_addr|default('0x00080000') %}
{% set booti_ramdisk_addr = booti_ramdisk_addr|default('0x02700000') %}
{% set booti_dtb_addr = booti_dtb_addr|default('0x02400000') %}

{% set bootm_kernel_addr = bootm_kernel_addr|default('0x00080000') %}
{% set bootm_ramdisk_addr = bootm_ramdisk_addr|default('0x02700000') %}
{% set bootm_dtb_addr = bootm_dtb_addr|default('0x02400000') %}

{% set uboot_mkimage_arch = 'arm64' %}

{% set bootloader_prompt = bootloader_prompt|default('U-Boot>') %}
{% set console_device = console_device|default('ttyS1') %}
```

--8<-- "refs.txt"

# OpenOCD

The `openocd` boot method is used to flash and boot devices using the
[OpenOCD](https://openocd.org/) tool. It takes no parameters.

```yaml
- boot:
    method: openocd
    timeout:
      minutes: 10
```

The method works by passing through the command line options defined in the
device type for the `openocd` command. It flashes the executable specified by
`binary` in the prior job deploy action. If `openocd_script` is also specified
in the deploy action, it uses that script file instead of the one specified by
the device type.

!!!note
    Both `binary` and `openocd_script` files must be downloaded by the
    preceding deploy action.

`board_selection_cmd` can be used in the device type to specify a command to
pass the board id/serial number to OpenOCD. See OpenOCD documentation for
details on the command to set the serial number for the interface your device
type is using.

# Dragonboard 410c (fastboot)

This guide covers setting up devices that use the fastboot protocol for
deployment and booting in LAVA, such as the DragonBoard 410c.

LAVA uses `fastboot` to deploy images to the device partitions and to boot the
device. LAVA supports running fastboot directly from the worker host or within a
Docker container. The Docker option is recommended as it allows using custom
fastboot binaries and provides better isolation.

## Hardware setup

Ensure your device is connected to the LAVA worker properly and prepared for
fastboot deployment and booting.

1. [Setup power control](common.md#power-control).
2. [Setup serial console](common.md#serial-console).

    Connect the USB to serial adapter adapter to the DB410c's Low Speed
    Expansion Connector UART1:

    | DB410c Pin | Signal | USB-Serial Adapter |
    |------------|--------|--------------------|
    | Pin 1      | GND    | GND                |
    | Pin 11     | TXD    | RXD                |
    | Pin 13     | RXD    | TXD                |

    A picture of the FTDI cable connected to the DragonBoard 410c is shown below:

    ![db410c-uart1](./db410c-uart1.jpg)

3. Connect the device's micro USB OTG port to your LAVA worker USB port.
4. Setup USB Ethernet.

    The board does not have an Ethernet port. If you want to use a USB Ethernet
    adapter, additional setup is required.

    The DragonBoard 410c hardware design prevents simultaneous use of the micro
    USB (device mode) and USB Type A (host mode) ports; they are mutually
    exclusive.

    To support both fastboot (device mode) and USB Ethernet (host mode), you
    must be able to toggle the OTG connection. This is typically achieved by connecting
    the OTG port to the LAVA worker through a controllable smart USB hub. The
    commands for the switches can be provided via the `pre_power_command` and
    `pre_os_command` variables in device dictionary. In the job definition, you
    can use the variable names in the
    [`command`](../../../technical-references/job-definition/actions/command.md)
    action to issue the commands.

5. Set dip switch 2 to ON position for booting from SD card or eMMC.
6. If needed, flash or upgrade device FW and partition table. These
[artifacts](https://storage.lavacloud.io/artifacts/dragonboard-410c/) can be used.
7. Ensure that the device boots to fastboot mode on every power-on. For additional
instructions and troubleshooting, see
[dragonboard 41c guides](https://github.com/96boards/documentation/tree/master/consumer/dragonboard/dragonboard410c/guides).

In general, setting up a fastboot device with an LAA is much simpler. It is
essentially plug-and-play. For an example, see the
[DragonBoard 410c](https://docs.lavacloud.io/devices/dragonboard-410c.html#dragonboard-410c)
guide.

## Creating device type

[Create the device type](common.md#create-device-type) using the name
`dragonboard-410c`.

## Creating device

1. [Add the device](common.md#add-device) using the following settings:
    * **Device Type:** `dragonboard-410c`
    * **Hostname:** A unique name (e.g., `dragonboard-410c-01`)

2. [Add the device configuration](common.md#add-device-configuration).

    Device dictionary example for a DragonBoard 410c:

    ```jinja
    {% extends 'dragonboard-410c.jinja2' %}

    {% set adb_serial_number = 'a2c22e48' %}
    {% set fastboot_serial_number = 'a2c22e48' %}
    {% set device_info = [{'board_id': 'a2c22e48'}] %}

    {% set connection_command = 'telnet localhost 2002' %}

    {% set hard_reset_command = ['usbrelay 1_2=0', 'sleep 1', 'usbrelay 1_2=1'] %}
    {% set power_off_command = 'usbrelay 1_2=0' %}
    {% set power_on_command =  'usbrelay 1_2=1'%}
    {% set pre_power_command = 'usbrelay 1_3=1' %}
    {% set pre_os_command = 'usbrelay 1_3=0' %}

    {% set flash_cmds_order = ['update', 'ptable', 'partition', 'hyp', 'modem',
                               'rpm', 'sbl1', 'sbl2', 'sec', 'tz', 'aboot',
                               'boot', 'rootfs', 'vendor', 'system', 'cache',
                               'userdata', ] %}
    ```

    !!! note
        Update the example to use your device's actual serial number, serial
        connection and power control commands.

Another device dictionary example for
[dragonboard-410c](../../../technical-references/configuration/device-dictionary.md#dragonboard-410c).

For more about fastboot device configures, see the
[fastboot configurations](../../../technical-references/configuration/device-dictionary.md#fastboot).

## Sample job definition

```yaml
device_type: dragonboard-410c
job_name: db410c fastboot example

timeouts:
  job:
    minutes: 30

priority: medium
visibility: public

actions:
- command:
    name: pre_power_command
    timeout:
      minutes: 1

- deploy:
    to: fastboot
    docker:
      image: linaro/noble-adb-fastboot
      local: true
    images:
      boot:
        url: https://storage.lavacloud.io/health-checks/dragonboard-410c/boot-linaro-buster-359.img.gz
        compression: gz
      rootfs:
        url: https://storage.lavacloud.io/health-checks/dragonboard-410c/rootfs-linaro-buster-359.img.gz
        compression: gz
        apply-overlay: true
    timeout:
      minutes: 15

- command:
    name: pre_os_command
    timeout:
      minutes: 1

- boot:
    method: fastboot
    docker:
      image: linaro/noble-adb-fastboot
      local: true
    prompts:
    - 'root@linaro-developer:~#'
    timeout:
      minutes: 5

- test:
    definitions:
    - from: inline
      repository:
        metadata:
          format: Lava-Test Test Definition 1.0
          name: test-definition-example
        run:
          steps:
          - lava-test-case run-uname-a --shell uname -a
          - lava-test-case check-os-id --shell 'cat /etc/os-release | grep "ID=debian"'
      path: inline/test-definition-example.yaml
      name: test-suite-example
    timeout:
      minutes: 5
```

--8<-- "refs.txt"

# NFS

The `nfs` deployment method is used to deploy a rootfs and optional kernel
modules to the NFS server that running on the LAVA worker.

## nfsrootfs

A compressed tarball containing the rootfs.

```yaml
deploy:
  to: nfs
  nfsrootfs:
    url: https://example.com/rootfs.tar.xz
    compression: xz
```

## modules

Optional. A compressed tarball containing kernel modules.

```yaml
deploy:
  to: nfs
  nfsrootfs:
    url: https://example.com/rootfs.tar.xz
    compression: xz
  modules:
    url: https://example.com/modules.tar.xz
    compression: xz
```

### overlays

The `nfsrootfs` action supports overlays with the `format` specified.

See [overlays](./index.md#overlays)

## images

Alternatively, you can specify a list of images to deploy. Here is an example
`nfs` deploy action for the `qemu-nfs` boot method.

```yaml
- deploy:
    to: nfs
    images:
      nfsrootfs:
        url: http://example.com/jessie-arm64-nfs.tar.gz
        image_arg: 'nfsroot={NFS_SERVER_IP}:{nfsrootfs},tcp,hard'
        compression: gz
      kernel:
        url: http://example.com/vmlinuz-4.9.0-2-arm64
        image_arg: -kernel {kernel}
      initrd:
        url: http://example.com/initrd.img-4.9.0-2-arm64
        image_arg: -initrd {initrd}
```

### image_arg

See [image_arg](./to-tmpfs.md#image_arg)

# NBD (Network Block Device)

The `nbd` deployment method is used to support NBD root deployments.  Files are
downloaded to a temporary directory on the worker. The rootfs is shared through
`xnbd-server` and the filenames are substituted into the
[bootloader commands](../boot/common.md#commands) specified in the device
configuration or overridden in the job definition.

```yaml
- deploy:
    to: nbd
    kernel:
      url: http://example.com/vmlinuz
      type: zimage
    initrd:
      url: http://example.com/initramfs.ext4.gz.u-boot
    nbdroot:
      url: http://example.com/rootfs.ext4.xz
      compression: xz
    dtb:
      url: http://example.com/dtb.dtb
```

## kernel

Required. The kernel to boot on the device.

### type

Specifies the type of kernel being downloaded. Supported types:

- `image`
- `uimage`
- `zimage`

## initrd

The initrd contains all necessary files, daemons and scripts to bring-up the
`nbd-client` and `pivot_root` to the final rootfs.

!!!note
    The nbdroot filesystem will not be modified prior to the boot. Since the
    filesystems are using security labels and the modification would alternate
    the FS. So the [LAVA overlay](./index.md#lava-overlay) needs to be transferred
    after the boot with [`transfer_overlay`](../boot/common.md#transfer_overlay).

## nbdroot

The root filesystem image shared over xnbd-server.

### url

See [url](./index.md#url)

### compression

The NBD filesystem image is unpacked into a temporary directory on the LAVA
worker in a location supported by the NBD server.

The compression method **must** be specified if the file is compressed. See
[compression](./index.md#compression).

## dtb

Optional Device Tree Blob.

## modules

Modules are not supported in the NBD deployment method. Modules must be part of
the filesystem already.

## Example job

```yaml
# NBD root deployment
job_name: standard Debian ARMMP nbd test on bbb
device_type: beaglebone-black

priority: medium
visibility: public

timeouts:
  job:
    minutes: 10
  action:
    minutes: 5
  connection:
    minutes: 2

actions:
# NBD_DEPLOY_BLOCK
- deploy:
    timeout:
      minutes: 5
    to: nbd
    kernel:
      url: http://example.com/vmlinuz
      type: zimage
    initrd:
      url: http://example.com/initramfs.ext4.gz.u-boot
    nbdroot:
      url: http://example.com/rootfs.ext4.xz
      compression: xz
    dtb:
      url: http://example.com/dtb.dtb

# NBD_BOOT_BLOCK
- boot:
    method: u-boot
    commands: nbd
    auto_login:
      login_prompt: 'login:'
      username: root
    prompts:
    - 'root@jessie:'
    timeout:
      minutes: 5

- test:
    definitions:
    - repository: https://github.com/Linaro/test-definitions
      from: git
      path: automated/linux/smoke/smoke.yaml
      name: smoke-tests
    timeout:
      minutes: 5
```

# DFU

The `dfu` boot method flashes and boots devices using the
[USB DFU (Device Firmware Upgrade)](https://en.wikipedia.org/wiki/USB#Device_Firmware_Upgrade_mechanism)
protocol.

```yaml
- boot:
    method: dfu
    timeout:
      minutes: 10
```

!!! note
    Images should be deployed using the [`tmpfs`](../deploy/to-tmpfs.md) deploy
    action before using this boot method. Each image entry in the deploy action
    should include an `image_arg` containing the DFU `--alt` and `--download`
    arguments.

## Installation

The boot method invokes `dfu-util` on the worker to download images to the
device over USB. The utility must be installed on the LAVA worker.

```shell
sudo apt-get install dfu-util
```

## Configuration

To add a new device, see [DFU](../../../configuration/device-dictionary.md#dfu).

To add a new device type, see [DFU](../../../configuration/device-type-template.md#dfu).

## Example jobs

### Hardware DFU

```yaml
device_type: arduino101
job_name: zephyr dfu

timeouts:
  job:
    minutes: 15
  action:
    minutes: 5
  actions:
    lava-test-monitor:
      seconds: 300
    wait-usb-device:
      seconds: 20
    flash-dfu:
      seconds: 60
  connections:
    lava-test-monitor:
      seconds: 300

priority: medium
visibility: public

actions:
- deploy:
    timeout:
      minutes: 3
    to: tmpfs
    images:
      app:
        image_arg: --alt x86_app --download {app}
        url: https://storage.lavacloud.io/health-checks/arduino_101/app_kernel-zephyr.bin
      sensor_core:
        image_arg: --alt sensor_core --download {sensor_core}
        url: https://storage.lavacloud.io/health-checks/arduino_101/sensor_core-arc.bin
      ble_core:
        image_arg: --alt ble_core --download {ble_core}
        url: https://storage.lavacloud.io/health-checks/arduino_101/ble_core-image.bin

- boot:
    method: dfu
    timeout:
      minutes: 10

- test:
    monitors:
    - name: simple-service
      start: S I M P L E   S E R V I C E    M E A S U R E M E N T S
      end: M A I L B O X   M E A S U R E M E N T S
      pattern: '\| (?P<test_case_id>[a-z ]+) +\| +(?P<measurement>\d+)\|'
```

### U-Boot DFU

```yaml
device_type: rzn1
job_name: flash rzn1 using u-boot dfu

timeouts:
  job:
    minutes: 10
  action:
    minutes: 5
  connection:
    minutes: 2
priority: medium
visibility: public

actions:
- deploy:
    namespace: dfu
    to: tmpfs
    images:
      sf_fsbl:
        image_arg: --alt sf_fsbl --download {sf_fsbl}
        url: http://.../rzn1d-snarc-fsbl-secmon.img
      sf_uboot0:
        image_arg: --alt sf_uboot0 --download {sf_uboot0}
        url: http://.../u-boot-lces2-ddr.itb
      sf_uboot1:
        image_arg: --alt sf_uboot1 --download {sf_uboot1}
        url: http://.../u-boot-lces2-ddr.itb
      n_kernel1:
        image_arg: --alt n_kernel1 --download {n_kernel1}
        url: http://.../fitImage-1.0-r0-rzn1-snarc.itb

- command:
    namespace: dfu
    name: set_boot_to_nand

- boot:
    namespace: dfu
    method: dfu
    timeout:
      minutes: 10

- command:
    namespace: test
    name: set_boot_to_qspi

- deploy:
    namespace: test
    to: overlay

- boot:
    namespace: test
    connection-namespace: dfu
    method: bootloader
    bootloader: u-boot
    prompts: ["=>"]
    commands: ["run linux_bestla"]

- boot:
    namespace: test
    timeout:
      minutes: 5
    method: minimal
    reset: false
    auto_login:
      login_prompt: 'login:'
      username: 'root'
      password_prompt: "Password:"
      password: "P@ssword-1"
      login_commands:
      - 'P@ssword-1'
      - 'azertAZERT12345'
      - 'azertAZERT12345'
      - 'azertAZERT12345'
    prompts:
    - "root@rzn1-snarc:~# "
    - "root@rzn1-snarc:/tmp# "
    - "Current password: "
    - "New password: "
    - "Retype new password: "
    transfer_overlay:
      download_command: unset http_proxy ; dhclient eth1 ; cd /tmp ; wget
      unpack_command: tar -C / -xzf

- test:
    namespace: test
    definitions:
    - repository: https://github.com/Linaro/test-definitions
      from: git
      path: automated/linux/busybox/busybox.yaml
      name: busybox
    timeout:
      minutes: 5
```

# USB Gadget Mass Storage

The `usbg-ms` deployment method exposes the download disk image to the DUT as a
USB mass storage device using the
[Linux USB Mass Storage Gadget](https://docs.kernel.org/usb/mass-storage.html).
USB mass storage emulation support is required on the LAVA worker, e.g.,
[using the LAA USB OTG port](https://docs.lavacloud.io/hardware/peripherals.html#laa-usb).
For the DUT, the emulated virtual device functions as a regular block device,
similar to a USB stick.

Note that the [usbg_ms_commands](../../../configuration/device-dictionary.md#usbg-ms)
device configuration is required for using the deployment method.

```yaml
- deploy:
    to: usbg-ms
    image:
      url: https://raspi.debian.net/tested/20231109_raspi_4_bookworm.img.xz
      compression: xz
    timeout:
      minutes: 5
```

## image

### url

See [url](./index.md#url).

### compression

See [compression](./index.md#compression).

If the image is compressed, the compression method **must** be specified.

## overlays

For applying overlays to the image, see [Overlays](./index.md#overlays).

The [LAVA overlay](./index.md#lava-overlay) is required when the job definition
contains tests.

## timeout

See [timeout](../../timeouts.md)

## Life cycle

By default, the gadget is removed at the end of the job. You can use the built-in
`usbg_ms_commands_disable` command to remove it earlier.

```yaml
- command:
    name: usbg_ms_commands_disable
```

See [RPi4b secondary media deployment example](./to-secondary.md#sample-job) for
how it is used in a pipeline.

## Sample job

This sample job demonstrates a highly efficient way to boot and test a disk image
via USB boot, as emulating a USB device is much faster than flashing a physical
USB disk.

```yaml
job_name: RPi4B USBG MS sample job
device_type: bcm2711-rpi-4-b

visibility: public
priority: medium

timeouts:
  job:
    minutes: 15
  connection:
    minutes: 2

actions:
- deploy:
    to: usbg-ms
    image:
      url: https://raspi.debian.net/tested/20231109_raspi_4_bookworm.img.xz
      compression: xz
      format: ext4
      partition: 1
      overlays:
        lava: true
    timeout:
      minutes: 5

- boot:
    method: minimal
    auto_login:
      login_prompt: 'login:'
      username: root
    prompts:
    - 'root@rpi4-20231108:'
    timeout:
      minutes: 5

- test:
    definitions:
    - from: inline
      repository:
        metadata:
          format: Lava-Test Test Definition 1.0
          name: test-definition-example
        run:
          steps:
          - lava-test-case run-uname-a --shell uname -a
          - lava-test-case check-os-id --shell 'cat /etc/os-release | grep "ID=debian"'
      path: inline/test-definition-example.yaml
      name: test-suite-example
    timeout:
      minutes: 5
```

--8<-- "refs.txt"

# Secondary media (USB / SATA / SD)

The `usb / sata / sd` deployment methods can be used to deploy an image to a
secondary media device on the DUT. The three methods use the same deployment
strategy. The only difference is which media is configured in the
[device dictionary](../../../configuration/device-dictionary.md#secondary-media)
and referenced in the job definition.

The DUT is first booted into a primary OS from a primary media device that contains
a pre-installed OS, or via netboot (for example, U-Boot ramdisk or NFS). Once the
primary OS is running, LAVA invokes `download` and `writer` tools on the DUT to
fetch the image that already downloaded to the LAVA worker and write it to the
secondary media device.

This guide describes how to write an image to a secondary media device using the
default `dd` tool and the custom `bmaptool` tool. Other download and writer tools
can be used similarly.

## dd

```yaml
- deploy:
    to: sd
    image:
      url: http://example.com/rootfs.img.gz
      compression: gz
      root_partition: 1
    device: sdcard
    download:
      tool: /usr/bin/wget
      options: --no-check-certificate --no-proxy --connect-timeout=30 -S --progress=dot:giga -O - {DOWNLOAD_URL}
      prompt: HTTP request sent, awaiting response
    timeout:
      minutes: 15
```

See also [SD device configuration](../../../configuration/device-dictionary.md#sd).

### image

A single disk image. Mutually exclusive with [`images`](#images).

#### url

See [url](./index.md#url).

#### compression

See [compression](./index.md#compression).

If the image is compressed and the writing tool doesn't support using a compressed
file, the compression method must be specified.

#### root_partition

When tests defined in the job definition, the `root_partition` index (0-based) of
the image must be specified for applying the LAVA overlay that contains the tests
and LAVA helpers.

### device

Required. The name of the secondary media device configured using the
[`<media>_label`](../../../configuration/device-dictionary.md#secondary-media) in
the device dictionary.

### download

Specifies how to download and pipe the image to the writing tool for flashing the
secondary device.

#### tool

Required. Absolute path to the download tool on the primary OS.

The tool must be pre-installed:

```shell title="Debian"
sudo apt install wget
```

#### options

Required. Options for the download tool.

LAVA substitutes the placeholder `{DOWNLOAD_URL}` with the HTTP URL served by
the LAVA worker for the downloaded (and decompressed if specified) image file.
For example:

```plain
http://198.18.0.1/tmp/529/storage-deploy-23634l0i/image/ubuntu-24.04.4-preinstalled-server-arm64+raspi.img
```

#### prompt

Required. A string that appears in the tool's output when the download starts. LAVA
waits for this string first before waiting for write completion prompts.

### writer

If not defined, `dd` is used for the flashing by default. The full pipeline sent
to the primary OS shell is:

```shell
/usr/bin/wget <options> | dd of='/dev/disk/by-id/<media>_uuid' bs=4M
```

For example:

```shell
/usr/bin/wget --no-check-certificate --no-proxy --connect-timeout=30 \
  -S --progress=dot:giga \
  -O - http://198.18.0.1/tmp/529/storage-deploy-23634l0i/image/ubuntu-24.04.4-preinstalled-server-arm64+raspi.img \
  | dd of='/dev/disk/by-id/mmc-SE16G_0x5005f804' bs=4M
```

### tool

Optional. Overrides the prompts LAVA uses to detect write completion.

#### prompts

If omitted, LAVA uses the default `DD_PROMPTS`:

```yaml
- "[0-9]+\+[0-9]+ records out"
- "[0-9]+ bytes \(.*\) copied"
```

## bmaptool

```yaml
- deploy:
    to: usb
    images:
      image:
        url: https://example.com/raspberrypi4-64/oniro-image-base-tests-raspberrypi4-64.rootfs.wic.gz
      bmap:
        url: https://example.com/raspberrypi4-64/oniro-image-base-tests-raspberrypi4-64.rootfs.wic.bmap
    uniquify: false
    device: SanDiskCruzerBlade
    writer:
      tool: /usr/bin/bmaptool
      options: copy {DOWNLOAD_URL} {DEVICE}
      prompt: 'bmaptool: info'
    tool:
      prompts: ['bmaptool: info: copying time: [0-9hms\.\ ]+, copying speed [0-9\.]+ [MKiBbytes]+\/sec']
    timeout:
      minutes: 30
```

See also [USB device configuration](../../../configuration/device-dictionary.md#usb).

### images

A dictionary of named files. Mutually exclusive with [`image`](#image). Must
contain the `image` key. Any additional keys (e.g. `bmap`) are downloaded as
companion files.

### uniquify

Optional boolean. By default, LAVA stores each downloaded image in a separate
subdirectory named after the image key to avoid filename collisions.

For this case, it must be set to `false` so that `bmaptool` can locate the
corresponding `bmap` file to accelerate the flashing. The `bmap` file must
share the same base name and reside in the same directory as the `image`.

### device

See [device](#device).

### writer

Specifies the image writing tool.

#### tool

Required. Absolute path to the writer tool on the primary OS.

The tools must be pre-installed:

```shell title="Debian"
sudo apt install bmap-tools
```

#### options

Required. Options for the writer tool.

Placeholders:

- `{DOWNLOAD_URL}` — See [options](#options). Additionally, the `bmaptool` fetches
  the `image` and the corresponding `bmap` files directly from the LAVA worker
  over HTTP. This means download tool is not needed in this case.
- `{DEVICE}` — LAVA substitutes the placeholder with the resolved block device
  path on the primary OS, e.g.
  `/dev/disk/by-id/usb-SanDisk_Cruzer_Blade_20060266531DA442AD42-0:0`. The
  `usb-SanDisk_Cruzer_Blade_20060266531DA442AD42-0:0` is configured using
  `usb_uuid` parameter in the device dictionary.

#### prompt

Required. The string that appears in the tool's output when the write operation
start. LAVA waits for this string before waiting for write completion prompts.

### tool

Optional. Overrides the prompts LAVA uses to detect write completion.

#### prompts

Overwrites the default [`DD_PROMPTS`](#prompts)

#### LAVA overlay

When using `bmaptool` to flash an image with a corresponding `bmap` file, the
LAVA overlay should be transferred after the secondary device flashing and boot
since the image shouldn't be modified before flashing.

```yaml
- boot:
    method: minimal
    auto_login:
      login_prompt: 'login:'
      username: root
    prompts:
      - 'root@device:~#'
    transfer_overlay:
      download_command: wget
      unpack_command: tar -C / -xzf
    timeout:
      minutes: 5
```

See also [transfer overlay](../boot/common.md#transfer_overlay).

## Sample job

```yaml
job_name: RPi4b secondary media deployment sample job
device_type: bcm2711-rpi-4-b

visibility: public
priority: medium

timeouts:
  job:
    minutes: 15
  connection:
    minutes: 2

actions:
# Attach USB mass storage gadget.
- deploy:
    to: usbg-ms
    image:
      url: https://raspi.debian.net/tested/20231109_raspi_4_bookworm.img.xz
      compression: xz
    timeout:
      minutes: 5

# Boot primary OS - Debian.
- boot:
    method: minimal
    auto_login:
      login_prompt: 'login:'
      username: root
    prompts:
    - 'root@rpi4-20231108:'
    timeout:
      minutes: 5

# 'wget | dd' Ubuntu image to secondary media SD Card
- deploy:
    timeout:
      minutes: 15
    to: sd
    image:
      url: https://cdimage.ubuntu.com/releases/24.04.4/release/ubuntu-24.04.4-preinstalled-server-arm64+raspi.img.xz
      compression: xz
      root_partition: 1
    device: sdcard
    download:
      tool: /usr/bin/wget
      options: --no-check-certificate --no-proxy --connect-timeout=30 -S --progress=dot:giga -O - {DOWNLOAD_URL}
      prompt: HTTP request sent, awaiting response

# Detach USB mass storage gadget.
- command:
    name: usbg_ms_commands_disable

# Boot Ubuntu from SD Card.
- boot:
    method: minimal
    auto_login:
      login_prompt: 'login:'
      username: 'ubuntu'
      password_prompt: 'Password:'
      password: 'ubuntu'
      login_commands:
      - ubuntu
      - linaro123
      - linaro123
      - sudo su
    prompts:
    - 'Current password:'
    - 'New password:'
    - 'Retype new password:'
    - 'ubuntu@ubuntu:'
    - 'root@ubuntu:'
    timeout:
      minutes: 5

# Run tests to check kernel version and OS ID.
- test:
    definitions:
    - from: inline
      repository:
        metadata:
          format: Lava-Test Test Definition 1.0
          name: test-definition-example
        run:
          steps:
          - lava-test-case run-uname-a --shell uname -a
          - lava-test-case check-os-id --shell 'cat /etc/os-release | grep "ID=ubuntu"'
      path: inline/test-definition-example.yaml
      name: test-suite-example
    timeout:
      minutes: 5
```

This job example utilizes the
[USB mass storage gadget](https://docs.lavacloud.io/hardware/peripherals.html#laa-usb)
feature of the LAA as the primary boot media to launch Debian. Once Debian is running,
it is used to flash the Ubuntu image onto the SD card. The pipeline can be useful
for testing if a disk image boots correctly from the SD card.

See also:

- [Prioritize RPi USB boot over SD](https://docs.lavacloud.io/devices/rpi-3-and-4.html#setup-a-simple-recovery-mechanism)
- [usbg-ms deployment method](./to-usbg-ms.md)

--8<-- "refs.txt"

# Running the LAVA dispatcher in a Docker container

Running a LAVA dispatcher in a Docker container requires quite a few special
options, to ensure that the dispatcher has access to system resources that
are, in most Docker use cases, invisible to containers. This is automated by
the `lava-docker-worker` program, provided by the `lava-dispatcher-host`
package.

`lava-docker-worker` should support all use cases that the regular LAVA worker
supports In particular, `lava-docker-work` **does support** running other
docker containers, QEMU/KVM, and other regular device types, as long as the
required user space tools are included in the LAVA worker docker image.

## 1: install and configure lava-dispatcher-host

**lava-dispatcher-host needs to be installed in the host machine** (bare metal)
for the LAVA dispatcher to work inside a container, as well as Docker itself.

### Debian package

If you are using the Debian package, just installing it will do all necessary
setup.

### Other means

If you are installing lava-dispatcher-host in some other way, you need to run
the following command, as root:

```
# lava-dispatcher-host rules install
```

This step needs to be repeated when lava-dispatcher-host is upgraded, to make
sure you have the most up to date udev rules file for it.

## 2: start lava-docker-worker

### Debian package

To run the Docker Dispatcher, first edit
`/etc/lava-dispatcher-host/lava-docker-worker` and set the connection
parameters. At least URL and TOKEN are probably needed:

```
WORKER_NAME="--name myworker"

# ...

# Server connection
URL="https://my.lava.server/"
TOKEN="--token 0123456789012345678901234567890123456789"
```

Make sure Docker get installed from Docker's official repo. At the time of
writing this doc, Docker version `18.09.1` installed from Debian's official
repo doesn't work with `lava-docker-worker`.

```
curl -fsSL https://get.docker.com -o get-docker.sh
sh get-docker.sh
```

??? tip "Debian Testing/Sid User"
    Install `docker.io` instead if the above docker CE installation script
    doesn't support your distribution.

    ```shell
    apt-get install docker.io
    ```

Then enable and start the `lava-docker-worker` service:

```
sudo systemctl enable lava-docker-worker.service
sudo systemctl start lava-docker-worker.service
```

### Other means

If you installed LAVA in a way that doesn't provide a lava-docker-worker
systemd service, you can just run `lava-docker-worker` as any user that can run
docker. You need to pass the relevant arguments to have it connect to your LAVA
server:

```
$ lava-docker-worker --url https://my.lavaserver.com/ --token 0123456789012345678901234567890123456789
```

## 3: build customized image

To build customized image for additional packages you might require, first edit
`/etc/lava-dispatcher-host/lava-docker-worker` and set the BUILD_DIR parameter
to a directory with a Dockerfile inside.

```
# Docker build directory
# Should be set to a directory with a Dockerfile inside.
BUILD_DIR="/etc/lava-dispatcher-host/build"
```

Then restart the `lava-docker-worker` service.

```
sudo systemctl restart lava-docker-worker.service
```

!!! note "Base image"

    A new `Dockerfile.lava` will be generated with the base image that matches
    your LAVA server version for building your customized image. All the `FROM`
    commands defined in your `Dockerfile` will be ignored.

## 4: provide custom tools in the container

If either of `/usr/share/lava-docker-worker` or
`/usr/local/share/lava-docker-worker` exists, they are bind monted into the
container by default, and if there is a `bin` subdirectory in any of then, that
gets added to the default $PATH.

Additionally, you can mount other host directories inside the container by
using the --mount option.  Edit `/etc/lava-dispatcher-host/lava-docker-worker`
and set the MOUNTS option:

```
MOUNTS="--mount /path/to/tools/bin:/usr/local/bin"
```

# Signing in

Each LAVA instance offers a local user account and may also be configured with
an alternative authentication method such as LDAP or social accounts.

LDAP authentication relies on local configuration and may involve obtaining a
privileged access token from the LDAP administrators. Local accounts remain
available for all instances.

Whichever authentication method is used, there is no difference in how users
interact with LAVA once signed in.

Click the "Sign In" button in the top-right corner to access the login screen.
It looks something like the following.

![User login](./user-login.png)

## Local user account

A local user account may be created by an administrator within LAVA. The
administrator may grant different permissions to the user to perform various
operations within LAVA.

To log in with a username and password, enter your credentials in the provided
text boxes and click the "Login" button.

Local user accounts can be particularly useful for automated users like build
systems or continuous integration scripts using the LAVA APIs. In this case,
a local admin can create the user with a secure password, sign in as that user
to create an authentication token and then supply the token to the scripts.

## Linaro lab users

Linaro's central validation lab in Cambridge is accessible via
[validation.linaro.org](https://validation.linaro.org/). It is Linaro's core
production instance of LAVA.

This lab is configured to use Linaro LDAP. To sign in, enter your details in the
LDAP fields.

!!! note
    If your Linaro email is `first.second@linaro.org` then use `first.second`
    as your username.

A successful authentication will redirect you to the LAVA dashboard.

![LAVA dashboard](./lava-dashboard.png)

# MultiNode job definition

A LAVA MultiNode job is a single LAVA job definition that runs multiple jobs
across multiple devices. It is implemented via the
[MultiNode](../../technical-references/job-definition/protocols.md#multinode)
protocol that allows the jobs to communicate via the protocol API and requests.

## How to avoid it

While the MultiNode support in LAVA is powerful and mature, it also introduces
significant complexity for writing and debugging the job definitions. We
recommend avoiding MultiNode unless it is truly required by your use case.

Since the introduction of the [docker test shell](./running-arbitrary-code-with-docker.md)
for single node jobs, MultiNode is no longer required in many scenarios.
Communicating with a DUT via a Docker container within a single node job is a
much simpler and more reliable approach. Give it a try before you start working
with the MultiNode.

## Dependency

[LAVA coordinator](../../technical-references/services/lava-coordinator.md)
service should be configured and reachable from all the workers involved in the
MultiNode jobs.

## Coordinating via APIs

The following example job uses the
[MultiNode APIs](../../technical-references/job-definition/protocols.md#api)
from the LAVA test shell to pass messages between the iPerf server and client for
starting the test.

```yaml hl_lines="71 72 125 128"
job_name: MultiNode iPerf test

visibility: public
priority: medium

timeouts:
  job:
    minutes: 30
  action:
    minutes: 5
  connection:
    minutes: 2
  actions:
    power-off:
      seconds: 30

# Define roles
protocols:
  lava-multinode:
    roles:
      server:
        device_type: ssh
        count: 1
      client:
        device_type: bcm2711-rpi-4-b
        count: 1
    timeout:
      minutes: 10

actions:
# Server actions
- deploy:
    role:
    - server
    to: ssh
    os: debian
    timeout:
      minutes: 5

- boot:
    role:
    - server
    method: ssh
    prompts:
    - 'root@x13:~#'
    timeout:
      minutes: 5

- test:
    role:
    - server
    protocols:
      lava-multinode:
      - action: multinode-test
        request: lava-wait
        messageID: start
        timeout:
          minutes: 10
    definitions:
    - from: inline
      repository:
        metadata:
          format: Lava-Test Test Definition 1.0
          name: iperf-server
        run:
          steps:
          - iperf -s &
          - echo $! > /tmp/iperf-server.pid
          - IP=$(ip route get 8.8.8.8 | grep -oP 'src \K\S+')
          - lava-test-case get-server-ip --shell echo "Server IP is ${IP}"
          - lava-send server-ready server_ip=${IP}
          - lava-wait client-done
          - kill -9 $(cat /tmp/iperf-server.pid)
      name: iperf-server
      path: inline/iperf-server.yaml
    timeout:
      minutes: 15

# Client actions
- deploy:
    role:
    - client
    to: usbg-ms
    image:
      url: http://198.18.0.1/tmp/images/rpi4/20231109_raspi_4_bookworm.img
      format: ext4
      partition: 1
      overlays:
        lava: true
    timeout:
      minutes: 5

- boot:
    role:
    - client
    method: minimal
    auto_login:
      login_prompt: 'login:'
      username: root
    prompts:
    - 'root@rpi4-20231108:'
    timeout:
      minutes: 5

- test:
    protocols:
      lava-multinode:
      - action: multinode-test
        request: lava-send
        messageID: start
        timeout:
          minutes: 10
    role:
    - client
    definitions:
    - from: inline
      repository:
        metadata:
          format: Lava-Test Test Definition 1.0
          name: iperf-client
        run:
          steps:
          - apt-get update -q
          - lava-test-case install-iperf --shell apt-get -q -y install iperf
          - lava-wait server-ready
          - SERVER_IP=$(cat /tmp/lava_multi_node_cache.txt | cut -d = -f 2)
          - lava-test-case iperf-to-server --shell iperf -c ${SERVER_IP} -t 10
          - lava-send client-done
      name: iperf-client
      path: inline/iperf-client.yaml
    timeout:
      minutes: 15
```

## Coordinating via requests

The following example job uses the
[MultiNode protocol requests](../../technical-references/job-definition/protocols.md#request)
across multiple job actions to configure the faster server waits for the slower
client.

```yaml hl_lines="36-42 109-116"
job_name: MultiNode iPerf test - wait for client

visibility: public
priority: medium

timeouts:
  job:
    minutes: 30
  action:
    minutes: 5
  connection:
    minutes: 2
  actions:
    power-off:
      seconds: 30

# Define roles
protocols:
  lava-multinode:
    roles:
      server:
        device_type: ssh
        count: 1
      client:
        device_type: bcm2711-rpi-4-b
        count: 1
    timeout:
      minutes: 10

actions:
# Server actions
- deploy:
    role:
    - server
    # Wait for the client.
    protocols:
      lava-multinode:
      - action: scp-overlay
        request: lava-wait
        messageID: start
        timeout:
          minutes: 10
    to: ssh
    os: debian
    timeout:
      minutes: 5

- boot:
    role:
    - server
    method: ssh
    prompts:
    - 'root@x13:~#'
    timeout:
      minutes: 5

- test:
    role:
    - server
    definitions:
    - from: inline
      repository:
        metadata:
          format: Lava-Test Test Definition 1.0
          name: iperf-server
        run:
          steps:
          - iperf -s &
          - echo $! > /tmp/iperf-server.pid
          - IP=$(ip route get 8.8.8.8 | grep -oP 'src \K\S+')
          - lava-test-case get-server-ip --shell echo "Server IP is ${IP}"
          - lava-send server-ready server_ip=${IP}
          - lava-wait client-done
          - kill -9 $(cat /tmp/iperf-server.pid)
      name: iperf-server
      path: inline/iperf-server.yaml
    timeout:
      minutes: 15

# Client actions
- deploy:
    role:
    - client
    to: usbg-ms
    image:
      url: http://198.18.0.1/tmp/images/rpi4/20231109_raspi_4_bookworm.img
      format: ext4
      partition: 1
      overlays:
        lava: true
    timeout:
      minutes: 5

- boot:
    role:
    - client
    method: minimal
    auto_login:
      login_prompt: 'login:'
      username: root
    prompts:
    - 'root@rpi4-20231108:'
    timeout:
      minutes: 5

- test:
    role:
    - client
    protocols:
      lava-multinode:
      - action: multinode-test
        # Notify sever to start.
        request: lava-send
        messageID: start
        timeout:
          minutes: 10
    definitions:
    - from: inline
      repository:
        metadata:
          format: Lava-Test Test Definition 1.0
          name: iperf-client
        run:
          steps:
          - apt-get update -q
          - lava-test-case install-iperf --shell apt-get -q -y install iperf
          - lava-wait server-ready
          - SERVER_IP=$(cat /tmp/lava_multi_node_cache.txt | cut -d = -f 2)
          - lava-test-case iperf-to-server --shell iperf -c ${SERVER_IP} -t 10
          - lava-send client-done
      name: iperf-client
      path: inline/iperf-client.yaml
    timeout:
      minutes: 15
```

# Job definition

The [job definition](../../technical-references/job-definition/job.md) is a
`yaml` file that you submit to the LAVA server. This file is describing everything
that LAVA needs to know to run your tests on a DUT.

## QEMU example

Let's look at this example:

```yaml
--8<-- "jobs/qemu.yaml"
```

## Structure

The job definition is made of:

* `device_type`: requested device-type
* `job_name`: name of the job, sometime called `description`
* `timeouts`: [timeout definition](../../technical-references/job-definition/timeouts.md)
* `priority`: [job priority](../../technical-references/job-definition/job.md#priority)
* `visibility`: [job visibility](../../technical-references/job-definition/job.md#visibility)
* `environment`: [job environment](../../technical-references/job-definition/environment.md)
* `actions`: list of actions to run

In this definition, we request to run a job called `simple qemu tjob` on a `qemu` DUT.
The job is visible to everyone (including anonymous users).
The job will not run for more than 10 minutes at `medium` priority.

## Actions

`actions` is a list of actions that LAVA will have to execute for the given
job.

Currently, LAVA support four type of actions:

* [command](../../technical-references/job-definition/actions/command.md): run commands on the dispatcher
* [deploy](../../technical-references/job-definition/actions/deploy/index.md): deploy software on the DUT
* [boot](../../technical-references/job-definition/actions/boot/common.md): boot the DUT
* [test](../../technical-references/job-definition/actions/test.md): run some tests

### Deploy

The deploy action will deploy the software on the DUT, using the method
specified by the `to`.

In this example, we are requesting to deploy to `tmpfs`. The `rootfs` will be
downloaded from the given url.

!!! info "deployment methods"
    LAVA supports many deployment methods. However, only some methods are
    supported by each device-type.

### Boot

The boot action will boot the DUT, using the method specified by the `method`.

In this example, we are requesting to boot `qemu` using the software from
`tmpfs`.

!!! info "boot methods"
    LAVA supports many boot methods. However, only some methods are supported
    by each device-type.

### Test

The test action will run the given tests on the DUT.

In this example, LAVA will download the [test definitions](./test-definition.md) from two git repositories.

!!! info "test methods"
    LAVA supports three test methods that do not depend on the device-type.

--8<-- "refs.txt"

# Dispatcher configuration

The dispatcher configuration files allow you to configure some of the behavior
of the dispatchers.

Available files are:

* dispatcher configuration
* environment
* DUT environment

## Dispatcher configuration

Environment configuration file is located on the server in:

* `/etc/lava-server/<hostname>.yaml` old configuration file
* `/etc/lava-server/dispatcher.d/<hostname>/dispatcher.yaml` new configuration file

When loading the configuration, LAVA will look at the new configuration file
and fallback to the old one

```yaml
--8<-- "config/dispatcher.yaml"
```

## Environment

The dispatcher environment is used to set the process environment when spawning
`lava-run`.

Environment configuration file is located on the server in:

* `/etc/lava-server/env.yaml` default, for every dispatcher
* `/etc/lava-server/dispatcher.d/<hostname>/env.yaml` for a specific dispatcher

When loading the configuration, LAVA will look at the dispatcher specific
configuration and fallback to the default configuration.

```yaml
--8<-- "config/env.yaml"
```

## DUT environment

The DUT environment is used to set some environment variable on the DUT when
running tests.

Environment configuration file is located on the server in:

* `/etc/lava-server/env-dut.yaml` default, for every dispatcher
* `/etc/lava-server/dispatcher.d/<hostname>/env-dut.yaml` for a specific dispatcher

When loading the configuration, LAVA will look at the dispatcher specific
configuration and fallback to the default configuration.

The format is the same as for the environment file.

--8<-- "refs.txt"

# Musca

The `musca` deployment method allows deployment of software to Musca devices.
Currently, supported Musca devices are:

* [Musca-A](https://developer.arm.com/Tools%20and%20Software/Musca-A%20Test%20Chip%20Board)
* [Musca-B1](https://developer.arm.com/Tools%20and%20Software/Musca-B1%20Test%20Chip%20Board)
* [Musca-S1](https://developer.arm.com/Tools%20and%20Software/Musca-S1%20Test%20Chip%20Board)

```yaml
- deploy:
    to: musca
    images:
      test_binary:
        url: https://storage.lavacloud.io/health-checks/musca-b/tfm.hex
```

The board is powered on and its mass storage device (MSD) is mounted. The test
binary is copied to the MSD and then the MSD is unmounted. When the board
processes it, the MSD will be re-exposed to the LAVA worker, at which point this
is re-mounted and LAVA will check for the presence of a ``FAIL.TXT`` file, in
case of errors.

## Setup

Some initial setup steps are required to ensure that the Musca device serves its
MSD when it is powered on. Check
[MSD COMMANDS](https://github.com/ARMmbed/DAPLink/blob/master/docs/MSD_COMMANDS.md)
for details on how to set up the board to auto-boot when it is programmed or
turned on. Ensure `DETAILS.TXT` on the MSD shows "Auto Reset" and "Auto Power"
are activated.

!!! note
    LAVA does not deploy firmware to the Musca board; firmware must be
    pre-installed on each device. Check the device pages for updating the firmware.

## images

### test_binary

The test binary to copy to the Musca mass storage device.

#### url

See [url](./index.md#url).

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/estimate_calculus.R
\name{merge.estimate}
\alias{merge.estimate}
\title{Merge estimate objects}
\usage{
\method{merge}{estimate}(
  x,
  y,
  ...,
  id,
  paired = FALSE,
  labels = NULL,
  keep = NULL,
  subset = NULL,
  regex = FALSE,
  sep = FALSE,
  drop.ic = FALSE,
  ignore.case = FALSE,
  sort = FALSE
)
}
\arguments{
\item{x}{Object of class \code{estimate}}

\item{y}{Object of class \code{estimate}}

\item{...}{Additional \code{estimate} objects or arguments}

\item{id}{Optional cluster variable}

\item{paired}{If TRUE a paired (matched) analysis is performed}

\item{labels}{Optional character vector of labels for the merged estimates}

\item{keep}{Optional character vector of parameter names to keep}

\item{subset}{Optional character vector of parameter names to subset}

\item{regex}{If TRUE, \code{keep} and \code{subset} are treated as regular expressions}

\item{sep}{Separator used for labeling}

\item{drop.ic}{If TRUE, drop the influence function from the result}

\item{ignore.case}{If TRUE, case is ignored in \code{keep}/\code{subset} matching}

\item{sort}{if true the returned influence function will be sorted according
to the id variables (lexigraphically)}
}
\value{
Object of class \code{estimate} (see \link{estimate.default}).
}
\description{
Merge estimate objects
}
\seealso{
\code{\link[=c.estimate]{c.estimate()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/estimate_calculus.R
\name{merge.estimate}
\alias{merge.estimate}
\title{Merge estimate objects}
\usage{
\method{merge}{estimate}(
  x,
  y,
  ...,
  id,
  paired = FALSE,
  labels = NULL,
  keep = NULL,
  subset = NULL,
  regex = FALSE,
  sep = FALSE,
  drop.ic = FALSE,
  ignore.case = FALSE
)
}
\arguments{
\item{x}{Object of class \code{estimate}}

\item{y}{Object of class \code{estimate}}

\item{...}{Additional \code{estimate} objects or arguments}

\item{id}{Optional cluster variable}

\item{paired}{If TRUE a paired (matched) analysis is performed}

\item{labels}{Optional character vector of labels for the merged estimates}

\item{keep}{Optional character vector of parameter names to keep}

\item{subset}{Optional character vector of parameter names to subset}

\item{regex}{If TRUE, \code{keep} and \code{subset} are treated as regular expressions}

\item{sep}{Separator used for labeling}

\item{drop.ic}{If TRUE, drop the influence function from the result}

\item{ignore.case}{If TRUE, case is ignored in \code{keep}/\code{subset} matching}
}
\value{
Object of class \code{estimate} (see \link{estimate.default}).
}
\description{
Merge estimate objects
}
\seealso{
\code{\link[=c.estimate]{c.estimate()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/baptize.R
\name{baptize}
\alias{baptize}
\title{Label elements of object}
\usage{
baptize(x, ...)
}
\arguments{
\item{x}{Object}

\item{\dots}{Additional arguments}
}
\description{
Generic method for labeling elements of an object
}
\author{
Klaus K. Holst
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zgetmplus.R
\name{getMplus}
\alias{getMplus}
\title{Read Mplus output}
\usage{
getMplus(infile = "template.out", coef = TRUE, ...)
}
\arguments{
\item{infile}{Mplus output file}

\item{coef}{Coefficients only}

\item{\dots}{additional arguments to lower level functions}
}
\description{
Read Mplus output files
}
\seealso{
getSAS
}
\author{
Klaus K. Holst
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/timedep.R
\name{timedep}
\alias{timedep}
\alias{timedep<-}
\title{Time-dependent parameters}
\usage{
timedep(object, formula, rate, timecut, type = "coxExponential.lvm", ...)
}
\arguments{
\item{object}{Model}

\item{formula}{Formula with rhs specifying time-varying covariates}

\item{rate}{Optional rate parameters. If given as a vector this
parameter is interpreted as the raw (baseline-)rates within each
time interval defined by \code{timecut}.  If given as a matrix the
parameters are interpreted as log-rates (and log-rate-ratios for
the time-varying covariates defined in the formula).}

\item{timecut}{Time intervals}

\item{type}{Type of model (default piecewise constant intensity)}

\item{...}{Additional arguments to lower level functions}
}
\description{
Add time-varying covariate effects to model
}
\examples{

## Piecewise constant hazard
m <- lvm(y~1)
m <- timedep(m,y~1,timecut=c(0,5),rate=c(0.5,0.3))

\dontrun{
d <- sim(m,1e4); d$status <- TRUE
dd <- mets::lifetable(Surv(y,status)~1,data=d,breaks=c(0,5,10));
exp(coef(glm(events ~ offset(log(atrisk)) + -1 + interval, dd, family=poisson)))
}


## Piecewise constant hazard and time-varying effect of z1
m <- lvm(y~1)
distribution(m,~z1) <- Binary.lvm(0.5)
R <- log(cbind(c(0.2,0.7,0.9),c(0.5,0.3,0.3)))
m <- timedep(m,y~z1,timecut=c(0,3,5),rate=R)

\dontrun{
d <- sim(m,1e4); d$status <- TRUE
dd <- mets::lifetable(Surv(y,status)~z1,data=d,breaks=c(0,3,5,Inf));
exp(coef(glm(events ~ offset(log(atrisk)) + -1 + interval+z1:interval, dd, family=poisson)))
}



## Explicit simulation of time-varying effects
m <- lvm(y~1)
distribution(m,~z1) <- Binary.lvm(0.5)
distribution(m,~z2) <- binomial.lvm(p=0.5)
#variance(m,~m1+m2) <- 0
#regression(m,m1[m1:0] ~ z1) <- log(0.5)
#regression(m,m2[m2:0] ~ z1) <- log(0.3)
regression(m,m1 ~ z1,variance=0) <- log(0.5)
regression(m,m2 ~ z1,variance=0) <- log(0.3)
intercept(m,~m1+m2) <- c(-0.5,0)
m <- timedep(m,y~m1+m2,timecut=c(0,5))

\dontrun{
d <- sim(m,1e5); d$status <- TRUE
dd <- mets::lifetable(Surv(y,status)~z1,data=d,breaks=c(0,5,Inf))
exp(coef(glm(events ~ offset(log(atrisk)) + -1 + interval + interval:z1, dd, family=poisson)))
}
}
\author{
Klaus K. Holst
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/subset.R
\name{subset.lvm}
\alias{subset.lvm}
\alias{measurement}
\title{Extract subset of latent variable model}
\usage{
\method{subset}{lvm}(x, vars, ...)
}
\arguments{
\item{x}{\code{lvm}-object.}

\item{vars}{Character vector or formula specifying variables to include in
subset.}

\item{\dots}{Additional arguments to be passed to the low level functions}
}
\value{
A \code{lvm}-object.
}
\description{
Extract measurement models or user-specified subset of model
}
\examples{

m <- lvm(c(y1,y2)~x1+x2)
subset(m,~y1+x1)

}
\author{
Klaus K. Holst
}
\keyword{models}
\keyword{regression}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/intervention.R
\name{intervention.lvm}
\alias{intervention.lvm}
\alias{intervention<-}
\alias{intervention}
\alias{intervention<-.lvm}
\title{Define intervention}
\usage{
\method{intervention}{lvm}(object, to, value, dist = none.lvm(), ...)
}
\arguments{
\item{object}{lvm object}

\item{to}{String defining variable or formula}

\item{value}{function defining intervention}

\item{dist}{Distribution}

\item{...}{Additional arguments to lower level functions}
}
\description{
Define intervention in a \code{lvm} object
}
\examples{
m <- lvm(y ~ a + x, a ~ x)
distribution(m, ~a+y) <- binomial.lvm()
mm <- intervention(m, "a", value=3)
sim(mm, 10)
mm <- intervention(m, a~x, function(x) (x>0)*1)
sim(mm, 10)
}
\seealso{
regression lvm sim
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/information.R
\name{information}
\alias{information}
\title{Extract information matrix}
\usage{
information(x, ...)
}
\arguments{
\item{x}{Model object}

\item{...}{Additional arguments to lower level functions}
}
\value{
Information matrix or its inverse (depending on arguments).
}
\description{
Extract the (observed or expected) information matrix from a model object.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/makemissing.R
\name{makemissing}
\alias{makemissing}
\title{Create random missing data}
\usage{
makemissing(
  data,
  p = 0.2,
  cols = seq_len(ncol(data)),
  rowwise = FALSE,
  nafun = function(x) x,
  seed = NULL
)
}
\arguments{
\item{data}{data.frame}

\item{p}{Fraction of missing data in each column}

\item{cols}{Which columns (name or index) to alter}

\item{rowwise}{Should missing occur row-wise (either none or all selected columns are missing)}

\item{nafun}{(Optional) function to be applied on data.frame before return (e.g. \code{na.omit} to return complete-cases only)}

\item{seed}{Random seed}
}
\value{
data.frame
}
\description{
Generates missing entries in data.frame/matrix
}
\author{
Klaus K. Holst
}
\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cancel.R
\name{cancel}
\alias{cancel}
\alias{cancel<-}
\title{Generic cancel method}
\usage{
cancel(x, ...)
}
\arguments{
\item{x}{Object}

\item{\dots}{Additioal arguments}
}
\description{
Generic cancel method
}
\author{
Klaus K. Holst
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/eventTime.R
\name{eventTime}
\alias{eventTime}
\alias{eventTime<-}
\title{Add an observed event time outcome to a latent variable model.}
\usage{
eventTime(object, formula, eventName = "status", ...)
}
\arguments{
\item{object}{Model object}

\item{formula}{Formula (see details)}

\item{eventName}{Event names}

\item{\dots}{Additional arguments to lower levels functions}
}
\description{
For example, if the model 'm' includes latent event time variables
are called 'T1' and 'T2' and 'C' is the end of follow-up (right censored),
then one can specify
}
\details{
\code{eventTime(object=m,formula=ObsTime~min(T1=a,T2=b,C=0,"ObsEvent"))}

when data are simulated from the model
one gets 2 new columns:
\itemize{
\item "ObsTime": the smallest of T1, T2 and C
\item "ObsEvent": 'a' if T1 is smallest, 'b' if T2 is smallest and '0' if C is smallest
}

Note that "ObsEvent" and "ObsTime" are names specified by the user.
}
\examples{

# Right censored survival data without covariates
m0 <- lvm()
distribution(m0,"eventtime") <- coxWeibull.lvm(scale=1/100,shape=2)
distribution(m0,"censtime") <- coxExponential.lvm(rate=1/10)
m0 <- eventTime(m0,time~min(eventtime=1,censtime=0),"status")
sim(m0,10)

# Alternative specification of the right censored survival outcome
## eventTime(m,"Status") <- ~min(eventtime=1,censtime=0)

# Cox regression:
# lava implements two different parametrizations of the same
# Weibull regression model. The first specifies
# the effects of covariates as proportional hazard ratios
# and works as follows:
m <- lvm()
distribution(m,"eventtime") <- coxWeibull.lvm(scale=1/100,shape=2)
distribution(m,"censtime") <- coxWeibull.lvm(scale=1/100,shape=2)
m <- eventTime(m,time~min(eventtime=1,censtime=0),"status")
distribution(m,"sex") <- binomial.lvm(p=0.4)
distribution(m,"sbp") <- normal.lvm(mean=120,sd=20)
regression(m,from="sex",to="eventtime") <- 0.4
regression(m,from="sbp",to="eventtime") <- -0.01
sim(m,6)
# The parameters can be recovered using a Cox regression
# routine or a Weibull regression model. E.g.,
\dontrun{
    set.seed(18)
    d <- sim(m,1000)
    library(survival)
    coxph(Surv(time,status)~sex+sbp,data=d)

    sr <- survreg(Surv(time,status)~sex+sbp,data=d)
    library(SurvRegCensCov)
    ConvertWeibull(sr)

}

# The second parametrization is an accelerated failure time
# regression model and uses the function weibull.lvm instead
# of coxWeibull.lvm to specify the event time distributions.
# Here is an example:

ma <- lvm()
distribution(ma,"eventtime") <- weibull.lvm(scale=3,shape=1/0.7)
distribution(ma,"censtime") <- weibull.lvm(scale=2,shape=1/0.7)
ma <- eventTime(ma,time~min(eventtime=1,censtime=0),"status")
distribution(ma,"sex") <- binomial.lvm(p=0.4)
distribution(ma,"sbp") <- normal.lvm(mean=120,sd=20)
regression(ma,from="sex",to="eventtime") <- 0.7
regression(ma,from="sbp",to="eventtime") <- -0.008
set.seed(17)
sim(ma,6)
# The regression coefficients of the AFT model
# can be tranformed into log(hazard ratios):
#  coef.coxWeibull = - coef.weibull / shape.weibull
\dontrun{
    set.seed(17)
    da <- sim(ma,1000)
    library(survival)
    fa <- coxph(Surv(time,status)~sex+sbp,data=da)
    coef(fa)
    c(0.7,-0.008)/0.7
}


# The following are equivalent parametrizations
# which produce exactly the same random numbers:

model.aft <- lvm()
distribution(model.aft,"eventtime") <- weibull.lvm(intercept=-log(1/100)/2,sigma=1/2)
distribution(model.aft,"censtime") <- weibull.lvm(intercept=-log(1/100)/2,sigma=1/2)
sim(model.aft,6,seed=17)

model.aft <- lvm()
distribution(model.aft,"eventtime") <- weibull.lvm(scale=100^(1/2), shape=2)
distribution(model.aft,"censtime") <- weibull.lvm(scale=100^(1/2), shape=2)
sim(model.aft,6,seed=17)

model.cox <- lvm()
distribution(model.cox,"eventtime") <- coxWeibull.lvm(scale=1/100,shape=2)
distribution(model.cox,"censtime") <- coxWeibull.lvm(scale=1/100,shape=2)
sim(model.cox,6,seed=17)

# The minimum of multiple latent times one of them still
# being a censoring time, yield
# right censored competing risks data

mc <- lvm()
distribution(mc,~X2) <- binomial.lvm()
regression(mc) <- T1~f(X1,-.5)+f(X2,0.3)
regression(mc) <- T2~f(X2,0.6)
distribution(mc,~T1) <- coxWeibull.lvm(scale=1/100)
distribution(mc,~T2) <- coxWeibull.lvm(scale=1/100)
distribution(mc,~C) <- coxWeibull.lvm(scale=1/100)
mc <- eventTime(mc,time~min(T1=1,T2=2,C=0),"event")
sim(mc,6)


}
\author{
Thomas A. Gerds, Klaus K. Holst
}
\keyword{models}
\keyword{regression}
\keyword{survival}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/diagtest.R
\name{diagtest}
\alias{diagtest}
\alias{odds}
\alias{riskcomp}
\alias{OR}
\alias{Ratio}
\alias{Diff}
\title{Calculate diagnostic tests for 2x2 table}
\usage{
diagtest(
  table,
  positive = 2,
  exact = FALSE,
  p0 = NA,
  confint = c("logit", "arcsin", "pseudoscore", "exact"),
  ...
)
}
\arguments{
\item{table}{Table or (matrix/data.frame with two columns)}

\item{positive}{Switch reference}

\item{exact}{If TRUE exact binomial proportions CI/test will be used}

\item{p0}{Optional null hypothesis (test prevalenc, sensitivity, ...)}

\item{confint}{Type of confidence limits}

\item{...}{Additional arguments to lower level functions}
}
\description{
Calculate prevalence, sensitivity, specificity, and positive and
negative predictive values
}
\details{
Table should be in the format with outcome in columns and
test in rows.  Data.frame should be with test in the first
column and outcome in the second column.
}
\examples{
M <- as.table(matrix(c(42,12,
                       35,28),ncol=2,byrow=TRUE,
                     dimnames=list(rater=c("no","yes"),gold=c("no","yes"))))
diagtest(M,exact=TRUE)
}
\author{
Klaus Holst
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/eventTime.R
\name{eventTime}
\alias{eventTime}
\alias{eventTime<-}
\title{Add an observed event time outcome to a latent variable model.}
\usage{
eventTime(object, formula, eventName = "status", ...)
}
\arguments{
\item{object}{Model object}

\item{formula}{Formula (see details)}

\item{eventName}{Event names}

\item{\dots}{Additional arguments to lower levels functions}
}
\description{
For example, if the model 'm' includes latent event time
variables are called 'T1' and 'T2' and 'C' is the end of follow-up (right
censored), then one can specify

\code{eventTime(object=m,formula=ObsTime~min(T1=a,T2=b,C=0,"ObsEvent"))}

when data are simulated from the model
one gets 2 new columns:
\itemize{
\item "ObsTime": the smallest of T1, T2 and C
\item "ObsEvent": 'a' if T1 is smallest, 'b' if T2 is smallest and '0' if C is smallest
}

Note that "ObsEvent" and "ObsTime" are names specified by the user.
}
\examples{

# Right censored survival data without covariates
m0 <- lvm()
distribution(m0,"eventtime") <- dist_cox_weibull(scale=1/100,shape=2)
distribution(m0,"censtime") <- dist_cox_exponential(rate=1/10)
m0 <- eventTime(m0,time~min(eventtime=1,censtime=0),"status")
sim(m0,10)

# Alternative specification of the right censored survival outcome
## eventTime(m,"Status") <- ~min(eventtime=1,censtime=0)

# Cox regression:
# lava implements two different parametrizations of the same
# Weibull regression model. The first specifies
# the effects of covariates as proportional hazard ratios
# and works as follows:
m <- lvm()
distribution(m,"eventtime") <- dist_cox_weibull(scale=1/100,shape=2)
distribution(m,"censtime") <- dist_cox_weibull(scale=1/100,shape=2)
m <- eventTime(m,time~min(eventtime=1,censtime=0),"status")
distribution(m,"sex") <- dist_bernoulli(p=0.4)
distribution(m,"sbp") <- dist_gaussian(mean=120,sd=20)
regression(m,from="sex",to="eventtime") <- 0.4
regression(m,from="sbp",to="eventtime") <- -0.01
sim(m,6)
# The parameters can be recovered using a Cox regression
# routine or a Weibull regression model. E.g.,
\dontrun{
    set.seed(18)
    d <- sim(m,1000)
    library(survival)
    coxph(Surv(time,status)~sex+sbp,data=d)

    sr <- survreg(Surv(time,status)~sex+sbp,data=d)
    library(SurvRegCensCov)
    ConvertWeibull(sr)

}

# The second parametrization is an accelerated failure time
# regression model and uses the function dist_weibull instead
# of dist_cox_weibull to specify the event time distributions.
# Here is an example:

ma <- lvm()
distribution(ma,"eventtime") <- dist_weibull(scale=3,shape=1/0.7)
distribution(ma,"censtime") <- dist_weibull(scale=2,shape=1/0.7)
ma <- eventTime(ma,time~min(eventtime=1,censtime=0),"status")
distribution(ma,"sex") <- dist_bernoulli(p=0.4)
distribution(ma,"sbp") <- dist_gaussian(mean=120,sd=20)
regression(ma,from="sex",to="eventtime") <- 0.7
regression(ma,from="sbp",to="eventtime") <- -0.008
set.seed(17)
sim(ma,6)
# The regression coefficients of the AFT model
# can be tranformed into log(hazard ratios):
#  coef.coxWeibull = - coef.weibull / shape.weibull
\dontrun{
    set.seed(17)
    da <- sim(ma,1000)
    library(survival)
    fa <- coxph(Surv(time,status)~sex+sbp,data=da)
    coef(fa)
    c(0.7,-0.008)/0.7
}


# The following are equivalent parametrizations
# which produce exactly the same random numbers:

model.aft <- lvm()
distribution(model.aft,"eventtime") <-
  dist_weibull(intercept=-log(1/100)/2,sigma=1/2)
distribution(model.aft,"censtime") <-
  dist_weibull(intercept=-log(1/100)/2,sigma=1/2)
sim(model.aft,6,seed=17)

model.aft <- lvm()
distribution(model.aft,"eventtime") <- dist_weibull(scale=100^(1/2), shape=2)
distribution(model.aft,"censtime") <- dist_weibull(scale=100^(1/2), shape=2)
sim(model.aft,6,seed=17)

model.cox <- lvm()
distribution(model.cox,"eventtime") <- dist_cox_weibull(scale=1/100,shape=2)
distribution(model.cox,"censtime") <- dist_cox_weibull(scale=1/100,shape=2)
sim(model.cox,6,seed=17)

# The minimum of multiple latent times one of them still
# being a censoring time, yield
# right censored competing risks data

mc <- lvm()
distribution(mc,~X2) <- dist_bernoulli()
regression(mc) <- T1~f(X1,-.5)+f(X2,0.3)
regression(mc) <- T2~f(X2,0.6)
distribution(mc,~T1) <- dist_cox_weibull(scale=1/100)
distribution(mc,~T2) <- dist_cox_weibull(scale=1/100)
distribution(mc,~C) <- dist_cox_weibull(scale=1/100)
mc <- eventTime(mc,time~min(T1=1,T2=2,C=0),"event")
sim(mc,6)


}
\author{
Thomas A. Gerds, Klaus K. Holst
}
\keyword{models}
\keyword{regression}
\keyword{survival}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fplot.R
\name{fplot}
\alias{fplot}
\title{fplot}
\usage{
fplot(
  x,
  y,
  z = NULL,
  xlab,
  ylab,
  ...,
  z.col = topo.colors(64),
  data = parent.frame(),
  add = FALSE,
  aspect = c(1, 1),
  zoom = 0.8
)
}
\arguments{
\item{x}{X variable}

\item{y}{Y variable}

\item{z}{Z variable (optional)}

\item{xlab}{x-axis label}

\item{ylab}{y-axis label}

\item{...}{additional arggument to lower-level plot functions}

\item{z.col}{color (use argument alpha to set transparency)}

\item{data}{data.frame}

\item{add}{if TRUE use current active device}

\item{aspect}{aspect ratio}

\item{zoom}{zoom level}
}
\description{
Faster plot via RGL
}
\examples{
if (interactive()) {
data(iris)
fplot(Sepal.Length ~ Petal.Length+Species, data=iris, size=2, type="s")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/twostage.R
\name{twostageCV}
\alias{twostageCV}
\title{Cross-validated two-stage estimator}
\usage{
twostageCV(
  model1,
  model2,
  data,
  control1 = list(trace = 0),
  control2 = list(trace = 0),
  knots.boundary,
  nmix = 1:4,
  df = 1:9,
  fix = TRUE,
  std.err = TRUE,
  nfolds = 5,
  rep = 1,
  messages = 0,
  ...
)
}
\arguments{
\item{model1}{model 1 (exposure measurement error model)}

\item{model2}{model 2}

\item{data}{data.frame}

\item{control1}{optimization parameters for model 1}

\item{control2}{optimization parameters for model 1}

\item{knots.boundary}{boundary points for natural cubic spline basis}

\item{nmix}{number of mixture components}

\item{df}{spline degrees of freedom}

\item{fix}{automatically fix parameters for identification (TRUE)}

\item{std.err}{calculation of standard errors (TRUE)}

\item{nfolds}{Number of folds (cross-validation)}

\item{rep}{Number of repeats of cross-validation}

\item{messages}{print information (>0)}

\item{...}{additional arguments to lower}
}
\description{
Cross-validated two-stage estimator for non-linear SEM
}
\examples{
\donttest{ ## Reduce Ex.Timings##'
m1 <- lvm( x1+x2+x3 ~ u, latent= ~u)
m2 <- lvm( y ~ 1 )
m <- functional(merge(m1,m2), y ~ u, value=function(x) sin(x)+x)
distribution(m, ~u1) <- uniform.lvm(-6,6)
d <- sim(m,n=500,seed=1)
nonlinear(m2) <- y~u1
if (requireNamespace('mets', quietly=TRUE)) {
  set.seed(1)
  val <- twostageCV(m1, m2, data=d, std.err=FALSE, df=2:6, nmix=1:2,
                  nfolds=2)
  val
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/twostage.R
\name{twostageCV}
\alias{twostageCV}
\title{Cross-validated two-stage estimator}
\usage{
twostageCV(
  model1,
  model2,
  data,
  control1 = list(trace = 0),
  control2 = list(trace = 0),
  knots.boundary,
  nmix = 1:4,
  df = 1:9,
  fix = TRUE,
  std.err = TRUE,
  nfolds = 5,
  rep = 1,
  messages = 0,
  ...
)
}
\arguments{
\item{model1}{model 1 (exposure measurement error model)}

\item{model2}{model 2}

\item{data}{data.frame}

\item{control1}{optimization parameters for model 1}

\item{control2}{optimization parameters for model 1}

\item{knots.boundary}{boundary points for natural cubic spline basis}

\item{nmix}{number of mixture components}

\item{df}{spline degrees of freedom}

\item{fix}{automatically fix parameters for identification (TRUE)}

\item{std.err}{calculation of standard errors (TRUE)}

\item{nfolds}{Number of folds (cross-validation)}

\item{rep}{Number of repeats of cross-validation}

\item{messages}{print information (>0)}

\item{...}{additional arguments to lower}
}
\description{
Cross-validated two-stage estimator for non-linear SEM
}
\examples{
\dontrun{ ## Reduce Ex.Timings
m1 <- lvm( x1+x2+x3 ~ u, latent= ~u)
m2 <- lvm( y ~ 1 )
m <- functional(merge(m1,m2), y ~ u, value=function(x) sin(x)+x)
distribution(m, ~u1) <- dist_uniform(-6, 6)
d <- sim(m,n=500,seed=1)
nonlinear(m2) <- y~u1
if (requireNamespace('mets', quietly=TRUE)) {
  set.seed(1)
  val <- twostageCV(m1, m2, data=d, std.err=FALSE, df=2:6, nmix=1:2,
                  nfolds=2)
  val
}
} ## Reduce Ex.Timings
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/csplit.R
\name{csplit}
\alias{csplit}
\alias{foldr}
\title{Split data into folds}
\usage{
csplit(x, p = NULL, replace = FALSE, return.index = FALSE, k = 2, ...)
}
\arguments{
\item{x}{Data or integer (size)}

\item{p}{Number of folds, or if a number between 0 and 1 is given two folds of size p and (1-p) will be returned}

\item{replace}{With or with-out replacement}

\item{return.index}{If TRUE index of folds are returned otherwise the actual data splits are returned (default)}

\item{k}{(Optional, only used when p=NULL) number of folds without shuffling}

\item{...}{additional arguments to lower-level functions}
}
\description{
Split data into folds
}
\examples{
foldr(5,2,rep=2)
csplit(10,3)
csplit(iris[1:10,]) ## Split in two sets 1:(n/2) and (n/2+1):n
csplit(iris[1:10,],0.5)
}
\author{
Klaus K. Holst
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summary.estimate.R
\name{c.summary.estimate}
\alias{c.summary.estimate}
\title{Concatenate summary.estimate objects}
\usage{
\method{c}{summary.estimate}(...)
}
\arguments{
\item{...}{\code{summary.estimate} objects and/or named numeric values}
}
\description{
When all arguments are \code{summary.estimate} objects, they are merged into a
single \code{summary.estimate} object. When some arguments are not
\code{summary.estimate} objects but are named numeric scalars/vectors, an object
of class \code{summary.estimate} is returned with an additional \code{extra} attribute
that contains the provided numeric scalars/vectors. This is useful for
bundling auxiliary per-iteration information (e.g., convergence status)
alongside an estimate for use with \code{\link[=sim.default]{sim.default()}}.
}
\examples{
e1 <- estimate(coef = 1, IC = scale(rnorm(10)), id = 1:10, labels = "a1")
e2 <- estimate(coef = 2, IC = scale(rnorm(10)), id = 1:10, labels = "a2")

# concatenating two summary.estimate objects
c(e1, e2)

# concatenating one summary.estimate object with one numerical variable
ss <- c(summary(e1), niter = 2)
print(ss)
attributes(ss)$extra
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/predict.R
\name{predictlvm}
\alias{predictlvm}
\title{Predict function for latent variable models}
\usage{
predictlvm(object, formula, p = coef(object), data = model.frame(object), ...)
}
\arguments{
\item{object}{Model object}

\item{formula}{Formula specifying which variables to predict and which to condition on}

\item{p}{Parameter vector}

\item{data}{Data.frame}

\item{...}{Additional arguments to lower level functions}
}
\description{
Predictions of conditinoal mean and variance and calculation of
jacobian with respect to parameter vector.
}
\examples{
m <- lvm(c(x1,x2,x3)~u1,u1~z,
         c(y1,y2,y3)~u2,u2~u1+z)
latent(m) <- ~u1+u2
d <- simulate(m,10,"u2,u2"=2,"u1,u1"=0.5,seed=123)
e <- estimate(m,d)

## Conditional mean given covariates
predictlvm(e,c(x1,x2)~1)$mean
## Conditional variance of u1,y1 given x1,x2
predictlvm(e,c(u1,y1)~x1+x2)$var
}
\seealso{
predict.lvm
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zib.R
\name{PD}
\alias{PD}
\title{Dose response calculation for binomial regression models}
\usage{
PD(
  model,
  intercept = 1,
  slope = 2,
  prob = NULL,
  x,
  level = 0.5,
  ci.level = 0.95,
  vcov,
  family,
  EB = NULL
)
}
\arguments{
\item{model}{Model object or vector of parameter estimates}

\item{intercept}{Index of intercept parameters}

\item{slope}{Index of intercept parameters}

\item{prob}{Index of mixture parameters (only relevant for
\code{zibreg} models)}

\item{x}{Optional weights
length(x)=length(intercept)+length(slope)+length(prob)}

\item{level}{Probability at which level to calculate dose}

\item{ci.level}{Level of confidence limits}

\item{vcov}{Optional estimate of variance matrix of parameter
estimates}

\item{family}{Optional distributional family argument}

\item{EB}{Optional ratio of treatment effect and adverse effects
used to find optimal dose (regret-function argument)}
}
\description{
Dose response calculation for binomial regression models
}
\author{
Klaus K. Holst
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Expand.R
\name{Expand}
\alias{Expand}
\title{Create a Data Frame from All Combinations of Factors}
\usage{
Expand(`_data`, ...)
}
\arguments{
\item{_data}{Data.frame}

\item{...}{vectors, factors or a list containing these}
}
\description{
Create a Data Frame from All Combinations of Factors
}
\details{
Simple wrapper of the 'expand.grid' function.  If x is a table
then a data frame is returned with one row pr individual
observation.
}
\examples{
dd <- Expand(iris, Sepal.Length=2:8, Species=c("virginica","setosa"))
summary(dd)

T <- with(warpbreaks, table(wool, tension))
Expand(T)
}
\author{
Klaus K. Holst
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/multipletesting.R
\name{closed_testing}
\alias{closed_testing}
\alias{alpha_zmax}
\title{Closed testing procedure}
\usage{
closed_testing(object, test = test_wald, ...)
}
\arguments{
\item{object}{\code{estimate} object}

\item{test}{function that conducts hypothesis test. See details below.}

\item{...}{Additional arguments passed to \code{test}}
}
\description{
Given p hypotheses \eqn{H_1, \ldots, H_p}{H1, ..., Hp} all
\eqn{2^p-1}{2^p-1} intersection hypotheses are calculated and adjusted
p-values are obtained for \eqn{H_j}{Hj} is calculated as the max p-value of
all intersection hypotheses containing Hj. Example, for p=3, the adjusted
p-value for \eqn{H_1}{H1} will be obtained from \eqn{\{(H1, H2, H3),
(H1,H2), (H1,H3), (H1)\}}.
}
\details{
\if{html}{
  \figure{closedtesting.svg}{options: width=500}
}
\if{latex}{
  \figure{closedtesting.pdf}{options: width=7cm}
}

The function \code{test} should be a function \verb{function(object, index, ...)} which as its first argument takes an \code{estimate} object and with
an argument \code{index} which is a integer vector specifying which
subcomponents of \code{object} to test. The ellipsis argument can be any other
arguments used in the test function. The function \code{\link[=test_wald]{test_wald()}} is an
example of valid test function (which has an additional argument \code{null} in
reference to the above mentioned ellipsis arguments).
}
\examples{
m <- lvm()
regression(m, c(y1,y2,y3,y4)~x) <- c(0, 0.25, 0, 0.25)
regression(m, to=endogenous(m), from="u") <- 1
variance(m,endogenous(m)) <- 1
set.seed(1)
d <- sim(m, 200)
l1 <- lm(y1~x,d)
l2 <- lm(y2~x,d)
l3 <- lm(y3~x,d)
l4 <- lm(y4~x,d)

(a <- merge(l1, l2, l3, l4, subset=2))
if (requireNamespace("mets",quietly=TRUE)) {
   alpha_zmax(a)
}
adj <- closed_testing(a)
adj
adj$p.value
summary(adj)
}
\references{
Marcus, R; Peritz, E; Gabriel, KR (1976).
"On closed testing procedures with special reference to ordered analysis of variance".
Biometrika. 63 (3): 655–660.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/multipletesting.R
\name{closed_testing}
\alias{closed_testing}
\alias{alpha_zmax}
\title{Closed testing procedure}
\usage{
closed_testing(object, test = test_wald, ...)
}
\arguments{
\item{object}{\code{estimate} object}

\item{test}{function that conducts hypothesis test. See details below.}

\item{...}{Additional arguments passed to \code{test}}
}
\description{
Given \eqn{k} hypotheses \eqn{H_1, \ldots, H_k}{H1, ..., Hp} all
\eqn{2^k-1}{2^k-1} intersection hypotheses are calculated and adjusted
p-values are obtained for \eqn{H_j}{Hj} is calculated as the max
\eqn{p}-value of all intersection hypotheses containing \eqn{H_j}{Hj}.
Example, for \eqn{k=3}, the adjusted \eqn{p}-value for \eqn{H_1}{H1} will be
obtained from \eqn{\{(H_1, H_2, H_3), (H_1,H_2), (H_1,H_3), (H_1)\}}.
}
\details{
\if{html}{
  \figure{closedtesting.svg}{options: width=500}
}
\if{latex}{
  \figure{closedtesting.pdf}{options: width=7cm}
}

The function \code{test} should be a function \verb{function(object, index, ...)} which as its first argument takes an \code{estimate} object and with
an argument \code{index} which is a integer vector specifying which
subcomponents of \code{object} to test. The ellipsis argument can be any other
arguments used in the test function. The function \code{\link[=test_wald]{test_wald()}} is an
example of valid test function (which has an additional argument \code{null} in
reference to the above mentioned ellipsis arguments).
}
\examples{
m <- lvm()
regression(m, c(y1,y2,y3,y4)~x) <- c(0, 0.25, 0, 0.25)
regression(m, to=endogenous(m), from="u") <- 1
variance(m,endogenous(m)) <- 1
set.seed(1)
d <- sim(m, 200)
l1 <- lm(y1~x,d)
l2 <- lm(y2~x,d)
l3 <- lm(y3~x,d)
l4 <- lm(y4~x,d)

(a <- merge(l1, l2, l3, l4, subset=2))
if (requireNamespace("mets",quietly=TRUE)) {
   alpha_zmax(a)
}
adj <- closed_testing(a, test = test_wald, null = 0)
adj
adj$p.value
summary(adj)
}
\references{
Marcus, R; Peritz, E; Gabriel, KR (1976).
"On closed testing procedures with special reference to ordered analysis of variance".
Biometrika. 63 (3): 655–660.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/graph.R
\name{Graph}
\alias{Graph}
\alias{Graph<-}
\title{Extract graph}
\usage{
Graph(x, ...)

Graph(x, ...) <- value
}
\arguments{
\item{x}{Model object}

\item{\dots}{Additional arguments to be passed to the low level functions}

\item{value}{New \code{graphNEL} object}
}
\description{
Extract or replace graph object
}
\examples{

m <- lvm(y~x)
Graph(m)

}
\seealso{
\code{\link{Model}}
}
\author{
Klaus K. Holst
}
\keyword{graphs}
\keyword{models}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/addhook.R
\name{lava.options}
\alias{lava.options}
\title{Set global options for \code{lava}}
\usage{
lava.options(...)
}
\arguments{
\item{\dots}{Arguments}
}
\value{
\code{list} of parameters
}
\description{
Extract and set global parameters of \code{lava}. In particular optimization
parameters for the \code{estimate} function.
}
\details{
\itemize{
\item \code{param}: 'relative' (factor loading and variance of one
endogenous variables in each measurement model are fixed to one), 'absolute'
(mean and variance of latent variables are set to 0 and 1, respectively),
'hybrid' (intercept of latent variables is fixed to 0, and factor loading of
at least one endogenous variable in each measurement model is fixed to 1),
'none' (no constraints are added)
\item \code{layout}: One of 'dot','fdp','circo','twopi','neato','osage'
\item \code{messages}: Set to 0 to disable various output messages
\item ...  }

see \code{control} parameter of the \code{estimate} function.
}
\examples{

\dontrun{
lava.options(iter.max=100,messages=0)
}

}
\author{
Klaus K. Holst
}
\keyword{models}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wait.R
\name{wait}
\alias{wait}
\alias{waitclick}
\title{Wait for user input (keyboard or mouse)}
\usage{
wait()
}
\description{
Wait for user input (keyboard or mouse)
}
\author{
Klaus K. Holst
}
\keyword{iplot}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/confint.R
\name{confint.lvmfit}
\alias{confint.lvmfit}
\alias{confint.multigroupfit}
\title{Calculate confidence limits for parameters}
\usage{
\method{confint}{lvmfit}(
  object,
  parm = seq_len(length(coef(object))),
  level = 0.95,
  profile = FALSE,
  curve = FALSE,
  n = 20,
  interval = NULL,
  lower = TRUE,
  upper = TRUE,
  ...
)
}
\arguments{
\item{object}{\code{lvm}-object.}

\item{parm}{Index of which parameters to calculate confidence limits for.}

\item{level}{Confidence level}

\item{profile}{Logical expression defining whether to calculate confidence
limits via the profile log likelihood}

\item{curve}{if FALSE and profile is TRUE, confidence limits are
returned. Otherwise, the profile curve is returned.}

\item{n}{Number of points to evaluate profile log-likelihood in
over the interval defined by \code{interval}}

\item{interval}{Interval over which the profiling is done}

\item{lower}{If FALSE the lower limit will not be estimated (profile intervals only)}

\item{upper}{If FALSE the upper limit will not be estimated (profile intervals only)}

\item{\dots}{Additional arguments to be passed to the low level functions}
}
\value{
A 2xp matrix with columns of lower and upper confidence limits
}
\description{
Calculate Wald og Likelihood based (profile likelihood) confidence intervals
}
\details{
Calculates either Wald confidence limits: \deqn{\hat{\theta} \pm
z_{\alpha/2}*\hat\sigma_{\hat\theta}} or profile likelihood confidence
limits, defined as the set of value \eqn{\tau}:
\deqn{logLik(\hat\theta_{\tau},\tau)-logLik(\hat\theta)< q_{\alpha}/2}

where \eqn{q_{\alpha}} is the \eqn{\alpha} fractile of the \eqn{\chi^2_1}
distribution, and \eqn{\hat\theta_{\tau}} are obtained by maximizing the
log-likelihood with tau being fixed.
}
\examples{

m <- lvm(y~x)
d <- sim(m,100)
e <- estimate(lvm(y~x), d)
confint(e,3,profile=TRUE)
confint(e,3)
\donttest{ ## Reduce Ex.timings
B <- bootstrap(e,R=50)
B
}
}
\seealso{
\code{\link{bootstrap}{lvm}}
}
\author{
Klaus K. Holst
}
\keyword{models}
\keyword{regression}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/toformula.R
\name{toformula}
\alias{toformula}
\title{Converts strings to formula}
\usage{
toformula(y = ".", x = ".")
}
\arguments{
\item{y}{vector of predictors}

\item{x}{vector of responses}
}
\value{
An object of class \code{formula}
}
\description{
Converts a vector of predictors and a vector of responses (characters) i#nto
a formula expression.
}
\examples{

toformula(c("age","gender"), "weight")

}
\seealso{
\code{\link{as.formula}},
}
\author{
Klaus K. Holst
}
\keyword{models}
\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ksmooth.R
\name{ksmooth2}
\alias{ksmooth2}
\alias{surface}
\title{Plot/estimate surface}
\usage{
ksmooth2(
  x,
  data,
  h = NULL,
  xlab = NULL,
  ylab = NULL,
  zlab = "",
  gridsize = rep(51L, 2),
  ...
)
}
\arguments{
\item{x}{formula or data}

\item{data}{data.frame}

\item{h}{bandwidth}

\item{xlab}{X label}

\item{ylab}{Y label}

\item{zlab}{Z label}

\item{gridsize}{grid size of kernel smoother}

\item{...}{Additional arguments to graphics routine (persp3d or persp)}
}
\description{
Plot/estimate surface
}
\examples{
if (requireNamespace("KernSmooth")) {##'
ksmooth2(rmvn0(1e4,sigma=diag(2)*.5+.5),c(-3.5,3.5),h=1,
        rgl=FALSE,theta=30)
##'
if (interactive()) {
    ksmooth2(rmvn0(1e4,sigma=diag(2)*.5+.5),c(-3.5,3.5),h=1)
    ksmooth2(function(x,y) x^2+y^2, c(-20,20))
    ksmooth2(function(x,y) x^2+y^2, xlim=c(-5,5), ylim=c(0,10))

    f <- function(x,y) 1-sqrt(x^2+y^2)
    surface(f,xlim=c(-1,1),alpha=0.9,aspect=c(1,1,0.75))
    surface(f,xlim=c(-1,1),clut=heat.colors(128))
    ##play3d(spin3d(axis=c(0,0,1), rpm=8), duration=5)
}

if (interactive()) {
    surface(function(x) dmvn0(x,sigma=diag(2)),c(-3,3),lit=FALSE,smooth=FALSE,box=FALSE,alpha=0.8)
    surface(function(x) dmvn0(x,sigma=diag(2)),c(-3,3),box=FALSE,specular="black")##'
}

if (!inherits(try(find.package("fields"),silent=TRUE),"try-error")) {
    f <- function(x,y) 1-sqrt(x^2+y^2)
    ksmooth2(f,c(-1,1),rgl=FALSE,image=fields::image.plot)
}

}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/estimate_calculus.R
\name{c.estimate}
\alias{c.estimate}
\title{Concatenate estimate objects}
\usage{
\method{c}{estimate}(..., as.list = FALSE)
}
\arguments{
\item{...}{\code{estimate} objects and/or named numeric values}

\item{as.list}{if TRUE the returned object will be of class \code{list} and not
an \code{estimate} object.}
}
\value{
An \code{estimate} object. If extra named numeric values were provided,
the result carries an \code{"extra"} attribute (a named numeric vector).
}
\description{
When all arguments are \code{estimate} objects, they are merged into a single
\code{estimate} object. When some arguments are not \code{estimate} objects but are
named numeric scalars/vectors, the merged \code{estimate} object is returned
with an \code{"extra"} attribute containing the provided values. This is useful
for bundling auxiliary per-iteration information (e.g., convergence status)
alongside an estimate for use with \code{\link[=sim.default]{sim.default()}}.
}
\details{
arguments \code{drop.ic}, \code{paired}, \code{sep} are passed to \link{merge.estimate}
}
\examples{
e <- estimate(coef = c(a = 1, b = 2), vcov = diag(2) * 0.1)
# Bundle estimate with extra information
res <- c(e, converged = 1, niter = 10)
attr(res, "extra")
}
\seealso{
\code{\link[=sim.default]{sim.default()}} \code{\link[=merge.estimate]{merge.estimate()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/By.R
\name{By}
\alias{By}
\title{Apply a Function to a Data Frame Split by Factors}
\usage{
By(x, INDICES, FUN, COLUMNS, array = FALSE, ...)
}
\arguments{
\item{x}{Data frame}

\item{INDICES}{Indices (vector or list of indices, vector of column names, or formula of column names)}

\item{FUN}{A function to be applied to data frame subsets of 'data'.}

\item{COLUMNS}{(Optional) subset of columns of x to work on}

\item{array}{if TRUE an array/matrix is always returned}

\item{...}{Additional arguments to lower-level functions}
}
\description{
Apply a Function to a Data Frame Split by Factors
}
\details{
Simple wrapper of the 'by' function
}
\examples{
By(datasets::CO2,~Treatment+Type,colMeans,~conc)
By(datasets::CO2,~Treatment+Type,colMeans,~conc+uptake)
}
\author{
Klaus K. Holst
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vars.R
\name{vars}
\alias{vars}
\alias{vars.lvm}
\alias{vars.lvmfit}
\alias{latent}
\alias{latent<-}
\alias{latent.lvm}
\alias{latent<-.lvm}
\alias{latent.lvmfit}
\alias{latent.multigroup}
\alias{manifest}
\alias{manifest.lvm}
\alias{manifest.lvmfit}
\alias{manifest.multigroup}
\alias{exogenous}
\alias{exogenous<-}
\alias{exogenous.lvm}
\alias{exogenous<-.lvm}
\alias{exogenous.lvmfit}
\alias{exogenous.multigroup}
\alias{endogenous}
\alias{endogenous.lvm}
\alias{endogenous.lvmfit}
\alias{endogenous.multigroup}
\title{Extract variable names from latent variable model}
\usage{
vars(x,...)

endogenous(x,...)

exogenous(x,...)

manifest(x,...)

latent(x,...)

\method{exogenous}{lvm}(x, xfree = TRUE,...) <- value

\method{exogenous}{lvm}(x,variable,latent=FALSE,index=TRUE,...)

\method{latent}{lvm}(x,clear=FALSE,...) <- value
}
\arguments{
\item{x}{\code{lvm}-object}

\item{\dots}{Additional arguments to be passed to the low level functions}

\item{variable}{list of variables to alter}

\item{latent}{Logical defining whether latent variables without parents
should be included in the result}

\item{index}{For internal use only}

\item{clear}{Logical indicating whether to add or remove latent variable
status}

\item{xfree}{For internal use only}

\item{value}{Formula or character vector of variable names.}
}
\value{
Vector of variable names.
}
\description{
Extract exogenous variables (predictors), endogenous variables (outcomes),
latent variables (random effects), manifest (observed) variables from a
\code{lvm} object.
}
\details{
\code{vars} returns all variables of the \code{lvm}-object including
manifest and latent variables. Similarily \code{manifest} and \code{latent}
returns the observered resp. latent variables of the model.
\code{exogenous} returns all manifest variables without parents, e.g.
covariates in the model, however the argument \code{latent=TRUE} can be used
to also include latent variables without parents in the result. Pr. default
\code{lava} will not include the parameters of the exogenous variables in
the optimisation routine during estimation (likelihood of the remaining
observered variables conditional on the covariates), however this behaviour
can be altered via the assignment function \code{exogenous<-} telling
\code{lava} which subset of (valid) variables to condition on.  Finally
\code{latent} returns a vector with the names of the latent variables in
\code{x}. The assigment function \code{latent<-} can be used to change the
latent status of variables in the model.
}
\examples{

g <- lvm(eta1 ~ x1+x2)
regression(g) <- c(y1,y2,y3) ~ eta1
latent(g) <- ~eta1
endogenous(g)
exogenous(g)
identical(latent(g), setdiff(vars(g),manifest(g)))

}
\seealso{
\code{\link{endogenous}}, \code{\link{manifest}},
\code{\link{latent}}, \code{\link{exogenous}}, \code{\link{vars}}
}
\author{
Klaus K. Holst
}
\keyword{models}
\keyword{regression}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/equivalence.R
\name{equivalence}
\alias{equivalence}
\title{Identify candidates of equivalent models}
\usage{
equivalence(x, rel, tol = 0.001, k = 1, omitrel = TRUE, ...)
}
\arguments{
\item{x}{\code{lvmfit}-object}

\item{rel}{Formula or character-vector specifying two variables to omit from
the model and subsequently search for possible equivalent models}

\item{tol}{Define two models as empirical equivalent if the absolute
difference in score test is less than \code{tol}}

\item{k}{Number of parameters to test simultaneously. For \code{equivalence}
the number of additional associations to be added instead of \code{rel}.}

\item{omitrel}{if \code{k} greater than 1, this boolean defines wether to
omit candidates containing \code{rel} from the output}

\item{\dots}{Additional arguments to be passed to the lower-level functions}
}
\description{
Identifies candidates of equivalent models
}
\seealso{
\code{\link{compare}}, \code{\link{modelsearch}}
}
\author{
Klaus K. Holst
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/blockdiag.R
\name{blockdiag}
\alias{blockdiag}
\title{Combine matrices to block diagonal structure}
\usage{
blockdiag(x, ..., pad = 0)
}
\arguments{
\item{x}{matrix}

\item{\dots}{additional matrices}

\item{pad}{value outside block-diagonal}
}
\description{
Combine matrices to block diagonal structure
}
\examples{
A <- diag(3)+1
blockdiag(A,A,A,pad=NA)
}
\author{
Klaus K. Holst
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/optims.R
\name{NR}
\alias{NR}
\title{Newton-Raphson method}
\usage{
NR(
  start,
  objective = NULL,
  gradient = NULL,
  hessian = NULL,
  control,
  args = NULL,
  ...
)
}
\arguments{
\item{start}{Starting value}

\item{objective}{Optional objective function (used for selecting step length)}

\item{gradient}{gradient}

\item{hessian}{hessian (if NULL a numerical derivative is used)}

\item{control}{optimization arguments (see details)}

\item{args}{Optional list of arguments parsed to objective, gradient and hessian}

\item{...}{additional arguments parsed to lower level functions}
}
\description{
Newton-Raphson method
}
\details{
\code{control} should be a list with one or more of the following components:
\itemize{
\item{trace} integer for which output is printed each 'trace'th iteration
\item{iter.max} number of iterations
\item{stepsize}: Step size (default 1)
\item{nstepsize}: Increase stepsize every nstepsize iteration (from stepsize to 1)
\item{tol}: Convergence criterion (gradient)
\item{epsilon}: threshold used in pseudo-inverse
\item{backtrack}: In each iteration reduce stepsize unless solution is improved according to criterion (gradient, armijo, curvature, wolfe)
}
}
\examples{
# Objective function with gradient and hessian as attributes
f <- function(z) {
   x <- z[1]; y <- z[2]
   val <- x^2 + x*y^2 + x + y
   structure(val, gradient=c(2*x+y^2+1, 2*y*x+1),
             hessian=rbind(c(2,2*y),c(2*y,2*x)))
}
NR(c(0,0),f)

# Parsing arguments to the function and
g <- function(x,y) (x*y+1)^2
NR(0, gradient=g, args=list(y=2), control=list(trace=1,tol=1e-20))


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modelsearch.R
\name{modelsearch}
\alias{modelsearch}
\title{Model searching}
\usage{
modelsearch(x, k = 1, dir = "forward", type = "all", ...)
}
\arguments{
\item{x}{\code{lvmfit}-object}

\item{k}{Number of parameters to test simultaneously. For \code{equivalence}
the number of additional associations to be added instead of \code{rel}.}

\item{dir}{Direction to do model search. "forward" := add
associations/arrows to model/graph (score tests), "backward" := remove
associations/arrows from model/graph (wald test)}

\item{type}{If equal to 'correlation' only consider score tests for covariance parameters. If equal to 'regression' go through direct effects only  (default 'all' is to do both)}

\item{...}{Additional arguments to be passed to the low level functions}
}
\value{
Matrix of test-statistics and p-values
}
\description{
Performs Wald or score tests
}
\examples{

m <- lvm();
regression(m) <- c(y1,y2,y3) ~ eta; latent(m) <- ~eta
regression(m) <- eta ~ x
m0 <- m; regression(m0) <- y2 ~ x
dd <- sim(m0,100)[,manifest(m0)]
e <- estimate(m,dd);
modelsearch(e,messages=0)
modelsearch(e,messages=0,type="cor")
}
\seealso{
\code{\link{compare}}, \code{\link{equivalence}}
}
\author{
Klaus K. Holst
}
\keyword{htest}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scheffe.R
\name{scheffe}
\alias{scheffe}
\title{Calculate simultaneous confidence limits by Scheffe's method}
\usage{
scheffe(model, newdata = model.frame(model), level = 0.95)
}
\arguments{
\item{model}{Linear model}

\item{newdata}{new data frame}

\item{level}{confidence level (0.95)}
}
\description{
Function to compute the Scheffe corrected confidence
interval for the regression line
}
\examples{
x <- rnorm(100)
d <- data.frame(y=rnorm(length(x),x),x=x)
l <- lm(y~x,d)
plot(y~x,d)
abline(l)
d0 <- data.frame(x=seq(-5,5,length.out=100))
d1 <- cbind(d0,predict(l,newdata=d0,interval="confidence"))
d2 <- cbind(d0,scheffe(l,d0))
lines(lwr~x,d1,lty=2,col="red")
lines(upr~x,d1,lty=2,col="red")
lines(lwr~x,d2,lty=2,col="blue")
lines(upr~x,d2,lty=2,col="blue")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/constrain.R
\name{Range.lvm}
\alias{Range.lvm}
\title{Define range constraints of parameters}
\usage{
Range.lvm(a = 0, b = 1)
}
\arguments{
\item{a}{Lower bound}

\item{b}{Upper bound}
}
\value{
function
}
\description{
Define range constraints of parameters
}
\author{
Klaus K. Holst
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.estimate.R
\name{plot.estimate}
\alias{plot.estimate}
\title{Plot method for 'estimate' objects}
\usage{
\method{plot}{estimate}(
  x,
  f,
  idx,
  intercept = FALSE,
  data,
  confint = TRUE,
  type = "l",
  xlab = "x",
  ylab = "f(x)",
  col = 1,
  add = FALSE,
  null = 0,
  ...
)
}
\arguments{
\item{x}{estimate object}

\item{f}{function of parameter coefficients and data parsed on to
'estimate'. If omitted a forest-plot will be produced.}

\item{idx}{Index of parameters (default all)}

\item{intercept}{include intercept in forest-plot}

\item{data}{data.frame}

\item{confint}{Add confidence limits}

\item{type}{plot type ('l')}

\item{xlab}{x-axis label}

\item{ylab}{y-axis label}

\item{col}{color}

\item{add}{add plot to current device}

\item{null}{null value for forest-plot}

\item{...}{additional arguments to lower-level functions}
}
\description{
Plot method for 'estimate' objects
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/multinomial.R
\name{multinomial}
\alias{multinomial}
\alias{kappa.multinomial}
\alias{kappa.table}
\alias{gkgamma}
\title{Estimate probabilities in contingency table}
\usage{
multinomial(
  x,
  data = parent.frame(),
  marginal = FALSE,
  transform,
  vcov = TRUE,
  IC = TRUE,
  ...
)
}
\arguments{
\item{x}{Formula (or matrix or data.frame with observations, 1 or 2 columns)}

\item{data}{Optional data.frame}

\item{marginal}{If TRUE the marginals are estimated}

\item{transform}{Optional transformation of parameters (e.g., logit)}

\item{vcov}{Calculate asymptotic variance (default TRUE)}

\item{IC}{Return ic decomposition (default TRUE)}

\item{...}{Additional arguments to lower-level functions}
}
\description{
Estimate probabilities in contingency table
}
\examples{
set.seed(1)
breaks <- c(-Inf,-1,0,Inf)
m <- lvm(); covariance(m,pairwise=TRUE) <- ~y1+y2+y3+y4
d <- transform(sim(m,5e2),
              z1=cut(y1,breaks=breaks),
              z2=cut(y2,breaks=breaks),
              z3=cut(y3,breaks=breaks),
              z4=cut(y4,breaks=breaks))

multinomial(d[,5])
(a1 <- multinomial(d[,5:6]))
(K1 <- kappa(a1)) ## Cohen's kappa

K2 <- kappa(d[,7:8])
## Testing difference K1-K2:
estimate(merge(K1,K2,id=TRUE),diff)

estimate(merge(K1,K2,id=FALSE),diff) ## Wrong std.err ignoring dependence
sqrt(vcov(K1)+vcov(K2))

## Average of the two kappas:
estimate(merge(K1,K2,id=TRUE),function(x) mean(x))
estimate(merge(K1,K2,id=FALSE),function(x) mean(x)) ## Independence
##'
## Goodman-Kruskal's gamma
m2 <- lvm(); covariance(m2) <- y1~y2
breaks1 <- c(-Inf,-1,0,Inf)
breaks2 <- c(-Inf,0,Inf)
d2 <- transform(sim(m2,5e2),
              z1=cut(y1,breaks=breaks1),
              z2=cut(y2,breaks=breaks2))

(g1 <- gkgamma(d2[,3:4]))
## same as
\dontrun{
gkgamma(table(d2[,3:4]))
gkgamma(multinomial(d2[,3:4]))
}

##partial gamma
d2$x <- rbinom(nrow(d2),2,0.5)
gkgamma(z1~z2|x,data=d2)
}
\author{
Klaus K. Holst
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lava-package.R
\docType{data}
\name{bmd}
\alias{bmd}
\title{Longitudinal Bone Mineral Density Data (Wide format)}
\format{
data.frame
}
\source{
Vonesh & Chinchilli (1997), Table 5.4.1 on page 228.
}
\description{
Bone Mineral Density Data consisting of 112 girls randomized to receive
calcium og placebo. Longitudinal measurements of bone mineral density
(g/cm^2) measured approximately every 6th month in 3 years.
}
\seealso{
calcium
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bootstrap.R
\name{bootstrap}
\alias{bootstrap}
\title{Generic bootstrap method}
\usage{
bootstrap(x, ...)
}
\arguments{
\item{x}{Model object}

\item{\dots}{Additional arguments}
}
\description{
Generic method for calculating bootstrap statistics
}
\seealso{
\code{bootstrap.lvm} \code{bootstrap.lvmfit}
}
\author{
Klaus K. Holst
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/napass0.R
\name{na.pass0}
\alias{na.pass0}
\title{Handle Missing Values in Objects}
\usage{
na.pass0(object, na.action = na.omit, ...)
}
\arguments{
\item{object}{data.frame, vector}

\item{na.action}{function used to identify missing values}

\item{...}{additional arguments to lower level functions}
}
\description{
Returns the object with missing data replaced by zeros. This is
sometimes useful for example when working with inverse probability
weighting of the complete-case data.
}
\examples{
d <- data.frame(y=c(1,1,NA,2,NA,2), r=c(1,1,0,1,1,1))
na.pass0(d)
glm(y ~ 1, weights=d$r, data=d, na.action=na.pass0)
}
\seealso{
\code{\link[=na.pass]{na.pass()}}, \code{\link[=na.omit]{na.omit()}}, \code{\link[=na.fail]{na.fail()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pdfconvert.R
\name{pdfconvert}
\alias{pdfconvert}
\title{Convert pdf to raster format}
\usage{
pdfconvert(
  files,
  dpi = 300,
  resolution = 1024,
  gs,
  gsopt,
  resize,
  format = "png",
  ...
)
}
\arguments{
\item{files}{Vector of (pdf-)filenames to process}

\item{dpi}{DPI}

\item{resolution}{Resolution of raster image file}

\item{gs}{Optional ghostscript command}

\item{gsopt}{Optional ghostscript arguments}

\item{resize}{Optional resize arguments (mogrify)}

\item{format}{Raster format (e.g. png, jpg, tif, ...)}

\item{\dots}{Additional arguments}
}
\description{
Convert PDF file to print quality png (default 300 dpi)
}
\details{
Access to ghostscript program 'gs' is needed
}
\seealso{
\code{dev.copy2pdf}, \code{printdev}
}
\author{
Klaus K. Holst
}
\keyword{iplot}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/constrain.R
\name{constrain<-}
\alias{constrain<-}
\alias{constrain}
\alias{constrain.default}
\alias{constrain<-.multigroup}
\alias{constrain<-.default}
\alias{constraints}
\alias{parameter<-}
\title{Add non-linear constraints to latent variable model}
\usage{
\method{constrain}{default}(x,par,args,endogenous=TRUE,...) <- value

\method{constrain}{multigroup}(x,par,k=1,...) <- value

constraints(object,data=model.frame(object),vcov=object$vcov,level=0.95,
                        p=pars.default(object),k,idx,...)
}
\arguments{
\item{x}{\code{lvm}-object}

\item{\dots}{Additional arguments to be passed to the low level functions}

\item{value}{Real function taking args as a vector argument}

\item{par}{Name of new parameter. Alternatively a formula with lhs
specifying the new parameter and the rhs defining the names of the
parameters or variable names defining the new parameter (overruling the
\code{args} argument).}

\item{args}{Vector of variables names or parameter names that are used in
defining \code{par}}

\item{endogenous}{TRUE if variable is endogenous (sink node)}

\item{k}{For multigroup models this argument specifies which group to
add/extract the constraint}

\item{object}{\code{lvm}-object}

\item{data}{Data-row from which possible non-linear constraints should be
calculated}

\item{vcov}{Variance matrix of parameter estimates}

\item{level}{Level of confidence limits}

\item{p}{Parameter vector}

\item{idx}{Index indicating which constraints to extract}
}
\value{
A \code{lvm} object.
}
\description{
Add non-linear constraints to latent variable model
}
\details{
Add non-linear parameter constraints as well as non-linear associations
between covariates and latent or observed variables in the model (non-linear
regression).

As an example we will specify the follow multiple regression model:

\deqn{E(Y|X_1,X_2) = \alpha + \beta_1 X_1 + \beta_2 X_2} \deqn{V(Y|X_1,X_2)
= v}

which is defined (with the appropiate parameter labels) as

\code{m <- lvm(y ~ f(x,beta1) + f(x,beta2))}

\code{intercept(m) <- y ~ f(alpha)}

\code{covariance(m) <- y ~ f(v)}

The somewhat strained parameter constraint \deqn{ v =
\frac{(beta1-beta2)^2}{alpha}}

can then specified as

\code{constrain(m,v ~ beta1 + beta2 + alpha) <- function(x)
(x[1]-x[2])^2/x[3] }

A subset of the arguments \code{args} can be covariates in the model,
allowing the specification of non-linear regression models.  As an example
the non-linear regression model \deqn{ E(Y\mid X) = \nu + \Phi(\alpha +
\beta X)} where \eqn{\Phi} denotes the standard normal cumulative
distribution function, can be defined as

\code{m <- lvm(y ~ f(x,0)) # No linear effect of x}

Next we add three new parameters using the \code{parameter} assigment
function:

\code{parameter(m) <- ~nu+alpha+beta}

The intercept of \eqn{Y} is defined as \code{mu}

\code{intercept(m) <- y ~ f(mu)}

And finally the newly added intercept parameter \code{mu} is defined as the
appropiate non-linear function of \eqn{\alpha}, \eqn{\nu} and \eqn{\beta}:

\code{constrain(m, mu ~ x + alpha + nu) <- function(x)
pnorm(x[1]*x[2])+x[3]}

The \code{constraints} function can be used to show the estimated non-linear
parameter constraints of an estimated model object (\code{lvmfit} or
\code{multigroupfit}). Calling \code{constrain} with no additional arguments
beyound \code{x} will return a list of the functions and parameter names
defining the non-linear restrictions.

The gradient function can optionally be added as an attribute \code{grad} to
the return value of the function defined by \code{value}. In this case the
analytical derivatives will be calculated via the chain rule when evaluating
the corresponding score function of the log-likelihood. If the gradient
attribute is omitted the chain rule will be applied on a numeric
approximation of the gradient.
}
\examples{
##############################
### Non-linear parameter constraints 1
##############################
m <- lvm(y ~ f(x1,gamma)+f(x2,beta))
covariance(m) <- y ~ f(v)
d <- sim(m,100)
m1 <- m; constrain(m1,beta ~ v) <- function(x) x^2
## Define slope of x2 to be the square of the residual variance of y
## Estimate both restricted and unrestricted model
e <- estimate(m,d,control=list(method="NR"))
e1 <- estimate(m1,d)
p1 <- coef(e1)
p1 <- c(p1[1:2],p1[3]^2,p1[3])
## Likelihood of unrestricted model evaluated in MLE of restricted model
logLik(e,p1)
## Likelihood of restricted model (MLE)
logLik(e1)

##############################
### Non-linear regression
##############################

## Simulate data
m <- lvm(c(y1,y2)~f(x,0)+f(eta,1))
latent(m) <- ~eta
covariance(m,~y1+y2) <- "v"
intercept(m,~y1+y2) <- "mu"
covariance(m,~eta) <- "zeta"
intercept(m,~eta) <- 0
set.seed(1)
d <- sim(m,100,p=c(v=0.01,zeta=0.01))[,manifest(m)]
d <- transform(d,
               y1=y1+2*pnorm(2*x),
               y2=y2+2*pnorm(2*x))

## Specify model and estimate parameters
constrain(m, mu ~ x + alpha + nu + gamma) <- function(x) x[4]*pnorm(x[3]+x[1]*x[2])
\donttest{ ## Reduce Ex.Timings
e <- estimate(m,d,control=list(trace=1,constrain=TRUE))
constraints(e,data=d)
## Plot model-fit
plot(y1~x,d,pch=16); points(y2~x,d,pch=16,col="gray")
x0 <- seq(-4,4,length.out=100)
lines(x0,coef(e)["nu"] + coef(e)["gamma"]*pnorm(coef(e)["alpha"]*x0))
}

##############################
### Multigroup model
##############################
### Define two models
m1 <- lvm(y ~ f(x,beta)+f(z,beta2))
m2 <- lvm(y ~ f(x,psi) + z)
### And simulate data from them
d1 <- sim(m1,500)
d2 <- sim(m2,500)
### Add 'non'-linear parameter constraint
constrain(m2,psi ~ beta2) <- function(x) x
## Add parameter beta2 to model 2, now beta2 exists in both models
parameter(m2) <- ~ beta2
ee <- estimate(list(m1,m2),list(d1,d2),control=list(method="NR"))
summary(ee)

m3 <- lvm(y ~ f(x,beta)+f(z,beta2))
m4 <- lvm(y ~ f(x,beta2) + z)
e2 <- estimate(list(m3,m4),list(d1,d2),control=list(method="NR"))
e2
}
\seealso{
\code{\link{regression}}, \code{\link{intercept}},
\code{\link{covariance}}
}
\author{
Klaus K. Holst
}
\keyword{models}
\keyword{regression}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/estimate.lvm.R
\name{estimate.lvm}
\alias{estimate.lvm}
\title{Estimation of parameters in a Latent Variable Model (lvm)}
\usage{
\method{estimate}{lvm}(
  x,
  data = parent.frame(),
  estimator = NULL,
  control = list(),
  missing = FALSE,
  weights,
  weightsname,
  data2,
  id,
  fix,
  index = !quick,
  graph = FALSE,
  messages = lava.options()$messages,
  quick = FALSE,
  method,
  param,
  cluster,
  p,
  ...
)
}
\arguments{
\item{x}{\code{lvm}-object}

\item{data}{\code{data.frame}}

\item{estimator}{String defining the estimator (see details below)}

\item{control}{control/optimization parameters (see details below)}

\item{missing}{Logical variable indiciating how to treat missing data.
Setting to FALSE leads to complete case analysis. In the other case
likelihood based inference is obtained by integrating out the missing data
under assumption the assumption that data is missing at random (MAR).}

\item{weights}{Optional weights to used by the chosen estimator.}

\item{weightsname}{Weights names (variable names of the model) in case
\code{weights} was given as a vector of column names of \code{data}}

\item{data2}{Optional additional dataset used by the chosen
estimator.}

\item{id}{Vector (or name of column in \code{data}) that identifies
correlated groups of observations in the data leading to variance estimates
based on a sandwich estimator}

\item{fix}{Logical variable indicating whether parameter restriction
automatically should be imposed (e.g. intercepts of latent variables set to
0 and at least one regression parameter of each measurement model fixed to
ensure identifiability.)}

\item{index}{For internal use only}

\item{graph}{For internal use only}

\item{messages}{Control how much information should be
printed during estimation (0: none)}

\item{quick}{If TRUE the parameter estimates are calculated but all
additional information such as standard errors are skipped}

\item{method}{Optimization method}

\item{param}{set parametrization (see \code{help(lava.options)})}

\item{cluster}{Obsolete. Alias for 'id'.}

\item{p}{Evaluate model in parameter 'p' (no optimization)}

\item{...}{Additional arguments to be passed to lower-level functions}
}
\value{
A \code{lvmfit}-object.
}
\description{
Estimate parameters. MLE, IV or user-defined estimator.
}
\details{
A list of parameters controlling the estimation and optimization procedures
is parsed via the \code{control} argument. By default Maximum Likelihood is
used assuming multivariate normal distributed measurement errors. A list
with one or more of the following elements is expected:

\describe{
\item{start:}{Starting value. The order of the parameters can be shown by
calling \code{coef} (with \code{mean=TRUE}) on the \code{lvm}-object or with
\code{plot(..., labels=TRUE)}. Note that this requires a check that it is
actual the model being estimated, as \code{estimate} might add additional
restriction to the model, e.g. through the \code{fix} and \code{exo.fix}
arguments. The \code{lvm}-object of a fitted model can be extracted with the
\code{Model}-function.}

\item{starterfun:}{Starter-function with syntax
\code{function(lvm, S, mu)}.  Three builtin functions are available:
\code{startvalues}, \code{startvalues0}, \code{startvalues1}, ...}

\item{estimator:}{ String defining which estimator to use (Defaults to
``\code{gaussian}'')}

\item{meanstructure}{Logical variable indicating
whether to fit model with meanstructure.}

\item{method:}{ String pointing to
alternative optimizer (e.g. \code{optim} to use simulated annealing).}

\item{control:}{ Parameters passed to the optimizer (default
\code{stats::nlminb}).}

\item{tol:}{ Tolerance of optimization constraints on lower limit of
variance parameters.  } }
}
\examples{
dd <- read.table(header=TRUE,
text="x1 x2 x3
 0.0 -0.5 -2.5
-0.5 -2.0  0.0
 1.0  1.5  1.0
 0.0  0.5  0.0
-2.5 -1.5 -1.0")
e <- estimate(lvm(c(x1,x2,x3)~u),dd)

## Simulation example
m <- lvm(list(y~v1+v2+v3+v4,c(v1,v2,v3,v4)~x))
covariance(m) <- v1~v2+v3+v4
dd <- sim(m,10000) ## Simulate 10000 observations from model
e <- estimate(m, dd) ## Estimate parameters
e

## Using just sufficient statistics
n <- nrow(dd)
e0 <- estimate(m,data=list(S=cov(dd)*(n-1)/n,mu=colMeans(dd),n=n))
rm(dd)

## Multiple group analysis
m <- lvm()
regression(m) <- c(y1,y2,y3)~u
regression(m) <- u~x
d1 <- sim(m,100,p=c("u,u"=1,"u~x"=1))
d2 <- sim(m,100,p=c("u,u"=2,"u~x"=-1))

mm <- baptize(m)
regression(mm,u~x) <- NA
covariance(mm,~u) <- NA
intercept(mm,~u) <- NA
ee <- estimate(list(mm,mm),list(d1,d2))

## Missing data
d0 <- makemissing(d1,cols=1:2)
e0 <- estimate(m,d0,missing=TRUE)
e0
}
\seealso{
estimate.default score, information
}
\author{
Klaus K. Holst
}
\keyword{models}
\keyword{regression}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/twostage.R
\name{twostage}
\alias{twostage}
\title{Two-stage estimator}
\usage{
twostage(object, ...)
}
\arguments{
\item{object}{Model object}

\item{...}{Additional arguments to lower level functions}
}
\description{
Generic function.
}
\seealso{
twostage.lvm twostage.lvmfit twostage.lvm.mixture twostage.estimate
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{plot.lvm}
\alias{plot.lvm}
\alias{plot.lvmfit}
\title{Plot path diagram}
\usage{
\method{plot}{lvm}(
  x,
  diag = FALSE,
  cor = TRUE,
  labels = FALSE,
  intercept = FALSE,
  addcolor = TRUE,
  plain = FALSE,
  cex,
  fontsize1 = 10,
  noplot = FALSE,
  graph = list(rankdir = "BT"),
  attrs = list(graph = graph),
  unexpr = FALSE,
  addstyle = TRUE,
  plot.engine = lava.options()$plot.engine,
  init = TRUE,
  layout = lava.options()$layout,
  edgecolor = lava.options()$edgecolor,
  graph.proc = lava.options()$graph.proc,
  ...
)
}
\arguments{
\item{x}{Model object}

\item{diag}{Logical argument indicating whether to visualize
variance parameters (i.e. diagonal of variance matrix)}

\item{cor}{Logical argument indicating whether to visualize
correlation parameters}

\item{labels}{Logical argument indiciating whether to add labels
to plot (Unnamed parameters will be labeled p1,p2,...)}

\item{intercept}{Logical argument indiciating whether to add
intercept labels}

\item{addcolor}{Logical argument indiciating whether to add colors
to plot (overrides \code{nodecolor} calls)}

\item{plain}{if TRUE strip plot of colors and boxes}

\item{cex}{Fontsize of node labels}

\item{fontsize1}{Fontsize of edge labels}

\item{noplot}{if TRUE then return \code{graphNEL} object only}

\item{graph}{Graph attributes (Rgraphviz)}

\item{attrs}{Attributes (Rgraphviz)}

\item{unexpr}{if TRUE remove expressions from labels}

\item{addstyle}{Logical argument indicating whether additional
style should automatically be added to the plot (e.g. dashed
lines to double-headed arrows)}

\item{plot.engine}{default 'Rgraphviz' if available, otherwise
visNetwork,igraph}

\item{init}{Reinitialize graph (for internal use)}

\item{layout}{Graph layout (see Rgraphviz or igraph manual)}

\item{edgecolor}{if TRUE plot style with colored edges}

\item{graph.proc}{Function that post-process the graph object
(default: subscripts are automatically added to labels of the
nodes)}

\item{...}{Additional arguments to be passed to the low level
functions}
}
\description{
Plot the path diagram of a SEM
}
\examples{

if (interactive()) {
m <- lvm(c(y1,y2) ~ eta)
regression(m) <- eta ~ z+x2
regression(m) <- c(eta,z) ~ x1
latent(m) <- ~eta
labels(m) <- c(y1=expression(y[scriptscriptstyle(1)]),
y2=expression(y[scriptscriptstyle(2)]),
x1=expression(x[scriptscriptstyle(1)]),
x2=expression(x[scriptscriptstyle(2)]),
eta=expression(eta))
edgelabels(m, eta ~ z+x1+x2, cex=2, lwd=3,
           col=c("orange","lightblue","lightblue")) <- expression(rho,phi,psi)
nodecolor(m, vars(m), border="white", labcol="darkblue") <- NA
nodecolor(m, ~y1+y2+z, labcol=c("white","white","black")) <- NA
plot(m,cex=1.5)

d <- sim(m,100)
e <- estimate(m,d)
plot(e)

m <- lvm(c(y1,y2) ~ eta)
regression(m) <- eta ~ z+x2
regression(m) <- c(eta,z) ~ x1
latent(m) <- ~eta
plot(lava:::beautify(m,edgecol=FALSE))
}
}
\author{
Klaus K. Holst
}
\keyword{hplot}
\keyword{regression}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/contr.R
\name{contr}
\alias{contr}
\alias{parsedesign}
\alias{pairwise.diff}
\title{Create contrast matrix}
\usage{
contr(p, n, diff = TRUE, ...)
}
\arguments{
\item{p}{index of non-zero entries (see example)}

\item{n}{Total number of parameters (if omitted the max number in p will be used)}

\item{diff}{If FALSE all non-zero entries are +1, otherwise the second non-zero element in each row will be -1.}

\item{...}{Additional arguments to lower level functions}
}
\description{
Create contrast matrix typically for use with 'estimate' (Wald tests).
}
\examples{
contr(2,n=5)
contr(as.list(2:4),n=5)
contr(list(1,2,4),n=5)
contr(c(2,3,4),n=5)
contr(list(c(1,3),c(2,4)),n=5)
contr(list(c(1,3),c(2,4),5))

parsedesign(c("aa","b","c"),"?","?",diff=c(FALSE,TRUE))

## All pairs comparisons:
pdiff <- function(n) lava::contr(lapply(seq(n-1), function(x) seq(x, n)))
pdiff(4)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/combine.R
\name{Combine}
\alias{Combine}
\title{Report estimates across different models}
\usage{
Combine(x, ...)
}
\arguments{
\item{x}{list of model objects}

\item{...}{additional arguments to lower-level functions}
}
\description{
Report estimates across different models
}
\examples{
data(serotonin)
m1 <- lm(cau ~ age*gene1 + age*gene2,data=serotonin)
m2 <- lm(cau ~ age + gene1,data=serotonin)
m3 <- lm(cau ~ age*gene2,data=serotonin)

Combine(list(A=m1,B=m2,C=m3),fun=function(x)
     c("_____"="",R2=" "\%++\%format(summary(x)$r.squared,digits=2)))
}
\author{
Klaus K. Holst
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/addvar.R
\name{addvar}
\alias{addvar}
\alias{addvar<-}
\title{Add variable to (model) object}
\usage{
addvar(x, ...)
}
\arguments{
\item{x}{Model object}

\item{\dots}{Additional arguments}
}
\description{
Generic method for adding variables to model object
}
\author{
Klaus K. Holst
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/contr.R
\name{contr}
\alias{contr}
\alias{parsedesign}
\alias{pairwise.diff}
\alias{pairwise_diff}
\title{Create contrast matrix}
\usage{
contr(p, n, diff = TRUE, ...)
}
\arguments{
\item{p}{index of non-zero entries (see example)}

\item{n}{Total number of parameters (if omitted the max number in p will be used)}

\item{diff}{If FALSE all non-zero entries are +1, otherwise the second non-zero element in each row will be -1.}

\item{...}{Additional arguments to lower level functions}
}
\description{
Create contrast matrix typically for use with 'estimate' (Wald tests).
}
\examples{
contr(2,n=5)
contr(as.list(2:4),n=5)
contr(list(1,2,4),n=5)
contr(c(2,3,4),n=5)
contr(list(c(1,3),c(2,4)),n=5)
contr(list(c(1,3),c(2,4),5))

parsedesign(c("aa","b","c"),"?","?",diff=c(FALSE,TRUE))

## All pairs comparisons:
pdiff <- function(n) lava::contr(lapply(seq(n-1), function(x) seq(x, n)))
pdiff(4)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zcolorbar.R
\name{colorbar}
\alias{colorbar}
\title{Add color-bar to plot}
\usage{
colorbar(
  clut = Col(rev(rainbow(11, start = 0, end = 0.69)), alpha),
  x.range = c(-0.5, 0.5),
  y.range = c(-0.1, 0.1),
  values = seq(clut),
  digits = 2,
  label.offset,
  srt = 45,
  cex = 0.5,
  border = NA,
  alpha = 0.5,
  position = 1,
  direction = c("horizontal", "vertical"),
  ...
)
}
\arguments{
\item{clut}{Color look-up table}

\item{x.range}{x range}

\item{y.range}{y range}

\item{values}{label values}

\item{digits}{number of digits}

\item{label.offset}{label offset}

\item{srt}{rotation of labels}

\item{cex}{text size}

\item{border}{border of color bar rectangles}

\item{alpha}{Alpha (transparency) level 0-1}

\item{position}{Label position left/bottom (1) or top/right (2) or no text (0)}

\item{direction}{horizontal or vertical color bars}

\item{\dots}{additional low level arguments (i.e. parsed to \code{text})}
}
\description{
Add color-bar to plot
}
\examples{
\dontrun{
plotNeuro(x,roi=R,mm=-18,range=5)
colorbar(clut=Col(rev(rainbow(11,start=0,end=0.69)),0.5),
         x=c(-40,40),y.range=c(84,90),values=c(-5:5))

colorbar(clut=Col(rev(rainbow(11,start=0,end=0.69)),0.5),
         x=c(-10,10),y.range=c(-100,50),values=c(-5:5),
         direction="vertical",border=1)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rotation.R
\name{rotate2}
\alias{rotate2}
\alias{rot2D}
\alias{rot3D}
\title{Performs a rotation in the plane}
\usage{
rotate2(x, theta = pi)
}
\arguments{
\item{x}{Matrix to be rotated (2 times n)}

\item{theta}{Rotation in radians}
}
\value{
Returns a matrix of the same dimension as \code{x}
}
\description{
Performs a rotation in the plane
}
\examples{
rotate2(cbind(c(1,2),c(2,1)))
}
\author{
Klaus K. Holst
}
\keyword{hplot}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/colorbar.R
\name{colorbar}
\alias{colorbar}
\title{Add color-bar to plot}
\usage{
colorbar(
  clut = Col(rev(rainbow(11, start = 0, end = 0.69)), alpha),
  x.range = c(-0.5, 0.5),
  y.range = c(-0.1, 0.1),
  values = seq(clut),
  digits = 2,
  label.offset,
  srt = 45,
  cex = 0.5,
  border = NA,
  alpha = 0.5,
  position = 1,
  direction = c("horizontal", "vertical"),
  ...
)
}
\arguments{
\item{clut}{Color look-up table}

\item{x.range}{x range}

\item{y.range}{y range}

\item{values}{label values}

\item{digits}{number of digits}

\item{label.offset}{label offset}

\item{srt}{rotation of labels}

\item{cex}{text size}

\item{border}{border of color bar rectangles}

\item{alpha}{Alpha (transparency) level 0-1}

\item{position}{Label position left/bottom (1) or top/right (2) or no text (0)}

\item{direction}{horizontal or vertical color bars}

\item{\dots}{additional low level arguments (i.e. parsed to \code{text})}
}
\description{
Add color-bar to plot
}
\examples{
\dontrun{
plotNeuro(x,roi=R,mm=-18,range=5)
colorbar(clut=Col(rev(rainbow(11,start=0,end=0.69)),0.5),
         x=c(-40,40),y.range=c(84,90),values=c(-5:5))

colorbar(clut=Col(rev(rainbow(11,start=0,end=0.69)),0.5),
         x=c(-10,10),y.range=c(-100,50),values=c(-5:5),
         direction="vertical",border=1)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lava-package.R
\docType{data}
\name{serotonin}
\alias{serotonin}
\title{Serotonin data}
\format{
data.frame
}
\source{
Simulated
}
\description{
This simulated data mimics a PET imaging study where the 5-HT2A
receptor and serotonin transporter (SERT) binding potential has
been quantified into 8 different regions. The 5-HT2A
cortical regions are considered high-binding regions
measurements.  These measurements can be regarded as proxy measures of
the extra-cellular levels of serotonin in the brain
\tabular{rll}{
day    \tab numeric \tab Scan day of the year \cr
age    \tab numeric \tab Age at baseline scan \cr
mem    \tab numeric \tab Memory performance score \cr
depr   \tab numeric \tab Depression (mild) status 500 days after baseline \cr
gene1  \tab numeric \tab Gene marker 1 (HTR2A) \cr
gene2  \tab numeric \tab Gene marker 2 (HTTTLPR) \cr
cau \tab numeric \tab SERT binding, Caudate Nucleus \cr
th  \tab numeric \tab SERT binding, Thalamus \cr
put \tab numeric \tab SERT binding, Putamen \cr
mid \tab numeric \tab SERT binding, Midbrain \cr
aci \tab numeric \tab 5-HT2A binding, Anterior cingulate gyrus \cr
pci  \tab numeric \tab 5-HT2A binding, Posterior cingulate gyrus \cr
sfc \tab numeric \tab 5-HT2A binding, Superior frontal cortex \cr
par \tab numeric \tab 5-HT2A binding, Parietal cortex \cr
}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/measurement_error.R
\name{measurement_error}
\alias{measurement_error}
\alias{measurement.error}
\title{Two-stage (non-linear) measurement error}
\usage{
measurement_error(
  model1,
  formula,
  data = parent.frame(),
  predictfun = function(mu, var, data, ...) mu[, 1]^2 + var[1],
  id1,
  id2,
  ...
)
}
\arguments{
\item{model1}{Stage 1 model}

\item{formula}{Formula specifying observed covariates in stage 2 model}

\item{data}{data.frame}

\item{predictfun}{Predictions to be used in stage 2}

\item{id1}{Optional id-vector of stage 1}

\item{id2}{Optional id-vector of stage 2}

\item{...}{Additional arguments to lower level functions}
}
\description{
Two-stage measurement error
}
\examples{
m <- lvm(c(y1,y2,y3)~u,c(y3,y4,y5)~v,u~~v,c(u,v)~x)
transform(m,u2~u) <- function(x) x^2
transform(m,uv~u+v) <- prod
regression(m) <- z~u2+u+v+uv+x
set.seed(1)
d <- sim(m,1000,p=c("u,u"=1))

## Stage 1
m1 <- lvm(c(y1[0:s],y2[0:s],y3[0:s])~1*u,c(y3[0:s],y4[0:s],y5[0:s])~1*v,u~b*x,u~~v)
latent(m1) <- ~u+v
e1 <- estimate(m1,d)

pp <- function(mu,var,data,...) {
    cbind(u=mu[,"u"],u2=mu[,"u"]^2+var["u","u"],
          v=mu[,"v"],uv=mu[,"u"]*mu[,"v"]+var["u","v"])
}
(e <- measurement_error(e1, z~1+x, data=d, predictfun=pp))

## uu <- seq(-1,1,length.out=100)
## pp <- estimate(e,function(p,...) p["(Intercept)"]+p["u"]*uu+p["u2"]*uu^2)$coefmat
if (interactive()) {
    plot(e,intercept=TRUE,line=0)
    dev.off()
    f <- function(p) p[1]+p["u"]*u+p["u2"]*u^2
    u <- seq(-1,1,length.out=100)
    plot(e, f, data=data.frame(u), ylim=c(-.5,2.5))
}
}
\seealso{
stack.estimate
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lava-package.R
\docType{package}
\name{lava-package}
\alias{lava-package}
\alias{lava}
\title{lava: Latent Variable Models}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

A general implementation of Structural Equation Models with latent variables (MLE, 2SLS, and composite likelihood estimators) with both continuous, censored, and ordinal outcomes (Holst and Budtz-Joergensen (2013) \doi{10.1007/s00180-012-0344-y}). Mixture latent variable models and non-linear latent variable models (Holst and Budtz-Joergensen (2020) \doi{10.1093/biostatistics/kxy082}). The package also provides methods for graph exploration (d-separation, back-door criterion), simulation of general non-linear latent variable models, and estimation of influence functions for a broad range of statistical models.

A general implementation of Structural Equation Models wth latent variables
(MLE, 2SLS, and composite likelihood estimators) with both continuous,
censored, and ordinal outcomes (Holst and Budtz-Joergensen (2013)
\url{doi:10.1007/s00180-012-0344-y}). Mixture latent variable models and
non-linear latent variable models (Holst and Budtz-Joergensen (2020)
\url{doi:10.1093/biostatistics/kxy082}). The package also provides methods for
graph exploration (d-separation, back-door criterion), simulation of general
non-linear latent variable models, and estimation of influence functions for
a broad range of statistical models.
}
\examples{
lava()
}
\seealso{
Useful links:
\itemize{
  \item \url{https://kkholst.github.io/lava/}
  \item \url{https://github.com/kkholst/lava}
  \item Report bugs at \url{https://github.com/kkholst/lava/issues}
}

}
\author{
\strong{Maintainer}: Klaus K. Holst \email{klaus@holst.it}

Other contributors:
\itemize{
  \item Brice Ozenne [contributor]
  \item Thomas Gerds [contributor]
}


Klaus K. Holst Maintainer: \href{mailto:klaus@holst.it}{klaus@holst.it}
}
\keyword{internal}
\keyword{package}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/measurement.error.R
\name{measurement.error}
\alias{measurement.error}
\title{Two-stage (non-linear) measurement error}
\usage{
measurement.error(
  model1,
  formula,
  data = parent.frame(),
  predictfun = function(mu, var, data, ...) mu[, 1]^2 + var[1],
  id1,
  id2,
  ...
)
}
\arguments{
\item{model1}{Stage 1 model}

\item{formula}{Formula specifying observed covariates in stage 2 model}

\item{data}{data.frame}

\item{predictfun}{Predictions to be used in stage 2}

\item{id1}{Optional id-vector of stage 1}

\item{id2}{Optional id-vector of stage 2}

\item{...}{Additional arguments to lower level functions}
}
\description{
Two-stage measurement error
}
\examples{
m <- lvm(c(y1,y2,y3)~u,c(y3,y4,y5)~v,u~~v,c(u,v)~x)
transform(m,u2~u) <- function(x) x^2
transform(m,uv~u+v) <- prod
regression(m) <- z~u2+u+v+uv+x
set.seed(1)
d <- sim(m,1000,p=c("u,u"=1))

## Stage 1
m1 <- lvm(c(y1[0:s],y2[0:s],y3[0:s])~1*u,c(y3[0:s],y4[0:s],y5[0:s])~1*v,u~b*x,u~~v)
latent(m1) <- ~u+v
e1 <- estimate(m1,d)

pp <- function(mu,var,data,...) {
    cbind(u=mu[,"u"],u2=mu[,"u"]^2+var["u","u"],
          v=mu[,"v"],uv=mu[,"u"]*mu[,"v"]+var["u","v"])
}
(e <- measurement.error(e1, z~1+x, data=d, predictfun=pp))

## uu <- seq(-1,1,length.out=100)
## pp <- estimate(e,function(p,...) p["(Intercept)"]+p["u"]*uu+p["u2"]*uu^2)$coefmat
if (interactive()) {
    plot(e,intercept=TRUE,line=0)
    dev.off()
    f <- function(p) p[1]+p["u"]*u+p["u2"]*u^2
    u <- seq(-1,1,length.out=100)
    plot(e, f, data=data.frame(u), ylim=c(-.5,2.5))
}
}
\seealso{
stack.estimate
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/estimate.list.R
\name{estimate.list}
\alias{estimate.list}
\title{Estimate method for lists}
\usage{
\method{estimate}{list}(x, ...)
}
\arguments{
\item{x}{list of model objects or \code{lvm} objects}

\item{...}{Additional arguments to lower level functions}
}
\value{
Object of class \code{estimate.list} (a list of \code{estimate} objects),
or a fitted multi-group model if input is a list of \code{lvm} objects.
}
\description{
Estimate method for lists
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/correlation.R
\name{correlation}
\alias{correlation}
\title{Generic method for extracting correlation coefficients of model object}
\usage{
correlation(x, ...)
}
\arguments{
\item{x}{Object}

\item{\dots}{Additional arguments}
}
\description{
Generic correlation method
}
\author{
Klaus K. Holst
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/devcoords.R
\name{devcoords}
\alias{devcoords}
\title{Returns device-coordinates and plot-region}
\usage{
devcoords()
}
\value{
A \code{list} with elements
\itemize{
\item \code{dev.x1}: device left x-coordinate
\item \code{dev.x2}: device right x-coordinate
\item \code{dev.y1}: device bottom y-coordinate
\item \code{dev.y2}: device top y-coordinate
\item \code{fig.x1}: plot left x-coordinate
\item \code{fig.x2}: plot right x-coordinate
\item \code{fig.y1}: plot bottom y-coordinate
\item \code{fig.y2}: plot top y-coordinate
}
}
\description{
Returns device-coordinates and plot-region
}
\author{
Klaus K. Holst
}
\keyword{hplot}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lava-package.R
\docType{package}
\name{lava-package}
\alias{lava-package}
\alias{lava}
\title{lava: Latent Variable Models}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

A general implementation of Structural Equation Models with latent variables (MLE, 2SLS, and composite likelihood estimators) with both continuous, censored, and ordinal outcomes (Holst and Budtz-Joergensen (2013) \doi{10.1007/s00180-012-0344-y}). Mixture latent variable models and non-linear latent variable models (Holst and Budtz-Joergensen (2020) \doi{10.1093/biostatistics/kxy082}). The package also provides methods for graph exploration (d-separation, back-door criterion), simulation of general non-linear latent variable models, and estimation of influence functions for a broad range of statistical models.

A general implementation of Structural Equation Models wth latent variables
(MLE, 2SLS, and composite likelihood estimators) with both continuous,
censored, and ordinal outcomes (Holst and Budtz-Joergensen (2013)
\url{doi:10.1007/s00180-012-0344-y}). Mixture latent variable models and
non-linear latent variable models (Holst and Budtz-Joergensen (2020)
\url{doi:10.1093/biostatistics/kxy082}). The package also provides methods for
graph exploration (d-separation, back-door criterion), simulation of general
non-linear latent variable models, and estimation of influence functions for
a broad range of statistical models.
}
\examples{
lava()
}
\seealso{
Useful links:
\itemize{
  \item \url{https://kkholst.github.io/lava/}
  \item \url{https://github.com/kkholst/lava}
  \item Report bugs at \url{https://github.com/kkholst/lava/issues}
}

}
\author{
\strong{Maintainer}: Klaus K. Holst \email{klaus@holst.it}

Authors:
\itemize{
  \item Klaus K. Holst \email{klaus@holst.it}
}

Other contributors:
\itemize{
  \item Benedikt Sommer [contributor]
  \item Brice Ozenne [contributor]
  \item Thomas Gerds [contributor]
}


Klaus K. Holst Maintainer: \href{mailto:klaus@holst.it}{klaus@holst.it}
}
\keyword{internal}
\keyword{package}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotConf.R
\name{plotConf}
\alias{plotConf}
\title{Plot regression lines}
\usage{
plotConf(
  model,
  var1 = NULL,
  var2 = NULL,
  data = NULL,
  ci.lty = 0,
  ci = TRUE,
  level = 0.95,
  pch = 16,
  lty = 1,
  lwd = 2,
  npoints = 100,
  xlim,
  col = NULL,
  colpt,
  alpha = 0.5,
  cex = 1,
  delta = 0.07,
  centermark = 0.03,
  jitter = 0.2,
  cidiff = FALSE,
  mean = TRUE,
  legend = ifelse(is.null(var1), FALSE, "topright"),
  trans = function(x) {
     x
 },
  partres = inherits(model, "lm"),
  partse = FALSE,
  labels,
  vcov,
  predictfun,
  plot = TRUE,
  new = TRUE,
  ...
)
}
\arguments{
\item{model}{Model object (e.g. \code{\link[=lm]{lm()}})}

\item{var1}{predictor (Continuous or factor)}

\item{var2}{Factor that interacts with \code{var1}}

\item{data}{data.frame to use for prediction (model.frame is used as default)}

\item{ci.lty}{Line type for confidence limits}

\item{ci}{Boolean indicating wether to draw pointwise 95\% confidence limits}

\item{level}{Level of confidence limits (default 95\%)}

\item{pch}{Point type for partial residuals}

\item{lty}{Line type for estimated regression lines}

\item{lwd}{Line width for regression lines}

\item{npoints}{Number of points used to plot curves}

\item{xlim}{Range of x axis}

\item{col}{Color (for each level in \code{var2})}

\item{colpt}{Color of partial residual points}

\item{alpha}{Alpha level}

\item{cex}{Point size}

\item{delta}{For categorical \code{var1}}

\item{centermark}{For categorical \code{var1}}

\item{jitter}{For categorical \code{var1}}

\item{cidiff}{For categorical \code{var1}}

\item{mean}{For categorical \code{var1}}

\item{legend}{Boolean (add legend)}

\item{trans}{Transform estimates (e.g. exponential)}

\item{partres}{Boolean indicating whether to plot partial residuals}

\item{partse}{.}

\item{labels}{Optional labels of \code{var2}}

\item{vcov}{Optional variance estimates}

\item{predictfun}{Optional predict-function used to calculate confidence limits and predictions}

\item{plot}{If FALSE return only predictions and confidence bands}

\item{new}{If FALSE add to current plot}

\item{\dots}{additional arguments to lower level functions}
}
\value{
list with following members:
\itemize{
\item \code{x} Variable on the x-axis \code{var1}
\item \code{y} Variable on the y-axis (partial residuals)
\item \code{predict} Matrix with confidence limits and predicted startvalues.R
}
}
\description{
Plot regression line (with interactions) and partial residuals.
}
\examples{
n <- 100
x0 <- rnorm(n)
x1 <- seq(-3,3, length.out=n)
x2 <- factor(rep(c(1,2),each=n/2), labels=c("A","B"))
y <- 5 + 2*x0 + 0.5*x1 + -1*(x2=="B")*x1 + 0.5*(x2=="B") + rnorm(n, sd=0.25)
dd <- data.frame(y=y, x1=x1, x2=x2)
lm0 <- lm(y ~ x0 + x1*x2, dd)
plotConf(lm0, var1="x1", var2="x2")
abline(a=5,b=0.5,col="red")
abline(a=5.5,b=-0.5,col="red")
### points(5+0.5*x1 -1*(x2=="B")*x1 + 0.5*(x2=="B") ~ x1, cex=2)

data(iris)
l <- lm(Sepal.Length ~ Sepal.Width*Species,iris)
plotConf(l,var2="Species")
plotConf(l,var1="Sepal.Width",var2="Species")

\dontrun{
## lme4 model
dd$Id <- rbinom(n, size = 3, prob = 0.3)
lmer0 <- lme4::lmer(y ~ x0 + x1*x2 + (1|Id), dd)
plotConf(lmer0, var1="x1", var2="x2")
}
}
\seealso{
\code{\link[=termplot]{termplot()}}
}
\author{
Klaus K. Holst
}
\keyword{hplot}
\keyword{regression}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/devcoords.R
\name{devcoords}
\alias{devcoords}
\title{Returns device-coordinates and plot-region}
\usage{
devcoords()
}
\value{
A \code{list} with elements
 \item{dev.x1}{Device: Left x-coordinate}
 \item{dev.x2}{Device: Right x-coordinate}
 \item{dev.y1}{Device Bottom y-coordinate}
 \item{dev.y2}{Device Top y-coordinate}
 \item{fig.x1}{Plot: Left x-coordinate}
 \item{fig.x2}{Plot: Right x-coordinate}
 \item{fig.y1}{Plot: Bottom y-coordinate}
 \item{fig.y2}{Plot: Top y-coordinate}
}
\description{
Returns device-coordinates and plot-region
}
\author{
Klaus K. Holst
}
\keyword{hplot}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/glmest.R
\name{predict_glm}
\alias{predict_glm}
\title{Predict from a GLM with modified coefficients}
\usage{
predict_glm(
  object,
  p = coef(object),
  data,
  offset = NULL,
  type = c("response", "link"),
  ...
)
}
\arguments{
\item{object}{A fitted \code{glm} object.}

\item{p}{Numeric vector of coefficients (defaults to \code{coef(x)}).}

\item{data}{Optional data frame for computing the model matrix and
response. If missing, the original model matrix from the fitted
object is used.}

\item{offset}{Optional offset vector. If \code{NULL} (default), the
offset stored in the fitted model (\code{x$offset}) is used.}

\item{type}{Character; \code{"response"} (default) returns predictions on
the response scale via the inverse link function, \code{"link"} returns
predictions on the linear predictor scale.}

\item{...}{Additional arguments (currently unused).}
}
\value{
Numeric vector of predictions with a \code{"grad"} attribute
containing the gradient (Jacobian) of predictions with respect to
the coefficients. When \code{type = "link"}, the gradient is simply the
model matrix \code{X}.
}
\description{
Compute predictions from a fitted \link[stats:glm]{stats::glm} object, optionally
substituting new parameter values. Unlike \link[stats:predict.glm]{stats::predict.glm}, this
function allows the user to supply an arbitrary coefficient vector \code{p},
which is useful for computing predictions at counterfactual parameter
values (e.g., during optimization or simulation). The returned value
also includes a \code{"grad"} attribute containing the Jacobian of predictions
with respect to the coefficients.
}
\examples{
m <- glm(mpg ~ hp + wt, data = mtcars)
p0 <- coef(m)
# Predictions at fitted coefficients match predict.glm
all.equal(as.numeric(predict_glm(m)), fitted(m))
# Predictions with modified coefficients
predict_glm(m, p = p0 * 1.1)
}
\seealso{
\link[stats:predict.glm]{stats::predict.glm}, \link{score.glm}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/predict.R
\name{predict.lvm}
\alias{predict.lvm}
\alias{predict.lvmfit}
\title{Prediction in structural equation models}
\usage{
\method{predict}{lvm}(
  object,
  x = NULL,
  y = NULL,
  residual = FALSE,
  p,
  data,
  path = FALSE,
  quick = is.null(x) & !(residual | path),
  ...
)
}
\arguments{
\item{object}{Model object}

\item{x}{optional list of (endogenous) variables to condition on}

\item{y}{optional subset of variables to predict}

\item{residual}{If true the residuals are predicted}

\item{p}{Parameter vector}

\item{data}{Data to use in prediction}

\item{path}{Path prediction}

\item{quick}{If TRUE the conditional mean and variance given covariates are returned (and all other calculations skipped)}

\item{\dots}{Additional arguments to lower level function}
}
\description{
Prediction in structural equation models
}
\examples{
m <- lvm(list(c(y1,y2,y3)~u,u~x)); latent(m) <- ~u
d <- sim(m,100)
e <- estimate(m,d)

## Conditional mean (and variance as attribute) given covariates
r <- predict(e)
## Best linear unbiased predictor (BLUP)
r <- predict(e,vars(e))
##  Conditional mean of y3 giving covariates and y1,y2
r <- predict(e,y3~y1+y2)
##  Conditional mean  gives covariates and y1
r <- predict(e,~y1)
##  Predicted residuals (conditional on all observed variables)
r <- predict(e,vars(e),residual=TRUE)

}
\seealso{
predictlvm
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dsep.R
\name{dsep.lvm}
\alias{dsep.lvm}
\alias{dsep}
\title{Check d-separation criterion}
\usage{
\method{dsep}{lvm}(object, x, cond = NULL, return.graph = FALSE, ...)
}
\arguments{
\item{object}{lvm object}

\item{x}{Variables for which to check for conditional independence}

\item{cond}{Conditioning set}

\item{return.graph}{If TRUE the moralized ancestral graph with the
conditioning set removed is returned}

\item{...}{Additional arguments to lower level functions}
}
\description{
Check for conditional independence (d-separation)
}
\details{
The argument \code{x} can be given as a formula, e.g. \code{x~y|z+v}
or \code{~x+y|z+v} With everything on the rhs of the bar defining the
variables on which to condition on.
}
\examples{
m <- lvm(x5 ~ x4+x3, x4~x3+x1, x3~x2, x2~x1)
if (interactive()) {
plot(m,layoutType='neato')
}
dsep(m,x5~x1|x2+x4)
dsep(m,x5~x1|x3+x4)
dsep(m,~x1+x2+x3|x4)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Inverse.R
\name{Inverse}
\alias{Inverse}
\title{Generalized matrix inverse}
\usage{
Inverse(
  X,
  tol = lava.options()$itol,
  det = TRUE,
  names = !chol,
  chol = FALSE,
  symmetric = FALSE
)
}
\arguments{
\item{X}{nxn matrix}

\item{tol}{tolerance for pseudo inverse}

\item{det}{logical, if true the determinant is returned}

\item{names}{preserve dimnames}

\item{chol}{use Cholesky decomposition for calculating inverse otherwise SVD}

\item{symmetric}{set to true if matrix is symmetric}
}
\description{
Generalized matrix inverse
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lava-package.R
\docType{data}
\name{hubble}
\alias{hubble}
\title{Hubble data}
\format{
data.frame
}
\source{
Freedman, W. L., et al. 2001, AstroPhysicalJournal, 553, 47.
}
\description{
Velocity (v) and distance (D) measures of 36 Type Ia super-novae from the
Hubble Space Telescope
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lava-package.R
\docType{data}
\name{brisa}
\alias{brisa}
\title{Simulated data}
\format{
data.frame
}
\source{
Simulated
}
\description{
Simulated data
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/img.R
\name{images}
\alias{images}
\title{Organize several image calls (for visualizing categorical data)}
\usage{
images(
  x,
  group,
  ncol = 2,
  byrow = TRUE,
  colorbar = 1,
  colorbar.space = 0.1,
  label.offset = 0.02,
  order = TRUE,
  colorbar.border = 0,
  main,
  rowcol = FALSE,
  plotfun = NULL,
  axis1,
  axis2,
  mar,
  col = list(c("#EFF3FF", "#BDD7E7", "#6BAED6", "#2171B5"), c("#FEE5D9", "#FCAE91",
    "#FB6A4A", "#CB181D"), c("#EDF8E9", "#BAE4B3", "#74C476", "#238B45"), c("#FEEDDE",
    "#FDBE85", "#FD8D3C", "#D94701")),
  ...
)
}
\arguments{
\item{x}{data.frame or matrix}

\item{group}{group variable}

\item{ncol}{number of columns in layout}

\item{byrow}{organize by row if TRUE}

\item{colorbar}{Add color bar}

\item{colorbar.space}{Space around color bar}

\item{label.offset}{label offset}

\item{order}{order}

\item{colorbar.border}{Add border around color bar}

\item{main}{Main title}

\item{rowcol}{switch rows and columns}

\item{plotfun}{Alternative plot function (instead of 'image')}

\item{axis1}{Axis 1}

\item{axis2}{Axis 2}

\item{mar}{Margins}

\item{col}{Colours}

\item{...}{Additional arguments to lower level graphics functions}
}
\description{
Visualize categorical by group variable
}
\examples{
X <- matrix(rbinom(400,3,0.5),20)
group <- rep(1:4,each=5)
images(X,colorbar=0,zlim=c(0,3))
images(X,group=group,zlim=c(0,3))
\dontrun{
images(X,group=group,col=list(RColorBrewer::brewer.pal(4,"Purples"),
                               RColorBrewer::brewer.pal(4,"Greys"),
                               RColorBrewer::brewer.pal(4,"YlGn"),
                               RColorBrewer::brewer.pal(4,"PuBuGn")),colorbar=2,zlim=c(0,3))
}
images(list(X,X,X,X),group=group,zlim=c(0,3))
images(list(X,X,X,X),ncol=1,group=group,zlim=c(0,3))
images(list(X,X),group,axis2=c(FALSE,FALSE),axis1=c(FALSE,FALSE),
      mar=list(c(0,0,0,0),c(0,0,0,0)),yaxs="i",xaxs="i",zlim=c(0,3))
}
\author{
Klaus Holst
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spaghetti.R
\name{spaghetti}
\alias{spaghetti}
\title{Spaghetti plot}
\usage{
spaghetti(
  formula,
  data = NULL,
  id = "id",
  group = NULL,
  type = "o",
  lty = 1,
  pch = NA,
  col = 1:10,
  alpha = 0.3,
  lwd = 1,
  level = 0.95,
  trend.formula = formula,
  tau = NULL,
  trend.lty = 1,
  trend.join = TRUE,
  trend.delta = 0.2,
  trend = !is.null(tau),
  trend.col = col,
  trend.alpha = 0.2,
  trend.lwd = 3,
  trend.jitter = 0,
  legend = NULL,
  by = NULL,
  xlab = "Time",
  ylab = "",
  add = FALSE,
  ...
)
}
\arguments{
\item{formula}{Formula (response ~ time)}

\item{data}{data.frame}

\item{id}{Id variable}

\item{group}{group variable}

\item{type}{Type (line 'l', stair 's', ...)}

\item{lty}{Line type}

\item{pch}{Colour}

\item{col}{Colour}

\item{alpha}{transparency (0-1)}

\item{lwd}{Line width}

\item{level}{Confidence level}

\item{trend.formula}{Formula for trendline}

\item{tau}{Quantile to estimate (trend)}

\item{trend.lty}{Trend line type}

\item{trend.join}{Trend polygon}

\item{trend.delta}{Length of limit bars}

\item{trend}{Add trend line}

\item{trend.col}{Colour of trend line}

\item{trend.alpha}{Transparency}

\item{trend.lwd}{Trend line width}

\item{trend.jitter}{Jitter amount}

\item{legend}{Legend}

\item{by}{make separate plot for each level in 'by' (formula, name of column, or vector)}

\item{xlab}{Label of X-axis}

\item{ylab}{Label of Y-axis}

\item{add}{Add to existing device}

\item{...}{Additional arguments to lower level arguments}
}
\description{
Spaghetti plot for longitudinal data
}
\examples{
if (interactive() & requireNamespace("mets")) {
K <- 5
y <- "y"\%++\%seq(K)
m <- lvm()
regression(m,y=y,x=~u) <- 1
regression(m,y=y,x=~s) <- seq(K)-1
regression(m,y=y,x=~x) <- "b"
N <- 50
d <- sim(m,N); d$z <- rbinom(N,1,0.5)
dd <- mets::fast.reshape(d); dd$num <- dd$num+3
spaghetti(y~num,dd,id="id",lty=1,col=Col(1,.4),
          trend.formula=~factor(num),trend=TRUE,trend.col="darkblue")
dd$num <- dd$num+rnorm(nrow(dd),sd=0.5) ## Unbalance
spaghetti(y~num,dd,id="id",lty=1,col=Col(1,.4),
          trend=TRUE,trend.col="darkblue")
spaghetti(y~num,dd,id="id",lty=1,col=Col(1,.4),
           trend.formula=~num+I(num^2),trend=TRUE,trend.col="darkblue")
}
}
\author{
Klaus K. Holst
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parpos.R
\name{parpos}
\alias{parpos}
\title{Generic method for finding indeces of model parameters}
\usage{
parpos(x, ...)
}
\arguments{
\item{x}{Model object}

\item{\dots}{Additional arguments}
}
\description{
Generic method for finding indeces of model parameters
}
\author{
Klaus K. Holst
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wkm.R
\name{wkm}
\alias{wkm}
\title{Weighted K-means}
\usage{
wkm(
  x,
  mu,
  data,
  weights = rep(1, NROW(x)),
  iter.max = 20,
  n.start = 5,
  init = "kmpp",
  ...
)
}
\arguments{
\item{x}{Data (or formula)}

\item{mu}{Initial centers (or number centers chosen randomly among x)}

\item{data}{optional data frmae}

\item{weights}{Optional weights}

\item{iter.max}{Max number of iterations}

\item{n.start}{Number of restarts}

\item{init}{method to create initial centres (default kmeans++)}

\item{...}{Additional arguments to lower level functions}
}
\description{
Weighted K-means via Lloyd's algorithm
}
\author{
Klaus K. Holst
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/backdoor.R
\name{backdoor}
\alias{backdoor}
\title{Backdoor criterion}
\usage{
backdoor(object, f, cond, ..., return.graph = FALSE)
}
\arguments{
\item{object}{lvm object}

\item{f}{formula. Conditioning, z, set can be given as y~x|z}

\item{cond}{Vector of variables to conditon on}

\item{...}{Additional arguments to lower level functions}

\item{return.graph}{Return moral ancestral graph with z and effects from x removed}
}
\description{
Check backdoor criterion of a lvm object
}
\examples{
m <- lvm(y~c2,c2~c1,x~c1,m1~x,y~m1, v1~c3, x~c3,v1~y,
         x~z1, z2~z1, z2~z3, y~z3+z2+g1+g2+g3)
ll <- backdoor(m, y~x)
backdoor(m, y~x|c1+z1+g1)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mixture.R
\name{mixture}
\alias{mixture}
\title{Estimate mixture latent variable model.}
\usage{
mixture(
  x,
  data,
  k = length(x),
  control = list(),
  vcov = "observed",
  names = FALSE,
  ...
)
}
\arguments{
\item{x}{List of \code{lvm} objects. If only a single \code{lvm} object is
given, then a \code{k}-mixture of this model is fitted (free parameters
varying between mixture components).}

\item{data}{\code{data.frame}}

\item{k}{Number of mixture components}

\item{control}{Optimization parameters (see details)
#type Type of EM algorithm (standard, classification, stochastic)}

\item{vcov}{of asymptotic covariance matrix (NULL to omit)}

\item{names}{If TRUE returns the names of the parameters (for defining starting values)}

\item{...}{Additional arguments parsed to lower-level functions}
}
\description{
Estimate mixture latent variable model
}
\details{
Estimate parameters in a mixture of latent variable models via the EM
algorithm.

The performance of the EM algorithm can be tuned via the \code{control}
argument, a list where a subset of the following members can be altered:

\describe{ \item{start}{Optional starting values} \item{nstart}{Evaluate
\code{nstart} different starting values and run the EM-algorithm on the
parameters with largest likelihood} \item{tol}{Convergence tolerance of the
EM-algorithm.  The algorithm is stopped when the absolute change in
likelihood and parameter (2-norm) between successive iterations is less than
\code{tol}} \item{iter.max}{Maximum number of iterations of the
EM-algorithm} \item{gamma}{Scale-down (i.e. number between 0 and 1) of the
step-size of the Newton-Raphson algorithm in the M-step} \item{trace}{Trace
information on the EM-algorithm is printed on every \code{trace}th
iteration} }

Note that the algorithm can be aborted any time (C-c) and still be saved
(via on.exit call).
}
\examples{

\donttest{
m0 <- lvm(list(y~x+z,x~z))
distribution(m0,~z) <- binomial.lvm()
d <- sim(m0,2000,p=c("y~z"=2,"y~x"=1),seed=1)

## unmeasured confounder example
m <- baptize(lvm(y~x, x~1));
intercept(m,~x+y) <- NA

if (requireNamespace('mets', quietly=TRUE)) {
  set.seed(42)
  M <- mixture(m,k=2,data=d,control=list(trace=1,tol=1e-6))
  summary(M)
  lm(y~x,d)
  estimate(M,"y~x")
  ## True slope := 1
}
}

}
\seealso{
\code{mvnmix}
}
\author{
Klaus K. Holst
}
\keyword{models}
\keyword{regression}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lava-package.R
\docType{data}
\name{hubble2}
\alias{hubble2}
\title{Hubble data}
\format{
data.frame
}
\description{
Hubble data
}
\seealso{
hubble
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wkm.R
\name{wkm}
\alias{wkm}
\title{Weighted K-means}
\usage{
wkm(
  x,
  mu,
  data,
  weights = rep(1, NROW(x)),
  iter.max = 20,
  n.start = 5,
  init = "kmpp",
  ...
)
}
\arguments{
\item{x}{Data (or formula)}

\item{mu}{Initial centers (or number centers chosen randomly among x)}

\item{data}{optional data frmae}

\item{weights}{Optional weights}

\item{iter.max}{Max number of iterations}

\item{n.start}{Number of restarts}

\item{init}{method to create initial centres (default kmeans++)}

\item{...}{Additional arguments to lower level functions}
}
\description{
Weighted K-means via Lloyd's algorithm
}
\examples{
## Two well-separated Gaussian blobs in 2-D
set.seed(1)
x <- rbind(matrix(rnorm(100, mean = -3), ncol = 2),
           matrix(rnorm(100, mean =  3), ncol = 2))
res <- wkm(x, mu = 2)
table(res$cluster)
res$center

## Supply explicit initial centers (as a list)
res2 <- wkm(x, mu = list(c(-3, -3), c(3, 3)))

## Weighted clustering: up-weight the second blob
w <- c(rep(1, 50), rep(10, 50))
res3 <- wkm(x, mu = 2, weights = w)

## Formula interface on a data.frame
wkm(~ Sepal.Length + Sepal.Width, data = iris, mu = 3)
}
\author{
Klaus K. Holst
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/partialcor.R
\name{partialcor}
\alias{partialcor}
\title{Calculate partial correlations}
\usage{
partialcor(formula, data, level = 0.95, ...)
}
\arguments{
\item{formula}{formula speciying the covariates and optionally the outcomes
to calculate partial correlation for}

\item{data}{data.frame}

\item{level}{Level of confidence limits}

\item{...}{Additional arguments to lower level functions}
}
\value{
A coefficient matrix
}
\description{
Calculate partial correlation coefficients and confidence limits via Fishers
z-transform
}
\examples{

m <- lvm(c(y1,y2,y3)~x1+x2)
covariance(m) <- c(y1,y2,y3)~y1+y2+y3
d <- sim(m,500)
partialcor(~x1+x2,d)

}
\author{
Klaus K. Holst
}
\keyword{models}
\keyword{regression}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bootstrap.R
\name{bootstrap.lvm}
\alias{bootstrap.lvm}
\alias{bootstrap.lvmfit}
\title{Calculate bootstrap estimates of a lvm object}
\usage{
\method{bootstrap}{lvm}(x,R=100,data,fun=NULL,control=list(),
                          p, parametric=FALSE, bollenstine=FALSE,
                          constraints=TRUE,sd=FALSE, mc.cores,
                          future.args=list(future.seed=TRUE),
                          ...)

\method{bootstrap}{lvmfit}(x,R=100,data=model.frame(x),
                             control=list(start=coef(x)),
                             p=coef(x), parametric=FALSE, bollenstine=FALSE,
                             estimator=x$estimator,weights=Weights(x),...)
}
\arguments{
\item{x}{\code{lvm}-object.}

\item{R}{Number of bootstrap samples}

\item{data}{The data to resample from}

\item{fun}{Optional function of the (bootstrapped) model-fit defining the
statistic of interest}

\item{control}{Options to the optimization routine}

\item{p}{Parameter vector of the null model for the parametric bootstrap}

\item{parametric}{If TRUE a parametric bootstrap is calculated. If FALSE a
non-parametric (row-sampling) bootstrap is computed.}

\item{bollenstine}{Bollen-Stine transformation (non-parametric bootstrap) for bootstrap hypothesis testing.}

\item{constraints}{Logical indicating whether non-linear parameter
constraints should be included in the bootstrap procedure}

\item{sd}{Logical indicating whether standard error estimates should be
included in the bootstrap procedure}

\item{mc.cores}{Optional number of cores for parallel computing. If omitted future.apply will be used (see future::plan)}

\item{future.args}{arguments to future.apply::future_lapply}

\item{\dots}{Additional arguments, e.g. choice of estimator.}

\item{estimator}{String definining estimator, e.g. 'gaussian' (see
\code{estimator})}

\item{weights}{Optional weights matrix used by \code{estimator}}
}
\value{
A \code{bootstrap.lvm} object.
}
\description{
Draws non-parametric bootstrap samples
}
\examples{
m <- lvm(y~x)
d <- sim(m,100)
e <- estimate(lvm(y~x), data=d)
\donttest{ ## Reduce Ex.Timings
B <- bootstrap(e,R=50,mc.cores=1)
B
}
}
\seealso{
\code{\link{confint.lvmfit}}
}
\author{
Klaus K. Holst
}
\keyword{models}
\keyword{regression}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/NA2x.R
\name{NA2x}
\alias{NA2x}
\alias{x2NA}
\title{Convert to/from NA}
\usage{
NA2x(s, x = 0)
}
\arguments{
\item{s}{The input vector (of arbitrary class)}

\item{x}{The elements to transform into \code{NA} resp. what to transform
\code{NA} into.}
}
\value{
A vector with same dimension and class as \code{s}.
}
\description{
Convert vector to/from NA
}
\examples{
##' 
x2NA(1:10, 1:5)
NA2x(x2NA(c(1:10),5),5)##'
}
\author{
Klaus K. Holst
}
\keyword{manip}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/index-method.R
\name{index}
\alias{index}
\alias{index<-}
\title{Generic method for extract index of an object}
\usage{
index(x, ...)

index(x, ...) <- value
}
\arguments{
\item{x}{object on which to set the index}

\item{...}{further arguments to be passed to or from other methods.}

\item{value}{the new index}
}
\description{
Generic method for extract index of an object

Generic method for setting the index of an object
}
\seealso{
\link{index<-}

\link{index}
}
\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zib.R
\name{zibreg}
\alias{zibreg}
\title{Regression model for binomial data with unkown group of immortals}
\usage{
zibreg(
  formula,
  formula.p = ~1,
  data,
  family = stats::binomial(),
  offset = NULL,
  start,
  var = "hessian",
  ...
)
}
\arguments{
\item{formula}{Formula specifying}

\item{formula.p}{Formula for model of disease prevalence}

\item{data}{data frame}

\item{family}{Distribution family (see the help page \code{family})}

\item{offset}{Optional offset}

\item{start}{Optional starting values}

\item{var}{Type of variance (robust, expected, hessian, outer)}

\item{...}{Additional arguments to lower level functions}
}
\description{
Regression model for binomial data with unkown group of immortals (zero-inflated binomial regression)
}
\examples{

## Simulation
n <- 2e3
x <- runif(n,0,20)
age <- runif(n,10,30)
z0 <- rnorm(n,mean=-1+0.05*age)
z <- cut(z0,breaks=c(-Inf,-1,0,1,Inf))
p0 <- lava:::expit(model.matrix(~z+age) \%*\% c(-.4, -.4, 0.2, 2, -0.05))
y <- (runif(n)<lava:::tigol(-1+0.25*x-0*age))*1
u <- runif(n)<p0
y[u==0] <- 0
d <- data.frame(y=y,x=x,u=u*1,z=z,age=age)
head(d)

## Estimation
e0 <- zibreg(y~x*z,~1+z+age,data=d)
e <- zibreg(y~x,~1+z+age,data=d)
compare(e,e0)
e
PD(e0,intercept=c(1,3),slope=c(2,6))

B <- rbind(c(1,0,0,0,20),
           c(1,1,0,0,20),
           c(1,0,1,0,20),
           c(1,0,0,1,20))
prev <- summary(e,pr.contrast=B)$prevalence

x <- seq(0,100,length.out=100)
newdata <- expand.grid(x=x,age=20,z=levels(d$z))
fit <- predict(e,newdata=newdata)
plot(0,0,type="n",xlim=c(0,101),ylim=c(0,1),xlab="x",ylab="Probability(Event)")
count <- 0
for (i in levels(newdata$z)) {
  count <- count+1
  lines(x,fit[which(newdata$z==i)],col="darkblue",lty=count)
}
abline(h=prev[3:4,1],lty=3:4,col="gray")
abline(h=prev[3:4,2],lty=3:4,col="lightgray")
abline(h=prev[3:4,3],lty=3:4,col="lightgray")
legend("topleft",levels(d$z),col="darkblue",lty=seq_len(length(levels(d$z))))
}
\author{
Klaus K. Holst
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model.R
\name{Model}
\alias{Model}
\alias{Model<-}
\title{Extract model}
\usage{
Model(x, ...)

Model(x, ...) <- value
}
\arguments{
\item{x}{Fitted model}

\item{\dots}{Additional arguments to be passed to the low level functions}

\item{value}{New model object (e.g. \code{lvm} or \code{multigroup})}
}
\value{
Returns a model object (e.g. \code{lvm} or \code{multigroup})
}
\description{
Extract or replace model object
}
\examples{

m <- lvm(y~x)
e <- estimate(m, sim(m,100))
Model(e)

}
\seealso{
\code{\link{Graph}}
}
\author{
Klaus K. Holst
}
\keyword{models}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/binomial.rrw.R
\name{binomial.rd}
\alias{binomial.rd}
\alias{binomial.rr}
\title{Define constant risk difference or relative risk association for binary exposure}
\usage{
binomial.rd(
  x,
  response,
  exposure,
  target.model,
  nuisance.model,
  exposure.model = dist_bernoulli(),
  ...
)
}
\arguments{
\item{x}{model}

\item{response}{response variable (character or formula)}

\item{exposure}{exposure variable (character or formula)}

\item{target.model}{variable defining the linear predictor for the target model}

\item{nuisance.model}{variable defining the linear predictor for the nuisance model}

\item{exposure.model}{model for exposure (default binomial logit link)}

\item{...}{additional arguments to lower level functions}
}
\description{
Set up model as defined in Richardson, Robins and Wang (2017).
}
\examples{
## ---------------------------------------------------------------
## binomial.rd: constant risk-difference model
##   P(Y=1|Z=1) - P(Y=1|Z=0) = tanh(lp)
## ---------------------------------------------------------------
m <- lvm()
regression(m) <- z ~ x
regression(m) <- lp ~ x
regression(m) <- op ~ x
intercept(m, ~lp) <- 0.4   ## constant linear predictor for RD
intercept(m, ~op) <- 0     ## odds product = exp(0) = 1
distribution(m, ~lp) <- dist_gaussian(sd = 0)
distribution(m, ~op) <- dist_gaussian(sd = 0)
m <- binomial.rd(m, response = "y", exposure = "z",
                 target.model = "lp", nuisance.model = "op")
set.seed(1)
d <- sim(m, n = 2000)
## Empirical risk difference should be close to tanh(0.4)
mean(d$y[d$z == 1]) - mean(d$y[d$z == 0])
tanh(0.4)

## Formula interface: response ~ exposure | target | nuisance
m2 <- lvm()
regression(m2) <- z ~ x
regression(m2) <- lp ~ x
regression(m2) <- op ~ x
m2 <- binomial.rd(m2, y ~ z | lp | op)

## ---------------------------------------------------------------
## binomial.rr: constant relative-risk model
##   log(P(Y=1|Z=1) / P(Y=1|Z=0)) = lp
## ---------------------------------------------------------------
m <- lvm()
regression(m) <- z ~ x
regression(m) <- lp ~ x
regression(m) <- op ~ x
intercept(m, ~lp) <- log(1.5)   ## constant log relative-risk
intercept(m, ~op) <- 0          ## odds product = 1
distribution(m, ~lp) <- dist_gaussian(sd = 0)
distribution(m, ~op) <- dist_gaussian(sd = 0)
m <- binomial.rr(m, response = "y", exposure = "z",
                 target.model = "lp", nuisance.model = "op")
set.seed(1)
d <- sim(m, n = 2000)
## Empirical log-RR should be close to log(1.5)
log(mean(d$y[d$z == 1]) / mean(d$y[d$z == 0]))
log(1.5)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/labels.R
\name{labels<-}
\alias{labels<-}
\alias{labels}
\alias{labels<-.default}
\alias{labels.lvm}
\alias{labels.lvmfit}
\alias{labels.graphNEL}
\alias{edgelabels}
\alias{edgelabels<-}
\alias{edgelabels<-.lvm}
\alias{nodecolor}
\alias{nodecolor<-}
\alias{nodecolor<-.default}
\title{Define labels of graph}
\usage{
\method{labels}{default}(object, ...) <- value
\method{edgelabels}{lvm}(object, to, ...) <- value
\method{nodecolor}{default}(object, var=vars(object),
border, labcol, shape, lwd, ...) <- value
}
\arguments{
\item{object}{\code{lvm}-object.}

\item{\dots}{Additional arguments (\code{lwd}, \code{cex}, \code{col},
\code{labcol}), \code{border}.}

\item{value}{node label/edge label/color}

\item{to}{Formula specifying outcomes and predictors defining relevant
edges.}

\item{var}{Formula or character vector specifying the nodes/variables to
alter.}

\item{border}{Colors of borders}

\item{labcol}{Text label colors}

\item{shape}{Shape of node}

\item{lwd}{Line width of border}
}
\description{
Alters labels of nodes and edges in the graph of a latent variable model
}
\examples{
m <- lvm(c(y,v)~x+z)
regression(m) <- c(v,x)~z
labels(m) <- c(y=expression(psi), z=expression(zeta))
nodecolor(m,~y+z+x,border=c("white","white","black"),
          labcol="white", lwd=c(1,1,5),
          lty=c(1,2)) <-  c("orange","indianred","lightgreen")
edgelabels(m,y~z+x, cex=c(2,1.5), col=c("orange","black"),labcol="darkblue",
           arrowhead=c("tee","dot"),
           lwd=c(3,1)) <- expression(phi,rho)
edgelabels(m,c(v,x)~z, labcol="red", cex=0.8,arrowhead="none") <- 2
if (interactive()) {
    plot(m,addstyle=FALSE)
}

m <- lvm(y~x)
labels(m) <- list(x="multiple\nlines")
if (interactive()) {
op <- par(mfrow=c(1,2))
plot(m,plain=TRUE)
plot(m)
par(op)

d <- sim(m,100)
e <- estimate(m,d)
plot(e,type="sd")
}
}
\author{
Klaus K. Holst
}
\keyword{aplot}
\keyword{graphs}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Missing.R
\name{Missing}
\alias{Missing}
\alias{Missing,}
\alias{Missing<-}
\title{Missing value generator}
\usage{
Missing(object, formula, Rformula, missing.name, suffix = "0", ...)
}
\arguments{
\item{object}{\code{lvm}-object.}

\item{formula}{The right hand side specifies the name of a latent
variable which is not always observed. The left hand side
specifies the name of a new variable which is equal to the latent
variable but has missing values.  If given as a string then this
is used as the name of the latent (full-data) name, and the
observed data name is 'missing.data'}

\item{Rformula}{Missing data mechanism with left hand side
specifying the name of the observed data indicator (may also just
be given as a character instead of a formula)}

\item{missing.name}{Name of observed data variable (only used if
'formula' was given as a character specifying the name of the
full-data variable)}

\item{suffix}{If missing.name is missing, then the name of the
oberved data variable will be the name of the full-data variable +
the suffix}

\item{...}{Passed to binomial.lvm.}
}
\value{
lvm object
}
\description{
Missing value generator
}
\details{
This function adds a binary variable to a given \code{lvm} model
and also a variable which is equal to the original variable where
the binary variable is equal to zero
}
\examples{
library(lava)
set.seed(17)
m <- lvm(y0~x01+x02+x03)
m <- Missing(m,formula=x1~x01,Rformula=R1~0.3*x02+-0.7*x01,p=0.4)
sim(m,10)


m <- lvm(y~1)
m <- Missing(m,"y","r")
## same as
## m <- Missing(m,y~1,r~1)
sim(m,10)

## same as
m <- lvm(y~1)
Missing(m,"y") <- r~x
sim(m,10)

m <- lvm(y~1)
m <- Missing(m,"y","r",suffix=".")
## same as
## m <- Missing(m,"y","r",missing.name="y.")
## same as
## m <- Missing(m,y.~y,"r")
sim(m,10)

}
\author{
Thomas A. Gerds \href{mailto:tag@biostat.ku.dk}{tag@biostat.ku.dk}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compare.R
\name{compare}
\alias{compare}
\alias{test_wald}
\title{Statistical tests}
\usage{
compare(object, ...)
}
\arguments{
\item{object}{\code{lvmfit}-object}

\item{\dots}{Additional arguments to low-level functions}
}
\value{
Matrix of test-statistics and p-values
}
\description{
Performs Likelihood-ratio, Wald and score tests
}
\examples{
m <- lvm();
regression(m) <- c(y1,y2,y3) ~ eta; latent(m) <- ~eta
regression(m) <- eta ~ x
m2 <- regression(m, c(y3,eta) ~ x)
set.seed(1)
d <- sim(m,1000)
e <- estimate(m,d)
e2 <- estimate(m2,d)

compare(e)

compare(e,e2) ## LRT, H0: y3<-x=0
compare(e,scoretest=y3~x) ## Score-test, H0: y3~x=0
compare(e2,par=c("y3~x")) ## Wald-test, H0: y3~x=0

B <- diag(2); colnames(B) <- c("y2~eta","y3~eta")
compare(e2,contrast=B,null=c(1,1))

B <- rep(0,length(coef(e2))); B[1:3] <- 1
compare(e2,contrast=B)

compare(e,scoretest=list(y3~x,y2~x))
}
\seealso{
\code{\link{modelsearch}}, \code{\link{equivalence}}
}
\author{
Klaus K. Holst
}
\keyword{htest}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Col.R
\name{Col}
\alias{Col}
\title{Generate a transparent RGB color}
\usage{
Col(col, alpha = 0.2, locate = 0)
}
\arguments{
\item{col}{color (numeric or character)}

\item{alpha}{degree of transparency (0-1)}

\item{locate}{Choose colour (with mouse)}
}
\value{
A character vector with elements of 7 or 9 characters, \verb{#} followed
by the red, blue, green and optionally alpha values in hexadecimal (after
rescaling to '0 ... 255').
}
\description{
This function transforms a standard color (e.g. "red") into an
transparent RGB-color (i.e. alpha-blend<1).
}
\details{
This only works for certain graphics devices (Cairo-X11 (x11 as of R>=2.7),
quartz, pdf, ...).
}
\examples{
plot(runif(1000),cex=runif(1000,0,4),
     col=Col(c("darkblue","orange"),0.5),pch=16)
}
\author{
Klaus K. Holst
}
\keyword{color}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/binomial.rrw.R
\name{binomial.rd}
\alias{binomial.rd}
\alias{binomial.rr}
\title{Define constant risk difference or relative risk association for binary exposure}
\usage{
binomial.rd(
  x,
  response,
  exposure,
  target.model,
  nuisance.model,
  exposure.model = binomial.lvm(),
  ...
)
}
\arguments{
\item{x}{model}

\item{response}{response variable (character or formula)}

\item{exposure}{exposure variable (character or formula)}

\item{target.model}{variable defining the linear predictor for the target model}

\item{nuisance.model}{variable defining the linear predictor for the nuisance model}

\item{exposure.model}{model for exposure (default binomial logit link)}

\item{...}{additional arguments to lower level functions}
}
\description{
Set up model as defined in Richardson, Robins and Wang (2017).
}
\examples{
## ---------------------------------------------------------------
## binomial.rd: constant risk-difference model
##   P(Y=1|Z=1) - P(Y=1|Z=0) = tanh(lp)
## ---------------------------------------------------------------
m <- lvm()
regression(m) <- z ~ x
regression(m) <- lp ~ x
regression(m) <- op ~ x
intercept(m, ~lp) <- 0.4   ## constant linear predictor for RD
intercept(m, ~op) <- 0     ## odds product = exp(0) = 1
distribution(m, ~lp) <- normal.lvm(sd = 0)
distribution(m, ~op) <- normal.lvm(sd = 0)
m <- binomial.rd(m, response = "y", exposure = "z",
                 target.model = "lp", nuisance.model = "op")
set.seed(1)
d <- sim(m, n = 2000)
## Empirical risk difference should be close to tanh(0.4)
mean(d$y[d$z == 1]) - mean(d$y[d$z == 0])
tanh(0.4)

## Formula interface: response ~ exposure | target | nuisance
m2 <- lvm()
regression(m2) <- z ~ x
regression(m2) <- lp ~ x
regression(m2) <- op ~ x
m2 <- binomial.rd(m2, y ~ z | lp | op)

## ---------------------------------------------------------------
## binomial.rr: constant relative-risk model
##   log(P(Y=1|Z=1) / P(Y=1|Z=0)) = lp
## ---------------------------------------------------------------
m <- lvm()
regression(m) <- z ~ x
regression(m) <- lp ~ x
regression(m) <- op ~ x
intercept(m, ~lp) <- log(1.5)   ## constant log relative-risk
intercept(m, ~op) <- 0          ## odds product = 1
distribution(m, ~lp) <- normal.lvm(sd = 0)
distribution(m, ~op) <- normal.lvm(sd = 0)
m <- binomial.rr(m, response = "y", exposure = "z",
                 target.model = "lp", nuisance.model = "op")
set.seed(1)
d <- sim(m, n = 2000)
## Empirical log-RR should be close to log(1.5)
log(mean(d$y[d$z == 1]) / mean(d$y[d$z == 0]))
log(1.5)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lvm.R
\name{lvm}
\alias{lvm}
\alias{print.lvm}
\alias{summary.lvm}
\title{Initialize new latent variable model}
\usage{
lvm(x = NULL, ..., latent = NULL, messages = lava.options()$messages)
}
\arguments{
\item{x}{Vector of variable names. Optional but gives control of the
sequence of appearance of the variables. The argument can be given as a
character vector or formula, e.g. \code{~y1+y2} is equivalent to
\code{c("y1","y2")}. Alternatively the argument can be a formula specifying
a linear model.}

\item{\dots}{Additional arguments to be passed to the low level functions}

\item{latent}{(optional) Latent variables}

\item{messages}{Controls what messages are printed (0: none)}
}
\value{
Returns an object of class \code{lvm}.
}
\description{
Function that constructs a new latent variable model object
}
\examples{

m <- lvm() # Empty model
m1 <- lvm(y~x) # Simple linear regression
m2 <- lvm(~y1+y2) # Model with two independent variables (argument)
m3 <- lvm(list(c(y1,y2,y3)~u,u~x+z)) # SEM with three items

}
\seealso{
\code{\link{regression}}, \code{\link{covariance}},
\code{\link{intercept}}, ...
}
\author{
Klaus K. Holst
}
\keyword{models}
\keyword{regression}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cv.R
\name{cv}
\alias{cv}
\title{Cross-validation}
\usage{
cv(modelList, data, K = 5, rep = 1, perf, seed = NULL, shared = NULL, ...)
}
\arguments{
\item{modelList}{List of fitting functions or models}

\item{data}{data.frame}

\item{K}{Number of folds (default 5, 0 splits in 1:n/2, n/2:n with last part
used for testing)}

\item{rep}{Number of repetitions (default 1)}

\item{perf}{Performance measure (default RMSE)}

\item{seed}{Optional random seed}

\item{shared}{function applied to each fold with results send to each model}

\item{...}{Additional arguments parsed to models in modelList and perf}
}
\description{
Cross-validation
}
\details{
Generic cross-validation function
}
\examples{
f0 <- function(data,...) lm(...,data)
f1 <- function(data,...) lm(Sepal.Length~Species,data)
f2 <- function(data,...) lm(Sepal.Length~Species+Petal.Length,data)
x <- cv(list(m0=f0,m1=f1,m2=f2),rep=10, data=iris, formula=Sepal.Length~.)
x2 <- cv(list(f0(iris),f1(iris),f2(iris)),rep=10, data=iris)
}
\author{
Klaus K. Holst
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Missing.R
\name{Missing}
\alias{Missing}
\alias{Missing,}
\alias{Missing<-}
\title{Missing value generator}
\usage{
Missing(object, formula, Rformula, missing.name, suffix = "0", ...)
}
\arguments{
\item{object}{\code{lvm}-object.}

\item{formula}{The right hand side specifies the name of a latent
variable which is not always observed. The left hand side
specifies the name of a new variable which is equal to the latent
variable but has missing values.  If given as a string then this
is used as the name of the latent (full-data) name, and the
observed data name is 'missing.data'}

\item{Rformula}{Missing data mechanism with left hand side
specifying the name of the observed data indicator (may also just
be given as a character instead of a formula)}

\item{missing.name}{Name of observed data variable (only used if
'formula' was given as a character specifying the name of the
full-data variable)}

\item{suffix}{If missing.name is missing, then the name of the
oberved data variable will be the name of the full-data variable +
the suffix}

\item{...}{Passed to \link{dist_bernoulli}}
}
\value{
lvm object
}
\description{
Missing value generator
}
\details{
This function adds a binary variable to a given \code{lvm} model
and also a variable which is equal to the original variable where
the binary variable is equal to zero
}
\examples{
library(lava)
set.seed(17)
m <- lvm(y0~x01+x02+x03)
m <- Missing(m,formula=x1~x01,Rformula=R1~0.3*x02+-0.7*x01,p=0.4)
sim(m,10)


m <- lvm(y~1)
m <- Missing(m,"y","r")
## same as
## m <- Missing(m,y~1,r~1)
sim(m,10)

## same as
m <- lvm(y~1)
Missing(m,"y") <- r~x
sim(m,10)

m <- lvm(y~1)
m <- Missing(m,"y","r",suffix=".")
## same as
## m <- Missing(m,"y","r",missing.name="y.")
## same as
## m <- Missing(m,y.~y,"r")
sim(m,10)

}
\author{
Thomas A. Gerds \href{mailto:tag@biostat.ku.dk}{tag@biostat.ku.dk}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/estimate.default.R
\name{estimate.array}
\alias{estimate.array}
\alias{estimate.data.frame}
\title{Estimate parameters and influence function.}
\usage{
\method{estimate}{array}(x, type = "mean", probs = 0.5, ...)
}
\arguments{
\item{x}{numeric matrix}

\item{type}{target parameter ("mean", "variance", "quantile")}

\item{probs}{numeric vector of probabilities (for type="quantile")}

\item{...}{Additional arguments to lower level functions (i.e.,
stats::density.default when type="quantile")}
}
\description{
Estimate parameters for the sample mean, variance, and quantiles
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/estimate.array.R
\name{estimate.array}
\alias{estimate.array}
\alias{estimate.data.frame}
\title{Estimate parameters and influence function.}
\usage{
\method{estimate}{array}(x, type = "mean", probs = 0.5, ...)
}
\arguments{
\item{x}{numeric matrix}

\item{type}{target parameter ("mean", "variance", "quantile")}

\item{probs}{numeric vector of probabilities (for type="quantile")}

\item{...}{Additional arguments to lower level functions (i.e.,
stats::density.default when type="quantile")}
}
\description{
Estimate parameters for the sample mean, variance, and quantiles
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lava-package.R
\docType{data}
\name{calcium}
\alias{calcium}
\title{Longitudinal Bone Mineral Density Data}
\format{
A data.frame containing 560 (incomplete) observations. The 'person'
column defines the individual girls of the study with measurements at
visiting times 'visit', and age in years 'age' at the time of visit. The
bone mineral density variable is 'bmd' (g/cm^2).
}
\source{
Vonesh & Chinchilli (1997), Table 5.4.1 on page 228.
}
\description{
Bone Mineral Density Data consisting of 112 girls randomized to receive
calcium og placebo. Longitudinal measurements of bone mineral density
(g/cm^2) measured approximately every 6th month in 3 years.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/curly.R
\name{curly}
\alias{curly}
\title{Adds curly brackets to plot}
\usage{
curly(
  x,
  y,
  len = 1,
  theta = 0,
  wid,
  shape = 1,
  col = 1,
  lwd = 1,
  lty = 1,
  grid = FALSE,
  npoints = 50,
  text = NULL,
  offset = c(0.05, 0)
)
}
\arguments{
\item{x}{center of the x axis of the curly brackets (or start end coordinates (x1,x2))}

\item{y}{center of the y axis of the curly brackets (or start end coordinates (y1,y2))}

\item{len}{Length of the curly brackets}

\item{theta}{angle (in radians) of the curly brackets orientation}

\item{wid}{Width of the curly brackets}

\item{shape}{shape (curvature)}

\item{col}{color (passed to lines/grid.lines)}

\item{lwd}{line width (passed to lines/grid.lines)}

\item{lty}{line type (passed to lines/grid.lines)}

\item{grid}{If TRUE use grid graphics (compatability with ggplot2)}

\item{npoints}{Number of points used in curves}

\item{text}{Label}

\item{offset}{Label offset (x,y)}
}
\description{
Adds curly brackets to plot
}
\examples{
if (interactive()) {
plot(0,0,type="n",axes=FALSE,xlab="",ylab="")
curly(x=c(1,0),y=c(0,1),lwd=2,text="a")
curly(x=c(1,0),y=c(0,1),lwd=2,text="b",theta=pi)
curly(x=-0.5,y=0,shape=1,theta=pi,text="c")
curly(x=0,y=0,shape=1,theta=0,text="d")
curly(x=0.5,y=0,len=0.2,theta=pi/2,col="blue",lty=2)
curly(x=0.5,y=-0.5,len=0.2,theta=-pi/2,col="red",shape=1e3,text="e")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lava-package.R
\docType{data}
\name{missingdata}
\alias{missingdata}
\title{Missing data example}
\format{
list of data.frames
}
\source{
Simulated
}
\description{
Simulated data generated from model
\deqn{E(Y_i\mid X) = X, \quad cov(Y_1,Y_2\mid X)=0.5}
}
\details{
The list contains four data sets
\enumerate{
\item Complete data
\item MCAR
\item MAR
\item MNAR (missing mechanism depends on variable V correlated with Y1,Y2)
}
}
\examples{
data(missingdata)
e0 <- estimate(lvm(c(y1,y2)~b*x,y1~~y2),missingdata[[1]]) ## No missing
e1 <- estimate(lvm(c(y1,y2)~b*x,y1~~y2),missingdata[[2]]) ## CC (MCAR)
e2 <- estimate(lvm(c(y1,y2)~b*x,y1~~y2),missingdata[[2]],missing=TRUE) ## MCAR
e3 <- estimate(lvm(c(y1,y2)~b*x,y1~~y2),missingdata[[3]]) ## CC (MAR)
e4 <- estimate(lvm(c(y1,y2)~b*x,y1~~y2),missingdata[[3]],missing=TRUE) ## MAR
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lava-package.R
\docType{data}
\name{deprdiag}
\alias{deprdiag}
\title{50 patients from Monash Medical Centre, Melbourne}
\format{
data.frame
}
\source{
Clarke, D. M., Smith, G. C., & Herrman, H. E. (1993). A comparative
study of screening instruments for mental disorders in general hospital
patients. International Journal Psychiatry in Medicine, 23, pp. 323-337.

McKenzie et al. (1996). Comparing correlated Kappas by resampling: Is one
level of agreement significantly different from another? J. Psychiat. Res.
30 (6), pp. 483-492.
}
\description{
Diagnosis of depression (DSM-III-R MDD, Dysthymia, Adjustment Disorder with
Depressed Mood, Depression NOS), Beck Depression Inventory (BDI) (Beck et
al., 1961) General Health Questionnaire (GHQ) (Goldberg & Williams, 1988)
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lava-package.R
\name{startvalues}
\alias{startvalues}
\alias{startvalues0}
\alias{startvalues1}
\alias{startvalues2}
\alias{starter.multigroup}
\alias{modelPar}
\alias{modelVar}
\alias{matrices}
\alias{pars}
\alias{pars.lvm}
\alias{regfix}
\alias{pars.lvmfit}
\alias{pars.glm}
\alias{score.glm}
\alias{procdata.lvmfit}
\alias{mat.lvm}
\alias{reorderdata}
\alias{graph2lvm}
\alias{igraph.lvm}
\alias{subgraph}
\alias{finalize}
\alias{randomslope}
\alias{randomslope<-}
\alias{lisrel}
\alias{variances}
\alias{offdiags}
\alias{describecoef}
\alias{parlabels}
\alias{rsq}
\alias{stdcoef}
\alias{CoefMat}
\alias{CoefMat.multigroupfit}
\alias{deriv}
\alias{updatelvm}
\alias{checkmultigroup}
\alias{profci}
\alias{estimate.MAR}
\alias{missingModel}
\alias{Identical}
\alias{gaussian_logLik.lvm}
\alias{addhook}
\alias{gethook}
\alias{multigroup}
\alias{Weights}
\alias{fixsome}
\alias{IV}
\alias{parameter}
\alias{Specials}
\alias{procformula}
\alias{getoutcome}
\alias{decomp.specials}
\alias{rmvn0}
\alias{dmvn0}
\alias{logit}
\alias{expit}
\alias{tigol}
\title{For internal use}
\description{
For internal use
}
\author{
Klaus K. Holst
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summary.estimate.R
\name{summary.estimate}
\alias{summary.estimate}
\title{Summary of estimate objects}
\usage{
\method{summary}{estimate}(
  object,
  contrast,
  ...,
  null = 0,
  level = 0.95,
  type,
  var.adj = 0.25,
  df,
  transform = NULL,
  print = NULL
)
}
\arguments{
\item{object}{an \code{estimate} object.}

\item{contrast}{(optional) contrast matrix for the final Wald test.}

\item{...}{additional arguments passed to \link{contr}.}

\item{null}{(optional) null hypothesis to test.}

\item{level}{level of confidence limits (default 0.95)}

\item{type}{type of small-sample correction. Requires the estimate to have
been computed with \code{IC=TRUE} (the default).}

\item{var.adj}{variance adjustment parameter for small-sample correction.
Requires the estimate to have been computed with \code{IC=TRUE} (the default).}

\item{df}{degrees of freedom for t-based inference (default: \code{NULL} for
Gaussian approximation; when set, confidence intervals and p-values use the
t-distribution with \code{df} degrees of freedom)}

\item{transform}{(optional) function applied to the point estimates and
confidence interval bounds \emph{after} inference is performed on the original
scale. Useful for variance-stabilizing transformations, e.g., compute CIs
on the \code{atanh} (Fisher z) scale and back-transform with \code{tanh}.}

\item{print}{(optional) custom print function for the resulting
\code{summary.estimate} object}
}
\description{
Computes hypothesis tests, contrasts, and small-sample corrections for
an \link{estimate} object. The arguments \code{null}, \code{contrast}, \code{type}, and
\code{var.adj} were previously available on \code{\link[=estimate.default]{estimate.default()}} and have
been moved here.
}
\seealso{
\code{\link[=estimate.default]{estimate.default()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/interactive.R
\name{click}
\alias{click}
\alias{idplot}
\alias{click.default}
\alias{colsel}
\title{Identify points on plot}
\usage{
\method{click}{default}(x, y=NULL, label=TRUE, n=length(x), pch=19, col="orange", cex=3, ...)
idplot(x, y ,..., id=list(), return.data=FALSE)
}
\arguments{
\item{x}{X coordinates}

\item{\dots}{Additional arguments parsed to \code{plot} function}

\item{y}{Y coordinates}

\item{label}{Should labels be added?}

\item{n}{Max number of inputs to expect}

\item{pch}{Symbol}

\item{col}{Colour}

\item{cex}{Size}

\item{id}{List of arguments parsed to \code{click} function}

\item{return.data}{Boolean indicating if selected points should be returned}
}
\description{
Extension of the \code{identify} function
}
\details{
For the usual 'X11' device the identification process is
terminated by pressing any mouse button other than the first. For
the 'quartz' device the process is terminated by pressing either
the pop-up menu equivalent (usually second mouse button or
'Ctrl'-click) or the 'ESC' key.
}
\examples{
if (interactive()) {
    n <- 10; x <- seq(n); y <- runif(n)
    plot(y ~ x); click(x,y)

    data(iris)
    l <- lm(Sepal.Length ~ Sepal.Width*Species,iris)
    res <- plotConf(l,var2="Species")## ylim=c(6,8), xlim=c(2.5,3.3))
    with(res, click(x,y))

    with(iris, idplot(Sepal.Length,Petal.Length))
}
}
\seealso{
\code{\link{idplot}}, \code{identify}
}
\author{
Klaus K. Holst
}
\keyword{iplot}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rbind.Surv.R
\name{rbind.Surv}
\alias{rbind.Surv}
\title{Appending \code{Surv} objects}
\usage{
\method{rbind}{Surv}(...)
}
\arguments{
\item{...}{\code{Surv} objects}
}
\value{
\code{Surv} object
}
\description{
\code{rbind} method for \code{Surv} objects
}
\examples{

y <- yl <- yr <- rnorm(10)
yl[1:5] <- NA; yr[6:10] <- NA
S1 <- survival::Surv(yl,yr,type="interval2")
S2 <- survival::Surv(y,y>0,type="right")
S3 <- survival::Surv(y,y<0,type="left")

rbind(S1,S1)
rbind(S2,S2)
rbind(S3,S3)

}
\author{
Klaus K. Holst
}
\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ordinal.R
\name{ordinal<-}
\alias{ordinal<-}
\alias{ordinal}
\title{Define variables as ordinal}
\usage{
ordinal(x, ...) <- value
}
\arguments{
\item{x}{Object}

\item{...}{additional arguments to lower level functions}

\item{value}{variable (formula or character vector)}
}
\description{
Define variables as ordinal in latent variable model object
}
\examples{
if (requireNamespace("mets")) {
m <- lvm(y + z ~ x + 1*u[0], latent=~u)
ordinal(m, K=3) <- ~y+z
d <- sim(m, 100, seed=1)
e <- estimate(m, d)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ordreg.R
\name{ordreg}
\alias{ordreg}
\title{Univariate cumulative link regression models}
\usage{
ordreg(
  formula,
  data = parent.frame(),
  offset,
  family = stats::binomial("probit"),
  start,
  fast = FALSE,
  ...
)
}
\arguments{
\item{formula}{formula}

\item{data}{data.frame}

\item{offset}{offset}

\item{family}{family (default proportional odds)}

\item{start}{optional starting values}

\item{fast}{If TRUE standard errors etc. will not be calculated}

\item{...}{Additional arguments to lower level functions}
}
\description{
Ordinal regression models
}
\examples{
m <- lvm(y~x)
ordinal(m,K=3) <- ~y
d <- sim(m,100)
e <- ordreg(y~x,d)
}
\author{
Klaus K. Holst
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/index-method.R
\name{index.lvm}
\alias{index.lvm}
\alias{index.lvmfit}
\alias{index<-.lvm}
\alias{index<-.lvmfit}
\title{Extract the parameter indicies of a lvm object}
\usage{
\method{index}{lvm}(x, ...)

\method{index}{lvm}(x, ...) <- value
}
\arguments{
\item{x}{object on which to set the index}

\item{...}{further arguments to be passed to or from other methods.}

\item{value}{new index}
}
\description{
Extracts the matrices with indices of model parameters from a
latent variable model (\code{lvm}). Returns a list with
\itemize{
\item A:   Matrix with fixed parameters and ones where parameters are free
\item J:   Manifest variable selection matrix
\item M0:  Index of free regression parameters
\item M1:  Index of free and \emph{unique} regression parameters
\item P:   Matrix with fixed variance parameters and ones where parameters are free
\item P0:  Index of free variance parameters
\item P1:  Index of free and \emph{unique} regression parameters
\item npar.var:   Number of covariance parameters
}
}
\seealso{
\code{\link[=modelPar]{modelPar()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/estimate.default.R
\name{estimate.default}
\alias{estimate.default}
\alias{estimate}
\alias{merge.estimate}
\alias{estimate.mlm}
\title{Influence function based inference}
\usage{
\method{estimate}{default}(
  x = NULL,
  f = NULL,
  ...,
  data,
  id,
  stack = TRUE,
  average = FALSE,
  subset,
  level = 0.95,
  IC = TRUE,
  type = c("robust", "df", "mbn"),
  var.adj,
  keep,
  use,
  regex = FALSE,
  ignore.case = FALSE,
  contrast,
  null,
  vcov,
  coef,
  df = NULL,
  print = NULL,
  labels,
  label.width,
  only.coef = FALSE,
  back.transform = NULL
)
}
\arguments{
\item{x}{model object (\code{glm}, \code{lvmfit}, ...) or an existing \code{estimate}
object. When two model objects are supplied (e.g., \code{estimate(g, g0)}) a
likelihood-ratio test is performed.}

\item{f}{transformation of model parameters. Accepts several input types:
\itemize{
\item A \strong{function} \code{f(p)} or \code{f(p, data)}: applies the delta method.
When \code{f} returns a named list the names are used as parameter labels.
\item A \strong{matrix}: used as a contrast (linear combination) matrix.
\item A \strong{numeric vector} of parameter indices: converted to a contrast
that selects and differences those parameters.
\item A \strong{list} of indices: each element selects one parameter.
\item \strong{Character} expressions: supports wildcards (\code{"?"}, \code{"*"}) and
arithmetic on parameter names (e.g., \code{"z" - "x"}, \code{2 * "z" - 3 * "x"}).
}}

\item{...}{additional arguments to lower level functions}

\item{data}{\code{data.frame} used by \code{f} when the transformation depends on
covariates (see \code{average}). Defaults to \code{model.frame(x)}.}

\item{id}{(optional) cluster identifier. Can be a vector of cluster IDs, a
one-sided formula (evaluated in \code{data}), a single character column name,
or a logical scalar (\code{TRUE} for one-to-one matching, \code{FALSE} for
independence). When supplied, the IF is aggregated within clusters to
produce cluster-robust standard errors.}

\item{stack}{if \code{TRUE} (default) the influence function contributions are
summed within each cluster defined by \code{id}. Set to \code{FALSE} to keep the
un-stacked (per-observation) decomposition.}

\item{average}{if \code{TRUE} the function computes the standardized
(marginalized) estimate \eqn{\hat\Psi = P_n f(X; \hat\theta)}, i.e.,
the empirical mean of \code{f(p, data)} over all rows of \code{data}. The
influence function accounts for both the empirical averaging and the
parameter estimation uncertainty (see Details).}

\item{subset}{(optional) logical vector, expression evaluated in \code{data}, or
column name. When used together with \code{average = TRUE}, the average is
conditioned on the subpopulation where \code{subset} is \code{TRUE}, yielding a
conditional marginalized estimate.}

\item{level}{level of confidence limits (default 0.95)}

\item{IC}{if \code{TRUE} (default) the influence function matrix is estimated and
stored in the returned object (extract with the \link{IC} method). Can also be
a user-supplied IF matrix (one row per observation, one column per
parameter), which is used directly instead of estimating it from \code{x}.}

\item{type}{type of small-sample correction for cluster-robust variance.
One of:
\itemize{
\item \code{"robust"} (default): no correction.
\item \code{"df"}: applies \eqn{n/(n-p)} correction (Mancl & DeRouen, 2001).
\item \code{"mbn"}: Morel-Bokossa-Neerchal (2003) correction.
\item \code{"hc3"}: leverage-adjusted HC3-type correction (blended with
\code{var.adj}).
\item \code{"hc4"}: Cribari-Neto (2004) leverage-adjusted correction.
}}

\item{var.adj}{blending parameter for the HC3 leverage adjustment
(default 0.25). Controls the weight between observation-level empirical
leverage and the average leverage \eqn{p/n}.}

\item{keep}{(optional) index of parameters to keep from final result.
Accepts integer indices, character names, or (with \code{regex = TRUE})
perl-compatible regular expressions.}

\item{use}{(optional) index of parameters to use in calculations. The
selected parameters are first extracted (via \code{keep}) and then the
remaining arguments (\code{f}, \code{contrast}, etc.) are applied to this subset.}

\item{regex}{if \code{TRUE} use perl-compatible regular expressions for \code{keep}
and \code{use} arguments}

\item{ignore.case}{ignore case in regular expressions}

\item{contrast}{(optional) contrast matrix for a final Wald test. When
supplied together with \code{null}, tests \eqn{H_0: B\theta = b_0}.}

\item{null}{(optional) null hypothesis vector \eqn{b_0} to test against
(default 0)}

\item{vcov}{(optional) covariance matrix of parameter estimates, or a
logical. If \code{TRUE}, \link[stats:vcov]{stats::vcov} is used to obtain the (model-based)
covariance matrix from \code{x}, yielding non-robust standard errors. If a
matrix is supplied it is used directly. When omitted or \code{FALSE}, robust
standard errors are computed from the influence function.}

\item{coef}{(optional) named parameter vector. Used instead of
\code{coef(x)} when constructing an \code{estimate} object without a model.}

\item{df}{degrees of freedom for t-based inference (default: \code{NULL} for
Gaussian approximation; when set, confidence intervals and p-values use
the t-distribution with \code{df} degrees of freedom)}

\item{print}{(optional) custom print function for the resulting \code{estimate}
object}

\item{labels}{(optional) character vector of coefficient names}

\item{label.width}{(optional) max display width of labels}

\item{only.coef}{(Deprecated) if \code{TRUE} only the coefficient matrix is
returned. Use \code{parameter(estimate(...))} instead.}

\item{back.transform}{(optional) function applied to the point estimates
and confidence interval bounds \emph{after} inference is performed on the
original scale. Useful for variance-stabilizing transformations, e.g.,
compute CIs on the \code{atanh} (Fisher z) scale and back-transform with
\code{tanh}.}
}
\description{
Primary tool for obtaining parameter estimates with robust (sandwich)
standard errors, applying the delta method, and testing linear hypotheses.
The function returns an object of class \code{estimate} which serves as a
general container for parameter estimates and their influence functions
(IFs). Three calling conventions are supported:
}
\details{
\itemize{
\item \code{estimate(x, ...)} -- extract estimates from a model object
\item \code{estimate(coef=, IC=, ...)} -- construct from coefficients and IF matrix
\item \code{estimate(coef=, vcov=, ...)} -- construct from coefficients and
covariance matrix
}
}
\section{Influence functions and robust standard errors}{
An estimator \eqn{\widehat{\theta}} is \emph{regular and asymptotically
linear} (RAL) when it admits the iid decomposition
\deqn{\sqrt{n}(\widehat{\theta}-\theta) =
\frac{1}{\sqrt{n}}\sum_{i=1}^n \mathrm{IC}(Z_i; P) + o_p(1)}
where \eqn{\mathrm{IC}} is the unique \emph{influence function} satisfying
\eqn{E\{\mathrm{IC}(Z; P)\} = 0}. By the central limit theorem
\deqn{\sqrt{n}(\widehat{\theta}-\theta)
\overset{d}{\longrightarrow}
N(0,\; \mathrm{Var}\{\mathrm{IC}(Z; P)\})}
and the asymptotic variance is consistently estimated by the empirical
variance of the plugin IF estimate, yielding robust (sandwich) standard
errors. The estimated IF can be extracted with the \link{IC} method.
}

\section{Parameter transformations (delta method)}{
When \code{f} is a function \eqn{\phi: R^p \to R^m}, the delta method is
applied:
\deqn{\sqrt{n}\{\phi(\widehat{\theta}) - \phi(\theta)\} =
\frac{1}{\sqrt{n}}\sum_{i=1}^n
\nabla\phi(\theta)\,\mathrm{IC}(Z_i; P) + o_p(1)}
Derivatives are computed numerically via \link[numDeriv:jacobian]{numDeriv::jacobian} unless the
function returns an attribute \code{"grad"} with the analytic Jacobian.

Alternatively, \code{estimate} objects support direct arithmetic operations
(e.g., \code{a * b}, \code{exp(a)}, \code{a^b}) which apply the delta method with
\emph{exact} (analytical) derivatives computed automatically. This influence
function calculus allows building complex transformations from simple
building blocks without numerical differentiation. See the last example
section ("influence function calculus") and
\code{vignette("influencefunction", package = "lava")} for details.
}

\section{Averaging and marginalization}{
When \code{average = TRUE} and \code{f(p, data)} depends on covariates, the
target parameter is the standardized (marginalized) estimate
\eqn{\Psi = E\{f(X;\theta)\}}. The IF for the averaged estimate
accounts for both the empirical averaging and parameter estimation
uncertainty:
\deqn{\mathrm{IC}_\Psi(Z; P) = f(X;\theta) - \Psi +
[E\nabla_\theta f(X;\theta)]\,\phi(Z; P)}
When \code{subset} is also specified, the average is conditioned on the
subpopulation, yielding a conditional marginalized estimate.
}

\section{Cluster-robust standard errors}{
When \code{id} is supplied, the per-observation IF contributions are summed
within clusters (when \code{stack = TRUE}), producing the cluster-level IF
\eqn{\widetilde{\mathrm{IC}}(Z_i; P) = \sum_{k=1}^{N_i}
\frac{n}{N}\mathrm{IC}(Z_{ik}; P)}.
The resulting variance estimate is equivalent to the GEE working
independence sandwich estimator.

For full theoretical background and worked examples see
\code{vignette("influencefunction", package = "lava")}.
}

\examples{

## Simulation from logistic regression model
m <- lvm(y~x+z);
distribution(m,y~x) <- binomial.lvm("logit")
d <- sim(m,1000)
g <- glm(y~z+x,data=d,family=binomial())
g0 <- glm(y~1,data=d,family=binomial())

## LRT
estimate(g, g0)

## Plain estimates (robust standard errors)
estimate(g)

## Testing contrasts
estimate(g, null=0)
estimate(g, rbind(c(1,1,0), c(1,0,2)))
estimate(g, rbind(c(1,1,0), c(1,0,2)), null=c(1,2))
estimate(g, 2:3) ## same as cbind(0,1,-1)
estimate(g, as.list(2:3)) ## same as rbind(c(0,1,0),c(0,0,1))
## Alternative syntax
estimate(g, "z", "z"-"x", 2*"z"-3*"x")
estimate(g, "?")  ## Wildcards
estimate(g, "*Int*", "z")
estimate(g, "1", "2"-"3", null = c(0,1))
estimate(g, 2, 3)

## Usual (non-robust) confidence intervals
estimate(g, vcov=TRUE)
estimate(g, vcov=vcov(g))

## Transformations
estimate(g, function(p) p[1]+p[2])

## Multiple parameters
e <- estimate(g, function(p) c(p[1]+p[2], p[1]*p[2]))
e
vcov(e)

## Label new parameters
estimate(g, function(p) list("a1"=p[1]+p[2], "b1"=p[1]*p[2]))
#'
## Multiple group
m <- lvm(y~x)
m <- baptize(m)
d2 <- d1 <- sim(m,50,seed=1)
e <- estimate(list(m,m),list(d1,d2))
estimate(e) ## Wrong
ee <- estimate(e, id=rep(seq(nrow(d1)), 2)) ## Clustered
ee
estimate(lm(y~x,d1))

## Marginalize
f <- function(p,data)
  list(p0=expit(p["(Intercept)"] + p["z"]*data[,"z"]),
       p1=expit(p["(Intercept)"] + p["x"] + p["z"]*data[,"z"]))
e <- estimate(g, f, average=TRUE)
e
estimate(e,diff)
estimate(e,cbind(1,1))

## Clusters and subset (conditional marginal effects)
d$id <- rep(seq(nrow(d)/4),each=4)
estimate(g,function(p,data)
         list(p0=expit(p[1] + p["z"]*data[,"z"])),
         subset=d$z>0, id=d$id, average=TRUE)

## More examples with clusters:
m <- lvm(c(y1,y2,y3)~u+x)
d <- sim(m,10)
l1 <- glm(y1~x,data=d)
l2 <- glm(y2~x,data=d)
l3 <- glm(y3~x,data=d)

## Some random id-numbers
id1 <- c(1,1,4,1,3,1,2,3,4,5)
id2 <- c(1,2,3,4,5,6,7,8,1,1)
id3 <- seq(10)

## Un-stacked and stacked i.i.d. decomposition
IC(estimate(l1,id=id1,stack=FALSE))
IC(estimate(l1,id=id1))

## Combined i.i.d. decomposition
e1 <- estimate(l1,id=id1)
e2 <- estimate(l2,id=id2)
e3 <- estimate(l3,id=id3)
(a2 <- merge(e1,e2,e3))

## If all models were estimated on the same data we could use the
## syntax:
## Reduce(merge,estimate(list(l1,l2,l3)))

## Same:
IC(a1 <- merge(l1,l2,l3,id=list(id1,id2,id3)))

IC(merge(l1,l2,l3,id=TRUE)) # one-to-one (same clusters)
IC(merge(l1,l2,l3,id=FALSE)) # independence


# ------ influence function calculus -------
a <- estimate(coef = c("a" = 0.5), IC = scale(rnorm(10), scale=FALSE), id = 1:10)
b <- estimate(coef = c("b" = 0.8), IC = scale(rnorm(10), scale=FALSE), id = 1:10)

e <- c(a, b) # merge
merge(a, b)
c(e1=a, b) # naming of par
labels(e, c("p1", "p2")) # renaming parameters
e["a"] # subset
subset(e, "a")

# pipes
# c(a, b) |>
#  transform(function(x) x^2) |>
#  subset("a") |>
#  labels("sq")

# Parameter transformation with automatic calculation of derivatives
a * b
(3 * cos(a) / sqrt(b) + 1) / a
expit(c(a,b))
c(sum=sum(e), sum2=a+b,
  prod=prod(e), prod2=a*b)
e \%*\% e # inner prod.
c(1, 2) \%*\% e
c(pow = a^b)
a^c(0.5, 2)
c(b=e["a"] * e["b"] / a, also.b=e["b"])

B <- rbind(c(1,-1), c(1,0), c(0,1))
B \%*\% e
e == 1 # wald-test, null-hypothesis H0: b=1
e == c(1,2)
B \%*\% e == 1
}
\seealso{
\link{estimate.array}, \link{merge.estimate}, \link{contr}, \link{parsedesign},
\link{pairwise.diff}, \code{summary.estimate}, \code{coef.estimate}, \code{vcov.estimate},
\code{transform.estimate}, \code{labels.estimate}, \code{IC.estimate}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gof.R
\name{gof}
\alias{gof}
\alias{gof.lvmfit}
\alias{moments}
\alias{moments.lvm}
\alias{information}
\alias{information.lvmfit}
\alias{score}
\alias{score.lvmfit}
\alias{logLik.lvmfit}
\title{Extract model summaries and GOF statistics for model object}
\usage{
gof(object, ...)

\method{gof}{lvmfit}(object, chisq=FALSE, level=0.90, rmsea.threshold=0.05,all=FALSE,...)

moments(x,...)

\method{moments}{lvm}(x, p, debug=FALSE, conditional=FALSE, data=NULL, latent=FALSE, ...)

\method{logLik}{lvmfit}(object, p=coef(object),
                      data=model.frame(object),
                      model=object$estimator,
                      weights=Weights(object),
                      data2=object$data$data2,
                          ...)

\method{score}{lvmfit}(x, data=model.frame(x), p=pars(x), model=x$estimator,
                   weights=Weights(x), data2=x$data$data2, ...)

\method{information}{lvmfit}(x,p=pars(x),n=x$data$n,data=model.frame(x),
                   model=x$estimator,weights=Weights(x), data2=x$data$data2, ...)
}
\arguments{
\item{object}{Model object}

\item{\dots}{Additional arguments to be passed to the low level functions}

\item{x}{Model object}

\item{p}{Parameter vector used to calculate statistics}

\item{data}{Data.frame to use}

\item{latent}{If TRUE predictions of latent variables are included in output}

\item{data2}{Optional second data.frame (only for censored observations)}

\item{weights}{Optional weight matrix}

\item{n}{Number of observations}

\item{conditional}{If TRUE the conditional moments given the covariates are
calculated. Otherwise the joint moments are calculated}

\item{model}{String defining estimator, e.g. "gaussian" (see
\code{estimate})}

\item{debug}{Debugging only}

\item{chisq}{Boolean indicating whether to calculate chi-squared
goodness-of-fit (always TRUE for estimator='gaussian')}

\item{level}{Level of confidence limits for RMSEA}

\item{rmsea.threshold}{Which probability to calculate, Pr(RMSEA<rmsea.treshold)}

\item{all}{Calculate all (ad hoc) FIT indices: TLI, CFI, NFI, SRMR, ...}
}
\value{
A \code{htest}-object.
}
\description{
Calculates various GOF statistics for model object including global
chi-squared test statistic and AIC. Extract model-specific mean and variance
structure, residuals and various predicitions.
}
\examples{
m <- lvm(list(y~v1+v2+v3+v4,c(v1,v2,v3,v4)~x))
set.seed(1)
dd <- sim(m,1000)
e <- estimate(m, dd)
gof(e,all=TRUE,rmsea.threshold=0.05,level=0.9)


set.seed(1)
m <- lvm(list(c(y1,y2,y3)~u,y1~x)); latent(m) <- ~u
regression(m,c(y2,y3)~u) <- "b"
d <- sim(m,1000)
e <- estimate(m,d)
rsq(e)
##'
rr <- rsq(e,TRUE)
rr
estimate(rr,contrast=rbind(c(1,-1,0),c(1,0,-1),c(0,1,-1)))

}
\author{
Klaus K. Holst
}
\keyword{methods}
\keyword{models}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/complik.R
\name{complik}
\alias{complik}
\title{Composite Likelihood for probit latent variable models}
\usage{
complik(
  x,
  data,
  k = 2,
  type = c("all", "nearest"),
  pairlist,
  messages = 0,
  estimator = "normal",
  quick = FALSE,
  ...
)
}
\arguments{
\item{x}{\code{lvm}-object}

\item{data}{data.frame}

\item{k}{Size of composite groups}

\item{type}{Determines number of groups. With \code{type="nearest"} (default)
only neighboring items will be grouped, e.g. for \code{k=2}
(y1,y2),(y2,y3),... With \code{type="all"} all combinations of size \code{k}
are included}

\item{pairlist}{A list of indices specifying the composite groups. Optional
argument which overrides \code{k} and \code{type} but gives complete
flexibility in the specification of the composite likelihood}

\item{messages}{Control amount of messages printed}

\item{estimator}{Model (pseudo-likelihood) to use for the pairs/groups}

\item{quick}{If TRUE the parameter estimates are calculated but all additional
information such as standard errors are skipped}

\item{\dots}{Additional arguments parsed on to lower-level functions}
}
\value{
An object of class \code{estimate.complik} inheriting methods from \code{lvm}
}
\description{
Estimate parameters in a probit latent variable model via a composite
likelihood decomposition.
}
\examples{
m <- lvm(c(y1,y2,y3)~b*x+1*u[0],latent=~u)
ordinal(m,K=2) <- ~y1+y2+y3
d <- sim(m,50,seed=1)
if (requireNamespace("mets", quietly=TRUE)) {
   e1 <- complik(m,d,control=list(trace=1),type="all")
}
}
\seealso{
estimate
}
\author{
Klaus K. Holst
}
\keyword{models}
\keyword{regression}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/regression.R
\name{regression<-}
\alias{regression<-}
\alias{regression}
\alias{regression<-.lvm}
\alias{regression.lvm}
\title{Add regression association to latent variable model}
\usage{
\method{regression}{lvm}(object = lvm(), to, from, fn = NA,
messages = lava.options()$messages, additive=TRUE, y, x, value, ...)
\method{regression}{lvm}(object, to=NULL, quick=FALSE, ...) <- value
}
\arguments{
\item{object}{\code{lvm}-object.}

\item{\dots}{Additional arguments to be passed to the low level functions}

\item{value}{A formula specifying the linear constraints or if
\code{to=NULL} a \code{list} of parameter values.}

\item{to}{Character vector of outcome(s) or formula object.}

\item{from}{Character vector of predictor(s).}

\item{fn}{Real function defining the functional form of predictors (for
simulation only).}

\item{messages}{Controls which messages are turned on/off (0: all off)}

\item{additive}{If FALSE and predictor is categorical a non-additive effect is assumed}

\item{y}{Alias for 'to'}

\item{x}{Alias for 'from'}

\item{quick}{Faster implementation without parameter constraints}
}
\value{
A \code{lvm}-object
}
\description{
Define regression association between variables in a \code{lvm}-object and
define linear constraints between model equations.
}
\details{
The \code{regression} function is used to specify linear associations
between variables of a latent variable model, and offers formula syntax
resembling the model specification of e.g. \code{lm}.

For instance, to add the following linear regression model, to the
\code{lvm}-object, \code{m}:
\deqn{ E(Y|X_1,X_2) = \beta_1 X_1 + \beta_2 X_2}
We can write

\code{regression(m) <- y ~ x1 + x2}

Multivariate models can be specified by successive calls with
\code{regression}, but multivariate formulas are also supported, e.g.

\code{regression(m) <- c(y1,y2) ~ x1 + x2}

defines
\deqn{ E(Y_i|X_1,X_2) = \beta_{1i} X_1 + \beta_{2i} X_2 }

The special function, \code{f}, can be used in the model specification to
specify linear constraints. E.g. to fix \eqn{\beta_1=\beta_2}
, we could write

\code{regression(m) <- y ~ f(x1,beta) + f(x2,beta)}

The second argument of \code{f} can also be a number (e.g. defining an
offset) or be set to \code{NA} in order to clear any previously defined
linear constraints.

Alternatively, a more straight forward notation can be used:

\code{regression(m) <- y ~ beta*x1 + beta*x2}

All the parameter values of the linear constraints can be given as the right
handside expression of the assigment function \code{regression<-} if the
first (and possibly second) argument is defined as well. E.g:

\code{regression(m,y1~x1+x2) <- list("a1","b1")}

defines \eqn{E(Y_1|X_1,X_2) = a1 X_1 + b1 X_2}. The rhs argument can be a
mixture of character and numeric values (and NA's to remove constraints).

The function \code{regression} (called without additional arguments) can be
used to inspect the linear constraints of a \code{lvm}-object.
}
\note{
Variables will be added to the model if not already present.
}
\examples{

m <- lvm() ## Initialize empty lvm-object
### E(y1|z,v) = beta1*z + beta2*v
regression(m) <- y1 ~ z + v
### E(y2|x,z,v) = beta*x + beta*z + 2*v + beta3*u
regression(m) <- y2 ~ f(x,beta) + f(z,beta)  + f(v,2) + u
### Clear restriction on association between y and
### fix slope coefficient of u to beta
regression(m, y2 ~ v+u) <- list(NA,"beta")

regression(m) ## Examine current linear parameter constraints

## ## A multivariate model, E(yi|x1,x2) = beta[1i]*x1 + beta[2i]*x2:
m2 <- lvm(c(y1,y2) ~ x1+x2)

}
\seealso{
\code{\link{intercept<-}}, \code{\link{covariance<-}},
\code{\link{constrain<-}}, \code{\link{parameter<-}},
\code{\link{latent<-}}, \code{\link{cancel<-}}, \code{\link{kill<-}}
}
\author{
Klaus K. Holst
}
\keyword{models}
\keyword{regression}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lava-package.R
\docType{data}
\name{nldata}
\alias{nldata}
\title{Example data (nonlinear model)}
\format{
data.frame
}
\source{
Simulated
}
\description{
Example data (nonlinear model)
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sim.default.R
\name{sim.default}
\alias{sim.default}
\alias{sim}
\alias{as.sim}
\title{Monte Carlo simulation}
\usage{
\method{sim}{default}(
  x = NULL,
  R = 100,
  f = NULL,
  colnames = NULL,
  seed = NULL,
  args = list(),
  iter = FALSE,
  mc.cores,
  progressr.message = NULL,
  estimate.index = 1:2,
  ...
)
}
\arguments{
\item{x}{function or 'sim' object}

\item{R}{Number of replications or data.frame with parameters}

\item{f}{Optional function (i.e., if x is a matrix)}

\item{colnames}{Optional column names}

\item{seed}{(optional) Seed (needed with cl=TRUE)}

\item{args}{(optional) list of named arguments passed to (mc)mapply}

\item{iter}{If TRUE the iteration number is passed as first argument to
(mc)mapply}

\item{mc.cores}{Optional number of cores. Will use parallel::mcmapply
instead of future}

\item{progressr.message}{Optional message for the progressr progress-bar}

\item{estimate.index}{If return object inherits from \code{estimate} then only
these column indices are extracted (estimate, se, lower, upper, p-val)}

\item{...}{Additional arguments to \code{\link[future.apply:future_mapply]{future.apply::future_mapply()}}}
}
\description{
Applies a function repeatedly for a specified number of replications or over
a list/data.frame with plot and summary methods for summarizing the Monte
Carlo experiment. Can be parallelized via the future package (use the
\code{\link[future:plan]{future::plan()}} function).
}
\details{
To parallelize the calculation use the \code{\link[future:plan]{future::plan()}} function
(e.g., \code{future::plan(multisession())} to distribute the calculations over
the \code{R} replications on all available cores). The output is controlled via
the progressr package (e.g., \code{progressr::handlers(global=TRUE)} to enable
progress information).
}
\examples{
m <- lvm(y~x+e)
distribution(m,~y) <- 0
distribution(m,~x) <- uniform.lvm(a=-1.1,b=1.1)
transform(m,e~x) <- function(x) (1*x^4)*rnorm(length(x),sd=1)

onerun <- function(iter=NULL,...,n=2e3,b0=1,idx=2) {
    d <- sim(m,n,p=c("y~x"=b0))
    l <- lm(y~x,d)
    res <- c(coef(summary(l))[idx,1:2],
             confint(l)[idx,],
             estimate(l,only.coef=TRUE)[idx,2:4])
    names(res) <- c("Estimate","Model.se","Model.lo","Model.hi",
                    "Sandwich.se","Sandwich.lo","Sandwich.hi")
    res
}
val <- sim(onerun,R=10,b0=1)
val

val <- sim(val,R=40,b0=1) ## append results
summary(val,estimate=c(1,1),confint=c(3,4,6,7),true=c(1,1))

summary(val,estimate=c(1,1),se=c(2,5),names=c("Model","Sandwich"))
summary(val,estimate=c(1,1),se=c(2,5),true=c(1,1),
        names=c("Model","Sandwich"),confint=TRUE)

if (interactive()) {
    plot(val,estimate=1,c(2,5),true=1,
         names=c("Model","Sandwich"),polygon=FALSE)
    plot(val,estimate=c(1,1),se=c(2,5),main=NULL,
         true=c(1,1),names=c("Model","Sandwich"),
         line.lwd=1,col=c("gray20","gray60"),
         rug=FALSE)
    plot(val,estimate=c(1,1),se=c(2,5),true=c(1,1),
         names=c("Model","Sandwich"))
}

f <- function(a=1, b=1) {
  rep(a*b, 5)
}
R <- Expand(a=1:3, b=1:3)
sim(f, R)
sim(function(a,b) f(a,b), 3, args=c(a=5,b=5))
sim(function(iter=1,a=5,b=5) iter*f(a,b), iter=TRUE, R=5)
}
\seealso{
\code{\link[=summary.sim]{summary.sim()}} \code{\link[=plot.sim]{plot.sim()}} \code{\link[=sim.lvm]{sim.lvm()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/estimate.default.R
\name{estimate.default}
\alias{estimate.default}
\alias{estimate}
\alias{estimate.mlm}
\title{Influence function based inference}
\usage{
\method{estimate}{default}(
  x = NULL,
  f = NULL,
  ...,
  data,
  id,
  coef,
  IC = TRUE,
  vcov,
  stack = TRUE,
  average = FALSE,
  subset,
  keep,
  use,
  regex = FALSE,
  ignore.case = FALSE,
  print = NULL,
  labels,
  label.width,
  contrast,
  null,
  level = NULL,
  type = NULL,
  var.adj = NULL,
  df = NULL,
  back.transform = NULL
)
}
\arguments{
\item{x}{model object (\code{glm}, \code{lvmfit}, ...) or an existing \code{estimate}
object. When two model objects are supplied (e.g., \code{estimate(g, g0)}) a
likelihood-ratio test is performed.}

\item{f}{transformation of model parameters. Accepts several input types: - A
\strong{function} \code{f(p)} or \code{f(p, data)}: applies the delta method. When \code{f}
returns a named list the names are used as parameter labels. - A
\strong{matrix}: used as a contrast (linear combination) matrix. - A \strong{numeric
vector} of parameter indices: converted to a contrast that selects and
differences those parameters. - A \strong{list} of indices: each element selects
one parameter. - \strong{Character} expressions: supports wildcards (\code{"?"},
\code{"*"}) and arithmetic on parameter names (e.g., \code{"z" - "x"}, \code{2 * "z" - 3 * "x"}).}

\item{...}{additional arguments to lower level functions}

\item{data}{\code{data.frame} used by \code{f} when the transformation depends on
covariates (see \code{average}). Defaults to \code{model.frame(x)}.}

\item{id}{(optional) cluster identifier. Can be a vector of cluster IDs, a
one-sided formula (evaluated in \code{data}), a single character column name, or
a logical scalar (\code{TRUE} for one-to-one matching, \code{FALSE} for
independence). When supplied, the IF is aggregated within clusters to
produce cluster-robust standard errors.}

\item{coef}{(optional) named parameter vector. Used instead of \code{coef(x)} when
constructing an \code{estimate} object without a model.}

\item{IC}{if \code{TRUE} (default) the influence function matrix is estimated and
stored in the returned object (extract with the \link{IC} method). Can also be a
user-supplied IF matrix (one row per observation, one column per
parameter), which is used directly instead of estimating it from \code{x}.}

\item{vcov}{(optional) covariance matrix of parameter estimates, or a
logical. If \code{TRUE}, \link[stats:vcov]{stats::vcov} is used to obtain the (model-based)
covariance matrix from \code{x}, yielding non-robust standard errors. If a
matrix is supplied it is used directly. When omitted or \code{FALSE}, robust
standard errors are computed from the influence function.}

\item{stack}{if \code{TRUE} (default) the influence function contributions are
summed within each cluster defined by \code{id}. Set to \code{FALSE} to keep the
un-stacked (per-observation) decomposition.}

\item{average}{if \code{TRUE} the function computes the standardized
(marginalized) estimate \eqn{\hat\Psi = P_n f(X; \hat\theta)}, i.e., the
empirical mean of \code{f(p, data)} over all rows of \code{data}. The influence
function accounts for both the empirical averaging and the parameter
estimation uncertainty (see Details).}

\item{subset}{(optional) logical vector, expression evaluated in \code{data}, or
column name. When used together with \code{average = TRUE}, the average is
conditioned on the subpopulation where \code{subset} is \code{TRUE}, yielding a
conditional marginalized estimate.}

\item{keep}{(optional) index of parameters to keep from final result. Accepts
integer indices, character names, or (with \code{regex = TRUE}) perl-compatible
regular expressions.}

\item{use}{(optional) index of parameters to use in calculations. The
selected parameters are first extracted (via \code{keep}) and then the remaining
arguments (\code{f}, \code{contrast}, etc.) are applied to this subset.}

\item{regex}{if \code{TRUE} use perl-compatible regular expressions for \code{keep} and
\code{use} arguments}

\item{ignore.case}{ignore case in regular expressions}

\item{print}{(optional) custom print function for the resulting \code{estimate}
object}

\item{labels}{(optional) character vector of coefficient names}

\item{label.width}{(optional) max display width of labels}

\item{contrast}{(deprecated, use summary method) contrast matrix for a final
Wald test. When supplied together with \code{null}, tests \eqn{H_0: B\theta =
b_0}.}

\item{null}{(deprecated, use summary method) null hypothesis vector \eqn{b_0}
to test against (default 0)}

\item{level}{(deprecated, use summary method) level of confidence limits
(default 0.95)}

\item{type}{(deprecated, use summary method) type of small-sample correction
for cluster-robust variance. One of: - \code{"robust"} (default): no correction.
\itemize{
\item \code{"df"}: applies \eqn{n/(n-p)} correction (Mancl & DeRouen, 2001). -
\code{"mbn"}: Morel-Bokossa-Neerchal (2003) correction. - \code{"hc3"}:
leverage-adjusted HC3-type correction (blended with \code{var.adj}). - \code{"hc4"}:
Cribari-Neto (2004) leverage-adjusted correction.
}}

\item{var.adj}{(deprecated, use summary method) blending parameter for the
HC3 leverage adjustment (default 0.25). Controls the weight between
observation-level empirical leverage and the average leverage \eqn{p/n}.}

\item{df}{(deprecated, use summary method) degrees of freedom for t-based
inference (default: \code{NULL} for Gaussian approximation; when set, confidence
intervals and p-values use the t-distribution with \code{df} degrees of
freedom).}

\item{back.transform}{(deprecated, use summary method) function applied to
the point estimates and confidence interval bounds \emph{after} inference is
performed on the original scale. Useful for variance-stabilizing
transformations, e.g., compute CIs on the \code{atanh} (Fisher z) scale and
back-transform with \code{tanh}.}
}
\value{
Object of class \code{estimate} with the following elements:
\item{coef}{Named vector of parameter estimates.}
\item{vcov}{Variance-covariance matrix.}
\item{IC}{Influence function matrix (observations x parameters).}
\item{coefmat}{Formatted coefficient table (estimate, std.err,
confidence limits, p-value).}
\item{id}{Cluster/id variable used.}
\item{ncluster}{Number of clusters.}
\item{n}{Number of observations.}
\item{compare}{(When \code{null} or contrasts are specified) Wald test result.}
}
\description{
Primary tool for obtaining parameter estimates with robust (sandwich)
standard errors, applying the delta method, and testing linear hypotheses.
The function returns an object of class \code{estimate} which serves as a
general container for parameter estimates and their influence functions
(IFs). Three calling conventions are supported:
}
\details{
\itemize{
\item \code{estimate(x, ...)} -- extract estimates from a model object
\item \code{estimate(coef=, IC=, ...)} -- construct from coefficients and IF matrix
\item \code{estimate(coef=, vcov=, ...)} -- construct from coefficients and
covariance matrix
}
}
\section{Influence functions and robust standard errors}{
An estimator \eqn{\widehat{\theta}} is \emph{regular and asymptotically
linear} (RAL) when it admits the iid decomposition
\deqn{\sqrt{n}(\widehat{\theta}-\theta) =
\frac{1}{\sqrt{n}}\sum_{i=1}^n \mathrm{IC}(Z_i; P) + o_p(1)}
where \eqn{\mathrm{IC}} is the unique \emph{influence function} satisfying
\eqn{E\{\mathrm{IC}(Z; P)\} = 0}. By the central limit theorem
\deqn{\sqrt{n}(\widehat{\theta}-\theta)
\overset{d}{\longrightarrow}
N(0,\; \mathrm{Var}\{\mathrm{IC}(Z; P)\})}
and the asymptotic variance is consistently estimated by the empirical
variance of the plugin IF estimate, yielding robust (sandwich) standard
errors. The estimated IF can be extracted with the \link{IC} method.
}

\section{Parameter transformations (delta method)}{
When \code{f} is a function \eqn{\phi: R^p \to R^m}, the delta method is
applied:
\deqn{\sqrt{n}\{\phi(\widehat{\theta}) - \phi(\theta)\} =
\frac{1}{\sqrt{n}}\sum_{i=1}^n
\nabla\phi(\theta)\,\mathrm{IC}(Z_i; P) + o_p(1)}
Derivatives are computed numerically via \link[numDeriv:jacobian]{numDeriv::jacobian} unless the
function returns an attribute \code{"grad"} with the analytic Jacobian.

Alternatively, \code{estimate} objects support direct arithmetic operations
(e.g., \code{a * b}, \code{exp(a)}, \code{a^b}) which apply the delta method with
\emph{exact} (analytical) derivatives computed automatically. This influence
function calculus allows building complex transformations from simple
building blocks without numerical differentiation. See the last example
section ("influence function calculus") and
\code{vignette("influencefunction", package = "lava")} for details.
}

\section{Averaging and marginalization}{
When \code{average = TRUE} and \code{f(p, data)} depends on covariates, the
target parameter is the standardized (marginalized) estimate
\eqn{\Psi = E\{f(X;\theta)\}}. The IF for the averaged estimate
accounts for both the empirical averaging and parameter estimation
uncertainty:
\deqn{\mathrm{IC}_\Psi(Z; P) = f(X;\theta) - \Psi +
[E\nabla_\theta f(X;\theta)]\,\phi(Z; P)}
When \code{subset} is also specified, the average is conditioned on the
subpopulation, yielding a conditional marginalized estimate.
}

\section{Cluster-robust standard errors}{
When \code{id} is supplied, the per-observation IF contributions are summed
within clusters (when \code{stack = TRUE}), producing the cluster-level IF
\eqn{\widetilde{\mathrm{IC}}(Z_i; P) = \sum_{k=1}^{N_i}
\frac{n}{N}\mathrm{IC}(Z_{ik}; P)}.
The resulting variance estimate is equivalent to the GEE working
independence sandwich estimator.

For full theoretical background and worked examples see
\code{vignette("influencefunction", package = "lava")}.
}

\examples{

## Simulation from logistic regression model
m <- lvm(y~x+z);
distribution(m,y~x) <- binomial.lvm("logit")
d <- sim(m,1000)
g <- glm(y~z+x,data=d,family=binomial())
g0 <- glm(y~1,data=d,family=binomial())

## LRT
estimate(g, g0)


estimate(g)

## Testing contrasts
summary(estimate(g), null=0)
estimate(g, rbind(c(1,1,0), c(1,0,2)))
summary(estimate(g, rbind(c(1,1,0), c(1,0,2))), null=c(1,2))
estimate(g, 2:3) ## same as cbind(0,1,-1)
estimate(g, as.list(2:3)) ## same as rbind(c(0,1,0),c(0,0,1))
## Alternative syntax
estimate(g, "z", "z"-"x", 2*"z"-3*"x")
estimate(g, "?")  ## Wildcards
estimate(g, "*Int*", "z")
summary(estimate(g, "1", "2"-"3"), null = c(0,1))
estimate(g, 2, 3)

## Usual (non-robust) confidence intervals
estimate(g, vcov=TRUE)
estimate(g, vcov=vcov(g))

## Transformations
estimate(g, function(p) p[1]+p[2])

## Multiple parameters
e <- estimate(g, function(p) c(p[1]+p[2], p[1]*p[2]))
e
vcov(e)

## Label new parameters
estimate(g, function(p) list("a1"=p[1]+p[2], "b1"=p[1]*p[2]))
#'
## Multiple group
m <- lvm(y~x)
m <- baptize(m)
d2 <- d1 <- sim(m,50,seed=1)
e <- estimate(list(m,m),list(d1,d2))
estimate(e) ## Wrong
ee <- estimate(e, id=rep(seq(nrow(d1)), 2)) ## Clustered
ee
estimate(lm(y~x,d1))

## Marginalize / standardization
f <- function(p,data)
  list(p0=expit(p["(Intercept)"] + p["z"]*data[,"z"]),
       p1=expit(p["(Intercept)"] + p["x"] + p["z"]*data[,"z"]))
e <- estimate(g, f, average=TRUE)
e
estimate(e,diff)
estimate(e,cbind(1,1))

## Clusters and subset (conditional marginal effects)
d$id <- rep(seq(nrow(d)/4),each=4)
estimate(g,function(p,data)
         list(p0=expit(p[1] + p["z"]*data[,"z"])),
         subset=d$z>0, id=d$id, average=TRUE)

## More examples with clusters:
m <- lvm(c(y1,y2,y3)~u+x)
d <- sim(m,10)
l1 <- glm(y1~x,data=d)
l2 <- glm(y2~x,data=d)
l3 <- glm(y3~x,data=d)

## Some random id-numbers
id1 <- c(1,1,4,1,3,1,2,3,4,5)
id2 <- c(1,2,3,4,5,6,7,8,1,1)
id3 <- seq(10)

## Un-stacked and stacked i.i.d. decomposition
IC(estimate(l1,id=id1,stack=FALSE))
IC(estimate(l1,id=id1))

## Combined i.i.d. decomposition
e1 <- estimate(l1,id=id1)
e2 <- estimate(l2,id=id2)
e3 <- estimate(l3,id=id3)
(a2 <- merge(e1,e2,e3))

## If all models were estimated on the same data we could use the
## syntax:
## Reduce(merge,estimate(list(l1,l2,l3)))

## Same:
IC(a1 <- merge(l1,l2,l3,id=list(id1,id2,id3)))

IC(merge(l1,l2,l3,id=TRUE)) # one-to-one (same clusters)
IC(merge(l1,l2,l3,id=FALSE)) # independence


# ------ influence function calculus -------
a <- estimate(coef = c("a" = 0.5), IC = scale(rnorm(10), scale=FALSE), id = 1:10)
b <- estimate(coef = c("b" = 0.8), IC = scale(rnorm(10), scale=FALSE), id = 1:10)

e <- c(a, b) # merge
merge(a, b)
c(e1=a, b) # naming of par
labels(e, c("p1", "p2")) # renaming parameters
e["a"] # subset
subset(e, "a")

# pipes
# c(a, b) |>
#  transform(function(x) x^2) |>
#  subset("a") |>
#  labels("sq")

# Parameter transformation with automatic calculation of derivatives
a * b
(3 * cos(a) / sqrt(b) + 1) / a
expit(c(a,b))
c(sum=sum(e), sum2=a+b,
  prod=prod(e), prod2=a*b)
e \%*\% e # inner prod.
c(1, 2) \%*\% e
c(pow = a^b)
a^c(0.5, 2)
c(b=e["a"] * e["b"] / a, also.b=e["b"])

B <- rbind(c(1,-1), c(1,0), c(0,1))
B \%*\% e
e == 1 # wald-test, null-hypothesis H0: b=1
e == c(1,2)
B \%*\% e == 1
}
\seealso{
\link{estimate.array}, \link{merge.estimate}, \link{contr}, \link{parsedesign},
\link{pairwise.diff}, \link{c.estimate}, \link{summary.estimate},
\code{coef.estimate},
\code{vcov.estimate}, \code{transform.estimate}, \code{labels.estimate},
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/revdiag.R
\name{revdiag}
\alias{revdiag}
\alias{revdiag<-}
\alias{offdiag}
\alias{offdiag<-}
\title{Create/extract 'reverse'-diagonal matrix or off-diagonal elements}
\usage{
revdiag(x,...)
offdiag(x,type=0,...)

revdiag(x,...) <- value
offdiag(x,type=0,...) <- value
}
\arguments{
\item{x}{vector}

\item{\dots}{additional arguments to lower level functions}

\item{value}{For the assignment function the values to put in the diagonal}

\item{type}{0: upper and lower triangular, 1: upper triangular, 2: lower triangular, 3: upper triangular + diagonal, 4: lower triangular + diagonal}
}
\description{
Create/extract 'reverse'-diagonal matrix or off-diagonal elements
}
\author{
Klaus K. Holst
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/kill.R
\name{rmvar}
\alias{rmvar}
\alias{rmvar<-}
\alias{kill}
\alias{kill<-}
\title{Remove variables from (model) object.}
\usage{
rmvar(x, ...) <- value
}
\arguments{
\item{x}{Model object}

\item{\dots}{additional arguments to lower level functions}

\item{value}{Vector of variables or formula specifying which nodes to
remove}
}
\description{
Generic method for removing elements of object
}
\examples{
m <- lvm()
addvar(m) <- ~y1+y2+x
covariance(m) <- y1~y2
regression(m) <- c(y1,y2) ~ x
### Cancel the covariance between the residuals of y1 and y2
cancel(m) <- y1~y2
### Remove y2 from the model
rmvar(m) <- ~y2

}
\seealso{
\code{cancel}
}
\author{
Klaus K. Holst
}
\keyword{models}
\keyword{regression}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrapvec.R
\name{wrapvec}
\alias{wrapvec}
\title{Wrap vector}
\usage{
wrapvec(x, delta = 0L, ...)
}
\arguments{
\item{x}{Vector or integer}

\item{delta}{Shift}

\item{...}{Additional parameters}
}
\description{
Wrap vector
}
\examples{
wrapvec(5,2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/twostage.R
\name{twostage.lvmfit}
\alias{twostage.lvmfit}
\alias{twostage.lvm}
\alias{twostage.lvm.mixture}
\alias{twostage.estimate}
\alias{nonlinear}
\alias{nonlinear<-}
\title{Two-stage estimator (non-linear SEM)}
\usage{
\method{twostage}{lvmfit}(
  object,
  model2,
  data = NULL,
  predict.fun = NULL,
  id1 = NULL,
  id2 = NULL,
  all = FALSE,
  formula = NULL,
  std.err = TRUE,
  ...
)
}
\arguments{
\item{object}{Stage 1 measurement model}

\item{model2}{Stage 2 SEM}

\item{data}{data.frame}

\item{predict.fun}{Prediction of latent variable}

\item{id1}{Optional id-variable (stage 1 model)}

\item{id2}{Optional id-variable (stage 2 model)}

\item{all}{If TRUE return additional output (naive estimates)}

\item{formula}{optional formula specifying non-linear relation}

\item{std.err}{If FALSE calculations of standard errors will be skipped}

\item{...}{Additional arguments to lower level functions}
}
\description{
Two-stage estimator for non-linear structural equation models
}
\examples{
m <- lvm(c(x1,x2,x3)~f1,f1~z,
         c(y1,y2,y3)~f2,f2~f1+z)
latent(m) <- ~f1+f2
d <- simulate(m,100,p=c("f2,f2"=2,"f1,f1"=0.5),seed=1)

## Full MLE
ee <- estimate(m,d)

## Manual two-stage
\dontrun{
m1 <- lvm(c(x1,x2,x3)~f1,f1~z); latent(m1) <- ~f1
e1 <- estimate(m1,d)
pp1 <- predict(e1,f1~x1+x2+x3)

d$u1 <- pp1[,]
d$u2 <- pp1[,]^2+attr(pp1,"cond.var")[1]
m2 <- lvm(c(y1,y2,y3)~eta,c(y1,eta)~u1+u2+z); latent(m2) <- ~eta
e2 <- estimate(m2,d)
}

## Two-stage
m1 <- lvm(c(x1,x2,x3)~f1,f1~z); latent(m1) <- ~f1
m2 <- lvm(c(y1,y2,y3)~eta,c(y1,eta)~u1+u2+z); latent(m2) <- ~eta
pred <- function(mu,var,data,...)
    cbind("u1"=mu[,1],"u2"=mu[,1]^2+var[1])
(mm <- twostage(m1,model2=m2,data=d,predict.fun=pred))

if (interactive()) {
    pf <- function(p) p["eta"]+p["eta~u1"]*u + p["eta~u2"]*u^2
    plot(mm,f=pf,data=data.frame(u=seq(-2,2,length.out=100)),lwd=2)
}

\donttest{ ## Reduce test timing
## Splines
f <- function(x) cos(2*x)+x+-0.25*x^2
m <- lvm(x1+x2+x3~eta1, y1+y2+y3~eta2, latent=~eta1+eta2)
functional(m, eta2~eta1) <- f
d <- sim(m,500,seed=1,latent=TRUE)
m1 <- lvm(x1+x2+x3~eta1,latent=~eta1)
m2 <- lvm(y1+y2+y3~eta2,latent=~eta2)
mm <- twostage(m1,m2,formula=eta2~eta1,type="spline")
if (interactive()) plot(mm)

nonlinear(m2,type="quadratic") <- eta2~eta1
a <- twostage(m1,m2,data=d)
if (interactive()) plot(a)

kn <- c(-1,0,1)
nonlinear(m2,type="spline",knots=kn) <- eta2~eta1
a <- twostage(m1,m2,data=d)
x <- seq(-3,3,by=0.1)
y <- predict(a, newdata=data.frame(eta1=x))

if (interactive()) {
  plot(eta2~eta1, data=d)
  lines(x,y, col="red", lwd=5)

  p <- estimate(a,f=function(p) predict(a,p=p,newdata=x))$coefmat
  plot(eta2~eta1, data=d)
  lines(x,p[,1], col="red", lwd=5)
  confband(x,lower=p[,3],upper=p[,4],center=p[,1], polygon=TRUE, col=Col(2,0.2))

  l1 <- lm(eta2~splines::ns(eta1,knots=kn),data=d)
  p1 <- predict(l1,newdata=data.frame(eta1=x),interval="confidence")
  lines(x,p1[,1],col="green",lwd=5)
  confband(x,lower=p1[,2],upper=p1[,3],center=p1[,1], polygon=TRUE, col=Col(3,0.2))
}
} ## Reduce test timing

\dontrun{ ## Reduce timing
 ## Cross-validation example
 ma <- lvm(c(x1,x2,x3)~u,latent=~u)
 ms <- functional(ma, y~u, value=function(x) -.4*x^2)
 d <- sim(ms,500)#,seed=1)
 ea <- estimate(ma,d)

 mb <- lvm()
 mb1 <- nonlinear(mb,type="linear",y~u)
 mb2 <- nonlinear(mb,type="quadratic",y~u)
 mb3 <- nonlinear(mb,type="spline",knots=c(-3,-1,0,1,3),y~u)
 mb4 <- nonlinear(mb,type="spline",knots=c(-3,-2,-1,0,1,2,3),y~u)
 ff <- lapply(list(mb1,mb2,mb3,mb4),
      function(m) function(data,...) twostage(ma,m,data=data,st.derr=FALSE))
 a <- cv(ff,data=d,rep=1)
 a
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getsas.R
\name{getSAS}
\alias{getSAS}
\title{Read SAS output}
\usage{
getSAS(infile, entry = "Parameter Estimates", ...)
}
\arguments{
\item{infile}{file (csv file generated by ODS)}

\item{entry}{Name of entry to capture}

\item{\dots}{additional arguments to lower level functions}
}
\description{
Run SAS code like in the following:
}
\details{
ODS CSVALL BODY="myest.csv";
proc nlmixed data=aj qpoints=2 dampstep=0.5;
...
run;
ODS CSVALL Close;

and read results into R with:

\code{getsas("myest.csv","Parameter Estimates")}
}
\author{
Klaus K. Holst
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.sim.R
\name{plot.sim}
\alias{plot.sim}
\alias{density.sim}
\title{Plot method for simulation 'sim' objects}
\usage{
\method{plot}{sim}(
  x,
  estimate,
  se = NULL,
  true = NULL,
  names = NULL,
  auto.layout = TRUE,
  byrow = FALSE,
  type = "p",
  ask = grDevices::dev.interactive(),
  col = c("gray60", "orange", "darkblue", "seagreen", "darkred"),
  pch = 16,
  cex = 0.5,
  lty = 1,
  lwd = 0.3,
  legend,
  legendpos = "topleft",
  cex.legend = 0.8,
  plot.type = c("multiple", "single"),
  polygon = TRUE,
  density = NULL,
  angle = -45,
  cex.axis = 0.8,
  alpha = 0.2,
  main,
  cex.main = 1,
  equal = FALSE,
  delta = 1.15,
  ylim = NULL,
  xlim = NULL,
  ylab = "",
  xlab = "",
  rug = FALSE,
  rug.alpha = 0.5,
  line.col = scatter.col,
  line.lwd = 1,
  line.lty = 1,
  line.alpha = 1,
  scatter.ylab = "Estimate",
  scatter.ylim = NULL,
  scatter.xlim = NULL,
  scatter.alpha = 0.5,
  scatter.col = col,
  border = col,
  true.lty = 2,
  true.col = "gray70",
  true.lwd = 1.2,
  density.plot = TRUE,
  scatter.plot = FALSE,
  running.mean = scatter.plot,
  add = FALSE,
  ...
)
}
\arguments{
\item{x}{sim object}

\item{estimate}{columns with estimates}

\item{se}{columns with standard error estimates}

\item{true}{(optional) vector of true parameter values}

\item{names}{(optional) names of estimates}

\item{auto.layout}{Auto layout (default TRUE)}

\item{byrow}{Add new plots to layout by row}

\item{type}{plot type}

\item{ask}{if TRUE user is asked for input, before a new figure is drawn}

\item{col}{colour (for each estimate)}

\item{pch}{plot symbol}

\item{cex}{point size}

\item{lty}{line type}

\item{lwd}{line width}

\item{legend}{legend}

\item{legendpos}{legend position}

\item{cex.legend}{size of legend text}

\item{plot.type}{'single' or 'multiple' (default)}

\item{polygon}{if TRUE fill the density estimates with colour}

\item{density}{if non-zero add shading lines to polygon}

\item{angle}{shading lines angle of polygon}

\item{cex.axis}{Font size on axis}

\item{alpha}{Semi-transparent level (1: non-transparent, 0: full)}

\item{main}{Main title}

\item{cex.main}{Size of title font}

\item{equal}{Same x-axis and y-axis for all plots}

\item{delta}{Controls the amount of space around axis limits}

\item{ylim}{y-axis limits}

\item{xlim}{x-axis limits}

\item{ylab}{y axis label}

\item{xlab}{x axis label}

\item{rug}{if TRUE add rug representation of data to x-axis}

\item{rug.alpha}{rug semi-transparency level}

\item{line.col}{line colour (running mean, only for scatter plots)}

\item{line.lwd}{line width (running mean, only for scatter plots)}

\item{line.lty}{line type (running mean, only for scatter plots)}

\item{line.alpha}{line transparency}

\item{scatter.ylab}{y label for density plots}

\item{scatter.ylim}{y-axis limits for density plots}

\item{scatter.xlim}{x-axis limits for density plots}

\item{scatter.alpha}{semi-transparency of scatter plot}

\item{scatter.col}{scatter plot colour}

\item{border}{border colour of density estimates}

\item{true.lty}{true parameter estimate line type}

\item{true.col}{true parameter colour}

\item{true.lwd}{true parameter line width}

\item{density.plot}{if TRUE add density plot}

\item{scatter.plot}{if TRUE add scatter plot}

\item{running.mean}{if TRUE add running average estimate to scatter plot}

\item{add}{if TRUE add to existing plot}

\item{...}{additional arguments to lower level functions}
}
\description{
Density and scatter plots
}
\examples{
n <- 1000
val <- cbind(est1=rnorm(n,sd=1),est2=rnorm(n,sd=0.2),est3=rnorm(n,1,sd=0.5),
             sd1=runif(n,0.8,1.2),sd2=runif(n,0.1,0.3),sd3=runif(n,0.25,0.75))

plot.sim(val,estimate=c(1,2),true=c(0,0),se=c(4,5),equal=TRUE,scatter.plot=TRUE)
plot.sim(val,estimate=c(1,3),true=c(0,1),se=c(4,6),xlim=c(-3,3),
	scatter.ylim=c(-3,3),scatter.plot=TRUE)
plot.sim(val,estimate=c(1,2),true=c(0,0),se=c(4,5),equal=TRUE,
	plot.type="single",scatter.plot=TRUE)
plot.sim(val,estimate=c(1),se=c(4,5,6),plot.type="single",scatter.plot=TRUE)
plot.sim(val,estimate=c(1,2,3),equal=TRUE,scatter.plot=TRUE)
plot.sim(val,estimate=c(1,2,3),equal=TRUE,byrow=TRUE,scatter.plot=TRUE)
plot.sim(val,estimate=c(1,2,3),plot.type="single",scatter.plot=TRUE)
plot.sim(val,estimate=1,se=c(3,4,5),plot.type="single",scatter.plot=TRUE)

density.sim(val,estimate=c(1,2,3),density=c(0,10,10), lwd=2, angle=c(0,45,-45),cex.legend=1.3)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.sim.R
\name{plot.sim}
\alias{plot.sim}
\alias{density.sim}
\title{Plot method for simulation 'sim' objects}
\usage{
\method{plot}{sim}(
  x,
  estimate,
  se = NULL,
  true = NULL,
  names = NULL,
  auto.layout = TRUE,
  byrow = FALSE,
  type = "p",
  ask = grDevices::dev.interactive(),
  col = c("gray60", "orange", "darkblue", "seagreen", "darkred"),
  pch = 16,
  cex = 0.5,
  lty = 1,
  lwd = 0.3,
  legend,
  legendpos = "topleft",
  cex.legend = 0.8,
  plot.type = c("multiple", "single"),
  polygon = TRUE,
  density = NULL,
  angle = -45,
  cex.axis = 0.8,
  alpha = 0.2,
  main,
  cex.main = 1,
  equal = FALSE,
  delta = 1.15,
  ylim = NULL,
  xlim = NULL,
  ylab = "",
  xlab = "",
  rug = FALSE,
  rug.alpha = 0.5,
  line.col = scatter.col,
  line.lwd = 1,
  line.lty = 1,
  line.alpha = 1,
  scatter.ylab = "Estimate",
  scatter.ylim = NULL,
  scatter.xlim = NULL,
  scatter.alpha = 0.5,
  scatter.col = col,
  border = col,
  true.lty = 2,
  true.col = "gray70",
  true.lwd = 1.2,
  density.plot = TRUE,
  scatter.plot = FALSE,
  running.mean = scatter.plot,
  ...
)
}
\arguments{
\item{x}{sim object}

\item{estimate}{columns with estimates}

\item{se}{columns with standard error estimates}

\item{true}{(optional) vector of true parameter values}

\item{names}{(optional) names of estimates}

\item{auto.layout}{Auto layout (default TRUE)}

\item{byrow}{Add new plots to layout by row}

\item{type}{plot type}

\item{ask}{if TRUE user is asked for input, before a new figure is drawn}

\item{col}{colour (for each estimate)}

\item{pch}{plot symbol}

\item{cex}{point size}

\item{lty}{line type}

\item{lwd}{line width}

\item{legend}{legend}

\item{legendpos}{legend position}

\item{cex.legend}{size of legend text}

\item{plot.type}{'single' or 'multiple' (default)}

\item{polygon}{if TRUE fill the density estimates with colour}

\item{density}{if non-zero add shading lines to polygon}

\item{angle}{shading lines angle of polygon}

\item{cex.axis}{Font size on axis}

\item{alpha}{Semi-transparent level (1: non-transparent, 0: full)}

\item{main}{Main title}

\item{cex.main}{Size of title font}

\item{equal}{Same x-axis and y-axis for all plots}

\item{delta}{Controls the amount of space around axis limits}

\item{ylim}{y-axis limits}

\item{xlim}{x-axis limits}

\item{ylab}{y axis label}

\item{xlab}{x axis label}

\item{rug}{if TRUE add rug representation of data to x-axis}

\item{rug.alpha}{rug semi-transparency level}

\item{line.col}{line colour (running mean, only for scatter plots)}

\item{line.lwd}{line width (running mean, only for scatter plots)}

\item{line.lty}{line type (running mean, only for scatter plots)}

\item{line.alpha}{line transparency}

\item{scatter.ylab}{y label for density plots}

\item{scatter.ylim}{y-axis limits for density plots}

\item{scatter.xlim}{x-axis limits for density plots}

\item{scatter.alpha}{semi-transparency of scatter plot}

\item{scatter.col}{scatter plot colour}

\item{border}{border colour of density estimates}

\item{true.lty}{true parameter estimate line type}

\item{true.col}{true parameter colour}

\item{true.lwd}{true parameter line width}

\item{density.plot}{if TRUE add density plot}

\item{scatter.plot}{if TRUE add scatter plot}

\item{running.mean}{if TRUE add running average estimate to scatter plot}

\item{...}{additional arguments to lower level functions}
}
\description{
Density and scatter plots
}
\examples{
n <- 1000
val <- cbind(est1=rnorm(n,sd=1),est2=rnorm(n,sd=0.2),est3=rnorm(n,1,sd=0.5),
             sd1=runif(n,0.8,1.2),sd2=runif(n,0.1,0.3),sd3=runif(n,0.25,0.75))

plot.sim(val,estimate=c(1,2),true=c(0,0),se=c(4,5),equal=TRUE,scatter.plot=TRUE)
plot.sim(val,estimate=c(1,3),true=c(0,1),se=c(4,6),xlim=c(-3,3),
	scatter.ylim=c(-3,3),scatter.plot=TRUE)
plot.sim(val,estimate=c(1,2),true=c(0,0),se=c(4,5),equal=TRUE,
	plot.type="single",scatter.plot=TRUE)
plot.sim(val,estimate=c(1),se=c(4,5,6),plot.type="single",scatter.plot=TRUE)
plot.sim(val,estimate=c(1,2,3),equal=TRUE,scatter.plot=TRUE)
plot.sim(val,estimate=c(1,2,3),equal=TRUE,byrow=TRUE,scatter.plot=TRUE)
plot.sim(val,estimate=c(1,2,3),plot.type="single",scatter.plot=TRUE)
plot.sim(val,estimate=1,se=c(3,4,5),plot.type="single",scatter.plot=TRUE)

density.sim(val,estimate=c(1,2,3),density=c(0,10,10), lwd=2, angle=c(0,45,-45),cex.legend=1.3)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sim.default.R
\name{sim.default}
\alias{sim.default}
\alias{sim}
\alias{as.sim}
\title{Monte Carlo simulation}
\usage{
\method{sim}{default}(
  x = NULL,
  R = 100,
  f = NULL,
  colnames = NULL,
  seed = NULL,
  args = list(),
  iter = FALSE,
  mc.cores,
  progressr.message = NULL,
  estimate.index = 1:2,
  ...
)
}
\arguments{
\item{x}{function or 'sim' object}

\item{R}{Number of replications or data.frame with parameters}

\item{f}{Optional function (i.e., if x is a matrix)}

\item{colnames}{Optional column names}

\item{seed}{(optional) Seed (needed with cl=TRUE)}

\item{args}{(optional) list of named arguments passed to (mc)mapply}

\item{iter}{If TRUE the iteration number is passed as first argument to
(mc)mapply}

\item{mc.cores}{Optional number of cores. Will use parallel::mcmapply
instead of future}

\item{progressr.message}{Optional message for the progressr progress-bar}

\item{estimate.index}{If return object inherits from \code{estimate} then only
these column indices are extracted (estimate, se, lower, upper, p-val)}

\item{...}{Additional arguments to \code{\link[future.apply:future_mapply]{future.apply::future_mapply()}}}
}
\description{
Applies a function repeatedly for a specified number of replications or over
a list/data.frame with plot and summary methods for summarizing the Monte
Carlo experiment. Can be parallelized via the future package (use the
\code{\link[future:plan]{future::plan()}} function).
}
\details{
To parallelize the calculation use the \code{\link[future:plan]{future::plan()}} function
(e.g., \code{future::plan(multisession())} to distribute the calculations over
the \code{R} replications on all available cores). The output is controlled via
the progressr package (e.g., \code{progressr::handlers(global=TRUE)} to enable
progress information).
}
\examples{
m <- lvm(y~x+e)
distribution(m,~y) <- 0
distribution(m,~x) <- dist_uniform(a=-1.1,b=1.1)
transform(m,e~x) <- function(x) (1*x^4)*rnorm(length(x),sd=1)

onerun <- function(iter=NULL,...,n=2e3,b0=1,idx=2) {
    d <- sim(m,n,p=c("y~x"=b0))
    l <- lm(y~x,d)
    res <- c(coef(summary(l))[idx,1:2],
             confint(l)[idx,],
             coef(estimate(l), mat=TRUE)[idx,2:4])
    names(res) <- c("Estimate","Model.se","Model.lo","Model.hi",
                    "Sandwich.se","Sandwich.lo","Sandwich.hi")
    res
}
val <- sim(onerun,R=10,b0=1)
val

val <- sim(val,R=40,b0=1) ## append results
summary(val,estimate=c(1,1),confint=c(3,4,6,7),true=c(1,1))

summary(val,estimate=c(1,1),se=c(2,5),names=c("Model","Sandwich"))
summary(val,estimate=c(1,1),se=c(2,5),true=c(1,1),
        names=c("Model","Sandwich"),confint=TRUE)

if (interactive()) {
    plot(val,estimate=1,c(2,5),true=1,
         names=c("Model","Sandwich"),polygon=FALSE)
    plot(val,estimate=c(1,1),se=c(2,5),main=NULL,
         true=c(1,1),names=c("Model","Sandwich"),
         line.lwd=1,col=c("gray20","gray60"),
         rug=FALSE)
    plot(val,estimate=c(1,1),se=c(2,5),true=c(1,1),
         names=c("Model","Sandwich"))
}

f <- function(a=1, b=1) {
  rep(a*b, 5)
}
R <- Expand(a=1:3, b=1:3)
sim(f, R)
sim(function(a,b) f(a,b), 3, args=c(a=5,b=5))
sim(function(iter=1,a=5,b=5) iter*f(a,b), iter=TRUE, R=5)

## Returning estimate objects with extra per-iteration information
onerun2 <- function(...) {
  x <- rnorm(50)
  y <- 0.5*x + rnorm(50)
  e <- estimate(lm(y ~ x))
  c(e, converged = 1, niter = sample(5:20, 1), drop.ic = TRUE)
  # works identically with summary.estimate objects
  # c(summary(e), converged = 1, niter = sample(5:20, 1), drop.ic = TRUE)
}
val2 <- sim(onerun2, R = 10)
val2
# extra columns are stored but excluded from default summary statistics
idx <- c("(Intercept)", "x", "niter")
summary(val2, estimate=idx, true=c(0,0.5,NA))
}
\seealso{
\code{\link[=summary.sim]{summary.sim()}} \code{\link[=plot.sim]{plot.sim()}} \code{\link[=sim.lvm]{sim.lvm()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/covariance.R
\name{covariance}
\alias{covariance}
\alias{covariance<-}
\alias{covariance.lvm}
\alias{covariance<-.lvm}
\alias{variance}
\alias{variance<-}
\alias{variance.lvm}
\alias{variance<-.lvm}
\title{Add covariance structure to Latent Variable Model}
\usage{
\method{covariance}{lvm}(object, var1=NULL, var2=NULL, constrain=FALSE, pairwise=FALSE,...) <- value
}
\arguments{
\item{object}{\code{lvm}-object}

\item{\dots}{Additional arguments to be passed to the low level functions}

\item{var1}{Vector of variables names (or formula)}

\item{var2}{Vector of variables names (or formula) defining pairwise
covariance between \code{var1} and \code{var2})}

\item{constrain}{Define non-linear parameter constraints to ensure positive definite structure}

\item{pairwise}{If TRUE and \code{var2} is omitted then pairwise correlation is added between all variables in \code{var1}}

\item{value}{List of parameter values or (if \code{var1} is unspecified)}
}
\value{
A \code{lvm}-object
}
\description{
Define covariances between residual terms in a \code{lvm}-object.
}
\details{
The \code{covariance} function is used to specify correlation structure
between residual terms of a latent variable model, using a formula syntax.

For instance, a multivariate model with three response variables,

\deqn{Y_1 = \mu_1 + \epsilon_1}

\deqn{Y_2 = \mu_2 + \epsilon_2}

\deqn{Y_3 = \mu_3 + \epsilon_3}

can be specified as

\code{m <- lvm(~y1+y2+y3)}

Pr. default the two variables are assumed to be independent. To add a
covariance parameter \eqn{r = cov(\epsilon_1,\epsilon_2)}, we execute the
following code

\code{covariance(m) <- y1 ~ f(y2,r)}

The special function \code{f} and its second argument could be omitted thus
assigning an unique parameter the covariance between \code{y1} and
\code{y2}.

Similarily the marginal variance of the two response variables can be fixed
to be identical (\eqn{var(Y_i)=v}) via

\code{covariance(m) <- c(y1,y2,y3) ~ f(v)}

To specify a completely unstructured covariance structure, we can call

\code{covariance(m) <- ~y1+y2+y3}

All the parameter values of the linear constraints can be given as the right
handside expression of the assigment function \code{covariance<-} if the
first (and possibly second) argument is defined as well. E.g:

\code{covariance(m,y1~y1+y2) <- list("a1","b1")}

\code{covariance(m,~y2+y3) <- list("a2",2)}

Defines

\deqn{var(\epsilon_1) = a1}

\deqn{var(\epsilon_2) = a2}

\deqn{var(\epsilon_3) = 2}

\deqn{cov(\epsilon_1,\epsilon_2) = b1}

Parameter constraints can be cleared by fixing the relevant parameters to
\code{NA} (see also the \code{regression} method).

The function \code{covariance} (called without additional arguments) can be
used to inspect the covariance constraints of a \code{lvm}-object.
}
\examples{

m <- lvm()
### Define covariance between residuals terms of y1 and y2
covariance(m) <- y1~y2
covariance(m) <- c(y1,y2)~f(v) ## Same marginal variance
covariance(m) ## Examine covariance structure


}
\seealso{
\code{\link{regression<-}}, \code{\link{intercept<-}},
\code{\link{constrain<-}} \code{\link{parameter<-}}, \code{\link{latent<-}},
\code{\link{cancel<-}}, \code{\link{kill<-}}
}
\author{
Klaus K. Holst
}
\keyword{models}
\keyword{regression}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stack.R
\name{stack.estimate}
\alias{stack.estimate}
\title{Stack estimating equations}
\usage{
\method{stack}{estimate}(
  x,
  model2,
  D1u,
  inv.D2u,
  propensity,
  dpropensity,
  U,
  keep1 = FALSE,
  propensity.arg,
  estimate.arg,
  na.action = na.pass,
  ...
)
}
\arguments{
\item{x}{Model 1}

\item{model2}{Model 2}

\item{D1u}{Derivative of score of model 2 w.r.t. parameter vector of model 1}

\item{inv.D2u}{Inverse of deri}

\item{propensity}{propensity score (vector or function)}

\item{dpropensity}{derivative of propensity score wrt parameters of model 1}

\item{U}{Optional score function (model 2) as function of all parameters}

\item{keep1}{If FALSE only parameters of model 2 is returned}

\item{propensity.arg}{Arguments to propensity function}

\item{estimate.arg}{Arguments to 'estimate'}

\item{na.action}{Method for dealing with missing data in propensity score}

\item{...}{Additional arguments to lower level functions}
}
\description{
Stack estimating equations (two-stage estimator)
}
\examples{
m <- lvm(z0~x)
Missing(m, z ~ z0) <- r~x
distribution(m,~x) <- binomial.lvm()
p <- c(r=-1,'r~x'=0.5,'z0~x'=2)
beta <- p[3]/2
d <- sim(m,500,p=p,seed=1)
m1 <- estimate(r~x,data=d,family=binomial)
d$w <- d$r/predict(m1,type="response")
m2 <- estimate(z~1, weights=w, data=d)
(e <- stack(m1,m2,propensity=TRUE))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/path.R
\name{path}
\alias{path}
\alias{effects}
\alias{path.lvm}
\alias{effects.lvmfit}
\alias{totaleffects}
\title{Extract pathways in model graph}
\usage{
\method{path}{lvm} (object, to = NULL, from, all=FALSE, ...)
\method{effects}{lvmfit} (object, to, from, ...)
}
\arguments{
\item{object}{Model object (\code{lvm})}

\item{\dots}{Additional arguments to be passed to the low level functions}

\item{to}{Outcome variable (string). Alternatively a formula specifying
response and predictor in which case the argument \code{from} is ignored.}

\item{from}{Response variable (string), not necessarily directly affected by
\code{to}.}

\item{all}{If TRUE all simple paths (in undirected graph) is returned
on/off.}
}
\value{
If \code{object} is of class \code{lvmfit} a list with the following
elements is returned \item{idx}{ A list where each element defines a
possible pathway via a integer vector indicating the index of the visited
nodes. } \item{V }{ A List of covariance matrices for each path. }
\item{coef }{A list of parameters estimates for each path} \item{path }{A
list where each element defines a possible pathway via a character vector
naming the visited nodes in order.  } \item{edges }{Description of 'comp2'}

If \code{object} is of class \code{lvm} only the \code{path} element will be
returned.

The \code{effects} method returns an object of class \code{effects}.
}
\description{
Extract all possible paths from one variable to another connected component
in a latent variable model. In an estimated model the effect size is
decomposed into direct, indirect and total effects including approximate
standard errors.
}
\note{
For a \code{lvmfit}-object the parameters estimates and their
corresponding covariance matrix are also returned.  The
\code{effects}-function additionally calculates the total and indirect
effects with approximate standard errors
}
\examples{

m <- lvm(c(y1,y2,y3)~eta)
regression(m) <- y2~x1
latent(m) <- ~eta
regression(m) <- eta~x1+x2
d <- sim(m,500)
e <- estimate(m,d)

path(Model(e),y2~x1)
parents(Model(e), ~y2)
children(Model(e), ~x2)
children(Model(e), ~x2+eta)
effects(e,y2~x1)
## All simple paths (undirected)
path(m,y1~x1,all=TRUE)

}
\seealso{
\code{children}, \code{parents}
}
\author{
Klaus K. Holst
}
\keyword{graphs}
\keyword{methods}
\keyword{models}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sim.default.R
\name{Print}
\alias{Print}
\title{Generic print method}
\usage{
Print(x, n = 5, digits = max(3, getOption("digits") - 3), ...)
}
\arguments{
\item{x}{object to print}

\item{n}{number of rows to show from top and bottom of tabular data}

\item{digits}{precision}

\item{...}{additional arguments to print method}
}
\description{
Nicer print method for tabular data. Falls back to standard print method for
all other data types.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sim.lvm.R
\name{sim.lvm}
\alias{sim.lvm}
\alias{sim.lvmfit}
\alias{simulate.lvmfit}
\alias{simulate.lvm}
\alias{transform<-}
\alias{transform<-.lvm}
\alias{transform.lvm}
\alias{functional}
\alias{functional<-}
\alias{functional.lvm}
\alias{functional<-.lvm}
\alias{distribution}
\alias{distribution<-}
\alias{distribution.lvm}
\alias{distribution<-.lvm}
\alias{heavytail}
\alias{heavytail<-}
\alias{weibull.lvm}
\alias{binomial.lvm}
\alias{poisson.lvm}
\alias{uniform.lvm}
\alias{multinomial.lvm}
\alias{beta.lvm}
\alias{normal.lvm}
\alias{mvn.lvm}
\alias{lognormal.lvm}
\alias{gaussian.lvm}
\alias{GM2.lvm}
\alias{GM3.lvm}
\alias{probit.lvm}
\alias{logit.lvm}
\alias{pareto.lvm}
\alias{student.lvm}
\alias{chisq.lvm}
\alias{coxGompertz.lvm}
\alias{coxWeibull.lvm}
\alias{coxExponential.lvm}
\alias{aalenExponential.lvm}
\alias{Gamma.lvm}
\alias{gamma.lvm}
\alias{loggamma.lvm}
\alias{categorical}
\alias{categorical<-}
\alias{threshold.lvm}
\alias{ones.lvm}
\alias{Binary.lvm}
\alias{binary.lvm}
\alias{Sequence.lvm}
\alias{none.lvm}
\alias{constant.lvm}
\alias{id.lvm}
\title{Simulate model}
\usage{
\method{sim}{lvm}(x, n = NULL, p = NULL, normal = FALSE, cond = FALSE,
sigma = 1, rho = 0.5, X = NULL, unlink=FALSE, latent=TRUE,
use.labels = TRUE, seed=NULL, ...)
}
\arguments{
\item{x}{Model object}

\item{n}{Number of simulated values/individuals}

\item{p}{Parameter value (optional)}

\item{normal}{Logical indicating whether to simulate data from a
multivariate normal distribution conditional on exogenous variables hence
ignoring functional/distribution definition}

\item{cond}{for internal use}

\item{sigma}{Default residual variance (1)}

\item{rho}{Default covariance parameter (0.5)}

\item{X}{Optional matrix of fixed values of variables (manipulation)}

\item{unlink}{Return Inverse link transformed data}

\item{latent}{Include latent variables (default TRUE)}

\item{use.labels}{convert categorical variables to factors before applying transformation}

\item{seed}{Random seed}

\item{\dots}{Additional arguments to be passed to the low level functions}
}
\description{
Simulate data from a general SEM model including non-linear effects and
general link and distribution of variables.
}
\examples{
##################################################
## Logistic regression
##################################################
m <- lvm(y~x+z)
regression(m) <- x~z
distribution(m,~y+z) <- binomial.lvm("logit")
d <- sim(m,1e3)
head(d)

e <- estimate(m,d,estimator="glm")
e
## Simulate a few observation from estimated model
sim(e,n=5)

##################################################
## Poisson
##################################################
distribution(m,~y) <- poisson.lvm()
d <- sim(m,1e4,p=c(y=-1,"y~x"=2,z=1))
head(d)
estimate(m,d,estimator="glm")
mean(d$z); lava:::expit(1)

summary(lm(y~x,sim(lvm(y[1:2]~4*x),1e3)))

##################################################
### Gamma distribution
##################################################
m <- lvm(y~x)
distribution(m,~y+x) <- list(Gamma.lvm(shape=2),binomial.lvm())
intercept(m,~y) <- 0.5
d <- sim(m,1e4)
summary(g <- glm(y~x,family=Gamma(),data=d))
\dontrun{MASS::gamma.shape(g)}

args(lava::Gamma.lvm)
distribution(m,~y) <- Gamma.lvm(shape=2,log=TRUE)
sim(m,10,p=c(y=0.5))[,"y"]

##################################################
### Beta
##################################################
m <- lvm()
distribution(m,~y) <- beta.lvm(alpha=2,beta=1)
var(sim(m,100,"y,y"=2))
distribution(m,~y) <- beta.lvm(alpha=2,beta=1,scale=FALSE)
var(sim(m,100))

##################################################
### Transform
##################################################
m <- lvm()
transform(m,xz~x+z) <- function(x) x[1]*(x[2]>0)
regression(m) <- y~x+z+xz
d <- sim(m,1e3)
summary(lm(y~x+z + x*I(z>0),d))

##################################################
### Non-random variables
##################################################
m <- lvm()
distribution(m,~x+z+v+w) <- list(Sequence.lvm(0,5),## Seq. 0 to 5 by 1/n
                               Binary.lvm(),       ## Vector of ones
                               Binary.lvm(0.5),    ##  0.5n 0, 0.5n 1
                               Binary.lvm(interval=list(c(0.3,0.5),c(0.8,1))))
sim(m,10)

##################################################
### Cox model
### piecewise constant hazard
################################################
m <- lvm(t~x)
rates <- c(1,0.5); cuts <- c(0,5)
## Constant rate: 1 in [0,5), 0.5 in [5,Inf)
distribution(m,~t) <- coxExponential.lvm(rate=rates,timecut=cuts)

\dontrun{
    d <- sim(m,2e4,p=c("t~x"=0.1)); d$status <- TRUE
    plot(timereg::aalen(survival::Surv(t,status)~x,data=d,
                        resample.iid=0,robust=0),spec=1)
    L <- approxfun(c(cuts,max(d$t)),f=1,
                   cumsum(c(0,rates*diff(c(cuts,max(d$t))))),
                   method="linear")
    curve(L,0,100,add=TRUE,col="blue")
}

##################################################
### Cox model
### piecewise constant hazard, gamma frailty
##################################################
m <- lvm(y~x+z)
rates <- c(0.3,0.5); cuts <- c(0,5)
distribution(m,~y+z) <- list(coxExponential.lvm(rate=rates,timecut=cuts),
                             loggamma.lvm(rate=1,shape=1))
\dontrun{
    d <- sim(m,2e4,p=c("y~x"=0,"y~z"=0)); d$status <- TRUE
    plot(timereg::aalen(survival::Surv(y,status)~x,data=d,
                        resample.iid=0,robust=0),spec=1)
    L <- approxfun(c(cuts,max(d$y)),f=1,
                   cumsum(c(0,rates*diff(c(cuts,max(d$y))))),
                   method="linear")
    curve(L,0,100,add=TRUE,col="blue")
}
## Equivalent via transform (here with Aalens additive hazard model)
m <- lvm(y~x)
distribution(m,~y) <- aalenExponential.lvm(rate=rates,timecut=cuts)
distribution(m,~z) <- Gamma.lvm(rate=1,shape=1)
transform(m,t~y+z) <- prod
sim(m,10)
## Shared frailty
m <- lvm(c(t1,t2)~x+z)
rates <- c(1,0.5); cuts <- c(0,5)
distribution(m,~y) <- aalenExponential.lvm(rate=rates,timecut=cuts)
distribution(m,~z) <- loggamma.lvm(rate=1,shape=1)
\dontrun{
mets::fast.reshape(sim(m,100),varying="t")
}

##################################################
### General multivariate distributions
##################################################
\dontrun{
m <- lvm()
distribution(m,~y1+y2,oratio=4) <- VGAM::rbiplackcop
ksmooth2(sim(m,1e4),rgl=FALSE,theta=-20,phi=25)

m <- lvm()
distribution(m,~z1+z2,"or1") <- VGAM::rbiplackcop
distribution(m,~y1+y2,"or2") <- VGAM::rbiplackcop
sim(m,10,p=c(or1=0.1,or2=4))
}

m <- lvm()
distribution(m,~y1+y2+y3,TRUE) <- function(n,...) rmvn0(n,sigma=diag(3)+1)
var(sim(m,100))

## Syntax also useful for univariate generators, e.g.
m <- lvm(y~x+z)
distribution(m,~y,TRUE) <- function(n) rnorm(n,mean=1000)
sim(m,5)
distribution(m,~y,"m1",0) <- rnorm
sim(m,5)
sim(m,5,p=c(m1=100))

##################################################
### Regression design in other parameters
##################################################
## Variance heterogeneity
m <- lvm(y~x)
distribution(m,~y) <- function(n,mean,x) rnorm(n,mean,exp(x)^.5)
if (interactive()) plot(y~x,sim(m,1e3))
## Alternaively, calculate the standard error directly
addvar(m) <- ~sd ## If 'sd' should be part of the resulting data.frame
constrain(m,sd~x) <- function(x) exp(x)^.5
distribution(m,~y) <- function(n,mean,sd) rnorm(n,mean,sd)
if (interactive()) plot(y~x,sim(m,1e3))

## Regression on variance parameter
m <- lvm()
regression(m) <- y~x
regression(m) <- v~x
##distribution(m,~v) <- 0 # No stochastic term
## Alternative:
## regression(m) <- v[NA:0]~x
distribution(m,~y) <- function(n,mean,v) rnorm(n,mean,exp(v)^.5)
if (interactive()) plot(y~x,sim(m,1e3))

## Regression on shape parameter in Weibull model
m <- lvm()
regression(m) <- y ~ z+v
regression(m) <- s ~ exp(0.6*x-0.5*z)
distribution(m,~x+z) <- binomial.lvm()
distribution(m,~cens) <- coxWeibull.lvm(scale=1)
distribution(m,~y) <- coxWeibull.lvm(scale=0.1,shape=~s)
eventTime(m) <- time ~ min(y=1,cens=0)

if (interactive()) {
    d <- sim(m,1e3)
    require(survival)
    (cc <- coxph(Surv(time,status)~v+strata(x,z),data=d))
    plot(survfit(cc) ,col=1:4,mark.time=FALSE)
}

##################################################
### Categorical predictor
##################################################
m <- lvm()
## categorical(m,K=3) <- "v"
categorical(m,labels=c("A","B","C")) <- "v"

regression(m,additive=FALSE) <- y~v
\dontrun{
plot(y~v,sim(m,1000,p=c("y~v:2"=3)))
}

m <- lvm()
categorical(m,labels=c("A","B","C"),p=c(0.5,0.3)) <- "v"
regression(m,additive=FALSE,beta=c(0,2,-1)) <- y~v
## equivalent to:
## regression(m,y~v,additive=FALSE) <- c(0,2,-1)
regression(m,additive=FALSE,beta=c(0,4,-1)) <- z~v
table(sim(m,1e4)$v)
glm(y~v, data=sim(m,1e4))
glm(y~v, data=sim(m,1e4,p=c("y~v:1"=3)))

transform(m,v2~v) <- function(x) x=='A'
sim(m,10)

##################################################
### Pre-calculate object
##################################################
m <- lvm(y~x)
m2 <- sim(m,'y~x'=2)
sim(m,10,'y~x'=2)
sim(m2,10) ## Faster
}
\author{
Klaus K. Holst
}
\keyword{datagen}
\keyword{models}
\keyword{regression}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/multipletesting.R
\name{closed.testing}
\alias{closed.testing}
\alias{p.correct}
\title{Closed testing procedure}
\usage{
closed.testing(
  object,
  idx = seq_along(coef(object)),
  null = rep(0, length(idx)),
  ...
)
}
\arguments{
\item{object}{estimate object}

\item{idx}{Index of parameters to adjust for multiple testing}

\item{null}{Null hypothesis value}

\item{...}{Additional arguments}
}
\description{
Closed testing procedure
}
\examples{
m <- lvm()
regression(m, c(y1,y2,y3,y4,y5,y6,y7)~x) <- c(0,0.25,0,0.25,0.25,0,0)
regression(m, to=endogenous(m), from="u") <- 1
variance(m,endogenous(m)) <- 1
set.seed(2)
d <- sim(m,200)
l1 <- lm(y1~x,d)
l2 <- lm(y2~x,d)
l3 <- lm(y3~x,d)
l4 <- lm(y4~x,d)
l5 <- lm(y5~x,d)
l6 <- lm(y6~x,d)
l7 <- lm(y7~x,d)

(a <- merge(l1,l2,l3,l4,l5,l6,l7,subset=2))
if (requireNamespace("mets",quietly=TRUE)) {
   p.correct(a)
}
as.vector(closed.testing(a))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Grep.R
\name{Grep}
\alias{Grep}
\title{Finds elements in vector or column-names in data.frame/matrix}
\usage{
Grep(x, pattern, subset = TRUE, ignore.case = TRUE, ...)
}
\arguments{
\item{x}{vector, matrix or data.frame.}

\item{pattern}{regular expression to search for}

\item{subset}{If TRUE returns subset of data.frame/matrix otherwise just the matching column names}

\item{ignore.case}{Default ignore case}

\item{...}{Additional arguments to 'grep'}
}
\value{
A data.frame with 2 columns with the indices in the first and the
matching names in the second.
}
\description{
Pattern matching in a vector or column names of a data.frame or matrix.
}
\examples{
data(iris)
head(Grep(iris,"(len)|(sp)"))
}
\seealso{
\code{\link{grep}}, and \code{\link{agrep}} for approximate string
matching,
}
\author{
Klaus K. Holst
}
\keyword{misc}
\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stack.R
\name{stack.estimate}
\alias{stack.estimate}
\title{Stack estimating equations}
\usage{
\method{stack}{estimate}(
  x,
  model2,
  D1u,
  inv.D2u,
  propensity,
  dpropensity,
  U,
  keep1 = FALSE,
  propensity.arg,
  estimate.arg,
  na.action = na.pass,
  ...
)
}
\arguments{
\item{x}{Model 1}

\item{model2}{Model 2}

\item{D1u}{Derivative of score of model 2 w.r.t. parameter vector of model 1}

\item{inv.D2u}{Inverse of deri}

\item{propensity}{propensity score (vector or function)}

\item{dpropensity}{derivative of propensity score wrt parameters of model 1}

\item{U}{Optional score function (model 2) as function of all parameters}

\item{keep1}{If FALSE only parameters of model 2 is returned}

\item{propensity.arg}{Arguments to propensity function}

\item{estimate.arg}{Arguments to 'estimate'}

\item{na.action}{Method for dealing with missing data in propensity score}

\item{...}{Additional arguments to lower level functions}
}
\value{
Object of class \code{estimate} (see \link{estimate.default}) with
inference adjusted for the two-stage estimation.
}
\description{
Stack estimating equations (two-stage estimator)
}
\examples{
m <- lvm(z0~x)
Missing(m, z ~ z0) <- r~x
distribution(m,~x) <- dist_bernoulli()
p <- c(r=-1,'r~x'=0.5,'z0~x'=2)
beta <- p[3]/2
d <- sim(m,500,p=p,seed=1)
m1 <- estimate(r~x,data=d,family=binomial)
d$w <- d$r/predict(m1,type="response")
m2 <- estimate(z~1, weights=w, data=d)
(e <- stack(m1,m2,propensity=TRUE))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sim.lvm.R
\name{sim.lvm}
\alias{sim.lvm}
\alias{sim.lvmfit}
\alias{simulate.lvmfit}
\alias{simulate.lvm}
\alias{transform<-}
\alias{transform<-.lvm}
\alias{transform.lvm}
\alias{functional}
\alias{functional<-}
\alias{functional.lvm}
\alias{functional<-.lvm}
\alias{distribution}
\alias{distribution<-}
\alias{distribution.lvm}
\alias{distribution<-.lvm}
\alias{heavytail}
\alias{heavytail<-}
\alias{categorical}
\alias{categorical<-}
\alias{dist_weibull}
\alias{weibull.lvm}
\alias{dist_bernoulli}
\alias{binomial.lvm}
\alias{probit.lvm}
\alias{logit.lvm}
\alias{dist_poisson}
\alias{poisson.lvm}
\alias{dist_uniform}
\alias{uniform.lvm}
\alias{dist_multinomial}
\alias{multinomial.lvm}
\alias{dist_beta}
\alias{beta.lvm}
\alias{dist_gaussian}
\alias{normal.lvm}
\alias{gaussian.lvm}
\alias{dist_mvn}
\alias{mvn.lvm}
\alias{dist_lognormal}
\alias{lognormal.lvm}
\alias{dist_gaussian_mixture2}
\alias{GM2.lvm}
\alias{dist_gaussian_mixture3}
\alias{GM3.lvm}
\alias{dist_pareto}
\alias{pareto.lvm}
\alias{dist_t}
\alias{student.lvm}
\alias{dist_chisq}
\alias{chisq.lvm}
\alias{dist_cox_gompertz}
\alias{coxGompertz.lvm}
\alias{dist_cox_weibull}
\alias{coxWeibull.lvm}
\alias{dist_cox_exponential}
\alias{coxExponential.lvm}
\alias{dist_aalen_exponential}
\alias{aalenExponential.lvm}
\alias{dist_gamma}
\alias{Gamma.lvm}
\alias{gamma.lvm}
\alias{dist_loggamma}
\alias{loggamma.lvm}
\alias{dist_threshold}
\alias{threshold.lvm}
\alias{dist_ones}
\alias{ones.lvm}
\alias{binary.lvm}
\alias{dist_seq}
\alias{Sequence.lvm}
\alias{dist_seqint}
\alias{id.lvm}
\alias{dist_none}
\alias{none.lvm}
\alias{dist_const}
\alias{constant.lvm}
\title{Simulate model}
\usage{
\method{sim}{lvm}(x, n = NULL, p = NULL, normal = FALSE, cond = FALSE,
sigma = 1, rho = 0.5, X = NULL, unlink=FALSE, latent=TRUE,
use.labels = TRUE, seed=NULL, ...)
}
\arguments{
\item{x}{Model object}

\item{n}{Number of simulated values/individuals}

\item{p}{Parameter value (optional)}

\item{normal}{Logical indicating whether to simulate data from a
multivariate normal distribution conditional on exogenous variables hence
ignoring functional/distribution definition}

\item{cond}{for internal use}

\item{sigma}{Default residual variance (1)}

\item{rho}{Default covariance parameter (0.5)}

\item{X}{Optional matrix of fixed values of variables (manipulation)}

\item{unlink}{Return Inverse link transformed data}

\item{latent}{Include latent variables (default TRUE)}

\item{use.labels}{convert categorical variables to factors before applying transformation}

\item{seed}{Random seed}

\item{\dots}{Additional arguments to be passed to the low level functions}
}
\description{
Simulate data from a general SEM model including non-linear effects and
general link and distribution of variables.
}
\examples{
##################################################
## Logistic regression
##################################################
m <- lvm(y~x+z)
regression(m) <- x~z
distribution(m,~y+z) <- dist_bernoulli("logit")
d <- sim(m,1e3)
head(d)

e <- estimate(m,d,estimator="glm")
e
## Simulate a few observation from estimated model
sim(e,n=5)

##################################################
## Poisson
##################################################
distribution(m,~y) <- dist_poisson()
d <- sim(m,1e4,p=c(y=-1,"y~x"=2,z=1))
head(d)
estimate(m,d,estimator="glm")
mean(d$z); lava:::expit(1)

summary(lm(y~x,sim(lvm(y[1:2]~4*x),1e3)))

##################################################
### Gamma distribution
##################################################
m <- lvm(y~x)
distribution(m,~y+x) <- list(dist_gamma(shape=2), dist_bernoulli())
intercept(m,~y) <- 0.5
d <- sim(m,1e4)
summary(g <- glm(y~x,family=Gamma(),data=d))
\dontrun{MASS::gamma.shape(g)}

args(lava::dist_gamma)
distribution(m,~y) <- dist_gamma(shape=2,log=TRUE)
sim(m,10,p=c(y=0.5))[,"y"]

##################################################
### Beta
##################################################
m <- lvm()
distribution(m,~y) <- dist_beta(alpha=2,beta=1)
var(sim(m,100,"y,y"=2))
distribution(m,~y) <- dist_beta(alpha=2,beta=1,scale=FALSE)
var(sim(m,100))

##################################################
### Transform
##################################################
m <- lvm()
transform(m,xz~x+z) <- function(x) x[1]*(x[2]>0)
regression(m) <- y~x+z+xz
d <- sim(m,1e3)
summary(lm(y~x+z + x*I(z>0),d))

##################################################
### Non-random variables
##################################################
m <- lvm()
distribution(m,~x+z+v+w) <- list(
  dist_seqint(0,5),## Seq. 0 to 5 by 1/n
  dist_ones(),       ## Vector of ones
  dist_ones(0.5),    ##  0.5n 0, 0.5n 1
  dist_ones(interval=list(c(0.3,0.5),c(0.8,1)))
)
sim(m,10)

##################################################
### Cox model
### piecewise constant hazard
################################################
m <- lvm(t~x)
rates <- c(1,0.5); cuts <- c(0,5)
## Constant rate: 1 in [0,5), 0.5 in [5,Inf)
distribution(m,~t) <- dist_cox_exponential(rate=rates,timecut=cuts)

\dontrun{
    d <- sim(m,2e4,p=c("t~x"=0.1)); d$status <- TRUE
    plot(timereg::aalen(survival::Surv(t,status)~x,data=d,
                        resample.iid=0,robust=0),spec=1)
    L <- approxfun(c(cuts,max(d$t)),f=1,
                   cumsum(c(0,rates*diff(c(cuts,max(d$t))))),
                   method="linear")
    curve(L,0,100,add=TRUE,col="blue")
}

##################################################
### Cox model
### piecewise constant hazard, gamma frailty
##################################################
m <- lvm(y~x+z)
rates <- c(0.3,0.5); cuts <- c(0,5)
distribution(m,~y+z) <- list(dist_cox_exponential(rate=rates,timecut=cuts),
                             dist_loggamma(rate=1,shape=1))
\dontrun{
    d <- sim(m,2e4,p=c("y~x"=0,"y~z"=0)); d$status <- TRUE
    plot(timereg::aalen(survival::Surv(y,status)~x,data=d,
                        resample.iid=0,robust=0),spec=1)
    L <- approxfun(c(cuts,max(d$y)),f=1,
                   cumsum(c(0,rates*diff(c(cuts,max(d$y))))),
                   method="linear")
    curve(L,0,100,add=TRUE,col="blue")
}
## Equivalent via transform (here with Aalens additive hazard model)
m <- lvm(y~x)
distribution(m,~y) <- dist_aalen_exponential(rate=rates, timecut=cuts)
distribution(m,~z) <- dist_gamma(rate=1, shape=1)
transform(m,t~y+z) <- prod
sim(m,10)
## Shared frailty
m <- lvm(c(t1,t2)~x+z)
rates <- c(1,0.5); cuts <- c(0,5)
distribution(m,~y) <- dist_aalen_exponential(rate=rates, timecut=cuts)
distribution(m,~z) <- dist_loggamma(rate=1,shape=1)
\dontrun{
mets::fast.reshape(sim(m,100),varying="t")
}

##################################################
### General multivariate distributions
##################################################
\dontrun{
m <- lvm()
distribution(m,~y1+y2,oratio=4) <- VGAM::rbiplackcop
ksmooth2(sim(m,1e4),rgl=FALSE,theta=-20,phi=25)

m <- lvm()
distribution(m,~z1+z2,"or1") <- VGAM::rbiplackcop
distribution(m,~y1+y2,"or2") <- VGAM::rbiplackcop
sim(m,10,p=c(or1=0.1,or2=4))
}

m <- lvm()
distribution(m,~y1+y2+y3,TRUE) <- function(n,...) rmvn0(n,sigma=diag(3)+1)
var(sim(m,100))

## Syntax also useful for univariate generators, e.g.
m <- lvm(y~x+z)
distribution(m,~y,TRUE) <- function(n) rnorm(n,mean=1000)
sim(m,5)
distribution(m,~y,"m1",0) <- rnorm
sim(m,5)
sim(m,5,p=c(m1=100))

##################################################
### Regression design in other parameters
##################################################
## Variance heterogeneity
m <- lvm(y~x)
distribution(m,~y) <- function(n,mean,x) rnorm(n,mean,exp(x)^.5)
if (interactive()) plot(y~x,sim(m,1e3))
## Alternaively, calculate the standard error directly
addvar(m) <- ~sd ## If 'sd' should be part of the resulting data.frame
constrain(m,sd~x) <- function(x) exp(x)^.5
distribution(m,~y) <- function(n,mean,sd) rnorm(n,mean,sd)
if (interactive()) plot(y~x,sim(m,1e3))

## Regression on variance parameter
m <- lvm()
regression(m) <- y~x
regression(m) <- v~x
##distribution(m,~v) <- 0 # No stochastic term
## Alternative:
## regression(m) <- v[NA:0]~x
distribution(m,~y) <- function(n,mean,v) rnorm(n,mean,exp(v)^.5)
if (interactive()) plot(y~x,sim(m,1e3))

## Regression on shape parameter in Weibull model
m <- lvm()
regression(m) <- y ~ z+v
regression(m) <- s ~ exp(0.6*x-0.5*z)
distribution(m,~x+z) <- dist_bernoulli()
distribution(m,~cens) <- dist_cox_weibull(scale=1)
distribution(m,~y) <- dist_cox_weibull(scale=0.1,shape=~s)
eventTime(m) <- time ~ min(y=1,cens=0)

if (interactive()) {
    d <- sim(m,1e3)
    require(survival)
    (cc <- coxph(Surv(time,status)~v+strata(x,z),data=d))
    plot(survfit(cc) ,col=1:4,mark.time=FALSE)
}

##################################################
### Categorical predictor
##################################################
m <- lvm()
## categorical(m,K=3) <- "v"
categorical(m,labels=c("A","B","C")) <- "v"

regression(m,additive=FALSE) <- y~v
\dontrun{
plot(y~v,sim(m,1000,p=c("y~v:2"=3)))
}

m <- lvm()
categorical(m,labels=c("A","B","C"),p=c(0.5,0.3)) <- "v"
regression(m,additive=FALSE,beta=c(0,2,-1)) <- y~v
## equivalent to:
## regression(m,y~v,additive=FALSE) <- c(0,2,-1)
regression(m,additive=FALSE,beta=c(0,4,-1)) <- z~v
table(sim(m,1e4)$v)
glm(y~v, data=sim(m,1e4))
glm(y~v, data=sim(m,1e4,p=c("y~v:1"=3)))

transform(m,v2~v) <- function(x) x=='A'
sim(m,10)

##################################################
### Pre-calculate object
##################################################
m <- lvm(y~x)
m2 <- sim(m,'y~x'=2)
sim(m,10,'y~x'=2)
sim(m2,10) ## Faster
}
\author{
Klaus K. Holst
}
\keyword{datagen}
\keyword{models}
\keyword{regression}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/trim.R
\name{trim}
\alias{trim}
\title{Trim string of (leading/trailing/all) white spaces}
\usage{
trim(x, all = FALSE, ...)
}
\arguments{
\item{x}{String}

\item{all}{Trim all whitespaces?}

\item{\dots}{additional arguments to lower level functions}
}
\description{
Trim string of (leading/trailing/all) white spaces
}
\author{
Klaus K. Holst
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lava-package.R
\docType{data}
\name{twindata}
\alias{twindata}
\title{Twin menarche data}
\format{
data.frame
}
\source{
Simulated
}
\description{
Simulated data
\tabular{rll}{
id    \tab numeric \tab Twin-pair id \cr
zyg    \tab character \tab Zygosity (MZ or DZ) \cr
twinnum    \tab numeric \tab Twin number (1 or 2) \cr
agemena    \tab numeric \tab Age at menarche (or censoring) \cr
status    \tab logical \tab Censoring status (observed:=T,censored:=F) \cr
bw    \tab numeric  \tab Birth weight \cr
msmoke    \tab numeric \tab Did mother smoke? (yes:=1,no:=0) \cr
}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vec.R
\name{vec}
\alias{vec}
\title{vec operator}
\usage{
vec(x, matrix = FALSE, sep = ".", ...)
}
\arguments{
\item{x}{Array}

\item{matrix}{If TRUE a row vector (matrix) is returned}

\item{sep}{Seperator}

\item{...}{Additional arguments}
}
\description{
vec operator
}
\details{
Convert array into vector
}
\author{
Klaus Holst
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tr.R
\name{tr}
\alias{tr}
\title{Trace operator}
\usage{
tr(x, ...)
}
\arguments{
\item{x}{Square numeric matrix}

\item{\dots}{Additional arguments to lower level functions}
}
\value{
\code{numeric}
}
\description{
Calculates the trace of a square matrix.
}
\examples{

tr(diag(1:5))
}
\seealso{
\code{\link{crossprod}}, \code{\link{tcrossprod}}
}
\author{
Klaus K. Holst
}
\keyword{algebra}
\keyword{math}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/iid.R
\name{iid}
\alias{iid}
\title{Extract i.i.d. decomposition from model object}
\usage{
iid(x, ...)
}
\arguments{
\item{x}{Model object}

\item{...}{Additional arguments (see the man-page of the IC method)}
}
\description{
This function extracts
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/IC.R
\name{IC.default}
\alias{IC.default}
\alias{IC}
\alias{var_ic}
\title{Extract influence function from model object}
\usage{
\method{IC}{default}(x, bread, id = NULL, ...)
}
\arguments{
\item{x}{model object}

\item{bread}{(optional) Inverse of derivative of mean score function}

\item{id}{(optional) id/cluster variable}

\item{...}{additional arguments}
}
\description{
Extract i.i.d. decomposition (influence function) from model object
}
\examples{
m <- lvm(y~x+z)
distribution(m, ~y+z) <- binomial.lvm("logit")
d <- sim(m,1e3)
g <- glm(y~x+z,data=d,family=binomial)
var_ic(IC(g))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sim.default.R
\name{summary.sim}
\alias{summary.sim}
\title{Summary method for 'sim' objects}
\usage{
\method{summary}{sim}(
  object,
  estimate = NULL,
  se = NULL,
  confint,
  true = NULL,
  fun,
  names = NULL,
  unique.names = TRUE,
  minimal = FALSE,
  level = 0.95,
  quantiles = c(0, 0.025, 0.5, 0.975, 1),
  df = Inf,
  ...
)
}
\arguments{
\item{object}{sim object}

\item{estimate}{(optional) columns with estimates}

\item{se}{(optional) columns with standard error estimates}

\item{confint}{(optional) list of pairs of columns with confidence limits}

\item{true}{(optional) vector of true parameter values}

\item{fun}{(optional) summary function}

\item{names}{(optional) names of estimates}

\item{unique.names}{if TRUE, unique.names will be applied to column names}

\item{minimal}{if TRUE, minimal summary will be returned}

\item{level}{confidence level (0.95)}

\item{quantiles}{quantiles (0,0.025,0.5,0.975,1)}

\item{df}{degrees of freedom in t-distribution used for constructing CIs
(default Gaussian approximation)}

\item{...}{additional levels to lower-level functions}
}
\description{
Summary method for 'sim' objects
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/score.R
\name{score}
\alias{score}
\title{Extract score function}
\usage{
score(x, ...)
}
\arguments{
\item{x}{Model object}

\item{...}{Additional arguments to lower level functions}
}
\value{
Matrix of score/gradient values. Rows correspond to
observations (if \code{indiv=TRUE}) or a single row, columns to parameters.
}
\description{
Extract the score function (gradient of the log-likelihood) from a
model object.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fix.R
\name{intercept}
\alias{intercept}
\alias{intercept<-}
\alias{intercept.lvm}
\alias{intercept<-.lvm}
\title{Fix mean parameters in 'lvm'-object}
\usage{
\method{intercept}{lvm}(object, vars, ...) <- value
}
\arguments{
\item{object}{\code{lvm}-object}

\item{\dots}{Additional arguments}

\item{vars}{character vector of variable names}

\item{value}{Vector (or list) of parameter values or labels (numeric or
character) or a formula defining the linear constraints (see also the
\code{regression} or \code{covariance} methods).}
}
\value{
A \code{lvm}-object
}
\description{
Define linear constraints on intercept parameters in a \code{lvm}-object.
}
\details{
The \code{intercept} function is used to specify linear constraints on the
intercept parameters of a latent variable model. As an example we look at
the multivariate regression model

\deqn{ E(Y_1|X) = \alpha_1 + \beta_1 X} \deqn{ E(Y_2|X) = \alpha_2 + \beta_2
X}

defined by the call

\code{m <- lvm(c(y1,y2) ~ x)}

To fix \eqn{\alpha_1=\alpha_2} we call

\code{intercept(m) <- c(y1,y2) ~ f(mu)}

Fixed parameters can be reset by fixing them to \code{NA}.  For instance to
free the parameter restriction of \eqn{Y_1} and at the same time fixing
\eqn{\alpha_2=2}, we call

\code{intercept(m, ~y1+y2) <- list(NA,2)}

Calling \code{intercept} with no additional arguments will return the
current intercept restrictions of the \code{lvm}-object.
}
\note{
Variables will be added to the model if not already present.
}
\examples{


## A multivariate model
m <- lvm(c(y1,y2) ~ f(x1,beta)+x2)
regression(m) <- y3 ~ f(x1,beta)
intercept(m) <- y1 ~ f(mu)
intercept(m, ~y2+y3) <- list(2,"mu")
intercept(m) ## Examine intercepts of model (NA translates to free/unique paramete##r)


}
\seealso{
\code{\link{covariance<-}}, \code{\link{regression<-}},
\code{\link{constrain<-}}, \code{\link{parameter<-}},
\code{\link{latent<-}}, \code{\link{cancel<-}}, \code{\link{kill<-}}
}
\author{
Klaus K. Holst
}
\keyword{models}
\keyword{regression}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lava-package.R
\docType{data}
\name{bmidata}
\alias{bmidata}
\title{Data}
\format{
data.frame
}
\description{
Description
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/confpred.R
\name{confpred}
\alias{confpred}
\title{Conformal prediction}
\usage{
confpred(object, data, newdata = data, alpha = 0.05, mad, ...)
}
\arguments{
\item{object}{Model object (lm, glm or similar with predict method) or formula (lm)}

\item{data}{data.frame}

\item{newdata}{New data.frame to make predictions for}

\item{alpha}{Level of prediction interval}

\item{mad}{Conditional model (formula) for the MAD (locally-weighted CP)}

\item{...}{Additional arguments to lower level functions}
}
\value{
data.frame with fitted (fit), lower (lwr) and upper (upr) predictions bands.
}
\description{
Conformal predicions using locally weighted conformal inference with a split-conformal algorithm
}
\examples{
set.seed(123)
n <- 200
x <- seq(0,6,length.out=n)
delta <- 3
ss <- exp(-1+1.5*cos((x-delta)))
ee <- rnorm(n,sd=ss)
y <- (x-delta)+3*cos(x+4.5-delta)+ee
d <- data.frame(y=y,x=x)

newd <- data.frame(x=seq(0,6,length.out=50))
cc <- confpred(lm(y~splines::ns(x,knots=c(1,3,5)),data=d), data=d, newdata=newd)
if (interactive()) {
plot(y~x,pch=16,col=lava::Col("black"),ylim=c(-10,10),xlab="X",ylab="Y")
with(cc,
     lava::confband(newd$x,lwr,upr,fit,
        lwd=3,polygon=TRUE,col=Col("blue"),border=FALSE))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/operators.R
\name{\%ni\%}
\alias{\%ni\%}
\alias{\%in.open\%}
\alias{\%in.closed\%}
\title{Matching operator (x not in y) oposed to the \code{\%in\%}-operator (x in y)}
\usage{
x \%ni\% y
}
\arguments{
\item{x}{vector}

\item{y}{vector of same type as \code{x}}
}
\value{
A logical vector.
}
\description{
Matching operator
}
\examples{

1:10 \%ni\% c(1,5,10)

}
\seealso{
\code{\link{match}}
}
\author{
Klaus K. Holst
}
\keyword{misc}
\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/iid.R
\name{iid}
\alias{iid}
\title{Extract i.i.d. decomposition from model object}
\usage{
iid(x, ...)
}
\arguments{
\item{x}{Model object}

\item{...}{Additional arguments (see the man-page of the IC method)}
}
\value{
Matrix (n x p) with the i.i.d. decomposition of the estimator
(scaled influence function, i.e., the influence function divided by n).
}
\description{
This function extracts
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/confband.R
\name{confband}
\alias{confband}
\alias{forestplot}
\alias{plot_region}
\title{Add Confidence limits bar to plot}
\usage{
confband(
  x,
  lower,
  upper,
  center = NULL,
  line = TRUE,
  delta = 0.07,
  centermark = 0.03,
  pch,
  blank = TRUE,
  vert = TRUE,
  polygon = FALSE,
  alpha = 1,
  step = FALSE,
  ...
)
}
\arguments{
\item{x}{Position (x-coordinate if vert=TRUE, y-coordinate otherwise)}

\item{lower}{Lower limit (if NULL no limits is added, and only the
center is drawn (if not NULL))}

\item{upper}{Upper limit}

\item{center}{Center point}

\item{line}{If FALSE do not add line between upper and lower bound}

\item{delta}{Length of limit bars}

\item{centermark}{Length of center bar}

\item{pch}{Center symbol (if missing a line is drawn)}

\item{blank}{If TRUE a white ball is plotted before the center is
added to the plot}

\item{vert}{If TRUE a vertical bar is plotted. Otherwise a horizontal
bar is used}

\item{polygon}{If TRUE polygons are added between 'lower' and 'upper'}

\item{alpha}{transparency of fill-color of polygon (alpha<1)}

\item{step}{Type of polygon (step-function or piecewise linear)}

\item{...}{Additional low level arguments (e.g. col, lwd, lty,...)}
}
\description{
Add Confidence limits bar to plot
}
\examples{
plot(0,0,type="n",xlab="",ylab="")
confband(0.5,-0.5,0.5,0,col="darkblue")
confband(0.8,-0.5,0.5,0,col="darkred",vert=FALSE,pch=1,cex=1.5)

set.seed(1)
K <- 20
est <- rnorm(K)
se <- runif(K,0.2,0.4)
x <- cbind(est,est-2*se,est+2*se,runif(K,0.5,2))
x[c(3:4,10:12),] <- NA
rownames(x) <- unlist(lapply(letters[seq(K)],function(x) paste(rep(x,4),collapse="")))
rownames(x)[which(is.na(est))] <- ""
signif <- sign(x[,2])==sign(x[,3])
forestplot(x,text.right=FALSE)
forestplot(x[,-4],sep=c(2,15),col=signif+1,box1=TRUE,delta=0.2,pch=16,cex=1.5)
forestplot(x,vert=TRUE,text=FALSE)
forestplot(x,vert=TRUE,text=FALSE,pch=NA)
##forestplot(x,vert=TRUE,text.vert=FALSE)
##forestplot(val,vert=TRUE,add=TRUE)

z <- seq(10)
zu <- c(z[-1],10)
plot(z,type="n")
confband(z,zu,rep(0,length(z)),col=Col("darkblue"),polygon=TRUE,step=TRUE)
confband(z,zu,zu-2,col=Col("darkred"),polygon=TRUE,step=TRUE)

z <- seq(0,1,length.out=100)
plot(z,z,type="n")
confband(z,z,z^2,polygon=TRUE,col="darkred", alpha=0.1)

set.seed(1)
k <- 10
x <- seq(k)
est <- rnorm(k)
sd <- runif(k)
val <- cbind(x,est,est-sd,est+sd)
par(mfrow=c(1,2))
plot(0,type="n",xlim=c(0,k+1),ylim=range(val[,-1]),axes=FALSE,xlab="",ylab="")
axis(2)
confband(val[,1],val[,3],val[,4],val[,2],pch=16,cex=2)
plot(0,type="n",ylim=c(0,k+1),xlim=range(val[,-1]),axes=FALSE,xlab="",ylab="")
axis(1)
confband(val[,1],val[,3],val[,4],val[,2],pch=16,cex=2,vert=FALSE)

x <- seq(0, 3, length.out=20)
y <- cos(x)
yl <- y - 1
yu <- y + 1
plot_region(x, y, yl, yu)
plot_region(x, y, yl, yu, type='s', col="darkblue", add=TRUE)

}
\seealso{
\code{confband}
}
\author{
Klaus K. Holst
}
\keyword{iplot}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/confband.R
\name{confband}
\alias{confband}
\alias{forestplot}
\alias{plot_region}
\title{Add Confidence limits bar to plot}
\usage{
confband(
  x,
  lower,
  upper,
  center = NULL,
  line = TRUE,
  delta = 0.07,
  centermark = 0.03,
  pch,
  blank = TRUE,
  vert = TRUE,
  polygon = FALSE,
  step = FALSE,
  ...
)
}
\arguments{
\item{x}{Position (x-coordinate if vert=TRUE, y-coordinate otherwise)}

\item{lower}{Lower limit (if NULL no limits is added, and only the
center is drawn (if not NULL))}

\item{upper}{Upper limit}

\item{center}{Center point}

\item{line}{If FALSE do not add line between upper and lower bound}

\item{delta}{Length of limit bars}

\item{centermark}{Length of center bar}

\item{pch}{Center symbol (if missing a line is drawn)}

\item{blank}{If TRUE a white ball is plotted before the center is
added to the plot}

\item{vert}{If TRUE a vertical bar is plotted. Otherwise a horizontal
bar is used}

\item{polygon}{If TRUE polygons are added between 'lower' and 'upper'.}

\item{step}{Type of polygon (step-function or piecewise linear)}

\item{...}{Additional low level arguments (e.g. col, lwd, lty,...)}
}
\description{
Add Confidence limits bar to plot
}
\examples{
plot(0,0,type="n",xlab="",ylab="")
confband(0.5,-0.5,0.5,0,col="darkblue")
confband(0.8,-0.5,0.5,0,col="darkred",vert=FALSE,pch=1,cex=1.5)

set.seed(1)
K <- 20
est <- rnorm(K)
se <- runif(K,0.2,0.4)
x <- cbind(est,est-2*se,est+2*se,runif(K,0.5,2))
x[c(3:4,10:12),] <- NA
rownames(x) <- unlist(lapply(letters[seq(K)],function(x) paste(rep(x,4),collapse="")))
rownames(x)[which(is.na(est))] <- ""
signif <- sign(x[,2])==sign(x[,3])
forestplot(x,text.right=FALSE)
forestplot(x[,-4],sep=c(2,15),col=signif+1,box1=TRUE,delta=0.2,pch=16,cex=1.5)
forestplot(x,vert=TRUE,text=FALSE)
forestplot(x,vert=TRUE,text=FALSE,pch=NA)
##forestplot(x,vert=TRUE,text.vert=FALSE)
##forestplot(val,vert=TRUE,add=TRUE)

z <- seq(10)
zu <- c(z[-1],10)
plot(z,type="n")
confband(z,zu,rep(0,length(z)),col=Col("darkblue"),polygon=TRUE,step=TRUE)
confband(z,zu,zu-2,col=Col("darkred"),polygon=TRUE,step=TRUE)

z <- seq(0,1,length.out=100)
plot(z,z,type="n")
confband(z,z,z^2,polygon="TRUE",col=Col("darkblue"))

set.seed(1)
k <- 10
x <- seq(k)
est <- rnorm(k)
sd <- runif(k)
val <- cbind(x,est,est-sd,est+sd)
par(mfrow=c(1,2))
plot(0,type="n",xlim=c(0,k+1),ylim=range(val[,-1]),axes=FALSE,xlab="",ylab="")
axis(2)
confband(val[,1],val[,3],val[,4],val[,2],pch=16,cex=2)
plot(0,type="n",ylim=c(0,k+1),xlim=range(val[,-1]),axes=FALSE,xlab="",ylab="")
axis(1)
confband(val[,1],val[,3],val[,4],val[,2],pch=16,cex=2,vert=FALSE)

x <- seq(0, 3, length.out=20)
y <- cos(x)
yl <- y - 1
yu <- y + 1
plot_region(x, y, yl, yu)
plot_region(x, y, yl, yu, type='s', col="darkblue", add=TRUE)

}
\seealso{
\code{confband}
}
\author{
Klaus K. Holst
}
\keyword{iplot}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pcor.R
\name{pcor}
\alias{pcor}
\title{Polychoric correlation}
\usage{
pcor(x, y, X, start, ...)
}
\arguments{
\item{x}{Variable 1}

\item{y}{Variable 2}

\item{X}{Optional covariates}

\item{start}{Optional starting values}

\item{...}{Additional arguments to lower level functions}
}
\description{
Maximum likelhood estimates of polychoric correlations
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/children.R
\name{children}
\alias{children}
\alias{parents}
\alias{ancestors}
\alias{descendants}
\alias{roots}
\alias{sinks}
\alias{adjMat}
\alias{edgeList}
\title{Extract children or parent elements of object}
\usage{
children(object, ...)
}
\arguments{
\item{object}{Object}

\item{\dots}{Additional arguments}
}
\description{
Generic method for memberships from object (e.g. a graph)
}
\author{
Klaus K. Holst
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/operators.R
\name{\%++\%}
\alias{\%++\%}
\title{Concatenation operator}
\usage{
x \%++\% y
}
\arguments{
\item{x}{First object}

\item{y}{Second object of same class}
}
\description{
For matrices a block-diagonal matrix is created. For all other
data types he operator is a wrapper of \code{paste}.
}
\details{
Concatenation operator
}
\examples{
## Block diagonal
matrix(rnorm(25),5)\%++\%matrix(rnorm(25),5)
## String concatenation
"Hello "\%++\%" World"
## Function composition
f <- log \%++\% exp
f(2)
}
\seealso{
\code{blockdiag}, \code{\link{paste}}, \code{\link{cat}},
}
\author{
Klaus K. Holst
}
\keyword{misc}
\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lava-package.R
\docType{data}
\name{semdata}
\alias{semdata}
\title{Example SEM data}
\format{
data.frame
}
\source{
Simulated
}
\description{
Simulated data
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/commutation.R
\name{commutation}
\alias{commutation}
\title{Finds the unique commutation matrix}
\usage{
commutation(m, n = m)
}
\arguments{
\item{m}{rows}

\item{n}{columns}
}
\description{
Finds the unique commutation matrix K:
\eqn{K vec(A) = vec(A^t)}
}
\author{
Klaus K. Holst
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/estimate.glm.R
\name{estimate.glm}
\alias{estimate.glm}
\title{Estimate method for GLM objects}
\usage{
\method{estimate}{glm}(x, ...)
}
\arguments{
\item{x}{Fitted \code{glm} object}

\item{...}{Additional arguments to \link{estimate.default}}
}
\value{
Object of class \code{estimate} (see \link{estimate.default}).
}
\description{
Applies \link{estimate.default} to obtain robust inference for a fitted
\code{glm} object.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/var_ic.R
\name{var_ic}
\alias{var_ic}
\title{Variance based on influence function}
\usage{
var_ic(x, ...)
}
\arguments{
\item{x}{Influence function matrix (observations x parameters), typically
obtained from \link{IC}.}

\item{...}{Additional arguments (currently not used)}
}
\value{
Covariance matrix (p x p).
}
\description{
Computes the empirical variance of the influence function, providing a
sandwich-type variance estimator.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/estimate.formula.R
\name{estimate.formula}
\alias{estimate.formula}
\title{Estimate method for formulas}
\usage{
\method{estimate}{formula}(
  x,
  data,
  weights,
  family = stats::gaussian,
  ...,
  model = "glm",
  lvm = FALSE
)
}
\arguments{
\item{x}{formula specifying the model}

\item{data}{data.frame}

\item{weights}{optional weights}

\item{family}{GLM family (default \code{gaussian})}

\item{...}{Additional arguments to \link{estimate.default}}

\item{model}{character string specifying the model type (default \code{"glm"})}

\item{lvm}{logical, if TRUE use \code{estimate.lvm} instead}
}
\value{
Object of class \code{estimate} (see \link{estimate.default}).
}
\description{
Fit a GLM model specified by a formula and apply \link{estimate.default}
for robust inference.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mvnmix.R
\name{mvnmix}
\alias{mvnmix}
\title{Estimate mixture latent variable model}
\usage{
mvnmix(
  data,
  k = 2,
  theta,
  steps = 500,
  tol = 1e-16,
  lambda = 0,
  mu = NULL,
  silent = TRUE,
  extra = FALSE,
  n.start = 1,
  init = "kmpp",
  ...
)
}
\arguments{
\item{data}{\code{data.frame}}

\item{k}{Number of mixture components}

\item{theta}{Optional starting values}

\item{steps}{Maximum number of iterations}

\item{tol}{Convergence tolerance of EM algorithm}

\item{lambda}{Regularisation parameter. Added to diagonal of covariance matrix (to avoid
singularities)}

\item{mu}{Initial centres (if unspecified random centres will be chosen)}

\item{silent}{Turn on/off output messages}

\item{extra}{Extra debug information}

\item{n.start}{Number of restarts}

\item{init}{Function to choose initial centres}

\item{...}{Additional arguments parsed to lower-level functions}
}
\value{
A \code{mixture} object
}
\description{
Estimate mixture latent variable model
}
\details{
Estimate parameters in a mixture of latent variable models via the EM
algorithm.
}
\examples{

data(faithful)
set.seed(1)
M1 <- mvnmix(faithful[,"waiting",drop=FALSE],k=2)
M2 <- mvnmix(faithful,k=2)
if (interactive()) {
    par(mfrow=c(2,1))
    plot(M1,col=c("orange","blue"),ylim=c(0,0.05))
    plot(M2,col=c("orange","blue"))
}

}
\seealso{
\code{mixture}
}
\author{
Klaus K. Holst
}
\keyword{models}
\keyword{regression}
